#!/usr/bin/env python
# 
"""A module to read Windoze style .ini files.

The main interface is an IniFile class that can read/write .ini files.

An IniFile looks like a dictionary (called the main dictionary) with
each "section" of the inifile becoming a key in the main dictionary.
How the values are stored is illustrated by the example below.

If the contents of the .ini file are as follows:

  key1 = val7
  key2 =val8
  key3= val9

  [section1]
  key1 = val1
  key2 = val2
  line 1
  line 2

  [section2]
  key1=val3
  line 3

Then the corresponding IniFile object would be as follows:

  {'':
      {'key1':'val7',
       'key2':'val8',
       'key3':'val9'
      },
   'section1':
      {'key1':'val1',
       'key2':'val2',
       '_values_':['line 1', 'line 2']
      },
   'section2':
      {'key1':'val3',
       '_values_':['line 3']
      }
   }


Terminology from above:
  Section Name : section1 and section2
  Section Keys : key1, key2, key3
  Section Value: val1, val2, etc

Lines in the .ini file which are either empty or for which the first
non-white-space character is # or ; are comment lines and are ignored.

Lines which become entries in the '_values_' list are maintained in the order
in which they were found in the .ini file.

Example usage:

  >>> ini = IniFile('c:/windows/win.ini')
  >>> # See what your wallpaper is set to...
  >>> wp = ini['Desktop']['Wallpaper']
  >>> print wp
  'c:\\windows\\spam.bmp'

There is also write_ini_file() function.

"""

__version__ = "$Revision: 1.1 $"[11:-2]

import re, string
from UserDict import UserDict

DefaultSectionKey = ''
DefaultValueKey = ''


class Substitute(UserDict):
    """IniFile class which acts as a dictionary of section names.

    methods:
      sections() return the list of section names
      update(other) update this dict based on OTHER
      read_file(file) update this dict by reading ini file FILE
      write_file(file) write this dict to file FILE
    
    """
    def __init__(self, file=None):
        apply(UserDict.__init__,(self,)) # initialize the base class
        if file is not None:             # read the file if one is specified
            self.read_file(file)
        
    def sections(self):
        """Returns list of section names."""
        return self.data.keys()

    def update(self,other):
        """Update this Substitute object by merging the sections from OTHER.

        If OTHER contains a section name not in this object, the
        entire section and all (key,value) pairs are added to this
        object.

        If OTHER shares a section with this object, the section
        dictionaries are merged. sectionKeys that exist in common will
        be updated to contain the value from OTHER.
        
        """

        # if the other object is also an Substitute (or at least has a
        # "sections" attribute), then we will merge the section dictionaries
        if hasattr(other,"sections"):
            for section in other.sections():
                if not self.data.has_key(section):
                    self.data[section] = {}
                self.data[section].update(other.data[section])
        else:
            apply(UserDict.update,(self,other))

    _section_re = re.compile(r'^\[(.*)\]')
    _option_re  = re.compile(r'^([^=]*\S)\s*=\s*(.*)$')

    def read_file(self,file):
        """Update this object by reading the ini file from FILE.

        If FILE is a string, then FILE is assumed to be a filename and
        that file is opened. Otherwise, file is assumed to be a
        read-able file object.

        """
        
        if type(file) == type(''):
            file = open(file,'r')

        option_re  = self._option_re
        section_re = self._section_re
        
        cur_section = DefaultSectionKey
        while 1:
            line = file.readline()
            if line == '': break        # end of file

            # remove trailing, leading white space
            line = string.strip(line)
            
            # check for a comment line
            # note: inclusion below of '#' is NOT windows standard
            if line == '' or line[0] == ';' or line[0] == '#':
                continue                # comment line, go to next line

            # see if current line is a new section
            section_match = section_re.match(line)
            if section_match is not None:
                cur_section = section_match.group(1)

                if not self.data.has_key(cur_section):
                    # found new section, need to create new section dict
                    self.data[cur_section] = {}

            else:
                # check if this line a line of the form X=Y
                option_match = option_re.match(line)

                if not self.data.has_key(cur_section):
                    # we have found an option line, but we're not in a section.
                    # so create a default section
                    self.data[cur_section] = {}
                if option_match is not None:
                    opt, val = option_match.group(1,2)
                    self.data[cur_section][opt] = val
                else:
                    sys.exit(1)
    

    def write_file(self, file):
        """Write the file object to FILE.

        if FILE is a string, the file with that name is opened. Otherwise,
        FILE must be a write-able file-like object.

        """

        
        if type(file) == type(''):
            file = open(file,'w')
            
        for section,section_dict in self.items():
            file.write("[%s]\n" % section) # write section header

            # write all values that have value,name pairs
            # we first remove the DefaultValueKey key from the
            # list of keys
            option_keys = section_dict.keys()
            option_keys.remove(DefaultValueKey)
            for option in option_keys:
                value = section_dict[option]
                file.write("%s=%s\n" % (option,value))

            # now print out any remaining values, as found in
            # DefaultValueKey
            values = section_dict.get(DefaultValueKey,[])
            for value in values:
                file.write("%s\n" % value)
    
      
def test(argv):
    for file in argv:
        ini = Substitute()
        ini.read_file(file)
        print ini.sections ()
        print ini
        ini.write_file ("output.ini")
        print "wrote ini file to output.ini"
    
def self_test():
    data = r"""# some file
[BMK]
Value=1
# Test that space is ok in key and values
D:\Windows\Program Files\Foobar.exe=Some Other Value
#Test that key and space is ok around =
  SomeKey = SomeValue 
Alpha
Beta
"""
    import tempfile,sys
    filename = tempfile.mktemp()
    print "saving the following data to file", filename
    print data
    try:
        fp = open(filename,"w")
        fp.write(data)
        fp.close()
        ini = Substitute()
        ini.read_file(filename)
        print "Read Init file back as dictionary:"
        print ini
        ini.write_file("output.ini")
    except:
        sys.stderr.write("test failed")
        raise
  
if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        test(sys.argv[1:])
    else:
        self_test()
