
#include "dgp_err_handle.h"
#include "silx_encode.h"
#include "DgpEncode.h"
#include <time.h>
#include <limits.h>
#include "PsgFix.h"
#include <stdio.h>

static Dgp_Float64 
get_utime()
{
	return (Dgp_Float64)clock()/CLOCKS_PER_SEC;
}

static CLOpts cl_opts = {   DEF_VERBOSE,
							DEF_LOSSLESS,
                            DEF_FRACHAUSD, 
                            DEF_WIDTHDIFF, 
                            DEF_HEIGHTDIFF, 
                            DEF_MAXDENSITY,
                            DEF_COMMITTHRESH,
                            DEF_FREEZEHISTTHRESH,
                            DEF_MAKEHISTTHRESH,
                            DEF_RETURNHISTTHRESH,
                            DEF_DILATIONRADIUS,
                            DEF_SIZETHRESH,
                            DEF_WRITETHRESH,
                            DEF_NEARESTNEIGHBORS,
                            DEF_XSHIFT,
                            DEF_XSLOPECHANGE,
                            DEF_XNEWSLOPE,
							DEF_LOWERDILATIONLIMIT,
                            DEF_NONTP,
                            DEF_INCRDICT,
                            DEF_CLEARDICT,
                            DEF_BASELINEORIENT,
                            DEF_DESKEW,
                            DEF_DESKEWACCURACY,
                            DEF_GRAYHALFTONE,
                            DEF_TOK_OUTPUT_RES,
                            DEF_JPEG_OUTPUT_RES,
							DEF_JPEG_QUALITY,
							DEF_ROTATION,
							DEF_TOKENSIZELIMIT,
                            DEF_NOALIGN,
                            DEF_G4TOKENS,
			                DEF_JBIG2_PROFILE,
							DEF_FILENAME,
                            DEF_ERRFILENAME,
                            DEF_NOERROROUT,
                            DEF_MEMBOND,
                            DEF_REDTOKNOISE,
                            DEF_JB2ONLY,
                            DEF_PDFOUT};


/****************
 *
 * The header displayed when showing usage information.
 */

static const Dgp_Int8 *cl_header = \
"Usage: DgpEncode [options] foo.tif [bar.tif...]\n \
\tEncodes input files to tfx.\n";

/****************
 *
 * The definition of the command line options.
 */

static DgpCl_Option cl_cmdln_options[] = {
  DGPCL_HELP_CMD,
  DGPCL_VERBOSE_CMD(cl_opts.verbose),
  DGPCL_OUTFNAME_CMD(cl_opts.filename, DEF_FILENAME),
  DGPCL_NEW_FLAG_CMD("ll", cl_opts.lossless, "Encode in full lossless mode (electronic source)"),
  DGPCL_INC_DICT_CMD(cl_opts.incrementalDict, DEF_INCRDICT_STR),
  DGPCL_CLR_DICT_CMD(cl_opts.clearDict, DEF_CLEARDICT_STR),
  DGPCL_RES_CMD(cl_opts.outputTokRes, DEF_TOK_OUTPUT_RES_STR),
  DGPCL_RESJ_CMD(cl_opts.jpegRes, DEF_JPEG_OUTPUT_RES_STR),
  DGPCL_JPEG_QUAL_CMD(cl_opts.jpegQual, DEF_JPEG_QUALITY_STR),
  DGPCL_NEW_INT_CMD("mem", cl_opts.membond, "Token dictionary memory upper limit, [-1\\disable, 1\\1 Meg, 2\\2 Meg, 3\\32 Meg]",DEF_MEMBOND_STR),
  DGPCL_NEW_FLAG_CMD("rn", cl_opts.redTokNoise, "Reduce token noise for each page. Color images only"),
  DGPCL_NEW_FLOAT_CMD("ph", cl_opts.fracHausd, "Pct. of Hausdorff distance to use for model match", DEF_FRACHAUSD_STR),
  DGPCL_NEW_INT_CMD("wd", cl_opts.widthDiff, "Highest allowed width difference for matching boxes", DEF_WIDTHDIFF_STR),
  DGPCL_NEW_INT_CMD("hd", cl_opts.heightDiff, "Highest allowed height difference for matching boxes", DEF_HEIGHTDIFF_STR),
  DGPCL_NEW_INT_CMD("md", cl_opts.maxDensity, "Allowed number of mismatches per 3 x 3 cell", DEF_MAXDENSITY_STR),
  DGPCL_NEW_INT_CMD("ct", cl_opts.commitThresh, "Population threshold for committing a class", DEF_COMMITTHRESH_STR),
  DGPCL_NEW_INT_CMD("ft", cl_opts.freezeHistThresh, "Population threshold for freezing class histogram", DEF_FREEZEHISTTHRESH_STR),
/* 10 */
  DGPCL_NEW_INT_CMD("mt", cl_opts.makeHistThresh, "Population threshold for class histogram creation", DEF_MAKEHISTTHRESH_STR),
  DGPCL_NEW_INT_CMD("rt", cl_opts.returnHistThresh, "Population threshold for class histogram return", DEF_RETURNHISTTHRESH_STR),
  DGPCL_NEW_INT_CMD("dr", cl_opts.dilationRadius, "Dilation radius (0..9). 10, 14, or 20 for non-tp, r = 1, sqrt(2), and 2", 
    DEF_DILATIONRADIUS_STR),
  DGPCL_NEW_INT_CMD("st", cl_opts.sizeThresh, "Minimum cc size, inclusive", DEF_SIZETHRESH_STR),
  DGPCL_NEW_INT_CMD("wt", cl_opts.writeThresh, "Population threshold for writing a class", DEF_WRITETHRESH_STR),
  DGPCL_NEW_INT_CMD("nn", cl_opts.nearestNeighbors, "Neighborhood size for matching shifts", DEF_NEARESTNEIGHBORS_STR),
  DGPCL_NEW_INT_CMD("xshift", cl_opts.xShift, "Small token error curve, offset parameter", DEF_XSHIFT_STR),
  DGPCL_NEW_INT_CMD("xsc", cl_opts.xSlopeChange, "Small token error curve, slope change parameter", DEF_XSLOPECHANGE_STR),
  DGPCL_NEW_FLOAT_CMD("xns", cl_opts.xNewSlope, "Small token error curve, new slope parameter", DEF_XNEWSLOPE_STR),
  DGPCL_NEW_INT_CMD("ldl", cl_opts.lowerDilationLimit, "When matching, only dilate tokens this size and larger", DEF_LOWERDILATIONLIMIT_STR),
/* 20 */
  DGPCL_NEW_FLAG_CMD("nontp", cl_opts.nontp, "Force non-topology-preserving dilation"),
  DGPCL_DESKEW_CMD(cl_opts.deskew),
  DGPCL_NEW_STRING_CMD("bo", cl_opts.baselineOrient, "Baseline deskewing orientation (\"d[own]\", \"l[eft]\", \"r[ight]\", or \"u[p]\")", 
    DEF_BASELINEORIENT),
  DGPCL_NEW_INT_CMD("da", cl_opts.deskewAccuracy, "Deskew accuracy: 0 (fast) to 5 (accurate)", DEF_DESKEWACCURACY_STR),
  DGPCL_NEW_FLAG_CMD("hlsy", cl_opts.grayHalftone, "Encode halftone lossyly in gray if this option is set"),
  DGPCL_NEW_INT_CMD("rot", cl_opts.rotation, "Rotate clockwise by this multiple of pi/2 before encoding", DEF_ROTATION_STR),
  DGPCL_NEW_INT_CMD("profile", cl_opts.jbig2Profile, "Jbig2 profile", DEF_JBIG2_PROFILE_STR),
  DGPCL_NEW_FLAG_CMD("noalign", cl_opts.noAlign, "Do not perform shifts when determining bounding box positions"),
/* 30 */
  DGPCL_NEW_FLAG_CMD("g4t", cl_opts.g4Tokens, "Encode singleton and outlier tokens as group 4"),
  DGPCL_NEW_STRING_CMD("e", cl_opts.err_filename, "Set error file name", DEF_ERRFILENAME),
  DGPCL_NEW_FLAG_CMD("noe", cl_opts.noe, "Suppress error output; overrides -e"),
  DGPCL_NEW_FLAG_CMD("noTokenSizeLimit", cl_opts.tokenSizeLimit, "Flag disables standard limit of 10% of the page size for max size of tokens"),
  DGPCL_NEW_FLAG_CMD("jb2only", cl_opts.jb2Only, "Only output jb2 stream instead of TFX file"),
  DGPCL_NEW_FLAG_CMD("pdfOut", cl_opts.pdfOut, "Output jb2 stream embedded in PDF"),
  {NULL}
};


static ErrCode
validateArgs(Dgp_State *dgp_state,
             Dgp_Int32 unparsed_argc, 
             Dgp_Int8 **unparsed_argv,
             FILE *err_out,
             void *validate_cookie)
{
  DGP_PROLOGUE("validateArgs()", DGP_PERRSTATE);
  CLOpts *my_opts = (CLOpts *)validate_cookie;

  unparsed_argc = unparsed_argc;
  unparsed_argv = unparsed_argv;

  if((my_opts->jpegQual > 100) || (my_opts->jpegQual < 0)) {
    DGPCL_PRINT_ERRF0(err_out, "Valid resolutions are 0 < qual < 100...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->incrementalDict < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid incremental dictionary parameter...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->clearDict < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid clear dictionary parameter...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if (my_opts->pdfOut)
  {
      if (my_opts->incrementalDict != DEF_INCRDICT)
      {
        DGPCL_PRINT_ERRF0(err_out, "No incremental dictionaries in PDF\n");
        DGP_ERR(cErrInvalidCmdLine);
      }
      /* no sections in PDF .... at least not in this early version */
      my_opts->incrementalDict = 65000;
      /*my_opts->clearDict = 65000; */
      if (!my_opts->jb2Only)
      {
        DGPCL_PRINT_ERRF0(err_out, "For PDF output -jb2only flag must be set\n");
        DGP_ERR(cErrInvalidCmdLine);
      }
  }
  if((my_opts->jbig2Profile > 3) || (my_opts->jbig2Profile < 1)) {
    DGPCL_PRINT_ERRF0(err_out, "Valid JBIG2 profiles  are 1, 2 or 3\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->outputTokRes > 2400) || (my_opts->outputTokRes < 0)) {
    DGPCL_PRINT_ERRF0(err_out, "Valid resolutions are 0 < res < 2400...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->fracHausd < (Dgp_Float32)0.0) || (my_opts->fracHausd > (Dgp_Float32)1.0)){
    my_opts->fracHausd = my_opts->fracHausd / (Dgp_Float32)100;
    DGPCL_PRINT_ERRF0(err_out, "Interpreting Hausdorff distance as percentage...\n");
    /*DGP_ERR(cErrInvalidCmdLine); */
  }
  if(my_opts->fracHausd < (Dgp_Float32)0.0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid Hausdorff distance...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->widthDiff < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid width difference...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->heightDiff < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid height difference...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->maxDensity < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid max density...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->commitThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid commit threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->freezeHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid historgram freeze threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->makeHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid histogram make threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->returnHistThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid histogram return threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->dilationRadius < 0) || ((my_opts->dilationRadius > 9) &&
      (my_opts->dilationRadius != 10) && (my_opts->dilationRadius != 14) && (my_opts->dilationRadius != 20))){
    DGPCL_PRINT_ERRF0(err_out, "Invalid dilation radius (use 0...9, 10, 14, or 20).\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->sizeThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid size threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->writeThresh < 0){
    DGPCL_PRINT_ERRF0(err_out, "Invalid write threshold...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->deskewAccuracy < 0) || (my_opts->deskewAccuracy > 5)){
    DGPCL_PRINT_ERRF0(err_out, "Invalid deskew accuracy...\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if((my_opts->nearestNeighbors != 0) && (my_opts->nearestNeighbors != 4) && (my_opts->nearestNeighbors != 8)
      && (my_opts->nearestNeighbors != 24)){
    DGPCL_PRINT_ERRF0(err_out, "Invalid nearest neighbor parameter; specify 0, 4, 8, or 24.\n");
    DGP_ERR(cErrInvalidCmdLine);
  }
  switch(my_opts->baselineOrient[0]){
  case 'd': case 'D':   case 'u':   case 'U':
  case 'l': case 'L':   case 'r':   case 'R':
      break;
  default:
    DGPCL_PRINT_ERRF0(err_out, "Invalid baseline orientation; specify \"down\", \"up\", \"left\", or \"right\", \
or at least the first letter.\n");
    DGP_ERR(cErrInvalidCmdLine);
  }

  NORMAL_RETURN;
}


int
main (int argc, char** argv)
{
    Dgp_State				*dgp_state = NULL;
    DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);
    /* For error & status reports... */
    Dgp_Int8				*outfname = NULL;
    Dgp_Int32               file_num = 0;
    Dgp_Float64				t0, t1;
    FILE                    *outfp = NULL;
    SilxEncCookie           sCookie = NULL;

    
    if(argc == 1){
        DgpCl_printHelp(stderr, cl_header, cl_cmdln_options);
        MAIN_NORMAL_RETURN;
    }
    
    /* Encode timer... */
    t0 = get_utime();
    
    /* Init dgp state */    
    DGP_ERR_PASS_CLEANUP(Dgp_initDgpState(&dgp_state, 
        NULL,
        NULL,
        stderr, 
        Dgp_fileErrPrinter, 
        DGP_TRUE, 
        DGP_TRUE));
    DGP_SET_ERR_STATE(DGP_PERRSTATE);
    
    /* parse command line options */
    DGP_ERR_PASS_CLEANUP(DgpCl_parse(dgp_state, (Dgp_Int32 *)&argc, argv, 
        cl_cmdln_options, stderr, cl_header, validateArgs, &cl_opts));
    if(argc < 1){
        DGPCL_PRINT_ERRF0(stderr, "Need at least one input file for encoding...\n");
        DGP_ERR(cErrInvalidCmdLine);
    }
    
    
    /* Concoct an output filename if the user didn't specify one... */
    if((cl_opts.filename == NULL) || (cl_cmdln_options[2].found == DGP_FALSE)) {
        DGP_CHECK_ALLOC_CLEANUP(outfname = (Dgp_Int8 *)DGP_MALLOC((strlen(argv[0]) + 5) * sizeof(Dgp_Int8)));
        outfname[0] = '\0';      
        if(cl_opts.jb2Only)
	  {
	    if (cl_opts.pdfOut)
	      sprintf(outfname, "%s.pdf", argv[0]);
	    else
	      sprintf(outfname, "%s.jb2", argv[0]);
	  }
        else
            sprintf(outfname, "%s.tfx", argv[0]);
    }
    else{
        DGP_CHECK_ALLOC_CLEANUP(outfname = (Dgp_Int8 *)DGP_MALLOC((strlen(cl_opts.filename) + 1) * sizeof(Dgp_Int8)));
        strcpy(outfname, cl_opts.filename);
    }
    
    DGP_SYS_ERR_GEN_CLEANUP(IS_NULL(outfp = fopen (outfname, "wb")));
    DGP_ERR_PASS_CLEANUP(silx_initEncState(dgp_state, &sCookie));
    
    /* API calls to initiate tokenizer */
    silx_setGeneralEncOptions(
                sCookie,
                &cl_opts.grayHalftone,
                &cl_opts.deskew,
                &cl_opts.deskewAccuracy,
                cl_cmdln_options[6].found == DGP_TRUE ? &(cl_opts.outputTokRes) : NULL,
                &cl_opts.baselineOrient,  
                &cl_opts.rotation,
                &cl_opts.jbig2Profile,
                &cl_opts.g4Tokens,
                &cl_opts.tokenSizeLimit,
                &cl_opts.verbose,
                &cl_opts.membond,
                &cl_opts.noe,
                &cl_opts.lossless,
                &cl_opts.incrementalDict,
                &cl_opts.clearDict,
                &cl_opts.redTokNoise,
                &cl_opts.jb2Only,
		&cl_opts.pdfOut);
     
    /* Fill in tokenizer options too... */
    silx_setTokenOptions(
                sCookie,
                &cl_opts.writeThresh,
                &cl_opts.fracHausd,
                &cl_opts.maxDensity,
                &cl_opts.widthDiff,
                &cl_opts.heightDiff,
                &cl_opts.commitThresh,
                &cl_opts.freezeHistThresh,
                &cl_opts.makeHistThresh,
                &cl_opts.returnHistThresh,
                &cl_opts.nearestNeighbors,
                &cl_opts.xShift,
                &cl_opts.xSlopeChange,
                &cl_opts.xNewSlope,
                &cl_opts.dilationRadius,
                &cl_opts.lowerDilationLimit,
                &cl_opts.sizeThresh);

    /* API calls to initiate Jpeg options  */
    silx_setJpegOptions(
                sCookie,
                &cl_opts.jpegQual,
                &cl_opts.jpegRes,
                NULL);

    /* Start encoding ... */
    DGP_ERR_PASS_CLEANUP(silx_encStart(dgp_state, sCookie, (void *)outfp, cl_opts.err_filename)); 

    /* Go through each document */
    while(file_num < argc){
        Dgp_Int32 retval;
        if(file_num < argc-1) /* encode one document at a time */
            retval = silx_encDocument(dgp_state, sCookie, argv[file_num], DGP_FALSE); 
        else /* encode last document */
            retval = (silx_encDocument(dgp_state, sCookie, argv[file_num], DGP_TRUE));
        /* If the TFX dictionary limit is exceeded. (Only valid when encode to TFX format) */
        if(retval == cTFXDictionaryLimitErr){
            DGP_ERR_PRINTERF0("Terminating encoder because TIFF-FX dictionary limit of 5 has been reached.");
            break;
	}
	else DGP_ERR_PASS_CLEANUP(retval);
        file_num ++;
    }

    /* Clean up encode cookie */
    DGP_ERR_PASS_CLEANUP(silx_cleanEncState(dgp_state, &sCookie));

    if(outfp) fclose(outfp);
    if(outfname) DGP_FREE(outfname);
    
    t1 = get_utime();
    DGP_REPORTF1(cl_opts.verbose, "Encoding time:  %.3f secs.\n", (t1 - t0));
    
    /*Clean up Dgp_State */
    Dgp_cleanupDgpState(&dgp_state);
    
    MAIN_NORMAL_RETURN;
    
ERR_CLEANUP:
    if(outfp) fclose(outfp);
    if(outfname) DGP_FREE(outfname);
    if(sCookie) silx_cleanEncState(dgp_state, &sCookie);
    if(dgp_state)  Dgp_cleanupDgpState(&dgp_state);
    MAIN_ERR_RETURN;
}


