#include "StdAfx.h"

#include "DibImgBuf.h"
#include "AsdMath.h"
	
using namespace Asd;
using namespace AsdMath;

DibImgBuf::DibImgBuf( void )
{
    // We have to make sure there is data available for us to use the bmih() functions!
    m_headerBuf.resize( sizeof(BITMAPINFOHEADER) +  2 * sizeof(RGBQUAD) );

    // Default to 24-bit RGB bitmap with no area at 100-dpi.
    bmih().biSize = sizeof( BITMAPINFOHEADER  );
    bmih().biWidth    = 0;
    bmih().biHeight   = 0;
    bmih().biPlanes   = 1;
    bmih().biCompression   = BI_RGB;
    bmih().biSizeImage     = 0;
    bmih().biXPelsPerMeter = 3937;
    bmih().biYPelsPerMeter = 3937;
    bmih().biClrUsed       = 0;
    bmih().biClrImportant  = 0;

    // This is a little trick to force setColorDepth() to do it's thing.
    // Basically, setColorDepth is a noop if the the depth is not
    // changing.  Since we have not 
    bmih().biBitCount = COLOR_DEPTH_ERROR;
    setColorDepth(COLOR_24_BIT);
}


UInt32 DibImgBuf::width( void ) const
{
    const long width = Abs( bmih().biWidth );

    return UInt32( width );
}

UInt32 DibImgBuf::height( void ) const
{
    const long height = Abs( bmih().biHeight );

    return UInt32( height );
}

ColorDepth DibImgBuf::depth( void ) const
{
    return ColorDepth(bmih().biBitCount);
}

const BITMAPINFO & DibImgBuf::getBmpInfo( void ) const
{
    DibImgBuf * me = const_cast<DibImgBuf *>(this);

    return *reinterpret_cast<const BITMAPINFO*>(&me->m_headerBuf[0]);
}

const BITMAPINFOHEADER & DibImgBuf::bmih( void ) const
{
    DibImgBuf * me = const_cast<DibImgBuf *>(this);

    return *reinterpret_cast<const BITMAPINFOHEADER*>(&me->m_headerBuf[0]);
}


UInt32 DibImgBuf::
calcBpl( const UInt32 width, const UInt32 bpp, const UInt32 align )
{
    UInt32 bpl  = 0;

    if( bpp >= 8 )
    {
        assert( (bpp % 8) == 0 );

        bpl = (bpp / 8) * width;
    }
    else // thus( bpp < 8 )
    {
        assert( (8 % bpp) == 0 );

        // make sure we have at least as many bits as needed
        bpl = (width + (8 / bpp - 1)) / (8 / bpp);
    }

    while( bpl % align )
    {
        bpl++;
    }

    return bpl;
}

// If we are not 32-bit aligned, then there is a chance that a given width
// is not actually possible for our object.  fixWidth()'s sole purpose
// is to help resize() choose a legal width.
void DibImgBuf::fixWidth( void )
{
    if( m_align != sizeof(Asd::UInt32) )
    {
        UInt32 dibBpl = calcBpl( width(), depth(), sizeof(Asd::UInt32) );
        UInt32 desBpl = calcBpl( width(), depth(), m_align       );

        // Ok.  This looks like it might be an infinite loop in some cases,
        // but it's really ok.  What is going on here is the following.
        // For a given width, either the bytes per line of a normal
        // 32-bit-aligned DIB is the same as it is for our alignment or
        // it's not.  If it's not, that means we have to add at least one
        // pixel to the bitmap and try again.  Since all pixels are 
        // 32-bits or less and only multiples of 32-bit alignment are
        // allowed, adding a single pixel in this situation will not
        // change the bytes per line of our bitmap, but it might change
        // the bytes per line of the normal DIB.  Thus, we can keep adding
        // a single pixel to the width until the two bytes per line values
        // are the same.
        while( dibBpl != desBpl )
        {
            ++bmih().biWidth;

            dibBpl = calcBpl( width(), depth(), sizeof(Asd::UInt32) );
        }
    }
}

void DibImgBuf::
resize( const Point32 & size, ColorDepth cd, Asd::UInt32 align )

{
    const UInt32 oldDataSize = dataSize();
    
    bmih().biWidth    =  Abs(size.getX());
    bmih().biHeight   = -Abs(size.getY());
    
    setColorDepth( cd );

    
    // JBTODO: This is for safety but may not be the best way to do this.
    if( (align ==    sizeof(Asd::UInt32))   ||
        (align == (2*sizeof(Asd::UInt32)))   )
    {
        m_align = align;
    }
    else
    {
        m_align = sizeof(Asd::UInt32);
    }


    fixWidth();

    const UInt32 newDataSize = height() * calcBpl( width(), depth(), m_align );

    if( oldDataSize != newDataSize )
    {
        m_data.resize( newDataSize );
    }
}

void DibImgBuf::setColorDepth( ColorDepth cd )
{
    if( cd == depth() )
    {
        return;
    }

    bmih().biBitCount = cd;

    switch( cd )
    {
        case   MONOCHROME:
        case COLOR_24_BIT:
        case COLOR_32_BIT:
        {
            const UInt32 newSize = sizeof(BITMAPINFOHEADER) +  2 * sizeof(RGBQUAD);

            BITMAPINFOHEADER hdr;

            assert( m_headerBuf.size() >= sizeof( hdr ) );

            memcpy( &hdr, &bmih(), sizeof( hdr ) );
            m_headerBuf.resize( newSize );
            memcpy( &bmih(), &hdr, sizeof( hdr ) );

            bmi().bmiColors[0].rgbBlue     = 0xff;
            bmi().bmiColors[0].rgbGreen    = 0xff;
            bmi().bmiColors[0].rgbRed      = 0xff;
            bmi().bmiColors[0].rgbReserved = 0x00;

            bmi().bmiColors[1].rgbBlue     = 0x00;
            bmi().bmiColors[1].rgbGreen    = 0x00;
            bmi().bmiColors[1].rgbRed      = 0x00;
            bmi().bmiColors[1].rgbReserved = 0x00;
        }
        break;

        case COLOR_8_BIT:
        {
            const UInt32 newSize = sizeof(BITMAPINFOHEADER) +  256 * sizeof(RGBQUAD);
            
            BITMAPINFOHEADER hdr;
            memcpy( &hdr, &bmih(), sizeof( hdr ) );
            m_headerBuf.resize( newSize );
            memcpy( &bmih(), &hdr, sizeof( hdr ) );

            for( UInt32 ii=0; ii < 256; ++ii )
            {                
                bmi().bmiColors[ii].rgbBlue     = UInt8(ii);
                bmi().bmiColors[ii].rgbGreen    = UInt8(ii);
                bmi().bmiColors[ii].rgbRed      = UInt8(ii);
                bmi().bmiColors[ii].rgbReserved = 0x00;
            }
        }
        break;
        
        case COLOR_4_BIT:
        case COLOR_DEPTH_ERROR:
        default:
                assert( NULL );
        break;
    }
}

void DibImgBuf::flipData( void )
{
    const UInt32 bytesPerLine = this->bpl();
    const UInt32 lastRow = this->height() - 1;
    const UInt32 MidRow  = this->height() / 2;


    for( UInt32 row = 0; row < MidRow; ++row )
    {
        UInt8 * r1Ptr = this->data() + (row * bytesPerLine);
        UInt8 * r2Ptr = this->data() + ((lastRow - row) * bytesPerLine);

        for( UInt32 byte = 0; byte < bytesPerLine; ++byte, ++r1Ptr, ++r2Ptr )
        {
            std::swap( *r1Ptr, *r2Ptr );
        }
    }

    isTopDown()?  setBottomUp() :  setTopDown();
}
