#ifdef _WIN32
#include <windows.h>
#pragma warning(disable:4786) // supress the annoying windows warning
#endif

#include <stdio.h>

#include "MrcFile.h"
#include "TfxPage.h"
#include "IPCoreImgBuf.h"
#include "DibImgBuf.h"

#include "Option.h"
using namespace std;

#ifdef TIMING
#include <time.h>
static double
get_utime()
{
        return (double)clock()/CLOCKS_PER_SEC;
}
#endif

const string dfOutFile("[infile].[outext]");

void 
addOptions(Option& op, const string& filename){
	op.addOption("o", filename, STRING_TYPE, "Output File Name");
	op.addOption("pscale", "1", FLOAT32_TYPE, "Output scale factor");
    op.addOption("pstart", "0", INT32_TYPE, "Requested start page number (starting from 0)");
    op.addOption("pend", "0", INT32_TYPE, "Requested end page number (starting from 0)");
    op.addOption("otype", "pnm", STRING_TYPE, "Output format [tif][pnm]");
    op.addOption("px", "0", INT32_TYPE, "Requested area x location");
	op.addOption("py", "0", INT32_TYPE, "Requested area y location");
    op.addOption("pw", "page width", INT32_TYPE,  "Requested area width");
    op.addOption("ph", "page height", INT32_TYPE, "Requested area height");
    op.addOption("help", false_string, FLAG_TYPE, "Print help");
    op.addOption("nout", false_string, FLAG_TYPE, "No output");

}


bool parseArguments(Option &op, int& argc, char**& argv){

    // parse options with marker
    if(!op.parseOptions(--argc, ++argv)) return false;
    //first looking for help option
    bool helpFlag = false;
    op.getValue("help", helpFlag);
    if(helpFlag) return false;
    return true;
}


int main(int argc, char *argv[])
{

    //create header string
    string header("Silx Decoder, copyright@ PARC Inc \n");
    header.append("SilxDecode [input TFX file] \n");
	
	// create Option and add custom options
    Option op(header, "-");	
	addOptions(op, dfOutFile);
	
    //parse Option
	if(! parseArguments(op, argc, argv)){
		op.printHelp();
		return 1;
	} 

  	//parse input file name	
    if(argc ==0) {
        	op.printHelp();
            return 1;
    }

	// get output type
    string ext;
    op.getValue("otype", ext);

	//get outfile name
    string outfile;
    op.getValue("o", outfile);

    // last argument has to be infile
    for(int i = 0; i< argc; i++){
        string infile(argv[i]);
        char index[20]; 
        // if no user setup for output, modify default outfile
        if(!outfile.compare(dfOutFile)){ 
            if(infile.find_last_of(".") != string::npos)
                outfile = infile.substr(0,infile.find_last_of("."));
            else
                outfile = infile;
        } else if(argc >1){
			sprintf(index,"%d",i);
            outfile += index;
        }

        //Open infile   
         MrcFile inFile(infile.data());

        // we don't want to switch RGB to BGR
         inFile.setBGR(false);

        //Get Page info
        int pagestart = 0;
        int pageend = 0;	
        op.getValue("pstart", pagestart);
		op.getValue("pend", pageend);
		if(pageend < pagestart) 
			pageend = pagestart;
		if((Asd::UInt32)(pageend) >= inFile.numPages()) 
			pageend = inFile.numPages()-1;

		// Get requested scale
		float pageScale = 1;
		op.getValue("pscale",pageScale);

		for(int j = pagestart; j <= pageend; j++){

			MrcPage *inPage = (MrcPage*) (inFile.getPage (j));
			if (inPage == NULL) return 1;

			// get page width and height
			int pageWidth = inPage->getSize ().getX();
			int pageHeight = inPage->getSize ().getY();
			Asd::Rect pageRect (pageWidth, pageHeight);

            // get page resolution
            int xRes = inPage->getXRes();
            int yRes = inPage->getYRes();

			// Get requested region
			int pX =0;
			int pY =0;
			op.getValue("px", pX);
			op.getValue("py", pY);
			int pWidth = pageWidth - pX;
			int pHeight = pageHeight - pY;
			op.getValue("pw", pWidth);
			op.getValue("ph", pHeight);
			Asd::Rect clipZone (pWidth, pHeight); 
			clipZone.offsetXY(pX, pY);
			
			//Clip the scaled page region with requested region
			clipZone.intersection(pageRect);
#ifdef TIMING
    Asd::Float64 t1, t2; 
    t1 = get_utime();
#endif
			//scale the region
			clipZone.scaleDown(pageScale);

			// create an image buffer 
			DibImgBuf dibBuf;

#ifdef TIMING
            t2 = get_utime();
            fprintf(stdout, "%f\t", t2-t1);
#endif
            bool noout = false;
            op.getValue("noout", noout);
            bool status;
            string outf = outfile;
            if(noout == false){
                sprintf(index,"_page%d.",j);
                outf += index;
                outf += ext;
                // render the reqested area;
                status = inPage->renderToFile(&dibBuf, clipZone, Asd::Rational<Asd::Int32>((Asd::Int32)(pageScale *100), 100), 
                    outf.data(), ext.data(), xRes, yRes); 
                
            }else
                status = inPage->render(&dibBuf, clipZone, Asd::Rational<Asd::Int32>((Asd::Int32)(pageScale *100), 100)); 
            if(!status) return 1;
        }
    }
    return 0;
}
