// SilxScrLiteCtl.cpp : Implementation of CSilxScrLiteCtl

#include "stdafx.h"
#include "SilxScreenLite.h"
#include "SilxScrLiteCtl.h"

#include "TifFile.h"
#include "TfxPage.h"
#include "MrcFile.h"

#include "AsdRect.h"
#include "AsdMath.h"

using namespace Asd;
using namespace ipcore;

/////////////////////////////////////////////////////////////////////////////
// CSilxScrLiteCtl

CSilxScrLiteCtl::CSilxScrLiteCtl( void )
:
 m_file( NULL ),
 m_dirty( true ),
 m_currentPage(0)
{
}

CSilxScrLiteCtl::~CSilxScrLiteCtl( void )
{
    delete m_file; m_file = NULL;
}

/*************************************************************************

  Draw the current image in the box requested.  If there is no current
  image, we draw a wash.
 
**************************************************************************/

HRESULT CSilxScrLiteCtl::OnDraw( ATL_DRAWINFO& di )
{
	TfxPage * pg = m_file? m_file->getPage(m_currentPage) : NULL;

    if( pg )
    {
        DrawFit( di, pg );
    }
    else
    {
        DrawWash( di );
    }

    return S_OK;
}

/*************************************************************************

 DrawWash draws a blue-yellow wash on the screen.  This is
 not really useful other than it allows us to see that the
 control is alive even if there is nothing being displayed.

*************************************************************************/

void CSilxScrLiteCtl::DrawWash( ATL_DRAWINFO& di )
{
    // Construct a Rect corresponding to the window area.
	const Rect bnd( *reinterpret_cast<const RECT*>(di.prcBounds) );

    if( m_dirty )
    {
        m_dirty = false;

		const Asd::UInt32 size = 256;

        Point32 hw( size, size );

        m_img.resize( hw, COLOR_24_BIT, sizeof(DWORD)  );
    
        Asd::UInt8 * rowPtr = m_img.data();

        for( Asd::UInt32 ii=0; ii < size; ++ii, rowPtr += m_img.bpl() )
        { 
            RgbPtr pixPtr( rowPtr );

            for( Asd::UInt32 jj=0; jj < size; ++jj, ++pixPtr )
            {          
                pixPtr.setRGB( jj, ii, ii );
            }
        }
    }

    StretchDIBits( di.hdcDraw, 
                   bnd.left(), bnd.top(),
                   bnd.width(), bnd.height(),
                   0, 0,
                   m_img.width(), m_img.height(),
                   m_img.data(), &m_img.getBmpInfo(),
                   DIB_RGB_COLORS, SRCCOPY );
}

/*************************************************************************

  Draw the image on the page pg such that it fills as much of
  the boundary region of the window as possible.  Fill in any
  unused space with the fill color.

*************************************************************************/

void CSilxScrLiteCtl::DrawFit( ATL_DRAWINFO & di, TfxPage * pg )
{

    // A rectangle representing the boundaries of the window.
	const Rect bnd( *reinterpret_cast<const RECT*>(di.prcBounds) );

    // If the boundaries are empty, there is nothing to draw.
	if( bnd.isEmpty() )
	{
		return;
	}

	// Fill in the background with the fill color.  We just paint
	// the whole screen because it takes much less code.  This is
    // slower than just filling in the unneed space, although not
	// measureably so.  The real issue is normally that it causes
	// flicker.  However, calls to OnDraw are relatively rare
	// for a web-page plug in.  So far, this has not been a problem.
    {
        COLORREF fillColor = RGB( 0, 0, 0 );
        OleTranslateColor( m_clrFillColor, NULL, &fillColor );
        HBRUSH hBrush = CreateSolidBrush( fillColor );
        FillRect( di.hdcDraw, &bnd.asRECT(), hBrush );
        DeleteObject( hBrush );  hBrush = NULL;
    }

	// Construct a rectangle the size of the image in raw image
	// coordinates.
	Rect region( pg->getSize() );

	// Find the zoom we will need to fit the image to the current window.
	const Asd::Float64 xRatio = Asd::Float64(bnd.width ()) / Asd::Float64(region.width ());
	const Asd::Float64 yRatio = Asd::Float64(bnd.height()) / Asd::Float64(region.height());
	
	const Asd::Float64 zoom = std::_cpp_min( xRatio, yRatio );

	// JBTODO:  This is not a rigourous error conidition.  Basically, there
	// is only so small we can allow an image to be and render it.  1% is
	// pretty darn small for most images.  Putting this line in here
	// prevents crashing in some situations.  In some others, it will prevent
	// the image from being rendered.  Note that some sort of check will
	// have to happen because zoom can clearly not be 0.
	if( zoom < 0.01 )
	{
		return;
	}
	
	// Change the units of region to window pixels.  (We have not moved the
	// origin though.)
	region.scaleUp( zoom );

	// Box represents the portion of the window we will actually
	// draw in.  This will be a subset of bnd because the image may
	// not fill bnd.
    Rect box( bnd );

    // Move box's origin so that it coinsides with the scaled up region.
	box.offset( -bnd.topLeft() );

    // Clip both the region and the box.  We need to clip both because
	// we will need representations in two different coordinate systems.
	region = region.intersection( box );
    box = box.intersection( region );

	// Move box back to it's original origin.  box now represents
	// the portion of the window rectangle containing image, and
	// in window coordinates.
    box.offset( bnd.topLeft() );

	// We only need to render if m_img does not already represent what we
	// want.
	if( m_dirty )
    {  
        m_dirty = false;
        pg->render( &m_img, region, 1.0/zoom, 0 );
    }

    SetDIBitsToDevice( di.hdcDraw,
                    box.left(), box.top(),
                    region.width(), region.height(),
                    region.left(), 0,
                    0, m_img.height(),
                    m_img.data(), &m_img.getBmpInfo(),
                    DIB_RGB_COLORS );
}


/*************************************************************************

  Get the name of the file we are currently viewing.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::get_FileName(BSTR *pVal)
{
    if( pVal )
    {
        *pVal = m_file? m_bstFileName : _bstr_t();
    }

	return S_OK;
}

/*************************************************************************

 View the file specified by newVal.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::put_FileName( BSTR newVal )
{
    m_bstFileName = newVal;

    std::string newName;

	// Convert newVal from 16-bit format to 8-bit format.
	// I am unsure if this is always needed, but it will
	// take so little time that writing code to figure out
	// if it needs to be done is a waste.
    {for( Asd::UInt32 ii=0; ii < m_bstFileName.length(); ++ii )
    {
        newName += char(newVal[ii]);
    }}

    // Clearly, a new file means that we must redraw the image.
	m_dirty = true;

    // Upon loading a file, we always go to the first page.
	m_currentPage = 0;

    // Get rid of the old file.
	delete m_file; m_file = NULL;

    // type is the extension of the file.  It is of the form
	// ".abcd...".
	std::string type;

	// Find the extension and set type.
    {
        Asd::UInt32 ii = newName.length() - 1;

        // Search for the ".".  Note that we start from the end
		// and work our way back.
		for( ; (ii > 0) && (newName[ii] != '.'); --ii )  {}
    
        
		// Copy the dot and everything after it into type.
		for( ; ii < newName.length(); ++ii )
        {
            type += newName[ii];
        }
    }

    // Depending on type, open the file.
    if( type.compare(".tfx") )
    {
        m_file = new MrcFile( newName.c_str() );
    }

    FireViewChange();

	return S_OK;
}


/*************************************************************************

  Return the current, 1-based page number.  Returns 0 if there
  is no file.  You can use this method to determine if there is
  a file or not because 0 will only be returned if there is no
  current file.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::get_CurrentPage(long *pVal)
{
	if( pVal )
    {
        if( m_file )
        {
            // Convert from 0-based to 1-based numbering.
            *pVal = m_currentPage + 1;
        }
        else
        {
            *pVal = 0;
        }
    }


    return S_OK;
}

/************************************************************************
 
  Set the page you wish to view.  If the page is out of range
  then nothing happens.

************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::put_CurrentPage(long newVal)
{
    // If there is no file, this method is pointless.,
	if( m_file )
    {
        // Convert from 1-based to 0-based numbering.
        newVal -= 1;

        // Make sure newVal is in the proper range.        
        if( (newVal >= 0) && (newVal < m_file->numPages()) )
        {
            // We only have to do something if our page actually changed.
            if( newVal != m_currentPage )
            {
                m_currentPage = newVal;
                m_dirty = true;
                FireViewChange();
            }
        }
    }

	return S_OK;
}

/*************************************************************************

 Change the page you are viewing to the next page.  If there is
 no next page, nothing happens.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::NextPage( void )
{
	long newPage = 0;

	get_CurrentPage( &newPage );
	put_CurrentPage( newPage + 1 );

	return S_OK;
}

/*************************************************************************

 Change the page you are viewing to the previous page.  If there is
 no previous page, nothing happens.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::PreviousPage( void )
{
	long newPage = 0;

	get_CurrentPage( &newPage );
	put_CurrentPage( newPage - 1 );

    return S_OK;
}

/*************************************************************************

  Change the page you are viewing to the first page.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::FirstPage( void )
{
	put_CurrentPage( 1 );

	return S_OK;
}

/*************************************************************************

  Change the page you are viewing to the last page.

*************************************************************************/

STDMETHODIMP CSilxScrLiteCtl::LastPage( void )
{
	long numPages = 0;

	get_NumPages( &numPages );
	put_CurrentPage( numPages );

	return S_OK;
}

/*************************************************************************

  Return the number of pages in the current document.  Return 0 if there
  is no file.  You can use this method to determine if there is a file or
  not because 0 will only be returned if there is no current file.

*************************************************************************/


STDMETHODIMP CSilxScrLiteCtl::get_NumPages( long *pVal )
{
	if( pVal )
    {
        if( m_file )
        {
            *pVal = long(m_file->numPages());
        }
        else
        {
            *pVal = 0; 
        }
    }

    return S_OK;
}
