/* cache_table.c
   
   Aaron Stump, Summer '97

   This file contains an implementation for the DgpAdt_CacheTable data structure
   (see cache_table.h for the specification of this data structure).
   */

#ifdef VXWORKS
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include "cache_table.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/cache_table.c,v 1.11 2003/01/27 21:48:28 awilliam Exp $");

typedef struct {
  void *key;
  void *element;
  Dgp_UInt32 count;
} DgpAdt_ct_elt;


#define DEFAULT_HASH_VALUE 100

static Dgp_Bool 
generic_isequal(void *cookie, DgpAdt_ht_elt e1, DgpAdt_ht_elt e2) {
  (void)cookie;
  return (e1 == e2);
}

static Dgp_UInt16
generic_hash(void *cookie, DgpAdt_ht_elt e) {
  Dgp_UInt32 l = (Dgp_UInt32)e;
  (void)cookie;
  return (Dgp_UInt16)(l % DEFAULT_HASH_VALUE);
}

static Dgp_Bool
isequal_wrapper(void *cookie, DgpAdt_ht_elt _e1, DgpAdt_ht_elt _e2) {
  DgpAdt_ct_elt *e1 = (DgpAdt_ct_elt *)_e1;
  DgpAdt_ct_elt *e2 = (DgpAdt_ct_elt *)_e2;
  DgpAdt_CacheTable ct = (DgpAdt_CacheTable)cookie;

  return ct->isKeyEqual(ct->user_cookie,e1->key,e2->key);
}

static Dgp_UInt16
hash_wrapper(void *cookie, DgpAdt_ht_elt _e) {
  DgpAdt_ct_elt *e = (DgpAdt_ct_elt *)_e;
  DgpAdt_CacheTable ct = (DgpAdt_CacheTable)cookie;

  return ct->hashKey(ct->user_cookie,e->key);
}

static void
last_call_wrapper(void *cookie) {
  DgpAdt_CacheTable ct = (DgpAdt_CacheTable)cookie;

  if (ct->last_call)
    ct->last_call(ct->user_cookie);
}

/* ctNew()

   Parameters:
   user_cookie        -- a void * that will be passed to all the functions
                         you supply whenever they're called.
   isKeyEqual         -- This htComparator function (see hash_table.h) should
                         treat its last two arguments as whatever kind of keys
			 the user intends to call ctGetElement() with.  
			 It should return TRUE if it wants to treat the two
			 keys it's given as equal, and FALSE otherwise.
   hashKey            -- This htHash function (again, see hash_table.h) should
                         hash keys of the sort that the user intends to call
		         ctGetElement() with.
   last_call          -- This htLastCall function will be called exactly once
                         by ctFree(), after the last call to the given 
			 disposer.
   getter             -- This function will be called by ctGetElement() to
                         get a new piece of data to cache.  Its first argument
			 will be the user_cookie passed in here, and its
			 second argument will be whatever key the user gives
			 ctGetElement().  It should return a void * that 
			 gives the data to cache.  getter will be called 
			 just when its key has expired (see ctTock()).
   disposer           -- This function will be called by ctTock() and ctFree()
                         to dispose of cached data (of the sort returned
			 by getter).  Its first argument will be the 
			 user_cookie given here, and its second argument will
			 be the piece of cached data to dispose of.
   flush_after        -- A piece of cached data will be disposed of (via 
                         disposer) after flush_after calls to ctTock()
			 without a call to ctGetElement() which returns
			 that piece of data.  If flush_after is less than 0,
			 cached data will never be disposed of.

   NOTE that getter and disposer cannot be NULL, but last_call, isKeyEqual,
   and hashKey may be.  If last_call is NULL, no such function will be called
   by ctFree().  If isKeyEqual or hashKey is NULL, a default function will be
   supplied.  */
ErrCode
DgpAdt_ctNew(DgpAdt_CacheTable *_ct, void *user_cookie, DgpAdt_htComparator isKeyEqual, 
			 DgpAdt_htHash hashKey,
      DgpAdt_htLastCall last_call, DgpAdt_ctElementGetter getter, 
      DgpAdt_ctElementDisposer disposer, Dgp_Int32 flush_after, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("ctNew()", DGP_PERRSTATE);
  DgpAdt_CacheTable ct = NULL;

  DGP_CHECK_NULL(_ct);

  DGP_CHECK_ALLOC_CLEANUP(ct = (DgpAdt_CacheTable)(DGP_CALLOC)((size_t)1,sizeof(*ct)));

  ct->dgp_state = dgp_state;

  DGP_ERR_PASS_CLEANUP(DgpAdt_htNew(&(ct->ht), (void *)ct, isequal_wrapper, hash_wrapper,
		       last_call_wrapper, DEFAULT_HASH_VALUE, dgp_state));

  DGP_ERR_CLEANUP(DgpAdt_liNew(&ct->elts, dgp_state));

  ct->isKeyEqual  = (isKeyEqual == (DgpAdt_htComparator)NULL) ? generic_isequal  
                                                       : isKeyEqual;
  ct->hashKey     = (hashKey    == (DgpAdt_htHash)      NULL) ? generic_hash   
                                                       : hashKey;
  ct->user_cookie = user_cookie;
  ct->getter      = getter;
  ct->disposer    = disposer;
  ct->last_call   = last_call;
  ct->flush_after = flush_after;

  *_ct = ct;
  ct = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(ct != NULL) {
    if(ct->ht != NULL) DgpAdt_htFree(ct->ht);
    if(ct->elts != NULL) DgpAdt_liFree(ct->elts);
    (DGP_FREE)(ct);
  }

  ERR_RETURN;
}
  
void
DgpAdt_ctFree(DgpAdt_CacheTable ct) {
  DgpAdt_ListNode ln;
  DgpAdt_List elts = ct->elts;
  DgpAdt_ct_elt *e;
  Dgp_Free tmp_free;

  if(ct == NULL) return;

  tmp_free = ct->DGP_FREE;

  DGPADT_FOREACH(ln,elts) {
    e = (DgpAdt_ct_elt *) DGPADT_LIGET(ln);
    ct->disposer(ct->user_cookie,e->element);
    tmp_free (e);
  }
  DgpAdt_liFree(ct->elts);
  DgpAdt_htFree(ct->ht); /* this calls last_call_wrapper(), so put it after
		     the call to ct->disposer (since that sees the
		     user's cookie) */
  (tmp_free)((void *)ct);
}

/* ctGetElement()

   If we have data cached with a key that matches (via the user's htComparator
   function) the given key, return that data.  Otherwise, call the user's
   ctElementGetter to get data to cache for that key.  If we have the data
   cached, set *key_in_table to be the key corresponding to that data. 
   Otherwise, set *key_in_table to NULL.

   Return error if failed. */
ErrCode
DgpAdt_ctGetElement(DgpAdt_CacheTable ct, void *key, void **key_in_table, void **element) 
{
  DGP_PROLOGUE("ctGetElement()", ct->DGP_PERRSTATE);
  DgpAdt_ct_elt *tmp, *from_table;

  *element = NULL;
  *key_in_table = NULL;
  DGP_CHECK_ALLOC(tmp = (DgpAdt_ct_elt*) ct->DGP_MALLOC (sizeof (DgpAdt_ct_elt)));
  tmp->key = key;
  tmp->element = NULL;
  tmp->count = 0;

  DGP_ERR_PASS(DgpAdt_htAddToTable(ct->ht,tmp,(void **)&from_table));

  if (from_table == NULL) {
    /* we don't have the data corresponding to the given
       key in this DgpAdt_CacheTable, so we need to call the
       ctElementGetter function to get it. We also need
       to add it to our linked list of elements in the
       table. */

	  DgpAdt_ListNode ln;
	  *element = tmp->element = ct->getter(ct->user_cookie,key);
	  DGP_ERR_CHECK (DgpAdt_liAdd(ct->elts, tmp, &ln)) {
      /* problem adding tmp to the DgpAdt_List */
      ct->disposer(ct->user_cookie,tmp->element);
      DGP_ERR(cErrNomem);
    }
  }
  else {
    /* we had a DgpAdt_ct_elt for this key in our DgpAdt_CacheTable already */

    *element = from_table->element;
    *key_in_table = from_table->key;
    from_table->count = 0; /* reset the count for this data */

    /* we should throw away the temporary DgpAdt_ct_elt we allocated for our
       query. */
    ct->DGP_FREE (tmp);
  }

  NORMAL_RETURN;
}

/* ctTock()

   This function increments a counter for each piece of cached data.
   Elements whose counter goes above the value of flush_after supplied
   for this DgpAdt_CacheTable when it was created with ctNew() will be flushed
   from the DgpAdt_CacheTable (each counter is incremented and then compared with
   flush_after; ctGetElement() sets the counter associated with the data
   it returns to 0). */
void
DgpAdt_ctTock(DgpAdt_CacheTable ct) {
  DgpAdt_ListNode ln;
  DgpAdt_List elts = ct->elts;
  DgpAdt_ct_elt *tmp, *dummy;
  Dgp_UInt32 flush_after;

  if (ct->flush_after < 0) 
    return;
  flush_after = (Dgp_UInt32)ct->flush_after;

  for (ln = DGPADT_LIFIRST(elts); ln != DGP_NULLLISTNODE;) {
    /* for each element in the DgpAdt_List (and hence, in the HashTable) */
    tmp = (DgpAdt_ct_elt *) DGPADT_LIGET(ln);
    if (++(tmp->count) > flush_after) {
      /* we need to flush the data associated with tmp from our cache */
      DgpAdt_htRemoveFromTable(ct->ht,(void *)tmp, (void **)&dummy);
      ct->disposer(ct->user_cookie,tmp->element);
      DgpAdt_liRemAndNext(ln, &ln);
    }
    else {
      ln = DGPADT_LINEXT(ln);
    }
  }
}
