/*
*
*	$Header: /project/icons/cvsroot/digipaper/src/libadt/priq.c,v 1.7 2003/01/27 21:48:29 awilliam Exp $
*
* Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
* Reserved.
*
* Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
*
* Use, reproduction and preparation of derivative works of this software is
* permitted.  Any copy of this software or of any derivative work must
* include both the above copyright notices of Cornell University and Xerox
* Corporation and this paragraph.  This software is made available AS IS, and
* XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
* WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
* FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
* HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
* EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
* NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGES.
*/
/*
* Xerox may use this code without restriction.
*/


/*
*	priq.c - Priority Queues
*
* This module handles generic priority queues.
*
* Exports:
*	type DgpAdt_PriQ
*
*	constant DgpAdt_PriQ DGP_NULLPRIQ
*	constant DgpAdt_PriQNode DGP_NULLPRIQNODE
*
*	DgpAdt_PriQ pqNew(Dgp_Bool (*compareGT)(void *a, void *b)) - create a new,
*		empty priority queue
*
*	void pqFree(DgpAdt_PriQ p) - free priority queue p
*
*	DgpAdt_PriQNode pqFirst(DgpAdt_PriQ p) - return the first entry in p
*
*	DgpAdt_PriQNode pqAdd(DgpAdt_PriQ p, void *userdata) - add a node containing
*		userdata to p
*
*	void pqRem(DgpAdt_PriQNode pqn) - remove pqn from the queue containing it
*
*	void *pqGet(DgpAdt_PriQNode pqn) - return pqn's userdata
*
*	int pqLen(DgpAdt_PriQ p) - return the number of nodes in p
*
*	Dgp_Bool (*)(void *, void *) pqGetCmpFunction(DgpAdt_PriQ p) - get the
*		comparison function from an existing priority queue.
*
* An object of type DgpAdt_PriQ is used as a handle on a priority queue.
*	Each element of the queue contains a void *
*	field, which is used for whatever data the user wishes to attach
*	to that node.
*
* A DgpAdt_PriQNode refers to an entry in a priority queue.
*
* pqNew creates an empty queue. If this is not possible, it returns DGP_NULLPRIQ.
*	The function which it takes is used as the comparison function to
*	sort the queue. It should be declared as
*	Dgp_Bool compareGT(void *a, void *b)
*	and it returns FALSE if "a" <= "b" and DGP_TRUE if "a" > "b".
*	It does not compare the pointers themselves (priq could do that
*	itself), but rather compares the user data values associated with
*	those pointers in whatever way the user desires (which should be
*	a total order). It will be called as required to ensure that
*	pqFirst returns the smallest valued entry.
*
* pqFree frees the storage associated with a priority queue.
*	Note that it does not free the userdata; the user
*	must do this before freeing the queue structure.
*
* pqFirst returns the first (smallest) entry of the given queue.
*	It does not guarantee any order among equal elements.
*	If the queue is empty, it returns DGP_NULLPRIQNODE.
*
* pqAdd adds an entry to the given priority queue. It returns the new
*	DgpAdt_PriQNode. If it could not allocate needed storage, it returns
*	DGP_NULLPRIQNODE.
*
* pqRem removes the given DgpAdt_PriQNode from its DgpAdt_PriQ. The DgpAdt_PriQNode is then
*	invalid.
*
* pqGet returns the userdata value of some DgpAdt_PriQNode.
*
* pqLen returns the number of nodes in a given DgpAdt_PriQ (i.e. the number of
*	successful pqAdds minus the number of pqRems).
*
* Changing the values pointed to by the userdata pointers may result
*	in pqFirst not returning the smallest current value. To change these,
*	perform a pqRem and a pqAdd.
*
* pqAdd, pqRem and other operations involving DgpAdt_PriQNodes may be arbitrarily
*	interleaved. A DgpAdt_PriQNode is valid from the moment pqAdd first returns
*	it until it is passed into pqRem, or its containing DgpAdt_PriQ is passed
*	to pqFree.
*
* pqGetCmpFunction gets the comparison function from a DgpAdt_PriQ. This lets you
*	set up a duplicate of a priority queue without having to know
*	from other sources what sorting function it is using.
*/

#include "util.h"
#include "priq.h"
#include <stdlib.h>
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/priq.c,v 1.7 2003/01/27 21:48:29 awilliam Exp $");
#include "PsgFix.h"

#define	DGP_NULLPRIQINTNODE	((DgpAdt_PriQIntNode)NULL)

static void swap(DgpAdt_PriQIntNode pqin1, DgpAdt_PriQIntNode pqin2);

/*
* This module uses the small pool to allocate and free various things
* efficiently.
*/
#define PQ_POOL_ID   DGP_SMALLPOOL_MIN20
#define PQN_POOL_ID  DGP_SMALLPOOL_MIN8
#define PQIN_POOL_ID DGP_SMALLPOOL_MIN28

#define PQ_SEC_MALLOC(_pppq) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, PQ_POOL_ID, (void **)(_pppq)))
#define PQ_FREE(_ppq) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, PQ_POOL_ID, (void *)(_ppq)))

#define PQN_SEC_MALLOC(_pppqn) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, PQN_POOL_ID, (void **)(_pppqn)))
#define PQN_FREE(_ppqn) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, PQN_POOL_ID, (void *)(_ppqn)))

#define PQIN_SEC_MALLOC(_pppqin) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, PQIN_POOL_ID, (void **)(_pppqin)))
#define PQIN_FREE(_ppqin) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, PQIN_POOL_ID, (void *)(_ppqin)))



/* static void * pqinChunk = (void *)NULL;
static struct _DgpAdt_PriQIntNode *malloc_pqin(Dgp_State *dgp_state)
{
struct _DgpAdt_PriQIntNode *newt;
if (pqinChunk == (void *)NULL) {
pqinChunk = (void *)chNew(sizeof(struct _DgpAdt_PriQIntNode), dgp_state);
if (pqinChunk == NULLCHUNK) {
return((struct _DgpAdt_PriQIntNode *)NULL);
}
if (!reRegister(&pqinChunk, chDestroy, dgp_state)) {
chDestroy(pqinChunk, dgp_state);
pqinChunk = (void *)NULL;
return((struct _DgpAdt_PriQIntNode *)NULL);
}
}
newt = (struct _DgpAdt_PriQIntNode *)chAlloc((Chunk)pqinChunk, dgp_state);
return(newt);
}

  static void free_pqin(struct _DgpAdt_PriQIntNode *oldt)
  {
  DGP_ASSERT(oldt != (struct _DgpAdt_PriQIntNode *)NULL);
  DGP_ASSERT(pqinChunk != (void *)NULL);
  chFree((Chunk)pqinChunk, (void *)oldt);
  return;
} */


ErrCode
DgpAdt_pqNew(Dgp_Bool (*compareGT)(void *a, void *b), Dgp_State *dgp_state, 
			 DgpAdt_PriQ *retpriq)
{
	
	DGP_PROLOGUE ("DgpAdt_pqNew", DGP_PERRSTATE);
	
  /* check small pool usage */
  DGP_SMALLPOOL_VALIDATE(struct _DgpAdt_PriQ, PQ_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(struct _DgpAdt_PriQNode, PQN_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(struct _DgpAdt_PriQIntNode, PQIN_POOL_ID);

    DGP_ASSERT(compareGT != (Dgp_Bool (*)(void *a, void *b))NULL);
	  PQ_SEC_MALLOC(retpriq);
	
    (*retpriq)->root = (*retpriq)->last = DGP_NULLPRIQINTNODE;
    (*retpriq)->compareGT = compareGT;
    (*retpriq)->count = 0;
    (*retpriq)->dgp_state = dgp_state;
	   
    NORMAL_RETURN;
}

ErrCode
DgpAdt_pqFree(DgpAdt_PriQ p)
{
    DgpAdt_PriQIntNode pqin, nextpqin;
	Dgp_State *dgp_state = p->dgp_state;
	
    DGP_PROLOGUE ("DgpAdt_pqFree", DGP_PERRSTATE);
    DGP_STOP_WHINE;
	
    DGP_ASSERT(p != DGP_NULLPRIQ);
    for (pqin = p->root; pqin != DGP_NULLPRIQINTNODE; pqin = nextpqin) {
		nextpqin = pqin->next;
		PQN_FREE(pqin->owner);
		PQIN_FREE(pqin);
	}
    PQ_FREE(p);
	
	NORMAL_RETURN;
}

ErrCode
DgpAdt_pqRem(DgpAdt_PriQNode pqn)
{
    DgpAdt_PriQIntNode pqlast;
    DgpAdt_PriQIntNode smallchild;
    DgpAdt_PriQIntNode pqin;
    DgpAdt_PriQ p;
    Dgp_State *dgp_state = pqn->owner->dgp_state;
	
    DGP_PROLOGUE ("DgpAdt_pqRem", DGP_PERRSTATE);
    DGP_STOP_WHINE;
	
    DGP_ASSERT(pqn != DGP_NULLPRIQNODE);
    DGP_ASSERT(pqn->val != DGP_NULLPRIQINTNODE);
    DGP_ASSERT(pqn->val->owner == pqn);
    p = pqn->owner;
    DGP_ASSERT(p != DGP_NULLPRIQ);
    DGP_ASSERT(p->root != DGP_NULLPRIQINTNODE);
    DGP_ASSERT(p->count > 0);
	
    /* Was that the last thing in the queue? */
    if (p->root == p->last) {
		DGP_ASSERT(pqn->val == p->root);
		DGP_ASSERT(p->count == 1);
		PQIN_FREE(p->root);
		PQN_FREE(pqn);
		p->root = p->last = DGP_NULLPRIQINTNODE;
		
		p->count--;
		NORMAL_RETURN;
	}
	
    if (pqn->val == p->last) {
	/*
	* They're about to remove the last element. Don't need
	* to reheapify - just patch the data structures
		*/
		pqlast = p->last;
		DGP_ASSERT(pqlast->parent != DGP_NULLPRIQINTNODE);
		if (pqlast->parent->right == pqlast) {
			pqlast->parent->right = DGP_NULLPRIQINTNODE;
		}
		else {
			DGP_ASSERT(pqlast->parent->left == pqlast);
			pqlast->parent->left = DGP_NULLPRIQINTNODE;
		}
		DGP_ASSERT(pqlast->prev != DGP_NULLPRIQINTNODE);
		pqlast->prev->next = DGP_NULLPRIQINTNODE;
		p->last = pqlast->prev;
		PQIN_FREE(pqlast);
		PQN_FREE(pqn);
		
		p->count--;
		NORMAL_RETURN;
	}
	
    /* Now, move the last element up to the to-be deleted element */
    pqlast = p->last;
    pqn->val->userdata = pqlast->userdata;
    DGP_ASSERT(pqlast->owner != DGP_NULLPRIQNODE);
    DGP_ASSERT(pqlast->owner->val == pqlast);
    pqlast->owner->val = pqn->val;
    pqn->val->owner = pqlast->owner;
	
    /* and delete the last element structure */
    DGP_ASSERT(pqlast->parent != DGP_NULLPRIQINTNODE);	/*
													* if it was the root,
													* it was handled above
	*/
    /* which child was it? */
    if (pqlast->parent->right == pqlast) {
		pqlast->parent->right = DGP_NULLPRIQINTNODE;
	}
    else {
		DGP_ASSERT(pqlast->parent->left == pqlast);
		pqlast->parent->left = DGP_NULLPRIQINTNODE;
	}
	
    /* if it was the root as well, it was handled above */
    DGP_ASSERT(pqlast->prev != DGP_NULLPRIQINTNODE);
    pqlast->prev->next = DGP_NULLPRIQINTNODE;
    p->last = pqlast->prev;
    PQIN_FREE(pqlast);
	
    /* Now reheapify. */
    /*
	* We have just taken some random value and stuck it into the middle
	* of the heap. It might have to move up or down
	*/
    pqin = pqn->val;
    if ((pqin->parent == DGP_NULLPRIQINTNODE) ||
		((*(p->compareGT))(pqin->userdata, pqin->parent->userdata))) {
		/* Reheapify down, if at all */
		while (DGP_TRUE) {
			if (pqin->left == DGP_NULLPRIQINTNODE) {
				/* Hit the bottom */
				break;
			}
			
			/* Find which child has the smaller value */
			if ((pqin->right == DGP_NULLPRIQINTNODE) ||
				((*(p->compareGT))(pqin->right->userdata,
				pqin->left->userdata))) {
				smallchild = pqin->left;
			}
			else {
				smallchild = pqin->right;
			}
			
			/* Swap if necessary */
			if ((*(p->compareGT))(pqin->userdata, smallchild->userdata)) {
				/* Swap the two */
				swap(pqin, smallchild);
				pqin = smallchild;
			}
			else {
				/* Not smaller - heapification has been done */
				break;
			}
		}
	}
    else {
		/* Need to reheapify upwards */
		while (DGP_TRUE) {
			if ((pqin->parent != DGP_NULLPRIQINTNODE) &&
				((*(p->compareGT))(pqin->parent->userdata, pqin->userdata))) {
				/* Swap it and its parent */
				swap(pqin, pqin->parent);
				pqin = pqin->parent;
			}
			else {
				break;
			}
		}
	}
	
    /* and get rid of the DgpAdt_PriQNode */
    PQN_FREE(pqn);
	
    /* and adjust the count */
    p->count--;
	NORMAL_RETURN;
}

ErrCode
DgpAdt_pqAdd(DgpAdt_PriQ p, void *userdata, DgpAdt_PriQNode *retpriqn)
{
	DgpAdt_PriQIntNode pqin;
	DgpAdt_PriQIntNode newpqin;
	Dgp_State *dgp_state = p->dgp_state;
	
	
	DGP_PROLOGUE ("pqAdd", DGP_PERRSTATE);
	
	DGP_ASSERT(p != DGP_NULLPRIQ);
	PQN_SEC_MALLOC(retpriqn);
	PQIN_SEC_MALLOC(&newpqin);
	
	newpqin->left = newpqin->right = newpqin->next = DGP_NULLPRIQINTNODE;
	newpqin->owner = (*retpriqn);
	newpqin->userdata = userdata;
	(*retpriqn)->val = newpqin;
	(*retpriqn)->owner = p;
	
	if (p->root == DGP_NULLPRIQINTNODE) {
		/* The first thing in the queue */
		DGP_ASSERT(p->count == 0);
		
		p->root = p->last = newpqin;
		newpqin->prev = DGP_NULLPRIQINTNODE;
		newpqin->parent = DGP_NULLPRIQINTNODE;
		p->count++;
		NORMAL_RETURN;
	}
	
	DGP_ASSERT(p->count > 0);
	
	newpqin->prev = p->last;
	if (p->last->parent == DGP_NULLPRIQINTNODE) {
		/* The last element is also the root */
		DGP_ASSERT(p->last == p->root);
		p->last->left = newpqin;
		newpqin->parent = p->last;
	}
	else if (p->last->parent->right == DGP_NULLPRIQINTNODE) {
		/* The last entry's parent has an empty slot - use it */
		p->last->parent->right = newpqin;
		newpqin->parent = p->last->parent;
	}
	else {
		DGP_ASSERT(p->last->parent->next != DGP_NULLPRIQINTNODE);
		DGP_ASSERT(p->last->parent->next->left == DGP_NULLPRIQINTNODE);
		p->last->parent->next->left = newpqin;
		newpqin->parent = p->last->parent->next;
	}
	p->last->next = newpqin;
	p->last = newpqin;
	
	/* Now reheapify */
	for (pqin = p->last; ; pqin = pqin->parent) {
		if ((pqin->parent != DGP_NULLPRIQINTNODE) &&
			((*(p->compareGT))(pqin->parent->userdata, pqin->userdata))) {
			/* Swap it and its parent */
			swap(pqin, pqin->parent);
		}
		else {
			break;
		}
	}
	
	/* Successfully added */
	p->count++;
	NORMAL_RETURN;
}

Dgp_Bool
(*DgpAdt_pqGetCmpFunction(DgpAdt_PriQ p))(void *, void *)
{
	Dgp_State *dgp_state = p->dgp_state;
	
	DGP_VOID_PROLOGUE ("pqGetCmpFunction", DGP_PERRSTATE);
	
	DGP_VOID_ASSERT(p != DGP_NULLPRIQ);
	return(p->compareGT);
}

static void
swap(DgpAdt_PriQIntNode pqin1, DgpAdt_PriQIntNode pqin2)
{
	void *temp;
	DgpAdt_PriQNode tempqn;
	Dgp_State *dgp_state = pqin1->owner->owner->dgp_state;
	
	DGP_VOID_PROLOGUE ("DgpAdt_PriQIntNode", DGP_PERRSTATE);
	
	DGP_VOID_ASSERT(pqin1 != DGP_NULLPRIQINTNODE);
	DGP_VOID_ASSERT(pqin2 != DGP_NULLPRIQINTNODE);
	
	/* Swap the values */
	temp = pqin1->userdata;
	pqin1->userdata = pqin2->userdata;
	pqin2->userdata = temp;
	
	/* Now, swap the owner's pointers to them */
	DGP_VOID_ASSERT(pqin1->owner != DGP_NULLPRIQNODE);
	DGP_VOID_ASSERT(pqin2->owner != DGP_NULLPRIQNODE);
	DGP_VOID_ASSERT(pqin1->owner->val == pqin1);
	DGP_VOID_ASSERT(pqin2->owner->val == pqin2);
	
	tempqn = pqin1->owner;
	pqin1->owner = pqin2->owner;
	pqin2->owner = tempqn;
	pqin1->owner->val = pqin1;
	pqin2->owner->val = pqin2;
}
