#include "seekable_buffer.h"
#include "util.h"
#ifdef VXWORKS
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include <stdio.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <string.h> /* for memcpy() */
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/seekable_buffer.c,v 1.14 2003/01/27 21:48:29 awilliam Exp $");

/* sbNew()

   Create a new DgpAdt_SeekableBuffer.

   Parameters:
   buf        -- the buffer to use
   buflen     -- the size of buf
   a          -- the suite of allocators to use.
   do_realloc -- reallocate the buffer (using a->realloc) when necessary
                 iff do_realloc is non-NULL.

   NOTE that buf may not be NULL. 
*/
ErrCode
DgpAdt_sbNew(DgpAdt_SeekableBuffer *_s, Dgp_Int8 *buf, Dgp_UInt32 buflen, Dgp_Bool do_realloc, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("sbNew()", DGP_PERRSTATE);
  DgpAdt_SeekableBuffer s = NULL;
  
  DGP_CHECK_NULL(_s);

  DGP_ERR_GEN((buf == NULL), cErrInvalidParm);

  DGP_CHECK_ALLOC_CLEANUP(s = (DgpAdt_SeekableBuffer)(DGP_MALLOC)(sizeof(*s)));

  /* max_cur tells us the max. position cur has ever been while doing writing
     (max_cur is not updated by sbRead()) */
  s->max_cur    = s->cur = s->buf = buf;
  s->buflen     = buflen;
  s->do_realloc = do_realloc;
  s->dgp_state = dgp_state;

  *_s = s;
  s = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(s) (DGP_FREE)(s);

  ERR_RETURN;
}

/* sbFree()
   
   Free a DgpAdt_SeekableBuffer.

   NOTE that the DgpAdt_SeekableBuffer's actual buffer in memory will _not_ be 
   freed by this function.  It is the caller's responsibility to get
   it using sbGetBuffer() to clean it up (particularly if the DgpAdt_SeekableBuffer
   was created to realloc the buffer if necessary). */
void 
DgpAdt_sbFree(DgpAdt_SeekableBuffer s) {
  Dgp_Free tmp_free;

  if(s == NULL) return;

  tmp_free = s->DGP_FREE;

  (tmp_free)((void *)s);
}

void 
DgpAdt_sbGetBuffer(DgpAdt_SeekableBuffer s, Dgp_Int8 **buf, Dgp_UInt32 *buflen, 
	    Dgp_UInt32 *len_written) {
  *buf         = s->buf;
  *buflen      = s->buflen;
  *len_written = s->max_cur - s->buf;
}

/* sbSeek()

   This function works just like fseek(), only on a memory buffer
   instead of a file. */
Dgp_Int32 
DgpAdt_sbSeek(DgpAdt_SeekableBuffer s, Dgp_Int32  offset, Dgp_Int32 origin) {
  Dgp_Int32  pos_to_seek_to = -1; /* if origin isn't anything we know about,
			       we'll automatically bail at the guard
			       if (pos_to_seek_to < 0) */
  
  switch(origin) {
  case SEEK_SET:
    pos_to_seek_to = offset;
    break;
  case SEEK_CUR:
    pos_to_seek_to = (s->cur - s->buf) + offset; /* these are (Dgp_Int8 *)'s,
						    so this does the right
						    thing */
    break;
  case SEEK_END:
    pos_to_seek_to = (s->max_cur - s->buf) + offset;
    break;
  }

  if (pos_to_seek_to < 0) 
    /* We can't seek to a position before the start of the buffer, so
       bail */
    return -1;
    
  /* NOTE on implementation: if the user told us not to realloc this
     buffer, then if s/he tries to seek past the end of it, we'll complain.
     If the user told us we could realloc this buffer and s/he tries to
     seek past the end of it, we'll let her/him do it.  We'll set our cur
     pointer to the appropriate position, and if s/he does an sbRead()
     or sbWrite(), we'll realloc the buffer then.  We won't realloc it
     now, because we might not need to. */
  if ((Dgp_UInt32)pos_to_seek_to > s->buflen) 
    if (!s->do_realloc) 
      /* we're going to complain here (see NOTE above) */
      return -1;

  s->cur = &s->buf[pos_to_seek_to];
  return 0;
}

/* sbTell()

   This function works just like ftell(), only on a memory buffer
   instead of a file. */
Dgp_Int32 
DgpAdt_sbTell(DgpAdt_SeekableBuffer s) {
  return (Dgp_Int32 )(s->cur - s->buf);
}

/* sbRead()

   This function works just like fread(), only on a memory buffer
   instead of a file. */
size_t
DgpAdt_sbRead(void *ptr, size_t size, size_t nobj, DgpAdt_SeekableBuffer s) {
  Dgp_UInt32 nobjRead;
  Dgp_UInt32 cur_ind;
  Dgp_Int8 *cptr = (Dgp_Int8 *)ptr;

  /* Part of the spec for fread*/
  if(!size || !nobj){
    return 0;
  }

  /* NOTE that cur may already be pointing past the end of the buffer,
     because of the way we do seeks.  See sbSeek(). */
  cur_ind = s->cur - s->buf;

  for(nobjRead = 0; nobjRead < nobj; nobjRead++) {
    /* We read as many obj as we can and
       return howmany we actually read */
    if(s->cur + size >= s->max_cur){
      /* We have or will hit the end of the file
	 set s->cur to the end of the file
	 and return the number of obj read
	 This is the spec for fread*/
      s->cur = s->max_cur;
      break;
    }
    memcpy(cptr, s->cur, size);
    s->cur += size;
    cptr += size;
  }

  return nobjRead;

}

/* sbWrite()

   This function works just like fwrite(), only on a memory buffer
   instead of a file. */
size_t
DgpAdt_sbWrite(void *ptr, size_t size, size_t nobj, DgpAdt_SeekableBuffer s) {
  Dgp_UInt32 i;
  Dgp_UInt32 len, cur_ind, max_cur_ind;
  void *tmp;
  Dgp_Int8 *cptr = (Dgp_Int8 *)ptr;
  
  /* NOTE that cur may already be pointing past the end of the buffer,
     because of the way we do seeks.  See sbSeek(). */
  cur_ind = s->cur - s->buf;
  /* this is how much space we'd have to have in our buffer
     to be able to perform this operation */
  len = cur_ind + nobj * size;

  if (len > s->buflen) {
    if (!s->do_realloc)
      /* we need more space, but the user told us not to realloc.
	       Bail. */
      return 0;

    /* realloc the buffer */
    if (!(tmp = s->DGP_REALLOC(s->buf, len)))
      /* we couldn't realloc the buffer, so bail */
      return 0;

    max_cur_ind = s->max_cur - s->buf;

    s->buf     = tmp;
    s->cur     = &s->buf[cur_ind];
    s->max_cur = &s->buf[max_cur_ind];
    s->buflen  = len;

  }

  for (i = 0; i < nobj; i++) {
    memcpy(s->cur,cptr,size);

    /* advance both pointers by size bytes */
    s->cur += size;
    cptr += size; 
  }

  /* max_cur tells us how far cur ever advanced while doing writing */
  s->max_cur = MAX(s->max_cur,s->cur);

  return nobj;
}

