
/* sm_image.c
   
   Aaron Stump, Summer '96

   (NOTE on implementation: We maintain the maximal y- and x-coordinates the
   user passes to us in smGet() and smSet(), and the actual domain for
   which we've allocated memory in do_*_allocation_*)
*/

#include "dgp_types.h"

#if defined (_WIN32) || defined (VXWORKS) || defined (EQUATOR_CA_ENV) || defined (LYNX)
/* MSDEV has a different include and a different name for
   UNIX's MAXSHORT */
#include <limits.h>
const Dgp_Int16 MAXSHORT = (SHRT_MAX);
#else
#include <values.h>
/* 26.10.2000 / dhtesq. */
/* Unix has SHRT_MAX here, so use that to stop compiler whine. */
#include <limits.h>
#endif

#include "sm_image.h"

#ifdef VXWORKS
#include <stdlib.h>
#else
#include "malloc.h"
#endif

#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/sm_image.c,v 1.19 2003/02/10 22:53:18 mkrocken Exp $");

/* when allocating memory for an array of DgpAdt_sm_elt_types for the first
   time, allocate as much of (2 * INIT_RADIUS) + 1 as possible */
#define INIT_RADIUS 10

/* always_clear -- If TRUE, then whenever additional memory is allocated 
                   for this DgpAdt_SMImage, calloc() is the allocator used. Otherwise,
		   malloc() may be used.
*/
ErrCode
DgpAdt_smNew(DgpAdt_SMImage *_sm, Dgp_Bool always_clear, Dgp_UInt32 init_maxy, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("smNew()", DGP_PERRSTATE);
  DgpAdt_SMImage sm = NULL;

  DGP_CHECK_NULL(_sm);

  DGP_CHECK_ALLOC_CLEANUP(sm = (DgpAdt_SMImage)(DGP_MALLOC)(sizeof(*sm)));
  sm->dgp_state = dgp_state;
  
  sm->always_clear = always_clear;
  sm->maxy = sm->maxy_alloced = -1;
  
  sm->init_maxy = ((init_maxy <= 0) ? 100 : init_maxy);

  *_sm = sm;
  sm = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(sm != NULL) (DGP_FREE)(sm);

  ERR_RETURN;
}

void 
DgpAdt_smFree(DgpAdt_SMImage sm) 
{
  Dgp_Free tmp_free;
  Dgp_Int32 y, yend;
  DgpAdt_sm_elt_type **x_arr;

  if(sm == NULL) return;

  tmp_free = sm->DGP_FREE;

  if (sm->maxy_alloced == -1) {
    /* We haven't allocated any memory for this DgpAdt_SMImage
       except the highest-level struct itself */
    (tmp_free)((void *)sm);
    return;
  }

  x_arr = sm->x_arr;

  /* Free the arrays of sm_elts */
  for (y = 0, yend = sm->maxy; y <= yend; y++) {
    if (x_arr[y])
      (tmp_free)((void *)x_arr[y]);
  }
  
  /* Free the other arrays */
  (tmp_free)((void *)x_arr);
  (tmp_free)((void *)sm->x_min);
  (tmp_free)((void *)sm->x_len);
  (tmp_free)((void *)sm);
}

static ErrCode
dgpadt_do_first_allocation_y(DgpAdt_SMImage sm, Dgp_Int32 new_maxy_alloced) 
{
  DGP_PROLOGUE("dgpadt_do_first_allocation_y()", sm->DGP_PERRSTATE);
  Dgp_Int32 len = new_maxy_alloced + 1;
  Dgp_Allocators *a = &(sm->DGP_ALLOCATORS);
  /* initialization for easy cleanup on failure */
  sm->x_arr = NULL;
  sm->x_min = NULL;
  sm->x_len = NULL;
  
  /* For all these allocations, we need to use calloc */
  
  /* allocate the arrays of pointers to arrays */
  DGP_CHECK_ALLOC_CLEANUP(sm->x_arr = 
    (DgpAdt_sm_elt_type **)(a->calloc)((size_t)len,sizeof(DgpAdt_sm_elt_type *)));
  
  /* allocate the array giving the minimum x-coordinate for each line */
  DGP_CHECK_ALLOC_CLEANUP(sm->x_min =
    (Dgp_UInt16 *)(a->calloc)((size_t)len,sizeof(Dgp_UInt16)));
  
  /* allocate the array giving the length of memory allocated
     for each line */
  DGP_CHECK_ALLOC_CLEANUP(sm->x_len =
    (Dgp_UInt16 *)(a->calloc)((size_t)len,sizeof(Dgp_UInt16)));
  
  sm->maxy_alloced = new_maxy_alloced;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (sm->x_arr)
    (a->free)((void *)sm->x_arr);
  if (sm->x_min)
    (a->free)((void *)sm->x_min);
  if (sm->x_len)
    (a->free)((void *)sm->x_len);

  ERR_RETURN;
}

/* NOTE that on failure, any of sm->x_arr, sm->x_min, sm->x_len may
   have greater than sm->maxy number of elements.  But this function is
   supposed to keep track of all memory even in the case of failure,
   so there should be no leaks and calling smFree() should still be OK. */
static ErrCode
dgpadt_do_allocation_y(DgpAdt_SMImage sm, Dgp_Int32 new_maxy) 
{
  DGP_PROLOGUE("dgpadt_do_allocation_y()", sm->DGP_PERRSTATE);
  DgpAdt_sm_elt_type **new_x_arr;
  Dgp_UInt16 *new_ushort, *new_ushort2;
  Dgp_Int32 y,len = new_maxy + 1;
  Dgp_Allocators *a = &(sm->DGP_ALLOCATORS);

  /* reallocate 1 */
  DGP_CHECK_ALLOC_CLEANUP(new_x_arr = 
    (DgpAdt_sm_elt_type **)
    (a->realloc)((void *)sm->x_arr,
			 (size_t)(len * sizeof(DgpAdt_sm_elt_type *))));
  sm->x_arr = new_x_arr;
  
  /* reallocate 2 */
  DGP_CHECK_ALLOC_CLEANUP(new_ushort = 
    (Dgp_UInt16 *)
    (a->realloc)((void *)sm->x_min,
			 (size_t)(len * sizeof(Dgp_UInt16))));
  sm->x_min = new_ushort;

  /* reallocate 3 */
  DGP_CHECK_ALLOC_CLEANUP(new_ushort2 = 
    (Dgp_UInt16 *)
    (a->realloc)((void *)sm->x_len,
			 (size_t)(len * sizeof(Dgp_UInt16))));
  sm->x_len = new_ushort2;

  /* Initialize all the new memory to zero or NULL */
  for (y = sm->maxy_alloced+1; y < len; y++) {
    new_x_arr[y]   = NULL;
    new_ushort[y]  = 0;
    new_ushort2[y] = 0;
  }

  sm->maxy_alloced = new_maxy;
  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

static ErrCode
dgpadt_do_first_allocation_x(DgpAdt_SMImage sm, Dgp_Int32 y, Dgp_Int32 x) 
{
  DGP_PROLOGUE("dgpadt_do_first_allocation_x()", sm->DGP_PERRSTATE);
  Dgp_Int32 new_minx, tempx;
  Dgp_Allocators *a = &(sm->DGP_ALLOCATORS);
  
  /* Handle some boundary cases */
  new_minx = x - INIT_RADIUS;
  if (new_minx < 0)
    new_minx = 0;
  
  tempx = x + INIT_RADIUS;
  /* 26.10.2000 / dhtesq.  Changed from MAXSHORT to SHRT_MAX (compiler whine). */
  if(tempx > (SHRT_MAX))
    tempx = (SHRT_MAX); 
  /* here tempx is maxx for this column */
  
  tempx = tempx - new_minx + 1; /* tempx is now the length of the memory
				   we'll allocate for this column */
  
  /* now allocate the memory */
  if (sm->always_clear) {
    DGP_CHECK_ALLOC_CLEANUP(sm->x_arr[y] = 
      (DgpAdt_sm_elt_type *)(a->calloc)((size_t)tempx, sizeof(DgpAdt_sm_elt_type)));
  } else {
    DGP_CHECK_ALLOC_CLEANUP(sm->x_arr[y] = 
      (DgpAdt_sm_elt_type *)(a->malloc)((size_t)(tempx 
						  * sizeof(DgpAdt_sm_elt_type))));
  }
  
  sm->x_min[y] = (Dgp_UInt16) new_minx;
  sm->x_len[y] = (Dgp_UInt16) tempx;
  NORMAL_RETURN;

ERR_CLEANUP:

  if(sm->x_arr[y] != NULL) (a->free)(sm->x_arr[y]);

  ERR_RETURN;
}

/* Reallocate the y'th column of sm. This function handles the
   case where x is less than x_min[y]. 

   Sadly, this operation involves copying the entire column that's already
   been allocated (since you can't tell realloc where in the new block of
   memory you want it to copy the old memory you're reallocating). */
static ErrCode
dgpadt_do_reallocation_x_min(DgpAdt_SMImage sm, Dgp_Int32 y, Dgp_Int32 x) 
{
  DGP_PROLOGUE("dgpadt_do_reallocation_x_min()", sm->DGP_PERRSTATE);
  Dgp_Int32 new_len,i,iend;
  DgpAdt_sm_elt_type *xarr = NULL;
  DgpAdt_sm_elt_type *ptr;
  Dgp_Allocators *a = &(sm->DGP_ALLOCATORS);
  
  /* figure out how much additional memory we need */
  new_len = sm->x_len[y] + (sm->x_min[y] - x);
  
  /* allocate it */
  if (sm->always_clear) {
    DGP_CHECK_ALLOC_CLEANUP(xarr = (DgpAdt_sm_elt_type *)(a->calloc)((size_t)new_len, 
					      sizeof(DgpAdt_sm_elt_type)));
  } else {
    DGP_CHECK_ALLOC_CLEANUP(xarr = (DgpAdt_sm_elt_type *)(a->malloc)((size_t)(new_len 
						       * sizeof(DgpAdt_sm_elt_type))));
  }
  
  /* copy the old array into the new memory, starting from the end with
     the higher address of each array and working down. */
  for(iend = 0, i = sm->x_len[y] - 1,ptr = &xarr[new_len - 1]; 
      i >= iend; 
      i--, ptr--)
    *ptr = sm->x_arr[y][i];
  
  (a->free)((void *)(sm->x_arr[y])); /* free the old array */
  sm->x_arr[y] = xarr;
  xarr = NULL;
  sm->x_min[y] = (Dgp_UInt16) x;
  sm->x_len[y] = (Dgp_UInt16) new_len;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(xarr != NULL) (a->free)(xarr);

  ERR_RETURN;
}

/* Reallocate the y'th column of sm. This function handles the
   case where x is greater than (x_min[y] + x_len[y] - 1). */
static ErrCode
dgpadt_do_reallocation_x_len(DgpAdt_SMImage sm, Dgp_Int32 y, Dgp_Int32 x) 
{
  DGP_PROLOGUE("dgpadt_do_reallocation_x_len()", sm->DGP_PERRSTATE);
  Dgp_Int32 new_len,i,iend;
  DgpAdt_sm_elt_type *xarr = NULL;
  
  /* figure out how much additional memory we need */
  new_len = x - sm->x_min[y] + 1;
  
  /* allocate it */
  DGP_CHECK_ALLOC_CLEANUP(xarr = (DgpAdt_sm_elt_type *)
    (sm->DGP_REALLOC)((void *)(sm->x_arr[y]),
			 (size_t)(new_len * sizeof(DgpAdt_sm_elt_type))));
  
  /* clear it if necessary */
  if (sm->always_clear) {
    for(i = sm->x_len[y], iend = x - sm->x_min[y]; i <= iend; i++)
      xarr[i] = NULL;
  }
  
  sm->x_len[y] = (Dgp_UInt16) new_len;
  sm->x_arr[y] = xarr;
  xarr = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:
  
  if(xarr != NULL) (sm->DGP_FREE)(xarr);
  
  ERR_RETURN;
}

ErrCode
DgpAdt_smGet(DgpAdt_SMImage sm, Dgp_UInt16 x, Dgp_UInt16 y, DgpAdt_sm_elt_type *e) 
{
  DGP_PROLOGUE("smGet()", sm->DGP_PERRSTATE);
  Dgp_Int32 temp;

  /* Y COORDINATE */

  if (y > sm->maxy_alloced) {
    if (sm->maxy_alloced == -1) {
      /* we haven't allocated any internal structures 
	 for this DgpAdt_SMImage yet: go and get some memory. */
      DGP_ERR_PASS_CLEANUP(dgpadt_do_first_allocation_y(sm,MAX(sm->init_maxy, y)));
    }
    else {
      /* we have allocated internal structures for this DgpAdt_SMImage
	 already: we just need to realloc them and initialize the
	 new parts to be zero or NULL */
      DGP_ERR_PASS_CLEANUP(dgpadt_do_allocation_y(sm,y));
    }
  }

  if (y > sm->maxy)
    sm->maxy = y;

  /* X COORDINATE */

  if (sm->x_arr[y] == (DgpAdt_sm_elt_type *)NULL) {
    /* we need to allocate memory for this array of DgpAdt_sm_elt_types. */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_first_allocation_x(sm,y,x));
  }
  else if ((temp = x - sm->x_min[y]) < 0) {
    /* the x-coordinate is less than the 
       minimum x-coordinate for this line. */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_reallocation_x_min(sm,y,x));
  }
  else if (temp > sm->x_len[y] - 1) {
    /* the x-coordinate is greater than the 
       maximum x-coordinate for this line */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_reallocation_x_len(sm,y,x));
  }

  *e = sm->x_arr[y][x - sm->x_min[y]];

  NORMAL_RETURN;
 
ERR_CLEANUP:

  ERR_RETURN;
}

ErrCode
DgpAdt_smSet(DgpAdt_SMImage sm, Dgp_UInt16 x, Dgp_UInt16 y, DgpAdt_sm_elt_type e) 
{
  DGP_PROLOGUE("smSet()", sm->DGP_PERRSTATE);
  Dgp_Int32 temp;

  /* Y COORDINATE */

  if (y > sm->maxy_alloced) {
    if (sm->maxy_alloced == -1) {
      /* we haven't allocated any internal structures 
	 for this DgpAdt_SMImage yet: go and get some memory. */
      DGP_ERR_PASS_CLEANUP(dgpadt_do_first_allocation_y(sm,MAX(sm->init_maxy, y)));
    }
    else {
      /* we have allocated internal structures for this DgpAdt_SMImage
	 already: we just need to realloc them and initialize the
	 new parts to be zero or NULL */
      DGP_ERR_PASS_CLEANUP(dgpadt_do_allocation_y(sm,y));
    }
  }

  if (y > sm->maxy)
    sm->maxy = y;

  /* X COORDINATE */

  if (sm->x_arr[y] == (DgpAdt_sm_elt_type *)NULL) {
    /* we need to allocate memory for this array of DgpAdt_sm_elt_types. */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_first_allocation_x(sm,y,x));
  }
  else if ((temp = x - sm->x_min[y]) < 0) {
    /* the x-coordinate is less than the 
       minimum x-coordinate for this line. */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_reallocation_x_min(sm,y,x));
  }
  else if (temp > sm->x_len[y] - 1) {
    /* the x-coordinate is greater than the 
       maximum x-coordinate for this line */
    DGP_ERR_PASS_CLEANUP(dgpadt_do_reallocation_x_len(sm,y,x));
  }

  sm->x_arr[y][x - sm->x_min[y]] = e;

  NORMAL_RETURN;
 
ERR_CLEANUP:
  ERR_RETURN;
}

Dgp_Bool
DgpAdt_smIsAllocated(DgpAdt_SMImage sm, Dgp_UInt16 x, Dgp_UInt16 y) {
  Dgp_Int32 temp;

  if (y > sm->maxy_alloced)
    return DGP_FALSE;

  if ((temp = x - sm->x_min[y]) < 0)
    /* the x-coordinate is less than the 
       minimum x-coordinate for this line. */
    return DGP_FALSE;
  
  if (temp > sm->x_len[y] - 1)
    /* the x-coordinate is greater than the 
       maximum x-coordinate for this line */
    return DGP_FALSE;

  return DGP_TRUE;
}

Dgp_UInt16 
DgpAdt_smGetNextY(DgpAdt_SMImage sm, Dgp_UInt16 y) {
  Dgp_Int32 nexty = y + 1;

  while (1) {
    if (nexty > sm->maxy)
      /* nexty has exceeded the max y coordinate that contains a value
	 inserted by the user */
      return (Dgp_UInt16) nexty;
    if (sm->x_arr[nexty] != (DgpAdt_sm_elt_type *)NULL)
      /* nexty gives a row which has some of the user's data in it */
      return (Dgp_UInt16) nexty;
    nexty++;
  }
}
