/* sm_image.h
   
   Aaron Stump, Summer '96

   This module implements a slightly weird data structure for use with 
   (although not dependent on) the lightweight version of the lossless
   tokenizer.

   DESCRIPTION:
       The data structure implemented in this module is meant to give
       a sparse (hence the s in sm) representation of a 2D array, where
       for any given line of the array, most of the points needed are
       expected to be within some distance of each other that is
       significantly less than the length of the line (the m in sm is
       for minimal distance), and all points are in the first quadrant.  
   
   INTENDED USE:
       This data structure is meant to use less memory, on average, than
       a straightforward implementation of a 2D array.
   
       Consequently, I've optimized to decrease memory use in my 
       implementation.  I've not made it use a minimal amount of memory,
       just a significantly lesser amount.  Furthermore, accessing an
       element of the array is not going to be anywhere near as efficient
       as, say, DGPIMG_REF() in libwjr.  On the other hand, the memory used is
       theta(number of rows * (diameter of bounding interval for each row)),
       while libwjr's implementation takes 
       theta((max x-coordinate) * (max y-coordinate)) units of memory.

       Don't use this data structure if you need to support arrays whose
       points could have a coordinate greater than can be stored in an
       Dgp_UInt16.

   IMPLEMENTATION:
       The basic idea is to have an array of pointers which is indexed by
       the y-value of the point to locate.  Each element of this array of
       pointers itself points to an array, whose elements are the actual values
       for the points of the image.  We'll only allocate memory for the
       array of pointers and the arrays the pointers point to when we have to.
       So our representation will be sparse, in the sense that if the maximum
       x-coordinate to represent is maxx and the maximum y-coordinate maxy,
       it will not necessarily be the case that we use O(maxx*maxy) memory for
       the image we're representing.  

*/

/* $Header: /project/icons/cvsroot/digipaper/src/libadt/sm_image.h,v 1.9 2000/03/04 00:22:16 mkrocken Exp $ */
   
#ifndef SM_IMAGE_H
#define SM_IMAGE_H

#include <sys/types.h>
#include "util.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef void *DgpAdt_sm_elt_type;

typedef struct {
  Dgp_UInt16 *x_min;
  Dgp_UInt16 *x_len;
  DgpAdt_sm_elt_type **x_arr;
  Dgp_Int32 maxy_alloced;   /* number of elements in the array x_arr */
  Dgp_Int32 maxy;           /* max. y-coordinate seen so far
			 (possibly != maxy_alloced) */
  Dgp_UInt32 init_maxy;
  Dgp_Bool always_clear;
  Dgp_State *dgp_state;
} *DgpAdt_SMImage;

DGP_LIBEXPORT void    
DgpAdt_smSetMaxy(Dgp_UInt32 init_maxy);

DGP_LIBEXPORT ErrCode 
DgpAdt_smNew(DgpAdt_SMImage *sm, Dgp_Bool always_clear, Dgp_UInt32 init_maxy, Dgp_State *dgp_state);

DGP_LIBEXPORT void    
DgpAdt_smFree(DgpAdt_SMImage sm);

DGP_LIBEXPORT ErrCode 
DgpAdt_smGet(DgpAdt_SMImage sm, Dgp_UInt16 x, Dgp_UInt16 y, 
			DgpAdt_sm_elt_type *e);

DGP_LIBEXPORT ErrCode 
DgpAdt_smSet(DgpAdt_SMImage sm, Dgp_UInt16 x, Dgp_UInt16 y, 
			DgpAdt_sm_elt_type e);
DGP_LIBEXPORT Dgp_Bool 
DgpAdt_smIsAllocated(DgpAdt_SMImage sm, Dgp_UInt16 x, 
				Dgp_UInt16 y);

DGP_LIBEXPORT Dgp_UInt16 
DgpAdt_smGetNextY(DgpAdt_SMImage sm, Dgp_UInt16 y);

/* some compilers doesn't like braces in these macros... */
#ifndef __GNUC__
#define DGPADT_SMGET_YMIN(sm)   DgpAdt_smGetNextY(sm,0)
#define DGPADT_SMGET_YMAX(sm)   ((sm)->maxy)
#define DGPADT_SMGET_XMIN(sm,y) (((y) > (sm)->maxy) ? -1 \
			 : (sm)->x_min[(y)])
#define DGPADT_SMGET_XMAX(sm,y) (((y) > (sm)->maxy) ? -1 \
			 : (sm)->x_min[(y)] + (sm)->x_len[(y)] - 1)
/* ...but gcc is cool with it */
#else
#define DGPADT_SMGET_YMIN(sm)   DgpAdt_smGetNextY(sm,0)
#define DGPADT_SMGET_YMAX(sm)   ({(sm)->maxy;})
#define DGPADT_SMGET_XMIN(sm,y) ({DgpAdt_SMImage _sm = sm; Dgp_UInt16 _y = y; \
                          (_y > _sm->maxy) \
                             ? -1 \
			     : _sm->x_min[_y];})
#define DGPADT_SMGET_XMAX(sm,y) ({DgpAdt_SMImage _sm = sm; Dgp_UInt16 _y = y; \
                          (y > _sm->maxy) \
                             ? -1 \
			     : _sm->x_min[_y] + _sm->x_len[_y] - 1;})
#endif

/* the following macro is guaranteed to set (x,y) to every point that has
   data which the user added to the DgpAdt_SMImage sm; but note that it will set 
   (x,y) to some other points as well */
#define DGPADT_SM_FOREACH(sm,x,y) \
for (y=DGPADT_SMGET_YMIN(sm), x=DGPADT_SMGET_XMIN(sm,y); \
     y <= DGPADT_SMGET_YMAX(sm); \
     (x>=DGPADT_SMGET_XMAX(sm,y)) ? (y=DgpAdt_smGetNextY(sm,y), x=DGPADT_SMGET_XMIN(sm,y)) : x++)

#ifdef __cplusplus
}
#endif

#endif
