/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 25.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: stack.c.
**
**  Handles stacks of generic items.  Exports:
**		type DgpAdt_Stack
**
**		constant DgpAdt_Stack DGP_NULLSTACK
**   
**		ErrCode DgpAdt_staNew(Dgp_State *dgp_state, Dgp_Int32 n, DgpAdt_Stack *_news);
**		void DgpAdt_staFree(Dgp_State *dgp_state, DgpAdt_Stack s);
**		void DgpAdt_staClear(DgpAdt_Stack s);
**		Dgp_Int32 DgpAdt_staRealloc(Dgp_State *dgp_state, DgpAdt_Stack s, Dgp_Int32 n);
**		void *DgpAdt_staPop(DgpAdt_Stack s);
**		Dgp_Int32 DgpAdt_staPush(DgpAdt_Stack s, void *userdata);
**		Dgp_Int32 DgpAdt_staLen(DgpAdt_Stack s);
**		Dgp_Int32 DgpAdt_staEmpty(DgpAdt_Stack s);
**		Dgp_Int32 DgpAdt_staFull(DgpAdt_Stack s);
**		Dgp_Int32 DgpAdt_staMaxNumNodes(DgpAdt_Stack s);
**
**	An object of type DgpAdt_Stack is used as a handle on a stack.  Each stack
**	is implemented as an array of generic pointers.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  impacts the prototypes for some functions, but
**		   I understand currently only libscancolor uses this adt.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**		o  Moved from libscancolor to libadt as part of overhaul effort.
**
**  
**************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "dgp_state.h"
#include "util.h"			/* For INLINE... / dhtesq. */
#include "stack.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/stack.c,v 1.9 2003/02/10 22:53:18 mkrocken Exp $");


/*************************************************************************
**
**  Function:  ErrCode DgpAdt_staNew(Dgp_State *dgp_state, Dgp_Int32 n, DgpAdt_Stack *_news)
**
**
**  Creates a new stack with room for n elements.  Note memory is allocated for
**	the data pointers only;  this function doesn't know anything about the data
**	type the caller wants to stack.
**
**	Arguments:
**		o  dgp_state points to valid Dgp_State{}.
**		o  n is the number of stack elements initially allocated.
**		o  _news will point to the new stack on successful return.
**
**	Return value:
**		o  An ErrCode if stack or data allocation fails.  On failure, *_news will be set to
**		   DGP_NULLSTACK.  If caller passes n <= 0, I will throw an error of type
**		   cADTNullStackRequested.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  now takes Dgp_State{}* as context.
**  
**/

ErrCode
DgpAdt_staNew(Dgp_State *dgp_state, Dgp_Int32 n, DgpAdt_Stack *_news) {
	DGP_PROLOGUE("DgpAdt_staNew()", DGP_PERRSTATE);

	DgpAdt_Stack news = DGP_NULLSTACK;

	*_news = DGP_NULLSTACK;

	DGP_ERR_GEN((n <= 0), cADTNullStackRequested);

	DGP_CHECK_ALLOC(news = (DgpAdt_Stack)DGP_MALLOC(sizeof(* news)));
	DGP_CHECK_ALLOC_CLEANUP(news->data = (_DgpAdt_Stacknode *)DGP_MALLOC(n * sizeof(_DgpAdt_Stacknode)));
		
	news->max_num_nodes = n;
	news->top = news->data - 1;
	*_news = news;
	NORMAL_RETURN;

ERR_CLEANUP:
	free(news);
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void DgpAdt_staFree(Dgp_State *dgp_state, DgpAdt_Stack s)
**
**
**  Frees the memory associated with the DgpAdt_Stack s, but
**  not the memory pointed to by the stack's generic 
**  pointers.
**
**	Arguments:
**		o  s is the stack to be freed.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void DgpAdt_staFree(Dgp_State *dgp_state, DgpAdt_Stack s)
{
  DGP_FREE((void *)(s->data));
  DGP_FREE((void *)s);
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void DgpAdt_staClear(DgpAdt_Stack s)
**
**
**  Clears the DgpAdt_Stack s, but does not free any memory.
**
**	Arguments:
**		o  s is the stack to be cleared.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void DgpAdt_staClear(DgpAdt_Stack s)
{
  s->top = s->data - 1;
  return;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  Dgp_Int32 DgpAdt_staMaxNumNodes(DgpAdt_Stack s)
**
**
**	Returns the maximum number of nodes that can be stored in the passed stack
**	with the memory currently allocated for it.
**
**	Arguments:
**		o  s is the stack to be queried.
**
**	Return value:
**		o  The Dgp_Int32 s->max_num_nodes.
**
**	Modification history:
**		o  None.
**  
**/

Dgp_Int32 DgpAdt_staMaxNumNodes(DgpAdt_Stack s) {
	return s->max_num_nodes; 
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode DgpAdt_staRealloc(Dgp_State *dgp_state, DgpAdt_Stack s, Dgp_Int32 n)
**
**
**  DgpAdt_staRealloc() reallocates the memory used to hold the elements of the DgpAdt_Stack
**  s so that it will now hold n elements.  If n is less than the number
**  of elements currently stored in s, or if there isn't enough memory
**  available, DgpAdt_staRealloc() leaves s unchanged and throws an error -- see return
**	value.
**
**  Note that if the user doesn't know in advance how many items she'll want
**  to put on a given DgpAdt_Stack s, she can check the return value of DgpAdt_staPush each
**  time she tries to push something onto s and call DgpAdt_staRealloc with s and 
**  some number greater than DgpAdt_staMaxNumNodes(s) whenever that return value is 
**  DGP_FALSE.
**
**	Arguments:
**		o  dgp_state points to valid Dgp_State{}.
**		o  s is the stack to be reallocated.
**		o  n is the new number of stack elements to be allocated for.
**
**	Return value:
**		o  An ErrCode on stack reallocation failure.  For n < num_nodes, throws error
**		   of type cADTSmallerStackRequested.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  now takes Dgp_State{}* as context.
**  
**/

ErrCode
DgpAdt_staRealloc(Dgp_State *dgp_state, DgpAdt_Stack s, Dgp_Int32 n) {
	DGP_PROLOGUE("DgpAdt_staRealloc()", DGP_PERRSTATE);

	Dgp_Int32	num_nodes;
    num_nodes = DgpAdt_staLen(s);

    DGP_ERR_GEN((n < num_nodes), cADTSmallerStackRequested);
	s->max_num_nodes = n;

	DGP_CHECK_ALLOC(s->data = (_DgpAdt_Stacknode *)DGP_REALLOC(s->data, s->max_num_nodes * sizeof(_DgpAdt_Stacknode)));
	s->top = &(s->data[num_nodes - 1]);
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int32 DgpAdt_staLen(DgpAdt_Stack s)
**
**
**	Returns the number of generic pointers in the passed stack.
**
**	Arguments:
**		o  s is the stack to be queried.
**
**	Return value:
**		o  Integer.
**
**	Modification history:
**		o  None.
**  
**/

Dgp_Int32 DgpAdt_staLen(DgpAdt_Stack s)
{
  return s->top - s->data + 1; 
} 


/*************************************************************************/
