#include "blit_mono.h"
#include "util.h"
#include "blit.h"
#include <string.h>
/* blit_mono.c
   wbell 10.19.99
   packed mono blitting into packed mono destination
 */

/* blitSetBinary()
   set a portion of a binary image to a particular value (0) or (1).
   will respect outside image data (like partial words) and will not modify
   outside the given rectangle.
   
   val - the value you want it to be set to, should be either 0 or 1.
   dst - pointer to the destination at dstMinX, dstMinY.
   dstBpl - bytes per line in the dest.
   dstX, dstY - where you want set, the upper left hand corner.
   dstW, dstH - how much you want cleared or set, must be positive
*/
void DgpBlat_blitSetBinary(Dgp_UInt8 val, 
		   Dgp_UInt32 * dst, Dgp_Int32 dstBpl,
		   Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		   Dgp_Int32 dstX, Dgp_Int32 dstY,
		   Dgp_Int32 dstW, Dgp_Int32 dstH)
{
  Dgp_UInt32 * dstRow;
  Dgp_UInt32 * dstCur;
  Dgp_UInt32 * dstEnd;
  Dgp_Int32 i = 0;
  Dgp_Int32 fullbytes = 0;
  Dgp_Int32 start = 0, end = 0;
  Dgp_UInt32 blitVal = 0;
  Dgp_UInt32 startMask = 0, endMask = 0;

  if (dstW <= 0 || dstH <= 0 || !dstBpl){
    return;
  }

  /* so here's the plan, for each line to clear, we've possibly got a 
     boundary case on both sides if we're not quite word aligned.
     so, we'll clear what we need of the first word by hand, and
     then just do a memset for all the full words we need to clear 
     on the line, and then we'll special case the final word as well.
     */
  if (val){
    blitVal = (Dgp_UInt32)-1;
  }

  /* how many full ulongs per line can we clear?  */
  /* what's the end of the last full ulong in raster coords */
  end = (dstX + dstW - dstMinX) - ((dstX + dstW - dstMinX) % DGP_BITS_UINT32);
  /* what's the beginning of the first fill ulong? */
  start = (dstX - dstMinX);
  /* well if this one's a partial, we'll advance it forward. */
  start = ((start % DGP_BITS_UINT32) > 0 ? start + DGP_BITS_UINT32 : start);

  /* well how many bytes is this? */
  fullbytes = (end - start) / 8;

  /* let's figure out where we start. */
  /* first get us on the correct row. */
  dstRow = (Dgp_UInt32 *)((Dgp_Int8 *)dst + (dstBpl * (dstY - dstMinY)));
  /* now get us on the correct word boundary. */
  dstRow = (Dgp_UInt32 *)((Dgp_Int8 *)dstRow + 
    ((start / DGP_BITS_UINT32) * sizeof(Dgp_UInt32)));
  /* end row is dstH rows below that. */
  dstEnd = (Dgp_UInt32 *)((Dgp_Int8 *)dstRow + (dstBpl * dstH));

  /* now we'll create a startMask and an end mask which we can use */
  /* to save the untouchable portions of the partial start and end */
  /* ulongs. */
  /* make the start mask, which is the first start % DGP_BITS_UINT32 all on,  */
  /* the rest 0. */
  start = DGP_BITS_UINT32 - (start % DGP_BITS_UINT32);
  if (start != DGP_BITS_UINT32){
    for (i = 0; i < start; i++){
      startMask = (startMask << 1) | 1;
    }
  }

  /* make the end mask, which is the last end % DGP_BITS_UINT32 all on, the rest 0. */
  end = end % DGP_BITS_UINT32;
  for (i = 0; i < end; i++){
    endMask = (endMask >> 1) | (1^(DGP_BITS_UINT32 - 1));
  } 
  
  for ( ; dstRow < dstEnd ; dstRow = (Dgp_UInt32 *)
	  ((Dgp_Int8 *)dstRow + dstBpl)){
    /* is the first ulong not full? */
    dstCur = dstRow;
    if(startMask){
      *dstCur = (*dstCur & ~startMask) | (blitVal & startMask);
      dstCur ++;
    }

    memset(dstCur, blitVal, fullbytes);
    dstCur = (Dgp_UInt32 *)((Dgp_Int8 *)dstCur + fullbytes);

    if(endMask){
      *dstCur = (*dstCur & ~endMask) | (blitVal & endMask);
    }
  }
}


#define IncPtr(xx, yy) ( (Dgp_UInt32 *) (((Dgp_UInt8 *) xx) + yy) )

/* blitBinaryOrExact()
   blits a binary image into a binary destination image. Source image must
   have no extra bytes on the end of the row (rows must only be padded to 
   32-bits) and padding must be set to 0. 
   Code stolen from Russ Atkinson's OrBlt
   dst - pointer to first word of destination.
   off - what bit of dst to put down first bit of src (31 is leftmost bit)
   dstBpl - bytes per line in the destination
   src - pointer to first word of source
   srcH - how many lines is the source token?

21.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...

*/
#ifdef _WIN32
DGP_LIBEXPORT
#endif
INLINE void 
DgpBlat_blitBinaryOrExact(Dgp_UInt32 * dst,
		       Dgp_UInt32 off,
		       Dgp_Int32 dstBpl,
		       Dgp_UInt32 *src,
		       Dgp_UInt32 srcW,
		       Dgp_Int32 srcH)
{
  Dgp_UInt32 sw;
  Dgp_UInt32 dw;
  Dgp_UInt32 shiftL;
  Dgp_UInt32 tw = (off + srcW);
  if (srcW <= 32) {
    /* Each src line fits in a word */ 
    if (tw <= 32) {
      /* Each dst line fits in a word */
      if (off == 0) {
	/* The src and dst are aligned */
	while (srcH >= 4) {
	  sw = src[0];
	  dw = dst[0];
	  dw = dw | sw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[1];
	  dw = dst[0];
	  dw = dw | sw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[2];
	  dw = dst[0];
	  dw = dw | sw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[3];
	  dw = dst[0];
	  dw = dw | sw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  src = (src + 4);
	  srcH = srcH - 4;
	};
	if (srcH <= 0) return;
	sw = src[0];
	dw = dst[0];
	dw = dw | sw;
	dst[0] = dw;
	if (srcH == 1) return;
	sw = src[1];
	dst = IncPtr(dst, dstBpl);
	dw = dst[0];
	dw = dw | sw;
	dst[0] = dw;
	if (srcH == 2) return;
	sw = src[2];
	dst = IncPtr(dst, dstBpl);
	dw = dst[0];
	dw = dw | sw;
	dst[0] = dw;
      }
      else {
	/* a shift is needed, but the dst srcH do not span words */
	while (srcH >= 4) {
	  sw = src[0];
	  dw = dst[0];
	  tw = sw >> off;
	  dw = dw | tw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[1];
	  dw = dst[0];
	  tw = sw >> off;
	  dw = dw | tw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[2];
	  dw = dst[0];
	  tw = sw >> off;
	  dw = dw | tw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  sw = src[3];
	  dw = dst[0];
	  tw = sw >> off;
	  dw = dw | tw;
	  dst[0] = dw;
	  dst = IncPtr(dst, dstBpl);
	  src = src + 4;
	  srcH = srcH - 4;
	};
	if (srcH <= 0) return;
	sw = src[0];
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	if (srcH == 1) return;
	sw = src[1];
	dst = IncPtr(dst, dstBpl);
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	if (srcH == 2) return;
	sw = src[2];
	dst = IncPtr(dst, dstBpl);
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
      };
    }
    else {
      /* The dst characters always span a word */
      shiftL = (32 - off);
      while (srcH >= 4) {
	sw = src[0];
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	dw = dst[1];
	tw = sw << shiftL;
	dw = dw | tw;
	dst[1] = dw;
	dst = IncPtr(dst, dstBpl);
	sw = src[1];
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	dw = dst[1];
	tw = sw << shiftL;
	dw = dw | tw;
	dst[1] = dw;
	dst = IncPtr(dst, dstBpl);
	sw = src[2];
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	dw = dst[1];
	tw = sw << shiftL;
	dw = dw | tw;
	dst[1] = dw;
	dst = IncPtr(dst, dstBpl);
	sw = src[3];
	dw = dst[0];
	tw = sw >> off;
	dw = dw | tw;
	dst[0] = dw;
	dw = dst[1];
	tw = sw << shiftL;
	dw = dw | tw;
	dst[1] = dw;
	dst = IncPtr(dst, dstBpl);
	src = (src + 4);
	srcH = srcH - 4;
      };
      if (srcH <= 0) return;
      sw = src[0];
      dw = dst[0];
      tw = sw >> off;
      dw = dw | tw;
      dst[0] = dw;
      dw = dst[1];
      tw = sw << shiftL;
      dw = dw | tw;
      dst[1] = dw;
      if (srcH == 1) return;
      sw = src[1];
      dst = IncPtr(dst, dstBpl);
      dw = dst[0];
      tw = sw >> off;
      dw = dw | tw;
      dst[0] = dw;
      dw = dst[1];
      tw = sw << shiftL;
      dw = dw | tw;
      dst[1] = dw;
      if (srcH == 2) return;
      sw = src[2];
      dst = IncPtr(dst, dstBpl);
      dw = dst[0];
      tw = sw >> off;
      dw = dw | tw;
      dst[0] = dw;
      dw = dst[1];
      tw = sw << shiftL;
      dw = dw | tw;
      dst[1] = dw;
    };
  }
  else {
    if (srcW <= 64) {
      /* Each source line fits in two words */
      if (off == 0) {
	while (srcH > 0) {
	  dw = dst[0];
	  sw = src[0];
	  dw = dw | sw;
	  dst[0] = dw;
	  dw = dst[1];
	  sw = src[1];
	  dw = dw | sw;
	  dst[1] = dw;
	  src = src + 2;
	  dst = IncPtr(dst, dstBpl);
	  srcH = srcH - 1;
	};
      }
      else {
	shiftL = (32 - off);
	if (tw <= 64) {
	  while (srcH > 0) {
	    dw = dst[0];
	    sw = src[0];
	    tw = sw >> off;
	    dw = dw | tw;
	    dst[0] = dw;
	    dw = dst[1];
	    tw = sw << shiftL;
	    dw = dw | tw;
	    sw = src[1];
	    tw = sw >> off;
	    dw = dw | tw;
	    dst[1] = dw;
	    src = src + 2;
	    dst = IncPtr(dst, dstBpl);
	    srcH = srcH - 1;
	  };
	}
	else {
	  while (srcH > 0) {
	    dw = dst[0];
	    sw = src[0];
	    tw = sw >> off;
	    dw = dw | tw;
	    dst[0] = dw;
	    dw = dst[1];
	    tw = sw << shiftL;
	    dw = dw | tw;
	    sw = src[1];
	    tw = sw >> off;
	    dw = dw | tw;
	    dst[1] = dw;
	    dw = dst[2];
	    tw = sw << shiftL;
	    dw = dw | tw;
	    dst[2] = dw;
	    src = src + 2;
	    dst = IncPtr(dst, dstBpl);
	    srcH = srcH - 1;
	  };
	};
      };
    }
    else {
      /* Each source line is more than two words */
      {
	if (off == 0) {
	  srcW = srcW - 32;
	  while (srcH > 0) {
	    Dgp_UInt32 * nextDst = IncPtr(dst, dstBpl);
	    Dgp_UInt32 rem = srcW;
	    dw = dst[0];
	    sw = src[0];
	    dw = dw | sw;
	    dst[0] = dw;
	    for (;;) {
	      dst = dst + 1;
	      src = src + 1;
	      dw = dst[0];
	      sw = src[0];
	      dw = dw | sw;
	      dst[0] = dw;
	      if (rem <= 32) break;
	      rem = rem - 32;
	    };
	    dst = nextDst;
	    src = src + 1;
	    srcH = srcH - 1;
	  };
	}
	else {
	  shiftL = (32 - off);
	  while (srcH > 0) {
	    Dgp_UInt32 * nextDst = IncPtr(dst, dstBpl);
	    Dgp_Int32 rem = (Dgp_Int32)srcW;
	    dw = dst[0];
	    for (;;) {
	      sw = src[0];
	      tw = sw >> off;
	      dw = dw | tw;
	      dst[0] = dw;
	      dst = dst + 1;
	      dw = dst[0];
	      tw = sw << shiftL;
	      dw = dw | tw;
	      src = src + 1;

              /* added by wbell
                 this for loop does batches of 2 words, but if there's 
                 one more word's worth (of dst) this'll go off the end of the
                 buffer. here we special case that case.
              */
              rem = rem - 32;
              if ((rem - (Dgp_Int32)shiftL) <= 0)
                break;
	    };
            /* here's the one lone word case on the end. */
            if (rem > 0){
              sw = src[0];
              tw = sw >> off;
              dw = dw | tw;
              src = src + 1;
            }
	    dst[0] = dw;
	    dst = nextDst;
	    srcH = srcH - 1;
	  };
	};
      };
    };
  };
}






