#include "scaleBlit.h"

#include <assert.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif  /* (!VXWORKS) */

/* ...for MIN/MAX */
#include "util.h"

/* blitBinaryRGB uses the 256x6*sizeof(Dgp_UInt32) table */
/* for working 6 words at a time. */
#include "blitTables.h"

static INLINE void blitBinaryRGBRow(DgpImage_RGB * dstCur, Dgp_UInt32 *srcCur,
			     Dgp_Int32 curSrcBitsLeft, Dgp_Int32 dstW);

static Dgp_Int32 calcRatLowDown (Dgp_Int32 dst, DgpCoord_Rat res) { return (Dgp_Int32) floor(dst * ((Dgp_Float64)res.den / (Dgp_Float64)res.num));}
static Dgp_Int32 calcRatLowUp   (Dgp_Int32 src, DgpCoord_Rat res) { return (Dgp_Int32) floor(src * ((Dgp_Float64)res.den / (Dgp_Float64)res.num));}
static Dgp_Int32 calcRatHighDown(Dgp_Int32 dst, DgpCoord_Rat res) { return (Dgp_Int32)  ceil(dst * ((Dgp_Float64)res.den / (Dgp_Float64)res.num));}
static Dgp_Int32 calcRatHighUp  (Dgp_Int32 src, DgpCoord_Rat res) { return (Dgp_Int32)  ceil(src * ((Dgp_Float64)res.den / (Dgp_Float64)res.num));}

void DgpBlat_scaleBlitCalcNeeded(Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			    DgpCoord_Resolution * res,
			    Dgp_Int32 srcImgW, Dgp_Int32 srcImgH,
			    Dgp_Int32 * srcX, Dgp_Int32 * srcY,
			    Dgp_Int32 * srcW, Dgp_Int32 * srcH)
{
  /* so what we want to calculate here is if we were asked to generate
   *	this rectangle of the destination image, which is the source image
   *	at some scale factor, which pixels of the source would we need in
   *	order to generate this destination rectangle?
   */
  Dgp_Int32 sxl,syl,sxh,syh;
  Dgp_Int32 sw,sh;

  if (DgpCoord_float(res->xRes) < 1.0)
  {
    sxl = calcRatLowUp (dstX,        res->xRes);
    sxh = calcRatHighUp(dstX + dstW, res->xRes) - 1;
  }
  else
  {
    sxl = calcRatLowDown (dstX,        res->xRes);
    sxh = calcRatHighDown(dstX + dstW, res->xRes) - 1;
  }

  if (DgpCoord_float(res->yRes) < 1.0)
  {
    syl = calcRatLowDown (dstY,        res->yRes);
    syh = calcRatHighDown(dstY + dstH, res->yRes) - 1;
  }
  else
  {
    syl = calcRatLowUp (dstY,        res->yRes);
    syh = calcRatHighUp(dstY + dstH, res->yRes) - 1;
  }

  sw = (sxh-sxl) + 1;
  sh = (syh-syl) + 1;

  *srcX = sxl;
  *srcY = syl;
  
  *srcW = (sxl+sw > srcImgW) ? (srcImgW - sxl) : sw;
  *srcH = (syl+sh > srcImgH) ? (srcImgH - syl) : sh;
}


void DgpBlat_scaleBlitCalcAffected(Dgp_Int32 srcX, Dgp_Int32 srcY, Dgp_Int32 srcW, Dgp_Int32 srcH,
			   DgpCoord_Resolution * res,
			   Dgp_Int32 dstImgW, Dgp_Int32 dstImgH,
			   Dgp_Int32 * dstX, Dgp_Int32 * dstY,
			   Dgp_Int32 * dstW, Dgp_Int32 * dstH)
{
  /* We want to calculate the rectangle of the destination that
     if we change any of the pixels in (srcX,srcY)
     (srcX+srcW-1,srcY+srcH-1) [inclusive] will be
     changed.
  */

  Dgp_Int32 dxl,dxh,dyl,dyh;
  
  if (DgpCoord_float(res->xRes) < 1.0)
  {
    dxl = calcRatLowDown(srcX, res->xRes);
    dxh = calcRatHighDown(srcX + (srcW - 1), res->xRes);
  }
  else
  {
    dxl = calcRatLowUp(srcX, res->xRes);
    dxh = calcRatHighUp(srcX + srcW, res->xRes);
  }
  
  if (DgpCoord_float(res->yRes) < 1.0)
  {
    dyl = calcRatLowDown(srcY, res->yRes);
    dyh = calcRatHighDown(srcY + (srcH - 1), res->yRes);
  }
  else
  {
    dyl = calcRatLowUp(srcY, res->yRes);
    dyh = calcRatHighUp(srcY + srcH, res->yRes);
  }
  
  if (dxh >= dstImgW)
  {
    dxh = dstImgW - 1;
  }
  
  if (dyh >= dstImgH)
  {
    dyh = dstImgH - 1;
  }

  *dstX = dxl;
  *dstY = dyl;
  *dstW = (dxh - dxl) + 1;
  *dstH = (dyh - dyl) + 1;

}


void DgpBlat_scaleBlitRGBBGRGen(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			DgpCoord_Resolution * res)
{
  /* Given dst, which corresponds to the poDgp_Int32 (dstMinX,dstMinY),
     fill in the rectangle (dstX,dstY) (dstX+dstW-1,dstY+dstH-1)
     [inclusive] with pixels from the source image at the specified
     scale. There must be enough source to perform this opeDGPCOORD_Ration 
     (can figure out how much source is needed from scaleBlitCalcNeeded)
     This is the general purpose scaling, no special cases.
  */
  DgpImage_RGB pix;
  Dgp_Int8 swap;
  Dgp_Int32 x,y;
  
  DgpImage_RGB *srcRow = src;
  /* srcOff is just a junk var used to compute offsets for the src. */
  Dgp_Int32 srcOff;
  
  /* the previous row of the source we just used. */
  DgpImage_RGB *srcPrev = NULL;
  
  Dgp_Float64 invScaleX = (Dgp_Float64)res->xRes.den / res->xRes.num;
  Dgp_Float64 invScaleY = (Dgp_Float64)res->yRes.den / res->yRes.num;
  
  DgpBlat_BGR *dstRow,*dstCur;
  
  /* dstRow is the first pixel in the first row to be blitted. */
  /* get the top row. */
  Dgp_Int32 dstOff = ((dstY-dstMinY) * dstBpl);
  /* advance the x position. */
  dstOff += ((dstX-dstMinX)*sizeof(DgpBlat_BGR));
  
  dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dst + dstOff);
  
  assert(dstW > 0 && dstH > 0);
  assert(srcBpl != 0 && src);
  assert(dstBpl != 0 && dst);
  assert(invScaleX > 0 && invScaleY > 0);
  assert((dstY - dstMinY) >= 0);
  assert(sizeof(DgpImage_RGB) == sizeof(DgpBlat_BGR));

  for (y = 0; y < dstH; y++){
    dstCur = dstRow;
    
    srcOff = ((Dgp_Int32)((dstY + y) * invScaleY)) - srcMinY;
    srcOff *= srcBpl;
    srcRow = (DgpImage_RGB *)((Dgp_Int8 *)src + srcOff);
    
    /* did we just compute this row? then we can just memcpy it from */
    /* the destination line above */
    if (srcPrev == srcRow){
      memcpy(dstCur,((Dgp_Int8 *)dstCur - dstBpl),(dstW * sizeof(DgpBlat_BGR)));
      /* advance the destination row. */
      dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dstRow + dstBpl);
      continue;
    }
    
    /* else we need to generate a new destination row. */
    for (x = 0; x<dstW; x++){
      srcOff = ((Dgp_Int32)((dstX + x) * invScaleX)) - srcMinX;
      srcOff *= sizeof(DgpImage_RGB);
      
      pix = *((DgpImage_RGB *)((Dgp_Int8 *)srcRow + srcOff));
      
      swap = pix.r;
      pix.r = pix.b;
      pix.b = swap;
      
      *((DgpImage_RGB *)dstCur) = pix;
      dstCur++;		
    }
    
    srcPrev = srcRow;
    dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dstRow + dstBpl);
  }
}

void DgpBlat_scaleBlitRGBBGR(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		     DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		     DgpCoord_Resolution * res)
{
  if(res->xRes.num == res->xRes.den && res->yRes.num == res->yRes.den){
    DgpBlat_blitRGBBGR(src, srcBpl, srcMinX, srcMinY,
	       dst, dstBpl, dstMinX, dstMinY,
	       dstX, dstY, dstW, dstH);
  }
  else{
    DgpBlat_scaleBlitRGBBGRGen(src,srcBpl,srcMinX,srcMinY,
		       dst,dstBpl,dstMinX,dstMinY,
		       dstX,dstY,dstW,dstH,
		       res);
  }	
}


/* this is an exact copy of the RGBBGR one, so any bug fixes */
/*  should be applied to this as well. */
void DgpBlat_scaleBlitRGBGen(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		     DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		     DgpCoord_Resolution * res)
{
  /* Given dst, which corresponds to the poDgp_Int32 (dstMinX,dstMinY), 
     fill in the rectangle (dstX,dstY) (dstX+dstW-1,dstY+dstH-1)
     [inclusive] with pixels from the source image at the specified
     scale. There must be enough source to perform this opeDGPCOORD_Ration 
     (can figure out how much source is needed from scaleBlitCalcNeeded)
     This is the general purpose scaling, no special cases.
  */
  DgpImage_RGB pix;

  Dgp_Int32 x,y;

  DgpImage_RGB *srcRow = src;
  /* srcOff is just a junk var used to compute offsets for the src. */
  Dgp_Int32 srcOff;

  /* the previous row of the source we just used. */
  DgpImage_RGB *srcPrev = NULL;

  Dgp_Float64 invScaleX = (Dgp_Float64)res->xRes.den / res->xRes.num;
  Dgp_Float64 invScaleY = (Dgp_Float64)res->yRes.den / res->yRes.num;
  
  DgpBlat_BGR *dstRow,*dstCur;
  
  /* dstRow is the first pixel in the first row to be blitted. */
  /* get the top row. */
  Dgp_Int32 dstOff = ((dstY-dstMinY) * dstBpl);
  /* advance the x position. */
  dstOff += ((dstX-dstMinX)*sizeof(DgpBlat_BGR));
		
  dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dst + dstOff);
	
  assert(dstW > 0 && dstH > 0);
  assert(srcBpl != 0 && src);
  assert(dstBpl != 0 && dst);
  assert(invScaleX > 0 && invScaleY > 0);
  assert((dstY - dstMinY) >= 0);
  assert(sizeof(DgpImage_RGB) == sizeof(DgpBlat_BGR));
  for (y = 0; y < dstH; y++){
    dstCur = dstRow;
    
    srcOff = ((Dgp_Int32)((dstY + y) * invScaleY)) - srcMinY;
    srcOff *= srcBpl;
    srcRow = (DgpImage_RGB *)((Dgp_Int8 *)src + srcOff);
    
    /* did we just compute this row? then we can just memcpy it from */
    /* the destination line above */
    if (srcPrev == srcRow){
      memcpy(dstCur,((Dgp_Int8 *)dstCur - dstBpl),(dstW * sizeof(DgpBlat_BGR)));
      /* advance the destination row. */
      dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dstRow + dstBpl);
      continue;
    }
    
    /* else we need to generate a new destination row. */
    for (x = 0; x < dstW; x++){
      srcOff = ((Dgp_Int32)((dstX + x) * invScaleX)) - srcMinX;
      srcOff *= sizeof(DgpImage_RGB);
      
      pix = *((DgpImage_RGB *)((Dgp_Int8 *)srcRow + srcOff));
      
      *((DgpImage_RGB *)dstCur) = pix;
      dstCur++;		
    }
    
    srcPrev = srcRow;
    dstRow = (DgpBlat_BGR *)((Dgp_Int8 *)dstRow + dstBpl);
  }
}


void DgpBlat_scaleBlitRGB(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		  DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		  Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		  DgpCoord_Resolution * res)
{

  if(res->xRes.num == res->xRes.den && res->yRes.num == res->yRes.den){
    DgpBlat_blitRGB(src, srcBpl, srcMinX, srcMinY,
	    dst, dstBpl, dstMinX, dstMinY,
	    dstX, dstY, dstW, dstH);
  }
  else{
    DgpBlat_scaleBlitRGBGen(src,srcBpl,srcMinX,srcMinY,
		    dst,dstBpl,dstMinX,dstMinY,
		    dstX,dstY,dstW,dstH,
		    res);
  }
}

void DgpBlat_copyRGBOverlapNoClip(
    DgpImage_RGB * baseMem,
    Dgp_Int32 baseBpl,
    Dgp_Int32 w,
    Dgp_Int32 h,
    Dgp_Int32 srcX,
    Dgp_Int32 srcY,
    Dgp_Int32 dstX,
    Dgp_Int32 dstY )
{
    /* The idea here is that you have one block of memory you've allocated
     *  and you have a source rectangle in it (srcX,srcY) (srcX+w-1,srcY+h-1)
     *	that you want to copy to (dstX,dstY) (dstX+w-1,dstY+h-1), but the 
     *  problem is that there could be overlap, and you want to copy so that
     *  you get the src into the dest. We don't know the w/h of the buffer, so
     * we assume that both rectangles fit in the buffer.
     */
    
    Dgp_Int32 y, count;

    void * dest, * src;

    assert( (w > 0) && (h > 0) );

    if( (srcX == dstX)  && (srcY == dstY) )
    {
        return;
    }
    /* if we are farther down with the source, there's no problem with */
    /* rows overlapping. */
    else if (srcY > dstY)
    {
        count = w * sizeof(DgpImage_RGB);
        dest  = (Dgp_Int8 *)baseMem + ((dstY * baseBpl) + (dstX * sizeof(DgpImage_RGB)));
        src   = (Dgp_Int8 *)baseMem + ((srcY * baseBpl) + (srcX * sizeof(DgpImage_RGB)));
    
        for( y = 0; y < h ; y++ )
        {
            memcpy(dest, src, count);
            dest = (Dgp_Int8 *)dest + baseBpl;
            src  = (Dgp_Int8 *)src  + baseBpl;
        }
    }
    /* if the src and dst start on the same row, we can use memmove to make sure */
    /* we get the right behavior for overlapping. */
    else if (srcY == dstY)
    {
        count = w * sizeof(DgpImage_RGB);
        dest  = (Dgp_Int8 *)baseMem + ((dstY * baseBpl) + (dstX * sizeof(DgpImage_RGB)));
        src   = (Dgp_Int8 *)baseMem + ((srcY * baseBpl) + (srcX * sizeof(DgpImage_RGB)));
        
        for (y = 0; y < h ; y++)
        {
            memmove(dest, src, count);
            dest = (Dgp_Int8 *)dest + baseBpl;
            src  = (Dgp_Int8 *)src  + baseBpl;
        }
    }
    else
    {
        count = w * sizeof(DgpImage_RGB);
        dest  = (Dgp_Int8 *)baseMem + (((dstY + (h - 1)) * baseBpl) + (dstX * sizeof(DgpImage_RGB)));
        src   = (Dgp_Int8 *)baseMem + (((srcY + (h - 1)) * baseBpl) + (srcX * sizeof(DgpImage_RGB)));
        
        for (y = 0; y < h ; y++)
        {
            memmove(dest, src, count);
            dest = (Dgp_Int8 *)dest - baseBpl;
            src  = (Dgp_Int8 *)src - baseBpl;
        }
    }
}


void DgpBlat_blitRGBBGR(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH)
{
  Dgp_Int32 y,x;
  DgpImage_RGB * srcCur;
  DgpBlat_BGR *dstCur;
  DgpImage_RGB swap;
  Dgp_Int8 temp;
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 off;

  off = (dstY - srcMinY) * srcBpl;
  off += ((dstX - srcMinX) * sizeof(DgpImage_RGB));
  srcRow = (DgpImage_RGB *) ((Dgp_Int8 *)src + off);

  off = (dstY - dstMinY) * dstBpl;
  off += ((dstX - dstMinX) * sizeof(DgpBlat_BGR));
  dstRow = (DgpBlat_BGR *) ((Dgp_Int8 *)dst + off);
	
  for (y = 0 ; y < dstH ; y++){
      srcCur=srcRow;
      dstCur=dstRow;
		
      for (x=0 ; x < dstW ; x++){
	  swap=*(srcCur++);
			
	  temp = swap.r;
	  swap.r = swap.b;
	  swap.b = temp;

	  *((DgpImage_RGB *)(dstCur++))=swap;
	}
      srcRow = (DgpImage_RGB *) ((Dgp_Int8 *)srcRow + srcBpl);
      dstRow = (DgpBlat_BGR *) ((Dgp_Int8 *)dstRow + dstBpl);
    }
}

void DgpBlat_blitRGB(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
	     DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
	     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH)
{
  Dgp_Int32 y, count;
  DgpImage_RGB * srcRow,* dstRow;
  Dgp_Int32 off;

  off = (dstY - srcMinY) * srcBpl;
  off += ((dstX- srcMinX) * sizeof(DgpImage_RGB));
  srcRow = (DgpImage_RGB *) ((Dgp_Int8 *)src + off);

  off = (dstY - dstMinY) * dstBpl;
  off += ((dstX- dstMinX) * sizeof(DgpImage_RGB));
  dstRow = (DgpImage_RGB *) ((Dgp_Int8 *)dst + off);
	
  count = dstW * sizeof(DgpImage_RGB);

  for (y = 0 ; y < dstH ; y++){
    memcpy(dstRow,srcRow,count);
    
    srcRow = (DgpImage_RGB *) ((Dgp_Int8 *)srcRow + srcBpl);
    dstRow = (DgpImage_RGB *) ((Dgp_Int8 *)dstRow + dstBpl);
  }
}



void DgpBlat_blitBinaryRGB(Dgp_UInt32 * src,
		   Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		   DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		   Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH )
{
  /* The general idea here is that we want to use an 8-bit lookup 
   * table to convert 8-bits of packed image data to 24 bytes of
   * RGB data (which is 8 rgbs, which is 6 words), so we need to
   * for each row, find the first dest (RGB) pixel that is word
   * aligned, and up to that point, the conversion one bit at a 
   * time, and then afterwards do the conversion 8-bits at a time.
   */
	
  Dgp_UInt32 *srcRow, *srcCur;
  DgpImage_RGB * dstRow, *dstCur;
	
  Dgp_Int32 dstOff, srcOff, srcLpl, curSrcBitsLeft;
  Dgp_Int32 y;

  srcLpl = srcBpl / sizeof(Dgp_UInt32);
 
  dstOff = ((dstY - dstMinY) * dstBpl) + ((dstX - dstMinX) * sizeof(DgpImage_RGB));
  dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dst + dstOff); 
	
  /* this gets us the beginning of the row of the source, but */
  /* now we need to find the first word we're concerned with, and */
  /* which pixel (bit) in that word is the first to be converted. */
  srcOff = (srcLpl * (dstY - srcMinY));
  srcOff += ((dstX - srcMinX) / 32);
  srcRow = src + srcOff;
	
  for (y = 0; y < dstH ; y ++){
    dstCur = dstRow;
    srcCur = srcRow;
    curSrcBitsLeft = 32 - (dstX - srcMinX) % 32;
    
    blitBinaryRGBRow(dstCur, srcCur, curSrcBitsLeft, dstW);

    dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + dstBpl);
    srcRow = srcRow + srcLpl;
  }
}


static INLINE void blitBinaryRGBRow(DgpImage_RGB * dstCur, Dgp_UInt32 *srcCur,
			     Dgp_Int32 curSrcBitsLeft, Dgp_Int32 dstW)
{
  Dgp_Int32 x = 0,temp;
  Dgp_Int32 wb;
  Dgp_UInt32 curSrcWord;
  Dgp_Int32 srcBitsUsed;
  Dgp_UInt32 * dstLong;
  const Dgp_UInt32 *dstLookup;

	
  /* do pixel by pixel until dstCur is on a word boundary. */
  wb = (Dgp_Int32)dstCur % sizeof(Dgp_UInt32);
  /* make sure this is less than the blatting width. */
  wb = MIN(wb,dstW);

  /* we can need to do either 0,1,2,3 pixels. We just special */
  /* case them all. */
  /* clear out all the bits we're not interested in. */
  curSrcWord = *(srcCur++) << (32 - curSrcBitsLeft);
  srcBitsUsed = 0;
			
  /* make sure we actually have enough bits in curSrcWord to */
  /* to do these few writes. */
  if (curSrcBitsLeft < wb)
    {
      /* this assumes srcBitsUsed = 0; */
      curSrcWord = curSrcWord | (*srcCur >> curSrcBitsLeft);
      srcBitsUsed = 32 - curSrcBitsLeft;
      curSrcBitsLeft = 32;
    }

  if (wb == 1)
    {
      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcBitsLeft--;
      curSrcWord = curSrcWord << 1;
      x++;
    }
  else if (wb == 2)
    {
      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;

      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;
		
      curSrcBitsLeft -= 2;
      x+=2;
    }
  else if (wb == 3)
    {
      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;

      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;
		
      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;
   
      curSrcBitsLeft -= 3;
      x+=3;
    }
	
  /* now we should be word aligned in the destination, or we should be done. */
  assert(!((Dgp_Int32)dstCur % sizeof(Dgp_UInt32)) || x==dstW);

  dstLong = (Dgp_UInt32 *)dstCur;
  /* now we're going to do this 32 pixels at a time in blocks of 4. */
  while ((x + 32) <= dstW)
    {
      /* fill the word up with 32 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if (curSrcBitsLeft != 32)
	{
	  if ((curSrcBitsLeft + (32 - srcBitsUsed)) < 32)
	    {
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	      if (srcBitsUsed < 32)
		{
		  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
		  curSrcBitsLeft += (32 - srcBitsUsed);
		}

	      srcCur ++;
	      srcBitsUsed = 0;
	    }
			
				/* no spanning a word. */
	  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	  srcBitsUsed += (32 - curSrcBitsLeft);
	  curSrcBitsLeft = 32;
				
	  assert(srcBitsUsed<=32 && srcBitsUsed>=0);
	  assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0);
	}
		
      /* we've got enough, now let's do the 4 8-bit blats. */
      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = dstLookup[0];
      dstLong[1] = dstLookup[1];
      dstLong[2] = dstLookup[2];
      dstLong[3] = dstLookup[3];
      dstLong[4] = dstLookup[4];
      dstLong[5] = dstLookup[5];
      dstLong += 6;

      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 16)];
      dstLong[0] = dstLookup[0];
      dstLong[1] = dstLookup[1];
      dstLong[2] = dstLookup[2];
      dstLong[3] = dstLookup[3];
      dstLong[4] = dstLookup[4];
      dstLong[5] = dstLookup[5];
      dstLong += 6;
		
      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 8)];
      dstLong[0] = dstLookup[0];
      dstLong[1] = dstLookup[1];
      dstLong[2] = dstLookup[2];
      dstLong[3] = dstLookup[3];
      dstLong[4] = dstLookup[4];
      dstLong[5] = dstLookup[5];
      dstLong += 6;

      dstLookup = monoToRGB8bit[(Dgp_UInt8)curSrcWord];
      dstLong[0] = dstLookup[0];
      dstLong[1] = dstLookup[1];
      dstLong[2] = dstLookup[2];
      dstLong[3] = dstLookup[3];
      dstLong[4] = dstLookup[4];
      dstLong[5] = dstLookup[5];
      dstLong += 6;
			
      curSrcBitsLeft = 0;
      curSrcWord = 0;
		
      x += 32;
    }


  while ((x + 8) <= dstW)
    {
      /* fill the word up with 8 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if (curSrcBitsLeft < 8)
	{
	  if ((8 - curSrcBitsLeft) > (32 - srcBitsUsed))
	    {
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	      if (srcBitsUsed < 32)
		{
		  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
		  curSrcBitsLeft += (32 - srcBitsUsed);
		}

	      srcCur ++;
	      srcBitsUsed = 0;
	    }
	  /* no spanning a word, just take as much as we can. */
	  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	  temp = MIN(32-srcBitsUsed,32-curSrcBitsLeft);
	  srcBitsUsed += temp;
	  curSrcBitsLeft += temp;
			
	  assert(srcBitsUsed<=32 && srcBitsUsed>=0);
	  assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0);
	}
		
      /* we've got enough, now let's do the 8-bit blat. */
      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = dstLookup[0];
      dstLong[1] = dstLookup[1];
      dstLong[2] = dstLookup[2];
      dstLong[3] = dstLookup[3];
      dstLong[4] = dstLookup[4];
      dstLong[5] = dstLookup[5];
      dstLong += 6;

      curSrcWord = curSrcWord << 8;
      curSrcBitsLeft -= 8;
      x += 8;
    }

  dstCur = (DgpImage_RGB *)dstLong;
  /* so now we could have at most 7 more pixels to do. */
  wb = (dstW - x);
  /* make sure we have enough bits in curSrcWord to do it. */
  if (curSrcBitsLeft < wb)
    {
      if ((wb - curSrcBitsLeft) > (32 - srcBitsUsed))
	{
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	  if (srcBitsUsed < 32)
	    {
	      curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	      curSrcBitsLeft += (32 - srcBitsUsed);
	    }

	  srcCur ++;
	  srcBitsUsed = 0;
	}
			
      /* no spanning a word, just take as much as we can. */
      curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
      temp = MIN(32-srcBitsUsed,32-curSrcBitsLeft);
      srcBitsUsed += temp;
      curSrcBitsLeft += temp;
			
      assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0);
    }

  /* now just special case them. */
  while (x < dstW)
    {
      dstCur->r=dstCur->g=dstCur->b=(Dgp_UInt8)(~curSrcWord >> 31)*255;
      dstCur++;
      curSrcWord = curSrcWord << 1;

      x++;
    }
}



/*#if 0   */
/* What's wrong?  Resurrecting for santa;  August 2000. / dhtesq. */

/* scaleBlitBinaryApprox uses this */
#define scaleBinaryLow 0xff
#define scaleBinaryHigh 0x00
static const Dgp_UInt8 scaleBinarySpacing[5][4]=
{{scaleBinaryLow, scaleBinaryLow, scaleBinaryLow, scaleBinaryLow},
 {scaleBinaryLow, scaleBinaryLow, scaleBinaryLow, scaleBinaryHigh},
 {scaleBinaryLow, scaleBinaryHigh, scaleBinaryLow, scaleBinaryHigh},
 {scaleBinaryLow, scaleBinaryHigh, scaleBinaryHigh, scaleBinaryHigh},
 {scaleBinaryHigh, scaleBinaryHigh, scaleBinaryHigh, scaleBinaryHigh}
};	
	
INLINE void createScale8BitTable(Dgp_UInt32 *tbl, Dgp_Int32 scaledNumBits)
{
  Dgp_UInt16 eightbitval,i;
  Dgp_UInt32 curWord, newBit;
  Dgp_Float64 factor;

  if (scaledNumBits)
    factor = 8.0 / scaledNumBits;
  else {
    memset(tbl,0,255*sizeof(long));
    return;
  }
  assert(scaledNumBits<=32);
  for (eightbitval = 0; eightbitval < 256; eightbitval++)
    {
      curWord = 0;
      for (i = 0; i<=scaledNumBits; i++)
	{
	  newBit = (eightbitval >> (8-((Dgp_Int32)(i*factor)+1))) & 0x1;
	  /*assert((i*factor)<=8); */
	  curWord = curWord | (newBit << (31-i));
	}
      tbl[eightbitval] = curWord;
    }
}


/* Blats a row of src packed bits to dst while scaling by 8-bit table lookup. Used as a helper
	function for scaleBlitBinary. tbl[i] is a table that maps an 8-bit input to tblBits[i] of
	scaled output. Note, this makes it so that we can only have scale factors between 3.125%
	and 400%. Anything over 400% can be done by multiple calls.
	srcRow	- a pointer to source data, containing the first pixel that needs to be scaled from
			src to dst.
	srcBitsLeft - how many bits are left to use for scaling in this word, 24 means the top eight
		pixels have been used, and the lower 24-bits need to be used.
	dstRow  - a pointed to the word with the first pixel to be blatted.
	dstBitsLeft specifies how many bits should be packed into this word. Note, we count bitsLeft
		from the top. so 28 bitsleft means the upper 4 pixels are full, and the lower 28 are free
	dstW - the width of the destination row.
	dstX - the first pixel in the destination that needs to be blatted (this corresponds to 
		dstRow,dstBitsLeft)
	dstMap - is the coordinate that srcRow maps to into the dst. Note dstMap<=dstX
	tbl - are the 4 tables that specify how 8-bit quantities should be scaled. We just cycle
		through all the tables in 32-bit rotations.
	tblBits - how many output bits each tbl[i] spits out.
	totalBits = tblBits[0]+tblBits[1]+tblBits[2]+tblBits[3]
 */

INLINE void scaleBlitBinaryApprox1Row(Dgp_UInt32 * srcRow, Dgp_Int32 srcBitsLeft, 
				      Dgp_UInt32 *dstRow, Dgp_Int32 dstBitsLeft,Dgp_Int32 dstW, Dgp_Int32 dstX, Dgp_Int32 dstMap,
				      Dgp_UInt32 *tbl[4], const Dgp_UInt16 tblBits[4],
				      const Dgp_UInt16 totalBits)
{
  Dgp_UInt32 * dstCur, *srcCur;
  Dgp_UInt32 srcCurWord, dstCurWord, newBits;
  Dgp_Int32 x, curTblIdx, newBitsLeft, needed;

  dstCur = dstRow;
  srcCur = srcRow;
  srcCurWord = *srcCur;
  dstCurWord = 0;
  x = 0;
  curTblIdx = 3-(Dgp_Int32)((srcBitsLeft-1) / 8.0);
  assert(curTblIdx<4 || curTblIdx >=0);


  /* So we're setup to go for a line. Now neither the srcCur or dstCur */
  /* is on a word boundary, and having the srcCur on a word boundary would */
  /* be nice, so we want to get srcCur on an 8-bit boundary. We'll do this by actually */
  /* computing the scale for the full byte of the src, and then just throwing away */
  /* what we don't want. if dstMap != dstX, then we need to do some partial */
  if((dstX - dstMap) != 0)
    {
      newBits = tbl[curTblIdx][(Dgp_UInt8)(srcCurWord >> (24 - (8*curTblIdx)))];
      newBitsLeft = tblBits[curTblIdx];

      /* So how many of these bits do we actually want to put in the dstCurWord? */
      needed = newBitsLeft -(dstX - dstMap);

      newBits = newBits << (newBitsLeft - needed);
		
      /* If dstW is really small, you might only want to write some of these bits. */
      newBitsLeft = MIN(needed,dstW);
		
      /* If the above min is dstW, we need to clear out the lower bits of this */
      /* that we don't want to blat. */
      newBits = newBits >> (32-newBitsLeft);
      newBits = newBits << (32-newBitsLeft);

      if(newBitsLeft == 0)
	newBits = 0;
      /* So now, we have tblBits[curTblIdx] bits of dest to fit into */
      /* curDstWord. We have 2 options. We can fit it all in, or we can't. */
      /* if we can't, we'll put as much as we can, and then advance the word. */
      /* else, we'll just put it all in. */
      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* We should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;
		
      srcBitsLeft -= 8;
      x += needed;
      curTblIdx++;
      curTblIdx = curTblIdx % 4; 

      /* If originally word aligned, srcBitsLeft can now be negative. / dhtesq. */
      /*if (srcBitsLeft == 0){   */
      if(srcBitsLeft <= 0){
          srcCurWord = *(++srcCur);
          srcBitsLeft = 32;
      }
    }



  if (srcBitsLeft != 32 && (x+tblBits[curTblIdx]) <= dstW)
    {
      while(srcBitsLeft != 0 && (x+tblBits[curTblIdx]) <= dstW)
	{
	  assert(curTblIdx !=0); /* if this was 0, we'd be on a 32-bit boundary. */
			
	  newBits = tbl[curTblIdx][(Dgp_UInt8)(srcCurWord >> (24 - (8*curTblIdx)))];
	  newBitsLeft = tblBits[curTblIdx];
	  /* so now, we have tblBits[curTblIdx] bits of dest to fit into */
	  /* curDstWord. We have 2 options. We can fit it all in, or we can't. */
	  /* if we can't, we'll put as much as we can, and then advance the word. */
	  /* else, we'll just put it all in. */
	  if (dstBitsLeft < newBitsLeft)
	    {
	      if (dstBitsLeft != 0)
		{
		  dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
		  newBits = newBits << dstBitsLeft;
		  newBitsLeft -= dstBitsLeft;
		  assert(newBitsLeft > 0);
		}
				
	      *dstCur = *dstCur | dstCurWord;
	      dstCur++;
	      dstCurWord = 0;
	      dstBitsLeft = 32;
	    }

	  /* we should be able to fit it all in. */

	  dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	  dstBitsLeft -= newBitsLeft;
			
			
	  srcBitsLeft -= 8;
	  x += tblBits[curTblIdx];
	  curTblIdx++;
	}
		
      /* in certain small width cases, you might not actually get to run this */
      /* loop until you get on a word boundary. */
      if (srcBitsLeft == 0)
	{
	  curTblIdx = 0;
	  /*assert(srcBitsLeft == 0); */

	  srcCurWord = *(++srcCur);
	  srcBitsLeft = 32;
	}
    }



  /* we're on a word boundary! No problem. Let's fly now. Doing 32-bits of source */
  /* at a time. */
  while ( (x+totalBits) <= dstW)
    {
		
      /* a quick check. Is this whole srcWord == 0? If so, we can skip it. */
      if (srcCurWord == 0)
	{
	  newBitsLeft = totalBits;
			
	  if (newBitsLeft > dstBitsLeft)
	    {
	      newBitsLeft -= dstBitsLeft;

	      *dstCur = *dstCur | dstCurWord;
	      dstCur++;

	      needed = newBitsLeft / 32;
	      dstCur += needed;
	      newBitsLeft -= (needed * 32);

	      dstCurWord = 0;
	      dstBitsLeft = 32;
	    }

	  dstBitsLeft -= newBitsLeft;

	  srcCurWord = *(++srcCur);
	  x += totalBits;
	  continue;
	}

      newBits = tbl[0][(Dgp_UInt8)(srcCurWord >> 24)];
      newBitsLeft = tblBits[0];

      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;

      newBits = tbl[1][(Dgp_UInt8)(srcCurWord >> 16)];
      newBitsLeft = tblBits[1];

      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;

      newBits = tbl[2][(Dgp_UInt8)(srcCurWord >> 8)];
      newBitsLeft = tblBits[2];

      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;

      newBits = tbl[3][(Dgp_UInt8)(srcCurWord)];
      newBitsLeft = tblBits[3];

      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;

      srcCurWord = *(++srcCur);
      x += totalBits;
    }

  /* how many 8-bits can we fit in? */
  while((x+tblBits[curTblIdx]) <= dstW)
    {
      assert(srcBitsLeft >= 8);
      newBits = tbl[curTblIdx][(Dgp_UInt8)(srcCurWord >> (24 - (8*curTblIdx)))];
      newBitsLeft = tblBits[curTblIdx];
      /* so now, we have tblBits[curTblIdx] bits of dest to fit into */
      /* curDstWord. We have 2 options. We can fit it all in, or we can't. */
      /* if we can't, we'll put as much as we can, and then advance the word. */
      /* else, we'll just put it all in. */
      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;
		
      srcBitsLeft -= 8;
      x += tblBits[curTblIdx];
      curTblIdx++;
      curTblIdx = curTblIdx % 4;
    }

  /* we could have a partial 8-bit value left to do. */
  if(x < dstW)
    {
      assert(srcBitsLeft >= 0);
      newBits = tbl[curTblIdx][(Dgp_UInt8)(srcCurWord >> (24 - (8*curTblIdx)))];
      newBitsLeft = tblBits[curTblIdx];

      /* so how many of these bits do we actually want to put in the dstCurWord? */
      needed=dstW-x;
      assert(needed<newBitsLeft);
      newBitsLeft = needed;

      /* clean out those bottom bits we don't want. */
      newBits = newBits >> (32 - needed);
      newBits = newBits << (32 - needed);

      /* so now, we have tblBits[curTblIdx] bits of dest to fit into */
      /* curDstWord. We have 2 options. We can fit it all in, or we can't. */
      /* if we can't, we'll put as much as we can, and then advance the word. */
      /* else, we'll just put it all in. */
      if (dstBitsLeft < newBitsLeft)
	{
	  if (dstBitsLeft != 0)
	    {
	      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
	      newBits = newBits << dstBitsLeft;
	      newBitsLeft -= dstBitsLeft;
	      assert(newBitsLeft > 0);
	    }
			
	  *dstCur = *dstCur | dstCurWord;
	  dstCur++;
	  dstCurWord = 0;
	  dstBitsLeft = 32;
	}

      /* we should be able to fit it all in. */
      dstCurWord = dstCurWord | (newBits >> (32-dstBitsLeft));
      dstBitsLeft -= newBitsLeft;
		
      x += needed;
    }

  /* make sure to flush the last word. */
  if (dstBitsLeft != 32)
    *dstCur = *dstCur | dstCurWord;
}



/* note, the beginning of a line of src must be on a word boundary, ie, srcMinX must be a multiple of 32. */
/* note, that because of the 3.1% precision in this scale, you must make sure there's enough source for */
/*  the scale factor that's up to 3.1 % smaller. */
Dgp_Int32 scaleBlitBinaryRGBApprox(Dgp_UInt32 * src,
			     Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			     DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			     Dgp_Float64 scaleX, Dgp_Float64 scaleY, Dgp_Allocators * a)
{
  /* See scaleBlitBinary and the helper function */
  /* the 2 12.5% intervals */
  Dgp_Int32 lowBits,highBits;
  Dgp_UInt32 lowTbl[256];
  Dgp_UInt32 highTbl[256];
	
  /* these arrays map which tables we use when processing. */
  /* tbl[i] is either lowTbl or highTbl, depending on position. */
  /* tblBits[i] is either lowBits or highBits, depending on tbl[i] */
  Dgp_UInt32 *tbl[4];
  Dgp_UInt16 tblBits[4];
  Dgp_UInt16 totalBits;

	
  Dgp_Int32 rdstW;
  Dgp_Int32 x,y;
  Dgp_UInt32 * tempRow, *srcRow;
  Dgp_Int32 srcLpl;
  Dgp_Int32 tempBpl;
  DgpImage_RGB *dstRow;
  Dgp_Int32 dstOff;
	
  Dgp_Int32 srcBitsLeft;

  Dgp_Int32 srcMaxX;

	/* these are used at the beginning of the row, because we might not be */
	/* perfectly aligned on an 8-bit boundary of source, we want to know where */
	/* the one before where we should start is (srcStartX, in src coords), and */
	/* what pixel that maps to in dst (dstMap, in dst coords), and how many */
	/* bits our first word of destination has left at the beginning of the line. */
  Dgp_Int32 srcStartX;
  Dgp_Int32 tempMap;	

	/* the actual scale we're going to generate this image at (a 3.1% boundary) */
  Dgp_Float64 actScaleX;
  Dgp_Int32 numStep;


	/* temporary variables inside loops. */
  Dgp_Int32 tempStep, srcStep;

  /* so what 2 table values do we need? */
  lowBits = (Dgp_Int32)floor(scaleX / .125);
  highBits = (Dgp_Int32)ceil(scaleX / .125);

	/* generate those 2 tables. If lowBits==highBits, we happen to  */
	/* be on a 12.5% boundary, and need only one. */
  if (lowBits == highBits)
    {
      createScale8BitTable(lowTbl, lowBits);
      tbl[0]=tbl[1]=tbl[2]=tbl[3]=lowTbl;
      tblBits[0]=tblBits[1]=tblBits[2]=tblBits[3] = (Dgp_UInt16)lowBits;
      actScaleX = scaleX;
    }
  else
    {
      /* figure out what scale we're really generating at? */
      /* (its the 3.1% boundary at or below the actual x-scale) */
		
      numStep = (Dgp_Int32)(scaleX / (1/32.0));
      actScaleX = numStep * (1/32.0);

      if(actScaleX != (lowBits * (1/8.0)))
	createScale8BitTable(highTbl, highBits);
		
      createScale8BitTable(lowTbl, lowBits);
				
      /* if you're actual scale is lower than 3.15%, we just can't do it. */
      if (actScaleX == 0.0)
	return 0;
      else 
	{
	  numStep -= (lowBits * 4);
	  assert(numStep >=0 && numStep<4);
	  for (x = 0; x < 4; x++)
	    {
	      tbl[x]=((scaleBinarySpacing[numStep][x]==scaleBinaryLow) ? lowTbl : highTbl);
	      tblBits[x] = (Dgp_UInt16)(tbl[x]==lowTbl ? lowBits : highBits);
	    }
	}
    }

  assert(actScaleX <= scaleX);
  /* So now we know the actual scale (which isn't exactly the real scale)
   * that we're going to generate, we have to make sure that the desired
   * destination coords are in our scale, and we don't wrap around reading
   * into the next line of the source.
	 */
  srcMaxX = (srcMinX + (srcBpl * 8));
  if ((Dgp_Int32)(srcMaxX*actScaleX) < (dstX + (dstW-1)))
    {
      rdstW = (Dgp_Int32)(srcMaxX*actScaleX) - dstX;
      rdstW = MAX(rdstW,0);
    }
  else
    rdstW = dstW;

  totalBits = tblBits[0]+tblBits[1]+tblBits[2]+tblBits[3];
	
  srcLpl = srcBpl / 4;

  tempBpl = (Dgp_Int32)(ceil(dstW/(8*sizeof(Dgp_UInt32)))+2)*sizeof(Dgp_UInt32);
  tempRow = (Dgp_UInt32 *)a->malloc(tempBpl);
  if (!tempRow) 
    return 0;
	
	/* find the 8-bit boundary of source that we want */
	/* to start on. */
  tempMap = dstX - (dstX % totalBits);
  srcStartX = (Dgp_Int32)(dstX / totalBits);
  srcStartX = srcStartX * 32;
	

  tempStep = 0;
  srcStep = 0;
  for ( x = 0; (tempMap + tblBits[x]) <= dstX; x++)
    {
      if (tblBits[x])
	{
	  tempMap += (tempStep + tblBits[x]);
	  srcStartX += (srcStep + 8);
	  tempStep = 0;
	  srcStep = 0;
	}
      else
	{
	  srcStep += 8;
	  tempStep += 0;
	}
    }


  dstOff=((dstY-dstMinY)*dstBpl)+((dstX-dstMinX)*sizeof(DgpImage_RGB));
  dstRow=(DgpImage_RGB *)((Dgp_Int8 *)dst+dstOff);
  for (y = 0 ; y < dstH ; y++)
    {
      srcRow = src + ((Dgp_Int32)((dstY+y)/scaleY)-srcMinY)*srcLpl;
      srcRow = srcRow + (Dgp_Int32)((srcStartX-srcMinX)/32);
		
      srcBitsLeft = 32-(srcStartX % 32);
		
      memset(tempRow, 0 , tempBpl);

      scaleBlitBinaryApprox1Row(srcRow, srcBitsLeft, tempRow, 32, rdstW,
				dstX, tempMap, tbl,tblBits,totalBits);

      blitBinaryRGBRow(dstRow, tempRow,32,dstW);
				
      dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + dstBpl);
    }
	
  a->free(tempRow);
  return 1;	
}


/* */
/* Note, the beginning of a line of src must be on a word boundary, ie, srcMinX must be a multiple of 32. */
/* writes with or. */
/* */
/* Trying to resurrect, August 2000 / dhtesq. */
/* Since no one else uses this function, I'm changing the internals to make it more friendly */
/* to token operations.  The other scaling functions assume that the src buffer fits into */
/* the (entire) destintation buffer at some scale, whereas when you start with a token bitmap */
/* you want the src buffer to fit into the *destination region* at the desired scale. */
/* */
/* I was only able to get this to work by ignoring the dstMap parameter in the call to */
/* scaleBlitBinaryApprox1Row() -- that is, by setting it equal to dstX.  Unclear (to me)  how WB */
/* was trying to handle bits hanging off the word boundaries. */
/* / dhtesq. */
/* */
void scaleBlitBinaryApprox(Dgp_UInt32 * src,
		     Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		     Dgp_UInt32 * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		     DgpCoord_Resolution * res)
{
  /* So this is the kicker. We want to scale this (x-direction)
   * by some amount, and we want to do it so that we take in an
   * 8-bit amount of source an spew out some number of bits in
   * the destination. So we're limited to a precision of 12.5%. What if
   * we wanted 120% scaling? if we spit out 9-bits, we'll get 112.5%, 
   * and the next step up is 10-bits is 125%. But we want better 
   * precision than that. so we'll make a table for the number of bits
   * below the scale factor, and one for the number of bits above, and then
   * we'll just alternate between tables in a pattern of 4 8-bit lookups.
   * This gives us a 3.1% precision if your image is over 32 pixels wide,
   * which isn't too bad.
   */
  
  /* Since we only have a 3.1% precision between the low and high values
   * and we want the scaling to look as linear as possible, we have a table
   * that just shows how the low/high tables should be used in the 3.1% steps
   * from lowBits/8.0 to highBits/8.0
   */
  /* the 2 12.5% intervals */
  Dgp_Int32 lowBits,highBits;
  Dgp_UInt32 lowTbl[256];
  Dgp_UInt32 highTbl[256];
	
  /* these arrays map which tables we use when processing. */
  /* tbl[i] is either lowTbl or highTbl, depending on position. */
  /* tblBits[i] is either lowBits or highBits, depending on tbl[i] */
  Dgp_UInt32 *tbl[4];
  Dgp_UInt16 tblBits[4];
  Dgp_UInt16 totalBits;
	
  Dgp_Int32 x, y;
  Dgp_UInt32 *dstRow, *srcRow;
  Dgp_Int32 dstLpl;
  Dgp_Int32 srcLpl;

  Dgp_Int32 dstBitsLeft;

  Dgp_Int32 srcBitsLeft;

  Dgp_Int32 srcMaxX;


  /* these are used at the beginning of the row, because we might not be */
  /* perfectly aligned on an 8-bit boundary of source, we want to know where */
  /* the one before where we should start is (srcStartX, in src coords), and */
  /* what pixel that maps to in dst (dstMap, in dst coords), and how many */
  /* bits our first word of destination has left at the beginning of the line. */
  Dgp_Int32 dstInitBitsLeft;
  Dgp_Int32 srcStartX = 0;
  Dgp_Int32 dstMap;	
  
  /* the actual scale we're going to generate this image at (a 3.1% boundary) */
  Dgp_Float64 scaleX, scaleY, actScaleX;
  Dgp_Int32 numStep;
  
    /* temporary variables inside loops. */
  Dgp_Int32 dstStep, srcStep;

  scaleX = DgpCoord_double(res->xRes);
  scaleY = DgpCoord_double(res->yRes);

  /* so what 2 table values do we need? */
  lowBits = (Dgp_Int32)floor(scaleX / .125);
  highBits = (Dgp_Int32)ceil(scaleX / .125);

	/* generate those 2 tables. If lowBits==highBits, we happen to  */
	/* be on a 12.5% boundary, and need only one. */
  if (lowBits == highBits){
      createScale8BitTable(lowTbl, lowBits);
      tbl[0]=tbl[1]=tbl[2]=tbl[3]=lowTbl;
      tblBits[0]=tblBits[1]=tblBits[2]=tblBits[3] = (Dgp_UInt16)lowBits;
      actScaleX = scaleX;
  }
  else {
      /* figure out what scale we're really generating at? */
      /* (its the 1.5% boundary at or below the actual x-scale) */
		
      numStep = (Dgp_Int32)(scaleX / (1/32.0));
      actScaleX = numStep * (1/32.0);

      if(actScaleX != (lowBits * (1/8.0))){
          createScale8BitTable(highTbl, highBits);
      }
		
      createScale8BitTable(lowTbl, lowBits);
				
      /* If you're actual scale is lower than 3.15%, we just can't do it. */
      if (actScaleX == 0.0){
          return;
      } else {
          numStep -= (lowBits * 4);
          assert(numStep >=0 && numStep<4);
          for (x = 0; x < 4; x++){
              tbl[x]=((scaleBinarySpacing[numStep][x]==scaleBinaryLow) ? lowTbl : highTbl);
              tblBits[x] = (Dgp_UInt16)(tbl[x]==lowTbl ? lowBits : highBits);
          }
      }
  }
	
  /* So now we know the actual scale (which isn't exactly the real scale)
   * that we're going to generate, we have to make sure that the desired
   * destination coords are in our scale, and we don't wrap around reading
   * into the next line of the source.
   */
  srcMaxX = (srcMinX + (srcBpl * 8));
  /* Changed this trap so that the destination region (dstX, dstY, dstX + dstW - 1, etc.) */
  /* is allowed to contain the entire source, rather than a scaled portion. / dhtesq. */
  if((Dgp_Int32)((srcMaxX - srcMinX) * actScaleX) < dstW){
      dstW = (Dgp_Int32)((srcMaxX - srcMinX) * actScaleX);
  }
  /*
  if ((Dgp_Int32)(srcMaxX * actScaleX) < (dstX + (dstW - 1))){
      dstW = (Dgp_Int32)(srcMaxX * actScaleX) - dstX;
      if (dstW <= 0){
          return;
      }
  }
  */

  totalBits = tblBits[0] + tblBits[1] + tblBits[2] + tblBits[3];
	
  dstLpl = dstBpl / 4;
  srcLpl = srcBpl / 4;

  /* Changing these will only affect the position of the blatted data. / dhtesq. */
  dstRow = dst + ((dstY - dstMinY) * dstLpl);
  dstRow += (Dgp_Int32)((dstX - dstMinX) / 32);
  dstInitBitsLeft = 32 - ((dstX - dstMinX) % 32);
  /*/ */

  /* Find the 8-bit boundary of source that we want */
  /* to start on. */
  dstMap = dstX - (dstX % totalBits);
  srcStartX = (Dgp_Int32)(dstX / totalBits);
  srcStartX = srcStartX * 32;

  dstStep = 0;
  srcStep = 0;
  for(x = 0; (dstMap + tblBits[x]) <= dstX; x++){
      if (tblBits[x]){
          dstMap += (dstStep + tblBits[x]);
          srcStartX += (srcStep + 8);
          dstStep = 0;
          srcStep = 0;
      }
      else {
          srcStep += 8;
          dstStep += 0;
      }
  }
	
  /* So now, srcStartX is the 8-bit boundary (in src coords) */
  /* of where we'll start generating 8-bit values on this line. */
  /* dstMap is what that maps to in the dst system. */

  /* srcStartX obsolesced... / dhtesq. */
  for(y = 0; y < dstH / scaleY; y++){
      /*srcRow = src + ((Dgp_Int32)((dstY + y) / scaleY) - srcMinY) * srcLpl; */
      /*srcRow = src + (Dgp_Int32)((srcMinY + (y / scaleY)) * srcLpl); */
      srcRow = src + (Dgp_Int32)((srcMinY + y) * srcLpl);
      /*srcRow = srcRow + (Dgp_Int32)((srcStartX - srcMinX) / 32); */
      srcRow += (Dgp_Int32)(srcMinX / 32);
		
      /*srcBitsLeft = 32 - (srcStartX % 32); */
      srcBitsLeft = 32;
      dstBitsLeft = dstInitBitsLeft;

      scaleBlitBinaryApprox1Row(srcRow, srcBitsLeft, dstRow, dstBitsLeft, dstW,	dstX, dstX/*dstMap*/, 
          tbl, tblBits, totalBits);

      /* Try blatting all lines, incrementing only when all scaling to a given destination row */
      /* have been done.  Takes longer but some quality improvement for downscaling. */
      if(!((y + 1) % (Dgp_Int32)(1 / scaleY))){
          dstRow += dstLpl;
      }
  }		

  return;
}
/*#endif // 0 */

