/**************************************************************** 
 *
 * dgcl.c
 *
 ****************************************************************/

#include "dgp_cl.h"

#include <stdlib.h>
#include <string.h>
#include "PsgFix.h"

/* internal string length (used by DgpCl_printHelp()) */
#define TEMP_STR_LEN (DGPCL_MAX_OPTION_LEN + 14)

/*
 * DgpCl_printHelp()
 *   Command line help printer.  Outputs program usage info 
 *   such as input, output, options...
 */
DGP_LIBEXPORT void
DgpCl_printHelp(FILE *fp, 
                const Dgp_Int8 *header, 
                DgpCl_Option *options)
{
  static const Dgp_Int8 *typeNames[] = {
    "<string>", "<int>", "<float>", "", "", ""
  };

  Dgp_Int32 i, len, maxlen;
  Dgp_Int8 temp[TEMP_STR_LEN];
  const Dgp_Int8 *tmpstr1, *tmpstr2;

  /* no file pointer, no output */
  if(fp == NULL) return;

  /* print header if there is one */
  if(header != NULL) {
    fprintf(fp, "\n%s\n\nOptions:\n", header);
  }

  /* if there aren't any options, return */
  if(options == NULL) return;

  /* set the last null character just in case */
  temp[TEMP_STR_LEN-1] = '\0';

  /* figure out indentation */
  for(i = 0, maxlen = 0; options[i].name != NULL; i++) {
    len = sprintf(temp, "-%.*s %s", DGPCL_MAX_OPTION_LEN, options[i].name, typeNames[options[i].type]);
    if (len > maxlen) maxlen = len;
  }

  /* print option help */
  for (i = 0; options[i].name != NULL; i++) {
    sprintf(temp, "-%.*s %s", DGPCL_MAX_OPTION_LEN, options[i].name, typeNames[options[i].type]);
    fprintf(fp, "  %-*s  ", (int)maxlen, temp);

    tmpstr1 = options[i].help;
    if(tmpstr1 != NULL) {

      while((tmpstr2 = strchr(tmpstr1, '\n')) != NULL) {
        fprintf(fp, "%.*s%*s", (tmpstr2 - tmpstr1 + 1), options[i].help, (int)(maxlen + 4), "");
        tmpstr1 = tmpstr2 + 1;
      }
      fprintf(fp, "%s\n", tmpstr1);

    } else {
      fprintf(fp, "\n");
    }

    /* print default value */
    switch(options[i].type) {
    case DGP_CL_STRING:

	    fprintf(fp, "%*s(Default = \"%s\")\n", (int)(maxlen + 4), "",
		    ((options[i].def_value != NULL) ? (options[i].def_value) : ""));
	    break;

	  case DGP_CL_INT:
	  case DGP_CL_FLOAT:
      
	    fprintf(fp, "%*s(Default = %s)\n", (int)(maxlen + 4), "",
		    options[i].def_value);
	    break;

	  case DGP_CL_FLAG:
    case DGP_CL_QUIET:
    case DGP_CL_HELP:

	    /*fprintf(fp, "\n"); */
      break;

    default:
      break;
    }

  }

  /*fprintf(fp, "\n"); */
}



/*
 * DgpCl_parse()
 *   Command line parser function.
 *   (see dgp_cl.h for details)
 */
DGP_LIBEXPORT ErrCode
DgpCl_parse(Dgp_State *dgp_state, 
            Dgp_Int32 *pargc, 
            Dgp_Int8 **argv, 
            DgpCl_Option *options,
            FILE *err_out, 
            const Dgp_Int8 *help_header,
            DgpCl_ValidatorType arg_validator,
            void *validate_cookie)
{
  DGP_STATE_PROLOGUE2("DgpCl_parse()");

  Dgp_Int32 i, j, numLeft = 0, argc;
  Dgp_Int8 *appName;
  Dgp_Int8 *cur_arg;

  /* validate params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(options);

  /* if there's no command line, we're done */
  if (argv == NULL) NORMAL_RETURN;

  /* validate more params */
  DGP_CHECK_NULL(pargc);

  argc = *pargc;
  appName = argv[0];

  /* start parsing */

  /* skip the first arg, it's the program name */
  for (i = 1; i < argc; i++) {

    for (j = 0; options[j].name != NULL; j++) {
      
      /* skip options we've already parsed */
      if(options[j].found) continue;

      /* compare current arg to option name */
#ifdef _WIN32
      cur_arg = (((argv[i][0] == '-') || (argv[i][0] == '/')) ? ((argv[i])+1) : (argv[i]));
#else
      cur_arg = ((argv[i][0] == '-') ? ((argv[i])+1) : (argv[i]));
#endif
      if(strncmp(cur_arg, options[j].name, DGPCL_MAX_OPTION_LEN) == 0) {

        /* figure out how to parse this arg */
	      switch (options[j].type) {

	      case DGP_CL_STRING:

          if(i >= (argc - 1)) {
            if(err_out != NULL) {
              fprintf(err_out, "%s: Error -- invalid use of option '%s'\n", appName, argv[i]);
            }
            DGP_ERR_CLEANUP(cErrInvalidCmdLine);
          }

	        *((Dgp_Int8 **) options[j].value) = argv[++i];
	        break;

	      case DGP_CL_INT:

          if(i >= (argc - 1)) {
            if(err_out != NULL) {
              fprintf(err_out, "%s: Error -- invalid use of option '%s'\n", appName, argv[i]);
            }
            DGP_ERR_CLEANUP(cErrInvalidCmdLine);
          }

	        *((Dgp_Int32 *) options[j].value) = atoi(argv[++i]);
	        break;

	      case DGP_CL_FLOAT:

          if(i >= (argc - 1)) {
            if(err_out != NULL) {
              fprintf(err_out, "%s: Error -- invalid use of option '%s'\n", appName, argv[i]);
            }
            DGP_ERR_CLEANUP(cErrInvalidCmdLine);
          }

	        *((Dgp_Float32 *) options[j].value) = (Dgp_Float32) atof(argv[++i]);
	        break;

	      case DGP_CL_FLAG:

	        *((Dgp_Bool *) options[j].value) = DGP_TRUE;
	        break;

        case DGP_CL_QUIET:

          /* clear this pointer, no output */
          err_out = NULL;

          *((Dgp_Bool *) options[j].value) = DGP_TRUE;
          break;

        case DGP_CL_HELP:

          DGP_ERR_CLEANUP_NO_MSG(cErrInvalidCmdLine);
          break;

        default:
          
          DGP_ERRF0(cErrInternal, "invalid option type");
          break;
	      }
    
        /* indicate that we've found this one */
        options[j].found = DGP_TRUE;

        /* parse next argument */
	      break;
      }
    }

    if (options[j].name == NULL) {
      
#ifdef _WIN32
      if((argv[i][0] == '-') || (argv[i][0] == '/')) {
#else
      if(argv[i][0] == '-') {
#endif
        /* this is an option we don't understand... */
        if(err_out != NULL) {
          fprintf(err_out, "%s: Warning -- unknown option '%s'\n\n", appName, argv[i]);
        }

      } else {

        /* this is extra info */
        argv[numLeft++] = argv[i];

      }
    }
  }

  /* Return number of unparsed args. */
  *pargc = numLeft;

  /* Now, allow the user to validate the new values. */
  if(arg_validator != NULL) {
    DGP_ERR_PASS_CLEANUP(arg_validator(dgp_state, numLeft, argv, err_out, validate_cookie)) ;
  }


  NORMAL_RETURN;

ERR_CLEANUP:

  if((err_out != NULL) && (DGP_ERROR_IS(cErrInvalidCmdLine))) {
    DgpCl_printHelp(err_out, help_header, options);
  }

  ERR_RETURN;
}
