/****************************************************************
 *
 * dgcl.h
 *
 * Generic command line parsing utilities.
 *
 ****************************************************************/

#ifndef _DGP_CL_H
#define _DGP_CL_H

#include "dgp_state.h"

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Maximum length of an option name (if you need more
 *   characters than this, you have other problems...)
 */
#define DGPCL_MAX_OPTION_LEN 30


/**
 * Command line arg data type constants 
 *   (assigned to the 'value' field of <DgpCl_Option>)
 */
typedef enum {
  DGP_CL_STRING,  /* value should point to a <Dgp_Int8 *>       */
  DGP_CL_INT,     /* value should point to a <Dgp_Int32>    */
  DGP_CL_FLOAT,   /* value should point to a <Dgp_Float32>  */
  DGP_CL_FLAG,    /* value should point to a <Dgp_Bool>     */
  DGP_CL_QUIET,   /* special value for "quiet", value should point to a <Dgp_Bool> */
  DGP_CL_HELP     /* special value for help flag, value field ignored */
} DgpCl_OpType;

/**
 * Structure which holds info used to parse an argument on
 *   the command line.
 */
typedef struct {
  const Dgp_Int8 *name;  /* name of the option 
                           (with no leading symbol such as '-' or '/') */
  DgpCl_OpType type;     /* one of the constants defined above */
  void *value;           /* pointer to a variable (of the correct type as defined by
                            the 'type' value) which will be assigned the value
                            found on the command line (if any), otherwise it will not
                            be altered */
  const Dgp_Int8 *help;  /* help string for this option, long strings should be broken up by
                            '\n' characters */
  const Dgp_Int8 *def_value;  /* a string representing the default value for this option
                                 if one isn't assigned on the command line (this is ignored
                                 for the types (DGP_CL_FLAG, DGP_CL_QUIET, DGP_CL_HELP) */
  Dgp_Bool found;        /* boolean indicating whether this option was found on the 
                            command line (should be DGP_FALSE initially) */
} DgpCl_Option;


/**
 * Type definition for the validator function which can
 * be passed to the DgpCl_parse() function below.
 *
 * Input:
 *   dgp_state     - digipaper passed in to the DgpCl_parse() function, unchanged.
 *   unparsed_argc - number of strings in unparsed_argv.
 *   unparsed_argv - arguments found on the command line which were not parsed.
 *   err_out       - either the err_out function passed to DgpCl_parse() or
 *                   NULL (could be set to NULL if an arg of type DGP_CL_QUIET
 *                   is parsed).
 *   validate_cookie - untouched cookie which was passed to DgpCl_parse().
 *
 * Output:
 *   Should do any checking of parsed command line arguments and print any
 *   problems to err_out (if not NULL).
 *
 * Returns:
 *   cSuccessful        - if no problems were encountered.
 *   cErrInvalidCmdLine - if any invalid values were passed on the command line,
 *                        this will result in "help" being printed to err_out by
 *                        the DgpCl_parse() function.  the error will then be
 *                        returned to the caller of DgpCl_parse().
 *   <err_code>         - will be passed to the caller to DgpCl_parse().
 */
typedef ErrCode (*DgpCl_ValidatorType)(Dgp_State *dgp_state,
                                       Dgp_Int32 unparsed_argc, 
                                       Dgp_Int8 **unparsed_argv,
                                       FILE *err_out,
                                       void *validate_cookie);

/**
 * DGPCL_PRINT_ERRF*
 *   Macro to be used to print errors if invalid command line values are
 *   encountered in the validator function.
 */
#define DGPCL_PRINT_ERRF0(_fperr_out, _err_str) \
do { \
  if((_fperr_out) != NULL) { \
    fprintf((_fperr_out), "Error -- " _err_str "\n"); \
  } \
} while(0)

#define DGPCL_PRINT_ERRF1(_fperr_out, _err_str, _arg1) \
do { \
  if((_fperr_out) != NULL) { \
    fprintf((_fperr_out), "Error -- " _err_str "\n", (_arg1)); \
  } \
} while(0)

#define DGPCL_PRINT_ERRF2(_fperr_out, _err_str, _arg1, _arg2) \
do { \
  if((_fperr_out) != NULL) { \
    fprintf((_fperr_out), "Error -- " _err_str "\n", (_arg1), (_arg2)); \
  } \
} while(0)


/**
 * DgpCl_parse()
 *   Command line parser function.
 *
 * Input:
 *   dgp_state - valid digipaper state
 *   pargc     - pointer to integer which should contain the number
 *               of args in argv.
 *   argv      - command line args to be parsed
 *   options   - an array of options used to parse argv.  the last
 *               option should have a name which is NULL, representing
 *               the end of the array of options.
 *   err_out   - file to which error output should be written if problems
 *               are encountered in the command line, and to which
 *               "help" info will be printed if a value of type DGP_CL_HELP
 *               is parsed or if there is an error on the command line.  the
 *               value may be NULL, in which no output will be printed.
 *   help_header - header for the "help" output (usually contains the
 *                 usage, input, and output info for the program).
 *   arg_validator   - function which will be called iff there were no problems
 *                     encountered while parsing the command line.  this function
 *                     should evaluate the parsed values and determine if any
 *                     invalid values were given.  (see description for 
 *                     DgpCl_ValidatorType above).
 *   validate_cookie - cookie which will be passed to the arg_validator function.
 *
 * Output:
 *   Fills in the value fields in the options array with the parsed values.  Any
 *   strings returned by this function should not be freed.
 *
 * Returns:
 *   cSuccessful - no problems were encountered while parsing the command line.
 *   cErrInvalidCmdLine - if incorrect values or args were given on the command line.
 *   <err_code>  - other error codes for internal errors and errors returned by the
 *                 arg_validator function.
 */
DGP_LIBEXPORT ErrCode
DgpCl_parse(Dgp_State *dgp_state, 
            Dgp_Int32 *pargc, 
            Dgp_Int8 **argv, 
            DgpCl_Option *options,
            FILE *err_out, 
            const Dgp_Int8 *help_header,
            DgpCl_ValidatorType arg_validator,
            void *validate_cookie);

/*
 * This wasn't originally in here, but it helps if you want to trap for certain 'help'
 * situations (like argc == 1) without throwing an error at the user.
 */
DGP_LIBEXPORT void
DgpCl_printHelp(FILE *fp, 
                const Dgp_Int8 *header, 
                DgpCl_Option *options);

/**
 * DGPCL_*_CMD()
 *   Default command line args (don't reinvent the wheel).  These
 *   should be used when defining the options array for your program.
 */

#define DGPCL_HELP_CMD \
{"h", DGP_CL_HELP, NULL, \
"Display usage information", NULL, DGP_FALSE}

#define DGPCL_QUIET_CMD(_bool_var) \
{"q", DGP_CL_QUIET, &(_bool_var), \
"Quiet output; stifle reports", NULL, DGP_FALSE}

#define DGPCL_RES_CMD(_int_var, _sdef_val) \
{"r", DGP_CL_INT, &(_int_var), \
"Default input file resolution. Only used if no resolution specified in file", (_sdef_val), DGP_FALSE}

#define DGPCL_RESJ_CMD(_int_var, _sdef_val) \
{"jr", DGP_CL_INT, &(_int_var), \
"Set resolution of output jpegs", (_sdef_val), DGP_FALSE}

#define DGPCL_RESX_CMD(_int_var, _sdef_val) \
{"rx", DGP_CL_INT, &(_int_var), \
"Set resolution for x dimension", (_sdef_val), DGP_FALSE}

#define DGPCL_RESY_CMD(_int_var, _sdef_val) \
{"ry", DGP_CL_INT, &(_int_var), \
"Set resolution for y dimension", (_sdef_val), DGP_FALSE}

#define DGPCL_OUTFNAME_CMD(_str_var, _sdef_val) \
{"o", DGP_CL_STRING, &(_str_var), \
"Set output file name", (_sdef_val), DGP_FALSE}

#define DGPCL_VERBOSE_CMD(_bool_var) \
{"v", DGP_CL_FLAG, &(_bool_var), \
"Verbose output", NULL, DGP_FALSE}

#define DGPCL_COLOR_CMD(_bool_var) \
{"color", DGP_CL_FLAG, &(_bool_var), \
"Force color output", NULL, DGP_FALSE}

#define DGPCL_MONO_CMD(_bool_var) \
{"mono", DGP_CL_FLAG, &(_bool_var), \
"Force monochrome output", NULL, DGP_FALSE}

#define DGPCL_INC_DICT_CMD(_int_var, _sdef_val) \
{"id", DGP_CL_INT, &(_int_var), \
"Page interval for incremental dictionaries", (_sdef_val), DGP_FALSE}

#define DGPCL_CLR_DICT_CMD(_int_var, _sdef_val) \
{"cd", DGP_CL_INT, &(_int_var), \
"Page interval for clearing dictionaries", (_sdef_val), DGP_FALSE}

#define DGPCL_JPEG_QUAL_CMD(_int_var, _sdef_val) \
{"jq", DGP_CL_INT, &(_int_var), \
"Set the quality of jpeg compression (range 0-100)", (_sdef_val), DGP_FALSE}

#define DGPCL_JPEG_SCALE_CMD(_float_var, _sdef_val) \
{"js", DGP_CL_FLOAT, &(_float_var), \
"Scale down large jpegs by this factor (relative to\n" \
"page resolution)", (_sdef_val), DGP_FALSE}

#define DGPCL_DESKEW_CMD(_bool_var) \
{"deskew", DGP_CL_FLAG, &(_bool_var), \
"Attempt to deskew pages", NULL, DGP_FALSE}

#define DGPCL_HALFTONE_CMD(_bool_var) \
{"half", DGP_CL_FLAG, &(_bool_var), \
"Attempt to do halftone detection", NULL, DGP_FALSE}




/**
 * DGPCL_NEW_*_CMD()
 *   Macros for creating custom command line args.  These
 *   should be used when defining the options array for your program.
 *   (Much safer than assigning the DgpCl_Options values directly)
 */

/* new DGP_CL_INT command line arg */
#define DGPCL_NEW_INT_CMD(_sarg, _int_var, _shelp, _sdef_val) \
{(_sarg), DGP_CL_INT, &(_int_var), (_shelp), (_sdef_val), DGP_FALSE}

/* new DGP_CL_FLOAT command line arg */
#define DGPCL_NEW_FLOAT_CMD(_sarg, _float_var, _shelp, _sdef_val) \
{(_sarg), DGP_CL_FLOAT, &(_float_var), (_shelp), (_sdef_val), DGP_FALSE}

/* new DGP_CL_STRING command line arg */
#define DGPCL_NEW_STRING_CMD(_sarg, _str_var, _shelp, _sdef_val) \
{(_sarg), DGP_CL_STRING, &(_str_var), (_shelp), (_sdef_val), DGP_FALSE}

/* new DGP_CL_FLAG command line arg */
#define DGPCL_NEW_FLAG_CMD(_sarg, _bool_var, _shelp) \
{(_sarg), DGP_CL_FLAG, &(_bool_var), (_shelp), NULL, DGP_FALSE}

/* new DGP_CL_QUIET command line arg (special quiet flag) */
#define DGPCL_NEW_QUIET_CMD(_sarg, _bool_var, _shelp) \
{(_sarg), DGP_CL_QUIET, &(_bool_var), (_shelp), NULL, DGP_FALSE}

/* new DGP_CL_HELP command line arg (special help flag) */
#define DGPCL_NEW_HELP_CMD(_sarg, _shelp) \
{(_sarg), DGP_CL_HELP, NULL, (_shelp), NULL, DGP_FALSE}


#ifdef __cplusplus
}
#endif

#endif /* _DGP_CL_H */
