
/*
   Pedro Felzenszwalb, Summer '98.
   New color extension to position block encoder & decoder.
*/

#include <stdio.h>
#include <stdlib.h>
#include "huff.h"
#include "bitfile.h"
#include "list.h"
#include "encode-colors.h"
#include "color-extension.h"
#include "dgp_allocators.h"
#include "hash_table.h"
#include "mmcompress.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcolcomp/encode-colors.c,v 1.32 2003/02/10 22:53:22 mkrocken Exp $");

/* represents a unique color to be encoded */
typedef struct {
	color_tag *tag;
	Dgp_Int32 count;
	DgpEncode_Huff *Huff_elt;
} unique_tag;

/* each run in the color extension is a color tag + its length */
typedef struct {
	color_tag *tag;
	Dgp_Int32 len;
} tag_run;

/* small pool stuff */
#define UTAG_POOL_ID DGP_SMALLPOOL_MIN16

#define UTAG_SEC_MALLOC(_pputag) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, UTAG_POOL_ID, (void **)(_pputag)))
#define UTAG_FREE(_putag) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, UTAG_POOL_ID, (void *)(_putag)))

#define TAGRUN_POOL_ID DGP_SMALLPOOL_MIN8

#define TAGRUN_SEC_MALLOC(_pptagrun) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, TAGRUN_POOL_ID, (void **)(_pptagrun)))
#define TAGRUN_FREE(_ptagrun) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, TAGRUN_POOL_ID, (void *)(_ptagrun)))


/* check if two color tags are identicals */
static Dgp_Bool 
tag_isequal(color_tag *c1, color_tag *c2) { 
	if ((c1->color_type == RGB_COLOR) && (c2->color_type == RGB_COLOR))
		return ((c1->value.rgb_color.r == c2->value.rgb_color.r) &&
				(c1->value.rgb_color.g == c2->value.rgb_color.g) &&
				(c1->value.rgb_color.b == c2->value.rgb_color.b));

	if ((c1->color_type == IMAGE_INDEX) && (c2->color_type == IMAGE_INDEX))
		return (c1->value.image_index == c2->value.image_index);

	if ((c1->color_type == SAMP_CLASS) && (c2->color_type == SAMP_CLASS))
		return (c1->value.samp_class->samptoknum == c2->value.samp_class->samptoknum);
	

	return DGP_FALSE;
}

/* used by the hashtable to find unique colors */
static Dgp_Bool 
ht_isequal(void *cookie, DgpAdt_ht_elt v1, DgpAdt_ht_elt v2) {
	color_tag *c1 = ((unique_tag *)v1)->tag;
	color_tag *c2 = ((unique_tag *)v2)->tag;

	/* don't bug me with warnings */
	(void)cookie;
  
	return (tag_isequal(c1, c2));
}

/* hash function for finding unique colors */
static Dgp_UInt16 
ht_hash(void *cookie, DgpAdt_ht_elt e) {
	unique_tag *c = (unique_tag *)e;
  
	/* don't bug me */
	(void)cookie;

	/* return the image index if this is such thing */
	if (c->tag->color_type == SAMP_CLASS)
		return (Dgp_UInt16)c->tag->value.samp_class->samptoknum;

	/* return the image index if this is such thing */
	if (c->tag->color_type == IMAGE_INDEX)
		return (Dgp_UInt16)c->tag->value.image_index;

	/* its got to be an value.rgb_color... return red component as hash value */
	return (Dgp_UInt16)c->tag->value.rgb_color.r;
}

/* 
	Find unique color tags given a number of color tag runs and return a list of 
	representatives together with a hastable that can find the representative
	of any given color tag. We also count how many runs refer to a given unique color tag.
	Actually, the list of unique colors is split in two, one for colors and one for image indices.
*/
static ErrCode
histogram_colors(color_cookie *ccookie, DgpAdt_List runs,
				 DgpAdt_HashTable *_hist, DgpAdt_List *_colors, DgpAdt_List *_images) 
{
  Dgp_State *dgp_state = ccookie->dgp_state;
  DGP_PROLOGUE("histogram_colors()", DGP_PERRSTATE);
  DgpAdt_HashTable ht = NULL;
	unique_tag *hist_elt = NULL;
	DgpAdt_ht_elt in_table;
	DgpAdt_List unique_colors = NULL;
	DgpAdt_List unique_images = NULL;
	DgpAdt_ListNode runnode;
	tag_run *run;

	/* get new lists */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&unique_colors, dgp_state));
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&unique_images, dgp_state));
  
	/* get a hashtable */
	DGP_ERR_PASS_CLEANUP(DgpAdt_htNew(&ht, NULL, ht_isequal, ht_hash, NULL, 0, dgp_state));
  
	/* get a unique color element */
	UTAG_SEC_MALLOC(&hist_elt);

	DGPADT_FOREACH (runnode, runs) {
		run = (tag_run *)DGPADT_LIGET(runnode);
		/* for each run of colors  */
		hist_elt->tag = run->tag;
		DGP_ERR_PASS_CLEANUP(DgpAdt_htAddToTable(ht, (DgpAdt_ht_elt)hist_elt, &in_table));

    if (!in_table) {
			DgpAdt_ListNode retnode;
			/* the element was added to the table so its a new unique tag */
			hist_elt->count = 1;
			/* put the element in the right list */
			if ((hist_elt->tag->color_type == IMAGE_INDEX) ||
				(hist_elt->tag->color_type == SAMP_CLASS)) {
				DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(unique_images, (void *)hist_elt, &retnode));
			} else {
				DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(unique_colors,(void *)hist_elt, &retnode));
			}
			/* get a new element */
			UTAG_SEC_MALLOC(&hist_elt);
		} else {
			/* the element was already in the table, only increas its count */
			((unique_tag *)in_table)->count++;
		}
	}

	/* free the element thats left */
	UTAG_FREE(hist_elt);
  hist_elt = NULL;

	*_hist = ht;
  ht = NULL;
	*_colors = unique_colors;
  unique_colors = NULL;
	*_images = unique_images;
  unique_images = NULL;

	NORMAL_RETURN;

ERR_CLEANUP:

  if(unique_colors != NULL) DgpAdt_liFree(unique_colors);
  if(unique_images != NULL) DgpAdt_liFree(unique_images);
  if(ht != NULL) DgpAdt_htFree(ht);
  if(hist_elt != NULL) UTAG_FREE(hist_elt);

  ERR_RETURN;
}

/* rl_encode_colors()
   
   Given an array of color tags, represented as an array of tokpos_extension
   generate a list of color_runs that gives a run-length encoding of the color tags.
*/
static ErrCode
rl_encode_colors(color_cookie *ccookie, tokpos **posext, Dgp_Int32 npos,
				 DgpAdt_List *_tag_runs) 
{
  Dgp_State *dgp_state = ccookie->dgp_state;
  DGP_PROLOGUE("rl_encode_colors()", DGP_PERRSTATE);
	DgpAdt_List tag_runs = NULL;
	DgpAdt_ListNode retnode;
	tag_run *run = NULL;
	color_tag *tag;
	Dgp_Int32 i;

	/* nothing to encode? */
	if (npos == 0)
		NORMAL_RETURN;
  
	DGP_REPORTF0(ccookie->report, "Run-length encoding the colors.");

	/* get a list... */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&tag_runs, dgp_state));
  
	/* start the first run */
	TAGRUN_SEC_MALLOC(&run);
	run->tag = (color_tag *)posext[0]->extdata;
	run->len = 1;

	for (i = 1; i < npos; i++) {
		tag = (color_tag *)posext[i]->extdata;
		if (tag_isequal(tag, run->tag)) {
			/* this color tag is the same as the last one. */
			run->len++;
		} else {
			/* put old run in run list & start a new one */
			DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(tag_runs, run, &retnode));
			TAGRUN_SEC_MALLOC(&run);
			run->tag = tag;
			run->len = 1;
		}
	}
	/* end last run */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(tag_runs, run, &retnode));
  run = NULL;

	*_tag_runs = tag_runs;
  tag_runs = NULL;

	NORMAL_RETURN;

ERR_CLEANUP:

  if(tag_runs != NULL) DgpAdt_liFree(tag_runs);
  if(run != NULL) TAGRUN_FREE(run);

	ERR_RETURN;
}


/* This little helper function reallocates (with ccookie->realloc()) the
   buffer *buf.  It returns the new buffer in *buf and the new
   length of the buffer in *buflen.  It also calls bfNewBuffer(), which
   tells the bitfile to use the new buffer it got from ccookie->realloc(). */
static ErrCode
realloc_bitfile(color_cookie *ccookie,
		DgpAdt_BitFile f, void **buf, size_t *buflen) 
{
  DGP_PROLOGUE("realloc_bitfile()", ccookie->DGP_PERRSTATE);
  void *tmp;
  Dgp_Int32 new_buflen = (3 * (*buflen + 1)) / 2;

  DGP_CHECK_ALLOC(tmp = ccookie->DGP_REALLOC(*buf, new_buflen));

  /* Ok, change the bitfile and set the stuff that's going
     to be returned. */
  DgpAdt_bfNewBuffer(f,
	      (Dgp_UInt8 *)(*buf = tmp), 
	      (*buflen = new_buflen));
  
  NORMAL_RETURN;
}


#define bfWriteCode_realloc(code,codelen,bf) \
  do { \
     DGP_ERR_CHECK_NO_MSG(DgpAdt_bfWriteCode((code),(codelen),(bf))) { \
       /* out of space in the bitfile, so try realloc'ing */ \
       DGP_ERR_PASS_CLEANUP(realloc_bitfile(ccookie,(bf),&buf, (size_t *)&buflen)); \
       DGP_ERR_PASS_CLEANUP(DgpAdt_bfWriteCode((code),(codelen),(bf))); \
     } \
  } \
  while(0)

#define bfWriteValue_realloc(code,bf) \
  do { \
     DGP_ERR_CHECK_NO_MSG(DgpAdt_bfWriteValue((code),(bf))) { \
       /* out of space in the bitfile, so try realloc'ing */ \
       DGP_ERR_PASS_CLEANUP(realloc_bitfile(ccookie,(bf),&buf, (size_t *)&buflen)); \
       DGP_ERR_PASS_CLEANUP(DgpAdt_bfWriteValue((code),(bf))); \
     } \
  } \
  while(0)

/* 
	This is the callback function that encodes the position block extension
	the extension data should be the color tags and the extension cookie is a color_cookie.
*/
ErrCode
encodeColors(tokpos **sorted_tokens, Dgp_Int32 npos, void *extcookie,
	     void **extension, size_t *extlen) 
{
	color_cookie *ccookie = extcookie;
  Dgp_State *dgp_state = ccookie->dgp_state;
  DGP_PROLOGUE("encodeColors()", DGP_PERRSTATE);
  DgpAdt_List tag_runs = NULL;
	tag_run *run;
	Dgp_Int32 num_runs;
	DgpAdt_List unique_colors = NULL;
	DgpAdt_List unique_images = NULL;
	Dgp_Int32 num_unique_colors;
	Dgp_Int32 num_unique_images;
	Dgp_Int32 num_unique_tags;
	unique_tag *utag;
	DgpAdt_HashTable hist = NULL;
	DgpAdt_ht_elt e;
	DgpEncode_Huff *vals = NULL, *cur_Huff;
	Dgp_Int32 i, buflen;
	DgpAdt_ListNode ln;
	unique_tag dummy;
	void *buf = NULL;
	DgpAdt_BitFile f = NULL;
  Dgp_Allocators *a = &(ccookie->DGP_ALLOCATORS);


	DGP_REPORTF0(ccookie->report, "encodeColors() starts.");

  /* check small pool usage */
  DGP_SMALLPOOL_VALIDATE(unique_tag, UTAG_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(tag_run, TAGRUN_POOL_ID);

	/* run length encode color info */
	DGP_ERR_PASS_CLEANUP(rl_encode_colors(ccookie, sorted_tokens, npos, &tag_runs));
	num_runs = DGPADT_LILEN(tag_runs);

	/* find unique tags */
	DGP_ERR_PASS_CLEANUP(histogram_colors(ccookie, tag_runs, &hist, &unique_colors, &unique_images));
	num_unique_colors = DGPADT_LILEN(unique_colors);
	num_unique_images = DGPADT_LILEN(unique_images);
	num_unique_tags = num_unique_colors+num_unique_images;

	/* the Huffman encoder takes an array of Huff structs, each of which */
	/* has a field giving the number of occurrences of the piece of data */
    /* that corresponds to that Huff struct. */
	DGP_CHECK_ALLOC_CLEANUP(vals = (DgpEncode_Huff *)(a->malloc)(sizeof(DgpEncode_Huff)*num_unique_tags));

  /* OK, now we're going to fill in the array of Huff structs to pass to
     the Huffman encoder.  ADgp_Int32 the way, we'll set a field of each
     unique_tag to point to the Huff struct that corresponds to it. Since all 
	 the unique tags are also still in our DgpAdt_HashTable (ht), we can use
     the DgpAdt_HashTable to look up the Huffman codes after we've done the encoding
     (since we still have to go back and write out the Huffman codes for
     each run, we'll need to be able to look up the Huffman code for any
     of our colors in some efficient way).
	 Note that the code space is shared by colors & image indices, the code
	 generation is specified as putting the colors before the image indices when
	 breaking ties */
	i = 0;
	/* first get colors */
	DGP_REPORTF0(ccookie->report, "\nUnique colors:");	
	DGPADT_FOREACH(ln, unique_colors) {
		/* DGPADT_FOREACH unique color */
		utag = (unique_tag *)DGPADT_LIGET(ln);
		vals[i].value = utag->count;
		utag->Huff_elt = &vals[i];
		DGP_REPORTF3(ccookie->report, "(%d,%d,%d)\n", 
			utag->tag->value.rgb_color.r, utag->tag->value.rgb_color.g, utag->tag->value.rgb_color.b);
		i++;
	}
	/* get image indices */
	DGP_REPORTF0(ccookie->report, "\nUnique image indices:");	
	DGPADT_FOREACH(ln, unique_images) {
		/* DGPADT_FOREACH unique index */
		utag = (unique_tag *)DGPADT_LIGET(ln);
		vals[i].value = utag->count;
		utag->Huff_elt = &vals[i];
		if(ccookie->report){
      if(utag->tag->color_type == IMAGE_INDEX) {
				DGP_REPORTF1(ccookie->report,"%d\n", (Dgp_Int32)utag->tag->value.image_index);
      } else {
				DGP_REPORTF1(ccookie->report,"%d\n", (Dgp_Int32)utag->tag->value.samp_class->samptoknum);
      }
		}
		i++;
	}

	/* call the Huffman encoder  */
	DGP_ERR_PASS_CLEANUP(DgpEncode_huffmanCode(dgp_state,vals,num_unique_tags));
 
	/* alloc destination buffer space, take a guess on the size... */
	buflen = (size_t)(5*(num_unique_tags+num_runs));
		
	DGP_CHECK_ALLOC_CLEANUP(buf = (a->malloc)(buflen));

	/* set up the bitfile  */
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfNew(&f, buf, buflen, dgp_state));

	/* write the distinct colors we have  */
	bfWriteValue_realloc(num_unique_colors,f);
	/* write the distinct image indices we have  */
	bfWriteValue_realloc(num_unique_images,f);
 	
	/* write the pairings of colors and length of Huffman code */
	DGPADT_FOREACH(ln, unique_colors) {
		utag = (unique_tag *)DGPADT_LIGET(ln);
		bfWriteCode_realloc(utag->tag->value.rgb_color.r, BITS_PER_SAMPLE, f);
		bfWriteCode_realloc(utag->tag->value.rgb_color.g, BITS_PER_SAMPLE, f);
		bfWriteCode_realloc(utag->tag->value.rgb_color.b, BITS_PER_SAMPLE, f);
		bfWriteValue_realloc(utag->Huff_elt->codelen,f);
	}
 
	/* write the pairings of index and length of Huffman code */
	DGPADT_FOREACH(ln, unique_images) {
		utag = (unique_tag *)DGPADT_LIGET(ln);
		if(utag->tag->color_type == IMAGE_INDEX)
			bfWriteValue_realloc(utag->tag->value.image_index, f);
		else 
			bfWriteValue_realloc(utag->tag->value.samp_class->samptoknum, f);
		bfWriteValue_realloc(utag->Huff_elt->codelen,f);
	}

	/* write the number of runs  */
	bfWriteValue_realloc(num_runs, f);

	/* For each run, write the Huffman code corresponding to its color + legth of run */
	DGPADT_FOREACH (ln, tag_runs) {
		run = (tag_run *)DGPADT_LIGET(ln);
		dummy.tag = run->tag;
		/* find the unique_tag for this color in the DgpAdt_HashTable  */
		DgpAdt_htIsInTable(hist, &dummy, &e);
		cur_Huff = ((unique_tag *)e)->Huff_elt; 
		/* write the Huffman code for the color  */
		bfWriteCode_realloc(cur_Huff->code,cur_Huff->codelen,f);

		/* write the length of the run  */
		bfWriteValue_realloc(run->len,f);
	}
  
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfFlush(f));

	/* free old buffer if there is one and stick the current buffer in the cookie */
	if(ccookie->colorbuf)
		(a->free)(ccookie->colorbuf);
	ccookie->colorbuf = *extension = buf;
  buf = NULL;

	/* free up */

	/* bit file */
	DgpAdt_bfFree(f,extlen);
  f = NULL;

	/* unique colors */
	DGPADT_FOREACH (ln, unique_colors) {
	  UTAG_FREE(DGPADT_LIGET(ln));
    DGPADT_LIPUT(ln, NULL);
	}
	DgpAdt_liFree(unique_colors);

	/* unique images */
	DGPADT_FOREACH (ln, unique_images) {
		UTAG_FREE(DGPADT_LIGET(ln));
    DGPADT_LIPUT(ln, NULL);
	}
	DgpAdt_liFree(unique_images);
  unique_images = NULL;

	/* hash table */
	DgpAdt_htFree(hist);
  hist = NULL;

	/* runs  */
	DGPADT_FOREACH (ln, tag_runs) {
		TAGRUN_FREE(DGPADT_LIGET(ln));
    DGPADT_LIPUT(ln, NULL);
	}
	DgpAdt_liFree(tag_runs);
  tag_runs = NULL;

	/* huffman code stuff */
	(a->free)(vals);
  vals = NULL;
	
	DGP_REPORTF0(ccookie->report,"encodeColors() ends (ok).");

	NORMAL_RETURN;

ERR_CLEANUP:

  if(buf != NULL) (a->free)(buf);
  if(f != NULL) DgpAdt_bfFree(f,extlen);

  if(unique_colors != NULL) {
	  DGPADT_FOREACH (ln, unique_colors) {
	    if(DGPADT_LIGET(ln) != NULL) UTAG_FREE(DGPADT_LIGET(ln));
	  }
	  DgpAdt_liFree(unique_colors);
  }
  if(unique_images != NULL) {
	  DGPADT_FOREACH (ln, unique_images) {
	    if(DGPADT_LIGET(ln) != NULL) UTAG_FREE(DGPADT_LIGET(ln));
	  }
	  DgpAdt_liFree(unique_images);
  }

  if(hist) DgpAdt_htFree(hist);

  if(tag_runs) {
	  DGPADT_FOREACH (ln, tag_runs) {
		  if(DGPADT_LIGET(ln) != NULL) TAGRUN_FREE(DGPADT_LIGET(ln));
	  }
	  DgpAdt_liFree(tag_runs);
  }

  if(vals != NULL) (a->free)(vals);

  DGP_REPORTF0(ccookie->report,"encodeColors() ends (bailing!!!).");
  
  ERR_RETURN;
}
