
#include <stdio.h>
#include <stdlib.h>
#include "huff.h"
#include "bitfile.h"
#include "list.h"
#include "color-extension.h"
#include "dgp_allocators.h"
#include "mmcompress.h"
#include "encode-colors.h"
#include "PsgFix.h"

/*
        This is the callback function that encodes the position block extension
        the extension data should be the color tags and the extension cookie is
a color_cookie.
*/

/* each run in the color extension is a color + its length */
typedef struct {
        DgpImage_RGB *color;
        Dgp_Int32 len;
} tag_run;

#define TAGRUN_POOL_ID DGP_SMALLPOOL_MIN8

#define TAGRUN_SEC_MALLOC(_pptagrun) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, TAGRUN_POOL_ID, (void **)(_pptagrun)))
#define TAGRUN_FREE(_ptagrun) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, TAGRUN_POOL_ID, (void*)(_ptagrun)))

/* check if two color tags are identicals */
/*static Dgp_Bool
tag_isequal(DgpImage_RGB *c1, DgpImage_RGB *c2) {
  const Dgp_Int32 tolerance = 0;
                return ((c1->r <= c2->r + tolerance && c1->r >= c2->r - tolerance) &&
                  (c1->g <= c2->g + tolerance && c1->g >= c2->g - tolerance) &&
                  (c1->b <= c2->b + tolerance && c1->b >= c2->b - tolerance) );
} */
static Dgp_Bool
tag_isequal(DgpImage_RGB *c1, DgpImage_RGB *c2) {
  return (c1->r == c2->r &&
          c1->g == c2->g &&
          c1->b == c2->b);
}

static ErrCode
rl_encode_colors(color_cookie *ccookie, tokpos **posext, Dgp_Int32 npos,
                 DgpAdt_List *_tag_runs)
{
    Dgp_State *dgp_state = ccookie->dgp_state;
    DGP_PROLOGUE("rl_encode_colors()", DGP_PERRSTATE);
    DgpAdt_List tag_runs = NULL;
    DgpAdt_ListNode retnode;
    tag_run *run = NULL;
    DgpImage_RGB *color;
    Dgp_Int32 i;
    
    /* nothing to encode? */
    if (npos == 0)
        NORMAL_RETURN;
    
    DGP_REPORTF0(ccookie->report, "Run-length encoding the colors.");
    
    /* get a list... */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&tag_runs, dgp_state));
    
    /* start the first run */
    TAGRUN_SEC_MALLOC(&run);
    run->color = (DgpImage_RGB *) posext[0]->extdata;
    run->len = 1;
    
    for (i = 1; i < npos; i++) {
        color = (DgpImage_RGB *) posext[i]->extdata;
        if (tag_isequal(color, run->color) && run->len < MAX_RUN_LEN -1) {
            /* this color tag is the same as the last one. */
            run->len++;
        } else {
            /* put old run in run list & start a new one */
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(tag_runs, run, &retnode));
            TAGRUN_SEC_MALLOC(&run);
            run->color = color;
            run->len = 1;
        }
    }
    /* end last run */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(tag_runs, run, &retnode));
    run = NULL;
    
    *_tag_runs = tag_runs;
    tag_runs = NULL;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(tag_runs != NULL) DgpAdt_liFree(tag_runs);
    if(run != NULL) TAGRUN_FREE(run);
    
    ERR_RETURN;
}

ErrCode
encodeJbig2Colors(tokpos **sorted_tokens, Dgp_Int32 npos, void *extcookie,
             void **extension, size_t *extlen, Dgp_State *dgp_state)
{
    color_cookie *ccookie = extcookie;
    DGP_PROLOGUE("encodeColors()", DGP_PERRSTATE);
    DgpAdt_List tag_runs = NULL;
    tag_run *run;
    Dgp_Int32 num_runs;
    Dgp_UInt32 buflen=0, buflen_est=0;
    DgpAdt_ListNode ln;
    Dgp_UInt8 *buf = NULL, *bufheader=NULL, *bufdata=NULL;
    Dgp_UInt32 nval, runlen;
    Dgp_Allocators *a = &(DGP_ALLOCATORS);
    
    
    DGP_REPORTF0(ccookie->report, "encodeColors() starts.");
    
    /* check small pool usage */
    DGP_SMALLPOOL_VALIDATE(tag_run, TAGRUN_POOL_ID);
    
    /* run length encode color info */
    DGP_ERR_PASS_CLEANUP(rl_encode_colors(ccookie, sorted_tokens, npos, &tag_runs));
    num_runs = DGPADT_LILEN(tag_runs);
    /* alloc destination buffer space, take a guess on the size... */
    buflen = 0;
    buflen_est = (size_t)(num_runs * (NCOMP*COMPLEN + 3) + 6);
    DGP_CHECK_ALLOC_CLEANUP(buf = (a->malloc)(buflen_est));
    
    /* fill in buffer header */
    bufheader = buf;
    *bufheader=NCOMP;
    *(++bufheader)=COMPLEN;

    /* write run lenth data */
    bufdata = bufheader + 4;
    nval = 0;
    buflen = 6;
    DGPADT_FOREACH(ln, tag_runs) {
        run = (tag_run *)DGPADT_LIGET(ln);
        runlen = run->len;
        nval += runlen;
        DGP_ASSERT(runlen >0);
        if( run->len < MIN_RUN_LEN){
            *(++bufdata) = (Dgp_UInt8)run->len;
            buflen +=1;
        }else if(run->len < MAX_RUN_LEN){
            *(++bufdata) =0;
            *(++bufdata) = run->len >> 8 & 0xff;
            *(++bufdata) = run->len & 0xff;
            buflen +=3;
        }
        *(++bufdata) = (Dgp_UInt8)run->color->r;
        *(++bufdata) = (Dgp_UInt8)run->color->g;
        *(++bufdata) = (Dgp_UInt8)run->color->b;
        buflen +=3;
    }

    /* store NVAL field in the header */
    *(++bufheader) = (Dgp_UInt8)(nval >> 24 & 0xff);
    *(++bufheader) = (Dgp_UInt8)(nval >> 16 & 0xff);
    *(++bufheader) = (Dgp_UInt8)(nval >> 8 & 0xff);
    *(++bufheader) = (Dgp_UInt8)(nval  & 0xff);
    /* runs */
    DGPADT_FOREACH (ln, tag_runs) {
        TAGRUN_FREE(DGPADT_LIGET(ln));
        DGPADT_LIPUT(ln, NULL);
    }
    DgpAdt_liFree(tag_runs);
    tag_runs = NULL;
    if(buflen < buflen_est){
        DGP_CHECK_ALLOC_CLEANUP(buf = (a->realloc)(buf,buflen));
    }
    DGP_REPORTF0(ccookie->report,"encodeColors() ends (ok).");
    *extension = buf;
    *extlen = buflen;

    NORMAL_RETURN;

ERR_CLEANUP:

  if(buf != NULL) (a->free)(buf);
  buf = NULL;
  if(tag_runs) {
          DGPADT_FOREACH (ln, tag_runs) {
                  if(DGPADT_LIGET(ln) != NULL) TAGRUN_FREE(DGPADT_LIGET(ln));
          }
          DgpAdt_liFree(tag_runs);
  }

  DGP_REPORTF0(ccookie->report,"encodeColors() ends (bailing!!!).");

  ERR_RETURN;
}
