#include <color-extension.h>
#include "dict-util.h"
#include "token-compress.h"
#include "run2bitmap.h"
#include <tok-lossless.h>
#include <list.h>
#include <mmcompress.h>
#include <insertbox.h>
#include <simple_utils.h>
#include "dgp_utils.h"
#include "image_defs.h"
#include "PsgFix.h"

#ifndef DGP_FAXMODE
#include <libjpegw.pub>
#endif
#include <props.pub>
#include <libpixrw.pub>
#include <libscalw.pub>


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcompdoc/dict-util.c,v 1.36 2003/02/10 22:53:22 mkrocken Exp $");

/*
 * This function allocates all the token arrays to be passed to writeDictionary
 * This is done by two passes through all the tokens.
 * 
 * The first pass is to determine array sizes for the 
 * per-page singleton dictionaries and the "global" or "partial" dictionary
 *
 * 19.4.2001 / dhtesq.
 * Modifications associated with lossy/lossless unification.  When we allocate here,
 * we'll lock down enough space for the lossy tokens as well.
 * nElemGlobalPadding:  number of lossy global tokens to make room for.
 * nElemPerPagePadding:  array of number of lossy residual tokens to make room for.  One array element
 *	for each page.  Can be NULL.
 */
ErrCode
createTokensNew(Dgp_State		*dgp_state, 
				DgpAdt_List		tokOutput, 
				DgpTdc_SectionBlock *sb, 
				token			**toks,
				Dgp_Int32		*num_non_singletons, 
				Dgp_Bool		final,
				Dgp_Int32		nElemGlobalPadding,
				Dgp_Int32		*nElemPerPagePadding)
{
  DGP_PROLOGUE("createTokensNew()", DGP_PERRSTATE);

  LlTokClass llTokClass, c;
  DgpAdt_ListNode classi, ln;
  Dgp_Int32 cur_non_singleton;
  token *tok;
  Dgp_Int32 i, perPagePadding;
  DgpCC_TokInfo *tokInfo;
  DgpCC_TokInfoBlock *tBlock = sb->tBlock;
  DgpTdc_PageInfo *pages = sb->pages;
  DgpCC_PosInfo *pInfo;
  DgpTdc_PageInfo *page;

  *num_non_singletons = 0;
  *toks = NULL;

  /* We should be able to loop through just the new tokens */
  DGPADT_FOREACH(classi, tokOutput) {
	  c = (LlTokClass)DGPADT_LIGET(classi);
	  DGP_ASSERT(c != (LlTokClass)NULL);
	  if(c->members != DGP_NULLLIST){
		  /* This is a new token */
		  if(c->num_instances == 1){
			  pInfo = &(sb->pBlock->poss[(Dgp_Int32)DGPADT_LIGET(DGPADT_LIFIRST(c->members))]);
			  pages[pInfo->page].numToks++;
		  } else {
			  (*num_non_singletons)++;
		  }
	  }
  }

  /*
   * Allocate the token arrays
   */
  cur_non_singleton = 0;
  if(*num_non_singletons != 0 || nElemGlobalPadding){
    DGP_SEC_MALLOC(*toks, (token *), sizeof(token) * (*num_non_singletons + nElemGlobalPadding));
  }

  for(i = 0; i < sb->numPages; i++){
    pages[i].toks = NULL;
	perPagePadding = nElemPerPagePadding ? nElemPerPagePadding[i]:0;
    if(pages[i].numToks != 0 || perPagePadding){
		DGP_SEC_MALLOC(pages[i].toks, (token *), sizeof(token) * (pages[i].numToks + perPagePadding));
    }
    pages[i].numToksAlloced = pages[i].numToks;
    pages[i].numToks = 0;
  }

  classi = DGPADT_LIFIRST(tokOutput);
  while(classi != DGP_NULLLISTNODE){
    llTokClass = (LlTokClass)DGPADT_LIGET(classi);

    tokInfo = (DgpCC_TokInfo *)(((Dgp_Int32)llTokClass->userdata) + tBlock->toks);
    if(llTokClass->members != DGP_NULLLIST){
      /* New token */
      if(llTokClass->num_instances == 1){
        pInfo = &(sb->pBlock->poss[(Dgp_Int32)DGPADT_LIGET(DGPADT_LIFIRST(llTokClass->members))]);
        page = &(pages[pInfo->page]);
        tok = &(page->toks[page->numToks]);
        page->numToks++;
        DGP_ASSERT(page->numToks <= page->numToksAlloced);
      } else {
        tok = &((*toks)[cur_non_singleton]);
		cur_non_singleton++;
        DGP_ASSERT(cur_non_singleton <= *num_non_singletons);
        tokInfo->token = tok;
      }

      /* Actually add it to the toks array */
      tok->width = (Dgp_UInt16)llTokClass->width;
      tok->height = (Dgp_UInt16)llTokClass->height;
      tok->uses = llTokClass->num_instances;
      tok->committed = DGP_FALSE;
      tok->extdata = NULL;
      tok->longsperline = (llTokClass->width + DGP_BITS_UINT32 - 1)/DGP_BITS_UINT32;
      DGP_ERR_PASS_CLEANUP(RunsToUnframedBitmap(	&(tok->bits), 
													llTokClass->runs, 
													llTokClass->num_runs, 
													llTokClass->height,
													tok->longsperline,
													0, 
													dgp_state));

      /* go through all the associated positions and set their */
      /* tok pointers; */
      ln = DGPADT_LIFIRST(llTokClass->members);
      while(ln != DGP_NULLLISTNODE){
        pInfo = &sb->pBlock->poss[(Dgp_Int32)DGPADT_LIGET(ln)];

        /* TODO */
        /* Might want to add a check here so the shapes of positions */
        /* that are sampled images are not added */
        pages[pInfo->page].boxes[pInfo->posOffset].tok = tok;

        /* If we merge lossy and lossless this will have to change */
        DgpAdt_liRemAndNext(ln, &ln);
      }

      /* If we merge lossy and lossless this will have to change */
      DgpAdt_liFree(llTokClass->members);
      llTokClass->members = DGP_NULLLIST;
      if(final || (llTokClass->num_instances == 1)){
        DgpLlTokFreeClass(llTokClass, dgp_state);
        DgpAdt_liRemAndNext(classi, &classi);
      } else {
        classi = DGPADT_LINEXT(classi);
      }
    } else {
      /*
       * This is an token that has already been written out
       * The token is setup and the postions are set correctly
       *
       * In the lossy case this will have to handled a bit differently
       */
      classi = DGPADT_LINEXT(classi);
    }
  }
  
  if(!final){
    /*
    * This is an interesting block of code
    *
    * The goal is to save all the needed tokInfo structs
    * by sticking them at the begining of the array located in 
    * the section block.
    *
    * We have the tokInfo structs in two places 
    * in their array off of the section block and in 
    * the userdata of the each of the classes in tokOutput
    *
    * tokOuput only contains classes that should be saved
    * the array of the section block contains all the class for
    * this section
    *
    * numOldToks starts being the number of saved tokInfo structs from 
    * the last partial endSection
    *
    * The positions list in the tokInfo struct becomes a flag, if it is 
    * DGP_NULLLIST then the tokInfo has already been saved, or it was discarded
    *
    * So we start by finding the first tokInfo that can be written over, and
    * we save everthing before that. (setting their positions "flag" to reflect
    * this
    *
    * Then we go through the tokOutput list and "save" the tokInfo's that hasn't
    * been saved.  Notice in this case we are "moving" the tokInfo so we must update
    * the userdata pointer in the class.
    *
    * After saving a tokInfo we need to find the next tokInfo that can be written over
    * 
    */
    while(tBlock->toks[sb->numOldToks].token != NULL){
      sb->numOldToks++;
    }
    DGPADT_FOREACH(ln, tokOutput){
      c = DGPADT_LIGET(ln);
      DGP_ASSERT(c != NULL);
      
      DGP_ASSERT(tBlock->toks[(Dgp_Int32)c->userdata].token != NULL);
      
      if(((Dgp_Int32)c->userdata) > sb->numOldToks){
        tokInfo = &(tBlock->toks[(Dgp_Int32)c->userdata]);
        tBlock->toks[sb->numOldToks] = *tokInfo;
        c->userdata = (void *)sb->numOldToks;
        sb->numOldToks++;
        while(tBlock->toks[sb->numOldToks].token != NULL){
          sb->numOldToks++;
        }
      }
    }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


ErrCode
makeSubIFDs(Dgp_State *dgp_state, tokpos* position, Dgp_Int32 *num_positions, 
          color_tag *colors, DgpAdt_List subIFDims)
{
  DGP_PROLOGUE("makeSubIFDs()", DGP_PERRSTATE);

  Dgp_Int32 i;
  DgpImage_Packed tokenImage;
  token *curTok;

  tokpos framePos = {NULL, NULL, cFrameBits,cFrameBits};
  Dgp_Int32 offset;
  samptokClass image;
  ImSubIFD *subIFD = NULL;

  for(i=0; i < *num_positions; i++) {
    while((i < *num_positions) &&
      (colors[i].color_type == SAMP_CLASS) &&
      !(colors[i].value.samp_class->flags & DGPTDC_SAMPTOK_IS_TOKEN)){
      image = colors[i].value.samp_class;
      curTok = position[i].tok;

       /* This is a subIFD image */
      SUBIFD_SEC_MALLOC(&subIFD);

      subIFD->xpos = position[i].xpos;
      subIFD->ypos = position[i].ypos;
      subIFD->samptok = image;        
      subIFD->mask = NULL;
      subIFD->buf = NULL;
      subIFD->size = 0;
      subIFD->width = image->width;
      subIFD->height = image->height;
        
      if((colors[i].value.samp_class->flags & SAMPTOK_IS_RECT) != 0){
  
        /* Need to make sure we take into account pixr alignment issues */
        DGP_ERR_PASS_CLEANUP(
          pimNewFrameAligned(dgp_state, curTok->width,
            curTok->height, &offset, &tokenImage));
        framePos.xpos = cFrameBits;
        framePos.tok = NULL;
        framePos.xpos += (Dgp_Int16)offset;
        subIFD->maskOffset = offset;
        InsertBoxPacked(tokenImage, &framePos, curTok, DGP_FALSE);
        subIFD->mask = tokenImage;
      }

      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(subIFDims, subIFD, NULL));
      subIFD = NULL;
      
      /* remove pos and color, carefully */
      /* this screws up the positions order, */
      /* but the backend resorts them  */
      (*num_positions)--;
      position[i] = position[*num_positions];
      colors[i] = colors[*num_positions];
      
      /*
       * We need to do something with the token for this 
       * image 
       *
       * The token might be a singleton or it might be
       * a global token
       *
       * In the first case it has already been written to the 
       * dictionary.  (Pending a smarter createTokensNew)
       * 
       * In the second case it hasn't been written yet
       * If it is marked commited then it shouldn't be written
       * (lets see if that works.)
       */
      /*curTok->committed = DGP_TRUE; */
      
    }  

  }

  NORMAL_RETURN;

ERR_CLEANUP:

  if(subIFD != NULL) SUBIFD_FREE(subIFD);

  ERR_RETURN;
}

typedef struct {
  Dgp_State *dgp_state;
  Dgp_Int8 *buf;
  Dgp_UInt32 size;
  Dgp_UInt32 alloced;
} jcookie;

static ErrCode 
jpeg_Compress_callback(IPWState *ipwstate, 
                              void *cookie, 
                              UInt8 *buf, 
                              UInt32 nbytes,
                              UInt32 *nwritten)
{
  jcookie *jc = (jcookie *)cookie;
  DGP_PROLOGUE("jpeg_Compress_callback()", jc->DGP_PERRSTATE);
  Dgp_UInt32 size;
  Dgp_Int8 *tmpptr;

  (void)ipwstate;

  size = jc->size + nbytes;
  if(size > jc->alloced){
    jc->alloced = (size * 3) / 2;
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)(jc->DGP_REALLOC)(jc->buf, jc->alloced));
    jc->buf = tmpptr;
  }

  memcpy(jc->buf + jc->size, buf, nbytes);
  jc->size += nbytes;

  *nwritten = nbytes;

  NORMAL_RETURN;
}

#ifndef DGP_FAXMODE
ErrCode
compressRgbImageJpeg(Dgp_State *dgp_state,
                     DgpImage_ColorRGB im, 
                     Dgp_Int32 xres, 
                     Dgp_Int32 yres, 
                     Dgp_Float32 scaleFactor,
                     Dgp_Int32 quality,
                     void **_cinfo,
                     Dgp_Int8 **buf, 
                     Dgp_Int32 *size)
{
  DGP_PROLOGUE("compressRgbImageJpeg()", DGP_PERRSTATE);

  jcookie jc;
  PIXR *pixr = (PIXR *)NULL;
  PIXR *tempPixr = NULL;
  UInt8 ***userBuf = NULL;
  Dgp_Int32 width, height;


  width = DGPIMG_WIDTH(im);
  height = DGPIMG_HEIGHT(im);

  /* set up compression cookie */
  jc.dgp_state = dgp_state;
  jc.size = 0;
  jc.alloced = width * height / 4;
  DGP_SEC_MALLOC(jc.buf, (Dgp_Int8 *), jc.alloced);


  if(scaleFactor != 1.00) {

    /* scale image dimensions */
    width = SCALE_DIMENSION(width, scaleFactor);
    height = SCALE_DIMENSION(height, scaleFactor);

    /* scale the resolution as well */
    xres = SCALE_RESOLUTION(xres, scaleFactor);
    yres = SCALE_RESOLUTION(yres, scaleFactor);

    /* in the scaling case, we need to convert to pixr's */
    DGP_ERR_PASS_CLEANUP(rgbimtorgbpixr(dgp_state, im, &pixr));

    /* scale rgb pixr */
    DGP_ERR_PASS_CLEANUP(
      w_scaleLinearSize(DGP_PIPWSTATE, pixr, width,
        height, &tempPixr));
    
    /* destroy unscaled pixr */
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &pixr));
    pixr = tempPixr;
    tempPixr = NULL;
    
    /* set resolution (scaled if necessary) */
    DGP_ERR_PASS_CLEANUP(
      w_pixrSetResolutionDPI(DGP_PIPWSTATE,pixr,
        Int32_TO_ScaledInt1616(xres), Int32_TO_ScaledInt1616(yres)));

    /* compress the scaled pixr */
#if !defined DGP_FAXMODE && !defined PSG
    DGP_ERR_PASS_CLEANUP(
      w_jCompress_callback(DGP_PIPWSTATE, pixr, jpeg_Compress_callback, 
        (void *)&jc, quality, JC_DEFAULT));
#endif

    /* ditch the pixr */
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &pixr));

  } else {

    Dgp_Int32 lperMCU, i;
    DgpImage_RGB **rows = im->data;
    CompressInfo *cinfo = (CompressInfo *)*_cinfo;

    
    /* create a compression info struct if we haven't yet */
    if(cinfo == NULL) {
      DGP_SEC_MALLOC(cinfo, (void *), sizeof(CompressInfo));
      *_cinfo = (void *)cinfo;
    }

    /* If no scaling is necessary, we don't need to convert to pixr's, 
       init ipcore jpeg compressor for interleaved rgb data. */
	/* 19.3.2001 / dhtesq. */
	/* Changed the colorspace options from JC_DEFAULT to itulab.  This is the only Digipaper call to w_jcInit. */
#if !defined DGP_FAXMODE && !defined PSG
    DGP_ERR_PASS_CLEANUP(w_jcInit(DGP_PIPWSTATE, cinfo, 
      width, height, 3, quality, JC_ITULAB, xres, yres, 
      jpeg_Compress_callback, (void *)&jc, NULL, JMEMORG_BIP));
#endif

    /* How many lines at a time are we doing...? */
    DGP_ERR_PASS_CLEANUP(w_jcGetImageLinesPerMCU(DGP_PIPWSTATE, cinfo, 
      &lperMCU));

    /* create a a buffer of row pointers */
    DGP_ERR_PASS_CLEANUP(w_jcAllocMCURowBuffer(DGP_PIPWSTATE, cinfo,
      cFalse, &userBuf));

    /* do all the rows except for the last few */
    while(height > lperMCU) {

      for(i = 0; i < lperMCU; i++) {
        userBuf[0][i] = (UInt8 *)*(rows++);
      }

      DGP_ERR_PASS_CLEANUP(w_jcMCURow(DGP_PIPWSTATE, cinfo, 
        userBuf));

      height -= lperMCU;

    }

    /* do the remaining lines (if any) */
    if(height > 0) {

      for(i = 0; i < height; i++) {
        userBuf[0][i] = (UInt8 *)*(rows++);
      }

      DGP_ERR_PASS_CLEANUP(w_jcMCURow(DGP_PIPWSTATE, cinfo, 
        userBuf));

    }

    /* free up row pointers */
	  DGP_ERR_PASS_CLEANUP(w_jcMCURowFree(DGP_PIPWSTATE, cinfo,
      cFalse, userBuf));
    userBuf = NULL;

    /* cleanup up jpeg state */
    DGP_ERR_PASS_CLEANUP(w_jcTerm(DGP_PIPWSTATE, cinfo));

  }

  /* shrink wrap the jpeg data */
  DGP_SEC_REALLOC(jc.buf, (Dgp_Int8 *), jc.size);

  /* return compressed info */
  *size = jc.size;
  *buf = jc.buf;
  jc.buf = NULL;


  NORMAL_RETURN;

ERR_CLEANUP:

  if(userBuf != NULL) w_jcMCURowFree(DGP_PIPWSTATE, (CompressInfo *)*_cinfo, cFalse, userBuf);
  if(jc.buf != NULL) DGP_FREE(jc.buf);
  if(pixr != NULL) w_destroyPixr(DGP_PIPWSTATE, &pixr);
  if(tempPixr != NULL) w_destroyPixr(DGP_PIPWSTATE, &tempPixr);

  ERR_RETURN;
}
#endif
