#ifndef __tokenCC_h__
#define __tokenCC_h__

/****************************

  This is the common interface that is used by the lossless token
  compression, and hopefully the lossy tokenizer will be converted to
  using this as well.

  It was developed to handle token pointers in positions.

  tokenAddCC is used to add a CC to the lossless(lossy)
  tokenizer.  The structures need to be used in combination with the 
  output of the tokenizer after a tokFinal, or tokPartial 
  (see tok-lossless.h or tok-lossy.h).
  
  Currently, the tokenAddCC function is used in the following places: 
  1. For monochrome lossless pages GetBoxes is called and it in turn
  calls tokenAddCC for each CC it finds.  
  2. For color lossless pages libcompdoc/tdoc_endPage gets a list of cc's from
  the colorCC finder and calls tokenAddCC for each of these.

  The structures are used with the token output in createTokensNew 
  (libcompdoc/dict-util.c).  

  Here is what is going on:
   * Each class in the tokenizer has two pieces of "user info": one is a
  a void *userdata, the other is a DgpAdt_List members.  
   * When tokenAddCC is called it puts a "TokInfo" in the userdata field, adds 
  the value of rTok.userdata to the members list for each rTok that matches that
  class.  The TokInfo is in quotes because it actually uses an offset into the 
  TokInfoBlock.toks array.
   * The "TokInfo" associated with each class is an accelerator that isn't always
  valid.  It contains a pointer to the token associated with the class.  In the
  current implementation (libcompdoc), a class only gets an associated token 
  after tokPartial or tokFinal is called.  Therefore the tokInfo is only useful
  if the a tokPartial was called the token* was set in the tokInfo, and the class 
  was given back to the tokenizer.
   * Also in the current implementation (libcompdoc and cc2box.c) the value of 
  rTok.userdata is a PosInfo offset. 

  You can look at libcompdoc/dict-util.c/createTokensNew for an example of what
  to do with the classes, pBlock, and tBlock after tokPartial/ tokFinal is called.
   * The basic idea is to go through all classes and make a token for each class.  
  At the same time go through the members of each class (using each value
  as an offset into the pBlock), and from that PosInfo determine which array of 
  tokpos* to use, and use the posOffset in that PosInfo to get the tokpos.  Then
  set the token* of that tokpos.
	 
  Notes:  the ownership of the members list is currently a bit weird.  Currently 
  the lossless tokenizer doesn't own it.  This will change in order to sync this 
  with the lossy tokenzier.

*****************************/

#include <tok-lossless.h>
#include <mmcompress.h>
#include "dgp_utils.h"
#include <dgp_state.h>

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct {
	token *token;
} TokInfo;

typedef struct {
	int num;
	int numAlloced;
	TokInfo *toks;
} TokInfoBlock;

typedef struct {
	int posOffset;
	int page;
} PosInfo;

typedef struct {
	int num;
	int numAlloced;
	PosInfo *poss;
} PosInfoBlock;

#define initPosBlock_err(args) \
	DGP_ERR_PASS_CLEANUP(initPosBlock##args)

ErrCode
initPosBlock(Dgp_State *dgp_state, PosInfoBlock **block, int num);

void
cleanupPosBlock(Dgp_State *dgp_state, PosInfoBlock *block);

#define initTokBlock_err(args) \
	DGP_ERR_PASS_CLEANUP(initTokBlock##args)

ErrCode 
initTokBlock(Dgp_State *dgp_state, TokInfoBlock **block, int numToks);

void
cleanupTokBlock(Dgp_State *dgp_state, TokInfoBlock *block);

#define tokenAddCC_err(args) \
	DGP_ERR_PASS_CLEANUP(tokenAddCC##args)

ErrCode
tokenAddCC(Dgp_State *dgp_state, TokState tokstate, TokInfoBlock *tBlock, LlRunToken *rTok,
           token **tok);

#ifdef __cplusplus
}
#endif

#endif
