/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#include <libiow.pub>
#include <libpixrw.pub>

#include <math.h>
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "santa_halftone.h"
#include "silx_decode_priv.h"
#include "silx_decode.h"

/*********************************************************************/



static Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType,  Dgp_Int8 *outExt);

/*************************************************************************
*
*  Function:  Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType, Dgp_Int8 *outExt)
*
*
*  Converts character-based orientation instructions to internal santa orientation.
*
*	Arguments:
*      o  pType is output type character array.
*      o  pOutType is the output type, internal representation.
*      o  outExt is the output extension character array, which gets filled on successful completion.
*         There should be enough space in it to hold four characters.
*
*	Return value:
*		o  Boolean true iff one of the valid output extensions has been passed:  pnm, tif, or bmp.
*
*	Modification history:
*      o  None.
*  
*/

static Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType, Dgp_Int8 *outExt) {
  if(!strcmp(pType, "PNM") || !strcmp(pType, "pnm")) {
      *pOutType = PNM;
      strcpy(outExt, "pnm");
      return DGP_TRUE;
  } else if(!strcmp(pType, "TIF") || !strcmp(pType, "tif")) {
      *pOutType = TIFF_G4;
      strcpy(outExt, "tif");
      return DGP_TRUE;
  } else if(!strcmp(pType, "BMP") || !strcmp(pType, "bmp")) {
      *pOutType = FF_BMP;
      strcpy(outExt, "bmp");
      return DGP_TRUE;
  } else {
    return DGP_FALSE;
  }
}
/*************************************************************************/


/************************************************************************
 * Just converts clockwise rotation, as a multiple of pi/2, to internal
 * orientation.  You can't get inverted orientations this way (e.g. RU)
 * but those haven't been supported in a long time anyway. / dhtesq.
 */
static DgpCoord_Orientation
rotationToOrientation(Dgp_Int32 rot)
{
	DgpCoord_Orientation o;

	rot %= 4;
	rot = rot < 0 ? rot + 4:rot;

	switch(rot){
	case 0:
		o = ORIENT_RD;
		break;
	case 1:
		o = ORIENT_UR;
		break;
	case 2:
		o = ORIENT_LU;
		break;
	case 3:
		o = ORIENT_DL;
		break;
	default:
		/* Should never get here... */
		o = ORIENT_RD;
		break;
	}

	return o;
}


/*********************************************************************************
 *
 * Decode doucment information.
 *
 * @param *dgp_state       pointer of Dgp_State
 * @param  *_sOpt,         pointer of SilxDecCookie.
 * @param  *fp             stream pointer, Default is FILE stream.
 *                         If you use other streams, you have to set the uds in SilxDecCookie to handle it.
 * @param  *numPages       returned number of pages in the document
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */

ErrCode silx_decDocumentHeader(
    Dgp_State       *dgp_state, 
    SilxDecCookie   _sOpt,
    void            *fp,
    Dgp_Int32       *numPages)
{
	DGP_PROLOGUE("slix_decDocumentHeader()", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;

	santaDoc *pdoc = NULL;

    DGP_ERR_GEN(!sOpt, cErrInvalidParm);
    DGP_ERR_GEN(!fp, cErrInvalidParm);

    sOpt->docUds.fp = fp;

    DGP_ERR_PASS_CLEANUPF0(santaOpenDocumentEx(&(sOpt->docUds), 1 << (DGP_BITS_UINT32 - 2), 
					                           santaDocOptsDefault,
					                           dgp_state,
					                           &pdoc),
 			   "santaUncompAll: Cannot open document\n");
    sOpt->pDoc = pdoc;
    *numPages = pdoc->npages;
    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}

/*********************************************************************************
 *
 * Decode one page information. Must run silx_decDucumentHeader first. Put NULL if
 * you don't want to return that value.
 *
 * @param *dgp_state       pointer of Dgp_State
 * @param  *_sOpt,         pointer of SilxDecCookie.
 * @param  pageNum         which page to decode
 * @param  *width          returned page width
 * @param  *height         returned page height
 * @param  *pagetype       returned page type (SILXPAGEMONO, SLIXPAGERGB, SLIXPAGECMYK)
 * @param  *res            returned page resolution (in dpi)
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
ErrCode silx_decPageHeader(
    Dgp_State *dgp_state, 
    SilxDecCookie _sOpt,
    Dgp_Int32 pageNum, 
    Dgp_Int32 *width,
    Dgp_Int32 *height,
    Dgp_Int32 *pagetype, 
    Dgp_Float64 *res)
{                              

	DGP_PROLOGUE("slix_decPageHeader()", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;
    santaPage *ppage;
    santaDoc *pdoc = sOpt->pDoc;

    DGP_ERR_GEN(!pdoc, cErrInvalidParm);
    DGP_ERR_PASS_CLEANUPF1(santaOpenPage(pdoc, pageNum, dgp_state, &ppage),
                            "I couldn't open page %d", pageNum);

    sOpt->pPage = ppage;
    if(width) *width = ppage->width;
    if(height) *height = ppage->height;
    if(pagetype){
        if (sOpt->force){
            if (sOpt->forceType == santaPageColor){
                *pagetype = SILXPAGERGB;
            }
            else if (sOpt->forceType == santaPageMono){
                *pagetype = SILXPAGEMONOPACKED;
            }
        } else if(sOpt->do_cmyk && (ppage->page_type == santaPageColor)){
            *pagetype = SILXPAGECMYK;
        } else { 
            if (ppage->page_type == santaPageColor){
               *pagetype = SILXPAGERGB;
            }
            else if (ppage->page_type == santaPageMono){
                *pagetype = SILXPAGEMONOPACKED;
            }
        }
     }
     if(res) *res = DgpCoord_float(ppage->res.xRes);

    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}

/*********************************************************************************
 *
 * Decode one page according to the parameters set by silx_setDecState. Must do 
 * silx_decDocumentHeader and silx_decPageHeader first.
 *
 * @param *dgp_state       pointer of Dgp_State
 * @param  *_sOpt,         pointer of SilxDecCookie.
 * @param  **image         output image buffer. Only valid if outToBuffer is set to TRUE in 
 *                         sOpt. Could be of type Dgp_ImageColorRGB / Dgp_ImagePacked / Dgp_ImageGray[4],
 *                         depending on the page type of  SLIXPAGERGB / SILXPAGEMONO / SLIXPAGECMYK.
 *                         Set to NULL is you don't want to return buffer.
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
ErrCode silx_decPage(
Dgp_State *dgp_state, 
SilxDecCookie _sOpt, 
void **image) 
{
	DGP_PROLOGUE("slix_decPage()", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;
	Dgp_Int32       pgNum;

	void *out = NULL, *binaryOut = NULL;
	void * dst = NULL;
    void *cmyk_img[4];
    void *cmyk_dst[4];
	Dgp_Int32 dstBpl = -1;
	Dgp_Int32 layerOptions = DGP_SHOWALL_MASK;

	Dgp_Int8        outName[255];
	Dgp_Int32       numDig;
	Dgp_Int8        *prefix = NULL;
	santaDoc * pdoc = NULL;
	santaPage * ppage = NULL;
	Dgp_Int32 x = 0, y = 0, w = 0, h = 0, i;
	Dgp_Int32 nw = 0, nh = 0;
	Dgp_UInt16 rgbImageOpts;
    Dgp_Float32 scale_factor;
	Dgp_Int32 renderedWidth=0, renderedHeight=0;
	santaInputType inputtype = -1;
    FILE            *outgoing[4];

    DGP_ERR_GEN(!sOpt, cErrInvalidParm);
    DGP_ERR_GEN(!(sOpt->pPage), cErrInvalidParm);
    DGP_ERR_GEN(!(sOpt->pDoc), cErrInvalidParm);
    DGP_ERR_GEN( sOpt->outToBuffer && !image, cErrInvalidParm);
    pdoc = sOpt->pDoc;
    ppage = sOpt->pPage;

    pgNum = ppage->page_num;
	numDig = (Dgp_Int32)ceil(log10(pdoc->npages));
    for(i = 0; i < (Dgp_Int32)(sizeof(cmyk_img)/sizeof(cmyk_img[0])); i++){
        cmyk_img[i] = NULL;
        cmyk_dst[i] = NULL;
        outgoing[i] = NULL;
    }
    if(sOpt->outToFile){
        if(sOpt->outFileName){
            prefix = sOpt->outFileName;
            if(sOpt->outtype == TIFF_G4){
                /* Multi-page output... */
                sprintf(outName, "%s.%s", prefix, sOpt->ext);
            } else if(sOpt->do_cmyk){
                /* Form will be foo.n.[c | m | y | k].pnm ...but hold off on the stuff after the page number. */
                sprintf(outName, "%s.%0*ld.", prefix, (int)numDig, pgNum);
            } else {
                sprintf(outName, "%s.%0*ld.%s", prefix, (int)numDig, pgNum, sOpt->ext);
            }
        }
        /* For stripwise decodes, we're going to open the outgoing files here and let santa write to them */
        /* as it fills the strips. */
        if(sOpt->stripwise){
            if(inputtype == santaInputCMYKAll){
                Dgp_Int8 fnm[256];
                Dgp_Int8 chars_cmyk[5] = "cmyk";
                for(i = 0; i < 4; i++){
                    if( sOpt->outToStdout != DGP_TRUE) 
                        sprintf(fnm, "%s%c.%s", outName, chars_cmyk[i], sOpt->ext);
                    DGP_CHECK_NULL_CLEANUP(outgoing[i] = sOpt->outToStdout ? stdout:fopen(fnm, "wb"));
                }
            } else {
                DGP_CHECK_NULL_CLEANUP(outgoing[0] = sOpt->outToStdout ? stdout:fopen(outName, "wb"));
            }
        }
    }
    
    if(DgpCoord_g(sOpt->output_res.xRes, ppage->res.xRes)){
        ppage->dst_res = ppage->res;
        sOpt->output_res = ppage->res;
        ppage->scale_factor = 1.0;
        ppage->scale_me = DGP_FALSE;
        sOpt->scale_me = DGP_FALSE;
		DGP_WARNF2("Page %d is only at %d dpi;  I can't upscale.", pgNum, (Dgp_Int32)DgpCoord_float(ppage->res.xRes));
	} else if(DgpCoord_equal(sOpt->output_res.xRes, ppage->res.xRes)){
        ppage->dst_res = ppage->res;
        sOpt->output_res = ppage->res;
        ppage->scale_factor = 1.0;
        ppage->scale_me = DGP_FALSE;
        sOpt->scale_me = DGP_FALSE;
		/*DGP_WARNF1("Requested resolution (%d dpi) is page resolution;  not scaling...", (Dgp_Int32)DgpCoord_float(ppage->res.xRes)); */
    } else {
        if((DgpCoord_float(sOpt->output_res.xRes) <= 0) || (DgpCoord_float(sOpt->output_res.xRes) <= 0)){
            ppage->dst_res = ppage->res;
            sOpt->output_res = ppage->res;
            ppage->scale_factor = 1.0;
            ppage->scale_me = DGP_FALSE;
            sOpt->scale_me = DGP_FALSE;
        } else {
            ppage->dst_res = sOpt->output_res;
            ppage->scale_factor = DgpCoord_float(DgpCoord_div(ppage->dst_res.xRes, ppage->res.xRes));
            ppage->scale_me = DGP_TRUE;
            sOpt->scale_me = DGP_TRUE;
        }
    }


    x = 0;
    y = 0;
    w = ppage->width;
    /* Note as currently constructed, santa_uncomp_all has to allocate a full page buffer to take advantage */
    /* of ipcore's file output functions.   When outputting tiff, setting stripSize will have no first-order  */
    /* effect on memory usage */
    /* and will certainly hurt performance. / dhtesq. */
    /*h = sOpt->stripSize; */
    h = ppage->height;

    if(sOpt->stripwise == DGP_FALSE){
        nw = (DgpCoord_isRotated(sOpt->o, ORIENT_RD) ? h : w);
        nh = (DgpCoord_isRotated(sOpt->o, ORIENT_RD) ? w : h);
    } else {
        nw = (DgpCoord_isRotated(sOpt->o, ORIENT_RD) ? h : w);
        nh = (DgpCoord_isRotated(sOpt->o, ORIENT_RD) ? w : h);
    }

    if((sOpt->stripSize <= 0) || (sOpt->stripSize > nh)){
        sOpt->stripSize = nh;
    }
    if(sOpt->stripwise){
        nh = sOpt->stripSize;
    }

    scale_factor = DgpCoord_float(DgpCoord_div(sOpt->output_res.xRes, ppage->res.xRes));
    nw = (Dgp_Int32)(nw * scale_factor);
    nh = (Dgp_Int32)(nh * scale_factor);

    if (sOpt->force){
        if (sOpt->forceType == santaPageColor){
            inputtype = santaInputRGB;
        }
        else if (sOpt->forceType == santaPageMono){
            inputtype = santaInputMonoPacked;
        }
    } else if(sOpt->do_cmyk && (ppage->page_type == santaPageColor)){
        inputtype = santaInputCMYKAll;
    } else { 
        if (ppage->page_type == santaPageColor){
            inputtype = santaInputRGB;
        }
        else if (ppage->page_type == santaPageMono){
            inputtype = santaInputMonoPacked;
        }
    }

    if(inputtype == santaInputRGB){
		rgbImageOpts = DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR;
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, rgbImageOpts, nw, nh, &out));
        
        dst = (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
        if (nh != 0){
            dstBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 1) - 
                (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
        }
        else {
            dstBpl = 0;
        }
    }
    else if(inputtype == santaInputMonoPacked){
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_PACKED, nw, nh, &out));
        dst = (void *)DGPIMG_GETROW((DgpImage_Packed)out, 0);
        dstBpl = DGPIMG_GETROWWIDTH((DgpImage_Packed)out);
    }
    else if(inputtype == santaInputCMYKAll){
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[0]));
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[1]));
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[2]));
        DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[3]));
        cmyk_dst[0] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[0], 0);
        cmyk_dst[1] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[1], 0);
        cmyk_dst[2] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[2], 0);
        cmyk_dst[3] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[3], 0);
        dst = cmyk_dst;
        out = cmyk_img;
        if (nh != 0){
            dstBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)cmyk_img[0], 0, 1) - 
                (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)cmyk_img[0], 0, 0);
        }
        else {
            dstBpl = 0;
        }
    }


	/* layerOptions... */
	if(sOpt->jbig2_only ^ sOpt->images_only){
		layerOptions = (!sOpt->jbig2_only << DGP_SHOWIMAGES_BITSHIFT) | (!sOpt->images_only << DGP_SHOWJBIG2_BITSHIFT);
	}

    DGP_ERR_PASS_CLEANUP(santaRenderInStrips(   pdoc, 
                                                ppage, 
                                                sOpt->o, 
                                                sOpt->stripSize, 
                                                x, y, 
                                                w, h,
                                                santaJobOptsDefault, 
                                                inputtype,
                                                dstBpl, 
                                                dst, 
                                                out, 
                                                DGP_FALSE,		/* Is buffer cleared? */
                                                dgp_state, 
                                                outgoing,
                                                (Dgp_Bool)(sOpt->stripwise || (Dgp_Bool)((!sOpt->outToFile) && (sOpt->stripSize < nh))),
                                                layerOptions));
    
    if(sOpt->outToFile){
        /* Output image is only monochrome if we haven't been asked to scale. */
        if(inputtype == santaInputMonoPacked){
			/* If forcing mono output on a color document, or if scaling to a new resolution, */
			/* we have to binarize santa's output. */
			binaryOut = out;
            if(!sOpt->outToStdout){
                handleMonoPage( sOpt, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->res.yRes),  
                                (DgpImage_Packed)binaryOut, 
                                dgp_state,
                                &(sOpt->tiff),
                                pgNum == sOpt->startPageNum? cTrue : cFalse,
                                pgNum == (sOpt->stopPageNum -1)? cTrue : cFalse);
			    renderedWidth = DGPIMG_WIDTH((DgpImage_Packed)binaryOut);
			    renderedHeight = DGPIMG_HEIGHT((DgpImage_Packed)binaryOut);
            }
			DGPIMAGE_FREE((DgpImage_Packed)binaryOut);
            out = NULL;
        }
        else if(inputtype == santaInputRGB){
            if(!sOpt->outToStdout){
                handleColorPage(sOpt, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->dst_res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->dst_res.yRes),  
                                (DgpImage_ColorRGB)out, 
                                dgp_state,
                                &(sOpt->tiff),
                                pgNum == sOpt->startPageNum ? cTrue : cFalse, 
                                pgNum == (sOpt->stopPageNum - 1) ? cTrue : cFalse);
			    renderedWidth = DGPIMG_WIDTH((DgpImage_ColorRGB)out);
			    renderedHeight = DGPIMG_HEIGHT((DgpImage_ColorRGB)out);
            }
            DGPIMAGE_FREE((DgpImage_ColorRGB)out);
            out = NULL;
        } else if(inputtype == santaInputCMYKAll){
            if(!sOpt->outToStdout){
                handleCMYKPage( sOpt, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->res.yRes),  
                                cmyk_img,
                                dgp_state,
                                &(sOpt->tiff),
                                pgNum == sOpt->startPageNum? cTrue : cFalse, 
                                pgNum == (sOpt->stopPageNum - 1) ? cTrue : cFalse);
                sprintf(outName, "%s.%0*ld.[c|m|y|k].%s", prefix, (int)numDig, pgNum, sOpt->ext);
            }
            DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[0]);
            DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[1]);
            DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[2]);
            DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[3]);
            out = NULL;
    }
  }

	DGP_REPORTF5(sOpt->verbose && sOpt->outToFile,
		"page %d, %d by %d, %d xspi, %d yspi", 
		(int)pgNum, (int)renderedWidth, (int)renderedHeight, 
		(int)DgpCoord_float(sOpt->output_res.xRes),
		(int)DgpCoord_float(sOpt->output_res.yRes));
	DGP_REPORTF1(sOpt->verbose && sOpt->outToFile,
		"%s written.", outName);
	DGP_REPORTF5(sOpt->verbose && !sOpt->outToFile,
		"page %d, %d by %d, %d xspi, %d yspi",
		(int)pgNum, (int)ppage->width, (int)ppage->height, 
		(int)DgpCoord_float(ppage->res.xRes),
		(int)DgpCoord_float(ppage->res.yRes));

    DGP_ERR_PASS_CLEANUP(santaClosePage(pdoc, &ppage, dgp_state));
    sOpt->pPage = NULL;
    if(image && out != NULL) *image = out;
    
    for(i = 0; i < 4; i++){
        if(outgoing[i]) fclose(outgoing[i]);
    }
	NORMAL_RETURN;

ERR_CLEANUP:
    if(sOpt->tiff) w_cleanupTiff(sOpt->tiff);
    if(out) DGP_FREE(out);
    if(cmyk_img[0]) DGP_FREE(cmyk_img[0]);
    if(cmyk_img[1]) DGP_FREE(cmyk_img[1]);
    if(cmyk_img[2]) DGP_FREE(cmyk_img[2]);
    if(cmyk_img[3]) DGP_FREE(cmyk_img[3]);
	ERR_RETURN;
}

/*********************************************************************************
 *
 * Decode the whole document according to the parameters set by silx_setDecState.
 *
 * @param *dgp_state       pointer of Dgp_State
 * @param  *_sOpt,         pointer of SilxDecCookie.
 * @param  *fp             stream pointer, Default is FILE stream.
 *                         If you use other streams, you have to set the uds in SilxDecCookie to handle it.
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */


ErrCode silx_decDocumentAll(   
    Dgp_State                   *dgp_state,
    SilxDecCookie               _sOpt,
    void                        *fp)
{
	DGP_PROLOGUE("slix_decDocument()", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;
	Dgp_Int32       pgNum, numPages;
    Dgp_Float64 pageRes;
	void * out = NULL;
	santaDoc * pdoc = NULL;
    Dgp_Int32 width, height, pageType;

    DGP_ERR_GEN(!sOpt, cErrInvalidParm);
    DGP_ERR_GEN(!fp, cErrInvalidParm);
    sOpt->docUds.fp = fp;

    if(!sOpt->pDoc)
	    DGP_ERR_PASS_CLEANUPF0(silx_decDocumentHeader(dgp_state, _sOpt, fp, &numPages),
 			           "santaUncompAll: Cannot open document\n");
    pdoc = sOpt->pDoc;

    /* decide start and stop page */
    sOpt->startPageNum = sOpt->startPageNum > 0 ? sOpt->startPageNum : 0;
  	if(sOpt->startPageNum> pdoc->npages)   sOpt->startPageNum = pdoc->npages; 
    sOpt->stopPageNum = sOpt->stopPageNum > 0 ? sOpt->stopPageNum : pdoc->npages; 
    if(sOpt->stopPageNum > pdoc->npages) sOpt->stopPageNum = pdoc->npages; 
  	if(sOpt->startPageNum > sOpt->stopPageNum)   sOpt->startPageNum = sOpt->stopPageNum; 

    if(sOpt->outToBuffer) sOpt->stopPageNum = sOpt->startPageNum +1;

    DGP_ERR_GENF0(!DgpCoord_equal(sOpt->output_res.xRes, sOpt->output_res.yRes),
        cErrInvalidParm,
        "Non-square output resolution specified...");

    for(pgNum = sOpt->startPageNum; pgNum < sOpt->stopPageNum; pgNum++){

        DGP_ERR_PASS_CLEANUPF1(silx_decPageHeader(
                                                    dgp_state, 
                                                    _sOpt, 
                                                    pgNum, 
                                                    &width,
                                                    &height,
                                                    &pageType,/*SILXPAGEMONO, SLIXPAGERGB, SLIXPAGECMYK */
                                                    &pageRes), 
                                                    "I couldn't open page %d", pgNum);
        /* all page related features can be set per page (res, force, forceType...) */
        /* by using silx_setDecState() here */
        DGP_ERR_PASS_CLEANUPF1(silx_decPage(dgp_state, _sOpt, &out),
            "I couldn't decode page %d", pgNum); /*Dgp_ImageColorRGB / Dgp_ImagePacked / Dgp_ImageGray[4] */
        if(out)  DGPIMAGE_FREE(out);
    }
	NORMAL_RETURN;

ERR_CLEANUP:
    if(out)  DGPIMAGE_FREE(out);
    out = NULL;
	ERR_RETURN;
}
/*************************************************************************/

/*********************************************************************************
 *
 * Init SilxDecCookie content to their default values. Memory is allocated for the cookie. 
 *
 * @param *dgp_state       pointer of Dgp_State
 * @param  *_sOpt,         pointer of SilxDecCookie.
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */

ErrCode
silx_initDecState(Dgp_State *dgp_state, SilxDecCookie *_sOpt)
{

    SilxDecState *sOpt = NULL;
    DGP_PROLOGUE("silx_initDecCookie", DGP_PERRSTATE);
    DGP_ERR_GEN(!_sOpt, cErrInvalidParm);

    DGP_CHECK_ALLOC_CLEANUP(sOpt = DGP_MALLOC(sizeof(*sOpt)));
    sOpt->force = DGP_FALSE;
    sOpt->stripSize = 1 << (DGP_BITS_UINT32 - 2); /* FIX: should be max Dgp_Int32 */
    sOpt->o = ORIENT_RD;
    sOpt->startPageNum = -1;
    sOpt->stopPageNum = -1;
    
    sOpt->outToFile = DGP_TRUE;        /* Default is output to stdout */
    sOpt->outToStdout = DGP_TRUE;
    sOpt->outToBuffer = DGP_FALSE;
    sOpt->outFileName = NULL;
    sOpt->outtype = PNM;
    
    sOpt->verbose = DGP_FALSE;
    
    sOpt->stripwise = DGP_FALSE;
    sOpt->do_cmyk = DGP_FALSE;
    sOpt->jbig2_only = DGP_FALSE;
    sOpt->images_only = DGP_FALSE;
    
    sOpt->output_res.xRes = DgpCoord_eq(0,1);
    sOpt->output_res.yRes = DgpCoord_eq(0,1);
    sOpt->pDoc = NULL;
    sOpt->pPage = NULL;
    sOpt->tiff = NULL;
    strcpy(sOpt->ext, "pnm");
    
    DGP_ERR_PASS(Dgp_initUds(dgp_state, &(sOpt->docUds),NULL)); 

    *_sOpt = (SilxDecCookie)sOpt;

    NORMAL_RETURN;
ERR_CLEANUP:
    if(sOpt) DGP_FREE(sOpt);
    ERR_RETURN;
 
}
/*********************************************************************************
 *
 * Set the content in SilxDecCookie, 
 * Note, you have to pass in pointers to those values.
 * Pass NULL if you don't want to set that value.
 *
 * @param  *dgp_state      pointer of Dgp_State
 * @param  _sOpt,          SilxDecCookie. Must be init first
 * @param  *docUds,        Stream handler uds
 */
ErrCode
silx_setDecStreamHandler(
               Dgp_State *dgp_state,       
               SilxDecCookie _sOpt,         
               void *docUds) 
{
    DGP_PROLOGUE("silx_setDecStreamHandler", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;

    DGP_ERR_GEN(!_sOpt, cErrInvalidParm);
    if(docUds){
           sOpt->docUds = *((uds *) docUds);
    }

    NORMAL_RETURN;
   
}
/*********************************************************************************
 *
 * Set the content in SilxDecCookie, 
 * Note, you have to pass in pointers to those values.
 * Pass NULL if you don't want to set that value.
 *
 * @param  *dgp_state      pointer of Dgp_State
 * @param  _sOpt,          SilxDecCookie. Must be init first
 * @param  *forceColor,   TRUE to force color output. Default is false.
 * @param  *forceMono,    TRUE to force mono output. Default is false
 * @param  *rotation,     output orientation (0-4 as 0 - 270). Default is 0.
 * @param  *startPage,    start page number. Default is 0.
 * @param  *stopPage,     stop page number.  Default is end of page.
 * @param  *verbose,      TRUE for verbose message output. Default is FALSE.
 * @param  *stripSize,    height of strip. Default is full page height.
 * @param  *jbig2Only,    TRUE to only return jbig2 text. Default is FALSE.
 * @param  *imagesOnly,   TRUE to only return back ground. Default is FALSE.
 * @param  *outputRes,    output resolution, same in x and y direction. Default is originally resolution.
 * @param  *outToBuffer,  TRUE for out to buffer, FALSE for out to file or out to stdout. Default is  FALSE.
 * @param  **outType,     ("TIF"/"BMP"/"PNM") only valid when outToBuffer is FALSE and outFileName is not stdout.
 * @param  **outFileName  ("stdout"/other file name) only valid when outToFile is TRUE.
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */
ErrCode
silx_setDecState(
               Dgp_State *dgp_state,       
               SilxDecCookie _sOpt,         
               Dgp_Bool  *forceColor,      
               Dgp_Bool  *forceMono,       
               Dgp_Int32 *rotation,        
               Dgp_Int32 *startPage,        
               Dgp_Int32 *stopPage,         
               Dgp_Bool  *verbose,         
               Dgp_Int32 *stripSize,       
               Dgp_Bool *jbig2Only,         
               Dgp_Bool *imagesOnly,        
               Dgp_Int32 *outputRes,    
               Dgp_Bool  *outToBuffer, 
               Dgp_Int8  **outType,     
               Dgp_Int8 **outFileName)
           
{
    DGP_PROLOGUE("silx_initDecCookie", DGP_PERRSTATE);
    SilxDecState *sOpt = (SilxDecState *)_sOpt;

    DGP_ERR_GEN(!_sOpt, cErrInvalidParm);
    DGP_ERR_GEN((forceMono && forceColor && *forceMono  && *forceColor), cErrInvalidParm);

    if(forceColor && *forceColor){
            sOpt->force = DGP_TRUE;
            sOpt->forceType = santaPageColor;
    }
    if(forceMono && *forceMono){
            sOpt->force = DGP_TRUE;
            sOpt->forceType = santaPageMono;
    }
    if(rotation)     sOpt->o = rotationToOrientation(*rotation);
    if(startPage && *startPage >= 0)    sOpt->startPageNum = *startPage;
    if(stopPage  && *stopPage >= 0)     sOpt->stopPageNum = *stopPage;
    if(verbose)      sOpt->verbose = *verbose;
    if(stripSize && *stripSize >0) sOpt->stripSize = *stripSize;
    /* For now we only support square-pixel destination resolutions... */
    if(outputRes){
        sOpt->output_res.xRes = DgpCoord_eqi(*outputRes);
        sOpt->output_res.yRes = DgpCoord_eqi(*outputRes);
    }
    /*sOpt->do_cmyk = cl_opts.do_cmyk; */
    if(jbig2Only)     sOpt->jbig2_only = *jbig2Only;
    if(imagesOnly)     sOpt->images_only = *imagesOnly;
    
    if(outToBuffer){
        if(*outToBuffer == DGP_FALSE){
            if(*outFileName){
                sOpt->outToFile = DGP_TRUE;
                if(!strcmp(*outFileName, "stdout")){ /* output To standard output */
                    /* This is the only case where we can do a true stripwise decomposition. */
                    sOpt->outToStdout = DGP_TRUE;
                    sOpt->outToBuffer = DGP_FALSE;
                    sOpt->stripwise = DGP_TRUE;
                    sOpt->outtype = PNM;
                    strcpy(sOpt->ext, "pnm");
                    if(sOpt->outFileName) DGP_FREE(sOpt->outFileName);
                    sOpt->outFileName = NULL;
                }else{ /* output To a file */
                    sOpt->outToBuffer = DGP_FALSE;
                    sOpt->outToStdout = DGP_FALSE;
                    if(sOpt->outFileName) DGP_FREE(sOpt->outFileName);
                    DGP_CHECK_ALLOC_CLEANUP(sOpt->outFileName = 
                        (Dgp_Int8 *)DGP_MALLOC((1 + strlen(*outFileName))* sizeof(Dgp_Int8)));
                    strcpy(sOpt->outFileName, *outFileName);
                    if(outType){
                        DGP_ERR_GENF1(!santaUncompGetType(*outType, &sOpt->outtype, sOpt->ext),
                            cErrInvalidParm,
                            "invalid output file type '%s'...", *outType);
                    }
                }
            }
        } else{  /* if not outToFile then outToBuffer */
            sOpt->outToFile = DGP_FALSE;
            sOpt->outToStdout = DGP_FALSE;
            sOpt->outToBuffer = DGP_TRUE;
            if(sOpt->outFileName) DGP_FREE(sOpt->outFileName);         
        }
    }
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(sOpt->outFileName) DGP_FREE(sOpt->outFileName);
    sOpt->outFileName = NULL;
    ERR_RETURN;
    
}

/****************
 *
 * Free the content in SilxDecCookie and the cookie itself
 *
 * @param dgp_state pointer of DgpState
 * @param sOpt      pointer of SilxDecCookie
 *
 * @return
 * <li> void
 *
 */
void
silx_cleanDecState(
    Dgp_State *dgp_state,
    SilxDecCookie *_sOpt)
{

    if(_sOpt){
        SilxDecState *sOpt = (SilxDecState *)(*_sOpt);
        if(sOpt){
            if(sOpt->outFileName)
              DGP_FREE(sOpt->outFileName);
            sOpt->outFileName = NULL;
            if(sOpt->pPage && sOpt->pDoc) santaClosePage((santaDoc *)(sOpt->pDoc), (santaPage **)(&(sOpt->pPage)), dgp_state);
            if(sOpt->pDoc) santaCloseDocument((santaDoc **)(&(sOpt->pDoc)), dgp_state);
            sOpt->tiff = NULL;
            sOpt->pDoc = NULL;     
            sOpt->pPage= NULL;
            DGP_FREE(sOpt);
            *_sOpt = NULL;
        }
    }
}

