/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#include <libiow.pub>
#include <libpixrw.pub>

#include <math.h>
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "santa_halftone.h"
#include "silx_decode_priv.h"
#include "silx_decode.h"
#include "tc-tiffio.h"



/*************************************************************************
**
**  Function:  ErrCode handleMonoPage( SilxDecState                *sOpt, 
**                                     Dgp_Int8                    *outName, 
**                                     Dgp_Int32                   pgXRes, 
**                                     Dgp_Int32                   pgYRes,
**                                     DgpImage_Packed             pageIm, 
**                                     Dgp_State                   *dgp_state
**                                     TIFF                        **tiff
**                                     Bool                        firstPage,
**                                     Bool                        lastPage)
**
**
**  Do the switch on type.  We just save it if it's a pnm output,
**  else, make a pixr, copy the packedimage, and ask ipcore to save it.
**
**  Writes/saves the passed page image, creating pixr representation as necessary. / dhtesq.
**
**	Arguments:
**      o  sOpt references the santa cookie.
**      o  outName is the output file name.
**      o  pgXRes, pgYRes in dpi.
**      o  pageIm is the packed image to be handled.
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode;  can fail on bad save/write or pixr creation.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Return type changed to ErrCode.  No longer outputs to console necessarily.
**  
**/

ErrCode
handleMonoPage(         SilxDecState               *sOpt, 
                        Dgp_Int8                    *outName, 
                        Dgp_Int32                   pgXRes, 
                        Dgp_Int32                   pgYRes,
                        DgpImage_Packed             pageIm, 
                        Dgp_State                   *dgp_state,
                        TIFF                        **tiff,
                        Bool                        firstPage,
                        Bool                        lastPage)
{
    DGP_PROLOGUE("handleMonoPage", DGP_PERRSTATE);
    PIXR *outIm = NULL;

  if(sOpt->outtype == PNM){
      DGP_ERR_PASS_CLEANUPF1(DgpImage_save(pageIm, outName), 
                     "santa-uncomp-all:  I couldn't save the pnm file '%s'.\n", outName);
  } else {
      Dgp_Int32 y;
      UInt32 *pPixrRow = NULL;
      ScaledInt1616 xdpi, ydpi;
      
      xdpi = Int32_TO_ScaledInt1616((Int32)pgXRes);
      ydpi = Int32_TO_ScaledInt1616((Int32)pgYRes);

      DGP_ERR_PASS_CLEANUPF1(w_createPixrExtended(DGP_PIPWSTATE, 
                                        DGPIMG_WIDTH(pageIm), 
                                        DGPIMG_HEIGHT(pageIm),
                                        1, 
                                        xdpi, 
                                        ydpi, 
                                        cFrameBits,
                                        1, 
                                        1, 
                                        NULL, 
                                        &outIm),
                     "santa-uncomp-all:  I couldn't create the pixr for '%s'.\n", outName);

      w_pixrSetType(DGP_PIPWSTATE, outIm, cColorNone);
      
      for(y = 0; y < DGPIMG_HEIGHT(pageIm); y++){
          w_getLine(DGP_PIPWSTATE, outIm, y,(UInt8 **)&pPixrRow);
          memcpy(pPixrRow, DGPIMG_GETROW(pageIm, y), DGPIMG_GETROWWIDTH(pageIm));
      }

      DGP_ERR_PASS_CLEANUPF1(w_writeEx(DGP_PIPWSTATE, 
                                       outName,
                                       outIm, 
                                       sOpt->outtype,
                                       tiff, 
                                       firstPage,
                                       lastPage),
                             "santa-uncomp-all:  I couldn't write the pixr for '%s'.\n", outName);
      
      w_destroyPixr(DGP_PIPWSTATE, &outIm);
  }
  NORMAL_RETURN;
ERR_CLEANUP:
  w_destroyPixr(DGP_PIPWSTATE, &outIm);
  ERR_RETURN;
}
/*************************************************************************/



/*************************************************************************
**
**  Function:  ErrCode handleColorPage( SilxDecState               *sOpt, 
**                                      Dgp_Int8                   *outName, 
**                                      Dgp_Int32                  pgXRes, 
**                                      Dgp_Int32                  pgYRes, 
**                                      RGBImage                   pageIm,
**                                      Dgp_State                  *dgp_state
**                                     TIFF                        ** tiff,
**                                     Bool                        firstPage,
**                                     Bool                        lastPage)
**
**
**  Writes/saves the passed page image, 
**
**	Arguments:
**      o  sOpt references the santa cookie.
**      o  outName is the output file name, good to go.
**      o  pgXRes, pgYRes in dpi.
**      o  pageIm is the RGBImage to be handled.
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode;  can fail on bad save/write or pixr creation.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Return type changed to ErrCode.  
**        No longer outputs to console necessarily.
**  
**/
ErrCode 
handleColorPage(         SilxDecState               *sOpt, 
                         Dgp_Int8                   *outName, 
                         Dgp_Int32                  pgXRes, 
                         Dgp_Int32                  pgYRes,
                         DgpImage_ColorRGB          pageIm,
                         Dgp_State                  *dgp_state,
                         TIFF                       **tif,
                         Bool                       firstPage,
                         Bool                       lastPage)
{
    DGP_PROLOGUE("handleColorPage", DGP_PERRSTATE);

    DGP_ERR_GEN_CLEANUP(!outName, cErrNullPtr);
    DGP_ERR_GEN_CLEANUP(!pageIm, cErrNullPtr);
    DGP_ERR_GEN_CLEANUP(!tif, cErrNullPtr);

    if(sOpt->outtype == PNM){
      DGP_ERR_PASS_CLEANUPF1(DgpImage_save(pageIm, outName),
                             "santa-uncomp-all:  I couldn't save the pnm file '%s'.\n", outName);
    } else {
        if(firstPage) DGP_ERR_PASS_CLEANUP(tc_writeTiffHeader (dgp_state,outName, (TCTIFFPtr *)tif));
        DGP_ERR_PASS_CLEANUP(
            tc_writeColorTiffDirectory(dgp_state, pageIm, 8, 3, pgXRes, pgYRes, DGP_TRUE, sOpt->stripSize, (TCTIFFPtr)*tif));
        if(lastPage) tc_finishTiff((TCTIFFPtr *)tif);
    }     
    NORMAL_RETURN;
ERR_CLEANUP:
    if(*tif) tc_cleanupTiff((TCTIFFPtr*)tif);
    ERR_RETURN;
}
/*************************************************************************/

/*************************************************************************/
ErrCode 
handleCMYKPage(          SilxDecState               *sOpt, 
                         Dgp_Int8                   *outName, 
                         Dgp_Int32                  pgXRes, 
                         Dgp_Int32                  pgYRes, 
                         void                       **pageIm,
                         Dgp_State                  *dgp_state,
                         TIFF                       **tiff,
                         Bool                       firstPage,
                         Bool                       lastPage)
{
    DGP_PROLOGUE("handleColorPage", DGP_PERRSTATE);
    Dgp_Int8 chars_cmyk[5] = "cmyk";
    Dgp_Int8 fnm[256];
    Dgp_Int32 i;

    pgXRes = pgXRes;
    pgYRes = pgYRes;
    tiff = tiff;
    firstPage = firstPage;
    lastPage = lastPage;

    if(sOpt->outtype == PNM){
        for(i = 0; i < 4; i++){
            sprintf(fnm, "%s%c.%s", outName, chars_cmyk[i], sOpt->ext);
            DGP_ERR_PASS_CLEANUPF2(DgpImage_save((DgpImage_Gray)pageIm[i], fnm),
                "I couldn't save the %s file '%s'.\n", sOpt->ext, fnm);
        }
    }

    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}

