/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libencode/huff.c,v 1.7 2003/02/11 01:31:41 mkrocken Exp $
*
*/

#include "util.h"
#include "list.h"
#include <stdlib.h>
#include "huff.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libencode/huff.c,v 1.7 2003/02/11 01:31:41 mkrocken Exp $");

typedef struct htree {
    struct htree *zero, *one;
    Dgp_Int32 totval;
    Dgp_Int32 index;
    Dgp_Int32 codelen;
} htree;

/*
* Given the code length in h, assign code lengths to all h's descendents.
* Fill MSB-to-LSB.
*/
static void
assignlengths(htree *h)
{

	if (!h) return;
	
    if (h->zero != (htree *)NULL) {
		h->zero->codelen = h->codelen + 1;
		assignlengths(h->zero);
	}
	
    if (h->one != (htree *)NULL) {
		h->one->codelen = h->codelen + 1;
		assignlengths(h->one);
	}
}


/*
* A use count we're not likely to see often
*/
#define	NBUCKETS	1000

/*
* Generate a Huffman encoding for the values passed in. Fill in the code and
* codelen (in bits) fields. The code is placed in the codelen LSBs of the
* code field, and should be read MSB-to-LSB.
*
* The algorithm used is:
* Scan the current array for the two lowest-value elements. In cases
* of ties, the element closest to the start of the array is used.
* Replace the first (closest to the start of those two) element with the
* node representing the merged elements, and delete the second.
* Continue until there's only one thing in the array.
*
* This process gives a tree; the code lengths are assigned based on the depth
* in the tree (i.e. how many merges it took to go from the original element
* to the single remaining thing in the array). The codes themselves are
* then assigned based on the lengths: the longest codes start at 000...0,
* and are assigned incrementally up from there (in the order that they occur
* in the array); the next shortest code starts at the lowest-valued possible
* unassigned codeword. These are the canonical Huffman codes described in
* "Managing Gigabytes".
*
* The tie-breaking algorithm is still pertinent, since it can affect which
* tokens get which which code lengths.
*
* The actual algorithm used is:
* Sort into buckets using radix sort
* Concatenate the buckets into a linked (sorted) list
* While the list contains at least two elements:
* - Merge the first two entries on the list
* - Insert the merged entry in the sorted list, starting the search for
*   the insertion point at the last insertion point.
* This is O(n).
*
* Return TRUE on success, FALSE on failure.
*/

ErrCode
DgpEncode_huffmanCode(Dgp_State *dgp_state, DgpEncode_Huff *vals, Dgp_Int32 nvals)
{
    htree *array = (htree *)NULL;
    htree *small, *small2;
    Dgp_Int32 i;
    Dgp_Int32 arraylen;
    Dgp_Int32 codecounts[DGP_MAX_CODELEN + 1];
    Dgp_Int32 firstcode[DGP_MAX_CODELEN + 1];
    /* DgpAdt_List buckets[NBUCKETS];  ...too much on stack / dhtesq. */
    DgpAdt_List *buckets = NULL;
    DgpAdt_List sorted = DGP_NULLLIST;
    DgpAdt_ListNode insertion, last, smallnode, temp;
    htree *node;

    DGP_PROLOGUE ("ENCODE_huffmanCode", DGP_PERRSTATE);

    if (nvals == 0) {
		/* We successfully did nothing. */
		NORMAL_RETURN;
	}
	
    /* Dynamically allocate buckets... */
    DGP_CHECK_ALLOC(buckets = (DgpAdt_List *)DGP_MALLOC(NBUCKETS * sizeof(DgpAdt_List)));

    for (i = 0; i < NBUCKETS; i++) {
		buckets[i] = DGP_NULLLIST;
	}
	
    DGP_CHECK_ALLOC (array = (htree *) DGP_MALLOC(nvals * 2 * sizeof(*array)));

    /* Read out the input into our array of buckets. */
    for (i = 0; i < nvals; i++) {
		array[i].zero = array[i].one = (htree *)NULL;
		array[i].totval = vals[i].value;
		array[i].index = i;
		array[i].codelen = 0;
		
		if (array[i].totval < NBUCKETS - 1) {
			if (buckets[array[i].totval] == DGP_NULLLIST) {
				DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&(buckets[array[i].totval]), dgp_state));
			}
			
			DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(buckets[array[i].totval], (void *)&array[i], &temp));

		}
		else {
		/*
		* The last bucket is a special case since I don't feel like
		* implementing a full radix sort.
			*/
			if (buckets[NBUCKETS - 1] == DGP_NULLLIST) {
				DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&(buckets[NBUCKETS - 1]), dgp_state));
			}
			
			/* Insert it just before the first item larger than it */
			last = DGP_NULLLISTNODE;
			DGPADT_FOREACH (insertion, buckets[NBUCKETS - 1]) {
				node = (htree *)DGPADT_LIGET(insertion);
				if (node->totval > array[i].totval) {
					/* We've found the right place */
					break;
				}
				last = insertion;
			}
			
			DGP_ERR_PASS_CLEANUP (DgpAdt_liAddAfter(buckets[NBUCKETS - 1], last,
				(void *)&array[i], &temp));
		}
	}
	

    /* Now concatenate the buckets */
    for (i = 0; i < NBUCKETS; i++) {
      if (buckets[i] != DGP_NULLLIST) {
	if (sorted == DGP_NULLLIST) {
	  /* First non-empty bucket */
	  sorted = buckets[i];
	  buckets[i] = DGP_NULLLIST;
	}
	else {
	  /* Append to sorted */
	  DGP_ERR_PASS_CLEANUP (DgpAdt_liApp(sorted, buckets[i]));
	  buckets[i] = DGP_NULLLIST;
	}
      }
    }
    arraylen = nvals;
	
    /* Here's our initial insertion point... */
    insertion = DGP_NULLLISTNODE;
    while (DGPADT_LILEN(sorted) > 1) {
		/* Get the smallest entry */
		smallnode = DGPADT_LIFIRST(sorted);
		DGP_ASSERT(smallnode != DGP_NULLLISTNODE);
		if (smallnode == insertion) {
		  /*
		* If this was our reference node, it's not going to be valid
		* anymore - just set our insertion search back to the start
		* of the list, since it was nearly there anyway.
			*/
			insertion = DGP_NULLLISTNODE;
		}
		small = (htree *)DGPADT_LIGET(smallnode);
		DGP_ASSERT(small != (htree *)NULL);
		DgpAdt_liRem(smallnode);
		
		/* and the second smallest */
		smallnode = DGPADT_LIFIRST(sorted);
		DGP_ASSERT(smallnode != DGP_NULLLISTNODE);
		if (smallnode == insertion) {
			insertion = DGP_NULLLISTNODE;
		}
		small2 = (htree *)DGPADT_LIGET(smallnode);
		DGP_ASSERT(small2 != (htree *)NULL);
		DgpAdt_liRem(smallnode);
		
		/* Create a merged node */
		array[arraylen].zero = small;
		array[arraylen].one = small2;
		/* The index in our theoretical array is used as a secondary sort key */
		array[arraylen].index = MIN(small->index, small2->index);
		array[arraylen].totval = small->totval + small2->totval;
		array[arraylen].codelen = 0;
		
		/*
		* Now find where to insert the merged node.
		* The tricky part is that we start our search at the last place
		* that we inserted anything, since this new merged node must have
		* totval at least as big as that.
		*
		* insertion might be equal to DGP_NULLLISTNODE, in which case we start
		* searching from the start of the list.
		*/
		last = insertion;
		if (insertion == DGP_NULLLISTNODE) {
			insertion = DGPADT_LIFIRST(sorted);
		}
		
		for (;
		insertion != DGP_NULLLISTNODE;
		last = insertion, insertion = DGPADT_LINEXT(insertion)) {
			node = (htree *)DGPADT_LIGET(insertion);
			
			if (node->totval >= array[arraylen].totval) {
				if (node->totval > array[arraylen].totval) {
					/* Here's where to put it */
					break;
				}
				else if (node->index > array[arraylen].index) {
					break;
				}
			}
		}
		
		/* Insert just before insertion */
		DGP_ERR_PASS_CLEANUP (DgpAdt_liAddAfter(sorted, last, (void *)&array[arraylen], &insertion));
		
		arraylen++;
	}


    /*
	* Done.
	* The only entry in sorted now refers to the single entry in the queue,
	* the root.
	*/
    small = DGPADT_LIGET(DGPADT_LIFIRST(sorted));
	
    /* We'd better have done nvals - 1 merges */
    DGP_ASSERT(arraylen == nvals * 2 - 1);
	
    /* Assign the code lengths. */
    assignlengths(small);

    /* count them */
    for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		codecounts[i] = 0;
	}
	
    for (i = 0; i < nvals; i++) {
		/* Count the leaves only */
		DGP_ASSERT(array[i].zero == (htree *)NULL);
		codecounts[array[i].codelen]++;
	}
	
    firstcode[DGP_MAX_CODELEN] = 0;
    for (i = DGP_MAX_CODELEN - 1; i >= 0; i--) {
		firstcode[i] = (firstcode[i+1] + codecounts[i+1]) / 2;
	}
	
    /* and read them out into the input */
    for (i = 0; i < nvals; i++) {
		vals[i].code = firstcode[array[i].codelen]++;
		vals[i].codelen = array[i].codelen;
	}
	
    (DGP_FREE)((void *)array);
    (DGP_FREE)((void *)buckets);
    DgpAdt_liFree(sorted);

    NORMAL_RETURN;
	
ERR_CLEANUP:
	
    if (array != (htree *)NULL) {
		(DGP_FREE)((void *)array);
	}
	
    for (i = 0; i < NBUCKETS; i++) {
		if (buckets[i] != DGP_NULLLIST) {
			DgpAdt_liFree(buckets[i]);
		}
	}
	
    if (sorted != DGP_NULLLIST) {
		DgpAdt_liFree(sorted);
	}

    if(buckets != NULL){
      DGP_FREE((void *)buckets);
    }
	
    ERR_RETURN;
}
	
/*
* Build some tables useful in decoding stuff encoded with the codes
* in *vals. The codes in *vals must have been filled in using huffmanCode
* above.
*
* The tables it builds are:
* - firstcode[i] is the minimum value seen in any i bit long code
* - index[i][j] is the index (in the Huff array) of the code i bits long
*   whose value is (j + firstcode[i]).
*
* This means that you can decode using the following algorithm:
* v = 0; l = 1;
* while (v < firstcode[l]) {
*     v = (v << 1) | nextbit(); l++;
*     }
* The code is now that represented by index[l][v - firstcode[l]].
*
* The caller is responsible for allocating the memory pointed at by
* firstcode and index (each should be DGP_MAX_CODELEN+1 elements long);
* this routine allocates the memory pointed at by index[i] for
* i=0..DGP_MAX_CODELEN. If a code length has 0 instances, then
* index[that length] may well be NULL. The caller is
* responsible for freeing the memory when it is done with it.
* The allocator supplied to libwjr_set_alloc is used.
*
* This routine returns FALSE on failure, TRUE on success. If it fails,
* the values in index[0..DGP_MAX_CODELEN] are all set to (Dgp_Int32 *)NULL.
*/

ErrCode
DgpEncode_huffBuildDecodeTables(Dgp_State *dgp_state, DgpEncode_Huff *vals, Dgp_Int32 nvals, Dgp_UInt32 *firstcode,
					  Dgp_Int32 **index)
{
	Dgp_Int32 i;
	Dgp_Int32 counts[DGP_MAX_CODELEN + 1];
	
	DGP_PROLOGUE ("ENCODE_huffBuildDecodeTables", DGP_PERRSTATE);

	DGP_ASSERT(vals != (DgpEncode_Huff *)NULL);
	DGP_ASSERT(firstcode != (Dgp_UInt32 *)NULL);
	DGP_ASSERT(index != (Dgp_Int32 **)NULL);
	
	/* Get the counts. Clear index[..] while we're at it. */
	for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		counts[i] = 0;
		index[i] = (Dgp_Int32 *)NULL;
	}
	for (i = 0; i < nvals; i++) {
		counts[vals[i].codelen]++;
	}
	
	/* Work out the firstcode values */
	firstcode[DGP_MAX_CODELEN] = 0;
	for (i = DGP_MAX_CODELEN - 1; i >= 0; i--) {
		firstcode[i] = (firstcode[i+1] + counts[i+1]) / 2;
	}
	
	/* Allocate the index space */
	for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		if (counts[i]) {
			DGP_CHECK_ALLOC_CLEANUP (index[i] = (Dgp_Int32 *) DGP_MALLOC(counts[i] * sizeof(**index)));
		}
	}

	/* Fill in the index values. Re-use counts. */
	for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		counts[i] = 0;
	}
	
	for (i = 0; i < nvals; i++) {
		index[vals[i].codelen][counts[vals[i].codelen]] = i;
		counts[vals[i].codelen]++;
	}

	NORMAL_RETURN;
	
ERR_CLEANUP:
	
	for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		if (index[i] != (Dgp_Int32 *)NULL) {
			(DGP_FREE)((void *)index[i]);
			index[i] = (Dgp_Int32 *)NULL;
		}
	}
	
	ERR_RETURN;
	
}


#if 0
Dgp_Int32 
main(Dgp_Int32 argc, char **argv)
{
	Dgp_Int32 i, nrep, j;
	Dgp_Int32 n;
	DgpEncode_Huff h[100000];
	Dgp_Int32 len;
	Dgp_Int32 sum;
	
	sum = 0;
	for (i = 0; ; i++) {
		if (scanf("%d %d", &h[i].value, &nrep) != 2) {
			break;
		}
		nrep = 1;
		
		sum += h[i].value * nrep;
		for (j = 1; j < nrep; j++) {
			h[i + j].value = h[i].value;
		}
		i += nrep - 1;
	}
	n = i;
	
	if (!huffmanCode(h, n)) {
		printf("oops\n");
		exit(1);
	}
	
	for (i = 0; i < n; i++) {
		printf("%4d %2d ", h[i].value, h[i].codelen);
		for (j = 0; j < h[i].codelen; j++) {
			if (h[i].code & (1ul << (31 - j))) {
				putchar('1');
			}
			else {
				putchar('0');
			}
		}
		putchar('\n');
	}
	
	len = (n < 256) ? 8 : ((n < 65536) ? 16 : 32);
	printf("old len: %d (%d)\n", len * sum, (len * sum + 7) / 8);
	for (i = 1, len = 0; i < n; i <<= 1, len++) {
	}
	printf("better len: %d (%d)\n", len * sum, (len * sum + 7) / 8);
	len = 0;
	for (i = 0; i < n; i++) {
		len += h[i].codelen * h[i].value;
	}
	printf("huff len: %d (%d)\n", len, (len + 7) / 8 + n);
	
	return(0);
}
#endif

#if 0
/*
Illustrative example of minimum-redundancy code calculation.

  Program reads a file of symbol frequencies
  -- one per line
  -- ascending frequency
  -- first line indicates upper bound on how many 
  frequencies appear;
  calculate the codeword lengths for a minimum-redundancy code;
  prDgp_Int32 out the codeword lengths;
  prDgp_Int32 out the average cost per symbol and the entropy.
  
	The method for calculating codelengths in function
	calculate_minimum_redundancy is described in 
	
	  @inproceedings{mk95:wads,
	  author = "A. Moffat and J. Katajainen",
	  title = "In-place calculation of minimum-redundancy codes",
	  booktitle = "Proc. Workshop on Algorithms and Data Structures",
	  address = "Queen's University, Kingston, Ontario",
	  publisher = "LNCS 955, Springer-Verlag",
	  Month = aug,
	  year = 1995,
	  editor = "S.G. Akl and F. Dehne and J.-R. Sack",
	  pages = "393-402",
	  }
	  
		The abstract of that paper may be fetched from 
		http://www.cs.mu.oz.au/~alistair/abstracts/wads95.html
		A revised version is currently being prepared.
		
		  Written by
		  Alistair Moffat, alistair@cs.mu.oz.au,
		  Jyrki Katajainen, jyrki@diku.dk
		  November 1996.
*/

/*** Function to calculate in-place a minimum-redundancy code
Parameters:
A        array of n symbol frequencies, non-decreasing
n        number of symbols in A
*/
void
calculate_minimum_redundancy(Dgp_Int32 A[], Dgp_Int32 n) {
	Dgp_Int32 root;                  /* next root node to be used */
	Dgp_Int32 leaf;                  /* next leaf to be used */
	Dgp_Int32 next;                  /* next value to be assigned */
	Dgp_Int32 avbl;                  /* number of available nodes */
	Dgp_Int32 used;                  /* number of internal nodes */
	Dgp_Int32 dpth;                  /* current depth of leaves */
	
	/* check for pathological cases */
	if (n==0) { return; }
	if (n==1) { A[0] = 0; return; }
	
	/* first pass, left to right, setting parent pointers */
	A[0] += A[1]; root = 0; leaf = 2;
	for (next=1; next < n-1; next++) {
		/* select first item for a pairing */
		if (leaf>=n || A[root]<A[leaf]) {
			A[next] = A[root]; A[root++] = next;
		} else
			A[next] = A[leaf++];
		
		/* add on the second item */
		if (leaf>=n || (root<next && A[root]<A[leaf])) {
			A[next] += A[root]; A[root++] = next;
		} else
			A[next] += A[leaf++];
	}
	
	/* second pass, right to left, setting internal depths */
	A[n-2] = 0;
	for (next=n-3; next>=0; next--)
		A[next] = A[A[next]]+1;
	
	/* third pass, right to left, setting leaf depths */
	avbl = 1; used = dpth = 0; root = n-2; next = n-1;
	while (avbl>0) {
		while (root>=0 && A[root]==dpth) {
			used++; root--;
		}
		while (avbl>used) {
			A[next--] = dpth; avbl--;
		}
		avbl = 2*used; dpth++; used = 0;
	}
}
#endif
