/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, February 1 1998
**
**  Filename: dgp_err_cotrol.h
**
**  Description of what's implemented in this file:
**    macros to simplify error reporting.
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_err_control.h,v 1.80 2003/01/31 19:15:43 mkrocken Exp $
 */

#ifndef __DGP_ERR_CONTROL_H
#define __DGP_ERR_CONTROL_H


#include "types.pub"
#include "ipwerrs.pub"
#include "ipserrs.pub"
#include "errmacw.pub"

/*  VxWorks needs explicit architecture definition for stdarg.h to work. / dhtesq.*/
#ifdef VXWORKS
#include <vxWorks.h>
#endif  /* (VXWORKS)*/

#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include "dgp_allocators.h"
#include "dgp_rcs.h"
#include "dgp_errs.h"
#include "dgp_types.h"
#include "util.h"

/*  
 * ***NOTE:  If you don't understand how to use this code, 
 *           please read "Error Handling.htm".***
 */


#ifdef __cplusplus
extern "C"
{
#endif


/* This special error code specifies a "Warning" to the user. */
#define DGP_WARNING     -2

/* This special error code specifies a "Report" to the user. */
#define DGP_REPORT      -3


#ifdef DGP_DBUG

/* function typedef for debug output error printer */
typedef void (*Dgp_ErrPrinterType)(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                          Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...);

/* function prototype for debug output */
DGP_LIBEXPORT void Dgp_errPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                          Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...);

/* debug arguments to dgpErrPriner function */
#define DGP_ERR_PRINTER_ARGS(funct_name, error) \
  __FILE__, (funct_name), __LINE__, (error)


#else /* DGP_DBUG is not defined */

/* function typedef for nondebug output error printer */
typedef void (*Dgp_ErrPrinterType)(void *cookie, ErrCode error, const Dgp_Int8 *format, ...);

/* function prototype for nondebug output */
DGP_LIBEXPORT void Dgp_errPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...);

/* nondebug arguments to dgpErrPriner function */
#define DGP_ERR_PRINTER_ARGS(funct_name, error) \
  (error)

#endif /* DGP_DBUG */

/* for asserts in void functions */
#ifdef __GNUC__
DGP_LIBEXPORT void Dgp_voidAssertPrinter(const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
				 Dgp_Int32 line_number, const Dgp_Int8 *cond) __attribute__ ((noreturn));
#else
DGP_LIBEXPORT void Dgp_voidAssertPrinter(const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                                 Dgp_Int32 line_number, const Dgp_Int8 *cond);
#endif


/* error state for debug output */
typedef struct _Dgp_ErrState {
  void *userCookie;
  Dgp_ErrPrinterType errPrinter;
} Dgp_ErrState;


/*
 * DGP_MAIN_PROLOGUE(_FUNC)
 *   May be used in "main" procedure of program in order to statically create
 *   a dgpErrStruct for error output.  The *_FUNC version allows you
 *   to specify a different error printer function than the default.
 *   The user cookie is the first parameter passed into the error
 *   printer function.  You can then get a pointer to the static struct
 *   using DGP_ERR_STATE_PTR.
 */
#ifdef DGP_DBUG

#define DGP_MAIN_PROLOGUE(funct_name, user_cookie, err_printer) \
  Dgp_ErrState __ERR_STATE__ = {(user_cookie), (err_printer)}; \
  Dgp_ErrState *__PERR_STATE__ = &__ERR_STATE__; \
  PROLOGUE; \
  Dgp_Int8* __FUNCT_NAME__ = (funct_name)

#else /* DGP_DBUG is not defined */

#define DGP_MAIN_PROLOGUE(funct_name, user_cookie, err_printer) \
  Dgp_ErrState __ERR_STATE__ = {(user_cookie), (err_printer)}; \
  Dgp_ErrState *__PERR_STATE__ = &__ERR_STATE__; \
  PROLOGUE

#endif /* DGP_DBUG */


/*
 * DGP_PROLOGUE
 *   A function prologue. It defines the "__error__" variable, and
 *   initializes it, as well as the error cookie and possibly function name.
 */
#ifdef DGP_DBUG

#define DGP_PROLOGUE(funct_name, pdgp_err_state)  \
  Dgp_ErrState *__PERR_STATE__ = (pdgp_err_state); \
  PROLOGUE; \
  Dgp_Int8* __FUNCT_NAME__ = (funct_name)

#else /* DGP_DBUG is not defined */

#define DGP_PROLOGUE(funct_name, pdgp_err_state)  \
  Dgp_ErrState *__PERR_STATE__ = (pdgp_err_state); \
  PROLOGUE

#endif /* DGP_DBUG */


/*
 * DGP_VOID_PROLOGUE
 *   A void function prologue. It defines the error cookie and possibly 
 *   function name.
 *   (useful for void functions which may print warnings or reports or use
 *    DGP_VOID_ASSERT's).
 */
#ifdef DGP_DBUG

#define DGP_VOID_PROLOGUE(funct_name, pdgp_err_state)  \
  Dgp_ErrState *__PERR_STATE__ = (pdgp_err_state); \
  Dgp_Int8* __FUNCT_NAME__ = (funct_name)

#else /* DGP_DBUG is not defined */

#define DGP_VOID_PROLOGUE(funct_name, pdgp_err_state)  \
  Dgp_ErrState *__PERR_STATE__ = (pdgp_err_state)

#endif /* DGP_DBUG */


/*
 * DGP_ERR_STATE
 *   Get a handle to the current dgpErrStruct
 */
#define DGP_ERR_STATE __PERR_STATE__

/*
 * DGP_CUR_ERROR
 *   Get the current error
 */
#define DGP_CUR_ERROR __error__

/*
 * DGP_CLEAR_CUR_ERROR
 *   Clear the current error value.
 */
#define DGP_CLEAR_CUR_ERROR __error__ = cSuccessful

/* 
 * DGP_SET_ERR_COOKIE
 *   Change the current user cookie for error printing
 */
#define DGP_SET_ERR_COOKIE(user_cookie) __PERR_STATE__->userCookie = (user_cookie)

/* 
 * DGP_SET_ERR_FUNC
 *   Change the current error printing function
 */
#define DGP_SET_ERR_FUNC(err_printer) __PERR_STATE__->errPrinter = (err_printer)

/*
 * DGP_SET_ERR_STATE
 *   Change the current error state variable.
 */
#define DGP_SET_ERR_STATE(err_state) __PERR_STATE__ = (err_state)


/*
 * DGP_STOP_WHINE
 *  Don't you hate it when the compiler whines about unused variables...
 */
#ifdef DGP_DBUG
#define DGP_STOP_WHINE \
  (void)__PERR_STATE__; \
  (void)__error__; \
  (void)__FUNCT_NAME__
#else
#define DGP_STOP_WHINE \
  (void)__PERR_STATE__; \
  (void)__error__
#endif

/*
 * MAIN_*_RETURN
 *   macros to return from "main" function.
 */
#define MAIN_NORMAL_RETURN return(EXIT_SUCCESS)
#define MAIN_ERR_RETURN return(EXIT_FAILURE)


/*
 * VOID_RETURN
 *   Used in functions which return void.
 */
#define VOID_RETURN \
  return


/* utility macros for the macros defined below */
#ifdef DGP_DBUG

#define DGP_ERR_PRINTER \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, __error__, \
                Dgp_getErrStr(__error__))

#define DGP_ERR_PRINTERF0(err_str) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, __error__, \
                (err_str))

#define DGP_ERR_PRINTERF1(err_str, arg1) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, __error__, \
                (err_str), (arg1))

#define DGP_ERR_PRINTERF2(err_str, arg1, arg2) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, __error__, \
                (err_str), (arg1), (arg2))

#define DGP_ERR_PRINTERF3(err_str, arg1, arg2, arg3) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, __error__, \
                (err_str), (arg1), (arg2), (arg3))

#define DGP_ERR_PASS_PRINTER DGP_ERR_PRINTER

#define DGP_ERR_PASS_PRINTERF0 DGP_ERR_PRINTERF0

#define DGP_ERR_PASS_PRINTERF1 DGP_ERR_PRINTERF1

#define DGP_ERR_PASS_PRINTERF2 DGP_ERR_PRINTERF2

#define DGP_ERR_PASS_PRINTERF3 DGP_ERR_PRINTERF3

#define DGP_NO_PRINTER ((void)__FUNCT_NAME__); ((void)__PERR_STATE__)

#else /* DGP_DBUG is not defined */

#define DGP_ERR_PRINTER \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __error__, Dgp_getErrStr(__error__))

#define DGP_ERR_PRINTERF0(err_str) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __error__, (err_str))

#define DGP_ERR_PRINTERF1(err_str, arg1) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __error__, (err_str), (arg1))

#define DGP_ERR_PRINTERF2(err_str, arg1, arg2) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __error__, (err_str), (arg1), (arg2))

#define DGP_ERR_PRINTERF3(err_str, arg1, arg2, arg3) \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __error__, (err_str), (arg1), (arg2), (arg3))

#define DGP_ERR_PASS_PRINTER ((void)__error__);  ((void)__PERR_STATE__)

#define DGP_ERR_PASS_PRINTERF0(err_str) ((void)__error__);  ((void)__PERR_STATE__)

#define DGP_ERR_PASS_PRINTERF1(err_str, arg1) ((void)__error__);  ((void)__PERR_STATE__)

#define DGP_ERR_PASS_PRINTERF2(err_str, arg1, arg2) ((void)__error__);  ((void)__PERR_STATE__)

#define DGP_ERR_PASS_PRINTERF3(err_str, arg1, arg2, arg3) ((void)__error__);  ((void)__PERR_STATE__)

#define DGP_NO_PRINTER ((void)__PERR_STATE__)

#endif /* DGP_DBUG */


/*
 * And now, on with the macros!!!
 */


/*
 * DGP_GOTO_CLEANUP
 *   Jump to the error cleanup code.
 */
#define DGP_GOTO_CLEANUP goto ERR_CLEANUP


/*
 * DGP_REPORTF*
 *   Print a warning to the user
 */
#define DGP_REPORTF0(do_report, rep_str) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str)); \
} while (0)

#define DGP_REPORTF1(do_report, rep_str, arg1) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str), (arg1)); \
} while (0)

#define DGP_REPORTF2(do_report, rep_str, arg1, arg2) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str), (arg1), (arg2)); \
} while (0)

#define DGP_REPORTF3(do_report, rep_str, arg1, arg2, arg3) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str), (arg1), (arg2), (arg3)); \
} while (0)

#define DGP_REPORTF4(do_report, rep_str, arg1, arg2, arg3, arg4) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, \
        DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str), (arg1), (arg2), (arg3), (arg4)); \
} while (0)

#define DGP_REPORTF5(do_report, rep_str, arg1, arg2, arg3, arg4, arg5) \
do { \
  if((do_report) && (__PERR_STATE__ != NULL)) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, \
        DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_REPORT), (rep_str), (arg1), (arg2), (arg3), (arg4), (arg5)); \
} while (0)


/*
 * DGP_WARNF*
 *   Print a warning to the user
 */
#define DGP_WARNF0(warn_str) \
do { \
  if(__PERR_STATE__ != NULL) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_WARNING), (warn_str)); \
} while (0)

#define DGP_WARNF1(warn_str, arg1) \
do { \
  if(__PERR_STATE__ != NULL) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_WARNING), (warn_str), (arg1)); \
} while (0)

#define DGP_WARNF2(warn_str, arg1, arg2) \
do { \
  if(__PERR_STATE__ != NULL) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_WARNING), (warn_str), (arg1), (arg2)); \
} while (0)

#define DGP_WARNF3(warn_str, arg1, arg2, arg3) \
do { \
  if(__PERR_STATE__ != NULL) \
    __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, DGP_WARNING), (warn_str), (arg1), (arg2), (arg3)); \
} while (0)


/*
 * DGP_ASSERT
 * Old:
 *  In the debug case, assert that the condition is true.  if not, print and
 *  error and return.  In the non-debug case, this code will do NOTHING.
 *
 * New:
 *  This caused a different behavior between the debug and non-debug
 *  versions (which is bad).  The idea behind an assertion is to let the
 *  debug execution stop and complain where the assertion failed, letting the
 *  programmer fix it before a non-debug version were given to a client.
 *
 *  The assert now throws an assertion in the debug case, and returns an
 *  error in the non-debug case (still a difference between the versions,
 *  but will cause the execution to stop in the debug version, so this is
 *  okay).
 */
#ifdef DGP_DBUG
#define DGP_ASSERT(cond) assert(cond)
#else /* DGP_DBUG is not defined */
#define DGP_ASSERT(cond)\
do { \
  if(!(cond)) { \
    __error__ = cErrAssertFailure;\
    return __error__;\
  } \
} while(0)
#endif /* DGP_DBUG */


/*
 * DGP_VOID_ASSERT
 *  In the debug case, assert that the condition is true.  if not, print and error
 *  and exit (since we are in a void function).  If the error state is null, the 
 *  assert failure info will be printed to stderr.  In the non-debug case, this code
 *  will do NOTHING.
 */
#ifdef DGP_DBUG
#define DGP_VOID_ASSERT(cond)\
do { \
  if(!(cond)) { \
    if(__PERR_STATE__ != NULL) { \
      __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, __FILE__, __FUNCT_NAME__, __LINE__, \
        cErrAssertFailure, "(" #cond ") assertion failed"); \
      exit(EXIT_FAILURE);\
    } else { \
      Dgp_voidAssertPrinter(__FILE__, __FUNCT_NAME__, __LINE__, #cond ); \
    } \
  } \
} while(0)

#else /* DGP_DBUG is not defined */
#define DGP_VOID_ASSERT(cond) \
  (void)__PERR_STATE__;
#endif /* DGP_DBUG */


/*
 * DGP_PANIC
 *  This is used for panics: internal consistency check failure.
 */
#define DGP_PANIC(err_str) \
do { \
  if(__PERR_STATE__ != NULL) __PERR_STATE__->errPrinter(__PERR_STATE__->userCookie, DGP_ERR_PRINTER_ARGS( \
    __FUNCT_NAME__, cErrPanic), (err_str)); \
  exit(EXIT_FAILURE);\
} while(0)


/*
 * DGP_PRINT_ERR
 *   For custom error handling, call the error printer with the given error info.
 */
#define DGP_PRINT_ERR(error, user_cookie, err_printer) \
do { \
  __error__ = (error); \
  (err_printer)(user_cookie, DGP_ERR_PRINTER_ARGS(__FUNCT_NAME__, __error__), Dgp_ErrStr(__error__)); \
} while(0)


/*
 * DGP_ERR(F*)
 *   Generate an error with the given error code
 */
#define DGP_ERR(error)\
do { \
  __error__ = (error);\
  DGP_ERR_PRINTER;\
  return __error__;\
} while(0)

#define DGP_ERRF0(error, err_str)\
do { \
  __error__ = (error);\
  DGP_ERR_PRINTERF0(err_str);\
  return __error__;\
} while(0)


/*
 * DGP_ERR_NO_MSG
 *   Generate an error with the given error code, but do not output anything.
 */
#define DGP_ERR_NO_MSG(error) \
do { \
  DGP_NO_PRINTER; \
  return (error); \
} while(0)


/*
 * DGP_ERR_CLEANUP(F*)
 *   Generate an error with the given error code then go to a cleanup
 *   action at the end of the function.
 */
#define DGP_ERR_CLEANUP(error)\
do { \
  __error__ = (error);\
  DGP_ERR_PRINTER;\
  goto ERR_CLEANUP;\
} while(0)

#define DGP_ERR_CLEANUPF0(error, err_str)\
do { \
  __error__ = (error);\
  DGP_ERR_PRINTERF0(err_str);\
  goto ERR_CLEANUP;\
} while(0)


/*
 * DGP_ERR_CLEANUP_NO_MSG
 *   Generate an error with the given error code, but do not output anything.
 */
#define DGP_ERR_CLEANUP_NO_MSG(error) \
do { \
  __error__ = (error);\
  DGP_NO_PRINTER; \
  goto ERR_CLEANUP; \
} while(0)


/*
 * DGP_ERR_GEN(F*)
 *   Generate an error. "cond" is some condition (probably, it's just tested
 *   for non-zero) and "error" is the error code to generate 
 */ 
#define DGP_ERR_GEN(cond, error)\
do { \
  if (cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTER;\
    return __error__;\
  }\
} while(0)

#define DGP_ERR_GENF0(cond, error, err_str)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF0(err_str);\
    return __error__;\
  }\
} while(0)

#define DGP_ERR_GENF1(cond, error, err_str, arg1)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF1(err_str, (arg1));\
    return __error__;\
  }\
} while(0)

#define DGP_ERR_GENF2(cond, error, err_str, arg1, arg2)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF2(err_str, (arg1), (arg2));\
    return __error__;\
  }\
} while(0)

#define DGP_ERR_GENF3(cond, error, err_str, arg1, arg2, arg3)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF3(err_str, (arg1), (arg2), (arg3));\
    return __error__;\
  }\
} while(0)


/*
 * DGP_ERR_GEN_NO_MSG
 *   Generate an error. "cond" is some condition (probably, it's just tested
 *   for non-zero) and "error" is the error code to generate, don't output
 *   anything.
 */ 
#define DGP_ERR_GEN_NO_MSG(cond, error)\
do { \
  if (cond) \
  {\
    __error__ = (error); \
    DGP_NO_PRINTER; \
    return __error__; \
  }\
} while(0)


/*
 * DGP_ERR_GEN_CLEANUP(F*)
 *   Error generation: instead of returning, as in DGP_ERR_GEN, go to a cleanup
 *   action at the end of the function. 
 */
#define DGP_ERR_GEN_CLEANUP(cond, error)\
do { \
  if (cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_ERR_GEN_CLEANUPF0(cond, error, err_str)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF0(err_str);\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_ERR_GEN_CLEANUPF1(cond, error, err_str, arg1)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF1((err_str), (arg1));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_ERR_GEN_CLEANUPF2(cond, error, err_str, arg1, arg2)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF2((err_str), (arg1), (arg2));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_ERR_GEN_CLEANUPF3(cond, error, err_str, arg1, arg2, arg3)\
do { \
  if(cond) \
  {\
    __error__ = (error);\
    DGP_ERR_PRINTERF3((err_str), (arg1), (arg2), (arg3));\
    goto ERR_CLEANUP;\
  }\
} while(0)


/*
 * DGP_ERR_GEN_CLEANUP_NO_MSG
 *   Generate an error. "cond" is some condition (probably, it's just tested
 *   for non-zero) and "error" is the error code to generate, don't output
 *   anything.
 */ 
#define DGP_ERR_GEN_CLEANUP_NO_MSG(cond, error)\
do { \
  if (cond) \
  {\
    __error__ = (error); \
    DGP_NO_PRINTER; \
    goto ERR_CLEANUP; \
  }\
} while(0)


/*
 * DGP_ERR_PASS(F*)
 *   This is an error check on a function call. It assumes that the
 *   function returns an error code, and if the error code is non-zero,
 *   returns from the function (after possibly printing an error message).
 */
#define DGP_ERR_PASS(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTER;\
    return __error__;\
  }\
} while (0)

#define DGP_ERR_PASSF0(functionCall, err_str)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTERF0(err_str);\
    return __error__;\
  }\
} while (0)


/*
 * DGP_ERR_PASS_NO_MSG
 *   This is an error check on a function call. It assumes that the
 *   function returns an error code, and if the error code is non-zero,
 *   returns from the function.
 */
#define DGP_ERR_PASS_NO_MSG(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_NO_PRINTER;\
    return __error__;\
  }\
} while (0)


/*
 * DGP_ERR_PASS_CLEANUP(F*)
 *   This is an error check on a function call. It is like DGP_ERR_PASS except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERR_RETURN" macro at the end of it.
 */
#define DGP_ERR_PASS_CLEANUP(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_ERR_PASS_CLEANUPF0(functionCall, err_str)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTERF0(err_str);\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_ERR_PASS_CLEANUPF1(functionCall, err_str, arg1)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTERF1((err_str), (arg1));\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_ERR_PASS_CLEANUPF2(functionCall, err_str, arg1, arg2)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTERF2((err_str), (arg1), (arg2));\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_ERR_PASS_CLEANUPF3(functionCall, err_str, arg1, arg2, arg3)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_ERR_PASS_PRINTERF3((err_str), (arg1), (arg2), (arg3));\
    goto ERR_CLEANUP;\
  }\
} while (0)


/*
 * DGP_ERR_PASS_CLEANUP_NO_MSG
 *   This is an error check on a function call. It is like DGP_ERR_PASS except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERR_RETURN" macro at the end of it.
 */
#define DGP_ERR_PASS_CLEANUP_NO_MSG(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != cSuccessful)\
  {\
    DGP_NO_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while (0)


/*
 * DGP_CHECK_ALLOC(F*)
 *   This is an error check on a memory allocation.  If the returned value is NULL,
 *   this macro returns from the function (after possibly printing an error message).
 */
#define DGP_CHECK_ALLOC(call)\
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNomem;\
    DGP_ERR_PRINTER;\
    return __error__;\
	}\
} while (0)

#define DGP_CHECK_ALLOCF0(call, err_str)\
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNomem;\
    DGP_ERR_PRINTERF0(err_str);\
    return __error__;\
	}\
} while (0)

/*
 * DGP_CHECK_ALLOC_CLEANUP(F*)
 *   This is an error check on a memory allocation.  It is like DGP_CHECK_ALLOC except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERR_RETURN" macro at the end of it.
 */
#define DGP_CHECK_ALLOC_CLEANUP(call)\
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNomem;\
    DGP_ERR_PRINTER;\
	  goto ERR_CLEANUP;\
	}\
} while (0)

#define DGP_CHECK_ALLOC_CLEANUPF0(call, err_str)\
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNomem;\
    DGP_ERR_PRINTERF0(err_str);\
	  goto ERR_CLEANUP;\
	}\
} while (0)


/*
 * DGP_CHECK_NULL
 *   Return an error if the value is null.
 */
#define DGP_CHECK_NULL(call) \
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNullPtr;\
    DGP_ERR_PRINTER;\
	  return __error__;\
	}\
} while (0)


/*
 * DGP_CHECK_NULL_CLEANUP(F*)
 *   Got to the cleanup code if the value is null.
 */
#define DGP_CHECK_NULL_CLEANUP(call) \
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNullPtr;\
    DGP_ERR_PRINTER;\
	  goto ERR_CLEANUP;\
	}\
} while (0)

#define DGP_CHECK_NULL_CLEANUPF0(call, err_str) \
do {\
	if ((call) == NULL) {\
	  __error__ = cErrNullPtr;\
    DGP_ERR_PRINTERF0(err_str);\
	  goto ERR_CLEANUP;\
	}\
} while (0)


/*
 * DGP_CHECK_NOT_NULL
 *   Return an error if the value is not null.
 */
#define DGP_CHECK_NOT_NULL(call) \
do {\
	if ((call) != NULL) {\
	  __error__ = cErrPtrNotNull;\
    DGP_ERR_PRINTER;\
	  return __error__;\
	}\
} while (0)


/*
 * DGP_CHECK_NOT_NULL_CLEANUP
 *   Got to the cleanup code if the value not is null.
 */
#define DGP_CHECK_NOT_NULL_CLEANUP(call) \
do {\
	if ((call) != NULL) {\
	  __error__ = cErrPtrNotNull;\
    DGP_ERR_PRINTER;\
	  goto ERR_CLEANUP;\
	}\
} while (0)


/*
 * DGP_ERR_CHECK
 *   This is an even simpler variant of an error check on a function call.
 *   It expects a statement or block to follow it, and will execute
 *   that statement or block if an error ocurred in the function call.
 *   This macro is used when we haven't decided whether there's really
 *   an error.
 */
#define DGP_ERR_CHECK(functionCall)\
   __error__ = (functionCall);\
   if (__error__ != cSuccessful){ \
     DGP_ERR_PRINTER;\
   } \
   if (__error__ != cSuccessful)


/*
 * DGP_ERR_CHECK_NO_MSG
 *   This is an even simpler variant of an error check on a function call.
 *   It expects a statement or block to follow it, and will execute
 *   that statement or block if an error ocurred in the function call.
 *   It differs from ERR_CHECK in that no error message is printed out.
 *   This macro is used when we haven't decided whether there's really
 *   an error.
 */
#define DGP_ERR_CHECK_NO_MSG(functionCall)\
    __error__ = (functionCall);\
    if (__error__ != cSuccessful)

/*
 * DGP_ERROR_IS
 *   True iff the current error equals the given error code.
 */
#define DGP_ERROR_IS(error) (__error__ == (error))


/*
 * DGP_ERROR_IS_NOT
 *   True iff the current error does NOT equal the given error code.
 */
#define DGP_ERROR_IS_NOT(error) (__error__ != (error))


/* return the default error string for a given digipaper, ipcore, or xif error code */
DGP_LIBEXPORT const Dgp_Int8 *
Dgp_getErrStr(ErrCode error);

#ifdef __cplusplus
}
#endif

#endif /* __DGP_ERR_CONTROL_H */




  


