/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_state.c
**
**  Description of what's implemented in this file:
**    digipaper state variable
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_state.c,v 1.32 2003/01/29 22:47:51 mkrocken Exp $
 *
 */


#include "dgp_state.h"
#include "dgp_err_control_priv.h"
#include "dgp_sm_alloc.h"
#include "PsgFix.h"

#ifndef PSG
#define StrelInfoPtr void *
#endif
#include "shrutils.pub"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_state.c,v 1.32 2003/01/29 22:47:51 mkrocken Exp $");



/* flags for the private_flags value */
#define DGPS_DEFAULT_SMALLALLOC_FLAG  0x00000001
#define DGPS_DEFAULT_ALLOCATORS_FLAG  0x00000002

#define DGPS_DEFAULT_ALLOCATORS {malloc, free, calloc, realloc}
#define DGPS_DEFAULT_SMALLOCATORS {NULL, dgp_mallocSmall, dgp_freeSmall}

#if defined EQUATOR_CA_ENV || defined PSG  /* use PSG IPCORE */

static void * 
dgp_ipc_malloc(size_t count, void *thing)
{
  return((((Dgp_Allocators *)thing)->malloc)(count));
}

static void *
dgp_ipc_calloc(size_t nel, size_t size, void *thing)
{
  return((((Dgp_Allocators *)thing)->calloc)(nel, size));
}

static void *
dgp_ipc_realloc(void *ptr, size_t size, void *thing)
{
  return((((Dgp_Allocators *)thing)->realloc)(ptr, size));
}

static void
dgp_ipc_free(void *ptr, void *thing)
{
  (((Dgp_Allocators *)thing)->free)(ptr);
}

#else /* use PARC IPCORE */

static void * 
dgp_ipc_malloc(size_t count, Bool hasPtrs, void *thing)
{
  (void)hasPtrs;
  return((((Dgp_Allocators *)thing)->malloc)(count));
}

static void *
dgp_ipc_calloc(size_t nel, size_t size, Bool hasPtrs, void *thing)
{
  (void)hasPtrs;
  return((((Dgp_Allocators *)thing)->calloc)(nel, size));
}

static void *
dgp_ipc_realloc(void *ptr, size_t size, Bool hasPtrs,void *thing)
{
  (void)hasPtrs;
  return((((Dgp_Allocators *)thing)->realloc)(ptr, size));
}

static void
dgp_ipc_free(void *ptr, size_t size, void *thing)
{
  (void)size;
  (((Dgp_Allocators *)thing)->free)(ptr);
}

#endif

static void
dgp_ipc_errPrinter(void *pThing, Dgp_Int8 *pErrorString) 
{
  DGP_PROLOGUE("dgpIpErrPrinter()", (Dgp_ErrState *)pThing);

  __error__ = cErrInternalIpcoreErr;
  DGP_ERR_PRINTERF0(pErrorString);
}

/*********************************************************************************
 *
 * Init dgp_state content to their default values. Memory is allocated for the cookie. 
 *
 * @param **dgp_state      pointer to pointer to Dgp_State
 * @param *_a              pointer to Dgp_Allocators. If NULL default value will be assigned.
 * @param *_sm_a           pointer to Dgp_SmallAllocators. If NULL defualt value will be assigned.
 * @param *userErrCookie   error output stream, can be  stderr, a certain file stream, 
 *                         or a string buffer.  
 * @param errPrinter       Dgp_ErrPrinterType. can be  Dgp_fileErrPrinter, or Dgp_strErrPrinter 
 *                         depend on the userErrCookie.
 * @param createIPWState   Flag to create internal IPWState  Most time set to DGP_TRUE.
 * @param createIPState    Flag to create internal IPState. Most time set to DGP_FALSE.
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */

ErrCode
Dgp_initDgpState(Dgp_State **_dgp_state, 
                 Dgp_Allocators *_a, 
                 Dgp_SmallAllocators *_sm_a,
                 void *userErrCookie, 
                 Dgp_ErrPrinterType errPrinter, 
                 Dgp_Bool createIPWState,
                 Dgp_Bool createIPState)
{
  Dgp_ErrState tmp_err_state;

  /* if the err printer isn't null, set up the error state for this
     function */
  DGP_PROLOGUE("Dgp_InitDgpState()", ((errPrinter != NULL) ? (&tmp_err_state) : NULL));
  Dgp_State *dgp_state = NULL;
  
  Dgp_Allocators def_a = DGPS_DEFAULT_ALLOCATORS;
  Dgp_SmallAllocators def_sm_a = DGPS_DEFAULT_SMALLOCATORS;
  Dgp_Allocators *a;
  Dgp_SmallAllocators *sm_a;

  DGP_CHECK_NULL(_dgp_state);

  tmp_err_state.userCookie = userErrCookie;
  tmp_err_state.errPrinter = errPrinter;

  /* verify the Dgp_Allocators (if provided) */
  if(_a != NULL) {
    DGP_CHECK_NULL(_a->malloc);
    DGP_CHECK_NULL(_a->calloc);
    DGP_CHECK_NULL(_a->realloc);
    DGP_CHECK_NULL(_a->free);
    a = _a;
  } else {
    a = &def_a;
  }

  /* verify the Dgp_SmallAllocators (if provided) */
  if(_sm_a != NULL) {
    DGP_CHECK_NULL(_sm_a->mallocSmall);
    DGP_CHECK_NULL(_sm_a->freeSmall);
    sm_a = _sm_a;
  } else {
    sm_a = &def_sm_a;
  }

  DGP_CHECK_ALLOC_CLEANUP(dgp_state = (Dgp_State *)(a->calloc)(1, sizeof(*dgp_state)));
  
  /* set allocators */
  dgp_state->a = *a;

  /* indicate using defaults */
  if(_a == NULL) {
    dgp_state->private_flags |= DGPS_DEFAULT_ALLOCATORS_FLAG;
  }

  /* set up err output */
  if(errPrinter != NULL) {
    DGP_ERR_PASS_CLEANUP(Dgp_initDgpErrState(&(DGP_PERRSTATE), a,
      userErrCookie, errPrinter));
  }

  /* set small allocators */
  dgp_state->sm_a = *sm_a;

  if(_sm_a == NULL) {

    /* indicate using defaults */
    dgp_state->private_flags |= DGPS_DEFAULT_SMALLALLOC_FLAG;

    /* init default small mem allocator */
    DGP_ERR_PASS_CLEANUP(dgp_initSmallMem(dgp_state, &(DGP_PSMALLCOOKIE)));

  }

  /* create ipw state if desired */
  if(createIPWState) {
    DGP_ERR_PASS_CLEANUP(Dgp_initIPWState(dgp_state));
  }
  
#ifndef EQUATOR_CA_ENV
  /* create ip state if desired */
  if(createIPState) {
    DGP_ERR_PASS_CLEANUP(Dgp_initIPState(dgp_state));
  }
#endif

  *_dgp_state = dgp_state;
  dgp_state = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  Dgp_cleanupDgpState(&dgp_state);

  ERR_RETURN;
}

/*********************************************************************************
 *
 * Clean up  dgp_state.  Memory will be freed here. 
 *
 * @param **dgp_state      pointer to pointer to Dgp_State
 * 
 * @return
 * <li>  cSuccessful
 * <li>  ErrCode
 */

ErrCode
Dgp_cleanupDgpState(Dgp_State **_dgp_state)
{
  DGP_PROLOGUE("Dgp_CleanupDgpState()", NULL);
  Dgp_Free tmpFree;
  Dgp_State *dgp_state;

  DGP_STOP_WHINE;

  if(_dgp_state == NULL) NORMAL_RETURN;

  dgp_state = *_dgp_state;
  *_dgp_state = NULL;

  if(dgp_state == NULL) NORMAL_RETURN;

  if((dgp_state->private_flags & DGPS_DEFAULT_SMALLALLOC_FLAG) != 0) {
    
    /* free our default small allocator stuff */
    dgp_cleanupSmallMem(&(DGP_PSMALLCOOKIE));

  }

  if(DGP_PIPSTATE != NULL) (DGP_FREE)(DGP_PIPSTATE);
  if(DGP_PIPWSTATE != NULL) w_IPWStateTerm(&(DGP_PIPWSTATE));
  if(DGP_PERRSTATE != NULL) Dgp_cleanupDgpErrState(&(DGP_PERRSTATE), DGP_PALLOCATORS);

  tmpFree = DGP_FREE;
  (tmpFree)(dgp_state);

  NORMAL_RETURN;
}


ErrCode
Dgp_initIPWState(Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("Dgp_InitIPWState()");

  DGP_ERR_BEGIN(dgp_state);

  if((dgp_state != NULL) && (DGP_PIPWSTATE == NULL)) {
#if defined EQUATOR_CA_ENV || defined PSG

    DGP_ERR_PASS_CLEANUP(w_IPWStateInit(
dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
dgp_ipc_errPrinter, DGP_PERRSTATE, 0, 0, NULL, NULL, &(DGP_PIPWSTATE)));

#else

    DGP_ERR_PASS_CLEANUP(w_IPWStateInit(dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
        dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
        dgp_ipc_errPrinter, DGP_PERRSTATE, 0, NULL, NULL, &(DGP_PIPWSTATE)));

#endif
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

ErrCode
Dgp_initIPState(Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("Dgp_InitIPState()");

  Int32 ipstate_size; /*when running with IPCORE-PSG compiler will complain that this doesn't get used, but that's ok*/

  DGP_ERR_BEGIN(dgp_state);

#if !defined EQUATOR_CA_ENV && !defined PSG
  

  if((dgp_state != NULL) && (DGP_PIPSTATE == NULL)) {
    DGP_ERR_PASS_CLEANUP(i_getSizeOfIPState(&ipstate_size));
    DGP_CHECK_ALLOC_CLEANUP(DGP_PIPSTATE = (IPState *)(DGP_MALLOC)(ipstate_size));
    DGP_ERR_PASS_CLEANUP(i_IPStateInit(DGP_PIPSTATE, dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
        dgp_ipc_malloc, dgp_ipc_calloc, dgp_ipc_realloc, dgp_ipc_free, DGP_PALLOCATORS,
        dgp_ipc_errPrinter, DGP_PERRSTATE, 0, NULL, NULL, ipstate_size));
  }
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  if(DGP_PIPSTATE != NULL) (DGP_FREE)(DGP_PIPSTATE);
  DGP_PIPSTATE = NULL;

  ERR_RETURN;
}


ErrCode
Dgp_cleanupIPWState(Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("Dgp_CleanupIPWState()");

  DGP_ERR_BEGIN(dgp_state);

  if((dgp_state != NULL) && (DGP_PIPWSTATE != NULL)) {
    w_IPWStateTerm(&(DGP_PIPWSTATE));
  }

  NORMAL_RETURN;
}


ErrCode
Dgp_cleanupIPState(Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("Dgp_CleanupIPState()");

  DGP_ERR_BEGIN(dgp_state);

  if((dgp_state != NULL) && (DGP_PIPSTATE != NULL)) {
    (DGP_FREE)(DGP_PIPSTATE);
    DGP_PIPSTATE = NULL;
  }

  NORMAL_RETURN;
}


ErrCode
Dgp_selectPools(Dgp_State *dgp_state, 
                Dgp_UInt32 selected_pools)
{
  DGP_STATE_PROLOGUE2("Dgp_selectPools()");

  DGP_ERR_BEGIN(dgp_state);

  if((dgp_state->private_flags & DGPS_DEFAULT_SMALLALLOC_FLAG) != 0) {
    DGP_CHECK_NULL(DGP_PSMALLCOOKIE);
    
    dgp_selectPools(DGP_PSMALLCOOKIE, selected_pools);
  }

  NORMAL_RETURN;
}

/* and their respective sizes (yeah, i know it's obvious) */
#define DGP_SMALLPOOL_MIN8_SIZE      8
#define DGP_SMALLPOOL_MIN16_SIZE    16
#define DGP_SMALLPOOL_MIN20_SIZE    20
#define DGP_SMALLPOOL_MIN24_SIZE    24
#define DGP_SMALLPOOL_MIN28_SIZE    28
#define DGP_SMALLPOOL_MIN32_SIZE    32
#define DGP_SMALLPOOL_MIN44_SIZE    44
#define DGP_SMALLPOOL_MIN128_SIZE  128

static Dgp_UInt8 chunk_sizes[DGP_NUM_SMALLPOOLS] = 
  {DGP_SMALLPOOL_MIN8_SIZE, 
   DGP_SMALLPOOL_MIN16_SIZE, 
   DGP_SMALLPOOL_MIN20_SIZE,
   DGP_SMALLPOOL_MIN24_SIZE,
   DGP_SMALLPOOL_MIN28_SIZE,
   DGP_SMALLPOOL_MIN32_SIZE,
   DGP_SMALLPOOL_MIN44_SIZE,
   DGP_SMALLPOOL_MIN128_SIZE};

Dgp_UInt32
Dgp_getPoolSize(Dgp_Int32 pool_id)
{
  if((pool_id < DGP_SMALLPOOL_MIN8) || (pool_id > DGP_SMALLPOOL_MIN128)) {
    return -1;
  }

  return((Dgp_UInt32)chunk_sizes[pool_id]);
}
