/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_state.h
**
**  Description of what's implemented in this file:
**    digipaper state definitions
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_state.h,v 1.24 2001/06/08 20:26:43 mkrocken Exp $
 */

#ifndef __DGP_STATE_H
#define __DGP_STATE_H

#include "util.h"
#include "dgp_err_control.h"
#include <libutilw.pub>

#ifdef __cplusplus
extern "C"
{
#endif


/* digipaper state cookie */
typedef struct _Dgp_State {
  Dgp_Allocators a;
  Dgp_SmallAllocators sm_a;
  Dgp_ErrState *err_state;
  IPWState *ipwstate;
  IPState *ipstate;
  Dgp_Int32 private_flags;
} Dgp_State;


/* useful macros */
#define DGP_ALLOCATORS dgp_state->a
#define DGP_PALLOCATORS (&(dgp_state->a))
#define DGP_SMALLALLOCATORS dgp_state->sm_a
#define DGP_PSMALLALLOCATORS (&(dgp_state->sm_a))
#define DGP_PERRSTATE dgp_state->err_state
#define DGP_PERRCOOKIE dgp_state->err_state->userCookie
#define DGP_PERRPRINTER dgp_state->err_state->errPrinter
#define DGP_PIPWSTATE dgp_state->ipwstate
#define DGP_PIPSTATE dgp_state->ipstate

#define DGP_MALLOC	dgp_state->a.malloc
#define DGP_FREE		dgp_state->a.free
#define DGP_CALLOC	dgp_state->a.calloc
#define DGP_REALLOC	dgp_state->a.realloc

#define DGP_PSMALLCOOKIE dgp_state->sm_a.smallCookie
#define DGP_MALLOCSMALL  dgp_state->sm_a.mallocSmall
#define DGP_FREESMALL		 dgp_state->sm_a.freeSmall


/* macros to complement the error handling stuff */

/*
 * DGP_STATE_PROLOGUE
 *   A function prologue. given a dgp_state pointer instead of a dgp_err_state
 *   pointer, it only dereferences the dgp_state if it is not null.
 */
#define DGP_STATE_PROLOGUE(funct_name, pdgp_state) \
  DGP_PROLOGUE((funct_name), (((pdgp_state) != NULL) ? ((pdgp_state)->err_state) : NULL))


/*
 * DGP_STATE_PROLOGUE2
 *   A function prologue for use with the DGP_ERR_BEGIN macro below.
 */
#define DGP_STATE_PROLOGUE2(funct_name) \
  DGP_PROLOGUE((funct_name), NULL)


/*
 * DGP_ERR_BEGIN
 *   Verify that the digipaper state is not null--if so, return an error,
 *   otherwise, set up error handling.
 */
#define DGP_ERR_BEGIN(pdgp_state) \
do { \
  DGP_ERR_GEN(((pdgp_state) == NULL), cErrInvalidParm); \
  DGP_SET_ERR_STATE((pdgp_state)->err_state); \
} while(0)



/*
 * dgp state creation and deletion functions
 *  after calling this function, you probably want
 *  to have the line:
 *    DGP_SET_ERR_STATE(DGP_PERRSTATE);
 *  to set the current error state to that newly
 *  created by this function.
 */
DGP_LIBEXPORT ErrCode
Dgp_initDgpState(Dgp_State **_dgp_state, 
                 Dgp_Allocators *a, 
                 Dgp_SmallAllocators *sm_a,
                 void *userErrCookie, 
                 Dgp_ErrPrinterType errPrinter, 
                 Dgp_Bool createIPWState, 
                 Dgp_Bool createIPState);

DGP_LIBEXPORT ErrCode
Dgp_cleanupDgpState(Dgp_State **_dgp_state);


/* dynamically create and delete ipstates and ipwstates in the 
   Dgp_State structure */
DGP_LIBEXPORT ErrCode
Dgp_initIPWState(Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
Dgp_cleanupIPWState(Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
Dgp_initIPState(Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
Dgp_cleanupIPState(Dgp_State *dgp_state);


/* change setting for internal small allocator functions.
   this is a null op if the small allocator functions
   are user defined */
#define DGP_USE_SMALLPOOL_MIN8    0x00000001
#define DGP_USE_SMALLPOOL_MIN16   0x00000002
#define DGP_USE_SMALLPOOL_MIN20   0x00000004
#define DGP_USE_SMALLPOOL_MIN24   0x00000008
#define DGP_USE_SMALLPOOL_MIN28   0x00000010
#define DGP_USE_SMALLPOOL_MIN32   0x00000020
#define DGP_USE_SMALLPOOL_MIN44   0x00000040
#define DGP_USE_SMALLPOOL_MIN128  0x00000080

DGP_LIBEXPORT ErrCode
Dgp_selectPools(Dgp_State *dgp_state, 
                Dgp_UInt32 selected_pools);


/* get the minimum number of valid bytes returned by
   a call to the small mem allocators using the 
   given pool id */
DGP_LIBEXPORT Dgp_UInt32
Dgp_getPoolSize(Dgp_Int32 pool_id);

#ifdef __cplusplus
}
#endif

#endif /* __DGP_STATE_H */


/********************************************
 * Running Log
 * $Log: dgp_state.h,v $
 * Revision 1.24  2001/06/08 20:26:43  mkrocken
 * changed small pool size from 40 to 44. This allows a group of other
 * structs to share this pool.
 *
 * Revision 1.23  2001/01/05 22:57:10  mkrocken
 * increased the size of the largest small pool from 124 to 128
 * This is required by the chains in libscancolor
 *
 * Revision 1.22  2000/11/06 22:03:13  mkrocken
 * added names of structures ( requested by scansoft )
 *
 * Revision 1.21  2000/06/08 18:47:06  jahlborn
 * small prototype changes
 *
 * Revision 1.20  2000/05/30 17:17:33  jahlborn
 * new Dgp_initDgpState() prototype
 *
 * Revision 1.19  2000/05/16 21:13:20  jahlborn
 * added small allocator suite
 *
 * Revision 1.18  2000/04/06 22:21:02  ramamurt
 * Error codes/messages added for libjpeg
 *
 * Revision 1.17  2000/03/04 00:22:21  mkrocken
 *
 * merged repositories
 *
 * Revision 1.1  2000/02/17 22:23:53  mkrocken
 * Initial Revision
 *
 * Revision 1.16  2000/01/14 22:41:27  jahlborn
 * reworked to comply with coding staqndards
 *
 * Revision 1.15  1999/12/29 20:35:11  jlin
 * Add DGP_EXPORT
 *
 *******************************************/
