/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, February 1 1998
**
**  Filename: dgp_sys_err_cotrol.h
**
**  Description of what's implemented in this file:
**    macros to simplify error reporting of system errors.
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_sys_err_control.h,v 1.6 2000/03/04 00:22:21 mkrocken Exp $
 */

#ifndef __DGP_SYS_ERR_CONTROL_H
#define __DGP_SYS_ERR_CONTROL_H

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "dgp_err_control.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * These macros wrap the system errors.  They can be used when the "errno" variable
 * is set on return from a system error.  The macros doen't test for a special
 * return value from function calls because every system call has a different way
 * of indicating an error!  so, you've got to write the tests yourself.
 */

/* 
 * Errors from 12,000 - 12,999 will be digipaper wrapped system errors. 
 */
#define cDgpSys 12000
#define maxDgpSys 12999


/*
 * Common error tests for system calls.
 */
#define IS_NEG_ONE(call)  ((call) == -1)
#define LT_ZERO(call)     ((call) < 0)
#define IS_ZERO(call)     ((call) == 0)
#define IS_NULL(call)     ((call) == NULL)
#define IS_EOF(call)      ((call) == EOF)


/*
 * DGP_SYS_ERR_GEN(F*)
 *   Generate a system error. "cond" is some condition (probably, it's just tested
 *   for -1).
 */ 
#define DGP_SYS_ERR_GEN(cond)\
do { \
  if (cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTER;\
    return __error__;\
  }\
} while(0)

#define DGP_SYS_ERR_GENF0(cond, err_str)\
do { \
  if(cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTERF0(err_str);\
    return __error__;\
  }\
} while(0)


/*
 * DGP_SYS_ERR_GEN_NO_MSG
 *   Generate a system error. "cond" is some condition (probably, it's just tested
 *   for -1).  Don't output anything.
 */ 
#define DGP_SYS_ERR_GEN_NO_MSG(cond)\
do { \
  if (cond) \
  {\
    __error__ = errno + cDgpSys; \
    DGP_NO_PRINTER; \
    return __error__; \
  }\
} while(0)


/*
 * DGP_SYS_ERR_GEN_CLEANUP(F*)
 *   System error generation: instead of returning, as in DGP_ERR_GEN, go to a cleanup
 *   action at the end of the function. 
 */
#define DGP_SYS_ERR_GEN_CLEANUP(cond)\
do { \
  if (cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_SYS_ERR_GEN_CLEANUPF0(cond, err_str)\
do { \
  if(cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTERF0(err_str);\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_SYS_ERR_GEN_CLEANUPF1(cond, err_str, arg1)\
do { \
  if(cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTERF1((err_str), (arg1));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_SYS_ERR_GEN_CLEANUPF2(cond, err_str, arg1, arg2)\
do { \
  if(cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTERF2((err_str), (arg1), (arg2));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_SYS_ERR_GEN_CLEANUPF3(cond, err_str, arg1, arg2, arg3)\
do { \
  if(cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_ERR_PRINTERF3((err_str), (arg1), (arg2), (arg3));\
    goto ERR_CLEANUP;\
  }\
} while(0)


/*
 * DGP_SYS_ERR_GEN_CLEANUP_NO_MSG
 *   Generate a system error. "cond" is some condition (probably, it's just tested
 *   for -1) and don't output anything.  This must be followed by braces {}.
 */ 
#define DGP_SYS_ERR_GEN_CLEANUP_NO_MSG(cond)\
do { \
  if (cond) \
  {\
    __error__ = errno + cDgpSys;\
    DGP_NO_PRINTER; \
    goto ERR_CLEANUP; \
  }\
} while(0)


#ifdef __cplusplus
}
#endif

#endif /* __DGP_SYS_ERR_CONTROL_H */







