
#include "ipwcore.prv"
/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"

/* PROTOTYPES */
#include "props.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.pub"

/* create a G4 (MMR) compressed buffer from a PIXR 
   This function should be part of IPCORE.
   Therefore the IPCORE typedefs are used throughout */

ErrCode CDECL
w_g4_enc_pixr (IPWState *pIPWState, PIXR* pPixr, UInt8** ppDst, 
	     UInt32* pdwBytesInDst)
{
    Int32 channel, channels;
    UInt32 rowsLeft;
    UInt32 bytesLeft;
    Int32 pixrWidth, dwWidth;
    Int32 pixrHeight, dwHeight;
    ScaledInt1616 xres, yres;
    Int32 dwRowOffset = 0, dwResolution;
    Int32 srcBPL;
    UInt32 bytesUsed = 0UL;
    Int32 bytesFilled;
    Int32 colorType;
    Int32 colorDepth;
    PIXR *pSubPixr;
    UInt8 *imageChannels[3];
    UInt8 *src, *dst;
    UInt8 continued = cFalse;
    UInt8 stage;
    UInt8 *newbuf;
    UInt32 newbufSize;
    void *compInfo = NULL;
    UInt32 imgExh, cmpExh;
    Int32 pct = 0;
    Int32 stateBlockSize;
    UInt32 padType = cNoPad;
    PROLOGUE;

    /* read the image dimensions */
    ERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    ERR_PASS_CLEANUP(w_pixrGetType(pIPWState, pPixr, &colorType));
    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, &pixrWidth));
    dwWidth = pixrWidth;
    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    dwHeight = pixrHeight;
    ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, pPixr, &colorDepth));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBPL));
    ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, pPixr, &xres, &yres));
    dwResolution = xres >> 16;

    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				       &pSubPixr));
	    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &imageChannels[channel]));
	} else {
	    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				     &imageChannels[channel]));
	}
    }

    /* If there's no memory preallocated, initialize to something */
    if (*pdwBytesInDst == 0 || *ppDst == NULL) {
      newbufSize = dwWidth * dwHeight / 8 / 10;    /* 10:1 compression */
		if (newbufSize < 10) newbufSize = 100;

        newbuf = IPW_MALLOC(newbufSize, cAllocatorData);

	/* Replace old buffer */
	if (*ppDst) {
	    IPW_FREE(*ppDst, cAllocatorData);
	}
	*ppDst = newbuf;
	*pdwBytesInDst = newbufSize;
    }
    
    rowsLeft = dwHeight;
    bytesLeft = *pdwBytesInDst;
    dst = *ppDst;

      
    /* Prepare for pre-decode */
    ERR_PASS_CLEANUP(i_getSizeOfCCITTCompressionInfo(pIPWState, 
      dwWidth + 2*cFrameBits, &stateBlockSize));
    
    /* Allocate temporary compression data, attach to image */
    compInfo = IPW_MALLOC(stateBlockSize, cAllocatorData);
    
    ERR_PASS_CLEANUP(i_initializeCCITTCompressionInfo( pIPWState, 
      compInfo, dwWidth + 2*cFrameBits, cTrue, (dwResolution > 150 ? 4 : 2), 
      padType, stateBlockSize));
  
  src = imageChannels[0] - (cFrameBits / 8);
  continued = cFalse;
  bytesUsed = 0UL;
  stage = 1;
  do {
    bytesFilled = 0;
    switch (stage) {
    case 1:
      /*compiler complains about incompatible pointer types for &imgExh
	and &cmpExh here when using
       ipcore-psg.  That's because they have defined Bool as int and we have
      it as an unsigned long int.  I don't think this will be a problem.*/
      /* Compress using MMR, don't flush yet */
      ERR_PASS_CLEANUP(i_compressRowsMMR (pIPWState, compInfo, src,
	 dwWidth + 2 * cFrameBits, dwRowOffset, rowsLeft,
         srcBPL, dst, bytesLeft, continued, continued,
	 &imgExh, &cmpExh, &pct, &bytesFilled));
      
      if (imgExh == 0) {
	ERR_GEN_CLEANUP((cmpExh == 0), 100);
      } else {
	stage++;
      }
      
      break;
    case 2:
      /* Add the two EOL markers required by TIFF G4 */
      ERR_PASS_CLEANUP(i_writeCCITTEOL (pIPWState, compInfo, cNoPad,
	 cPlainEOL, dst, bytesLeft, &bytesFilled));
      stage++;
      break;
    case 3:
      /* Flush the rest of the data */
      ERR_PASS_CLEANUP(i_flushCCITT (pIPWState,
				      compInfo, dst, bytesLeft,
				      &bytesFilled));
      stage++;
      break;
    }
    bytesUsed += bytesFilled;
    bytesLeft -= bytesFilled;
    dst += bytesFilled;
    
    /* Increase the compressed buffer size, if needed */
    if ((imgExh == 0 && cmpExh == 1) || bytesLeft < 10) {
      
      ERR_GEN_CLEANUP(imgExh && cmpExh == 0 && bytesLeft >= 10,
		       100);
      
      /* Filled the compressed buffer, reallocate the compressed
       * data buffer */
      
      /* Anticipate the compression due to the percentage, plus a
       * fudge */
      if (pct > 0 && pct <= 100) {
	newbufSize = *pdwBytesInDst * (100 + 10) / pct;
      } else {
	newbufSize = *pdwBytesInDst * (200 - pct + 10) / 100;
      }
      newbuf = IPW_MALLOC(newbufSize, cAllocatorData);
      memcpy(newbuf, *ppDst, bytesUsed);
      
      /* Replace old buffer */
      IPW_FREE(*ppDst, cAllocatorData);
      *ppDst = newbuf;
      *pdwBytesInDst = newbufSize;
      dst = newbuf + bytesUsed;
      bytesLeft = newbufSize - bytesUsed;
      continued = cTrue;
    }
  } while (imgExh == 0 || stage < 4);


  IPW_FREE (compInfo, cAllocatorData);
  
  /* Can't compress given the buffer size, reallocate and go back
   * in */
  ERR_GEN_CLEANUP((cmpExh && imgExh == 0), 100);

  *pdwBytesInDst = bytesUsed;
  /*  IPW_REALLOC (*ppDst, bytesUsed, cAllocatorData); */


  NORMAL_RETURN;

 ERR_CLEANUP:

  ERR_RETURN;

}
