#include <stdio.h>
#include <stdlib.h>

#include <dgp_state.h>
#include "image_defs.h"
#include "dgp_utils.h"
#include "halftone.h"
#include "list.h"
#include "image.h"
#include "geomadt.h"

#include "libiow.pub"
#include "iaerror.pub"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "librastw.pub"
#include "libscalw.pub"
#include "endianness.h"
#include "PsgFix.h"

#ifdef _LITTLE_ENDIAN
#define DGP_UCHAR_ACCESS(pUChar) \
    *((UInt8*)(((UInt32)(pUChar)) ^ 3))
#else 
#define DGP_UCHAR_ACCESS(pUChar) *((UInt8*)(pUChar))
#endif

/* PROTOTYPES */
#include "props.pub"
#include "rast.pub"



static Dgp_Bool overlaps(HALFTONEPIC p1, HALFTONEPIC p2  )
{
    if ( (p1.x < p2.x + p2.w) &&
           ( p2.x < p1.x + p1.w) &&
           ( p1.y < p2.y + p2.h) &&
           ( p2.y < p1.y+ p1.h) ) return DGP_TRUE;
    else return DGP_FALSE;

}

/*
 * Merge pics that have overlaping bounding boxes
 */
ErrCode
mergePics(
     Dgp_State *dgp_state,
     DgpAdt_List pics
     )
 {
    pHALFTONEPIC pic = NULL, pic2 = NULL;
    DgpAdt_ListNode ln, ln2;
    Dgp_Bool merged;

    DGP_PROLOGUE("mergePics()", DGP_PERRSTATE);
    DGP_STOP_WHINE;
    ln = DGPADT_LIFIRST(pics);
    while(ln!= NULL){
        pic = (pHALFTONEPIC)DGPADT_LIGET(ln);
        while(DGP_TRUE){
            merged = DGP_FALSE;
            ln2 = DGPADT_LIFIRST(pics);
            while ( ln2 != DGP_NULLLISTNODE ){
                pic2 = (pHALFTONEPIC)DGPADT_LIGET(ln2);
                if(pic2 != pic && (overlaps(*pic, *pic2) == DGP_TRUE)){ /* merge bounding box */
                    pic->w = MAX(pic->x + pic->w, pic2->x + pic2->w) -
                                     MIN(pic->x, pic2->x);
                    pic->h = MAX(pic->y + pic->h, pic2->y + pic2->h) -
                                     MIN(pic->y, pic2->y);
                    pic->x = MIN(pic->x, pic2->x);
                    pic->y = MIN(pic->y, pic2->y);
                    merged = DGP_TRUE;
                    DgpAdt_liRemAndNext(ln2, &ln2);
                    DGP_FREE(pic2);
                }else ln2 = DGPADT_LINEXT(ln2);
            }
            if(merged == DGP_FALSE) break;
        }
        ln = DGPADT_LINEXT(ln);
    }
    NORMAL_RETURN;
 }


/* 
 *lookup table to find number of 1s for 4 1-bit pixel 
 */
ErrCode
createTable4bppArb(Dgp_State *dgp_state,
                   Dgp_UInt32 **table)

{
    Dgp_Int32 i,shift, value;
    Dgp_UInt32 *table2=NULL;
    Dgp_Allocators *a = &(dgp_state->a);

    DGP_PROLOGUE("createTable4bppArb()", DGP_PERRSTATE);
    DGP_ASSERT(table);
    
    DGP_CHECK_ALLOC_CLEANUP(table2 = (Dgp_UInt32*)a->malloc(16*sizeof(Dgp_UInt32)));
    for(i=0; i<16; i++){
        value =0;
        for(shift=0; shift<4; shift++){
            value += (i >> shift) & 0x01;
        }
        table2[i]=value;
    }
    *table = table2;
    NORMAL_RETURN;
ERR_CLEANUP:
    if(table2) DGP_FREE(table2);
    ERR_RETURN;
}

/* 
 * Lookup table to find number of 1s in a byte
 */
ErrCode
createTable8bppArb(Dgp_State *dgp_state,
                   Dgp_UInt32 **table)
{
    Dgp_Int32 i,shift, value;
    Dgp_UInt32 *table2=NULL;
    Dgp_Allocators *a = &(dgp_state->a);

    DGP_PROLOGUE("createTable8bppArb()", DGP_PERRSTATE);
    DGP_ASSERT(table);
    
    DGP_CHECK_ALLOC_CLEANUP(table2 = (Dgp_UInt32*)a->malloc(256*sizeof(Dgp_UInt32)));
    for(i=0; i<256; i++){
        value =0;
        for(shift=0; shift<8; shift++){
            value += (i >> shift) & 0x01;
        }
        table2[i]=value;
    }
    *table = table2;
    NORMAL_RETURN;
ERR_CLEANUP:
    if(table2) DGP_FREE(table2);
    ERR_RETURN;
}

/*
 *Simple gray filter restrict gray level <= 16 for cell_size <= 4
 * and gray level <= 256 for cell_size <=8
 */
ErrCode
createFilter(
Dgp_State *dgp_state,
Dgp_Int32 cell_size,
Dgp_UInt32 **filter)
{
    Dgp_Int32 i;
    Dgp_UInt32 *table2=NULL;
    Dgp_Allocators *a = &(dgp_state->a);
    Dgp_Int32 total, half, total2;

    DGP_PROLOGUE("createFilter()", DGP_PERRSTATE);
    DGP_ASSERT(filter);
    DGP_ASSERT(cell_size >0 && cell_size <= 16);

    total = cell_size*cell_size+1;
    half = total/2;
    total2 = total-1;
    DGP_CHECK_ALLOC_CLEANUP(table2 = (Dgp_UInt32*)a->malloc(total*sizeof(Dgp_UInt32)));
    if(cell_size ==4 || cell_size ==16){
        for(i=0; i<total; i++){
            table2[i]= (i*total2 + half)/total;
        }
    }else{
        for(i=0; i<total; i++){
            table2[i]= i;
        }
    }
    *filter = table2;
    NORMAL_RETURN;
ERR_CLEANUP:
    if(table2) DGP_FREE(table2);
    ERR_RETURN;
}

/* 
 * convert 1 bpp image to 8 bpp gray image
 */
ErrCode
scaleToGray8bppArb(Dgp_State *dgp_state,
                  PIXR *inPixr,
                  Dgp_Int32 cell_size,
                  PIXR *outPixr)
{
    Dgp_Int32 wD, hD, srcBpl, dstBpl;
    Dgp_UInt8 *srcLine=NULL;
    Dgp_UInt8 *dstLine=NULL;
    Dgp_UInt32 *srcPtr;
    Dgp_Int32 i=0, j=0, row, shift, srcWpl;
    Dgp_UInt32 andbits, bitsleft=0, totalbits=0;
    Dgp_Int32  bits,  index;
    register Dgp_UInt32 value8=0;
    Dgp_UInt32 value32[16];
    Dgp_UInt32 *table = NULL;
    Dgp_UInt32 *filter = NULL;
    IPWState *pIPWState = DGP_PIPWSTATE;

    DGP_PROLOGUE("scaleToGray8bppArb()", DGP_PERRSTATE);
    /*Make sure cell size is with range */
    DGP_ASSERT(cell_size >0 && cell_size <= 16);
    DGP_ASSERT(inPixr);
    DGP_ASSERT(outPixr);

    bits = 32;
    andbits = (1 << cell_size) - 1;
    DGP_ERR_PASS_CLEANUP(createTable8bppArb(dgp_state,&table));
    DGP_ERR_PASS_CLEANUP(createFilter(dgp_state, cell_size, &filter));

    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, inPixr, &srcLine));
    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, outPixr, &dstLine));
    DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, outPixr, &hD));
    DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, outPixr, &wD));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, outPixr, &dstBpl));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, inPixr, &srcBpl));
    srcWpl = srcBpl/4;

    if(cell_size <=8){
        for(i=0; i<hD; i++){
            srcPtr = (Dgp_UInt32 *)srcLine;
            bits=0;
            while(j<wD){
                for(row =0; row < cell_size; row++){    
                    value32[row] = *(srcPtr + srcWpl * row);
                }
                /* each time we have 32 + (bits left from last word) bits */
                bits=32 + bitsleft;  
                /*while we are finding bits with width of a cell */
                while((bits -= cell_size) >=0 ){ 
                    totalbits = cell_size - bitsleft;
                    for(row = 0; row < cell_size; row++){
                        index=((value32[row] >> bits) & andbits) ;
                        value8 += totalbits - table[index]; 
                    }
                    /* write out one pixel (a byte) to the output */
                    DGP_UCHAR_ACCESS(dstLine+j) = (Dgp_UInt8)filter[value8];
                    value8 = 0;
                    bitsleft=0;
                    j++;
                }
                if(bits <0){    /* if and end of a word there still some bits left */
                    shift = -bits;
                    bitsleft = cell_size - shift;
                    for(row=0; row < cell_size; row++){
                        index = (value32[row] << shift) & andbits;
                        value8 += bitsleft -table[index] ;                 /* we don't clean this value */
                    }
                }
                srcPtr++;
            }
            j=0;
            srcLine += srcBpl*cell_size;
            dstLine += dstBpl;
        }
   }else{
        for(i=0; i<hD; i++){
            srcPtr = (Dgp_UInt32 *)srcLine;
            bits=0;
            while(j<wD){
                for(row =0; row < cell_size; row++){    
                    value32[row] = *(srcPtr + srcWpl * row);
                }
                bits=32 + bitsleft;  /* each time we have 32 + (bits left from last word) bits */
                while((bits -= cell_size) >=0 ){ /* while we are finding bits with width of a whole cell in a word */
                    totalbits = cell_size - bitsleft;
                    for(row = 0; row < cell_size; row++){
                        index=((value32[row] >> bits) & andbits) ;
                        value8 += totalbits - (table[index>>8] + table[index &0xff]); /* for cell_size >8, we look up twice */
                    }
                    /* write out one pixel ( a byte) to the output */
                    /*DGP_ERR_PASS_CLEANUP(w_pixrPutPixel (pIPWState, outPixr, j, i, value8)); */
                    DGP_UCHAR_ACCESS(dstLine+j) = (Dgp_UInt8)filter[value8];
                    value8 = 0;
                    bitsleft=0;
                    j++;
                }
                if(bits <0){    /* if and end of a word there still some bits left */
                    shift = -bits;
                    bitsleft = cell_size - shift;
                    for(row=0; row < cell_size; row++){
                        index = (value32[row] << shift) & andbits;
                        value8 += bitsleft- (table[index>>8] + table[index &0xff]) ; /* we don't clean this value */
                    }
                }
                srcPtr++;
            }
            j=0;
            srcLine += srcBpl*cell_size;
            dstLine += dstBpl;
        }
    }
    DGP_FREE(filter);
    DGP_FREE(table);
    NORMAL_RETURN;
ERR_CLEANUP:
    if(filter)
        DGP_FREE(filter);
    if(table)
        DGP_FREE(table);
    ERR_RETURN;
}

/* 
 * convert 1 bpp image to 4 bpp gray image
 */
ErrCode
scaleToGray4bppArb(Dgp_State *dgp_state,
                  PIXR *inPixr,
                  Dgp_Int32 cell_size,
                  PIXR *outPixr)
{
    Dgp_Int32 wD, hD, srcBpl, dstBpl,srcWpl;
    Dgp_UInt8 *srcLine=NULL;
    Dgp_UInt8 *dstLine=NULL;
    Dgp_UInt32 *srcPtr;
    register Dgp_Int32 i=0, j=0, row, shift;
    Dgp_UInt32 andbits, bitsleft=0, totalbits=0;
    Dgp_Int32  bits,  index;
    register Dgp_UInt32 value8=0;
    register Dgp_UInt8 byteValue=0;
    Dgp_UInt32 value32[4];
    Dgp_UInt32 *table = NULL;
    Dgp_UInt32 *filter = NULL;
    Dgp_UInt8 count=4;
    IPWState *pIPWState = DGP_PIPWSTATE;

    DGP_PROLOGUE("scaleToGray4bppArb()", DGP_PERRSTATE);
    /*Make sure cell size is with range */
    DGP_ASSERT(cell_size >0 && cell_size <=4);
    DGP_ASSERT(inPixr);
    DGP_ASSERT(outPixr);

    bits = 32;
    andbits = (1 << cell_size) - 1;
    DGP_ERR_PASS_CLEANUP(createTable4bppArb(dgp_state,&table));
    DGP_ERR_PASS_CLEANUP(createFilter(dgp_state, cell_size, &filter));

    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, inPixr, &srcLine));
    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, outPixr, &dstLine));
    DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, outPixr, &hD));
    DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, outPixr, &wD));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, outPixr, &dstBpl));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, inPixr, &srcBpl));
    srcWpl = srcBpl/4;

    for(i=0; i<hD; i++){
        srcPtr = (Dgp_UInt32 *)srcLine;
        bits=0;
        while(j<wD){
            for(row =0; row < cell_size; row++){    
                value32[row] = *(srcPtr + srcWpl * row);
            }
            bits=32 + bitsleft;  /* each time we have 32 + (bits left from last word) bits */
            while((bits -= cell_size) >=0 ){ /* while we are finding bits with width of a whole cell in a word */
                totalbits = cell_size - bitsleft;
                for(row = 0; row < cell_size; row++){
                    index=((value32[row] >> bits) & andbits) ;
                    value8 += totalbits - table[index]; 
                }
                /* write out two pixel ( a byte) to the output */
                if(count == 4){
                    byteValue += (Dgp_UInt8)filter[value8] << count;
                    count = 0;
                }else{
                    byteValue += (Dgp_UInt8)filter[value8];
                    DGP_UCHAR_ACCESS(dstLine+j/2) = byteValue;
                    byteValue =0;
                    count = 4;
                }
                value8 = 0;
                bitsleft=0;
                j++;
            }
            if(bits <0){    /* if and end of a word there still some bits left */
                shift = -bits;
                bitsleft = cell_size - shift;
                for(row=0; row < cell_size; row++){
                    index = (value32[row] << shift) & andbits;
                    value8 += bitsleft -table[index] ;                 /* we don't clean this value */
                }
            }
            srcPtr++;
        }
        j=0;
        srcLine += srcBpl*cell_size;
        dstLine += dstBpl;
    }
    DGP_FREE(filter);
    DGP_FREE(table);
    NORMAL_RETURN;
ERR_CLEANUP:
    if(filter)
        DGP_FREE(filter);
    if(table)
        DGP_FREE(table);
    ERR_RETURN;
}

/*
 * Convert 1bpp halftone image to 4pp or 8bpp gray image base on cell_size
 * for cell_size <=4, convert to 4bpp gray
 * for cell_size <=8, convert to 8bpp gray
 */
DGP_LIBEXPORT ErrCode
compressHalftoneToGray(
    Dgp_State *dgp_state,
    PIXR *inPixr,
    DgpAdt_List pics,
    Dgp_Int32 cell_size,
    DgpAdt_List grayImageList
    )
{
    DGP_PROLOGUE("compressHalftoneToGray()", DGP_PERRSTATE);
    Dgp_Int32 bpp  = 0, bppS=0;
    IPWState *pIPWState = DGP_PIPWSTATE;
    ScaledInt1616        xdpi, ydpi;    
    PIXR *prTmpIn = NULL, *prTmpOut = NULL;
    DgpAdt_ListNode ln, retnode;
    pHALFTONEPIC pic = NULL;

    DGP_CHECK_NULL(inPixr);
    DGP_CHECK_NULL(grayImageList);
    DGP_CHECK_NULL(pics);
    
    if(cell_size ==0) cell_size = 4;
    if(cell_size <=4) bpp = 4;
    else bpp = 8;

    mergePics(dgp_state, pics);
          
    DGPADT_FOREACH(ln, pics) {
        if(cell_size !=4 && cell_size !=16)
            ((pHALFTONEPIC)DGPADT_LIGET(ln))->npattern=cell_size*cell_size+1;
        else 
            ((pHALFTONEPIC)DGPADT_LIGET(ln))->npattern=cell_size*cell_size;
    }

    DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, inPixr, &xdpi, &ydpi));
    DGP_ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, inPixr, &bppS));
    
    ln = DGPADT_LIFIRST(pics);
    while(ln!= NULL){
        pic = (pHALFTONEPIC)DGPADT_LIGET(ln);
        prTmpIn = NULL;
        prTmpOut = NULL;
        /* create pixr for each input halftone area */
        DGP_ERR_PASS_CLEANUP(w_createPixr(pIPWState, 
                                          pic->w, pic->h, bppS, xdpi, ydpi, 
                                          cFrameBits, 
                                          &prTmpIn));
        DGP_ERR_PASS_CLEANUP (w_rasterop (pIPWState,
                              pic->x,
                              pic->y, 
                              inPixr,
                              PIX_SRC, 
                              0, 0, pic->w, pic->h,
                              prTmpIn));

        /*create pixr for each output halftone area */
        DGP_ERR_PASS_CLEANUP(w_createPixr(pIPWState, 
                                          pic->w / cell_size, 
                                          pic->h / cell_size, 
                                          bpp, xdpi, ydpi, 
                                          cFrameBits, 
                                          &prTmpOut));
        if(cell_size <=4){/* convert to 4bpp gray */
            DGP_ERR_PASS_CLEANUP(scaleToGray4bppArb(dgp_state, 
                                                    prTmpIn, 
                                                    cell_size, 
                                                    prTmpOut)); 
        }else if(cell_size >4){/* convert to 8bpp gray */
            DGP_ERR_PASS_CLEANUP(scaleToGray8bppArb(dgp_state, 
                                                    prTmpIn, 
                                                    cell_size, 
                                                    prTmpOut)); 
        }
#ifdef DGP_HALFTONE_DBUG
        w_writeTiff (pIPWState, prTmpOut, "grays.tif", COMPRESSION_NONE, cTrue, 0);
#endif
        DGP_ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prTmpIn));
        ln = DGPADT_LINEXT(ln);
        DgpAdt_liAdd(grayImageList, prTmpOut, &retnode);
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    if(prTmpIn) w_destroyPixr(pIPWState, &prTmpIn);
    if(prTmpOut) w_destroyPixr(pIPWState, &prTmpOut);
    ERR_RETURN;

}

/*
 * Convert into a list of halftone pixrs 
 */
DGP_LIBEXPORT ErrCode
makeListofHalftoneRegions(
    Dgp_State *dgp_state,
    PIXR *inPixr,
    DgpAdt_List pics,
    DgpAdt_List imageList
    )
{
    DGP_PROLOGUE("makeListofHalftonRegions()", DGP_PERRSTATE);
    Dgp_Int32 bppS=0;
    IPWState *pIPWState = DGP_PIPWSTATE;
    ScaledInt1616        xdpi, ydpi;    
    PIXR *prTmpIn = NULL;
    DgpAdt_ListNode ln, retnode;
    pHALFTONEPIC pic = NULL;

    DGP_CHECK_NULL(inPixr);
    DGP_CHECK_NULL(imageList);
    DGP_CHECK_NULL(pics);
               
    mergePics(dgp_state, pics);
    DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, inPixr, &xdpi, &ydpi));
    DGP_ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, inPixr, &bppS));
   
    ln = DGPADT_LIFIRST(pics);
    while(ln!= NULL){
        pic = (pHALFTONEPIC)DGPADT_LIGET(ln);
        prTmpIn = NULL;
        /* create pixr for each input halftone area */
        DGP_ERR_PASS_CLEANUP(w_createPixr(pIPWState, 
                                          pic->w, pic->h, bppS, xdpi, ydpi, 
                                          cFrameBits, 
                                          &prTmpIn));
        DGP_ERR_PASS_CLEANUP (w_rasterop (pIPWState,
                              pic->x,
                              pic->y, 
                              inPixr,
                              PIX_SRC, 
                              0, 0, pic->w, pic->h,
                              prTmpIn));
       ln = DGPADT_LINEXT(ln);
       DgpAdt_liAdd(imageList, prTmpIn, &retnode);  
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if(prTmpIn) w_destroyPixr(pIPWState, &prTmpIn);
    ERR_RETURN;

}


/* 
 * Create a bit conversion table to convert 1 4bpp image to 4 1bpp image
 */
#if 0
static ErrCode
createBitTable4To1(
	 Dgp_State *dgp_state,
     Dgp_UInt8 ** table4To1)
{
    DGP_PROLOGUE("createBitTable4To1()", DGP_PERRSTATE);
    Dgp_UInt8 *table = NULL, *tableTop, temp;
    Dgp_UInt32 i;
    Dgp_Int32 sDepth =4;
    Dgp_Allocators *a = &(dgp_state->a);

    DGP_ASSERT(table4To1);
    DGP_CHECK_ALLOC_CLEANUP(table = (Dgp_UInt8 *) a->malloc(sizeof(Dgp_UInt8 *) * 256 *4)) ;

    tableTop = table;
    for(i=0; i<256; i++){
        temp = (Dgp_UInt8)(i & 0x88);
        temp = (temp << 3) | (temp & 0xf0);
        table[i] = temp;
    }
    table += 256;
    for(i=0; i<256; i++){
        temp = (Dgp_UInt8)(i & 0x44);
        temp = ((temp & 0xf0)<<1) + (temp << 4);
        table[i] = temp;
    }
    table += 256;
    for(i=0; i<256; i++){
        temp = (Dgp_UInt8)(i & 0x22);
        temp = ((temp & 0xf0) <<2) + (temp << 5);
        table[i] = temp;
    }
    table += 256;
    for(i=0; i<256; i++){
        temp = (Dgp_UInt8)(i & 0x11);
        temp = ((temp & 0xf0) <<3) + (temp << 6);
        table[i] = temp;
    }
    *table4To1 = tableTop;
    table = NULL;
    tableTop = NULL;
    NORMAL_RETURN;
ERR_CLEANUP:
    if(table)  DGP_FREE(table);
    ERR_RETURN;
}

/*
 * convert 4bpp gray image into 4 1bpp planes by using table lookup
 */
ErrCode
convert4bppTo1bppPlanes(
	Dgp_State *dgp_state,
   	PIXR *gray_image,
    DgpAdt_List pics,
    DgpAdt_List *planes
    )
{
    DGP_PROLOGUE("convertTo1bppPlanes()", DGP_PERRSTATE);
    DgpAdt_List output;
    DgpAdt_ListNode ln;
    Dgp_Int32 srcWidth, srcHeight, sDepth, srcBpl, row, col, dstBpl,dstWidth, dstHeight;
    IPWState *pIPWState = DGP_PIPWSTATE;
    ScaledInt1616        xdpi, ydpi;    
    register Dgp_UInt8 *pS;
    Dgp_UInt8 *table, *tableTop;
    Dgp_UInt8 *dstLine[4], *srcLine,i;
    Dgp_UInt8 *pD0, *pD1, *pD2, *pD3;
    register Dgp_UInt8 srcValue0, srcValue1, srcValue2, srcValue3;
    PIXR *dstPixr = NULL;
    Dgp_UInt32 value0, value1, value2,value3, tempvalue;
    Dgp_UInt32 srcValue;
    
    DGP_ASSERT(gray_image);
    DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, gray_image, &srcHeight));
    DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, gray_image, &srcWidth));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, gray_image, &srcBpl));
    DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, gray_image, &xdpi, &ydpi));
    DGP_ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, gray_image, &sDepth));
    sDepth = 4;

    /* create look up table  */
    DGP_ERR_PASS_CLEANUP( createBitTable4To1(dgp_state, &tableTop));
    
    /* create a list of 1 bpp pixr */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&output, dgp_state));
    for(i=0; i<sDepth; i++){
        DGP_ERR_PASS_CLEANUP(w_createPixr(  pIPWState, 
                                            srcWidth, 
                                            srcHeight, 
                                            1, 
                                            xdpi, 
                                            ydpi, 
                                            32,
                                            &dstPixr));
        DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, dstPixr, &dstLine[i]));
        DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, dstPixr, &dstHeight));
        DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, dstPixr, &dstWidth));
        DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, dstPixr, &dstBpl));
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(output, (void *)dstPixr, &ln));
        dstPixr = NULL;
    }

    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, gray_image, &srcLine));

    /* convert to 1 bpp planes */
    for(row =0; row < dstHeight; row ++){
        pS = srcLine;
        pD0 = dstLine[0];
        pD1 = dstLine[1];
        pD2 = dstLine[2];
        pD3 = dstLine[3];
        for(col = 0; col< dstWidth; col += 32){
                value0 = 0;
                value1 = 0;
                value2 = 0;
                value3 = 0;
                for( i=0; i< 4; i++){   
                    table = tableTop;
                    srcValue = *((UInt32Ptr)(pS));  /* read32 bit at a time, */
                    srcValue0 = (Dgp_UInt8)((srcValue & 0xff000000) >> 24);
                    srcValue1 = (Dgp_UInt8)((srcValue & 0x00ff0000) >> 16);
                    srcValue2 = (Dgp_UInt8)((srcValue & 0x0000ff00) >> 8);
                    srcValue3 = (Dgp_UInt8)(srcValue & 0x000000ff);
                    pS +=4;

                    tempvalue = table[srcValue0] + (table[srcValue1] >>2) + (table[srcValue2] >>4) + (table[srcValue3] >>6);
                    value0 =  (value0 << 8) + tempvalue;
                    
                    table += 256;
                    tempvalue = table[srcValue0] + (table[srcValue1] >>2) + (table[srcValue2] >>4) + (table[srcValue3] >>6);
                    value1 =  (value1 << 8) + tempvalue;
                    
                    table += 256;
                    tempvalue = table[srcValue0] + (table[srcValue1] >>2) + (table[srcValue2] >>4) + (table[srcValue3] >>6);
                    value2 =  (value2 << 8) + tempvalue;
                    
                    table += 256;
                    tempvalue = table[srcValue0] + (table[srcValue1] >>2) + (table[srcValue2] >>4) + (table[srcValue3] >>6);
                    value3 =  (value3 << 8) + tempvalue;
                }
                *((UInt32Ptr)(pD0))  = value0;  /* write 32 bit at a time */
                *((UInt32Ptr)(pD1))  = value1;
                *((UInt32Ptr)(pD2))  = value2;
                *((UInt32Ptr)(pD3))  = value3;
                pD0 += 4;
                pD1 += 4;
                pD2 += 4;
                pD3 += 4;
                    
        }
        srcLine += srcBpl;
        dstLine[0] += dstBpl;
        dstLine[1] += dstBpl;
        dstLine[2] += dstBpl;
        dstLine[3] += dstBpl;
     }

     *planes = output;

    /*if (output != DGP_NULLLIST) {
        DGPADT_FOREACH (ln, output) {
            dstPixr = (PIXR *)DGPADT_LIGET(ln);
            DGP_ERR_PASS_CLEANUP(
              w_writeTiff (pIPWState, dstPixr, "planes.tif", COMPRESSION_CCITTFAX4 , cTrue, 0));
        }
    }*/
    output = NULL;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (output != DGP_NULLLIST) {
        DGPADT_FOREACH (ln, output) {
            dstPixr = (PIXR *)DGPADT_LIGET(ln);
            if(dstPixr) w_destroyPixr(pIPWState, &dstPixr);
        }
        dstPixr = NULL;
        DgpAdt_liFree(output);
    }
    ERR_RETURN;
}


/*
 * will be used by 8bppto1 convertion

          for( i=0; i< sDepth; i++) {
                pD[i] = table[srcValue[0]][i] + table[srcValue[1]][i] >>1 +
                                       table[srcValue[2]][i] >>2 + table[srcValue[3]][i] >>3 +
                                       table[srcValue[4]][i] >>4 + table[[srcValue[5]][i] >>5 +
                                       table[srcValue[6]][i]>>6 + table[srcValue[7]][i] >>7; 
            }
*/
#endif
