/* */
/* */
/*	The functions for dealing with halftoning. */
/*	Roman Yelensky	Monday, June 8 1998 */
/* */
#include <stdio.h>
#include <stdlib.h>

#include <dgp_state.h>
#include "image_defs.h"
#include "dgp_utils.h"
#include "logicops.h"
#include "morph.h"
#include "morphg.h"
#include "halftone.h"
#include "list.h"
#include "image.h"
#include "geomadt.h"

#include "libiow.pub"
#include "iaerror.pub"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "librastw.pub"
#include "imginfo.pub"
#include "halftone.pub"
#include "fill.pub"
#include "hutils.pub"
#include "prxform.pub"
#include "histeq.pub"
#include "PsgFix.h"

/* PROTOTYPES */
#include "props.pub"
#include "rast.pub"
#include "ccs.pub"
#include "reduce.pub"
#include "morph.pub"
#include "expand.pub"
#include "rectops.pub"
#include "logicops.pub"
#include "reduce.pub"
#include "intops.pub"
#include "strel.pub"
#include "shrrast.pub"
/*#include "imageref.pub" */

struct final_cookie { /*with chocolate chips */
    Int32 cell_size;
    Int32 minWidth;      /* smallest cc size to keep */
    Int32 minHeight;
    DgpAdt_List pictures;     /*we're going to keep a list of images */
    PIXR * pNewMask;
    Dgp_Allocators *a;
};

/* */
/*  Remove small CCs */
/* */
static ErrCode perCCRemoveSmall(IPWState *pIPWState,
                      void *cookie,
                      Int32 x,
                      Int32 y,
                      Int32 w,
                      Int32 h,
                      Int32 m,
                      Int32 numRuns,
                      TCScanLineRun *pRuns)
{
    PIXR *pImarkIn=((struct final_cookie *)cookie)->pNewMask;
    /*Int32 cell_size=((struct final_cookie *)cookie)->cell_size; */
    Int32 minWidth = ((struct final_cookie *)cookie)->minWidth;
    Int32 minHeight = ((struct final_cookie *)cookie)->minHeight;
    ErrCode err;

    (void)m;
    (void)numRuns;
    (void)pRuns;

    /*lets eliminate anything smaller or equal to 8x8 sets (in resolution reduced) */
    if ((w>minWidth) && (h>minHeight)) return cSuccessful;

    /*mark in our pre-final mask */
    /*luckily, rasterop doesn't puke when given invalid */
    /*coords, but rather just marks what it can. Good, */
    /*I don't have to do time consuming boundary checks here. */
    err=w_rasterop(pIPWState,
                   0,
                   0,
                   NULL,
                   PIX_CLR,
                   x-32,
                   y-32,
                   w,
                   h,
                   pImarkIn);
    if(err!=cSuccessful) {
        return err;
    }

    return cSuccessful;
}

/* */
/*	FindHTSize - */
/* */
/*		Input:  */
/*			The PIXR in which to find the halftone cell size. */
/*			The IPWState maintained by the client of this function */
/*			necessary for calls to IPCore  */
/*		Ouptut: */
/*			The determined size of the halftone cell.  */
/* */
/*		Description: */
/*			The size of the halftone cell is found by Russ Atkinson's */
/*			algorithm: */
/*			At a set of sample points in the image (possibly all) */
/*			an 8 pixel row segment is compared to the corresponding  */
/*			3 through 16 rows below it to try to find repeating patterns. */
/*			We count the repeating patterns gotten at each of these */
/*			shifts and then take the closest peak.  */
/*			Note that this means that the halftone cells we are looking */
/*			for need to be square.(or else we just get the vertical phase) */
/*			In most cases they are, so we are ok. When they aren't. */
/*			we just won't get as good compression. Note that we only go  */
/*			up to 8x8 halftone subcell. You usually never get bigger  */
/*			than 8x8 halftone subcells. The spacial resolution would just not */
/*			be there. */
/*				 */
/* */

UInt32 CountMatches(IPWState *pIPWState, 
                    PIXR *pbCompared, 
                    PIXR *pbOriginal);

UInt32 
FindHTSize(
           Dgp_State *dgp_state, 
           PIXR *pbS, 
           Dgp_Int32 MinCellSize, 
           Dgp_Int32 MaxCellSize,
           UInt8 *_cellsize) 
{  
    
    DGPERR_INIT(FindHTSize);
    IPWState *pIPWState = DGP_PIPWSTATE;
    PIXR *pImCompare=NULL;
    int rowsdown, colright; 
    Dgp_UInt8 cellsize=0;
    Dgp_UInt32 count;
    Dgp_UInt32 maxcount=150; /*so the cell_size 0 sticks if there are no images */
    /*note that this also means that if there is a valid */
    /*image and it produces less than 150  */
    /*correlations, we will judge that there is no image. */
    Dgp_Int32 sHeight,sWidth, sBpl, sDepth;
    Dgp_UInt8 *ptrIm = NULL;
    
    DGP_ERR_PASS( w_pixrGetHeight(pIPWState, pbS, &sHeight));
    DGP_ERR_PASS( w_pixrGetWidth(pIPWState, pbS, &sWidth));
    DGP_ERR_PASS( w_pixrGetBpl(pIPWState, pbS, &sBpl));
    DGP_ERR_PASS( w_pixrGetDepth(pIPWState, pbS, &sDepth));
    /*first, create the copy of the given Image we'll */
    /*use for comparisons. */
    
    DGP_ERR_PASS(w_copy(pIPWState, pbS, &pImCompare));
    
    
    /*Now start comparing image to its shifted self */
    /*to determine similarites. We are going to be  */
    /*using IPCore functions here as much as possible */
    /*because they are really fast. */
    
    /*right now, we only go up to 16 big halftone cell (8 subcell) */
    rowsdown = 0;   /* we only do horrizontal correlation */
    for (colright = MinCellSize; colright <= MaxCellSize; colright += 2) {
        if(!pImCompare)  DGP_ERR_PASS(w_copy(pIPWState, pbS, &pImCompare));
        /*Get the image ptr to the compared result */
        DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, pImCompare, &ptrIm));
        DGP_ERR_PASS_CLEANUP(w_rasterop(pIPWState, 
            0, 
            0, 
            pbS, 
            PIX_DST^PIX_SRC, 
            colright, 
            rowsdown, 
            sWidth-colright,
            sHeight-rowsdown,
            pImCompare));       
        /* Set the left strip to 1, so it won't be in the count match result*/
        DGP_ERR_PASS_CLEANUP( i_rasterOp(
            pIPWState, 
            ptrIm,	                /* dest pointer */
            cFrameBits,             /* dest rect offset*/
            cFrameBits,             /* dest rect offset*/
            colright,               /* dest rect width*/
            sHeight,	            /* dest rect height*/
            sDepth,	                /* dest depth */
            sBpl,		            /* dest bytes/line */
            PIX_SET,		        /* operation */
            NULL,                   /* src pointer */
            0, 0, 0, 0));     
        
#ifdef DGP_HALFTONE_DBUG
            DGP_ERR_PASS_CLEANUP(
                  w_writeTiff(pIPWState, pImCompare, "Correlated.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
#endif
        /*count the matches */
        count=CountMatches(pIPWState, pImCompare, pbS);	
        
        /*keeping the maximum works. */
        if (count>maxcount) {
            maxcount=count;
            cellsize=rowsdown;  
            cellsize=colright;  
            /*this is actually the haltfone cell size corresponding to */
            /*2 black and 2 white sub-cells arranged as a checkers board.  */
            /*This corresponds to the normal clustered dot ordered dither for gray  */
            /*levels, where the 45 degree grid is used.  */
        }
        
        /*our pImCompare was blown away though by the rasterop. */
        /*make a new one from the master. */
        w_destroyPixr(pIPWState, &pImCompare);
        ptrIm = NULL;
        pImCompare = NULL;
    }
    
    /*we are only dealing now with regular graylevel 45 clustered dot ordered */
    /*dither. */
    DGP_ASSERT((cellsize%2)==0); /*this check is currently redundant */
    /*because the loop increments by 2 anyway */
    
    /*we are actually returning the size of one black or white cell */
    *_cellsize = cellsize/2;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(pImCompare) w_destroyPixr(pIPWState, &pImCompare);
    ERR_RETURN;
           }
           
/* */
/*	CountMatches - */
/*		Input: The usual IPWState */
/*			The image in which to count 0 bytes */
/*			The image that was autocorrelated */
/*			their offset */
/*	 */
/*		Output: The count of 0 bytes */
/* */
/*		Description: Basically, the above function just autocorrelated */
/*			     the image we are trying to compute the halftone  */
/*			     size for at a certain displacement. It is now  */
/*			     this function's job to count those matches, */
/*			     excluding those bytes that were 0 in the  */
/*			     original image. */

Dgp_UInt32 CountMatches(IPWState *pIPWState, 
                        PIXR *pbCompared, 
                        PIXR *pbOriginal)
{ 
    
    Int32 line, colmset, fincolmset;
    Int32 sHeight,sWidth, sBpl,sWplcomp,sWplorig; 
    UInt32 *pCompLine;
    UInt32 *pOrigLine;
    UInt32 *pnextCompLine;
    UInt32 *pnextOrigLine;
    UInt32 answer=0;
    UInt32 compval;
    UInt32 origval;
    
    /*I am not going to check any of the parameters in this  */
    /*function because it will slow down an inner part of the */
    /*loop and because they were already checked above. */
    w_pixrGetHeight(pIPWState, pbCompared, &sHeight);
    w_pixrGetWidth(pIPWState, pbCompared, &sWidth);
    
    w_pixrGetBpl(pIPWState, pbCompared, &sBpl);
    sWplcomp=sBpl/sizeof(UInt32);
    
    w_pixrGetBpl(pIPWState, pbCompared, &sBpl);
    sWplorig=sBpl/sizeof(UInt32);
    
    /*we're going to be accessing the packed binary columns in sets of 32. */
    fincolmset=sWidth/32; 
    
    w_getLine(pIPWState, pbCompared, 0, (UInt8 **)&pCompLine);
    w_getLine(pIPWState, pbOriginal, 0, (UInt8 **)&pOrigLine);
    
    pnextCompLine=pCompLine+sWplcomp;
    pnextOrigLine=pOrigLine+sWplorig;
    
    
    for (line=0; line<sHeight; line++) {
        
        for (colmset=0; colmset<fincolmset; colmset++) {
            
            compval=*pCompLine;
            origval=*pOrigLine;
            
            /*if there was a match and the original thing is not all 0s */
            /*if ((*pCompLine==0) && (*pOrigLine!=0)) answer++; */
            
            answer+=((!(compval&0x000000ff)) && (origval&0x000000ff));
            answer+=((!(compval&0x0000ff00)) && (origval&0x0000ff00));
            answer+=((!(compval&0x00ff0000)) && (origval&0x00ff0000));
            answer+=((!(compval&0xff000000)) && (origval&0xff000000));
            
            /*read the next set */
            pCompLine++;
            pOrigLine++;
        }
        
        pCompLine=pnextCompLine;
        pOrigLine=pnextOrigLine;
        
        pnextCompLine+=sWplcomp;
        pnextOrigLine+=sWplorig;
    }	    
    
    return answer;
}

ErrCode
halftoneThresh (
                Dgp_State         *dgp_state,  
                PIXR              *prSrc,
                UInt32            redFactor,
                UInt32            ruleWidth,
                UInt32            redThreshold,
                UInt32            minWidth,
                UInt32            minHeight,
                RECTANGL          **ppRects,
                PIXR              **pPrResultMask)
{
    IPWState        *pIPWState = dgp_state->ipwstate;
    PIXR            *prTemp = NULL, *prSeed = NULL, *prMask = NULL;
    PIXR            *prTempMask = NULL, *prTextMask = NULL;
    RECTANGL        *rectList;
    STREL           *horizRule, *vertRule;
    STREL           *pStrel5h = NULL, *pStrel5v = NULL, *pStrel2 = NULL,
                    *pS_pepper_1 = NULL, *pS_pepper_2 = NULL, *pStrel3 = NULL;
    Bool             isZero;
    struct final_cookie rabCookie;
    void        *pNewState=NULL;
    ErrCode     err2,err3;
    Dgp_Int32   width, height;
    
    DGPERR_INIT(halftoneThresh);
    
    prTemp = prSeed = prMask = NULL;	/* initialize pointers so we
                                        * can call destroyPixr on them
    * without fear. */
    rectList = NULL;			/* ditto */
    horizRule = vertRule = NULL;	/* ditto */
    
    
    /* Validate the PIXR passed */
    /*ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1, cFrame32); */
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5h));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5h, &pStrel5h));
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel3));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel3, &pStrel3));
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5v));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5v, &pStrel5v));
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel2));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel2, &pStrel2));
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, s_pepper_1));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, s_pepper_1, &pS_pepper_1));
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, s_pepper_2));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, s_pepper_2, &pS_pepper_2));
    
    /* test the value of the redution factor */
    ERR_GEN_CLEANUP ((redFactor != 1) && (redFactor != 2) &&
        (redFactor != 4), cErrBadReduction);
    
    
    
        /* We generally want to make the seed at about 1/4 the scale
        * of the mask.  So, if redFactor is 1, we reduce the seed
        * by 4 and so on.
    */
    prTemp = NULL;
    switch (redFactor)
    {
    case 1:	/* only reduce the seed by 4x */
        DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 4, 0, 0, &prTemp));
        break;
        
    case 2:	/* reduce the seed by 8x */
        DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 1, 4, 0, &prTemp));
        break;
        
    case 4:	/* reduce the seed by 16x */
        DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 1, 1, 4, 4, &prTemp));
        break;
    }
    
    DGP_ERR_PASS_CLEANUP (w_open (pIPWState, prTemp, pStrel5h, &prTemp));
    DGP_ERR_PASS_CLEANUP (w_open (pIPWState, prTemp, pStrel5v, &prTemp));
    prSeed = NULL;
    DGP_ERR_PASS_CLEANUP (w_expand(pIPWState, prTemp, 4, &prSeed));
    
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));
    
    
    
    /* test for NULL result */
    DGP_ERR_PASS_CLEANUP (w_isZero(pIPWState, prSeed, &isZero));
    if (isZero)
    {
        goto nullFinish;
    }
    
    /* Here's where things get a little tricky.  We want to either
    * not reduce at all, reduce by 2x, 4x or 8x.  Since the reduction
    * routines take a variable number of args depending on the
    * reduction factor, we just have three cases.  We also use
    * a threshold of 3, instead of 1.  With EZScan's fancy background
    * correction code, they seem to find pretty solid blocks for
    * halftones.  Using a threshold of 1 means we tend to connect
    * text to the images, which is no good.
    */
    prMask = NULL;
    switch (redFactor)
    {
    case 1:
        /* just make a copy of the source image */
        DGP_ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prSrc, &prMask));
        DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrc, COPY, prMask));
        break;
        
    case 2: /* use 2x reduction */
        DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, redThreshold, 
            0, 0, 0, &prMask));
        break;
        
    case 4:
        DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, redThreshold, 
            redThreshold, 0, 0, &prMask));
        break;
        
    }
    
    
    
    /* Prepare to remove horizontal and vertical rules */
    prTemp = NULL;
    DGP_ERR_PASS_CLEANUP (w_copy(pIPWState, prMask, &prTemp));
    
    /* To remove the horizontal and vertical rules from the mask,
    * we have to create a couple of strels, one for horizontal
    * stuff and one for vertical stuff.  The strels have hits down
    * the centers and misses along the sides (or along the top
    * and bottom for horizontal strels).  These select things that
    * are 1/2 the distance between the rows or columns of misses.
    * Therefore, we should make the strels twice the width of
    * the rules at the current resolution.
    */
    if (ruleWidth > 0)
    {
        horizRule = NULL;
        DGP_ERR_PASS_CLEANUP (w_createHorizRuleStrel(pIPWState, ruleWidth, 
            &horizRule));
        
        vertRule = NULL;
        DGP_ERR_PASS_CLEANUP (w_createVertRuleStrel(pIPWState, ruleWidth,
            &vertRule));
       
        /* get rid of horizontal and vertical lines */
        DGP_ERR_PASS_CLEANUP (w_open (pIPWState, prMask, horizRule, &prTemp));
        
        DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prMask));
        
        
        DGP_ERR_PASS_CLEANUP (w_open (pIPWState, prMask, vertRule, &prTemp));
        
        
        DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prMask));
        
        
        /* we're done with the strels, so delete them */
        DGP_ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &horizRule));
        horizRule = NULL;
        DGP_ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &vertRule));
        vertRule = NULL;
    }
    
    /* expand seed into mask, max of 10 filling iterations */
    /*DGP_ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState, prSeed, prMask, 10,  */
    /*				&prTemp)); */
    DGP_ERR_PASS_CLEANUP (w_fillClipSweep8(pIPWState, prSeed, prMask, 1, &prTemp));
    /* test for NULL result */
    DGP_ERR_PASS_CLEANUP (w_isZero(pIPWState, prTemp, &isZero));
    if (isZero)      
    {
        /* This is not an "error", but a common result */
        goto nullFinish;
    }
    
    /* Get rid of the isolated 1x1 and 2x2 dots.  This greatly reduces the
    * amount of work rabb will do on most images.
    */
    
    /* find the single dots */
    DGP_ERR_PASS_CLEANUP (w_erode (pIPWState, prTemp, pS_pepper_1, &prMask)); 
    
    /* get rid of them */
    DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, SUBTRACT, prTemp));
    
    /* get rid of 2x1, 1x2, 2x2 */
    DGP_ERR_PASS_CLEANUP (w_erode (pIPWState, prTemp, pS_pepper_2, &prMask));
    
    /* prSeed was free */
    DGP_ERR_PASS_CLEANUP (w_dilate (pIPWState, prMask, pStrel2, &prSeed));
    
    /* get rid of those dots too */
    DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, SUBTRACT, prTemp));
    
    
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    prSeed = NULL;
#ifndef PSG /*ipcore-psg doesn't have w_getFrameMask */
    DGP_ERR_PASS_CLEANUP (w_getFrameMask( pIPWState, prTemp,
        minWidth, minHeight*2,
        &prSeed));
#endif
    
    /* Take out frame before filling hole */
    if( prSeed != NULL )
        DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, SUBTRACT, prTemp));
    /* make hole overlay image and plug the "holes" */
    /* fill from outside */    
    DGP_ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, prTemp, 1, &prMask));
    
    /* get interior pixels (holes) */
    DGP_ERR_PASS_CLEANUP (w_invert(pIPWState, prMask, &prMask));
    
    
    /* find text mask in side holes*/
    {
        Int32 sWidth, sHeight;
        
        w_pixrGetWidth (pIPWState, prSrc, &sWidth);
        w_pixrGetHeight(pIPWState, prSrc, &sHeight);
        if(redFactor !=1)
            DGP_ERR_PASS_CLEANUP (w_expand (pIPWState, prMask, redFactor, &prTempMask));
        else{
            DGP_ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prMask, &prTempMask));
            DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMask, COPY, prTempMask));
        }

        DGP_ERR_PASS_CLEANUP(w_rasterop(pIPWState, 
            0, 
            0, 
            prSrc, 
            PIX_SRC&PIX_DST, 
            0, 
            0, 
            sWidth,
            sHeight,
            prTempMask));

        DGP_ERR_PASS_CLEANUP(w_textMask(pIPWState, prTempMask, &prTextMask,0));
        
        DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTempMask));
        prTempMask = NULL;
        DGP_ERR_PASS_CLEANUP (w_isZero (pIPWState, prTextMask, &isZero));
        if (!isZero) {
        /* Because w_textMask sometimes chops off punctuation at the
        * ends of lines, make the masks bigger.
            */
#ifdef DGP_HALFTONE_DBUG
            DGP_ERR_PASS_CLEANUP(
                w_writeTiff(pIPWState, prTextMask, "textMask.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
#endif
            DGP_ERR_PASS_CLEANUP (w_reduce (pIPWState, prTextMask, 1, 1, 1, 1,
                &prTempMask));
            DGP_ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
            prTextMask = NULL;
            DGP_ERR_PASS_CLEANUP (w_dilate (pIPWState, prTempMask, pStrel3, &prTextMask));
            DGP_ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTempMask));
            prTempMask = NULL;
            DGP_ERR_PASS_CLEANUP (w_invert (pIPWState, prTextMask, &prTextMask));
            DGP_ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, prMask, prTextMask, &prMask));
            DGP_ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
        }else
            DGP_ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTextMask));
    }
    
    /* plug the holes */
    DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, OR, prMask));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));
    

    /* Put frame back to mask */
    if( prSeed )
    {
        DGP_ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSeed, OR, prMask));
        DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    }

    /*what IPCore wants */
    DGP_ERR_PASS_CLEANUP(w_createRabbState(pIPWState, &pNewState));
    
    /*set up the second cookie. */
    rabCookie.pNewMask = prMask;
    rabCookie.minWidth = minWidth;
    rabCookie.minHeight = minHeight;
    
    w_pixrGetWidth(pIPWState, prMask, &width);
    w_pixrGetHeight(pIPWState, prMask, &height);
    /* try to remove small stuff in the hole pluged mask */
    DGP_ERR_PASS_CLEANUP(w_rabbRunsLite(
        pIPWState,
        prMask,
        pNewState,          /*void          *pState */
        0,                  /*UInt32         connectType */
        0,                  /*Int32          clipX */
        0,                  /*Int32          clipY */
        width,            /*Int32          clipW */
        height,           /*Int32          clipH */
        1,                  /*UInt32         minWidth */
        1,                  /*UInt32         minHeight */
        NULL,               /*void          *pRunData */
        NULL,               /*Int32          (CDECL *runFindingProc)(IPWState *pIPWState, */
        &rabCookie,         /*void          *pCallbackData */
        perCCRemoveSmall,   /*Int32          (CDECL *callbackProc)(IPWState *pIPWState, */
        &err2,              /*Int32         *pRunReturnCode */
        &err3               /*Int32         *pCallbackReturnCode */
        ));

    /* make and save list of rectangles for the connected components.*/
    rectList = NULL;
    DGP_ERR_PASS_CLEANUP (w_rabb(pIPWState, prMask, minWidth, minHeight, 
			     &rectList));
    if (rectList == NULL)
    {
        /* This isn't an error, so don't complain. */
        goto nullFinish;
    }
    
    DGP_ERR_PASS_CLEANUP (w_xformRectList (pIPWState, 0, 0,
        (Float32) redFactor, 
        (Float32) redFactor, 
        rectList));
    
    /* return the list of rectangles */
    *ppRects = rectList;
    
    /* expand mask to full size */
    if (redFactor > 1)
    {
        prTemp = NULL;
        DGP_ERR_PASS_CLEANUP (w_expand(pIPWState, prMask, redFactor, &prTemp));
        DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prMask));
        prMask = NULL;
        *pPrResultMask = prTemp;
    }
    else	/* redFactor == 1 */
    {
        *pPrResultMask = prMask;
    }
    NORMAL_RETURN;
    
#if  DEBUG_HT_THRESH
    {
        PIXR     *prt11;
        if ((prt11 = copyPixr(prTemp)) == NULL)
            return abortP("prt11 not made", procName, NULL);
        opPrPr(prt11, prSrc, AND);  /* halftone image part */
        writePr("c:\\temp\\junkhalftone", prt11, 1);
        opPrPr(prt11, prSrc, XOR);  /* non-image part */
        writePr("c:\\temp\\junkother", prt11, 1);
        DESTROY_PIXR(prt11);
    }
#endif  /* DEBUG_HT_THRESH */
    
    /* 
    * In this case, we didn't really failed, we just didn't get
    * a result. 
    */
nullFinish:
    /* clean up everything */
    if (horizRule != NULL)
        DGP_ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &horizRule));
    if (vertRule != NULL)
        DGP_ERR_PASS_CLEANUP (w_destroyStrel(pIPWState, &vertRule));
    
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prSeed));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prMask));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prTemp));
    if (rectList != NULL)
        DGP_ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));
    *ppRects = NULL;
    *pPrResultMask = NULL;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    /* clean up everything */
    if (horizRule != NULL)
        w_destroyStrel (pIPWState, &horizRule);
    if (vertRule != NULL)
        w_destroyStrel (pIPWState, &vertRule);
    
    if(prSeed)         w_destroyPixr (pIPWState, &prSeed);
    if(prMask)         w_destroyPixr (pIPWState, &prMask);
    if(prTemp)         w_destroyPixr (pIPWState, &prTemp);
    if(prTempMask)     w_destroyPixr (pIPWState, &prTempMask);
    if(prTextMask)     w_destroyPixr (pIPWState, &prTextMask);
    if (rectList != NULL)
        w_destroyRectList(pIPWState, &rectList);
    *ppRects = NULL;
    
    ERR_RETURN;
}


/*************************************************************************
*
*   Halftone segmantaion
*************************************************************************/
ErrCode
GetAllHalftoneInfo(
                   Dgp_State *dgp_state, 
                   PIXR *pbS, 
                   PIXR **ppMask, 
                   DgpAdt_List *hInfo, 
                   UInt8 *cell_size,
                   Dgp_Bool *_reverseVideo) 
{
    DGPERR_INIT(GetAllHalftoneInfo);
    
    IPWState    *pIPWState = dgp_state->ipwstate;
    Dgp_Int32   sHeight,sWidth,sDepth,sChannels,dpi=0;
    Dgp_Int32   rrHeight, rrWidth;
    Dgp_Int32   redFactor, minWidth, minHeight, ruleWidth, redThreshold = 1;
    Dgp_Int32   MinCellSize, MaxCellSize;
    Dgp_Int32   minTextAreaWidth, minTextAreaHeight;
    ScaledInt1616 xdpi, ydpi;
    
    STREL       *pStrel5h=NULL, *pStrel5v=NULL;
    ConnectedComponent *pCCList = NULL, *p;
    PIXR        *prMask = NULL;
    PIXR        *prMyCopy=NULL;     
    PIXR        *prTempMask=NULL; 
    Dgp_Bool    reverseVideo = DGP_FALSE;
    RECTANGL     *pRect = NULL, *rectList = NULL;
    Dgp_Int32 mass, i;
    
    /* related to convert RECTANGL to list of pHALFTONEPIC */
    DgpAdt_List picList = DGP_NULLLIST;
    DgpAdt_ListNode retnode;
    pHALFTONEPIC ppic = NULL;
    
    /*Assertion */
    DGP_ASSERT(pbS);
    DGP_ASSERT(ppMask);
    
    /*Initialize */
    *cell_size = 0;
    *ppMask = NULL;
    /*if the user wanted us to allocate his mask */
    mass = 0;
    
    /*Get info from pixr */
    DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pbS, &sHeight));
    DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pbS, &sWidth));
    DGP_ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, pbS, &sDepth));
    DGP_ERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pbS, &sChannels));
    
    DGP_ASSERT(sChannels==1);
    DGP_ASSERT(sDepth==1);      /*This is indeed a binary */
    
    DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, pbS, &xdpi, &ydpi));
    dpi = MAX (ScaledInt1616_TO_Int32 (xdpi), ScaledInt1616_TO_Int32 (ydpi));
    if (dpi == 0) {
        dpi = 300;
    }
    
    /* Initiate structure element */
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5h));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5h, &pStrel5h));
    
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5v));
    DGP_ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5v, &pStrel5v));
    
    MinCellSize = 4;
    
    if (dpi >= 300) {
        redFactor = 4;
        MaxCellSize = 16;
        minWidth = 99;
        minHeight = 15;
        ruleWidth = 8;
        minTextAreaWidth = dpi/50 *5; /*(min 5 letter); */
        minTextAreaHeight = dpi/25; /*(one line); */
    }else if (dpi >= 200)	{
        redFactor = 2;
        MaxCellSize = 12;
        minWidth = 66;
        minHeight = 10;
        ruleWidth = 4;
        minTextAreaWidth = dpi/50 *5; /*(min 5 letter); */
        minTextAreaHeight = dpi/25; /*(one line); */
    } else {
        redFactor = 1;
        MaxCellSize = 12;
        minWidth = 33;
        minHeight = 5;
        ruleWidth = 1;
        minTextAreaWidth = 200/50 *5; /*(min 5 letter); */
        minTextAreaHeight = 200/25; /*(one line); */
    }
    /* find out if the image is reverse video */
    DGP_ERR_PASS_CLEANUP (w_reduce(pIPWState, pbS, 4, 4, 4, 1, &prTempMask));
    DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, prTempMask, &rrWidth));
    DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, prTempMask, &rrHeight));
    w_open(pIPWState, prTempMask, pStrel5h, &prTempMask); 
    w_open(pIPWState, prTempMask, pStrel5v, &prTempMask); 
    DGP_ERR_PASS_CLEANUP (w_findConnectedComponents (pIPWState, prTempMask,minWidth, minHeight,0, &pCCList));
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prTempMask));
    prTempMask = NULL;
    
    if (pCCList)
    {
        for ( p = pCCList; p; p = p->next)
        {
            if( p->width == rrWidth && p->height == rrHeight) {
                for(i=0; i< p->runCount; i++){
                    mass += p->runs[i].runRight - p->runs[i].runLeft;
                }
                if(mass > rrWidth * rrHeight /3)  reverseVideo = DGP_TRUE;
                break;
            }
        }
    }
    DGP_ERR_PASS_CLEANUP (w_destroyConnectedComponents (pIPWState, &pCCList));
    if(reverseVideo == DGP_TRUE) {    
        DGP_ERR_PASS_CLEANUP(w_invert(pIPWState, pbS, &prMyCopy));
    }
    else  DGP_ERR_PASS_CLEANUP(w_copy(pIPWState, pbS, &prMyCopy));
    
    /*halftone segmentation by threshold-reduction and seed-mask filling */
    DGP_ERR_PASS_CLEANUP( halftoneThresh (
        dgp_state,
        prMyCopy,
        redFactor,
        ruleWidth,
        redThreshold,
        minWidth,
        minHeight,
        &rectList,
        &prMask));
    /*return if there is no halftone */
    if(!prMask){
        *hInfo = NULL;
        *ppMask = NULL;
        *cell_size=0;
        *_reverseVideo = reverseVideo;
        if(prMyCopy) w_destroyPixr(pIPWState, &prMyCopy);
        if (rectList)
            DGP_ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));
        
        NORMAL_RETURN;
    }
#ifdef DGP_HALFTONE_DBUG
       DGP_ERR_PASS_CLEANUP(
            w_writeTiff(pIPWState, prMask, "hMask.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
#endif
    /* get the halftone image */
    DGP_ERR_PASS_CLEANUP(w_rasterop(pIPWState, 
        0, 
        0, 
        prMask, 
        PIX_SRC&PIX_DST, 
        0, 
        0, 
        sWidth,
        sHeight,
        prMyCopy));
    
    /* find halftone image correlation */
    DGP_ERR_PASS_CLEANUP(FindHTSize(dgp_state, prMyCopy, MinCellSize, MaxCellSize, cell_size)); 
    
    /*convert pRects to pics */
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&picList, dgp_state));
    pRect = rectList;
    while(pRect){
        DGP_CHECK_ALLOC_CLEANUP(ppic = (pHALFTONEPIC) DGP_MALLOC(sizeof(HALFTONEPIC)));    
        ppic->cell_size = *cell_size;
        ppic->w = pRect->w;
        ppic->h = pRect->h;
        ppic->x = pRect->x;
        ppic->y = pRect->y;
        pRect = pRect->next;
        DgpAdt_liAdd(picList, ppic, &retnode);
    }
    if (rectList != NULL)
        DGP_ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));
    if (prMyCopy) 
        w_destroyPixr(pIPWState, &prMyCopy);

    *hInfo = picList;
    *ppMask = prMask;
    *_reverseVideo = reverseVideo;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (prMyCopy) 
        w_destroyPixr(pIPWState, &prMyCopy);
    if (prTempMask) 
        w_destroyPixr(pIPWState, &prTempMask);
    if(prMask)
        w_destroyPixr(pIPWState, &prMask);
    if (rectList != NULL)
        DGP_ERR_PASS_CLEANUP (w_destroyRectList(pIPWState, &rectList));
    if(picList){
        DgpAdt_ListNode ln;
        DGPADT_FOREACH (ln, picList) {
            ppic = (pHALFTONEPIC)DGPADT_LIGET(ln);
            DGP_FREE(ppic);
        }
        DgpAdt_liFree(picList);
    }
    ERR_RETURN;
}
