/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/imconv.c,v 1.2 2003/01/27 21:48:32 awilliam Exp $
 *
 * Copyright (c) 1990, 1991 Cornell University.  All Rights Reserved.
 *
 * Copyright (c) 1991 Xerox Corporation.  All Rights Reserved.
 *
 * Use, reproduction and preparation of derivative works of this software is
 * permitted.  Any copy of this software or of any derivative work must
 * include both the above copyright notices of Cornell University and Xerox
 * Corporation and this paragraph.  This software is made available AS IS, and
 * XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
 * WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
 * HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

#include "util.h"
#include "image.h"
#include "image_defs.h"
#include "imconv.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/imconv.c,v 1.2 2003/01/27 21:48:32 awilliam Exp $");


/*
 *	imconv.c - Image conversion.
 *
 * This file contains routines to convert between various forms of Image.
 *
 * Exports:
 *	DgpImage_Gray imcLtoG(DgpImage_Long lim, Dgp_Int32 low, Dgp_Int32 high) - convert the
 *		given DgpImage_Long to a DgpImage_Gray, with low mapping to zero and
 *		high mapping to 255; values in between scaled linearly.
 *		If low >= high, all values are set to zero. The output
 *		image has the same size and origin.
 *
 *	DgpImage_Gray imcLtoGAuto(DgpImage_Long lim) - convert the given DgpImage_Long
 *		to a DgpImage_Gray with the largest value mapping to 255 and
 *		the smallest to 0.
 *
 *	DgpImage_Gray imcFtoG(DgpImage_Float fim, Dgp_Float64 low, Dgp_Float64 high) -
 *		convert the given DgpImage_Float to a DgpImage_Gray with low
 *		mapping to 0 and high mapping to 255.
 *
 *	DgpImage_Gray imcFtoGAuto(DgpImage_Float fim) - convert the given DgpImage_Float
 *		to a DgpImage_Gray with the largest value mapping to 255 and
 *		the smallest to 0.
 *
 *	DgpImage_Gray imcDtoG(DgpImage_Double fim, Dgp_Float64 low, Dgp_Float64 high) -
 *		convert the given DgpImage_Double to a DgpImage_Gray with low
 *		mapping to 0 and high mapping to 255.
 *
 *	DgpImage_Gray imcDtoGAuto(DgpImage_Double fim) - convert the given DgpImage_Double
 *		to a DgpImage_Gray with the largest value mapping to 255 and
 *		the smallest to 0.
 *
 *	DgpImage_Gray imcRGBtoG(DgpImage_ColorRGB rgbim) - convert the given RGB
 *		image to a DgpImage_Gray by computing an appropriate weighted
 *		average of the R, G and B components.
 *
 *	DgpImage_Double imcGtoD(DgpImage_Gray gim) - convert the given DgpImage_Gray
 *		to a DgpImage_Double by simply copying the values across.
 *
 * All routines here return NULL (cast to the appropriate type) on
 *	failure.
 *
 */

ErrCode
DgpImage_imcLtoG(Dgp_State *dgp_state, DgpImage_Long lim, Dgp_Int32 low, Dgp_Int32 high, DgpImage_Gray *gim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;
    Dgp_Float64 conv;

	DGP_PROLOGUE ("DgpImage_imcLtoG", DGP_PERRSTATE);

    DGP_ASSERT(lim != (DgpImage_Long)NULL);

    DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_GRAY, DGPIMG_WIDTH(lim), 
			DGPIMG_HEIGHT(lim), DGPIMG_XBASE(lim), DGPIMG_YBASE(lim), (void**) &gim));

    if (high <= low) {
	/* All done: it's already all zero */
	NORMAL_RETURN;
	}

    top = DGPIMG_HEIGHT(lim) + DGPIMG_YBASE(lim);
    right = DGPIMG_WIDTH(lim) + DGPIMG_XBASE(lim);
    for (y = DGPIMG_YBASE(lim); y < top; y++) {
	for (x = DGPIMG_XBASE(lim); x < right; x++) {
	    conv = ((DGPIMG_REF(lim, x, y) - low) * DGPIMG_COLRNG) / (high - low);

	    if (conv < 0) {
		conv = 0;
		}
	    else if (conv >= DGPIMG_COLRNG) {
		conv = DGPIMG_COLRNG - 1;
		}
	    DGPIMG_REF(*gim, x, y) = (Dgp_UInt8) conv;
	    }
	}

    NORMAL_RETURN;
}

ErrCode
DgpImage_imcLtoGAuto(Dgp_State *dgp_state, DgpImage_Long lim, DgpImage_Gray *gim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 low;
    Dgp_Int32 high;
    Dgp_Int32 top;
    Dgp_Int32 right;

	DGP_PROLOGUE ("DgpImage_imcLtoGAuto", DGP_PERRSTATE);

    DGP_ASSERT(lim != (DgpImage_Long)NULL);

    if ((DGPIMG_WIDTH(lim) != 0) && (DGPIMG_HEIGHT(lim) != 0)) {
	low = high = DGPIMG_REF(lim, DGPIMG_XBASE(lim), DGPIMG_YBASE(lim));
	}
    else {
	DGP_ERR_PASS (DgpImage_imcLtoG(dgp_state, lim, 0, 0, gim));
	NORMAL_RETURN;
	}
    top = DGPIMG_HEIGHT(lim) + DGPIMG_YBASE(lim);
    right = DGPIMG_WIDTH(lim) + DGPIMG_XBASE(lim);
    for (y = DGPIMG_YBASE(lim); y < top; y++) {
	for (x = DGPIMG_XBASE(lim); x < right; x++) {
	    low = MIN(low, DGPIMG_REF(lim, x, y));
	    high = MAX(high, DGPIMG_REF(lim, x, y));
	    }
	}

    DGP_ERR_PASS (DgpImage_imcLtoG(dgp_state, lim, low, high, gim));
	NORMAL_RETURN;
}

/*
 * This routine converts a DgpImage_Float into a DgpImage_Gray. It scales the
 * float values to the range (0..DGPIMG_COLRNG-1), with values at low or below
 * going to zero, values at high or above going to DGPIMG_COLRNG-1, and values
 * in between to the appropriate linearly interpolated value.
 *
 * It returns the newly created DgpImage_Gray on success, and (DgpImage_Gray)NULL
 * on failure.
 */
ErrCode
DgpImage_imcFtoG(Dgp_State *dgp_state, DgpImage_Float fim, Dgp_Float64 low, Dgp_Float64 high, DgpImage_Gray *gim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;
    Dgp_Int32 conv;

	DGP_PROLOGUE ("DgpImage_imcFtoG", DGP_PERRSTATE);

    DGP_ASSERT(fim != (DgpImage_Float)NULL);

    /* Allocate the new image */
	DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_GRAY, DGPIMG_WIDTH(fim), DGPIMG_HEIGHT(fim),
		      DGPIMG_XBASE(fim), DGPIMG_YBASE(fim), (void**) gim));

    if (low >= high) {
	/*
	 * Nothing to be done - there's nothing there to convert
	 * (or rather, it's all the same, and will be squashed to
	 * zero, which is what is in the image now, as it's a new image).
	 */
	NORMAL_RETURN;
	}
	
    /* Convert the image */
    top = DGPIMG_HEIGHT(fim) + DGPIMG_YBASE(fim);
    right = DGPIMG_WIDTH(fim) + DGPIMG_XBASE(fim);
    for (y = DGPIMG_YBASE(fim); y < top; y++) {
	for (x = DGPIMG_XBASE(fim); x < right; x++) {
	    conv = (int)((DGPIMG_REF(fim, x, y) - low) / (high - low) * DGPIMG_COLRNG);

	    /* Check to see if it went out of range */
	    if (conv < 0) {
		conv = 0;
		}
	    else if (conv >= DGPIMG_COLRNG) {
		conv = DGPIMG_COLRNG - 1;
		}
	    DGPIMG_REF(*gim, x, y) = (Dgp_UInt8) conv;
	    }
	}

    NORMAL_RETURN;
}

/*
 * This routine converts a DgpImage_Float into a DgpImage_Gray, with the lowest
 * value in the DgpImage_Float becoming 0, and the highest value becoming
 * DGPIMG_COLRNG-1.
 *
 * It returns the new DgpImage_Gray on success, and (DgpImage_Gray)NULL on failure.
 */
ErrCode
DgpImage_imcFtoGAuto(Dgp_State *dgp_state, DgpImage_Float fim, DgpImage_Gray *gim)
{
    Dgp_Float64 low, high;
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;

	DGP_PROLOGUE ("DgpImage_imcFtoGAuto", DGP_PERRSTATE);

    DGP_ASSERT(fim != (DgpImage_Float)NULL);

    if ((DGPIMG_WIDTH(fim) != 0) && (DGPIMG_HEIGHT(fim) != 0)) {
	low = high = DGPIMG_REF(fim, DGPIMG_XBASE(fim), DGPIMG_YBASE(fim));
	}
    else {
	DGP_ERR_PASS (DgpImage_imcFtoG(dgp_state, fim, 0., 0., gim));
	NORMAL_RETURN;
	}

    top = DGPIMG_HEIGHT(fim) + DGPIMG_YBASE(fim);
    right = DGPIMG_WIDTH(fim) + DGPIMG_XBASE(fim);
    for (y = DGPIMG_YBASE(fim); y < top; y++) {
	for (x = DGPIMG_XBASE(fim); x < right; x++) {
	    low = MIN(low, DGPIMG_REF(fim, x, y));
	    high = MAX(high, DGPIMG_REF(fim, x, y));
	    }
	}

    /* Do the scaling itself */
    DGP_ERR_PASS (DgpImage_imcFtoG(dgp_state, fim, low, high, gim));
	NORMAL_RETURN;
}

/*
 * This routine converts a DgpImage_Double into a DgpImage_Gray. It scales the
 * Dgp_Float64 values to the range (0..DGPIMG_COLRNG-1), with values at low or below
 * going to zero, values at high or above going to DGPIMG_COLRNG-1, and values
 * in between to the appropriate linearly interpolated value.
 *
 * It returns the newly created DgpImage_Gray on success, and (DgpImage_Gray)NULL
 * on failure.
 */
ErrCode
DgpImage_imcDtoG(Dgp_State *dgp_state, DgpImage_Double dim, Dgp_Float64 low, Dgp_Float64 high, DgpImage_Gray *gim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;
    Dgp_Int32 conv;

	DGP_PROLOGUE ("DgpImage_imcDtoG", DGP_PERRSTATE);
    
	DGP_ASSERT(dim != (DgpImage_Double)NULL);

    /* Allocate the new image */
    DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_GRAY, DGPIMG_WIDTH(dim), DGPIMG_HEIGHT(dim),
		      DGPIMG_XBASE(dim), DGPIMG_YBASE(dim), (void**) gim));

    if (low >= high) {
	/*
	 * Nothing to be done - there's nothing there to convert
	 * (or rather, it's all the same, and will be squashed to
	 * zero, which is what is in the image now, as it's a new image).
	 */
	NORMAL_RETURN;
	}
	
    /* Convert the image */
    top = DGPIMG_HEIGHT(dim) + DGPIMG_YBASE(dim);
    right = DGPIMG_WIDTH(dim) + DGPIMG_XBASE(dim);
    for (y = DGPIMG_YBASE(dim); y < top; y++) {
	for (x = DGPIMG_XBASE(dim); x < right; x++) {
	    conv = (Dgp_UInt8)((DGPIMG_REF(dim, x, y) - low) / (high - low) * DGPIMG_COLRNG);

	    /* Check to see if it went out of range */
	    if (conv < 0) {
		conv = 0;
		}
	    else if (conv >= DGPIMG_COLRNG) {
		conv = DGPIMG_COLRNG - 1;
		}
	    DGPIMG_REF(*gim, x, y) = (Dgp_UInt8) conv;
	    }
	}

    NORMAL_RETURN;
}

/*
 * This routine converts a DgpImage_Double into a DgpImage_Gray, with the lowest
 * value in the DgpImage_Double becoming 0, and the highest value becoming
 * DGPIMG_COLRNG-1.
 *
 * It returns the new DgpImage_Gray on success, and (DgpImage_Gray)NULL on failure.
 */
ErrCode
DgpImage_imcDtoGAuto(Dgp_State *dgp_state, DgpImage_Double dim, DgpImage_Gray *gim)
{
    Dgp_Float64 low, high;
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;

	DGP_PROLOGUE ("DgpImage_imcDtoGAuto", DGP_PERRSTATE);

    DGP_ASSERT(dim != (DgpImage_Double)NULL);

    if ((DGPIMG_WIDTH(dim) != 0) && (DGPIMG_HEIGHT(dim) != 0)) {
	low = high = DGPIMG_REF(dim, DGPIMG_XBASE(dim), DGPIMG_YBASE(dim));
	}
    else {
	DGP_ERR_PASS (DgpImage_imcDtoG(dgp_state, dim, 0., 0., gim));
	NORMAL_RETURN;
	}

    top = DGPIMG_HEIGHT(dim) + DGPIMG_YBASE(dim);
    right = DGPIMG_WIDTH(dim) + DGPIMG_XBASE(dim);
    for (y = DGPIMG_YBASE(dim); y < top; y++) {
	for (x = DGPIMG_XBASE(dim); x < right; x++) {
	    low = MIN(low, DGPIMG_REF(dim, x, y));
	    high = MAX(high, DGPIMG_REF(dim, x, y));
	    }
	}

    /* Do the scaling itself */
    DGP_ERR_PASS (DgpImage_imcDtoG(dgp_state, dim, low, high, gim));
	NORMAL_RETURN;
}

/*
 * These weights are magic numbers which apparently are appropriate
 * for converting RGB values into a single gray-level value.
 */
#define	RED_WEIGHT	(0.299)
#define GREEN_WEIGHT	(0.587)
#define BLUE_WEIGHT	(0.114)

/*
 * This routine converts an RGB image into a DgpImage_Gray. It does this
 * by computing, for each pixel, a weighted average of the R, G, and B
 * components of the input image.
 *
 * It returns the new DgpImage_Gray on success, and (DgpImage_Gray)NULL on
 * failure.
 */
ErrCode
DgpImage_imcRGBtoG(Dgp_State *dgp_state, DgpImage_ColorRGB rgbim, DgpImage_Gray *gim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;

	DGP_PROLOGUE ("DgpImage_imcRGBtoG", DGP_PERRSTATE);

    DGP_ASSERT(rgbim != (DgpImage_ColorRGB)NULL);

    /* Allocate the new image */
	DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_GRAY, DGPIMG_WIDTH(rgbim), DGPIMG_HEIGHT(rgbim),
		      DGPIMG_XBASE(rgbim), DGPIMG_YBASE(rgbim), (void**) gim));

    /* Compute the values */
    top = DGPIMG_HEIGHT(rgbim) + DGPIMG_YBASE(rgbim);
    right = DGPIMG_WIDTH(rgbim) + DGPIMG_XBASE(rgbim);
    for (y = DGPIMG_YBASE(rgbim); y < top; y++) {
	for (x = DGPIMG_XBASE(rgbim); x < right; x++) {
	    DGPIMG_REF(*gim, x, y) = (Dgp_UInt8) 
			(DGPIMG_REF(rgbim, x, y).r * RED_WEIGHT +
		               DGPIMG_REF(rgbim, x, y).g * GREEN_WEIGHT +
			       DGPIMG_REF(rgbim, x, y).b * BLUE_WEIGHT + 0.5);
	    }
	}

    /* Return the new image */
    NORMAL_RETURN;
}

/*
 * This routine converts a DgpImage_Gray to a DgpImage_Double by simply copying
 * the values across.
 *
 * It returns the new DgpImage_Double on success, and (DgpImage_Double)NULL on
 * failure.
 */
ErrCode
DgpImage_imcGtoD(Dgp_State *dgp_state, DgpImage_Gray gim, DgpImage_Double *dim)
{
    Dgp_Int32 x, y;
    Dgp_Int32 top;
    Dgp_Int32 right;

	DGP_PROLOGUE ("DgpImage_imcGtoD", DGP_PERRSTATE);

    DGP_ASSERT(gim != (DgpImage_Gray)NULL);

    /* Allocate the new image */
    DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_DOUBLE,
				   DGPIMG_WIDTH(gim), DGPIMG_HEIGHT(gim),
				   DGPIMG_XBASE(gim), DGPIMG_YBASE(gim), (void**) dim));

    /* Compute the values */
    top = DGPIMG_HEIGHT(gim) + DGPIMG_YBASE(gim);
    right = DGPIMG_WIDTH(gim) + DGPIMG_XBASE(gim);
    for (y = DGPIMG_YBASE(gim); y < top; y++) {
	for (x = DGPIMG_XBASE(gim); x < right; x++) {
	    DGPIMG_REF(*dim, x, y) = DGPIMG_REF(gim, x, y);
	    }
	}

    /* Return the new image */
    NORMAL_RETURN;
}
