/*
*
* $header: g:/spl-2/hook-7/tokenize/cvsroot/libdgpimage/src/packedim.c,v 1.1.1.1 2000/02/01 00:43:54 mkrocken Exp $
*
* Copyright (c) 1994 Xerox Corporation.  All Rights Reserved.
*  
*/

static char rcsid[] = "@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/packedim.c,v 1.1 2000/03/04 00:20:52 mkrocken Exp $";

/*
*	packedim.c - Packed images.
*
* This module deals with binary images stored in a packed manner (using all
* the bits in each byte).
*
* Exports:
*	type DgpImage_Packed
*
*	constant DgpImage_Type DGPIMAGE_PACKED
*
*	macro DGPIMG_GETBIT(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y) - return zero if the bit
*		at (x,y) is clear, non-zero if it is set.
*
*	macro DGPIMG_GETBIT01(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y) - return zero if the
*		bit at (x,y) is clear, 1 if it is set.
*
*	macro DGPIMG_SETBIT(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y) - set the bit at (x,y).
*
*	macro DGPIMG_CLEARBIT(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y) - clear the bit
*		at (x,y).
*
*	macro DGPIMG_ORBIT(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 bit) - OR bit
*		into im at (x,y).
*
*	macro Dgp_UInt32 *DGPIMG_GETROW(DgpImage_Packed im, Dgp_Int32 y) - return a
*		pointer to the first long in row y.
*
*	Dgp_Int32 DGPIMG_GETROWWIDTH(DgpImage_Packed im) - return the number of bytes
*		in each row of im.
*
*	DgpImage_Packed pimWrap(Dgp_UInt32 *data, Dgp_Int32 width, Dgp_Int32 height,
*		Dgp_Int32 longsperline) - wrap a DgpImage_Packed around a data block
*
*	Dgp_UInt32 *pimUnwrap(DgpImage_Packed im) - unwrap a DgpImage_Packed,
*		leaving only the data behind.
*
*	imNew, imNewOffset, imFree, imSetOffset, imLoad, imLoadF, imSave,
*		imSaveF, imDup, imPromote, imDemote, imConvert, DGPIMG_WIDTH,
*		DGPIMG_HEIGHT, DGPIMG_XBASE, DGPIMG_YBASE, DGPIMG_XMAX, DGPIMG_YMAX,
*		DGPIMG_TYPE - see the image.c documentation.
*
* A DgpImage_Packed is a rectangular storage area of bits, with the bits being
* packed at 8 to the byte. The storage is arranged in row-major order.
* The bits are arranged to be accessed long-by-long: the unit of bit
* packing is the Dgp_UInt32. The most significant bit in each long is
* the leftmost. The rows are allocated contiguously, from top to bottom,
* and in such a way that each row begins on a long-aligned boundary.
* This means that usually the row widths are rounded up to the next
* multiple of 32 bits, but on some machines, they might be rounded up to 64.
* Because of this, don't try to calculate the length of a row yourself;
* use DGPIMG_GETROWWIDTH.  Note that DGPIMG_GETROWWIDTH returns the number of
* *bytes* in the row, not the number of longs.
*
* Note that the size of an Dgp_UInt32 varies from machine to machine,
* as does the byte endianness. Your code must access the images by longs
* and not bytes, shorts, ints or anything else.
*
* Packed images are allocated, stored, etc using the standard image.c
* routines. Loading and storing is done to/from PBM files. There are some
* caveats, though:
* 1) The only "real" conversions (for imPromote, imDemote, imConvert) are
*    between DgpImage_Binary and DgpImage_Packed; these are both considered
*    promotions. However, other conversions will be attempted by
*    trying to go via a DgpImage_Binary.
* 2) Comments are not supported in DgpImage_Packeds.
* 3) The standard access functions (DGPIMG_REF, DGPIMG_PTRREF, etc) do not work.
*    Use DGPIMG_GETBIT, DGPIMG_SETBIT, DGPIMG_CLEARBIT, or access the data area yourself
*    using DGPIMG_GETROW.
* 4) DGPIMAGE_CHECKALL is not supported. In other words, if you include
*    packedim.h, and also define DGPIMAGE_CHECKALL, a number of things will
*    break, possibly unpleasantly (about the best you can hope for is
*    the DGPIMAGE_CHECKALL not working).
* 5) It is still possible to have non-(0,0) based images. The leftmost bit
*    in each row will always be the DGPIMG_XBASE() bit.
* 6) A lot more is guaranteed about the internal structure of a DgpImage_Packed.
*    After all, you wouldn't use one of these unless you cared about speed...
*    Thus, it is guaranteed how the storage area is laid out, alignment, etc.
*
* You can take a bit array you've allocated yourself and turn it into a
* DgpImage_Packed using pimWrap.  Pass in the array, its width and height in
* pixels, and its row pitch (in longs), and get back a DgpImage_Packed,
* with an offset of (0, 0).  In case of failure, pimWrap returns
* (DgpImage_Packed)NULL, and has not done anything to the data block.  In case of
* success, the data block has been attached to the image, and so if you do
* an imFree on it, that block will be freed.
*
* pimUnwrap does the opposite of pimWrap - it strips off the image structure
* and hands you the data block.  Note that you're expected to have noted
* the image width, height and so on before doing the pimUnwrap.  After
* pimUnwrap, the data block is your responsibility; also, the original
* DgpImage_Packed is no longer valid.
*
* DGPIMG_GETBIT and DGPIMG_GETBIT01 perform the same function; the only difference
* is that DGPIMG_GETBIT01 always returns 0 or 1 (rather than 0 or non-zero)
* and may be useful in contexts where the value is used numerically
* rather than as a flag.  DGPIMG_GETBIT01 may be slightly slower.
*
* DGPIMG_ORBIT ORs a 1-bit value (0 or 1) into the image; it's essentially
* a conditional DGPIMG_SETBIT (set the bit or do nothing).
*/

#include "image.h"
#include "image_defs.h"
#include <stdio.h>
/*#include "undecl-stdio.h"*/
#include <stdlib.h>
#include <ctype.h>
#include "pbmstuff.h"
#include <string.h>
/*#include "undecl-string.h"*/
#include <memory.h>
#include "image-private.h"

#include "endianness.h"


