/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/packedim_new.c,v 1.1 2000/03/04 00:20:52 mkrocken Exp $
 *
 * Copyright (c) 1994 Xerox Corporation.  All Rights Reserved.
 *  
 */

static char rcsid[] = "@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/packedim_new.c,v 1.1 2000/03/04 00:20:52 mkrocken Exp $";

/*
 *	packedim.c - Packed images.
 *
 * This module deals with binary images stored in a packed manner (using all
 * the bits in each byte).
 *
 * Exports:
 *	type Dgpimage_Packed
 *
 *	constant Dgpimage_Type DGPIMAGE_PACKED
 *
 *	macro pimGetBit(Dgpimage_Packed im, int x, int y) - return zero if the bit
 *		at (x,y) is clear, non-zero if it is set.
 *
 *	macro pimGetBit01(Dgpimage_Packed im, int x, int y) - return zero if the
 *		bit at (x,y) is clear, 1 if it is set.
 *
 *	macro pimSetBit(Dgpimage_Packed im, int x, int y) - set the bit at (x,y).
 *
 *	macro pimClearBit(Dgpimage_Packed im, int x, int y) - clear the bit
 *		at (x,y).
 *
 *	macro pimOrBit(Dgpimage_Packed im, int x, int y, int bit) - OR bit
 *		into im at (x,y).
 *
 *	macro unsigned long *pimGetRow(Dgpimage_Packed im, int y) - return a
 *		pointer to the first long in row y.
 *
 *	int pimGetRowWidth(Dgpimage_Packed im) - return the number of bytes
 *		in each row of im.
 *
 *	Dgpimage_Packed pimWrap(unsigned long *data, int width, int height,
 *		int longsperline) - wrap a Dgpimage_Packed around a data block
 *
 *	unsigned long *pimUnwrap(Dgpimage_Packed im) - unwrap a Dgpimage_Packed,
 *		leaving only the data behind.
 *
 *	imNew, imNewOffset, imFree, imSetOffset, imLoad, imLoadF, imSave,
 *		imSaveF, imDup, imPromote, imDemote, imConvert, imGetWidth,
 *		imGetHeight, imGetXBase, imGetYBase, imGetXMax, imGetYMax,
 *		imGetType - see the image.c documentation.
 *
 * A Dgpimage_Packed is a rectangular storage area of bits, with the bits being
 * packed at 8 to the byte. The storage is arranged in row-major order.
 * The bits are arranged to be accessed long-by-long: the unit of bit
 * packing is the unsigned long. The most significant bit in each long is
 * the leftmost. The rows are allocated contiguously, from top to bottom,
 * and in such a way that each row begins on a long-aligned boundary.
 * This means that usually the row widths are rounded up to the next
 * multiple of 32 bits, but on some machines, they might be rounded up to 64.
 * Because of this, don't try to calculate the length of a row yourself;
 * use pimGetRowWidth.  Note that pimGetRowWidth returns the number of
 * *bytes* in the row, not the number of longs.
 *
 * Note that the size of an unsigned long varies from machine to machine,
 * as does the byte endianness. Your code must access the images by longs
 * and not bytes, shorts, ints or anything else.
 *
 * Packed images are allocated, stored, etc using the standard image.c
 * routines. Loading and storing is done to/from PBM files. There are some
 * caveats, though:
 * 1) The only "real" conversions (for imPromote, imDemote, imConvert) are
 *    between Dgpimage_Binary and Dgpimage_Packed; these are both considered
 *    promotions. However, other conversions will be attempted by
 *    trying to go via a Dgpimage_Binary.
 * 2) Comments are not supported in Dgpimage_Packeds.
 * 3) The standard access functions (imRef, imPtrRef, etc) do not work.
 *    Use pimGetBit, pimSetBit, pimClearBit, or access the data area yourself
 *    using pimGetRow.
 * 4) DGPIMAGE_CHECKALL is not supported. In other words, if you include
 *    packedim.h, and also define DGPIMAGE_CHECKALL, a number of things will
 *    break, possibly unpleasantly (about the best you can hope for is
 *    the DGPIMAGE_CHECKALL not working).
 * 5) It is still possible to have non-(0,0) based images. The leftmost bit
 *    in each row will always be the imGetXBase() bit.
 * 6) A lot more is guaranteed about the internal structure of a Dgpimage_Packed.
 *    After all, you wouldn't use one of these unless you cared about speed...
 *    Thus, it is guaranteed how the storage area is laid out, alignment, etc.
 *
 * You can take a bit array you've allocated yourself and turn it into a
 * Dgpimage_Packed using pimWrap.  Pass in the array, its width and height in
 * pixels, and its row pitch (in longs), and get back a Dgpimage_Packed,
 * with an offset of (0, 0).  In case of failure, pimWrap returns
 * (Dgpimage_Packed)NULL, and has not done anything to the data block.  In case of
 * success, the data block has been attached to the image, and so if you do
 * an imFree on it, that block will be freed.
 *
 * pimUnwrap does the opposite of pimWrap - it strips off the image structure
 * and hands you the data block.  Note that you're expected to have noted
 * the image width, height and so on before doing the pimUnwrap.  After
 * pimUnwrap, the data block is your responsibility; also, the original
 * Dgpimage_Packed is no longer valid.
 *
 * pimGetBit and pimGetBit01 perform the same function; the only difference
 * is that pimGetBit01 always returns 0 or 1 (rather than 0 or non-zero)
 * and may be useful in contexts where the value is used numerically
 * rather than as a flag.  pimGetBit01 may be slightly slower.
 *
 * pimOrBit ORs a 1-bit value (0 or 1) into the image; it's essentially
 * a conditional pimSetBit (set the bit or do nothing).
 */

#include "image.h"
#include "image_defs.h"
#include <stdio.h>
/*#include "undecl-stdio.h"*/
#include <stdlib.h>
#include <ctype.h>
#include "pbmstuff.h"
#include <string.h>
/*#include "undecl-string.h"*/
#include <memory.h>
#include "image-private.h"

#include "endianness.h"

/* Declare internal routines */
static unsigned long **allocPackedData(int width, int height,
				       int xbase, int ybase,
				       dgp_boolean dontclear, DgpState *dgp_state);
static unsigned long **allocPackedDataPointers(int width, int height,
					       unsigned long *data,
					       int longsperline,
					       int xbase, int ybase, DgpState *dgp_state);
static Dgpimage_Packed readRaw(FILE *inFile, DgpState *dgp_state);
static Dgpimage_Packed readCooked(FILE *inFile, DgpState *dgp_state);
static Dgpimage_Binary packedToBinary(Dgpimage_Packed im);
static Dgpimage_Packed binaryToPacked(Dgpimage_Binary im);

ErrCode
DGPIMAGE_pNew(DgpState *dgp_state, Dgpimage_Type type, int width, int height, void** image)
{

	DGP_PROLOGUE ("DGPIMAGE_pNew", dgp_pErrState);

    DGP_ERR_PASS (DGPIMAGE_pNewOffset(dgp_state, type, width, height, 0, 0, image));
	NORMAL_RETURN;
}


ErrCode
DGPIMAGE_pNewOffset(DgpState *dgp_state, Dgpimage_Type type, int width, int height,
	     int xbase, int ybase, void** image)
{
    Dgpimage_Header *newHeader;
    Dgpimage_Packed newPacked;
    dgp_boolean dontclear = FALSE;

	DGP_PROLOGUE ("DGPIMAGE_pNewOffset", dgp_pErrState);

    if (type & DGPIMAGE_DONTCLEAR) {
	type &= (~DGPIMAGE_DONTCLEAR);
	dontclear = TRUE;
	}

    /*
     * Allocate the data area for the appropriate type of image, 
     * allocate and initialise the actual *Image and return it.
     * This exists in one form for each type of Image.
     */
#define	ALLOC_INIT(imtype) \
    { \
	imtype newim; \
	DGP_CHECK_ALLOC (newim = (imtype)(dgp_Malloc)(sizeof(* newim))); \
	areaPtr = allocImage(dgp_state, width, height, xbase, ybase, \
			     sizeof(**newim->data), dontclear); \
	if (areaPtr == (void **)NULL) { \
	    (dgp_Free)((void *)newim); \
	    DGP_ERR (cErrInsufficientMem); \
	    } \
	newHeader = &newim->header; \
	newim->data = (void *)areaPtr; /* A better cast is possible */ \
	newim->dgp_state = dgp_state;		\
	*image = (void *)newim; \
	}

    switch(type) {
	case DGPIMAGE_GRAY:		ALLOC_INIT(Dgpimage_Gray);		break;
	case DGPIMAGE_FLOAT:		ALLOC_INIT(Dgpimage_Float);		break;
	case DGPIMAGE_RGB:		ALLOC_INIT(Dgpimage_ColorRGB);		break;
	case DGPIMAGE_DOUBLE:	ALLOC_INIT(Dgpimage_Double);	break;
	case DGPIMAGE_BINARY:	ALLOC_INIT(Dgpimage_Binary);	break;
	case DGPIMAGE_LONG:		ALLOC_INIT(Dgpimage_Long);		break;
	case DGPIMAGE_PTR:		ALLOC_INIT(Dgpimage_Ptr);		break;
	case DGPIMAGE_SHORT:		ALLOC_INIT(Dgpimage_Short);		break;
	case DGPIMAGE_RGBFLOAT:	ALLOC_INIT(Dgpimage_ColorRGBFloat);	break;
	case DGPIMAGE_PACKED:
		
		DGP_CHECK_ALLOC (newPacked = (Dgpimage_Packed)(dgp_Malloc)(sizeof(*newPacked)));
		
		newHeader = &newPacked->header;
		newPacked->dgp_state = dgp_state;
		newPacked->data = allocPackedData(width, height, xbase, ybase, dontclear, dgp_state);
		if (newPacked->data == (unsigned long **)NULL) {
			(dgp_Free)((void *)newPacked);
			DGP_ERR (cErrInsufficientMem);
		}
		newHeader->row_width = ROW_WIDTH(width);
		break;
	default: {
		DGP_ERR	(cErrInvalidCmdLine);
			 }
	}


#undef	ALLOC_INIT

    newHeader->width = width;
    newHeader->height = height;
    newHeader->xbase = xbase;
    newHeader->ybase = ybase;
    newHeader->tag = type;
    newHeader->comment = (char *)NULL;

    NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pFree (void *im, Dgpimage_Type tag)
{
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pFree", dgp_pErrState);

    if (tag != DGPIMAGE_PACKED) {
		Dgpimage_Header *header;
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
		dgpimage_unp_Free_(im, header);
		NORMAL_RETURN;
	}

    DGP_ASSERT(pim != (Dgpimage_Packed)NULL);
    DGP_ASSERT(pim->data != (unsigned long **)NULL);

    /* Free the data area, the pointer block, the header, and the image */
    (dgp_Free)((void *)pim->data[pim->header.ybase]);
    (dgp_Free)((void *)(pim->data + pim->header.ybase));
    (dgp_Free)((void *)pim);

	NORMAL_RETURN;
}ErrCode

DGPIMAGE_Free (void *im)
{
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;
	Dgpimage_header *header;

	DGP_PROLOGUE ("DGPIMAGE_pFree", dgp_pErrState);
    DGP_ASSERT(pim != (Dgpimage_Packed)NULL);
    DGP_ASSERT(pim->data != (unsigned long **)NULL);

	header = &pim->header;
	DGP_ASSERT(header != (Dgpimage_Header *)NULL);
	if (header->comment != (char *)NULL) {
		(dgp_Free)((void *)header->comment);


    if (header->tag != DGPIMAGE_PACKED) {
		void **areaPtr;
		size_t el_size;
		int xbase, ybase;
		Dgpimage_Gray gimage = (Dgpimage_Gray) im;		
		
		/* Pull out the data area and the size of the element */
		getAreaAndSize(im, header, &areaPtr, &el_size, dgp_state);
		
		/* and get rid of all of them */
		xbase = header->xbase;
		ybase = header->ybase;
		}
		(dgp_Free)((void *)im);
		freeImage (dgp_state, areaPtr, xbase, ybase, el_size);
		
		NORMAL_RETURN;
	}

    /* Free the data area, the pointer block, the header, and the image */
    (dgp_Free)((void *)pim->data[pim->header.ybase]);
    (dgp_Free)((void *)(pim->data + pim->header.ybase));
    (dgp_Free)((void *)pim);

	NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pSetOffset (void *im, Dgpimage_Type tag, int newxbase, int newybase)
{
    int oldybase;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pSetOffset_", dgp_pErrState);

    if (tag != DGPIMAGE_PACKED) {
		Dgpimage_Header *header;
		DGPIMAGE_getHeader(im, tag, &header);
		dgpimage_unp_SetOffset_(im, header, newxbase, newybase);
		NORMAL_RETURN;
	}

    DGP_ASSERT(pim != (Dgpimage_Packed)NULL);
    DGP_ASSERT(pim->data != (unsigned long **)NULL);

    /*
     * Since the row pointers aren't offset, just fix the pointer
     * block pointer and update the header.
     */
    oldybase = pim->header.ybase;
    pim->data += (oldybase - newybase);
    pim->header.xbase = newxbase;
    pim->header.ybase = newybase;

	NORMAL_RETURN;
}

ErrCode
DGPIMAGE_SetOffset (void *im, int newxbase, int newybase)
{
    int oldybase;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;
	Dgpimage_Type tag;
	
	DGP_PROLOGUE ("DGPIMAGE_pSetOffset_", dgp_pErrState);
    DGP_ASSERT(pim != (Dgpimage_Packed)NULL);
    DGP_ASSERT(pim->header != (Dgpimage_Header)NULL);
    DGP_ASSERT(pim->data != (unsigned long **)NULL);
	tag = pim->header.tag;
	
    if (tag != DGPIMAGE_PACKED) {
		int y;
		int oldxbase;
		int oldybase;
		Dgpimage_Gray gimage = (Dgpimage_Gray) im;
		
		oldxbase = header->xbase;
		oldybase = header->ybase;
		
#define	FIX_PTRS(imtype) \
		{ \
		imtype vim = (imtype)(im); \
		for (y = imGetYBase(vim); y <= imGetYMax(vim); y++) { \
		vim->data[y] += (oldxbase - newxbase); \
		} \
		vim->data += (oldybase - newybase); \
		}
		
		switch(header->tag) {
		case DGPIMAGE_GRAY:		FIX_PTRS(Dgpimage_Gray);	break;
		case DGPIMAGE_FLOAT:		FIX_PTRS(Dgpimage_Float);	break;
		case DGPIMAGE_RGB:		FIX_PTRS(Dgpimage_ColorRGB);	break;
		case DGPIMAGE_DOUBLE:	FIX_PTRS(Dgpimage_Double);	break;
		case DGPIMAGE_BINARY:	FIX_PTRS(Dgpimage_Binary);	break;
		case DGPIMAGE_LONG:		FIX_PTRS(Dgpimage_Long);	break;
		case DGPIMAGE_PTR:		FIX_PTRS(Dgpimage_Ptr);	break;
		case DGPIMAGE_SHORT:		FIX_PTRS(Dgpimage_Short);	break;
		case DGPIMAGE_RGBFLOAT:	FIX_PTRS(Dgpimage_ColorRGBFloat);break;
		default: {
			DGP_ERR	(cErrInvalidCmdLine);
				 }
		}
		
#undef	FIX_PTRS
		
		header->xbase = newxbase;
		header->ybase = newybase;
		
		NORMAL_RETURN;
		
	}
	
	
    /*
	* Since the row pointers aren't offset, just fix the pointer
	* block pointer and update the header.
	*/
    oldybase = pim->header.ybase;
    pim->data += (oldybase - newybase);
    pim->header.xbase = newxbase;
    pim->header.ybase = newybase;
	
	NORMAL_RETURN;
}

ErrCode
DGPIMAGE_Load (DgpState *dgp_state, Dgpimage_Type type, char *filename, void **image)
{
    FILE *inFile;
	
	DGP_PROLOGUE ("DGPIMAGE_Load", dgp_pErrState);
	
    /* Open the file, imLoadF on it, then close it */
    DGP_ASSERT(filename != (char *)NULL);
    DGP_CHECK_NULL (inFile = fopen(filename, "rb"));
	
	char fileHeader[PBM_MAX_LINELEN];
	int c;
	int i;
				
	/*
	* Read up to the first whitespace in order to find the file's header.
	*/
	i = 0;
	do {
		c = getc(inFile);
		if (c == EOF) {
			fclose (inFile);
			DGP_ERR (cErrFileNotFound);
		}
		fileHeader[i++] = c;
	} while (!isspace(c));
	fileHeader[--i] = '\0';		/* Clobber the whitespace */
	
	if (type == DGPIMAGE_PACKED)
	{
		
		if (!strcmp(fileHeader, TAG_BINARY)) {
			*image = readRaw(inFile, dgp_state);
		}
		else if (!strcmp(fileHeader, TAG_BINARY_COOKED)) {
			*image = readCooked(inFile, dgp_state);
		}
		else {
			/* Not a PBM file */
			DGP_ERR (cErrFileNotFound);	
		}
	}
	else
	{
		Dgpimage_Type filetype = (Dgpimage_Type)0;
		void * (*readfunc)(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader) = NULL;
		void *readim;
		Dgpimage_Header *readimHeader;
		
		
		/* Determine who is supposed to handle this sort of file */
		for (i = 0; i < NREADERS; i++) {
			if (strcmp(fileHeader, fileReaders[i].filetag) == 0) {
				/* Found it */
				filetype = fileReaders[i].filetype;
				readfunc = fileReaders[i].readfunc;
				break;
			}
		}
		if (i == NREADERS) {
			/* Didn't match any of the ones in the table - unknown type. */
			DGP_ERR (cErrFileNotFound);
		}
		
		/* Clear the comment buffer */
		incommentsize = 0;
		
		/* Read in the file */
		readim = (*readfunc)(dgp_state, inFile, &readimHeader);
		if (readim == (void *)NULL) {
			/* Didn't load for whatever reason */
			DGP_ERR (cErrFileNotFound);
		}
		
		if (filetype == type) {
			/* The file was of the expected type - no conversion required */
			if (!setComment(readimHeader, dgp_state)) {
				/* Sigh - stumbled over the comment */
				dgpimage_unp_Free_(readim, readimHeader);
				DGP_ERR (cErrFileNotFound);;
			}
			*image = readim;
			NORMAL_RETURN;
		}
		
		/*
		* OK. We read in a file, and it wasn't what we expected. We get to
		* promote it to the expected image type, if that's defined.
		*
		* If that conversion isn't a promotion, kick out.
		*/
		if (!convertTable[filetype][type].isPromotion) {
			dgpimage_unp_Free_(readim, readimHeader);
			DGP_ERR (cErrInvalidCmdLine);
		}
		
		/* Convert... */
		DGP_ERR_CHECK (dgpimage_unp_Promote_(readim, readimHeader, type, image)) {
			/* Didn't work. */
			dgpimage_unp_Free_(readim, readimHeader);
			ERR_RETURN;
		}
		
		/* Kill the image we read in, and return the promoted one. */
		dgpimage_unp_Free_(readim, readimHeader);
		
		/* Stick the comment from the file into the header */
		dgpimage_unp_getHeader(*image, type, &readimHeader);
		if (!setComment(readimHeader, dgp_state)) {
			/* Died in the comment copy */
			dgpimage_unp_Free_(*image, readimHeader);
			DGP_ERR (cErrInvalidCmdLine);
		}
		
	}
	
    if (fclose(inFile) == EOF) {
	/*
	* It would be a strange day that fclose on a read-only file
	* gave an error, but...
		*/
		if (*image) {
			DGPIMAGE_Free (*image);
		}
		DGP_ERR (cErrFileNotFound);
	}
	
    NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pLoadF(DgpState *dgp_state, Dgpimage_Type type, FILE *inFile, void** image)
{
    char fileHeader[PBM_MAX_LINELEN];
    int c;
    int i;

	DGP_PROLOGUE ("DGPIMAGE_pLoadF", dgp_pErrState);

    if (type != DGPIMAGE_PACKED) {
		DGP_ERR_PASS (DGPIMAGE_LOADF(dgp_state, type, inFile, image));
		NORMAL_RETURN;
	}

    /*
     * Read up to the first whitespace in order to find the file's header.
     */
    i = 0;
    do {
	c = getc(inFile);
	if (c == EOF) {
		fclose (inFile);
		DGP_ERR (cErrFileNotFound);
	}
	fileHeader[i++] = c;
	} while (!isspace(c));
    fileHeader[--i] = '\0';		/* Clobber the whitespace */

    if (!strcmp(fileHeader, TAG_BINARY)) {
		*image = readRaw(inFile, dgp_state);
	}
    else if (!strcmp(fileHeader, TAG_BINARY_COOKED)) {
		*image = readCooked(inFile, dgp_state);
	}
    else {
	/* Not a PBM file */
	DGP_ERR (cErrFileNotFound);	
	}

	NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pSave (void *im, Dgpimage_Type tag, char *filename)
{
    FILE *outFile;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pSave_", dgp_pErrState);

    DGP_ASSERT(filename != (char *)NULL);
    DGP_ASSERT(im != (void *)NULL);

    if (tag != DGPIMAGE_PACKED) {
		Dgpimage_Header *header;
		DGPIMAGE_getHeader(im, tag, &header);
		dgpimage_unp_Save_(im, header, filename);
		NORMAL_RETURN;
	}

    /* Open the file, imSaveF on it, then close it */
    outFile = fopen(filename, "wb");
    DGP_CHECK_NULL (outFile);

    /* Write the image */
    DGP_ERR_PASS (DGPIMAGE_pSaveF (im, DGPIMAGE_PACKED, outFile)); 

    /* and close the file */
    if (fclose(outFile) == EOF) {
	/* Bah. */
	DGP_ERR (cErrFileNotFound);	
	}

    NORMAL_RETURN;
    }
	       
ErrCode
DGPIMAGE_pSaveF (void *im, Dgpimage_Type tag, FILE *outFile)
{

    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;
	DGP_PROLOGUE ("DGPIMAGE_pSaveF_", dgp_pErrState);

    if (tag != DGPIMAGE_PACKED) {
		Dgpimage_Header *header;
		DGPIMAGE_getHeader(im, tag, &header);
		dgpimage_unp_SaveF_(im, header, outFile);
		NORMAL_RETURN;
	}

    /* First, spew the header */
    if (fprintf(outFile, "%s\n%d %d\n",
		TAG_BINARY, imGetWidth((Dgpimage_Packed)im),
		imGetHeight((Dgpimage_Packed)im)) == EOF) {
		DGP_ERR (cErrFileNotFound);	
	}
    
    if (!dgpimage_savePackedRaw((Dgpimage_Packed)im, outFile, 0)) {
	DGP_ERR (cErrFileNotFound);	
	}

	NORMAL_RETURN;
}
	       
ErrCode
DGPIMAGE_pDup (void *im, Dgpimage_Type tag, dgp_boolean copycomment, void** image)
{
    Dgpimage_Packed dup;
    unsigned long *p, *q;
    int y;
    int orowwidth, nrowwidth;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pDup_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);

    if (tag != DGPIMAGE_PACKED) {
		Dgpimage_Header *header;
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
		DGP_ERR_PASS (dgpimage_unp_Dup_(im, header, copycomment, image));
		NORMAL_RETURN;
	}

    DGP_ERR_PASS (DGPIMAGE_pNewOffset(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR,
				    imGetWidth(pim), imGetHeight(pim),
				    imGetXBase(pim), imGetYBase(pim), (void**) &dup));
    
    /* They might have passed in a padded image to imWrap... */
    if (pim->header.row_width != ROW_WIDTH(imGetWidth(pim))) {
	/* Sigh... */
	orowwidth = pim->header.row_width / sizeof(unsigned long);
	nrowwidth = ROW_WIDTH(imGetWidth(pim)) / sizeof(unsigned long);
	for (y = imGetYBase(pim), p = dup->data[y], q = pim->data[y];
	     y <= imGetYMax(pim);
	     y++, p += nrowwidth, q += orowwidth) {
	    (void)memcpy(p, q, nrowwidth * sizeof(unsigned long));
	    }
	}
    else {
	/* Aaah... */
	(void)memcpy(dup->data[imGetYBase(dup)], pim->data[imGetYBase(pim)],
		     pimGetRowWidth(pim) * imGetHeight(pim));
	}

    *image = (void *)dup;
	NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pPromote (void *im, Dgpimage_Type tag, Dgpimage_Type promoteto, void** image)
{
    Dgpimage_Binary tmp;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pPromote_", dgp_pErrState);

    if ((tag != DGPIMAGE_PACKED) && (promoteto != DGPIMAGE_PACKED)) {
		Dgpimage_Header *header;
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
		DGP_ERR_PASS (dgpimage_unp_Promote_(im, header, promoteto, image));
		NORMAL_RETURN;
	}

    if (promoteto == DGPIMAGE_PACKED) {
	if (tag == DGPIMAGE_BINARY) {
	    *image = (void *)binaryToPacked((Dgpimage_Binary)im);
		NORMAL_RETURN;
	    }
	else {
		Dgpimage_Header *header;
		
		DGP_ASSERT(tag != DGPIMAGE_PACKED);

	    /* Try to make it into a Dgpimage_Binary. */
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
	    DGP_ERR_PASS (dgpimage_unp_Promote_(im, header,
					   DGPIMAGE_BINARY, (void**) &tmp));

		*image = (void *)binaryToPacked(tmp);
	    dgpimage_unp_Free_(tmp, &tmp->header);
	    NORMAL_RETURN;
	    }
	}
    else if (promoteto == DGPIMAGE_BINARY) {
	DGP_ASSERT(tag == DGPIMAGE_PACKED);
	*image = (void *)packedToBinary((Dgpimage_Packed)im);
	NORMAL_RETURN;
	}
    else {
	DGP_ASSERT(tag == DGPIMAGE_PACKED);

	tmp = packedToBinary((Dgpimage_Packed)im);
	DGP_CHECK_NULL (tmp);

	dgpimage_unp_Promote_(tmp, &tmp->header, promoteto, image);
	dgpimage_unp_Free_(tmp, &tmp->header);
	NORMAL_RETURN;
	}
}

ErrCode
DGPIMAGE_pDemote (void *im, Dgpimage_Type tag, Dgpimage_Type demoteto, void** image)
{
    Dgpimage_Binary tmp;
	Dgpimage_Header *header;
    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pDemote_", dgp_pErrState);

    if ((tag != DGPIMAGE_PACKED) && (demoteto != DGPIMAGE_PACKED)) {
		Dgpimage_Header *header;
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
		DGP_ERR_PASS (dgpimage_unp_Demote_(im, header, demoteto, image));
		NORMAL_RETURN;
	}

    if (demoteto == DGPIMAGE_PACKED) {
	DGP_ASSERT((tag != DGPIMAGE_PACKED) && (tag != DGPIMAGE_BINARY));

	/* Try to make it into a Dgpimage_Binary. */
	DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
	DGP_ERR_PASS (dgpimage_unp_Demote_(im, header,
				      DGPIMAGE_BINARY, (void**) &tmp));

	*image = (void *)binaryToPacked(tmp);
	dgpimage_unp_Free_(tmp, &tmp->header);
	NORMAL_RETURN;
	}
    else {
	DGP_ASSERT(tag == DGPIMAGE_PACKED);
	DGP_ASSERT(demoteto != DGPIMAGE_BINARY);

	tmp = packedToBinary((Dgpimage_Packed)im);
	DGP_CHECK_NULL (tmp);

	DGP_ERR_PASS (dgpimage_unp_Demote_(tmp, &tmp->header, demoteto, image));
	dgpimage_unp_Free_(tmp, &tmp->header);
	NORMAL_RETURN;
	}
}

ErrCode
DGPIMAGE_pConvert (void *im, Dgpimage_Type tag, Dgpimage_Type convertto, void** image)
{

    Dgpimage_Packed pim = (Dgpimage_Packed) im;
	DgpState *dgp_state = pim->dgp_state;
	DGP_PROLOGUE ("DGPIMAGE_pConvert_", dgp_pErrState);

    if ((tag != DGPIMAGE_PACKED) && (convertto != DGPIMAGE_PACKED)) {
		Dgpimage_Header *header;
		DGP_ERR_PASS (DGPIMAGE_getHeader(im, tag, &header));
		DGP_ERR_PASS (dgpimage_unp_Convert_(im, header, convertto, image));
		NORMAL_RETURN;
	}

    if ((tag == DGPIMAGE_BINARY) || (convertto == DGPIMAGE_BINARY)) {
	DGP_ERR_PASS (DGPIMAGE_pPromote (im, tag, convertto, image));
	NORMAL_RETURN;
	}
    else if ((tag == DGPIMAGE_PACKED) && (convertto == DGPIMAGE_PACKED)) {
	DGP_ERR_PASS (DGPIMAGE_pDup (im, tag, FALSE, image));
	NORMAL_RETURN;
	}
    else {
	DGP_ERR_PASS (DGPIMAGE_pDemote (im, tag, convertto, image));
	NORMAL_RETURN;
	}
}

ErrCode
DGPIMAGE_pWrap(DgpState *dgp_state, unsigned long *data, int width, 
			   int height, int longsperline, Dgpimage_Packed *newim)
{
    Dgpimage_Header *newHeader;

	DGP_PROLOGUE ("DGPIMAGE_pWrap", dgp_pErrState);

    DGP_ASSERT(data != (unsigned long *)NULL);
    DGP_ASSERT(longsperline * (int)sizeof(unsigned long) >= ROW_WIDTH(width));

    DGP_CHECK_ALLOC (*newim = (Dgpimage_Packed)(dgp_Malloc)(sizeof(*newim)));

    newHeader = &(*newim)->header;
    (*newim)->data = allocPackedDataPointers(width, height, data, longsperline,
					0, 0, dgp_state);
    if ((*newim)->data == (unsigned long **)NULL) {
	(dgp_Free)((void *) *newim);
	DGP_ERR (cErrNullPtr);
	}

    newHeader->width = width;
    newHeader->row_width = longsperline * sizeof(unsigned long);
    newHeader->height = height;
    newHeader->xbase = 0;
    newHeader->ybase = 0;
    newHeader->tag = DGPIMAGE_PACKED;

    NORMAL_RETURN;
}

ErrCode
DGPIMAGE_pUnwrap(Dgpimage_Packed im, unsigned long **retval)
{

	DgpState *dgp_state = im->dgp_state;

	DGP_PROLOGUE ("DGPIMAGE_pUnwrap", dgp_pErrState);

    DGP_ASSERT(im != (Dgpimage_Packed)NULL);
    DGP_ASSERT(im->data != (unsigned long **)NULL);

    /* Free the pointer block and the image, and return the data area */
    *retval = im->data[im->header.ybase];
    (dgp_Free)((void *)(im->data + im->header.ybase));
    (dgp_Free)((void *)im);

    NORMAL_RETURN;
}

/*
 * Allocate and set up a block of memory to hold the image, plus
 * a block of pointers into it. Offset the row pointer array by ybase;
 * don't worry about xbase.
 */
static unsigned long **
allocPackedData(int width, int height, int xbase, int ybase,
		dgp_boolean dontclear, DgpState *dgp_state)
{
    unsigned long *mainArea;
    unsigned long **ptrArea;
    int row_width = ROW_WIDTH(width);
    int lrow_width = row_width / sizeof(unsigned long);

    /* Allocate the image area */
    if (!dontclear) {
	mainArea =
	    (unsigned long *)(dgp_Calloc)((size_t)height,
						     lrow_width *
						     sizeof(*mainArea));
	}
    else {
	mainArea =
	    (unsigned long *)(dgp_Malloc)(height * lrow_width *
						     sizeof(*mainArea));
	}
    if (mainArea == (unsigned long *)NULL) {
	return((unsigned long **)NULL);
	}

    /* Allocate a pointer for each row */
    ptrArea = allocPackedDataPointers(width, height, mainArea, lrow_width,
				      xbase, ybase, dgp_state);
    if (ptrArea == (unsigned long **)NULL) {
	(dgp_Free)(mainArea);
	return((unsigned long **)NULL);
	}

    return(ptrArea);
    }

static unsigned long **
allocPackedDataPointers(int width, int height, unsigned long *data,
			int longsperline, int xbase, int ybase, DgpState *dgp_state)
{
    unsigned long **ptrArea;
    int i;
    unsigned long *p, **q;

    /* kill a couple of warnings */
    width = width;
    xbase = xbase;

    /* Allocate the row pointers */
    ptrArea = (unsigned long **)(dgp_Malloc)(height *
							sizeof(*ptrArea));

    /* Set up the row pointers to point to the rows */
    for (i = 0, q = ptrArea, p = data;
	 i < height;
	 i++, q++, p += longsperline) {
	*q = p;
	}

    /* and return the (offset) row pointer area */
    return(ptrArea - ybase);
    }

/*
 * Read in a raw (non-printable) PBM file. Someone has already eaten the
 * file tag.
 */
Dgpimage_Packed
readRaw(FILE *inFile, DgpState *dgp_state)
{
    int width, height;
    dgp_boolean failed;
    Dgpimage_Packed im;

    /*
     * File format:
     * P4 <width> <height>
     * binary image, packed, high bit is leftmost; rows begin on byte
     * boundaries.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, (char *)NULL, (int *)NULL, dgp_state);
    if (failed) {
	return((Dgpimage_Packed)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, (char *)NULL, (int *)NULL, dgp_state);
    if (failed) {
	return((Dgpimage_Packed)NULL);
	}
    
    /* Allocate it; no need to clear it. */
    DGPIMAGE_pNew(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_Packed)NULL) {
	return((Dgpimage_Packed)NULL);
	}

    /* call out to rawimage.c for the rest */
    if (!dgpimage_readPackedRaw(im, width, height, inFile, 0)) {
	DGPIMAGE_pFree(im, DGPIMAGE_PACKED);
	return((Dgpimage_Packed)NULL);
	}

    return(im);
    }

/*
 * Read in a cooked (printable) PBM file.
 */
static Dgpimage_Packed
readCooked(FILE *inFile, DgpState *dgp_state)
{
    int width, height;
    dgp_boolean failed;
    Dgpimage_Packed im;
    int x, y;

    /*
     * File format:
     * P1 <width> <height>
     * bunches of 0[<ws>] and 1[<ws>], suitable for reading with
     * dgpimage_pbmGetPBMToken.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, (char *)NULL, (int *)NULL, dgp_state);
    if (failed) {
	return((Dgpimage_Packed)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, (char *)NULL, (int *)NULL, dgp_state);
    if (failed) {
	return((Dgpimage_Packed)NULL);
	}
    
    /* Allocate and clear it. */
    DGPIMAGE_pNew(dgp_state, DGPIMAGE_PACKED, width, height, (void**) &im);
    if (im == (Dgpimage_Packed)NULL) {
	return((Dgpimage_Packed)NULL);
	}
	
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    /* Can't use dgpimage_pbmGetToken - may not be separating whitespace */
	    if (dgpimage_pbmGetPBMToken(inFile, &failed, (char *)NULL, (int *)NULL, dgp_state)) {
		pimSetBit(im, x, y);
		}
	    if (failed) {
		DGPIMAGE_pFree(im, DGPIMAGE_PACKED);
		return((Dgpimage_Packed)NULL);
		}
	    }
	}

    return(im);
    }

static Dgpimage_Binary
packedToBinary(Dgpimage_Packed im)
{
    Dgpimage_Binary bim;
    Dgpimage_BinaryPixPtr bptr;
    int x, y;
    unsigned long *p;
    unsigned long c = 0;
    int bitsleft;
    int xend, yend;
    int xstart;
	DgpState *dgp_state = im->dgp_state;

	DGP_PROLOGUE ("packedToBinary", dgp_pErrState);
    DGP_ASSERT(im != (Dgpimage_Packed)NULL);

    DGPIMAGE_pNewOffset(dgp_state, DGPIMAGE_BINARY,
				    imGetWidth(im), imGetHeight(im),
				    imGetXBase(im), imGetYBase(im), (void**) &bim);
    if (bim == (Dgpimage_Binary)NULL) {
	return((Dgpimage_Binary)NULL);
	}

    xstart = imGetXBase(im);
    xend = imGetXMax(im);
    yend = imGetYMax(im);

    if (imGetWidth(im) == 0) {
	/* Silly case, but should be handled. */
	return(bim);
	}

    for (y = imGetYBase(im); y <= yend; y++) {
	bitsleft = 0;
	p = pimGetRow(im, y);
	for (x = xstart, bptr = imGetPixPtr(bim, xstart, y);
	     x <= xend;
	     x++, imPtrRight(bim, bptr)) {
	    if (bitsleft == 0) {
		c = *p++;
		bitsleft = _PIM_BITS_ULONG;
		}
	    /* High bit first */
	    if (c & (1ul << (_PIM_BITS_ULONG - 1))) {
		imPtrRef(bim, bptr) = 1;
		}
	    else if (c == 0) {
		/* It's already zero by default, and we've run out of bits */
		imPtrRightN(im, bptr, bitsleft - 1);
		x += bitsleft - 1;
		bitsleft = 0;
		continue;
		}
	    bitsleft--;
	    c <<= 1;
	    }
	}

    return(bim);
    }

static Dgpimage_Packed
binaryToPacked(Dgpimage_Binary im)
{
    Dgpimage_Packed pim;
    int x, y, bits, endx;
    unsigned long *p;
    unsigned long b;
    Dgpimage_BinaryPixPtr bp;
	DgpState *dgp_state = im->dgp_state;

    DGP_PROLOGUE ("binaryToPacked", dgp_pErrState);

    DGP_ASSERT(im != (Dgpimage_Binary)NULL);

    DGPIMAGE_pNewOffset(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR,
				    imGetWidth(im), imGetHeight(im),
				    imGetXBase(im), imGetYBase(im), (void**) &pim);
    if (pim == (Dgpimage_Packed)NULL) {
	return((Dgpimage_Packed)NULL);
	}

    p = pimGetRow(pim, imGetYBase(im));
    endx = imGetXMax(im);
    for (y = imGetYBase(im); y <= imGetYMax(im); y++) {
	for (x = imGetXBase(im), bits = _PIM_BITS_ULONG - 1, b = 0,
	     bp = imGetPixPtr(im, imGetXBase(im), y);
	     x <= endx;
	     x++, bits--, imPtrRight(im, bp)) {
	    if (imPtrRef(im, bp)) {
		b |= 1;
		}
	    if (bits == 0) {
		*p++ = b;
		bits = _PIM_BITS_ULONG;
		b = 0;
		}
	    else {
		b <<= 1;
		}
	    }
	if (bits != _PIM_BITS_ULONG - 1) {
	    b <<= bits;
	    *p++ = b;
	    }
	}

    return(pim);
    }

