/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libimage/rawimage.c,v 1.7 2003/01/27 21:48:32 awilliam Exp $
*
*/

/*
* rawimage - I/O routines for dealing with raw images.
*
* Exports:
*
*	constant DgpImage_Type DGPIMAGE_BIGENDIAN
*	constant DgpImage_Type DGPIMAGE_LITTLEENDIAN
*
*	void *imLoadRaw(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height,
*			Dgp_Int8 *filename) - load an image of the appropriate
*		type, with no header.  type may be ORed with DGPIMAGE_BIGENDIAN
*		or DGPIMAGE_LITTLEENDIAN to force it to treat the input
*		image as being stored with that endianness convention.
*		By default, the input image is treated as being stored
*		in the machine's native format.  Endianness is only
*		significant for multi-byte image types: DgpImage_Float,
*		DgpImage_Double, DgpImage_Long, DgpImage_Short, DgpImage_ColorRGBFloat.
*
*	void *imLoadRawF(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height,
*			 FILE *inFile) - as imLoadRaw, but reading from
*		a stream.
*
*	Dgp_Bool imSaveRaw(void *im, DgpImage_Type endianness, Dgp_Int8 *filename) -
*		save an image to a file, with no file header.
*		endianness must be either DGPIMAGE_BIGENDIAN, DGPIMAGE_LITTLEENDIAN,
*		or 0; these correspond to the file being written out in
*		big-endian, little-endian, or native-endian order.  Return
*		DGP_TRUE on success, DGP_FALSE on failure.
*
*	Dgp_Bool imSaveRawF(void *im, DgpImage_Type endianness, FILE *outFile) -
*		as imSaveRaw, but writing to a stream.
*
*/

#include "rawimage.h"
#include "image.h"
#include "image_defs.h"
#include "image-private.h"
#include "PsgFix.h"

#include <stdio.h>
#include <stdlib.h>

#include <endianness.h>

#ifndef DGP_FAXMODE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/rawimage.c,v 1.7 2003/01/27 21:48:32 awilliam Exp $");


static DgpImage_Gray readGrayRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
							 DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Float readFloatRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
							   DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_ColorRGB readRGBRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
						   DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Double readDoubleRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
								 DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Binary readBinaryRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
								 DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Long readLongRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
							 DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Short readShortRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
							   DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_ColorRGBFloat readRGBFloatRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
									 DgpImage_Type endian, Dgp_State *dgp_state);
static DgpImage_Packed readPackedRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile,
								 DgpImage_Type endian, Dgp_State *dgp_state);

ErrCode
DgpImage_loadRaw(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, Dgp_Int8 *filename, void **image, Dgp_State *dgp_state)
{
    FILE *inFile = NULL;
    void *im = NULL;
	
	DGP_PROLOGUE ("DgpImage_loadRaw", DGP_PERRSTATE);
	
    /* Open the file, imLoadRawF on it, then close it */
    DGP_ASSERT(filename != (Dgp_Int8 *)NULL);
    inFile = fopen(filename, "rb");
    if (inFile == (FILE *)NULL) {
		DGP_ERR ( cErrFileNotFound );
	}
	
    DGP_ERR_PASS_CLEANUP (DgpImage_loadRawF(type, width, height, inFile, &im, dgp_state));
	
    DGP_ERR_GEN_CLEANUP ((fclose(inFile) == EOF), cErrIO_ReadError);
	
	  *image = im;
    im = NULL;

    NORMAL_RETURN;

ERR_CLEANUP:

    if(im != NULL) DgpImage_free(im);

    if(inFile != NULL) fclose(inFile);

    ERR_RETURN;
}

ErrCode
DgpImage_loadRawF(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, void **image, Dgp_State *dgp_state)
{
    DgpImage_Type endian;
	
	DGP_PROLOGUE ("DgpImage_loadRawF", DGP_PERRSTATE);
	
    DGP_ASSERT(inFile != (FILE *)NULL);
	
    /* Peel off the endianness directive */
    endian = type & (DGPIMAGE_LITTLEENDIAN | DGPIMAGE_BIGENDIAN);
    type &= ~endian;
	
    switch (type) {
	case DGPIMAGE_GRAY: {
		*image = (void *)readGrayRaw(width, height, inFile, endian, dgp_state);
		break;
					 }
		
	case DGPIMAGE_FLOAT: {
		*image = (void *)readFloatRaw(width, height, inFile, endian, dgp_state);
		break;
					  }
		
	case DGPIMAGE_RGB: {
		*image = (void *)readRGBRaw(width, height, inFile, endian, dgp_state);
		break;
					}
		
	case DGPIMAGE_DOUBLE: {
		*image = (void *)readDoubleRaw(width, height, inFile, endian, dgp_state);
		break;
					   }
		
	case DGPIMAGE_BINARY: {
		*image = (void *)readBinaryRaw(width, height, inFile, endian, dgp_state);
		break;
					   }
		
	case DGPIMAGE_LONG: {
		*image = (void *)readLongRaw(width, height, inFile, endian, dgp_state);
		break;
					 }
		
	case DGPIMAGE_SHORT: {
		*image = (void *)readShortRaw(width, height, inFile, endian, dgp_state);
		break;
					  }
		
	case DGPIMAGE_RGBFLOAT: {
		*image = (void *)readRGBFloatRaw(width, height, inFile, endian, dgp_state);
		break;
						 }
		
	case DGPIMAGE_PACKED: {
		*image = (void *)readPackedRaw(width, height, inFile, endian, dgp_state);
		break;
					   }
		
	default: {
		DGP_ERRF0 ( cErrInternal, "Invalid image type" );	
			 }
	}

    NORMAL_RETURN;
}

/*
* Shuffle each of nelem size-wide entities, starting at ptr, assuming
* that they are initially set up according to endian.
*/
static void
shuffle(Dgp_UInt8 *ptr, size_t size, Dgp_Int32 nelem, DgpImage_Type endian)
{
    Dgp_Int32 i, j;
    Dgp_UInt8 tmp;
	
    /*
	* Do we have to do anything?  endian==0 means "data are in native
	* endianness", so we only have to shuffle if endian explicitly
	* indicates that the data are in the opposite endianness
	*/
#ifdef	_BIG_ENDIAN
    if (endian == DGPIMAGE_LITTLEENDIAN)
#else
		if (endian == DGPIMAGE_BIGENDIAN)
#endif
		{
			for (i = 0; i < nelem; i++, ptr += size) {
				for (j = 0; j < (int)size / 2; j++) {
					tmp = ptr[j];
					ptr[j] = ptr[size - 1 - j];
					ptr[size - 1 - j] = tmp;
				}
			}
		}
}

/*
* Write out nelem size-wide elems, starting at ptr, forcing the output
* to be according to endian.
*/
static Dgp_Bool
writeShuffling(Dgp_Int8 *ptr, size_t size, size_t nelem, FILE *outFile,
			   DgpImage_Type endian)
{
    Dgp_Int32 i, j;
	
#ifdef	_BIG_ENDIAN
    if (endian == DGPIMAGE_LITTLEENDIAN)
#else
		if (endian == DGPIMAGE_BIGENDIAN)
#endif
		{
			for (i = 0; i < (int)nelem; i++, ptr += size) {
				for (j = 0; j < (int)size; j++) {
					if (putc(ptr[size - 1 - j], outFile) == EOF) {
						return(DGP_FALSE);
					}
				}
			}
		}
		else {
			/* We can write it out as a single block */
			if (fwrite(ptr, size, nelem, outFile) != nelem) {
				return(DGP_FALSE);
			}
		}
		
		return(DGP_TRUE);    
}

/*
* Lots of reader functions.  Most of them are pretty simple: allocate
* the image, and inhale straight into it.  Then run through and do
* endian switching if required.
*/
static DgpImage_Gray
readGrayRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{

	DgpImage_Gray im;
    
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY | DGPIMAGE_DONTCLEAR,
				    width, height, (void**) &im);
	
    if (im == (DgpImage_Gray)NULL) {
		return((DgpImage_Gray)NULL);
	}
    
    if (!DgpImage_readGrayRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Gray)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readGrayRaw(DgpImage_Gray im, Dgp_Int32 width, Dgp_Int32 height,
				FILE *inFile, DgpImage_Type endian)
{
    /* stop compiler from whining */
    (void)endian;
	
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    /* No endian shuffling required */
    return(DGP_TRUE);
}

static DgpImage_Float
readFloatRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_Float im;
	
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_FLOAT | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_Float)NULL) {
		return((DgpImage_Float)NULL);
	}
    
    if (!DgpImage_readFloatRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Float)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readFloatRaw(DgpImage_Float im, Dgp_Int32 width, Dgp_Int32 height,
				 FILE *inFile, DgpImage_Type endian)
{
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    shuffle((Dgp_UInt8 *)im->data[0], sizeof(float),
		width * height, endian);
	
    return(DGP_TRUE);
}

static DgpImage_ColorRGB
readRGBRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_ColorRGB im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_ColorRGB)NULL) {
		return((DgpImage_ColorRGB)NULL);
	}
    
    if (!DgpImage_readRGBRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_ColorRGB)NULL);
	}
	
    return(im);
}

/*
* This one is slightly complicated by the fact that on some machines,
* sizeof(RGB) is 4 and on some machines it's 3.
*/
Dgp_Bool
DgpImage_readRGBRaw(DgpImage_ColorRGB im, Dgp_Int32 width, Dgp_Int32 height,
			   FILE *inFile, DgpImage_Type endian)
{
    /* stop compiler from whining */
    (void)endian;
	
    if (sizeof(DgpImage_RGB) == 3) {
		/* It's stored in the right way - just suck it all in. */
		if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
			(size_t)(width * height), inFile) != (size_t)(width * height)) {
			return(DGP_FALSE);
		}
	}
    else {
		/* This might go faster if I did some temporary buffering. */
		Dgp_Int32 x, y;
		DgpImage_RGB *p;
		Dgp_Int32 c;
		
		for (y = 0; y < height; y++) {
			for (x = 0, p = im->data[y]; x < width; x++, p++) {
				c = getc(inFile);
				if (c == EOF) {
					return(DGP_FALSE);
				}
				p->r = (Dgp_UInt8) c;
				
				c = getc(inFile);
				if (c == EOF) {
					return(DGP_FALSE);
				}
				p->g = (Dgp_UInt8) c;
				
				c = getc(inFile);
				if (c == EOF) {
					return(DGP_FALSE);
				}
				p->b = (Dgp_UInt8) c;
			}
		}
	}
	
    /* No endian shuffling required */
    return(DGP_TRUE);
}

static DgpImage_Double
readDoubleRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_Double im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_DOUBLE | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_Double)NULL) {
		return((DgpImage_Double)NULL);
	}
    
    if (!DgpImage_readDoubleRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Double)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readDoubleRaw(DgpImage_Double im, Dgp_Int32 width, Dgp_Int32 height,
				  FILE *inFile, DgpImage_Type endian)
{
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    shuffle((Dgp_UInt8 *)im->data[0], sizeof(double),
		width * height, endian);
	
    return(DGP_TRUE);
}

static DgpImage_Binary
readBinaryRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
/*
* A raw binary image is packed 8 bits per byte, so we read it
* as a DgpImage_Packed then convert.
	*/
    DgpImage_Packed pim = readPackedRaw(width, height, inFile, endian, dgp_state);
    DgpImage_Binary im;
	
    if (pim == (DgpImage_Packed)NULL) {
		return((DgpImage_Binary)NULL);
	}
	
    DGPIMAGE_CONVERT(pim, DGPIMAGE_BINARY, (void**) &im);
    if (im == (DgpImage_Binary)NULL) {
		DGPIMAGE_FREE(pim);
		return((DgpImage_Binary)NULL);
	}
	
    return(im);
}

static DgpImage_Long
readLongRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_Long im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_LONG | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_Long)NULL) {
		return((DgpImage_Long)NULL);
	}
    
    if (!DgpImage_readLongRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Long)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readLongRaw(DgpImage_Long im, Dgp_Int32 width, Dgp_Int32 height,
				FILE *inFile, DgpImage_Type endian)
{
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    shuffle((Dgp_UInt8 *)im->data[0], sizeof(long),
		width * height, endian);
	
    return(DGP_TRUE);
}

static DgpImage_Short
readShortRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_Short im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_SHORT | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_Short)NULL) {
		return((DgpImage_Short)NULL);
	}
    
    if (!DgpImage_readShortRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Short)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readShortRaw(DgpImage_Short im, Dgp_Int32 width, Dgp_Int32 height,
				 FILE *inFile, DgpImage_Type endian)
{
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    shuffle((Dgp_UInt8 *)im->data[0], sizeof(short),
		width * height, endian);
	
    return(DGP_TRUE);
}

static DgpImage_ColorRGBFloat
readRGBFloatRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_ColorRGBFloat im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGBFLOAT | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_ColorRGBFloat)NULL) {
		return((DgpImage_ColorRGBFloat)NULL);
	}
    
    if (!DgpImage_readRGBFloatRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_ColorRGBFloat)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readRGBFloatRaw(DgpImage_ColorRGBFloat im, Dgp_Int32 width, Dgp_Int32 height,
					FILE *inFile, DgpImage_Type endian)
{
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(im->data[0][0]),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		return(DGP_FALSE);
	}
	
    /* slightly unusual shuffle here */
    shuffle((Dgp_UInt8 *)im->data[0], sizeof(float),
		3 * width * height, endian);
	
    return(DGP_TRUE);
}


static DgpImage_Packed
readPackedRaw(Dgp_Int32 width, Dgp_Int32 height, FILE *inFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
    DgpImage_Packed im;
	DGPIMAGE_NEW(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR,
		width, height, (void**) &im);
	
    if (im == (DgpImage_Packed)NULL) {
		return((DgpImage_Packed)NULL);
	}
    
    if (!DgpImage_readPackedRaw(im, width, height, inFile, endian)) {
		DGPIMAGE_FREE(im);
		return((DgpImage_Packed)NULL);
	}
	
    return(im);
}

Dgp_Bool
DgpImage_readPackedRaw(DgpImage_Packed im, Dgp_Int32 width, Dgp_Int32 height,
				  FILE *inFile, DgpImage_Type endian)
{
    Dgp_Int32 x, y;
    Dgp_UInt8 *p;
#ifdef	_LITTLE_ENDIAN
    Dgp_UInt32 *q;
#endif
    Dgp_Int32 extrapad;
    Dgp_Int32 row_width;
    Dgp_Int32 file_row_width;
    Dgp_Int32 padmask;
	Dgp_State *dgp_state = im->dgp_state;

	DGP_VOID_PROLOGUE ("DgpImage_readPackedRaw", DGP_PERRSTATE);
	
    /* stop compiler from whining */
    (void)endian;
	
    /*
	* This one is harder because of the alignment issues of
	* DgpImage_Packeds - we can't just inhale.
	*/
	
    /*
	* See if we can inhale it as a single blob. We'll be able to do this
	* when the row length, as we padded it, is the same as the row length,
	* as PBM pads it, AND we're running on a bigendian machine.
	*/
    row_width = DGPIMG_ROWLEN(width);
    file_row_width = (width + 7) / 8;
    extrapad = row_width - file_row_width;
    if (extrapad == 0) {
		/* Yup. */
		p = (Dgp_UInt8 *)im->data[0];
		if (fread((Dgp_Int8 *)p, sizeof(Dgp_UInt8),
			(size_t)(row_width * height), inFile) !=
			(size_t)(row_width * height)) {
			return(DGP_FALSE);
		}
	}
    else {
		/* Have to do it a row at a time. */
		p = (Dgp_UInt8 *)im->data[0];
		for (y = 0; y < height; y++) {
			DGP_VOID_ASSERT(p == (Dgp_UInt8 *)im->data[y]);
			
			if (fread((Dgp_Int8 *)p, sizeof(Dgp_UInt8),
				(size_t)file_row_width, inFile) !=
				(size_t)file_row_width) {
				return(DGP_FALSE);
			}
			
			/* Clear out the extra padding */
			p += file_row_width;
			for (x = 0; x < extrapad; x++) {
				*p++ = 0;
			}
		}
	}
	
    /*
	* In either case, the file might have contained some '1' bits in the last
	* byte of each row, outside the image bounds. Argh.  Also, we need
	* to go through and flip the order of the bytes within each long, to
	* make sure it's set up for long-by-long access.
	*/
    if (width % 8) {
		/* There are pad bits */
		
		/* Mask off the pad bits */
		padmask = ~((1 << (8 - (width % 8))) - 1);
		for (y = 0, p = (Dgp_UInt8 *)im->data[0] + (width / 8);
		y < height;
		y++, p += row_width) {
			*p &= padmask;
		}
	}
	
#ifdef	_LITTLE_ENDIAN
    for (y = 0, p = (Dgp_UInt8 *)im->data[0], q = im->data[0];
	y < height;
	y++) {
		for (x = 0; x < row_width; x += sizeof(Dgp_UInt32)) {
			Dgp_UInt32 tmp = 0;
			Dgp_Int32 ix;
			
			for (ix = 0; ix < (int)sizeof(Dgp_UInt32); ix++) {
				tmp |= (*p++) << (DGP_BITS_UINT32 - (ix + 1) * 8);
			}
			*q++ = tmp;
		}
	}
#endif
	
    return(DGP_TRUE);
}

ErrCode
DgpImage_saveRaw_(void *im, DgpImage_Type tag, Dgp_Int8 *filename, DgpImage_Type endian, Dgp_State *dgp_state)
{
    FILE *outFile = NULL;
    
	DGP_PROLOGUE ("DgpImage_saveRaw_", DGP_PERRSTATE);
	
    DGP_ASSERT(filename != (Dgp_Int8 *)NULL);
    DGP_ASSERT(im != (void *)NULL);
	
    /* Open the file, imSaveRawF on it, then close it */
    outFile = fopen(filename, "wb");
  if (outFile == (FILE *)NULL) {
		DGP_ERR ( cErrFileNotFound );
	}
	
    /* Write the image */
  DGP_ERR_PASS_CLEANUP (DgpImage_saveRawF_(im, tag, outFile, endian, dgp_state));
	
    /* and close the file */
  DGP_ERR_GEN_CLEANUP ((fclose(outFile) == EOF), cErrIO_WriteError);
	
    NORMAL_RETURN;

ERR_CLEANUP:

    if(outFile != NULL) fclose(outFile);

    ERR_RETURN;
}

ErrCode
DgpImage_saveRawF_(void *im, DgpImage_Type tag, FILE *outFile, DgpImage_Type endian, Dgp_State *dgp_state)
{
	
	DGP_PROLOGUE ("DgpImage_saveRawF_", DGP_PERRSTATE);
	
    switch (tag) {
	case DGPIMAGE_GRAY: {
		DGP_ERR_GEN (!(DgpImage_saveGrayRaw((DgpImage_Gray)im, outFile, endian)),
      cErrIO_WriteError);
    break;
                      }
	case DGPIMAGE_FLOAT: {
		DGP_ERR_GEN (!(DgpImage_saveFloatRaw((DgpImage_Float)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					  }
		
	case DGPIMAGE_RGB: {
		DGP_ERR_GEN (!(DgpImage_saveRGBRaw((DgpImage_ColorRGB)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					}
		
	case DGPIMAGE_DOUBLE: {
		DGP_ERR_GEN (!(DgpImage_saveDoubleRaw((DgpImage_Double)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					   }
		
	case DGPIMAGE_BINARY: {
		DGP_ERR_GEN (!(DgpImage_saveBinaryRaw((DgpImage_Binary)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					   }
		
	case DGPIMAGE_LONG: {
		DGP_ERR_GEN (!(DgpImage_saveLongRaw((DgpImage_Long)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					 }
		
	case DGPIMAGE_SHORT: {
		DGP_ERR_GEN (!(DgpImage_saveShortRaw((DgpImage_Short)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					  }
		
	case DGPIMAGE_RGBFLOAT: {
		DGP_ERR_GEN (!(DgpImage_saveRGBFloatRaw((DgpImage_ColorRGBFloat)im, outFile, endian)),
      cErrIO_WriteError);
    break;
						 }
	case DGPIMAGE_PACKED: {
		DGP_ERR_GEN (!(DgpImage_savePackedRaw((DgpImage_Packed)im, outFile, endian)),
      cErrIO_WriteError);
    break;
					   }
		
	default: {
		DGP_ERRF0 ( cErrInternal, "Invalid image type" );
			 }
	}

    NORMAL_RETURN;
}
#define DGPIMAGE_SAVEBYLINE(size, nelem) \
{ \
    Dgp_Int32 y; \
    Dgp_UInt8 *p; \
    for (y = 0; y < im->header.height; y++) { \
        p = (Dgp_Int8 *) (im->data[y + im->header.ybase] + im->header.xbase); \
	if ( fwrite (p, 1, size * nelem, outFile) !=  (size_t)( nelem *  size) ) \
            return(DGP_FALSE); \
    } \
}
#define DGPIMAGE_SAVEBYLINE_SHUFFLE(size, nelem, endian) \
{ \
    Dgp_Int32 y; \
    Dgp_UInt8 *p; \
    for (y = 0; y < im->header.height; y++) { \
        p = (Dgp_Int8 *) (im->data[y + im->header.ybase] + im->header.xbase); \
	if ( writeShuffling (p, 1, size * nelem, outFile, endian) !=  DGP_TRUE ) \
            return(DGP_FALSE); \
    } \
}

/*
* Writer functions.  Those that don't have to shuffle the bytes
* just spew the whole memory area.  Those that have to shuffle end up
* doing byte-by-byte writes - slow.
*/
Dgp_Bool
DgpImage_saveGrayRaw(DgpImage_Gray im, FILE *outFile, DgpImage_Type endian)
{
    /* stop compiler from whining */
    (void)endian;
    DGPIMAGE_SAVEBYLINE(sizeof(im->data[0][0]), im->header.width); 
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveFloatRaw(DgpImage_Float im, FILE *outFile, DgpImage_Type endian)
{
    DGPIMAGE_SAVEBYLINE_SHUFFLE(sizeof(im->data[0][0]), im->header.width, endian); 
    return(DGP_TRUE);
}

/*
* This one is slightly complicated by the fact that on some machines,
* sizeof(RGB) is 4 and on some machines it's 3.
*/
Dgp_Bool
DgpImage_saveRGBRaw(DgpImage_ColorRGB im, FILE *outFile, DgpImage_Type endian)
{
    /* stop compiler from whining */
    (void)endian;
	
    if (sizeof(DgpImage_RGB) == 3) {
       DGPIMAGE_SAVEBYLINE(sizeof(im->data[0][0]), im->header.width); 
    } else {
        Dgp_Int32 x, y;
        DgpImage_RGB *p;
	/*
	* Do it Dgp_Int8-by-Dgp_Int8 - slow but safe.
	* Some explicit buffering here might speed it up.
		*/
		for (y = 0; y < im->header.height; y++) {
			for (x = 0, p = im->data[y + im->header.ybase] + im->header.xbase;
			x < im->header.width; x++, p++) {
				if ((putc(p->r, outFile) == EOF) ||
					(putc(p->g, outFile) == EOF) ||
					(putc(p->b, outFile) == EOF)) {
					return(DGP_FALSE);
				}
			}
		}
	}
	
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveDoubleRaw(DgpImage_Double im, FILE *outFile, DgpImage_Type endian)
{

    DGPIMAGE_SAVEBYLINE_SHUFFLE(sizeof(im->data[0][0]), im->header.width, endian); 
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveBinaryRaw(DgpImage_Binary im, FILE *outFile, DgpImage_Type endian)
{
    /* For binary images, have to write differently */
    Dgp_Int32 i, j, c, bitsleft, height, width;
	
    /* stop compiler from whining */
    (void)endian;
	
    height = (int)im->header.height + im->header.ybase;
    width = (int)im->header.width + im->header.xbase;
	
    for (i = im->header.ybase; i < height; i++) {
		bitsleft = 8;
		c = 0;
		for (j = im->header.xbase; j < width; j++) {
			--bitsleft;
			if (DGPIMG_REF((DgpImage_Binary)im, j, i)) {
				c |= 1 << bitsleft;
			}
			if ((bitsleft == 0) || (j == width - 1)) {
				if (putc(c, outFile) == EOF) {
					return(DGP_FALSE);
				}
				bitsleft = 8;
				c = 0;
			}
		}
	}
	
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveLongRaw(DgpImage_Long im, FILE *outFile, DgpImage_Type endian)
{

    DGPIMAGE_SAVEBYLINE_SHUFFLE(sizeof(im->data[0][0]), im->header.width, endian); 
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveShortRaw(DgpImage_Short im, FILE *outFile, DgpImage_Type endian)
{

    DGPIMAGE_SAVEBYLINE_SHUFFLE(sizeof(im->data[0][0]), im->header.width, endian); 
    return(DGP_TRUE);
}

Dgp_Bool
DgpImage_saveRGBFloatRaw(DgpImage_ColorRGBFloat im, FILE *outFile, DgpImage_Type endian)
{
    /* slightly unusual here */

    DGPIMAGE_SAVEBYLINE_SHUFFLE(sizeof(float), 3 * im->header.width, endian); 
    return(DGP_TRUE);
}


#endif

Dgp_Bool
DgpImage_savePackedRaw(DgpImage_Packed im, FILE *outFile, DgpImage_Type endian)
{
    Dgp_Int32 y;
    Dgp_Int32 row_width, file_row_width, imlrow_width;
    Dgp_Int32 height;
    Dgp_UInt8 *p;
#ifdef	_LITTLE_ENDIAN
    Dgp_UInt32 *q, *lq;
    Dgp_Int32 x;
#endif

	Dgp_State *dgp_state = im->dgp_state;

	DGP_VOID_PROLOGUE ("DgpImage_savePackedRaw", DGP_PERRSTATE);

    /* stop compiler from whining */
    (void)endian;
	
    DGP_VOID_ASSERT(im != (DgpImage_Packed)NULL);
    DGP_VOID_ASSERT(outFile != (FILE *)NULL);
	
    row_width = DGPIMG_ROWLEN(DGPIMG_WIDTH(im));	/* # longs containing data */
    /* how many actual longs/row */
    imlrow_width = im->header.row_width / sizeof(Dgp_UInt32);
    /* and bytes/row in the file */
    file_row_width = (DGPIMG_WIDTH(im) + 7) / 8;
    height = DGPIMG_HEIGHT(im);
	
#ifdef	_LITTLE_ENDIAN
    /*
	* Run through and convert it for byte-by-byte access.
	*/
    for (y = 0, lq = im->data[im->header.ybase];
	y < height;
	y++, lq += imlrow_width) {
		for (x = 0, q = lq, p = (Dgp_UInt8 *)lq;
		x < row_width;
		x += sizeof(Dgp_UInt32)) {
			Dgp_UInt32 tmp = *q++;
			Dgp_Int32 ix;
			
			for (ix = 0; ix < (int)sizeof(Dgp_UInt32); ix++) {
				*p++ = (Dgp_UInt8) ((tmp >> (DGP_BITS_UINT32 - (ix + 1) * 8)) & 0xff);
			}
		}
	}
#endif
	
    /* Can we spew one chunk? */
    if (im->header.row_width == file_row_width) {
		/* Yup. */
		p = (Dgp_UInt8 *)im->data[DGPIMG_YBASE(im)];
		if (fwrite((Dgp_Int8 *)p, sizeof(Dgp_UInt8),
			(size_t)(row_width * height), outFile) !=
			(size_t)(row_width * height)) {
			return(DGP_FALSE);
		}
	}
    else {

		for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
			p = (Dgp_UInt8 *)im->data[y];
			if (fwrite((Dgp_Int8 *)p, sizeof(Dgp_UInt8),
				(size_t)file_row_width, outFile) !=
				(size_t)file_row_width) {
				return(DGP_FALSE);
			}
		}
	}
	
#ifdef	_LITTLE_ENDIAN
    /*
	* Now go through and undo that byte fiddling - yeah, this is
	* suboptimal, but the other choices are doing byte-by-byte I/O
	* or allocating extra memory...
	*/
    for (y = 0, lq = im->data[im->header.ybase];
	y < height;
	y++, lq += imlrow_width) {
		for (x = 0, q = lq, p = (Dgp_UInt8 *)lq;
		x < row_width;
		x += sizeof(Dgp_UInt32)) {
			Dgp_UInt32 tmp = 0;
			Dgp_Int32 ix;
			
			for (ix = 0; ix < (int)sizeof(Dgp_UInt32); ix++) {
				tmp |= (*p++) << (DGP_BITS_UINT32 - (ix + 1) * 8);
			}
			*q++ = tmp;
		}
	}
#endif
	
    return(DGP_TRUE);
}
