/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/readpts.c,v 1.3 2003/01/27 21:48:32 awilliam Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
 *  
 * Use, reproduction, preparation of derivative works, and distribution
 * of this software is permitted.  Any copy of this software or of any
 * derivative work must include both the above copyright notices of
 * Cornell University and Xerox Corporation and this paragraph.  Any
 * distribution of this software or derivative works must comply with all
 * applicable United States export control laws.  This software is made
 * available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
 * LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#include "util.h"
#include <stdio.h>
#include <stdlib.h>
#include "image.h"
#include "image-private.h"
#include "image_defs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/readpts.c,v 1.3 2003/01/27 21:48:32 awilliam Exp $");


/*
 * This routine reads a list of points from the given stream. It expects
 * the format
 *	xsize ysize
 *	npts
 *	p1.x p1.y
 *	p2.x p2.y
 *	...
 *
 * Alternately, it can be fed a pbm file and will generate the list itself.
 * This may take longer. It might also be faster because it doesn't have
 * to parse all those decimal numbers; it depends on how sparse the image
 * is. For regular text images, reading the pbm file is probably faster.
 *
 * It returns the appropriate values in the given variables, as you would
 * expect.
 *
 * If pointim is NULL, it is ignored.
 * If it is non-NULL, it is taken to be a pointer to a DgpImage_Binary variable,
 * and on return the variable will contain a DgpImage_Binary which represents
 * the points read from the image.
 *
 * It returns NULL if something goes wrong. In this case, the values of the
 * parameters passed are undefined (including the file position).
 *
 */
ErrCode
DgpImage_read_pts(FILE *file, Dgp_Int32 *xsize, Dgp_Int32 *ysize, Dgp_Int32 *npts,
	 DgpImage_Binary *pointim, DgpImage_Point **retpoint, Dgp_State *dgp_state)
{
	DgpImage_Point *data;
    Dgp_Int32 i;
    Dgp_Int32 c;
    DgpImage_Binary im;
    Dgp_Int32 x, y;

	DGP_PROLOGUE ("DgpImage_read_pts", DGP_PERRSTATE);

    DGP_ASSERT(file != (FILE *)NULL);
    DGP_ASSERT(xsize != (Dgp_Int32 *)NULL);
    DGP_ASSERT(ysize != (Dgp_Int32 *)NULL);
    DGP_ASSERT(npts != (Dgp_Int32 *)NULL);

    /*
     * We have to figure out if we're reading a pointlist or a pbm
     * file. Unfortunately, the file we're reading from might be a pipe -
     * we can't read too much to decide. In fact, one Dgp_Int8acter...
     */
    c = getc(file);
    if (c == EOF) {
		DGP_ERR (cErrFileNotFound);
	}
    /* One Dgp_Int8acter of pushback is always OK */
    ungetc(c, file);

    if (c == 'P') {
	/* We've got a pbm file */
	DGP_ERR_PASS (DgpImage_pLoadF (dgp_state, DGPIMAGE_BINARY, file, (void**) &im));
	if (im == (DgpImage_Binary)NULL) {
	    DGP_ERR (cErrFileNotFound);
	    }

	*npts = 0;
	for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
	    for (x = DGPIMG_XBASE(im); x <= DGPIMG_XMAX(im); x++) {
		if (DGPIMG_REF(im, x, y) == 1) {
		    (*npts)++;
		    }
		}
	    }

	DGP_CHECK_ALLOC (data = (DgpImage_Point *)(DGP_MALLOC)(*npts * sizeof(DgpImage_Point)));

	i = 0;
	for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
	    for (x = DGPIMG_XBASE(im); x <= DGPIMG_XMAX(im); x++) {
		if (DGPIMG_REF(im, x, y) == 1) {
		    data[i].x = x;
		    data[i].y = y;
		    i++;
		    }
		}
	    }

	*xsize = DGPIMG_WIDTH(im);
	*ysize = DGPIMG_HEIGHT(im);

	if (pointim != (DgpImage_Binary *)NULL) {
	    *pointim = im;
	    }
	else {
	    DGPIMAGE_FREE(im);
	    }
	}
    else {	    
	/* A point list file */
	if (fscanf(file, "%d %d", (int *)xsize, (int *)ysize) != 2) {
	    DGP_ERR (cErrFileNotFound);
	    }

	if (fscanf(file, "%d", (int *)npts) != 1) {
	    DGP_ERR (cErrFileNotFound);
	    }

	DGP_CHECK_ALLOC (data = (DgpImage_Point *)(DGP_MALLOC)(*npts * sizeof(DgpImage_Point)));

	for (i = 0; i < *npts; i++) {
	    if (fscanf(file, "%ld %ld", &(data[i].x), &(data[i].y)) != 2) {
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}
	    if ((data[i].x < 0) || (data[i].x >= *xsize) ||
		(data[i].y < 0) || (data[i].y >= *ysize)) {
		/* Bad poDgp_Int32 */
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}
	    }
    
	if (pointim != (DgpImage_Binary *)NULL) {
	    /* Gotta convert it into a DgpImage_Binary */
	    DGP_ERR_PASS (DgpImage_pts_to_bim(*xsize, *ysize, 0, 0, *npts, data, &im, dgp_state));
	    if (im == (DgpImage_Binary)NULL) {
		/* I bet they regret they ever asked... */
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}

	    *pointim = im;
	    }
	}

	*retpoint = data;
    NORMAL_RETURN;
}

/*
 * Convert the poDgp_Int32 list to a binary image of the given shape.
 * Points must be in range. If anything goes wrong, (DgpImage_Binary)NULL
 * is returned.
 */
ErrCode
DgpImage_pts_to_bim(Dgp_Int32 width, Dgp_Int32 height,
	   Dgp_Int32 xbase, Dgp_Int32 ybase,
	   Dgp_Int32 npts, DgpImage_Point *pts, DgpImage_Binary *bim, Dgp_State *dgp_state)
{
    Dgp_Int32 i;
    DgpImage_Binary im;

	DGP_PROLOGUE ("DgpImage_pts_to_bim", DGP_PERRSTATE);

    DGP_ASSERT(pts != (DgpImage_Point *)NULL);

    DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_BINARY, width, height, 
		xbase, ybase, (void**) &im));

    for (i = 0; i < npts; i++) {
	DGP_ASSERT(((pts[i].x >= xbase) && (pts[i].x < xbase + (Dgp_Int32)width)) &&
	       ((pts[i].y >= ybase) && (pts[i].y < ybase + (Dgp_Int32)height)));
	DGPIMG_REF(im, pts[i].x, pts[i].y) = 1;
	}

	*bim = im;
    NORMAL_RETURN;
    }

/*
 * Same as read_pts, but it produces a DgpImage_Packed.
 */
ErrCode
DgpImage_read_pts_packed(FILE *file, Dgp_Int32 *xsize, Dgp_Int32 *ysize, Dgp_Int32 *npts,
		DgpImage_Packed *pointim, DgpImage_Point **retpoint, Dgp_State *dgp_state)
{
    DgpImage_Point *data;
    Dgp_Int32 i;
    Dgp_Int8 c;
    DgpImage_Packed im;
    Dgp_Int32 x, y;

	DGP_PROLOGUE ("DGPIMAGE_read_pts_packed", DGP_PERRSTATE);

    DGP_ASSERT(file != (FILE *)NULL);
    DGP_ASSERT(xsize != (Dgp_Int32 *)NULL);
    DGP_ASSERT(ysize != (Dgp_Int32 *)NULL);
    DGP_ASSERT(npts != (Dgp_Int32 *)NULL);

    /*
     * We have to figure out if we're reading a pointlist or a pbm
     * file. Unfortunately, the file we're reading from might be a pipe -
     * we can't read too much to decide. In fact, one Dgp_Int8acter...
     */
    c = getc(file);
    if (c == EOF) {
	DGP_ERR (cErrFileNotFound);
	}
    /* One Dgp_Int8acter of pushback is always OK */
    ungetc(c, file);

    if (c == 'P') {
	/* We've got a pbm file */
	DGP_ERR_PASS (DgpImage_pLoadF (dgp_state, DGPIMAGE_PACKED, file, (void**) &im));
	if (im == (DgpImage_Packed)NULL) {
	    DGP_ERR (cErrFileNotFound);
	    }

	*npts = 0;
	for (y = DGPIMG_YBASE(im);
	     y < DGPIMG_YBASE(im) + (Dgp_Int32)DGPIMG_HEIGHT(im);
	     y++) {
	    for (x = DGPIMG_XBASE(im);
		 x < DGPIMG_XBASE(im) + (Dgp_Int32)DGPIMG_WIDTH(im);
		 x++) {
		if (DGPIMG_GETBIT(im, x, y)) {
		    (*npts)++;
		    }
		}
	    }

	DGP_CHECK_ALLOC (data = (DgpImage_Point *)(DGP_MALLOC)(*npts * sizeof(DgpImage_Point)));

	i = 0;
	for (y = DGPIMG_YBASE(im);
	     y < DGPIMG_YBASE(im) + (Dgp_Int32)DGPIMG_HEIGHT(im);
	     y++) {
	    for (x = DGPIMG_XBASE(im);
		 x < DGPIMG_XBASE(im) + (Dgp_Int32)DGPIMG_WIDTH(im);
		 x++) {
		if (DGPIMG_GETBIT(im, x, y)) {
		    data[i].x = x;
		    data[i].y = y;
		    i++;
		    }
		}
	    }

	*xsize = DGPIMG_WIDTH(im);
	*ysize = DGPIMG_HEIGHT(im);

	if (pointim != (DgpImage_Packed *)NULL) {
	    *pointim = im;
	    }
	else {
	    DGPIMAGE_FREE(im);
	    }
	}
    else {	    
	/* A point list file */
	if (fscanf(file, "%d %d", (int *)xsize, (int *)ysize) != 2) {
	    DGP_ERR (cErrFileNotFound);
	    }

	if (fscanf(file, "%d", (int *)npts) != 1) {
	    DGP_ERR (cErrFileNotFound);
	    }

	DGP_CHECK_ALLOC (data = (DgpImage_Point *)(DGP_MALLOC)(*npts * sizeof(DgpImage_Point)));

	for (i = 0; i < *npts; i++) {
	    if (fscanf(file, "%ld %ld", &(data[i].x), &(data[i].y)) != 2) {
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}
	    if ((data[i].x < 0) || (data[i].x >= *xsize) ||
		(data[i].y < 0) || (data[i].y >= *ysize)) {
		/* Bad poDgp_Int32 */
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}
	    }
    
	if (pointim != (DgpImage_Packed *)NULL) {
	    /* Gotta convert it into a DgpImage_Packed */
	    DGP_ERR_PASS (DgpImage_pts_to_pim(*xsize, *ysize, 0, 0, *npts, data, &im, dgp_state));
	    if (im == (DgpImage_Packed)NULL) {
		/* I bet they regret they ever asked... */
		(DGP_FREE)((void *)data);
		DGP_ERR (cErrFileNotFound);
		}

	    *pointim = im;
	    }
	}

	*retpoint = data;
    NORMAL_RETURN;
}

/*
 * Convert the poDgp_Int32 list to a packed binary image of the given shape.
 * Points must be in range. If anything goes wrong, (DgpImage_Packed)NULL
 * is returned.
 */
ErrCode
DgpImage_pts_to_pim(Dgp_Int32 width, Dgp_Int32 height,
	   Dgp_Int32 xbase, Dgp_Int32 ybase,
	   Dgp_Int32 npts, DgpImage_Point *pts, DgpImage_Packed *pim, Dgp_State *dgp_state)
{
    Dgp_Int32 i;
    DgpImage_Packed im;

	DGP_PROLOGUE ("DgpImage_pts_to_pim", DGP_PERRSTATE);

    DGP_ASSERT(pts != (DgpImage_Point *)NULL);

    DGP_ERR_PASS (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_PACKED, width, height, 
		xbase, ybase, (void**) &im));

    for (i = 0; i < npts; i++) {
	DGP_ASSERT(((pts[i].x >= xbase) && (pts[i].x < xbase + width)) &&
	       ((pts[i].y >= ybase) && (pts[i].y < ybase + height)));
	DGPIMG_SETBIT(im, pts[i].x, pts[i].y);
	}

	*pim = im;
    NORMAL_RETURN;
}

/*
 * This routine takes the array of points it is given and randomises them.
 * This is useful for many applications.
 */
ErrCode
DgpImage_shuffle_pts(DgpImage_Point *pts, Dgp_Int32 npts, Dgp_State *dgp_state)
{
    Dgp_Int32 i;
    Dgp_Int32 j;
    DgpImage_Point p;

    DGP_PROLOGUE ("DgpImage_shuffle_pts", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    for (i = 0; i < npts - 1; i++) {
	j = (rand() % (npts - i)) + i;
	p = pts[i];
	pts[i] = pts[j];
	pts[j] = p;
    }
    NORMAL_RETURN;
}
