/*
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/smooth.c,v 1.5 2003/02/10 22:53:25 mkrocken Exp $
 */
/*
 * Perform Gaussian convolution on images.
 */

#include "smooth.h"
#include "image_defs.h"
#include <math.h>
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/smooth.c,v 1.5 2003/02/10 22:53:25 mkrocken Exp $");

static ErrCode 
dgpimage_makeGaussian(Dgp_Float64 sigma, Dgp_Int32 *width, Dgp_Float64 **result, Dgp_State *dgp_state);


ErrCode
DgpImage_gaussConvol(DgpImage_Gray im, Dgp_Float64 sigma, DgpImage_Float *fImage)
{
    Dgp_Int32 width;
    Dgp_Int32 height;
    DgpImage_Float intermed = NULL;
    DgpImage_Float final = NULL;
    Dgp_Float64 *gaussian = NULL;
    Dgp_Int32 sspread;    
    Dgp_Int32 x, y, j;
    register Dgp_Float32 gaussval0;
	Dgp_State *dgp_state = im->dgp_state;

	DGP_PROLOGUE ("dsutil_gaussConvol", DGP_PERRSTATE);
    DGP_ASSERT (im != (DgpImage_Gray)NULL);

    width = DGPIMG_WIDTH(im);
    height = DGPIMG_HEIGHT(im);

    /* Allocate work areas */
    DGP_ERR_PASS (DGPIMAGE_NEW(dgp_state, DGPIMAGE_FLOAT, width, height, (void**) &intermed));

    DGP_ERR_PASS_CLEANUP (DGPIMAGE_NEW(dgp_state, DGPIMAGE_FLOAT, width, height, (void**) &final));

    /* Generate the (one-sided) Gaussian distribution */
    DGP_ERR_PASS_CLEANUP (dgpimage_makeGaussian(sigma, &sspread, &gaussian, dgp_state));

    /*
     * We do the convolution in two passes: once convolving
     * horizontally with a gaussian, then re-convolving vertically.
     * Since a gaussian is separable, this gives the required
     * two-dimensional convolution.
     *
     * We do this by, for the horizontal direction:
     *  for each x value in the output of the convolution,
     *    for each value in the convolution mask,
     *      find out which x value in the input mask corresponds to
     *      that position in the mask and that position in the output
     *      image
     *      for each y position,
     *        add into the output image the appropriate pixel of the input
     *        image, multiplied by that part of the convolution mask.
     *
     * The vertical direction is similar.
     *
     * Note that since we traverse the image both horizontally and
     * vertically, we are bound to hit unfortunate cache and VM behaviour,
     * due to having a large stride on one of the passes.
     * Since the original image pixel values are smaller than the
     * intermediate or final pixel values (they are small integers,
     * not Dgp_Float32ing poDgp_Int32 numbers), the "worse" direction is computed
     * first to minimise this effect.
     */

    /* Now convolve with a gaussian in the X direction. */
    gaussval0 = (Dgp_Float32) gaussian[0];
    for (x = 0; x < width; x++) {
	/*
	 * When we hit the edge of the image, we want the edge row
	 * or column to continue to infinity, so there is no sharp
	 * transition at the edge, as there would be if we treated all
	 * pixels outside the image as being zero, or if we replicated
	 * the image.
	 */

	/* Do the centre bin of the Gaussian */
	for (y = 0; y < height; y++) {
	    DGPIMG_REF(intermed, x, y) = DGPIMG_REF(im, x, y) * gaussval0;
	    }

	/* and do the side bins */
	for (j = 1; j <= sspread; j++) {
	    /* Find out the actual values of x+-j, with borders sticking. */
	    Dgp_Int32 xval1 = MIN(width - 1, x + j);
	    Dgp_Int32 xval2 = MAX(0, x - j);
	    register Dgp_Float32 gaussval = (Dgp_Float32) gaussian[j];

	    for (y = 0; y < height; y++) {
		DGPIMG_REF(intermed, x, y) +=
		    (DGPIMG_REF(im, xval1, y) + DGPIMG_REF(im, xval2, y)) * gaussval;
		}
	    }
	}

    /* Now convolve with a gaussian in the Y direction */
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    DGPIMG_REF(final, x, y) = DGPIMG_REF(intermed, x, y) * gaussval0;
	    }

	for (j = 1; j <= sspread; j++) {
	    /* Find out the actual values of y+-j, with borders sticking. */
	    Dgp_Int32 yval1 = MIN((int)height - 1, y + j);
	    Dgp_Int32 yval2 = MAX(0, y - j);
	    register Dgp_Float32 gaussval = (Dgp_Float32) gaussian[j];

	    for (x = 0; x < width; x++) {
		DGPIMG_REF(final, x, y) +=
		    (DGPIMG_REF(intermed, x, yval1) + DGPIMG_REF(intermed, x, yval2))
			* gaussval;
		}
	    }
	}

    /* Free all intermediate storage */
    DGPIMAGE_FREE(intermed);
	DGP_FREE (gaussian);
	*fImage = final;
	NORMAL_RETURN;

ERR_CLEANUP:

	if (intermed) DGPIMAGE_FREE(intermed);
	if (final) DGPIMAGE_FREE(final);
	if (gaussian) DGP_FREE (gaussian);

	ERR_RETURN;
}

#ifdef _WIN32
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif
#endif

/* Also VxWorks... 14.3.2000 / dhtesq. */
#ifdef VXWORKS
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif
#endif


/*
 * In many applications, this routine will be called many times with the same
 * value of sigma (or several such values). We want to avoid regenerating
 * the gaussian every time.
 */


/*
 * How many standard deviations out we should generate.
 * 8 is sufficient, since the values out there are about 10^{-15}.
 * Making this too large gives us denorms out in the tails, which can slow
 * processing immensely.
 */
#define SIGMA_SPREAD 8

/*
 * This function generates one side of a Gaussian distribution with standard
 * deviation sigma. The Gaussian is normalised so that its integral is one.
 * The Gaussian returned will have width+1 entries (i.e. will be represented
 * from -width..width) (via symmetry).
 * In other words, it will generate v[0]..v[*width] such that
 * v[0] + 2 * v[1] + ... + 2 * v[*width] = 1
 * and v[*width],...,v[1],v[0],v[1],...,v[*width] is a Gaussian
 * distribution.
 * 
 * Note that the pointer returned *MUST NOT* be freed, and the values it
 * points at should not be modified.
 */

static ErrCode
dgpimage_makeGaussian(Dgp_Float64 sigma, Dgp_Int32 *width, Dgp_Float64 **result, Dgp_State *dgp_state)
{
    Dgp_Float64 *gaussian;
    Dgp_Float64 accum;
    Dgp_Float64 sigma2;
    Dgp_Float64 r2psigma;
	Dgp_Float64 r2p;
    Dgp_Int32 j;

	DGP_PROLOGUE ("dsutil_makeGaussian", DGP_PERRSTATE);

    DGP_ASSERT (sigma >= 0.);
    DGP_ASSERT (width != (Dgp_Int32 *)NULL);

    /* Calculate sqrt(2*M_PI) once only. */
	r2p = sqrt(2 * M_PI);

    sigma2 = -2. * sigma * sigma;
    r2psigma = r2p * sigma;
	

    /*
     * Now find out how wide to make the actual gaussian, based on sigma.
     * SIGMA_SPREAD should be high enough to make the tails insignificant.
     * However, making it too large can slow down computation.
     */
    *width = (int) (SIGMA_SPREAD * sigma + 0.5);
    DGP_CHECK_ALLOC (gaussian =
	(Dgp_Float64 *)(DGP_MALLOC)((*width + 1) * sizeof(*gaussian)));

    /* Now fill in the gaussian array */
    if (sigma == 0.) {
	/* A delta function */
	gaussian[0] = 1.;
	}
    else {
	accum = 0.;
	for (j = 0; j <= *width; j++) {
	    gaussian[j] = 1 / r2psigma * exp(((Dgp_Float64)j) * j / sigma2);
	    accum += (j == 0) ? gaussian[j] : (2 * gaussian[j]);
	    }
	/* Normalise the gaussian so that its total sum is 1. */
	for (j = 0; j <= *width; j++) {
	    gaussian[j] /= accum;
	    }
	}
    
    *result = gaussian;
	NORMAL_RETURN;
}
