#include "mrc2file.h"
#include "PsgFix.h"

#ifndef DGP_FAXMODE
/* create a TIFF-FX file and return a pointer to the data structure */
ErrCode
Dgp_createCompressedDocument (Dgp_State *dgp_state, uds *u, 
			      Dgp_CompressedDocument **cDocument)
{
  Dgp_CompressedDocument *cDoc = NULL;
  DGP_PROLOGUE ("Dgp_createCompressedDocument", DGP_PERRSTATE);
  DOCTYPE outDocType = DOCTYPE_TIFFFX_M1;
  Dgp_UInt8 version = 0;
  Dgp_UInt32 options = 0;
  Dgp_UInt32 myInstanceID = 0;

  DGP_ASSERT (u != NULL);
  DGP_ASSERT (u->fp != NULL);
  DGP_ASSERT (cDocument != NULL);

  /* create structure */
  DGP_CHECK_ALLOC (cDoc = DGP_MALLOC (sizeof (Dgp_CompressedDocument)));
  cDoc->dgp_state = dgp_state;

  /* set default vals */
  cDoc->n_pages = 0;
  cDoc->currentSection = NULL;
  cDoc->n_dict = 0;

  /* Begin a MRC session */
  cDoc->outfile.fp = (FILE *)u->fp;
  cDoc->outfile.pos = 0UL;
  cDoc->outfile.u = u;
  DGP_ERR_PASS_CLEANUP(IO_InitInstance(myInstanceID, 
				       &(cDoc->ioInstanceID)));
  DGP_ERR_PASS_CLEANUP(IO_CreateDocument(cDoc->ioInstanceID, outDocType, 
					 (Dgp_Int32)&(cDoc->outfile), version, 
					 options, (IO_READFUNC) fileRead, 
					 (IO_WRITEFUNC) fileWrite, fileSeek, 
					 fileFileSize, fileTell, 
					 &(cDoc->outIoDoc)));

  /* create section structure */
  DGP_CHECK_ALLOC (cDoc->currentSection = 
   DGP_MALLOC (sizeof (Dgp_CompressedSection)));

  cDoc->currentSection->dict = NULL;
  cDoc->currentSection->dict_size = 0;
  DGP_ERR_PASS_CLEANUP (DgpAdt_liNew (&(cDoc->currentSection->pages), dgp_state));

  *cDocument = cDoc;
  NORMAL_RETURN;

 ERR_CLEANUP:
  if (cDoc)
    {
      if (cDoc->currentSection) DGP_FREE (cDoc->currentSection);
      DGP_FREE (cDoc);
    }
  ERR_RETURN;

}

/* close the TIFF-FX file and free all remaining memory */
ErrCode
Dgp_closeCompressedDocument (Dgp_CompressedDocument *cDocument)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_closeCompressedDocument", DGP_PERRSTATE);

  /* close MRC file */
  DGP_ERR_PASS (IO_CloseDocument (cDocument->outIoDoc));
  DGP_ERR_PASS (IO_EndInstance(cDocument->ioInstanceID));
  fclose(cDocument->outfile.fp);

  /* should confirm that current section is empty */
  DGP_FREE (cDocument->currentSection);
  DGP_FREE (cDocument);

  NORMAL_RETURN;
}

/* write a section to file and free all memory associated with this section */
ErrCode
Dgp_writeSection (Dgp_CompressedDocument *cDocument, Dgp_Bool clear_dict)
{

  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_writeSection", DGP_PERRSTATE);

  IO_COLORSPACE colorSpace;
  IO_COMPRESSION compression;
  /* Xerox's private jbig2 code */
  Dgp_UInt16 jbig2_code = 34715;
  IO_COLORTYPE ioColorType;
  IO_PAGEHANDLE outIoPage;
  IO_IMGHANDLE outIoImg, newFGImg;

  Dgp_UInt32 dictID;
  Dgp_Bool dict_written = DGP_FALSE;

  /* IO stuff */
  Dgp_UInt32 bytes_written;


  /* ****NEW**** */
  Dgp_CompressedSection *sec = cDocument->currentSection;
  Dgp_CompressedPage *page;
  Dgp_CompressedImage *im;
  DgpAdt_ListNode ln_page, ln_jpeg;

  DGPADT_FOREACH (ln_page, sec->pages){
      /* write out the page */
      page = (Dgp_CompressedPage*) ln_page->userdata;
      DGP_ERR_PASS_CLEANUP(IO_AddPageStart(cDocument->outIoDoc,
				       -1 /* Append to the end */,
				       &outIoPage));
      
      if (page->posBlock){
		  /* set up for saving JBIG2 data */
		  colorSpace = IO_COLORSPACE_RGB;
		  compression = IO_COMP_JBIG2;
		  ioColorType = IO_COLORTYPE_BINARY;
		  
		  /* write position block and dictionary */
		  im = page->posBlock;
		  DGP_ERR_PASS_CLEANUP(IO_AddImage(	outIoPage,
											im->layer, ioColorType,
											im->x, im->y, im->xres,
											im->yres, im->width, im->height,
											im->height /*BestStripHeight*/, 
											NULL, 
											&outIoImg));
		  DGP_ERR_PASS_CLEANUP(IO_SetCompression (outIoImg, jbig2_code));
		  DGP_ERR_PASS_CLEANUP(IO_ImageWriteStart(outIoImg, 0, &compression,
							   &im->quality, &colorSpace));

		  if (!dict_written && sec->dict_size > 0){
			  DGP_ERR_PASS_CLEANUP(IO_StoreSymbolDictionary (cDocument->outIoDoc, 
				  sec->dict, 
				  sec->dict_size, sec->dict_size, &dictID));

			  /* store the dictID in the array */ 
			  cDocument->dict_list[cDocument->n_dict] = dictID;
			  DGP_ERR_GEN_CLEANUPF0 ((cDocument->n_dict >= DGP_MAXDICT - 1), 1, "Too many dictionaries ( > 256)" ); 
			  cDocument->n_dict++;
			  DGP_FREE (sec->dict);
			  sec->dict = NULL;
			  sec->dict_size = 0;
			  dict_written = DGP_TRUE;
		  }

		  /* store the position block for the first page on the list */
		  DGP_ERR_PASS_CLEANUP(IO_StorePositionBlock (outIoImg, im->data, 
			  im->data_size + im->extData_size, cDocument->dict_list, 
			  cDocument->n_dict));

		  /* save the token colors, if we have them */
		  if (im->extData_size > 0){
			  DGP_ERR_PASS_CLEANUP(IO_ShareColorTokenImage(outIoImg, 
				  im->data_size, IO_COLORTYPE_COLOR24, &newFGImg));
		  }
  
		  /* finish up token image */
		  DGP_FREE (im->data);
		  DGP_FREE (im);
		  DGP_ERR_PASS_CLEANUP(IO_ImageWriteFinish(outIoImg));
	  }

      /* write jpeg images */
      colorSpace = IO_COLORSPACE_YCBCR;
      compression = IO_COMP_JPEG;
      ioColorType = IO_COLORTYPE_COLOR24;

      DGPADT_FOREACH (ln_jpeg, page->jpegImages){
		  im = (Dgp_CompressedImage*) ln_jpeg->userdata;


		  DGP_ERR_PASS_CLEANUP(IO_AddImage(outIoPage, im->layer,
				     ioColorType,
				     im->x , im->y, im->xres, im->yres, 
				     im->width, im->height,
				     im->height /*BestStripHeight*/, NULL, 
				     &outIoImg));
		  DGP_ERR_PASS_CLEANUP(IO_ImageWriteStart(outIoImg, 0, &compression,
						    &im->quality, &colorSpace));

		  DGP_ERR_PASS_CLEANUP(IO_CompressedWriteStrip (outIoImg, 
				     0 /* row offset */,
				     im->height /* strip height */, im->data, 
				     im->data_size, &bytes_written));

		  DGP_ERR_PASS_CLEANUP(IO_ImageWriteFinish (outIoImg));
	  
		  DGP_FREE (im->data);
		  DGP_FREE (im);
	  }
      
      DgpAdt_liFree (page->jpegImages);
      page->jpegImages = DGP_NULLLIST;
      DGP_ERR_PASS_CLEANUP(IO_AddPageFinish(outIoPage));
      Dgp_freePage(cDocument, &page);
  }

  cDocument->n_pages += (Dgp_UInt16) sec->pages->nelt;
  DgpAdt_liClear (sec->pages);

  if (clear_dict) {
    /* Restart the dictionary list */
    cDocument->n_dict = 0;
  }

  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}
#endif
ErrCode
Dgp_addNewPage (Dgp_CompressedDocument *cDocument, Dgp_CompressedPage *cPage)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_createCompressedPage", DGP_PERRSTATE);

  DGP_ASSERT (cPage != NULL);
  DGP_ASSERT(cDocument != NULL);
  /* add page to current section */
  DGP_ERR_PASS (DgpAdt_liAdd(cDocument->currentSection->pages, 
			     (void*) cPage, NULL));
  NORMAL_RETURN;
}

/*create page with out add to the cDocument */
ErrCode
Dgp_createNewPage (Dgp_CompressedDocument *cDocument, Dgp_UInt32 width, 
		  Dgp_UInt32 height, Dgp_CompressedPage **cPage)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  Dgp_CompressedPage *cPageTemp;
  DGP_PROLOGUE ("Dgp_createCompressedPage", DGP_PERRSTATE);

  DGP_ASSERT (cPage != NULL);

  /* create structure */
  DGP_CHECK_ALLOC (cPageTemp = DGP_MALLOC (sizeof (Dgp_CompressedPage)));
  cPageTemp->dgp_state = dgp_state;
  cPageTemp->width = width;
  cPageTemp->height = height;
  DGP_ERR_PASS (DgpAdt_liNew (&(cPageTemp->jpegImages), dgp_state));
  DGP_ERR_PASS (DgpAdt_liNew (&(cPageTemp->fgImages), dgp_state));
  /* predictable state */
  cPageTemp->posBlock = NULL;

  /* return */
  *cPage = cPageTemp;
  NORMAL_RETURN;
}


ErrCode
Dgp_freePage (Dgp_CompressedDocument *cDocument, Dgp_CompressedPage **cPage)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_createCompressedPage", DGP_PERRSTATE);
  DgpAdt_ListNode ln_jpeg;
  Dgp_CompressedImage* im = NULL;

  DGP_STOP_WHINE;

  DGP_ASSERT (cPage != NULL);
  if(*cPage){
      if((*cPage)->jpegImages != DGP_NULLLIST){
          DGPADT_FOREACH (ln_jpeg, (*cPage)->jpegImages){
	          im = (Dgp_CompressedImage*) ln_jpeg->userdata;
              if(im){
                  if(im->data)  DGP_FREE (im->data);
		          DGP_FREE (im);
              }
          }
            DgpAdt_liFree ((*cPage)->jpegImages);
      }
      if((*cPage)->fgImages != DGP_NULLLIST){
          DGPADT_FOREACH (ln_jpeg, (*cPage)->fgImages){
	          im = (Dgp_CompressedImage*) ln_jpeg->userdata;
              if(im){
                  if(im->data)  DGP_FREE (im->data);
		          DGP_FREE (im);
              }
          }
          DgpAdt_liFree ((*cPage)->fgImages);
      }
      DGP_FREE(*cPage);
  }
  /* return */
  *cPage = NULL;
  NORMAL_RETURN;
}
  
ErrCode
Dgp_startNewPage (Dgp_CompressedDocument *cDocument, Dgp_UInt32 width,
                  Dgp_UInt32 height, Dgp_CompressedPage **cPage)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  Dgp_CompressedPage *cPageTemp;
  DGP_PROLOGUE ("Dgp_createCompressedPage", DGP_PERRSTATE);

  DGP_ASSERT (cPage != NULL);

  /* create structure */
  DGP_CHECK_ALLOC (cPageTemp = DGP_MALLOC (sizeof (Dgp_CompressedPage)));
  cPageTemp->dgp_state = dgp_state;
  cPageTemp->width = width;
  cPageTemp->height = height;
  DGP_ERR_PASS (DgpAdt_liNew (&(cPageTemp->jpegImages), dgp_state));

  /* predictable state */
  cPageTemp->posBlock = NULL;

  /* add page to current section */
  DGP_ERR_PASS (DgpAdt_liAdd(cDocument->currentSection->pages,
                             (void*) cPageTemp, NULL));

  /* return */
  *cPage = cPageTemp;
  NORMAL_RETURN;
}
ErrCode
Dgp_storeCompPosBlock (Dgp_CompressedPage *cPage, 
		       Dgp_CompressedImage *posBlock)
{
  Dgp_State *dgp_state = cPage->dgp_state;
  DGP_PROLOGUE ("Dgp_storeCompPosBlock", DGP_PERRSTATE);

  DGP_STOP_WHINE;

  DGP_ASSERT (posBlock != NULL);
  
  cPage->posBlock = posBlock;

  NORMAL_RETURN;
}

ErrCode
Dgp_storeJpegImage (Dgp_CompressedPage *cPage, 
		    Dgp_CompressedImage *jpeg)
{
  Dgp_State *dgp_state = cPage->dgp_state;
  DGP_PROLOGUE ("Dgp_storeJpegImage", DGP_PERRSTATE);

  if(!jpeg) NORMAL_RETURN;

  DGP_ERR_PASS (DgpAdt_liAdd (cPage->jpegImages, (void*) jpeg, NULL));

  NORMAL_RETURN;
}

ErrCode
Dgp_storeFgImage (
Dgp_CompressedPage *cPage, 
DgpAdt_List fgImages)
{
    Dgp_State *dgp_state = cPage->dgp_state;
    DgpAdt_ListNode ln_jpeg;
    Dgp_CompressedImage* im = NULL;    

    DGP_PROLOGUE ("Dgp_storeFgImage", DGP_PERRSTATE);

    if(!fgImages) NORMAL_RETURN;

    DGPADT_FOREACH (ln_jpeg, fgImages){
        im = (Dgp_CompressedImage*) ln_jpeg->userdata;
        if(im){
            DGP_ERR_PASS (DgpAdt_liAdd (cPage->fgImages, (void*) im, NULL));
        }
    }
  NORMAL_RETURN;
}



ErrCode
Dgp_storeCompTokDict (Dgp_CompressedDocument *cDocument, 
		      Dgp_UInt8 *dict, Dgp_UInt32 dict_size)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_storeCompTokDict", DGP_PERRSTATE);

  DGP_STOP_WHINE;

/*  DGP_ASSERT (dict != NULL); */
/*  DGP_ASSERT (cDocument->currentSection->dict == NULL); */

  cDocument->currentSection->dict_size = dict_size;
  if (dict_size > 0)
  {
    cDocument->currentSection->dict = dict;
  }

  NORMAL_RETURN;
}

ErrCode
Dgp_getPageOfSection (Dgp_CompressedDocument *cDocument, Dgp_UInt32 pagenum, 
		      Dgp_CompressedPage **cPage)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_getFirstPageOfSection", DGP_PERRSTATE);
  
  DgpAdt_List li = cDocument->currentSection->pages;
  DgpAdt_ListNode ln;
  Dgp_UInt32 i = 1;

  DGP_STOP_WHINE;

  DGP_ASSERT ((Dgp_UInt32)li->nelt >= pagenum);

  ln = li->first;
  while (i < pagenum) 
    {
      ln = ln->next;
      i++;
    }
  *cPage = (Dgp_CompressedPage*) ln->userdata;

  NORMAL_RETURN;
}

