/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/hufftable.c,v 1.20 2003/02/10 22:53:26 mkrocken Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "hufftable.h" 
#include "jbitfile.h" 
#include <stdlib.h> 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/hufftable.c,v 1.20 2003/02/10 22:53:26 mkrocken Exp $"); 
 
/* 
 * Routines for dealing with Huffman tables 
 */ 
 
/* 
 * Given the table specification, where none of the actual prefix codes 
 * have been assigned, figure out what the codes should be and store them 
 * into the specification.  The lengths must be valid, of course.  If 
 * something's wrong, like there not being enough code space, it returns 
 * DGP_FALSE. 
 */ 
ErrCode 
jhtSetCodes(hufftable_spec *spec, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtSetCodes()", DGP_PERRSTATE); 
 
    Dgp_Int32 i; 
    Dgp_Int32 counts[JHT_MAXCODELEN]; 
    Dgp_UInt32 firstcode[JHT_MAXCODELEN]; 
 
     
    /* Make a histogram of the code lengths */ 
    for (i = 0; i < JHT_MAXCODELEN; i++) { 
	counts[i] = 0; 
	} 
 
    for (i = 0; i < spec->nlines; i++) { 
	counts[(Dgp_UInt8)(spec->lines[i].preflen)]++; 
	} 
 
    /* Assign the base values */ 
    spec->firstcode[0] = firstcode[0] = 0; 
    counts[0] = 0; 
    for (i = 1; i < JHT_MAXCODELEN; i++) { 
	spec->firstcode[i] = firstcode[i] = 
	    2 * (firstcode[i - 1] + counts[i - 1]); 
 
	/* Bad - there's a code of this length, and no room for it */ 
	DGP_ERR_GEN(((firstcode[i] >= (1ul << i)) && (counts[i] > 0)), 
		    cErrInvalidParm); 
 
	spec->lastcode[i] = firstcode[i] + counts[i] - 1; 
	} 
 
    /* Assign the codes themselves */ 
    for (i = 0; i < spec->nlines; i++) { 
	if (spec->lines[i].preflen > 0) { 
	    /* Assign it the next code of the given length */ 
	    spec->lines[i].prefix = 
		firstcode[(Dgp_UInt8)(spec->lines[i].preflen)]++; 
	    } 
	else { 
	    /* Zero-length means don't assign it any code space */ 
	    spec->lines[i].prefix = 0ul; 
	    } 
	} 
 
    /* Find the Dgp_Int16est code length */ 
    spec->minlen = JHT_MAXCODELEN - 1; 
    for (i = 1; i < JHT_MAXCODELEN; i++) { 
	if (counts[i] > 0) { 
	    spec->minlen = i; 
	    break; 
	    } 
	} 
 
    spec->compiled = DGP_FALSE; 
    spec->decoded[0] = (void **)NULL; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Given the table specification, which must have had jhtSetCodes 
 * called on it, build some tables that make decoding faster. 
 * Call jhtCompile on a table, providing one void * per line of the table. 
 * After that, call jhtReadCode on the table; each call will return one 
 * of the void *s. 
 * 
 * The array "lines" is not needed after this routine returns. 
 */ 
ErrCode 
jhtCompile(hufftable_spec *spec, void **lines, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtCompile()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i; 
    void **newdec = (void **)NULL; 
    Dgp_Int32 count; 
    Dgp_Int32 indexes[JHT_MAXCODELEN]; 
 
    /* If it was already compiled, nothing to do. */ 
    if (spec->compiled) { 
	NORMAL_RETURN; 
	} 
 
    /* Allocate the (single) array of decoded values */ 
    newdec = (void **)(a->malloc)(spec->nlines * sizeof(*newdec)); 
    DGP_CHECK_ALLOC_CLEANUP(newdec); 
 
    /* Set the initial indexes into that array */ 
    indexes[0] = 0; 
    spec->decoded[0] = newdec; 
    for (i = 1; i < JHT_MAXCODELEN; i++) { 
	/* How many codes of the previous length? */ 
	count = spec->firstcode[i] / 2 - spec->firstcode[i - 1]; 
 
	/* Set the initial index pointer */ 
	indexes[i] = indexes[i - 1] + count; 
 
	/* and the pointer into the decoded values array */ 
	spec->decoded[i] = newdec + indexes[i]; 
	} 
 
    /* Now just copy the pointers in */ 
    for (i = 0; i < spec->nlines; i++) { 
	int preflen = spec->lines[i].preflen; 
 
	/* 
	 * For a table line, we can look at its code length and code 
	 * and immediately find its slot in decoded: decoded[preflen] points 
	 * to the first slot for codes of that length, so just by 
	 * subtracting the first code of that length from the actual code, 
	 * we get the address of the right slot. 
	 */ 
	if (preflen > 0) { 
	    spec->decoded[preflen][spec->lines[i].prefix - 
				  spec->firstcode[preflen]] = lines[i]; 
	    } 
	} 
 
    spec->compiled = DGP_TRUE; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newdec != (void **)NULL) { 
	(a->free)((void *)newdec); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Read bits until it matches one of the line prefixes; return 
 * the void * associated with that line by the call to jhtCompile. 
 * If there's a problem with the jbitfile, set *end to DGP_TRUE. 
 * If you get an illegal code, return NULL. 
 * 
 * We can speed up the processing: once we know the code length, then 
 * we can quickly zoom to the right table line.  Two approaches for 
 * doing this: 
 * - read one bit at a time, comparing the code-so-far with the 
 *   valid code range for codes of the current length; finish when 
 *   they match. 
 * - build the code by peeking at the next 8 bits of the code stream, 
 *   and doing a quick binary search to see if the current code, plus 
 *   1-8 of the peeked bits, fits the range for current length+(1-8). 
 *   If you can't do the peeking, fall back to the read-one-bit method. 
 * 
 * In either case, we can start by reading minlen bits. 
 * 
 * This code strongly relies on the fact that if you take all the codes 
 * of length 1-l, shift them all left so that they're all l bits Dgp_Int32, 
 * and sort the results, the codes end up sorted by length: all the 1-bit 
 * codes, then all the 2-bit codes, and so on.  Also, all the codes 
 * Dgp_Int32er than l, if shifted right, would all be larger than the last l-bit 
 * code.  This means that we can check any code against the lastcode entry 
 * for a Dgp_Int32er or Dgp_Int16er code (by right-shifting or left-shifting 
 * appropriately), and get an indication that the code's true length 
 * is <= or > the length we're comparing it against. 
 */ 
ErrCode 
jhtReadCode(const hufftable_spec *spec, jbitfile f, Dgp_State *dgp_state, 
	    void **_code) 
{ 
  DGP_PROLOGUE("jhtReadCode()", DGP_PERRSTATE); 
 
    Dgp_UInt32 code, tmpcode; 
    Dgp_Int32 newcode; 
    Dgp_Int32 codelen; 
 
    /* 
     * Read bits one at a time, and for each bit check to see if the 
     * code-so-far falls into the correct range for codes of that length. 
     * Start by reading minlen bits. 
     */ 
    DGP_ERR_PASS_CLEANUP(jbfReadBits(f, spec->minlen, &code)); 
 
    /* firstcode[minlen] must be zero, so code >= firstcode[minlen] */ 
    for (codelen = spec->minlen; codelen < JHT_MAXCODELEN - 1;) { 
	/* 
	 * At this point, we know that code >= firstcode[codelen].  If 
	 * code > lastcode[codelen], then this can't be a valid code 
	 * for codelen. 
	 * Note that firstcode[codelen+1] = (lastcode[codelen] + 1) << 1. 
	 */ 
	if (code <= spec->lastcode[codelen]) { 
	    /* 
	     * It's not valid for the next code length - must be OK for this 
	     * length. 
	     */ 
/*	    DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
	    *_code = (spec->decoded[codelen][code - spec->firstcode[codelen]]); 
	    NORMAL_RETURN; 
	    } 
 
	/* 
	 * At this point, we know that (code << 1) >= firstcode[codelen+1], 
	 * so its length is > codelen. 
	 */ 
	newcode = jbfPeek(f); 
	if ((newcode < 0) || (codelen + DGP_BITS_UINT8 >= DGP_BITS_UINT32)) { 
	    /* Couldn't peek - read the one bit we know we need */ 
	    DGP_ERR_PASS_CLEANUP(jbfReadBits(f, 1, &tmpcode)); 
	    code = (code << 1) | tmpcode; 
 
	    codelen++; 
	    continue; 
	    } 
 
	/* Tack on the 8 bits we just peeked */ 
	code = (code << DGP_BITS_UINT8) | newcode; 
	if (code > spec->lastcode[codelen + DGP_BITS_UINT8]) { 
	    /* It's definitely more than codelen+(1-8) bits Dgp_Int32 */ 
	    (void)jbfReadBits(f, DGP_BITS_UINT8, &tmpcode); 
	    codelen += DGP_BITS_UINT8; 
	    continue; 
	    } 
 
	if ((code >> 4) <= spec->lastcode[codelen + 4]) { 
	    /* It's codelen+(1-4) bits */ 
	    if ((code >> 6) <= spec->lastcode[codelen + 2]) { 
		/* It's codelen+(1-2) bits */ 
		if ((code >> 7) <= spec->lastcode[codelen + 1]) { 
		    /* It's codelen+1 bits */ 
		    (void)jbfReadBits(f, 1, &tmpcode); 
		    code >>= 7; 
		    codelen = codelen + 1; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code - 
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		else { 
		    /* It's codelen+2 bits */ 
		    (void)jbfReadBits(f, 2, &tmpcode); 
		    code >>= 6; 
		    codelen = codelen + 2; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code -  
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		} 
	    else { 
		/* It's codelen+(3-4) bits */ 
		if ((code >> 5) <= spec->lastcode[codelen + 3]) { 
		    /* It's codelen+3 bits */ 
		    (void)jbfReadBits(f, 3, &tmpcode); 
		    code >>= 5; 
		    codelen = codelen + 3; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code -  
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		else { 
		    /* It's codelen+4 bits */ 
		    (void)jbfReadBits(f, 4, &tmpcode); 
		    code >>= 4; 
		    codelen = codelen + 4; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code -  
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		} 
	    } 
	else { 
	    /* It's codelen+(5-8) bits */ 
	    if ((code >> 2) <= spec->lastcode[codelen + 6]) { 
		/* It's codelen+(5-6) bits */ 
		if ((code >> 3) <= spec->lastcode[codelen + 5]) { 
		    /* It's codelen+5 bits */ 
		    (void)jbfReadBits(f, 5, &tmpcode); 
		    code >>= 3; 
		    codelen = codelen + 5; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code - 
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		else { 
		    /* It's codelen+6 bits */ 
		    (void)jbfReadBits(f, 6, &tmpcode); 
		    code >>= 2; 
		    codelen = codelen + 6; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code - 
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		} 
	    else { 
		/* It's codelen+(7-8) bits */ 
		if ((code >> 1) <= spec->lastcode[codelen + 7]) { 
		    /* It's codelen+7 bits */ 
		    (void)jbfReadBits(f, 7, &tmpcode); 
		    code >>= 1; 
		    codelen = codelen + 7; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code - 
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		else { 
		    /* It's codelen+8 bits */ 
		    (void)jbfReadBits(f, 8, &tmpcode); 
		    codelen = codelen + 8; 
		    /*DGP_REPORTF2("jht %lx(%d)\n", code, codelen);  */
		    *_code = (spec->decoded[codelen][code - 
						    spec->firstcode[codelen]]); 
		    NORMAL_RETURN; 
		    } 
		} 
	    } 
	} 
 
    /* Theoretically, this could be a real 31-bit code. FIX. */ 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
void 
jhtCleanup(hufftable_spec *spec, Dgp_State *dgp_state) 
{ 
    if (spec->compiled) { 
	/* 
	 * Free the compiled decoding tables.  decoded[0] was set up 
	 * to be the base pointer. 
	 */ 
	(DGP_FREE)((void *)spec->decoded[0]); 
	spec->decoded[0] = (void **)NULL; 
	spec->compiled = DGP_FALSE; 
	} 
    } 
 
void 
jhtFree(hufftable_spec *spec, Dgp_State *dgp_state) 
{ 
  Dgp_Allocators *a = &(dgp_state->a); 
 
    jhtCleanup(spec, dgp_state); 
    (a->free)((void *)spec->lines); 
    (a->free)((void *)spec); 
    } 
