/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/hufftable_util.c,v 1.14 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
#include "list.h" 
#include <stdlib.h> 
#include "hufftable.h" 
#include "encint.h" 
#include "hufftable_util.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/hufftable_util.c,v 1.14 2003/01/27 21:48:34 awilliam Exp $"); 
 
/* 
 * A useful constant: about how many bits will a Huffman table segment 
 * take up in the bitstream? 
 * The segment header will be 11 bytes; the body will be 9 bytes plus 
 * about one byte per table line.  That's 160 bits overhead, 8 bits per 
 * line. 
 */ 
#define	JHTSEGOVERHEAD	160 
#define	JHTSEGLINEBITS	8 
 
typedef struct htree { 
    struct htree *zero, *one; 
    Dgp_Int32 totval; 
    Dgp_Int32 index; 
    Dgp_Int32 codelen; 
    } htree; 
 
static void assignlengths(htree *h); 
 
/* 
 * A use count we're not likely to see often.  This limits the size 
 * of the bucket sort; things with counts greater than this get stuffed 
 * into a single bucket. 
 */ 
#define	NBUCKETS	1000 
 
/* 
 * Given the distribution of occurrence counts in counts, set the array 
 * lengths to the code lengths of a Huffman table tuned for counts. 
 * 
 * The algorithm used is: 
 * Scan the current array for the two lowest-value elements. In cases 
 * of ties, the element closest to the start of the array is used. 
 * Replace the first (closest to the start of those two) element with the 
 * node representing the merged elements, and delete the second. 
 * Continue until there's only one thing in the array. 
 * 
 * This process gives a tree; the code lengths are assigned based on the depth 
 * in the tree (i.e. how many merges it took to go from the original element 
 * to the single remaining thing in the array). 
 * 
 * The tie-breaking algorithm is still pertinent, since it can affect which 
 * tokens get which which code lengths. 
 * 
 * The actual algorithm used is: 
 * Sort into buckets using radix sort 
 * Concatenate the buckets into a linked (sorted) list 
 * While the list contains at least two elements: 
 * - Merge the first two entries on the list 
 * - Insert the merged entry in the sorted list, starting the search for 
 *   the insertion point at the last insertion point. 
 * This is O(n). 
 * 
 * If counts[i]=0, then lengths[i] will be set to zero.  No other 
 * zero-length codes will be assigned. 
 * 
 * Return DGP_TRUE on success, DGP_FALSE on failure. 
 */ 
ErrCode 
jhtGetCodeLengths(Dgp_Int32 *counts, Dgp_Int32 *lengths, Dgp_Int32 ncounts, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtGetCodeLengths()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    htree *array = (htree *)NULL; 
    htree *small, *small2; 
    Dgp_Int32 i; 
    Dgp_Int32 nzcounts; 
    Dgp_Int32 arraylen; 
    DgpAdt_List buckets[NBUCKETS]; 
    DgpAdt_List sorted = DGP_NULLLIST; 
    DgpAdt_ListNode insertion, last, smallnode; 
    htree *node; 
 
    if (ncounts == 0) { 
	/* We successfully did nothing. */ 
      NORMAL_RETURN; 
	} 
 
    for (i = 0; i < NBUCKETS; i++) { 
	buckets[i] = DGP_NULLLIST; 
	} 
 
    array = (htree *)(a->malloc)(ncounts * 2 * sizeof(*array)); 
    DGP_CHECK_ALLOC_CLEANUP(array); 
 
    /* Read out the input into our array of buckets. */ 
    for (arraylen = 0, i = 0; i < ncounts; i++) { 
	/* Ignore zero-count entries */ 
	if (counts[i] > 0) { 
		DgpAdt_ListNode retnode; 
	    array[arraylen].zero = array[arraylen].one = (htree *)NULL; 
	    array[arraylen].totval = counts[i]; 
	    array[arraylen].index = i; 
	    array[arraylen].codelen = 0; 
 
	    if (array[arraylen].totval < NBUCKETS - 1) { 
		if (buckets[array[arraylen].totval] == DGP_NULLLIST) { 
		    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&buckets[array[arraylen].totval], dgp_state)); 
		    } 
 
		DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(buckets[array[arraylen].totval], 
				      (void *)&array[arraylen], &retnode)); 
		} 
	    else { 
		/* 
		 * The last bucket is a special case since I don't feel like 
		 * implementing a full radix sort. 
		 */ 
		if (buckets[NBUCKETS - 1] == DGP_NULLLIST) { 
		    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&buckets[NBUCKETS - 1], dgp_state)); 
		    } 
	     
		/* Insert it just before the first item larger than it */ 
		last = DGP_NULLLISTNODE; 
		DGPADT_FOREACH (insertion, buckets[NBUCKETS - 1]) { 
		    node = (htree *)DGPADT_LIGET(insertion); 
		    if (node->totval > array[arraylen].totval) { 
			/* We've found the right place */ 
			break; 
			} 
		    last = insertion; 
		    } 
 
		DGP_ERR_PASS_CLEANUP(DgpAdt_liAddAfter(buckets[NBUCKETS - 1], last, 
						   (void *)&array[arraylen], &retnode)); 
		} 
 
	    arraylen++; 
	    } 
	else { 
	    /* We already know what the answer will be */ 
	    lengths[i] = 0; 
	    } 
	} 
 
    /* How many non-zero counts? */ 
    nzcounts = arraylen; 
 
    /* Special case: all buckets were empty */ 
    if (nzcounts == 0) { 
	/* 
	 * Nothing to do - just report lengths of 0 for everything, 
	 * and those have already been filled in. 
	 */ 
	(a->free)((void *)array); 
	NORMAL_RETURN; 
	} 
 
    /* Now concatenate the buckets */ 
    for (i = 0; i < NBUCKETS; i++) { 
	if (buckets[i] != DGP_NULLLIST) { 
	    if (sorted == DGP_NULLLIST) { 
		/* First non-empty bucket */ 
		sorted = buckets[i]; 
		buckets[i] = DGP_NULLLIST; 
		} 
	    else { 
		/* Append to sorted */ 
		DgpAdt_liApp(sorted, buckets[i]); 
		buckets[i] = DGP_NULLLIST; 
		} 
	    } 
	} 
 
    /* Here's our initial insertion point */ 
    insertion = DGP_NULLLISTNODE; 
    while (DGPADT_LILEN(sorted) > 1) { 
	/* Get the smallest entry */ 
	smallnode = DGPADT_LIFIRST(sorted); 
	DGP_ASSERT(smallnode != DGP_NULLLISTNODE); 
	if (smallnode == insertion) { 
	    /* 
	     * If this was our reference node, it's not going to be valid 
	     * anymore - just set our insertion search back to the start 
	     * of the list, since it was nearly there anyway. 
	     */ 
	    insertion = DGP_NULLLISTNODE; 
	    } 
	small = (htree *)DGPADT_LIGET(smallnode); 
	DGP_ASSERT(small != (htree *)NULL); 
	DgpAdt_liRem(smallnode); 
 
	/* and the second smallest */ 
	smallnode = DGPADT_LIFIRST(sorted); 
	DGP_ASSERT(smallnode != DGP_NULLLISTNODE); 
	if (smallnode == insertion) { 
	    insertion = DGP_NULLLISTNODE; 
	    } 
	small2 = (htree *)DGPADT_LIGET(smallnode); 
	DGP_ASSERT(small2 != (htree *)NULL); 
	DgpAdt_liRem(smallnode); 
 
	/* Create a merged node */ 
	array[arraylen].zero = small; 
	array[arraylen].one = small2; 
	/* The index in our theoretical array is used as a secondary sort key */ 
	array[arraylen].index = MIN(small->index, small2->index); 
	array[arraylen].totval = small->totval + small2->totval; 
	array[arraylen].codelen = 0; 
 
	/* 
	 * Now find where to insert the merged node. 
	 * The tricky part is that we start our search at the last place 
	 * that we inserted anything, since this new merged node must have 
	 * totval at least as big as that. 
	 * 
	 * insertion might be equal to DGP_NULLLISTNODE, in which case we start 
	 * searching from the start of the list. 
	 */ 
	last = insertion; 
	if (insertion == DGP_NULLLISTNODE) { 
	    insertion = DGPADT_LIFIRST(sorted); 
	    } 
 
	for (; 
	     insertion != DGP_NULLLISTNODE; 
	     last = insertion, insertion = DGPADT_LINEXT(insertion)) { 
	    node = (htree *)DGPADT_LIGET(insertion); 
 
	    if (node->totval >= array[arraylen].totval) { 
		if (node->totval > array[arraylen].totval) { 
		    /* Here's where to put it */ 
		    break; 
		    } 
		else if (node->index > array[arraylen].index) { 
		    break; 
		    } 
		} 
	    } 
 
	/* Insert just before insertion */ 
	DGP_ERR_PASS_CLEANUP (DgpAdt_liAddAfter(sorted, last, (void *)&array[arraylen], &insertion)); 
 
	arraylen++; 
	} 
 
    /* 
     * Done. 
     * The only entry in sorted now refers to the single entry in the queue, 
     * the root. 
     */ 
    small = DGPADT_LIGET(DGPADT_LIFIRST(sorted)); 
 
    /* We'd better have done nzcounts - 1 merges */ 
    DGP_ASSERT(arraylen == nzcounts * 2 - 1); 
	     
    /* Assign the code lengths. */ 
    assignlengths(small); 
 
    /* and read them out into lengths */ 
    for (i = 0; i < nzcounts; i++) { 
	/* The thing at array[i] came from array[i].index */ 
	lengths[array[i].index] = array[i].codelen; 
 
	/* 
	 * Special case: there was only one valid code.  In this case, 
	 * we theoretically could give it a code length of zero, but 
	 * other bits of the code get unhappy about that, since codelen==0 
	 * indicates "unused".  So we give it a code length of 1. 
	 */ 
	if (lengths[array[i].index] == 0) { 
	    lengths[array[i].index] = 1; 
	    } 
	} 
 
    (a->free)((void *)array); 
    DgpAdt_liFree(sorted); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (array != (htree *)NULL) { 
	(a->free)((void *)array); 
	} 
 
    for (i = 0; i < NBUCKETS; i++) { 
	if (buckets[i] != DGP_NULLLIST) { 
	    DgpAdt_liFree(buckets[i]); 
	    } 
	} 
 
    if (sorted != DGP_NULLLIST) { 
	DgpAdt_liFree(sorted); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Given the code length in h, assign code lengths to all h's descendents. 
 */ 
static void 
assignlengths(htree *h) 
{ 
  DGP_VOID_PROLOGUE("assignlengths()", NULL); 
 
    DGP_VOID_ASSERT(h != (htree *)NULL); 
 
    if (h->zero != (htree *)NULL) { 
	h->zero->codelen = h->codelen + 1; 
	assignlengths(h->zero); 
	} 
 
    if (h->one != (htree *)NULL) { 
	h->one->codelen = h->codelen + 1; 
	assignlengths(h->one); 
	} 
    } 
 
/* 
 * Given the Huffman table described in spec, and the counts in *counts, 
 * re-assign the prefix code lengths and codes. 
 * 
 * If lengths is not NULL, then it is assumed to hold correct lengths 
 * for the given counts, and the counts are ignored (may be NULL). 
 * If lengths is not NULL, the counts are used to generate lengths. 
 * 
 * This is a destructive operation: it overwrites information in spec. 
 * Do *NOT* call this on any of the standard Huffman tables, as that will 
 * cause memory faults; instead, copy those tables and call this on the 
 * copy.  See htTuneStandardTable. 
 * 
 * If this returns DGP_FALSE, the table may have already been damaged. 
 * 
 * If the table was compiled before, it has to be recompiled. 
 */ 
ErrCode 
jhtTuneTable(hufftable_spec *spec, Dgp_Int32 *counts, Dgp_Int32 *lengths,  
	     Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtTuneTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 *mylengths = (Dgp_Int32 *)NULL; 
    Dgp_Int32 i; 
 
    /* Get the new prefix lengths */ 
    if (lengths == (Dgp_Int32 *)NULL) { 
	mylengths = (Dgp_Int32 *)(a->malloc)(spec->nlines * sizeof(*mylengths)); 
	DGP_CHECK_ALLOC_CLEANUP(mylengths); 
 
	DGP_ERR_PASS_CLEANUP(jhtGetCodeLengths(counts, mylengths, 
					       spec->nlines, dgp_state)); 
	} 
    else { 
	mylengths = lengths; 
	} 
 
    /* Install them into the table */ 
 
    /* Get rid of any compiled information */ 
    jhtCleanup(spec, dgp_state); 
 
    /* Install the new lengths */ 
    for (i = 0; i < spec->nlines; i++) { 
		spec->lines[i].preflen = (Dgp_Int8)mylengths[i]; 
	} 
 
    /* Figure out the new codes */ 
    DGP_ERR_PASS_CLEANUP(jhtSetCodes(spec, dgp_state)); 
 
    if (mylengths != lengths) { 
	(a->free)((void *)mylengths); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if ((mylengths != (Dgp_Int32 *)NULL) && (mylengths != lengths)) { 
	(a->free)((void *)mylengths); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * This re-tunes the given numeric table.  It calls jhtTuneTable and 
 * then recompiles the table in a way that decint will like. 
 */ 
ErrCode 
jhtTuneNumericTable(hufftable_spec *spec, Dgp_Int32 *counts, Dgp_Int32 *lengths, 
		    Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtTuneNumericTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i; 
    void **lineptrs = (void **)NULL; 
 
    DGP_ERR_PASS_CLEANUP(jhtTuneTable(spec, counts, lengths, dgp_state)); 
     
    /* Make up the line pointers array: pointers back to the lines themselves */ 
    lineptrs = (void **)(a->malloc)(spec->nlines * sizeof(*lineptrs)); 
    DGP_CHECK_ALLOC_CLEANUP(lineptrs); 
 
    for (i = 0; i < spec->nlines; i++) { 
	lineptrs[i] = (void *)&spec->lines[i]; 
	} 
     
    DGP_ERR_PASS_CLEANUP(jhtCompile(spec, lineptrs, dgp_state)); 
 
    /* free lineptrs - no Dgp_Int32er needed */ 
    (a->free)((void *)lineptrs); 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (lineptrs != (void **)NULL) { 
	(a->free)((void *)lineptrs); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * This re-tunes the given standard table.  It copies the table and 
 * calls jhtTuneNumericTable.  It returns the new table spec on success, 
 * NULL on failure. 
 */ 
ErrCode 
jhtTuneStandardTable(hufftable_spec *spec, Dgp_Int32 *counts, Dgp_Int32 *lengths, 
		     Dgp_State *dgp_state, hufftable_spec **_newspec) 
{ 
  DGP_PROLOGUE("jhtTuneStandardTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    hufftable_spec *newspec = (hufftable_spec *)NULL; 
    table_line *newlines = (table_line *)NULL; 
    Dgp_Int32 i; 
 
    *_newspec = (hufftable_spec *)NULL; 
 
    /* Create a new table */ 
    newspec = (hufftable_spec *)(a->malloc)(sizeof(*newspec)); 
    DGP_CHECK_ALLOC_CLEANUP(newspec); 
 
    /* Copy over some information */ 
    *newspec = *spec; 
 
    /* Can't use the compiled information */ 
    newspec->compiled = DGP_FALSE; 
    newspec->decoded[0] = (void **)NULL; 
 
    /* Create a new line array */ 
    newlines = (table_line *)(a->malloc)(spec->nlines * sizeof(*newlines)); 
    DGP_CHECK_ALLOC_CLEANUP(newlines); 
 
    /* and copy the old information into it */ 
    for (i = 0; i < spec->nlines; i++) { 
	newlines[i] = spec->lines[i]; 
	} 
    newspec->lines = newlines; 
 
    /* Now we can re-tune the table */ 
    DGP_ERR_PASS_CLEANUP(jhtTuneNumericTable(newspec, counts, lengths, 
					     dgp_state)); 
 
    /* all done */ 
    *_newspec = newspec; 
    newspec = (hufftable_spec *)NULL; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (newspec != (hufftable_spec *)NULL) { 
	(a->free)((void *)newspec); 
	} 
 
    if (newlines != (table_line *)NULL) { 
	(a->free)((void *)newlines); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Given the Huffman table spec and various counts, count up how many 
 * bits are used, and how many bits would be used by a tuned version of 
 * that table.  A side effect of this routine is that the array lengths is 
 * filled in with the prefix lengths of that tuned table.  *normalbits 
 * and *tunedbits are the bit counts of the original and tuned tables. 
 */ 
ErrCode 
jhtCountBits(hufftable_spec *spec, Dgp_Int32 *counts, Dgp_Int32 *lengths, 
	    Dgp_Int32 *normalbits, Dgp_Int32 *tunedbits, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtCountBits()", DGP_PERRSTATE); 
    Dgp_Int32 i; 
    Dgp_Int32 nb, tb; 
 
    nb = tb = 0; 
 
    /* Compute the tuned bit lengths */ 
    DGP_ERR_PASS_CLEANUP(jhtGetCodeLengths(counts, lengths, spec->nlines, 
					   dgp_state)); 
 
    for (i = 0; i < spec->nlines; i++) { 
	nb += counts[i] * (spec->lines[i].preflen + spec->lines[i].rangelen); 
	tb += counts[i] * (lengths[i] + spec->lines[i].rangelen); 
	} 
	 
    *normalbits = nb; 
    *tunedbits = tb; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * For each table in tables, set lengths[i] and counts[i] to 
 * newly-allocated, cleared, arrays of ints the same size as tables[i]; 
 * set eis[i] to a newly-allocated eiEncoder using tables[i]. 
 * 
 * lengths[i] and counts[i] are assumed to be initialised to 
 * (int *)NULL; eis[i] is assumed to be NULLEIENCODER.  If this routine 
 * fails, then those values are still present. 
 * 
 * The caller is responsible for freeing everything allocated. 
 */ 
ErrCode 
jhtAllocTableInfo(hufftable_spec **tables, Dgp_Int32 **lengths, Dgp_Int32 **counts, 
		 eiEncoder *eis, Dgp_Int32 ntables, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtAllocTableInfo()", DGP_PERRSTATE); 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i; 
 
    /* Allocate the count and length arrays, plus the eiEncoders */ 
    for (i = 0; i < ntables; i++) { 
	lengths[i] = (Dgp_Int32 *)(a->malloc)(tables[i]->nlines * sizeof(*lengths[i])); 
	DGP_CHECK_ALLOC_CLEANUP(lengths[i]); 
	 
	counts[i] = (Dgp_Int32 *)(a->calloc)(tables[i]->nlines, sizeof(*counts[i])); 
	DGP_CHECK_ALLOC_CLEANUP(counts[i]); 
 
	DGP_ERR_PASS_CLEANUP(eiNewEncoder(tables[i], NULLJBITFILE, dgp_state, 
					  &(eis[i]))); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    for (i = 0; i < ntables; i++) { 
	if (lengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)lengths[i]); 
	    lengths[i] = (Dgp_Int32 *)NULL; 
	    } 
	if (counts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)counts[i]); 
	    counts[i] = (Dgp_Int32 *)NULL; 
	    } 
	if (eis[i] != NULLEIENCODER) { 
	    eiFreeEncoder(eis[i]); 
	    eis[i] = NULLEIENCODER; 
	    } 
	} 
    ERR_RETURN; 
    } 
 
/* 
 * Given the table specs in tables[i], accumulated counts in counts[i], 
 * lengths[i] (tuned table lengths for table i, according to 
 * counts[i]), normal[i] and tuned[i] (bit counts for the normal and tuned 
 * table), decide which table is best, whether that best 
 * table is a tuned version of one of the standard tables (in *bestnew), 
 * and the actual best table (either one of tables[i] or a newly 
 * allocated table) in *bestTable. 
 * 
 * If custom_allowed is DGP_FALSE, this will not consider tuned tables. 
 */ 
ErrCode 
jhtDecideTable(hufftable_spec **tables, Dgp_Int32 **counts, Dgp_Int32 **lengths, 
	       Dgp_Int32 *normal, Dgp_Int32 *tuned, Dgp_Bool *bestnew, 
	       hufftable_spec **bestTable, Dgp_Int32 ntables, 
	       Dgp_Bool custom_allowed, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jhtDecideTable()", DGP_PERRSTATE); 
    Dgp_Int32 i; 
    Dgp_Int32 bestbits; 
    Dgp_Int32 besti; 
 
    if (ntables == 0) { 
	/* What did they mean? */ 
      NORMAL_RETURN; 
	} 
 
    /* Check the normal tables */ 
    *bestnew = DGP_FALSE; 
    bestbits = normal[0]; 
    besti = 0; 
    for (i = 1; i < ntables; i++) { 
	if (normal[i] < bestbits) { 
	    bestbits = normal[i]; 
	    besti = i; 
	    } 
	} 
 
    if (custom_allowed) { 
	/* 
	 * and the tuned tables, assuming JHTSEGOVERHEAD bits of overhead 
	 * plus JHTSEGLINEBITS per line 
	 */ 
	for (i = 0; i < ntables; i++) { 
	    if (tuned[i] + JHTSEGOVERHEAD + 
		JHTSEGLINEBITS * tables[i]->nlines < 
		bestbits) { 
		bestbits = tuned[i] + JHTSEGOVERHEAD + 
		    JHTSEGLINEBITS * tables[i]->nlines; 
		besti = i; 
		*bestnew = DGP_TRUE; 
		} 
	    } 
	} 
 
    /* Communicate back the decision */ 
    if (*bestnew) { 
	/* Best choice was a custom table - build it */ 
	DGP_ERR_PASS_CLEANUP(jhtTuneStandardTable(tables[besti], 
					  counts[besti], lengths[besti], 
					  dgp_state, bestTable)); 
	} 
    else { 
	*bestTable = tables[besti]; 
	} 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
