/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jbitfile.c,v 1.22 2003/01/27 21:48:35 awilliam Exp $ 
 * 
 */ 
 
#include "jbitfile.h" 
#include <sys/types.h> 
#include <stdlib.h> 
#include "datasource.h" 
#include "datasink.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/jbitfile.c,v 1.22 2003/01/27 21:48:35 awilliam Exp $"); 
 
 
 
 
/* 
 * This module provides bit-level I/O on an in-memory file. 
 * Exports: 
 *	type jbitfile 
 * 
 *	constant jbitfile NULLJBITFILE 
 * 
 *	jbitfile jbfOpen(Dgp_UInt8 *buf, size_t buflen, 
 *			 Dgp_Bool (*more_data)(Dgp_UInt8 **buf, 
 *					      size_t *buflen, 
 *				     	      void *userdata), 
 *			 void *userdata, Dgp_Bool forWrite) - 
 *		open a jbitfile for reading or writing. 
 * 
 *	jbitfile jbfOpenFromDataSource(DataSource data) - open a jbitfile 
 *		for reading from the given DataSource. 
 * 
 *	jbitfile jbfOpenFromDataSink(DataSink data) - open a jbitfile 
 *		for writing to the given DataSink. 
 * 
 *	void jbfFlushByte(jbitfile f) - flush to a byte boundary. 
 * 
 *	Dgp_Bool jbfWriteBits(jbitfile f, Dgp_UInt32 code, Dgp_Int32 codelen) - 
 *		write out some bits. 
 * 
 *	Dgp_UInt32 jbfReadBits(jbitfile f, Dgp_Int32 codelen, Dgp_Bool *end) - 
 *		read some bits. 
 * 
 *	Dgp_UInt8 *jbfPos(jbitfile f) - return the location of the next 
 *					    untouched byte. 
 * 
 *	Dgp_Bool jbfClose(jbitfile f) - finish up the file. 
 * 
 *	Dgp_Bool jbfCloseFromDataSource(jbitfile f) - finish up a file that 
 *		was opened with jbfOpenFromDataSource. 
 * 
 *	Dgp_Bool jbfCloseFromDataSink(jbitfile f) - finish up a file that 
 *		was opened with jbfOpenFromDataSink. 
 * 
 *	void jbfSuspend(jbitfile f) - suspend a jbitfile, so that you can 
 *		mess with the underlying buffer memory 
 * 
 *	void jbfResume(jbitfile f) - let the jbitfile sync up with an 
 *		underlying DataSource or DataSink. 
 * 
 *	int jbfPeek(jbitfile f) - peek at the next 8 bits, if possible 
 * 
 * To write to a jbitfile: 
 * - open the file with jbfOpen.  Pass in a buffer (aDgp_Int32 with its length) 
 *   to be written to.  The more_data pointer should be a callback routine 
 *   that is called when the buffer has been filled.  more_data will be 
 *   called with (*buf, *buflen) equal to the location and size of the 
 *   full buffer; it should return with (*buf, *buflen) equal to the 
 *   location and size of an empty buffer (which can be the same as the one 
 *   that was just filled, if more_data has copied its contents). 
 *   more_data can be NULL; in that case, only the initial buffer will 
 *   be used. 
 *   The initial buffer can be NULL; in that case, more_data will be 
 *   called immediately to provide some data area. 
 *   more_data can return DGP_FALSE if there is no more output space. 
 * - call jbfWriteBits a number of times.  jbfWriteBits returns DGP_FALSE if 
 *   it cannot output the code.  jbfWriteBits writes into the bytes of 
 *   the output buffer from MSB to LSB.  The bits that jbfWriteBits 
 *   writes are the lowest codelen bits of code. 
 * - call jbfFlushByte as required; jbfFlushByte fills the remaining bits 
 *   of any partially-filled byte with zero, 
 *   and advances the write pointer to the MSB of the next byte. 
 * - call jbfPos as required; jbfPos performs a jbfFlushByte and returns 
 *   a pointer to the next byte after the most-recently filled byte. 
 * - call jbfSkipBytes as required; jbfSkipBytes performs a jbfFlushByte 
 *   and then advances the write pointer by skip bytes. 
 * - finally, call jbfClose; this performs a jbfFlushByte and destroys the 
 *   jbitfile. 
 * 
 * You can call jbfOpenFromDataSink to create a jbitfile that writes into 
 * a DataSink.  Call jbfCloseFromDataSink when you're done with this jbitfile, 
 * so that all the buffer pointers get properly synchronised.  While the 
 * jbitfile is active, don't mess with the DataSink, unless you bracket 
 * such messing with jbfSuspend and jbfResume.  jbfCloseFromDataSink can 
 * fail, in which case the jbitfile is destroyed, but some data may have been 
 * lost. 
 * 
 * To read from a jbitfile: 
 * - as before, call jbfOpen.  In this case, more_data will be called 
 *   whenever there are no bytes remaining to be read; it is called 
 *   with (*buf, *buflen) equal to the location and size of the 
 *   empty buffer, and should return with (*buf, *buflen) equal to the 
 *   location and size of a full buffer (which can be the same as the one 
 *   that was just emptied). 
 *   As in the write case, more_data or the initial buffer can be NULL, and 
 *   more_data can return DGP_FALSE if there is nothing more to be read. 
 * - call jbfReadBits a number of times.  jbfReadBits returns the bits read 
 *   in the low-order codelen bits of the return value.  If there was 
 *   some sort of error (no more data), it sets *end to DGP_TRUE. 
 * - jbfFlushByte skips over the remaining bits in the current 
 *   partially-consumed byte. 
 * - jbfPos does a jbfFlushByte and returns a pointer to the first 
 *   untouched byte or NULL on failure. 
 * - jbfSkipBytes does a jbfFlushByte and skips over skip additional bytes. 
 * - jbfClose destroys the jbitfile. 
 * - while the file is open, you can call jbfPeek to examine the next 
 *   8 bits in the file without changing the file's (externally-visible) 
 *   state.  jbfPeek may (at its discretion) return -1; this 
 *   is not an error, but merely an indication that it can't safely peek 
 *   at the next 8 bits without changing the file's state. 
 * 
 * jbfOpenFromDataSource creates a jbitfile from the given DataSource. 
 * If you use this, you should also use jbfCloseFromDataSource to close 
 * the file; this ensures that the file pointers stay correct.  And don't 
 * do anything to that DataSource while the jbitfile exists, of course. 
 * Or at least if you do, call jbfSuspend before doing anything, and 
 * jbfResume after messing with the DataSource. 
 */ 
 
struct _jbitfile { 
    Dgp_UInt8 *buf; 
    Dgp_UInt8 *bufpos; 
    size_t bufleft; 
    size_t count; 
    Dgp_Int32 bitsleft; 
    Dgp_UInt32 curint; 
    ErrCode (*more_data)(Dgp_UInt8 **buf, size_t *buflen, void *userdata); 
    void *userdata; 
    Dgp_State *dgp_state; 
    Dgp_Bool forWrite; 
    }; 
 
static ErrCode clearcurint(jbitfile f); 
 
ErrCode 
jbfOpen(Dgp_UInt8 *buf, size_t buflen, 
	ErrCode (*more_data)(Dgp_UInt8 **buf, size_t *buflen, 
			     void *userdata), 
	void *userdata, Dgp_Bool forWrite,Dgp_State *dgp_state, jbitfile *f) 
{ 
  DGP_PROLOGUE("jbfOpen()", DGP_PERRSTATE); 
  Dgp_Allocators *a = &(dgp_state->a); 
    jbitfile newjbf = NULLJBITFILE; 
 
 
    *f = NULLJBITFILE; 
 
    /* Make sure we have some data */ 
    if ((buf == (Dgp_UInt8 *)NULL) || (buflen == 0)) { 
      DGP_ERR_PASS((*more_data)(&buf, &buflen, userdata)); 
      DGP_ERR_GEN((buf == (Dgp_UInt8 *)NULL) || 
		  (buflen == 0), cErrIO_BadFileLength); 
    } 
 
    newjbf = (jbitfile)(a->malloc)(sizeof(*newjbf)); 
    DGP_CHECK_ALLOC(newjbf); 
 
    newjbf->buf = buf; 
    newjbf->more_data = more_data; 
    newjbf->userdata = userdata; 
    newjbf->dgp_state = dgp_state; 
    newjbf->bufleft = buflen; 
    newjbf->forWrite = forWrite; 
    newjbf->count = 0; 
 
    newjbf->curint = 0; 
    if (forWrite) { 
	newjbf->bitsleft = DGP_BITS_UINT32; 
	} 
    else { 
	newjbf->bitsleft = 0; 
	} 
 
    /* Point to the next untouched byte */ 
    newjbf->bufpos = buf; 
 
    *f = newjbf; 
    newjbf = NULLJBITFILE; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * A callback: given a DataSource, find some more data and return it. 
 */ 
static ErrCode 
dsrc_callback(Dgp_UInt8 **buf, size_t *buflen, void *userdata) 
{ 
    DataSource ds = (DataSource)userdata; 
    DGP_PROLOGUE("dsrc_callback()", ds->DGP_PERRSTATE); 
 
    DGP_ERR_GEN((ds->more_data == (ErrCode (*)(DataSource))NULL), cErrNullPtr); 
    DGP_ERR_PASS((*ds->more_data)(ds)); 
    *buf = ds->buf + ds->bufpos; 
    *buflen = ds->bufsize - ds->bufpos; 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jbfOpenFromDataSource(DataSource data, jbitfile *_f) 
{ 
  DGP_PROLOGUE("jbfOpenFromDataSource()", data->DGP_PERRSTATE); 
  jbitfile f = NULLJBITFILE; 
 
    /* Find the data buffer; create a callback that will do the right thing. */ 
    DGP_ERR_PASS(jbfOpen(data->buf + data->bufpos, 
			 data->bufsize - data->bufpos, 
			 dsrc_callback, (void *)data, DGP_FALSE, data->dgp_state,  
			 &f)); 
 
    *_f = f; 
    f = NULLJBITFILE; 
 
    NORMAL_RETURN; 
} 
 
/* 
 * A callback: given a DataSink, find some more data and return it. 
 * We have to update the data sink first to indicate that the buffer is full. 
 */ 
static ErrCode 
dsink_callback(Dgp_UInt8 **buf, size_t *buflen, void *userdata) 
{ 
    DataSink ds = (DataSink)userdata; 
    DGP_PROLOGUE("dsink_callback()", ds->DGP_PERRSTATE); 
 
    /* We've filled the entire buffer */ 
    ds->bufpos = ds->bufsize; 
 
    DGP_ERR_GEN((ds->more_data == (ErrCode (*)(DataSink))NULL), cErrNullPtr); 
    DGP_ERR_PASS((*ds->more_data)(ds)); 
 
    *buf = ds->buf + ds->bufpos; 
    *buflen = ds->bufsize - ds->bufpos; 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jbfOpenFromDataSink(DataSink data, jbitfile *_f) 
{ 
  DGP_PROLOGUE("jbfOpenFromDataSource()", data->DGP_PERRSTATE); 
  jbitfile f = NULLJBITFILE; 
 
    /* Find the data buffer; create a callback that will do the right thing. */ 
    DGP_ERR_PASS(jbfOpen(data->buf + data->bufpos, 
			 data->bufsize - data->bufpos, 
			 dsink_callback, (void *)data, DGP_TRUE, data->dgp_state, 
			 &f)); 
 
    *_f = f; 
    f = NULLJBITFILE; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Get some more data. 
 */ 
static ErrCode 
getmore(jbitfile f) 
{ 
  DGP_PROLOGUE("getmore()", f->DGP_PERRSTATE); 
   
    DGP_ASSERT(f->bufleft == 0); 
 
    DGP_ERR_GEN((f->more_data == (ErrCode (*)(Dgp_UInt8 **, size_t *, 
					       void *))NULL), 
		 cErrNullPtr); 
    DGP_ERR_PASS((*f->more_data)(&f->buf, &f->bufleft, f->userdata)); 
 
    DGP_ASSERT(f->buf != (Dgp_UInt8 *)NULL); 
    DGP_ASSERT(f->bufleft > 0); 
    f->bufpos = f->buf; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Load up to a full Dgp_Int32 into curint.  Trash all the state associated 
 * with curint. 
 */ 
static ErrCode 
loadbuf(jbitfile f) 
{ 
  DGP_PROLOGUE("loadbuf()", f->DGP_PERRSTATE); 
    Dgp_UInt8 *p = f->bufpos; 
    Dgp_UInt32 l; 
 
    if (f->bufleft >= sizeof(Dgp_UInt32)) { 
	/* Enough data left to fill the entire buffer */ 
	/* 
	 * We have to split into L32 and L64 cases... 
	 */ 
	if (DGP_BITS_UINT32 == 32) { 
	    f->curint = (p[0] << (DGP_BITS_UINT8 * 3)) | 
		(p[1] << (DGP_BITS_UINT8 * 2)) | 
		(p[2] << DGP_BITS_UINT8) | 
		p[3]; 
	    } 
	else { 
	    /* 
	     * The odd-looking shift amounts here are a hack to shut up 
	     * the compiler: if DGP_BITS_UINT32 is 32, then if I just said 
	     * p[0] << (DGP_BITS_UINT8 * 7), that would be a left shift by 56, 
	     * and the compiler would complain that it's illegal.  If 
	     * I AND the shift amounts with DGP_BITS_UINT32 - 1, then 
	     * - when DGP_BITS_UINT32 is 32, all the shifts are < 32.  They're 
	     *   incorrect, but this is dead code then anyway 
	     * - when DGP_BITS_UINT32 is 64, all the shifts are unaffected. 
	     *   Of course, the compiler has to do at least the simplest 
	     *   sort of constant expression evaluation; it's hard to find 
	     *   a compiler that doesn't. 
	     */ 
	    f->curint = 
		(p[0] << ((DGP_BITS_UINT8 * 7) & (DGP_BITS_UINT32 - 1))) | 
		(p[1] << ((DGP_BITS_UINT8 * 6) & (DGP_BITS_UINT32 - 1))) | 
		(p[2] << ((DGP_BITS_UINT8 * 5) & (DGP_BITS_UINT32 - 1))) | 
		(p[3] << ((DGP_BITS_UINT8 * 4) & (DGP_BITS_UINT32 - 1))) | 
		(p[4] << (DGP_BITS_UINT8 * 3)) | 
		(p[5] << (DGP_BITS_UINT8 * 2)) | 
		(p[6] << DGP_BITS_UINT8) | p[7]; 
	    } 
	f->bufpos += sizeof(Dgp_UInt32); 
	f->bufleft -= sizeof(Dgp_UInt32); 
	f->bitsleft = DGP_BITS_UINT8 * sizeof(Dgp_UInt32); 
	} 
    else if (f->bufleft == 0) { 
	/* No data at all - get some */ 
      DGP_ERR_PASS(getmore(f)); 
	/* Now there's something - recurse to get it */ 
      DGP_ERR_PASS(loadbuf(f)); 
      NORMAL_RETURN; 
	} 
    else { 
	/* Only a partial buffer's worth */ 
	/* Load all the bytes we've got, right-justified. */ 
	l = 0; 
	switch(f->bufleft) { 
	  case 7: 
	    l |= *p++ << ((DGP_BITS_UINT8 * 6) & (DGP_BITS_UINT32 - 1)); 
	  case 6: 
	    l |= *p++ << ((DGP_BITS_UINT8 * 5) & (DGP_BITS_UINT32 - 1)); 
	  case 5: 
	    l |= *p++ << ((DGP_BITS_UINT8 * 4) & (DGP_BITS_UINT32 - 1)); 
	  case 4: 
	    l |= *p++ << (DGP_BITS_UINT8 * 3); 
	  case 3: 
	    l |= *p++ << (DGP_BITS_UINT8 * 2); 
	  case 2: 
	    l |= *p++ << DGP_BITS_UINT8; 
	  case 1: 
	    l |= *p++; 
	  } 
	f->curint = l; 
	f->bitsleft = DGP_BITS_UINT8 * f->bufleft; 
	f->bufpos += f->bufleft; 
	f->bufleft = 0; 
	} 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Write out all the stuff in curint.  Make curint empty. 
 * Actually, if writing curint would overfill the buffer, then 
 * curint might not actually end up empty - some bits might be left 
 * in curint. 
 */ 
static ErrCode 
storebuf(jbitfile f) 
{ 
  DGP_PROLOGUE("storebuf()", f->DGP_PERRSTATE); 
 
    Dgp_UInt8 *p = f->bufpos; 
    Dgp_UInt32 l = f->curint; 
 
    if (f->bufleft >= sizeof(Dgp_UInt32)) { 
	/* Enough room to store it all */ 
	if (sizeof(Dgp_UInt32) == 4) { 
	    p[0] = (Dgp_UInt8)(l >> (DGP_BITS_UINT8 * 3)); 
	    p[1] = (Dgp_UInt8)((l >> (DGP_BITS_UINT8 * 2)) & 0xff); 
	    p[2] = (Dgp_UInt8)((l >> DGP_BITS_UINT8) & 0xff); 
	    p[3] = (Dgp_UInt8)(l & 0xff); 
	    } 
	else { 
	    p[0] = (Dgp_UInt8)(l >> ((DGP_BITS_UINT8 * 7) & (DGP_BITS_UINT32 - 1))); 
	    p[1] = (Dgp_UInt8)((l >> ((DGP_BITS_UINT8 * 6) & (DGP_BITS_UINT32 - 1))) & 0xff); 
	    p[2] = (Dgp_UInt8)((l >> ((DGP_BITS_UINT8 * 5) & (DGP_BITS_UINT32 - 1))) & 0xff); 
	    p[3] = (Dgp_UInt8)((l >> ((DGP_BITS_UINT8 * 4) & (DGP_BITS_UINT32 - 1))) & 0xff); 
	    p[4] = (Dgp_UInt8)((l >> (DGP_BITS_UINT8 * 3)) & 0xff); 
	    p[5] = (Dgp_UInt8)((l >> (DGP_BITS_UINT8 * 2)) & 0xff); 
	    p[6] = (Dgp_UInt8)((l >> DGP_BITS_UINT8) & 0xff); 
	    p[7] = (Dgp_UInt8)(l & 0xff); 
	    } 
	f->bufpos += sizeof(Dgp_UInt32); 
	f->bufleft -= sizeof(Dgp_UInt32); 
	f->bitsleft = DGP_BITS_UINT8 * sizeof(Dgp_UInt32); 
	f->curint = 0; 
	f->count += sizeof(Dgp_UInt32); 
	} 
    else if (f->bufleft == 0) { 
	/* Get some room to write */ 
      DGP_ERR_PASS(getmore(f)); 
	/* and recurse to write */ 
      DGP_ERR_PASS(storebuf(f)); 
      NORMAL_RETURN; 
	} 
    else { 
	/* 
	 * Can only store some of curint.  Store as many as possible of 
	 * the most significant bytes, shifting up as we go. 
	 */ 
	switch(f->bufleft) { 
	  case 7: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 6: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 5: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 4: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 3: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 2: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  case 1: 
	    *p++ = (Dgp_UInt8)(l >> (DGP_BITS_UINT32 - DGP_BITS_UINT8)); 
	    l <<= DGP_BITS_UINT8; 
	  } 
	/* The high-order bytes of l now contain the unwritten data. */ 
	f->curint = l; 
	f->bitsleft = DGP_BITS_UINT8 * f->bufleft; 
	f->bufpos = p; 
	f->count += f->bufleft; 
	f->bufleft = 0; 
	} 
 
    NORMAL_RETURN; 
    } 
 
void 
jbfFlushByte(jbitfile f) 
{ 
    /* How many bits to skip? */ 
    Dgp_Int32 bitskip = f->bitsleft & 7; 
 
    if (bitskip > 0) { 
	f->bitsleft -= bitskip; 
	if (f->forWrite) { 
	    /* 
	     * Clear bits bitskip + bitsleft - 1 ... bitsleft 
	     * (1 << bitskip) - 1 is bitskip 1s at the LS end; 
	     * shifting that left by bitsleft puts it in the right place; 
	     * negating it and ANDing it clears the right bits. 
	     */ 
	    f->curint &= ~(((1ul << bitskip) - 1) << f->bitsleft); 
	    } 
	} 
    } 
 
ErrCode 
jbfWriteBits(jbitfile f, Dgp_UInt32 code, Dgp_Int32 codelen) 
{ 
  DGP_PROLOGUE("jbfWriteBits()", f->DGP_PERRSTATE); 
 
    /* 
     * codelen can be as large as 32.  We need to ensure that there are 
     * no unprotected shifts by codelen, because those won't do the right 
     * thing. 
     */ 
    while (codelen > f->bitsleft) { 
	/* Fill any space left in curint */ 
	if (f->bitsleft > 0) { 
	    codelen -= f->bitsleft; 
	    /* Put the upper few bits of code into the lower bits of curint */ 
	    /* 
	     * These shifts are codelen-32 safe because codelen just got 
	     * decremented by a non-zero bitsleft 
	     */ 
	    f->curint |= code >> codelen; 
	    /* and give back just the remaining bits of code */ 
	    code &= ((1ul << codelen) - 1); 
	    f->bitsleft = 0; 
	    } 
 
	/* 
	 * Now make some more room in curint.  Note that 
	 * storebuf isn't guaranteed to give back an absolutely empty 
	 * curint - thus the "while" above. 
	 */ 
	DGP_ERR_PASS(storebuf(f)); 
	} 
 
    if (codelen > 0) { 
	f->bitsleft -= codelen; 
 
	/* bitsleft can never be 32 because it just got decremented */ 
	f->curint |= code << f->bitsleft; 
	} 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jbfReadBits(jbitfile f, Dgp_Int32 codelen, Dgp_UInt32 *val) 
{ 
  DGP_PROLOGUE("jbfReadBits()", f->DGP_PERRSTATE); 
 
    Dgp_UInt32 code = 0; 
 
    *val = 0; 
 
    /* As above, we need to be safe when codelen==32 or bitsleft==32 */ 
    while (codelen > f->bitsleft) { 
	if (f->bitsleft > 0) { 
	    codelen -= f->bitsleft; 
	    /* Read out all the low-order bits of curint into code */ 
	    /* Shifts are safe because bitsleft < codelen <= 32 */ 
	    code <<= f->bitsleft; 
	    code |= f->curint & ((1ul << f->bitsleft) - 1); 
 
	    /* Now get some more data */ 
	    f->bitsleft = 0; 
	    } 
 
	DGP_ERR_PASS(loadbuf(f)); 
	} 
 
    if (codelen > 0) { 
	/* Read bits bitsleft ... bitsleft-codelen+1 out */ 
	f->bitsleft -= codelen; 
 
	/* Special-case codelen==DGP_BITS_UINT32 */ 
	if (codelen == DGP_BITS_UINT32) { 
	    /* 
	     * We can only get to this case when codelen==DGP_BITS_UINT32, and 
	     * before the decrement above, bitsleft was also DGP_BITS_UINT32. 
	     * This means that bitsleft is now zero, and the code we want 
	     * to read is *all* of curint. 
	     */ 
	    code = f->curint; 
	    } 
	else { 
	    code <<= codelen; 
	    code |= (f->curint >> f->bitsleft) & ((1ul << codelen) - 1); 
	    } 
	} 
 
    *val = code; 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jbfPos(jbitfile f, Dgp_UInt8 **pos) 
{ 
  DGP_PROLOGUE("jbfPos()", f->DGP_PERRSTATE); 
 
    /* On read, we do read-ahead; on write we do write-behind */ 
    if (f->forWrite) { 
	/* 
	 * We can't get a valid pointer without clearing the curint 
	 * buffer: as we're doing write-behind, we might already have 
	 * more data in curint than can be accomodated in the current 
	 * buffer. 
	 */ 
      DGP_ERR_PASS(clearcurint(f)); 
 
	/* curint is now empty and f->bufpos is the next untouched byte */ 
      *pos = f->bufpos; 
	} 
    else { 
	/* 
	 * This one's easier: just flush to a byte boundary, count the 
	 * number of unconsumed bytes in curint, and offset that from the 
	 * address of the next untouched byte. 
	 */ 
	jbfFlushByte(f); 
 
	*pos = (f->bufpos - f->bitsleft / DGP_BITS_UINT8); 
	} 
 
    NORMAL_RETURN; 
    } 
 
static ErrCode 
clearcurint(jbitfile f) 
{ 
  DGP_PROLOGUE("clearcurint()", f->DGP_PERRSTATE); 
 
    Dgp_Int32 toWrite; 
    Dgp_UInt8 b; 
 
    if (!f->forWrite) { 
	/* 
	 * This is the easy case - just point to the next untouched 
	 * byte and set curint to 0, bitsleft to 0. 
	 */ 
	DGP_ERR_PASS(jbfPos(f, &(f->bufpos))); 
	f->curint = 0; 
	f->bitsleft = 0; 
 
	NORMAL_RETURN; 
	} 
 
    /* 
     * We're now dealing with a jbitfile open for writing.  This 
     * is a bit messy: we need to complete the current byte, and then 
     * store all completed bytes.  There may not be enough room in the 
     * current buffer to store all of them. 
     * 
     * storebuf stores all the bytes in curint, possibly including some 
     * uncompleted ones, so we can't use it. 
     */ 
    jbfFlushByte(f); 
 
    toWrite = sizeof(Dgp_UInt32) - f->bitsleft / DGP_BITS_UINT8; 
 
    for (; toWrite > 0; toWrite--) { 
		/* Pull out the top byte */ 
		b = (Dgp_UInt8)(f->curint >> (DGP_BITS_UINT8 * (sizeof(Dgp_UInt32) - 1))); 
		/* and shift the next byte up to the top */ 
		f->curint <<= DGP_BITS_UINT8; 
		f->bitsleft += DGP_BITS_UINT8; 
 
		/* Now write b */ 
		if (f->bufleft == 0) { 
		    /* No room - get some more */ 
		  DGP_ERR_PASS(getmore(f)); 
	    } 
 
		*f->bufpos++ = b; 
		f->bufleft--; 
		f->count++; 
	} 
     
    /* Ensure that the buffer is empty */ 
    f->curint = 0; 
    f->bitsleft = DGP_BITS_UINT32; 
 
    NORMAL_RETURN; 
    } 
 
void 
jbfClose(jbitfile f) 
{ 
  Dgp_Free tmp_free; 
 
  if(f == NULL) return; 
 
  tmp_free = f->DGP_FREE; 
 
    jbfFlushByte(f); 
    (tmp_free)((void *)f); 
    } 
 
ErrCode 
jbfCloseFromDataSource(jbitfile f) 
{ 
  DGP_PROLOGUE("jbfCloseFromDataSource()", f->DGP_PERRSTATE); 
 
    Dgp_UInt8 *bp; 
    DataSource ds = (DataSource)f->userdata; 
    Dgp_Free tmp_free; 
 
    /* Get the correct buffer position */ 
    DGP_ERR_PASS(jbfPos(f, &bp)); 
    /* and set that as the DataSource's buffer position */ 
    ds->bufpos = bp - ds->buf; 
     
    tmp_free = f->DGP_FREE; 
 
    (tmp_free)((void *)f); 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jbfCloseFromDataSink(jbitfile f) 
{ 
  DGP_PROLOGUE("jbfCloseFromDataSink()", f->DGP_PERRSTATE); 
 
    Dgp_UInt8 *bp; 
    DataSink ds = (DataSink)f->userdata; 
    Dgp_Free tmp_free; 
 
    tmp_free = f->DGP_FREE; 
 
    /* Get the correct buffer position */ 
    DGP_ERR_CHECK_NO_MSG(jbfPos(f, &bp)) { 
	/* Failure to flush the last byte in the buffer */ 
	(tmp_free)((void *)f); 
	DGP_ERR(DGP_CUR_ERROR); 
	} 
 
    /* and set that as the DataSink's buffer position */ 
    ds->bufpos = bp - ds->buf; 
 
    /* adjust the DataSink's count based on how many bytes we put in */ 
    dsAdjustCount(ds, f->count); 
 
    (tmp_free)((void *)f); 
    NORMAL_RETURN; 
    } 
 
/* f->userdata is either a DataSource or a DataSink */ 
ErrCode 
jbfSuspend(jbitfile f) 
{ 
  DGP_PROLOGUE("jbfSuspend()", f->DGP_PERRSTATE); 
 
  Dgp_UInt8 *bp; 
 
  DGP_ERR_PASS(clearcurint(f)); 
 
    if (f->forWrite) { 
	DataSink ds = (DataSink)f->userdata; 
 
	/* Get the buffer position and stuff it into the DataSink */ 
	DGP_ERR_PASS(jbfPos(f, &bp)); 
	ds->bufpos = bp - ds->buf; 
	/* and adjust the count */ 
	dsAdjustCount(ds, f->count); 
	} 
    else { 
	DataSource ds = (DataSource)f->userdata; 
 
	/* Get the buffer position and stuff it into the DataSource */ 
	DGP_ERR_PASS(jbfPos(f, &bp));	 
	ds->bufpos = bp - ds->buf; 
	} 
    NORMAL_RETURN; 
    } 
 
void 
jbfResume(jbitfile f) 
{ 
    if (f->forWrite) { 
	DataSink ds = (DataSink)f->userdata; 
 
	/* Load buffer pointers from the DataSink */ 
	f->buf = ds->buf + ds->bufpos; 
	f->bufleft = ds->bufsize - ds->bufpos; 
	f->bufpos = f->buf; 
	/* reset the count since jbfSuspend already took it into account */ 
	f->count = 0; 
	f->bitsleft = DGP_BITS_UINT32; 
	} 
    else { 
	DataSource ds = (DataSource)f->userdata; 
 
	/* Load buffer pointers from the DataSource */ 
	f->buf = ds->buf + ds->bufpos; 
	f->bufleft = ds->bufsize - ds->bufpos; 
	f->bufpos = f->buf; 
	f->bitsleft = 0; 
	} 
 
    /* and clear our working Dgp_Int32word */ 
    f->curint = 0; 
    } 
 
/* 
 * Strategy: if there's at least 8 bits in curint, return them. 
 * If there's fewer than 8 bits in curint, then try to load up to 
 * sizeof(Dgp_UInt32)-1 more bytes into curint.  Since we already 
 * do read-ahead, this is safe. 
 * 
 * If there's fewer than 8 bits in curint, and no more bytes in the buffer, 
 * return -1, as getting 8 bits to peek at would involve calling more_data, 
 * and this routine can't be that intrusive. 
 */ 
Dgp_Int32
jbfPeek(jbitfile f) 
{ 
    if (f->bitsleft >= 8) { 
	return((f->curint >> (f->bitsleft - 8)) & 0xff); 
	} 
 
    if (f->bufleft > 0) { 
	int toload = MIN(f->bufleft, sizeof(Dgp_UInt32) - 1); 
	int i; 
 
	for (i = 0; i < toload; i++) { 
	    f->curint = (f->curint << 8) + *(f->bufpos++); 
	    } 
	f->bufleft -= toload; 
	f->bitsleft += toload * DGP_BITS_UINT8; 
 
	return((f->curint >> (f->bitsleft - 8)) & 0xff); 
	} 
 
    return(-1); 
    } 
