/*
 * jccolor.c
 *
 * Copyright (C) 1991-1996, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains input colorspace conversion routines.
 *
 * 16.3.2001 / dhtesq.
 * Integrating Dave Abercrombie's itu-lab colorspace modifications to the ipcore
 * jpeg code.
 *
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeg6blib.h"


/* Private subobject */

typedef struct {
  struct jpeg_color_converter pub; /* public fields */

  /* Private state for RGB->YCC conversion */
  INT32 * rgb_ycc_tab;		/* => table for RGB to YCbCr conversion */
  Int32 * rgb_lab_tab;		/* => table for RGB to LAB f() conversion */
} my_color_converter;

typedef my_color_converter * my_cconvert_ptr;

/* Prototypes:
 */
METHODDEF(void)
rgb_lab_start (
    j_compress_ptr cinfo);
METHODDEF(void)
rgb_icclab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows);
METHODDEF(void)
rgb_itulab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows);

/**************** RGB -> YCbCr conversion: most common case **************/

/*
 * YCbCr is defined per CCIR 601-1, except that Cb and Cr are
 * normalized to the range 0..MAXJSAMPLE rather than -0.5 .. 0.5.
 * The conversion equations to be implemented are therefore
 *	Y  =  0.29900 * R + 0.58700 * G + 0.11400 * B
 *	Cb = -0.16874 * R - 0.33126 * G + 0.50000 * B  + CENTERJSAMPLE
 *	Cr =  0.50000 * R - 0.41869 * G - 0.08131 * B  + CENTERJSAMPLE
 * (These numbers are derived from TIFF 6.0 section 21, dated 3-June-92.)
 * Note: older versions of the IJG code used a zero offset of MAXJSAMPLE/2,
 * rather than CENTERJSAMPLE, for Cb and Cr.  This gave equal positive and
 * negative swings for Cb/Cr, but meant that grayscale values (Cb=Cr=0)
 * were not represented exactly.  Now we sacrifice exact representation of
 * maximum red and maximum blue in order to get exact grayscales.
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^16 (about 4 digits precision); we have to divide
 * the products by 2^16, with appropriate rounding, to get the correct answer.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants times R,G,B for all possible values.
 * For 8-bit JSAMPLEs this is very reasonable (only 256 entries per table);
 * for 12-bit samples it is still acceptable.  It's not very reasonable for
 * 16-bit samples, but if you want lossless storage you shouldn't be changing
 * colorspace anyway.
 * The CENTERJSAMPLE offsets and the rounding fudge-factor of 0.5 are included
 * in the tables to save adding them separately in the inner loop.
 */

#define SCALEBITS	16	/* speediest right-shift on some machines */
#define CBCR_OFFSET	((INT32) CENTERJSAMPLE << SCALEBITS)
#define ONE_HALF	((INT32) 1 << (SCALEBITS-1))
#define FIX(x)		((INT32) ((x) * (1L<<SCALEBITS) + 0.5))

/* We allocate one big table and divide it up into eight parts, instead of
 * doing eight alloc_small requests.  This lets us use a single table base
 * address, which can be held in a register in the inner loops on many
 * machines (more than can hold all eight addresses, anyway).
 */

#define R_Y_OFF		0			/* offset to R => Y section */
#define G_Y_OFF		(1*(MAXJSAMPLE+1))	/* offset to G => Y section */
#define B_Y_OFF		(2*(MAXJSAMPLE+1))	/* etc. */
#define R_CB_OFF	(3*(MAXJSAMPLE+1))
#define G_CB_OFF	(4*(MAXJSAMPLE+1))
#define B_CB_OFF	(5*(MAXJSAMPLE+1))
#define R_CR_OFF	B_CB_OFF		/* B=>Cb, R=>Cr are the same */
#define G_CR_OFF	(6*(MAXJSAMPLE+1))
#define B_CR_OFF	(7*(MAXJSAMPLE+1))
#define TABLE_SIZE	(8*(MAXJSAMPLE+1))


/*
 * Initialize for RGB->YCC colorspace conversion.
 */

METHODDEF(void)
rgb_ycc_start (j_compress_ptr cinfo)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  INT32 * rgb_ycc_tab;
  INT32 i;

  /* Allocate and fill in the conversion tables. */
  cconvert->rgb_ycc_tab = rgb_ycc_tab = (INT32 *)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				(TABLE_SIZE * SIZEOF(INT32)));

  for (i = 0; i <= MAXJSAMPLE; i++) {
    rgb_ycc_tab[i+R_Y_OFF] = FIX(0.29900) * i;
    rgb_ycc_tab[i+G_Y_OFF] = FIX(0.58700) * i;
    rgb_ycc_tab[i+B_Y_OFF] = FIX(0.11400) * i     + ONE_HALF;
    rgb_ycc_tab[i+R_CB_OFF] = (-FIX(0.16874)) * i;
    rgb_ycc_tab[i+G_CB_OFF] = (-FIX(0.33126)) * i;
    /* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
     * This ensures that the maximum output will round to MAXJSAMPLE
     * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
     */
    rgb_ycc_tab[i+B_CB_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
/*  B=>Cb and R=>Cr tables are the same
    rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
*/
    rgb_ycc_tab[i+G_CR_OFF] = (-FIX(0.41869)) * i;
    rgb_ycc_tab[i+B_CR_OFF] = (-FIX(0.08131)) * i;
  }
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 *
 * Note that we change from the application's interleaved-pixel format
 * to our internal noninterleaved, one-plane-per-component format.
 * The input buffer is therefore three times as wide as the output buffer.
 *
 * A starting row offset is provided only for the output buffer.  The caller
 * can easily adjust the passed input_buf value to accommodate any row
 * offset required on that side.
 */

METHODDEF(void)
/*rgb_ycc_convert (j_compress_ptr cinfo,
		 JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
		 JDIMENSION output_row, int num_rows)*/
rgb_ycc_convert (j_compress_ptr cinfo,
		 JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
		 JDIMENSION output_row, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register INT32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0, outptr1, outptr2;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = GETJSAMPLE(inptr[RGB_RED]);
      g = GETJSAMPLE(inptr[RGB_GREEN]);
      b = GETJSAMPLE(inptr[RGB_BLUE]);
      inptr += RGB_PIXELSIZE;
      /* If the inputs are 0..MAXJSAMPLE, the outputs of these equations
       * must be too; we do not need an explicit range-limiting operation.
       * Hence the value being shifted is never negative, and we don't
       * need the general RIGHT_SHIFT macro.
       */
      /* Y */
      outptr0[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
      /* Cb */
      outptr1[col] = (JSAMPLE)
		((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
		 >> SCALEBITS);
      /* Cr */
      outptr2[col] = (JSAMPLE)
		((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
		 >> SCALEBITS);
    }
  }
}


/**************** Cases other than RGB -> YCbCr **************/


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles RGB->grayscale conversion, which is the same
 * as the RGB->Y portion of RGB->YCbCr.
 * We assume rgb_ycc_start has been called (we only use the Y tables).
 */

METHODDEF(void)
rgb_gray_convert (j_compress_ptr cinfo,
		  JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
		  JDIMENSION output_row, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register INT32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr = output_buf[0][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = GETJSAMPLE(inptr[RGB_RED]);
      g = GETJSAMPLE(inptr[RGB_GREEN]);
      b = GETJSAMPLE(inptr[RGB_BLUE]);
      inptr += RGB_PIXELSIZE;
      /* Y */
      outptr[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
    }
  }
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles Adobe-style CMYK->YCCK conversion,
 * where we convert R=1-C, G=1-M, and B=1-Y to YCbCr using the same
 * conversion as above, while passing K (black) unchanged.
 * We assume rgb_ycc_start has been called.
 */

METHODDEF(void)
cmyk_ycck_convert (j_compress_ptr cinfo,
		   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
		   JDIMENSION output_row, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r, g, b;
  register INT32 * ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0, outptr1, outptr2, outptr3;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    outptr3 = output_buf[3][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      r = MAXJSAMPLE - GETJSAMPLE(inptr[0]);
      g = MAXJSAMPLE - GETJSAMPLE(inptr[1]);
      b = MAXJSAMPLE - GETJSAMPLE(inptr[2]);
      /* K passes through as-is */
      outptr3[col] = inptr[3];	/* don't need GETJSAMPLE here */
      inptr += 4;
      /* If the inputs are 0..MAXJSAMPLE, the outputs of these equations
       * must be too; we do not need an explicit range-limiting operation.
       * Hence the value being shifted is never negative, and we don't
       * need the general RIGHT_SHIFT macro.
       */
      /* Y */
      outptr0[col] = (JSAMPLE)
		((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
		 >> SCALEBITS);
      /* Cb */
      outptr1[col] = (JSAMPLE)
		((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
		 >> SCALEBITS);
      /* Cr */
      outptr2[col] = (JSAMPLE)
		((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
		 >> SCALEBITS);
    }
  }
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles grayscale output with no conversion.
 * The source can be either plain grayscale or YCbCr (since Y == gray).
 */

METHODDEF(void)
grayscale_convert (j_compress_ptr cinfo,
		   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
		   JDIMENSION output_row, int num_rows)
{
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  int instride = cinfo->input_components;

  while (--num_rows >= 0) {
    inptr = *input_buf++;
    outptr = output_buf[0][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++) {
      outptr[col] = inptr[0];	/* don't need GETJSAMPLE() here */
      inptr += instride;
    }
  }
}


/*
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles multi-component colorspaces without conversion.
 * We assume input_components == num_components.
 */

METHODDEF(void)
null_convert (j_compress_ptr cinfo,
	      JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
	      JDIMENSION output_row, int num_rows)
{
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  register int ci;
  int nc = cinfo->num_components;
  JDIMENSION num_cols = cinfo->image_width;

  while (--num_rows >= 0) {
    /* It seems fastest to make a separate pass for each component. */
    for (ci = 0; ci < nc; ci++) {
      inptr = *input_buf;
      outptr = output_buf[ci][output_row];
      for (col = 0; col < num_cols; col++) {
	outptr[col] = inptr[ci]; /* don't need GETJSAMPLE() here */
	inptr += nc;
      }
    }
    input_buf++;
    output_row++;
  }
}


/*
 * Empty method for start_pass.
 */

METHODDEF(void)
null_method (j_compress_ptr cinfo)
{
  cinfo = cinfo;
  /* no work needed */
}

/*
 * (Added 16.3.2001 / dhtesq.)
 *
 * Convert some rows of samples to the JPEG colorspace.
 * This version handles RGB->grayscale conversion, which is the same
 * as the RGB->Y portion of RGB->YCbCr.
 * We assume rgb_ycc_start has been called (we only use the Y tables).
 *
 * Note: I've never tested this code (aber)
 */

METHODDEF(void)
grayscale_itulab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  register JSAMPLE tmp, Lmin, Lmax;
  JDIMENSION num_cols = cinfo->image_width;
  float range;

  if (cinfo->Lmax - cinfo->Lmin == 0) {
      grayscale_convert(cinfo, input_buf, output_buf, output_row, num_rows);
  }
  Lmin = (JSAMPLE)(cinfo->Lmin + 0.5);
  Lmax = (JSAMPLE)(cinfo->Lmax + 0.5);
  range = (float)(Lmax - Lmin);

  while (--num_rows >= 0) {
      inptr = *input_buf++;
      outptr = output_buf[0][output_row];
      output_row++;
      for (col = 0; col < num_cols; col++) {
	  tmp = GETJSAMPLE(*inptr);
	  if (tmp < Lmin) {
	      *outptr++ = Lmin;
	  } else if (tmp > Lmax) {
	      *outptr++ = Lmax;
	  } else {
	      *outptr++ = (JSAMPLE)(((float)((tmp - Lmin) * 255)) / range + 0.5);
	  }
	  inptr++;
      }
  }

  return;
}


/*
 * Module initialization routine for input colorspace conversion.
 */

GLOBAL(void)
jinit_color_converter (j_compress_ptr cinfo)
{
  my_cconvert_ptr cconvert;

  cconvert = (my_cconvert_ptr)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				SIZEOF(my_color_converter));
  cinfo->cconvert = (struct jpeg_color_converter *) cconvert;
  /* set start_pass to null method until we find out differently */
  cconvert->pub.start_pass = null_method;

  /* Make sure input_components agrees with in_color_space */
  switch (cinfo->in_color_space) {
  case JCS_GRAYSCALE:
    if (cinfo->input_components != 1)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  case JCS_RGB:
#if RGB_PIXELSIZE != 3
    if (cinfo->input_components != RGB_PIXELSIZE)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;
#endif /* else share code with YCbCr */

  case JCS_YCbCr:
    if (cinfo->input_components != 3)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  case JCS_CMYK:
  case JCS_YCCK:
    if (cinfo->input_components != 4)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;

  default:			/* JCS_UNKNOWN can be anything */
    if (cinfo->input_components < 1)
      ERREXIT(cinfo, JERR_BAD_IN_COLORSPACE);
    break;
  }

  /* Check num_components, set conversion method based on requested space */
  switch (cinfo->jpeg_color_space) {
  case JCS_GRAYSCALE:
    if (cinfo->num_components != 1)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_GRAYSCALE)
      cconvert->pub.color_convert = grayscale_convert;
    else if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_gray_convert;
    } else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = grayscale_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_RGB:
    if (cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_RGB && RGB_PIXELSIZE == 3)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_YCbCr:
    if (cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_ycc_convert;
    } else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_CMYK:
    if (cinfo->num_components != 4)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_CMYK)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_YCCK:
    if (cinfo->num_components != 4)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    if (cinfo->in_color_space == JCS_CMYK) {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = cmyk_ycck_convert;
    } else if (cinfo->in_color_space == JCS_YCCK)
      cconvert->pub.color_convert = null_convert;
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

/* Added 16.3.2001 (DA's) / dhtesq. */
/*///////////////////////////////////////// */
  case JCS_CIELAB:
  case JCS_ICCLAB:
    /* Technically, TIFF 6.0 spells out that these values are different by
     * the bounding, where CIELAB is bounded to [-127,127], and ICCLAB is
     * bounded to [-128,127].  But it appears that use of the values in JPEG
     * are such that they are not left as signed values, so both require the
     * addition of 128.  This would imply that the -128 could reasonably be
     * used for both, and that there is no difference between the
     * implementations of CIELAB and ICCLAB.
     */
    if ((cinfo->num_components != 3) &&
        (cinfo->num_components != 1)) {
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    }
    if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_lab_start;
      cconvert->pub.color_convert = rgb_icclab_convert;
    }
    else if (cinfo->in_color_space == JCS_GRAYSCALE) {
      cconvert->pub.start_pass = rgb_lab_start;
      cconvert->pub.color_convert = grayscale_convert;
    } else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_ITULAB:
    if ((cinfo->num_components != 3) &&
        (cinfo->num_components != 1)) {
        ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    }
    if (cinfo->in_color_space == JCS_RGB) {
      cconvert->pub.start_pass = rgb_lab_start;
      cconvert->pub.color_convert = rgb_itulab_convert;
    }
    else if (cinfo->in_color_space == JCS_GRAYSCALE) {
      cconvert->pub.start_pass = rgb_lab_start;
      cconvert->pub.color_convert = grayscale_itulab_convert;
    }
    else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;
/*///////////////////////////////////////// */

  default:			/* allow null conversion of JCS_UNKNOWN */
    if (cinfo->jpeg_color_space != cinfo->in_color_space ||
	cinfo->num_components != cinfo->input_components)
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    cconvert->pub.color_convert = null_convert;
    break;
  }
}

/* Added 16.3.2001 (DA's) / dhtesq. */
/**************** RGB -> ITULab conversion **************/

/*
 * ITULab is defined somewhere in T.42, but I haven't read it far enough
 * to see that an actual definition exists there.  My understanding is that
 * ITULab was a better name picked for something called FaxLab, which had
 * the obvious connotation of being tied to fax data.  FaxLab is, again
 * according to my understanding, CIELab with some gamut limitations.
 *
 * The following code was taken from the International Electrotechnical
 * Commission's: Default RGB Colour Space - sRGB.
 * Due to the non-linearities in the sRGB -> XYZ conversion, and the fact that
 * there is an optional gamut range, there is no straight conversion from
 * sRGB to FaxLab; instead the conversion is a two-step process:
 *	sRGB -> XYZ -> FaxLab.
 *
 * The normalization of sRGB8 -> sRGB' -> sRGB:
 *	For each x = R, G, and B:
 *	x' = x8 / 255;
 *	if (x' <= 0.04045) {
 *		x = x' / 12.92;
 *	} else {
 *		x = ((x' + 0.055) / 1.055) ** 2.4;
 *	}
 * Since each x8 is 8-bit integers, this can be rewritten as:
 *	if (x8 <= 10) {
 *		x = (x8 / 255) / 12.92;
 *	} else {
 *		x = (((x8 / 255) + 0.055) / 1.055) ** 2.4;
 *	}
 *
 * Here's the transformation from sRGB -> XYZ:
 *	X =  0.4124 * R + 0.3576 * G + 0.1805 * B
 *	Y =  0.2126 * R + 0.7152 * G + 0.0722 * B
 *	Z =  0.0193 * R + 0.1192 * G + 0.9505 * B
 * And here's the tranformation from XYZ -> ITULab:
 *	L =  295.8 * g(Y) - 40.3, bounded to [0,255]
 *	a =  750.0 * (g(X / 0.96391) - g(Y)) + 128.0, bounded to [0,255]
 *	b =  255.0 * (g(Y) - g(Z / 0.824704)) + 96.125, bounded to [0,255]
 *	where g(x) =
 *				x <= 0.008856	7.787 * x + 0.137931034
 *				x > 0.008856	x^(1/3)
 *
 *
 * What I've done is implement the actual formulas, allowing for
 * arbitrary gamuts (after all, TIFF-FX allows the gamut to be specified
 * via the Decode TIFF field).
 *
 * Here's the conversion XYZ -> Lab:
 *	fy = f(Y / Yn)
 *	L = 116.0 * fy - 16.0
 *	a = 500.0 * (f(X / Xn) - fy)
 *	b = 200.0 * (fy - f(Z / Zn))
 * where
 *	f(x) is ((x) <= 0.008856? 7.787 * (x) + 16.0/116.0: x ^ 1/3)
 *
 * Then taking the gamut ([Lmin,Lmax], [amin,amax], [bmin,bmax] into account,
 * the 8-bit convert Lab -> L8,a8,b8 is:
 *	L8 = eightbit((L - Lmin) * 256.0 / (Lmax - Lmin))
 *	a8 = eightbit((a - amin) * 256.0 / (amax - amin))
 *	b8 = eightbit((b - bmin) * 256.0 / (bmax - bmin))
 * where
 *	eightbit(x) is ((x) < 0.0? 0.0: (x) > 255.0? 255.0: (x))
 *
 * OPTIMIZATIONS, AND FLOATING-POINT MINIMIZATION
 * I did a lot of work to minimize the floating point, and to speed things
 * up.  Here's a list:
 *	1. Simplify some algebra, so that the number of mathematical operations
 *	   required inside the loops is minimized.
 *	2. Precompute some constants before the loops, for portions of
 *	   the equations that don't change inside the loops.
 *	3. Use shifted integers, instead of floats.  The precision on the
 *	   integers varies for different portions of the equations.  I varied
 *	   the precision until I found the optimal sRGB->ITULAB->sRGB result.
 *	4. Precompute the f() function, given a fixed-precision input, and
 *	   a fixed-precision output.  This helps most size a cuberoot() is
 *	   involved.
 *	5. Avoid divides where a bitshift can be used.
 *
 * CIELab vs ICCLAB vs ITULAB (FaxLab)
 * The difference between these three color spaces is the gamut range given
 * to the three, and the way the final values are stored:
 *	CIELab has the gamut L* [0,100], a* [-127,127], b* [-127,127].
 *		The values are stored in 8-bit values where:
 *		L8 [0,255] = L* * 256.0 / 100.0
 * 		a8 [-127,127] = a* railed to [-127,127]
 * 		b8 [-127,127] = b* railed to [-127,127]
 *	(Note that CIELab has one 8-bit value unused for a8 and b8: -128.)
 *
 *	ICCLAB has the gamut L* [0,100], a* [-128,127], b* [-128,127].
 *		The values are stored in 8-bit values where:
 *		L8 [0,255] = L* * 255.0 / 100.0
 * 		a8 [0,255] = a* + 128 railed to [0,255]
 * 		b8 [0,255] = b* + 128 railed to [0,255]
 *
 *	ITULAB has the DEFAULT gamut L* [0,100], b* [-85,85], b* [-75,125];
 *		but in TIFF-FX they are approximated to
 *			L* [0,100]
 *			a* [-21760/255, 21590/255]
 *			b* [-19200/255, 31800/255];
 *		and the gamut values can be parameters (in TIFF-FX they
 *		are specified in the Decode TIFF field).
 *		The values are stored in 8-bit values where:
 *		L8 [0,255] = (L* - Lmin) * 256.0 / (Lmax - Lmin), railed
 * 		a8 [0,255] = (a* - amin) * 256.0 / (amax - amin), railed
 * 		b8 [0,255] = (b* - bmin) * 256.0 / (bmax - bmin), railed
 *
 *
 * TODO:
 * Nothing has been done to accommodate 12-bit samples; nor 16-bit samples.
 */

#undef R_Y_OFF
#undef G_Y_OFF
#undef B_Y_OFF
#undef TABLE_SIZE

#define R_X_OFF		0			/* offset to R => X section */
#define G_X_OFF		(1*(MAXJSAMPLE+1))	/* offset to G => X section */
#define B_X_OFF		(2*(MAXJSAMPLE+1))	/* etc. */
#define R_Y_OFF		(3*(MAXJSAMPLE+1))
#define G_Y_OFF		(4*(MAXJSAMPLE+1))
#define B_Y_OFF		(5*(MAXJSAMPLE+1))
#define R_Z_OFF		(6*(MAXJSAMPLE+1))
#define G_Z_OFF		(7*(MAXJSAMPLE+1))
#define B_Z_OFF		(8*(MAXJSAMPLE+1))
#define TABLE_SIZE	(9*(MAXJSAMPLE+1))


/* Fast computation of the cube root, computed until accurate within an error
 * tolerance.
 * Note that for normal computation, you'd use error = 0.00001; however,
 * for L*a*b* conversion, you could set error = 0.001, but due to the few
 * number of computations, I've picked 0.00001.
 */
static float CDECL
cuberoot(float x, float error)
{
    float root = 0.6f, cube, diff;
    int negFlag = 0, count = 0;

    if (x < 0.0f) {
	negFlag = 1;
	x = -x;
    } else if (x == 0.0f) {
	return 0.0f;
    }
    do {
	root = (x / (root * root) + root) / 2.0f;
	count++;
	if (count > 25)
	    break;
	cube = root * root * root;
	if (cube > x)
	    diff = cube - x;
	else
	    diff = x - cube;
    } while (diff >= error);
    if (negFlag) {
	root = -root;
    }
    return root;
}


#ifndef USELINEARRGB

#define N_BITS_F	12	/* Works best at 14, but costs extra time */
#define N_F		(1 << N_BITS_F)
#define KZERO_BITS	12	/* If you change, see srgbLUT[] */
#define KONE_BITS	14
#define KTWO_BITS	3
#define KZERO		(1 << KZERO_BITS)
#define KONE		(1 << KONE_BITS)
#define KTWO		(1 << KTWO_BITS)

#define EIGHTBITFROMINT(x)	(unsigned char)((x) < 0? 0: (x) > 255? 255: ((x)))
#define INTLAB(x)	((x) >> (KONE_BITS + KTWO_BITS))
#define KONMO		(1 << (KZERO_BITS + KONE_BITS - N_BITS_F - 1))
#define f(x)		(ctab[((x) + KONMO) >> (KZERO_BITS + KONE_BITS - N_BITS_F)])

#else	/* Here's the linear-RGB values: (might need optimiztion, see comments
	 * above) */

#define N_BITS_F	10
#define N_F		(1 << N_BITS_F)
#define KONE_BITS	21
#define KTWO_BITS	1
#define KONE		(1 << KONE_BITS)
#define KTWO		(1 << KTWO_BITS)

#define EIGHTBITLAB(x)	(unsigned char)((x) < 0? 0: (x) > (KONE * KTWO * 255)? 255: \
			(x) >> (KONE_BITS + KTWO_BITS))
#define KONMO		(1 << (KONE_BITS - N_BITS_F - 1))
#define f(x)		(ctab[((x) + KONMO) >> (KONE_BITS - N_BITS_F)])
#endif


METHODDEF(void)
rgb_lab_start (
    j_compress_ptr cinfo)
{
    Int32 i;
    float f;
    Int32 *rgb_lab_tab;

    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;

    /* Allocate and fill in the conversion tables. */
	rgb_lab_tab = (Int32 *)((*cinfo->mem->alloc_small) ((j_common_ptr) cinfo,
		JPOOL_IMAGE, ((N_F + 1) * SIZEOF(Int32))));
    cconvert->rgb_lab_tab = rgb_lab_tab;

    for (i = 0; i < N_F+1; i++) {
		f = (float)i / (float)(N_F - 1);
		if (f <= 0.008856f) {
			f = 7.787f * f + 16.0f / 116.0f;
		} else {
			f = cuberoot(f, 0.00001f);
		}
		rgb_lab_tab[i] = (Int32)(f * (float)KONE);
    }

    return;
}


/**************** RGB -> ICCLAB conversion **************/

/* Note, I equate CIELAB and ICCLAB, since they differ only in
 * thier use of -128.
 */

METHODDEF(void)
rgb_icclab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{

    /* Use the full gamut */
    cinfo->Lmin = 0.0;
    cinfo->Lmax = 100.0;
    cinfo->amin = -128.0;
    cinfo->amax = 127.0;
    cinfo->bmin = -128.0;
    cinfo->bmax = 127.0;
    /* Technically, TIFF 6.0 spells out that these values are bounded to
     * [-127,127], but it appears that use of the values in JPEG are such
     * that they are not left as signed values, so the addition of 128
     * would imply that the -128 could reasonably be used.  This means that
     * there is no difference between the implementations of CIELAB and ICCLAB.
     */

    /* Use the ITULAB conversion, with the gamut just set */
    rgb_itulab_convert(cinfo, input_buf, output_buf, output_row, num_rows);
	return;
}


#ifndef USELINEARRGB

/*
 * Convert some rows of samples to the JPEG colorspace.
 *
 * Note that we change from the application's interleaved-pixel format
 * to our internal noninterleaved, one-plane-per-component format.
 * The input buffer is therefore three times as wide as the output buffer.
 *
 * A starting row offset is provided only for the output buffer.  The caller
 * can easily adjust the passed input_buf value to accommodate any row
 * offset required on that side.
 */

/* Do an in-place conversion from sRGB to ITULAB */

METHODDEF(void)
rgb_itulab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
    register int R, G, B;
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->rgb_lab_tab;
    register JSAMPROW inptr;
    register JSAMPROW outptr0, outptr1, outptr2;
    register JDIMENSION col;
    JDIMENSION num_cols = cinfo->image_width;
    Int32 X, Y, Z, L, a, b;
    Int32 x1, x2, x3, y1, y2, y3, z1, z2, z3;
    Int32 fy;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;	/* White point */
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */
    Int32 const1, const2, const3, const4, const5, const6;
    Int32 ebL, eba, ebb;	/* Eight-Bit L, a, b */
    /* sRGB8 to sRGB', precomputed for KZERO_BITS = 12 */
    Int32 srgbLUT[256] = {
	0,     1,     2,     4,     5,     6,     7,     9,
	10,    11,    12,    14,    15,    16,    18,    20,
	21,    23,    25,    27,    29,    31,    33,    35,
	37,    40,    42,    45,    48,    50,    53,    56,
	59,    62,    66,    69,    72,    76,    79,    83,
	87,    91,    95,    99,   103,   107,   112,   116,
	121,   126,   131,   136,   141,   146,   151,   156,
	162,   168,   173,   179,   185,   191,   197,   204,
	210,   217,   223,   230,   237,   244,   251,   258,
	265,   273,   280,   288,   296,   304,   312,   320,
	329,   337,   346,   354,   363,   372,   381,   390,
	400,   409,   419,   429,   438,   448,   458,   469,
	479,   490,   500,   511,   522,   533,   544,   556,
	567,   579,   590,   602,   614,   626,   639,   651,
	664,   676,   689,   702,   715,   729,   742,   756,
	769,   783,   797,   811,   826,   840,   855,   869,
	884,   899,   914,   930,   945,   961,   976,   992,
	1008,  1025,  1041,  1058,  1074,  1091,  1108,  1125,
	1142,  1160,  1177,  1195,  1213,  1231,  1249,  1268,
	1286,  1305,  1324,  1343,  1362,  1381,  1400,  1420,
	1440,  1460,  1480,  1500,  1521,  1541,  1562,  1583,
	1604,  1625,  1647,  1668,  1690,  1712,  1734,  1756,
	1778,  1801,  1824,  1846,  1869,  1893,  1916,  1940,
	1963,  1987,  2011,  2035,  2060,  2084,  2109,  2134,
	2159,  2184,  2210,  2235,  2261,  2287,  2313,  2339,
	2366,  2392,  2419,  2446,  2473,  2501,  2528,  2556,
	2584,  2612,  2640,  2668,  2697,  2725,  2754,  2783,
	2813,  2842,  2872,  2902,  2931,  2962,  2992,  3022,
	3053,  3084,  3115,  3146,  3178,  3209,  3241,  3273,
	3305,  3338,  3370,  3403,  3436,  3469,  3502,  3535,
	3569,  3603,  3637,  3671,  3705,  3740,  3775,  3810,
	3845,  3880,  3916,  3951,  3987,  4023,  4060,  4096
    };	

    /* If a different gamut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
	Lmin = cinfo->Lmin;
	Lmax = cinfo->Lmax;
	if (cinfo->amax - cinfo->amin > 0) {
	    amin = cinfo->amin;
	    amax = cinfo->amax;
	    if (cinfo->bmax - cinfo->bmin > 0) {
		bmin = cinfo->bmin;
		bmax = cinfo->bmax;
	    }
	}
    }

    /* Precompute the constants */
    const1 = (Int32)((116.0f * 256.0f / (Lmax - Lmin)) * KTWO);
    const2 = (Int32)((-(16.0f + Lmin) * 256.0f / (Lmax - Lmin)) * KONE * KTWO);
    const3 = (Int32)((500.0f * 256.0f / (amax - amin)) * KTWO);
    const4 = (Int32)((-amin * 256.0f / (amax - amin)) * KONE * KTWO);
    const5 = (Int32)((200.0f * 256.0f / (bmax - bmin)) * KTWO);
    const6 = (Int32)((-bmin * 256.0f / (bmax - bmin)) * KONE * KTWO);
    x1 = (UInt32)((0.4124f / Xn) * KONE);
    x2 = (UInt32)((0.3576f / Xn) * KONE);
    x3 = (UInt32)((0.1805f / Xn) * KONE);
    y1 = (UInt32)((0.2126f / Yn) * KONE);
    y2 = (UInt32)((0.7152f / Yn) * KONE);
    y3 = (UInt32)((0.0722f / Yn) * KONE);
    z1 = (UInt32)((0.0193f / Zn) * KONE);
    z2 = (UInt32)((0.1192f / Zn) * KONE);
    z3 = (UInt32)((0.9505f / Zn) * KONE);

    /* Create a LUT to go from 8-bit sRGB to sRGB */
#if 0
    {FILE *fp = fopen("srgb2srgbp.h", "w");
    Int32 i;
    double tmp;
    for (i = 0; i < 256; i++) {
	/* Convert from 8-bit sRGB to sRGB' to sRGB */
	if (i <= 10) {
	    srgbLUT[i] = (Int32)((float)i / 255.0f / 12.92f * (float)KZERO +
		0.5f);
	} else {
	    tmp = (double)((((float)i / 255.0f) + 0.055f) / 1.055f);
	    srgbLUT[i] = (Int32)((float)pow(tmp, 2.4) * (float)KZERO + 0.5f);
	}
	fprintf(fp, "%5d, ", srgbLUT[i]);
    }
    fclose(fp);
    }
#endif

    /* Convert the image */
    while (--num_rows >= 0) {
	inptr = *input_buf++;
	outptr0 = output_buf[0][output_row];
	outptr1 = output_buf[1][output_row];
	outptr2 = output_buf[2][output_row];
	output_row++;
	for (col = 0; col < num_cols; col++) {
	    R = GETJSAMPLE(inptr[RGB_RED]);
	    G = GETJSAMPLE(inptr[RGB_GREEN]);
	    B = GETJSAMPLE(inptr[RGB_BLUE]);
	    inptr += RGB_PIXELSIZE;

	    /* Convert from RGB -> XYZ -> L*a*b* */
/* These were rounded!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
	    X = srgbLUT[R] * x1 + srgbLUT[G] * x2 + srgbLUT[B] * x3;
	    Y = srgbLUT[R] * y1 + srgbLUT[G] * y2 + srgbLUT[B] * y3;
	    Z = srgbLUT[R] * z1 + srgbLUT[G] * z2 + srgbLUT[B] * z3;

	    fy = f(Y);
	    L = const1 * fy + const2;
	    a = const3 * (f(X) - fy) + const4;
	    b = const5 * (fy - f(Z)) + const6;

	    /* Compute the integer value */
	    ebL = INTLAB(L);
	    eba = INTLAB(a);
	    ebb = INTLAB(b);

	    /* If mapping back to gamut, do here */

	    /* Output the 8-bit result */
	    outptr0[col] = (JSAMPLE) EIGHTBITFROMINT(ebL);
	    outptr1[col] = (JSAMPLE) EIGHTBITFROMINT(eba);
	    outptr2[col] = (JSAMPLE) EIGHTBITFROMINT(ebb);
	}
    }

    return;
}


#else	/* Here's the linear-RGB: */

/**************** RGB -> ITULAB conversion **************/

METHODDEF(void)
rgb_itulab_convert (
    j_compress_ptr cinfo,
    JSAMPARRAY	input_buf,
    JSAMPIMAGE	output_buf,
    JDIMENSION	output_row,
    int		num_rows)
{
    register int R, G, B;
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->rgb_lab_tab;
    register JSAMPROW inptr;
    register JSAMPROW outptr0, outptr1, outptr2;
    register JDIMENSION col;
    JDIMENSION num_cols = cinfo->image_width;
    Int32 X, Y, Z, L, a, b;
    Int32 x1, x2, x3, y1, y2, y3, z1, z2, z3;
    Int32 fy;
    float Xn = 0.96422f, Yn = 1.000f, Zn = 0.82521f;	/* White point */
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    Int32 const1, const2, const3, const4, const5, const6;

    /* If a different gamut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
	Lmin = cinfo->Lmin;
	Lmax = cinfo->Lmax;
	if (cinfo->amax - cinfo->amin > 0) {
	    amin = cinfo->amin;
	    amax = cinfo->amax;
	    if (cinfo->bmax - cinfo->bmin > 0) {
		bmin = cinfo->bmin;
		bmax = cinfo->bmax;
	    }
	}
    }

    /* Precompute the constants */
    const1 = (Int32)((116.0f * 256.0f / (Lmax - Lmin)) * KTWO);
    const2 = (Int32)((-(16.0f + Lmin) * 256.0f / (Lmax - Lmin)) * KONE * KTWO);
    const3 = (Int32)((500.0f * 256.0f / (amax - amin)) * KTWO);
    const4 = (Int32)((-amin * 256.0f / (amax - amin)) * KONE * KTWO);
    const5 = (Int32)((200.0f * 256.0f / (bmax - bmin)) * KTWO);
    const6 = (Int32)((-bmin * 256.0f / (bmax - bmin)) * KONE * KTWO);

    x1 = (UInt32)((0.468682f / (255.0f * Xn)) * KONE);
    x2 = (UInt32)((0.356608f / (255.0f * Xn)) * KONE);
    x3 = (UInt32)((0.138620f / (255.0f * Xn)) * KONE);
    y1 = (UInt32)((0.252939f / (255.0f * Yn)) * KONE);
    y2 = (UInt32)((0.684458f / (255.0f * Yn)) * KONE);
    y3 = (UInt32)((0.062603f / (255.0f * Yn)) * KONE);
    z1 = (UInt32)((0.022318f / (255.0f * Zn)) * KONE);
    z2 = (UInt32)((0.109283f / (255.0f * Zn)) * KONE);
    z3 = (UInt32)((0.693103f / (255.0f * Zn)) * KONE);

    /* Convert the image */
    while (--num_rows >= 0) {
	inptr = *input_buf++;
	outptr0 = output_buf[0][output_row];
	outptr1 = output_buf[1][output_row];
	outptr2 = output_buf[2][output_row];
	output_row++;
	for (col = 0; col < num_cols; col++) {
	    R = GETJSAMPLE(inptr[RGB_RED]);
	    G = GETJSAMPLE(inptr[RGB_GREEN]);
	    B = GETJSAMPLE(inptr[RGB_BLUE]);
	    inptr += RGB_PIXELSIZE;

	    /* Convert from RGB -> XYZ -> L*a*b* */
	    X = R * x1 + G * x2 + B * x3;
	    Y = R * y1 + G * y2 + B * y3;
	    Z = R * z1 + G * z2 + B * z3;

	    fy = f(Y);
	    L = const1 * fy + const2;
	    a = const3 * (f(X) - fy) + const4;
	    b = const5 * (fy - f(Z)) + const6;

	    /* Rail, and output */
	    outptr0[col] = (JSAMPLE) EIGHTBITLAB(L);
	    outptr1[col] = (JSAMPLE) EIGHTBITLAB(a);
	    outptr2[col] = (JSAMPLE) EIGHTBITLAB(b);
	}
    }

    return;
}
#endif	/* USELINEARRGB */


/*
$Log: jccolor.c,v $
Revision 1.8  2003/02/11 01:31:41  mkrocken
fixed problems with comments caused by previous commit

Revision 1.7  2003/02/10 22:53:27  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.6  2001/03/20 01:36:46  dtracy
Lots of mods, mostly related to itulab colorspace for jpegs.  Decoder
now handles Scansoft-style tfx (g4 selector) and can handle stripwise
decodes of subsampled jpegs.

Revision 1.14  2001/03/12 23:02:26  dabercro
Use the TIFF-FX default gamut range.

Revision 1.13  2001/03/07 18:34:21  dabercro
Linear RGB replaced by sRGB.  A little code cleanup.

Revision 1.12  2001/01/19 23:59:42  dabercro
Remove editing mishap.
Remove unused function - both ICCLAB and CIELAB share the
same conversion, since the CIELAB was meant to do what
ICCLAB corrected (range was meant to be [-128,127] instead of
[-127,127].

Revision 1.11  2001/01/18 23:29:26  dabercro
Handle 1-component CIELAB/ICCLAB (gray).

Revision 1.10  2001/01/18 00:29:27  dabercro
Fix precision problem.  When quantizing to 8-bit values, I was
scaling to [0,255], rounding, then quantizing.  Problem was that
gave me 255 bins, and the first and last bin corresponded to only
half a pre-quantized value ([0,0.5] went to 0, [0.5,1.5] went to 1,
... [254.5,255.0] went to 255).  Now 256 evenly distributed bins.
Added support for encoding to a non-default gamut.
Added grayscale ITULAB support, and non-default gamut
support for this as well.

Revision 1.9  2000/09/21 21:24:22  dabercro
Avoid unnecessary function call.

Revision 1.8  2000/09/08 22:04:49  dabercro
Slight precision change, to avoid an overflow.

Revision 1.7  2000/09/06 01:28:44  dabercro
Huge optimizations for Lab color space conversion.  12x speed-up
for RGB->LAB, and 4x speed-up for LAB->RGB.

Revision 1.6  2000/08/01 21:46:58  dabercro
Handle grayscale for ITULAB also. Changes from Mark Tobin.

Revision 1.5  2000/04/07 23:59:42  dabercro
Added support for CIELAB and ICCLAB color spaces.

Revision 1.4  2000/02/26 01:28:37  dabercro
Avoid sometime predefined version of fabs().

Revision 1.3  2000/02/26 00:10:54  dabercro
Updated the ITULAB conversion formula, after discussions with Rob Buckley.

Revision 1.2  2000/02/24 00:48:23  dabercro
Changes to support ITULAB color conversion

Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.13  1998/01/13 23:50:42  ddavies
jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.12  1996/04/22 17:30:58  xzhu
xzhu on Mon Apr 22 10:29:55 PDT 1996

jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

Added cinfo = cinfo for no work function to keep compiler happy.

Revision 1.11  1996/04/20 01:33:39  xzhu
xzhu on Fri Apr 19 18:33:15 PDT 1996

jccolor.c -> /product/ipcore/ipshared/src/RCS/jccolor.c,v

new jpeg version

*/
