/*
 * jdcolor.c
 *
 * Copyright (C) 1991-1997, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains output colorspace conversion routines.
 *
 * Adding Dave Abercrombie's code for additional colorspace support -- 
 * specifically tfx's itulab.
 *
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeg6blib.h"
#include "mask_table.h"

#include <assert.h>
#include <math.h>

/* 8.8.2000 / dhtesq. */
/* Adding transparency mask support. */
#include "endianness.h"
/* 5.9.2000 / dhtesq. */
/* Beefing up the cmyk support. */
#include "image.h"

/* 22.8.2000 / dhtesq. */
/* Define these for masking.  They're undef'd at the end of the file. */
#define DGP_OPAQUE          0x00
#define DGP_TRANSPARENT     0xff

/* Private subobject */

typedef struct {
  struct jpeg_color_deconverter pub; /* public fields */

  /* Private state for YCC->RGB conversion */
  int * Cr_r_tab;		/* => table for Cr to R conversion */
  int * Cb_b_tab;		/* => table for Cb to B conversion */
  INT32 * Cr_g_tab;		/* => table for Cr to G conversion */
  INT32 * Cb_g_tab;		/* => table for Cb to G conversion */
  INT32 * lab_rgb_tab;		/* f_inv conversion table */
  UINT8 * srp_sr8_tab;		/* sR' to sR8 conversion table */
} my_color_deconverter;

typedef my_color_deconverter * my_cconvert_ptr;

METHODDEF(void)
itulab_rgb_convert_0 (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int	num_rows);

/* Just a horseshit color converter until we get Accucolor plugged in. */
static INLINE void
doLutInterpolation(DgpImage_RGB *rgb, DgpImage_CMYK *cmyk)
{
    cmyk->k = 0;
    cmyk->c = (rgb->b + rgb->g) / 2;
    cmyk->m = (rgb->b + rgb->r) / 2;
    cmyk->y = (rgb->r + rgb->g) / 2;
    return;
}

/**************** YCbCr -> RGB conversion: most common case **************/

/*
 * YCbCr is defined per CCIR 601-1, except that Cb and Cr are
 * normalized to the range 0..MAXJSAMPLE rather than -0.5 .. 0.5.
 * The conversion equations to be implemented are therefore
 *	R = Y                + 1.40200 * Cr
 *	G = Y - 0.34414 * Cb - 0.71414 * Cr
 *	B = Y + 1.77200 * Cb
 * where Cb and Cr represent the incoming values less CENTERJSAMPLE.
 * (These numbers are derived from TIFF 6.0 section 21, dated 3-June-92.)
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^16 (about 4 digits precision); we have to divide
 * the products by 2^16, with appropriate rounding, to get the correct answer.
 * Notice that Y, being an integral input, does not contribute any fraction
 * so it need not participate in the rounding.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants times Cb and Cr for all possible values.
 * For 8-bit JSAMPLEs this is very reasonable (only 256 entries per table);
 * for 12-bit samples it is still acceptable.  It's not very reasonable for
 * 16-bit samples, but if you want lossless storage you shouldn't be changing
 * colorspace anyway.
 * The Cr=>R and Cb=>B values can be rounded to integers in advance; the
 * values for the G calculation are left scaled up, since we must add them
 * together before rounding.
 */

#define SCALEBITS	16	/* speediest right-shift on some machines */
#define ONE_HALF	((INT32) 1 << (SCALEBITS-1))
#define FIX(x)		((INT32) ((x) * (1L<<SCALEBITS) + 0.5))

/*
 * Initialize tables for YCC->RGB colorspace conversion.
 */

LOCAL(void)
build_ycc_rgb_table (j_decompress_ptr cinfo)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  int i;
  INT32 x;
  SHIFT_TEMPS

  cconvert->Cr_r_tab = (int *)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				(MAXJSAMPLE+1) * SIZEOF(int));
  cconvert->Cb_b_tab = (int *)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				(MAXJSAMPLE+1) * SIZEOF(int));
  cconvert->Cr_g_tab = (INT32 *)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				(MAXJSAMPLE+1) * SIZEOF(INT32));
  cconvert->Cb_g_tab = (INT32 *)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				(MAXJSAMPLE+1) * SIZEOF(INT32));

  for (i = 0, x = -CENTERJSAMPLE; i <= MAXJSAMPLE; i++, x++) {
    /* i is the actual input pixel value, in the range 0..MAXJSAMPLE */
    /* The Cb or Cr value we are thinking of is x = i - CENTERJSAMPLE */
    /* Cr=>R value is nearest int to 1.40200 * x */
    cconvert->Cr_r_tab[i] = (int)
		    RIGHT_SHIFT(FIX(1.40200) * x + ONE_HALF, SCALEBITS);
    /* Cb=>B value is nearest int to 1.77200 * x */
    cconvert->Cb_b_tab[i] = (int)
		    RIGHT_SHIFT(FIX(1.77200) * x + ONE_HALF, SCALEBITS);
    /* Cr=>G value is scaled-up -0.71414 * x */
    cconvert->Cr_g_tab[i] = (- FIX(0.71414)) * x;
    /* Cb=>G value is scaled-up -0.34414 * x */
    /* We also add in ONE_HALF so that need not do it in inner loop */
    cconvert->Cb_g_tab[i] = (- FIX(0.34414)) * x + ONE_HALF;
  }
}

METHODDEF(void)
ycc_rgb_convert_180 (j_decompress_ptr cinfo,
		 JSAMPIMAGE input_buf, JDIMENSION input_row,
		 dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr,col;
  register JSAMPROW outptr;
  register JSAMPROW inptr0, inptr1, inptr2;
  register JDIMENSION input;
  JDIMENSION num_cols = cinfo->output_width;
  JDIMENSION in_cols = cinfo->image_width;
  static int check_cnt=0;
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);
  int r_val,g_val,b_val; int outptr_r;
  unsigned int copy_cnt=0;

  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask;
  unsigned long tr_col, tr_row_long_offset;
  unsigned char transparency;

SHIFT_TEMPS

  row_cnt = row_cnt;

  /*Top-bottom crop */
  if( ((int)(cinfo->output_scanline + cinfo->scale_num - 1) < destination.crop_coords.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_coords.end_y) ) {
    return;
  } 

  r_val = g_val = b_val = 0;
  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          
          /*Shift to the last column but one */
          outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * 3; 
          outptr = *(output_buf) + outptr_r;

          /*The pointer is already at the right hand corner of */
          /*the appropriate row - which is where it should be */
          
          for(col = 0, input = 0; input < in_cols; col += cinfo->scale_num, input++) {
              
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x) {
                  /*fprintf(stderr,"Pixels written is %d\n",check_cnt); */
                  return;
              }
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              
              outptr[RGB_RED] =  r_val = range_limit[y + Crrtab[cr]];
              outptr[RGB_GREEN] = g_val = range_limit[y + ((int) RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              outptr[RGB_BLUE] =  b_val = range_limit[y + Cbbtab[cb]];
              outptr -= RGB_PIXELSIZE; check_cnt++;

              if(cinfo->scale_num > 1){
				  Dgp_UInt32 numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (unsigned int)(destination.crop_coords.end_x - col + 1));

                  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
                      if((int)(col+copy_cnt) < destination.crop_coords.end_x){
                          outptr[RGB_RED] = r_val;
                          outptr[RGB_GREEN] = g_val;
                          outptr[RGB_BLUE] = b_val;
                          outptr -= RGB_PIXELSIZE;
                      }
                  }
              } 
              r_val = g_val = b_val = 0;
          } /*end of for(col...) */
      } /*end of while loop */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          
          /*Shift to the last column but one */
          outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * RGB_PIXELSIZE; 
          outptr = *(output_buf) + outptr_r;
          /* Find the right column for the transparency mask... */
          tr_col = MIN((int)(destination.crop_coords.end_x), (int)num_cols - 1) - MAX(destination.crop_coords.start_x, 0);
          tr_row_long_offset = (destination.crop_coords.end_y - destination.crop_coords.start_y) 
			  * (destination.trans_bpl / sizeof(*mask));
          mask = destination.trans_mask + tr_row_long_offset;

          /*The pointer is already at the right hand corner of */
          /*the appropriate row - which is where it should be */
          for (input = 0, col = 0; input < in_cols; input++, col += cinfo->scale_num) {
              
              /*In the normal orientation case, just don't write out what is not needed. */
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x){
                  return;
              }
              
              /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
              /* The masks coming out of the print driver seem to be big-endian by construction. */
              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              /* Range-limiting is essential due to noise introduced by DCT losses. */
			  r_val = range_limit[y + Crrtab[cr]];
			  g_val = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
			  b_val = range_limit[y + Cbbtab[cb]];
              outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
              outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
              outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
              outptr -= RGB_PIXELSIZE;
              tr_col--;

              if(cinfo->scale_num > 1){
				  int j;
				  unsigned int startCol = tr_col + 1;
				  JSAMPROW startOutptr = outptr + RGB_PIXELSIZE;
				  Dgp_UInt32 numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  Dgp_UInt32 numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (unsigned int)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, destination.crop_coords.end_y - cinfo->output_scanline + 1);

				  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
		              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
		              tr_col--;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
					  outptr -= RGB_PIXELSIZE;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < (int)numToCopyVert; j++){
					  tr_col = startCol;
					  outptr = startOutptr - (j * in_cols * cinfo->scale_num * RGB_PIXELSIZE);
					  for(copy_cnt = 0; copy_cnt < numToCopy; copy_cnt++){
						  transparency = (mask[-(int)(j * destination.trans_bpl / sizeof(*mask)) + (tr_col / 32)] 
							  >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
						  tr_col--;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
						  outptr -= RGB_PIXELSIZE;
					  }					  
				  }
				  outptr = startOutptr - (numToCopy * RGB_PIXELSIZE);
              } 
              r_val = g_val = b_val = 0;
          } /*end of for */
      } /*end of while loop */
  } /* end of if... */
  return;
} /*end of function ycc_rgb_convert_180 */

METHODDEF(void)
ycc_rgb_convert_90 (j_decompress_ptr cinfo,
		 JSAMPIMAGE input_buf, JDIMENSION input_row,
		 dgp_dest_buf dest_buf, int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr;
  register JSAMPROW outptr;
  register JSAMPROW inptr0, inptr1, inptr2;
  register JDIMENSION input, col;
  /*JDIMENSION num_cols = cinfo->output_width; */
  JDIMENSION in_cols = cinfo->image_width;

  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);
  int r_val, g_val, b_val; 
  unsigned int copy_cnt=0, bpr = 0;
  /*static int pixels_written=0; */

  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask;
  unsigned long tr_col;
  unsigned char transparency;
  unsigned long tLpl = destination.trans_bpl / sizeof(*mask);

  SHIFT_TEMPS

  row_cnt = row_cnt;

  /*Top-bottom crop */
  if( ((int)(cinfo->output_scanline + cinfo->scale_num - 1) < destination.crop_coords.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_coords.end_y) ) {
    return;
  }

  r_val = g_val = b_val = 0;
  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  bpr = destination.bytes_per_row;

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;

          outptr = *(output_buf);

          for (col = 0, input = 0; input < in_cols; col += cinfo->scale_num, input++) {
              
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x){
                  return;
              }
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              outptr[RGB_RED] =  r_val = range_limit[y + Crrtab[cr]];
              outptr[RGB_GREEN] = g_val = range_limit[y + ((int) RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              outptr[RGB_BLUE] =  b_val = range_limit[y + Cbbtab[cb]];
              
              outptr += bpr; /*Remember that bpr includes the number of output components - 3, in this case */

              if(cinfo->scale_num > 1){
				  Dgp_UInt32 numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, destination.crop_coords.end_x - col + 1);
                  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
					  outptr[RGB_RED] = r_val;
					  outptr[RGB_GREEN] = g_val;
					  outptr[RGB_BLUE] = b_val;
					  outptr += bpr;
                  } /*end of the for loop */
              }
              r_val = g_val = b_val = 0;
          } /*end of for */
      } /*end of while loop */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          
          outptr = *(output_buf);
          
          /* Get to the right row of the transparency mask... */
          tr_col = destination.crop_coords.end_y - cinfo->output_scanline;
          mask = destination.trans_mask;

          for(input = 0, col = 0; input < in_cols; input++, col += cinfo->scale_num) {
              
              /*In the normal orientation case, just don't write out what is not needed. */
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x){
                  return;
              }
              
              /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
              /* The masks coming out of the print driver seem to be big-endian by construction. */
              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
			  r_val = range_limit[y + Crrtab[cr]];
			  g_val = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
			  b_val = range_limit[y + Cbbtab[cb]];
              outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
              outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
              outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
              outptr += bpr;
			  mask += tLpl;

              if(cinfo->scale_num > 1){
				  unsigned int j;
				  unsigned long *startMask = mask - tLpl;
				  JSAMPROW startOutptr = outptr - bpr;
				  Dgp_UInt32 numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  Dgp_UInt32 numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (Dgp_UInt32)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, destination.crop_coords.end_y - cinfo->output_scanline + 1);

				  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
		              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
		              mask += tLpl;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
					  outptr += bpr;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < numToCopyVert; j++){
					  mask = startMask;
					  outptr = startOutptr - (j * RGB_PIXELSIZE);
					  for(copy_cnt = 0; copy_cnt < numToCopy; copy_cnt++){
			              transparency = (mask[(tr_col - j) / 32] >> (31 - ((tr_col - j) % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
			              mask += tLpl;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
						  outptr += bpr;
					  }					  
				  }
				  mask = startMask + (numToCopy * tLpl);
				  outptr = startOutptr + (numToCopy * bpr);
              } 
              r_val = g_val = b_val = 0;
          } /*end of for */
      } /*end of while loop */
  } /* end of if... */
  return;
} /*end of function ycc_rgb_convert_90 */

METHODDEF(void)
ycc_rgb_convert_270 (j_decompress_ptr cinfo,
		 JSAMPIMAGE input_buf, JDIMENSION input_row,
		 dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr,col;
  register JSAMPROW outptr;
  register JSAMPROW inptr0, inptr1, inptr2;
  register JDIMENSION input;
  JDIMENSION num_cols = cinfo->output_width;
  JDIMENSION in_cols = cinfo->image_width;

  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);
  int r_val,g_val,b_val; int outptr_r;
  unsigned int copy_cnt=0, bpr = 0;

  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask;
  unsigned long tr_row_long_offset;
  unsigned long tr_col;
  unsigned char transparency;
  unsigned long tLpl = destination.trans_bpl / sizeof(*mask);

  SHIFT_TEMPS

  row_cnt = row_cnt;

  /*Top-bottom crop */
  if( ((int)(cinfo->output_scanline + cinfo->scale_num - 1) < destination.crop_coords.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_coords.end_y) ) {
    return;
  } 

  r_val = g_val = b_val = 0;
  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  bpr = destination.bytes_per_row;

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          
          outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * bpr;
          *(output_buf) += outptr_r;
          outptr = *(output_buf);
          
          for(input = 0, col = 0; input < in_cols; input++, col += cinfo->scale_num) {
                            
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x)
                  return;

              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              
              outptr[RGB_RED] =  r_val = range_limit[y + Crrtab[cr]];
              outptr[RGB_GREEN] = g_val = range_limit[y + ((int) RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              outptr[RGB_BLUE] =  b_val = range_limit[y + Cbbtab[cb]];
              outptr -= bpr;

              if(cinfo->scale_num > 1){
				  Dgp_UInt32 numToCopy = col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (Dgp_UInt32)(destination.crop_coords.end_x - col + 1));
                  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
					  outptr[RGB_RED] = r_val;
					  outptr[RGB_GREEN] = g_val;
					  outptr[RGB_BLUE] = b_val;
					  outptr -= bpr;
                  }
              } 
              r_val = g_val = b_val = 0;
          } /*end of for */
      } /*end of while loop */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          
          outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * bpr;
          *(output_buf) += outptr_r;
          outptr = *(output_buf);
          
          /* Get to the right row of the transparency mask... */
          tr_row_long_offset = (MIN((int)(num_cols - 1), destination.crop_coords.end_x) - 
			  MAX(0, destination.crop_coords.start_x)) * tLpl;
          tr_col = cinfo->output_scanline - destination.crop_coords.start_y;
          mask = destination.trans_mask + tr_row_long_offset;

          for(input = 0, col = 0; input < in_cols; input++, col += cinfo->scale_num) {
              
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x){
                  return;
              }
              
              /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
              /* The masks coming out of the print driver seem to be big-endian by construction. */
              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              /* Range-limiting is essential due to noise introduced by DCT losses. */
			  r_val = range_limit[y + Crrtab[cr]];
			  g_val = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
			  b_val = range_limit[y + Cbbtab[cb]];
              outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
              outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
              outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
              outptr -= bpr;
              mask -= tLpl;

              if(cinfo->scale_num > 1){
				  unsigned int j;
				  unsigned long *startMask = mask + tLpl;
				  JSAMPROW startOutptr = outptr + bpr;
				  Dgp_UInt32 numToCopy = col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  Dgp_UInt32 numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (Dgp_UInt32)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, destination.crop_coords.end_y - cinfo->output_scanline + 1);

				  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
		              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
		              mask -= tLpl;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
					  outptr -= bpr;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < numToCopyVert; j++){
					  mask = startMask;
					  outptr = startOutptr + (j * RGB_PIXELSIZE);
					  for(copy_cnt = 0; copy_cnt < numToCopy; copy_cnt++){
			              transparency = (mask[(tr_col + j) / 32] >> (31 - ((tr_col + j) % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
			              mask -= tLpl;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (r_val & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
						  outptr -= bpr;
					  }					  
				  }
				  mask = startMask - (numToCopy * tLpl);
				  outptr = startOutptr - (numToCopy * bpr);
              } 
              r_val = g_val = b_val = 0;
          } /*end of for */
      } /*end of while loop */
  } /* end of if... */
  return;
} /*end of function ycc_rgb_convert_270 */


/*
 * Convert some rows of samples to the output colorspace.
 *
 * Note that we change from noninterleaved, one-plane-per-component format
 * to interleaved-pixel format.  The output buffer is therefore three times
 * as wide as the input buffer.
 * A starting row offset is provided only for the input buffer.  The caller
 * can easily adjust the passed output_buf value to accommodate any row
 * offset required on that side.
 *
 * 8.8.2000 / dhtesq.
 * Implementing a 1-bpp mask, arriving here as a void * in the dest_buf.
 * No guarantees as to what this will do to performance.
 *
 */

/*METHODDEF(void)
ycc_rgb_convert (j_decompress_ptr cinfo,
		 JSAMPIMAGE input_buf, JDIMENSION input_row,
		 JSAMPARRAY output_buf, int num_rows) 10/12/99 */

METHODDEF(void)
ycc_rgb_convert (j_decompress_ptr cinfo,
		 JSAMPIMAGE input_buf, JDIMENSION input_row,
		 dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr;
  register JSAMPROW outptr;
  register JSAMPROW inptr0, inptr1, inptr2;
  register JDIMENSION col,input;
  /*JDIMENSION num_cols = cinfo->output_width; */
  JDIMENSION in_cols = cinfo->image_width;

  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);
  register Dgp_UInt8 r_val,g_val,b_val;
  /*static int check_cnt=0; */
  unsigned int copy_cnt=0;

  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask = destination.trans_mask;

  unsigned char transparency;
  Dgp_UInt32 dstRGBpl,  maskWpl;
  Dgp_UInt32 col_start, col_end, input_start;

  SHIFT_TEMPS
  row_cnt = row_cnt;

#define YCBCR_TO_RGB(input, r,g,b)      y  = GETJSAMPLE(inptr0[(input)]); \
                                        cb = GETJSAMPLE(inptr1[(input)]); \
                                        cr = GETJSAMPLE(inptr2[(input)]); \
                                        r = range_limit[y + Crrtab[cr]]; \
                                        g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))]; \
                                        b = range_limit[y + Cbbtab[cb]]; 
  /*Top-bottom crop */
  if( ((int)(cinfo->output_scanline + cinfo->scale_num - 1) < destination.crop_coords.start_y) ||
      ((int)(cinfo->output_scanline) > destination.crop_coords.end_y) ) {
    return;
  } 
 
  /* destination byte perl line */
  dstRGBpl = in_cols * cinfo->scale_num * RGB_PIXELSIZE;

  /* selector layer DWORD per line */
  maskWpl = destination.trans_bpl / sizeof(Dgp_UInt32);

  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  
  /* destination end postion croped by start_x. */
  col_end = destination.crop_coords.end_x;

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
	  Dgp_UInt32 numToCopy;
	  while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          outptr = *output_buf; 
          for(col = 0, input = 0; input < in_cols; col+= cinfo->scale_num, input++) {
              /* In the normal orientation case, just don't write out what is not needed. */
              if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
              if((int)col > destination.crop_coords.end_x){
                  /*fprintf(stderr,"HOHO chk_cnt = %d\n",check_cnt); */
                  return;
              }
			  YCBCR_TO_RGB(input, r_val, g_val, b_val);
			  /*switch from rgb to bgr if asked so, need to do this for all color space
 */
			  /* and angles---- FIXME
 */
			  if(cinfo->BGR) {	
				  outptr[RGB_RED] = b_val;
				  outptr[RGB_GREEN] = g_val;
				  outptr[RGB_BLUE] = r_val;
			  }else{
				  outptr[RGB_RED] = r_val;
				  outptr[RGB_GREEN] = g_val;
				  outptr[RGB_BLUE] = b_val;
			  }
              outptr += RGB_PIXELSIZE;
              if(cinfo->scale_num >1){
                  numToCopy = (int)col < destination.crop_coords.start_x ?
                      cinfo->scale_num - (destination.crop_coords.start_x - col) : cinfo->scale_num;
                  numToCopy = MIN(numToCopy, destination.crop_coords.end_x - col + 1);
                  for(copy_cnt = 1; copy_cnt < numToCopy; copy_cnt++){
					  if(cinfo->BGR) {	/*switch from rgb to bgr if asked so
 */
						  outptr[RGB_RED] = b_val;
						  outptr[RGB_GREEN] = g_val;
						  outptr[RGB_BLUE] = r_val;
					  }else{
						  outptr[RGB_RED] = r_val;
						  outptr[RGB_GREEN] = g_val;
						  outptr[RGB_BLUE] = b_val;
					  }
                      outptr += RGB_PIXELSIZE;
                  } 
              }
          } /*end of for(col...) */
          *output_buf += dstRGBpl;
      } /*end of while loop  */
  } else {
      /* We have a mask...  */
      register int j, col2;  
      register Dgp_UInt8  numToCopyVert;
      register Dgp_UInt32 maskvalue;
      register JSAMPROW tempptr;
      register Dgp_UInt32 wStart, shift;
      register Dgp_UInt32 in_cols_aligned, col_end_aligned;
      register Dgp_UInt8 r_val2, g_val2, b_val2;             
      Dgp_UInt8 r_vall[4], g_vall[4],b_vall[4];

      /* no matter if the scale number is 1, 2 or 4, we always read 4 bit each time */
      /* from the selector and look up the pixel assignment in a table of functions. */
      /* for scale number 1, pass 4 color pixels each time.( 1 bg pixel - 1 selector bit) */
      /* for scale number 2, pass 2 color pixels each time ( 1 bg pixel - 2 selector bit) */
      /* for scale number 4, pass 1 color pixels each time ( 1 bg pixel - 4 selector bit) */
      /* for scale 1, 2, the remaining pixels each line will be handled seperately. */
      /* for scale number 3, we use the tradional way of assigning pixel based on each slector bit */
      


      /* in_cols_aligned and col_end_aligned is 4 pixel boundary */
      switch (cinfo->scale_num){
      case 2:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* 2 input match 4 output */
          in_cols_aligned =  in_cols - in_cols % 2;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          col_end_aligned =  col_end - (destination.crop_coords.end_x + 1) % 4  ;
          break;
      case 1:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* 4 input match 4 output */
          in_cols_aligned = in_cols - in_cols % 4;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          col_end_aligned =  col_end - (destination.crop_coords.end_x + 1) % 4  ;
          break;
      case 4:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* 1 input match 4 output */
          in_cols_aligned = in_cols;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          /* for scale num = 4, there should be no left over pixels */
          col_end_aligned = col_end / 4 * 4  + 3;
          break;
      default: /* for other scale factor use the old method. */
          /* dest and src start position croped by start_x  */
          col_start = destination.crop_coords.start_x;
          input_start = destination.crop_coords.start_x / cinfo->scale_num;
          break;
          
      }
      /* crop destination by start_y and end_y.  */
	  numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
		  cinfo->output_scanline - destination.crop_coords.start_y + cinfo->scale_num:cinfo->scale_num;
	  numToCopyVert = MIN(numToCopyVert, destination.crop_coords.end_y - cinfo->output_scanline + 1);
      /* in this case num_rows is usually just 1. This while loop is not useful. Also if num_rows is */
      /* not one than the numToCopyVert should be calculated for each row. */
      while (num_rows--) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          outptr = *output_buf + col_start * RGB_PIXELSIZE;
          switch(cinfo->scale_num){
          case 4:
              for(col = col_start, input = input_start; input < in_cols_aligned, col <= col_end_aligned; col +=4, input++) {
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
              
                  YCBCR_TO_RGB(input, r_val, g_val, b_val);
              
                  tempptr = outptr;
                  wStart = col / 32;
                  shift  = 28 - col % 32; 
                  for(j = 0; j < numToCopyVert; j++){
                      maskvalue = (mask[wStart + maskWpl * j] >> shift) & 0x0f;
					  if(cinfo->BGR)
						pixel_assign4[maskvalue](tempptr, b_val, g_val, r_val);  
					  else
						pixel_assign4[maskvalue](tempptr, r_val, g_val, b_val);  
                      tempptr += dstRGBpl;
                  }
                  outptr += 4 * RGB_PIXELSIZE;
              } /*end of for(col...) */
              break;
          case 2:
              for(col = col_start, input = input_start; input < in_cols_aligned, col<= col_end_aligned; col +=4, input++) {
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
                  YCBCR_TO_RGB(input, r_val, g_val, b_val);
                  input ++;
                  YCBCR_TO_RGB(input, r_val2, g_val2, b_val2);
                  tempptr = outptr;
                  wStart = col / 32;
                  shift  = 28 - col % 32; 
                  for(j = 0; j < numToCopyVert; j++){
                      maskvalue = (mask[wStart + maskWpl * j] >> shift) & 0x0f;
					  if(cinfo->BGR)
						pixel_assign2[maskvalue](tempptr, b_val, g_val, r_val, b_val2, g_val2, r_val2);  
					  else
						pixel_assign2[maskvalue](tempptr, r_val, g_val, b_val, r_val2, g_val2, b_val2);  
                      tempptr += dstRGBpl;
                  }
                  outptr += 4 * RGB_PIXELSIZE;
              }/* end of for(col.. */
              break;
          case 1:
              for(col = col_start, input = input_start; input < in_cols_aligned, col<= col_end_aligned; col +=4, input++) {
              
                  /* convert YcBcr to RGB */
                  YCBCR_TO_RGB(input, r_vall[0], g_vall[0], b_vall[0]);
                  input ++;
                  YCBCR_TO_RGB(input, r_vall[1], g_vall[1], b_vall[1]);
                  input ++;
                  YCBCR_TO_RGB(input, r_vall[2], g_vall[2], b_vall[2]);
                  input ++;
                  YCBCR_TO_RGB(input, r_vall[3], g_vall[3], b_vall[3]);
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
                  maskvalue = (mask[ col / 32] >> (28 - col % 32)) & 0x0f;
				  if(cinfo->BGR)
					  pixel_assign1[maskvalue](outptr, b_vall, g_vall, r_vall); 
				  else
					  pixel_assign1[maskvalue](outptr, r_vall, g_vall, b_vall); 
                  outptr += 4 * RGB_PIXELSIZE;
              }/* end of for(col.. */
              break;
          default:
              col = col_start;
              input = input_start;
              break;
          }/*end of switch(cinfo_scale_num) */

          /* for remaining columns that is not dword aligned. or for the other scalefactors that are not 1.2 or4 */
          /* we use the old method of layer merging one bit at a time without table look up. */
          while(input < in_cols && col <= col_end) {
              Dgp_UInt32 numToCopy = cinfo->scale_num - (int)col % cinfo->scale_num;
              numToCopy = MIN(numToCopy, destination.crop_coords.end_x - col + 1);
              /* get maskvalue in unit of 32 bit */
              /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
              /* The masks coming out of the print driver seem to be big-endian by construction. */
              YCBCR_TO_RGB(input, r_val, g_val, b_val);  
              for(j = 0; j < numToCopyVert; j++){
                  col2 = col;
                  tempptr = outptr + j * dstRGBpl;
                  for(copy_cnt = 0; copy_cnt < numToCopy; copy_cnt++){
                      transparency = (mask[(j * destination.trans_bpl / sizeof(Dgp_UInt32)) + (col2 / 32)] 
                          >> (31 - (col2 % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
                      col2++;
					  if(cinfo->BGR){
						  tempptr[RGB_RED] = (tempptr[RGB_RED] & (~transparency)) | (b_val & transparency);
						  tempptr[RGB_GREEN] = (tempptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  tempptr[RGB_BLUE] = (tempptr[RGB_BLUE] & (~transparency)) | (r_val & transparency);
					  }else{
						  tempptr[RGB_RED] = (tempptr[RGB_RED] & (~transparency)) | (r_val & transparency);
						  tempptr[RGB_GREEN] = (tempptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  tempptr[RGB_BLUE] = (tempptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
					  }
                      tempptr += RGB_PIXELSIZE;
                  }					  
              }
              outptr += numToCopy * RGB_PIXELSIZE;
              input++;
              col += numToCopy; 
                  
          } /*end of while(input<in_cols... */
          *output_buf += dstRGBpl;
      } /*end of while loop */
  } /* end of if(mask == NULL)...  */

  return;
}/*end of function ycc_rgb_convert */


/**************** Cases other than YCbCr -> RGB **************/


/*
 * Color conversion for no colorspace change: just copy the data,
 * converting from separate-planes to interleaved representation.
 */

/* 10/12/99 METHODDEF(void)
null_convert (j_decompress_ptr cinfo,
	      JSAMPIMAGE input_buf, JDIMENSION input_row,
	      JSAMPARRAY output_buf, int num_rows)*/
METHODDEF(void)
null_convert (j_decompress_ptr cinfo,
	      JSAMPIMAGE input_buf, JDIMENSION input_row,
	      dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  register JSAMPROW inptr, outptr;
  register JDIMENSION count;
  register int num_components = cinfo->num_components;
  JDIMENSION num_cols = cinfo->output_width;
  int ci;
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);

  row_cnt = row_cnt;

  while (--num_rows >= 0) {
    for (ci = 0; ci < num_components; ci++) {
      inptr = input_buf[ci][input_row];
      outptr = output_buf[0] + ci;
      for (count = num_cols; count > 0; count--) {
	*outptr = *inptr++;	/* needn't bother with GETJSAMPLE() here */
	outptr += num_components;
      }
    }
    input_row++;
    output_buf++;
  }
}


/*
 * Color conversion for grayscale: just copy the data.
 * This also works for YCbCr -> grayscale conversion, in which
 * we just copy the Y (luminance) component and ignore chrominance.
 */


METHODDEF(void)
/*grayscale_convert (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   JSAMPARRAY output_buf, int num_rows) 10/12/99 */
grayscale_convert (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   dgp_dest_buf dest_buf,int row_cnt, int num_rows)

{
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);                      
  row_cnt = row_cnt;

  jcopy_sample_rows(input_buf[0], (int)input_row, (JSAMPARRAY)output_buf, 0,
		    num_rows, cinfo->output_width);
}


/*
 * Convert grayscale to RGB: just duplicate the graylevel three times.
 * This is provided to support applications that don't want to cope
 * with grayscale as a separate case.
 */

/*METHODDEF(void)
gray_rgb_convert (j_decompress_ptr cinfo,
		  JSAMPIMAGE input_buf, JDIMENSION input_row,
		  JSAMPARRAY output_buf, int num_rows) 10/12/99 */
METHODDEF(void)
gray_rgb_convert (j_decompress_ptr cinfo,
		  JSAMPIMAGE input_buf, JDIMENSION input_row,
		  dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  register JSAMPROW inptr, outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->output_width;
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);

  row_cnt = row_cnt;

  while (--num_rows >= 0) {
    inptr = input_buf[0][input_row++];
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++) {
      /* We can dispense with GETJSAMPLE() here */
      outptr[RGB_RED] = outptr[RGB_GREEN] = outptr[RGB_BLUE] = inptr[col];
      outptr += RGB_PIXELSIZE;
    }
  }
}


/*
 * Adobe-style YCCK->CMYK conversion.
 * We convert YCbCr to R=1-C, G=1-M, and B=1-Y using the same
 * conversion as above, while passing K (black) unchanged.
 * We assume build_ycc_rgb_table has been called.
 */

/*METHODDEF(void)
ycck_cmyk_convert (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   JSAMPARRAY output_buf, int num_rows) 10/12/99 */

#if 0
METHODDEF(void)
ycck_cmyk_convert (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr;
  register JSAMPROW outptr;
  register JSAMPROW inptr0, inptr1, inptr2, inptr3;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->output_width;
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);

  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;
  SHIFT_TEMPS

  row_cnt = row_cnt;

  while (--num_rows >= 0) {
    inptr0 = input_buf[0][input_row];
    inptr1 = input_buf[1][input_row];
    inptr2 = input_buf[2][input_row];
    inptr3 = input_buf[3][input_row];
    input_row++;
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++) {
      y  = GETJSAMPLE(inptr0[col]);
      cb = GETJSAMPLE(inptr1[col]);
      cr = GETJSAMPLE(inptr2[col]);
      /* Range-limiting is essential due to noise introduced by DCT losses. */
      outptr[0] = range_limit[MAXJSAMPLE - (y + Crrtab[cr])];	/* red */
      outptr[1] = range_limit[MAXJSAMPLE - (y +			/* green */
			      ((int) RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr],
						 SCALEBITS)))];
      outptr[2] = range_limit[MAXJSAMPLE - (y + Cbbtab[cb])];	/* blue */
      /* K passes through unchanged */
      outptr[3] = inptr3[col];	/* don't need GETJSAMPLE here */
      outptr += 4;
    }
  }
}
#endif

#if 0
METHODDEF(int)
get_min_cmy(int c, int m,int y){

    if( (c <= m) && (c <= y))
      return c;
    else if( (m <= c) && (m <= y))
      return m;
    else /*if( (y <= c) && (y <= m)) ...dhtesq. */
      return y;
}
#endif

METHODDEF(void)
ycbcr_ycck_cmyk_convert_270 (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr, bpr, height_val, index_value; 
  register JSAMPROW inptr0, inptr1, inptr2, inptr3;
  register JDIMENSION col,input;
  JDIMENSION num_cols = cinfo->output_width;
  JDIMENSION in_cols = cinfo->image_width;
  /*int min_cmy = 0;  */
  /*int count=0;  */
  unsigned int copy_cnt=0;
  dgp_cmyk_dest_buf destination = dest_buf.cmyk_buf;
  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask;
  unsigned long tr_row_long_offset;
  unsigned long tr_col;
  unsigned char transparency;
  DgpImage_RGB          rgb, last_rgb;
  DgpImage_CMYK         cmyk, last_cmyk;

  SHIFT_TEMPS

  Cbbtab = Cbbtab;
  row_cnt = row_cnt;

  /* Prepare first color;  guess black. */
  last_rgb.r = 0;   last_rgb.g = 0;     last_rgb.b = 0;
  last_cmyk.c = 0;  last_cmyk.m = 0;    last_cmyk.y = 0;    last_cmyk.k = 255;

  /*Top-bottom crop */
  if( ((int)cinfo->output_scanline < destination.crop_vals.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_vals.end_y) ) {
    return;
  }

  bpr = dest_buf.cmyk_buf.bpr;
  height_val = destination.crop_vals.end_x;

  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
          /*for(input = 0, col = 0; input < in_cols, col < cinfo->output_width; input++, col += cinfo->scale_num) { */
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col > destination.crop_vals.end_x) return; */

              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;

              switch(destination.imgType){
              case DGP_CMYK:
                  index_value = (height_val - col)*bpr;
                  destination.c[index_value] = cmyk.c;
                  destination.m[index_value] = cmyk.m;
                  destination.y[index_value] = cmyk.y;
                  destination.k[index_value] = cmyk.k;
                  
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.c[index_value] = cmyk.c;
                          destination.m[index_value] = cmyk.m;
                          destination.y[index_value] = cmyk.y;
                          destination.k[index_value] = cmyk.k;
                      }
                  }
                  break;
              case DGP_C:
                  index_value = (height_val - col)*bpr;
                  destination.c[index_value] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.c[index_value] = cmyk.c;
                      }
                  }
                  break;
              case DGP_M:
                  index_value = (height_val - col)*bpr;
                  destination.m[index_value] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.m[index_value] = cmyk.m;
                      }
                  }
                  break;
              case DGP_Y:
                  index_value = (height_val - col)*bpr;
                  destination.y[index_value] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.y[index_value] = cmyk.y;
                      }
                  }
                  break;
              case DGP_K:
                  index_value = (height_val - col)*bpr;
                  destination.k[index_value] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.k[index_value] = cmyk.k;
                      }
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          /* Get to the right row of the transparency mask... */
          tr_row_long_offset = (MIN((int)(num_cols - 1), destination.crop_vals.end_x) - 
                               MAX(0, destination.crop_vals.start_x)) * 
                               (destination.trans_bpl / sizeof(*mask));
          tr_col = cinfo->output_scanline - destination.crop_vals.start_y;
          mask = destination.trans_mask + tr_row_long_offset;

          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++, mask -= (destination.trans_bpl / sizeof(*mask))) {
          /*for(input = 0, col = 0; input < in_cols, col < cinfo->output_width; input++, col += cinfo->scale_num) { */
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col > destination.crop_vals.end_x) return; */

              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              /*mask -= (destination.trans_bpl / sizeof(*mask)); */
              if(transparency == DGP_OPAQUE) continue;

              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;

              switch(destination.imgType){
              case DGP_CMYK:
                  index_value = (height_val - col)*bpr;
                  destination.c[index_value] = cmyk.c;
                  destination.m[index_value] = cmyk.m;
                  destination.y[index_value] = cmyk.y;
                  destination.k[index_value] = cmyk.k;
                  
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.c[index_value] = cmyk.c;
                          destination.m[index_value] = cmyk.m;
                          destination.y[index_value] = cmyk.y;
                          destination.k[index_value] = cmyk.k;
                      }
                  }
                  break;
              case DGP_C:
                  index_value = (height_val - col)*bpr;
                  destination.c[index_value] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.c[index_value] = cmyk.c;
                      }
                  }
                  break;
              case DGP_M:
                  index_value = (height_val - col)*bpr;
                  destination.m[index_value] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.m[index_value] = cmyk.m;
                      }
                  }
                  break;
              case DGP_Y:
                  index_value = (height_val - col)*bpr;
                  destination.y[index_value] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.y[index_value] = cmyk.y;
                      }
                  }
                  break;
              case DGP_K:
                  index_value = (height_val - col)*bpr;
                  destination.k[index_value] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          index_value = (height_val - col - copy_cnt)*bpr;
                          destination.k[index_value] = cmyk.k;
                      }
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } /* end of if(mask == NULL)... */
  return;
} /*end of function ycbcr_ycck_cmyk_convert_270 */

METHODDEF(void)
ycbcr_ycck_cmyk_convert_90 (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr, bpr, sub_val;
  register JSAMPROW inptr0, inptr1, inptr2,inptr3;
  register JDIMENSION col,input;
  /*JDIMENSION num_cols = cinfo->output_width; */
  JDIMENSION in_cols = cinfo->image_width;
  /*int min_cmy = 0; */
  /*int count = 0;  */
  unsigned int copy_cnt=0;
  dgp_cmyk_dest_buf destination = dest_buf.cmyk_buf;
  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long *mask;
  unsigned long tr_col;
  unsigned char transparency;
  DgpImage_RGB          rgb, last_rgb;
  DgpImage_CMYK         cmyk, last_cmyk;

  SHIFT_TEMPS

  row_cnt = row_cnt;
  Cbbtab = Cbbtab;

  /* Prepare first color;  guess black. */
  last_rgb.r = 0;   last_rgb.g = 0;     last_rgb.b = 0;
  last_cmyk.c = 0;  last_cmyk.m = 0;    last_cmyk.y = 0;    last_cmyk.k = 255;

  /*Top-bottom crop */
  if( ((int)cinfo->output_scanline < destination.crop_vals.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_vals.end_y) ) {
    return;
  }

  bpr = dest_buf.cmyk_buf.bpr;
  sub_val = destination.crop_vals.start_x;

  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
          /*for (col = 0,input=0; col < (cinfo->output_width),input < in_cols; col+=cinfo->scale_num, input++) { */
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col > destination.crop_vals.end_x) return; */

              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;
      
              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[(col-sub_val)*bpr] = cmyk.c;
                  destination.m[(col-sub_val)*bpr] = cmyk.m;
                  destination.y[(col-sub_val)*bpr] = cmyk.y;
                  destination.k[(col-sub_val)*bpr] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.c[(col-sub_val+copy_cnt)*bpr] = cmyk.c;
                              destination.m[(col-sub_val+copy_cnt)*bpr] = cmyk.m;
                              destination.y[(col-sub_val+copy_cnt)*bpr] = cmyk.y;
                              destination.k[(col-sub_val+copy_cnt)*bpr] = cmyk.k;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_C:
                  destination.c[(col-sub_val)*bpr] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.c[(col-sub_val+copy_cnt)*bpr] = cmyk.c;
                          } else {
                              return;
                          }
                      } 
                  }
                  break;
              case DGP_M:
                  destination.m[(col-sub_val)*bpr] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.m[(col-sub_val+copy_cnt)*bpr] = cmyk.m;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_Y:
                  destination.y[(col-sub_val)*bpr] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.y[(col-sub_val+copy_cnt)*bpr] = cmyk.y;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_K:
                  destination.k[(col-sub_val)*bpr] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.k[(col-sub_val+copy_cnt)*bpr] = cmyk.k;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
         
          /* Get to the right row of the transparency mask...               */
          tr_col = destination.crop_vals.end_y - cinfo->output_scanline;
          mask = destination.trans_mask;

          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++, mask += (destination.trans_bpl / sizeof(*mask))) {
          /*for(col = 0, input = 0; col < (cinfo->output_width), input < in_cols; col += cinfo->scale_num, input++) { */
              
              if((int)col < destination.crop_vals.start_x) continue;
              if((int)col > destination.crop_vals.end_x) return;

              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              /*mask += (destination.trans_bpl / sizeof(*mask)); */
              if(transparency == DGP_OPAQUE) continue;

              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;
      
              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[(col-sub_val)*bpr] = cmyk.c;
                  destination.m[(col-sub_val)*bpr] = cmyk.m;
                  destination.y[(col-sub_val)*bpr] = cmyk.y;
                  destination.k[(col-sub_val)*bpr] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.c[(col-sub_val+copy_cnt)*bpr] = cmyk.c;
                              destination.m[(col-sub_val+copy_cnt)*bpr] = cmyk.m;
                              destination.y[(col-sub_val+copy_cnt)*bpr] = cmyk.y;
                              destination.k[(col-sub_val+copy_cnt)*bpr] = cmyk.k;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_C:
                  destination.c[(col-sub_val)*bpr] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.c[(col-sub_val+copy_cnt)*bpr] = cmyk.c;
                          } else {
                              return;
                          }
                      } 
                  }
                  break;
              case DGP_M:
                  destination.m[(col-sub_val)*bpr] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.m[(col-sub_val+copy_cnt)*bpr] = cmyk.m;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_Y:
                  destination.y[(col-sub_val)*bpr] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.y[(col-sub_val+copy_cnt)*bpr] = cmyk.y;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              case DGP_K:
                  destination.k[(col-sub_val)*bpr] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          if((int)(col+copy_cnt) < destination.crop_vals.end_x){
                              destination.k[(col-sub_val+copy_cnt)*bpr] = cmyk.k;
                          } else {
                              return;
                          }
                      }
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  }/*end of if(mask == NULL)... */
} /*end of function ycbcr_ycck_cmyk_convert_90 */

METHODDEF(void)
ycbcr_ycck_cmyk_convert_180 (j_decompress_ptr cinfo,
		   JSAMPIMAGE input_buf, JDIMENSION input_row,
		   dgp_dest_buf dest_buf,int row_cnt, int num_rows)
{
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int y, cb, cr, width_val;
  register JSAMPROW inptr0, inptr1, inptr2,inptr3;
  register JDIMENSION col, input;
  JDIMENSION num_cols = cinfo->output_width;
  JDIMENSION in_cols = cinfo->image_width;
  /*int min_cmy = 0;  */
  /*int count = 0;  */
  unsigned int copy_cnt = 0;
  dgp_cmyk_dest_buf destination = dest_buf.cmyk_buf;
  /* copy these pointers into registers if possible */
  register JSAMPLE * range_limit = cinfo->sample_range_limit;
  register int * Crrtab = cconvert->Cr_r_tab;
  register int * Cbbtab = cconvert->Cb_b_tab;
  register INT32 * Crgtab = cconvert->Cr_g_tab;
  register INT32 * Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long         *mask;
  unsigned char         transparency;
  unsigned long         tr_col, tr_row_long_offset;
  DgpImage_RGB          rgb, last_rgb;
  DgpImage_CMYK         cmyk, last_cmyk;

  SHIFT_TEMPS

  Cbbtab = Cbbtab;
  row_cnt = row_cnt;

  /* Prepare first color;  guess black. */
  last_rgb.r = 0;   last_rgb.g = 0;     last_rgb.b = 0;
  last_cmyk.c = 0;  last_cmyk.m = 0;    last_cmyk.y = 0;    last_cmyk.k = 255;

  /*Top-bottom crop */
  if( ((int)cinfo->output_scanline < destination.crop_vals.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_vals.end_y) ) {
    return;
  }

  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }

  /*NOTE THAT here we're at the start of the row.. */

  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          width_val = destination.crop_vals.end_x;/* - destination.crop_vals.start_x; */

          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
          /*for(input = 0, col = 0; input < in_cols, col < num_cols; input++, col += cinfo->scale_num){ */
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col >= destination.crop_vals.end_x) return; // was continue ROSHNI */
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;
            
              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[width_val - col] = cmyk.c;
                  destination.m[width_val - col] = cmyk.m;
                  destination.y[width_val - col] = cmyk.y;
                  destination.k[width_val - col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[width_val - col-copy_cnt] = cmyk.c;
                          destination.m[width_val - col-copy_cnt] = cmyk.m;
                          destination.y[width_val - col-copy_cnt] = cmyk.y;
                          destination.k[width_val - col-copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_C:
                  destination.c[width_val - col] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[width_val - col-copy_cnt] = cmyk.c;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_M:
                  destination.m[width_val - col] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.m[width_val - col-copy_cnt] = cmyk.m;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_Y:
                  destination.y[width_val - col] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.y[width_val - col-copy_cnt] = cmyk.y;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_K:
                  destination.k[width_val - col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.k[width_val - col-copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          width_val = destination.crop_vals.end_x;/* - destination.crop_vals.start_x; */

          /* Find the right column for the transparency mask... */
          tr_col = MIN(destination.crop_vals.end_x, (int)num_cols - 1) - 
                   MAX(destination.crop_vals.start_x, 0);
          tr_row_long_offset = (destination.crop_vals.end_y - destination.crop_vals.start_y) * (destination.trans_bpl / sizeof(*mask));
          mask = destination.trans_mask + tr_row_long_offset;

          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
          /*for(input = 0, col = 0; input < in_cols, col < num_cols; input++, col += cinfo->scale_num){ */
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col >= destination.crop_vals.end_x) return; // was continue ROSHNI */
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);

              transparency = (mask[(width_val - col) / 32] >> (31 - ((width_val - col) % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
              /* Under ordinary circumstances, this branch wouldn't save us much, but here we might get to skip */
              /* the color lookup. */
              if(transparency == DGP_OPAQUE) continue;

              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;
            
              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[width_val - col] = cmyk.c;
                  destination.m[width_val - col] = cmyk.m;
                  destination.y[width_val - col] = cmyk.y;
                  destination.k[width_val - col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[width_val - col-copy_cnt] = cmyk.c;
                          destination.m[width_val - col-copy_cnt] = cmyk.m;
                          destination.y[width_val - col-copy_cnt] = cmyk.y;
                          destination.k[width_val - col-copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_C:
                  destination.c[width_val - col] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[width_val - col-copy_cnt] = cmyk.c;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_M:
                  destination.m[width_val - col] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.m[width_val - col-copy_cnt] = cmyk.m;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_Y:
                  destination.y[width_val - col] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.y[width_val - col-copy_cnt] = cmyk.y;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_K:
                  destination.k[width_val - col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.k[width_val - col-copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } /* end of if(mask == NULL)... */

  return;
} /*end of function ycbcr_ycck_cmyk_convert_180 */

/*
 * 5.9.2000 / dhtesq.
 * Implementing a 1-bpp mask, arriving here as a void * in the dest_buf.
 * No guarantees as to what this will do to performance.  Also, cmyk
 * conversion is compartmentalized to its own function, in anticipation
 * of using Sandpiper's tables.  The same goes for all the rotated (90, 180, 270)
 * versions of this function.
 *
 */
METHODDEF(void)
ycbcr_ycck_cmyk_convert(j_decompress_ptr        cinfo,
                        JSAMPIMAGE              input_buf, 
                        JDIMENSION              input_row,
                        dgp_dest_buf            dest_buf, 
                        int                     row_cnt, 
                        int                     num_rows)
{
  my_cconvert_ptr       cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int          y, cb, cr;
  register              JSAMPROW inptr0, inptr1, inptr2, inptr3;
  register JDIMENSION   col,input;
  /*JDIMENSION            num_cols = cinfo->output_width; */
  JDIMENSION            in_cols = cinfo->image_width;
  /*int                   min_cmy = 0;  */
  /*int                   count = 0;  */
  unsigned int          copy_cnt=0;
  dgp_cmyk_dest_buf     destination = dest_buf.cmyk_buf;
  /* Copy these pointers into registers if possible. */
  register JSAMPLE      *range_limit = cinfo->sample_range_limit;
  register int          *Crrtab = cconvert->Cr_r_tab;
  register int          *Cbbtab = cconvert->Cb_b_tab;
  register INT32        *Crgtab = cconvert->Cr_g_tab;
  register INT32        *Cbgtab = cconvert->Cb_g_tab;

  /* For mask..."transparent" and "opaque" are with respect to jpeg, not background. */
  unsigned long         *mask = destination.trans_mask;
  unsigned char         transparency;
  DgpImage_RGB          rgb, last_rgb;
  DgpImage_CMYK         cmyk, last_cmyk;

  SHIFT_TEMPS

  Cbbtab = Cbbtab;
  row_cnt = row_cnt;

  /* Prepare first color;  guess black. */
  last_rgb.r = 0;   last_rgb.g = 0;     last_rgb.b = 0;
  last_cmyk.c = 0;  last_cmyk.m = 0;    last_cmyk.y = 0;    last_cmyk.k = 255;

  /* Top-bottom crop. */
  if( ((int)cinfo->output_scanline < destination.crop_vals.start_y) ||
      ((int)cinfo->output_scanline > destination.crop_vals.end_y) ) {
    return;
  }

  if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
  }
  
  /* Don't make the poor loop carry the mask overhead if we don't even have a mask... */
  if(destination.trans_mask == NULL){
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          /*for(col = 0, input = 0; col < num_cols, input < in_cols; col += cinfo->scale_num, input++) { */
          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col > destination.crop_vals.end_x) return; */
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;
              
              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[col] = cmyk.c;
                  destination.m[col] = cmyk.m;
                  destination.y[col] = cmyk.y;
                  destination.k[col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[col + copy_cnt] = cmyk.c;
                          destination.m[col + copy_cnt] = cmyk.m;
                          destination.y[col + copy_cnt] = cmyk.y;
                          destination.k[col + copy_cnt] = cmyk.k;
                      }
                  }
                  break;
              case DGP_C:
                  destination.c[col] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[col + copy_cnt] = cmyk.c;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_M:
                  destination.m[col] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.m[col + copy_cnt] = cmyk.m;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_Y:
                  destination.y[col] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.y[col + copy_cnt] = cmyk.y;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_K:
                  destination.k[col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.k[col + copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } else {
      /* We have a mask... */
      while (--num_rows >= 0) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          if(cinfo->jpeg_color_space == JCS_YCCK){
              inptr3 = input_buf[3][input_row];
          }
          
          input_row++;
          
          /*for(col = 0, input = 0; col < num_cols, input < in_cols; col += cinfo->scale_num, input++) { */
          for(col = destination.crop_vals.start_x, input = destination.crop_vals.start_x; 
             ((col <= (unsigned)destination.crop_vals.end_x) && (input < in_cols)); 
             col += cinfo->scale_num, input++) {
              
              /*if((int)col < destination.crop_vals.start_x) continue; */
              /*if((int)col > destination.crop_vals.end_x) return; */
              
              transparency = (mask[col / 32] >> (31 - (col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;

              /* Under ordinary circumstances, this branch wouldn't save us much, but here we might get to skip */
              /* the color lookup. */
              if(transparency == DGP_OPAQUE) continue;
              
              y  = GETJSAMPLE(inptr0[input]);
              cb = GETJSAMPLE(inptr1[input]);
              cr = GETJSAMPLE(inptr2[input]);
              /* Range-limiting is essential due to noise introduced by DCT losses. */
              /* Any way to go straight from ycbcr to cmyk? */
              rgb.r = range_limit[y + Crrtab[cr]];
              rgb.g = range_limit[y + ((int)RIGHT_SHIFT(Cbgtab[cb] + Crgtab[cr], SCALEBITS))];
              rgb.b = range_limit[y + Cbbtab[cb]];
              
              if((rgb.r == last_rgb.r) && (rgb.g == last_rgb.g) && (rgb.b == last_rgb.b)){
                  /* Just use last color... */
                  cmyk = last_cmyk;
              } else if((rgb.r == rgb.g) && (rgb.g == rgb.b)){
                  /* Neutral detection... */
                  cmyk.c = 0;   cmyk.m = 0;     cmyk.y = 0;
                  cmyk.k = 255 - rgb.r;
              } else {
                  /* Resort to lookup... */
                  doLutInterpolation(&rgb, &cmyk);
              }
              last_rgb = rgb;
              last_cmyk = cmyk;              

              switch(destination.imgType){
              case DGP_CMYK:
                  destination.c[col] = cmyk.c;
                  destination.m[col] = cmyk.m;
                  destination.y[col] = cmyk.y;
                  destination.k[col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[col + copy_cnt] = cmyk.c;
                          destination.m[col + copy_cnt] = cmyk.m;
                          destination.y[col + copy_cnt] = cmyk.y;
                          destination.k[col + copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_C:
                  destination.c[col] = cmyk.c;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.c[col + copy_cnt] = cmyk.c;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_M:
                  destination.m[col] = cmyk.m;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.m[col + copy_cnt] = cmyk.m;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_Y:
                  destination.y[col] = cmyk.y;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.y[col + copy_cnt] = cmyk.y;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              case DGP_K:
                  destination.k[col] = cmyk.k;
                  if(cinfo->scale_num > 1){
                      for(copy_cnt = 1; copy_cnt < cinfo->scale_num; copy_cnt++){
                          destination.k[col + copy_cnt] = cmyk.k;
                      }
                      /*col += cinfo->scale_num; */
                  }
                  break;
              default:
                  break;
              } /*end of switch statement */
          } /*end of for loop */
      } /*end of while(num_rows) */
  } /* end of if(mask == NULL)... */

  return;
} /*end of function ycbcr_ycck_cmyk_convert */


/*
 * Empty method for start_pass.
 */

METHODDEF(void)
start_pass_dcolor (j_decompress_ptr cinfo)
{
  cinfo = cinfo;
  /* no work needed */
}


/*///////////////////////////////// */
/* Added 16.3.2001 (DA's) / dhtesq. */
/*
 * Convert some rows of samples to the output colorspace.
 *
 * For ICCLAB, the values [0,255] correspond to L [0,100], a [-128,127], and
 * b [-128,127].
 *
 * Note, I equate CIELAB and ICCLAB, since they differ only in
 * thier use of -128.
 */

METHODDEF(void)
icclab_rgb_convert (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int	num_rows)
{

    /* Setup the gammut */
    cinfo->Lmin = 0.0f;
    cinfo->Lmax = 100.0f;
    cinfo->amin = -128.0f;
    cinfo->amax = 127.0f;
    cinfo->bmin = -128.0f;
    cinfo->bmax = 127.0f;

    /* Use the ITULAB conversion, with the gamut just set */
    itulab_rgb_convert_0(cinfo, input_buf, input_row, dest_buf, row_cnt, num_rows);
}


/**************** CIELAB -> RGB conversion **************/

/*
 * See jccolor.c for more information on CIELAB.
 *
 * The conversion is a 2-step process, 8-bit -> CIELAB -> XYZ -> RGB.
 *
 * Here's the 8-bit conversion to ABC -> L*a*b*:
 *	L = A - 127
 *	a = B - 127
 *	b = C - 127
 *
 * Here's the conversion from CIELAB -> XYZ:
 *	X =  ((a + 128.5) / 750.0 + (L + 40.3) / 295.8) ^ 3
 *	Y =  ((L + 40.3) / 295.8) ^ 3
 *	Z =  ((96.5 - b) / 255.0 + (L + 40.3) / 295.8) ^ 3
 * And here's the tranformation from XYZ -> RGB:
 *	R = ( 2.93225894531 * X - 1.45508080078 * Y - 0.45502558593 * Z) * 256
 *	G = (-1.09070074218 * X + 2.01783791015 * Y + 0.03588427734 * Z) * 256
 *	B = ( 0.07755386718 * X - 0.27130406250 * Y + 1.44614982421 * Z) * 256
 *		each bounded from 0 to 255
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^something (about 4 digits precision); we have to
 * divide the products by 2^something to get the correct answer.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants.  And for even more speed, we precomput the
 * power function.
 */

#define EIGHTBIT(x)	(unsigned char)((x) < 0.0f? 0: (x) > 255.0f? 255: (x))

#ifndef USELINEARRGB

#define N_BITS_F_INV	10
#define N_F_INV		(1 << N_BITS_F_INV)
#define KTHREE_BITS	22
#define KFOUR_BITS	15
#define KFIVE_BITS	9
#define KSIX_BITS	10  /* works best at 14, but costs extra time */
#define KTHREE		(1 << KTHREE_BITS)
#define KFOUR		(1 << KFOUR_BITS)
#define KFIVE		(1 << KFIVE_BITS)
#define KSIX		(1 << KSIX_BITS)

#define EIGHTBITRGB(x)	(unsigned char)((x) < 0? 0: (x) > (KFOUR * KFIVE * 255)? 255: \
			(x) >> (KFOUR_BITS + KFIVE_BITS))
/* If negative, use the first array element.
 * If greater than the maximum value less the added rounding factor, use the
 * last array element.
 * Otherwise, round, shift to KSIX elements, and use the LUT.
 */
#define ARRAY_MASK_INV	~(0xffffffff << (KTHREE_BITS))
#define f_inv(x)	(((x) < 0)? ctab[0]: ((unsigned)(x) > (ARRAY_MASK_INV)? ctab[N_F_INV]: ctab[(x) >> (KTHREE_BITS - N_BITS_F_INV)]))

/* If negative, use the first array element.
 * If greater than the maximum value, less the added rounding factor, use the
 * last array element.
 * Otherwise, round, and shift to KSIX elements, and use the LUT.
 */
#define KSMO		(1 << (KSIX_BITS - 1))
#define ARRAY_MASK_SRGB	~(0xffffffff << (KFOUR_BITS + KFIVE_BITS))
#define sRGBp_to_sRGB8(x)	(((x) < 0)? ptab[0]: \
	((unsigned)(x) > (ARRAY_MASK_SRGB - KSMO)? ptab[KSIX - 1]: \
	ptab[((x) + KSMO) >> (KFOUR_BITS + KFIVE_BITS - KSIX_BITS)]))


#else	/* Here's the linear-RGB values: (might need optimiztion, see comments
	 * above) */

#define N_BITS_F_INV	10
#define N_F_INV		(1 << N_BITS_F_INV)
#define KTHREE_BITS	22
#define KFOUR_BITS	15
#define KFIVE_BITS	4
#define KTHREE		(1 << KTHREE_BITS)
#define KFOUR		(1 << KFOUR_BITS)
#define KFIVE		(1 << KFIVE_BITS)

#define EIGHTBITRGB(x)	(unsigned char)((x) < 0? 0: (x) > (KFOUR * KFIVE * 255)? 255: \
			(x) >> (KFOUR_BITS + KFIVE_BITS))
#define KTNMO		(1 << (KTHREE_BITS - N_BITS_F_INV - 1))
#define ARRAY_MASK_INV	(0xffffffffUL << (KTHREE_BITS))
#define f_inv(x)	(((unsigned)(x) & ARRAY_MASK_INV)? \
	(((x) < 0)? ctab[0]: ctab[N_F_INV]): \
	ctab[((x) + KTNMO) >> (KTHREE_BITS - N_BITS_F_INV)])

#endif	/* USELINEARRGB */

#ifndef USELINEARRGB

/*
 * Initialize tables for ITULAB->RGB colorspace conversion.
 */
LOCAL(void)
build_lab_rgb_table (
    j_decompress_ptr cinfo)
{
    Int32 i;
    float f;
    Int32 *lab_rgb_tab;
    UInt8 *srp_sr8_tab;
	DGP_VOID_PROLOGUE("build_lab_rgb_table()", NULL);

    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;

    /* Allocate and fill in the conversion tables. */
	lab_rgb_tab = (Int32 *)((*cinfo->mem->alloc_small)((j_common_ptr) cinfo, JPOOL_IMAGE, ((N_F_INV + 1) * SIZEOF(Int32))));
	/* What to do, what to do...assert is silly;  panic will not print for lack of an error state. */
	if(!lab_rgb_tab){
		DGP_PANIC("I was unable to allocate for the lab_rgb_tab conversion table.");
	}
    cconvert->lab_rgb_tab = lab_rgb_tab;

    for (i = 0; i < N_F_INV+1; i++) {
		f = (float)i / (float)(N_F_INV - 1);
		if (f < 0.20689303f) {
			f = ((f) - 16.0f / 116.0f) / 7.787f;
		} else {
			f = f * f * f;
		}
		lab_rgb_tab[i] = (Int32)(f * (float)KFOUR);
    }

    srp_sr8_tab = (UInt8 *)((*cinfo->mem->alloc_small)((j_common_ptr) cinfo, JPOOL_IMAGE, ((KSIX) * SIZEOF(Int32))));
    cconvert->srp_sr8_tab = srp_sr8_tab;

    for (i = 0; i < KSIX; i++) {
		f = (float)i / (float)(KSIX - 1);
		if (f <= 0.00313) {
			f = f * 12.92f * 256.0f;
		} else {
			f = ((float)pow((double)f, 1.0 / 2.4) * 1.055f - 0.055f) * 256.0f;
		}
		if (f > 255.0f) {
			srp_sr8_tab[i] = 255;
		} else if (f < 0.0f) {
			srp_sr8_tab[i] = 0;
		} else {
			srp_sr8_tab[i] = (UInt8)f;
		}
    }

    return;
}

#define  PRE_COMPUTE_ITU_CONST             \
                                                /* If a different gamut exists, grab it */                      \
                                                {                                                               \
                                                if (cinfo->Lmax - cinfo->Lmin > 0) {                            \
		                                            Lmin = cinfo->Lmin;                                         \
		                                            Lmax = cinfo->Lmax;                                         \
		                                            if (cinfo->amax - cinfo->amin > 0) {                        \
			                                            amin = cinfo->amin;                                     \
			                                            amax = cinfo->amax;                                     \
			                                            if (cinfo->bmax - cinfo->bmin > 0) {                    \
				                                            bmin = cinfo->bmin;                                 \
				                                            bmax = cinfo->bmax;                                 \
			                                            }                                                       \
		                                            }                                                           \
                                                }else{                                                          \
                                                    Xn = 0.9505f; Yn = 1.000f; Zn = 1.089f;                     \
                                                    Lmin = 0.0f; Lmax = 100.0f;                                 \
                                                    amin = -21760.0f/255.0f; amax = 21590.0f/255.0f;            \
                                                    bmin = -19200.0f/255.0f; bmax = 31800.0f/255.0f;            \
                                                }                                                               \
                                                const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE); \
                                                const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);           \
                                                const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE); \
                                                const4 = (Int32)((amin / 500.0f) * KTHREE);                     \
                                                const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE); \
                                                const6 = (Int32)((-bmin / 200.0f) * KTHREE);                    \
                                                    \
                                                r1 = (Int32)(( 3.2410f * Xn) * KFIVE);                          \
                                                r2 = (Int32)((-1.5374f * Yn) * KFIVE);                          \
                                                r3 = (Int32)((-0.4986f * Zn) * KFIVE);                          \
                                                g1 = (Int32)((-0.9692f * Xn) * KFIVE);                          \
                                                g2 = (Int32)(( 1.8760f * Yn) * KFIVE);                          \
                                                g3 = (Int32)(( 0.0416f * Zn) * KFIVE);                          \
                                                b1 = (Int32)(( 0.0556f * Xn) * KFIVE);                          \
                                                b2 = (Int32)((-0.2040f * Yn) * KFIVE);                          \
                                                b3 = (Int32)(( 1.0570f * Zn) * KFIVE);                          \
                                                }  

/* CONVERT FROM ITULAB TO sRGB
 *
 * Conversion from XYZ to sRGB:
 * sR =  3.2410 * X - 1.5374 * Y - 0.4986 * Z;
 * sG = -0.9692 * X + 1.8760 * Y + 0.0416 * Z;
 * sB =  0.0556 * X - 0.2040 * Y + 1.0570 * Z;
 *
 * Conversion from sRGB to sRGB':
 *	For each x = sR, sG, sB:
 * if (x <= 0.00313) {
 *	x' = x * 12.92;
 * } else {
 *	x' = pow(x, (1.0 / 2.4)) * 1.055 - 0.055;
 * }
 * Conversion from sRGB' to 8-bit sRGB:
 *	For each x = sR, sG, sB:
 * 8-bit x = eightbit(x * 256.0);
 * where eightbit(x) = x railed to [0,255]
 *
 */


METHODDEF(void)
itulab_rgb_convert_0 (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int	num_rows)
{
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->lab_rgb_tab;
    register UInt8 * ptab = cconvert->srp_sr8_tab;
    register JSAMPROW outptr;
    register JSAMPROW inptr0, inptr1, inptr2;
    register JDIMENSION col, input;
    JDIMENSION num_cols = cinfo->output_width;
    JDIMENSION in_cols = cinfo->image_width;
    UInt8 L, a, b;
    Int32 X, Y, Z;
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn, Yn, Zn;
    float Lmin, Lmax;
    float amin, amax;
    float bmin, bmax;
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */

    dgp_other_dest_buf destination = dest_buf.other_buf;
    char** output_buf = &(destination.buffer);
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */
    Int32 const1, const2, const3, const4, const5, const6;
    Int32 c, t;
    Int32 sR, sG, sB;
	Dgp_UInt8 r_val, g_val, b_val;
	unsigned long *mask = destination.trans_mask;
	unsigned int tLpl = destination.trans_bpl / sizeof(*mask);
	unsigned char transparency;
	unsigned int bpr = destination.bytes_per_row;
	Dgp_UInt32 copy_cnt;
    Dgp_UInt32 dstRGBpl,  maskWpl;
    Dgp_UInt32 col_start, col_end, input_start;

#define ITULAB_TO_RGB(input, rVal, gVal, bVal)      \
                                            L = GETJSAMPLE(inptr0[input]);  \
				                            a = GETJSAMPLE(inptr1[input]);  \
				                            b = GETJSAMPLE(inptr2[input]);  \
				                            c = const1 * L + const2;        \
  				                            t = const3 * a + const4 + c;    \
                                            X = f_inv(t);                   \
                                            Y = f_inv(c);                   \
			                                t = c - b * const5 + const6;    \
                                            Z = f_inv(t);                   \
				                            sR = r1 * X + r2 * Y + r3 * Z;  \
				                            sG = g1 * X + g2 * Y + g3 * Z;  \
				                            sB = b1 * X + b2 * Y + b3 * Z;  \
                                            rVal = sRGBp_to_sRGB8(sR);      \
                                            gVal = sRGBp_to_sRGB8(sG);      \
                                            bVal = sRGBp_to_sRGB8(sB);      \

	/*Top-bottom crop */
	if( ((int)(cinfo->output_scanline +  cinfo->scale_num - 1) < destination.crop_coords.start_y) ||
		((int)(cinfo->output_scanline) > destination.crop_coords.end_y) ) {
		return;
	} 
    if(cinfo->scale_num == 1){
    in_cols = cinfo->output_width;
    }

    PRE_COMPUTE_ITU_CONST;
    
    /* selector layer DWORD per line */
    maskWpl = destination.trans_bpl / sizeof(Dgp_UInt32);

    /* destination byte perl line */
    dstRGBpl = in_cols * cinfo->scale_num * RGB_PIXELSIZE;

    /* destination start and end postion croped by start_x. */
    col_end = destination.crop_coords.end_x;

    if (destination.trans_mask == NULL) {
        Dgp_UInt32 numToCopy = 1;
        /* dest and src start position croped by start_x  */
        col_start  = destination.crop_coords.start_x;
        input_start = destination.crop_coords.start_x / cinfo->scale_num;
        while (--num_rows >= 0) {
            inptr0 = input_buf[0][input_row];
            inptr1 = input_buf[1][input_row];
            inptr2 = input_buf[2][input_row];
            input_row++;
            outptr = *output_buf; 
            for(col = 0, input = 0; input < in_cols; col+= cinfo->scale_num, input++) {
                /* In the normal orientation case, just don't write out what is not needed. */
                if((int)(col + cinfo->scale_num - 1) < destination.crop_coords.start_x)
                  continue;
                if((int)col > destination.crop_coords.end_x){
                  /*fprintf(stderr,"HOHO chk_cnt = %d\n",check_cnt); */
                  return;
                }
                ITULAB_TO_RGB(input, r_val, g_val, b_val);
				if(cinfo->BGR){
					outptr[RGB_RED] = b_val;
					outptr[RGB_GREEN] = g_val;
					outptr[RGB_BLUE] = r_val;
				}else{
					outptr[RGB_RED] = r_val;
					outptr[RGB_GREEN] = g_val;
					outptr[RGB_BLUE] = b_val;
				}
                outptr += RGB_PIXELSIZE;
                if(cinfo->scale_num >1){
                    numToCopy = (int)col < destination.crop_coords.start_x ?
                      cinfo->scale_num - (destination.crop_coords.start_x - col) : cinfo->scale_num;
                    numToCopy = MIN(numToCopy, destination.crop_coords.end_x - col + 1);
                    for(copy_cnt = 1; copy_cnt <numToCopy; copy_cnt++){
						if(cinfo->BGR){
							outptr[RGB_RED] = b_val;
							outptr[RGB_GREEN] = g_val;
							outptr[RGB_BLUE] = r_val;
						}else{
							outptr[RGB_RED] = r_val;
							outptr[RGB_GREEN] = g_val;
							outptr[RGB_BLUE] = b_val;
						}
                        outptr += RGB_PIXELSIZE;
                    } 
                }
            } /*end of for(col...) */
            *output_buf += dstRGBpl;
        } /*end of while loop  */
    } else {
      /* We have a mask...  */
      register int j, col2;  
      register Dgp_UInt8  numToCopyVert;
      register Dgp_UInt32 maskvalue;
      register JSAMPROW tempptr;
      register Dgp_UInt32 wStart, shift;
      register Dgp_UInt32 in_cols_aligned, col_end_aligned;
      register Dgp_UInt8 r_val2, g_val2, b_val2;             
      Dgp_UInt8 r_vall[4], g_vall[4],b_vall[4];

      /* no matter if the scale number is 1, 2 or 4, we always read 4 bit each time */
      /* from the selector and look up the pixel assignment in a table of functions. */
      /* for scale number 1, pass 4 color pixels each time.( 1 bg pixel - 1 selector bit) */
      /* for scale number 2, pass 2 color pixels each time ( 1 bg pixel - 2 selector bit) */
      /* for scale number 4, pass 1 color pixels each time ( 1 bg pixel - 4 selector bit) */
      /* for scale 1, 2, the remaining pixels each line will be handled seperately. */
      /* for scale number 3, we use the tradional way of assigning pixel based on each slector bit */
      
      /* in_cols_aligned and col_end_aligned is 4 pixel boundary */
      switch (cinfo->scale_num){
      case 2:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* every 2 input match 4 dest */
          in_cols_aligned =  in_cols - in_cols % 2;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          col_end_aligned =  col_end - (destination.crop_coords.end_x + 1) % 4  ;
          break;
      case 1:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* every 4 input match 4 dest */
          in_cols_aligned = in_cols - in_cols % 4;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          col_end_aligned =  col_end - (destination.crop_coords.end_x + 1) % 4  ;
          break;
      case 4:
          /* back ground layer start and end position croped by start_x */
          input_start = destination.crop_coords.start_x / 4;
          /* every 4 input match 4 dest */
          in_cols_aligned = in_cols;
          /* destination start and end position align on 4 pixels boundary */
          col_start =  input_start * 4;
          /* for scale num = 4, there should be no left over pixels */
          col_end_aligned = col_end / 4 * 4  + 3;
          break;
      default: /* for other scale factor use the old method. */
          /* dest and src start position croped by start_x  */
          col_start = destination.crop_coords.start_x;
          input_start = destination.crop_coords.start_x / cinfo->scale_num;
          break;
          
      }
      /* crop destination by start_y and end_y.  */
	  numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
		  cinfo->output_scanline - destination.crop_coords.start_y + cinfo->scale_num:cinfo->scale_num;
	  numToCopyVert = MIN(numToCopyVert, destination.crop_coords.end_y - cinfo->output_scanline + 1);
      /* in this case num_rows is usually just 1. This while loop is not useful. Also if num_rows is */
      /* not one than the numToCopyVert should be calculated for each row. */
      while (num_rows--) {
          inptr0 = input_buf[0][input_row];
          inptr1 = input_buf[1][input_row];
          inptr2 = input_buf[2][input_row];
          input_row++;
          outptr = *output_buf + col_start * RGB_PIXELSIZE;
          switch(cinfo->scale_num){
          case 4:
              for(col = col_start, input = input_start; input < in_cols_aligned, col <= col_end_aligned; col +=4, input++) {
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
              
                  ITULAB_TO_RGB(input, r_val, g_val, b_val);
              
                  tempptr = outptr;
                  wStart = col / 32;
                  shift  = 28 - col % 32; 
                  for(j = 0; j < numToCopyVert; j++){
                      maskvalue = (mask[wStart + maskWpl * j] >> shift) & 0x0f;
					  if(cinfo->BGR)
						pixel_assign4[maskvalue](tempptr, b_val, g_val, r_val);  
					  else
						pixel_assign4[maskvalue](tempptr, r_val, g_val, b_val);  
                      tempptr += dstRGBpl;
                  }
                  outptr += 4 * RGB_PIXELSIZE;
              } /*end of for(col...) */
              break;
          case 2:
              for(col = col_start, input = input_start; input < in_cols_aligned, col<= col_end_aligned; col +=4, input++) {
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
                  ITULAB_TO_RGB(input, r_val, g_val, b_val);
                  input ++;
                  ITULAB_TO_RGB(input, r_val2, g_val2, b_val2);
                  tempptr = outptr;
                  wStart = col / 32;
                  shift  = 28 - col % 32; 
                  for(j = 0; j < numToCopyVert; j++){
                      maskvalue = (mask[wStart + maskWpl * j] >> shift) & 0x0f;
					  if(cinfo->BGR)
						pixel_assign2[maskvalue](tempptr, b_val, g_val, r_val, b_val2, g_val2, r_val2);  
					  else
						pixel_assign2[maskvalue](tempptr, r_val, g_val, b_val, r_val2, g_val2, b_val2);  
                      tempptr += dstRGBpl;
                  }
                  outptr += 4 * RGB_PIXELSIZE;
              }/* end of for(col.. */
              break;
          case 1:
              for(col = col_start, input = input_start; input < in_cols_aligned, col<= col_end_aligned; col +=4, input++) {
              
                  /* convert ITULAB to RGB */
                  ITULAB_TO_RGB(input, r_vall[0], g_vall[0], b_vall[0]);
                  input ++;
                  ITULAB_TO_RGB(input, r_vall[1], g_vall[1], b_vall[1]);
                  input ++;
                  ITULAB_TO_RGB(input, r_vall[2], g_vall[2], b_vall[2]);
                  input ++;
                  ITULAB_TO_RGB(input, r_vall[3], g_vall[3], b_vall[3]);
                  /* get maskvalue in unit of 32 bit */
                  /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
                  /* The masks coming out of the print driver seem to be big-endian by construction. */
                  maskvalue = (mask[ col / 32] >> (28 - col % 32)) & 0x0f;
				  if(cinfo->BGR)
					  pixel_assign1[maskvalue](outptr, b_vall, g_vall, r_vall); 
				  else
					  pixel_assign1[maskvalue](outptr, r_vall, g_vall, b_vall); 
                  outptr += 4 * RGB_PIXELSIZE;
              }/* end of for(col.. */
              break;
          default:
              col = col_start;
              input = input_start;
              break;
          }/*end of switch(cinfo_scale_num) */

          /* for remaining columns that is not dword aligned. or for the other scalefactors that are not 1.2 or4 */
          /* we use the old method of layer merging one bit at a time without table look up. */
          while(input < in_cols && col <= col_end) {
              Dgp_UInt32 numToCopy = cinfo->scale_num - (int)col % cinfo->scale_num;
              numToCopy = MIN(numToCopy, destination.crop_coords.end_x - col + 1);
              /* get maskvalue in unit of 32 bit */
              /* We assume a 1-bpp decompressed mask in the same orientation as the jpeg. */
              /* The masks coming out of the print driver seem to be big-endian by construction. */
              ITULAB_TO_RGB(input, r_val, g_val, b_val);  
              for(j = 0; j < numToCopyVert; j++){
                  col2 = col;
                  tempptr = outptr + j * dstRGBpl;
                  for(copy_cnt = 0; copy_cnt < numToCopy; copy_cnt++){
                      transparency = (mask[(j * destination.trans_bpl / sizeof(Dgp_UInt32)) + (col2 / 32)] 
                          >> (31 - (col2 % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
                      col2++;
					  if(cinfo->BGR){
						  tempptr[RGB_RED] = (tempptr[RGB_RED] & (~transparency)) | (b_val & transparency);
						  tempptr[RGB_GREEN] = (tempptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  tempptr[RGB_BLUE] = (tempptr[RGB_BLUE] & (~transparency)) | (r_val & transparency);
					  }else{
						  tempptr[RGB_RED] = (tempptr[RGB_RED] & (~transparency)) | (r_val & transparency);
						  tempptr[RGB_GREEN] = (tempptr[RGB_GREEN] & (~transparency)) | (g_val & transparency);
						  tempptr[RGB_BLUE] = (tempptr[RGB_BLUE] & (~transparency)) | (b_val & transparency);
					  }
                      tempptr += RGB_PIXELSIZE;
                  }					  
              }
              outptr += numToCopy * RGB_PIXELSIZE;
              input++;
              col += numToCopy; 
                  
          } /*end of while(input<in_cols... */
          *output_buf += dstRGBpl;
      } /*end of while loop */
  } /* end of if(mask == NULL)...  */
    return;
}

/* ORIENT_DL case... */
METHODDEF(void)
itulab_rgb_convert_270 (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int num_rows)
{
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->lab_rgb_tab;
    register UInt8 * ptab = cconvert->srp_sr8_tab;
    register JSAMPROW outptr;
    register JSAMPROW inptr0, inptr1, inptr2;
    register JDIMENSION col, input;
	JDIMENSION num_cols = cinfo->output_width;
	JDIMENSION in_cols = cinfo->image_width;
    UInt8 L, a, b;
    Int32 X, Y, Z;
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    dgp_other_dest_buf destination = dest_buf.other_buf;
    char** output_buf = &(destination.buffer);
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */
    Int32 const1, const2, const3, const4, const5, const6;
    Int32 c, t;
    Int32 sR, sG, sB;
	int rVal, gVal, bVal;
	unsigned long *mask = destination.trans_mask;
	unsigned long tr_col, tr_row_long_offset;
	unsigned int tLpl = destination.trans_bpl / sizeof(*mask);
	unsigned char transparency;
	unsigned int bpr = destination.bytes_per_row;
	int copyCount = 0;
	int nominalCopies = cinfo->scale_num - 1;

	/*Top-bottom crop */
	if( ((int)(cinfo->output_scanline + nominalCopies) < destination.crop_coords.start_y) ||
		((int)(cinfo->output_scanline) > destination.crop_coords.end_y) ) {
		return;
	} 

	if(cinfo->scale_num == 1){
		in_cols = cinfo->output_width;
	}

    /* If a different gamut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
		Lmin = cinfo->Lmin;
		Lmax = cinfo->Lmax;
		if (cinfo->amax - cinfo->amin > 0) {
			amin = cinfo->amin;
			amax = cinfo->amax;
			if (cinfo->bmax - cinfo->bmin > 0) {
				bmin = cinfo->bmin;
				bmax = cinfo->bmax;
			}
		}
    }

    /* Precompute the constants */
    const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Int32)((amin / 500.0f) * KTHREE);
    const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Int32)((-bmin / 200.0f) * KTHREE);

    r1 = (Int32)(( 3.2410f * Xn) * KFIVE);
    r2 = (Int32)((-1.5374f * Yn) * KFIVE);
    r3 = (Int32)((-0.4986f * Zn) * KFIVE);
    g1 = (Int32)((-0.9692f * Xn) * KFIVE);
    g2 = (Int32)(( 1.8760f * Yn) * KFIVE);
    g3 = (Int32)(( 0.0416f * Zn) * KFIVE);
    b1 = (Int32)(( 0.0556f * Xn) * KFIVE);
    b2 = (Int32)((-0.2040f * Yn) * KFIVE);
    b3 = (Int32)(( 1.0570f * Zn) * KFIVE);

	if(destination.trans_mask == NULL){
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;

			*(output_buf) += (destination.crop_coords.end_x - destination.crop_coords.start_x) * bpr;
			outptr = *(output_buf);

			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    outptr[RGB_RED] = rVal = sRGBp_to_sRGB8(sR);
				outptr[RGB_GREEN] = gVal = sRGBp_to_sRGB8(sG);
				outptr[RGB_BLUE] = bVal = sRGBp_to_sRGB8(sB);
				outptr -= bpr;

				if(nominalCopies){
					int numToCopy = (int)col < destination.crop_coords.start_x ? 
						cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
					numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
					for(copyCount = 1; copyCount < numToCopy; copyCount++){
						outptr[RGB_RED] = rVal;
						outptr[RGB_GREEN] = gVal;
						outptr[RGB_BLUE] = bVal;
						outptr -= bpr;
					}
				}
			}
		}
	} else {
      /* We have a mask... */
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;

			*(output_buf) += (destination.crop_coords.end_x - destination.crop_coords.start_x) * bpr;
			outptr = *(output_buf);
			
			/* Get to the right row of the transparency mask... */
			tr_row_long_offset = (MIN((int)(num_cols - 1), destination.crop_coords.end_x) - 
				MAX(0, destination.crop_coords.start_x)) * tLpl;
			tr_col = cinfo->output_scanline - destination.crop_coords.start_y;
			mask = destination.trans_mask + tr_row_long_offset;

			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				/* We assume a 1-bpp decompressed mask in the same orientation as the jpeg destination. */
				/* The masks coming out of the print driver seem to be big-endian by construction. */
				transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;

				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    rVal = sRGBp_to_sRGB8(sR);
				gVal = sRGBp_to_sRGB8(sG);
				bVal = sRGBp_to_sRGB8(sB);
				outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
				outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
				outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
				outptr -= bpr;
				mask -= tLpl;

				if(nominalCopies){
				  int j;
				  unsigned long *startMask = mask + tLpl;
				  JSAMPROW startOutptr = outptr + bpr;
				  int numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  int numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, (int)(destination.crop_coords.end_y - cinfo->output_scanline + 1));

				  for(copyCount = 1; copyCount < numToCopy; copyCount++){
		              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
		              mask -= tLpl;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
					  outptr -= bpr;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < numToCopyVert; j++){
					  mask = startMask;
					  outptr = startOutptr + (j * RGB_PIXELSIZE);
					  for(copyCount = 0; copyCount < numToCopy; copyCount++){
			              transparency = (mask[(tr_col + j) / 32] >> (31 - ((tr_col + j) % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
			              mask -= tLpl;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
						  outptr -= bpr;
					  }					  
				  }
				  mask = startMask - (numToCopy * tLpl);
				  outptr = startOutptr - (numToCopy * bpr);
				}
			}
		}
	}

    return;
}


/* ORIENT_UR case... */
METHODDEF(void)
itulab_rgb_convert_90 (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int num_rows)
{
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->lab_rgb_tab;
    register UInt8 * ptab = cconvert->srp_sr8_tab;
    register JSAMPROW outptr;
    register JSAMPROW inptr0, inptr1, inptr2;
    register JDIMENSION col, input;
	JDIMENSION num_cols = cinfo->output_width;
	JDIMENSION in_cols = cinfo->image_width;
    UInt8 L, a, b;
    Int32 X, Y, Z;
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    dgp_other_dest_buf destination = dest_buf.other_buf;
    char** output_buf = &(destination.buffer);
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */
    Int32 const1, const2, const3, const4, const5, const6;
    Int32 c, t;
    Int32 sR, sG, sB;
	int rVal, gVal, bVal;
	unsigned long *mask = destination.trans_mask;
	unsigned long tr_col;
	unsigned int tLpl = destination.trans_bpl / sizeof(*mask);
	unsigned char transparency;
	unsigned int bpr = destination.bytes_per_row;
	int copyCount = 0;
	int nominalCopies = cinfo->scale_num - 1;

	/*Top-bottom crop */
	if( ((int)(cinfo->output_scanline + nominalCopies) < destination.crop_coords.start_y) ||
		((int)(cinfo->output_scanline) > destination.crop_coords.end_y) ) {
		return;
	} 

	if(cinfo->scale_num == 1){
		in_cols = cinfo->output_width;
	}

    /* If a different gamut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
		Lmin = cinfo->Lmin;
		Lmax = cinfo->Lmax;
		if (cinfo->amax - cinfo->amin > 0) {
			amin = cinfo->amin;
			amax = cinfo->amax;
			if (cinfo->bmax - cinfo->bmin > 0) {
				bmin = cinfo->bmin;
				bmax = cinfo->bmax;
			}
		}
    }

    /* Precompute the constants */
    const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Int32)((amin / 500.0f) * KTHREE);
    const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Int32)((-bmin / 200.0f) * KTHREE);

    r1 = (Int32)(( 3.2410f * Xn) * KFIVE);
    r2 = (Int32)((-1.5374f * Yn) * KFIVE);
    r3 = (Int32)((-0.4986f * Zn) * KFIVE);
    g1 = (Int32)((-0.9692f * Xn) * KFIVE);
    g2 = (Int32)(( 1.8760f * Yn) * KFIVE);
    g3 = (Int32)(( 0.0416f * Zn) * KFIVE);
    b1 = (Int32)(( 0.0556f * Xn) * KFIVE);
    b2 = (Int32)((-0.2040f * Yn) * KFIVE);
    b3 = (Int32)(( 1.0570f * Zn) * KFIVE);

	if(destination.trans_mask == NULL){
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;

			outptr = *(output_buf);

			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    outptr[RGB_RED] = rVal = sRGBp_to_sRGB8(sR);
				outptr[RGB_GREEN] = gVal = sRGBp_to_sRGB8(sG);
				outptr[RGB_BLUE] = bVal = sRGBp_to_sRGB8(sB);
				outptr += bpr;

				if(nominalCopies){
					int numToCopy = (int)col < destination.crop_coords.start_x ? 
						cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
					numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
					for(copyCount = 1; copyCount < numToCopy; copyCount++){
						outptr[RGB_RED] = rVal;
						outptr[RGB_GREEN] = gVal;
						outptr[RGB_BLUE] = bVal;
						outptr += bpr;
					}
				}
			}
		}
	} else {
      /* We have a mask... */
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;

			outptr = *(output_buf);
			
			/* Get to the right row of the transparency mask... */
			tr_col = destination.crop_coords.end_y - cinfo->output_scanline;

			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				/* We assume a 1-bpp decompressed mask in the same orientation as the jpeg destination. */
				/* The masks coming out of the print driver seem to be big-endian by construction. */
				transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;

				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    rVal = sRGBp_to_sRGB8(sR);
				gVal = sRGBp_to_sRGB8(sG);
				bVal = sRGBp_to_sRGB8(sB);
				outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
				outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
				outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
				outptr += bpr;
				mask += tLpl;

				if(nominalCopies){
				  int j;
				  unsigned long *startMask = mask - tLpl;
				  JSAMPROW startOutptr = outptr - bpr;
				  int numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  int numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, (int)(destination.crop_coords.end_y - cinfo->output_scanline + 1));

				  for(copyCount = 1; copyCount < numToCopy; copyCount++){
		              transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
		              mask += tLpl;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
					  outptr += bpr;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < numToCopyVert; j++){
					  mask = startMask;
					  outptr = startOutptr - (j * RGB_PIXELSIZE);
					  for(copyCount = 0; copyCount < numToCopy; copyCount++){
			              transparency = (mask[(tr_col - j) / 32] >> (31 - ((tr_col - j) % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
			              mask += tLpl;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
						  outptr += bpr;
					  }					  
				  }
				  mask = startMask + (numToCopy * tLpl);
				  outptr = startOutptr + (numToCopy * bpr);
				}
			}
		}
	}

    return;
}

/* ORIENT_LU case... */
METHODDEF(void)
itulab_rgb_convert_180 (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int	num_rows)
{
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->lab_rgb_tab;
    register UInt8 * ptab = cconvert->srp_sr8_tab;
    register JSAMPROW outptr;
    register JSAMPROW inptr0, inptr1, inptr2;
    register JDIMENSION col, input;
	JDIMENSION num_cols = cinfo->output_width;
	JDIMENSION in_cols = cinfo->image_width;
    UInt8 L, a, b;
    Int32 X, Y, Z;
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    dgp_other_dest_buf destination = dest_buf.other_buf;
    char** output_buf = &(destination.buffer);
    /* Note, here the computed full spectrum sRGB gamut
     * float amin = -88.0f, amax = 99.0f;
     * float bmin = -108.8f, bmax = 95.2f;
     */
    Int32 const1, const2, const3, const4, const5, const6;
    Int32 c, t;
    Int32 sR, sG, sB;
	int rVal, gVal, bVal, outptr_r;
	unsigned long *mask = destination.trans_mask;
	unsigned int tLpl = destination.trans_bpl / sizeof(*mask);
	unsigned char transparency;
	unsigned long tr_col, tr_row_long_offset;
	unsigned int bpr = destination.bytes_per_row;
	int copyCount = 0;
	int nominalCopies = cinfo->scale_num - 1;

	/* Top-bottom crop */
	if( ((int)(cinfo->output_scanline + nominalCopies) < destination.crop_coords.start_y) ||
		((int)(cinfo->output_scanline) > destination.crop_coords.end_y) ) {
		return;
	} 

	if(cinfo->scale_num == 1){
		in_cols = cinfo->output_width;
	}

    /* If a different gamut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
		Lmin = cinfo->Lmin;
		Lmax = cinfo->Lmax;
		if (cinfo->amax - cinfo->amin > 0) {
			amin = cinfo->amin;
			amax = cinfo->amax;
			if (cinfo->bmax - cinfo->bmin > 0) {
				bmin = cinfo->bmin;
				bmax = cinfo->bmax;
			}
		}
    }

    /* Precompute the constants */
    const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Int32)((amin / 500.0f) * KTHREE);
    const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Int32)((-bmin / 200.0f) * KTHREE);

    r1 = (Int32)(( 3.2410f * Xn) * KFIVE);
    r2 = (Int32)((-1.5374f * Yn) * KFIVE);
    r3 = (Int32)((-0.4986f * Zn) * KFIVE);
    g1 = (Int32)((-0.9692f * Xn) * KFIVE);
    g2 = (Int32)(( 1.8760f * Yn) * KFIVE);
    g3 = (Int32)(( 0.0416f * Zn) * KFIVE);
    b1 = (Int32)(( 0.0556f * Xn) * KFIVE);
    b2 = (Int32)((-0.2040f * Yn) * KFIVE);
    b3 = (Int32)(( 1.0570f * Zn) * KFIVE);

	if(destination.trans_mask == NULL){
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;

			/* Shift to the last column but one... */
			outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * RGB_PIXELSIZE;
			outptr = *(output_buf) + outptr_r;

			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    outptr[RGB_RED] = rVal = sRGBp_to_sRGB8(sR);
				outptr[RGB_GREEN] = gVal = sRGBp_to_sRGB8(sG);
				outptr[RGB_BLUE] = bVal = sRGBp_to_sRGB8(sB);
				outptr -= RGB_PIXELSIZE;

				if(nominalCopies){
					int numToCopy = (int)col < destination.crop_coords.start_x ? 
						cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
					numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
					for(copyCount = 1; copyCount < numToCopy; copyCount++){
						outptr[RGB_RED] = rVal;
						outptr[RGB_GREEN] = gVal;
						outptr[RGB_BLUE] = bVal;
						outptr -= RGB_PIXELSIZE;
					}
				}
			}
		}
	} else {
      /* We have a mask... */
		while (--num_rows >= 0) {
			inptr0 = input_buf[0][input_row];
			inptr1 = input_buf[1][input_row];
			inptr2 = input_buf[2][input_row];
			input_row++;
			outptr_r = (destination.crop_coords.end_x - destination.crop_coords.start_x) * RGB_PIXELSIZE; 
			outptr = *(output_buf) + outptr_r;
			/* Find the right column for the transparency mask... */
			tr_col = MIN((int)(destination.crop_coords.end_x), (int)num_cols - 1) - MAX(destination.crop_coords.start_x, 0);
			tr_row_long_offset = (destination.crop_coords.end_y - destination.crop_coords.start_y) * tLpl;
			mask = destination.trans_mask + tr_row_long_offset;
			for (col = 0, input = 0; input < in_cols; input++, col += cinfo->scale_num) {
				/* In the normal orientation case, just don't write out what is not needed. */
				if((int)(col + nominalCopies) < destination.crop_coords.start_x)
					continue;
				if((int)col > destination.crop_coords.end_x){
					return;
				}
				/* We assume a 1-bpp decompressed mask in the same orientation as the jpeg destination. */
				/* The masks coming out of the print driver seem to be big-endian by construction. */
				transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;

				L = GETJSAMPLE(inptr0[input]);
				a = GETJSAMPLE(inptr1[input]);
				b = GETJSAMPLE(inptr2[input]);

				/* Convert from ITULAB -> XYZ: */
				c = const1 * L + const2;
				t = const3 * a + const4 + c;
				X = f_inv(t);
				Y = f_inv(c);
				t = c - b * const5 + const6;
				Z = f_inv(t);
				
				/* Convert from XYZ to RGB */
				sR = r1 * X + r2 * Y + r3 * Z;
				sG = g1 * X + g2 * Y + g3 * Z;
				sB = b1 * X + b2 * Y + b3 * Z;
				
				/* Convert from sRGB to sRGB' to 8-bit sRGB */
				/* Rail and integerize to 8-bit */
			    rVal = sRGBp_to_sRGB8(sR);
				gVal = sRGBp_to_sRGB8(sG);
				bVal = sRGBp_to_sRGB8(sB);
				outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
				outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
				outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
				outptr -= RGB_PIXELSIZE;
				tr_col--;

				if(nominalCopies){
				  int j;
				  int startCol = tr_col + 1;
				  JSAMPROW startOutptr = outptr + RGB_PIXELSIZE;
				  int numToCopy = (int)col < destination.crop_coords.start_x ? 
					  cinfo->scale_num - (destination.crop_coords.start_x - col):cinfo->scale_num;
				  int numToCopyVert = (int)cinfo->output_scanline < destination.crop_coords.start_y ? 
					  cinfo->scale_num - (destination.crop_coords.start_y - cinfo->output_scanline):cinfo->scale_num;
				  numToCopy = MIN(numToCopy, (int)(destination.crop_coords.end_x - col + 1));
				  numToCopyVert = MIN(numToCopyVert, (int)(destination.crop_coords.end_y - cinfo->output_scanline + 1));

				  for(copyCount = 1; copyCount < numToCopy; copyCount++){
					  transparency = (mask[tr_col / 32] >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
					  tr_col--;
					  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
					  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
					  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
					  outptr -= RGB_PIXELSIZE;
				  }
				  /* Adding vertical loop too; no row copies... / dhtesq.				   */
				  for(j = 1; j < numToCopyVert; j++){
					  tr_col = startCol;
					  outptr = startOutptr - (j * in_cols * cinfo->scale_num * RGB_PIXELSIZE);
					  for(copyCount = 0; copyCount < numToCopy; copyCount++){
						  transparency = (mask[-(int)(j * tLpl) + (tr_col / 32)] 
							  >> (31 - (tr_col % 32))) & 0x1 ? DGP_TRANSPARENT:DGP_OPAQUE;
						  tr_col--;
						  outptr[RGB_RED] = (outptr[RGB_RED] & (~transparency)) | (rVal & transparency);
						  outptr[RGB_GREEN] = (outptr[RGB_GREEN] & (~transparency)) | (gVal & transparency);
						  outptr[RGB_BLUE] = (outptr[RGB_BLUE] & (~transparency)) | (bVal & transparency);
						  outptr -= RGB_PIXELSIZE;
					  }					  
				  }
				  outptr = startOutptr - (numToCopy * RGB_PIXELSIZE);
				}
			}
		}
	}

    return;
}

#else	/* Here's the linear-RGB: */


/**************** ITULAB -> RGB conversion **************/

/*
 * See jccolor.c for more information on ITULAB.
 *
 * The conversion is a 2-step process, ITULAB -> XYZ -> RGB.
 *
 * Here's the conversion from ITULAB -> XYZ:
 *	X =  ((a + 128.5) / 750.0 + (L + 40.3) / 295.8) ^ 3
 *	Y =  ((L + 40.3) / 295.8) ^ 3
 *	Z =  ((96.5 - b) / 255.0 + (L + 40.3) / 295.8) ^ 3
 * And here's the tranformation from XYZ -> RGB:
 *	R = ( 2.93225894531 * X - 1.45508080078 * Y - 0.45502558593 * Z) * 256
 *	G = (-1.09070074218 * X + 2.01783791015 * Y + 0.03588427734 * Z) * 256
 *	B = ( 0.07755386718 * X - 0.27130406250 * Y + 1.44614982421 * Z) * 256
 *		each bounded from 0 to 255
 *
 */

/*
 * Initialize tables for ITULAB->RGB colorspace conversion.
 */

LOCAL(void)
build_lab_rgb_table (
    j_decompress_ptr cinfo)
{
    Int32 i;
    float f;
    Int32 *lab_rgb_tab;

    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;

    /* Allocate and fill in the conversion table. */
	lab_rgb_tab = (Int32 *)((*cinfo->mem->alloc_small)((j_common_ptr) cinfo, JPOOL_IMAGE, ((N_F_INV + 1) * SIZEOF(Int32)));
    cconvert->lab_rgb_tab = lab_rgb_tab;

    for (i = 0; i < N_F_INV+1; i++) {
		f = (float)i / (float)(N_F_INV - 1);
		if (f < 0.20689303f) {
			f = ((f) - 16.0f / 116.0f) / 7.787f;
		} else {
			f = f * f * f;
		}
		lab_rgb_tab[i] = (Int32)(f * (float)KFOUR);
    }

    return;
}


/*
 * Convert some rows of samples to the output colorspace.
 *
 * Note that we change from noninterleaved, one-plane-per-component format
 * to interleaved-pixel format.  The output buffer is therefore three times
 * as wide as the input buffer.
 * A starting row offset is provided only for the input buffer.  The caller
 * can easily adjust the passed output_buf value to accommodate any row
 * offset required on that side.
 */

METHODDEF(void)
itulab_rgb_convert (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int		num_rows)
{
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
    register Int32 * ctab = cconvert->lab_rgb_tab;
    register JSAMPROW outptr;
    register JSAMPROW inptr0, inptr1, inptr2;
    register JDIMENSION col;
    JDIMENSION num_cols = cinfo->output_width;
    UInt8 L, a, b;
    Int32 X, Y, Z, R, G, B;
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn = 0.96422f, Yn = 1.000f, Zn = 0.82521f;
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    Int32 const1,const2,const3,const4,const5,const6;
    Int32 c, t;
	dgp_other_dest_buf destination = dest_buf.other_buf;
	char** output_buf = &(destination.buffer);

    SHIFT_TEMPS

    /* If a different gammut exists, grab it */
    if (cinfo->Lmax - cinfo->Lmin > 0) {
	Lmin = cinfo->Lmin;
	Lmax = cinfo->Lmax;
	if (cinfo->amax - cinfo->amin > 0) {
	    amin = cinfo->amin;
	    amax = cinfo->amax;
	    if (cinfo->bmax - cinfo->bmin > 0) {
		bmin = cinfo->bmin;
		bmax = cinfo->bmax;
	    }
	}
    }

    /* Precompute the constants */
    const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Int32)((amin / 500.0f) * KTHREE);
    const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Int32)((-bmin / 200.0f) * KTHREE);

    r1 = (Int32)(( 2.93225894531f * 256.0f * Xn) * KFIVE);
    g1 = (Int32)((-1.09070074218f * 256.0f * Xn) * KFIVE);
    b1 = (Int32)(( 0.07755386718f * 256.0f * Xn) * KFIVE);
    r2 = (Int32)((-1.45508080078f * 256.0f * Yn) * KFIVE);
    g2 = (Int32)(( 2.01783791015f * 256.0f * Yn) * KFIVE);
    b2 = (Int32)((-0.27130406250f * 256.0f * Yn) * KFIVE);
    r3 = (Int32)((-0.45502558593f * 256.0f * Zn) * KFIVE);
    g3 = (Int32)(( 0.03588427734f * 256.0f * Zn) * KFIVE);
    b3 = (Int32)(( 1.44614982421f * 256.0f * Zn) * KFIVE);

    while (--num_rows >= 0) {
	inptr0 = input_buf[0][input_row];
	inptr1 = input_buf[1][input_row];
	inptr2 = input_buf[2][input_row];
	input_row++;
	outptr = *output_buf++;
	for (col = 0; col < num_cols; col++) {
	    L = GETJSAMPLE(inptr0[col]);
	    a = GETJSAMPLE(inptr1[col]);
	    b = GETJSAMPLE(inptr2[col]);

	    /* Convert from ITULAB -> XYZ: */
	    c = const1 * L + const2;
	    t = const3 * a + const4 + c;
	    X = f_inv(t);
	    Y = f_inv(c);
	    t = c - b * const5 + const6;
	    Z = f_inv(t);

	    /* Convert from XYZ to RGB */
	    R = r1 * X + r2 * Y + r3 * Z;
	    G = g1 * X + g2 * Y + g3 * Z;
	    B = b1 * X + b2 * Y + b3 * Z;

	    /* Rail and integerize to 8-bit */
	    outptr[RGB_RED] = EIGHTBITRGB(R);
	    outptr[RGB_GREEN] = EIGHTBITRGB(G);
	    outptr[RGB_BLUE] = EIGHTBITRGB(B);
	    outptr += RGB_PIXELSIZE;
	}
    }

    return;
}

#endif /* !USELINEARRGB */

/*
 * ITULAB conversion for grayscale: simple scale and offset.
 * Note, in the case of full gammut range for L*, we'll by-pass
 * this routine up-stream, and just use grayscale_convert().
 */

    /* TODO:
     * make use of the precomputed table, and find the mapping from each
     * byte [0,255] to L* [Lmin,Lmax].
     * What would really be useful is if we actually implemented the 12-bit
     * version, in which case a gammut range might actually make sense.
     *
     * However, this is not a real costly loop, we're probably just as well
     * off to multiply integers and do some bit shifts.
     */

METHODDEF(void) 
itulab_grayscale_convert (
    j_decompress_ptr cinfo,
    JSAMPIMAGE	input_buf,
    JDIMENSION	input_row,
	dgp_dest_buf dest_buf,
	int row_cnt,
    int		num_rows)
{
  register JSAMPROW outptr;
  register JSAMPROW inptr;
  register JDIMENSION col;
  register float fact;
  JDIMENSION num_cols = cinfo->output_width;
  float Lmin = 0.0f, Lmax = 100.0f;
  float L;
  dgp_other_dest_buf destination = dest_buf.other_buf;
  char** output_buf = &(destination.buffer);

  SHIFT_TEMPS

  if (cinfo->Lmax - cinfo->Lmin > 0) {
      Lmin = cinfo->Lmin;
      Lmax = cinfo->Lmax;
  }
  fact = (Lmax - Lmin) / 100.0f;

  while (--num_rows >= 0) {
    inptr = input_buf[0][input_row];
    input_row++;
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++) {
      L = (float)GETJSAMPLE(inptr[col]);

      /* Reduce the intensity range based on the gammut range */
      L = Lmin + L * fact;

      /* Bound */
      *outptr++ = EIGHTBIT(L);
    }
  }

  return;
}

/*
 * Module initialization routine for output colorspace conversion.
 */

GLOBAL(void)
jinit_color_deconverter (j_decompress_ptr cinfo)
{
  my_cconvert_ptr cconvert;
  int ci;

  cconvert = (my_cconvert_ptr)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				SIZEOF(my_color_deconverter));
  cinfo->cconvert = (struct jpeg_color_deconverter *) cconvert;
  cconvert->pub.start_pass = start_pass_dcolor;


  /* Make sure num_components agrees with jpeg_color_space */
  switch (cinfo->jpeg_color_space) {
  case JCS_GRAYSCALE:
    if (cinfo->num_components != 1)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    break;

  case JCS_RGB:
  case JCS_YCbCr:
  case JCS_CIELAB:
  case JCS_ICCLAB:
  case JCS_ITULAB:
    if (cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    break;

#if 0
  case JCS_ITULAB:
    if (cinfo->num_components != 1 && cinfo->num_components != 3)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);

    /* If the gammut is full range, it's really just grayscale */
    if (cinfo->num_components == 1
		&& cinfo->Lmin < (0.5 / 255.0)
		&& cinfo->Lmax > (99 + 0.5 / 255.0)) {
		ERR_PASS(jd_force_grayscale_output(pIPState, cinfo));
    }
    break;
#endif /* 0 */

  case JCS_CMYK:
  case JCS_YCCK:
    if (cinfo->num_components != 4)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    break;

  default:			/* JCS_UNKNOWN can be anything */
    if (cinfo->num_components < 1)
      ERREXIT(cinfo, JERR_BAD_J_COLORSPACE);
    break;
  }

  /* Set out_color_components and conversion method based on requested space.
   * Also clear the component_needed flags for any unused components,
   * so that earlier pipeline stages can avoid useless computation.
   */

  switch (cinfo->out_color_space) {
  case JCS_GRAYSCALE:
    cinfo->out_color_components = 1;
    switch (cinfo->jpeg_color_space) {
    case JCS_GRAYSCALE:
	case JCS_YCbCr:
	case JCS_CIELAB:
	case JCS_ICCLAB:
      cconvert->pub.color_convert = grayscale_convert;
      /* For color->grayscale conversion, only the Y (0) component is needed */
      for (ci = 1; ci < cinfo->num_components; ci++)
		  cinfo->comp_info[ci].component_needed = FALSE;
	  break;
    case JCS_ITULAB:
      cconvert->pub.color_convert = itulab_grayscale_convert;
      break;
    default:
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    }
    break;

  case JCS_RGB:
    cinfo->out_color_components = RGB_PIXELSIZE;
    if (cinfo->jpeg_color_space == JCS_YCbCr) {

      switch(cinfo->output_orientation){
        case ORIENT_RD:
          cconvert->pub.color_convert = ycc_rgb_convert;
          break;
        case ORIENT_DL:
          /*cconvert->pub.color_convert = ycc_rgb_convert_90; */
          cconvert->pub.color_convert = ycc_rgb_convert_270;
          break;
        case ORIENT_LU:
          cconvert->pub.color_convert = ycc_rgb_convert_180;
          break;
        case ORIENT_UR:
          cconvert->pub.color_convert = ycc_rgb_convert_90;
          /*cconvert->pub.color_convert = ycc_rgb_convert_270; */
          break;
        default:
          cconvert->pub.color_convert = ycc_rgb_convert;
          break;
       }/*end of switch statement */

       build_ycc_rgb_table(cinfo);
    } else if (cinfo->jpeg_color_space == JCS_GRAYSCALE) {
      cconvert->pub.color_convert = gray_rgb_convert;
    } else if (cinfo->jpeg_color_space == JCS_CIELAB) {
      cconvert->pub.color_convert = icclab_rgb_convert;
      build_lab_rgb_table (cinfo);
    } else if (cinfo->jpeg_color_space == JCS_ICCLAB) {
      cconvert->pub.color_convert = icclab_rgb_convert;
      build_lab_rgb_table (cinfo);
    } else if (cinfo->jpeg_color_space == JCS_ITULAB) {
      switch(cinfo->output_orientation){
        case ORIENT_RD:
          cconvert->pub.color_convert = itulab_rgb_convert_0;
          break;
        case ORIENT_DL:
          cconvert->pub.color_convert = itulab_rgb_convert_270;
          break;
        case ORIENT_LU:
          cconvert->pub.color_convert = itulab_rgb_convert_180;
          break;
        case ORIENT_UR:
          cconvert->pub.color_convert = itulab_rgb_convert_90;
          break;
        default:
          cconvert->pub.color_convert = itulab_rgb_convert_0;
          break;
       }/*end of switch statement */
      build_lab_rgb_table (cinfo);
    } else if (cinfo->jpeg_color_space == JCS_RGB && RGB_PIXELSIZE == 3) {
      cconvert->pub.color_convert = null_convert;
    } else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  case JCS_CMYK:
    cinfo->out_color_components = 3; 
    if (cinfo->jpeg_color_space == JCS_YCCK) {

      switch(cinfo->output_orientation){
        case ORIENT_RD:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert;
          break;
        case ORIENT_DL:
          /*cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_90; */
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_270;
          break;
        case ORIENT_LU:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_180;
          break;
        case ORIENT_UR:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_90;
          /*cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_270; */
          break;
        default:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert;
          break;
       }/*end of switch statement */
      build_ycc_rgb_table(cinfo);

    } else if (cinfo->jpeg_color_space == JCS_CMYK) {
      cconvert->pub.color_convert = null_convert;
    } else if( cinfo->jpeg_color_space == JCS_YCbCr){ /*Ranjani.new */

      switch(cinfo->output_orientation){
        case ORIENT_RD:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert;
          break;
        case ORIENT_DL:
          /*cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_90; */
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_270;
          break;
        case ORIENT_LU:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_180;
          break;
        case ORIENT_UR:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_90;
          /*cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert_270; */
          break;
        default:
          cconvert->pub.color_convert = ycbcr_ycck_cmyk_convert;
          break;
       }/*end of switch statement */

      build_ycc_rgb_table(cinfo);
    } else
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;

  default:
    /* Permit null conversion to same output space */
    if (cinfo->out_color_space == cinfo->jpeg_color_space) {
      cinfo->out_color_components = cinfo->num_components;
      cconvert->pub.color_convert = null_convert;
    } else			/* unsupported non-null conversion */
      ERREXIT(cinfo, JERR_CONVERSION_NOTIMPL);
    break;
  }

  if (cinfo->quantize_colors)
    cinfo->output_components = 1; /* single colormapped output component */
  else
    cinfo->output_components = cinfo->out_color_components;
}

/* Undefine constants used for masking. */
#undef DGP_OPAQUE
#undef DGP_TRANSPARENT
