/*************************************************************************
**  Copyright 1999-2000 Xerox Corporation.All rights reserved.
**
**  DigiPaper Codebase
**    
**  Ranjani Ramamurthy, October 1999
**
**  Filename: jpeg_interface.c
**
**  Description of what's implemented in this file:
**    Interface to the jpeg6b libraries are implemented. This is the
**    heart of the code.
**    The four functions that are exported by the jpeg6b library are
**    defined here. They are:
**
**    dgp_init_jpeg_vars
**    dgp_destroy_jpeg_vars
**    dgp_check_jpeg_render
**    dgp_render_jpeg
**
**
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#if !defined(VXWORKS)
#include <memory.h>
#endif /* (VXWORKS) */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeg6blib.h"
#include "jpeg_mem_consts.h"
#include "jerror.h"
#include "jversion.h"
#include "cdjpeg.h"	
#include <assert.h>

#include "coord.h"
#include "image.h"
#include "PsgFix.h"

extern void
InitDGPUdsSrcMgr(j_decompress_ptr cinfo,
                 uds* in_uds,
		 int size);
extern void
InitDGPMemSrcMgr(j_decompress_ptr cinfo,
                 Dgp_UInt8 * buf,
		 int size);
extern void
DGP_jpeg_buf_dest(j_compress_ptr cinfo,
                  DGP_JPEGOutputDef* output_buf);

#ifdef MEM_DATA_COLLECTION
extern FILE* mem_file;
#endif

#ifdef CREATE_BUFFER_FILE
extern FILE* raw_buffer_fp;
#endif

/******************** STATIC FUNCTIONS ********************/

static ErrCode
dgp_translate_err(struct jpeg_error_mgr* err,Dgp_State* dgp_state)
{
 
  DGP_PROLOGUE("dgp_translate_err()",DGP_PERRSTATE);

  if( (err->msg_code > JERR_DUMMY_BEGIN_IO_MSGS) &&
      (err->msg_code < JERR_DUMMY_END_IO_MSGS)) {
    DGP_ERR_GENF0((err->num_errors > 0),cJpegIOErr,err->err_buffer);
    ERR_RETURN;
  } else if( err->msg_code == JERR_NO_SOI) {
    DGP_ERR_GENF0((err->num_errors > 0),cJpegNotAJpegFile,
                  err->err_buffer);
    ERR_RETURN;
  } else if( err->msg_code == JERR_OUT_OF_MEMORY) {
    DGP_ERR_GENF0((err->num_errors > 0),cJpegOutOfMemoryErr,
                  err->err_buffer);
    ERR_RETURN;
  } else {
    DGP_ERR_GENF0((err->num_errors > 0),cJpegInternalPanicErr,
                   err->err_buffer);
    ERR_RETURN;
  }
}

static ErrCode
dgp_translate_err_decomp(j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
  struct jpeg_error_mgr * err = cinfo->err;

  DGP_PROLOGUE("dgp_translate_err_decomp()",DGP_PERRSTATE);

  DGP_ERR_PASS(dgp_translate_err(err,dgp_state));

  NORMAL_RETURN;

}

static ErrCode
dgp_translate_err_comp(j_compress_ptr cinfo,Dgp_State* dgp_state)
{
  struct jpeg_error_mgr * err = cinfo->err;

  DGP_PROLOGUE("dgp_translate_err_comp()",DGP_PERRSTATE);

  DGP_ERR_PASS(dgp_translate_err(err,dgp_state));

  NORMAL_RETURN;
}

static ErrCode
dgp_internal_decomp_check(j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
  struct jpeg_error_mgr * err = cinfo->err;

  DGP_PROLOGUE("dgp_internal_decomp_check()",DGP_PERRSTATE);

  if(err->num_warnings > 0){
    DGP_WARNF0("JPEG data warning.Expect an incorrect image");
    NORMAL_RETURN;
  }

  if(err->num_errors > 0){
    DGP_ERR_PASS(dgp_translate_err_decomp(cinfo,dgp_state));
    ERR_RETURN;
  }

  NORMAL_RETURN;
}

static ErrCode
dgp_internal_comp_check(j_compress_ptr cinfo,Dgp_State* dgp_state)
{
  struct jpeg_error_mgr * err = cinfo->err;

  DGP_PROLOGUE("dgp_internal_comp_check()",DGP_PERRSTATE);

  if(err->num_warnings > 0){
    DGP_WARNF0("JPEG data warning.Expect an incorrect image");
    NORMAL_RETURN;
  }

  if(err->num_errors > 0){
    DGP_ERR_PASS(dgp_translate_err_comp(cinfo,dgp_state));
    ERR_RETURN;
  }

  NORMAL_RETURN;
}

#if 0
static ErrCode
dgp_jpeg_set_cinfo_vals(j_decompress_ptr cinfo,
               Dgp_State* dgp_state)
{
  DGP_PROLOGUE("dgp_jpeg_set_cinfo_vals()",DGP_PERRSTATE);
  DGP_STOP_WHINE;
  cinfo->output_width = cinfo->crop_width;
  cinfo->output_height = cinfo->crop_height;

  NORMAL_RETURN;
}
#endif

static ErrCode
dgp_jpeg_assert_vals(DGP_JPEGImageDef *image_def, Dgp_State *dgp_state)
{

  DGP_PROLOGUE("dgp_jpeg_assert_vals()",DGP_PERRSTATE);
  DGP_STOP_WHINE;

  image_def = image_def;

#ifndef STANDALONE_TESTING
  DGP_ASSERT(image_def->imgCoOrds.bytesPerRow >= 0);
#endif
  DGP_ASSERT(image_def->imgCoOrds.dstH >= 0);
  DGP_ASSERT(image_def->imgCoOrds.dstW >= 0);
  DGP_ASSERT(image_def->imgCoOrds.dstMinX >= 0);
  DGP_ASSERT(image_def->imgCoOrds.dstMinY >= 0);
  DGP_ASSERT(image_def->imgCoOrds.dstX >= 0);
  DGP_ASSERT(image_def->imgCoOrds.dstY >= 0);
  DGP_ASSERT(image_def->imgCoOrds.srcOriginX >= 0);
  DGP_ASSERT(image_def->imgCoOrds.srcOriginY >= 0);
  DGP_ASSERT(image_def->imgSrc.size > 0);
  /*
  These quantities do not correspond exactly to their semantic counterparts in libsanta,
  so keep a sense of humor when dealing with these.

  srcOrigins are in page coordinates, whereas dsts are in
  image coordinates.
  / dhtesq.
  */
  DGP_ASSERT((image_def->imgCoOrds.dstX - 
          image_def->imgCoOrds.srcOriginX) >= 0);
  DGP_ASSERT((image_def->imgCoOrds.dstY - 
          image_def->imgCoOrds.srcOriginY) >= 0);
  
  DGP_ASSERT((image_def->imgCoOrds.dstX - 
          image_def->imgCoOrds.dstMinX) >= 0);
  DGP_ASSERT((image_def->imgCoOrds.dstY - 
          image_def->imgCoOrds.dstMinY) >= 0);
  /*check resolution ROTATION_SUPPORT */
  DGP_ASSERT(DgpCoord_equal(image_def->imgResolution.srcRes.xRes,
			    image_def->imgResolution.srcRes.yRes));
  DGP_ASSERT(DgpCoord_equal(image_def->imgResolution.destRes.xRes,
			    image_def->imgResolution.destRes.yRes));
  NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_recheck_jpeg_render(DGP_JPEGImageDef* image_def,
				  Dgp_UInt32* ret_val,
                  j_decompress_ptr cinfo, Dgp_State* dgp_state)
{

  Dgp_UInt32 ret = 0;
  /*struct jpeg_error_mgr * err = cinfo->err; */

  DGP_PROLOGUE("dgp_jpeg_recheck_jpeg_render()",DGP_PERRSTATE);

  if(cinfo->src == NULL) {
     if(image_def->imgSrc.u)
         InitDGPUdsSrcMgr(cinfo,image_def->imgSrc.u,
                            image_def->imgSrc.size);
     else if (image_def->imgSrc.buf)
         InitDGPMemSrcMgr(cinfo,image_def->imgSrc.buf,
                            image_def->imgSrc.size);
     else DGP_ERR_GEN( TRUE, cJpegIOErr);
  }

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

  DGP_ERR_PASS(dgp_jpeg_assert_vals(image_def,dgp_state));

  if( (cinfo->image_height == 0) && (cinfo->image_width == 0)){
   (void) jpeg_read_header(cinfo,DGP_TRUE);
   DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));
  } 

  /*check type */
  switch(image_def->imgType){
    case DGP_RGB:
    case DGP_CMYK:
    case DGP_C:
    case DGP_M:
    case DGP_Y:
    case DGP_K:
      break;
    case DGP_PACKED:
    default:
       ret |= DGP_IMAGE_TYPE_BIT;
       break;
  }

  switch(image_def->imgOrientation){
    case ORIENT_RD: /*no rotation */
	case ORIENT_DL: /*rotation clockwise by 90 degrees */
    case ORIENT_LU: /*rotation clockwise by 180 degrees */
    case ORIENT_UR: /*rotation clockwise by 270 degrees */
       cinfo->output_orientation = image_def->imgOrientation;
       break;
    default:
       ret |= DGP_ORIENTATION_BIT;
  }

  *(ret_val) = ret;
  NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_check_coords(DGP_JPEGImageDef* image_def, Dgp_Bool* ret_val,
             Dgp_State* dgp_state)
{
  Dgp_Bool image_def_changed = DGP_FALSE;

  DGP_PROLOGUE("dgp_jpeg_check_coords()",DGP_PERRSTATE);
  DGP_STOP_WHINE;

  if(image_def->imgCoOrds.dstH < 0){
    image_def->imgCoOrds.dstH = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.dstW < 0){
    image_def->imgCoOrds.dstW = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.dstMinX < 0){
    image_def->imgCoOrds.dstMinX = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.dstMinY < 0){
    image_def->imgCoOrds.dstMinY = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.dstX < 0){
    image_def->imgCoOrds.dstX = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.dstY < 0){
    image_def->imgCoOrds.dstY = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.srcOriginX < 0){
    image_def->imgCoOrds.srcOriginX = 0;
    image_def_changed = DGP_TRUE;
  }
  if(image_def->imgCoOrds.srcOriginY < 0){
    image_def->imgCoOrds.srcOriginY = 0;
    image_def_changed = DGP_TRUE;
  }
  *(ret_val) = image_def_changed;
  NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_check_coord_diff(DGP_JPEGImageDef* image_def,Dgp_Bool* ret_val,
                 Dgp_State* dgp_state)
{
  Dgp_Bool invalid_diff = DGP_FALSE;

  DGP_PROLOGUE("dgp_jpeg_check_coord_diff()",DGP_PERRSTATE);
  DGP_STOP_WHINE;

  /*
  srcOriginX and srcOriginY have been defined in santa in terms of image (not page) coordinates,
  so I believe this comparison is invalid.  See slbz_determine_desired_img_portion().

  I'm trying a workaround by redefining these quantities in libzephyr/jpeg.c.  This at least
  keeps the asserts from triggering.
  / dhtesq.
  */
  if((image_def->imgCoOrds.dstX - 
          image_def->imgCoOrds.srcOriginX) < 0){
    image_def->imgCoOrds.dstX = DGP_INVALID_COORD;
    image_def->imgCoOrds.srcOriginX = 
                           DGP_INVALID_COORD;
    invalid_diff = DGP_TRUE;
  }
  if((image_def->imgCoOrds.dstY - 
          image_def->imgCoOrds.srcOriginY) < 0){
    image_def->imgCoOrds.dstY = DGP_INVALID_COORD;
    image_def->imgCoOrds.srcOriginY = 
                           DGP_INVALID_COORD;
    invalid_diff = DGP_TRUE;
  }
  
  if((image_def->imgCoOrds.dstX - 
          image_def->imgCoOrds.dstMinX) < 0){
    image_def->imgCoOrds.dstX = DGP_INVALID_COORD;
    image_def->imgCoOrds.dstMinX = 
                           DGP_INVALID_COORD;
    invalid_diff = DGP_TRUE;
  }
  if((image_def->imgCoOrds.dstY - 
          image_def->imgCoOrds.dstMinY) < 0){
    image_def->imgCoOrds.dstY = DGP_INVALID_COORD;
    image_def->imgCoOrds.dstMinY = 
                           DGP_INVALID_COORD;
    invalid_diff = DGP_TRUE;
  }
  *(ret_val) = invalid_diff;

  NORMAL_RETURN;

}

/************************************************************
* Look at jpeg_mem_consts.h for an explanation of how this *
* memory estimate is done. jpeg_memory_consts.h contains *
* data out of libjpeg.doc as well as tests I ran with the *
* MEM_DATA_COLLECTION flag set. For later versions of jpeg *
* please be sure to look at memory requirements documentation *
* closely. */

static ErrCode
dgp_estimate_memory(j_decompress_ptr cinfo,
    DGP_JPEGImageDef* image_def,Dgp_State* dgp_state)
{
  Dgp_Int32 base_mem=0;

  DGP_PROLOGUE("dgp_estimate_memory()",DGP_PERRSTATE);
  DGP_STOP_WHINE;

  base_mem = DGP_JPEG_BASE_MEMORY + 
             DGP_JPEG_IDCT_STRIP * cinfo->image_width ; /*in bytes */
  image_def->memData.mem_estimate = base_mem;

  if(cinfo->progressive_mode){
    /*The image is a progressive jpeg image */
    image_def->memData.jpeg_type = DGP_PROGRESSIVE;
    if((cinfo->max_h_samp_factor == 1) &&
       (cinfo->max_v_samp_factor == 1)){
      /*worst case sampling.. 6 bytes per pixel of image */
      image_def->memData.mem_estimate +=
          DGP_WORST_CASE_SAMPLING * 
          (cinfo->image_height * cinfo->image_width);
    } else {
      image_def->memData.mem_estimate +=
          DGP_AVERAGE_CASE_SAMPLING * 
          (cinfo->image_height * cinfo->image_width);
    }
  } else {
    image_def->memData.jpeg_type = DGP_BASELINE;
  }
  /*Now add overhead - as the smaller chunks are */
  /*not taken into account in this computation */

  image_def->memData.mem_estimate +=
    (Dgp_Int32)(DGP_ALLOC_SMALL_OVERHEAD * image_def->memData.mem_estimate);

  NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_check_input_vals(DGP_JPEGImageDef* image_def, 
                 Dgp_UInt32* ret_val, Dgp_State* dgp_state)
{
  Dgp_Bool invalid_coord = DGP_FALSE;
  Dgp_Bool invalid_diff=DGP_FALSE;
  Dgp_Bool invalid_resolution = DGP_FALSE;
  /*Dgp_Bool invalid_jpeg_sz = DGP_FALSE; */
  Dgp_UInt32 ret=0;

  DGP_PROLOGUE("dgp_jpeg_check_input_vals()",DGP_PERRSTATE);
#ifndef STANDALONE_TESTING
  DGP_ASSERT(image_def->imgCoOrds.bytesPerRow >= 0);
#endif

  DGP_ERR_PASS(dgp_jpeg_check_coords(image_def,&invalid_coord,dgp_state));

  if(image_def->imgSrc.size <= 0){
    ret  |= DGP_SZ_BIT;
  }

  DGP_ERR_PASS(dgp_jpeg_check_coord_diff(image_def,&invalid_diff,dgp_state));
  
  /*check resolution ROTATION_SUPPORT */
  if(!DgpCoord_equal(image_def->imgResolution.srcRes.xRes,
			    image_def->imgResolution.srcRes.yRes)){
    invalid_resolution = DGP_TRUE;
    ret |= DGP_RESOLUTION_BIT;
  }
  if(!DgpCoord_equal(image_def->imgResolution.destRes.xRes,
			    image_def->imgResolution.destRes.yRes)){
    invalid_resolution = DGP_TRUE;
    ret |= DGP_RESOLUTION_BIT;
  }
  /* Make sure the scale is in integer times. Both upscale and down scale */
  if(image_def->imgResolution.destRes.xRes.den ==
			    image_def->imgResolution.srcRes.xRes.den &&
     (image_def->imgResolution.destRes.xRes.num % 
			    image_def->imgResolution.srcRes.xRes.num != 0 &&
       image_def->imgResolution.srcRes.xRes.num % 
			    image_def->imgResolution.destRes.xRes.num != 0 )){
    invalid_resolution = DGP_TRUE;
    ret |= DGP_RESOLUTION_BIT;
  }

  if(invalid_coord)
    ret |= DGP_COORD_BIT;
  if(invalid_diff)
    ret |= DGP_COORD_DIFF_BIT;

  *(ret_val) = ret;

  NORMAL_RETURN;
}

#ifdef STANDALONE_TESTING
static ErrCode
dgp_jpeg_convert_buffer_normal(Dgp_Int8** buffer,dgp_dest_buf* in_dest, 
                      int act_img_bpr, int img_type,Dgp_State* dgp_state)
{
   Dgp_Int32 i,j; Dgp_Int8* local_buf;
   Dgp_Int8 c,m,y,k;
#ifdef CREATE_BUFFER_FILE
   int fwrite_ret,fflush_ret;
#endif  

   DGP_PROLOGUE("dgp_jpeg_convert_buffer_normal()",DGP_PERRSTATE);

   local_buf = *(buffer); 
   if(img_type == DGP_CMYK){
     for(i = 0,j=0; i < act_img_bpr, j < (act_img_bpr * 3);i++, j+=3) /*was 4 */
     {
       c = in_dest->cmyk_buf.c[i];
       m = in_dest->cmyk_buf.m[i];
       y = in_dest->cmyk_buf.y[i];
       k = in_dest->cmyk_buf.k[i];
       local_buf[j] = 255 - (c+k); /*r value */
       local_buf[j+1] = 255 - (m+k); /*g value */
       local_buf[j+2] = 255 - (y+k); /* b value */
#ifdef CREATE_BUFFER_FILE
     /* Create a raw CMYK buffer  */
     if(raw_buffer_fp != NULL){
       /*fwrite(c,sizeof(Dgp_Int8),1,raw_buffer_fp);
       fwrite(m,sizeof(Dgp_Int8),1,raw_buffer_fp);
       fwrite(y,sizeof(Dgp_Int8),1,raw_buffer_fp);
       fwrite(k,sizeof(Dgp_Int8),1,raw_buffer_fp);*/
       fputc(c,raw_buffer_fp);fputc(m,raw_buffer_fp);
       fputc(y,raw_buffer_fp); fputc(k,raw_buffer_fp);
       fflush_ret = fflush(raw_buffer_fp);
       if(fflush_ret != 0){
         ERR_RETURN;
       }
     }
#endif
     }
   } else if(img_type == DGP_RGB){
     memcpy(local_buf,in_dest->other_buf.buffer,act_img_bpr);
#ifdef CREATE_BUFFER_FILE
     /* Create a raw RGB buffer */
     if(raw_buffer_fp != NULL){
       fwrite_ret = fwrite(local_buf,1,act_img_bpr,raw_buffer_fp);
       if(fwrite_ret != act_img_bpr){
         ERR_RETURN;
       }
       fflush_ret = fflush(raw_buffer_fp);
       if(fflush_ret != 0){
         ERR_RETURN;
       }
     }
#endif
   }

   NORMAL_RETURN;
}
#endif

static ErrCode
dgp_jpeg_shift_to_x_offset(dgp_img_type img_type, dgp_dest_buf *in_dest,
                  Dgp_Int32 offset, Dgp_State* dgp_state)
{
    Dgp_Int8 **local_buf, **c, **m, **y, **k;

    DGP_PROLOGUE("dgp_jpeg_shift_to_x_offset()",DGP_PERRSTATE);
    DGP_STOP_WHINE;
    switch(img_type){
      case DGP_RGB:
        local_buf = &(in_dest->other_buf.buffer);
        *(local_buf) += offset;
        break;
      case DGP_CMYK:
        c = (Dgp_Int8 **)&(in_dest->cmyk_buf.c) ; *c += offset;
        m = (Dgp_Int8 **)&(in_dest->cmyk_buf.m) ; *m += offset;
        y = (Dgp_Int8 **)&(in_dest->cmyk_buf.y) ; *y += offset;
        k = (Dgp_Int8 **)&(in_dest->cmyk_buf.k) ; *k += offset;
        break;
      case DGP_C:
        c =(Dgp_Int8 **)&(in_dest->cmyk_buf.c) ; *c += offset;
        break;
      case DGP_M:
        m = (Dgp_Int8 **)&(in_dest->cmyk_buf.m) ; *m += offset;
        break;
      case DGP_Y:
        y = (Dgp_Int8 **)&(in_dest->cmyk_buf.y) ; *y += offset;
        break;
      case DGP_K:
        k = (Dgp_Int8 **)&(in_dest->cmyk_buf.k) ; *k += offset;
        break;
	  default:
	    break;
    }
    NORMAL_RETURN;
}

#if 0
static ErrCode
dgp_jpeg_shift_to_row(dgp_img_type img_type,dgp_dest_buf* in_dest,
             Dgp_Int32 offset,Dgp_State* dgp_state)
{
    DGP_PROLOGUE("dgp_jpeg_shift_to_row()",DGP_PERRSTATE);
    DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(img_type,in_dest,offset,dgp_state));
    NORMAL_RETURN;
}
#endif

static ErrCode
dgp_jpeg_shift_to_x_offset_rotated(DGP_JPEGImageDef* image_def,
          dgp_dest_buf* in_dest,Dgp_Int32 write_rownum,Dgp_State* dgp_state)
{

    Dgp_Int32 out_color_components, offset = 0;
    DGP_PROLOGUE("dgp_jpeg_shift_to_x_offset_rotated()",DGP_PERRSTATE);
    
    if(image_def->imgType == DGP_RGB) {
      out_color_components = 3;
    } else {
      out_color_components = 1;
    }

    switch(image_def->imgOrientation){
      case ORIENT_DL: /*(not 90) 270 */
        offset = out_color_components *
          ((image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)
          + write_rownum);
        break;
      case ORIENT_UR: /*(not 270) 90 */
        offset = out_color_components *
          ((image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)
         +(image_def->imgCoOrds.dstH - 1 - write_rownum));
        break;
      case ORIENT_RD: /*no rotation */
      case ORIENT_LU:
      default:
        offset = out_color_components *
         (image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX);
        break;
    }
    DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,offset,dgp_state));

    NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_shift_to_end_of_row(DGP_JPEGImageDef* image_def,
          dgp_dest_buf* in_dest,
          j_decompress_ptr cinfo,
	  Dgp_State* dgp_state)
{
    Dgp_Int32 out_color_components,sub_val=0;

    DGP_PROLOGUE("dgp_jpeg_shift_to_end_of_row()",DGP_PERRSTATE);
    cinfo = cinfo;

    if(image_def->imgType == DGP_RGB)
      out_color_components = 3;
    else
      out_color_components = 1;

   switch(image_def->imgOrientation){

     case ORIENT_DL: /*rotation by 270 */
     case ORIENT_UR: /*rotation by 90 */
       /*For these two cases, do nothing.. */
       break;
     case ORIENT_LU: /*rotation by 180 */
       sub_val = (image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)
                 *out_color_components + image_def->imgCoOrds.bytesPerRow;
       DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest, -sub_val,dgp_state));

       if(image_def->imgType == DGP_CMYK){
           in_dest->cmyk_buf.trans_mask -= in_dest->cmyk_buf.trans_bpl / sizeof(Dgp_UInt32);
       } else {
           in_dest->other_buf.trans_mask -= in_dest->other_buf.trans_bpl / sizeof(Dgp_UInt32);
       }
       break;
     case ORIENT_RD: /*no rotation */
     default:
       DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,
          (image_def->imgCoOrds.bytesPerRow -
          (image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)*
           out_color_components),dgp_state));

       if(image_def->imgType == DGP_CMYK){
           in_dest->cmyk_buf.trans_mask += in_dest->cmyk_buf.trans_bpl / sizeof(Dgp_UInt32);
       } else {
           in_dest->other_buf.trans_mask += in_dest->other_buf.trans_bpl / sizeof(Dgp_UInt32);
       }
       break;
   }
   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_shift_to_init_offset_180(DGP_JPEGImageDef* image_def,
        dgp_dest_buf* in_dest,Dgp_Int32 bytes_per_row,Dgp_State* dgp_state)
{
   Dgp_Int32 offset=0;
   DGP_PROLOGUE("dgp_jpeg_shift_to_init_offset_180()",DGP_PERRSTATE);
   /*Point it to the beginning of the last row */
   offset = bytes_per_row *
                    ((image_def->imgCoOrds.dstY -
                      image_def->imgCoOrds.dstMinY) +
                      (image_def->imgCoOrds.dstH-1));

   DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,
                                 offset,dgp_state));
   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_shift_to_init_offset_default(DGP_JPEGImageDef* image_def,
        dgp_dest_buf* in_dest, Dgp_Int32 bytes_per_row, Dgp_State* dgp_state)
{
   Dgp_Int32 offset = 0;
   DGP_PROLOGUE("dgp_jpeg_shift_to_init_offset_default()",DGP_PERRSTATE);
   offset = bytes_per_row * (image_def->imgCoOrds.dstY - image_def->imgCoOrds.dstMinY);
   DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest, offset, dgp_state));
   NORMAL_RETURN;
}

/*
13.7.2000 / dhtesq.
Changed behavior for DL, UR, LU cases;  didn't work with banding.
*/
static ErrCode
dgp_jpeg_shift_to_initial_row(DGP_JPEGImageDef* image_def,
        dgp_dest_buf* in_dest,Dgp_Int32 bytes_in_jpeg_row,Dgp_State* dgp_state){
  /*Dgp_Int32 offset = 0; */
   DGP_PROLOGUE("dgp_jpeg_shift_to_initial_row()",DGP_PERRSTATE);
   switch(image_def->imgOrientation){
      case ORIENT_LU: /*rotation cw by 180 */
        DGP_ERR_PASS(dgp_jpeg_shift_to_init_offset_180(image_def,in_dest,
                                bytes_in_jpeg_row,dgp_state));
        break;
      case ORIENT_DL:
          /*
          offset = bytes_in_jpeg_row * (image_def->imgCoOrds.dstY -
                              image_def->imgCoOrds.dstMinX);
          DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,
                                  offset,dgp_state));
          break;
            */
      case ORIENT_UR: /*rotation cw by 90 */
          /*
          offset = bytes_in_jpeg_row * (image_def->imgCoOrds.dstY -
                              image_def->imgCoOrds.dstMinX);
          DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,
                                  offset,dgp_state));
          break;
            */
      case ORIENT_RD: /*Normal orientation */
      default:
        DGP_ERR_PASS(dgp_jpeg_shift_to_init_offset_default(image_def,in_dest,
                     bytes_in_jpeg_row,dgp_state));
        break;
   }
   NORMAL_RETURN;
}

#if 0
static ErrCode
dgp_jpeg_set_addresses_to_beg_of_img(DGP_JPEGImageDef* image_def,
        dgp_dest_buf* in_dest,Dgp_Int32 bytes_in_jpeg_row,Dgp_State* dgp_state)
{
   /*Never mind the orientation. Treat it like a normal */
   /*unrotated image - only then will the right thing get */
   /*written out. */
   DGP_PROLOGUE("dgp_jpeg_set_addresses_to_beg_of_img()",DGP_PERRSTATE);
   DGP_ERR_PASS(dgp_jpeg_shift_to_row(image_def->imgType, in_dest,
            (bytes_in_jpeg_row * 
            (image_def->imgCoOrds.dstY - 
             image_def->imgCoOrds.dstMinY)),dgp_state));       
   NORMAL_RETURN;
}
#endif

#if 0
static ErrCode
dgp_jpeg_shift_to_end_of_row_testing(DGP_JPEGImageDef* image_def,
          dgp_dest_buf* in_dest,
          j_decompress_ptr cinfo, Dgp_State* dgp_state){

    Dgp_Int32 out_color_components;

    DGP_PROLOGUE("dgp_jpeg_shift_to_end_of_row_testing()",DGP_PERRSTATE);
    cinfo = cinfo;

    if(image_def->imgType == DGP_RGB)
      out_color_components = 3;
    else
      out_color_components = 1;
    
    DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType,in_dest,
          (image_def->imgCoOrds.bytesPerRow -
          (image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)*
                                           out_color_components
          ),dgp_state));
    NORMAL_RETURN;
}
#endif

/*--------------------------------------------------------------*/
static ErrCode
dgp_jpeg_set_address(dgp_img_type img_type,dgp_dest_buf* in_dest,
   Dgp_Int8** other_ptr, Dgp_Int8** c_ptr, Dgp_Int8** m_ptr, Dgp_Int8** y_ptr,
   Dgp_Int8** k_ptr,Dgp_State* dgp_state)
{
   DGP_PROLOGUE("dgp_jpeg_set_address()",DGP_PERRSTATE);
   DGP_STOP_WHINE;
   switch(img_type){
     case DGP_RGB:
       *(other_ptr) = in_dest->other_buf.buffer;
       break;
     case DGP_CMYK:
       *(c_ptr) = in_dest->cmyk_buf.c;
       *(m_ptr) = in_dest->cmyk_buf.m;
       *(y_ptr) = in_dest->cmyk_buf.y;
       *(k_ptr) = in_dest->cmyk_buf.k;
       break;
     case DGP_C:
       *(c_ptr) = in_dest->cmyk_buf.c;
       break;
     case DGP_M:
       *(m_ptr) = in_dest->cmyk_buf.m;
       break;
     case DGP_Y:
       *(y_ptr) = in_dest->cmyk_buf.y;
       break;
     case DGP_K:
       *(k_ptr) = in_dest->cmyk_buf.k;
       break;
     default:
       break;
   }      
   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_reset_address(dgp_img_type img_type,dgp_dest_buf* in_dest,
   Dgp_Int8* other_ptr, Dgp_Int8* c_ptr, Dgp_Int8* m_ptr, Dgp_Int8* y_ptr,
   Dgp_Int8* k_ptr,Dgp_State* dgp_state)
{
   Dgp_Int8 **local, **c_buf, **m_buf, **y_buf, **k_buf;
   DGP_PROLOGUE("dgp_jpeg_reset_address()",DGP_PERRSTATE);
   DGP_STOP_WHINE;
   switch(img_type){
     case DGP_RGB:
       local = &(in_dest->other_buf.buffer);
       *(local) = other_ptr;
       break;
     case DGP_CMYK:
       c_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.c); *(c_buf) = c_ptr;
       m_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.m); *(m_buf) = m_ptr;
       y_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.y); *(y_buf) = y_ptr;
       k_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.k); *(k_buf) = k_ptr;
       break;
     case DGP_C:
       c_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.c); *(c_buf) = c_ptr;
       break;
     case DGP_M:
       m_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.m); *(m_buf) = m_ptr;
       break;
     case DGP_Y:
       y_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.y); *(y_buf) = y_ptr;
       break;
     case DGP_K:
       k_buf = (Dgp_Int8 **)&(in_dest->cmyk_buf.k); *(k_buf) = k_ptr;
       break;
     default:
       break;
   }      
   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_set_resolution_vals(DGP_JPEGImageDef* image_def,
                    j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
   Dgp_Int32 destres1, destres2, srcres;
   Dgp_Int32 scalenum, scaledenom;

   DGP_PROLOGUE("dgp_jpeg_set_resolution_vals()",DGP_PERRSTATE);

   DGP_STOP_WHINE;
   scalenum = scaledenom = 1; /*default value */

   if(DgpCoord_equal(image_def->imgResolution.destRes.xRes,
			    image_def->imgResolution.destRes.yRes)){

      destres1 = DgpCoord_floor(image_def->imgResolution.destRes.xRes);     
      destres2 = DgpCoord_floor(image_def->imgResolution.destRes.yRes);     
      srcres = DgpCoord_floor(image_def->imgResolution.srcRes.xRes);
      
      if(destres1 > srcres){ /*expanding image */
        scalenum = destres1 / srcres;
        scaledenom = 1;
        DGP_ASSERT(scalenum <= 6);
        DGP_ASSERT(scalenum >= 1);
      } else { /*staying at 1/1 or reducing the sz of the image */
        scaledenom = srcres /destres1;
        scalenum = 1;
        DGP_ASSERT(scaledenom >= 1);
        DGP_ASSERT(scaledenom <= 8);
      }

      cinfo->scale_num = scalenum;
      cinfo->scale_denom = scaledenom;
   } else {
        ERR_RETURN;
   }
   
#ifdef MEM_DATA_COLLECTION
   fprintf(mem_file,"Scale num/denom = %d/%d::",scalenum,scaledenom);
   fflush(mem_file);
#endif

   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_re_check_image_def_vals(DGP_JPEGImageDef* image_def,
                    j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
   DGP_PROLOGUE("dgp_jpeg_re_check_image_def_vals()",DGP_PERRSTATE);
   DGP_STOP_WHINE;
   /*Takes care of incorrect parameters! */
   if(image_def->imgCoOrds.dstW > (int)cinfo->output_width){
     image_def->imgCoOrds.dstW = cinfo->output_width;
   }
   if(image_def->imgCoOrds.dstH > (int)cinfo->output_height){
     image_def->imgCoOrds.dstH = cinfo->output_height;
   } /*whew! */

   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_copy_row_horizontal(dgp_dest_buf* to_buf,
         dgp_dest_adds* from_buf,dgp_img_type img_type,
         Dgp_Int32 copy_length,Dgp_State* dgp_state)
{
   DGP_PROLOGUE("dgp_jpeg_copy_row_horizontal()",DGP_PERRSTATE);

   DGP_STOP_WHINE;

   switch(img_type){
     case DGP_RGB:
        memcpy(to_buf->other_buf.buffer, from_buf->other_buf.buffer, copy_length); 
        break;
     case DGP_CMYK:
        memcpy(to_buf->cmyk_buf.c,
               from_buf->cmyk_buf.c,
               copy_length);
        memcpy(to_buf->cmyk_buf.m,
               from_buf->cmyk_buf.m,
               copy_length);
        memcpy(to_buf->cmyk_buf.y,
               from_buf->cmyk_buf.y,
               copy_length);
        memcpy(to_buf->cmyk_buf.k,
               from_buf->cmyk_buf.k,
               copy_length);
        break;
     case DGP_C:
        memcpy(to_buf->cmyk_buf.c,
               from_buf->cmyk_buf.c,
			       copy_length);
        break;
     case DGP_M:
        memcpy(to_buf->cmyk_buf.m,
               from_buf->cmyk_buf.m,
               copy_length);
        break;
     case DGP_Y:
        memcpy(to_buf->cmyk_buf.y,
               from_buf->cmyk_buf.y,
               copy_length);
        break;
     case DGP_K:
        memcpy(to_buf->cmyk_buf.k,
               from_buf->cmyk_buf.k,
               copy_length);
        break;
     default:
        ERR_RETURN;
   }
   NORMAL_RETURN;
} /*end of function copy_row_horizontal() */

static ErrCode
dgp_jpeg_copy_row_vertical(dgp_dest_buf* to_buf,
            dgp_dest_adds* from_buf,
            DGP_JPEGImageDef* image_def,
            j_decompress_ptr cinfo,Dgp_State* dgp_state)
{ 
   Dgp_Int32 i,bpr,out_color_components;
   /*From and to buffers have to be exactly a byte */
   /*apart .. sanity check */

   DGP_PROLOGUE("dgp_jpeg_copy_row_vertical()", DGP_PERRSTATE);
   
   DGP_STOP_WHINE;
   cinfo = cinfo;

   if(image_def->imgType == DGP_RGB){
     out_color_components = 3;
   } else {
     out_color_components = 1;
   }

   if(image_def->imgType == DGP_RGB){
       DGP_ASSERT(abs(from_buf->other_buf.buffer - to_buf->other_buf.buffer) == out_color_components); 
       bpr = to_buf->other_buf.bytes_per_row;
   } else {
       DGP_ASSERT(abs(from_buf->cmyk_buf.c -
              to_buf->cmyk_buf.c) == out_color_components);
       DGP_ASSERT(abs(from_buf->cmyk_buf.m -
              to_buf->cmyk_buf.m) == out_color_components);
       DGP_ASSERT(abs(from_buf->cmyk_buf.y -
              to_buf->cmyk_buf.y) == out_color_components);
       DGP_ASSERT(abs(from_buf->cmyk_buf.k -
              to_buf->cmyk_buf.k) == out_color_components);
       bpr = to_buf->cmyk_buf.bpr; 
   }

   switch(image_def->imgType){
     case DGP_RGB:
       /*unfortunately we need to go one pixel at a time */
       /*as we are copying along a column */
       for(i = 0 ; i < image_def->imgCoOrds.dstW; i++){

         DGP_ASSERT(abs((to_buf->other_buf.buffer+i*bpr) -
               (from_buf->other_buf.buffer+i*bpr)) == out_color_components);

         to_buf->other_buf.buffer[i*bpr] = 
                from_buf->other_buf.buffer[i*bpr];
         to_buf->other_buf.buffer[i*bpr+1] = 
                from_buf->other_buf.buffer[i*bpr+1];
         to_buf->other_buf.buffer[i*bpr+2] = 
                from_buf->other_buf.buffer[i*bpr+2];        
       }
	   from_buf->other_buf.buffer = to_buf->other_buf.buffer;
       break;
     case DGP_CMYK:
       for(i = 0;i < image_def->imgCoOrds.dstW; i++){ 
         to_buf->cmyk_buf.c[i*bpr] = 
                from_buf->cmyk_buf.c[i*bpr];
         to_buf->cmyk_buf.m[i*bpr] = 
                from_buf->cmyk_buf.m[i*bpr];
         to_buf->cmyk_buf.y[i*bpr] = 
                from_buf->cmyk_buf.y[i*bpr];
         to_buf->cmyk_buf.k[i*bpr] = 
                from_buf->cmyk_buf.k[i*bpr];
       }
       break;
     case DGP_C:
       for(i = 0; i < image_def->imgCoOrds.dstW; i++){
         to_buf->cmyk_buf.c[i*bpr] = 
                from_buf->cmyk_buf.c[i*bpr];
       }
       break;
     case DGP_M:
       for(i = 0; i < image_def->imgCoOrds.dstW; i++){
         to_buf->cmyk_buf.m[i*bpr] = 
                from_buf->cmyk_buf.m[i*bpr];
       }
       break;
     case DGP_Y:
       for(i = 0; i < image_def->imgCoOrds.dstW; i++){
         to_buf->cmyk_buf.y[i*bpr] = 
                from_buf->cmyk_buf.y[i*bpr];
       }
       break;
     case DGP_K:
       for(i = 0; i < image_def->imgCoOrds.dstW; i++){
         to_buf->cmyk_buf.k[i*bpr] = 
                from_buf->cmyk_buf.k[i*bpr];
       }
       break;
     default:
       break;
   }
   NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_copy_row(dgp_dest_buf* in_dest,
                 dgp_dest_adds* copy_address,
                 DGP_JPEGImageDef* image_def,
                 j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
  Dgp_Int32 out_color_components=0;

  DGP_PROLOGUE("dgp_jpeg_copy_row()",DGP_PERRSTATE);

  if(image_def->imgType == DGP_RGB)
    out_color_components = 3;
  else
    out_color_components = 1;

   switch(image_def->imgOrientation){
     case ORIENT_DL: /*270 */
     case ORIENT_UR: /*90 */
       DGP_ERR_PASS(dgp_jpeg_copy_row_vertical(in_dest,copy_address,image_def,cinfo,
                    dgp_state));
       break;
     case ORIENT_LU: /*180 */
     case ORIENT_RD: /*No rotation */
     default:
       DGP_ERR_PASS(dgp_jpeg_copy_row_horizontal(in_dest,copy_address,image_def->imgType,
               cinfo->crop_width*out_color_components,dgp_state));
       break;
   }
   NORMAL_RETURN;
}  /*end of function copy_row() */

static ErrCode
dgp_jpeg_store_address(dgp_dest_adds* copy_address,
              dgp_img_type   img_type,
              dgp_dest_buf*  in_dest, Dgp_State* dgp_state)
{
   Dgp_Int8 **other_ptr, **c_ptr, **m_ptr, **y_ptr, **k_ptr;

   DGP_PROLOGUE("dgp_jpeg_store_address()",DGP_PERRSTATE);
   DGP_STOP_WHINE;

   switch(img_type){
     case DGP_RGB:
        other_ptr = &copy_address->other_buf.buffer;
        *(other_ptr) = in_dest->other_buf.buffer;
        break;
     case DGP_CMYK:
        c_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.c;
        *(c_ptr) = in_dest->cmyk_buf.c;
        m_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.m;
        *(m_ptr) = in_dest->cmyk_buf.m;
        y_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.y;
        *(y_ptr) = in_dest->cmyk_buf.y;
        k_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.k;
        *(k_ptr) = in_dest->cmyk_buf.k;
        break;
     case DGP_C:
        c_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.c;
        *(c_ptr) = in_dest->cmyk_buf.c;
        break;
     case DGP_M:
        m_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.m;
        *(m_ptr) = in_dest->cmyk_buf.m;
        break;
     case DGP_Y:
        y_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.y;
        *(y_ptr) = in_dest->cmyk_buf.y;
        break;
     case DGP_K:
        k_ptr = (Dgp_Int8 **)&copy_address->cmyk_buf.k;
        *(k_ptr) = in_dest->cmyk_buf.k;
        break;
     default:
        ERR_RETURN;;
   }
   NORMAL_RETURN;
} /*end of function store_address() */

#if 0 
static ErrCode
dgp_jpeg_set_num_lines(j_decompress_ptr cinfo,DGP_JPEGImageDef* image_def, 
              Dgp_Int32* num_lines, Dgp_State* dgp_state)
{
   Dgp_Int32 ret_val=0;
   DGP_PROLOGUE("dgp_jpeg_set_num_lines()",DGP_PERRSTATE);
   DGP_STOP_WHINE;
   switch(image_def->imgOrientation){
     case ORIENT_DL: /*270 */
     case ORIENT_UR: /*90 */
     {
       ret_val = (cinfo->crop_height + 
         (image_def->imgCoOrds.dstY - image_def->imgCoOrds.dstMinY));
       break;
     }
     case ORIENT_LU: /*180 */
     case ORIENT_RD: /*no rotation */
     default:
     {
       ret_val = (cinfo->crop_height - cinfo->output_scanline +
         (image_def->imgCoOrds.dstY - image_def->imgCoOrds.dstMinY));
       break;
     }
   }
   *(num_lines) = ret_val;
   NORMAL_RETURN;
}
#endif

static ErrCode
dgp_jpeg_set_cropped_vals(DGP_JPEGImageDef* image_def,
                 j_decompress_ptr cinfo,
                 crop_data* crop_vals,Dgp_State* dgp_state)
{
   /*This gives the extent of the jpeg image that needs to */
   /*be decoded. - in scaled destination pixels. */

   DGP_PROLOGUE("dgp_jpeg_set_cropped_vals()",DGP_PERRSTATE);

   DGP_STOP_WHINE;
   crop_vals->start_x = image_def->imgCoOrds.dstX - image_def->imgCoOrds.srcOriginX;
   crop_vals->start_y = image_def->imgCoOrds.dstY - image_def->imgCoOrds.srcOriginY;

   DGP_ASSERT(crop_vals->start_x >= 0);
   DGP_ASSERT(crop_vals->start_y >= 0);

   /*It does not matter if the image is rotated or not - */
   /*as, we decode in the normal way and only write out those */
   /*we want. Note however that this is in the destination  */
   crop_vals->end_x = crop_vals->start_x + image_def->imgCoOrds.dstW - 1;
   crop_vals->end_y = crop_vals->start_y + image_def->imgCoOrds.dstH - 1;
   
   if(crop_vals->end_x >= (int)cinfo->output_width)
       crop_vals->end_x = cinfo->output_width - 1;        /* Added - 1 / dhtesq. */
   if(crop_vals->end_y >= (int)cinfo->output_height)       
       crop_vals->end_y = cinfo->output_height - 1;           /* Added - 1 / dhtesq. */
	   
   /*
   if(crop_vals->end_x >= image_def->imgCoOrds.nominal_strip_width)
       crop_vals->end_x = image_def->imgCoOrds.nominal_strip_width - 1;       
   if(crop_vals->end_y >= image_def->imgCoOrds.nominal_strip_height)       
       crop_vals->end_y = image_def->imgCoOrds.nominal_page_height - 1;       
	   */

   NORMAL_RETURN;
}


/*Check that there is enough space allocated in the buffer */
/*to hold the image required. */

static ErrCode
dgp_jpeg_check_buffer_width(DGP_JPEGImageDef* image_def,
                   Dgp_Int32 bytes_in_jpeg_row,Dgp_State* dgp_state)
{
  Dgp_Int32 num_components = 0, tmp_val= 0;  

  DGP_PROLOGUE("dgp_jpeg_check_buffer_width()",DGP_PERRSTATE);
  DGP_STOP_WHINE;
  
  if(image_def->imgType == DGP_RGB){
    num_components = 3;
  } else {
    num_components = 1;
  }

  if((image_def->imgOrientation == ORIENT_RD) ||
     (image_def->imgOrientation == ORIENT_LU) ) {
     tmp_val = ((image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)
            + image_def->imgCoOrds.dstW ) * num_components;
  } else {
     tmp_val = ((image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX)
            + image_def->imgCoOrds.dstH ) * num_components;
  }
  
  if(tmp_val <= bytes_in_jpeg_row)
    NORMAL_RETURN;
  else
    ERR_RETURN;
}

static ErrCode
dgp_jpeg_cropped_image_setup(DGP_JPEGImageDef* image_def,dgp_dest_buf* in_dest,
                    j_decompress_ptr cinfo, crop_data* crop_vals,
                    Dgp_Int32* act_img_bpr,Dgp_State* dgp_state)
{

  DGP_PROLOGUE("dgp_jpeg_cropped_image_setup()",DGP_PERRSTATE);
  in_dest = in_dest;

  DGP_ERR_PASS(dgp_jpeg_set_cropped_vals(image_def,cinfo,crop_vals,dgp_state));

  if((image_def->imgOrientation == ORIENT_RD) ||
    (image_def->imgOrientation == ORIENT_LU)) {
    cinfo->crop_width = crop_vals->end_x - crop_vals->start_x+1;
    cinfo->crop_height = crop_vals->end_y - crop_vals->start_y+1;
  } else {
    cinfo->crop_width = crop_vals->end_y - crop_vals->start_y+1;
    cinfo->crop_height = crop_vals->end_x - crop_vals->start_x+1;
  }

  if(image_def->imgType == DGP_RGB) {
    *(act_img_bpr) = cinfo->crop_width * 3;
  }
  else if(image_def->imgType == DGP_CMYK){
    *(act_img_bpr) = cinfo->crop_width;
  }

  NORMAL_RETURN;
}

static ErrCode
dgp_jpeg_set_out_color_space(dgp_img_type im_type,
                    j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
  DGP_PROLOGUE("dgp_jpeg_set_out_color_space()", DGP_PERRSTATE);
  DGP_STOP_WHINE;

  if(im_type == DGP_RGB){
    cinfo->out_color_space = JCS_RGB;
#ifdef MEM_DATA_COLLECTION
  fprintf(mem_file,"Output format is RGB : ");
#endif
  } else {
    cinfo->out_color_space = JCS_CMYK;
#ifdef MEM_DATA_COLLECTION
  fprintf(mem_file,"Output format is CMYK : ");
#endif
  }

  NORMAL_RETURN;

}

static ErrCode
dgp_jpeg_check_comp_params(DGP_JPEGCompressDef* image_def,
                           DGP_JPEGOutputDef* output_def,
                           Dgp_State* dgp_state)
{
  DGP_PROLOGUE("dgp_jpeg_check_comp_params()", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  image_def = image_def;
  output_def = output_def;
  

  DGP_ASSERT(image_def->imgWidth > 0);
  DGP_ASSERT(image_def->imgHeight > 0);
  DGP_ASSERT(image_def->imgSrc.buffer != NULL);
  DGP_ASSERT(image_def->imgSrc.size > 0);
  DGP_ASSERT(output_def->buffer != NULL);
  DGP_ASSERT(output_def->outLen > 0);
  NORMAL_RETURN;
}

static ErrCode
dgp_convert_rows(DGP_JPEGCompressDef* image_def,
         JSAMPROW to_buffer,
         Dgp_State* dgp_state)
{
  int row_stride_rgb, row_stride_cmyk,i,j,row_num;
  Dgp_Int8 c,m,y,k;

  DGP_PROLOGUE("dgp_convert_rows()", DGP_PERRSTATE);
  DGP_STOP_WHINE;

  row_stride_rgb = image_def->imgWidth * 3;
  row_stride_cmyk = image_def->imgWidth * 4;

  for(row_num = 0; row_num < image_def->imgHeight; row_num++){
    for(i = 0, j = 0; /*i < row_stride_rgb,*/ j < row_stride_cmyk; i+=3, j+=4){
       c = image_def->imgSrc.buffer[row_num*row_stride_cmyk+j]; 
       m = image_def->imgSrc.buffer[row_num*row_stride_cmyk+j+1]; 
       y = image_def->imgSrc.buffer[row_num*row_stride_cmyk+j+2];
       k = image_def->imgSrc.buffer[row_num*row_stride_cmyk+j+3];

       to_buffer[row_num*row_stride_rgb+i] = 255 - (c+k); /* R value */
       to_buffer[row_num*row_stride_rgb+i+1] = 255 - (m+k); /*G value */
       to_buffer[row_num*row_stride_rgb+i+2] = 255 - (y+k); /* B value */
    }
  }
  NORMAL_RETURN;
}

/********** COMPRESSION FUNCTIONS EXPORTED BY THE LIBRARY ******
**  DGP_JPEG_init_jcompress
**  Dgp_JPEG_destroy_jcompress
**  Dgp_JPEG_create_jpeg
**/

/***********************************************************
**
**  Function name: DGP_JPEG_init_jcompress
**  Initializes the cinfo data structure
**/
ErrCode
DGP_JPEG_init_jcompress(j_compress_ptr cinfo, Dgp_State* dgp_state)
{
  DGP_PROLOGUE("DGP_JPEG_init_jcompress()", DGP_PERRSTATE);

  jpeg_create_compress(cinfo);

  /* the in_color_space will be reset to the right values when */
  /* we do the actual compression. This is more to just fill */
  /* a value to call jpeg_set_defaults() */
  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));

  cinfo->in_color_space = JCS_RGB; /* arbitrary guess */
  jpeg_set_defaults(cinfo);

  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));

  NORMAL_RETURN;
}

/***********************************************************
**
**  Function name: Dgp_JPEG_destroy_jcompress
**  Destroys the cinfo data structure
**/
ErrCode
Dgp_JPEG_destroy_jcompress(j_compress_ptr cinfo, Dgp_State* dgp_state)
{
  DGP_PROLOGUE("Dgp_JPEG_destroy_jcompress()",DGP_PERRSTATE);

  jpeg_destroy_compress(cinfo);

  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));  

  NORMAL_RETURN;
}


/***********************************************************
**
**  Function name: Dgp_JPEG_create_jpeg
**  Destroys the cinfo data structure
**
**  This function creates the required JPEG file. The input
**  is a raw buffer either of CMYK or RGB data. If it is
**  RGB data, the output is of the form YUV(YcbCr) or
**  Grayscale, as specified. If the input is of CMYK form,
**  the output is in YCCK form.
**
**  The 'output' contains a pre-allocated output buffer that 
**  will hold the compressed image and when this function is
**  called, output->outLen indicates the size of the allocated
**  buffer. When the call returns, it contains the size of
**  the data that is actually written. Of course, if outLen is
**  smaller than the size needed, an error is returned.
**
*/
ErrCode
Dgp_JPEG_create_jpeg(DGP_JPEGCompressDef* image_def,
            DGP_JPEGOutputDef* output,
            j_compress_ptr cinfo, Dgp_State* dgp_state)
{

  JSAMPROW* row_pointer;	/* pointer to a single row */
  JSAMPROW  temp_row = NULL;		
  JSAMPROW  temp_buffer = NULL;
  int row_stride,buf_sz;    /* physical row width in buffer */

  DGP_PROLOGUE("Dgp_JPEG_create_jpeg()",DGP_PERRSTATE);

  DGP_ERR_PASS(dgp_jpeg_check_comp_params(image_def,output,dgp_state));

  /* Step 1: First set the destination */

  DGP_jpeg_buf_dest(cinfo,output);
  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));                              

  /* Step 2: set parameters for compression */

  cinfo->image_width = image_def->imgWidth;
  cinfo->image_height = image_def->imgHeight;
  cinfo->input_components = 3; 
  cinfo->in_color_space = JCS_RGB;

  if(image_def->imgType == DGP_CMYK){
    buf_sz = cinfo->image_width * cinfo->image_height * cinfo->input_components;
    temp_buffer = DGP_CALLOC(1, buf_sz);
    DGP_ERR_PASS(dgp_convert_rows(image_def,temp_buffer,
                                  dgp_state));
  }

  jpeg_set_defaults(cinfo);

  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));
  /* Step 3 : TRUE ensures that a complete interchange-JPEG is written */
  jpeg_start_compress(cinfo, TRUE);
  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));

  /* Step 4 : Do the actual compression */

  row_stride = cinfo->image_width * cinfo->input_components;
  while (cinfo->next_scanline < cinfo->image_height){
    if(image_def->imgType == DGP_RGB){
      temp_row = image_def->imgSrc.buffer  + 
               cinfo->next_scanline * row_stride;
    } else if(image_def->imgType == DGP_CMYK) {
      temp_row = temp_buffer + cinfo->next_scanline * row_stride;
    }
    row_pointer = &temp_row;      
    jpeg_write_scanlines(cinfo,row_pointer,1);
    DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));
  }

  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));
  /* Step 5: Finish compression  */

  jpeg_finish_compress(cinfo);
  DGP_ERR_PASS(dgp_internal_comp_check(cinfo,dgp_state));

  output->outLen = cinfo->dest->free_in_buffer;                        

  if(image_def->imgType == DGP_CMYK){
    if(temp_buffer != NULL){
      DGP_FREE(temp_buffer);
    }
  }

  NORMAL_RETURN;
}

/********** DECOMPRESSION FUNCTIONS EXPORTED BY THE LIBRARY ******
**  DGP_JPEG_init_jdecompress
**  Dgp_JPEG_destroy_jdecompress
**  DGP_JPEG_check_jdecompress
**  Dgp_JPEG_decompress_jpeg
**/

/***********************************************************
**
**  Function name: DGP_JPEG_init_jdecompress
**  Initializes the cinfo data structure
**/
ErrCode
DGP_JPEG_init_jdecompress(j_decompress_ptr cinfo, Dgp_State* dgp_state)
{
  DGP_PROLOGUE("DGP_JPEG_init_jdecompress()",DGP_PERRSTATE);
  DGP_STOP_WHINE;
#ifdef PROGRESS_REPORT
  struct cdjpeg_progress_mgr *progress;       /*dhtesq... */
#endif

  jpeg_create_decompress(cinfo);

#ifdef PROGRESS_REPORT
  /* dhtesq. */
  progress = (struct cdjpeg_progress_mgr *)DGP_MALLOC(sizeof(struct cdjpeg_progress_mgr));
  start_progress_monitor(cinfo, progress);
#endif


  NORMAL_RETURN;
}

/***********************************************************
**
**  Function name: Dgp_JPEG_destroy_jdecompress
**  Destroys the cinfo data structure
**/
ErrCode
Dgp_JPEG_destroy_jdecompress(j_decompress_ptr cinfo, Dgp_State* dgp_state)
{
  DGP_PROLOGUE("Dgp_JPEG_destroy_jdecompress()",DGP_PERRSTATE);
  DGP_STOP_WHINE;

#ifdef PROGRESS_REPORT
  /* Free progress context if we've been using it. / dhtesq.  */
  DGP_FREE(cinfo->progress);
#endif

  jpeg_destroy_decompress(cinfo);
  
  NORMAL_RETURN;
}

/***********************************************************
**
** function name: DGP_JPEG_check_jdecompress
**
** This function checks the parameters passed in
** and tells the calling program what can and cannot
** be done
**
** ret_val is the return value. Things that it checks are
** co-ordinate values: asserts that all need to be >= 0
** co-ordinate differences: no negative differences
**    (when the co-ordinate is invalid it is set to 
**     DGP_INVALID_COORD)
** orientation - has to be a valid orientation
** resolution - x and y resolutions have to be equal. can be
**    scaled down by 1/2 and 1/4. can be scaled up 2,3,4,5,6 times
**    original size.
** other sanity checks as well:
**    that if the image is cropped, it is a sane crop, that the
**    buffer size (bytes_per_row and num_rows together) is enough to
**    hold the decompressed image, etc.
**
** the return value, if it is 0, indicates that the parameters are
** valid and can be handled. If not, see which bits are set (to 1)
** to indicate where the problem is.Look at jpeg_interface.h to
** see which bit implies what problem.
**/

ErrCode
DGP_JPEG_check_jdecompress(DGP_JPEGImageDef* image_def,
				  Dgp_UInt32* ret_val,
                  j_decompress_ptr cinfo, Dgp_State* dgp_state)
{

  Dgp_UInt32 ret = 0;
  /*struct jpeg_error_mgr * err = cinfo->err; */

  DGP_PROLOGUE("DGP_JPEG_check_jdecompress()",DGP_PERRSTATE);

  if(cinfo->src == NULL) {
     if(image_def->imgSrc.u)
         InitDGPUdsSrcMgr(cinfo,image_def->imgSrc.u,
                            image_def->imgSrc.size);
     else if (image_def->imgSrc.buf)
         InitDGPMemSrcMgr(cinfo,image_def->imgSrc.buf,
                            image_def->imgSrc.size);
     else DGP_ERR_GEN( TRUE, cJpegIOErr);
  }

  DGP_ERR_PASS(dgp_jpeg_check_input_vals(image_def, &ret, dgp_state));

  if( (cinfo->image_height == 0) && (cinfo->image_width == 0)){
   (void) jpeg_read_header(cinfo,DGP_TRUE);
  } 

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

  DGP_ERR_PASS(dgp_estimate_memory(cinfo,image_def,dgp_state));

#ifdef MEM_DATA_COLLECTION
  if(mem_file != NULL){
    fprintf(mem_file,"Mem estimate is %d\n",
                    image_def->memData.mem_estimate);
    fflush(mem_file);
  }
#endif
  /*check type */
  switch(image_def->imgType){
    case DGP_RGB:
    case DGP_CMYK:
    case DGP_C:
    case DGP_M:
    case DGP_Y:
    case DGP_K:
      break;
    case DGP_PACKED:
    default:
       ret |= DGP_IMAGE_TYPE_BIT;
       break;
  }

  switch(image_def->imgOrientation){
      /*
      Actually all of these rotations are anticlockwise.
      As originally implemented, images would rotate in the direction opposite the page rotation
      Fixed throughout by swapping code for DL and UR cases...
      / dhtesq.
      */
    case ORIENT_RD: /*no rotation */
	case ORIENT_DL: /*rotation clockwise by 90 degrees */
    case ORIENT_LU: /*rotation clockwise by 180 degrees */
    case ORIENT_UR: /*rotation clockwise by 270 degrees */

       cinfo->output_orientation = image_def->imgOrientation;
       break;
    default:
        /* So I guess libjpeg6b can't do "inverting" orientations like ORIENT_DR. / dhtesq. */
       ret |= DGP_ORIENTATION_BIT;
  }

  *(ret_val) = ret;
  
  NORMAL_RETURN;
}

/***********************************************************
**
**  Function name: Dgp_JPEG_decompress_jpeg
**  
**  Main decompress function. Rechecks the image parameters and
**  produces the output image in the format requested, in
**  the dgp_dest_buf variable.
**/

#ifdef STANDALONE_TESTING
ErrCode
Dgp_JPEG_decompress_jpeg(DGP_JPEGImageDef* image_def,
            dgp_dest_buf* in_dest,
            j_decompress_ptr cinfo,Dgp_State* dgp_state,
            FILE* output_file)
{
#else
ErrCode
Dgp_JPEG_decompress_jpeg(DGP_JPEGImageDef* image_def,
            dgp_dest_buf* in_dest,
            j_decompress_ptr cinfo,Dgp_State* dgp_state)
{
#endif
  Dgp_UInt32 ret_val=0;
  Dgp_Int32 bytes_in_jpeg_row=0,skip_rows=0, skip_row_cnt=0;
  JDIMENSION num_scanlines;
  Dgp_Bool need_to_skip = DGP_FALSE;
  Dgp_Int8 *orig_ptr, *c_ptr, *m_ptr, *y_ptr, *k_ptr;
  dgp_dest_adds  copy_address;
  /*Dgp_Int32 out_color_components=0; */
  crop_data* crop_vals;
  struct jpeg_error_mgr* err = cinfo->err;
  Dgp_Int32 act_img_bpr, temp;

  Dgp_Int32 whocares_w, whocares_h, fo, so;
  Dgp_Int32 scanlinesReadInThisStrip, stripStartRef;

#ifdef STANDALONE_TESTING
  djpeg_dest_ptr dest_mgr = NULL;
  Dgp_Int32 i;
#endif

  /*Just call recheck_jpeg_render to make sure that the */
  /*caller has indeed fixed all issues before calling the */
  /*decompress function */

  DGP_PROLOGUE("Dgp_JPEG_decompress_jpeg()",DGP_PERRSTATE);

  DGP_ERR_PASS(dgp_jpeg_recheck_jpeg_render(image_def,&ret_val,
                                   cinfo,dgp_state));
  
  if(ret_val != 0)
    ERR_RETURN;

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

  /* This is where the scale factor is computed... / dhtesq. */
  DGP_ERR_PASS(dgp_jpeg_set_resolution_vals(image_def,cinfo,dgp_state));

  DGP_ERR_PASS(dgp_jpeg_set_out_color_space(image_def->imgType,cinfo,
               dgp_state));

  /* We could support v. short strips (< scale_num), but performance is ugly */
  /* and will only get worse with the added bounds checking. */
  DGP_ERR_GENF2(image_def->imgCoOrds.nominal_strip_height < (int)cinfo->scale_num, cErrDecodeFailure,
	  "The strip height (%d) is shorter than the image's scale factor (%d); you really want me to support this case?", 
		  image_def->imgCoOrds.nominal_strip_height, cinfo->scale_num);

  jpeg_calc_output_dimensions(cinfo);

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

  DGP_ERR_PASS(dgp_jpeg_re_check_image_def_vals(image_def,cinfo,dgp_state));

  if(image_def->imgType == DGP_RGB){
    crop_vals = &in_dest->other_buf.crop_coords;
  } else {
    crop_vals = &in_dest->cmyk_buf.crop_vals;
  }

  DGP_ERR_PASS(dgp_jpeg_cropped_image_setup(image_def,in_dest,cinfo,
                         crop_vals,&act_img_bpr,dgp_state));
  
#ifdef STANDALONE_TESTING
  dest_mgr = jinit_write_bmp(cinfo, DGP_FALSE);
  dest_mgr->output_file = output_file;
#endif
  if(image_def->imgType != DGP_RGB){
     in_dest->cmyk_buf.imgType = image_def->imgType;
  }

  /* Image coordinates still aren't in the rotated frame at this point.  If you don't transform, */
  /* images will appear on a rotated page as if their UL corners were pinned to their RD locations */
  /* (but rotated).  I don't know where this is "supposed" to happen. / dhtesq. */
  /* */
  /* The only thing we are really interested in here is the image origin;  everything else is already correct. */
    
  DGP_ERR_PASS(DgpCoord_coordOrientTransform(image_def->imgCoOrds.dstMinX, image_def->imgCoOrds.dstMinY,
					      image_def->imgCoOrds.nominal_page_width, image_def->imgCoOrds.nominal_page_height,
                          image_def->imgCoOrds.dstX, 
                          image_def->imgCoOrds.dstY,
                          image_def->imgCoOrds.dstW, image_def->imgCoOrds.dstH,
					      image_def->imgOrientation,
					      &fo, &so,
					      &image_def->imgCoOrds.dstX, &image_def->imgCoOrds.dstY,
					      &whocares_w, &whocares_h,
					      dgp_state));

  /* The rotation takes place in global (page) coordinates.  Get dstX and dstY back into strip coordinates... */
  switch(image_def->imgOrientation){
    case ORIENT_RD:
        break;
    case ORIENT_LU:
        image_def->imgCoOrds.dstX = image_def->imgCoOrds.nominal_strip_width + image_def->imgCoOrds.dstMinX - 
            (image_def->imgCoOrds.nominal_page_width - image_def->imgCoOrds.dstX);
        image_def->imgCoOrds.dstY = image_def->imgCoOrds.nominal_strip_height + image_def->imgCoOrds.dstMinY -                               
            (image_def->imgCoOrds.nominal_page_height - image_def->imgCoOrds.dstY);

        image_def->imgCoOrds.dstX += image_def->imgCoOrds.dstMinX;
        /*image_def->imgCoOrds.dstY += image_def->imgCoOrds.dstMinY; */
        /*image_def->imgCoOrds.dstY = image_def->imgCoOrds.dstMinY; */
        break;
    case ORIENT_UR:
        image_def->imgCoOrds.dstX = image_def->imgCoOrds.nominal_strip_width + image_def->imgCoOrds.dstMinY -                               
            (image_def->imgCoOrds.nominal_page_height - image_def->imgCoOrds.dstX);
        /* image_def->imgCoOrds.dstY -= image_def->imgCoOrds.dstMinX; */

        image_def->imgCoOrds.dstX += image_def->imgCoOrds.dstMinY;
        /* image_def->imgCoOrds.dstY += image_def->imgCoOrds.dstMinX; */

        /* Swap dstMinX and dstMinY for jpeg lib... */
        temp = image_def->imgCoOrds.dstMinX;
        image_def->imgCoOrds.dstMinX = image_def->imgCoOrds.dstMinY;
        image_def->imgCoOrds.dstMinY = temp;

		/* dhtesq. */
		/*image_def->imgCoOrds.dstX = image_def->imgCoOrds.dstMinX; */
		/*image_def->imgCoOrds.dstY = image_def->imgCoOrds.dstMinY; */

        break;
    case ORIENT_DL:
        /*image_def->imgCoOrds.dstX = image_def->imgCoOrds.nominal_strip_width + image_def->imgCoOrds.dstMinY -                                */
        /*    (image_def->imgCoOrds.nominal_page_height - image_def->imgCoOrds.dstX); */
        image_def->imgCoOrds.dstY = image_def->imgCoOrds.nominal_strip_height + image_def->imgCoOrds.dstMinX - 
            (image_def->imgCoOrds.nominal_page_width - image_def->imgCoOrds.dstY);

        /*image_def->imgCoOrds.dstX += image_def->imgCoOrds.dstMinY; */
        /*image_def->imgCoOrds.dstY += image_def->imgCoOrds.dstMinX; */
        /*image_def->imgCoOrds.dstY += image_def->imgCoOrds.dstW; */

        /* Swap dstMinX and dstMinY for jpeg lib... */
        temp = image_def->imgCoOrds.dstMinX;
        image_def->imgCoOrds.dstMinX = image_def->imgCoOrds.dstMinY;
        image_def->imgCoOrds.dstMinY = temp;

		/* dhtesq. */
		/*image_def->imgCoOrds.dstX = image_def->imgCoOrds.dstMinX; */
		/*image_def->imgCoOrds.dstY = image_def->imgCoOrds.dstMinY; */

        break;
    default:
        DGP_ERR_GENF0(DGP_TRUE, cErrInvalidParm, "Invalid orientation.");
  }
  /*///////////////////////// */

  /*Begin */
  jpeg_start_decompress(cinfo);

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

  if(image_def->imgCoOrds.bytesPerRow != bytes_in_jpeg_row){
    bytes_in_jpeg_row = image_def->imgCoOrds.bytesPerRow;
     DGP_ERR_PASS(dgp_jpeg_check_buffer_width(image_def,bytes_in_jpeg_row,dgp_state));
#ifndef STANDALONE_TESTING
    if(image_def->imgType == DGP_CMYK){
      in_dest->cmyk_buf.bpr = bytes_in_jpeg_row;
    } else {
      in_dest->other_buf.bytes_per_row = bytes_in_jpeg_row;
    }
#endif
  } /*end of if(bytes_per_row != bytes_in_jpeg_row) */

#ifdef STANDALONE_TESTING

  DGP_ERR_PASS(dgp_jpeg_set_num_lines(cinfo,image_def,&num_lines,dgp_state));
  if(image_def->imgType == DGP_RGB)
    out_color_components = 3;
  else
    out_color_components = 1;

  if(image_def->imgType == DGP_CMYK){
    in_dest->cmyk_buf.c = DGP_CALLOC( 1,bytes_in_jpeg_row * num_lines);
    in_dest->cmyk_buf.m = DGP_CALLOC( 1,bytes_in_jpeg_row * num_lines);
    in_dest->cmyk_buf.y = DGP_CALLOC( 1,bytes_in_jpeg_row * num_lines);
    in_dest->cmyk_buf.k = DGP_CALLOC( 1,bytes_in_jpeg_row * num_lines);
    in_dest->cmyk_buf.bpr = bytes_in_jpeg_row;
  } else {
    in_dest->other_buf.buffer = DGP_CALLOC( 1,bytes_in_jpeg_row * num_lines);
    in_dest->other_buf.bytes_per_row = bytes_in_jpeg_row;
  }
#endif

  DGP_ERR_PASS(dgp_jpeg_set_address(image_def->imgType, in_dest,
               &orig_ptr,&c_ptr,&m_ptr,&y_ptr,&k_ptr,dgp_state));
  
#ifdef STANDALONE_TESTING
 (*dest_mgr->start_output) (cinfo, dest_mgr);
#endif


  /*Shift buffer pointers to appropriate row */
  /* 12.7.2000 / dhtesq. */
  /* This shift needs to be done even in cases where this condition is not true, i.e. */
  /* LU orientation by bands. */
/*  if((image_def->imgCoOrds.dstY - image_def->imgCoOrds.dstMinY) != 0){ */
    DGP_ERR_PASS(dgp_jpeg_shift_to_initial_row(image_def, in_dest, bytes_in_jpeg_row, dgp_state));
/*  } */

  if(cinfo->scale_num > 1){
    skip_rows = cinfo->scale_num - 1;
    skip_row_cnt=0;
    need_to_skip = DGP_FALSE;
  }  

  /*
  Since crop_vals->end_y is defined as the ending coordinate, this comparison
  will miss the last line. / dhtesq.

  while (((int)cinfo->output_scanline < crop_vals->end_y) &&
          (err->num_errors == 0)) {    
  */
  scanlinesReadInThisStrip = 0;
  while (((int)cinfo->output_scanline <=  crop_vals->end_y) && (err->num_errors == 0) /*&& ((int)cinfo->output_scanline < 100)*/){    

    if((image_def->imgOrientation == ORIENT_DL) || (image_def->imgOrientation == ORIENT_UR)){
        if((int)cinfo->output_scanline >= crop_vals->start_y) {
            DGP_ERR_PASS(dgp_jpeg_reset_address(image_def->imgType, in_dest, orig_ptr,c_ptr,m_ptr,y_ptr,k_ptr,dgp_state));
            DGP_ERR_PASS(dgp_jpeg_shift_to_initial_row(image_def,in_dest, bytes_in_jpeg_row,dgp_state));
        }
		stripStartRef = crop_vals->start_x;
    } else {
		stripStartRef = crop_vals->start_y;
	}

    /*Shift buffer pointer to the appropriate x position. */
    if((int)cinfo->output_scanline >= crop_vals->start_y) {
		/* 7.3.2000 / dhtesq.  */
		/* This inequality used to read "> 0", with the consequence that */
		/* UR and DL would only have their first columns rendered. */
		/* I think -- think -- we only caught this recently because of MK's */
		/* full page jpegs, where dstX == dstMinX. */
      if((image_def->imgCoOrds.dstX  - image_def->imgCoOrds.dstMinX) >= 0){
          DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset_rotated(image_def, in_dest, cinfo->output_scanline - crop_vals->start_y, dgp_state));
      }
    }

    if((skip_rows > 0) && (need_to_skip)){
      if((int)cinfo->output_scanline > crop_vals->start_y) {
		  /* 6.3.2001 / dhtesq. */
		  /* Row copying is too crude a scaling method to accommodate masks and jpegs at different resolutions. */
		  /* The color conversion routines will handle the scaling if there's a mask. */
		  if(in_dest->other_buf.trans_mask == NULL){
			  DGP_ERR_PASS(dgp_jpeg_copy_row(in_dest, &copy_address, image_def, cinfo, dgp_state));   
		  }
      }
      cinfo->output_scanline += 1;
      skip_row_cnt++;
      if(skip_rows == skip_row_cnt){
        skip_row_cnt = 0;
        need_to_skip = DGP_FALSE;
      }
      num_scanlines = 1;
    } else {
	  /* This clause handles the case at the beginning of a strip... */
	  if(skip_rows && ((int)(cinfo->output_scanline + 1 - crop_vals->start_y) <= 0) && 
		  ((int)(cinfo->output_scanline + 1 - crop_vals->start_y) > -((int)cinfo->scale_num - 1))){
		  /* Depending on where the strip boundary falls, we may not want to repeat the full number of times */
		  /* indicated by the downsampling.  Preset the counter. */
		  skip_row_cnt = skip_rows - (cinfo->scale_num - (crop_vals->start_y - cinfo->output_scanline + 1));
		  cinfo->output_scanline = crop_vals->start_y;
		  need_to_skip = skip_row_cnt < skip_rows ? DGP_TRUE:DGP_FALSE;
		  if(need_to_skip == DGP_FALSE) skip_row_cnt = 0;
	  }
      num_scanlines = jpeg_read_scanlines(cinfo, *(in_dest), 1); 
	  /* 9.3.2001 / dhtesq. */
	  /* We were handling stripwise decodes, and subsampled jpegs, but not both at the same time. */
	  /* Make sure we don't read too many scanlines before we get to the useful part of the image. */
	  if((int)cinfo->output_scanline < crop_vals->start_y){
		  cinfo->output_scanline += cinfo->scale_num - 1;/*MIN(cinfo->scale_num - 1, crop_vals->start_y - cinfo->output_scanline); */
	  }
	  /*////// */
	  scanlinesReadInThisStrip++;
    }

    if((need_to_skip == DGP_TRUE) || (((cinfo->output_scanline) % cinfo->scale_num == 1) && (skip_rows > 0) 
		&& ((int)cinfo->output_scanline >= crop_vals->start_y))){
      /*first time for this row, then! */
      need_to_skip = DGP_TRUE;
      DGP_ERR_PASS(dgp_jpeg_store_address(&copy_address, image_def->imgType, in_dest, dgp_state));
    }

    /*Now shift to the end of the line - from whereever you */
    /*started to the end of the line */
    /*Do nothing for 90 and 270 */
    if((int)cinfo->output_scanline > crop_vals->start_y) {
	    DGP_ERR_PASS(dgp_jpeg_shift_to_end_of_row(image_def, in_dest, cinfo, dgp_state));
    }


  } /*end of while loop */

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

#ifdef STANDALONE_TESTING
    DGP_ERR_PASS(dgp_jpeg_reset_address(image_def->imgType, in_dest,
                 orig_ptr,c_ptr,m_ptr,y_ptr,k_ptr,dgp_state));
    DGP_ERR_PASS(dgp_jpeg_set_addresses_to_beg_of_img(image_def,in_dest,
                 bytes_in_jpeg_row,dgp_state));     
    DGP_ERR_PASS(dgp_jpeg_set_cinfo_vals(cinfo,dgp_state));

    for(i = 0; i < cinfo->crop_height; i++){
      if((image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX) > 0){
        DGP_ERR_PASS(dgp_jpeg_shift_to_x_offset(image_def->imgType, in_dest,
			out_color_components * (image_def->imgCoOrds.dstX - image_def->imgCoOrds.dstMinX),
			dgp_state));
      }
      if((image_def->imgType == DGP_RGB) ||
         (image_def->imgType == DGP_CMYK)){ 
        DGP_ERR_PASS(dgp_jpeg_convert_buffer_normal((Dgp_Int8 **)dest_mgr->buffer,in_dest,
                   act_img_bpr,image_def->imgType,dgp_state));
        (*dest_mgr->put_pixel_rows) (cinfo, dest_mgr,
                                      1);
      }
      DGP_ERR_PASS(dgp_jpeg_shift_to_end_of_row_testing(image_def,in_dest,cinfo,
                      dgp_state));
    } /*end of the for loop */
#endif

  DGP_ERR_PASS(dgp_jpeg_reset_address(image_def->imgType, in_dest, orig_ptr, c_ptr, m_ptr, y_ptr, k_ptr, dgp_state));

  /*Finish decompression */
#ifdef STANDALONE_TESTING
  (*dest_mgr->finish_output) (cinfo, dest_mgr,num_lines);
  fclose(output_file);
#endif

  (void) jpeg_finish_decompress(cinfo);

  DGP_ERR_PASS(dgp_internal_decomp_check(cinfo,dgp_state));

#ifdef STANDALONE_TESTING
  cinfo->output_height = cinfo->original_output_height;
  cinfo->output_width = cinfo->original_output_width;
  /*Now deallocate the buffers you allocated */
  if(image_def->imgType == DGP_CMYK){
    if(in_dest->cmyk_buf.c)
      DGP_FREE(in_dest->cmyk_buf.c);
    if(in_dest->cmyk_buf.m)
      DGP_FREE(in_dest->cmyk_buf.m);
    if(in_dest->cmyk_buf.y)
      DGP_FREE(in_dest->cmyk_buf.y);
    if(in_dest->cmyk_buf.k)
      DGP_FREE(in_dest->cmyk_buf.k);
  } else {
    if(in_dest->other_buf.buffer)
      DGP_FREE(in_dest->other_buf.buffer);
  }
  /*Deallocate the uds pointer as well */
  if(image_def->imgSrc.u)
    DGP_FREE(image_def->imgSrc.u);
#endif

  NORMAL_RETURN;  
}
