/*************************************************************************
**  Copyright 1999-2000 Xerox Corporation.All rights reserved.
**
**  DigiPaper Codebase
**    
**  Ranjani Ramamurthy, October 1999
**
**  Filename: jpeg_interface.h
**
**  Description of what's implemented in this file:
**    Interface to the jpeg6b libraries
**  
**  4.8.2000 / dhtesq.
**  Linkage note for digipaper:  horrible namespace conflicts with ipcore.
**  MK changed the names of the upper-level api functions.  We compile
**  with NEED_SHORT_EXTERNAL_NAMES as a workaround for the lower-level
**  ones.  Avoid renaming all functions, as this will make it difficult 
**  to incorporate new jpeg releases.
**************************************************************************/

#ifndef _JPEG_INTERFACE_H_
#define _JPEG_INTERFACE_H_

#define NEED_SHORT_EXTERNAL_NAMES       /* See note above. */

#include "dgp_types.h"
#include "resolution_t.h" 
#include "orientation_t.h"
#include "util.h" /*for the BITS_ULONG definition */
#include "uds.h"
#include "dgp_state.h" /*includes dgp_err_control.h */

#ifdef __cplusplus
extern "C"
{
#endif

#define DGP_INVALID_COORD          -999

/*Think of the jpeg image as a smaller rectangle embedded */
/*within a page, the larger rectangle. All values are in the */
/*destination co-ordinate space */

typedef struct{
 Dgp_Int32 dstMinX; /*X co-ord of top left corner of the buffer */
 Dgp_Int32 dstMinY; /*Y co-ord of top left corner of the buffer */
 Dgp_Int32 dstX;    /*X co-ord of the start of the jpeg image */
 Dgp_Int32 dstY;    /*Y co-ord of the start of the jpeg image */
 Dgp_Int32 srcOriginX; /*X co-ord of top left corner of source frame */
                 /*Note that this has to be in *scaled* up or down */
                 /*as you want it. If you are scaling up by 2, make */
                 /*sure that it srcOriginX is scaled up by 2 */
			     /*May indicate something other than 0 */
 Dgp_Int32 srcOriginY; /*Y co-ord of top left corner of source frame */
 Dgp_Int32 dstW;     /*Width of the destination jpeg image that is reqd */
 Dgp_Int32 dstH;     /*Height of the dest jpeg image that is reqd */
               /*Note that dstW and dstH will be different from */
			   /*the true width of the jpeg image only in the case */
			   /*of cropping or scaling */
 Dgp_Int32 bytesPerRow; /*This is of the buffer */
 Dgp_Int32 numRows; /*The height of the buffer */

 /* Added 21.6.2000 / dhtesq. */
 Dgp_Int32 nominal_page_width;  /* Width of page in RD orientation... */
 Dgp_Int32 nominal_page_height; /* Height of page in RD orientation... */
 Dgp_Int32 nominal_strip_width;  /* Strip width in pixels... */
 Dgp_Int32 nominal_strip_height; /* Strip height in pixels... */
}dgp_co_ords;

/*Note that we need the x and y resolution to be equal */
typedef struct{
  DgpCoord_Resolution srcRes;
  DgpCoord_Resolution destRes;
} dgp_resolution;

typedef enum{
  DGP_BASELINE,
  DGP_PROGRESSIVE
} dgp_jpeg_type;

typedef struct{
  Dgp_Int32      mem_estimate;
  dgp_jpeg_type  jpeg_type;
} dgp_mem_data;

/*Note that the DGP_CMYK is non-interleaved as a default */
/*and DGP_RGB is interleaved */
typedef enum{
  DGP_PACKED, DGP_RGB,
  DGP_CMYK, DGP_C,
  DGP_M, DGP_Y,
  DGP_K
} dgp_img_type;


typedef struct{
  uds* u;
  Dgp_UInt8 *buf;
  Dgp_Int32 size;
} dgp_src_def;

typedef struct{
  Dgp_Int8* buffer;
  Dgp_Int32 size;
} dgp_src_buffer;

typedef struct{
  Dgp_Int32 start_x;
  Dgp_Int32 start_y;
  Dgp_Int32 end_x;
  Dgp_Int32 end_y;
} crop_data;


/*For the dgp_cmyk_dest_buf, I need the imgType to be filled */
/*in, for the code to know which buffers to use, and which */
/*to ignore */
typedef struct{
  Dgp_UInt8* c;
  Dgp_UInt8* m;
  Dgp_UInt8* y;
  Dgp_UInt8* k;
  dgp_img_type  imgType;
  Dgp_Int32   bpr; /*bytes per row */
  crop_data crop_vals;

  /* 7.8.2000 / dhtesq. */
  /* Adding transparency mask support. */
  Dgp_UInt32 *trans_mask;
  Dgp_Int32 trans_bpl;
} dgp_cmyk_dest_buf;

typedef struct{
  Dgp_Int8* buffer;
  Dgp_Int32   bytes_per_row; /*need it for the rotation as we need this */
					         /*data in the color conversion functions. */
  crop_data crop_coords;

  /* 7.8.2000 / dhtesq. */
  /* Adding transparency mask support. */
  Dgp_UInt32 *trans_mask;
  Dgp_Int32 trans_bpl;
} dgp_other_dest_buf;

typedef union{
  dgp_cmyk_dest_buf cmyk_buf;
  dgp_other_dest_buf other_buf;
} dgp_dest_buf;

#define dgp_dest_adds dgp_dest_buf
#define dgp_save_adds dgp_dest_buf

typedef struct{
  dgp_co_ords               imgCoOrds;
  dgp_resolution            imgResolution;
  dgp_img_type              imgType;
  dgp_mem_data              memData;
  dgp_src_def               imgSrc;
  DgpCoord_Orientation      imgOrientation;
} DGP_JPEGImageDef;

#define DGP_IMAGE_TYPE_BIT           0x00000001
#define DGP_ORIENTATION_BIT          0x00000010
#define DGP_RESOLUTION_BIT           0x00000100
#define DGP_COORD_BIT				 0x00001000
#define DGP_COORD_DIFF_BIT   		 0x00010000
#define DGP_SZ_BIT				     0x00100000

typedef struct{
  Dgp_Int32       imgWidth;    /*Width of JPEG image */
  Dgp_Int32       imgHeight;   /*Height of JPEG image */
  dgp_img_type    imgType;     /*Should be DGP_RGB */
  dgp_src_buffer  imgSrc;      /*Source buffer to compress */
} DGP_JPEGCompressDef;

typedef struct{
  Dgp_Int8*   buffer;
  Dgp_Int32   outLen;
} DGP_JPEGOutputDef;

/* COMPRESSION FUNCTIONS */
extern ErrCode 
DGP_JPEG_init_jcompress(j_compress_ptr cinfo, Dgp_State* dgp_state);

extern ErrCode 
Dgp_JPEG_destroy_jcompress(j_compress_ptr cinfo, Dgp_State* dgp_state);

extern ErrCode
Dgp_JPEG_create_jpeg(DGP_JPEGCompressDef* image_def,
            DGP_JPEGOutputDef* output,
            j_compress_ptr cinfo, Dgp_State* dgp_state);


/* DECOMPRESSION FUNCTIONS */

extern ErrCode
DGP_JPEG_check_jdecompress(DGP_JPEGImageDef* image_def,
				  Dgp_UInt32* ret_val,
                  j_decompress_ptr cinfo, Dgp_State* dgp_state);
extern ErrCode 
DGP_JPEG_init_jdecompress(j_decompress_ptr cinfo, Dgp_State* dgp_state);

extern ErrCode 
Dgp_JPEG_destroy_jdecompress(j_decompress_ptr cinfo, Dgp_State* dgp_state);

#ifdef STANDALONE_TESTING
extern ErrCode
Dgp_JPEG_decompress_jpeg(DGP_JPEGImageDef* image_def,
            dgp_dest_buf* in_dest,
            j_decompress_ptr cinfo, Dgp_State* dgp_state,
            FILE* output_file);
#else
extern ErrCode
Dgp_JPEG_decompress_jpeg(DGP_JPEGImageDef* image_def,
            dgp_dest_buf* in_dest,
            j_decompress_ptr cinfo, Dgp_State* dgp_state);
#endif

#ifdef __cplusplus
}
#endif

#endif

