/*************************************************************************
**  Copyright 1999-2000 Xerox Corporation.All rights reserved.
**
**  DigiPaper Codebase
**    
**  Ranjani Ramamurthy, October 1999
**
**  Filename: jpeg_mem_mgr.c
**
**  Description of what's implemented in this file:
**    
**    Two source managers that could be used. THe UDS source
**    manager and the Memory Source manager (in case you just
**    want to use a memory mapped file as input).
**
**************************************************************************/

#include "jpeg_mem_mgr.h"
#include "jpeg6blib.h"  /* ...for fill_input_buffer. / dhtesq. */
#include "uds.h"

/* A data src needs the following functions defined */
/* */
/* 1) METHODDEF void init_source (j_decompress_ptr cinfo) */
/*    Any initialization of src data structure */
/* 2) METHODDEF boolean fill_input_buffer(j_decompress_ptr) */
/* 3) METHODDEF void skip_input_data(j_decompress_ptr cinfo, */
/*								long num_bytes) */
/* 4) METHODDEF void term_source (j_decompress_ptr cinfo) */
/* 5) GLOBAL void jpeg_***_src(j_decompress_ptr cinfo, */
/*							   input_data */
/* */

#define DGP_INPUT_BUF_SIZE  4096

/*Called by jpeg_read_header */
METHODDEF (void)
DGP_MSM_init_source(j_decompress_ptr cinfo)
{
  DGPMemSrcMgr *src = (DGPMemSrcMgr*) cinfo->src;

  /* we reset the empty input file flag for each image, but we
  dont clear the input buffer. this is correct behavior for reading
  a series of images from one source
  */
  src->start_of_file = TRUE;

  /* Insert a fake EOT marker */
  src->eof_buffer[0] = (JOCTET) 0xFF;
  src->eof_buffer[1] = (JOCTET) JPEG_EOI;
}

/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

METHODDEF(boolean)
DGP_MSM_fill_input_buffer (j_decompress_ptr cinfo)
{
  DGPMemSrcMgr* src = (DGPMemSrcMgr*) cinfo->src;
  size_t nbytes;

  if(src->currentOffset < src->size){
    nbytes = src->size;
    src->buffer = (const JOCTET*) src->MMapAddress;
  } else {
    nbytes = 0;
  }

  if (nbytes <= 0) {
    if (src->start_of_file)	/* Treat empty input file as fatal error */
      ERREXIT(cinfo, JERR_INPUT_EMPTY);
    WARNMS(cinfo, JWRN_JPEG_EOF);
    /* Insert a fake EOI marker */
    src->buffer = src->eof_buffer;
    nbytes = 2;
  }

  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file = FALSE;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

METHODDEF(void)
DGP_MSM_skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  DGPMemSrcMgr* src = (DGPMemSrcMgr*) cinfo->src;
  if (num_bytes > 0) {
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF(void)
DGP_MSM_term_source (j_decompress_ptr cinfo)
{
  cinfo = cinfo;
  /* no work necessary here */
}

void
InitDGPMemSrcMgr(j_decompress_ptr cinfo,
                 char* mmapaddress,
		 int size)
{
  DGPMemSrcMgr* src;
  
  if(cinfo->src == NULL){
    cinfo->src =(struct jpeg_source_mgr*)
      (*cinfo->mem->alloc_small)((j_common_ptr) cinfo,
       JPOOL_PERMANENT, SIZEOF(DGPMemSrcMgr));
    /*No Need to allocate buffer, instead point to appropriate */
    /*area of memory */
    src = (DGPMemSrcMgr*) cinfo->src;
    src->buffer = (const JOCTET*) src->MMapAddress;
  }
  
  src = (DGPMemSrcMgr*) cinfo->src;
  src->pub.init_source = DGP_MSM_init_source;
  src->pub.fill_input_buffer = DGP_MSM_fill_input_buffer;
  src->pub.skip_input_data =DGP_MSM_skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /*default */
  src->pub.term_source = DGP_MSM_term_source;

  src->MMapAddress = (JOCTET*)mmapaddress;
  src->size = size;
  src->currentOffset = 0;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on */
						        /* first read */
  src->pub.next_input_byte = NULL; /*till buffer loaded */
  return;
}

/*Called by jpeg_read_header */
METHODDEF (void)
DGP_UDS_init_source(j_decompress_ptr cinfo)
{
  DGPUdsSrcMgr *src = (DGPUdsSrcMgr*) cinfo->src;

  /* we reset the empty input file flag for each image, but we
  dont clear the input buffer. this is correct behavior for reading
  a series of images from one source
  */
  src->start_of_file = TRUE;

  /* Insert a fake EOT marker */
  src->eof_buffer[0] = (JOCTET) 0xFF;
  src->eof_buffer[1] = (JOCTET) JPEG_EOI;
}

METHODDEF(void)
DGP_UDS_term_source (j_decompress_ptr cinfo)
{
  cinfo = cinfo;
  /* no work necessary here */
}

boolean
DGP_UDS_fill_input_buffer (j_decompress_ptr cinfo)
{

  DGPUdsSrcMgr* src = (DGPUdsSrcMgr*) cinfo->src;
  uds* uds_ptr = src->uds_ptr;
  size_t nbytes; int cur_ptr=0;
  /*void* buf; */

  /*buf = malloc(DGP_INPUT_BUF_SIZE * sizeof(char)); */
  
  cur_ptr =(uds_ptr->ftell)(src->address);
  nbytes = (uds_ptr->fread)(src->buffer,1,DGP_INPUT_BUF_SIZE,src->address);
  cur_ptr =(uds_ptr->ftell)(src->address);
  /*nbytes = fread(src->buffer,(size_t)1,DGP_INPUT_BUF_SIZE,(FILE*)src->address); */
  if (nbytes <= 0) {
    if (src->start_of_file)	/* Treat empty input file as fatal error */
      ERREXIT(cinfo, JERR_INPUT_EMPTY);
    WARNMS(cinfo, JWRN_JPEG_EOF);
    /* Insert a fake EOI marker */
    src->buffer[0] = (JOCTET) 0xFF;
    src->buffer[1] = (JOCTET) JPEG_EOI;
    nbytes = 2;
  }

  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->currentOffset += nbytes;
  src->start_of_file = FALSE;

  return TRUE;
}

/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

#include "jpeg_mem_mgr.h"

void
DGP_UDS_skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  DGPUdsSrcMgr* src = (DGPUdsSrcMgr*) cinfo->src;
  /*uds* uds_ptr = src->uds_ptr; */

  /*
  28.6.2000 / dhtesq.
  This appears not to handle the case where num_bytes is greater than src->pub.bytes_in_buffer.
  I think they were trying for efficiency by using fseek, but this doesn't do enough bookkeeping.
  Fixing...
  */
  if (num_bytes > 0) {
    while (num_bytes > (long) src->pub.bytes_in_buffer) {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      (void) DGP_UDS_fill_input_buffer(cinfo);
      /* note we assume that fill_input_buffer will never return FALSE,
       * so suspension need not be handled.
       */
    }

    /* (uds_ptr->fseek)(src->address,num_bytes,SEEK_CUR); */
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }
}

void
InitDGPUdsSrcMgr(j_decompress_ptr cinfo,
                 uds* in_uds,
				 int size)
{
  DGPUdsSrcMgr* src;
  
  if(cinfo->src == NULL){
    cinfo->src =(struct jpeg_source_mgr*)
       (*cinfo->mem->alloc_small)((j_common_ptr) cinfo,
       JPOOL_PERMANENT, SIZEOF(DGPUdsSrcMgr));
    src = (DGPUdsSrcMgr*) cinfo->src;
    src->buffer = (JOCTET *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  DGP_INPUT_BUF_SIZE * SIZEOF(JOCTET));
  }
  
  src = (DGPUdsSrcMgr*) cinfo->src;
  src->pub.init_source = DGP_UDS_init_source;
  src->pub.fill_input_buffer = DGP_UDS_fill_input_buffer;
  src->pub.skip_input_data =DGP_UDS_skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /*default */
  src->pub.term_source = DGP_UDS_term_source;

  src->address = (JOCTET*)in_uds->fp;
  src->size = size;
  src->currentOffset = 0;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on */
						        /* first read */
  src->pub.next_input_byte = NULL; /*till buffer loaded */
  src->uds_ptr = in_uds;  
}
