/*************************************************************************
**
**  DigiPaper Codebase
**    
**  Dillon H. Tracy esq., 5 April 2001
**
**  Filename: dgpjpegls.c
**
**  Description of what's implemented in this file:
**    'main' testing function for libjpegls.
**  
**
**************************************************************************/

#include "dgp_cl.h"
#include "image.h"
#include "image_defs.h"
#include "iowrap.h"
#include "dgpjpegls.h"
#include <stdio.h>
#include "dgp_state.h"
#include "dgp_err_printers.h"


/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libjpegls/dgpjpegls.c,v 1.2 2003/02/10 22:53:29 mkrocken Exp $
 *
 */


/*
 * Default values for command line arguments.
 */
#define DEF_OUTFILENAME "a.pgm"
#define DEF_VERBOSE     DGP_TRUE

typedef struct {
  Dgp_Int8 *out_filename;
  Dgp_Bool verbose;

} CLOpts;

static CLOpts cl_opts = {   DEF_OUTFILENAME, DEF_VERBOSE };

/* Displayed when showing usage information. */
static const Dgp_Int8 *cl_header = "Usage: dpgjpegls [options] foo.jls";

/*
 * The definition of the command line options.
 */
static DgpCl_Option cl_cmdln_options[] = {
  DGPCL_HELP_CMD,
  DGPCL_VERBOSE_CMD(cl_opts.verbose),
  DGPCL_OUTFNAME_CMD(cl_opts.out_filename, DEF_OUTFILENAME),
  {NULL}
};

/*************************************************************************/
/* An argument validator... */
static ErrCode
validateArgs(Dgp_State *dgp_state,
             Dgp_Int32 unparsed_argc, 
             Dgp_Int8 **unparsed_argv,
             FILE *err_out,
             void *validate_cookie)
{
  DGP_PROLOGUE("validateArgs()", DGP_PERRSTATE);
  CLOpts *my_opts = (CLOpts *)validate_cookie;

  unparsed_argc = unparsed_argc;
  unparsed_argv = unparsed_argv;
      
  NORMAL_RETURN;
}
/*************************************************************************/


Dgp_Int32
main(Dgp_Int32 argc, Dgp_Int8 **argv)
{
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);

  Dgp_Int8			*infname = NULL, *outfname = NULL, *errfname = NULL, *minus_path = NULL;
  uds				in, out;
  DgpJpegLsContext	*c = NULL;

  Dgp_State			*dgp_state = NULL;
  Dgp_Float64		timeTaken = 0.0;
  DgpImage_Gray		hostImage = NULL;
  Dgp_Int32			frameMargin = 100;
  Dgp_Int32			dstLpl;

#ifdef _WIN32
  Dgp_Int8          filesep = '\\';
#else
  Dgp_Int8          filesep = '/';
#endif

  if(argc == 1){
      DgpCl_printHelp(stderr, cl_header, cl_cmdln_options);
      MAIN_NORMAL_RETURN;
  }
  in.fp = NULL;

  /* Init dgp state. */
  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, NULL, NULL, stderr, Dgp_fileErrPrinter, DGP_TRUE, DGP_FALSE));
  DGP_SET_ERR_STATE(DGP_PERRSTATE);

  DGP_ERR_CHECK_NO_MSG(DgpCl_parse(dgp_state, &argc, argv, cl_cmdln_options, stderr, cl_header, validateArgs, &cl_opts)){
      if(DGP_ERROR_IS(cErrInvalidCmdLine)){
          MAIN_NORMAL_RETURN;
      }
  }
  if(!argc){
	  DGPCL_PRINT_ERRF0(stderr, "No input file specified...");
	  DGP_ERR(cErrInvalidCmdLine);
  }

  infname = argv[0];

  if((cl_opts.out_filename == NULL) || (cl_cmdln_options[2].found == DGP_FALSE)) {
      /* The user has not specified the output filename... */
      DGP_CHECK_ALLOC_CLEANUP(outfname = (Dgp_Int8 *)DGP_MALLOC((strlen(infname) + 5) * sizeof(Dgp_Int8)));
      outfname[0] = '\0';      
      sprintf(outfname, "%s.pgm", infname);
  } else {
      /* Copy the command line string, unless stdout has been specified. */
      if(!strcmp(cl_opts.out_filename, "stdout")){
          outfname = NULL;
      } else {
          /* -o argument could be a path or a destination filename. */
          /* If it ends in a filesep, use it as a prepended path. */
          if(cl_opts.out_filename[strlen(cl_opts.out_filename) - 1] == filesep){
              /* Prepend path... */
              minus_path = strrchr(infname, filesep);
              if(minus_path == NULL){
                  minus_path = infname;
              } else {
                  minus_path += 1;
              }
              DGP_CHECK_ALLOC_CLEANUP(outfname = (Dgp_Int8 *)DGP_MALLOC((strlen(cl_opts.out_filename) + strlen(minus_path) + 6) 
                  * sizeof(Dgp_Int8)));
              strcpy(outfname, cl_opts.out_filename);
              outfname = strcat(outfname, minus_path);
              outfname = strcat(outfname, ".pgm");
          } else {
              /* -o argument is destination filename, no manipulation required. */
              DGP_CHECK_ALLOC_CLEANUP(outfname = (Dgp_Int8 *)DGP_MALLOC((strlen(cl_opts.out_filename) + 1) * sizeof(Dgp_Int8)));
              strcpy(outfname, cl_opts.out_filename);
          }
      }
  }

  in.fseek  = (fseek_func)fseek;
  in.ftell  = (ftell_func)ftell;
  in.fread  = (fread_func)fread;
  in.fwrite = (fwrite_func)fwrite;
  in.ffeof = (feof_func)feof;

  /* Open all files needed. */
  DGP_ERR_GEN_CLEANUPF1((!(in.fp = fopen(infname, "rb"))), cErrFileNotFound, "I can't open file \'%s\' for reading.", infname);

  /* Initialize decoder. */
  DGP_ERR_PASS_CLEANUP(DgpJpegLsInitializeDecode(&c, dgp_state, &in, cl_opts.verbose));

  /* A destination image... */
  DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(	dgp_state, 
										DGPIMAGE_GRAY, 
										c->columns + frameMargin, 
										c->rows + frameMargin, 
										(void **)&hostImage));
  dstLpl = (((Dgp_Int8 *)DGPIMG_PIXPTR(hostImage, 0, 1) - (Dgp_Int8 *)DGPIMG_PIXPTR(hostImage, 0, 0)) + 3) / 4;

  /* Generate output uds... */
  DGP_ERR_PASS_CLEANUP(DgpJpegLsFOpenImage(	dgp_state, 
											DGPIMG_PIXPTR(hostImage, frameMargin / 2, frameMargin / 2),
											DGPIMG_WIDTH(hostImage),	/* width */
											DGPIMG_HEIGHT(hostImage),	/* height */
											dstLpl,						/* longs per row */
											frameMargin / 2,			/* x offset into host image */
											DGPIMG_XMAX(hostImage),		/* max x */
											DGPIMG_YMAX(hostImage),		/* max y */
											8,							/* bpp */
											&out));

  /* Decode input file. */
  DGP_ERR_PASS_CLEANUP(DgpJpegLsDecode(c, &out, &timeTaken));
  /* N.B. context has now been destroyed... */

  DGP_REPORTF2(cl_opts.verbose, "%.3f secs to decode %s.", timeTaken, infname);

  DGP_ERR_PASS_CLEANUP(DGPIMAGE_SAVE((void *)hostImage, outfname));

  /* Close all files. */
  DGP_ERR_GEN_CLEANUPF0((fclose(in.fp) < 0), cErrIO_ReadError, "I can't close the input file.");
  DGP_ERR_PASS_CLEANUP(DgpJpegLsFCloseImage(dgp_state, &out));


  if(outfname) DGP_FREE(outfname);
  DGPIMAGE_FREE(hostImage);

  Dgp_cleanupDgpState(&dgp_state);
  
  MAIN_NORMAL_RETURN;

ERR_CLEANUP:
  if(outfname) DGP_FREE(outfname);
  if(in.fp) fclose(in.fp);
  if(dgp_state) Dgp_cleanupDgpState(&dgp_state);

  MAIN_ERR_RETURN;
}
