/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#ifndef IOFILE_H_INCLUDED
#define IOFILE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(iofile_hRCSInfo, "$RCSfile: iofile.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* iofile.h
 *
 * DESCRIPTION
 *	 Declarations for IO file level structures (and above, like instance
 *	 handles).
 *
 */

/*
 * INCLUDES
 */
#include "iomrc.h"

/*
 * CONSTANTS
 */
#define IO_MAGIC_NUM	0x123abc00		/* Instance identifier */
#define IO_HANDLE_TYPE	(0x00010000 * 'I' + 0x00000100 * 'O')

/*
 * MACROS
 */

/*
 * XERR_GEN
 *  Generate an error. "cond" is some condition (probably, it's just tested
 *  for non-zero) and "err" is the error code to generate 
 */
    
#ifdef SHOWERR
#define XERR_GEN(cond, err)\
do { \
    if (cond) \
    {\
        if (err) { \
	    io_error(err, __FILE__, __LINE__);\
	} \
        return err;\
    }\
} while(0)
#else
#define XERR_GEN(cond, err)\
do { \
    if (cond) \
    {\
        return err;\
    }\
} while(0)
#endif

/*
 * XERR_GEN_CLEANUP
 *   Error generation: instead of returning, as in XERR_GEN, go to a cleanup
 *   action at the end of the function. 
 */

#ifdef SHOWERR
#define XERR_GEN_CLEANUP(cond, err)\
do {\
    if (cond)\
    {\
        if (err) { \
	    io_error(err, __FILE__, __LINE__);\
	} \
        __error__ = err;\
        goto ERR_CLEANUP;\
    }\
} while(0)
#else
#define XERR_GEN_CLEANUP(cond, err)\
do {\
    if (cond)\
    {\
        __error__ = err;\
        goto ERR_CLEANUP;\
    }\
} while(0)
#endif

/*
 * XERR_PASS
 *   This is an error check on a function call. It assumes that the
 *   function returns an error code, and if the error code is non-zero,
 *   returns from the function (after possibly printing an error message).
 */

#ifdef SHOWERR
#define XERR_PASS(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
	io_error(__error__, __FILE__, __LINE__);\
        return __error__;\
    }\
} while (0)
#else
#define XERR_PASS(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
        return __error__;\
    }\
} while (0)
#endif

/*
 * XERR_PASS_CLEANUP
 *   This is an error check on a function call. It is like XERR_PASS except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERROR_RETURN" macro at the end of it.
 */

#ifdef SHOWERR
#define XERR_PASS_CLEANUP(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
	io_error(__error__, __FILE__, __LINE__);\
	goto ERR_CLEANUP;\
    }\
} while (0)
#else
#define XERR_PASS_CLEANUP(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
        goto ERR_CLEANUP;\
    }\
} while (0)
#endif

/*
 * XERR_WARNOK_CLEANUP
 *   This is an error check on a function call. It is like XERR_PASS_CLEANUP
 *   except that warnings are allowed.
 */

#ifdef SHOWERR
#define XERR_WARNOK_CLEANUP(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
	io_error(__error__, __FILE__, __LINE__);\
        if (__error__ < IO_WARNINGS)\
	    goto ERR_CLEANUP;\
	__error__ = 0;\
    }\
} while (0)
#else
#define XERR_WARNOK_CLEANUP(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != IO_NOERROR)\
    {\
        if (__error__ < IO_WARNINGS)\
            goto ERR_CLEANUP;\
	__error__ = 0;\
    }\
} while (0)
#endif


/*
 * XERR_IGNORE
 *  Ignore the return error.
 *
 *  This is used basically to keep the source code browsing parser happy,
 *  just to know that the programmer didn't miss catching a function return.
 */ 

#define XERR_IGNORE(functionCall)	functionCall


/* I/O Functions */
#define IO_FILETOKEN(ioFile)			(&(ioFile)->fileToken)
#define IO_READ(ioFile, buffer, bytes)		\
	if (IO_FILETOKEN(ioFile)->FileRead(IO_FILETOKEN(ioFile)->dwClientFileID, \
	    (Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = IO_READ_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define IO_WRITE(ioFile, buffer, bytes)		\
	if (IO_FILETOKEN(ioFile)->FileWrite(IO_FILETOKEN(ioFile)->dwClientFileID, \
	    (const Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = IO_WRITE_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define IO_SEEK(ioFile, offset)			\
	if (IO_FILETOKEN(ioFile)->FileSeek(IO_FILETOKEN(ioFile)->dwClientFileID, \
	    offset) != (Int32)(offset)) {	\
	    __error__ = IO_SEEK_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define IO_FILESIZE(ioFile, size)		\
	if (((size) = IO_FILETOKEN(ioFile)->FileSize(IO_FILETOKEN(ioFile)->\
		dwClientFileID)) < 0) {		\
	    __error__ = IO_FILELEN_ERR;		\
	    goto ERR_CLEANUP;			\
	}

/* Malloc and free, must have a local variable *instData which is active. */
#ifdef SHOWERR
#define IO_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID \
    , __FILE__, __LINE__    \
)
#else
#define IO_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID \
)
#endif
#define IPCORE_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID)

#define IO_FREE(buffer)			\
	instData->free(buffer, 0, instData->mallocID)

#define IO_CHECKMALLOC(call)			\
	if ((call) == NULL) {			\
	    __error__ = IO_NOMEMORY;		\
	    goto ERR_CLEANUP;			\
	}

#ifdef SHOWERR
#define IO_CheckHandleValidity(structPtr)		\
	if (structPtr == 0 ||			\
		((IO_INSTANCE *)structPtr)->handleID != (UInt32)structPtr) {  \
		io_error(__error__, __FILE__, __LINE__);\
		__error__ = IO_BADPARAMETER;	\
		ERR_RETURN;			\
	} else if (((IO_INSTANCE *)structPtr)->handleType != (UInt32)IO_HANDLE_TYPE) {  \
		__error__ = IO_WARN_NOSUPPORT;	\
		ERR_RETURN;			\
	}
#define IO_CheckStructureSize(structPtr)		\
	if (structPtr == NULL ||		\
		structPtr->dwSize != sizeof(*structPtr)) {	\
		io_error(__error__, __FILE__, __LINE__);\
		__error__ = IO_BADPARAMETER;	\
		ERR_RETURN;			\
	}
#else
#define IO_CheckHandleValidity(structPtr)		\
	if (structPtr == 0 ||			\
		((IO_INSTANCE *)structPtr)->handleID != (UInt32)structPtr) {  \
		__error__ = IO_BADPARAMETER;	\
		ERR_RETURN;			\
	} else if (((IO_INSTANCE *)structPtr)->handleType != (UInt32)IO_HANDLE_TYPE) {  \
		__error__ = IO_WARN_NOSUPPORT;	\
		ERR_RETURN;			\
	}
#define IO_CheckStructureSize(structPtr)		\
	if (structPtr == NULL ||		\
		structPtr->dwSize != sizeof(*structPtr)) {	\
		__error__ = IO_BADPARAMETER;	\
		ERR_RETURN;			\
	}
#endif

/*
 * TYPEDEFS
 */

/* Document token
 */
typedef struct IO_TOKEN
{
    UInt32 dwSize;		/* Size of this structure */
    Int32 dwClientFileID;	/* Client's file identifier (fd, fp, structure,
				 * etc.) */
    IO_READFUNC FileRead;	/* Read function, set by client */
    IO_WRITEFUNC FileWrite;	/* Write function, set by client */
    IO_SEEKFUNC FileSeek;	/* Seek function, set by client */
    IO_SIZEFUNC FileSize;	/* File size function, set by client */
    IO_TELLFUNC FileTell;	/* File tell function, set by client */
    UInt32 dwCurPos;		/* Current file position */
    Bool bSwapBytes;		/* File must be byte swapped for endian
				 * differences */
} IO_TOKEN;


typedef struct IO_FILE {
    UInt32 handleID;		/* Handle ID, consistency check */
    IO_TOKEN fileToken;		/* File token (see io.prv) */
    struct IO_INSTANCE *instance;	/* IO INSTANCE */
} IO_FILE;

typedef struct IO_INSTANCE {
    UInt32 handleID;		/* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    UInt32 nextDocID;		/* Unique ID assigned to each doc */
    UInt32 ioMagicNum;		/* IO instance identifier */
    UInt32 clientInstID;	/* Client passed instance identifier */
    IO_MALLOCFUNC malloc;	/* Client malloc function */
    IO_CALLOCFUNC calloc;	/* Calloc function, that runs on top of malloc */
    IO_REALLOCFUNC realloc;	/* Realloc function, that runs on top of
				 * malloc and free */
    IO_FREEFUNC free;		/* Client free function */
    void *mallocID;		/* Malloc ID sent to malloc/free */
    void *defaultMallocID;	/* Reserved default Malloc ID */
    ClientErrorProc errProc;	/* Client error callback function */
    void *errInfo;		/* Client error cookie */
    UInt32 pIPWState;		/* Handle for IPCore IPWrap layer */
    UInt32 myIPWState;		/* Boolean for private (vs. client)
				 * IPWrap handle */
    UInt32 xifInstance;		/* XIF instance handle */
    UInt32 mrcInstance;		/* MRC instance handle */
    UInt32 tfxInstance;		/* TIFF-FX/Coh TIFF instance handle */
} IO_INSTANCE;

/*
 * PROTOTYPES
 */

void wp_swaps(Int8 *x);
void wp_swapl(Int8 *x);
void wp_swapd(Int8 *x);

#ifdef SHOWERR
void io_error(int error, const char *file, long line);
#endif


#ifdef __cplusplus
}
#endif

#endif /* IOFILE_H_INCLUDED */
