#include "reorient.h"
#include "rot_binary.h"
#include "orientation.h"
#include "PsgFix.h"

#include <math.h>
/* reorient.h
   higher level primitives that use the rotate primitives; presents an 
   interface for converting images from one DgpCoord_Orientation to another.
*/


/* reorientBinaryNeeded()
   If I'm planning on reorienting a binary image, what kind of output do I 
   need to allocate.

   in - the DgpCoord_Orientation of the input image
   out - the desired output DgpCoord_Orientation.
   srcW - the image's fast width in it's input DgpCoord_Orientation.
   srcH - the image's slow width in it's input DgpCoord_Orientation.
   srcBpl - bytes per line in the source.
   will be set if valid pointers:
   dstW - the fast width of the output image 
   dstH - the slow width of the output image
   dstBpl - bytes per line in the dest.
   inplace - can this transformation be done in place with the rotation
     primitives?
*/
ErrCode
DgpRot_reorientBinaryNeeded(DgpCoord_Orientation in, DgpCoord_Orientation out, 
                     Dgp_Int32 srcW, Dgp_Int32 srcH, Dgp_Int32 srcBpl,
                     Dgp_Int32 * dstW, Dgp_Int32 * dstH, Dgp_Int32 * dstBpl,
                     Dgp_Bool * inplace, Dgp_State * dgp_state)
{
  Dgp_Bool rotated;
  DGP_STATE_PROLOGUE("reorientBinaryNeeded()", dgp_state);
  DGP_STOP_WHINE;
  DGP_ASSERT(dgp_state);


  if (in == out){
    /* ok, this is a real easy one. */
    if (dstW){
      *dstW = srcW;
    }
    if (dstH){
      *dstH = srcH;
    }
    if (dstBpl){
      *dstBpl = srcBpl;
    }
    if (inplace){
      *inplace = DGP_TRUE;
    }
  }
  else {
    rotated = DgpCoord_isRotated(in, out);
    if (rotated){
      if (dstW){
        *dstW = srcH;
      }
      if (dstH){
        *dstH = srcW;
      }
      /* we have no special alignment constraints on rotation. */
      if (dstBpl){
        *dstBpl = (int)ceil(srcH / (8.0 * sizeof(Dgp_UInt32)))
                       * sizeof(Dgp_UInt32);
      }
      if (inplace){
        *inplace = DGP_FALSE;
      }
    }
    else {
      if (dstW){
        *dstW = srcW;
      }
      if (dstH){
        *dstH = srcH;
      }
      if (dstBpl){
        *dstBpl = srcBpl;
      }

      if (inplace){
        *inplace = DGP_TRUE;
      }
    }
  }

  NORMAL_RETURN;
}



/* reorientBinary()
   Reorient the following binary image into the given output image.

   in - the DgpCoord_Orientation of the input image.
   inp - a pointer to (0,0) of the input image.
   srcW - the image's fast width in it's input DgpCoord_Orientation.
   srcH - the image's slow width in it's input DgpCoord_Orientation.
   srcBpl - bytes per line in the source.
   dstW - the fast width of the output image 
   dstH - the slow width of the output image
   dstBpl - bytes per line in the dest.

   If reorientBinaryNeeded() told you that this could be done in place, feel
   free to pass in the same pointers.
*/

ErrCode
DgpRot_reorientBinary(DgpCoord_Orientation in, Dgp_UInt32 * inp,
               Dgp_Int32 srcW, Dgp_Int32 srcH, Dgp_Int32 srcBpl,
               DgpCoord_Orientation out, Dgp_UInt32 * outp,
               Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Int32 dstBpl, 
               Dgp_State * dgp_state)
{
  Dgp_Bool rotleft90 = DGP_FALSE;
  Dgp_Bool rotright90 = DGP_FALSE;
  Dgp_Bool rot180 = DGP_FALSE;
  DGP_STATE_PROLOGUE("reorientBinary()", dgp_state);
  DGP_ASSERT(outp);
  DGP_ASSERT(inp);
  DGP_ASSERT(srcBpl);
  DGP_ASSERT(dstBpl);
  DGP_ASSERT(dstW > 0 && dstH > 0);
  DGP_ASSERT(srcW > 0 && srcH > 0);

  (void)dstW;
  (void)dstH;
  
  if (in == out){
    NORMAL_RETURN;
  }
    
  switch(in){
  case ORIENT_RD:
    switch (out){
    case ORIENT_UR:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_DL:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_LU:
      /* 180; */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_UR:
    switch (out){
    case ORIENT_LU:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_RD:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_DL:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_LU:
    switch (out){
    case ORIENT_DL:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_UR:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_RD:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_DL:
    switch (out){
    case ORIENT_RD:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_LU:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_UR:
      /* 180 */
      rot180 = DGP_TRUE;
      break;      
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_LD:
    switch (out){
    case ORIENT_UL:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_DR:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_RU:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_UL:
    switch (out){
    case ORIENT_RU:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_LD:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_DR:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_RU:
    switch (out){
    case ORIENT_DR:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_UL:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_LD:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  case ORIENT_DR:
    switch (out){
    case ORIENT_LD:
      /* right 90; */
      rotright90 = DGP_TRUE;
      break;
    case ORIENT_RU:
      /* left 90; */
      rotleft90 = DGP_TRUE;
      break;
    case ORIENT_UL:
      /* 180 */
      rot180 = DGP_TRUE;
      break;
    default:
      DGP_WARNF0("reorientBinary: unsupported transformation\n");
      DGP_ASSERT(0);
      break;
    }
    break;
  }

  if (rotright90){
    DgpRot_rotateBinary90CW(inp, srcW, srcH, srcBpl,
                     outp, dstBpl);
  }
  else if (rotleft90){
    DgpRot_rotateBinary90CCW(inp, srcW, srcH, srcBpl,
                      outp, dstBpl);
  }
  else if (rot180){
    /* FIX: we don't handle 180 rotations that aren't in place. */
    /* DGP_WARNF0("reorientBinary: 180 degree rotations that are not inplace are not supported.\n"); */
    DGP_ASSERT(inp == outp && dstBpl == srcBpl);
    DgpRot_rotateBinary180InPlace(inp, srcW, srcH, srcBpl);
  }
  NORMAL_RETURN;
}
              
