/*************************************************************************
**	  
**  Author:  Dave Abercrombie
**
**  Filename:  decode_api.c
**
**  
**************************************************************************/

#include "decode_api.h"

#include "uds.h"
#include "dgp_state.h"
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "image_defs.h"
#include "jb2mmuncompress.h"
#include "blit.h"


/* This is a decode API for decoding JBIG2 dictionary and position block
 * streams.
 *
 * The following functions can be used, and in this order:
 *	DgpDec_init(DecodeHandle *decodeHandle, Dgp_Allocators *a,
 *		Dgp_SmallAllocators *sm_a)
 *	DgpDec_decodeDict(DecodeHandle decodeHandle, Dgp_UInt8 *dictData,
 *		Dgp_UInt32 dictBytes, Dgp_UInt32 *dictHandle)
 *	DgpDec_buildDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList,
 *		Dgp_UInt32 dictHandle)
 *	DgpDec_renderBinaryImage(DecodeHandle decodeHandle,
 *		Dgp_UInt32 dictList, Dgp_UInt8 *posBlock,
 *		Dgp_UInt32 posBlockBytes, Dgp_UInt8 *image,
 *		Dgp_UInt32 width, Dgp_UInt32 height, Dgp_UInt32 longsPerLine)
 *	DgpDec_destroyDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList)
 *	DgpDec_destroyDict(DecodeHandle decodeHandle, Dgp_UInt32 *dictHandle)
 *	DgpDec_close(DecodeHandle *decodeHandle)
 *
 * A further explanation is warranted, but I haven't yet written it.  For
 * an example of its use, you could browse through mrcio/iolib/tifffx/tfxrjb2.c
 */

/* Structures:
 */

/* Internal version of decode handle */
typedef struct _DecodeHandle {
    Dgp_State  *dgp_state;	/* digipaper handle */
    DgpAdt_List dicts_list;	/* decoded dictionary segments list */
    DgpAdt_List dictLists_list;	/* decoded dictionary lists list
				 * (a list of lists) */
} _DecodeHandle;



/********************************************************
** Function: generate DictList and Dgp_State
** input:  None
** out:    DictList *
*********************************************************/
/* Initialize */
DGP_LIBEXPORT ErrCode
DgpDec_init(
    DecodeHandle *decodeHandle,
    Dgp_Allocators *a,
    Dgp_SmallAllocators *sm_a
)
{
    _DecodeHandle *decHandle = NULL;
    Dgp_State *dgp_state;
    DGP_MAIN_PROLOGUE("DgpDec_init()", stderr, Dgp_fileErrPrinter);

    /* Check the parameter */
    DGP_ERR_GEN(decodeHandle == NULL, cErrInvalidParm);
    *decodeHandle = 0UL;	/* Clear the return pointer */

    /* Construct a new decode handle */
    if (a) {
	decHandle = (_DecodeHandle *)a->calloc(1, sizeof(_DecodeHandle));
    } else {
	decHandle = (_DecodeHandle *)calloc(1, sizeof(_DecodeHandle));
    }
    DGP_ERR_GEN_CLEANUP(decHandle == NULL, cErrInsufficientMem);
    *decodeHandle = (DecodeHandle)decHandle;

    /* Initialize the Digipaper state */
    DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, a, sm_a, stderr,
	Dgp_fileErrPrinter, 1, 0));

    /* Retain the Digipaper state */
    decHandle->dgp_state = dgp_state;

    /* Create the decoded dictionary list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&decHandle->dicts_list, dgp_state));

    /* Create the decoded dictionary lists, list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&decHandle->dictLists_list, dgp_state));


ERR_CLEANUP:
    return 0;
}


/* Decode a dictionary (to be retained for possible multiple renders)
 */
DGP_LIBEXPORT ErrCode
DgpDec_decodeDict(DecodeHandle decodeHandle, Dgp_UInt8 *dictData,
    Dgp_UInt32 dictBytes, Dgp_UInt32 *dictHandle)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    jb2Segment *decodedSegment;
    DGP_PROLOGUE( "DgpDec_decodeDict()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN(decodeHandle == 0UL, cErrInvalidParm);

    /* Decode the dictionary segment(s): looks like this actually picks out
     * the dictionary from the segments */
    DGP_ERR_PASS_CLEANUP(jb2_decodeTokDictSeg(&decodedSegment, 
	dictData, dictBytes, dgp_state));

    /* Retain the decoded dictionaries in the global list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(decHandle->dicts_list, decodedSegment,
	NULL));

    /* Return the decoded segment */
    *dictHandle = (Dgp_UInt32)decodedSegment;

ERR_CLEANUP:
    ERR_RETURN;
}


/* Build, or add to, a decoded dictionary list
 */
DGP_LIBEXPORT ErrCode
DgpDec_buildDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList,
    Dgp_UInt32 dictHandle)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    DgpAdt_List seg_list = (DgpAdt_List)*dictList;
    DGP_PROLOGUE( "DgpDec_buildDictList()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN(decodeHandle == 0UL, cErrInvalidParm);
    DGP_ERR_GEN(dictList == NULL, cErrInvalidParm);

    /* If there's no list, start one */
    if (*dictList == 0UL) {
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&seg_list, dgp_state));
	*dictList = (Dgp_UInt32)seg_list;

	/* Retain this list in the global list of lists (follow that?) */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(decHandle->dictLists_list, seg_list,
	    NULL));
    }

    /* Add the dictionary to the end of the list (must be maintained
     * in that order, for the decode) */
    if (dictHandle) {
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(seg_list, (DgpAdt_List)dictHandle, NULL));
    }

ERR_CLEANUP:
    ERR_RETURN;
}


/* Render an image from a decoded dictionary list and an encoded position
 * block.
 */
DGP_LIBEXPORT ErrCode
DgpDec_renderBinaryImage(DecodeHandle decodeHandle, Dgp_UInt32 dictList,
    Dgp_UInt8 *posBlock, Dgp_UInt32 posBlockBytes, Dgp_UInt8 *image,
    Dgp_UInt32 width, Dgp_UInt32 height, Dgp_UInt32 longsPerLine)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    Dgp_UInt32 nboxes, boxnum;
    jb2SymbolInstance *positions, *inst;
    jb2Symbol *sym_img;
    DgpAdt_List seg_list;
    Dgp_UInt32 dictSegs = 0UL;
    jb2Segment *curseg = (jb2Segment *)dictList;
    Dgp_Allocators *a = DGP_PALLOCATORS;
    DgpAdt_ListNode ln;
    DGP_PROLOGUE( "DgpDec_renderBinaryImage()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN(decodeHandle == 0UL, cErrInvalidParm);

    /* Generate a new list of segments, with the dictionary segments seeding
     * the list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&seg_list, dgp_state));
    DGPADT_FOREACH (ln, (DgpAdt_List)dictList) {
	curseg = (jb2Segment *)DGPADT_LIGET(ln);
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(seg_list, curseg, NULL));
	dictSegs++;
    }

    /* Decode the position block segments into a list of positions and IDs */
    DGP_ERR_PASS_CLEANUP(jb2_decodeTokPositions(&positions, 0, 0, 0, 0,
	posBlock, posBlockBytes, seg_list, &nboxes, dgp_state));

    /* Stamp each instance of a symbol onto the image */
    inst = positions;
    for (boxnum = 0; boxnum < nboxes; boxnum++) {
	sym_img = inst->sym;
	DgpBlat_blitBitsOp((Dgp_UInt32 *)image, longsPerLine,
	    0, 0, width, height, sym_img->bits, sym_img->longsperline,
	    sym_img->width, sym_img->height, inst->xpos, inst->ypos,
	    DGPBLAT_OP_OR, dgp_state);

	/* Go to the next instance */
	inst++;
    }

    /* Free the non-dictionary segments (and the data associated with them) */
    while ((ln = DGPADT_LIFIRST(seg_list)) != DGP_NULLLISTNODE) {
	/* Ignore the first segments, which really belong to the client's
	 * dictionary list */
	if (dictSegs) {
	    /* Remove it from this list, it really belongs to another */
	    DGP_ERR_PASS_CLEANUP(DgpAdt_liRem(ln));
	    dictSegs--;
	} else {
	    curseg = (jb2Segment *)DGPADT_LIGET(ln);
	    jb2FreeSegmentInfo(curseg, dgp_state);
	    DGP_ERR_PASS_CLEANUP(DgpAdt_liRem(ln));
	    (a->free)((void *)curseg);
	}
    }

    /* Free the segment list */
    DgpAdt_liFree(seg_list);

ERR_CLEANUP:
    ERR_RETURN;
}


/* Destroy a decoded dictionary list (but not the decoded dictionaries
 * themselves).
 * Note, DgpDec_close() will destroy all dictionary lists also.
 */
DGP_LIBEXPORT ErrCode
DgpDec_destroyDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    DgpAdt_List listOfLists = (DgpAdt_List)*dictList;
    DgpAdt_ListNode ln;
    DGP_PROLOGUE( "DgpDec_destroyDictList()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN_CLEANUP(decodeHandle == 0UL, cErrInvalidParm);

    /* If the list is empty, there's nothing to do */
    if (listOfLists == DGP_NULLLIST) return 0;

    /* Remove it from the list */
    if (listOfLists->nelt) {
	while ((ln = DGPADT_LIFIRST(listOfLists)) != DGP_NULLLISTNODE) {
	    DgpAdt_liRem(ln);
	}
    }

    /* Free this list */
    DgpAdt_liFree(listOfLists);

    /* Remove the pointer from the handle */
    *dictList = 0;

ERR_CLEANUP:
    ERR_RETURN;
}


/* Destroy the decoded dictionary.
 * Note, DgpDec_close() will destroy all decoded dictionaries also.
 */
DGP_LIBEXPORT ErrCode
DgpDec_destroyDict(DecodeHandle decodeHandle, Dgp_UInt32 *dictHandle)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    jb2Segment *decodedSegment = (jb2Segment *)*dictHandle;
    Dgp_Allocators *a = DGP_PALLOCATORS;
    DgpAdt_ListNode ln;
    jb2Segment *curseg = NULL;
    DGP_PROLOGUE( "DgpDec_destroyDict()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN_CLEANUP(decodeHandle == 0UL, cErrInvalidParm);

    /* If the segment is empty, there's nothing to do */
    if (decodedSegment == (jb2Segment *)NULL) return 0;

    /* Find the dict in the list of retained dicts */
    DGPADT_FOREACH (ln, decHandle->dicts_list) {
        curseg = (jb2Segment *)DGPADT_LIGET(ln);
	if (curseg == decodedSegment) {
	    break;
	}
    }

    /* Removed it from that list (so we don't try to remove it again later) */
    DGP_ERR_GEN(curseg != decodedSegment, cErrInvalidParm);
    DGP_ERR_PASS_CLEANUP(DgpAdt_liRem(ln));

    /* Free the segment, and the data associated with it */
    jb2FreeSegmentInfo(decodedSegment, dgp_state);
    (a->free)((void *)decodedSegment);

    /* Remove the pointer from the handle */
    *dictHandle = 0;

ERR_CLEANUP:
    ERR_RETURN;
}


/* De-initialize */
DGP_LIBEXPORT ErrCode
DgpDec_close(DecodeHandle *decodeHandle)
{
    _DecodeHandle *decHandle = (_DecodeHandle *)*decodeHandle;
    Dgp_State *dgp_state = decHandle->dgp_state;
    Dgp_Allocators *a = DGP_PALLOCATORS;
    DgpAdt_ListNode ln;
    jb2Segment *curseg;
    DGP_PROLOGUE( "DgpDec_close()", DGP_PERRSTATE );

    /* Check the parameter */
    DGP_ERR_GEN(decodeHandle == 0UL, cErrInvalidParm);

    /* Free the decoded dictionarys */
    if (decHandle->dicts_list) {
	if (decHandle->dicts_list->nelt) {

	    /* Free the data in each node of the list */
	    while ((ln = DGPADT_LIFIRST(decHandle->dicts_list)) != DGP_NULLLISTNODE) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		DGP_ERR_PASS_CLEANUP(DgpDec_destroyDict(*decodeHandle, (Dgp_UInt32 *)&curseg));
	    }

	    DGP_ERR_GEN_CLEANUP(decHandle->dicts_list->nelt, cErrInternal);
	}

	/* Free the list, and all the nodes in it, including the data in each node */
	DgpAdt_liFree(decHandle->dicts_list);
	decHandle->dicts_list = NULL;
    }

    /* Free the dictionary lists */
    if (decHandle->dictLists_list) {
	DGP_ERR_PASS_CLEANUP(DgpDec_destroyDictList(*decodeHandle,
	    (UInt32 *)&decHandle->dictLists_list));
    }

    /* Free the decode handle itself */
    (a->free)((void *)decHandle);

    /* Zero out the reference */
    *decodeHandle = 0;

    /* Close the API */
    DGP_ERR_PASS_CLEANUP(Dgp_cleanupDgpState(&dgp_state));

ERR_CLEANUP:
    ERR_RETURN;
}


/*******************************************************************
 * Running Log
 * $Log: decode_api.c,v $
 * Revision 1.4  2002/10/18 00:13:01  jlin
 * rm gcc warnings
 *
 * Revision 1.3  2002/03/29 00:24:44  dabercro
 * Add dict handle only if it exists.
 *
 * Revision 1.2  2002/03/08 20:08:33  dabercro
 * Move part of the cleanup until later, since it clobbers the allocation routines.
 *
 * Revision 1.1  2001/09/19 00:36:04  dabercro
 * Initial decoder api, taken from work done by Ming Xu.
 *
 *******************************************************************/
