/************************************************************
 * decode_api.h
 * $Header: /project/icons/cvsroot/digipaper/src/libsanta/decode_api.h,v 1.2 2001/10/05 18:33:31 dabercro Exp $
 ***********************************************************/

#ifndef __DECODE_API_H
#define __DECODE_API_H

/* This is the decoder API, written by Dave Abercrombie, adapted from work
 * done by Ming Xu of Cadisys.  It's written on top of a layer of the Digipaper
 * code such that we can pass our own decoded chunks of data in, and get image
 * data back out.  It wouldn't be hard to get back out a list of symbols, and
 * a pointer to a reference image (the symbol, which shouldn't be free'd by
 * the user, since it may be referenced again), instead of an image.
 *
 * The image, like so many other things related to the Digipaper code via
 * the coupling with the IPCore code, must be in IPCore's PIXR layout.  Namely,
 * it must be an image large enough to include padding on the right so that
 * scanlines are long word aligned, then ALSO have a 32-pixel frame around the
 * whole thing.
 *
 * Some things to know, if you're not really familiar with JBIG2:
 *
 * 1. JBIG2 compression has two parts: 1) compressed image data; and 2) a
 *	set of symbol dictionaries, any of which could be used by the image
 *	data
 * 2. The knowledge of which dictionaries are used by an individual image
 *	is usually known at the file format level.  In the case of TIFF-FX,
 *	there is a list of dictionaries that precede each block of compressed
 *	image data; and that list refers to the dictionaries stored in the
 *	SharedData items in the Shared Resources, which are maintained globally
 *	in the file via the GlobalParametersIFD.
 * 3. A dictionary may be reused from one image to the next.  This is actually
 *	one of the features that makes JBIG2 compress so much better than other
 *	binary algorithms: the symbols are stored once, then reused for every
 *	page that uses them.
 *
 * Here's the scoop on the way to call this API:
 *
 * 1. Call DgpDec_init(), send in some allocators (if you want), and get back
 *	a handle to use for all calls to this API.
 * 2. Call DgpDec_decodeDict() to decode 1 or more dictionaries (or don't call
 *	it if you have none).  Each of these dictionaries can be decoded only
 *	once, and reused when needed by an image stream.
 * 3. Add the dictionaries decoded in #2 to a list of dictionaries to be
 *	used to decode the image, using DgpDec_buildDictList().  If the
 *	list is empty, the list is started, otherwise the list is added to.
 *	The list is later emptied by calling DgpDec_destroyDictList().
 *	I used to think the dictionaries should be in exactly the same order
 *	as they were first generated by the JBIG2 encoder, but there's actually
 *	an ID internal to the dictionary that allows them to be used in the
 *	right order.
 * 4. Decode the position block (compressed image block) via 
 *	DgpDec_renderBinaryImage().  This routine is called with the dictionary
 *	list, the compressed image data, and an image with the IPCore PIXR
 *	structure.
 * 5. If the same exact dictionaries are to be used for the next position
 *	block, AND they're in the same exact order, then you can just reuse
 *	it (the dictionaries are already decoded).
 *	Otherwise, empty the dictionary list via DgpDec_destroyDictList().
 * 6. Of the dictionaries decoded, you can throw away the ones not used (or
 *	not) for the next position block, using DgpDec_destroyDict(); then
 *	load the new dictionaries as in #2.
 * 7. As before, make sure you have all the dictionaries used for the position
 *	block added to a dictionary list, as in #3.
 * 8. When you're completely done with decoding JBIG2 blocks, free everything
 *	up with a call to DgpDec_close().
 */

/* Includes:
 */
#include "dgp_err_control.h"
#include "dgp_allocators.h"


/* Defines:
 */
typedef Dgp_UInt32		DecodeHandle;


/* Prototypes:
 */

/* Initialize */
DGP_LIBEXPORT ErrCode
DgpDec_init(DecodeHandle *decodeHandle, Dgp_Allocators *a,
    Dgp_SmallAllocators *sm_a);


/* Decode a dictionary (to be retained for possible multiple renders)
 */
DGP_LIBEXPORT ErrCode
DgpDec_decodeDict(DecodeHandle decodeHandle, Dgp_UInt8 *dictData,
    Dgp_UInt32 dictBytes, Dgp_UInt32 *dictHandle);


/* Build, or add to, a decoded dictionary list
 */
DGP_LIBEXPORT ErrCode
DgpDec_buildDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList,
    Dgp_UInt32 dictHandle);


/* Render an image from a decoded dictionary list and an encoded position
 * block.
 */
DGP_LIBEXPORT ErrCode
DgpDec_renderBinaryImage(DecodeHandle decodeHandle, Dgp_UInt32 dictList,
    Dgp_UInt8 *posBlock, Dgp_UInt32 posBlockBytes, Dgp_UInt8 *image,
    Dgp_UInt32 width, Dgp_UInt32 height, Dgp_UInt32 longsPerLine);


/* Destroy a decoded dictionary list (but not the decoded dictionaries
 * themselves).
 * Note, DgpDec_close() will destroy all dictionary lists also.
 */
DGP_LIBEXPORT ErrCode
DgpDec_destroyDictList(DecodeHandle decodeHandle, Dgp_UInt32 *dictList);


/* Destroy the decoded dictionary.
 * Note, DgpDec_close() will destroy all decoded dictionaries also.
 */
DGP_LIBEXPORT ErrCode
DgpDec_destroyDict(DecodeHandle decodeHandle, Dgp_UInt32 *dictHandle);


/* De-initialize */
DGP_LIBEXPORT ErrCode
DgpDec_close(DecodeHandle *decodeHandle);


#endif /* __DECODE_API_H */


/*******************************************************************
 * Running Log
 * $Log: decode_api.h,v $
 * Revision 1.2  2001/10/05 18:33:31  dabercro
 * Add API description.
 *
 * Revision 1.1  2001/09/19 00:36:04  dabercro
 * Initial decoder api, taken from work done by Ming Xu.
 *
 *******************************************************************/
