#include "santa_private_t.h"
#include "santa.h"
#include "doc_helpers.h"
#include "memory_sentry.h"
#include "zephdocument.h"
#include "santa_macros.h"
#include "sucache.h"
/* doc_helpers.h 
   santaDoc helper functions.
   wbell 10/16/99
 */

/* docOpen()
   Given a previously allocated and cleared santaDoc, docOpen will
   fill it in and open the necessary document wide data structures.
   also initializes the document memory manager, which controls overall
   memory consumption for rendering processes on the document.
   
   pdoc - previously created document structure.
   mem_limit - the memory limit for rendering in this document, not counting
    santaRenderJobs and the memory associated with them (which comes from
    the caller)
   mem_used - how much memory has already been used (because presumably, this
    isn't the first thing you've allocated.)
   opts - the document rendering options.
   docuds - a uds structure that has an opened file pointer and
    will be shallowly copied
*/
ErrCode 
docOpen(santaDoc * pdoc, uds * docuds, Dgp_Int32 mem_limit, Dgp_Int32 mem_used, 
	santaDocOptions opts,
	Dgp_State * dgp_state)
{
  suOptions options;
  Dgp_Bool suMade = DGP_FALSE;
  Dgp_Bool zephDocMemAccounted = DGP_FALSE;
  DGP_STATE_PROLOGUE("docOpen()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!docuds, cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaDocOptions(opts), cErrInvalidParm);
  /* right now we just make sure we have enough memory for the */
  /* memory sentry, and then anything other than that is checked */
  /* by it.  */
  DGP_ERR_GEN(mem_limit <= mem_used, cErrInvalidParm);
  DGP_ERR_GEN(mem_used < 0, cErrInvalidParm);
  
  /* make sure the uds is ok. */
  DGP_ERR_GEN(!(docuds->fp), cErrInvalidParm);
  DGP_ERR_GEN(!(docuds->fseek), cErrInvalidParm);
  DGP_ERR_GEN(!(docuds->ftell), cErrInvalidParm);
  DGP_ERR_GEN(!(docuds->fread), cErrInvalidParm);
  
  /* params look good, lets make the memorySentry */
  DGP_ERR_PASS_CLEANUP(memorySentryCreate(pdoc, mem_limit, mem_used,
					  &(doc_private(pdoc)->memmgr),
					  dgp_state));


  doc_private(pdoc)->u = *docuds;
  doc_private(pdoc)->opts = opts;

  /* open the zeph document, and claim the memory it uses. We make */
  /* the assumption that the zeph doc doesn't allocate much memory. */
  DGP_ERR_PASS_CLEANUP(zephOpenDocument(dgp_state, &(doc_private(pdoc)->u),
			       &(doc_private(pdoc)->zephDoc)));
  
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
		   zephGetDocumentMemoryUsage(doc_private(pdoc)->zephDoc), 
					       dgp_state));
  zephDocMemAccounted = DGP_TRUE;

  DGP_ERR_PASS_CLEANUP(suSetDefaults(&options, dgp_state));
  options.psentry = doc_private(pdoc)->memmgr;
  /* FIX: might want to figure out how to readjust the supercache suggested */
  /* usage. */
  /* options.maxMemUsage = ??? */
  DGP_ERR_PASS_CLEANUP(suInit(&options,
			      doc_private(pdoc)->zephDoc,
			      &(doc_private(pdoc)->cache)));
  suMade = DGP_TRUE;
  /* see if we have enough memory. */
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
				      doc_private(pdoc)->zephDoc->pageCount *
						 sizeof(santaPage *),
					       dgp_state));

  DGP_CHECK_ALLOC_CLEANUP(doc_private(pdoc)->pppages = 
			  DGP_CALLOC(doc_private(pdoc)->zephDoc->pageCount,
				     sizeof(santaPage *)));
  
  pdoc->npages = doc_private(pdoc)->zephDoc->pageCount;
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (doc_private(pdoc)){
    if (doc_private(pdoc)->pppages){
      DGP_FREE(doc_private(pdoc)->pppages);
      doc_private(pdoc)->pppages = NULL;
      
      /* release the memory for this. */
      memorySentryReleaseMemory(doc_private(pdoc)->memmgr, 
				doc_private(pdoc)->zephDoc->pageCount
				* sizeof(santaPage *),
				dgp_state);
    }
    if (suMade){
      suDestroy(doc_private(pdoc)->cache);
    }
    if (doc_private(pdoc)->zephDoc){
      if (zephDocMemAccounted){
	memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
		   zephGetDocumentMemoryUsage(doc_private(pdoc)->zephDoc), 
				  dgp_state);
      }
      zephCloseDocument(&(doc_private(pdoc)->zephDoc));
    }
    
    if (doc_private(pdoc)->memmgr){
      memorySentryDestroy(&(doc_private(pdoc)->memmgr), dgp_state);
    }
  }
  CLEAR_STRUCT(*pdoc);
  ERR_RETURN;
}

/* docClose()
   Given a document opened with docOpen, this'll clean up all the state
   that's document wide. Yes, all pages should be closed before doing this.
   
   pdoc - the opened doc
*/
ErrCode
docClose(santaDoc * pdoc, Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("docClose()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);

  DGP_ERR_CHECK(suDestroy(doc_private(pdoc)->cache)){
    UPDATE_ERR(firstErr);
  }

  /* free the memory associated with the zephDoc. */
  DGP_ERR_CHECK(memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
		   zephGetDocumentMemoryUsage(doc_private(pdoc)->zephDoc), 
					  dgp_state)){
    UPDATE_ERR(firstErr);
  }
		
  /* free the zephdoc. */
  DGP_ERR_CHECK(zephCloseDocument(&(doc_private(pdoc)->zephDoc))){
    UPDATE_ERR(firstErr);
  }
  
  
  /* release this memory. */
  DGP_ERR_CHECK(memorySentryReleaseMemory(doc_private(pdoc)->memmgr, 
				pdoc->npages * sizeof(santaPage *),
					  dgp_state)){
    UPDATE_ERR(firstErr);
  }
  
  /* free the pages pointer. */
  DGP_FREE(doc_private(pdoc)->pppages);
  doc_private(pdoc)->pppages = NULL;

  /* destroy the memory sentry. */
  DGP_ERR_CHECK(memorySentryDestroy(&(doc_private(pdoc)->memmgr), 
				    dgp_state)){
    UPDATE_ERR(firstErr);
  }


  CLEAR_STRUCT(*pdoc);
  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}







