#include "memory_sentry.h"
#include "santa_layer.h"
#include "santa_private_t.h"
#include "santa_image.h"
#include "sucache.h"
/* memory_sentry.c
   wbell 10/12/99 
   the memory manager datatype manipulation functions.
   Note: the memory sentry mearly presents an interface for keeping track
   of memory that diff'rent parts of the decompresser are using at any point
   in time. It also provides a policy for purging memory from the santaLayers
   when memory is needed to continue.

   It *only* deals with the supercache in the document, and for each
   open page of the document, deals with the santaLayers for each layer
   in each open page. There are special hooks in santaLayer and supercache
   to facilitate memory querying 
*/

/* msGetAvailableMemory()
   gets the currently available memory limit from the sentry, any calls 
   to memorySentryClaimMemory for temporary storage will succeed. Any
   additional memory beyond this point will require flushing of temporary
   data from some part of the page / document.
   this is important to inform low level layers of how much buffer space
   they can use while decompressing into a specified output image, etc.

   psentry - a previously opened sentry.
   plimit - the current limit, in bytes
*/
static
ErrCode msGetAvailableMemory(memorySentry * psentry,
				       Dgp_Int32 * plimit,
				       Dgp_State * dgp_state)
{
  memorySentrySpecific * pspecific = (memorySentrySpecific *)psentry->specific;
  DGP_STATE_PROLOGUE("msGetAvailableMemory()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pspecific, cErrInvalidParm);
  DGP_ERR_GEN(!plimit, cErrInvalidParm);

  *plimit = pspecific->mem_limit - pspecific->mem_used;

  NORMAL_RETURN;
}

/* msClaimMemory()
   claim a certain amount of memory from the free pool. If this succeeds,
   this grants the user the ability to allocate nbytes of memory from the 
   heap. We do not take into account things such as fragmentation...
   If nbytes is greater than the current available memory, the memory
   sentry will attempt to purge cached information in order to fufill the
   request. If enough memory is available, the amount will just be
   removed from the pool.
    This presents some challenges in the low memory situations; say
    we need some memory to decompress an image to place on a temporary
    buffer in the layer that is to be a selector: we don't want the
    memory sentry to purge the layer's temporary buffer, or the 
    already allocated memory in the image. Likewise, we wouldn't want
    the temporary buffer in the selector purged when we went to decompress
    a foreground image. We placed a hook in the santaLayer type that allows
    one to 'pin' pieces of memory as being essential to the current operation
    and will not be removed when memory conditions get low.

    psentry - the sentry
    nbyes - number of bytes requested

    returns an insufficient memory area if the request could not be
    satisified when all non-essential buffer space was purged.
 */
static
ErrCode msClaimMemory(memorySentry * psentry,
				Dgp_Int32 nbytes,
				Dgp_State * dgp_state)
{
#ifndef SIMPLE_MEM_SENTRY
  Dgp_Int32 page, layer,img;
  santaDoc * pdoc;
  santaPage * ppage;
  santaLayer * player;
  santaImage * pimg;
#endif
  memorySentrySpecific * pspecific = (memorySentrySpecific *)psentry->specific;
  DGP_STATE_PROLOGUE("msClaimMemory()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pspecific, cErrInvalidParm);
  DGP_ERR_GEN(nbytes < 0, cErrInvalidParm);

#ifdef SIMPLE_MEM_SENTRY
  DGP_ERR_GEN((pspecific->mem_limit - (pspecific->mem_used + nbytes) < 0),
	      cErrInsufficientMem);
  pspecific->mem_used += nbytes;
#else
  if ((pspecific->mem_limit - (pspecific->mem_used + nbytes) < 0)){
    /* ok, we don't have enough. Our purging method is as follows:
       1) cleanup unused stuff in the suXiffCache
       2) remove the layersave from each layer.
       3) remove the pmask & prendered from each image
       4) remove tokenized image data for each tokenized image
       This'll only purge stuff that's not locked down (pinned) by people
       already.
    */

    /* 1) */
    pdoc = pspecific->pdoc;

    DGP_ERR_PASS(suClean(doc_private(pdoc)->cache));
    if ((pspecific->mem_limit - pspecific->mem_used) > nbytes){
      pspecific->mem_used += nbytes;
      NORMAL_RETURN;
    }
    /* 2) */
    for (page = 0 ; page < pdoc->npages ; page++){
      ppage = doc_private(pdoc)->pppages[page];
      if (ppage){
        for (layer = 0 ; layer < page_private(ppage)->nlayers ; layer++){
          player = &(page_private(ppage)->players[layer]);
          if (player){
            DGP_ERR_PASS(santaLayerPurgeSave(player, psentry, dgp_state));
            if ((pspecific->mem_limit - pspecific->mem_used) > nbytes){
              pspecific->mem_used += nbytes;
              NORMAL_RETURN;
            }
          }
        }
      }
    }
    /* 3) */
    for (page = 0 ; page < pdoc->npages ; page++){
      ppage = doc_private(pdoc)->pppages[page];
      if (ppage){
        for (layer = 0 ; layer < page_private(ppage)->nlayers ; layer++){
          player = &(page_private(ppage)->players[layer]);
          if (player){
            for (img = 0 ; img < player->nimages ; img++){
              pimg = &(player->pimgs[img]);
              if (pimg){
                DGP_ERR_PASS(santaImagePurgeSave(pimg, psentry, dgp_state));
                if ((pspecific->mem_limit - pspecific->mem_used) > nbytes){
                  pspecific->mem_used += nbytes;
                  NORMAL_RETURN;
                }
              }
            }
          }
        }
      }
    }
    /* 4) */
    for (page = 0 ; page < pdoc->npages ; page++){
      ppage = doc_private(pdoc)->pppages[page];
      if (ppage){
        for (layer = 0 ; layer < page_private(ppage)->nlayers ; layer++){
          player = &(page_private(ppage)->players[layer]);
          if (player){
            for (img = 0 ; img < player->nimages ; img++){
              pimg = &(player->pimgs[img]);
              if (pimg){
                DGP_ERR_PASS(santaImagePurgeTokenized(pimg, psentry, 
                                                      dgp_state));
                DGP_ERR_PASS(suClean(doc_private(pdoc)->cache));
                if ((pspecific->mem_limit - pspecific->mem_used) > nbytes){
                  pspecific->mem_used += nbytes;
                  NORMAL_RETURN;
                }
              }
            }
          }
        }
      }
    }
    /* if we're here, we've got issues. No memory. */
    DGP_ERR_GEN((pspecific->mem_limit - (pspecific->mem_used + nbytes) < 0),
                cErrInsufficientMem);
  }
  pspecific->mem_used += nbytes;
#endif

  /*    fprintf(stdout, "\t\t%d total, %d claimed\n", pspecific->mem_used, nbytes); */

  NORMAL_RETURN;		
}

/* msReleaseMemory()
   when you're done with a segment of memory that you got granted from
   memorySentryClaimMemory, call this to return it to the free pool for
   others to use. Always nice to play well with others. And hey, this
   isn't java, so it ain't getting cleaned up for you.
*/
static
ErrCode msReleaseMemory(memorySentry * psentry,
				  Dgp_Int32 nbytes,
				  Dgp_State * dgp_state)
{
  memorySentrySpecific * pspecific = (memorySentrySpecific *)psentry->specific;
  DGP_STATE_PROLOGUE("msReleaseMemory()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pspecific, cErrInvalidParm);
  DGP_ERR_GEN(nbytes < 0, cErrInvalidParm);

  pspecific->mem_used -= nbytes;
  /*    fprintf(stdout, "\t\t%d total, %d released\n", pspecific->mem_used, nbytes); */
  DGP_ASSERT(pspecific->mem_used >= 0);

  NORMAL_RETURN;		
}


/* memorySentryCreate()
   create a memorySentry, which keeps track of large memory allocations
   in santa. Any requests for temporary space for rendering or 
   permanent save space must go through the memory sentry for the job.
   It also has the hooks to purge cached memory when times get rough
   and we really need it.

   pdoc - the current doc.
   ppage - the page in question.
   mem_limit - the limit for large block allocation, in bytes.
   mem_used - how much memory has already been used (because presumably, this
    isn't the first thing you've allocated.)
*/
ErrCode memorySentryCreate(santaDoc * pdoc, Dgp_Int32 mem_limit, Dgp_Int32 mem_used, 
			   memorySentry ** ppsentry,
			   Dgp_State * dgp_state)
{
  memorySentry * psentry = NULL;
  memorySentrySpecific * pspecific = NULL;
  DGP_STATE_PROLOGUE("memorySentryCreate()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ppsentry, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN((mem_limit - mem_used) <= 
	      (Dgp_Int32)(sizeof(memorySentry) + sizeof(memorySentrySpecific)), 
	      cErrInsufficientMem);

  DGP_CHECK_ALLOC_CLEANUP(psentry = DGP_CALLOC(1, sizeof(memorySentry)));
  DGP_CHECK_ALLOC_CLEANUP(pspecific = 
			  DGP_CALLOC(1, sizeof(memorySentrySpecific)));


  pspecific->mem_used = mem_used + sizeof(memorySentry) +
                          sizeof(memorySentrySpecific);
  pspecific->mem_limit = mem_limit;
  pspecific->pdoc = pdoc;
  psentry->specific = (void *)pspecific;
  psentry->getAvailableMemory = msGetAvailableMemory;
  psentry->claimMemory = msClaimMemory;
  psentry->releaseMemory = msReleaseMemory;

  *ppsentry = psentry;
  NORMAL_RETURN;
ERR_CLEANUP:
  if (pspecific){
    DGP_FREE(pspecific);
  }
  if (psentry){
    DGP_FREE(psentry);
  }
  if (ppsentry){
    *ppsentry = NULL;
  }
  ERR_RETURN;
}



/* memorySentryDestroy()
   destroy a memory sentry.

   ppsentry - a sentry object, will set *ppsentry = NULL;
*/
ErrCode memorySentryDestroy(memorySentry ** ppsentry, 
			    Dgp_State * dgp_state)
{
    /*memorySentrySpecific * pspecific = (memorySentrySpecific *)(*ppsentry)->specific; */
  DGP_STATE_PROLOGUE("memorySentryDestroy()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ppsentry, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppsentry), cErrInvalidParm);


  DGP_FREE((*ppsentry)->specific);
  DGP_FREE(*ppsentry);
  *ppsentry = NULL;
  
  NORMAL_RETURN;
}


