#ifndef MEMORY_SENTRY_T_H
#define MEMORY_SENTRY_T_H

#include "santa_t.h"
#include "dgp_state.h"
#ifdef __cplusplus
extern "C"
{
#endif

/* memory_sentry_t.h
   wbell 10/12/99
   the memory manager datatype.
   the memory sentry keeps track of data in 3 structures which we consider
   to be the significant usages of memory in santa: the supercache, 
   the images which can store temporary data for faster decompression,
   and the layers which can store higher level temporary information such
   as portions of an assembled selectors, etc. The memory sentry has
   the hooks to selectively tell the 3 types to purge all their memory,
   this is how we keep the decompressing process confined to a particular
   memory limit at runtime.
*/

typedef struct memorySentry_
{
  /* every memory sentry is allowed to have specific data. */
  void * specific;

/* getAvailableMemory()
   gets the currently available memory limit from the sentry, any calls 
   to memorySentryClaimMemory for temporary storage will succeed. Any
   additional memory beyond this point will require flushing of temporary
   data from some part of the page / document.
   this is important to inform low level layers of how much buffer space
   they can use while decompressing into a specified output image, etc.

   psentry - a previously opened sentry.
   plimit - the current limit, in bytes
*/
  ErrCode (* getAvailableMemory)(struct memorySentry_ * psentry,
				 Dgp_Int32 * plimit,
				 Dgp_State * dgp_state);

  /* claimMemory()
   claim a certain amount of memory from the free pool. If this succeeds,
   this grants the user the ability to allocate nbytes of memory from the 
   heap. We do not take into account things such as fragmentation...
   If nbytes is greater than the current available memory, the memory
   sentry will attempt to purge cached information in order to fufill the
   request. If enough memory is available, the amount will just be
   removed from the pool.
    This presents some challenges in the low memory situations; say
    we need some memory to decompress an image to place on a temporary
    buffer in the layer that is to be a selector: we don't want the
    memory sentry to purge the layer's temporary buffer, or the 
    already allocated memory in the image. Likewise, we wouldn't want
    the temporary buffer in the selector purged when we went to decompress
    a foreground image. We placed a hook in the santaLayer type that allows
    one to 'pin' pieces of memory as being essential to the current operation
    and will not be removed when memory conditions get low.

    psentry - the sentry
    nbyes - number of bytes requested

    returns an insufficient memory area if the request could not be
    satisified when all non-essential buffer space was purged.
 */
  ErrCode (* claimMemory)(struct memorySentry_ * psentry,
			  Dgp_Int32 nbytes,
			  Dgp_State * dgp_state);
				
  /* releaseMemory()
   when you're done with a segment of memory that you got granted from
   memorySentryClaimMemory, call this to return it to the free pool for
   others to use. Always nice to play well with others. And hey, this
   isn't java, so it ain't getting cleaned up for you.
*/
  ErrCode (* releaseMemory)(struct memorySentry_ * psentry,
			    Dgp_Int32 nbytes,
			    Dgp_State * dgp_state);


} memorySentry;

/* memorySentryGetAvailableMemory()
   gets the currently available memory limit from the sentry, any calls 
   to memorySentryClaimMemory for temporary storage will succeed. Any
   additional memory beyond this point will require flushing of temporary
   data from some part of the page / document.
   this is important to inform low level layers of how much buffer space
   they can use while decompressing into a specified output image, etc.

   psentry - a previously opened sentry.
   plimit - the current limit, in bytes
*/
#define memorySentryGetAvailableMemory(psentry, plimit, dgp_state) \
 ((psentry)->getAvailableMemory((psentry), (plimit), (dgp_state)))

/* memorySentryClaimMemory()
   claim a certain amount of memory from the free pool. If this succeeds,
   this grants the user the ability to allocate nbytes of memory from the 
   heap. We do not take into account things such as fragmentation...
   If nbytes is greater than the current available memory, the memory
   sentry will attempt to purge cached information in order to fufill the
   request. If enough memory is available, the amount will just be
   removed from the pool.
    This presents some challenges in the low memory situations; say
    we need some memory to decompress an image to place on a temporary
    buffer in the layer that is to be a selector: we don't want the
    memory sentry to purge the layer's temporary buffer, or the 
    already allocated memory in the image. Likewise, we wouldn't want
    the temporary buffer in the selector purged when we went to decompress
    a foreground image. We placed a hook in the santaLayer type that allows
    one to 'pin' pieces of memory as being essential to the current operation
    and will not be removed when memory conditions get low.

    psentry - the sentry
    nbyes - number of bytes requested

    returns an insufficient memory area if the request could not be
    satisified when all non-essential buffer space was purged.
 */
#define memorySentryClaimMemory(psentry, nbytes, dgp_state) \
 ((psentry)->claimMemory((psentry), (nbytes), (dgp_state)))
				
/* memorySentryReleaseMemory()
   when you're done with a segment of memory that you got granted from
   memorySentryClaimMemory, call this to return it to the free pool for
   others to use. Always nice to play well with others. And hey, this
   isn't java, so it ain't getting cleaned up for you.
*/

#define memorySentryReleaseMemory(psentry, nbytes, dgp_state) \
 ((psentry)->releaseMemory((psentry), (nbytes), (dgp_state)))


/* if you don't want to clutter your code on the if switch, just call this 
   with a positive or negative number and the right thing will happen.
*/
#define memorySentryUpdateMemory(psentry, nbytes, dgp_state) \
(((nbytes) > 0) ? \
 memorySentryClaimMemory((psentry), (nbytes), (dgp_state)) : \
 memorySentryReleaseMemory((psentry), -(nbytes), (dgp_state)))


#ifdef __cplusplus
}
#endif
#endif
