#include "santa_private_t.h"
#include "page_helpers.h"
#include "santa_layer.h"
#include "memory_sentry.h"
#include "zephpage.h"
#include "santa_macros.h"
#include "sucache.h"

/* page_helpers.c
   santaPage helper functions.
   wbell 10/16/99
 */

/* pageOpen()
   takes a previously allocated santaPage structure and fills it in with the
   initialization information for page page_num (note pages start numbering
   at 0)

   pdoc - the document already opened and initialized
   ppage - the page structure that you want initialized.
   page_num - what page do you want it initialized to? starts from 0.
*/
ErrCode
pageOpen(santaDoc * pdoc, santaPage * ppage, Dgp_Int32 page_num,
	 Dgp_State * dgp_state)
{
  Dgp_Int32 i = 0;
  Dgp_Bool evenLayer = DGP_FALSE;
  Dgp_Bool zephPageMemAccounted = DGP_FALSE;
  DGP_STATE_PROLOGUE("pageOpen()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(page_num < 0 || page_num >= pdoc->npages, cErrInvalidParm);

  /* no references to this page; the santa layer handles that. */
  page_private(ppage)->nrefs = 0;

  /* we assume that the zephPage open won't allocate too much memory,  */
  /* and since it doesn't allow us to know how much memory we'll need */
  /* before opening it, we just open the puppy and try and account for */
  /* it later. */
  DGP_ERR_PASS_CLEANUP(zephOpenPage(doc_private(pdoc)->zephDoc, (Dgp_UInt16)page_num,
				    &(page_private(ppage)->zephPage)));

  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
		   zephGetPageMemoryUsage(page_private(ppage)->zephPage), 
					       dgp_state));
  zephPageMemAccounted = DGP_TRUE;

  page_private(ppage)->nlayers = 
    page_private(ppage)->zephPage->layerCount;

  /* we need an array of layers. clear it with the memory sentry. */
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
					       page_private(ppage)->nlayers *
					       sizeof(santaLayer), 
					       dgp_state));

  DGP_CHECK_ALLOC_CLEANUP(page_private(ppage)->players = 
			  DGP_CALLOC(page_private(ppage)->nlayers,
				     sizeof(santaLayer)));

  ppage->page_num = page_num;
  ppage->width = page_private(ppage)->zephPage->width;
  ppage->height = page_private(ppage)->zephPage->height;
  ppage->res = page_private(ppage)->zephPage->res;

  /* create the layer objects, and check for the page type at the same time. */
  /* If the document contains only even layers (odd in our frame because */
  /* we're computer scientists and count from 0), then it's a monochrome */
  /* page. Else, it's a color one. */
  for (i = 0; i < page_private(ppage)->nlayers; i ++){
    DGP_ERR_PASS_CLEANUP(santaLayerInit(pdoc, ppage, i, 
					&(page_private(ppage)->players[i]),
					doc_private(pdoc)->memmgr, dgp_state));
    
    /* if it has images in an even layer, then we'll assume it's a color */
    /* page. This really isn't quite right, because we could have a page */
    /* that has even layers that have binary images. But that's a pretty  */
    /* messed up page. No sense in optimizing that case. */
    if ((page_private(ppage)->zephPage->imageCounts[i] > 0) &&
	((i % 2) == 0)){
      evenLayer = DGP_TRUE;
    }
  }
  if (!evenLayer){
    ppage->page_type = santaPageMono;
  }
  else {
    ppage->page_type = santaPageColor;
  }
  
  NORMAL_RETURN;
ERR_CLEANUP:
  /* did it die somewhere along the layer creation? */
  if (i != 0){
    Dgp_Int32 j = 0;
    /* we'll clean up from 0 to i-1. */
    for (j = 0; j < i; j++){
      santaLayerDestroy(pdoc, ppage, &(page_private(ppage)->players[i]),
			doc_private(pdoc)->memmgr, dgp_state);
    }
  }

  page_private(ppage)->nlayers = 0;
  if (page_private(ppage)->players){
    DGP_FREE(page_private(ppage)->players);
    page_private(ppage)->players = NULL;

    /* return this memory to the memorySentry. */
    memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
			    page_private(ppage)->nlayers *
			    sizeof(santaLayer), 
			    dgp_state);
  }


  if (page_private(ppage)->zephPage){
    if (zephPageMemAccounted){
	memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
		   zephGetPageMemoryUsage(page_private(ppage)->zephPage), 
				  dgp_state);
    }
    zephClosePage(&(page_private(ppage)->zephPage));
  }
  CLEAR_STRUCT(*ppage);
  ERR_RETURN;
}




/* pageClose()
   destroys a page structure initialized with pageOpen, closes the relevent
   opened layers and images. make sure all jobs are done with this page 
   before you close it. looking at the ref count is a good idea
   to determine that.

   pdoc - the document for this page.
   ppage - the page structure that has been initialized.
*/
ErrCode
pageClose(santaDoc * pdoc, santaPage * ppage, 
	  Dgp_State * dgp_state)
{
  Dgp_Int32 i = 0;
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("pageClose()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ASSERT(page_private(ppage)->nrefs == 0);

  /* free the zephpage and release the zephpage memory. */
  DGP_ERR_CHECK(memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
		   zephGetPageMemoryUsage(page_private(ppage)->zephPage), 
					  dgp_state)){
    UPDATE_ERR(firstErr);
  }

  DGP_ERR_CHECK(zephClosePage(&page_private(ppage)->zephPage)){
    UPDATE_ERR(firstErr);
  }

  for (i = 0; i < page_private(ppage)->nlayers; i++){
    DGP_ERR_CHECK(santaLayerDestroy(pdoc, ppage,
				    &(page_private(ppage)->players[i]),
				    doc_private(pdoc)->memmgr, dgp_state)){
    UPDATE_ERR(firstErr);
    }
  }

  DGP_FREE(page_private(ppage)->players);
  page_private(ppage)->players = NULL;

  DGP_ERR_CHECK(memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
			    page_private(ppage)->nlayers * 
					  sizeof(santaLayer), dgp_state)){
    UPDATE_ERR(firstErr);
  }

  DGP_ERR_CHECK(suTock(doc_private(pdoc)->cache)){
    UPDATE_ERR(firstErr);
  }

  page_private(ppage)->nlayers = 0;
  CLEAR_STRUCT(*ppage);
  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN; 
}

