#include "page_strip_order.h"
#include "page_strip.h"
#include "rendered_image.h"
#include "orientation.h"
#include "santa_macros.h"
#include <math.h>
/* File page_strip_order.c
   wbell 6/8/98
   primitive for deciding how PageStrips will be generated for a particular 
   job, either by progressing down the page, or progressing through the color
   space (single/multiple Separations of CMYK) and then down the page, or
   the reverse. Note, going down the page means nothing about the orientation
   (it just means that progressive strips represent a different physical area
    of the page.)

   Revised for the millenium by wbell on 10/3/99
 */

static void
pageStripOrderGetBandCoords(pageStripOrder * porder, Dgp_Int32 num,
			     Dgp_Int32 * px, Dgp_Int32 * py, Dgp_Int32 * pw, Dgp_Int32 * ph,
			     renderedImageType * ptype,
			     Dgp_State * dgp_state);

ErrCode
pageStripOrderCreate(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
		     DgpCoord_Orientation o, DgpCoord_Resolution * res,
		     stripAdvanceOrderType advOrder,
		     orderInputType inputType,
		     pageAreaAdvanceOrderType pageAdvOrder,
		     Dgp_Int32 slowWidth,
		     pageStripOrder ** pporder,
		     Dgp_State * dgp_state)
{
  pageStripOrder * porder = NULL;
  DGP_STATE_PROLOGUE("pageStripOrderCreate()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pporder, cErrInvalidParm);
  DGP_ERR_GEN(!res, cErrInvalidParm);
  DGP_ERR_GEN(w <= 0, cErrInvalidParm);
  DGP_ERR_GEN(h <= 0, cErrInvalidParm);
  DGP_ERR_GEN(slowWidth <= 0, cErrInvalidParm);
  DGP_ERR_GEN(!ISVALIDORIENT(o), cErrInvalidParm);
  DGP_ERR_GEN(!isValidStripAdvanceOrder(advOrder), cErrInvalidParm); 
  DGP_ERR_GEN(!isValidOrderInputType(inputType), cErrInvalidParm);
  DGP_ERR_GEN(!isValidPageAreaAdvanceOrder(pageAdvOrder),
		      cErrInvalidParm);

  /* create the struct. */
  DGP_CHECK_ALLOC_CLEANUP(porder = (pageStripOrder *)
			  DGP_CALLOC(1,sizeof(pageStripOrder)));
  
  /* how many total strips are there? Well, there are  */
  /* the size of the slow direction / slowWidth. */
  /* if the DgpCoord_Orientation is rotated, then it's the w, else it's the height */
  if (DgpCoord_isRotated(ORIENT_RD, o)){
    porder->numStrips = (Dgp_Int32)ceil((double)w / slowWidth);
    porder->fastWidth = h;
    porder->slowWidth = MIN(slowWidth,w);
  }
  else{
    porder->numStrips = (Dgp_Int32)ceil((double)h / slowWidth);
    porder->fastWidth = w;
    porder->slowWidth = MIN(slowWidth,h);
  }

  porder->advOrder = advOrder;
  porder->inputType = inputType;
  porder->pageAdvOrder = pageAdvOrder;
  porder->o = o;
  porder->res = *res;
  porder->x = x;
  porder->y = y;
  porder->w = w;
  porder->h = h;

  *pporder = porder;
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (porder){
    DGP_FREE(porder);
  }
  if (pporder){
    *pporder = NULL;
  }
  ERR_RETURN;
}

ErrCode
pageStripOrderDestroy(pageStripOrder ** pporder,
		      Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("pageStripOrderDestroy()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pporder, cErrInvalidParm);

  if (*pporder){
    DGP_FREE(*pporder);
  }
  *pporder = NULL;
    
  NORMAL_RETURN;
}


ErrCode 
pageStripOrderGetNextBand(pageStripOrder * porder,
		   void * data,
		   Dgp_Int32 bpl,
		   Dgp_Bool cleared,
		   pageStrip ** ppstrip,
           DgpCoord_Resolution *dst_res,
		   Dgp_State * dgp_state)
{
  renderedImage * pImg = NULL;
  pageStrip * pStrip = NULL;
  Dgp_Int32 x,y,w,h;
  renderedImageType type;
  DGP_STATE_PROLOGUE("pageStripOrderGetNextBand()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  /* data/bpl can be NULL/0 -- this is the signal that renderedImageCreate() */
  /* should allocate storage. */
  DGP_ERR_GEN(!ppstrip, cErrInvalidParm);
  
  /* are we done? */
  if (porder->curStrip < porder->numStrips){
    /* get the coords for this band. */
    pageStripOrderGetBandCoords(porder, porder->curStrip, &x, &y, &w, &h, 
				 &type, dgp_state);
    DGP_ERR_PASS_CLEANUP(renderedImageCreate(x, y, w, h, &porder->res,
                         dst_res,
					     porder->o, bpl, type,
					     data, &pImg, dgp_state));

    /* 20.7.2000 / dhtesq. */
    /* Trap for sub-nominal sized bands at the end of a decode loop. */
    /* First saw this when debugging Dixie -- last band rendered will typically */
    /* be less than the nominal size, but porder does not reflect this, */
    /* confusing the jpeg placement downstream. */
    switch(porder->o){
        case ORIENT_RD:
            if(pImg->y == porder->y){
                porder->slowWidth = pImg->h;
            }
            break;
        case ORIENT_DL:
            if(pImg->x == porder->x){
                porder->slowWidth = pImg->w;
            }
            break;
        case ORIENT_UR:
            if(pImg->x + pImg->w == porder->x + porder->w){
                porder->slowWidth = pImg->w;
            }
            break;
        case ORIENT_LU:
            if(pImg->y == porder->y){
                porder->slowWidth = pImg->h;
            }
            break;
        default:
            break;
    }

    DGP_ERR_PASS_CLEANUP(pageStripCreate(pImg, cleared, &pStrip, dgp_state));
    *ppstrip = pStrip;
  }
  else{
    *ppstrip = NULL;
  }
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (ppstrip){
    *ppstrip = NULL;
  }
  if (pStrip){
    pageStripDestroy(&pStrip, dgp_state);
  }
  if (pImg){
    renderedImageDestroy(&pImg, dgp_state);
  }
  ERR_RETURN;
}

ErrCode
pageStripOrderReturnBand(pageStripOrder * porder,
			 pageStrip ** ppstrip,
			 Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  renderedImage * pimg;
  DGP_STATE_PROLOGUE("pageStripOrderReturnBand()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!ppstrip, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppstrip), cErrInvalidParm);

  pimg = (*ppstrip)->pimg;
  DGP_ERR_CHECK(pageStripDestroy(ppstrip, dgp_state)){
    UPDATE_ERR(firstErr);
  }
  DGP_ERR_CHECK(renderedImageDestroy(&pimg, dgp_state)){
    UPDATE_ERR(firstErr);
  }
  porder->curStrip++;

  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;   
}


ErrCode
pageStripOrderNextOn(pageStripOrder * porder,
		     Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
		     Dgp_Int32 * pnext,
		     Dgp_State * dgp_state)
{
  Dgp_Int32 bx, by, bw, bh, i;
  DGP_STATE_PROLOGUE("pageStripOrderNextOn()", dgp_state);
  DGP_ERR_GEN_CLEANUP(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!porder, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!pnext, cErrInvalidParm);

  *pnext = -1;
  for (i = porder->curStrip + 1; i < porder->numStrips; i++){
    pageStripOrderGetBandCoords(porder, i, &bx, &by, &bw, &bh, NULL, 
				 dgp_state);
    if ((x < (bx + bw)) && (bx < (x + w)) &&
	(y < (by + bh)) && (by < (y + h))){
      break;
    }
  }
  
  if (i < porder->numStrips){
    *pnext = i - porder->curStrip + 1;
  }

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pnext){
    *pnext = -1;
  }
  ERR_RETURN;
}

/* Internal function that gives you the page coordinates of a  */
/* specific strip in a PageOrder. */
void pageStripOrderGetBandCoords(pageStripOrder * porder, Dgp_Int32 num,
				  Dgp_Int32 * px, Dgp_Int32 * py, Dgp_Int32 * pw, Dgp_Int32 * ph,
				  renderedImageType * ptype,
				  Dgp_State * dgp_state)
{
  Dgp_Int32 x = 0, y = 0, w = 0, h = 0;
  renderedImageType type = 0;
  Dgp_Int32 advY = 0,advX = 0;

  DGP_VOID_PROLOGUE("pageStripOrderGetBandCoords()", DGP_PERRSTATE);
  DGP_VOID_ASSERT(dgp_state);
  DGP_VOID_ASSERT(porder);  

  /* coordinates are always normalized (upper-left to lower-right) */
  /* figure out the x,y,w,h of band 0, and what distance the bands progress */
  /* across the page. Note, we know that slowWidth - the span of the bands */
  /* in the slow direction is > 0, so x,y will never be less than 0. */
  switch(porder->o)
    {
    case ORIENT_RD:
    case ORIENT_LD:
      advY = 1 * porder->slowWidth;
      advX = 0;
      w = porder->fastWidth; 
      h = porder->slowWidth;
      x = porder->x;
      y = porder->y;
      break;

    case ORIENT_UR:
    case ORIENT_DR:
      advY = 0;
      advX = 1 * porder->slowWidth;
      w = porder->slowWidth;
      h = porder->fastWidth;
      x = porder->x;
      y = porder->y;
      break;

    case ORIENT_RU:
    case ORIENT_LU:
      advY = -1 * porder->slowWidth;
      advX = 0;
      w = porder->fastWidth;
      h = porder->slowWidth;
      x = porder->x;
      y = porder->y + (porder->h - 1) - (porder->slowWidth - 1);
      break;

    case ORIENT_DL:
    case ORIENT_UL:
      advY = 0;
      advX = -1 * porder->slowWidth;
      w = porder->slowWidth;
      h = porder->fastWidth;
      x = porder->x + (porder->w - 1) - (porder->slowWidth - 1);
      y = porder->y;
      break;
    }

  /* so we know where the origin of the first band is, so now let's figure */
  /* out where the num'th band is. Now we've got two cases, depending on if */
  /* we're doing something that has multiple bands in color space  */
  /* (cmyk Separations) */
  /* or we're doing something that just progresses across the page. */
  /* are we doing something that can actually progress across the color space? */
  if (porder->advOrder == saoColorSpaceProgression && 
      porder->inputType == oiSingleSeparationsCMYK){
    switch(num % 4){
    case 0:
      type = riCSeparation;
      break;
    case 1:
      type = riMSeparation;
      break;
    case 2:
      type = riYSeparation;
      break;
    case 3:
      type = riKSeparation;
      break;
    }
    
    x = x + (advX * (num % 4));
    y = y + (advY * (num % 4));
  }
  /* else, we move across the page first, and then progress across the color */
  /* space. */
  else{
    x = x + (advX * (num % porder->numStrips));
    y = y + (advY * (num % porder->numStrips));
    
    /* are we something that has a progression through the color space? */
    switch(porder->inputType){
    case oiSingleSeparationsCMYK:
      switch(num / porder->numStrips){
      case 0:
	type = riCSeparation;
	break;
      case 1:
	type = riMSeparation;
	break;
      case 2:
	type = riYSeparation;
	break;
      case 3:
	type = riKSeparation;
	break;
      }
    case oiAllSeparationsCMYK:
      type = riCMYKSeparations;
      break;
    case oiMonoPacked:
      type = riMonoPacked;
      break;
    case oiRGBInterleaved:
      type = riRGBInterleaved;
      break;
    }
  }     
  
  if (px){
    if (x < porder->x){
      w = w - (porder->x - x);
    }
    *px = MAX(x, porder->x);
  }
  if (py){
    if (y < porder->y){
      h = h - (porder->y - y);
    }
    *py = MAX(y, porder->y);
  }
  if (pw){
    *pw = MIN(w, (porder->w + porder->x) - x);
  }
  if (ph){
    *ph = MIN(h, (porder->h + porder->y) - y);
  }
  if (ptype){
    *ptype = type;
  }
}




