#ifndef __PAGE_STRIP_ORDER_H__
#define __PAGE_STRIP_ORDER_H__

#include "page_strip_order_t.h"
#include "page_strip_t.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* File page_strip_order.h
   wbell 6/8/98
   primitive for deciding how PageStrips will be generated for a particular 
   job, either by progressing down the page, or progressing through the color
   space (single/multiple seperations of CMYK) and then down the page, or
   the reverse. Note, going down the page means nothing about the DgpCoord_Orientation
   (it just means that progressive strips represent a different physical area
    of the page.)

   Revised for the millenium by wbell on 10/3/99
 */

/* check if a value is ok for these types. */
#define isValidStripAdvanceOrder(a) \
((a) == saoPageAreaProgression || (a) == saoColorSpaceProgression)

#define isValidOrderInputType(c) \
(((c) == oiMonoPacked) || ((c) == oiRGBInterleaved) || \
 ((c) == oiSingleSeparationsCMYK) || ((c) == oiAllSeparationsCMYK)) \

#define isValidPageAreaAdvanceOrder(aa) \
((aa) == paoAdvanceSlow)


/* Create a pageStripOrder:
   x,y,w,h - the entire area that you wish to have rendered in page
     coordinates, native DgpCoord_Orientation.
   o - DgpCoord_Orientation you wish to render in.
   res - resolution of the page.
   advOrder - how do you want strips to advance? in the color space
     then the page area, or vice-versa. Only applies if it's possible
     to advance through the color space (cmyk)
   colorAdvOrder - how do you advance through the color space?
    cmyk? ycmk? Only applies if it's possible to advance through
    the color space (cmyk)
   pageAdvOrder - how do you advance down the page? advancing in the
    slow direction?
   slowWidth - what band size do you wish to use?
   ppOrder - the PageStripOrder will be created and returned here.
*/   
ErrCode
pageStripOrderCreate(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
		     DgpCoord_Orientation o, DgpCoord_Resolution * res,
		     stripAdvanceOrderType advOrder,
		     orderInputType inputType,
		     pageAreaAdvanceOrderType pageAdvOrder,
		     Dgp_Int32 slowWidth,
		     pageStripOrder ** pporder,
		     Dgp_State * dgp_state);


/* Destroy a pageStripOrder:
   ppOrder - the order to destroy. Will set the pointer to NULL when done.
   
   Note: return all bands before you destroy the call this.
*/
ErrCode
pageStripOrderDestroy(pageStripOrder ** pporder,
		      Dgp_State * dgp_state);

/* Get the first/next band, following the progression you specified in 
   pageStripOrderCreate. You can have multiple bands out at once.
   
   pOrder - a page order created with pageStripOrderCreate
   data -  image data, you have already allocated it. For CMYK seperations,
     if you're just doing one, just pass it. If you're doing all 4 at once,
     pass an array of 4 Dgp_UInt8 pointers, which correspond to c,m,y,k.
     packed / rgb, just pass the pointer. 
   bpl - number of bytes per line in the source buffer. Note that all cmyk
     seperations must have the same
     bpl.
   cleared - is this memory cleared (an optimization)? 
   ppStrip - the page strip will be returned here. If this function succeeds
    and ppStrip is NULL, the pageStripOrder is finished.
*/

ErrCode 
pageStripOrderGetNextBand(pageStripOrder * porder,
			  void * data,
			  Dgp_Int32 bpl,
			  Dgp_Bool cleared,
			  pageStrip ** ppstrip,
              DgpCoord_Resolution *dst_res,
			  Dgp_State * dgp_state);
			  
  
/* Return this band. Does not free the memory that you passed in. It's
   your responsibility.
*/
ErrCode
pageStripOrderReturnBand(pageStripOrder * porder,
			 pageStrip ** ppstrip,
			 Dgp_State * dgp_state);

/* Here's a neat one; when will this rectangle (x,y,x+w,y+h) be on a pageStrip?
   (not including the last band retrieved from pageStripOrderGetNextBand)
   x,y,w,h - a rectange on page coordinates that you wish to check.
   pNextStrip - a pointer to an integer that will be set to how many pageStrip
    before any part of this rectangle will appear on a pageStrip. 0 means the
    next pageStrip, etc. a negative value means that this rectangle will
    never appear on a future pageStrip in this PageOrder.
*/
ErrCode
pageStripOrderNextOn(pageStripOrder * porder,
		     Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
		     Dgp_Int32 * pnext,
		     Dgp_State * dgp_state);

  



#ifdef __cplusplus
}
#endif
#endif 

