#include "rendered_image.h"
#include "orientation.h"
#include <math.h>

/* File rendered_image.c
   wbell 10/5/99
   access functions for rendered images.
*/
  
ErrCode isRenderedImageValid(renderedImage * ri, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("isRenderedImageValid()", dgp_state);
  DGP_ERR_GEN_CLEANUP(!dgp_state, cErrInvalidParm);

  DGP_ERR_GEN_CLEANUP(!ri, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(ri->w <= 0, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(ri->h <= 0, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(DgpCoord_le(ri->dst_res.yRes, DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(DgpCoord_le(ri->dst_res.xRes, DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(ri->bpl <= 0, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!isValidRenderedImageType(ri->type), cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!(ri->type == riMonoPacked
			&& ri->buffer.packed != NULL) &&
		      !(ri->type == riRGBInterleaved
			&& ri->buffer.rgb != NULL)
		      && !(((ri->type & riCSeparation)
			    && ri->buffer.cmyk[0] != NULL) &&
			   ((ri->type & riMSeparation)
			    && ri->buffer.cmyk[1] != NULL) &&
			   ((ri->type & riYSeparation)
			    && ri->buffer.cmyk[2] != NULL) &&
			   ((ri->type & riKSeparation)
			    && ri->buffer.cmyk[3] != NULL)),
		      cErrInvalidParm);
  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}

ErrCode renderedImageCreate(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    DgpCoord_Resolution * src_res,
			    DgpCoord_Resolution * dst_res,
			    DgpCoord_Orientation orient,
			    Dgp_Int32 bpl,
			    renderedImageType type,
			    void * src,
			    renderedImage ** ppri,
			    Dgp_State * dgp_state)
{
  renderedImage * pri = NULL;
  Dgp_Int32 allocrows = 0, alloccols = 0;
  DGP_STATE_PROLOGUE("renderedImageCreate()", dgp_state);
  DGP_ERR_GEN_CLEANUP(!dgp_state, cErrInvalidParm);
  
  DGP_ERR_GEN_CLEANUP(w <= 0, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(h <= 0, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!dst_res, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(DgpCoord_le(dst_res->xRes,DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(DgpCoord_le(dst_res->yRes,DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(bpl == 0 && src, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!ppri, cErrInvalidParm);
  
  DGP_CHECK_ALLOC_CLEANUP(pri = (renderedImage *)
			  DGP_CALLOC(sizeof(renderedImage),1));

  pri->x = x;
  pri->y = y;
  pri->w = w;
  pri->h = h;
  pri->src_res = *src_res;
  pri->dst_res = *dst_res;
  pri->orient = orient;
  pri->bpl = bpl;
  
  pri->type = type;

  pri->trans_src = NULL;
  pri->trans_bpl = 0;

  allocrows = (DgpCoord_isRotated(ORIENT_RD, orient) ? w : h);
  alloccols = (DgpCoord_isRotated(ORIENT_RD, orient) ? h : w);

  if (src){
    if (type == riMonoPacked){
      pri->buffer.packed = (Dgp_UInt32 *)src;
    }
    else if (type == riRGBInterleaved){
      pri->buffer.rgb = (DgpImage_RGB *)src;
    }
    else if (type & riCMYKSeparations){
      DGP_ERR_GEN_CLEANUP((type & riCSeparation)
			  && !((Dgp_UInt8 **)src)[0],
			  cErrInvalidParm);
      pri->buffer.cmyk[0] = ((Dgp_UInt8 **)src)[0];
      
      DGP_ERR_GEN_CLEANUP((type & riMSeparation)
			  && !((Dgp_UInt8 **)src)[1],
			  cErrInvalidParm);
      pri->buffer.cmyk[1] = ((Dgp_UInt8 **)src)[1];

      DGP_ERR_GEN_CLEANUP((type & riYSeparation)
			  && !((Dgp_UInt8 **)src)[2],
			  cErrInvalidParm);
      pri->buffer.cmyk[2] = ((Dgp_UInt8 **)src)[2];

      DGP_ERR_GEN_CLEANUP((type & riKSeparation)
			  && !((Dgp_UInt8 **)src)[3],
			  cErrInvalidParm);
      pri->buffer.cmyk[3] = ((Dgp_UInt8 **)src)[3];
    }
    else {
      DGP_ERR_GEN_CLEANUP(DGP_TRUE,cErrInvalidParm);
    }
    pri->freeMem = DGP_FALSE;
  }
  else{
    if (type == riMonoPacked){
      if (!bpl){
        bpl = (Dgp_Int32)ceil((double)alloccols / 32) * sizeof(Dgp_UInt32);
      }
      DGP_CHECK_ALLOC_CLEANUP(pri->buffer.packed = 
			      (Dgp_UInt32 *)DGP_CALLOC(1,bpl * allocrows));
    }
    else if (type == riRGBInterleaved){
      if (!bpl){
        bpl = alloccols * sizeof(DgpImage_RGB);
      }
      DGP_CHECK_ALLOC_CLEANUP(pri->buffer.rgb = 
			      (DgpImage_RGB *)DGP_MALLOC(bpl * allocrows));
    }
    else if (type & riCMYKSeparations){     
      if (!bpl){
        bpl = alloccols * sizeof(Dgp_UInt8);
      }
      if (type & riCSeparation){
	DGP_CHECK_ALLOC_CLEANUP(pri->buffer.cmyk[riCSepIdx] =
				(Dgp_UInt8 *)DGP_MALLOC(bpl * allocrows));
      }
      if (type & riMSeparation){
	DGP_CHECK_ALLOC_CLEANUP(pri->buffer.cmyk[riMSepIdx] =
				(Dgp_UInt8 *)DGP_MALLOC(bpl * allocrows));
      }
      if (type & riYSeparation){
	DGP_CHECK_ALLOC_CLEANUP(pri->buffer.cmyk[riYSepIdx] =
				(Dgp_UInt8 *)DGP_MALLOC(bpl * allocrows));
      }
      if (type & riKSeparation){
	DGP_CHECK_ALLOC_CLEANUP(pri->buffer.cmyk[riKSepIdx] =
				(Dgp_UInt8 *)DGP_MALLOC(bpl * allocrows));
      }
    }
    else {
      DGP_ERR_GEN_CLEANUP(DGP_TRUE,cErrInvalidParm);
    }
    pri->bpl = bpl;
    pri->freeMem = DGP_TRUE;
  }

  DGP_ASSERT(cSuccessful == isRenderedImageValid(pri,dgp_state));

  *ppri = pri;

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pri){
    /* note that we're taking advantage of the union here. */
    /* we know that buffer.packed and buffer.DgpImage_RGB all fall in the same */
    /* slot as buffer.cmyk[0] */
    if (pri->freeMem){
      if (pri->buffer.cmyk[0]){
	DGP_FREE(pri->buffer.cmyk[0]);
      }
      if (pri->buffer.cmyk[0]){
	DGP_FREE(pri->buffer.cmyk[1]);
      }
      if (pri->buffer.cmyk[0]){
	DGP_FREE(pri->buffer.cmyk[2]);
      }
      if (pri->buffer.cmyk[0]){
	DGP_FREE(pri->buffer.cmyk[3]);
      }
    }
    DGP_FREE(pri);
  }
  ERR_RETURN;
}

ErrCode renderedImageDestroy(renderedImage **ppri, Dgp_State * dgp_state)
{
  renderedImage * pri = NULL;
  DGP_STATE_PROLOGUE("renderedImageDestroy()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ppri, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppri), cErrInvalidParm);
  
  pri = * ppri;
  DGP_ASSERT(cSuccessful == isRenderedImageValid(pri,dgp_state));

  /* note that we're taking advantage of the union here. */
  /* we know that buffer.packed and buffer.DgpImage_RGB all fall in the same */
  /* slot as buffer.cmyk[0] */
  if (pri->freeMem){
    if (pri->buffer.cmyk[0]){
      DGP_FREE(pri->buffer.cmyk[0]);
    }
    if (pri->buffer.cmyk[0]){
      DGP_FREE(pri->buffer.cmyk[1]);
    }
    if (pri->buffer.cmyk[0]){
      DGP_FREE(pri->buffer.cmyk[2]);
    }
    if (pri->buffer.cmyk[0]){
      DGP_FREE(pri->buffer.cmyk[3]);
    }
  }
  DGP_FREE(pri);

  *ppri = NULL;

  NORMAL_RETURN;
}



/*	
  renderedImageDoesOverlap()
    determines if these coordinates (page coordinates forming a rectangle)
    have any points on the given renderedImage.
    
    returns DGP_TRUE if part of this rectangle overlaps the image
    DGP_FALSE if not
	
    pri - the strip of the page in question.
    x,y,w,h - coordinates at page resolution to check if they're 
      on the pageStrip.
*/
Dgp_Bool 
renderedImageDoesOverlap(renderedImage * pri,
		     Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h)
{
  DGP_VOID_PROLOGUE("renderedImageDoesOverlap",NULL);
  DGP_VOID_ASSERT(pri);
  if(pri && (x < (pri->x + pri->w)) && (pri->x < (x + w))
     && (y < (pri->y + pri->h)) && (pri->y < (y + h))){
    return DGP_TRUE;
  }
  return DGP_FALSE;
}



/* renderedImageSize()
   given a renderedimage *, this will tell you how many bytes it have
   been allocated for it (including the struct itself) for the call
   to renderedImageCreate. Note that if you passed in your own memory
   into renderedImageCreate, this is not counted, because you allocated
   it.
*/
ErrCode
renderedImageSize(renderedImage * pri, Dgp_UInt32 * psize,
		  Dgp_State * dgp_state)
{
  Dgp_Int32 i = 0, allocrows = 0;
  DGP_STATE_PROLOGUE("renderedImageSize()", dgp_state);
  DGP_ERR_GEN(!psize, cErrInvalidParm);
  DGP_ERR_PASS(isRenderedImageValid(pri, dgp_state));


  *psize = sizeof(renderedImage);
  if (pri->freeMem) {
    allocrows = (DgpCoord_isRotated(ORIENT_RD, pri->orient) ? pri->w : pri->h);
    if (pri->type != riCMYKSeparations) {
      *psize += allocrows * pri->bpl;
    }
    else {
      for (i = 0 ; i < 4 ; i++) {
	if (pri->buffer.cmyk[i]){
	  *psize += allocrows * pri->bpl;
	}
      }
    }
  }

  NORMAL_RETURN;
}

ErrCode
renderedImageCreateSize(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
                        DgpCoord_Resolution * res,
                        DgpCoord_Orientation orient,
                        Dgp_Int32 bpl,
                        renderedImageType type,
                        void * src,
                        Dgp_UInt32 * psize,
                        Dgp_State * dgp_state)
{
  Dgp_Int32 mem = 0;
  Dgp_Int32 allocrows = 0, alloccols = 0;
  DGP_STATE_PROLOGUE("renderedImageCreateSize()", dgp_state);
  
  DGP_ERR_GEN(w <= 0, cErrInvalidParm);
  DGP_ERR_GEN(h <= 0, cErrInvalidParm);
  DGP_ERR_GEN(!res, cErrInvalidParm);
  DGP_ERR_GEN(DgpCoord_le(res->xRes,DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN(DgpCoord_le(res->yRes,DGPCOORD_RAT_ZERO), cErrInvalidParm);
  DGP_ERR_GEN(bpl == 0 && src, cErrInvalidParm);

  (void)x;
  (void)y;
  mem = sizeof(renderedImage);

  allocrows = (DgpCoord_isRotated(ORIENT_RD, orient) ? w : h);
  alloccols = (DgpCoord_isRotated(ORIENT_RD, orient) ? w : h);

  if (!src){
    if (type == riMonoPacked){
      if (!bpl){
        bpl = (Dgp_Int32)ceil((double)alloccols / 32) * sizeof(Dgp_UInt32);
      }
      mem += (bpl * allocrows);
    }
    else if (type == riRGBInterleaved){
      if (!bpl){
        bpl = alloccols * sizeof(DgpImage_RGB);
      }
      mem += (bpl * allocrows);
    }
    else if (type & riCMYKSeparations){     
      if (!bpl){
        bpl = alloccols * sizeof(Dgp_UInt8);
      }
      if (type & riCSeparation){
        mem += (bpl * allocrows);
      }
      if (type & riMSeparation){
        mem += (bpl * allocrows);
      }
      if (type & riYSeparation){
        mem += (bpl * allocrows);
      }
      if (type & riKSeparation){
        mem += (bpl * allocrows);
      }
    }
    else {
      DGP_ERR_GEN(DGP_TRUE,cErrInvalidParm);
    }
  }

  if (psize){
    *psize = mem;
  }
  NORMAL_RETURN;
}
