#ifndef RENDERED_IMAGE_H
#define RENDERED_IMAGE_H

#include "rendered_image_t.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"

#endif

  
/* File rendered_image.h
   wbell 10/5/99
   access functions for rendered images.
*/
  
/* Verifies that a particular renderedImage is valid: width/height >= 0
   resolution is ok, bpl is non-zero, orientation is ok, buffer is non-null.
   Returns an error if the structure is bad.
*/
ErrCode isRenderedImageValid(renderedImage * ri, Dgp_State * dgp_state);

/* is this a valid renderedImageType */
#define isValidRenderedImageType(t) \
   (((t) == riMonoPacked) || ((t) == riRGBInterleaved) \
   || ((t) & riCMYKSeparations))

/* Creates a rendered image. If src is non-null and bpl is non-zero, we'll
   use what you gave us as memory, else we'll malloc some (note, not clear it.)

   x,y,w,h - represent the size/offset of the renderedImage in some coordinate
      system of your choosing. Commonly the page coordinate system, but you
      can be creative. (usually relative to ORIENT_RD)
   src_res / dst_res - source and destination resolutions of the image in the x and y directions.
   orient - the orientation of the image, relative to ORIENT_RD.
   bpl - number of bytes per line in the source buffer, if zero, will allocate
     the buffer ourselves. Note that all cmyk seperations must have the same
     bpl.
   type - specifies the type of the image, from one of the renderedImageTypes;
     note for CMYK, or the seperations that you'd like together.
   src - image data, if you have already allocated it. You must pass a bpl
     if this is set.  For CMYK seperations, the void * is actually
     an array of 4 Dgp_UInt8 pointers, c,m,y,k which we will copy.
   ppri - a pointer that will be set to the rendered image that
     will be allocated and returned.
     
     if we are to create the memory, and it is to be binary packed, it 
     will be cleared. Otherwise it will not be cleared.
*/
ErrCode
renderedImageCreate(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    DgpCoord_Resolution * src_res,
			    DgpCoord_Resolution * dst_res,
			   DgpCoord_Orientation orient,
			   Dgp_Int32 bpl,
			   renderedImageType type,
			   void * src,
			   renderedImage ** ppri,
			   Dgp_State * dgp_state);


/* Destroys a rendered image created with renderedImageCreate.
   If you passed in memory in renderedImageCreate to be used, it is not freed.
   Otherwise the image data is freed.
   *ppri is set to NULL.
 */
ErrCode
renderedImageDestroy(renderedImage **ppri, Dgp_State * dgp_state);

/*	
  renderedImageDoesOverlap()
    determines if these coordinates (page coordinates forming a rectangle)
    have any points on the given renderedImage.
    
    returns DGP_TRUE if part of this rectangle overlaps the image
    DGP_FALSE if not
	
    pri - the strip of the page in question.
    x,y,w,h - coordinates at page resolution to check if they're 
      on the pageStrip.
*/
Dgp_Bool 
renderedImageDoesOverlap(renderedImage * pri,
		     Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h);


/* renderedImageSize()
   given a renderedimage *, this will tell you how many bytes it have
   been allocated for it (including the struct itself) for the call
   to renderedImageCreate. Note that if you passed in your own memory
   into renderedImageCreate, this is not counted, because you allocated
   it.
*/
ErrCode
renderedImageSize(renderedImage * pri, Dgp_UInt32 * psize,
		  Dgp_State * dgp_state);


/* renderedImageCreateSize()
   Sometimes you're careful. Sometimes you want to kepe track of all your
   memory before you allocate it. We understand your paranoia. This function
   will tell you, if you made the same call to renderedImageCreate, how
   much memory would be allocated. From there you can make your decision.
   Good luck.
*/
ErrCode
renderedImageCreateSize(Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
                        DgpCoord_Resolution * res,
                        DgpCoord_Orientation orient,
                        Dgp_Int32 bpl,
                        renderedImageType type,
                        void * src,
                        Dgp_UInt32 * psize,
                        Dgp_State * dgp_state);


#ifdef __cplusplus
}
#endif
#endif
