#include "rendered_image_manip.h"
#include "zephimage.h"
#include "rendered_image.h"
#include "blit_mono.h"
#include "blit_byte.h"
#include "blit_rgb.h"
#include "blit.h"
#include "coord.h"


/* File rendered_image_manip.c
   wbell 10/11/99
   rendered image manipulation routines.
 */

/* A recursive helper function, we use start and end to tell us where in
   the ppimgs array we consider for clearing on each call.  We do it this
   way rather than just calling renderedImageClear() with one less nimgs
   because we get optimizations for the ordering of the images in how
   they're put down, we want to look at them in the order they go down
   so we don't end up clearing too much for transparent images that
   actually have another image under them.  End is inclusive. 
   x,y,w,h are the region of the rendered image (in that coordinate space)
   that we're looking at.
*/
ErrCode renderedImageClearRec(renderedImage * pri, santaImage * pimgs,
			      Dgp_Int32 start, Dgp_Int32 end,
			      Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			      Dgp_Bool cleared, 
			      Dgp_Int32 layer_num,
			      Dgp_State * dgp_state)
{
  Dgp_Int32 t, l, b, r;
  Dgp_Int32 ximg, yimg, wimg, himg;
  Dgp_Int32 fs, fo, fw;
  Dgp_Int32 ss, so, sw;
  DGP_STATE_PROLOGUE("renderedImageClearRec()", dgp_state);
  DGP_ERR_GEN_CLEANUP(!dgp_state, cErrInvalidParm);

  if ((start - 1) == end){
    /* we're at the end of the recursion. */
    /* actually clear this rectangle out. */
    /* first thing we need to do is get this thing in our orientation frame, */
    /* all the stuff so far has been in terms of ORIENT_RD, and now we */
    /* need to move over to the rendered image's native orientation. */
    DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(pri->x, pri->y,
					      pri->w, pri->h,
					      x, y, w, h, 
					      pri->orient,
					      &fo, &so,
					      &fs, &ss,
					      &fw, &sw,
					      dgp_state));
    if(pri->type == riMonoPacked){
      /* layer_num has no effect on mono buffers. */
      DgpBlat_blitSetBinary((Dgp_UInt8)-1, pri->buffer.packed, pri->bpl,
		    fo, so, fs, ss, fw, sw);
    }
    else if(pri->type == riRGBInterleaved){
      DgpImage_RGB color;
	  /* 5.3.2001 / dhtesq. */
	  /* Bizarre case.  I'm not even sure this is tfx-compliant as it stands */
	  /* (white on layer_num == 0, black otherwise), but it has to be modified */
	  /* for rendering a mono document into a color buffer.  This requires white */
	  /* for layer_num == 1 too. */
      color.r = color.g = color.b = (((layer_num == 0) || (layer_num == 1)) ? 0xff : 0x00);
      /* if we're just going to set this to 0, and it's already cleared, */
      /* we don't have to do anything. */
      if (!(cleared && color.r == 0)){
        DgpBlat_blitSetRGB(color, pri->buffer.rgb, pri->bpl,
		   fo, so, fs, ss, fw, sw);
      }
    }
    else if(pri->type & riCMYKSeparations){
      Dgp_Int32 i;
      if (!(cleared && layer_num != 0)){
	for (i = 0 ; i < 3 ; i++){
	  if (pri->buffer.cmyk[i]){
	    DgpBlat_blitSetByte((Dgp_UInt8)(layer_num == 0 ? 0xff : 0x00),
			pri->buffer.cmyk[i], pri->bpl, 
			fo, so, fs, ss, fw, sw);
	  }
	}
      }
    }
  }
  
  /* if it has doesn't have a transparency mask, then it'll cover this */
  /* area, and it doesn't need to be blanked. */
  /* this is kind of questionable: */
  /*  for mono rendered images, we're going to blank out any place where */
  /*  there exists an image, and make it selected anywhere there's not, which */
  /*  might get us into trouble if we do color pages in mono on the fly, */
  /*  and we decide the default mono for that case should be different. */
  else if ((pri->type == riMonoPacked) ||
	   (pimgs[start].zephImage->maskInfo == NULL && 
	    pimgs[start].zephImage->compType != IO_COMP_TOKENIZATION &&
	    pimgs[start].zephImage->compType != IO_COMP_JBIG2 &&
	    pimgs[start].zephImage->compType != IO_COMP_LLTOKENIZATION)) {

    /* remove the start'th image from the rectangle, */
    /* splitting it into 4 rectangles (some of which may be empty) */
    /* and recurse on them. */
    
    /* where does the image lie in this coordinate space  */
    /* (basically taking into effect the scale.) */
   DGP_ERR_PASS_CLEANUP(zephCalcPagePosition(pimgs[start].zephImage,
						&ximg, &yimg,
						&wimg, &himg));

    /* if this is mono, we clear this area. */
    if (pri->type == riMonoPacked && !cleared){
      /* crop the cleared area to the size of the rendered image; it */
      /* doesn't need to be the right size  */
      fs = MAX(ximg, pri->x);
      ss = MAX(yimg, pri->y);
      fw = MIN(wimg, (pri->x + pri->w) - fs);
      sw = MIN(himg, (pri->y + pri->h) - ss);
      DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(pri->x, pri->y,
                        pri->w, pri->h,
						fs, ss, fw, sw,
						pri->orient,
						&fo, &so,
						&fs, &ss,
						&fw, &sw,
						dgp_state));
      /* 30.8.2000 / dhtesq.
      In running kringle_test, reached this code for the first time I'm aware of (cleared flag not set).
      I can't figure out how it was ever supposed to work.

      DgpBlat_blitSetBinary((Dgp_UInt8)0, pri->buffer.packed, pri->bpl,
		    fo, so, fs, ss, fw, sw);
      */
      DgpBlat_blitSetBinary((Dgp_UInt8)0, pri->buffer.packed, pri->bpl,
		    0, 0, 0, 0, fw, sw);
    }

    
    /* now figure out how many rows we need on the top/left/bottom/right */
    /* ...taking into account the band size.  Image might not even lie on the band.  / dhtesq. */
    t = yimg - y;
    b = (y + h) - (yimg + himg);
    l = ximg - x;
    r = (x + w) - (ximg + wimg);

    if(t > h){
        t = h;
    }
    if(t < 0){
        t = 0;
    }

    if(b > h){
        b = h;
    }
    if(b < 0){
        b = 0;
    }

    if(l < 0){
        l = 0;
    }
    if(l > w){
        l = w;
    }

    if(r < 0){
        r = 0;
    }
    if(r > w){
        r = w;
    }


    /* "Optimization" for cases where image does not lie on band... */
    if((t == h) || (b == h)){
        l = r = 0;
    } else if((l == w) || (r == w)){
        t = b = 0;
    }

    /* So what's the rectangle on the top? */
    if (t > 0){        
      renderedImageClearRec(pri, pimgs, start + 1, end,
			    x, y, w, t, cleared, layer_num, dgp_state);                
    }	 
    /* ...the bottom? */
    if (b > 0){        
      renderedImageClearRec(pri, pimgs, start + 1, end,
			    x, MAX((yimg + himg), y), w, b, cleared,
			    layer_num, dgp_state);                
    }

    /* the left; now note that we don't want to recurse on the section */
    /* of the top and bottom we just did. */
    if (l > 0){
      renderedImageClearRec(pri, pimgs, start + 1, end,
                x,
                l == w ? y:MAX(y, yimg),
                l, 
                l == w ? h:MIN((y + h), (yimg + himg)) - MAX(y, yimg),
			    cleared, layer_num, dgp_state);
    }
    /* same for the right. */
    if (r > 0){
        /* 8.6.2000 / dhtesq. */
        /* This call used to pass 'x + wimg' as the x-coordinate.  This does not account for the */
        /* width of the left margin.  Check this if you have funny black chunks scattered among */
        /* your images. / dhtesq. */
      renderedImageClearRec(pri, pimgs, start + 1, end,
			    MAX(ximg + wimg, x),
                r == w ? y:MAX(y, yimg),
                r, 
                r == w ? h:MIN((y + h), (yimg + himg)) - MAX(y, yimg),
			    cleared, layer_num, dgp_state);
    }
  }
  /* has a transparency mask or is a tokenized image; */
  /* just recurse downward on the same rect. */
  else {
    renderedImageClearRec(pri, pimgs, start + 1, end,
			  x, y, w, h, cleared, layer_num, dgp_state);
  }
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}


ErrCode
renderedImageClear(renderedImage * pri, santaImage * pimgs,
		   Dgp_Int32 nimgs, Dgp_Bool cleared, 
		   Dgp_Int32 layer_num, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("renderedImageClear()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(layer_num < 0, cErrInvalidParm);
  DGP_ERR_PASS(isRenderedImageValid(pri, dgp_state));
  
  DGP_ERR_GEN(!pimgs, cErrInvalidParm);
  DGP_ERR_PASS(renderedImageClearRec(pri, pimgs, 0, nimgs - 1,
				     pri->x, pri->y, 
				     pri->w, pri->h,
				     cleared,
				     layer_num,
				     dgp_state));
  
  NORMAL_RETURN;
}


/* renderedImageIsImgOn()
   Determines if a specified image is on this renderedImage.
   pri - the rendered image
   pimg - the image.
   pon - a Dgp_Bool that is set to DGP_TRUE if the image is on the
    strip, DGP_FALSE otherwise.
*/
ErrCode
renderedImageIsImgOn(renderedImage * pri, santaImage * pimg,
		 Dgp_Bool * pon, Dgp_State * dgp_state)
{
  Dgp_Int32 x, y, w, h;
  DGP_STATE_PROLOGUE("renderedImageIsImgOn", dgp_state);
  DGP_ERR_GEN_CLEANUP(!pri, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!pimg, cErrInvalidParm);
  DGP_ERR_GEN_CLEANUP(!pon, cErrInvalidParm);

  DGP_ERR_PASS_CLEANUP(zephCalcPagePosition(pimg->zephImage, &x, &y, &w, &h));

  *pon = renderedImageDoesOverlap(pri, x, y, w, h);
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}



/* renderedImageCrop()
   given two rendered images of the same type, copy a portion of src
   into dst. 
   pdst - the destination image.
   dstX, dstY - where the first pixel of the cropped src image will fall in
    the dst image (relative to native RD orientation)
   psrc - the source image
   srcX, srcY, srcW, srcH - what portion of the src image is to be cropped
    and inserted into the dst image (coordinates are relative to RD 
    orientation)
   note: if dst is binary, will blit with or.
*/
ErrCode
renderedImageCrop(renderedImage * pdst, Dgp_Int32 dstX, Dgp_Int32 dstY,
                  renderedImage * psrc, Dgp_Int32 srcX, Dgp_Int32 srcY,
                  Dgp_Int32 srcW, Dgp_Int32 srcH, 
                  santaPage *ppage,
                  Dgp_State * dgp_state)
{
  Dgp_Int32 df, ds, dfo, dso, sf, ss, sfo, sso, fw, sw; 
  Dgp_Int32 of, os, ofo, oso, tf, ts, tfo, tso;

  DGP_STATE_PROLOGUE("renderedImageCrop()", dgp_state);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!psrc, cErrInvalidParm);
  /* they have to be the same type. */
  DGP_ERR_GEN(psrc->type != pdst->type, cErrInvalidParm);
  DGP_ERR_PASS(isRenderedImageValid(psrc, dgp_state));
  DGP_ERR_PASS(isRenderedImageValid(pdst, dgp_state));
  
 

  /* ADD: right now they need to be the same orientation. There's
     no real reason for this. Someone needs to implement the correct
     transformations.
  */
  if (psrc->orient != pdst->orient){
    DGP_WARNF0("renderedImageCrop: orientations must be the same\n");
    DGP_ERR_GEN(psrc->orient != pdst->orient, cErrInvalidParm);
  }

  /* Transform source image coordinates... */
  DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(   0, 0,
                                                        ppage->width, ppage->height,
                                                        psrc->x, psrc->y,
                                                        psrc->w, psrc->h,
                                                        pdst->orient,
                                                        &dfo, &dso, 
                                                        &df, &ds, 
                                                        NULL, NULL,
                                                        dgp_state)); 

  /* Transform dst image coordinates... */
  DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(   0, 0,
                                                        ppage->width, ppage->height,
                                                        pdst->x, pdst->y,
                                                        pdst->w, pdst->h,
                                                        pdst->orient,
                                                        &tfo, &tso, 
                                                        &tf, &ts, 
                                                        NULL, NULL,
                                                        dgp_state)); 

  /* Transform selection coordinates... */
  DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(   0, 0,
                                                        ppage->width, ppage->height,
                                                        srcX, srcY, 
                                                        srcW, srcH, 
                                                        pdst->orient,
                                                        &sfo, &sso, 
                                                        &sf, &ss, 
                                                        &fw, &sw,
                                                        dgp_state)); 

  /* Transform destination start point coordinates... */
  DGP_ERR_PASS_CLEANUP(DgpCoord_coordOrientTransform(   0, 0,
                                                        ppage->width, ppage->height,
                                                        dstX, dstY, 
                                                        srcW, srcH, 
                                                        pdst->orient,
                                                        &ofo, &oso, 
                                                        &of, &os, 
                                                        NULL, NULL,
                                                        dgp_state)); 
 
  if (psrc->type == riMonoPacked){
    DgpBlat_blitBitsClipLR( pdst->buffer.packed,
                            pdst->bpl / sizeof(Dgp_UInt32),
                            (Dgp_UInt32 *)((Dgp_Int8 *)psrc->buffer.packed + ((Dgp_Int32)abs(ss - ds) * psrc->bpl)),
                            psrc->bpl / sizeof(Dgp_UInt32),
                            (Dgp_Int32)abs(sf - df), 
                            fw, sw, 
                            (Dgp_Int32)abs(tf - of), (Dgp_Int32)abs(ts - os),
                            dgp_state);
  }
  else {
    /* ADD: need more image types here. */
    DGP_WARNF0("renderedImageCrop: unsupported image type\n");
    DGP_ASSERT(0);
  }

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}
