#ifndef RENDERED_IMAGE_MANIP_H
#define RENDERED_IMAGE_MANIP_H

#include "dgp_state.h"
#include "rendered_image_t.h"
#include "santa_image_t.h"

/* File rendered_image_manip.h
   wbell 10/11/99
   rendered image manipulation routines.
 */

#ifdef __cplusplus
extern "C"
{
#endif

/* renderedImageClear takes a rendered_image and clears it using the
   following rules:
    for a mono rendered_image, it is set to entirely selected (1)
     except for the regions occupied by images,
     where the rendered_image is set to non-selected (0) so that 
     the images can be put down with or correctly.
    for a color rendered_image, it is set to white (255,255,255) for
     all areas on the rendered image where there is no image covering
     or there is only an image covering that has a transparency mask.
     if we're dealing with a non-background layer (layer_num != 0), then
     this empty area is set to black (0,0,0) as per the xiff spec.
   This is useful for prepping a selector layer before putting down the
    images, and for clearing PageStrips.
   We assume that the renderedImage coordinates are relative to 
    orient_rd, although we do handle the coordinate transformation
    to clear the region in the rendered_image's orientation space.

    pri    - a pointer to a created rendered image.
    pimgs -  an array of images that are to be considered
           (keep in mind that these images should all have no other selection
	   than possible transparency masks or this doesn't work. Usually,
	   the images you'd want to use are the ones in the background layer
	   [for color pages], or the ones in a selector [mono pages/ selector]
    nimgs  - the number of images in the array.
    cleared - if this buffer has been zeroed out.
    layer_num - the number layer we will be putting down (0 is background),
           this just affects the default clear color.
*/
ErrCode
renderedImageClear(renderedImage * pri, santaImage * pimgs,
		   Dgp_Int32 nimgs, Dgp_Bool cleared, Dgp_Int32 layer_num,
		   Dgp_State * dgp_state);

/* renderedImageIsImgOn()
   Determines if a specified image is on this renderedImage.
   pri - the rendered image
   pimg - the image.
   pon - a Dgp_Bool that is set to DGP_TRUE if the image is on the
    strip, DGP_FALSE otherwise.
*/
ErrCode
renderedImageIsImgOn(renderedImage * pri, santaImage * pimg,
		 Dgp_Bool * pon, Dgp_State * dgp_state);


/* renderedImageCrop()
   given two rendered images of the same type, copy a portion of src
   into dst. 
   pdst - the destination image.
   dstX, dstY - where the first pixel of the cropped src image will fall in
    the dst image (relative to native RD orientation)
   psrc - the source image
   srcX, srcY, srcW, srcH - what portion of the src image is to be cropped
    and inserted into the dst image (coordinates are relative to RD 
    orientation)
   note: if dst is binary, will blit with or.
*/
ErrCode
renderedImageCrop(renderedImage * pdst, Dgp_Int32 dstX, Dgp_Int32 dstY,
                  renderedImage * psrc, Dgp_Int32 srcX, Dgp_Int32 srcY,
                  Dgp_Int32 srcW, Dgp_Int32 srcH, 
                  santaPage *ppage,
                  Dgp_State * dgp_state);

#ifdef __cplusplus
}
#endif
#endif
