#include "santa_private_t.h"
#include "santa.h"
#include "santa_layer.h"
#include "memory_sentry.h"
#include "page_strip_order.h"
#include "page_strip.h"
#include "doc_helpers.h"
#include "page_helpers.h"
#include "rendered_image_manip.h"
#include "orientation.h"
#include "santa_macros.h"
#include "santa_macros.h"
#include "santa_halftone.h"

#include "scaleBlit.h"
#include "scaleBlitAvgColor.h"
#include "scaleBlitAvgMono.h"
#include "image_defs.h"

/* santa.c
   wbell 6/8/98
   operations on core santa datatypes.
   visible to the outside world.
   reincarnated 10.11.99
*/

/* note that the santaDoc memorySentry allocation and memory counting 
   happens at the levels below this, so it's important not to allocate
   anything other than the base structures (which are kept track of below)
   here, or it won't get counted. santaJob memory is not counted.
*/



/*	santaOpenDocument()
	starts off the whole process of decompressing your favorite XIFF file
        creates the SantaDoc cookie that we'll carry around for the rest
	of the document decompression. 
	
	filename	-	the file to be opened.
	memlimit - how much memory can we use [bytes] (set an upper bound on
	 the amount of memory that rendering this page will ever consume.
	 note, this does not include jobs, nor the buffer memory
	 (the destination band memory) that a job needs for destination.
	opts - any document specific options you want to pass in. Else, pass 
	 in santaDocOptsDefault.
	ppdoc		-	a pointer to a SantaDocInfo *. 
	We will allocate a XiffDocInfo and initialize it and set
	this pointer to be valid. Your double-pointer can
	point to a pointer to anything
*/
ErrCode santaOpenDocument(Dgp_Int8 * filename, 
			  Dgp_Int32 memlimit, santaDocOptions opts,
			  Dgp_State * dgp_state,
			  santaDoc ** ppdoc)
{
  FILE * fp;
  uds tmpuds;
  
  DGP_STATE_PROLOGUE("santaOpenDocument()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!filename, cErrInvalidParm);
  DGP_ERR_GEN(!ppdoc, cErrInvalidParm);
  DGP_ERR_GEN(memlimit <= 0, cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaDocOptions(opts), cErrInvalidParm);
  
  fp = fopen(filename,"rb");
  DGP_ERR_GEN_CLEANUP(!fp, cErrFileNotFound);
  
  tmpuds.fp     = (void *)fp;
  tmpuds.fseek  = (fseek_func)fseek;
  tmpuds.ftell  = (ftell_func)ftell;
  tmpuds.fread  = (fread_func)fread;
  tmpuds.fwrite = (fwrite_func)fwrite;
  
  DGP_ERR_PASS_CLEANUP(santaOpenDocumentEx(&tmpuds, memlimit, opts,
					   dgp_state, ppdoc));

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (fp){
    fclose(fp);
  }
  ERR_RETURN;
}


/*	santaOpenDocumentEx()
	just like santaOpenDocument, but lets you use your own uds 
	(a file you've already opened up.) We will store away a copy
	of the uds structure, don't free it or nothing.
	
	docuds	        -	the opened file, with it's seek/read 
                            	functions, will be shallowly copied
				and saved.
	memlimit - how much memory can we use [bytes] (set an upper bound on
	 the amount of memory that rendering this page will ever consume.
	 note, this does not include jobs, nor the buffer memory
	 (the destination band memory) that a job needs for destination.
	opts - any document specific options you want to pass in. Else, pass 
	 in santaDocOptsDefault.
	ppdoc		-	a pointer to a santaDoc *. 
	We will allocate a santaDoc and initialize it and set
	this pointer to be valid. Your double-pointer can point to a 
	pointer to anything.
*/
#define DOCSIZE (Dgp_Int32)(sizeof(santaDoc) +  sizeof(_santaPrivateDoc))
ErrCode santaOpenDocumentEx(uds * docuds, Dgp_Int32 memlimit, 
			    santaDocOptions opts,
			    Dgp_State * dgp_state, 
			    santaDoc ** ppdoc)
{
  santaDoc * pdoc = NULL;
  DGP_STATE_PROLOGUE("santaOpenDocumentEx()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!docuds, cErrInvalidParm);
  DGP_ERR_GEN(!ppdoc, cErrInvalidParm);
  /* need to have at least enough memory to make the structure */
  /* and then make recursive call. */
  DGP_ERR_GEN(memlimit < DOCSIZE, cErrInsufficientMem);  
  DGP_ERR_GEN(!isValidSantaDocOptions(opts), cErrInvalidParm);
  
  /* calloc the job and its private data structure */
  DGP_CHECK_ALLOC_CLEANUP(pdoc = (santaDoc *)
			  DGP_CALLOC(1, DOCSIZE));
  doc_private(pdoc) = (_santaPrivateDoc *)((Dgp_Int8 *)pdoc + sizeof(santaDoc));
  
  DGP_ERR_PASS_CLEANUP(docOpen(pdoc, docuds, memlimit, DOCSIZE, 
			       opts, dgp_state));

  *ppdoc = pdoc;
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pdoc){
    DGP_FREE(pdoc);
  }
  ERR_RETURN;
}

/*	santaCloseDocument()
	cleans up all the document related info; destroys the document cookie
	and closes the opened file. 
	use for files that were opened with santaOpenDocument
	
	ppdoc		-	a double-pointer to a santaDoc that
	was created with santaOpenDocument(). Will be set to NULL.
*/
ErrCode santaCloseDocument(santaDoc ** ppdoc, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("santaCloseDocument()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ppdoc, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppdoc), cErrInvalidParm);

  fclose(doc_private(*ppdoc)->u.fp);
  DGP_ERR_PASS(santaCloseDocumentEx(ppdoc, dgp_state));

  NORMAL_RETURN;
}

/*	santaCloseDocumentEx()
	just like santaCloseDocument, but doesn't close the file.
	that's your job.
	
	ppdoc		-	a double-pointer to a santaDoc
	that was created with santaOpenDocumentEx(). Will be set to NULL.
*/
ErrCode santaCloseDocumentEx(santaDoc ** ppdoc, Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("santaCloseDocumentEx()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ppdoc, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppdoc), cErrInvalidParm);
    
  DGP_ERR_CHECK(docClose(*ppdoc, dgp_state)){
    UPDATE_ERR(firstErr);
  }
  
  DGP_FREE(*ppdoc);
  *ppdoc=NULL;

  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}


/*	santaOpenPage()
	creates a santaPage cookie about a particular page in 
	a previously opened XIFF file; returns interesting information
	about the page that can help one decided orientation and such
	before starting a rendering job.
	
	pdoc -a previously opened XIFF document
	 (opened with santaOpenDocument*())
	page_num - the page number to open (note pages start numbering at 0)
	pppage - a double-pointer to where we'll set a pointer to our created
	and initialized SantaPage cookie.
*/
#define PAGESIZE (Dgp_Int32)(sizeof(santaPage) +  sizeof(_santaPrivatePage))
ErrCode santaOpenPage(santaDoc * pdoc, Dgp_Int32 page_num, Dgp_State * dgp_state,
		      santaPage ** pppage)
{
  santaPage *ppage = NULL;	
  DGP_STATE_PROLOGUE("santaOpenPage()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(page_num < 0 || page_num >= pdoc->npages, cErrInvalidParm);
  DGP_ERR_GEN(!pppage, cErrInvalidParm);

  /* so we might have this page opened already, if so, we just want */
  /* to give back a pointer to the currently opened one. */
  if (doc_private(pdoc)->pppages[page_num]){
    ppage = doc_private(pdoc)->pppages[page_num];
  }
  else{
    /* make sure there's enough memory available to make this struct */
    DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
						 PAGESIZE,
						 dgp_state));
    /* calloc the user visible side + the non-visible part */
    DGP_CHECK_ALLOC_CLEANUP(ppage = (santaPage *)
			    DGP_CALLOC(1, PAGESIZE));
    page_private(ppage) = (_santaPrivatePage *)((Dgp_Int8 *)ppage +
						sizeof(santaPage));
    
    DGP_ERR_PASS_CLEANUP(pageOpen(pdoc, ppage, page_num, dgp_state));
  }

  /* up the reference count. */
  page_private(ppage)->nrefs ++;

  *pppage = ppage;
  NORMAL_RETURN;
ERR_CLEANUP:
  if (ppage){
    DGP_FREE(ppage);
    *pppage = NULL;

    /* free the memory that's been claimed in the memorySentry. */
    memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
			      PAGESIZE, dgp_state);
  }
  ERR_RETURN;
}


/*	santaClosePage()
	destroys a santaPage cookie about a particular page that 
	has been previously opened.
	
	pppage -  a double pointer to the previously opened
	page that will be closed. Will be set to NULL.
*/
ErrCode santaClosePage(santaDoc * pdoc, santaPage ** pppage,
		       Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("santaClosePage()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!pppage, cErrInvalidParm);
  DGP_ERR_GEN(!(*pppage), cErrInvalidParm);

  /* we'll check the reference count, and make sure that */
  /* there are no more references, and if so, we'll close it. */
  page_private(*pppage)->nrefs --;

  if (!page_private(*pppage)->nrefs){
    /* remove the entry in the doc page table. */
    doc_private(pdoc)->pppages[(*pppage)->page_num] = NULL;

    DGP_ERR_CHECK(pageClose(pdoc, *pppage, dgp_state)){
      UPDATE_ERR(firstErr);
    }
    
    DGP_FREE(*pppage);

    /* free the memory that's been claimed in the memorySentry. */
    DGP_ERR_CHECK(memorySentryReleaseMemory(doc_private(pdoc)->memmgr,
					    PAGESIZE, dgp_state)){
      UPDATE_ERR(firstErr);
    }

  }
  
  *pppage = NULL;

  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}


/*	santaStartRenderJob()
	Starts a job to render a portion of the page at in a specified
	orientation.

	pdoc - the document in question.
	ppage -  the page to start the rendering job on.
	x,y,w,h - the coordinates of the page that you wish to
	 generate in ORIENT_RD orientation.
	o - the page is in native (ORIENT_RD) orientation, which
	 orientation do you wish to render in.
	bandsize - the job will be done in bands of this size in the slow
	 direction (they will be done full width in the fast direction)
	advorder - what criteria do we use to go from one band to the next?
	 advance down the slow direction? across the color space?
        inputtype - what kind of memory will you be passing in to the
	 band render? cmyk seperations one at a time? rgb interleaved?
	opts - what job-level options do you want. If unsure, just pass
	 in santaJobOptsDefault
	ppjob - pointer to where the render job state will be stored.
*/
ErrCode santaStartRenderJob(santaDoc * pdoc, santaPage * ppage,
			    Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    DgpCoord_Orientation o, Dgp_Int32 bandsize,
			    santaJobAdvanceOrder advorder,
			    santaInputType inputtype,
			    santaJobOptions opts,
			    Dgp_State * dgp_state,
			    santaRenderJob ** ppjob)
{
  santaRenderJob * pjob = NULL;
  stripAdvanceOrderType stripAdvOrder = -1;
  orderInputType stripInputType = -1;
  pageAreaAdvanceOrderType stripPageAdvOrder = -1;

  DGP_STATE_PROLOGUE("santaStartRenderJob()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!ppjob, cErrInvalidParm);
  DGP_ERR_GEN(x < 0 || x >= (Dgp_Int32)page_private(ppage)->zephPage->width, 
	      cErrInvalidParm);
  DGP_ERR_GEN(y < 0 || y >= (Dgp_Int32)page_private(ppage)->zephPage->height,
	      cErrInvalidParm);
  DGP_ERR_GEN(w <= 0 || (x + w) > (Dgp_Int32)page_private(ppage)->zephPage->width,
	      cErrInvalidParm);
  DGP_ERR_GEN(h <= 0 || (y + h) > (Dgp_Int32)page_private(ppage)->zephPage->height,
	      cErrInvalidParm);
  DGP_ERR_GEN(bandsize <= 0, cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaAdvanceOrder(advorder), cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaInputType(inputtype), cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaJobOptions(opts), cErrInvalidParm);
  DGP_ERR_GEN(!ISVALIDORIENT(o), cErrInvalidParm);

  switch(inputtype)
    {
    case santaInputMonoPacked:
      stripInputType = oiMonoPacked;
      break;
    case santaInputRGB:
      stripInputType = oiRGBInterleaved;
      break;
    case santaInputCMYKSingle:
      stripInputType = oiSingleSeparationsCMYK;
      break;
    case santaInputCMYKAll:
      stripInputType = oiAllSeparationsCMYK;
      break;
    }

  switch(advorder)
    {
    case santaJobAdvanceSlow:
      stripAdvOrder = saoPageAreaProgression;
      stripPageAdvOrder = paoAdvanceSlow;
      break;
    case santaJobAdvanceCMYKSlow:
      stripAdvOrder = saoColorSpaceProgression;
      stripPageAdvOrder = paoAdvanceSlow;
      break;
    case santaJobAdvanceSlowCMYK:
      stripAdvOrder = saoPageAreaProgression;
      stripPageAdvOrder = paoAdvanceSlow;
      break;
    }

  /* calloc the job and the private structure. */
  DGP_CHECK_ALLOC_CLEANUP(pjob = (santaRenderJob *)
			  DGP_CALLOC(1,sizeof(santaRenderJob) + 
				     sizeof(_santaPrivateRenderJob)));
  job_private(pjob) = (_santaPrivateRenderJob *)((Dgp_Int8 *)pjob +
				  sizeof(_santaPrivateRenderJob));
  job_private(pjob)->opts = opts;
  pjob->ppage = ppage;
  /* make the page order. */
  DGP_ERR_PASS_CLEANUP(pageStripOrderCreate(x, y, w, h, o, 
					    &page_private(ppage)->
					       zephPage->res,
					    stripAdvOrder,
					    stripInputType,
					    stripPageAdvOrder,
					    bandsize,
					    &(job_private(pjob))->order,
					    dgp_state));
		       
  *ppjob = pjob;
  
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pjob){
    if (job_private(pjob)){
      if (job_private(pjob)->order){
	pageStripOrderDestroy(&(job_private(pjob)->order), dgp_state);
      }
    }
    DGP_FREE(pjob);
  }
  if (ppjob){
    *ppjob = NULL;
  }
  ERR_RETURN;
}


/*	santaEndRenderJob()
	Ends a rendering job of a portion of the page.

	pdoc - a previously opened XIFF document
	(opened with santaOpenDocument*())
	ppage -  the page to a end rendering job on.
	ppjob - the job to end. 
*/
ErrCode santaEndRenderJob(santaDoc * pdoc, santaPage *ppage,
			  santaRenderJob ** ppjob, Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("santaEndRenderJob()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!ppjob, cErrInvalidParm);
  DGP_ERR_GEN(!(*ppjob), cErrInvalidParm);

  DGP_ERR_CHECK(pageStripOrderDestroy(&(job_private(*ppjob)->order),
					     dgp_state)){
    UPDATE_ERR(firstErr); 
  }

  DGP_FREE(*ppjob);
  *ppjob = NULL;
  
  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}


/*      santaJobDoBand()
	Does the next band of a rendering job.
	pdoc - the SantaDoc for the job in question.
	ppage - the SantaPage for the job in question.
	pjob - the job we're talking about.
	data - where the raster data is for this band;
	 this void * will be interpreted according to the
	 rules in the jobs inputtype.
	bpl  - number of bytes per line in the source data.
	cleared - is this memory cleared? if so, we won't clear
	 the appropriate portions. (note, having us clear what we
	 need is generally better than preclearing the image.)
/* return values, can be null: */
        px, py, pw, ph - the position on the page that this band
	 corresponds to in native orientation. santaJobDoBand is called
	 on a job that has already had all it's bands rendered, but
	 has not been ended, these will be set to -1.

	 if this fails, the buffer pointed to by data is in some 
	 indeterminant state (some of the images may be put down, some
	 not, etc...)
*/
ErrCode santaRenderJobDoBand(santaDoc * pdoc, santaPage * ppage,
			     santaRenderJob * pjob, void * data, Dgp_Int32 bpl,
			     Dgp_Bool cleared, 
			     Dgp_Int32 * px, Dgp_Int32 * py, Dgp_Int32 * pw, Dgp_Int32 * ph,
                 Dgp_Int32 layerOptions,
			     Dgp_State * dgp_state)
{
  pageStrip * pstrip = NULL;
  Dgp_Int32 i;
  DgpCoord_Resolution scale_res;
  Dgp_Int32 rows = 0, cols = 0;
  Dgp_Bool halftoneTheOutput = DGP_FALSE;
  /* In case we need to halftone... */
  DgpImage_ColorRGB rgbImage, tempRGBImage = NULL;
  DgpImage_Packed binaryImage, tempBinaryImage;
  PIXR *tempBinaryPixr;
  void *tempData;
  Dgp_Int32 tempBpl;
  Dgp_Int32 *dataArea;
  Dgp_Bool showJbig2 = (layerOptions & DGP_SHOWJBIG2_MASK) && DGP_TRUE;
  Dgp_Bool showImages = (layerOptions & DGP_SHOWIMAGES_MASK) && DGP_TRUE;
  /*//// */

  DGP_STATE_PROLOGUE("santaRenderJobDoBand()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pjob, cErrInvalidParm);
  DGP_ERR_GEN(!data, cErrInvalidParm);
  /* note bpl can be positive/negative, not zero. */
  DGP_ERR_GEN(!bpl, cErrInvalidParm);

  /* Create a page strip to represent this portion of the page */
  /* and this raster data. */

  /* Halftone the output if going from a color document to a mono buffer. */
  if((ppage->page_type == santaInputRGB) && (ppage->outputType == santaInputMonoPacked)){
	  halftoneTheOutput = DGP_TRUE;
  }
  if((ppage->scale_me == DGP_FALSE) && (halftoneTheOutput == DGP_FALSE)){
	  /* Same old same-o. */
      DGP_ERR_PASS_CLEANUP(pageStripOrderGetNextBand(job_private(pjob)->order,
                            data, bpl, cleared,
                            &pstrip, &ppage->dst_res, dgp_state));
  } else {
      /* We have to create an image, packed or RGB or CMYK, at native resolution. */
      /* Let the embedded renderedImageCreate() call do the work for us. */
      /* ppage->scale_factor has been set before this call... */
      /*ppage->scale_factor = DgpCoord_float(DgpCoord_div(ppage->dst_res.xRes, ppage->res.xRes)); */
      DGP_ERR_PASS_CLEANUP(pageStripOrderGetNextBand(job_private(pjob)->order,
                            NULL, 0, cleared,
                            &pstrip, &ppage->dst_res, dgp_state));
  }

  /* is this render job done? if so, pstrip will be null despite the */
  /* the last call succeding. we set the output page coords to -1 and */
  /* the w/h to 0 to let the outside guy know that it's all done. */
  if (!pstrip){
    if (px){
      *px = -1;
    }
    if (py){
      *py = -1;
    }
    if (pw){
      *pw = 0;
    }
    if (ph){
      *ph = 0;
    }
    DGP_ERR_GEN_NO_MSG(DGP_TRUE, cSuccessful);
  }
		       
  /* lets get down to the dirt. If this is a santaPageMono, we're just */
  /* going to put down layer 2 (the selector, but starting at 0) on */
  /* this pageStrip. Else, we're going to do the full blown xiff thing */
  /* where we put down the background and each successive layer pair */
  /* after that. */
  if (ppage->page_type == santaPageMono){
    /* we only put down the even ones. note we count from zero though. */
    for (i = 1; i < (page_private(ppage)->nlayers); i += 2){
      DGP_ERR_PASS_CLEANUP(santaLayerAddLayer(pdoc, ppage,
					      i, pstrip,
					      job_private(pjob)->order,
					      doc_private(pdoc)->memmgr,
					      dgp_state));
    }
  }
  else if (ppage->page_type == santaPageColor){
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
	  /* we'll put down the background. */
	  Dgp_Bool layerIsJbig2;
	  Dgp_Int32 jbig2LayersRendered = 0;
	  
	  santaLayer *player = &(page_private(ppage)->players[0]);
	  
	  /* check if there is anything in the background */
	  if (player->nimages > 0)
	  {
		  layerIsJbig2 = (player->pimgs->zephImage->compType == IO_COMP_JBIG2);
		  
		  if((layerIsJbig2 && showJbig2) || (!layerIsJbig2 && showImages)){
			  if(layerIsJbig2) jbig2LayersRendered++;
			  if((jbig2LayersRendered == 1) && !showImages &&
				  (job_private(pjob)->order->inputType == oiRGBInterleaved)){
				  
				  DGP_ERR_PASS_CLEANUP(renderedImageClear(pstrip->pimg, 
					  player->pimgs,
					  player->nimages,
					  pstrip->cleared, 
					  0,
					  dgp_state));
				  
			  }
			  DGP_ERR_PASS_CLEANUP(santaLayerAddLayer(	pdoc, ppage,
				  0, pstrip, 
				  job_private(pjob)->order,
				  doc_private(pdoc)->memmgr,
				  dgp_state));
		  }
	  }
	  else
	  {
		  /* if there is no background specified, set it to white */
		 /* memset((void *)pstrip->pimg->buffer.rgb, 0xff, pstrip->pimg->bpl * pstrip->pimg->h); */
	  	  DGP_ERR_PASS_CLEANUP(renderedImageClear(pstrip->pimg, 
		  player->pimgs,
		  player->nimages,
		  pstrip->cleared, 
		  0,
		  dgp_state));

      }
	  

	  /* Now put down successive selector / foreground pairs. */
	  for (i = 1; i  < (page_private(ppage)->nlayers); i += 2){
		  santaLayer *player = &(page_private(ppage)->players[i]);
		  layerIsJbig2 = (player->pimgs->zephImage->compType == IO_COMP_JBIG2);
		  
		  if((layerIsJbig2 && showJbig2) || (!layerIsJbig2 && showImages)){
			  if(layerIsJbig2) jbig2LayersRendered++;
			  if((jbig2LayersRendered == 1) && !showImages &&
				  (job_private(pjob)->order->inputType == oiRGBInterleaved)){
				  
				  DGP_ERR_PASS_CLEANUP(renderedImageClear(	pstrip->pimg, 
															player->pimgs,
															player->nimages,
															pstrip->cleared, 
															i,
															dgp_state));
                                                      
				  /*memset((void *)pstrip->pimg->buffer.rgb, 0xff, pstrip->pimg->bpl * pstrip->pimg->h); */
			  }
			  DGP_ERR_PASS_CLEANUP(santaLayerAddPairedLayers(pdoc, ppage, 
															 i, i+1,
															 pstrip, 
															 job_private(pjob)->order,
															 doc_private(pdoc)->memmgr,
															 dgp_state));
		  }
	  }

#endif
#if 0
          DGP_ERR_PASS_CLEANUP(santaLayerAddLayer(pdoc, ppage,
					    0, pstrip, 
					    job_private(pjob)->order,
					    doc_private(pdoc)->memmgr,
					    dgp_state));

          /* Now put down successive selector / foreground pairs. */
          for (i = 1; i  < (page_private(ppage)->nlayers); i += 2){
              DGP_ERR_PASS_CLEANUP(santaLayerAddPairedLayers(pdoc, ppage, 
                                                       i, i+1,
                                                       pstrip, 
                                                       job_private(pjob)->order,
                                                       doc_private(pdoc)->memmgr,
                                                       dgp_state));
          }
    
          /*
          for(i = 0; i < (page_private(ppage)->nlayers); i++){
          DGP_ERR_PASS_CLEANUP(santaLayerAddLayer(pdoc, ppage, i, pstrip, job_private(pjob)->order, 
                doc_private(pdoc)->memmgr, dgp_state));
                }
          */
          /* so we could have a single selector as the last layer, so we don't */
          /* want to put it down as a pair, we'll just put that down. */
          if (i != (page_private(ppage)->nlayers)){
          /*
          DGP_ERR_PASS_CLEANUP(santaLayerAddLayer(pdoc, ppage,
					      i, pstrip, 
					      job_private(pjob)->order,
					      doc_private(pdoc)->memmgr,
					      dgp_state));
                          */
          }
#endif /* 0 */

  } /* end page_type == santaPageColor */

  /* We're done. set the return variables with this strip's position */
  /* on the page. */
  DGP_ERR_PASS_CLEANUP(pageStripGetCoords(pstrip, px, py, pw, ph, dgp_state));

  if(ppage->scale_me == DGP_FALSE){
      /* Nothing to do... */
  } else {
      scale_res.xRes = DgpCoord_div(DgpCoord_eqi((Dgp_Int32)(ppage->scale_factor * DgpCoord_float(ppage->res.xRes))), ppage->res.xRes);
      scale_res.yRes = DgpCoord_div(DgpCoord_eqi((Dgp_Int32)(ppage->scale_factor * DgpCoord_float(ppage->res.yRes))), ppage->res.yRes);
      rows = (Dgp_Int32)((DgpCoord_isRotated(ORIENT_RD, pstrip->pimg->orient) ? pstrip->pimg->w : pstrip->pimg->h)
          * ppage->scale_factor);
      cols = (Dgp_Int32)((DgpCoord_isRotated(ORIENT_RD, pstrip->pimg->orient) ? pstrip->pimg->h : pstrip->pimg->w)
          * ppage->scale_factor);

      switch(ppage->page_type){
      case santaPageMono:
		  /* We need memory here. */
		  if(ppage->outputType == santaInputMonoPacked){
			  DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, cols, rows, (void **)&tempRGBImage));
			  tempData = (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 0);
			  if (rows != 0){
				  tempBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 1) - 
					  (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 0);
			  }
			  else {
				  tempBpl = 0;
			  }
		  } else {
			  tempData = data;
			  tempBpl = bpl;
		  }

          /* Do a binary to RGB scale... */
          if(!DgpBlat_scaleBinaryRGBAvgSub(   pstrip->pimg->buffer.packed,
                                              pstrip->pimg->bpl,
                                              0, 0,
                                              (DgpImage_RGB *)tempData, 
                                              tempBpl, 
                                              0, 0,
                                              0, 0,
                                              cols,
                                              rows,
                                              &scale_res, 
                                              &DGP_ALLOCATORS)){
              DGP_ERR_GENF0(DGP_TRUE, cErrPanic, "Error scaling binary->RGB...");
          }

		  /* Binarize the results into passed memory. */
		  if(ppage->outputType == santaInputMonoPacked){
			  Dgp_Int8 *dst = (Dgp_Int8 *)data;
			  DGP_ERR_PASS_CLEANUP(binarizeRGBImage(dgp_state, tempRGBImage, &tempBinaryPixr, (void **)&tempBinaryImage));

			  /* HORRIBLE, BAD, ABOMINABLE, OUCH, FIXME */
			  /* Binarizer allocates its own page;  copy it into passed memory. */
			  for(i = 0; i < rows; i++, dst += bpl){
				  memcpy((void *)dst, (void *)DGPIMG_PIXPTR(tempBinaryImage, 0, i), bpl);
			  }
			  
			  DGP_ERR_PASS_CLEANUP(DgpImage_unwrap((DgpImage_Packed)tempBinaryImage, (Dgp_UInt32 **)&dataArea));
			  DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &tempBinaryPixr));
		  }	

          break;
      case santaPageColor:
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
          /* Do an RGB to RGB scale... */
          /* There's also an RGB->BGR scale available, e.g. for Windows... */

		  /* If eventually destined for a mono buffer, we need memory here. */
		  if(halftoneTheOutput){
			  DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, cols, rows, (void **)&tempRGBImage));
			  tempData = (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 0);
			  if (rows != 0){
				  tempBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 1) - 
					  (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)tempRGBImage, 0, 0);
			  }
			  else {
				  tempBpl = 0;
			  }
		  } else {
			  tempData = data;
			  tempBpl = bpl;
		  }

          if(!DgpBlat_scaleRGBRGBAvgSub(pstrip->pimg->buffer.rgb,
                                        pstrip->pimg->bpl, 
                                        0, 0,
                                        (DgpImage_RGB *)tempData,
                                        tempBpl,
                                        0, 0,
                                        0, 0,
                                        cols,
                                        rows,
                                        &scale_res,
                                        &DGP_ALLOCATORS)){
              DGP_ERR_GENF0(DGP_TRUE, cErrPanic, "Error scaling RGB->RGB...");
          }
#endif
          break;
      default:
          DGP_ERR_GENF0(DGP_TRUE, cErrInvalidParm, "Unknown page type...");
      }
  }

  if(halftoneTheOutput){
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
      rows = (Dgp_Int32)((DgpCoord_isRotated(ORIENT_RD, pstrip->pimg->orient) ? pstrip->pimg->w : pstrip->pimg->h)
          * ppage->scale_factor);
      cols = (Dgp_Int32)((DgpCoord_isRotated(ORIENT_RD, pstrip->pimg->orient) ? pstrip->pimg->h : pstrip->pimg->w)
          * ppage->scale_factor);
	  DGP_ERR_PASS_CLEANUP(DgpImage_wrap(dgp_state, (Dgp_UInt32 *)data, cols, rows, bpl / sizeof(Dgp_Int32), &binaryImage));
	  if(tempRGBImage){
		  /* Reaching here implies scaled output... */
		  halftone3(tempRGBImage, cols, rows, binaryImage);
	  } else {
		  /* Reaching here means we haven't had to scale;  halftone from strip memory. */
		  DGP_ERR_PASS_CLEANUP(DgpImage_wrapRGB(dgp_state, 
												(Dgp_UInt32 *)pstrip->pimg->buffer.rgb, 
												cols, 
												rows, 
												pstrip->pimg->bpl,
												&rgbImage));
		  halftone3(rgbImage, cols, rows, binaryImage);
		  DGP_ERR_PASS_CLEANUP(DgpImage_unwrapRGB((DgpImage_ColorRGB)rgbImage, (Dgp_UInt32 **)&dataArea));
	  }
	  DGP_ERR_PASS_CLEANUP(DgpImage_unwrap(binaryImage, (Dgp_UInt32 **)&dataArea));
#endif
}

  if(tempRGBImage) DGPIMAGE_FREE((DgpImage_ColorRGB)tempRGBImage);

  /* free the page strip. */
  /* This will also free any native-resolution memory we may have used. */
  DGP_ERR_PASS_CLEANUP(pageStripOrderReturnBand(job_private(pjob)->order,
						&pstrip,
						dgp_state));
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pstrip){
    pageStripOrderReturnBand(job_private(pjob)->order, &pstrip, dgp_state);
  }
  ERR_RETURN;
}
