#ifndef SANTA_H
#define SANTA_H

#include "santa_t.h"
#include "resolution_t.h"
#include "orientation_t.h"
#include "dgp_state.h"
#include "uds.h"

#ifdef __cplusplus
extern "C"
{
#endif

  /* santa.h
     wbell 6/8/98
     operations on core santa datatypes.
     visible to the outside world.
     reincarnated 10.11.99
  */

/* For manipulating layer options... / dhtesq. */
#define DGP_SHOWJBIG2_BITSHIFT	0
#define DGP_SHOWIMAGES_BITSHIFT	1
#define DGP_SHOWJBIG2_MASK		(1 << DGP_SHOWJBIG2_BITSHIFT)
#define DGP_SHOWIMAGES_MASK		(1 << DGP_SHOWIMAGES_BITSHIFT)
#define DGP_SHOWALL_MASK		(DGP_SHOWJBIG2_MASK | DGP_SHOWIMAGES_MASK)


#define isValidSantaAdvanceOrder(a) \
  ((a) == santaJobAdvanceSlow || (a) == santaJobAdvanceCMYKSlow || \
   (a) == santaJobAdvanceSlowCMYK)

#define isValidSantaInputType(t) \
  ((t) == santaInputMonoPacked || (t) == santaInputRGB || \
   (t) == santaInputCMYKSingle || (t) == santaInputCMYKAll)

#define isValidSantaDocOptions(o) \
  ((o) == santaDocOptsDefault || ((o) == santaDocOptsSafe))

#define isValidSantaJobOptions(o) \
  ((o) == santaJobOptsDefault || ((o) == santaJobOptsXiff3))

/*	santaOpenDocument()
	starts off the whole process of decompressing your favorite XIFF file
        creates the SantaDoc cookie that we'll carry around for the rest
	of the document decompression. 
	
	filename	-	the file to be opened.
	memlimit - how much memory can we use [bytes] (set an upper bound on
	 the amount of memory that rendering this page will ever consume.
	 note, this does not include jobs, nor the buffer memory
	 (the destination band memory) that a job needs for destination.
	opts - any document specific options you want to pass in. Else, pass 
	 in santaDocOptsDefault.
	ppdoc		-	a pointer to a SantaDocInfo *. 
	We will allocate a XiffDocInfo and initialize it and set
	this pointer to be valid. Your double-pointer can
	point to a pointer to anything
*/
ErrCode santaOpenDocument(Dgp_Int8 * filename, 
			  Dgp_Int32 memlimit, santaDocOptions opts,
			  Dgp_State * dgp_state,
			  santaDoc ** ppdoc);

/*	santaOpenDocumentEx()
	just like santaOpenDocument, but lets you use your own uds 
	(a file you've already opened up.) We will store away a copy
	of the uds structure, don't free it or nothing.
	
	docuds	        -	the opened file, with it's seek/read 
                            	functions, will be shallowly copied
				and saved.
	memlimit - how much memory can we use [bytes] (set an upper bound on
	 the amount of memory that rendering this page will ever consume.
	 note, this does not include jobs, nor the buffer memory
	 (the destination band memory) that a job needs for destination.
	opts - any document specific options you want to pass in. Else, pass 
	 in santaDocOptsDefault.
	ppdoc		-	a pointer to a santaDoc *. 
	We will allocate a santaDoc and initialize it and set
	this pointer to be valid. Your double-pointer can point to a 
	pointer to anything.
*/
ErrCode santaOpenDocumentEx(uds * docuds, Dgp_Int32 memlimit, 
			    santaDocOptions opts,
			    Dgp_State * dgp_state, 
			    santaDoc ** ppdoc);

/*	santaCloseDocument()
	cleans up all the document related info; destroys the document cookie
	and closes the opened file. 
	use for files that were opened with santaOpenDocument
	
	ppdoc		-	a double-pointer to a santaDoc that
	was created with santaOpenDocument(). Will be set to NULL.
*/
ErrCode santaCloseDocument(santaDoc ** ppdoc, Dgp_State * dgp_state);

/*	santaCloseDocumentEx()
	just like santaCloseDocument, but doesn't close the file.
	that's your job.
	
	ppdoc		-	a double-pointer to a santaDoc
	that was created with santaOpenDocumentEx(). Will be set to NULL.
*/
ErrCode santaCloseDocumentEx(santaDoc ** ppdoc, Dgp_State * dgp_state);


/*	santaOpenPage()
	creates a santaPage cookie about a particular page in 
	a previously opened XIFF file; returns interesting information
	about the page that can help one decided orientation and such
	before starting a rendering job.
	
	pdoc -a previously opened XIFF document
	 (opened with santaOpenDocument*())
	page_num - the page number to open (note pages start numbering at 0)
	pppage - a page cookie that will be created and returned;
*/
ErrCode santaOpenPage(santaDoc * pdoc, Dgp_Int32 page_num, Dgp_State * dgp_state,
		      santaPage ** pppage);


/*	santaClosePage()
	destroys a santaPage cookie about a particular page that 
	has been previously opened.
	
	pppage -  a double pointer to the previously opened
	page that will be closed. Will be set to NULL.
*/
ErrCode santaClosePage(santaDoc * pdoc, santaPage ** pppage,
		       Dgp_State * dgp_state);

/*	santaStartRenderJob()
	Starts a job to render a portion of the page at in a specified
	orientation.

	pdoc - the document in question.
	ppage -  the page to start the rendering job on.
	x,y,w,h - the coordinates of the page that you wish to
	 generate in ORIENT_RD orientation.
	o - the page is in native (ORIENT_RD) orientation, which
	 orientation do you wish to render in.
	bandsize - the job will be done in bands of this size in the slow
	 direction (they will be done full width in the fast direction)
	advorder - what criteria do we use to go from one band to the next?
	 advance down the slow direction? across the color space?
        inputtype - what kind of memory will you be passing in to the
	 band render? cmyk seperations one at a time? rgb interleaved?
	opts - what job-level options do you want. If unsure, just pass
	 in santaJobOptsDefault
	ppjob - pointer to where the render job state will be stored.
*/
ErrCode santaStartRenderJob(santaDoc * pdoc, santaPage * ppage,
			    Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    DgpCoord_Orientation o, Dgp_Int32 bandsize,
			    santaJobAdvanceOrder advorder,
			    santaInputType inputtype,
			    santaJobOptions opts,
			    Dgp_State * dgp_state,
			    santaRenderJob ** ppjob);


/*	santaEndRenderJob()
	Ends a rendering job of a portion of the page.

	pdoc - a previously opened XIFF document
	(opened with santaOpenDocument*())
	ppage -  the page to a end rendering job on.
	ppjob - the job to end, will be set to NULL.
*/
ErrCode santaEndRenderJob(santaDoc * pdoc, santaPage * ppage,
			  santaRenderJob ** pjob, Dgp_State * dgp_state);



/*      santaJobDoBand()
	Does the next band of a rendering job.
	pdoc - the SantaDoc for the job in question.
	ppage - the SantaPage for the job in question.
	pjob - the job we're talking about.
	data - where the raster data is for this band;
	 this void * will be interpreted according to the
	 rules in the jobs inputtype.
	bpl  - number of bytes per line in the source data.
	cleared - is this memory cleared? if so, we won't clear
	 the appropriate portions. (note, having us clear what we
	 need is generally better than preclearing the image.)
/* return values, can be null: */
        px, py, pw, ph - the position on the page that this band
	 corresponds to in native orientation. santaJobDoBand is called
	 on a job that has already had all it's bands rendered, but
	 has not been ended, these will be set to -1, -1, 0, 0

     jbig2_only -- if true, santa will only decode the tokenized layer(s).
	 ...replaced 4.5.2001 / dhtesq:
	 layerOptions -- contains bits set which are interpreted by the decoder
	 to ignore certain layers.

	 if this fails, the buffer pointed to by data is in some 
	 indeterminant state (some of the images may be put down, some
	 not, etc...)
*/
ErrCode santaRenderJobDoBand(santaDoc * pdoc, santaPage * ppage,
			     santaRenderJob * pjob, void * data, Dgp_Int32 bpl,
			     Dgp_Bool cleared, 
			     Dgp_Int32 * px, Dgp_Int32 * py, Dgp_Int32 * pw, Dgp_Int32 * ph,
                 Dgp_Int32 layerOptions,
			     Dgp_State * dgp_state);
#ifdef __cplusplus
}
#endif
#endif










