/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszwalb, Summer 1998
**
**  File:  halftone.c
**
**  We want to produce binary output given an rgb image....Halftone!
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  3.5.2001 / dhtesq.
**	This completely disappeared with the "new" santa.  Reinstating, so we 
**  can force mono output again.
**  
**************************************************************************/

#include "santa_halftone.h"

/* Order-3 clustered dithering matrix. */
static Dgp_Int32 cluster3[6][6] = {
  {128,156,142,114, 85,100},
  {171,242,228, 71,  0, 14},
  {185,199,213, 57, 43, 28},
  {114, 85,100,128,156,142},
  {71,   0, 14,171,242,228},
  {57,  43, 28,185,199,213} };

/*************************************************************************
**
**  Function:  void halftone3(...)
**
**
**  Converts a color image to a halftone gray one, using the "Order-3 clustered dithering matrix".
**
**	Arguments:
**      o  src is a color image...
**      o  w, h...
**      o  dst...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

void
halftone3(DgpImage_ColorRGB src, Dgp_Int32 w, Dgp_Int32 h, DgpImage_Packed dst) {
  Dgp_Int32         x, y, y_cell;
  DgpImage_RGB      *psrc;
  Dgp_Int32         value;
  
  for (y = 0; y < h; y++) {
    psrc = DGPIMG_PIXPTR(src, 0, y);
    y_cell = y % 6;
    for (x = 0; x < w; x++, psrc++) {
      /* Get gray value given r,g,b.  Cheap version. */
      value = (psrc->g >> 1) + (psrc->r >> 2) + (psrc->b >> 2);
      if (value < cluster3[y_cell][x % 6]){
          DGPIMG_SETBIT(dst, x, y);
      }
    }
  }
  return;
}
/*************************************************************************/


/*
3.5.2001 / dhtesq.
A sort-of related function to binarize an rgb image.  Santa uses this for scaling
mono documents:  first scale to RGB, then binarize.

For convenience, I give back a pixr and a dgp image, but to clean up you have to unwrap the image
and destroy the pixr, both.
*/
ErrCode
binarizeRGBImage( Dgp_State		*dgp_state, 
				  DgpImage_ColorRGB	rgbImage, 
				  PIXR			**_binaryPixr,
				  void			**_binaryImage) {
	DGP_PROLOGUE("binarizeRGBImage()", DGP_PERRSTATE);

	PIXR    			*rgbPixr = NULL, *gray = NULL;
	PIXR				*binaryPixr = NULL;
	DgpImage_Packed		binaryImage;
	Dgp_Int32			width, height;
	Dgp_Int32			longsperline;
	Dgp_UInt8			*ptr;
	IPWState    		*ipstate = dgp_state->ipwstate;

	DGP_ASSERT(_binaryImage);
	DGP_ASSERT(_binaryPixr);
	DGP_ASSERT(rgbImage);
	DGP_ASSERT(ipstate);
    
	DGP_ERR_PASS_CLEANUP(rgbimtorgbpixr(dgp_state, rgbImage, &rgbPixr));

    /* Convert to gray... */
	DGP_ERR_PASS(w_RGBToY(ipstate, rgbPixr, &gray));

	/* Binarize...note this function has an iffy reputation. */
	DGP_ERR_PASS_CLEANUP(w_binarize2(ipstate, gray, 0, SANTA_BINARIZATION_THRESHOLD, &binaryPixr));

	/*	Having used the grayscale image to get the binary one, can now dispose of the gray one. */
	DGP_ERR_PASS(w_destroyPixr(ipstate, &gray));

	/* Wrap a packed image around the ipcore one. */
	DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(ipstate, binaryPixr, &width));
	DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(ipstate, binaryPixr, &height));
	DGP_ERR_PASS_CLEANUP(w_pixrGetImage(ipstate, binaryPixr, &ptr));
	DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(ipstate, binaryPixr, &longsperline));
	longsperline /= sizeof(Dgp_Int32);

	DGP_ERR_PASS_CLEANUP(DgpImage_wrap(dgp_state, (Dgp_UInt32 *)ptr, width, height, longsperline, &binaryImage));

	*_binaryPixr = binaryPixr;
	*_binaryImage = (void *)binaryImage;

	NORMAL_RETURN;

ERR_CLEANUP:
	if(gray != NULL){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &gray));
	}
	if(binaryPixr != NULL){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &binaryPixr));
	}

	ERR_RETURN;
}

