#include "santa_image.h"
#include "zephimage.h"
#include "santa_macros.h"
#include "santa_private_t.h"
#include "rendered_image.h"
#include "memory_sentry.h"
#include "sucache.h"
#include "sutokposblock.h"
#include "santa_layer_blat_tokenized.h"
/* santa_image.c
   wbell 11.27.99
   Maintenance functions for the santaImage datatype.
*/


/* santaImageInit()
   Initializes a previously allocated santaImage, opening the desired
   zephImageInfo, and recording the allocations in the memory sentry.

   pdoc - the document for this image.
   ppage - the page for this image.
   layer_num - the layer number that this image is in. Layers start
     counting from 0.
   image_num - the image in that layer. Images also count from zero.
   pimg - the previously allocated santaImage structure that will be 
     filled in.
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode
santaImageInit(santaDoc * pdoc, santaPage * ppage, Dgp_Int32 layer_num,
	       Dgp_Int32 image_num, santaImage * pimg, memorySentry * psentry,
	       Dgp_State * dgp_state)
{
  Dgp_Bool zephImageMemAccounted = DGP_FALSE;
  DGP_STATE_PROLOGUE("santaImageInit()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  CLEAR_STRUCT(*pimg);
  
  /* aw yeah. open the zephImage. */
  DGP_ERR_PASS_CLEANUP(zephOpenImage(page_private(ppage)->zephPage,
				     layer_num, image_num, 
				     &(pimg->zephImage)));
  /* account for this memory in the memory sentry. */
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(doc_private(pdoc)->memmgr,
		   zephGetImageMemoryUsage(pimg->zephImage), 
					       dgp_state));
  zephImageMemAccounted = DGP_TRUE;
  /* all the pins are set to zero and the renderedImages are NULLed thanks */
  /* to the memset. As is all the tokenized image stuff. */

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pimg->zephImage){
    if (zephImageMemAccounted){
      memorySentryReleaseMemory(psentry,
		   zephGetImageMemoryUsage(pimg->zephImage), 
				  dgp_state);
    }
    zephCloseImage(&pimg->zephImage);
  }
  CLEAR_STRUCT(*pimg);
  ERR_RETURN;
}

/* santaImageDestroy()
   Destroys and invalidates the state in a santaImage that has been
   opened. All pins on particular portions of the image should
   have been removed.
   
   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image that is to be destroyed. Does *not* free the 
     santaImage structure.
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode
santaImageDestroy(santaDoc * pdoc, santaPage * ppage, santaImage * pimg,
		  memorySentry * psentry, Dgp_State * dgp_state)
{
  Dgp_Int32 imgsize = 0, i;
  ErrCode firstErr = cSuccessful;
  DGP_STATE_PROLOGUE("santaImageInit()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);


  /* release the memory associated with the zephImage. */
  DGP_ERR_CHECK(memorySentryReleaseMemory(psentry,
		   zephGetImageMemoryUsage(pimg->zephImage), 
					  dgp_state)){
    UPDATE_ERR(firstErr);
  }
  
  /* close the image. */
  DGP_ERR_CHECK(zephCloseImage(&(pimg->zephImage))){
    UPDATE_ERR(firstErr);
  }

  /* these all should be 0, because noone should be wanting to hold onto */
  /* this image if we're closing it. */
  DGP_ASSERT(pimg->pin_image == 0);
  DGP_ASSERT(pimg->pin_rendered == 0);
  DGP_ASSERT(pimg->pin_mask == 0);
  pimg->pin_image = 0;
  pimg->pin_rendered = 0;
  pimg->pin_mask = 0;

  /* if there's a rendered portion of the image, kill it. */
  if (pimg->prendered){
    DGP_ERR_CHECK(renderedImageSize(pimg->prendered, &imgsize,
				    dgp_state)){
      imgsize = 0;
      UPDATE_ERR(firstErr);
    }
    
    DGP_ERR_CHECK(memorySentryReleaseMemory(psentry, imgsize, dgp_state)){
      UPDATE_ERR(firstErr);
    }

    /* image data for these rendered images is made in renderedImageCreate */
    /* so it'll get freed. */
    DGP_ERR_CHECK(renderedImageDestroy(&pimg->prendered, dgp_state)){
      UPDATE_ERR(firstErr);
    }
  }

  /* same with a rendered portion of the mask. */
  if (pimg->pmask){
    DGP_ERR_CHECK(renderedImageSize(pimg->pmask, &imgsize, dgp_state)){
      imgsize = 0;
      UPDATE_ERR(firstErr);
    }
    
    DGP_ERR_CHECK(memorySentryReleaseMemory(psentry, imgsize, dgp_state)){
      UPDATE_ERR(firstErr);
    }

    /* image data for these rendered images is made in renderedImageCreate */
    /* so it'll get freed. */
    DGP_ERR_CHECK(renderedImageDestroy(&pimg->pmask, dgp_state)){
      UPDATE_ERR(firstErr);
    }
  }
  
  /* BAD BAD BAD FIXME FIXME FIXME */
  /* 30.8.2000 / dhtesq. */
  /* Memory ownership issues with libjbig2, which won't know if we've rotated the token bitmaps. */
  /* */
  /* 30.4.2001 / dhtesq. */
  /* Upstream changes??  This appears to be no longer necessary.  Freeing is done in libjbig2, as per design. */
  if((pimg->prendered) && ((pimg->prendered->orient == ORIENT_DL) || (pimg->prendered->orient == ORIENT_UR))){
      for(i = 0; i < pimg->pos_block->posCount; i++){
          if(pimg->pos_block->pos[i].sym->bits != NULL) {
			  DGP_WARNF0("Deprecated clause;  santa is trying to free tokens that belong to jbig2.");
              /*DGP_FREE((void *)pimg->pos_block->pos[i].sym->bits); */
              /*pimg->pos_block->pos[i].sym->bits = NULL; */
          }
      }
  }
  /*/////////////////////////////// */
  /* if this is a tokenized image, we need to make sure we free all the */
  /* tokenized image state. */
  DGP_ERR_CHECK(santaImagePurgeTokenized(pimg, psentry, dgp_state)){
    UPDATE_ERR(firstErr);
  }
  
  CLEAR_STRUCT(*pimg);
  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}


/* santaImageFetchTokenized()
   For tokenized images, fetches the position blocks and required 
   dictionaries so that it is prepared to be blatted.

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode
santaImageFetchTokenized(santaDoc * pdoc, santaPage * ppage,
			 santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state)
{
  Dgp_Bool fetched = DGP_FALSE;
  DGP_STATE_PROLOGUE("santaImageFetchTokenized()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  /* do we have the tokens to do this? */
  if (!pimg->tokimg_entry){
    DGP_ERR_PASS_CLEANUP(suGetTokenPositionBlockH(doc_private(pdoc)->cache,
						  ppage->page_num,
						  pimg->zephImage,
						  &(pimg->tokimg_entry)));
    fetched = DGP_TRUE;
    DGP_ERR_PASS_CLEANUP(suExtractTokenPositionBlock(pimg->tokimg_entry,
						     &pimg->pos_block));
  }
  
  NORMAL_RETURN;
ERR_CLEANUP:
  if (fetched){
    suReleaseReference(pimg->tokimg_entry);
    pimg->tokimg_entry = NULL;
    pimg->pos_block = NULL;
  }
  ERR_RETURN;
}

/* santaImageFetchColorTokenized()
   For color tokenized images, fetches the position blocks and required 
   dictionaries so that it is prepared to be blatted.

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.

  Added 16.6.2000 / dhtesq.
*/
ErrCode
santaImageFetchColorTokenized(santaDoc * pdoc, santaPage * ppage,
			 santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state)
{
  Dgp_Bool fetched = DGP_FALSE;
  ColorTokenizedImageAddition *pExt = pimg->zephImage->colorTokImgExt;

  color_tag       *color_tags;
  colorDictInfo   *color_dictinfo;          /* color dictionary is not implemented in tfx */
  color_token     *color_dict = NULL;       /* right now we keep these structures to avoid supercache error */
                                            
  DGP_STATE_PROLOGUE("santaImageFetchColorTokenized()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  /* Do we have the tokens to do this? */
  if (!pimg->tokimg_entry){
    DGP_ERR_PASS_CLEANUP(suGetTokenPositionBlockH(doc_private(pdoc)->cache,
						  ppage->page_num,
						  pimg->zephImage,
						  &(pimg->tokimg_entry)));
    fetched = DGP_TRUE;
    DGP_ERR_PASS_CLEANUP(suExtractTokenPositionBlock(pimg->tokimg_entry,
						     &pimg->pos_block));

    DGP_ERR_PASS_CLEANUPF0(suGetColorTokenPositionExtensionH(   doc_private(pdoc)->cache,
                                                                ppage->page_num,
                                                                pimg->zephImage->layerNum,
                                                                pimg->zephImage->imageNum,
                                                                pimg->zephImage,
                                                                &pExt->pColorTagEntry),
                           "santaImageFetchColorTokenized(): error getting color extension data\n"); 

    DGP_ERR_PASS_CLEANUPF0(suExtractColorTokenPositionExtension(pExt->pColorTagEntry,
                                                                &color_tags,
                                                                &color_dictinfo),
                           "santaImageFetchColorTokenized(): error reading color extension data\n");
   
  /* Copy the position block pointer into the extension?...don't seem to match. */
    pExt->pPosEntry = pimg->tokimg_entry;

  /* color dictionary is not implemented in tfx
    DGP_ERR_PASS_CLEANUPF0(suGetColorDictionaryUnion(   doc_private(pdoc)->cache,
                                                        ppage->page_num,
                                                        color_dictinfo,
                                                        &pExt->pColorDictEntry),
                           "santaImageFetchColorTokenized(): error getting color dictionary data\n");

    DGP_ERR_PASS_CLEANUPF0(suExtractColorDictionaryUnion(   pExt->pColorDictEntry,
		                                                    &color_dict,
                                                            &numImages),
                           "santaImageFetchColorTokenized(): error reading color dictionary data\n"); 
   */
    pExt->color_tags = color_tags;
    pExt->color_dict = &color_dict;
    pExt->numBoxes = pimg->pos_block->posCount;
    pExt->boxes = pimg->pos_block->pos;

    /* Sort the positions... */
    DGP_ERR_PASS_CLEANUPF0(SortPositions(   pdoc,
                                            ppage,
                                            pimg,
                                            dgp_state),
                           "santaImageFetchColorTokenized() / couldn't sort position block...\n");

  }
  

  NORMAL_RETURN;
ERR_CLEANUP:
  if (fetched){
    suReleaseReference(pimg->tokimg_entry);
    pimg->tokimg_entry = NULL;
    pimg->pos_block = NULL;
  }
  ERR_RETURN;
}


/* santaImageUpdateTokenized()
   For tokenized images, we sometimes alter the tokens. The memory consumption
   for this is all correctly calculated in the tokenized_blat_* functions
   and is reflected in the memorySentry correctly, but we do want to
   provide feedback to supercache to have it recalculate its usages. This 
   call tells supercache to recalculate things so that in low memory 
   situations we can accurately ask supercache "How much memory is this image
   using?"

   pimg - the image in question.
*/
ErrCode
santaImageUpdateTokenized(santaImage * pimg, Dgp_State * dgp_state)
{
  
  DGP_STATE_PROLOGUE("santaImageUpdateTokenized()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);

  suRecalculateUsage(pimg->tokimg_entry);
  
  NORMAL_RETURN;
}


/* santaImagePurgeTokenized()
   For tokenized images, frees the memory allocated in
   santaImageFetchTokenized and returns the image to a valid state, if the
   number of pins == 0;

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode 
santaImagePurgeTokenized(santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state)
{
  suCacheEntry entry = NULL;
  DGP_STATE_PROLOGUE("santaImagePurgeTokenized()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (pimg->pin_image == 0){
    entry = pimg->tokimg_entry;
    pimg->pos_block = NULL;
    pimg->tokimg_entry = NULL;
    pimg->prendered = NULL;
    
    if (entry){
      DGP_ERR_PASS_CLEANUP(suReleaseReference(entry));
    }
    
    /* There might be various color cache entries associated with this image.  Dereference those too. */
    /* dhtesq. */
    if(pimg->zephImage && pimg->zephImage->colorTokImgExt){
        entry = pimg->zephImage->colorTokImgExt->pPosEntry;
        if(entry){
            DGP_ERR_PASS_CLEANUP(suReleaseReference(entry));
            pimg->zephImage->colorTokImgExt->pPosEntry = NULL;
        }
        /*Color token dictionary is not implemented in tfx
        entry = pimg->zephImage->imgAddition.colorTokImgExt->pColorDictEntry;
        if(entry){
            DGP_ERR_PASS_CLEANUP(suReleaseReference(entry));
            pimg->zephImage->imgAddition.colorTokImgExt->pColorDictEntry = NULL;
        }*/
        entry = pimg->zephImage->colorTokImgExt->pColorTagEntry;
        if(entry){
            DGP_ERR_PASS_CLEANUP(suReleaseReference(entry));
            pimg->zephImage->colorTokImgExt->pColorTagEntry = NULL;
        }
    }
  }
  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}

/* santaImagePinImage()
   Pin the image data (the zephyr level and below) memory from being
   purged by the memory sentry. It does keep a reference count, so the 
   same number of pins should be unpinned.
   
   pimg - the image in question.
*/
ErrCode 
santaImagePinImage(santaImage * pimg, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("santaImagePinImage()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);

  DGP_ASSERT(pimg->pin_image < (Dgp_UInt32)-1);
  pimg->pin_image ++;

  NORMAL_RETURN;
}

/* santaImageUnpinImage()
   unpin the image data (the zephyr level and below) memory from being
   purged by the memory sentry.
   
   pimg - the image in question.
*/
ErrCode 
santaImageUnpinImage(santaImage * pimg, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("santaImageUnpinImage()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);

  DGP_ASSERT(pimg->pin_image > 0);
  pimg->pin_image --;

  NORMAL_RETURN;
}



/* santaImagePurgeSave()
   If the number of pins on the rendered and mask are 0, this function
   will remove it and put the memory back in the sentry.

   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode 
santaImagePurgeSave(santaImage * pimg, memorySentry * psentry,
                    Dgp_State * dgp_state)
{
  Dgp_Int32 sz;
  DGP_STATE_PROLOGUE("santaImagePurgeSave()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (pimg->pin_rendered == 0 && pimg->prendered){
    DGP_ERR_PASS(renderedImageSize(pimg->prendered, &sz, dgp_state));
    renderedImageDestroy(&(pimg->prendered), dgp_state);
    memorySentryReleaseMemory(psentry, sz, dgp_state);
  }
  if (pimg->pin_mask == 0 && pimg->pmask){
    DGP_ERR_PASS(renderedImageSize(pimg->pmask, &sz, dgp_state));
    renderedImageDestroy(&(pimg->pmask), dgp_state);
    memorySentryReleaseMemory(psentry, sz, dgp_state);
  }

  NORMAL_RETURN;
}
