#ifndef SANTA_IMAGE_H
#define SANTA_IMAGE_H

#include "santa_image_t.h"
#include "santa_t.h"
#include "memory_sentry_t.h"

#ifdef __cplusplus
extern "C"

#endif

/* santa_image.h
   wbell 11.27.99
   Maintenance functions for the santaImage datatype.

*/

/* santaImageInit()
   Initializes a previously allocated santaImage, opening the desired
   zephImageInfo, and recording the allocations in the memory sentry.

   pdoc - the document for this image.
   ppage - the page for this image.
   layer_num - the layer number that this image is in. Layers start
     counting from 0.
   image_num - the image in that layer. Images also count from zero.
   pimg - the previously allocated santaImage structure that will be 
     filled in.
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode
santaImageInit(santaDoc * pdoc, santaPage * ppage, Dgp_Int32 layer_num,
	       Dgp_Int32 image_num, santaImage * pimg, memorySentry * psentry,
	       Dgp_State * dgp_state);

/* santaImageDestroy()
   Destroys and invalidates the state in a santaImage that has been
   opened. All pins on particular portions of the image should
   have been removed.
   
   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image that is to be destroyed. Does *not* free the 
     santaImage structure.
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/

ErrCode
santaImageDestroy(santaDoc * pdoc, santaPage * ppage, santaImage * pimg,
		  memorySentry * psentry, Dgp_State * dgp_state);


/* santaImageFetchTokenized()
   For tokenized images, fetches the position blocks and required 
   dictionaries so that it is prepared to be blatted.

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode
santaImageFetchTokenized(santaDoc * pdoc, santaPage * ppage,
			 santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state);


/* santaImageFetchColorTokenized()
   For color tokenized images, fetches the position blocks and required 
   dictionaries so that it is prepared to be blatted.

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.

  Added 16.6.2000 / dhtesq.
*/
ErrCode
santaImageFetchColorTokenized(santaDoc * pdoc, santaPage * ppage,
			 santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state);

/* santaImageUpdateTokenized()
   For tokenized images, we sometimes alter the tokens. The memory consumption
   for this is all correctly calculated in the tokenized_blat_* functions
   and is reflected in the memorySentry correctly, but we do want to
   provide feedback to supercache to have it recalculate its usages. This 
   call tells supercache to recalculate things so that in low memory 
   situations we can accurately ask supercache "How much memory is this image
   using?"

   pimg - the image in question.
*/
ErrCode
santaImageUpdateTokenized(santaImage * pimg, Dgp_State * dgp_state);
   

/* santaImagePurgeTokenized()
   For tokenized images, frees the memory allocated in
   santaImageFetchTokenized and returns the image to a valid state, if
   the number of pins is 0.

   pdoc - the document for this image.
   ppage - the page for this image.
   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode 
santaImagePurgeTokenized(santaImage * pimg, memorySentry * psentry,
			 Dgp_State * dgp_state);

/* santaImagePurgeSave()
   If the number of pins on the rendered and mask are 0, this function
   will remove it and put the memory back in the sentry.

   pimg - the image in question
   psentry - the memory sentry that will grant us the ability to keep
     track of memory usage.
*/
ErrCode 
santaImagePurgeSave(santaImage * pimg, memorySentry * psentry,
                    Dgp_State * dgp_state);


/* santaImagePinImage()
   Pin the image data (the zephyr level and below) memory from being
   purged by the memory sentry. It does keep a reference count, so the 
   same number of pins should be unpinned.
   
   pimg - the image in question.
*/
ErrCode 
santaImagePinImage(santaImage * pimg, Dgp_State * dgp_state);

/* santaImageUnpinImage()
   unpin the image data (the zephyr level and below) memory from being
   purged by the memory sentry.
   
   pimg - the image in question.
*/
ErrCode 
santaImageUnpinImage(santaImage * pimg, Dgp_State * dgp_state);



#ifdef __cplusplus
}
#endif
#endif
