#include "santa_layer.h"
#include "santa_image.h"
#include "santa_layer_blat.h"
#include "santa_private_t.h"
#include "memory_sentry.h"
#include "rendered_image_manip.h"
#include "rendered_image.h"
#include "zephimage.h"
#include "zephpage_t.h"
#include "santa_macros.h"
/* 12.6.2000 / dhtesq.  Color support... */
#include "selector_blat.h"
/* libblat... */
#include "colorBlit.h"

/* santa_layer.h
   wbell 10/12/99
   the santa page layer datatype outside manipulation functions.
*/


/*  santaLayerInit()
    initializes a santaLayer structure for a particular layer in
    a particular page. 

    pdoc      - the document in question
    ppage     - the page in question
    layer_num - the layer which you want to be initialized for in the page
      note: layers start numbering at zero.
    player    - a previously allocated layer structure which will be filled in
    psentry   - someone to ask about allocating memory.
*/    
ErrCode 
santaLayerInit(santaDoc * pdoc, santaPage * ppage, Dgp_Int32 layer_num,
	       santaLayer * player, memorySentry * psentry,
	       Dgp_State * dgp_state)
{
  Dgp_Int32 i = 0;
  DGP_STATE_PROLOGUE("santaLayerInit()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(layer_num < 0 || layer_num >= page_private(ppage)->nlayers,
	      cErrInvalidParm);
  DGP_ERR_GEN(!player, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);

  CLEAR_STRUCT(*player);

  player->layer_num = layer_num;
  player->nimages = page_private(ppage)->zephPage->imageCounts[layer_num];

  /* let's allocate an array of santaImageInfo's which we'll open */
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, 
					       player->nimages
					       * sizeof(santaImage),
					       dgp_state));
  DGP_CHECK_ALLOC_CLEANUP(player->pimgs = DGP_CALLOC(player->nimages,
						 sizeof(santaImage)));

  for (i = 0; i < player->nimages; i++){
    DGP_ERR_PASS_CLEANUP(santaImageInit(pdoc, ppage, layer_num,
					i, &(player->pimgs[i]), psentry,
					dgp_state));
  }
  
  /* layerpin and playersave are both initialized to null by the memset. */

  NORMAL_RETURN;
 ERR_CLEANUP:
  
  if (player->pimgs){
    /* need to close all the zephImages, and then destroy the pimgs */
    for (i = 0; i < player->nimages; i++){
      santaImageDestroy(pdoc, ppage, &(player->pimgs[i]), psentry, dgp_state);
    }
    
    memorySentryReleaseMemory(psentry, player->nimages
			      * sizeof(santaImage),
			      dgp_state);
    DGP_FREE(player->pimgs);
    player->pimgs = NULL;
  }

  player->layer_num = -1;
  player->nimages = 0;
  ERR_RETURN;
}

/* santaLayerDestroy()
   destroys a santaLayer structure and cleans up related information
   created in the init function.
   
   pdoc       - the document
   ppage      - the page
   player     - the layer to clean up information for
   psentry    - the memory manager to indicate what memory we've freed.
*/
ErrCode
santaLayerDestroy(santaDoc * pdoc, santaPage * ppage, santaLayer * player,
		  memorySentry * psentry, Dgp_State * dgp_state)
{
  ErrCode firstErr = cSuccessful;
  Dgp_Int32 i = 0;
  Dgp_UInt32 imgsize = 0;
  DGP_STATE_PROLOGUE("santaLayerDestroy()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!player, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (player->pimgs){
    /* need to close all the santaImages, and then destroy the pimgs  */
    for (i = 0; i < player->nimages; i++){
      
      DGP_ERR_CHECK(santaImageDestroy(pdoc, ppage, &(player->pimgs[i]),
				      psentry, dgp_state)){
          UPDATE_ERR(firstErr); 
      }
    }
    
    DGP_ERR_CHECK(memorySentryReleaseMemory(psentry, player->nimages
					    * sizeof(santaImage),
					    dgp_state)){
      UPDATE_ERR(firstErr);
    }

    DGP_FREE(player->pimgs);
    player->pimgs = NULL;
  }

  /* and now the possible layer save. */
  if (player->playersave){
    DGP_ERR_CHECK(renderedImageSize(player->playersave, &imgsize,
				    dgp_state)){
      imgsize = 0;
      UPDATE_ERR(firstErr); 
    }
    
    DGP_ERR_CHECK(memorySentryReleaseMemory(psentry, imgsize, dgp_state)){
      UPDATE_ERR(firstErr);
    }

    /* image data for these rendered images is made in renderedImageCreate */
    /* so it'll get freed. */
    DGP_ERR_CHECK(renderedImageDestroy(&player->playersave, dgp_state)){
      UPDATE_ERR(firstErr);
    }
  }
  
  player->layer_num = -1;
  player->nimages = 0;

  DGP_ERR_PASS(firstErr);
  NORMAL_RETURN;
}



/* santaLayerAddLayer()
   put down a layer of a xiff page on a pageStrip. Note, this follows
   the xiff spec and clears / initializes the pageStrip as if this was
   a single layer being put down (non-covered areas on binary layers are
   selected, and non-covered areas in color layers are white
   or black [for background layers and foreground layers respectively] )

   pdoc       - the document
   ppage      - the page
   layer_idx   - the index of they layer that you wish put down
   pstrip     - where you wish to put it
   porder     - for memory awareness, you can look ahead to see if
                particular pieces of the page will be needed in the near
		future of the strip order.
  psentry    - how to handle memory
*/
ErrCode
santaLayerAddLayer(santaDoc * pdoc, santaPage * ppage,
		   Dgp_Int32 layer_idx, pageStrip * pstrip, 
		   pageStripOrder * porder, memorySentry * psentry,
		   Dgp_State * dgp_state)
{
  santaDocOptions opts;
  santaLayer * player = NULL;
  Dgp_Int32 i = 0;
  DGP_STATE_PROLOGUE("santaLayerAddLayer()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(layer_idx < 0 || layer_idx >= page_private(ppage)->nlayers,
	      cErrInvalidParm);
  DGP_ERR_GEN(!pstrip, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  opts = doc_private(pdoc)->opts;
  /* while putting down this layer, we need to keep in mind the */
  /* current document options; if we're in safe mode, we bail on *any* */
  /* error. If we're not, we'll try and deal with errors and continue; */
  /* *no whammy*! */

  /* do we have any portion of this layer stored around completely? */
  /* if so, we can just copy it in. */

  /* we don't have do deal with xiff3 compatibility mode, because nothing */
  /* special takes place when we're putting down just one layer. */
  player = &(page_private(ppage)->players[layer_idx]);

  /* first thing we need to do is clear out this rendered image for where */
  /* the images go; in the color case, we just clear out (set the default */
  /* color for non-covered areas, selected for the uncovered portion of mono) */
  DGP_ERR_CHECK(renderedImageClear(pstrip->pimg, 
				   player->pimgs,
				   player->nimages,
				   pstrip->cleared, 
				   layer_idx,
				   dgp_state)){
    /* something went wrong. Now some would say we should bail, because */
    /* anything that happens after here is probably going to be ugly */
    /* (real ugly in mono, only ugly where there are no images in color), */
    /* I'm going to go out on a limb, and let this go through. */
    if (opts & santaDocOptsSafe){
      DGP_ERR(DGP_CUR_ERROR);
    }
  }
		
  /* so we've cleared all we can. Now if we have a portion of the entire */
  /* layer stored that covers a portion of this pageStrip, then we'll */
  /* just want to put it down */
  if (player->playersave){
    /* ADD: we don't actually save any full layers yet, but if we do, we */
    /* want to handle this case appropriately. */
    /* make sure to lock down the layer save when you do something here, */
    /* you wouldn't want it to disappear. */
    DGP_WARNF0("santaLayerAddLayer: portion of the layer is saved, we don't know how to use it\n");
  }
  else{
    /* we just want to put down the appropriate images [0...nimages - 1] */
    for (i = 0; i < player->nimages; i++){
      DGP_ERR_CHECK(santaLayerBlatImage(pdoc, ppage, player, i, pstrip->pimg, 
					porder, psentry, NULL, NULL, dgp_state)){
	/* should we bail? */
	if (opts & santaDocOptsSafe){
	  DGP_ERR(DGP_CUR_ERROR);
	}
      }
    }
  }
  
  NORMAL_RETURN;
}


/* santaLayerAddPairedLayers()
   puts down a set of paired layers (a selector and a foreground layer)
   according to the xiff specification on the current pageStrip.

   pdoc       - the document
   ppage      - the page
   sel_idx  - the index of the selector layer of the pair (odd)
   for_idx  - the index of the foreground layer of the pair (sel_idx + 1)
   pstrip     - the destination for the blatting
   porder     - for memory awareness, you can look ahead to see if
                particular pieces of the page will be needed in the near
		future of the strip order.
   psentry    - how to handle memory
*/
ErrCode
santaLayerAddPairedLayers(santaDoc * pdoc, santaPage * ppage,
			  Dgp_Int32 sel_idx, Dgp_Int32 for_idx,
			  pageStrip * pstrip, pageStripOrder * porder,
			  memorySentry * psentry, Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("santaLayerAddPairedLayers()", dgp_state);
  renderedImage     *selector = NULL;
  santaLayer        *player_for, *player_sel;
  Dgp_Int32         i, index;
  Dgp_Bool          needDrawSelector = DGP_TRUE;

  /*DGP_WARNF0("santaLayerAddPairedLayer: flailing...\n"); */
  /* ADD: paired layers */
  /* 12.6.2000 / dhtesq. */
  /* Trying to implement... */
  
    /* Selector index should be odd... */
    DGP_ASSERT(sel_idx % 2);
    DGP_ASSERT(pstrip);
    DGP_ASSERT(ppage);
    DGP_ASSERT(dgp_state);

	/* 6.3.2001 / dhtesq. */
	/* We're passing the selector layer down into zephyr as an image mask, so this call becomes */
	/* useless.  If this transparency mask implementation works out I'll ditch it altogether. */
/*
    DGP_ERR_PASS_CLEANUP(stripDecodeSelectorLayer(sel_idx,
                                                  ppage,
                                                  pstrip,
                                                  &selector,
                                                  porder,
                                                  psentry,
                                                  pdoc,
                                                  dgp_state));*/
    /*
       So the MRC model says that wherever a selector selects an area of the
     page not covered by a foreground image, it needs to be drawn black.
     There are a lot of optimizations we can make here, but the only one 
     we'll do is if there is a foreground image that covers the whole strip,
     then we don't blat the selector in black. If there is[n't? sic?], we need to blat
     the selector in black.

     If there are no images in the selector, we don't want to put it down,
     because then we'd be blacking out the entire page [sic?  Blackout should only
     occur where selector is true.  This trap just saves time.  Also, if there are no images
     on the selector, why did we just waste 1MB allocating it? / dhtesq.].
    */


    player_sel = &page_private(ppage)->players[sel_idx];
    /* if there is no more layer above the selector we just put the selector down */
    if(!selector && page_private(ppage)->nlayers == sel_idx+1){
      for(i = 0; i < player_sel->nimages; i++)
	  {
  		DGP_ERR_PASS_CLEANUP(santaLayerBlatImage(	pdoc, 
                                                    ppage, 
                                                    player_sel,     /* selector layer first */
                                                    i, 
                                                    pstrip->pimg, 
                                                    porder,
                                                    psentry,
													selector,
                                                    &(player_sel->pimgs[i]),
			                                        dgp_state));
	  }
	NORMAL_RETURN;
    }

    
  if(!player_sel->nimages){
      needDrawSelector = DGP_FALSE;
  } 

  player_for = &page_private(ppage)->players[for_idx];

  if(page_private(ppage)->nlayers >= for_idx){
      for(i = 0; i < player_for->nimages; i++){
          zephImageInfo image;
          image = player_for->pimgs[i].zephImage;
          if(image->xPos == 0 && image->yPos == 0 && image->width == (Dgp_UInt32)pstrip->pimg->w && 
              image->height == (Dgp_UInt32)pstrip->pimg->h){
              needDrawSelector = DGP_FALSE;
          }
      }
  }

  /* The point is to black out regions of the pstrip->pimg where the selector is true. */
  /*	This call taken from BlatMaskRGB() from old santa.  / dhtesq. */
  if(selector){
      if(needDrawSelector){
          Dgp_Int32 dstX, dstY, dstMaxX, dstMaxY;
          DgpImage_RGB black;

          dstX = MAX(selector->x, pstrip->pimg->x);
          dstY = MAX(selector->y, pstrip->pimg->y);
  
          dstMaxX = MIN((selector->x + selector->w - 1), (pstrip->pimg->x + pstrip->pimg->w - 1));
          dstMaxY = MIN((selector->y + selector->h - 1), (pstrip->pimg->y + pstrip->pimg->h - 1));

          black.r = 0;  black.g = 0;    black.b = 0;

          DgpBlat_colorBlitBinaryRGB(   selector->buffer.packed,
                                        selector->bpl,
                                        selector->x,
                                        selector->y,
                                        pstrip->pimg->buffer.rgb,
                                        pstrip->pimg->bpl,
                                        pstrip->pimg->x,
                                        pstrip->pimg->y,
                                        dstX,
                                        dstY,
                                        dstMaxX - dstX + 1,
                                        dstMaxY - dstY + 1,
                                        black);
      }
  }
  
  /* if there are some selector strips without images, we need to put them down */
  if (player_sel->nimages == 1 && for_idx < page_private(ppage)->nlayers &&
	  player_for->nimages > 0 && player_sel->pimgs[0].zephImage->height > player_sel->pimgs[0].zephImage->stripHeight)
  {
	  Dgp_UInt32 y_end = 0;
	  Dgp_Bool needBlit = DGP_FALSE;
      for(i = 0; i < player_for->nimages; i++)
	  {

		if (y_end < player_for->pimgs[i].zephImage->yPos)
		{
			needBlit = DGP_TRUE;
			break;
		}
		y_end = player_for->pimgs[i].zephImage->yPos + player_for->pimgs[i].zephImage->height;
	  }

	  if (y_end < player_sel->pimgs[0].zephImage->height)
	  {
		needBlit = DGP_TRUE;
	  }
	  if (needBlit)
	  {
  		DGP_ERR_PASS_CLEANUP(santaLayerBlatImage(	pdoc, 
                                                    ppage, 
                                                    player_sel,     /* selector layer first */
                                                    0, 
                                                    pstrip->pimg, 
                                                    porder,
                                                    psentry,
													selector,
                                                    &(player_sel->pimgs[0]),
			                                        dgp_state));
	  }
  }

  /* Now we can put down all the images, if there is a layer.    */
  /* */
  /* For each image in the foreground, set the mask info to point to the selector layer. */
  /* As a first-order implementation, this will only work if the selector and image have  */
  /* the same size (with a scale factor). / dhtesq. */
  if (page_private(ppage)->nlayers >= for_idx){
	  zephImageInfo image;
	  santaLayer *theLayer;
	  index = 0;
      for(i = 0; i < player_for->nimages; i++){
          image = player_for->pimgs[i].zephImage;

		  /* if there are more selectors than images, we need to make sure we get the correct selector */
		  if (player_sel->nimages > player_for->nimages)
		  {
			  Dgp_UInt32 y_pos = player_for->pimgs[i].zephImage->yPos;
			  for (index = 0; index < player_sel->nimages; index++)
			  {
				  if (player_sel->pimgs[index].zephImage->yPos == y_pos) break;
			  }
		  }
		  else if (player_sel->nimages == player_for->nimages)
		  {
			  index = i;
		  }
			  
		  if(needDrawSelector){
			image->maskInfo = player_sel->pimgs[index].zephImage;
		  }
		  /* 14.3.2001 / dhtesq. */
		  /* Color jbig2 is recorded in tfx as a "foreground," which contains the b & w tokens, */
		  /* and a "selector," which points to the foreground position block and includes the color */
		  /* extensions.  So, if we have two jbig2 images in a row, we should just render the "selector." */
		  theLayer = (image->compType == IO_COMP_JBIG2) && (image->compType == IO_COMP_JBIG2) ? player_sel:player_for;

          DGP_ERR_PASS_CLEANUP(santaLayerBlatImage( pdoc, 
                                                    ppage, 
                                                    theLayer,
                                                    i, 
                                                    pstrip->pimg, 
                                                    porder,
                                                    psentry,
													selector,
                                                    &(player_sel->pimgs[index]),
                                                    dgp_state));

		  /* Null out that mask info.  Since, in tfx, the mask is a layer unto itself, layer cleanup will take care of it. */
		  if(needDrawSelector){
			image->maskInfo = NULL;
		  }
      }
  } 


  if(selector){
      DGP_ERR_PASS(renderedImageDestroy(&selector, dgp_state));
  }
  NORMAL_RETURN;  
ERR_CLEANUP:
  if(selector){
      DGP_ERR_PASS(renderedImageDestroy(&selector, dgp_state));
  }
  ERR_RETURN;
}



/* santaLayerPurgeSave()
   Purge as much memory as possible from the layer level (do not recurse
   into the images on the layer), meaning remove an unpinned partial
   layer image if there is one.
*/

ErrCode
santaLayerPurgeSave(santaLayer * player, memorySentry * psentry, 
                    Dgp_State * dgp_state)
{
  Dgp_Int32 sz;
  DGP_STATE_PROLOGUE("santaLayerPurgeSave()", dgp_state);
  DGP_ERR_GEN(!player, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (player->layerpin == 0 && player->playersave){
    DGP_ERR_PASS(renderedImageSize(player->playersave, &sz, dgp_state));
    renderedImageDestroy(&(player->playersave), dgp_state);
    memorySentryReleaseMemory(psentry, sz, dgp_state);
  }

  NORMAL_RETURN;
}













