#ifndef SANTA_LAYER_H
#define SANTA_LAYER_H

#include "santa_layer_t.h"
#include "dgp_state.h"
#include "page_strip_t.h"
#include "santa_t.h"
#include "memory_sentry_t.h"
#include "page_strip_order_t.h"
#ifdef __cplusplus
extern "C"
{
#endif

/* santa_layer.h
   wbell 10/12/99
   the santa page layer datatype outside manipulation functions.
   
   This is the interface that the santa layer above sees. It handles
   all memory stuff and caching. The pdoc pointer is exclusively used
   for access to the supercache for tokenized image decompression and 
   memory policy.
*/


/*  santaLayerInit()
    initializes a santaLayer structure for a particular layer in
    a particular page. 

    pdoc      - the document in question
    ppage     - the page in question
    layer_num - the layer which you want to be initialized for in the page
      note: layers start numbering at zero.
    player    - a previously allocated layer structure which will be filled in
    psentry   - someone to ask about allocating memory.
*/    
ErrCode 
santaLayerInit(santaDoc * pdoc, santaPage * ppage, Dgp_Int32 layer_num,
	       santaLayer * player, memorySentry * psentry,
	       Dgp_State * dgp_state);

/* santaLayerDestroy()
   destroys a santaLayer structure and cleans up related information
   created in the init function.
   
   pdoc       - the document
   ppage      - the page
   player     - the layer to clean up information for
   psentry    - the memory manager to indicate what memory we've freed.
*/
ErrCode
santaLayerDestroy(santaDoc * pdoc, santaPage * ppage, santaLayer * player,
		  memorySentry * psentry, Dgp_State * dgp_state);

/* santaLayerAddPairedLayers()
   puts down a set of paired layers (a selector and a foreground layer)
   according to the xiff specification on the current pageStrip.

   pdoc       - the document
   ppage      - the page
   sel_idx  - the index of the selector layer of the pair
   for_idx  - the index of the foreground layer of the pair
   pstrip     - the destination for the blatting
   porder     - for memory awareness, you can look ahead to see if
                particular pieces of the page will be needed in the near
		future of the strip order.
   psentry    - how to handle memory
*/
ErrCode
santaLayerAddPairedLayers(santaDoc * pdoc, santaPage * ppage,
			  Dgp_Int32 sel_idx, Dgp_Int32 for_idx,
			  pageStrip * pstrip, pageStripOrder * porder,
			  memorySentry * psentry, Dgp_State * dgp_state);



/* santaLayerAddLayer()
   put down a layer of a xiff page on a pageStrip. Note, this follows
   the xiff spec and clears / initializes the pageStrip as if this was
   a single layer being put down (non-covered areas on binary layers are
   selected, and non-covered areas in color layers are white
   or black [for background layers and foreground layers respectively] )

   pdoc       - the document
   ppage      - the page
   layer_idx   - the index of they layer that you wish put down
   pstrip     - where you wish to put it
   porder     - for memory awareness, you can look ahead to see if
                particular pieces of the page will be needed in the near
		future of the strip order.
   psentry    - how to handle memory
*/
ErrCode
santaLayerAddLayer(santaDoc * pdoc, santaPage * ppage,
		   Dgp_Int32 layer_idx, pageStrip * pstrip, 
		   pageStripOrder * porder, memorySentry * psentry,
		   Dgp_State * dgp_state);

/* santaLayerPurgeSave()
   Purge as much memory as possible from the layer level (do not recurse
   into the images on the layer), meaning remove an unpinned partial
   layer image if there is one.
*/

ErrCode
santaLayerPurgeSave(santaLayer * player, memorySentry * psentry, 
                    Dgp_State * dgp_state);

#ifdef __cplusplus
}
#endif
#endif



