#include "santa_layer_blat.h"
#include "santa_layer_blat_tokenized.h"
#include "santa_layer_blat_zephyr.h"
#include "santa_private_t.h"
#include "rendered_image.h"
#include "rendered_image_manip.h"

/* For debugging... / dhtesq. */
#include <libiow.pub>
#include <libpixrw.pub>
/*//////////////////////// */

/* santa_layer_blat.c
   wbell 11.6.99
   blatting an image in a santa page layer onto a rendered image.
*/


/* santaLayerBlatImage()
   blats an image in a layer on a renderedImage. Handles dispatching
   on image type / output type / scale / transparency / rotation / cropping.
   takes care of memory allocation with the memorySentry, and uses the
   pageStripOrder to make intelligent decisions what to keep (if any)
   of partial images.

   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   player - the layer to which the image belongs. uses the layer as a place
    to cache information about partially decoded images when we're all done.
   img_idx - which image in that layer you want to put down, starting
    from 0
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
   selector - plane in structure form. / dhtesq.
*/


ErrCode
santaLayerBlatImage(santaDoc * pdoc, santaPage * ppage,
		    santaLayer * player,
		    Dgp_Int32 img_idx, renderedImage * pdst, 
		    pageStripOrder * porder, memorySentry * psentry,
			renderedImage *selector,
            santaImage *selectorImage,
		    Dgp_State * dgp_state)
{
  santaImage *pimg = NULL;
  Dgp_Bool on;
  renderedImage *ri_mask = NULL;

  DGP_STATE_PROLOGUE("santaLayerBlatImage()", dgp_state);

  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!player, cErrInvalidParm);
  DGP_ERR_GEN(img_idx < 0 || img_idx >= player->nimages, cErrInvalidParm);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);

  /* so what we're doing in this function is basically dispatching on if 
     we have a transparency mask or not. If no transparency mask, we
     just pass this rendered image down the pipe and let the world take 
     its way and see if it can be rendered in this output buffer, if not,
     somewhere down the pipe a temporary buffer will be allocated and
     it'll do the copy/scale/rotate.
     
     For the case with the transparency mask, we need to figure out a good
     band size to decode them both in, and get portions of the original image
     and transparency, and do the blat.
  */
  /* a good first thing to check is if this image even intersects  */
  /* a portion of this destination rendered image. */
  pimg = &(player->pimgs[img_idx]);
  DGP_ERR_GEN(!pimg, cErrAssertFailure);

  DGP_ERR_PASS(renderedImageIsImgOn(pdst, pimg, &on, dgp_state));
  if (on){
    /* ADD: We need to handle the transparency mask case. */
    /* 4.8.2000 / dhtesq.   */
    /* Attempted implementation.  To keep memory usage */
    /* & copying down, we 1) only allocate for the part of the mask necessary for */
    /* this strip, and 2) require some sort of mask interface in our image */
    /* decoders.  The old santa used to blat image and mask into distinct PIXRs, */
    /* and then copy the result into the strip memory. */

    pdst->trans_src = NULL;
    pdst->trans_bpl = 0;
    if(pimg->zephImage->maskInfo){
        Dgp_Int32 x, y, w, h;
		Dgp_Int32 scaleFactor;

        /* For purposes of memory allocation, let's just get the part of the mask */
        /* that lies on this strip.  A note on scaling:  it's very possible that we have */
		/* a mask at page resolution masking a low-res jpeg.  Compare the resolutions */
		/* and get the mask dimensions accordingly. */
		scaleFactor = DgpCoord_floor(selectorImage->zephImage->res.xRes) / DgpCoord_floor(pimg->zephImage->res.xRes);
        x = MAX((Dgp_Int32)(pimg->zephImage->xPos), (Dgp_Int32)pdst->x);
        y = MAX((Dgp_Int32)(pimg->zephImage->yPos), (Dgp_Int32)pdst->y);
        w = MIN((Dgp_Int32)(pdst->x + pdst->w), (Dgp_Int32)((pimg->zephImage->xPos) + (pimg->zephImage->width * scaleFactor))) - x;
        h = MIN((Dgp_Int32)(pimg->zephImage->yPos + (pimg->zephImage->height * scaleFactor)), (Dgp_Int32)(pdst->y + pdst->h)) - y;

        /* Create the renderedImage{} and let it do the image allocation. */
        DGP_ERR_PASS_CLEANUP(renderedImageCreate(   x, y,
                                                    w, h,
                                                    &selectorImage->zephImage->res,
                                                    &pdst->dst_res,
                                                    pdst->orient,
                                                    0,
                                                    riMonoPacked,       /* We only accept 1-bit-depth masks. */
                                                    NULL,
                                                    &ri_mask,
                                                    dgp_state));

        DGP_ERR_PASS_CLEANUP(santaLayerBlatTransparencyMask(pdoc, 
                                                            ppage,
                                                            pimg,
                                                            ri_mask, 
                                                            porder,
                                                            psentry,
                                                            selector,
															selectorImage,
                                                            dgp_state));

#ifdef DGP_DBUG
		/*
		DGP_ERR_PASS_CLEANUP(writeOutTiff(	"c:\\testbed\\tif\\g4.tif", 
											DgpCoord_floor(selectorImage->zephImage->res.xRes), 
											DgpCoord_floor(selectorImage->zephImage->res.yRes), 
											ri_mask->buffer.packed, 
											ri_mask->bpl, 
											ri_mask->w, ri_mask->h, 
											dgp_state));*/
#endif

        pdst->trans_src = (void *)ri_mask->buffer.packed;
        pdst->trans_bpl = ri_mask->bpl;
	}

    DGP_ERR_PASS_CLEANUP(santaLayerBlatSingleImage(pdoc, ppage,
						     player, img_idx,
						     pdst, porder,
						     psentry,
                             selector,
						     dgp_state));

	if(ri_mask){
		DGP_ERR_PASS(renderedImageDestroy(&ri_mask, dgp_state));
		ri_mask = NULL;
	}
  }

  
  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}



/* santaLayerBlatSingleImage()
   blats an image in a layer on a renderedImage, ignores transparency 
   masks. Handles dispatching on input type and deciding what to cache
   while decoding.


   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   player - the layer to which the image belongs. uses the layer as a place
    to cache information about partially decoded images when we're all done.
   img_idx - which image in that layer you want to put down, starting
    from 0
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatSingleImage(santaDoc * pdoc, santaPage * ppage,
		    santaLayer * player,
		    Dgp_Int32 img_idx, renderedImage * pdst, 
		    pageStripOrder * porder, memorySentry * psentry,
            renderedImage *selector,
		    Dgp_State * dgp_state)
{
  santaImage * pimg = NULL;
  DGP_STATE_PROLOGUE("santaLayerBlatSingleImage()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!player, cErrInvalidParm);
  DGP_ERR_GEN(img_idx < 0 || img_idx >= player->nimages, cErrInvalidParm);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);

  /* first thing we check is if we have the desired region of the image */
  /* already stored in the layer, and if so, we can just blat that puppy  */
  /* onto the output image. */
  if (player->pimgs[img_idx].prendered){
    /* ADD: we don't currently handle keeping around partially decoded images */
    DGP_WARNF0("santaLayerBlatSingleImage: We do not handle using partial images that are kept around.\n");
    /* note, we could just have a portion of the image around, and then we'd */
    /* have to render possibly 4 other rectangles. Take care of that here. */
    
    /* note that when we implement this case, we don't really want it to fall */
    /* through and render the thing after all, but that's ok now. */
  }
  
  pimg = &(player->pimgs[img_idx]);
  DGP_ERR_GEN(!pimg, cErrAssertFailure);
  /* So what we do here is dispatch on the type of the zephImage, and  */
  /* let the particular pipe for each of the input types handle the */
  /* rest. */
  switch (pimg->zephImage->compType){
  case IO_COMP_TOKENIZATION:
  case IO_COMP_LLTOKENIZATION:
  case IO_COMP_JBIG2:
    DGP_ERR_PASS_CLEANUP(santaLayerBlatTokenized(pdoc, ppage,
						 pimg, pdst, 
						 porder, psentry, selector,
						 dgp_state));
    break;
  case IO_COMP_MMR: /* G4 */
  case IO_COMP_MH: /* G3_1D */
  case IO_COMP_MR: /* G3_2D */

  case IO_COMP_JPEG:
    /* Here's the general case for all image types that'll go through
       the zephyr decoding mechanism. Right now we only handle g4 & jpeg, so
       those'll be the only ones we pass down.
    */
    DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyr(pimg, pdst, porder, ppage, psentry,
                                              dgp_state));
    
    break;
    /* ADD: We need to handle decoding of other input types. */
  default:
    DGP_WARNF0("santaLayerBlatSingleImage: image input type not handled.\n");

    break;
  }

  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
  
}


/* santaLayerBlatTransparencyMask().
    Actually this just tries to render the zephImage pointed to by
    pimg->zephImage->maskInfo.  Typically it will be g4 encoded,
    but we don't care.

	(added to tfx version 5 March 2001 / dhtesq.)

   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   pimg -- this is the only one of the passed parameters that has been prepared
    specially for the mask.  The dimensions of the buffer in here fit snugly
    around the mask -- it won't have the same width as a strip.
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatTransparencyMask(santaDoc * pdoc, 
                               santaPage * ppage,
                               santaImage * pimg,
                               renderedImage * pdst, 
                               pageStripOrder * porder, 
                               memorySentry * psentry,
                               renderedImage *selector,
					           santaImage *selectorImage,
                               Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("santaLayerBlatTransparencyMask()", dgp_state);

  zephImageInfo the_mask = pimg->zephImage->maskInfo;

  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);

  DGP_ASSERT(the_mask != NULL);
  selector = selector;
  
  /* So what we do here is dispatch on the type of the zephImage, and  */
  /* let the particular pipe for each of the input types handle the */
  /* rest. */
  switch (the_mask->compType){
  case IO_COMP_TOKENIZATION:
  case IO_COMP_LLTOKENIZATION:
  case IO_COMP_JBIG2:
  case IO_COMP_JPEG:
    DGP_WARNF1("santaLayerBlatTransparencyMask:  mask compression type %d not supported.\n", the_mask->compType);
    break;
  case IO_COMP_MMR: /* G4 */
    /* Here's the general case for all image types that'll go through
       the zephyr decoding mechanism. Right now we only handle g4 & jpeg, so
       those'll be the only ones we pass down.
    */
    DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyr(selectorImage, pdst, porder, ppage, psentry,
                                              dgp_state));
    
    break;
    /* ADD: We need to handle decoding of other input types. */

  /* Case added by Eli */
  case IO_COMP_MH: /* G3_1D */
    /* Copy of the IO_COMP_MMR case */
    DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyr(selectorImage, pdst, porder, ppage, psentry, dgp_state));
    
    break;
  case IO_COMP_MR: /* G3_2D */
    /* Copy of the IO_COMP_MMR case */
    DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyr(selectorImage, pdst, porder, ppage, psentry, dgp_state));
    
    break;
  /* Case added by Eli */

  default:
    DGP_WARNF0("santaLayerBlatTransparencyMask: image input type not handled.\n");

    break;
  }

  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
  
}

#ifdef DGP_DBUG
/*static */
ErrCode writeOutTiff(	Dgp_Int8                    *outName, 
                        Dgp_Int32                   pgXRes, 
                        Dgp_Int32                   pgYRes,
                        Dgp_UInt32		            *pageIm, 
						Dgp_Int32					bpl,
						Dgp_Int32					w,
						Dgp_Int32					h,
                        Dgp_State                   *dgp_state)
{
    DGP_PROLOGUE("writeOutTiff", DGP_PERRSTATE);
    PIXR *outIm = NULL;
    TIFF *tiff = NULL;

	Dgp_Int32 y;
	UInt32 *pPixrRow = NULL;
	ScaledInt1616 xdpi, ydpi;
	Dgp_Int32 lpl = bpl / sizeof(Dgp_UInt32);
    	
	xdpi = Int32_TO_ScaledInt1616((Int32)pgXRes);
	ydpi = Int32_TO_ScaledInt1616((Int32)pgYRes);

	DGP_ERR_PASS_CLEANUPF1(w_createPixrExtended(DGP_PIPWSTATE, 
                                        w,
                                        h,
                                        1, 
                                        xdpi, 
                                        ydpi, 
                                        cFrameBits,
                                        1, 
                                        1, 
                                        NULL, 
                                        &outIm),
                     "DgpDecode:  I couldn't create the pixr for '%s'.\n", outName);

	w_pixrSetType(DGP_PIPWSTATE, outIm, cColorNone);
    	
	for(y = 0; y < h; y++){
		w_getLine(DGP_PIPWSTATE, outIm, y,(UInt8 **)&pPixrRow);
		memcpy(pPixrRow, pageIm + (y * lpl), bpl);
	}

	DGP_ERR_PASS_CLEANUPF1(w_writeEx(	DGP_PIPWSTATE, 
										outName,
										outIm, 
										TIFF_G4,
										&tiff, 
										cTrue,
										cTrue),
		"DgpDecode:  I couldn't write the pixr for '%s'.\n", outName);
      
	w_destroyPixr(DGP_PIPWSTATE, &outIm);
  NORMAL_RETURN;

ERR_CLEANUP:
  w_destroyPixr(DGP_PIPWSTATE, &outIm);
  ERR_RETURN;
}
/*************************************************************************/
#endif /* DGP_DBUG */
