#include "santa_layer_blat_tokenized.h"
#include "tokenized_blat_mono.h"
#include "tokenized_blat_color.h"
#include "page_strip_order.h"
#include "zephimage.h"
#include "santa_image.h"
#include "resolution.h"

#include "tokpos_quicksort.h"

/* santa_layer_blat_tokenized.c
   wbell 11.22.99
   blatting a black & white tokenized image onto a rendered image
   in various orientations and output types.

   dhtesq June 2000
   Color too...
*/

/* santaLayerBlatTokenized()
   takes an tokenized image, and blates it onto the given renderedImage.
   takes care of output type dispatch, scaling, and rotation.

   pdoc - where to get the tokenized info from.
   ppage - the page this image is on.
   pimg - the zephImage that we're blatting down.
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.

  16.6.2000 / dhtesq.
  Adding color support, incl. decoding color token dictionaries...
*/

ErrCode
santaLayerBlatTokenized(santaDoc * pdoc, santaPage * ppage,
			santaImage * pimg,
			renderedImage * pdst, 
			pageStripOrder * porder,
			memorySentry * psentry,
			renderedImage *selector,
			Dgp_State * dgp_state)
{
  Dgp_Bool modified = DGP_FALSE;
  Dgp_Int32 next_on = - 1;
  Dgp_Int32 x = 0, y = 0, w = 0, h = 0;
  Dgp_Bool pinned = DGP_FALSE;
  DGP_STATE_PROLOGUE("santaLayerBlatTokenized()", dgp_state);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);

  /* let's fetch the necessary positions and dictionaries. */
  /* pin the image so that nothing gets freed in the ensuing mallocs. */
  DGP_ERR_PASS_CLEANUP(santaImagePinImage(pimg, dgp_state));
  pinned = DGP_TRUE;

  /* first thing we do is switch on the output type  */
  if (pdst->type == riMonoPacked){
      /* dhtesq. */
      pimg->prendered = pdst;

    /* Fetch the tokenized data if we haven't already. */
    DGP_ERR_PASS_CLEANUP(santaImageFetchTokenized(pdoc, ppage, 
						pimg, psentry,
						dgp_state));
    if (RES_EQUAL(pdst->src_res, pimg->zephImage->res)){
      /* looks like we're actually going to blat it right here and now. */
      /* The tokenizeBlat function can reallocate tokens when it 
         needs to rotate them, which we have to notify supercache of it
         happens. So we need to know if any tokens were modified and if so, we 
         notify supercache
      */
      DGP_ERR_PASS_CLEANUP(tokenizeBlatMono(pdst->buffer.packed,
					    pdst->bpl,
					    pdst->x, pdst->y,
					    pdst->orient,
					    pdst->x, pdst->y,
					    pdst->w, pdst->h,
					    pimg->pos_block->posCount,
					    pimg->pos_block->pos,
                        psentry, &modified,
					    dgp_state));

      if (modified){
        DGP_ERR_PASS_CLEANUP(santaImageUpdateTokenized(pimg, dgp_state));
      }
    }
    else {
      /* ADD: we don't handle tokenized layers that are not at the same resolution as the destination image. */
      DGP_WARNF0("santaLayerBlatTokenized: We only currently blat tokenized images at the same resolution as the destination.\n");
    }
  }
  else {
    /* ADD: we don't currently handle tokenized images with output types other than mono packed. */
      /* 16.6.2000 / dhtesq.  Trying to implement... */
      /*DGP_WARNF0("santaLayerBlatTokenized():  you want me to blat tokens on a color image.  Flailing...\n"); */

    /* Now that we know we're going to make use of the color apparatus, allocate for the color */
    /* additions to the zephyr info. */
    /* Only retrieve & sort once though;  subsequent bands will use the same information. */
    if(!pimg->tokimg_entry){
        DGP_CHECK_ALLOC_CLEANUP(
            pimg->zephImage->colorTokImgExt = DGP_CALLOC(1, sizeof(ColorTokenizedImageAddition))
            );
    }

    /* Fetch the tokenized data if we haven't already. */
    DGP_ERR_PASS_CLEANUP(santaImageFetchColorTokenized(pdoc, ppage, 
                                                       pimg, psentry,
                                                       dgp_state));

    if (RES_EQUAL(pdst->src_res, pimg->zephImage->res)){
      DGP_ERR_PASS_CLEANUP(tokenizeBlatColor(   pdst->buffer.packed,
                                                pdst->bpl,
                                                pdst->x, pdst->y,
                                                pdst->orient,
                                                pdst->x, pdst->y,
                                                pdst->w, pdst->h,
                                                pimg,
                                                pimg->zephImage->colorTokImgExt->numBoxes,
                                                pimg->zephImage->colorTokImgExt->boxes,
                                                psentry, 
                                                &modified,
                                                selector,
                                                dgp_state));
      if (modified){
        DGP_ERR_PASS_CLEANUP(santaImageUpdateTokenized(pimg, dgp_state));
      }
    }
    else {
      /* ADD: we don't handle tokenized layers that are not at the same resolution as the destination image. */
      DGP_WARNF0("santaLayerBlatTokenized():  We only currently blat tokenized images at the same resolution as the destination (color case).\n");
    }
  }

  /* unpin the image */
  DGP_ERR_PASS_CLEANUP(santaImageUnpinImage(pimg, dgp_state));
  pinned = DGP_FALSE;

  /* now, let's see if we need this puppy again. If not, we're going to */
  /* purge it. */
  DGP_ERR_PASS_CLEANUP(zephCalcPagePosition(pimg->zephImage,
					    &x, &y, &w, &h));
  DGP_ERR_PASS_CLEANUP(pageStripOrderNextOn(porder, x, y, w, h,
					    &next_on, dgp_state));
  
  /* if it's not going to be seen again, axe it. */
  if (next_on <= 0){
      DGP_ERR_PASS_CLEANUP(santaImagePurgeTokenized(pimg, psentry,
					  dgp_state));
  }
  /* dhtesq. */
  pimg->prendered = NULL;

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pinned){
    santaImageUnpinImage(pimg, dgp_state);
  }

  /* our default failure behavour is to purge this image in case we never */
  /* use it again; it shouldn't hamper our memory usage. */
  santaImagePurgeTokenized(pimg, psentry, dgp_state);

  ERR_RETURN;
}

/*************************************************************************
**
**  Function:  ErrCode SortPositions(...)
**
**  ...
**
**	Arguments:
**      o  pdoc...
**      o  ppage...
**      o  pimg...
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Implemented liberr exception handling;  return type Dgp_Bool -> ErrCode.
**      June 2000
**      o  Imported from old santa to get color working again...
**		30.4.2001 / dhtesq
**		o  Turned off sorting, since there's no more hole-filling upstream.
**  
**/

ErrCode
SortPositions(santaDoc       *pDoc,
              santaPage      *pInfo, 
              santaImage     *pImInfo,
              Dgp_State      *dgp_state){
  DGP_PROLOGUE("SortPositions", DGP_PERRSTATE);  
  Dgp_Int32 i;
  color_tag *color_tags = pImInfo->zephImage->colorTokImgExt->color_tags;
  jb2SymbolInstance *boxes = pImInfo->zephImage->colorTokImgExt->boxes;
  Dgp_Int32 numBoxes = pImInfo->zephImage->colorTokImgExt->numBoxes;
  
  /* Unused... / dhtesq. */
  pInfo = pInfo;
  pDoc = pDoc;

  DGP_ASSERT(pDoc);
  DGP_ASSERT(pInfo);
  DGP_ASSERT(pImInfo);
  DGP_ASSERT(pImInfo->zephImage->colorTokImgExt->pPosEntry);
    
  DGP_CHECK_ALLOC_CLEANUP(pImInfo->zephImage->colorTokImgExt->sorted_positions =
                          (jb2SymbolInstance **)dgp_state->a.malloc(numBoxes * sizeof(jb2SymbolInstance *)));
  
  /* Sort the positions.
     
     We need to sort the positions, because our color CC finder has
     filled in holes for us.  A hole is a region where one CC was contained
     inside another CC; when we pull out the CC that was contained, we
     have a hole in the CC that was doing the containing.  Filling in
     such holes helps make the group-4 compression do better, but it
     also means that we need to make sure not to blat CCs with holes
     filled in overtop of CCs they contained.  So we'll use the fact that
     any CC that had a hole will be have an upper left corner that is
     less than (in terms of the measure implemented in bbGT() above --
     in brief, (a, b) is less than (c, d) iff (a < c) and (b < d))
     the upper left corner of any CC that was in the hole. We'll just
     sort our jb2SymbolInstance's so that they're in increasing order by this
     measure of size, and then we'll blat them in that order.

	30.4.2001 / dhtesq.
	Since MK turned off hole-filling upstream, there's no longer any need to sort
	tokens.  Hardly does anything for compression.
  */
  
  for(i = 0; i < numBoxes; i++){
      boxes[i].userdata = &(color_tags[i]);
      pImInfo->zephImage->colorTokImgExt->sorted_positions[i] = &boxes[i];
  }
  
  /*tokpos_ext_qsort(pImInfo->zephImage->colorTokImgExt->sorted_positions, numBoxes); */
  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}
/*************************************************************************/
