#include "santa_layer_blat_zephyr.h"
#include "rendered_image.h"
#include "rendered_image_manip.h"
#include "santa_image.h"
#include "reorient.h"
#include "zephconst.h"
#include "zephimage.h"
#include "dgp_state.h"
#include "coord.h"

static ErrCode 
santaLayerBlatZephyrRec(santaImage * pimg, 
                        renderedImage * pdst,
                        pageStripOrder * porder,
                        santaPage *ppage,
                        memorySentry * psentry,
                        Dgp_State * dgp_state);

static
ErrCode santaLayerBlatZephyrRecLowMem(santaImage * pimg, renderedImage * pdst,
                                      pageStripOrder * porder, 
                                      santaPage *ppage,
                                      memorySentry * psentry,
                                      Dgp_Int32 memreq, 
                                      Dgp_State * dgp_state);

/* Determine what portion of the image we need for this destination
   blat 
*/
static
ErrCode slbz_determine_desired_img_portion(renderedImage * pdst,
                                           santaImage * pimg,
                                           Dgp_Int32 * px, Dgp_Int32 * py,
                                           Dgp_Int32 * pw, Dgp_Int32 * ph, 
                                           Dgp_State * dgp_state)
{
  Dgp_Int32 x,y,w,h;
  DGP_STATE_PROLOGUE("slbz_determine_desired_img_portion()", dgp_state);
  
  DGP_ERR_PASS(zephCalcPagePosition(pimg->zephImage, &x, &y, &w, &h));
  w += x;
  /* This used to read 'h += h;' -- a typo I think. / dhtesq. */
  h += y;
  /*// */
  x = MAX(x, (Dgp_Int32)pdst->x);
  y = MAX(y, (Dgp_Int32)pdst->y);
  w = MIN(w, (Dgp_Int32)(pdst->x + pdst->w)) - x;
  h = MIN(h, (Dgp_Int32)(pdst->y + pdst->h)) - y;
  /* these are the *page* postion of the image region we want.
     convert it to the actual image region we want.
  */
  x -= pimg->zephImage->xPos;
  y -= pimg->zephImage->yPos;
  *px = x;
  *py = y;
  *pw = w;
  *ph = h;
  NORMAL_RETURN;
}



/* Here's how we map our internal santa renderedImage types to 
   the zephReadTypes for decoding.
*/
static 
ErrCode slbz_transform_type_to_zephrt(renderedImage * pdst,
                                      zephReadType * prtype,
                                      Dgp_State * dgp_state)
{
  zephReadType rtype = 0;
  DGP_STATE_PROLOGUE("slbz_transform_type_to_zephrt()", dgp_state);
  if (pdst->type == riMonoPacked){
    rtype = ZEPHRT_MONO_PACKED;
  }
  else if (pdst->type == riRGBInterleaved){
    rtype = ZEPHRT_RGB_INTERLEAVED;
  }
  else if (pdst->type & riCMYKSeparations){
    rtype = ZEPHRT_CMYK_8BIT_SEPARATIONS;
  }
  else {
    DGP_WARNF0("santaLayerBlatZephyrRec: unknown destination type.\n");
    DGP_ERR_GEN(0, cErrDecodeFailure);
  }
  *prtype = rtype;
  NORMAL_RETURN;
}



/* 9.8.2000 / dhtesq.
 * Our MO here is to blat the problem image full size at ORIENT_RD,
 * then rotate it into the desired orientation, and then crop it
 * to what we need.
 *
 * Shouldn't we crop first, then rotate?  Could be very advantageous for
 * small strips.
 */
static ErrCode 
santaLayerBlatZephyrOrient(santaImage * pimg, renderedImage * pdst,
                           pageStripOrder * porder,
                           santaPage *ppage,
                           memorySentry * psentry,
                           zephReadType rtype, Dgp_State * dgp_state)
{
  DgpCoord_Orientation ro;
  DgpCoord_Resolution rres;
  Dgp_UInt32 memreq;
  Dgp_Int32 x, y, w, h;
  Dgp_Int32 xc, yc, wc, hc;
  Dgp_Int32 dw, dh, dbpl;
  Dgp_Bool inplace;
  renderedImage * prot = NULL, *pprerot = NULL;
  Dgp_Int32 rotmem = 0, prerotmem = 0;
  Dgp_Bool gotrotmem = DGP_FALSE, gotprerotmem = DGP_FALSE;
  Dgp_UInt8 cantdo;
  Dgp_UInt32 before, after;
  DGP_STATE_PROLOGUE("santaLayerBlatZephyrOrient()", dgp_state);

  /* FIX: this does *way* too many copies. Fix. */
  /* since we're here because we can't do it native orientation,
     we'll try and do it ORIENT_RD
  */
  ro = ORIENT_RD;
  rres = pdst->src_res;
  DGP_ERR_PASS(memorySentryGetAvailableMemory(psentry, (Dgp_Int32 *)&memreq, 
                                              dgp_state));
  
  DGP_ERR_PASS_CLEANUP(slbz_determine_desired_img_portion(pdst, pimg,
                                                          &x, &y, &w, &h,
                                                          dgp_state));

  
  before = zephGetImageMemoryUsage(pimg->zephImage);

  DGP_ERR_PASS(zephReadImageAbilities(pimg->zephImage,
                                      x, y, &rtype, &ro,
                                      &rres, &x, &y, &w, &h,
                                      &memreq, 
                                      -x, -y,
                                      &cantdo));
  after = zephGetImageMemoryUsage(pimg->zephImage);
  DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                ((Dgp_Int32)after - (Dgp_Int32)before),
                                                dgp_state));  

  /* FIX: might band prerot for low memory conditions. */

  /* Well, let's decode into a new renderedImage. */
  /* How big is this puppy going to be? */
  
  /* Reset to page coordinates... / dhtesq. */
  x += pimg->zephImage->xPos;
  y += pimg->zephImage->yPos;

  DGP_ERR_PASS_CLEANUP(renderedImageCreateSize((Dgp_Int32)x, (Dgp_Int32)y, (Dgp_Int32)w, (Dgp_Int32)h,
                                           &rres, ro, 0, pdst->type,
                                               NULL, &prerotmem, dgp_state));
  DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, prerotmem,
                                               dgp_state));
  gotprerotmem = DGP_TRUE;
  DGP_ERR_PASS_CLEANUP(renderedImageCreate((Dgp_Int32)x, (Dgp_Int32)y, (Dgp_Int32)w, (Dgp_Int32)h,
                                           &rres, &pdst->dst_res, ro, 0, pdst->type,
                                           NULL, &pprerot, dgp_state));

  /* Here we'll fill it in. */
  DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRec(pimg, pprerot, porder, ppage, psentry,
                                               dgp_state));

  /* now we need another to rotate it into. */
  /* ADD: other buffer types for rotation. */
  switch(pdst->type){
  case riMonoPacked:
  /* FIX: we could eliminate one of the temporaries in certain cases if our
     rotation code wrote with or and not set.
  */
    DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinaryNeeded(ORIENT_RD, pdst->orient,
                                              (Dgp_Int32)w, (Dgp_Int32)h, pprerot->bpl,
                                              &dw, &dh, &dbpl, &inplace,
                                              dgp_state));
    
    if (inplace){
      /* create a shadow renderedImage that really just uses the same 
         memory as the prerot, but in a different orientation.
         it's unlikely that this'll fail, because it shouldn't alloc
         any real memory.
      */
      DGP_ERR_PASS_CLEANUP(renderedImageCreateSize((Dgp_Int32)x, (Dgp_Int32)y, 
                                                   (Dgp_Int32)w, (Dgp_Int32)h,
                                                   &rres, pdst->orient,
                                                   pprerot->bpl,
                                                   pdst->type, 
                                                   pprerot->buffer.packed,
                                                   &rotmem, dgp_state));
      DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, rotmem,
                                                   dgp_state));
      gotrotmem = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(renderedImageCreate((Dgp_Int32)x, (Dgp_Int32)y, (Dgp_Int32)w, (Dgp_Int32)h,
                                               &rres, &pdst->dst_res, pdst->orient,
                                               pprerot->bpl,
                                               pdst->type, 
                                               pprerot->buffer.packed, &prot,
                                               dgp_state));


      DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinary(ORIENT_RD, pprerot->buffer.packed,
                                          (Dgp_Int32)w, (Dgp_Int32)h, pprerot->bpl,
                                          pdst->orient, pprerot->buffer.packed,
                                          dw, dh, prot->bpl,
                                          dgp_state));
    }
    else {
      /* FIX: eventually we'd prolly want to be able to band this if 
         we don't have enough memory.
      */
      DGP_ERR_PASS_CLEANUP(renderedImageCreateSize((Dgp_Int32)x, (Dgp_Int32)y,
                                                   (Dgp_Int32)w, (Dgp_Int32)h,
                                                   &rres, pdst->orient, dbpl,
                                                   pdst->type, NULL, &rotmem,
                                                   dgp_state));
      DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, rotmem,
                                                   dgp_state));
      gotrotmem = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(renderedImageCreate((Dgp_Int32)x, (Dgp_Int32)y, (Dgp_Int32)w, (Dgp_Int32)h,
                                               &rres, &pdst->dst_res, pdst->orient, dbpl,
                                               pdst->type, NULL, &prot, 
                                               dgp_state));

      DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinary(ORIENT_RD, pprerot->buffer.packed,
                                          (Dgp_Int32)w, (Dgp_Int32)h, pprerot->bpl,
                                          pdst->orient, prot->buffer.packed,
                                          dw, dh, prot->bpl, dgp_state)); 
    }
    break;
  default:
    DGP_WARNF0("santaLayerBlatZephyrOrient: only binary packed orientation changes are currently supported.\n");
    DGP_ERR_GEN_CLEANUP(0, cErrDecodeFailure);
  }

  /* now do the crop into the dst */
  /* transform back into page coords. */
  /* Already done... / dhtesq. */
  /*x += pimg->zephImage->xPos; */
  /*y += pimg->zephImage->yPos; */
  
  xc = MAX(pdst->x, x);
  yc = MAX(pdst->y, y);
  wc = MIN(pdst->x + pdst->w, x + w) - xc;
  hc = MIN(pdst->y + pdst->h, y + h) - yc;
  DGP_ERR_PASS_CLEANUP(renderedImageCrop(pdst, xc, yc, prot,
                                         xc,
                                         yc,
                                         wc, hc, 
                                         ppage,
                                         dgp_state));

  DGP_ERR_PASS_CLEANUP(memorySentryReleaseMemory(psentry, prerotmem,
                                                 dgp_state));
  gotprerotmem = DGP_FALSE;
  DGP_ERR_PASS_CLEANUP(renderedImageDestroy(&pprerot, dgp_state));
  
  DGP_ERR_PASS_CLEANUP(memorySentryReleaseMemory(psentry, rotmem, dgp_state));
  gotrotmem = DGP_FALSE;
  DGP_ERR_PASS_CLEANUP(renderedImageDestroy(&prot, dgp_state));
  

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (gotprerotmem){
    memorySentryReleaseMemory(psentry, prerotmem, dgp_state);
  }
  if (pprerot){
    renderedImageDestroy(&pprerot, dgp_state);
  }
  if (gotrotmem){
    memorySentryReleaseMemory(psentry, rotmem, dgp_state);
  }
  if (prot){
    renderedImageDestroy(&prot, dgp_state);
  }
  ERR_RETURN;
}



static ErrCode 
santaLayerBlatZephyrRec(santaImage * pimg, 
                        renderedImage * pdst,
                        pageStripOrder * porder,
                        santaPage *ppage,
                        memorySentry * psentry,
                        Dgp_State * dgp_state)
{
  zephReadType rtype;
  zephReadState rs;
  DgpCoord_Orientation ro;
  DgpCoord_Resolution rres;

  renderedImage * pcrop = NULL;
  Dgp_Int32 cropmem = 0;
  Dgp_Bool gotcropmem = DGP_FALSE;

  Dgp_Int32 x,y,w,h;
  Dgp_UInt32 before, after;
  Dgp_UInt32 memreq;
  Dgp_UInt8 cantdo;

  /* this is definitely an internal function. You call this with bad
     params, we cut off your legs.
  */
  DGP_STATE_PROLOGUE("santaLayerBlatZephyrRec()", dgp_state);

  /* ok, figure out what portion of the image we want / need to put in the
     given destination. Also map our destination type to the zephyr type.
     Figure out how much memory we have to give to rendering temp space.
     If these fail, game over man. No coming back from here. 
  */
  ro = pdst->orient;
  rres = pdst->src_res;
  DGP_ERR_PASS(slbz_transform_type_to_zephrt(pdst, &rtype, dgp_state));
  DGP_ERR_PASS(slbz_determine_desired_img_portion(pdst, pimg,
                                                  &x, &y, &w, &h, dgp_state));
  DGP_ERR_PASS(memorySentryGetAvailableMemory(psentry, (Dgp_Int32 *)&memreq, 
                                              dgp_state));
  
  /* we want to keep track of the zephImage memory before and after.
   */
  before = zephGetImageMemoryUsage(pimg->zephImage);
  
  
  /* ADD: the zeph layer is quite vague about how this will work for
     CMYK single bands. We'll handle that bridge when we get there.
  */

  /* We're at the mercy of the decompressor here. What can you do? */
  DGP_ERR_PASS(zephReadImageAbilities(pimg->zephImage,
                                      pdst->x, pdst->y, &rtype, &ro,
                                      &rres, &x, &y, &w, &h, &memreq, 
                                      pimg->zephImage->xPos,
                                      pimg->zephImage->yPos,
                                      &cantdo));
  /*first hope; can we do this natively? */
  if (cantdo == 0){
    /* heck yes. Native baby. */
    DGP_ERR_PASS_CLEANUP(zephReadImageStart(pimg->zephImage,
                                            pdst->type == riCMYKSeparations ? (void *)pdst->buffer.cmyk:(void *)pdst->buffer.packed,
                                            pdst->bpl, 
                                            pdst->trans_src,
                                            pdst->trans_bpl,
                                            pdst->x, pdst->y, 
                                            rtype, ro, &rres, x, y, w, h,
                                            &memreq, 
                                            pimg->zephImage->xPos,
                                            pimg->zephImage->yPos,
                                            &rs));

    DGP_ERR_CHECK(zephReadImageEnd(pimg->zephImage, &rs, &memreq, &rres, ppage->width, ppage->height, porder->fastWidth,
        porder->slowWidth)){
      /* if it's an out of memory error, we can try to get some more
         memory and retry the render. */
      if (DGP_CUR_ERROR == cErrInsufficientMem){ 
        DGP_CLEAR_CUR_ERROR;
        after = zephGetImageMemoryUsage(pimg->zephImage);
        DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                    ((Dgp_Int32)after - (Dgp_Int32)before),
                                                      dgp_state));
        DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRecLowMem(pimg, pdst,
                                                           porder,
                                                           ppage,
                                                           psentry,
                                                           memreq,
                                                           dgp_state));
        before = zephGetImageMemoryUsage(pimg->zephImage);
      }
      else {
        /* any other errors, we bail on. */
        DGP_ERR_GEN(0, DGP_CUR_ERROR);
      }
    }
  }
  /* can't do it natively. */
  else {
    /* now we figure out how we have to do this. */
   
    /* ADD: any special cases we want to add such as scale and rotate
       or whatnot, you can add here, and the right thing will happen.
    */

    /* right now, we're just going to assume each of these transformations
       is a copy. but you can easily see how adding more conditionals here
       will get you the right thing.
    */
    if (cantdo & ZEPH_CANT_DO_TYPE){
      DGP_WARNF0("santaLayerBlatZephyr: We only currently blat zeph images into the same destination type as the source.\n");
      /* assert will do what we want. Alert in debug, but just draw the image */
      /* blank in release. */
      DGP_ASSERT(0);
    }
    else if (cantdo & ZEPH_CANT_DO_RESOLUTION){
      DGP_WARNF0("santaLayerBlatZephyr: We only currently blat zeph images at the same resolution.\n");
      DGP_ASSERT(0);
    }
    else if (cantdo & ZEPH_CANT_DO_ORIENT){
      /* We'll fold the orientation as well as the rectangle in one.
         try again, with the native orientation.

         Since we're making a recursive call, we want to update the amount of 
         memory that we're using in the memorySentry for the image.
      */
      after = zephGetImageMemoryUsage(pimg->zephImage);
      DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                    ((Dgp_Int32)after - (Dgp_Int32)before),
                                                    dgp_state));
      DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrOrient(pimg, pdst,
                                                      porder, ppage,
                                                      psentry,
                                                      rtype, dgp_state));
      before = zephGetImageMemoryUsage(pimg->zephImage);
    }
    /* if it's just a cropping issue, we'll have to do the copy. */
    else if (cantdo == ZEPH_CANT_DO_RECT){
      Dgp_Int32 xc, yc, wc, hc;
      Dgp_Bool lowmem = DGP_FALSE;

      /* Reset to page coordinates... / dhtesq. */
      x += pimg->zephImage->xPos;
      y += pimg->zephImage->yPos;

      /* make a new renderedImage. How big would it be? */
      DGP_ERR_PASS_CLEANUP(renderedImageCreateSize((Dgp_Int32)x, (Dgp_Int32)y,
                                                   (Dgp_Int32)w, (Dgp_Int32)h,
                                                   &rres, ro, 0, pdst->type,
                                                   NULL, &cropmem, dgp_state));
      DGP_ERR_CHECK(memorySentryClaimMemory(psentry,
                                            cropmem,
                                            dgp_state))
        {
          if (DGP_CUR_ERROR == cErrInsufficientMem){ 
            DGP_CLEAR_CUR_ERROR;
            lowmem = DGP_TRUE;
          }
          else {
            /* any other errors, we bail on. */
            DGP_ERR_GEN(0, DGP_CUR_ERROR);
          }
        }
      
      if (lowmem){
        after = zephGetImageMemoryUsage(pimg->zephImage);
        DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                    ((Dgp_Int32)after - (Dgp_Int32)before),
                                                      dgp_state));
        DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRecLowMem(pimg, pdst,
                                                           porder, 
                                                           ppage,
                                                           psentry,
                                                           0,
                                                           dgp_state));
        before = zephGetImageMemoryUsage(pimg->zephImage);
      }
      else {
        gotcropmem = DGP_TRUE;
        DGP_ERR_PASS_CLEANUP(renderedImageCreate((Dgp_Int32)x, (Dgp_Int32)y, 
                                                 (Dgp_Int32)w, (Dgp_Int32)h,
                                                 &rres, &pdst->dst_res, ro, 0, pdst->type,
                                                 NULL, &pcrop, dgp_state));
       
       
        /* this call'll fill it in. */
        after = zephGetImageMemoryUsage(pimg->zephImage);
        DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry,
                                                    ((Dgp_Int32)after - (Dgp_Int32)before),
                                                      dgp_state));
        DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRec(pimg, 
                                                     pcrop,
                                                     porder, 
                                                     ppage,
                                                     psentry, 
                                                     dgp_state));
        before = zephGetImageMemoryUsage(pimg->zephImage);
       
        /* now copy it over with the crop. */
        /* transform back into page coords. */
        /* ...already done / dhtesq. */
        /*x += pimg->zephImage->xPos; */
        /*y += pimg->zephImage->yPos; */

        xc = MAX(pdst->x, x);
        yc = MAX(pdst->y, y);
        wc = MIN(pdst->x + pdst->w, x + w) - xc;
        hc = MIN(pdst->y + pdst->h, y + h) - yc;
       
        DGP_ERR_PASS_CLEANUP(renderedImageCrop(pdst, 
                                               xc, yc,
                                               pcrop,
                                               xc,
                                               yc,
                                               wc, hc,
                                               ppage,
                                               dgp_state));
        gotcropmem = DGP_FALSE;
        memorySentryReleaseMemory(psentry, cropmem, dgp_state);
        cropmem = 0;
        renderedImageDestroy(&pcrop, dgp_state);
      }
    }
    else if (cantdo == ZEPH_CANT_DO_MEMORY){
      after = zephGetImageMemoryUsage(pimg->zephImage);
      DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                    ((Dgp_Int32)after - (Dgp_Int32)before),
                                                    dgp_state));
      DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRecLowMem(pimg, pdst,
                                                         porder, 
                                                         ppage,
                                                         psentry,
                                                         memreq,
                                                         dgp_state));
      before = zephGetImageMemoryUsage(pimg->zephImage);
    }
    else {
      DGP_WARNF0("santaLayerBlatZephyr: Unknown rendering constraint.\n");
      DGP_ASSERT(0);
    }
  }

  after = zephGetImageMemoryUsage(pimg->zephImage);
  DGP_ERR_PASS_CLEANUP(memorySentryUpdateMemory(psentry, 
                                                ((Dgp_Int32)after - (Dgp_Int32)before),
                                                dgp_state));
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (gotcropmem){
      memorySentryReleaseMemory(psentry, cropmem, dgp_state);
  }
  if (pcrop){
    renderedImageDestroy(&pcrop, dgp_state);
  }

  after = zephGetImageMemoryUsage(pimg->zephImage);
  memorySentryUpdateMemory(psentry, ((Dgp_Int32)after - (Dgp_Int32)before), dgp_state);

  ERR_RETURN;                         
}


/* santaLayerBlatZephyr()
   takes an zephyr image, and blats it onto the given renderedImage.
   takes care of output type dispatch, scaling, and rotation.

   pimg - the zephImage that we're blatting down.
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatZephyr(santaImage * pimg,
                     renderedImage * pdst, 
                     pageStripOrder * porder,
                     santaPage *ppage,
                     memorySentry * psentry,
                     Dgp_State * dgp_state)
{
  Dgp_Bool pinned = DGP_FALSE;

  DGP_STATE_PROLOGUE("santaLayerBlatZephyr()", dgp_state);
  DGP_ERR_GEN(!pimg, cErrInvalidParm);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!porder, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  /* pin the image so that the memory sentry won't throw anything related
     to it away in a low memory situation.
  */
  DGP_ERR_PASS(santaImagePinImage(pimg, dgp_state));
  pinned = DGP_TRUE;
  DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRec(pimg, pdst,
                                               porder, ppage, psentry,
                                               dgp_state));

  pinned = DGP_FALSE;
  DGP_ERR_PASS_CLEANUP(santaImageUnpinImage(pimg, dgp_state));
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pinned){
    santaImageUnpinImage(pimg, dgp_state);
  }
  ERR_RETURN;
}


static
ErrCode santaLayerBlatZephyrRecLowMem(santaImage * pimg, renderedImage * pdst,
                                      pageStripOrder * porder, 
                                      santaPage *ppage,
                                      memorySentry * psentry,
                                      Dgp_Int32 memreq, 
                                      Dgp_State * dgp_state)
{
  Dgp_Bool lowmem = DGP_FALSE;
  DGP_STATE_PROLOGUE("santaLayerBlatZephyr()", dgp_state);
  /* if memreq is 0, a claim failed and we just want to try and band
     this baby. If it's non-zero, we can try and claim it as temporary space
     and see if we can make this puppy work.
  */
  if (memreq != 0){
    /* just make it available. */
    DGP_ERR_CHECK(memorySentryClaimMemory(psentry, memreq, dgp_state))
      {
        if (DGP_CUR_ERROR == cErrInsufficientMem){ 
          DGP_CLEAR_CUR_ERROR;
          lowmem = DGP_TRUE;
          /* ok, we're really in low memory, call ourselves recursively. */
          DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRecLowMem(pimg, pdst,
                                                             porder, 
                                                             ppage,
                                                             psentry,
                                                             0, dgp_state));
        }
        else {
          /* any other errors, we bail on. */
          DGP_ERR_GEN(0, DGP_CUR_ERROR);
        }
      }
    if (!lowmem){
      DGP_ERR_PASS_CLEANUP(memorySentryReleaseMemory(psentry, memreq,
                                                     dgp_state));
      /* retry the render. */
      DGP_ERR_PASS_CLEANUP(santaLayerBlatZephyrRec(pimg, pdst, porder, ppage, psentry,
                                                   dgp_state));
    }
  }
  else {
    /* FIX: we want to do some kind of banding probably in the fast direction.
       It actually won't help us now with the black and white setup because 
       the g4 stuff has really bad granularity of what we can render, and
       in the case where memory is going to be an issue (rotated orientations)
       we need 2 temporary buffers to get it done. That coupled with the
       fact that our rotation code sets and does not or, makes life difficult.
    */
       
  }

  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}
