/* FIX: remove */
#include "memory_sentry_t.h"
#include "santa_private_t.h"

#include "santa_page.h"
#include "santa.h"
#include "orientation.h"
#include "santa_macros.h"
#include "rawimage.h"
#include "image.h"
#include "image_defs.h"
/* santa_page.c 
   easier primitives for generating a page.
   10.21.99
 */

/* santaRenderInStrips()
   Given an open document and an open page, render this portion of the
   page in this buffer in this orientation, given the entire destination
   buffer, but do it in bandsize strips to save decompression memory.

   pdoc - the document for ppage
   ppage - the page in question
   o - what orientation do you want to render this in. You pass in
    a pointer to data that represents the point in the rendering area that
    is closest to the origin of the full page in that orientation.
   x, y, w, h - what region of the page (in native coordinates) do you want
    to render.
   opts - any job specific options. If unsure, pass santaJobOptsDefault
   inputtype - what is the inputtype of this data (how is the dst pointer
    to be interpreted?)
   dstBpl - what is the number of bytes per line in the source?
   dst - the destination pointer.
   cleared - is this memory cleared.

  17.7.2000 / dhtesq.
   out -- the image structure;  typically will be wrapped around dst.  Need this for DgpImage_() fns.
   isStripwise -- i.e., has the user opted for true decompression in bands.  If so, we'll output each
    complete strip to stdout.  Currently only pnm is supported.

  4.5.2001 / dhtesq.
   layerOptions -- contains bit settings which direct the decoder to ignore certain layers
   when rendering.  Replaces 'jbig2_only' flag.

   santaInputCMYKSingle is not supported for this inputtype, because that
   would imply that there is only 1 pointer in dst. We don't currently
   support doing one separation of a CMYK page without doing the others.
*/
ErrCode santaRenderInStrips(santaDoc * pdoc,
			    santaPage * ppage,
			    DgpCoord_Orientation o, Dgp_Int32 bandsize,
			    Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    santaJobOptions opts,
			    santaInputType inputtype,
			    Dgp_Int32 dstBpl, void * dst,
			    void *out,
			    Dgp_Bool cleared,
			    Dgp_State * dgp_state,
			    FILE **outgoing,
			    Dgp_Bool force_stripwise,
			    Dgp_Int32 layerOptions)
{
  ErrCode firstErr = cSuccessful;
  santaRenderJob * pjob = NULL;
  void * curdata;
  /* Needed for the cmyk stuff. */
  Dgp_Int8 * cmyk[4];
  void *cmyk_img[4];
  Dgp_Bool is_stripwise = DGP_FALSE;
  santaInputType tellDecoder = santaInputMonoPacked;

  Dgp_Int32 cw, ch, i, band_index = 0;

  DGP_STATE_PROLOGUE("santaRenderInStrips()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!pdoc, cErrInvalidParm);
  DGP_ERR_GEN(!ppage, cErrInvalidParm);
  DGP_ERR_GEN(!ISVALIDORIENT(o), cErrInvalidParm);
  DGP_ERR_GEN(bandsize <= 0, cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaJobOptions(opts), cErrInvalidParm);
  DGP_ERR_GEN(!isValidSantaInputType(inputtype), cErrInvalidParm);
  DGP_ERR_GEN(!dstBpl, cErrInvalidParm);
  DGP_ERR_GEN(!dst, cErrInvalidParm);

  force_stripwise = force_stripwise;

  /* ADD: we're not going to handle doing a single CMYK strip here. It's */
  /* the santa layer only lets you do all 4 one at a time as of now. */
  /* we could expand it, but see no need now. */
  DGP_ERR_GEN(inputtype == santaInputCMYKSingle, cErrInvalidParm);

  /* Since we can't blat color stuff into a mono buffer, be careful what inputtype you give this function... */
  ppage->outputType = inputtype;
  if(ppage->page_type == santaInputRGB){
	  /* Color page.  Decoder should blat into a color buffer in all cases. */
	  tellDecoder = santaInputRGB;
  } else if(ppage->page_type == santaInputMonoPacked){
	  /* Mono page.  If forcing, go to color -- unless scaling. */
	  if(ppage->scale_me){
		  tellDecoder = santaInputMonoPacked;
	  }
	  else if(ppage->outputType == santaInputRGB){
		  tellDecoder = santaInputRGB;
	  } 
  }
  DGP_ERR_PASS_CLEANUP(santaStartRenderJob(pdoc, ppage, x, y, w, h, o,
					   bandsize, 
					   (inputtype == santaInputCMYKAll ? santaJobAdvanceSlowCMYK : santaJobAdvanceSlow), 
					   tellDecoder,
					   opts,
					   dgp_state, &pjob));

  if (inputtype == santaInputCMYKAll){
    for (i = 0; i < 4; i++){
      cmyk[i] = ((Dgp_Int8**)dst)[i];
    }
    curdata = (void *)cmyk;
  }
  else {
    curdata = (void *)dst;
  }

  do{

    DGP_ERR_CHECK(santaRenderJobDoBand(pdoc, ppage, pjob, curdata, dstBpl,
				       cleared, NULL, NULL, &cw, &ch,
                       layerOptions,
				       dgp_state)){
      UPDATE_ERR(firstErr);
      break;
    }
    
    /* Now advance the curdata pointer. */
    if (inputtype == santaInputCMYKAll){
      for (i = 0; i < 4; i++){
          if(((Dgp_Int8 **)curdata)[i]){
              ((Dgp_Int8 **)curdata)[i] = ((Dgp_Int8 **)curdata)[i] + (bandsize * dstBpl);
          }
      }
    }
    else {
      curdata = (void *)((Dgp_Int8 *)curdata + ((Dgp_Int32)(bandsize * ppage->scale_factor) * dstBpl));
    }

    /* If the user has passed in good outgoing file pointers, that's our directive to write out */
    /* stripwise pnms. */
    /* Special case first band. */
    if((outgoing[0] != NULL) && (cw != 0 && ch != 0) && (inputtype != santaInputCMYKAll)){
        if(band_index == 0){
            /* Fake out the header...these will get restored in the next clause. */
            DGPIMG_WIDTH((DgpImage_Packed)out) = DgpCoord_isRotated(o, ORIENT_RD) ? ppage->height:ppage->width;
            DGPIMG_HEIGHT((DgpImage_Packed)out) = DgpCoord_isRotated(o, ORIENT_RD) ? ppage->width:ppage->height;

            fflush(stdout);

            /* Just write the pnm header... */
            DGP_ERR_PASS_CLEANUP(DgpImage_pSaveHeaderF(out, outgoing[0]));

            /* Restore... */
            DGPIMG_WIDTH((DgpImage_Packed)out) = DgpCoord_isRotated(o, ORIENT_RD) ? ch:cw;
            DGPIMG_HEIGHT((DgpImage_Packed)out) = DgpCoord_isRotated(o, ORIENT_RD) ? cw:ch;
        }

        /* Write out the data block; force it to be bigendian. */
        DGP_ERR_PASS_CLEANUP(DgpImage_saveRawF_(out, 
                                                DGPIMG_TYPE((DgpImage_Packed)out),
                                                outgoing[0],
                                                DGPIMAGE_BIGENDIAN,
                                                dgp_state));
    } else if((outgoing[0] != NULL) && (cw != 0 && ch != 0) && (inputtype == santaInputCMYKAll)){
        DGP_CHECK_NULL_CLEANUP(outgoing[1]);
        DGP_CHECK_NULL_CLEANUP(outgoing[2]);
        DGP_CHECK_NULL_CLEANUP(outgoing[3]);

        for(i = 0; i < 4; i++){
            if(band_index == 0){
                /* Fake out the header...these will get restored in the next clause. */
                DGPIMG_WIDTH((DgpImage_Gray)cmyk_img[i]) = DgpCoord_isRotated(o, ORIENT_RD) ? ppage->height:ppage->width;
                DGPIMG_HEIGHT((DgpImage_Gray)cmyk_img[i]) = DgpCoord_isRotated(o, ORIENT_RD) ? ppage->width:ppage->height;
                
                fflush(stdout);
                
                /* Just write the pnm header... */
                DGP_ERR_PASS_CLEANUP(DgpImage_pSaveHeaderF(cmyk_img[i], outgoing[i]));                                
            }
            /* Restore... */
            DGPIMG_WIDTH((DgpImage_Gray)cmyk_img[i]) = DgpCoord_isRotated(o, ORIENT_RD) ? ch:cw;
            DGPIMG_HEIGHT((DgpImage_Gray)cmyk_img[i]) = DgpCoord_isRotated(o, ORIENT_RD) ? cw:ch;
            
            /* Write out the data block; force it to be bigendian. */
            DGP_ERR_PASS_CLEANUP(DgpImage_saveRawF_(cmyk_img[i], 
                                                    DGPIMG_TYPE((DgpImage_Gray)cmyk_img[i]),
                                                    outgoing[i],
                                                    DGPIMAGE_BIGENDIAN,
                                                    dgp_state));
        }
    }

    /* Recycle the band memory... */
    if(is_stripwise){
        if((cw != 0 && ch != 0) && (inputtype != santaInputCMYKAll)){
            memset(dst, 0, dstBpl * DGPIMG_HEIGHT((DgpImage_Packed)out));
            curdata = (void *)dst;
        }
        else if((cw != 0 && ch != 0) && (inputtype == santaInputCMYKAll)){
            for(i = 0; i < 4; i++){
                memset(((Dgp_Int8**)dst)[i], 0, dstBpl * DGPIMG_HEIGHT((DgpImage_Gray)cmyk_img[i]));
                cmyk[i] = ((Dgp_Int8**)dst)[i];
            }
            curdata = (void *)cmyk;
        }
    }
    band_index++;
  } while(cw != 0 && ch != 0);

  DGP_ERR_PASS_CLEANUP(santaEndRenderJob(pdoc, ppage, &pjob, dgp_state));
  
  DGP_ERR_PASS_CLEANUP(firstErr);
  NORMAL_RETURN;
ERR_CLEANUP:
  if (pjob){
    santaEndRenderJob(pdoc, ppage, &pjob, dgp_state);
  }
  ERR_RETURN;
}
    
