#include "tokenized_blat_mono.h"
#include "memory_sentry.h"
#include "util.h"
#include "blit.h"
#include "blit_mono.h"
#include "reorient.h"
#include "orientation.h"
#include "g4.h"

#include <math.h>
/* tokenized_blat_mono.c
   wbell 11.29.99
   blatting a set of tokens down into a mono packed image; this level
   no longer deals with data structures, just a simple representation
   of the output buffer and the tokens.
*/


/* tokenizeBlatDecompressG4()
   Decompresses a g4 token for usage when blatting.
*/
/* FIX: it'd be nice if these functions could clear the memory with the sentry 
   before they allocate it...
*/

ErrCode
tokenizeBlatDecompressG4(jb2Symbol * ptok, Dgp_Int32 * mem_diff, 
                         Dgp_State *dgp_state)
{
  Dgp_UInt32 *pbits = NULL;
  Dgp_UInt16 lpl;
  Dgp_UInt32 uses = 0;
  DGP_STATE_PROLOGUE("tokenizeBlatDecompressG4()", dgp_state);
  DGP_ASSERT(ptok);
  DGP_ASSERT(mem_diff);

  /* how big is this puppy gonna be? decide and allocate the memory */
  /* (note, it must be set to zero, so we calloc) */
  lpl = (Dgp_UInt16)ceil(((float)ptok->width / DGP_BITS_UINT32));
  DGP_CHECK_ALLOC(pbits = (Dgp_UInt32 *)DGP_CALLOC(lpl*ptok->height,
			       sizeof(Dgp_Int32)));

  DGP_ERR_PASS_CLEANUP(g4Decode(pbits, ptok->width, 
                                 ptok->height, lpl,0, ptok->width, 
                                 ptok->height,
                                 (Dgp_UInt8 *)ptok->bits,
                                 uses, DGP_FALSE, NULL, dgp_state));
  DGP_FREE(ptok->bits);
  ptok->bits = pbits;
  ptok->longsperline = lpl;
  *mem_diff = (lpl * ptok->height) - uses;
  NORMAL_RETURN;
 ERR_CLEANUP:
  if (pbits){
    DGP_FREE(pbits);
  }
  ERR_RETURN;
}

/* tokenizeBlatRotateToken()
   Does the actual rotation of a token's bits into the desired orientation,
   and lets you know how much more/less memory the new token bits use
   inplace of the old ones.
*/
ErrCode
tokenizeBlatRotateBinaryToken(jb2Symbol * ptok, DgpCoord_Orientation o, Dgp_Int32 * mem_diff,
                        Dgp_State * dgp_state)
{
  Dgp_Int32 dw, dh, dbpl, sw, sh;
  Dgp_UInt32 * dst = NULL;
  Dgp_Bool inplace;
  Dgp_Bool rotated;
  DGP_STATE_PROLOGUE("tokenizeBlatRotateBinaryToken()", dgp_state);
  DGP_ASSERT(ptok);
  DGP_ASSERT(mem_diff);
  /* this will only fire it the token is already in the right orientation. */
  DGP_ASSERT(ptok->userflags != o);

  rotated = DgpCoord_isRotated(ptok->userflags, ORIENT_RD);
  sw = (rotated ? ptok->height : ptok->width);
  sh = (rotated ? ptok->width : ptok->height);

  DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinaryNeeded(ptok->userflags, o, 
                                            sw, sh, (ptok->longsperline 
                                                     * sizeof(Dgp_UInt32)),
                                            &dw, &dh, &dbpl, &inplace,
                                            dgp_state));

  if (inplace){
    /* heck yes...we can do this inplace. Just pass in src as dst. */
    DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinary(ptok->userflags, ptok->bits,
                                        sw, sh,  (ptok->longsperline 
                                                     * sizeof(Dgp_UInt32)),
                                        o, ptok->bits, dw, dh, dbpl,
                                        dgp_state));
    *mem_diff = 0;
  }
  else {
    /* we need to malloc a new token. */
    DGP_CHECK_ALLOC_CLEANUP(dst = DGP_MALLOC(dh * dbpl));
    /* let's do the rotation baby. */
    DGP_ERR_PASS_CLEANUP(DgpRot_reorientBinary(ptok->userflags, ptok->bits,
                                        sw, sh, (ptok->longsperline 
                                                     * sizeof(Dgp_UInt32)),
                                        o, dst, dw, dh, dbpl,
                                        dgp_state));

    *mem_diff = ((dh * dbpl) - (Dgp_Int32)ptok->userdata);
    ptok->userdata = (void *)(dh * dbpl);

    DGP_FREE(ptok->bits);

    ptok->bits = dst;
    ptok->longsperline = (dbpl / sizeof(Dgp_UInt32));
  }

  ptok->userflags = o;

  NORMAL_RETURN;
 ERR_CLEANUP:
  if (dst){
    DGP_FREE(dst);
  }
  ERR_RETURN;
}

/* tokenizeBlatMono()
     given some jb2SymbolInstance and an output buffer at some orientation, blats
     these tokens onto the output buffer with no scaling, cropped to 
     a region inside the buffer. jb2SymbolInstance are considered to be in the
     page coordinate space, and will be cropped as such.

     pdst - a pointer to (x,y) in the destination buffer.
     bpl - number of bytes per line in one row of the dest buffer
     x, y - in the page coordinate space, what point does pdst represent
     o - what orientation is this buffer in?
     cropx, cropy, cropw, croph - in the page coordinate space, 
      what rectangle inside the destination do you want to blat with
      tokens (gives you the ability to crop)
     nboxes - how many tokens do we want to blat?
     pboxes - an array of jb2SymbolInstance, where the required dictionaries have already
      been fetched. Does not assume tokens are in the right orientation;
      will rotate to suit. 
     psentry - a memory sentry object to keep track of changes in usage
      of memory.
     pmod - did we modify any tokens?
*/
ErrCode
tokenizeBlatMono(Dgp_UInt32 * pdst, Dgp_Int32 bpl, Dgp_Int32 x, Dgp_Int32 y, DgpCoord_Orientation o, 
		 Dgp_Int32 cropx, Dgp_Int32 cropy, Dgp_Int32 cropw, Dgp_Int32 croph,
		 Dgp_Int32 nboxes, jb2SymbolInstance * pboxes, memorySentry * psentry,
                 Dgp_Bool * pmod, Dgp_State * dgp_state)
{
  Dgp_Int32 toknum;
  jb2Symbol * cur_tok;
  jb2SymbolInstance * cur_pos;
  Dgp_Int32 curxl, curxh, curyl, curyh;
  Dgp_Int32 cur_min;
  Dgp_Bool mod = DGP_FALSE;
  Dgp_UInt32 * cur_src = NULL;
  Dgp_UInt32 * cur_dst = NULL;
  Dgp_UInt32 curlpl, dstlpl;
  Dgp_UInt8 cur_clip;
  
  Dgp_Int32 mem_limit = 0;
  Dgp_Int32 mem_chg = 0;
  Dgp_Int32 mem_diff = 0;

    /* crop rectangle */
  Dgp_Int32 cxl, cxh, cyl, cyh;

  DGP_STATE_PROLOGUE("tokenizeBlatMono()", dgp_state);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!bpl, cErrInvalidParm);
  DGP_ERR_GEN(!cropw < 0, cErrInvalidParm);
  DGP_ERR_GEN(!croph < 0, cErrInvalidParm);
  DGP_ERR_GEN(nboxes != 0 && !pboxes, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (cropw == 0 || croph == 0 || nboxes == 0){
    NORMAL_RETURN;
  }

  /* find out the position of the rectangle in absolute inclusive */
  /* coordinates. */
  cxl = cropx;
  cyl = cropy;
  cxh = cropx + cropw - 1;
  cyh = cropy + croph - 1;

  DGP_ASSERT((bpl % sizeof(Dgp_UInt32)) == 0);
  dstlpl = bpl / sizeof(Dgp_UInt32);

  /* get our memory limit, we won't check with the supercache each time we need */
  /* some memory, we'll just do it in batches, and only talk to the cache when */
  /* the batch gets bigger than this limit, or we're done. */
  DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                      &mem_limit,
                                                      dgp_state));



  /* I know this is real ugly, but the cost of doing the switch
     on each token is horrible, so we duplicate the loop and the initial
     checks. Make sure to change all of them if you change any of the
     boilerplate code.
  */
  switch(o){
  case ORIENT_RD:
    for ( toknum = 0 ; toknum < nboxes ; toknum++){
      cur_pos = pboxes + toknum;
      cur_tok = cur_pos->sym;
      
      DGP_ASSERT(cur_tok);
      /* first thing to check is if this is in the crop rectangle. */
      if(!((cur_pos->xpos <= cxh)
           && (cur_pos->ypos <= cyh)
           && (cxl < (cur_pos->xpos + cur_tok->width))
           && (cyl < (cur_pos->ypos + cur_tok->height)))){
        continue;
      }

      /* It's possible that this is a g4 token; which means that it's a single
         use token thats left compressed for us to uncompress. We're going
         to uncompress it, and leave it around. The rationale behind is that
         if it's truly a single use token, it'll be in the per page dictionary
         and will be purged once we leave this page.
      */
      if (cur_tok->longsperline == 0){
        DGP_ERR_CHECK(tokenizeBlatDecompressG4(cur_tok, &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }
        
        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
            
          }
          mod = DGP_TRUE;
        }
      }
      /* rotate the token if we need to. We store the current orientation in */
      /* userflags. So if anyone mucks with userflags,  */
      /* I'll be extremely unhappy. */
      if (cur_tok->userflags != o){
        DGP_ERR_CHECK(tokenizeBlatRotateBinaryToken(cur_tok, o,
                                                    &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }

        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
       
          }
          mod = DGP_TRUE;
        }
      }

      cur_clip = 0;
      cur_min = 0;
      cur_src = cur_tok->bits;
      curlpl = cur_tok->longsperline;
      curyl = cur_pos->ypos;
      curyh = curyl + cur_tok->height - 1;
      curxl = cur_pos->xpos;
      curxh = curxl + cur_tok->width - 1;

      /* ORIENT_RD */
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        cur_src += (curlpl * (cyl - curyl));
        curyl = cyl;
      }
      if (curyh > cyh){
        curyh = cyh;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        cur_min = cxl - curxl;
        curxl = cxl;
        cur_clip = 1;
      }
      if (curxh > cxh){
        curxh = cxh;
        cur_clip = 1;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curxh - curxl) + 1, (curyh - curyl) + 1,
                       curxl - x, curyl - y, dgp_state);
      }
      else{
        cur_dst = pdst + ((curyl - y) * dstlpl) + ((curxl - x) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (curxl - x) % 32,
                          bpl, cur_src, 
                          (curxh - curxl) + 1, (curyh - curyl) + 1);
      }
    }
    break;
  case ORIENT_UR:

    for ( toknum = 0 ; toknum < nboxes ; toknum++){
      cur_pos = pboxes + toknum;
      cur_tok = cur_pos->sym;

      DGP_ASSERT(cur_tok);
      /* first thing to check is if this is in the crop rectangle. */
      if(!((cur_pos->xpos <= cxh)
           && (cur_pos->ypos <= cyh)
           && (cxl < (cur_pos->xpos + cur_tok->width))
           && (cyl < (cur_pos->ypos + cur_tok->height)))){
        continue;
      }


      /* It's possible that this is a g4 token; which means that it's a single
         use token thats left compressed for us to uncompress. We're going
         to uncompress it, and leave it around. The rationale behind is that
         if it's truly a single use token, it'll be in the per page dictionary
         and will be purged once we leave this page.
      */
      if (cur_tok->longsperline == 0){
        DGP_ERR_CHECK(tokenizeBlatDecompressG4(cur_tok, &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }
        
        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
            
          }
          mod = DGP_TRUE;
        }
      }

      /* rotate the token if we need to. We store the current orientation in */
      /* userflags. So if anyone mucks with userflags, */
      /* I'll be extremely unhappy. */
      if (cur_tok->userflags != o){
        DGP_ERR_CHECK(tokenizeBlatRotateBinaryToken(cur_tok, o,
                                                    &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }

        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
          }
          mod = DGP_TRUE;
        }
      }

      cur_clip = 0;
      cur_min = 0;
      cur_src = cur_tok->bits;
      curlpl = cur_tok->longsperline;
      curyl = cur_pos->ypos;
      curyh = curyl + cur_tok->height - 1;
      curxl = cur_pos->xpos;
      curxh = curxl + cur_tok->width - 1;

      /* ORIENT_UR */
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        curyl = cyl;
        cur_clip = 1;
      }
      if (curyh > cyh){
        cur_min = curyh - cyh;
        curyh = cyh;
        cur_clip = 1;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        cur_src += (curlpl * (cxl - curxl));
        curxl = cxl;
      }
      if (curxh > cxh){
        curxh = cxh;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curyh - curyl) + 1, (curxh - curxl) + 1,
                       (cyh - curyh), curxl - x, dgp_state);
      }
      else{
        cur_dst = pdst + ((curxl - x) * dstlpl) + ((cyh - curyh) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (cyh - curyh) % 32,
                          bpl, cur_src, 
                          (curyh - curyl) + 1, (curxh - curxl) + 1);
      }

    }
    break;
  case ORIENT_RU:
    DGP_ERR_GEN(DGP_TRUE, cErrNoImpl);
    break;
  case ORIENT_LU:
    for ( toknum = 0 ; toknum < nboxes ; toknum++){
      cur_pos = pboxes + toknum;
      cur_tok = cur_pos->sym;

      DGP_ASSERT(cur_tok);
      /* first thing to check is if this is in the crop rectangle. */
      if(!((cur_pos->xpos <= cxh)
           && (cur_pos->ypos <= cyh)
           && (cxl < (cur_pos->xpos + cur_tok->width))
           && (cyl < (cur_pos->ypos + cur_tok->height)))){
        continue;
      }

      /* It's possible that this is a g4 token; which means that it's a single
         use token thats left compressed for us to uncompress. We're going
         to uncompress it, and leave it around. The rationale behind is that
         if it's truly a single use token, it'll be in the per page dictionary
         and will be purged once we leave this page.
      */
      if (cur_tok->longsperline == 0){
        DGP_ERR_CHECK(tokenizeBlatDecompressG4(cur_tok, &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }
        
        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
            
          }
          mod = DGP_TRUE;
        }
      }

      /* rotate the token if we need to. We store the current orientation in */
      /* userflags. So if anyone mucks with userflags,  */
      /* I'll be extremely unhappy. */
      if (cur_tok->userflags != o){
        DGP_ERR_CHECK(tokenizeBlatRotateBinaryToken(cur_tok, o,
                                                    &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }

        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
       
          }
          mod = DGP_TRUE;
        }
      }


      cur_clip = 0;
      cur_min = 0;
      cur_src = cur_tok->bits;
      curlpl = cur_tok->longsperline;
      curyl = cur_pos->ypos;
      curyh = curyl + cur_tok->height - 1;
      curxl = cur_pos->xpos;
      curxh = curxl + cur_tok->width - 1;
      /* ORIENT_LU */
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        curyl = cyl;
      }
      if (curyh > cyh){
        cur_src += (curlpl * (curyh - cyh));
        curyh = cyh;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        curxl = cxl;
        cur_clip = 1;
      }
      if (curxh > cxh){
        cur_min = curxh - cxh;
        curxh = cxh;
        cur_clip = 1;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curxh - curxl) + 1, (curyh - curyl) + 1,
                       cxh - curxh, cyh - curyh, dgp_state);
      }
      else{
        cur_dst = pdst + ((cyh - curyh) * dstlpl) + ((cxh - curxh) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (cxh - curxh) % 32,
                          bpl, cur_src, 
                          (curxh - curxl) + 1, (curyh - curyl) + 1);
      }
    }
    break;
  case ORIENT_DL:
    for ( toknum = 0 ; toknum < nboxes ; toknum++){
      cur_pos = pboxes + toknum;
      cur_tok = cur_pos->sym;

      DGP_ASSERT(cur_tok);
      /* first thing to check is if this is in the crop rectangle. */
      if(!((cur_pos->xpos <= cxh)
           && (cur_pos->ypos <= cyh)
           && (cxl < (cur_pos->xpos + cur_tok->width))
           && (cyl < (cur_pos->ypos + cur_tok->height)))){
        continue;
      }

      /* It's possible that this is a g4 token; which means that it's a single
         use token thats left compressed for us to uncompress. We're going
         to uncompress it, and leave it around. The rationale behind is that
         if it's truly a single use token, it'll be in the per page dictionary
         and will be purged once we leave this page.
      */
      if (cur_tok->longsperline == 0){
        DGP_ERR_CHECK(tokenizeBlatDecompressG4(cur_tok, &mem_diff, dgp_state)){
          /* just print something nasty if fails. */
          continue;
        }
        
        /* it now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* so this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
            
          }
          mod = DGP_TRUE;
        }
      }

      /* Rotate the token if we need to.  We store the current orientation in */
      /* userflags.  So if anyone mucks with userflags,  */
      /* I'll be extremely unhappy. */
      if (cur_tok->userflags != o){
        DGP_ERR_CHECK(tokenizeBlatRotateBinaryToken(cur_tok, o,
                                                    &mem_diff, dgp_state)){
          /* Just print something nasty if fails. */
          continue;
        }

        /* It now uses a different amount of memory. Keep update our counter. */
        if (mem_diff){
          mem_chg += mem_diff;
          /* So this might be a little sloppy, it could go a little over our */
          /* memory limit, but we've decided that's ok (we've already malloced */
          /* the memory anyways.) */
          if (mem_chg >= mem_limit){
            DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                         dgp_state));
            DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                                &mem_limit,
                                                                dgp_state));
            mem_chg = 0;
          }
          mod = DGP_TRUE;
        }
      }

      cur_clip = 0;
      cur_min = 0;
      cur_src = cur_tok->bits;
      curlpl = cur_tok->longsperline;
      curyl = cur_pos->ypos;
      curyh = curyl + cur_tok->height - 1;
      curxl = cur_pos->xpos;
      curxh = curxl + cur_tok->width - 1;

      /* ORIENT_DL */
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        cur_min = cyl - curyl;
        curyl = cyl;
        cur_clip = 1;
      }
      if (curyh > cyh){
        curyh = cyh;
        cur_clip = 1;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        curxl = cxl;
      }
      if (curxh > cxh){
        cur_src += (curlpl * (curxh - cxh));
        curxh = cxh;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);

      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curyh - curyl) + 1, (curxh - curxl) + 1,
                       curyl - cyl, (cxh - curxh),  dgp_state);
      }
      else{
        cur_dst = pdst + ((cxh - curxh) * dstlpl) + ((curyl - cyl) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (curyl - cyl) % 32,
                          bpl, cur_src, 
                          (curyh - curyl) + 1, (curxh - curxl) + 1);
      }

    }
    break;
  case ORIENT_DR:
    DGP_ERR_GEN(DGP_TRUE, cErrNoImpl);
    break;
  case ORIENT_UL:
    DGP_ERR_GEN(DGP_TRUE, cErrNoImpl);
    break;
  case ORIENT_LD:
    DGP_ERR_GEN(DGP_TRUE, cErrNoImpl);
    break;
  default:
    DGP_ERR_GEN(DGP_TRUE, cErrDecodeFailure);
    break;
  }
  
  if (mem_chg){
    if (mem_chg > 0){
      cur_min = mem_chg;
      mem_chg = 0;
      DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, cur_min, 
                                                   dgp_state));
    }
    else { /* mem_chg < 0 */
      cur_min = -mem_chg;
      mem_chg = 0;
      DGP_ERR_PASS_CLEANUP(memorySentryReleaseMemory(psentry, cur_min, 
                                                     dgp_state));
    }
  }
  
  if (pmod){
    *pmod = mod;
  }
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  
  if (mem_chg){
    if (mem_chg > 0){
      cur_min = mem_chg;
      mem_chg = 0;
      memorySentryClaimMemory(psentry, cur_min, dgp_state);
    }
    else { /* mem_chg < 0 */
      cur_min = -mem_chg;
      mem_chg = 0;
      memorySentryReleaseMemory(psentry, cur_min, dgp_state);
    }
  }

  if (pmod){
    *pmod = mod;
  }

  ERR_RETURN;
}


#if 0
/* tokenizeBlatMono()
     given some jb2SymbolInstance and an output buffer at some orientation, blats
     these tokens onto the output buffer with no scaling, cropped to 
     a region inside the buffer. jb2SymbolInstance are considered to be in the
     page coordinate space, and will be cropped as such.

     pdst - a pointer to (x,y) in the destination buffer.
     bpl - number of bytes per line in one row of the dest buffer
     x, y - in the page coordinate space, what point does pdst represent
     o - what orientation is this buffer in?
     cropx, cropy, cropw, croph - in the page coordinate space, 
      what rectangle inside the destination do you want to blat with
      tokens (gives you the ability to crop)
     nboxes - how many tokens do we want to blat?
     pboxes - an array of jb2SymbolInstance, where the required dictionaries have already
      been fetched. Does not assume tokens are in the right orientation;
      will rotate to suit. 
     psentry - a memory sentry object to keep track of changes in usage
      of memory.
     pmod - did we modify any tokens?
*/
ErrCode
tokenizeBlatMono(Dgp_UInt32 * pdst, Dgp_Int32 bpl, Dgp_Int32 x, Dgp_Int32 y, DgpCoord_Orientation o, 
		 Dgp_Int32 cropx, Dgp_Int32 cropy, Dgp_Int32 cropw, Dgp_Int32 croph,
		 Dgp_Int32 nboxes, jb2SymbolInstance * pboxes, memorySentry * psentry,
                 Dgp_Bool * pmod, Dgp_State * dgp_state)
{
  Dgp_Int32 toknum;
  token * cur_tok;
  jb2SymbolInstance * cur_pos;
  Dgp_Int32 curxl, curxh, curyl, curyh;
  Dgp_Int32 cur_min;
  Dgp_Bool mod = DGP_FALSE;
  Dgp_UInt32 * cur_src = NULL;
  Dgp_UInt32 * cur_dst = NULL;
  Dgp_UInt32 curlpl, dstlpl;
  Dgp_UInt8 cur_clip;
  
  Dgp_Int32 mem_limit = 0;
  Dgp_Int32 mem_chg = 0;
  Dgp_Int32 mem_diff = 0;

    /* crop rectangle */
  Dgp_Int32 cxl, cxh, cyl, cyh;

  DGP_STATE_PROLOGUE("tokenizeBlatMono()", dgp_state);
  DGP_ERR_GEN(!pdst, cErrInvalidParm);
  DGP_ERR_GEN(!bpl, cErrInvalidParm);
  DGP_ERR_GEN(!cropw < 0, cErrInvalidParm);
  DGP_ERR_GEN(!croph < 0, cErrInvalidParm);
  DGP_ERR_GEN(nboxes != 0 && !pboxes, cErrInvalidParm);
  DGP_ERR_GEN(!psentry, cErrInvalidParm);

  if (cropw == 0 || croph == 0 || nboxes == 0){
    NORMAL_RETURN;
  }

  /* find out the position of the rectangle in absolute inclusive */
  /* coordinates. */
  cxl = cropx;
  cyl = cropy;
  cxh = cropx + cropw - 1;
  cyh = cropy + croph - 1;

  DGP_ASSERT((bpl % sizeof(Dgp_UInt32)) == 0);
  dstlpl = bpl / sizeof(Dgp_UInt32);

  /* get our memory limit, we won't check with the supercache each time we need */
  /* some memory, we'll just do it in batches, and only talk to the cache when */
  /* the batch gets bigger than this limit, or we're done. */
  DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                      &mem_limit,
                                                      dgp_state));




  for ( toknum = 0 ; toknum < nboxes ; toknum++){
    cur_pos = pboxes + toknum;
    cur_tok = cur_pos->tok;

    DGP_ASSERT(cur_tok);
    /* first thing to check is if this is in the crop rectangle. */
    if(!((cur_pos->xpos <= cxh)
	 && (cur_pos->ypos <= cyh)
	 && (cxl < (cur_pos->xpos + cur_tok->width))
	 && (cyl < (cur_pos->ypos + cur_tok->height)))){
      continue;
    }

    /* It's possible that this is a g4 token; which means that it's a single
       use token thats left compressed for us to uncompress. We're going
       to uncompress it, and leave it around. The rationale behind is that
       if it's truly a single use token, it'll be in the per page dictionary
       and will be purged once we leave this page.
    */
    if (cur_tok->longsperline == 0){
      DGP_ERR_CHECK(tokenizeBlatDecompressG4(cur_tok, &mem_diff, dgp_state)){
        /* just print something nasty if fails. */
        continue;
      }
      
      /* it now uses a different amount of memory. Keep update our counter. */
      if (mem_diff){
        mem_chg += mem_diff;
        /* so this might be a little sloppy, it could go a little over our */
        /* memory limit, but we've decided that's ok (we've already malloced */
        /* the memory anyways.) */
        if (mem_chg >= mem_limit){
          DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                       dgp_state));
          DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                              &mem_limit,
                                                              dgp_state));
          mem_chg = 0;
          
        }
        mod = DGP_TRUE;
      }
    }
    
    
    /* rotate the token if we need to. We store the current orientation in */
    /* userflags. So if anyone mucks with userflags, I'll be extremely unhappy. */
    if (cur_tok->userflags != o){
      DGP_ERR_CHECK(tokenizeBlatRotateBinaryToken(cur_tok, o,
                                                  &mem_diff, dgp_state)){
        /* just print something nasty if fails. */
        continue;
      }

      /* it now uses a different amount of memory. Keep update our counter. */
      if (mem_diff){
        mem_chg += mem_diff;
        /* so this might be a little sloppy, it could go a little over our */
        /* memory limit, but we've decided that's ok (we've already malloced */
        /* the memory anyways.) */
        if (mem_chg >= mem_limit){
          DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, mem_chg, 
                                                       dgp_state));
          DGP_ERR_PASS_CLEANUP(memorySentryGetAvailableMemory(psentry,
                                                              &mem_limit,
                                                              dgp_state));
          mem_chg = 0;
       
        }
        mod = DGP_TRUE;
      }
    }


    cur_clip = 0;
    cur_min = 0;
    cur_src = cur_tok->bits;
    curlpl = cur_tok->longsperline;
    curyl = cur_pos->ypos;
    curyh = curyl + cur_tok->height - 1;
    curxl = cur_pos->xpos;
    curxh = curxl + cur_tok->width - 1;


    switch (o){
    case ORIENT_RD:
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        cur_src += (curlpl * (cyl - curyl));
        curyl = cyl;
      }
      if (curyh > cyh){
        curyh = cyh;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        cur_min = cxl - curxl;
        curxl = cxl;
        cur_clip = 1;
      }
      if (curxh > cxh){
        curxh = cxh;
        cur_clip = 1;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curxh - curxl) + 1, (curyh - curyl) + 1,
                       curxl - x, curyl - y,  dgp_state);
      }
      else{
        cur_dst = pdst + ((curyl - y) * dstlpl) + ((curxl - x) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (curxl - x) % 32,
                          bpl, cur_src, 
                          (curxh - curxl) + 1, (curyh - curyl) + 1);
      }
      break;
    case ORIENT_UR:
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        curyl = cyl;
        cur_clip = 1;
      }
      if (curyh > cyh){
        cur_min = curyh - cyh;
        curyh = cyh;
        cur_clip = 1;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        cur_src += (curlpl * (cxl - curxl));
        curxl = cxl;
      }
      if (curxh > cxh){
        curxh = cxh;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curyh - curyl) + 1, (curxh - curxl) + 1,
                       (cyh - curyh), curxl - x,  dgp_state);
      }
      else{
        cur_dst = pdst + ((curxl - x) * dstlpl) + ((cyh - curyh) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (cyh - curyh) % 32,
                          bpl, cur_src, 
                          (curyh - curyl) + 1, (curxh - curxl) + 1);
      }
      break;
    case ORIENT_LU:
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        curyl = cyl;
      }
      if (curyh > cyh){
        cur_src += (curlpl * (curyh - cyh));
        curyh = cyh;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        curxl = cxl;
        cur_clip = 1;
      }
      if (curxh > cxh){
        cur_min = curxh - cxh;
        curxh = cxh;
        cur_clip = 1;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);
      
      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curxh - curxl) + 1, (curyh - curyl) + 1,
                       cxh - curxh, cyh - curyh,  dgp_state);
      }
      else{
        cur_dst = pdst + ((cyh - curyh) * dstlpl) + ((cxh - curxh) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (cxh - curxh) % 32,
                          bpl, cur_src, 
                          (curxh - curxl) + 1, (curyh - curyl) + 1);
      }
      break;
    case ORIENT_DL:
      /* do we need to crop the top or bottom off this token? */
      if (curyl < cyl){
        cur_min = cyl - curyl;
        curyl = cyl;
        cur_clip = 1;
      }
      if (curyh > cyh){
        curyh = cyh;
        cur_clip = 1;
      }
      /* we should have some rows, or we have an issue. */
      DGP_ASSERT((curyh - curyl) >= 0);
      
      /* do we need to crop the left or right off this token? */
      if (curxl < cxl){
        curxl = cxl;
      }
      if (curxh > cxh){
        cur_src += (curlpl * (curxh - cxh));
        curxh = cxh;
      }
      /* we should have some columns, or we have an issue. */
      DGP_ASSERT((curxh - curxl) >= 0);

      if (cur_clip){
        DgpBlat_blitBitsClipLR(pdst, dstlpl,
                       cur_src, curlpl,
                       cur_min, (curyh - curyl) + 1, (curxh - curxl) + 1,
                       curyl - cyl, (cxh - curxh),  dgp_state);
      }
      else{
        cur_dst = pdst + ((cxh - curxh) * dstlpl) + ((curyl - cyl) / 32);
        DgpBlat_blitBinaryOrExact(cur_dst,  (curyl - cyl) % 32,
                          bpl, cur_src, 
                          (curyh - curyl) + 1, (curxh - curxl) + 1);
      }
      break;
    }
  }

  if (mem_chg){
    if (mem_chg > 0){
      cur_min = mem_chg;
      mem_chg = 0;
      DGP_ERR_PASS_CLEANUP(memorySentryClaimMemory(psentry, cur_min, 
                                                   dgp_state));
    }
    else { /* mem_chg < 0 */
      cur_min = -mem_chg;
      mem_chg = 0;
      DGP_ERR_PASS_CLEANUP(memorySentryReleaseMemory(psentry, cur_min, 
                                                     dgp_state));
    }
  }
  
  
  if (pmod){
    *pmod = mod;
  }
  
  NORMAL_RETURN;
 ERR_CLEANUP:
  
  if (mem_chg){
    if (mem_chg > 0){
      cur_min = mem_chg;
      mem_chg = 0;
      memorySentryClaimMemory(psentry, cur_min, dgp_state);
    }
    else { /* mem_chg < 0 */
      cur_min = -mem_chg;
      mem_chg = 0;
      memorySentryReleaseMemory(psentry, cur_min, dgp_state);
    }
  }

  if (pmod){
    *pmod = mod;
  }

  ERR_RETURN;
}

#endif
