/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  wbell? 1998?
**
**  File:  tokpos_quicksort.c
**
**  ...
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**      o  liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#include "tokpos_quicksort.h"
#include "util.h"
#include "dgp_types.h"

/* This parameter defines the cutoff between using quick sort and
   insertion sort for arrays;  arrays with lengths shorter or equal to the
   below value use insertion sort. */
#define CUTOFF 8            /* Testing shows that this is good value. */


/*************************************************************************
**
**  Function:  static INLINE Dgp_Int32 tokpos_comp(...)
**
**  This is a comparator given to qsort() to sort the tokpos's (see the
**  comment immediately above the actual call to qsort() in
**  DecompressAndWrite() below).
**   
**  Details of comparison:
** 
**  Is the top edge of the bounding box (bb) of the first tokpos closer
**  to the top of the page than the top edge of the bb of the second?
**
**  [If not, return DGP_FALSE.
**   
**  If so, is the left edge of the bb of the first tokpos closer to
**  the left of the page than the left edge of the bb of the second?
**   
**  If not, return DGP_FALSE.
**   
**  Otherwise, return DGP_TRUE.] / This seems incorrect, given qsort comparator
**  must return a signed quantity;  see code below. / dhtesq.
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**		o  Signed integer.
**
**	Modification history:
**      o  None.
**  
**/

static INLINE Dgp_Int32
tokpos_comp(jb2SymbolInstance *p1, jb2SymbolInstance *p2){
    if ((p1->ypos) > (p2->ypos)){
        return 1;
    }
    if (p1->ypos == p2->ypos) {
        if (p1->xpos > p2->xpos) {
            return 1;
        } else if (p1->xpos == p2->xpos) {
            return 0;
        }
    }
  return -1;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE void tokpos_swap(...)
**
**  ...
**
**	Arguments:
**      o  a, b...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static INLINE void 
tokpos_swap(jb2SymbolInstance **a, jb2SymbolInstance **b){
  jb2SymbolInstance *tmp;
  tmp = *a;
  *a = *b;
  *b = tmp;
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE void tokpos_shortsort(...)
**
**  ...
**
**	Arguments:
**      o  lo, hi...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static INLINE void 
tokpos_shortsort(jb2SymbolInstance **lo, jb2SymbolInstance **hi){
  jb2SymbolInstance **p, **max;
  
  while (hi > lo) {
    max = lo;
    for (p = lo + 1; p <= hi; p++) {
        if (tokpos_comp(*p, *max) > 0) {
            max = p;
        }
    }
    
    tokpos_swap(max, hi);
    hi--;
  }
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void tokpos_ext_qsort(...)
**
**  ...
**
**	Arguments:
**      o  base...
**      o  num...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

void
tokpos_ext_qsort(jb2SymbolInstance **base, unsigned num){
  jb2SymbolInstance **lo, **hi;              /* Ends of sub-array currently sorting. */
  jb2SymbolInstance **mid;                   /* Points to middle of subarray. */
  jb2SymbolInstance **loguy, **higuy;        /* Traveling pointers for partition step. */
  unsigned size;		          /* Size of the sub-array. */
  jb2SymbolInstance **lostk[30], **histk[30];
  Dgp_Int32 stkptr;               /* Stack for saving sub-array to be processed. */
  
  /* Note: the number of stack entries required is no more than
     1 + log2(size), so 30 is sufficient for any array. */
  
  if (num < 2){
      return;                     /* Nothing to do... */
  }
  
  stkptr = 0;                     /* Initialize stack. */
  
  lo = base;
  hi = base + num - 1;
  
  /* This entry point is for pseudo-recursion calling:  setting
     lo and hi and jumping to here is like recursion, but stkptr is
     preserved, locals aren't, so we preserve stuff on the stack. */
recurse:
  
  /* Number of el's to sort. */
  size = ((Dgp_Int32)hi - (Dgp_Int32)lo) / sizeof(jb2SymbolInstance) + 1;
  
  /* Below a certain size, it is faster to use a O(n^2) sorting method. */
  if (size <= CUTOFF) {
      tokpos_shortsort(lo, hi);
  } else {
    /* First we pick a partititioning element.  The efficiency of the
       algorithm demands that we find one that is approximately the
       median of the values, but also that we select one fast.  Using
       the first one produces bad performace if the array is already
       sorted, so we use the middle one, which would require a very
       wierdly arranged array for worst case performance.  Testing shows
       that a median-of-three algorithm does not, in general, increase
       performance. */
      
      mid = lo + (size / 2);			    /* Find middle element. */
      tokpos_swap(mid, lo);                 /* Swap it to beginning of array. */
    
    /* We now wish to partition the array into three pieces, one
       consisting of elements <= partition element, one of elements
       equal to the parition element, and one of element >= to it.  This
       is done below; comments indicate conditions established at every
       step. */
    
    loguy = lo;
    higuy = hi + 1;
    
    /* Note that higuy decreases and loguy increases on every iteration,
       so loop must terminate. */
    for (;;) {
      /* lo <= loguy < hi, lo < higuy <= hi + 1,
        A[i] <= A[lo] for lo <= i <= loguy,
        A[i] >= A[lo] for higuy <= i <= hi */
      
      do  {
          loguy++;
      } while (loguy <= hi && tokpos_comp(*loguy, *lo) <= 0);
      
      /* lo < loguy <= hi+1, A[i] <= A[lo] for lo <= i < loguy,
    	 either loguy > hi or A[loguy] > A[lo] */
      
      do  {
          higuy--;
      } while (higuy > lo && tokpos_comp(*higuy, *lo) >= 0);
      
      /* lo-1 <= higuy <= hi, A[i] >= A[lo] for higuy < i <= hi,
        either higuy <= lo or A[higuy] < A[lo] */
      
      if (higuy < loguy){
          break;
      }
      
      /* if loguy > hi or higuy <= lo, then we would have exited, so
      A[loguy] > A[lo], A[higuy] < A[lo],
      loguy < hi, highy > lo */
      
      tokpos_swap(loguy, higuy);
      
      /* A[loguy] < A[lo], A[higuy] > A[lo]; so condition at top
      of loop is re-established */
    }
    
    /* A[i] >= A[lo] for higuy < i <= hi,
	   A[i] <= A[lo] for lo <= i < loguy,
	   higuy < loguy, lo <= higuy <= hi
           implying:
	   A[i] >= A[lo] for loguy <= i <= hi,
	   A[i] <= A[lo] for lo <= i <= higuy,
	   A[i] = A[lo] for higuy < i < loguy */
    
    tokpos_swap(lo, higuy);     /* Put partition element in place. */
    
    /* OK, now we have the following:
       A[i] >= A[higuy] for loguy <= i <= hi,
       A[i] <= A[higuy] for lo <= i < higuy
       A[i] = A[lo] for higuy <= i < loguy    */
    
    /* We've finished the partition, now we want to sort the subarrays
       [lo, higuy-1] and [loguy, hi].
       We do the smaller one first to minimize stack usage.
       We only sort arrays of length 2 or more.*/
    
    if (higuy - 1 - lo >= hi - loguy ) {
        if (lo + 1 < higuy) {
            lostk[stkptr] = lo;
            histk[stkptr] = higuy - 1;
            ++stkptr;
        }                           /* Save big recursion for later. */

        if (loguy < hi) {
            lo = loguy;
            goto recurse;           /* Do small recursion. */
        }
    } else {
        if (loguy < hi) {
            lostk[stkptr] = loguy;
            histk[stkptr] = hi;
            ++stkptr;               /* Save big recursion for later. */
        }
        if (lo + 1 < higuy) {
            hi = higuy - 1;
            goto recurse;           /* Do small recursion. */
        }
    }
  }
  
  /* We have sorted the array, except for any pending sorts on the stack.
     Check if there are any, and do them. */
  
  --stkptr;
  if (stkptr >= 0) {
    lo = lostk[stkptr];
    hi = histk[stkptr];
    goto recurse;           /* Pop subarray from stack. */
  } else {
      return;
  }                         /* All subarrays done */
}
/*************************************************************************/

