/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 21.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  chains.c.
**
**  Code for chain manipulation.  A "chain" is a set of components connected by edges.
**	In this implementation, components that align well within some baseline 
**	and have similar size and color will be granted chain membership.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

#ifndef VXWORKS
#include <memory.h>
#endif  /* !(VXWORKS) */
#include <stdlib.h>
#include <math.h>
#include "image.h"
#include "image_defs.h"
#include "dgp_state.h"		/* For liberr exception handling. */
/*#include <hist.h>			// Used? dhtesq. */
#include "list.h"
#include "ccs.h"
#include "imutil.h"
#include "fit.h"
#include "edges.h"
#include "chains.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/chains.c,v 1.13 2003/02/11 01:31:43 mkrocken Exp $");


/* small pool for allocation and deletion of chain structs */
#define CHAIN_POOL_ID DGP_SMALLPOOL_MIN128

#define CHAIN_SEC_MALLOC(_ppchn) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, CHAIN_POOL_ID, (void **)(_ppchn)))
#define CHAIN_FREE(_pchn) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, CHAIN_POOL_ID, (void *)(_pchn)))


/*************************************************************************
**
**  Function:  void defaultChainOptions(chainOptions *p, Dgp_Int32 xres, Dgp_Int32 yres)
**
**
**  Sets thresholds and boundaries for various heuristics in this module, e.g., how far
**	can a component lie from a chain baseline and still be admitted?  Very unclear as of this
**	writing how satisfactory these numbers are, and to what extent they are optimized for Latin
**	character sets.
**
**	Arguments:
**		o  p points to a preallocated chainOptions{}.
**		o  xres is the horizontal resolution in dpi (?).
**		o  yres is the vertical resolution in dpi (?).
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**
**  
**/

void
defaultChainOptions(chainOptions *p, Dgp_Int32 xres, Dgp_Int32 yres) {
	p->minChainSize = 5;					/* Minimum size of a chain to be considered 'good' (in components?). */
	p->alignEpsilon = 1.0F;					/* How offset from the baseline a component can be */
											/*	relative to the height of the chain seed (in pixels?) */
	p->alignInlierEpsilon = 0.5F;			/* How offset from the baseline a component can be */
											/*	and still be included in the baseline estimation (in pixels?) */
	p->baseDot = 0.99F;						/* The maximum dot product for two baselines to be  */
											/*	considered parallel. -- for vectors normalized to unit length / dhtesq. */
											/*	Also -- isn't it the minimum? */
	p->minAvgPts = xres*yres/1500;			/* Minimum average mass of a chain. -- in "points/component" / dhtesq. */
	p->maxAvgPts = xres*yres/50;			/* Maximum average mass of a chain.	--	ditto */
	p->sizeEpsilon = 5.0F;					/* How much size can vary in the chain	-- from component to component, */
											/*	in pixels? */
	p->colorEpsilon = 100;					/* How much color can vary in the chain	-- units?  24 bit color implied elsewhere. */
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode freeChain(chain *c)
**
**
**  Frees three lists in the chain:  the connected components, the upper left positions, and
**	the lower right positions;  frees the passed chain.
**
**
**	Arguments:
**		o  c points to the chain to be freed.
**
**	Return value:
**		o  ErrCode;  can fail on list free operations.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'void' to 'ErrCode'.
**
**  
**/

ErrCode
freeChain(chain *c) {
	Dgp_State *dgp_state = c->dgp_state;
	DGP_PROLOGUE("freeChain()", DGP_PERRSTATE);

	DGP_ERR_PASS(DgpAdt_liFree(c->ccs));
	DGP_ERR_PASS(DgpAdt_liFree(c->ul));
	DGP_ERR_PASS(DgpAdt_liFree(c->dr));

	CHAIN_FREE(c);
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  INLINE ErrCode addComponent(ComponentInfo *cc, chain *pchain)
**
**
**  Adds the passed component to the passed chain if the component fits the various criteria
**	for size, color, etc.
**
**	Arguments:
**		o  cc points to the component to be added.
**		o  pchain points to the desired destination chain.
**
**	Return value:
**		o  ErrCode.  DgpAdt_List-related errors will be passed.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**		o  Removed the argument 'DgpAdt_ListNode node', which appeared to be unused.
**
**  
**/

INLINE ErrCode
addComponent(ComponentInfo *cc, chain *pchain) {
	Dgp_State	*dgp_state = pchain->dgp_state;
	DGP_PROLOGUE("addComponent()", DGP_PERRSTATE);
	Dgp_Float32		dr_baseDist, ul_baseDist, baseDist;
	DgpAdt_ListNode	node;

	/* Check size range. */
	if ((cc->size < pchain->minSize) || (cc->size > pchain->maxSize)){
		NORMAL_RETURN;
	}

	/* Check color range. */
	if ((cc->color.r < pchain->minred) || (cc->color.r > pchain->maxred) ||
		(cc->color.g < pchain->mingreen) || (cc->color.g > pchain->maxgreen) ||
		(cc->color.b < pchain->minblue) || (cc->color.b > pchain->maxblue)){
		NORMAL_RETURN;
	}

	/* Check distance to baseline of chain. */
	dr_baseDist = (Dgp_Float32)lineDistance((Dgp_Float64)cc->maxX, 
                                            (Dgp_Float64)cc->maxY, 
                                            pchain->dr_cog_x, 
                                            pchain->dr_cog_y,
                                            pchain->dr_base_x, 
                                            pchain->dr_base_y);
	ul_baseDist = (Dgp_Float32)lineDistance((Dgp_Float64)cc->minX, 
                                            (Dgp_Float64)cc->minY, 
                                            pchain->ul_cog_x, 
                                            pchain->ul_cog_y, 
                                            pchain->ul_base_x, 
                                            pchain->ul_base_y);
	baseDist = MAX(dr_baseDist, ul_baseDist);
	if (baseDist > pchain->maxBaseDist){
		NORMAL_RETURN;
	}

	/* Okay;  add component to chain. */
	DGP_ERR_PASS(DgpAdt_liAdd(pchain->ccs, (void *)cc, &node));

	/* And maybe add it to the baseline estimation too... */
	if (dr_baseDist <= pchain->maxInlierBaseDist) {
		DGP_ERR_PASS(DgpAdt_liAdd(pchain->dr, cc, &node));
		eigenLineFitDR( pchain->dr, &(pchain->dr_base_x), &(pchain->dr_base_y), 
						&(pchain->dr_cog_x), &(pchain->dr_cog_y));
	}
	if (ul_baseDist <= pchain->maxInlierBaseDist) {
		DGP_ERR_PASS(DgpAdt_liAdd(pchain->ul, cc, &node));
		eigenLineFitUL( pchain->ul, &(pchain->ul_base_x), &(pchain->ul_base_y), 
						&(pchain->ul_cog_x), &(pchain->ul_cog_y));
	}

	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode liCopy(Dgp_State *dgp_state, DgpAdt_List in, DgpAdt_List *_out)
**
**
**  Copies the passed list into a new list created in the function.  This involves one
**	traversal of the passed list.  The list interface in libwjr is still under construction,
**	so this function is not to be regarded as complete.
**
**	Arguments:
**		o  dgp_state points to valid Dgp_State.
**		o  in is the list to be copied.
**		o  _out will point to the list location on successful completion.
**
**	Return value:
**		o  ErrCode.  DgpAdt_List-related errors will be passed;  can fail on list creation or node addition.
**		   In the event of failure I will try to destroy the outgoing list.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'DgpAdt_List' to 'ErrCode'.
**
**  
**/

ErrCode
liCopy(Dgp_State *dgp_state, DgpAdt_List in, DgpAdt_List *_out) {
	DGP_PROLOGUE("liCopy()", DGP_PERRSTATE);
	DgpAdt_List		out;
	DgpAdt_ListNode	ln;
	DgpAdt_ListNode	retnode;

	/* Create a new list. */
	DGP_ERR_PASS(DgpAdt_liNew(&out, dgp_state));

	/* Copy the nodes. */
	DGPADT_FOREACH(ln, in) {
		DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(out, DGPADT_LIGET(ln), &retnode));
	}
	*_out = out;
	NORMAL_RETURN;

ERR_CLEANUP:
	/*	Try to destroy the outgoing list. */
	DGP_ERR_PASS(DgpAdt_liFree(out));
	*_out = out;		/* I assume the above call nulls the outgoing list... */
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  INLINE ErrCode newChain(	Dgp_State		*dgp_state,
**										ComponentInfo	*cc1,
**										ComponentInfo	*cc2,
**										chainOptions	*p,
**										chain			**_pchain)
**
**
**  Tries to create a new chain using the two passed components as seeds.  The two cc's will define
**	the baseline of the new chain.
**
**	Arguments:
**		o  dgp_state points to valid Dgp_State.
**		o  cc1 and cc2 are two candidate components for the new chain.
**		o  p points to previously initialized chain options.
**		o  _pchain will point to the location of the new chain on successful return.
**
**	Return value:
**		o  ErrCode.  Can fail on allocation and list-related errors.  If the two components do not
**		   fit the criteria for a new chain, newChain() will not throw an error.  In this case
**		   *_pchain will be set to NULL.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**		o  Patched some memory leaks on bailouts.
**
**  
**/

INLINE ErrCode
newChain(Dgp_State *dgp_state, ComponentInfo *cc1, ComponentInfo *cc2, chainOptions *p, chain **_pchain) {
	DGP_PROLOGUE("newChain()", DGP_PERRSTATE);

	chain		    *c;
	Dgp_Float64		par;
	Dgp_Int32		height;
	DgpAdt_ListNode	retnode;

	/* Get new chain.  Interface in flux, but currently has return type ErrCode. / dhtesq. */
	/*	Second argument is a void *page that used to be global;  will have to come back to this. */
	CHAIN_SEC_MALLOC(&c);
	c->dgp_state = dgp_state;

	/* Create the component list for the new chain. */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&c->ccs, dgp_state));

	/* Add the two components to the new chain's list. */
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(c->ccs, (void *)cc1, &retnode));
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(c->ccs, (void *)cc2, &retnode));

	/* Calculate baseline for new chain. */
	/* */
	/* Start by copying the existing component list:  once for the upper left series, */
	/*	again for the lower right. */
	DGP_ERR_PASS_CLEANUP(liCopy(dgp_state, c->ccs, &c->ul));
	DGP_ERR_PASS_CLEANUP(liCopy(dgp_state, c->ccs, &c->dr));

	eigenLineFitDR(c->dr, &(c->dr_base_x), &(c->dr_base_y), &(c->dr_cog_x), &(c->dr_cog_y));
	eigenLineFitUL(c->ul, &(c->ul_base_x), &(c->ul_base_y), &(c->ul_cog_x), &(c->ul_cog_y));

	/* Use a dot product to see if the two fit lines are approximately parallel. */
	/* If they aren't, we leave. / dhtesq. */
	par = ABS(c->dr_base_x*c->ul_base_x + c->dr_base_y*c->ul_base_y);
	if (par < p->baseDot) {
		freeChain(c);			/* Note this frees the chain's constituent lists. */
		*_pchain = NULL;
		NORMAL_RETURN;
	}

	/* Is the text horizontal or vertical? */
	/*	Calculate the 'height', which will be used in turn to calculate some of the thresholds */
	/*	for the chain (maxBaseDist, maxInlierBaseDist...). */
	if (ABS(c->dr_base_x) > ABS(c->dr_base_y)){
		c->horizontal = DGP_TRUE;
		height = cc1->height;
	}
	else{
		c->horizontal = DGP_FALSE;
		height = cc1->width;
	}

	/* Max baseline distance... */
	c->maxBaseDist = height * p->alignEpsilon;
	c->maxInlierBaseDist = height * p->alignInlierEpsilon;
	
	/* Size range... */
	/* ...outside of which new components are not to be considered part of the chain? dhtesq. */
	c->minSize = (Dgp_Int32)(cc1->size/p->sizeEpsilon);
	c->maxSize = (Dgp_Int32)(cc1->size*p->sizeEpsilon);

	/* Color range... */
	c->minred = cc1->color.r - p->colorEpsilon;
	c->maxred = cc1->color.r + p->colorEpsilon;
	c->mingreen = cc1->color.g - p->colorEpsilon;
	c->maxgreen = cc1->color.g + p->colorEpsilon;
	c->minblue = cc1->color.b - p->colorEpsilon;
	c->maxblue = cc1->color.b + p->colorEpsilon;

	/*	N.B.  This clause is asymmetric wrt cc1 and cc2.  Consider:  cc1->size == (p->sizeEpsilon/2) * cc2->size. */
	/*	cc2 will pass the trap, but c->minSize and c->maxSize are set wrt cc1...?? / dhtesq. */
	/* */
	/*	Also:  why check the size relationships at all when parallel vector check has done that implicitly? */
	/* */
	/* Check component 2... */
	if ((cc2->size < c->minSize) || (cc2->size > c->maxSize) ||
		(cc2->color.r < c->minred) || (cc2->color.r > c->maxred) ||
		(cc2->color.g < c->mingreen) || (cc2->color.g > c->maxgreen) ||
		(cc2->color.b < c->minblue) || (cc2->color.b > c->maxblue)) {
		freeChain(c);
		*_pchain = NULL;
		NORMAL_RETURN;
	}
	*_pchain = c;
	NORMAL_RETURN;


ERR_CLEANUP:
	/*	Free chain on a bailout;  member lists are freed in freeChain(). */
	freeChain(c);
	*_pchain = NULL;

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode growChain(chain *pchain)
**
**
**  Attempts to grow the passed chain by looking at the edges of each component in the chain.
**	Conditions for admission to the chain are expressed in addComponent().
**
**	Arguments:
**		o  pchain points to the chain to be grown.
**
**	Return value:
**		o  ErrCode.  Passes failures from dependencies.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**  
**/

ErrCode
growChain(chain *pchain) {
	Dgp_State		*dgp_state = pchain->dgp_state;
	DGP_PROLOGUE("growChain()", DGP_PERRSTATE);

	DgpAdt_ListNode		ln, retnode = NULL;
	edge			    *pedge;			/* An 'edge' is a struct containing a 'cc' field and a 'Dgp_Float32 dist' field. / dhtesq. */
	ComponentInfo	    *seed_cc;
	DgpAdt_ListNode		seed_node;
	/*static int          i = 0; */

	for (seed_node = DGPADT_LIFIRST(pchain->ccs); seed_node != NULL; seed_node = DGPADT_LINEXT(seed_node)) {
		seed_cc = (ComponentInfo *)DGPADT_LIGET(seed_node);
		/*	Each component itself maintains a list of edges. */
		DGPADT_FOREACH (ln, seed_cc->edges) {
			pedge = (edge *)DGPADT_LIGET(ln);
			/* If this edge component is not already in the chain, attach it. */
			DGP_ERR_PASS(DgpAdt_liIsIn(pchain->ccs, (void *)pedge->cc, &retnode));
			if(retnode){
				continue;
			}
			DGP_ERR_PASS(addComponent(pedge->cc, pchain));
		}
/*      printf("%d / %d / %d\n", i++, pchain->ccs->nelt, seed_cc->edges->nelt); */
	}
	NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  Dgp_Int32 totalPts(DgpAdt_List ccs)
**
**
**  Returns total chain mass in points (pixels?).
**
**	Arguments:
**		o  ccs is a list of all connected components in the chain.
**
**	Return value:
**		o  An integer containing the mass of pchain in points.
**
**	Modification history:
**		o  None.
**  
**/

Dgp_Int32
totalPts(DgpAdt_List ccs) {
	DgpAdt_ListNode		ln;
	ComponentInfo	*cc;
	Dgp_Int32				pts = 0;

	DGPADT_FOREACH (ln, ccs) {
		cc = DGPADT_LIGET(ln);
		pts += cc->num_pts;
	}

	return pts;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int32 checkChain(chain *pchain, chainOptions *p)
**
**
**  Performs final chain checks.  The chain should have a length of at least
**	p->minChainSize, and the average points per component should fall between 
**	p->minAvgPts and p->maxAvgPts.
**
**	Arguments:
**		o  pchain points to the chain to be checked.
**		o  p points to previously initialized chain options.
**
**	Return value:
**		o  An integer:  1 for pass, 0 for fail.
**
**	Modification history:
**		o  None.
**
**  
**/

Dgp_Int32
checkChain(chain *pchain, chainOptions *p) {
	Dgp_Int32 avgPts;

	/* Check minimum length... */
	/* Oddly phrased / dhtesq. */
/*	if (!(DGPADT_LILEN(pchain->ccs) >= p->minChainSize)){
		return 0;
	}
*/
	if (DGPADT_LILEN(pchain->ccs) < (Dgp_UInt32)p->minChainSize){
		return 0;
	}

	/* Make sure the average chain mass is not too small (or too large). */
	avgPts = totalPts(pchain->ccs)/DGPADT_LILEN(pchain->ccs);
	if ((avgPts < p->minAvgPts) || (avgPts > p->maxAvgPts)){
		return 0;
	}
	return 1;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode getChain(Dgp_State *dgp_state, ComponentInfo *cc1, chainOptions *p, chain **_pchain)
**
**
**  Tries to create a valid chain starting with the passed connected component.
**	I look at the first listed edge component and, using these two as seeds, attempt
**	to construct a chain.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  cc1 is the connected component seed.
**		o  p points to previously initialized chain options.
**		o  _pchain points to the location of the new chain on successful return.
**
**	Return value:
**		o  ErrCode.  Passes failures from dependencies.  *_pchain will be set to NULL on nominal failure,
**		   but an error will not be thrown.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**  
**/

ErrCode
getChain(Dgp_State *dgp_state, ComponentInfo *cc1, chainOptions *p, chain **_pchain) {
	DGP_PROLOGUE("getChain()", DGP_PERRSTATE);

	DgpAdt_ListNode		ln;
	edge			*pedge;
	chain			*pchain;

	/* Try only the closest edge component as a second seed... */
	/* (why? performance compromise? / dhtesq.) */
	ln = DGPADT_LIFIRST(cc1->edges);
	if (ln) {
		pedge = (edge *)DGPADT_LIGET(ln);

		DGP_ERR_PASS(newChain(dgp_state, cc1, pedge->cc, p, &pchain));

		/*	newChain reports nominal failure by setting pchain to NULL. */
		if (pchain == NULL) {
			*_pchain = NULL;
			NORMAL_RETURN;
		}

		/* Try to grow the newly-created chain. */
		DGP_ERR_PASS(growChain(pchain));

		/* Is it good enough? */
		if (checkChain(pchain, p)) {
			*_pchain = pchain;
			NORMAL_RETURN;
		} else {
			/* Forget it... */
			freeChain(pchain);
		}
	}

	*_pchain = NULL;
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode getChains(Dgp_State		*dgp_state,
**								 ComponentInfo	*components,
**								 Dgp_Int32			numccs,
**								 DgpAdt_List			*_chains,
**								 chainOptions	*p)
**
**
**  Oversees the creation of chains for a given components array (in practice presumably
**	containing all cc's on a given page).  Creates a list of chains which on successful
**	return are nominally tokenizable text.
**
**	Chains created by this function will carry the passed Dgp_State{}.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  components is an array of cc's.
**		o  numccs is length(components).
**		o  _chains points to a list of chains on successful return.
**		o  p points to previously initialized chain options.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**  
**/

ErrCode
getChains(Dgp_State *dgp_state, ComponentInfo *components, Dgp_Int32 numccs, DgpAdt_List *_chains, chainOptions *p) {
	DGP_PROLOGUE("getChains()", DGP_PERRSTATE);

	Dgp_Int32			cc;
	ComponentInfo		*pcc;
	DgpAdt_ListNode		ln, retnode = NULL;
	chain				*pchain;
	DgpAdt_List			chains = NULL;


  /* validate small pool usage */
  DGP_SMALLPOOL_VALIDATE(chain, CHAIN_POOL_ID);
  
  /* Here is the list of chains. */
	/* Each chain itself contains a list of connected component pointers. */
	DGP_ERR_PASS(DgpAdt_liNew(&chains, dgp_state));

	/* Make sure no cc in the passed list is marked as text. */
	for (cc = 0; cc < numccs; cc++) {
		components[cc].text = TEXT_0;
	}

	/*	Loop through all cc's in the array.  If the cc looks suitable, */
	/*	try to seed a chain with it. */
	for (cc = 0; cc < numccs; cc++) {
		/* Could this be a seed? */
		/*	...We require the component has not yet been marked as text, and also that the component's mass */
		/*	exceed p->minAvgPts.  Note that individual components with less mass than this could wind up in */
		/*	the chain. */
		/* */
		if ((components[cc].text == TEXT_0) && (components[cc].num_pts >= p->minAvgPts)) {
			/* Let him try to get a chain for himself... */
			DGP_ERR_PASS_CLEANUP(getChain(dgp_state, &(components[cc]), p, &pchain));

			/* Did we get a new chain? */
			if (pchain != NULL) {
				/* Yes;  mark everyone in the chain as being in some chain.  Requires another list traverse. */
				/* */
				/*	Can we do this marking during the chain growth? / dhtesq. */
				DGPADT_FOREACH (ln, pchain->ccs) {
					pcc = (ComponentInfo *)DGPADT_LIGET(ln);
					pcc->text = TEXT_1;
				}
				DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(chains, (void *)pchain, &retnode));
			}
		}
	}

	*_chains = chains;
	NORMAL_RETURN;

ERR_CLEANUP:
	if(chains != NULL){
		DGP_ERR_PASS(freeChains(dgp_state, chains));
	}

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode freeChains(Dgp_State *dgp_state, DgpAdt_List chains)
**
**
**  Call this to get rid of a list of chains obtained using getChains().  Frees each chain
**	in the list, then the list itself.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  chains is the list of chains to be freed.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**  
**/

ErrCode
freeChains(Dgp_State *dgp_state, DgpAdt_List chains) {
	DGP_PROLOGUE("freeChains()", DGP_PERRSTATE);

	DgpAdt_ListNode ln;
	chain *pchain;

	DGPADT_FOREACH (ln, chains) {
		pchain = (chain *)DGPADT_LIGET(ln);
		DGP_ERR_PASS(freeChain(pchain));
	}
	DGP_ERR_PASS(DgpAdt_liFree(chains));

	NORMAL_RETURN;
}
/*************************************************************************/


/*
 * The following is a bunch of routines to output chains for debugging purposes.
 */

/*	These allow quick modification if you instead want ul, dr: */
#define xfield centerX
#define yfield centerY

/*************************************************************************
**
**  Function:  void drawChainRGB(DgpImage_ColorRGB out, DgpImage_RGB *color, chain *pchain)
**
**
**  Draws a chain as a series of DgpImage_RGB lines joining the member components.
**
**	Arguments:
**		o  out is the image to be drawn to.
**		o  color points to the output color.
**		o  pchain points to the chain to be drawn.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
drawChainRGB(DgpImage_ColorRGB out, DgpImage_RGB *color, chain *pchain) {
	ComponentInfo *cc1, *cc2;
	DgpAdt_ListNode ln;

	ln = DGPADT_LIFIRST(pchain->ccs);
	cc1 = (ComponentInfo *)DGPADT_LIGET(ln);
	for (ln = DGPADT_LINEXT(ln); ln != NULL; ln = DGPADT_LINEXT(ln)) {
		cc2 = (ComponentInfo *)DGPADT_LIGET(ln);
		drawLineRGB(out, 
                    color, 
                    (Dgp_Int32)cc1->xfield, 
                    (Dgp_Int32)cc1->yfield, 
                    (Dgp_Int32)cc2->xfield, 
                    (Dgp_Int32)cc2->yfield);
		cc1 = cc2;
	}
	return;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  void drawChainB(DgpImage_Binary out, chain *pchain)
**
**
**  Draws a chain as a series of binary lines joining the member components.
**
**	Arguments:
**		o  out is the image to be drawn to.
**		o  pchain points to the chain to be drawn.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
drawChainB(DgpImage_Binary out, chain *pchain) {
	ComponentInfo *cc1, *cc2;
	DgpAdt_ListNode ln;

	ln = DGPADT_LIFIRST(pchain->ccs);
	cc1 = (ComponentInfo *)DGPADT_LIGET(ln);
	for (ln = DGPADT_LINEXT(ln); ln != NULL; ln = DGPADT_LINEXT(ln)) {
		cc2 = (ComponentInfo *)DGPADT_LIGET(ln);
		drawLineB(out, (Dgp_Int32)cc1->xfield, (Dgp_Int32)cc1->yfield, (Dgp_Int32)cc2->xfield, (Dgp_Int32)cc2->yfield);
		cc1 = cc2;
	}
	return;
}
/*************************************************************************/

#undef xfield
#undef yfield

/*************************************************************************
**
**  Function:  void drawChains(DgpImage_Binary out, DgpAdt_List chains)
**
**
**  Draws a list of chains to the passed binary image.
**
**	Arguments:
**		o  out is the image to be drawn to.
**		o  chains is the list of chains to be drawn.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
drawChains(DgpImage_Binary out, DgpAdt_List chains) {
	DgpAdt_ListNode ln;
	chain *pchain;

	DGPADT_FOREACH (ln, chains) {
		pchain = (chain *)DGPADT_LIGET(ln);
		drawChainB(out, pchain);
	}
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  void blatChain(DgpImage_ColorRGB out, DgpImage_Long compMap, chain *pchain, ComponentInfo *components)
**
**
**  Blats the components in the passed chain to the passed DgpImage_RGB image.  Outputs text to console.
**
**	Arguments:
**		o  out is the DgpImage_RGB image to be drawn to.
**		o  compMap is a valid component map.
**		o  pchain points to the chain to be blatted.
**		o  components is the array of cc's in the compMap.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
blatChain(DgpImage_ColorRGB out, DgpImage_Long compMap, chain *pchain, ComponentInfo *components) {
	DgpAdt_ListNode ln;
	ComponentInfo *cc;

	DGPADT_FOREACH (ln, pchain->ccs) {
		cc = (ComponentInfo *)DGPADT_LIGET(ln);
		printf("cc: %d\n", (int)cc->ccnum);
		blatComponent(out, compMap, components, cc->ccnum);
	}
	return;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  void debugChain(Dgp_Int32 dcc, DgpAdt_List chains, ComponentInfo *components, DgpImage_Long compMap, DgpImage_ColorRGB out)
**
**
**  Blats and draws the passed component to the passed image iff the component lies somewhere
**	on the passed list of chains. Outputs text to console.
**
**	Arguments:
**		o  dcc is the index of the component to be drawn/blatted.
**		o  chains is the list of chains to be checked for the component.
**		o  components is the cc array for the compMap.
**		o  compMap is a valid component map for the image.
**		o  out is the DgpImage_RGB image to be drawn to.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
debugChain(Dgp_Int32 dcc, DgpAdt_List chains, ComponentInfo *components, DgpImage_Long compMap, DgpImage_ColorRGB out) {
	chain			*pchain;
	DgpAdt_ListNode		ln1;
	DgpAdt_ListNode		ln2;
	ComponentInfo	*cc;
	DgpImage_RGB				color;

	color.r = 255;
	color.g = 255;
	color.b = 255;

	imClearRGB(out);
	DGPADT_FOREACH (ln1, chains) {
		pchain = (chain *)DGPADT_LIGET(ln1);
		DGPADT_FOREACH (ln2, pchain->ccs) {
			cc = (ComponentInfo *)DGPADT_LIGET(ln2);
			if (cc->ccnum == dcc) {
				blatChain(out, compMap, pchain, components);
				drawChainRGB(out, &color, pchain);
				printf("line fit, dr: %f, %f\n", pchain->dr_base_x, pchain->dr_base_y);
				printf("line fit, ul: %f, %f\n", pchain->ul_base_x, pchain->ul_base_y);
				return;
			}
		}
	}
	return;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode saveChains(Dgp_State *dgp_state, DgpImage_Long compMap, DgpAdt_List chains, ComponentInfo *components)
**
**
**  Blats & draws the passed list of chains to an DgpImage_RGB image allocated here, then saves the image
**	as 'chains.ppm'.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  compMap is a valid component map for the image.
**		o  chains is the list of chains to be checked for the component.
**		o  components is the cc array for the compMap.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**  
**/

ErrCode
saveChains(Dgp_State *dgp_state, DgpImage_Long compMap, DgpAdt_List chains, ComponentInfo *components) {
	DGP_PROLOGUE("saveChains()", DGP_PERRSTATE);

	DgpAdt_ListNode			ln;
	chain				    *pchain;
	DgpImage_RGB			c;
	Dgp_Int32				width = DGPIMG_WIDTH(compMap) - 2;			/* compMap has a border of one pixel. */
	Dgp_Int32				height = DGPIMG_HEIGHT(compMap) - 2;
	DgpImage_ColorRGB		out;

	DGP_ERR_PASS(DgpImage_new(dgp_state, (Dgp_UInt16)DGPIMAGE_RGB, width, height, (void **)&out));

	c.r = c.g = c.b = 255;

	DGPADT_FOREACH (ln, chains) {
		pchain = DGPADT_LIGET(ln);
		blatChain(out, compMap, pchain, components);
		drawChainRGB(out, &c, pchain);		
	}

/*	#define	DgpImage_Save(i, f)			DgpImage_pSave_((void *)(i), (i)->header.tag, (f)) */

	DGP_ERR_PASS_CLEANUP(DgpImage_save(out, "chains.ppm"));

	DGP_ERR_PASS(DgpImage_free(out));

	NORMAL_RETURN;

ERR_CLEANUP:
	DGP_ERR_PASS(DgpImage_free(out));

	ERR_RETURN;
}
/*************************************************************************/
