/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 24.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  complexity.c.
**
**  Code for calculation of component complexity.  For these purposes, complexity
**	is defined as the number of 4-connected subcomponents in an 8-connected component (see below).
**	I gather text components have low complexity, making this an effective heuristic for
**	segmentation.
**
**
**	More detail:
**	The complexity measure we use is the number of 4-connected pieces in each component (the pieces are
**	themselves 8-connected).
**
**	Pipeline position in normal usage seems to be in sc_findTextPass1(). / dhtesq.
**	Currently it appears the only way a cc can have complexity > 1 is to be consist of pieces connected only
**	at the ul, ur, lr, or ll corners:
**
**	-->	*   * <--
**		  * 
**	-->	*   * <--
**
**	This seems queer:  1) if the "bridge" to the next piece is over one pixel long, each subsequent pixel will
**	count as a piece, and increase the complexity measure by one;  2) if the pieces are connected by a horizontal or 
**	vertical bridge, they will be have the same complexity as a single piece (one).  Does all this have some
**	justification, or is it for ease of calculation?
**
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/


#include <math.h>
#include "dgp_state.h"		/* For liberr exception handling. */
#include "image.h"
#include "image_defs.h"
#include "list.h"
#include "ccs.h"			
#include "stack.h"
#include "findCC.h"
#include "complexity.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/complexity.c,v 1.6 2003/02/10 22:53:36 mkrocken Exp $");

#define STACK_SIZE 100000

/*	This macro pushes the image point (x, y) onto stack, reallocating if necessary. */
/*	Modified for liberr, though stack.c has not been changed as of this writing. */
/* */
/*	x and y are cast as void*, so the native stack memory management is sufficient; */
/*	could break on some weird machines in extraordinary circumstances. */
/* */
/*	Cleaner as inline? / dhtesq. */
#define pushPoint(x,y) \
if ((!DGPIMG_REF(visited, x, y)) && (DGPIMG_REF(image, x, y) == v)) { \
	if (!DgpAdt_staPush(xstack,(void *)(x))) { \
		DGP_ERR_PASS(DgpAdt_staRealloc(dgp_state, xstack, 2*DgpAdt_staMaxNumNodes(xstack))); \
        DgpAdt_staPush(xstack,(void *)(x)); \
		DGP_ERR_PASS(DgpAdt_staRealloc(dgp_state, ystack, 2*DgpAdt_staMaxNumNodes(ystack))); \
        DgpAdt_staPush(ystack,(void *)(y)); \
	} else \
        DgpAdt_staPush(ystack,(void *)(y)); \
	DGPIMG_REF(visited, x, y) = 1; }


/*************************************************************************
**
**  Function:  ErrCode markCC(	Dgp_State		*dgp_state,
**								DgpImage_Long		image,
**								Dgp_Int32				sx,
**								Dgp_Int32				sy,
**								Dgp_Int32			v,
**								DgpImage_Binary		visited,
**								Stack			xstack,
**								Stack			ystack)
**
**
**  Marks the four-connected neighbors of a passed point on a passed image iff
**	they possess the passed value.  Neighbors are marked as 'visited' on the passed binary image.
**
**	(Should this be static?)
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  image is the image to be explored.
**		o  (sx, sy) are the coordinates of the seed point on the image.
**		o  v is the value the point's neighbors are required to have in order to be marked as visited.
**		o  visited is the binary image where visits will be marked.  Prepared by caller.
**		o  xstack and ystack are prepared stacks used for connection explorations.
**
**	Return value:
**		o  ErrCode on stack reallocation failure.  If this happens, markCC() will not destroy the
**		   passed stacks.
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**  
**/

ErrCode
markCC(Dgp_State        *dgp_state, 
       DgpImage_Long    image, 
       Dgp_Int32        sx, 
       Dgp_Int32        sy, 
       Dgp_Int32        v, 
       DgpImage_Binary  visited, 
       DgpAdt_Stack     xstack,
       DgpAdt_Stack     ystack) {
	DGP_PROLOGUE("markCC()", DGP_PERRSTATE);
	Dgp_Int32	x, y;

	/* Add the passed point to our stack. */
	pushPoint(sx, sy);

	while (!DgpAdt_staEmpty(xstack)) {
		/* Pop the stack to get the current point. */
		x = (Dgp_Int32)DgpAdt_staPop(xstack);
		y = (Dgp_Int32)DgpAdt_staPop(ystack);

		/* Push neighbors. */
		pushPoint(x - 1, y);
		pushPoint(x + 1, y);
		pushPoint(x, y + 1);
		pushPoint(x, y - 1);
    }

	NORMAL_RETURN;
}
/*************************************************************************/

#undef pushPoint


/*************************************************************************
**
**  Function:  ErrCode getComplexity(	Dgp_State		*dgp_state,
**										DgpImage_Long		compMap,
**										ComponentInfo	*components,
**										Dgp_Int32		numccs)
**
**
**  For each 8-connected component in the passed map, finds the number of 4-connected components
**	inside of it.  We call this the "complexity" of the component.
**
**	The scancolor pipeline uses this interface for the complexity module.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  compMap is the component map to be examined.
**		o  components is the cc array for compMap.
**		o  numccs is length(compMap).
**
**	Return value:
**		o  An ErrCode.  Fails on bad allocation of visited image, bad stack allocation,
**		   or on bad stack reallocation via markCC().
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**  
**/

ErrCode
getComplexity(Dgp_State *dgp_state, DgpImage_Long compMap, ComponentInfo *components, Dgp_Int32 numccs) {
	DGP_PROLOGUE("getComplexity()", DGP_PERRSTATE);

	Dgp_Int32			width = DGPIMG_WIDTH(compMap) - 2;		/* (-2 because compMap has border of 1 pixel.) */
	Dgp_Int32			height = DGPIMG_HEIGHT(compMap) - 2;
	Dgp_Int32			x, y;
	DgpAdt_Stack		xstack = NULL, ystack = NULL;
	DgpImage_Binary     visited;
	Dgp_Int32			cc;

	/* Initialize our complexity measure. */
	for (x = 0; x < numccs; x++) {
		components[x].complexity = 0;
	}
	/* Try ignoring complexity...v. expensive, and studies show <2% of components get rejected anyway. */
	NORMAL_RETURN;

	/* Get some memory for marking visited spots. */
	/*	We'll place a one pixel border around this one, just like the compMap. */
	DGP_ERR_PASS(DgpImage_newOffset(dgp_state, DGPIMAGE_BINARY, width + 2, height + 2, -1, -1, (void **)&visited));

	/* Mark the border as already visited. */
	/* */
	/*	Note since the 'visited' image has been allocated with a safe border around it, */
	/*	we can use negative array indices on it with impunity. / dhtesq. */
	for (x = -1; x <= width; x++) {
		DGPIMG_REF(visited, x, -1) = 1;
		DGPIMG_REF(visited, x, height) = 1;
	}
	for (y = -1; y <= height; y++) {
		DGPIMG_REF(visited, -1, y) = 1;
		DGPIMG_REF(visited, width, y) = 1;
	}

	/* Stacks for findCC... --?  You mean markCC()? / dhtesq. */
	DGP_ERR_PASS_CLEANUP(DgpAdt_staNew(dgp_state, STACK_SIZE, &xstack));
	DGP_ERR_PASS_CLEANUP(DgpAdt_staNew(dgp_state, STACK_SIZE, &ystack));

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			/* Have we touched this pixel? and does it belong to some component? */
			cc = DGPIMG_REF(compMap, x, y);
			if ((!DGPIMG_REF(visited, x, y)) && (cc != NOT_ASSIGNED)) {
				DGP_ERR_PASS_CLEANUP(markCC(dgp_state, compMap, x, y, cc, visited, xstack, ystack));
				components[cc].complexity++;
			}
		}
	}

	/* Thank purify... */
	DgpAdt_staFree(dgp_state, xstack);
	DgpAdt_staFree(dgp_state, ystack);
	DGP_ERR_PASS(DgpImage_free(visited));

	NORMAL_RETURN;

ERR_CLEANUP:
	if(xstack != NULL){
		DgpAdt_staFree(dgp_state, xstack);
	}
	if(ystack != NULL){
		DgpAdt_staFree(dgp_state, ystack);
	}
	DGP_ERR_PASS(DgpImage_free(visited));

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int32 remComplex(ComponentInfo *components, Dgp_Int32 numccs, Dgp_Int32 complexthresh)
**
**
**  Sets the 'size' field of components to zero if their complexity exceeds the passed threshold.
**	This effectively removes them from consideration by further stages of the scancolor pipeline.
**
**	Arguments:
**		o  components is the cc array to be scanned.
**		o  numccs is length(components).
**		o  complexthresh is the complexity measure above which a component is to be zeroed.
**
**	Return value:
**		o  An integer;  number of removed components.
**
**	Modification history:
**		o  None.
**
**    
**/

Dgp_Int32
remComplex(ComponentInfo *components, Dgp_Int32 numccs, Dgp_Int32 complexthresh) {
	Dgp_Int32 cc;
	Dgp_Int32 num = 0;

	for (cc = 0; cc < numccs; cc++) {
		if (components[cc].complexity > complexthresh) {
			components[cc].size = 0;
			num++;
		}
	}
	return num;
}
/*************************************************************************/
