/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 25.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  edges.c.
**
**
** Given a component, we want to know its neighbors of similar color and size.
**
** For each range of component size and color we make an image marking the center
** of the bounding box of the components fitting that range. We figure out the Voronoi
** neighbors in this image by propagating the component labels and finding the neighboring
** labels in the resulting image. We create edges between Voronoi neighbors.
**				-- any guarantee the full component will not extend beyond the Voronoi region? / dhtesq.
**
** The edges are represented as a list of edge{}s hanging off the ComponentInfo{} of each
** component.  They are sorted by distance (the distance between the center of the two 
** components it connects).
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "dgp_err_control.h"		/* For liberr exception handling. */
#include "image.h"
#include "image_defs.h"
#include "list.h"
#include "ccs.h"
#include "findCC.h"
/*#include "pixr.h" */
#include "imutil.h"
#include "propagate.h"
#include "edges.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/edges.c,v 1.10 2003/02/10 22:53:36 mkrocken Exp $");


/* small pool for allocation and deletion of chain structs */
#define EDGE_POOL_ID DGP_SMALLPOOL_MIN8

#define EDGE_SEC_MALLOC(_ppedge) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, EDGE_POOL_ID, (void **)(_ppedge)))
#define EDGE_FREE(_pedge) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, EDGE_POOL_ID, (void *)(_pedge)))


/*************************************************************************
**
**  Function:  void defaultEdgeOptions(chainOptions *p, Dgp_Int32 xres, Dgp_Int32 yres)
**
**
**  Sets default options for the edge finder.
**
**	Arguments:
**		o  p points to a prepared chainOptions{}.
**		o  xres is the horizontal resolution in dpi (?).
**		o  yres is the vertical resolution in dpi (?).
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**
**  
**/

void
defaultEdgeOptions(edgeOptions *p, Dgp_Int32 xres, Dgp_Int32 yres) {
	Dgp_Int32 res = MIN(xres, yres);	/* In dpi. */
	p->scale = 60.0F/res;		/* We compute the Voronoi neighboors in a scaled down image. */
	p->sizeEpsilon = 3.0F;		/* Size range to include in each Voronoi diagram. */
	p->colorEpsilon = 100;		/* Color range to include in each Voronoi diagram. */
	p->maxDist = (Dgp_Int32)(0.5*res*p->scale);	/* Maximum edge distance we are interested in.  */
											/* -- I think this is the distance _to_ the edge -- and only along the x or y axes. */
											/* ...? dhtesq. */
	p->maxSize = res;			/* Maximum component size to connect with edges. */
	p->minSize = (Dgp_Int32)(0.03*res);	/* Minimum component size to connect with edges. */
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  INLINE ErrCode insertEdge(Dgp_State *dgp_state, DgpAdt_List edges, edge *pedge)
**
**
**  Inserts the passed edge into a position in the passed list sorted in order of increasing distance.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  edges is a pre-sorted list of edges.
**		o  pedge points to the edge to be inserted.
**
**	Return value:
**		o  ErrCode;  can fail on list add operations.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**/

INLINE ErrCode
insertEdge(Dgp_State *dgp_state, DgpAdt_List edges, edge *pedge) {
	DGP_PROLOGUE("insertEdge()", DGP_PERRSTATE);
	edge			*pcur_edge;
	DgpAdt_ListNode		ln, retnode;

	DGPADT_FOREACH (ln, edges) {
		pcur_edge = (edge *)DGPADT_LIGET(ln);
		if (pedge->dist <= pcur_edge->dist) {
			/* Time to insert this edge... */
			DGP_ERR_PASS(DgpAdt_liAddAfter(edges, DGPADT_LIPREV(ln), (void *)pedge, &retnode));
			NORMAL_RETURN;
		}
	}

	/* Add edge to end of list. */
	DGP_ERR_PASS(DgpAdt_liAdd(edges, (void *)pedge, &retnode));
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  INLINE ErrCode addEdge(Dgp_State *dgp_state, ComponentInfo *cc1, ComponentInfo *cc2)
**
**
**	This function adds an edge between two cc's. It adds the edge to its correct
**  position sorted by distance between components. It does not insert duplicate edges.
**  Also, it only adds the edge to the list hanging out of the cc with smaller memory address.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  cc1 and cc2 point to the two components whose edge is to be added to the list.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**/

INLINE ErrCode
addEdge(Dgp_State *dgp_state, ComponentInfo *cc1, ComponentInfo *cc2) {
	DGP_PROLOGUE("addEdge()", DGP_PERRSTATE);
	DgpAdt_ListNode			ln;
	edge				*pedge, *pcur_edge;
	ComponentInfo		*tmp;

	/* Make sure cc1 is the 'smaller component';  swap pointers. */
	if (cc1 > cc2) {
		tmp = cc1;
		cc1 = cc2;
		cc2 = tmp;
	}

	/* Check if edge is already in the list.  If so, I don't want to calculate the distance. */
    /*  Added check for nulllist here, as I don't think it's guaranteed that cc1 will have */
	DGPADT_FOREACH (ln, cc1->edges) {
		pcur_edge = DGPADT_LIGET(ln);
		if (pcur_edge->cc == cc2){	/* A duplicate!  Bail. */
			NORMAL_RETURN;
		}
	}

	/* Make a new edge. */
	EDGE_SEC_MALLOC(&pedge);
	pedge->cc = cc2;
	pedge->dist = ccDistance(cc1, cc2);

	/* Add edge to its position. */
	DGP_ERR_PASS_CLEANUP(insertEdge(dgp_state, cc1->edges, pedge));
	NORMAL_RETURN;

ERR_CLEANUP:
	/* Free allocated edge. */
	EDGE_FREE(pedge);
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode addSecondEdge(Dgp_State *dgp_state, ComponentInfo *components, Dgp_Int32 numccs)
**
**
**  When we create edges between components, each edge only gets stuck in one of the component's
**  edge lists.  Here we fix this.  The original edges get stuck in the list of the component 
**  with the smaller memory address.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  components is a cc array.
**		o  numccs is length(components).
**
**	Return value:
**		o  An ErrCode.  On an edge insertion error, addSecondEdge() will free the current edge, but
**		   not others that have already been successfully inserted.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**/

ErrCode
addSecondEdge(Dgp_State *dgp_state, ComponentInfo *components, Dgp_Int32 numccs) {
	DGP_PROLOGUE("addSecondEdge()", DGP_PERRSTATE);
	Dgp_Int32					cc;
	DgpAdt_ListNode			ln;
	edge				*pedge, *pcur_edge;
	ComponentInfo		*cc1, *cc2;

	for (cc = 0; cc < numccs; cc++) {
		cc1 = &(components[cc]);
		DGPADT_FOREACH (ln, cc1->edges) {
			pcur_edge = (edge *)DGPADT_LIGET(ln);
			cc2 = pcur_edge->cc;
			if (cc2 > cc1) {
				/* Make a new edge. */
				EDGE_SEC_MALLOC(&pedge);
				pedge->cc = cc1;
				pedge->dist = pcur_edge->dist;
				DGP_ERR_PASS_CLEANUP(insertEdge(dgp_state, cc2->edges, pedge));
			}
		}
	}
	NORMAL_RETURN;

ERR_CLEANUP:
	/*	Free the edge allocated immediately prior to error -- but no others. */
	EDGE_FREE(pedge);
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  INLINE ErrCode getSimilarEdge(	Dgp_State		*dgp_state,
**												ComponentInfo	*components,
**												Dgp_Int32				x1,
**												Dgp_Int32				y1,
**												Dgp_Int32				x2,
**												Dgp_Int32				y2,
**												DgpImage_Long		blobMap)
**
**
**  Adds the edge between the two components at the two passed positions on the passed map.
**  If the two points belong to the same component, or belong to an unassigned component, 
**  the edge will not be added.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  components is a cc array.
**		o  (x1, y1) and (x2, y2) are points lying in two components whose edge is to be added.
**		o  blobMap is an image of component indices.
**
**	Return value:
**		o  An ErrCode passed up from addEdge().
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**/

INLINE ErrCode
getSimilarEdge(Dgp_State        *dgp_state, 
               ComponentInfo    *components, 
               Dgp_Int32        x1, 
               Dgp_Int32        y1, 
               Dgp_Int32        x2, 
               Dgp_Int32        y2, 
               DgpImage_Long    blobMap) {
	DGP_PROLOGUE("getSimilarEdge()", DGP_PERRSTATE);
	Dgp_Int32 cc1, cc2;			/* Note these are not *ComponentInfo{}. */

	cc1 = DGPIMG_REF(blobMap, x1, y1);
	cc2 = DGPIMG_REF(blobMap, x2, y2);
	if (cc1 == cc2) {
		NORMAL_RETURN;
	}
	if ((cc1 == NOT_ASSIGNED) || (cc2 == NOT_ASSIGNED)) {
		NORMAL_RETURN;
	}
	DGP_ERR_PASS(addEdge(dgp_state, &(components[cc1]), &(components[cc2])));
	NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode getSimilarEdges(	Dgp_State		*dgp_state,
**										DgpImage_Long		similar,
**										ComponentInfo	*components,
**										Dgp_Int32				maxDist);
**
**
**  Prepares a Voronoi diagram for the passed cc map and uses it to find edges between
**  adjoining components.  The Voronoi propagation proceeds only up to the passed distance
**  parameter, so cc's separated by greater distances than this will have no recorded edge.
**	...aka 'Delaunay triangulation'.  In practice, the passed component map only has 'similar'
**	components, i.e, components with a specified size and color relationship.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  similar is the cc map to be examined.
**		o  components is the cc array for similar.
**		o  maxDist is the separation, in pixels, above which two cc's will not bother to record an edge.
**
**	Return value:
**		o  An ErrCode passed up from addEdge().
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'Dgp_Int32' to 'ErrCode'.
**
**/

ErrCode
getSimilarEdges(Dgp_State *dgp_state, DgpImage_Long similar, ComponentInfo *components, Dgp_Int32 maxDist) {
	DGP_PROLOGUE("getSimilarEdges()", DGP_PERRSTATE);
	Dgp_Int32				x, y;
	Dgp_Int32				width = DGPIMG_WIDTH(similar);
	Dgp_Int32				height = DGPIMG_HEIGHT(similar);
	Dgp_Int32				xmax = width - 1;
	Dgp_Int32				ymax = height - 1;
	DgpImage_Long		dist;

	/* Propagate the component labels... */

	/* Make initial distance image. */
	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG, width, height, (void **)&dist));

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			if (DGPIMG_REF(similar, x, y) == NOT_ASSIGNED) {
				DGPIMG_REF(dist, x, y) = maxDist + 1;
			} 
		}
	}
	propagate_l(similar, dist);				/* See propagate.c / dhtesq. */
	/*	Done with dist image... */
	DGP_ERR_PASS(DgpImage_free(dist));

	/* Find neighbors & make an edge for each. */
	for (y = 0; y < ymax; y++) {
		for (x = 0; x < xmax; x++) {
			/* x - x */
			DGP_ERR_PASS(getSimilarEdge(dgp_state, components, x, y, x+1, y, similar));
			/*  x
			    |
				x  */
			DGP_ERR_PASS(getSimilarEdge(dgp_state, components, x, y, x, y+1, similar));
		}
	}
	
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode getSimilarComponents(	Dgp_State		*dgp_state,
**												ComponentInfo	*components,
**												Dgp_Int32				numccs,
**												Dgp_Int32				cc,
**												Dgp_Int32				width,
**												Dgp_Int32				height,
**												edgeOptions		*p,
**												DgpImage_Long		*_out)
**
**
**  Makes a component map of cc's resembling a passed component.  Color and size range
**  criteria are contained in the passed options.
**
**	The generated image is scaled down by a factor specified in the options p.  As of this writing,
**	integral addressing of the scaled-down image can result in overwritten components.  Unsure
**	what consequences this has / dhtesq.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  components is the cc array to compare against.
**		o  numccs is length(components).
**		o  cc is the exemplar component.
**		o  width and height pertain to the image cc is drawn from.
**		o  p points to previously initialized edgeOptions{}.
**		o  _out points to the location of the generated image on successful completion.
**
**	Return value:
**		o  An ErrCode;  can fail on allocation of outgoing image.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  return type changed from 'DgpImage_Long' to 'ErrCode'.
**
**/

ErrCode
getSimilarComponents(	Dgp_State		*dgp_state,
						ComponentInfo	*components,
						Dgp_Int32		numccs,
						Dgp_Int32		cc,
						Dgp_Int32		width,
						Dgp_Int32		height,
						edgeOptions		*p,
						DgpImage_Long   *_out) {
	DGP_PROLOGUE("getSimilarComponents()", DGP_PERRSTATE);
	Dgp_Int32			minsize = MAX((Dgp_Int32)(components[cc].size/p->sizeEpsilon), p->minSize);
	Dgp_Int32			maxsize = MIN((Dgp_Int32)(components[cc].size*p->sizeEpsilon), p->maxSize);
	Dgp_Int32			minred = components[cc].color.r - p->colorEpsilon;
	Dgp_Int32			maxred = components[cc].color.r + p->colorEpsilon;
	Dgp_Int32			mingreen = components[cc].color.g - p->colorEpsilon;
	Dgp_Int32			maxgreen = components[cc].color.g + p->colorEpsilon;
	Dgp_Int32			minblue = components[cc].color.b - p->colorEpsilon;
	Dgp_Int32			maxblue = components[cc].color.b + p->colorEpsilon;
	Dgp_Int32			scc;
	DgpImage_Long	out;
	Dgp_Float32		scale = p->scale;
	
	/* Scale width & height. */
	width = (Dgp_Int32)((width - 1) * scale) + 1;
	height = (Dgp_Int32)((height - 1) * scale) + 1;

	/* Get a new image and set all of it to NOT_ASSIGNED. */

	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG | DGPIMAGE_DONTCLEAR, width, height, (void **)&out));
	imSetL(out, NOT_ASSIGNED);		/* From imutil.c / dhtesq. */

	for (scc = 0; scc < numccs; scc++) {
		/* Is it a similar component? */
		if ((components[scc].color.r >= minred) && (components[scc].color.r <= maxred) &&
			(components[scc].color.g >= mingreen) && (components[scc].color.g <= maxgreen) &&
			(components[scc].color.b >= minblue) && (components[scc].color.b <= maxblue) &&
			(components[scc].size >= minsize) && (components[scc].size <= maxsize)) {			
			
			/* This may overwrite some other component.. but it shouldn't happen very often (ugh) */
			/*	...because of the type cast to Dgp_Int32 (?) / dhtesq. */
			/*	Note 'out' contains an index value only, not a ComponentInfo{}.  'out' is not a component map. */
			DGPIMG_REF(out, (Dgp_Int32)(components[scc].centerX*scale), (Dgp_Int32)(components[scc].centerY*scale)) = scc;

			/* Mark as considered. */
			components[scc].userdata = (void *)1;
		}
	}
	*_out = out;
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode getEdges(Dgp_State		*dgp_state,
**								DgpImage_Long		compMap,
**								ComponentInfo	*components,
**								Dgp_Int32				numccs,
**								edgeOptions		*p)
**
**
**  Oversees edge computation for all components in the passed map.  Edges are only computed
**  for subimages of similar components generated by getSimilarComponents().
**
**	The userdata field of cc ComponentInfo{}'s is bashed to mark cc's as 'considered'.
**
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  compMap is the cc map of the image to be analyzed.
**		o  components is a cc array.
**		o  numccs is length(components).
**		o  p points to previously initialized edgeOptions{}.
**
**	Return value:
**		o  An ErrCode.  If failure occurs in the main loop, I will free the list being filled at time of the error,
**		   but previously found edges will remain in place on return and will not be erased.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**		o  Moved list creation to main loop to simplify error handling.
**
**/

ErrCode
getEdges(Dgp_State *dgp_state, DgpImage_Long compMap, ComponentInfo *components, Dgp_Int32 numccs, edgeOptions *p) {
	DGP_PROLOGUE("getEdges()", DGP_PERRSTATE);

	Dgp_Int32			width = DGPIMG_WIDTH(compMap) - 2;	/* compMap has 1 pixel border... */
	Dgp_Int32			height = DGPIMG_HEIGHT(compMap) - 2;
	DgpImage_Long   	similar = NULL;
	Dgp_Int32			cc;
	DgpImage_Binary     tmp = NULL;		/* For debug block. */

  /* validate small pool usage */
  DGP_SMALLPOOL_VALIDATE(edge, EDGE_POOL_ID);

	/* Initialize one list of edges for each component.  Also make sure userdata is NULL because we */
	/* will use it to mark components as considered. */
	for (cc = 0; cc < numccs; cc++) {
		/*	Create a list for this cc (may not be used). */
		DGP_ERR_PASS(DgpAdt_liNew(&(components[cc].edges), dgp_state));
        /*  Null userdata field. */
		components[cc].userdata = NULL;
	}

	for (cc = 0; cc < numccs; cc++) {
		/* Check if this component 1) has not been considered and 2) is within size bounds... */
		if ((components[cc].userdata == NULL) && (components[cc].size >= p->minSize) && 
			(components[cc].size <= p->maxSize)) {
			/* Make image of similar components. */
			DGP_ERR_PASS_CLEANUP(getSimilarComponents(dgp_state, components, numccs, cc, width, height, p, &similar));

			/* Get edges by making Delaunay triangulation. */
			/*	The Delaunay triangulation is built by drawing a line segment between any two sites whose */
			/*	Voronoi regions share an edge. / dhtesq. */
			DGP_ERR_PASS_CLEANUP(getSimilarEdges(dgp_state, similar, components, p->maxDist));

			/* Free the similar component map... */
			DGP_ERR_PASS_CLEANUP(DgpImage_free(similar));
		}
	}
	/* Now make sure both ccs in each edge get a handle to that edge.  So far we've only added edges to the cc occurring */
	/*	at the lower address in memory. */
	DGP_ERR_PASS_CLEANUP(addSecondEdge(dgp_state, components, numccs));

#if defined(DGP_DBUG) && defined(SAVE_INTERMEDIATE_IMAGES)
	/* Save image of the edges -- it looks cool! */
	do {
		DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_BINARY, width, height, (void **)&tmp));
		drawEdges(tmp, components, numccs);
		DGP_ERR_PASS_CLEANUP(DgpImage_save(tmp, "edges.pbm"));
		DGP_ERR_PASS_CLEANUP(DgpImage_free(tmp));
	} while(0);
#endif
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(tmp != NULL){
		DGP_ERR_PASS(DgpImage_free(tmp));
	}
	
	/*	If we failed in the main loop, there will be a list to free.  Maybe an image too. */
	if(cc < numccs){
		DGP_ERR_PASS(DgpAdt_liFree(components[cc].edges));
		
		if(similar != NULL){
			DGP_ERR_PASS(DgpImage_free(similar));
		}
	}

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode freeEdges(	Dgp_State		*dgp_state,
**									ComponentInfo	*components,
**									Dgp_Int32				numccs);
**
**
**  Call this to cleanly get rid of all edges created by getEdges();  lists are also
**  decommissioned.
**
**	The cc userdata field is not touched.  Also, this function assumes one edge list exists for
**	all passed cc's.  Exercise caution when calling this after getEdges() has returned with an error.
**
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  components is the cc array whose edges are to be freed.
**		o  numccs is length(components).
**
**	Return value:
**		o  An ErrCode on list deallocation failure.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**
**/

ErrCode
freeEdges(Dgp_State *dgp_state, ComponentInfo *components, Dgp_Int32 numccs) {
	DGP_PROLOGUE("freeEdges()", DGP_PERRSTATE);
	Dgp_Int32				i;
	DgpAdt_ListNode		ln;

	for (i = 0; i < numccs; i++) {
		DGPADT_FOREACH (ln, components[i].edges) {
			EDGE_FREE(DGPADT_LIGET(ln));
		}
		DGP_ERR_PASS(DgpAdt_liFree(components[i].edges));
	}
    NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void drawEdges(DgpImage_Binary out, ComponentInfo *components, Dgp_Int32 numccs);
**
**
**  Draws a binary image of the passed components;  one line segment will connect
**  the centers of similar components.  Technically, this doesn't draw edges -- vertices
**	lie on component centers
**
**
**	Arguments:
**		o  out is the image to be drawn to.
**		o  components is the cc array whose edges are to be drawn.
**		o  numccs is length(components).
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None. 
**
**/

void
drawEdges(DgpImage_Binary out, ComponentInfo *components, Dgp_Int32 numccs) {
	Dgp_Int32 cc1;
	ComponentInfo *cc2;
	DgpAdt_ListNode ln;

	for (cc1 = 0; cc1 < numccs; cc1++) {
		DGPADT_FOREACH (ln, components[cc1].edges) {
			cc2 = ((edge *)DGPADT_LIGET(ln))->cc;
			drawLineB(out, (Dgp_Int32)components[cc1].centerX, (Dgp_Int32)components[cc1].centerY, 
				(Dgp_Int32)cc2->centerX, (Dgp_Int32)cc2->centerY);
		}
	}
}
/*************************************************************************/
