/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 21.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  fit.c.
**
**
** These are functions to find the best fit line passing through the upper left,
** down (lower) right, or center coordinates of a list of components.
**
** The fit used here is called an 'eigenvector fit.'  It minimizes the sum the of squares of
** the perpendicular distances from the points to the line, as opposed to the normal 
** mse method which minimizes the sum of squares of the vertical or horizontal distances.
** The eigenvector fit is independent of the choice of axis and is particularly useful
** when the error in the measurements is in both the x and y coordinates.
**
**	(This is like the principal axes problem from rigid body dynamics, laminar body case -- dhtesq.)
**
**	Insofar as this module assumes passed lists of ComponentInfo{}'s, generality is
**	poor.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000v
**		o  Already liberr-compliant;  no exceptions to handle.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/* This file has public dependencies via image.h, via ccs.h (at least). dhtesq. */
#include <stdlib.h>
#include <math.h>
#include "list.h"
#include "dgp_state.h"
#include "image.h"
#include "ccs.h"
#include "fit.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/fit.c,v 1.4 2003/02/10 22:53:36 mkrocken Exp $");

#define SQUARE(x) ((x)*(x))

/*************************************************************************
**
**  Function: Dgp_Float64 lineDistance(Dgp_Float64      x, 
**                                     Dgp_Float64      y, 
**                                     Dgp_Float64      xbar, 
**                                     Dgp_Float64      ybar, 
**                                     Dgp_Float64      line_x, 
**                                     Dgp_Float64      line_y)
**
**
**	This function (external) returns the perpendicular distance of point (x, y) from the line containing 
**	the vector (line_x, line_y)	and the point (xbar, ybar).  Although (xbar, ybar) is nominally
**  the cog of some distribution, this will return the correct result for any point (xbar, ybar) on the line.
**
**	This follows from cross product of 1) vector from (xbar, ybar) to (x, y), and 2) vector (line_x, line_y).
**	It seems like d has the wrong dimensionality, but units have disappeared in the normalization of line_n. / dhtesq.
**
**
**	Arguments:
**		o  (x, y) is the point off the line from which the distance is to be calculated.
**		o  (xbar, ybar) is nominally the cog of the point distribution, but may be any point lying on the line.
**		o  (line_x, line_y) is a unit vector lying in the line.
**
**	Return value:
**		o  Double-precision distance.
**
**	Modification history:
**		o  None.
**  
**/

Dgp_Float64
lineDistance(Dgp_Float64        x, 
             Dgp_Float64        y, 
             Dgp_Float64        xbar, 
             Dgp_Float64        ybar, 
             Dgp_Float64        line_x, 
             Dgp_Float64        line_y) {
	Dgp_Float64 xp, yp;
	Dgp_Float64 d;

	xp = x - xbar;
	yp = y - ybar;
	d = ABS(yp*line_x - xp*line_y);	/* ABS via util.h, via image.h.  ABS is also defined in defines.pub. */
	return d;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: static void cogXX(DgpAdt_List ccs, Dgp_Float64 *xbar, Dgp_Float64 *ybar)
**			...where 'XX' is one of 'CE', 'UL', or 'DR'.
**
**
**  Finds center of gravity of the passed connected component list, returned in (xbar, ybar).
**	Function will query the ccs' centerN, minN, or maxN fields for component positions, for cogCE,
**	cogUL, and cogDR, respectively.
**
**
**	Arguments:
**		o  ccs are the cc's whose cog is to be found.
**		o  xbar will point to the x component of the center of mass of the cc's on return.
**		o  ybar will point to the y component of the center of mass of the cc's on return.
**
**	Return value:
**		o  None explicit.
**
**	Modification history:
**		o  None;  might be wise to retype 'length' as a 64-bit Dgp_Int32 or as a Dgp_Float64, though as is it's
**		   not likely to fail in real-world examples with English text.
**  
**/

#define cog(ccs, xbar, ybar, xfield, yfield) \
	DgpAdt_ListNode		ln; \
	Dgp_Float64			xsum, ysum; \
	Dgp_Int32			length; \
	ComponentInfo	    *cc; \
\
	xsum = 0; \
	ysum = 0; \
	length = 0; \
    DGPADT_FOREACH (ln, ccs) { \
		cc = (ComponentInfo *)DGPADT_LIGET(ln); \
		xsum += cc->xfield; \
		ysum += cc->yfield; \
		length++; \
	} \
	*xbar = xsum/length; \
	*ybar = ysum/length;

static void
cogCE(DgpAdt_List ccs, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	cog(ccs, xbar, ybar, centerX, centerY);
}

static void
cogUL(DgpAdt_List ccs, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	cog(ccs, xbar, ybar, minX, minY);
}

/* 'DR' == 'down right' (lower right) / dhtesq. */
static void
cogDR(DgpAdt_List ccs, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	cog(ccs, xbar, ybar, maxX, maxY);
}
/*************************************************************************/

#undef cog

/*************************************************************************
**
**  Function: static void sMatrixXX(DgpAdt_List         ccs, 
**                                  Dgp_Float64         xbar, 
**                                  Dgp_Float64         ybar, 
**                                  Dgp_Float64         *_s11, 
**                                  Dgp_Float64         *_s22, 
**                                  Dgp_Float64         *_s12_s21)
**			...where 'XX' is one of 'CE', 'UL', or 'DR'.
**
**
**  These functions find the sMatrix elements based on center, ul and lr positions in the passed cc's,
**	respectively.  The cog (xbar, ybar) of the cc's must be passed to the function.  On return, the s-matrix
**	looks like
**
**		_s11		_s12_s21
**		_s12_s21	_s22
**
**	where
**		_s11 is the sum of the squares of the ccs' x_distances to the cog,
**		_s22 is the sum of the squares of the ccs' y_distances to the cog, and
**		_s12_s21 is the sum of the products (x_distance)(y_distance) for all cc's.
**
**
**	Arguments:
**		o  ccs are the cc's whose s-matrix is to be calculated.
**		o  (xbar, ybar) is the precomputed cog of the cc list.
**		o  _sxx are the matrix elements that will be filled on successful return.
**
**	Return value:
**		o  None explicit.
**
**	Modification history:
**		o  None.
**  
**/

#define sMatrix(ccs, xbar, ybar, _s11, _s22, _s12_s21, xfield, yfield) \
	Dgp_Float64         xp, yp; \
	Dgp_Float64         s11, s22, s12_s21; \
	DgpAdt_ListNode     ln; \
	ComponentInfo       *cc; \
\
	s11 = 0; \
	s22 = 0; \
	s12_s21 = 0; \
	\
	DGPADT_FOREACH (ln, ccs) { \
		cc = (ComponentInfo *)DGPADT_LIGET(ln); \
		xp = (Dgp_Float64)cc->xfield - xbar; \
		yp = (Dgp_Float64)cc->yfield - ybar; \
		s11 += xp*xp; \
		s22 += yp*yp; \
		s12_s21 += xp*yp; \
	} \
	*_s11 = s11; \
	*_s22 = s22; \
	*_s12_s21 = s12_s21; 

/*	These three find the sMatrix elements based on center, ul, and lr positions in the connected components, */
/*	respectively. / dhtesq. */
static void
sMatrixCE(DgpAdt_List       ccs, 
          Dgp_Float64       xbar, 
          Dgp_Float64       ybar, 
          Dgp_Float64       *_s11, 
          Dgp_Float64       *_s22, 
          Dgp_Float64       *_s12_s21) {
	sMatrix(ccs, xbar, ybar, _s11, _s22, _s12_s21, centerX, centerY);
}

static void
sMatrixUL(DgpAdt_List       ccs, 
          Dgp_Float64       xbar, 
          Dgp_Float64       ybar, 
          Dgp_Float64       *_s11, 
          Dgp_Float64       *_s22, 
          Dgp_Float64       *_s12_s21) {
	sMatrix(ccs, xbar, ybar, _s11, _s22, _s12_s21, minX, minY);
}

static void
sMatrixDR(DgpAdt_List       ccs, 
          Dgp_Float64       xbar, 
          Dgp_Float64       ybar, 
          Dgp_Float64       *_s11, 
          Dgp_Float64       *_s22, 
          Dgp_Float64       *_s12_s21) {
	sMatrix(ccs, xbar, ybar, _s11, _s22, _s12_s21, maxX, maxY);
}
/*************************************************************************/

#undef sMatrix

/*************************************************************************
**
**  Function: static Dgp_Float64 greatestEigenvalue(Dgp_Float64 s11, Dgp_Float64 s22, Dgp_Float64 s12_21)
**
**
**  Returns the greater of the two eigenvalues of the passed 2 x 2 real, Hermitian 
**	(?;  equal off-diagonal elements) matrix:
**
**		_s11		_s12_21
**		_s12_21		_s22
**
**
**	Arguments:
**		o  sxx are the matrix elements.
**
**	Return value:
**		o  Double precision eigenvalue.
**
**	Modification history:
**		o  None.
**  
**/

static Dgp_Float64
greatestEigenvalue(Dgp_Float64 s11, Dgp_Float64 s22, Dgp_Float64 s12_21) {
	Dgp_Float64 lambda;

	/*	Assuming real, Hermitian matrix (?;  equal off-diagonal elements) / dht esq. */
	lambda = (s11 + s22 + (Dgp_Float64)sqrt(SQUARE(s22-s11) + 4*SQUARE(s12_21)))/2;
	return lambda;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: static void eigenvector(Dgp_Float64 s11, Dgp_Float64 s12, Dgp_Float64 lambda, Dgp_Float64 *x, Dgp_Float64 *y)
**
**
**  This computes the eigenvector of the passed matrix corresponding to the passed eigenvalue.
**	In practice, the eigenvalue is returned by greatestEigenvalue().  Only the first row of the matrix
**	is required to calculate the x-component of the vector;  normalization is enforced to compute
**	the y-component.
**
**	The formula for *x follows from the definition of eigenvector;  enforce normalization
**	and substitute for y.  If the passed matrix is the s-matrix computed elsewhere in this module,
**	then the computed eigenvector will be the best-fit line.
**
**	Arguments:
**		o  sxx are the matrix elements, first row.
**		o  lambda is the eigenvalue corresponding to the desired eigenvector.
**		o  (x, y) will point to the normalized eigenvector on return.
**
**	Return value:
**		o  None explicit.
**
**	Modification history:
**		o  None.
**  
**/

static void 
eigenvector(Dgp_Float64 s11, Dgp_Float64 s12, Dgp_Float64 lambda, Dgp_Float64 *x, Dgp_Float64 *y) {
	if ((s11 == lambda) && (s12 == 0)) {
		/*	Avoids division by zero case if matrix is already diagonal / dhtesq. */
		*x = 1;
		*y = 0;
		return;
	}
	*x = s12 / (sqrt(SQUARE(s11-lambda) + SQUARE(s12)));
	*y = sqrt(1 - SQUARE(*x));
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: void eigenLineFitXX(DgpAdt_List ccs, Dgp_Float64 *x, Dgp_Float64 *y, Dgp_Float64 *xbar, Dgp_Float64 *ybar)
**			...where 'XX' is one of 'CE', 'UL', or 'DR'.
**
**
**  Principal entry point to this module.  Places the eigenvector fit and cog for the passed cc's
**	in the passed storage.  The three functions will compute the fit based on the ccs' center, ul,
**	and lr positions, respectively.
**
**
**	Arguments:
**		o  ccs is the component list to be fit.
**		o  (x, y) will on return point to a normalized eigenvector fit to the ccs' positions.
**		o  (xbar, ybar) will on return point to the list's cog.
**
**	Return value:
**		o  Double precision eigenvalue.
**
**	Modification history:
**		o  None.
**  
**/

/*	## is the token-pasting operator, used to concatenate VAR with CE, UL, or DR and yield */
/*	the name of one of the statics in this file. / dhtesq. */
#define eigenLineFit(ccs, x, y, xbar, ybar, VAR) \
	Dgp_Float64 s11, s22, s12_21; \
	Dgp_Float64 lambda; \
\
	cog ## VAR(ccs, xbar, ybar); \
	sMatrix ## VAR(ccs, *xbar, *ybar, &s11, &s22, &s12_21); \
	lambda = greatestEigenvalue(s11, s22, s12_21); \
	eigenvector(s11, s12_21, lambda, x, y);

/* These and lineDistance() are the only externals in the module. / dhtesq. */
void 
eigenLineFitCE(DgpAdt_List ccs, Dgp_Float64 *x, Dgp_Float64 *y, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	eigenLineFit(ccs, x, y, xbar, ybar, CE);
}

void 
eigenLineFitUL(DgpAdt_List ccs, Dgp_Float64 *x, Dgp_Float64 *y, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	eigenLineFit(ccs, x, y, xbar, ybar, UL);
}

void 
eigenLineFitDR(DgpAdt_List ccs, Dgp_Float64 *x, Dgp_Float64 *y, Dgp_Float64 *xbar, Dgp_Float64 *ybar) {
	eigenLineFit(ccs, x, y, xbar, ybar, DR);
}
/*************************************************************************/

#undef eigenLineFit
