
#include "core.h"
#include "periphery.h"

#include <limits.h>
#include "dgp_err_control.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercache/core.c,v 1.23 2003/02/10 22:53:37 mkrocken Exp $");

/* Static functions. */
static ErrCode
suCleanOnceInternal( suXiffCache xc, Dgp_Int32 currentPage,
	suCacheEntry currentEntry, Dgp_Bool *removed);

#ifdef SU_PRINT_STATUS
static void
suPrintStatus( suXiffCache xc, Dgp_Int32 currentPage);
#endif


/* Publicly visible functions as declared in supercache.h. */

ErrCode
suSetDefaults( suOptions *options, Dgp_State *dgp_state) {

DGP_PROLOGUE( "suSetDefaults()", DGP_PERRSTATE);
DGP_STOP_WHINE;

DGP_ASSERT( options != (suOptions *)NULL);

  options->maxMemUsage = SU_DEFAULT_MAX_USAGE;
  
  options->psentry = (memorySentry *)NULL;
  
  options->dgp_state = dgp_state;

  NORMAL_RETURN;
}


ErrCode
suInit( suOptions *options, zephDocumentInfo zephDoc, suXiffCache *xc) {
  _suXiffCache *_xc = (_suXiffCache *)NULL;
  Dgp_State *dgp_state;

DGP_PROLOGUE( "suInit()", NULL);
DGP_ERR_GEN( options == (suOptions *)NULL, cErrInvalidParm);
  dgp_state = options->dgp_state;
DGP_ERR_BEGIN( dgp_state);

  if( options->psentry != (memorySentry *)NULL) { 
DGP_ERR_PASS_CLEANUP(
    memorySentryClaimMemory( options->psentry, sizeof( *_xc), dgp_state));
  }
DGP_CHECK_ALLOC_CLEANUP(
  _xc = (_suXiffCache *)DGP_MALLOC( sizeof( *_xc)));
  _xc->entries = DGP_NULLLIST; 

  /* Copy all of the options over. */
  _xc->options.maxMemUsage = options->maxMemUsage;
  _xc->options.psentry = options->psentry;
  _xc->options.dgp_state = options->dgp_state;

  _xc->currentUsage = sizeof( *_xc);
  _xc->timeToClean = SU_ROUTINE_CLEAN_PERIOD;

  _xc->lastUsedEntry = (_suCacheEntry *)NULL;

  SU_CLAIM_MEMORY( sizeof( struct _DgpAdt_List));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&_xc->entries, dgp_state));

  _xc->currentUsage += sizeof( struct _DgpAdt_List);
  _xc->zephDoc = zephDoc;
  
  *xc = (suXiffCache)_xc;

  SU_DEBUG1( "size of empty cache: %d\n", _xc->currentUsage);

  NORMAL_RETURN;

ERR_CLEANUP:
  if( _xc != (_suXiffCache *)NULL) {
    if( _xc->entries != DGP_NULLLIST) {
      DgpAdt_liFree( _xc->entries);
      SU_RELEASE_MEMORY( sizeof( struct _DgpAdt_List));
    }
    DGP_FREE( (void *)_xc);
    if( options->psentry != (memorySentry *)NULL) {
      memorySentryReleaseMemory( options->psentry, sizeof( *_xc), dgp_state);
    }
  }
  ERR_RETURN;
}

ErrCode
suAddReference( suCacheEntry entry, Dgp_Int32 currentPage) {
  _suCacheEntry *_entry = (_suCacheEntry *)entry;

SU_PROLOGUE_ENTRY( "suAddReference()", _entry);
DGP_ERR_GEN( currentPage < 0, cErrInvalidParm);

  _entry->refCount++;
  _entry->lastPage = currentPage;

  NORMAL_RETURN;
}

ErrCode
suReleaseReference( suCacheEntry entry) {
  _suCacheEntry *_entry = (_suCacheEntry *)entry;

SU_PROLOGUE_ENTRY( "suReleaseReference()", _entry);
DGP_ERR_GEN( _entry == (_suCacheEntry *)NULL, cErrInvalidParm);

  if( _entry->refCount > 0) {
    _entry->refCount--;
  }

  NORMAL_RETURN;
}

void
suRecalculateUsage( suCacheEntry entry) {
  _suXiffCache *_xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry;

  _xc = (_suXiffCache *)_entry->xc;
  suRecalculateUsageInternal( (suXiffCache)_xc, (suCacheEntry)_entry); 
}

Dgp_Int32
suInquireCacheUsage( suXiffCache xc)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  if( _xc == (_suXiffCache *)NULL) {
    return 0;
  } 
  else {
    return _xc->currentUsage;
  }
}

Dgp_Int32
suInquireEntryUsage( suCacheEntry entry)
{
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  if( _entry == (_suCacheEntry *)NULL) {
    return 0;
  } else {
    return _entry->memUsage;
  }
}


ErrCode
suTock( suXiffCache xc)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;

SU_PROLOGUE_XC( "suTock()", _xc);

DGP_ERR_PASS(
  suTockInternal( xc, NULL));

  NORMAL_RETURN;
}

ErrCode
suClean( suXiffCache xc) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;

SU_PROLOGUE_XC( "suClean()", _xc);

  if( _xc->lastUsedEntry != (_suCacheEntry *)NULL) {
DGP_ERR_PASS(
    suCleanInternal( xc, _xc->lastUsedEntry->lastPage, DGP_TRUE, NULL));
  } else {
DGP_ERR_PASS(
    suCleanInternal( xc, 0, DGP_TRUE, NULL));
  }
  
  NORMAL_RETURN;
}

ErrCode
suFlush( suXiffCache xc)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry;
  DgpAdt_ListNode ln;
  Dgp_Bool freed = DGP_FALSE;

SU_PROLOGUE_XC( "suFlush()", _xc);

  do
  {
    ln = DGPADT_LIFIRST( _xc->entries);
    while( ln != DGP_NULLLISTNODE) {
      _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
      
      if( _entry->refCount == 0) {
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_TRUE);
        freed = DGP_TRUE;
        DgpAdt_liRemAndNext( ln, &ln);
      }
      else {
        ln = DGPADT_LINEXT( ln);
      }
    }
  }
  while( freed && DGPADT_LILEN( _xc->entries) > 0);
  
  NORMAL_RETURN;
}

ErrCode
suDestroy( suXiffCache xc) {
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry;
  DgpAdt_ListNode ln;
  Dgp_Int32 remaining = 0;
  memorySentry *psentry;

SU_PROLOGUE_XC( "suDestroy()", _xc);

  while( DGPADT_LILEN( _xc->entries) != 0) {

    ln = DGPADT_LIFIRST( _xc->entries);
    while( ln != DGP_NULLLISTNODE) {
      _entry = (_suCacheEntry *)DGPADT_LIGET( ln);

      if( _entry->refCount == 0) {
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_TRUE);
        DgpAdt_liRemAndNext( ln, &ln);
      }
      else {
        ln = DGPADT_LINEXT( ln);
      }
    }

    /* If we haven't made any progress, then break out. */
    if( remaining == (Dgp_Int32)DGPADT_LILEN( _xc->entries)) {
      break;
    }

    remaining = DGPADT_LILEN( _xc->entries);
  }

  /* If there are still entries remaining, then someone outside must still
   * have a reference, or there is something messed up inside the cache.
   * The latter shouldn't happen, so print out a warning. */
  if( remaining > 0) {
    DGP_WARNF0( "Attempt to destory cache while still holding references.");

    ln = DGPADT_LIFIRST( _xc->entries);
    while( ln != DGP_NULLLISTNODE) {
      _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
      
      /* Mark the entry so that it knows that the cache
       * is being destroyed and that is shouuld be worried about
       * cleanliness. */
      _entry->uses = -1;
      suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_TRUE);
      DgpAdt_liRemAndNext( ln, &ln);
    }
  }  

  _xc->currentUsage -= sizeof( struct _DgpAdt_List) + sizeof( *_xc);
  SU_DEBUG1( "destroyed xc->currentUsage: %d\n", _xc->currentUsage);
DGP_ASSERT( _xc->currentUsage == 0);

  /* Finally, destroy the cache itself. */
  psentry = _xc->options.psentry;
  
  DgpAdt_liFree( _xc->entries);
  DGP_FREE( (void *)_xc);
  if( psentry != (memorySentry *)NULL) {
DGP_ERR_PASS( 
    memorySentryReleaseMemory( psentry, sizeof( struct _DgpAdt_List) + sizeof( *_xc), 
                               dgp_state));
  }

  if( remaining > 0) { 
    DGP_WARNF0( "Cache successfully destroyed anyway.");
  }

  NORMAL_RETURN;
}



/* Internally declared functions (in core.h) for use by cache type modules. */

ErrCode
suGetEntryInternal( suXiffCache xc, suType type, void *identifier, 
                    Dgp_Int32 currentPage, suCacheEntry *entry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  DgpAdt_ListNode ln;

SU_PROLOGUE_XC( "suGetEntryInternal()", _xc);
DGP_ASSERT( identifier != (void *)NULL);
DGP_ASSERT( type != SU_UNKNOWN_TYPE);
DGP_ASSERT( type < SU_NUMBER_OF_TYPES);
DGP_ASSERT( currentPage >= 0);

  SU_DEBUG( "Entering suGetEntryInternal()...\n");

  /* Look and see if this entry is already in the cache. */
  DGPADT_FOREACH( ln, _xc->entries) {
    _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
DGP_ASSERT( _entry != (_suCacheEntry *)NULL);

     SU_DEBUG2( "Comparing types: %d and %d\n", type, _entry->type);

     if( _entry->type == type)
     {
       /* Use the custom comparator. */
       if( (su_entry_ops[type]->equals)( identifier, (suCacheEntry)_entry)) {
         break;
       }
     }
     _entry = (_suCacheEntry *)NULL;
  }

  if( _entry == (_suCacheEntry *)NULL) {
    /* The entry doesn't currently exist in the cache, so ask the entry
     * type module to create it. */
    SU_DEBUG( "Entry not in cache, ask the type module.\n");

DGP_ERR_PASS_CLEANUP(
    (su_entry_ops[type]->get)( xc, identifier, currentPage,
                               (suCacheEntry *)&_entry));
  }
  else {
    SU_DEBUG( "Entry found in cache!\n");
  }

DGP_CHECK_NULL( _entry);

  /* Update entry history. */
  _entry->age = 0;
  _entry->uses++;
  _entry->lastPage = currentPage;
  
  _entry->refCount++;
    
  *entry = (suCacheEntry)_entry;

  /* Should we do a routine cleanup? */
  _xc->timeToClean--;
  if( _xc->timeToClean == 0) {
DGP_ERR_PASS_CLEANUP(
    suCleanInternal( xc, currentPage, DGP_FALSE, *entry));
  }

  NORMAL_RETURN;

ERR_CLEANUP:
  
  ERR_RETURN;
}

ErrCode
suContainsEntryInternal( suXiffCache xc, suType type, void *identifier,
                         Dgp_Bool *contains)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry;
  DgpAdt_ListNode ln;

SU_PROLOGUE_XC( "suGetEntryInternal()", _xc);
DGP_ASSERT( identifier != (void *)NULL);
DGP_ASSERT( type != SU_UNKNOWN_TYPE);
DGP_ASSERT( type < SU_NUMBER_OF_TYPES);

  SU_DEBUG( "Entering suContainsEntryInternal()...\n");

  DGPADT_FOREACH( ln, _xc->entries) {
    _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
DGP_ASSERT( _entry != (_suCacheEntry *)NULL);

     SU_DEBUG2( "Comparing types: %d and %d\n", type, _entry->type);

     if( _entry->type == type)
     {
       /* Use the custom comparitor. */
       if( (su_entry_ops[type]->equals)( identifier, (suCacheEntry)_entry)) {
         SU_DEBUG( "Entry found in cache!\n");
         
         *contains = DGP_TRUE;
         NORMAL_RETURN;
       }
     }
     _entry = (_suCacheEntry *)NULL;
  }

  SU_DEBUG( "Cache does not contain entry.\n");
  
  *contains = DGP_FALSE;
  NORMAL_RETURN;
}

ErrCode
suAddEntryInternal( suXiffCache xc, suCacheEntry entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  DgpAdt_ListNode notused = DGP_NULLLISTNODE;
SU_PROLOGUE_XC( "suAddEntryInternal()", _xc);
DGP_ASSERT( _entry != (_suCacheEntry *)NULL);
DGP_ASSERT( _entry->memUsage > 0);

  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd( _xc->entries, _entry, &notused));

  _xc->currentUsage += _entry->memUsage;
  SU_DEBUG2( "Adding entry of type: %d and size: %d to the cache.\n", 
             _entry->type, _entry->memUsage);

#ifdef SU_PRINT_STATUS
  if( _xc->lastUsedEntry != (_suCacheEntry *)NULL) {
    suPrintStatus( xc, _xc->lastUsedEntry->lastPage);
  } else {
    suPrintStatus( xc, 0);
  }
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

ErrCode
suRemoveEntryInternal( suXiffCache xc, suCacheEntry entry, DgpAdt_ListNode *next)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  DgpAdt_ListNode ln;

SU_PROLOGUE_XC( "suRemoveEntryInternal()", _xc);
DGP_ASSERT( _entry != (_suCacheEntry *)NULL);
DGP_ASSERT( _entry->refCount == 0);

  SU_DEBUG2( "Removing entry of type: %d and size: %d from the cache.\n", 
             _entry->type, _entry->memUsage);

  /* Now remove the entry from the cache DgpAdt_List. */
  DgpAdt_liIsIn( _xc->entries, _entry, &ln);
DGP_ASSERT( ln != DGP_NULLLISTNODE);

  if( next != NULL) {
    DgpAdt_liRemAndNext( ln, next);
  }
  else {
    DgpAdt_liRem( ln);
  }
  
  /* Cleanup the entry on a per-type basis. */
  suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_TRUE);
  
  NORMAL_RETURN;
}


void
suFreeEntryInternal( suXiffCache xc, suCacheEntry entry, Dgp_Bool inCache) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  Dgp_State *dgp_state = _xc->options.dgp_state;

  size_t b;
DGP_VOID_PROLOGUE( "suFreeEntryInternal()", DGP_PERRSTATE);

  if( _entry->refCount > 0) {
    DGP_WARNF0( "Attempting to free referenced entry.");
  }

  if( inCache)
  {
    _xc->currentUsage -= _entry->memUsage;
    
    /* Destroy per-type data. */
    su_entry_ops[ _entry->type]->destroy( xc, entry);
  }

  /* We don't need to free the userdata fields of these DgpAdt_List nodes,
   * since it's only a bunch of pointers to other entries. Hopefully,
   * the individual destroyer function cleaned up as much as
   * possible. */
  if( DGPADT_LILEN( _entry->relations) != 0)
  {
    DGP_WARNF0( "Entry relation DgpAdt_List in inconsistent state; freeing anyway.");
  }

  DgpAdt_liFree( _entry->relations);

  b = sizeof(struct _DgpAdt_List);
  b += sizeof(*_entry);
  _entry->memUsage -= b; /*sizeof( struct _DgpAdt_List) + sizeof( *_entry); */

  SU_DEBUG1( "destroyed entry->memUsage: %d\n", _entry->memUsage);

DGP_VOID_ASSERT( _entry->memUsage == 0);

  DGP_FREE( (void *)_entry);

  SU_RELEASE_MEMORY_VOID( sizeof( *_entry) + sizeof( struct _DgpAdt_List));

  VOID_RETURN;
}

void
suRecalculateUsageInternal( suXiffCache xc, suCacheEntry entry)
{
  _suCacheEntry *_entry = (_suCacheEntry *)entry;

  su_entry_ops[ _entry->type]->calculate( xc, entry);
}

ErrCode
suTockInternal( suXiffCache xc, suCacheEntry entry) {
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry,
    *_otherEntry = (_suCacheEntry *)NULL;
  DgpAdt_ListNode ln;

SU_PROLOGUE_XC( "suTockInternal()", _xc);

  DGPADT_FOREACH( ln, _xc->entries) {
    _otherEntry = (_suCacheEntry *)DGPADT_LIGET( ln);
    if( _otherEntry != _entry) {
      _otherEntry->age++;
    }
  }
  NORMAL_RETURN;
}

ErrCode
suCleanInternal( suXiffCache xc, Dgp_Int32 currentPage, Dgp_Bool aggressive,
                 suCacheEntry currentEntry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  DgpAdt_ListNode ln;
  Dgp_Bool removed;
  Dgp_Int32 nentries = 0;

SU_PROLOGUE_XC( "suCleanInternal()", _xc);
DGP_ASSERT( currentPage >= 0);

#ifdef SU_PRINT_STATUS
  suPrintStatus( xc, currentPage);
#endif

  if( !aggressive ) {
    while( _xc->currentUsage > _xc->options.maxMemUsage ) {

DGP_ERR_PASS_CLEANUP(
      suCleanOnceInternal( xc, currentPage, currentEntry, &removed));

      if( !removed) {
        break;
      }

    } /* end while */
  }
  else /* aggressive */
  {
    Dgp_Int32 i = 0, average = 0;
    nentries = DGPADT_LILEN( _xc->entries);
    if( nentries > 0) {
      /* First find the average value. */
      DGPADT_FOREACH( ln, _xc->entries) {
        _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
        average += (su_valuator)( (suCacheEntry)_entry, currentPage);
        i++;
      }
      
      average /= nentries;
      i = 0;
      
      SU_DEBUG1( "aggressive clean: average value = %d\n", average);

      /* Now throw out anything that has less that the average value. */
      ln = DGPADT_LIFIRST( _xc->entries);
      while( ln != DGP_NULLLISTNODE) {
        Dgp_Int32 value;
        _entry = (_suCacheEntry *)DGPADT_LIGET( ln);
        value = (su_valuator)( (suCacheEntry)_entry, currentPage);

        if( (value <= average) &&
            (_entry != (_suCacheEntry *)currentEntry) &&
            (_entry->refCount == 0) )
        {
          DgpAdt_ListNode next = DGPADT_LINEXT( ln); 
DGP_ERR_PASS_CLEANUP(
          suRemoveEntryInternal( xc, (suCacheEntry)_entry, &ln));
          ln = next;
          i++;
        }
        else {
          ln = DGPADT_LINEXT( ln);
          i++;
        }
      }
      
      /* Just make sure that we're under the limit. */
      DGP_ERR_PASS_CLEANUP(
    suCleanInternal( xc, currentPage, DGP_FALSE, currentEntry));
    }

  } /* end aggressive */

  _xc->timeToClean = SU_ROUTINE_CLEAN_PERIOD;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


ErrCode
suInitEntryInternal( suXiffCache xc, suType type, suCacheEntry *entry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;

SU_PROLOGUE_XC( "suInitEntry()", _xc);
DGP_ASSERT( type != SU_UNKNOWN_TYPE);
DGP_ASSERT( type < SU_NUMBER_OF_TYPES);
  
  SU_CLAIM_MEMORY( sizeof( *_entry) + sizeof( struct _DgpAdt_List));
DGP_CHECK_ALLOC_CLEANUP(
  _entry = (_suCacheEntry *)DGP_MALLOC( sizeof( *_entry)));
  _entry->relations = DGP_NULLLIST; 

  _entry->type = type;
  _entry->data = (void *)NULL;
  
  _entry->age = 0;
  _entry->uses = 0;
  _entry->lastPage = -1;
  
  _entry->memUsage = sizeof( *_entry);
  _entry->refCount = 0;

  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&_entry->relations, dgp_state));

  _entry->memUsage += sizeof( struct _DgpAdt_List);
  _entry->xc = xc;

  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( _entry != (_suCacheEntry *)NULL) {
    if( _entry->relations != DGP_NULLLIST) {
      DgpAdt_liFree( _entry->relations);
      SU_RELEASE_MEMORY_VOID( sizeof( *_entry));
    }
    DGP_FREE( (void *)entry);
    SU_RELEASE_MEMORY_VOID( sizeof( *_entry));
  }

  ERR_RETURN;
}


static ErrCode
suCleanOnceInternal( suXiffCache xc, Dgp_Int32 currentPage, 
                     suCacheEntry currentEntry, Dgp_Bool *removed)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  DgpAdt_ListNode ln, minLn = DGP_NULLLISTNODE;
  Dgp_Int32 val, minVal = INT_MAX;

SU_PROLOGUE_XC( "suCleanOnceInternal()", _xc);

  *removed = DGP_FALSE;

  DGPADT_FOREACH( ln, _xc->entries) {
    _entry = (_suCacheEntry *)DGPADT_LIGET( ln);

    /* Make sure that there are no user references to this entry and
     * that no other entries make reference to it. */
    if( _entry->refCount == 0 )
    {
      /* Don't throw out the current entry! */
      if( _entry != (_suCacheEntry *)currentEntry)
      {
        val = (su_valuator)( (suCacheEntry)_entry, currentPage);

        if( val < minVal)
        {
          minVal = val;
          minLn = ln;
        }
      }
    }
  }

  /* Not unless all entries have extremely high values, or if all
   * the entries have a reference count greater than zero. */
  if( minLn == DGP_NULLLISTNODE) {
    NORMAL_RETURN;
  }

  _entry = (_suCacheEntry *)DGPADT_LIGET( minLn);
    
  /* Got a candidate, now remove it. */
  *removed = DGP_TRUE;

DGP_ERR_PASS(
  suRemoveEntryInternal( xc, (suCacheEntry)_entry, NULL));

#ifdef SU_PRINT_STATUS
  suPrintStatus( xc, currentPage);
#endif

  NORMAL_RETURN;
}

#ifdef SU_PRINT_STATUS

static void
suPrintStatus( suXiffCache xc, Dgp_Int32 currentPage)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  DgpAdt_ListNode ln;

  fprintf( stderr, "supercache status:  %d entries  %d bytes usage\n",
           DGPADT_LILEN( _xc->entries), _xc->currentUsage);
  fprintf( stderr, "        type    age     uses    last    mem     refs    rels    value\n");
  fprintf( stderr, "        ----    ---     ----    ----    ---     ----    ----    -----\n");

  DGPADT_FOREACH( ln, _xc->entries) {
    _entry = (_suCacheEntry *)DGPADT_LIGET( ln);


    fprintf( stderr, "\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
            _entry->type, _entry->age, _entry->uses, _entry->lastPage,
            _entry->memUsage, _entry->refCount, DGPADT_LINLEN( _entry->relations),
            (su_valuator)( (suCacheEntry)_entry, currentPage));
  }
}

#endif
