
#ifndef CORE_H
#define CORE_H

#include "sucache_t.h"
#include "list.h"		/* DgpAdt_List */
#include "util.h"		/* Dgp_Bool */

#include "memory_sentry_t.h"
#include "zephdocument_t.h"


/* Supercache debugging macros. */

#ifdef SU_ENABLE_DEBUG
#define SU_DEBUG( x) 		    fprintf( stderr, "xc: "); printf( x);
#define SU_DEBUG1( x, y) 	    fprintf( stderr, "xc: "); printf( x, y);
#define SU_DEBUG2( x, y, z) 	fprintf( stderr, "xc: "); printf( x, y, z);
#define SU_DEBUG3( x, y, z, w) 	fprintf( stderr, "xc: "); printf( x, y, z, w);

#define SU_PRINT_STATUS

#else
#define SU_DEBUG( x)
#define SU_DEBUG1( x, y)
#define SU_DEBUG2( x, y, z)
#define SU_DEBUG3( x, y, z, w)

#endif /* SU_ENABLE_DEBUG */


/* Used to extend the Digipaper function prologue macros. */
#define SU_PROLOGUE_XC( funct_name, _xc) \
    Dgp_State *dgp_state; \
    DGP_PROLOGUE( funct_name, NULL); \
    DGP_ERR_GEN( (_xc) == (_suXiffCache *)NULL, cErrInvalidParm); \
    dgp_state = (_xc)->options.dgp_state; \
    DGP_ERR_BEGIN( dgp_state)

#define SU_PROLOGUE_ENTRY( funct_name, _entry) \
    Dgp_State *dgp_state; \
    DGP_PROLOGUE( funct_name, NULL); \
    DGP_ERR_GEN( (_entry) == (_suCacheEntry *)NULL, cErrInvalidParm); \
    DGP_ERR_GEN( (_entry)->xc == (suXiffCache)NULL, cErrInvalidParm); \
    dgp_state = ((_suXiffCache *)(_entry)->xc)->options.dgp_state; \
    DGP_ERR_BEGIN( dgp_state)

/* For dealing with the memory sentry. */

#define SU_CLAIM_MEMORY( nbytes) \
  if( _xc->options.psentry != (memorySentry *)NULL) { \
    SU_DEBUG3( "CLAIM: %d at line %d of %s\n", nbytes, __LINE__, \
               __FILE__); \
    DGP_ERR_PASS_CLEANUP( memorySentryClaimMemory( _xc->options.psentry, \
                                                   nbytes, dgp_state)); \
  }

#define SU_RELEASE_MEMORY( nbytes) \
  if( _xc->options.psentry != (memorySentry *)NULL) { \
    SU_DEBUG3( "RELEASE: %d at line %d of %s\n", nbytes, __LINE__, \
               __FILE__); \
    DGP_ERR_PASS_CLEANUP( memorySentryReleaseMemory( _xc->options.psentry, \
                                                     nbytes, dgp_state)); \
  }

#define SU_RELEASE_MEMORY_VOID( nbytes) \
  if( _xc->options.psentry != (memorySentry *)NULL) { \
    SU_DEBUG3( "RELEASE: %d at line %d of %s\n", nbytes, __LINE__, \
               __FILE__); \
    memorySentryReleaseMemory( _xc->options.psentry, nbytes, dgp_state); \
  }

/* The next few defines and typedefs create a uniform interface to all
 * the different type modules that may exist in the cache. */
#define SU_UNKNOWN_TYPE		0

typedef Dgp_UInt8 suType;

/*
 * Return DGP_TRUE if two entries (the given entry and the desired one
 * denoted by "identifier") are equivalent.
 * This function will only be called if the two entries have the same type.
 *
 * Preconditions:
 * - "identifer" is non-null.
 * - "entry" is a valid cache entry.
 */
typedef Dgp_Bool(*suEntryComparitor)( void *identifier, suCacheEntry entry);

/*
 * Given an identifier create a new entry. When this function is called
 * the desired entry is guarenteed not to exist in the cache.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "identifer" is non-null.
 * - "currentPage" is non-negative.
 *
 * Postconditions:
 *  - the following need not be set (and in fact will be bashed):
 *    o age
 *    o uses
 *    o lastPage
 *    o refCount
 *  - "entry" is a member of "xc->entries".
 *  - "xc->currentUsage" reflects "entry->memUsage"
 *
 * As noted, this function should add the new entry to the cache's DgpAdt_List
 * and update the cache's memory usage. The function suAddEntryInternal
 * is provided to accomplish these two tasks.
 */
typedef ErrCode(*suEntryGetter)( suXiffCache xc, void *identifier,
		Dgp_Int32 currentPage, suCacheEntry *entry);

/*
 * Should free any data specific to the entry type.
 *
 * This function should _not_ check the refCount field of the entry.
 * Either the core functions have already checked this value, or the
 * whole cache is going down and it doesn't really matter.
 *
 * NB: If the uses count of the current entry is negative then the
 * cache is being destroyed and the the destroy function shouldn't be
 * worried about keeping things clean, as long as "entry->data" is freed.
 * In fact, it shouldn't be surprised in this case if the state is a
 * little screwy (i.e. some of the entries it depends on have been freed.)
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "entry" is a valid suCacheEntry.
 *
 * Postconditions:
 * - "entry->data" and all structures hanging of "data" should be
 *      freed.
 * - "entry" should NOT be freed.
 * - "xc->currentUsage" should be left UNCHANGED.
 * - "xc->entries" should be left UNCHANGED.
 * - if "entry->uses" >= 0 (when entering the function) then
 *      o all items should be removed from "entry->relations"
 * - "entry->relations" should be a valid DgpAdt_List.
 */
typedef void(*suEntryDestroyer)( suXiffCache xc, suCacheEntry entry);

/*
 * This function requires entries to (re)calculate the amount of memory
 * that they are currently holding on to. Though cache data is usually
 * treated as immutable, there are cases where applications may make small
 * "adjustments" to such data. This function should scan through whatever
 * data is contained by this entry and recompute its current usage. Each
 * entry should propagate calls to "suEntryCalculator" to other entries
 * whose data may be accessible from the current entry. These other entries
 * will quite often be exactly the dependencies of the current entry.
 * (For example, a positions block entry should call the calculator of each
 * token entry that it references.)
 */
typedef void(*suEntryCalculator)( suXiffCache xc, suCacheEntry entry);


/* Assesses a value for a particular entry. Entries with high values
 * are kept longer than those with lower values. */
typedef Dgp_Int32(*suValuator)( suCacheEntry entry, Dgp_Int32 currentPage);

typedef struct {
  suEntryComparitor equals;
  suEntryGetter get;
  suEntryDestroyer destroy;
  suEntryCalculator calculate;   
} _suEntryOps;

typedef _suEntryOps * suEntryOps;

typedef struct {
  
  /* Data */
  suType type;
  void *data;
  
  /* History */
  Dgp_Int32 age;
  Dgp_Int32 uses;
  Dgp_Int32 lastPage;
  
  /* Usage */
  Dgp_UInt32 memUsage;
  Dgp_Int32 refCount;

  /* (inter)Dependencies */
  DgpAdt_List relations;
  suXiffCache xc;
} _suCacheEntry;


typedef struct {
  DgpAdt_List entries;
  Dgp_UInt32 currentUsage;
  Dgp_Int16 timeToClean;

  _suCacheEntry *lastUsedEntry;
  
  suOptions options;
  zephDocumentInfo zephDoc;
} _suXiffCache;


#define isUnknownType( entry) ((entry)->type == SU_UNKNOWN_TYPE)

#define hasRelations( entry) (liLen( (entry)->relations) > 0)


/*
 * When requesting an entry from inside supercache code, this is the
 * way to go. It checks the cache first and then asks the type
 * module to create a new one if necessary. Don't ever call the
 * "getter" function (in su_entry_ops) directly.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "type" is a known type with appriate operators defined in
 *      periphery.h.
 * - "identifier" is a valid and appropriate identifer for "type"
 * - "currentPage" is non-negative.
 *
 * Postconditions:
 * - "entry" is a pointer to a valid and complete cache entry.
 */
extern ErrCode
suGetEntryInternal( suXiffCache xc, suType type, void *identifier,
                    Dgp_Int32 currentPage, suCacheEntry *entry);

/*
 * Queries the cache to see if it contains a particular entry.
 * Sets "*contains" to DGP_TRUE if the cache contains a matching
 * entry.
 */
extern ErrCode
suContainsEntryInternal( suXiffCache xc, suType type, void *identifier,
                         Dgp_Bool *contains);

/*
 * Adds a new entry to the cache.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "entry" is a valid suCacheEntry.
 * - "entry" is not yet a member of "xc->entries".
 * - "entry->mamUsage" is positive.
 *
 * Postconditions: (on success)
 * - "entry" is a member of "xc->entries".
 * - "xc->currentUsage" has been appropriately updated.
 */
extern ErrCode
suAddEntryInternal( suXiffCache xc, suCacheEntry entry);

/*
 * Both removes an entry from the cache and frees
 * any memory associated with the entry. Also cleans
 * up a bunch of cache state.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "entry" is a valid suCacheEntry.
 * - "entry->refCount" == 0;
 *
 * Postconditions: (on success)
 * - "entry" and all data hanging from it has been freed.
 * - "entry" is no longer a member of "xc->entries".
 * - "xc->currentUsage" is up to date.
 * - if "next" was non-NULL then "next" points to the next
 *      member of "xc->entries".
 */
extern ErrCode
suRemoveEntryInternal( suXiffCache xc, suCacheEntry entry, DgpAdt_ListNode *next);

/*
 * Frees an entry. Never call the "destroyer" function
 * in su_entry_ops directly, rather, use this function.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "entry" is a valid suCacheEntry.
 * - "entry->refCount" == 0;
 * - "inCache" is DGP_TRUE iff this entry has been successfully added to the cache.
 *
 * Postconditions: (on success)
 * - "entry" and all data hanging from it has been freed.
 */
extern void
suFreeEntryInternal( suXiffCache xc, suCacheEntry entry, Dgp_Bool inCache);

/*
 * Call calculate usage on the given entry.
 */
extern void
suRecalculateUsageInternal( suXiffCache xc, suCacheEntry entry);

/*
 * See the comment for suTock in supercache.h. This
 * function increments the age for all entries except
 * "entry" if "entry" is non-null.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 */
extern ErrCode
suTockInternal( suXiffCache xc, suCacheEntry entry);

/*
 * See suClean in supercache.h. This function will never throw out
 * "currentEntry". Non-aggressive cleaning means that the cache
 * will only throw out entries in order to meet memory use
 * restrictions. Aggressive cleaning means that the cache may take
 * more time and throw out anything that's old.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "currentPage" is non-negative.
 */
extern ErrCode
suCleanInternal( suXiffCache xc, Dgp_Int32 currentPage,
                 Dgp_Bool aggressive, suCacheEntry currentEntry);

/*
 * Initializes a new cache entry.
 *
 * Preconditions:
 * - "xc" is a valid suXiffCache.
 * - "type" is a known type with appriate operators defined in
 *      periphery.h.
 *
 * Postconsditions: (on success)
 * - "entry" points to a new allocated and initialized entry.
 * - "entry->type" == "type"
 * - "entry->relations" is a valid (and empty) DgpAdt_List.
 * - the remainder of the fields of "entry" are initialized to
 *      default values.
 */
extern ErrCode
suInitEntryInternal( suXiffCache xc, suType type, suCacheEntry *entry);

#endif /* CORE_H */
