/* $Header: /project/icons/cvsroot/digipaper/src/libsupercache/libsupercache.h,v 1.5 2003/02/10 22:53:37 mkrocken Exp $ */

#ifndef LIBSUPERCACHE_H
#define LIBSUPERCACHE_H

/*
 * libsupercache currently defines one caching data type. that is,
 * suXiffCache. This data structure caches dictionary/token and
 * related info. In fact, this caching mechinism was designed to
 * cache just about any arbirary data type. This leads to two
 * distinct sets of functions, Core and Periphery, described
 * below.
 *
 * The basic functionality of the cache goes as follows:
 *  - you (the user of the cache) ask for some piece of data.
 *  - the cache looks around and finds this piece of data. it
 *      might be in the cache or maybe not. the cache may have
 *      to make calls into libtokcomp or io libs. that's life.
 *  - the cache keeps an extra reference to this piece of data.
 *  - it also store information about the last time this piece
 *      of data was used, how many times it has been used, etc.
 *      in all, a bunch of information that can be used to
 *      assign a numerical _value_ to all the data in the cache.
 *  - finally the cache gives you the piece of data you asked
 *      for. it remembers that you have it. when you are done
 *      with it, you should "release" your reference to it.
 *      that way the cache knows that you don't expect it
 *      to be around very much longer.
 *
 * This is all good you say, but what's this whole "value"
 * business? And what is this field called "maxMemUsage" in the
 * suOptions struct? Well, the in deciding what to throw out and
 * what to keep, the cache _tries_ to limit the memory usage of
 * the data inside of it. So when ever you call suClean it assesses
 * the _value_ of each piece data in the cache. If we are over the
 * limit and there is some piece of data that is not being
 * referenced, then this data will be destroyed. If there are many
 * such candidates, then the one with the lowest value is thrown out
 * first.
 *
 * The cache can't always keep memory usage below the desired
 * level though. If you see your usage start to explode, check and
 * make sure that you are releasing all of your references.
 *
 * It gets a little more complicated in other ways too. For
 * instance, you are not the only thing that can  reference data-
 * data within the cache can also hold references to other pieces
 * of data. But that's beyond the scope of what I'm talking about
 * here. See core.h for details.
 *
 * A note on memory ownership. Initially you need to allocate a
 * suOptions struct. You own the struct and must free it after suInit.
 * All suXiffCache and suCacheEntry structs are owned by the cache. Also 
 * just about anything else you get back from the cache continues to be 
 * under the control of the cache, 'cause, I mean, hey, it's a cache,
 * right? The sole exception here is Claim. Obviously anything returned
 * from a Claim function is owned by the caller.
 */

/* Supercache type definitions. */
#include "sucache_t.h"
#include "sutokdict_t.h"
#include "sutokposblock_t.h"

/*
 * Like I said, there are basically two types of functions.
 * There are those that are independent of the actual types of
 * data stored in the cache, and there are those that do
 * depend on these types. The first set are called the "Core"
 * functions. These functions deal with day-to-day cache
 * maintainence and cleanup.
 *
 * The second type, the "Periphery," provides specific
 * functionality with respect to specific types. See below.
 */

/* The Core Cache Functions */
#include "sucache.h"

/*
 * The following structure and definitions are used to keep track of
 * color image tokens. The "color_token" struct is analogous to the
 * "token" struct in libtokcomp. Here the image may be represented in one
 * of several forms, however. You _may_ change the format and resolution
 * of these structs _even_after_ calling the Extract method (and not the
 * Claim). However, all images must be in one of the formats listed below
 * and "imagetype" must reflect the current format. The reason for this
 * exception (to the Extract/Claim rule) is that these images may be returned
 * in PIXR format, which is not very useful to blatters. So in order to
 * make the cache a useful tool for color tokens, users can convert the
 * format of the image. This means though that if users do, in fact, change
 * the type, then they should be able to handle both types.
 *
 * (So without a policy about modifying cache-owned data, in this case,
 * everyone would just Claim and that would be the end of it.)
 */
#define SU_COLOR_TOKEN_IMAGETYPE_PIXR		1
#define SU_COLOR_TOKEN_IMAGETYPE_RGBIMAGE	2


/* The Periphery Functions */

/*
 * Notes in general:
 *
 * By calling a Get function below you are implicitly Adding a
 * reference to the suCacheEntry that is returned. You must
 * later release this reference by calling suReleaseReference.
 * (number of Gets) + (number of Adds)
 *	== (number of Releases) + (number of Claims)
 *
 * When calling an Extract function note that you CANNOT
 * release the associated entry until you are finished with
 * whatever information is returned from Extract. For example
 * the following code is (very) unsafe:
 * 	suGetTokenDictionary( xc, i, uids, n, &entry);
 * 	suExtractTokenData( entry, &my_toks,...);
 * 	suReleaseReference( entry);
 *	...
 *      /* bad, because  my_toks may have been freed */
 *	if( my_toks[j].committed)... 
 * Short story: as soon as you release the last reference to
 * an entry, that entry _and_ any info Extracted from it may
 * be freed at any time!
 */
#include "sutokdict.h"
/*#include "sujb2SymbolInstanceblock.h" */

#endif /* LIBSUPERCACHE_H */
