/* $Header: /project/icons/cvsroot/digipaper/src/libsupercache/sucache.h,v 1.1 2000/05/18 19:46:58 wbell Exp $ */

#ifndef SUCACHE_H
#define SUCACHE_H

#include "sucache_t.h"
#include "zephdocument_t.h"

#ifdef __cplusplus
extern "C"
{
#endif


/*
 * Set the default options. 
 *
 * Preconditions:
 *  - "options" has been allocated.
 *  - "dgp_state" is a valid Digipaper state cookie.
 */
extern ErrCode
suSetDefaults( suOptions *options, Dgp_State* dgp_state);


/*
 * Allocates and initializes the internal structures used
 * by the cache. This function will immediately allocate a
 * new options structure and copy over all the parameters.
 * So the caller should free the set of options passed in
 * after the call is completed.
 * 
 * Preconditions:
 *  - "options" is a valid set of options.
 *  - "zephDoc" is the valid state of an open document.
 *
 * Postconditions:
 *  - on success, "xc" points to an initialized suXiffCache.
 */
extern ErrCode
suInit( suOptions *options, zephDocumentInfo zephDoc, suXiffCache *xc);


/*
 * Adds a reference to an existing cache entry.
 *
 * Preconditions:
 *  - "entry" is a valid suCacheEntry.
 *  - "currentPage" is non-negative.
 */
extern ErrCode
suAddReference( suCacheEntry entry, Dgp_Int32 currentPage);


/*
 * Removes a reference from an existing cache entry.
 *
 * Preconditions:
 *  - "entry" is a valid suCacheEntry.
 */
extern ErrCode
suReleaseReference( suCacheEntry entry);

/*
 * Recalculate memory usage for this entry and its dependencies.
 */
extern void
suRecalculateUsage( suCacheEntry entry);

/*
 * Returns the amount of memory (in bytes) used by
 * the cache, its entries and the various data structures
 * hanging off these entries.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 */
extern Dgp_UInt32
suInquireCacheUsage( suXiffCache xc);


/*
 * Returns the amount of memory (in bytes) used by this entry
 * and the various data structures hanging off of it.
 *
 * Preconditions:
 *  - "entry" is a valid suCacheEntry.
 */
extern Dgp_UInt32
suInquireEntryUsage( suCacheEntry entry);


/*
 * Increment the age of all entries currently in the cache.
 * Older enties are thrown out before newer ones.
 *
 * This function is entirely optional. See suClean for details.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 */
extern ErrCode
suTock( suXiffCache xc);


/*
 * A note on cleaning. Even without calling suClean, the cache
 * will limit memory usage. If you are especially worried about
 * usage than you can call suClean. suClean is very aggressive and
 * will probably free even more entries than necessary to get below
 * the memory restriction. In fact, suClean will throw out anything
 * it considers to be "old"- probably around half the stuff in the
 * cache.
 *
 * You can make entries get "older" faster by calling suTock at
 * appropriate times. Therefore the most aggressive cleaning
 * will be done by the cache when suTock is called often, along
 * with periodic calls to suClean.
 *
 * On the other hand, in order to achieve the most passive cleaning
 * (i.e. staying just below the limit), you don't need to call either
 * suTock or suClean. This cleaning will be done automatically.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 */
extern ErrCode
suClean( suXiffCache xc);

/*
 * Removes everything from the cache without a reference. That is,
 * this function will make the cache as small as possible, given
 * the current outstanding references.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 */
extern ErrCode
suFlush( suXiffCache xc);


/*
 * Frees the cache along with all of its entries and all
 * of the data associated with these entries. suDestroy
 * is not clean- it ignores any references held to entries
 * in this cache and frees everything. It will however 
 * print a warning if an attempt is made to destroy the 
 * cache while still holding on to references to entries
 * still inside the cache.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 */
extern ErrCode
suDestroy( suXiffCache xc);

#ifdef __cplusplus
}
#endif

#endif /* SUCACHE_H */
