/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 31 2000
**
**  Filename: colorCCfill.c
**
**  Description of what's implemented in this file:
**    functions to do fill detection during color cc finding.
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCfill.c,v 1.13 2003/02/10 22:53:38 mkrocken Exp $
 *
 */

#include "colorCC-priv.h"
#include "dgp_utils.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCfill.c,v 1.13 2003/02/10 22:53:38 mkrocken Exp $");


#define FILL_MERGE_PERCENT 40

#define ccstate_ccstatesMatch(ccstate1, ccstate2) \
  ((((ccstate1)->color == (ccstate2)->color) && \
    ((ccstate_isNormal(ccstate1) || ccstate_isBackground(ccstate1)) && \
     (ccstate_isNormal(ccstate2) || ccstate_isBackground(ccstate2)))) || \
  (ccstate_isSampled(ccstate1) && ccstate_isSampled(ccstate2)) || \
  (ccstate_isWash(ccstate1) && ccstate_isWash(ccstate2)))


static ErrCode
dgpcc_fillCheck(CCstate *outer_ccstate, 
                CCstate *ccstate,
                CCstate *inner_ccstate,
                Dgp_Bool may_delay,
                colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_fillCheck()", cookie->DGP_PERRSTATE);

  Dgp_UInt32 area_percent;
  cccMapRowData *prow_data;


  /* this should be true by the time we get here */
  DGP_ASSERT(inner_ccstate->num_potential_fills == 1);

  /* determine percent of inner ccstate's area to outer ccstate's area */
  area_percent = ((inner_ccstate->lrX - inner_ccstate->ulX + 1) *
                  (inner_ccstate->lrY - inner_ccstate->ulY + 1) * 100) / 
                 ((ccstate->lrX - ccstate->ulX + 1) *
                  (ccstate->lrY - ccstate->ulY + 1));

  if(area_percent >= FILL_MERGE_PERCENT) {

    /* in this case, we'd like to merge the inner ccstate with the ccstate
       outside of the current outer ccstate, however, this is a bit tricky */
    
    /* if there isn't an outer_ccstate yet, and we are allowed to wait for
       one, just return. */
    if((outer_ccstate == NULL) && (may_delay)) NORMAL_RETURN;

    /* at this point, if outer_ccstate is still NULL, there is *no* 
       outer_ccstate, so suck up and deal.  otherwise, check to
       see if the inner and outer ccstates match */
    if((outer_ccstate != NULL) &&
       ccstate_ccstatesMatch(outer_ccstate, inner_ccstate)) {

      /* merge the inner and outer ccstates */
      DGP_ERR_PASS(ccstate_merge(outer_ccstate, &inner_ccstate, NULL, NULL, cookie));

      /* we're done */
      NORMAL_RETURN;
    }

  }
  
  prow_data = cccmap_getRowData(cookie->cccmap, inner_ccstate->map_index, ccstate->map_index);
  
  /* if there is still the potential of a wash between these two ccstates,
     we can't underfill yet */
  if(!cccmaprd_isPotentialWash(prow_data)) {

#if 0
    /* okay, we're not going to merge the inner ccstate with the outer ccstate, so
       we'll just fill underneath the inner_ccstate */
    DGP_ERR_PASS(ccstate_underFill(ccstate, inner_ccstate, cookie));
#endif

    /* clear potential fill status */
	  cccmap_updateNeighborNoPotentialFill(cookie->cccmap, prow_data, inner_ccstate->map_index, 
      ccstate->map_index, inner_ccstate, ccstate);

  }

  NORMAL_RETURN;
}


static ErrCode
_dgpcc_doFillHoles1(CCstate *way_outer_ccstate,
                    CCstate *ccstate,
                    Dgp_Int32 cur, 
                    Dgp_Int32 first, 
                    Dgp_Int32 last, 
	                  cccMapColData *col,
                    Dgp_Bool may_delay,
                    colorCC_state *cookie);


static INLINE ErrCode
dgpcc_doFillHoles1(CCstate *way_outer_ccstate,
                   CCstate *ccstate, 
                   Dgp_Int32 cur, 
                   Dgp_Int32 first, 
                   Dgp_Int32 last, 
	                 cccMapColData *col,
                   Dgp_Bool may_delay,
                   colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_doFillHoles1()", NULL);
  Dgp_Int32 j, num_potential_fills;
  /*RunImage *rim = cookie->rim; */
  cccMapRowData *row;
  CCstate *inner_ccstate, *outer_ccstate = NULL;
  DgpAdt_ListNode ln;
  DgpAdt_List inner_ccstates = DGP_NULLLIST;
  /*Dgp_Bool ccstate_merged = DGP_FALSE; */
  

  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&inner_ccstates, cookie->dgp_state));
  num_potential_fills = ccstate->num_potential_fills;

  /* search for potential fill relationships */
  row = col[cur].row;
  for(j = first; j < cur; j++) {
    if(cccmaprd_isPotentialFill(&row[j])) {

      if(!cccmaprd_isPotentialFillConflict(&row[j])) {

        /* figure out which is outer and which is inner */
        if(col[j].ccstate->ulX > ccstate->ulX) {
          DGP_ASSERT(col[j].ccstate->lrX < ccstate->lrX);
          DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(inner_ccstates, col[j].ccstate, NULL));
        } else {
          DGP_ASSERT(col[j].ccstate->lrX > ccstate->lrX);
          outer_ccstate = col[j].ccstate;
        }

      } else {

        /* for now, we're just going to ditch these, because there are many
           difficulties in actually determining if there is a problem or not */
        cccmaprd_clearPotentialFill(&row[j]);
        cccmaprd_clearPotentialFillConflict(&row[j]);
        ccstate->num_potential_fills--;
        col[j].ccstate->num_potential_fills--;

      }

      if(--num_potential_fills == 0) goto FOUND_ALL_FILLS;
    }
  }

  for(++j; j <= last; j++) {
    if(cccmaprd_isPotentialFill(&col[j].row[cur])) {

      if(!cccmaprd_isPotentialFillConflict(&col[j].row[cur])) {

        /* figure out which is outer and which is inner */
        if(col[j].ccstate->ulX > ccstate->ulX) {
          DGP_ASSERT(col[j].ccstate->lrX < ccstate->lrX);
          DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(inner_ccstates, col[j].ccstate, NULL));
        } else {
          DGP_ASSERT(col[j].ccstate->lrX > ccstate->lrX);
          outer_ccstate = col[j].ccstate;
        }

      } else {

        /* for now, we're just going to ditch these, because there are many
           difficulties in actually determining if there is a problem or not */
        cccmaprd_clearPotentialFill(&col[j].row[cur]);
        cccmaprd_clearPotentialFillConflict(&col[j].row[cur]);
        ccstate->num_potential_fills--;
        col[j].ccstate->num_potential_fills--;

      }

      if(--num_potential_fills == 0) goto FOUND_ALL_FILLS;

    }
  }

  /* jump here after we've found all of our fill neighbors */
FOUND_ALL_FILLS:

  /* recursively fill in holes inside this ccstate */
  DGPADT_FOREACH(ln, inner_ccstates) {
    inner_ccstate = (CCstate *)DGPADT_LIGET(ln);

    if(inner_ccstate->num_potential_fills > 1) {

      /* the inner ccstate has some holes in it, so recurse on this inner ccstate */
      DGP_ERR_PASS_CLEANUP(_dgpcc_doFillHoles1(outer_ccstate, inner_ccstate, inner_ccstate->map_index,
        first, last, col, DGP_FALSE, cookie));

    } else {

      /* this should be true, i think... */
      DGP_ASSERT(inner_ccstate->num_potential_fills == 1);

      /* if all the holes in this hole have been taken care of, 
         fill check this one */
      DGP_ERR_PASS_CLEANUP(dgpcc_fillCheck(outer_ccstate, ccstate, inner_ccstate, DGP_FALSE, cookie));

    }
  }

  /* done with the list */
  DgpAdt_liFree(inner_ccstates);
  inner_ccstates = DGP_NULLLIST;

  /* we can only fill check this ccstate and it's outer_ccstate if all of ccstate's 
     holes have been taken care of... */
  if((ccstate->num_potential_fills == 1) && (outer_ccstate != NULL)) {

    DGP_ERR_PASS_CLEANUP(dgpcc_fillCheck(way_outer_ccstate, outer_ccstate, ccstate, may_delay, cookie));

  }

  NORMAL_RETURN;
ERR_CLEANUP:

  if(inner_ccstates != DGP_NULLLIST) DgpAdt_liFree(inner_ccstates);

  ERR_RETURN;
}

static ErrCode
_dgpcc_doFillHoles1(CCstate *way_outer_ccstate,
                    CCstate *ccstate, 
                    Dgp_Int32 cur, 
                    Dgp_Int32 first, 
                    Dgp_Int32 last, 
	                  cccMapColData *col,
                    Dgp_Bool may_delay,
                    colorCC_state *cookie)
{
  DGP_PROLOGUE("_dgpcc_doFillHoles1()", cookie->DGP_PERRSTATE);

  DGP_ERR_PASS(dgpcc_doFillHoles1(way_outer_ccstate, ccstate, cur, first, last, col, may_delay, cookie));
  NORMAL_RETURN;
}


ErrCode
dgpcc_fillHoles(colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_fillHoles()", cookie->DGP_PERRSTATE);
  cccMap *cccmap = cookie->cccmap;
  CCstate *ccstate;
  Dgp_Int32 i, first = -1, last = cccmap->max_num_els;
  cccMapColData *col = cccmap->col;

  /* get the bounds of the cccmap */
  while(col[++first].ccstate == NULL);
  while(col[--last].ccstate == NULL);

  /* the cccmap holds pointers to all ccstates, so we'll
     use this to move through and check all newly
     inactive ccstates for active neighbors.  */

  for(i = first; i <= last; i++) {

    ccstate = col[i].ccstate;
  
    /* if this is an empty row, the ccstate is not newly inactive,
       or the ccstate doesn't have any fill potential, skip it */
    if((ccstate == NULL) || !ccstate_isNewInactive(ccstate)
       || !ccstate_isPotentialFill(ccstate)) continue;

    /* this is the "replaceable" hole filling algorithm 
       (it will probably need to be updated as difficulties
        appear with the current algorithm) */
    DGP_ERR_PASS(dgpcc_doFillHoles1(NULL, ccstate, i, first, last, col, DGP_TRUE, cookie));

  }

  NORMAL_RETURN;
}
