/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 31 2000
**
**  Filename: colorCCread.c
**
**  Description of what's implemented in this file:
**    functions to read runs from a user defined input source, and do
**    initial color cc finding.
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCread.c,v 1.22 2003/02/10 22:53:38 mkrocken Exp $
 *
 */

#include <util.h>
#include "colorCC-priv.h"
#include "dgp_utils.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCread.c,v 1.22 2003/02/10 22:53:38 mkrocken Exp $");

#define initRun(_run) \
do { \
  (_run)->cccrun.colorInfo.rgbColor = 0; \
  (_run)->cccrun.length = 0; \
  (_run)->cccrun.sampled = DGP_FALSE; \
  (_run)->ccstate = NULL; \
} while(0)

#define readFirstRun(_first_run, _xpos, _eol) \
do { \
  initRun(_first_run); \
  (_first_run)->xpos = 0; \
  DGP_ERR_PASS((cookie->readRun)(&((_first_run)->cccrun), (_eol), cookie->runCookie, cookie->dgp_state)); \
  if(!run_isSampled(_first_run)) { \
    (_first_run)->cccrun.colorInfo.rgbColor &= RGB_LONG_MASK; \
  } \
  *(_xpos) = (_first_run)->cccrun.length; \
} while(0)

/* compute the color distance between two lrgb colors */
/* 30.10.2000 / dhtesq. */
/* Moved from colorCCutil.h... */
static INLINE Dgp_Int32 
lrgb_colorDist(Dgp_UInt32 color1, Dgp_UInt32 color2)
{
  Dgp_Int32 tempColor;
  Dgp_Int32 total;

  tempColor = DGPCC_LRGB_R(color1) - DGPCC_LRGB_R(color2);
  total = tempColor * tempColor;

  tempColor = DGPCC_LRGB_G(color1) - DGPCC_LRGB_G(color2);
  total += tempColor * tempColor;

  tempColor = DGPCC_LRGB_B(color1) - DGPCC_LRGB_B(color2);
  total += tempColor * tempColor;

  return total;
}


static ErrCode
run_merge(colorCC_state *cookie, Run *cur_run, Run *next_run)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("readRun()", DGP_PERRSTATE);

  /* update length */
  cur_run->cccrun.length += next_run->cccrun.length;

  /* are they sampled runs? */
  if(run_isSampled(cur_run)) {

    /* reallocate the current run's sampled data */
    DGP_SEC_REALLOC(cur_run->cccrun.colorInfo.rgbSamples, (DgpImage_RGB *), 
      (sizeof(DgpImage_RGB) * (cur_run->cccrun.length + next_run->cccrun.length)));
  
    /* copy the bits over */
    memcpy((void *)(cur_run->cccrun.colorInfo.rgbSamples + cur_run->cccrun.length), 
      (void *)next_run->cccrun.colorInfo.rgbSamples,
      (sizeof(DgpImage_RGB) * next_run->cccrun.length));

    /* free old bits */
    DGP_FREE(next_run->cccrun.colorInfo.rgbSamples);
    next_run->cccrun.colorInfo.rgbSamples = (DgpImage_RGB *)NULL;

  }

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}


static INLINE ErrCode 
readNextRun(colorCC_state *cookie, Run *cur_run, Dgp_UInt16 *xpos, Run *next_run, Dgp_Bool *eol)
{  
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("readRun()", DGP_PERRSTATE);

  do {

    /* Read and partially initialize the next run */
    initRun(next_run);
    next_run->xpos = *xpos;
    DGP_ERR_PASS((cookie->readRun)(&(next_run->cccrun), eol, cookie->runCookie, cookie->dgp_state));

    if(*eol) {
      /* this is the last run in a row
         the current run is the last run in the row */
      NORMAL_RETURN;
    }

    /* make sure we aren't getting trash in upper bits */
    if(!run_isSampled(next_run)) {
      next_run->cccrun.colorInfo.rgbColor &= RGB_LONG_MASK;
    }

    /* update xpos */
    *xpos += next_run->cccrun.length;

    /* do the runs match? */
    if((run_isSampled(cur_run) != run_isSampled(next_run)) || 
       (cur_run->cccrun.colorInfo.rgbColor != next_run->cccrun.colorInfo.rgbColor)) {

      /* the two runs are different, return with the current one */
      NORMAL_RETURN;

    }

    /* the two runs match, combine them */
    DGP_ERR_PASS(run_merge(cookie, cur_run, next_run));
    
  } while(!(*eol));
          
  NORMAL_RETURN;
}

ErrCode
dgpcc_readLine(colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_readLine()", cookie->DGP_PERRSTATE);
  Dgp_Int32 cur_map_index = 0, tmp_map_index, cur_run_index, tmp_run_index;
  Dgp_Int32 match_run_index = 0;
  Dgp_Bool eol = DGP_FALSE, potential_fill_conflict = DGP_FALSE;
  Dgp_UInt16 xpos;
  RunImage *rim = cookie->rim;
  cccMap *cccmap = cookie->cccmap;
  RunRow *cur_row, *above_row;
  Run *cur_row_runs, *above_row_runs, *cur_run;
  Dgp_Int32 cur_start_above_run, last_above_run, cur_end_above_run;
  Dgp_Int32 step_back;
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);
  CCstate *ccstate, *tmp_ccstate;
  cccMapRowData *prow_data;


  /* get current row */
  cur_row = rim_getCurRow(rim);
  cur_row_runs = cur_row->runs;

  /* get previous row */
  above_row = rim_getPrevRow(rim);
  above_row_runs = above_row->runs;

  cur_start_above_run = 0;
  last_above_run = above_row->num_runs;

  /* read first run */
  rim_getNewRun(rim, cur_row, cur_row_runs, cur_run_index);
  cur_run = &cur_row_runs[cur_run_index];
  readFirstRun(cur_run, &xpos, &eol);

  while(!eol) {

    /* get next new run in the row */
    rim_getNewRun(rim, cur_row, cur_row_runs, tmp_run_index);

    /* get reference to the current run */
    cur_run = &cur_row_runs[cur_run_index];

    /* read the next run's info */
    DGP_ERR_PASS(readNextRun(cookie, cur_run, &xpos, &cur_row_runs[tmp_run_index], &eol));
    
    /* find the first and last runs above that overlap the current run */
    if (cookie->is_firstLine || cur_start_above_run < 0) 
	{
		cur_start_above_run = 0;
	}
    cur_end_above_run = cur_start_above_run;
    step_back = 1;
    while((cur_end_above_run < last_above_run) && 
          (above_row_runs[cur_end_above_run].xpos < 
          (cur_run->xpos + cur_run->cccrun.length))) {
      cur_end_above_run++;
    }

    /* check for the "special" case (since this is 8-connected component
       finding, diagonal connections count.  the "special" case is when
       two runs are only touching on their edges) */
    if((cur_end_above_run < last_above_run) && 
          (above_row_runs[cur_end_above_run].xpos == 
          (cur_run->xpos + cur_run->cccrun.length))) {
      cur_end_above_run++;
      step_back = 2;
    }

    /* first, find the initial ccstate for this run */
    for(tmp_run_index = cur_start_above_run; (tmp_run_index < cur_end_above_run); tmp_run_index++) {

      if(run_runsMatch(cur_run, &above_row_runs[tmp_run_index])) {

        /* save the matching index for the fill check code below */
        match_run_index = tmp_run_index;

        /* update the ccstate */
        ccstate = above_row_runs[tmp_run_index].ccstate;
        ccstate_addRun(ccstate, cur_run, cur_run_index, cookie->curLine);
        
        /* now, get the ccstates map index */
        cur_map_index = ccstate->map_index;

        break;
      }

    }

    if(cur_run->ccstate == NULL) {

      match_run_index = 0;

      /* uh oh, no matches, make a new ccstate */
      DGP_ERR_PASS_CLEANUP(ccstate_init(&(cur_run->ccstate), cur_row_runs, 
        cur_run_index, cookie));

      /* add to the cccmap */
      cccmap_add(cccmap, cur_run->ccstate, cur_map_index);

    }

    /* compute distance from previous run, same line */
    if(cur_run_index > 0) {

      /* get info for previous run in row */
      tmp_run_index = cur_run_index - 1;
      tmp_ccstate = cur_row_runs[tmp_run_index].ccstate;
      tmp_map_index = tmp_ccstate->map_index;

      cccmap_updateNeighbor(cccmap, prow_data, tmp_map_index, cur_map_index, 
        tmp_ccstate, cur_run->ccstate,
        ((ccstate_isNormal(tmp_ccstate) && ccstate_isNormal(cur_run->ccstate)) ? 
          lrgb_colorDist(cur_run->cccrun.colorInfo.rgbColor, 
                  cur_row_runs[tmp_run_index].cccrun.colorInfo.rgbColor) : CCCMAP_MAX_DISTANCE));

    }

    /* compare to above, overlapping runs--merge if match found, otherwise, update
       max distance */
    for(tmp_run_index = cur_start_above_run; (tmp_run_index < cur_end_above_run); tmp_run_index++) {
      
      /* if the runs are part of the same ccstate, look for potential fills */
      if(above_row_runs[tmp_run_index].ccstate == cur_run->ccstate) {
        
        /* only check for fills if this above run isn't the run we originally 
           matched against */
        if(tmp_run_index > match_run_index) {
          Dgp_Int32 tmp_run_index2;

          /* get the initial above ccstate */
          tmp_run_index2 = tmp_run_index - 1;
          tmp_ccstate = above_row_runs[tmp_run_index2].ccstate;

          while(tmp_ccstate != cur_run->ccstate) {

            /* this is a hole... */
            tmp_map_index = tmp_ccstate->map_index;

	    /*            cccmap_updateNeighborPotentialFill(cccmap, prow_data, tmp_map_index, cur_map_index, 
              tmp_ccstate, cur_run->ccstate); */

            /* get previous above ccstate */
            tmp_ccstate = above_row_runs[--tmp_run_index2].ccstate;

          }
        }
        
        /* we're done with this above run, and we can skip to the next one */
        continue;
      }
      
      if(run_runsMatch(cur_run, &above_row_runs[tmp_run_index])) {
        
        /* quick and dirty count to combine smaller ccstate into larger ccstate */
        if(ccstate_approxArea(cur_run->ccstate) < 
           ccstate_approxArea(above_row_runs[tmp_run_index].ccstate)) {
          ccstate = above_row_runs[tmp_run_index].ccstate;
          tmp_ccstate = cur_run->ccstate;
        } else {
          ccstate = cur_run->ccstate;
          tmp_ccstate = above_row_runs[tmp_run_index].ccstate;
        }

        /* merge ccstates */
        DGP_ERR_PASS_CLEANUP(ccstate_merge(ccstate, &tmp_ccstate, cur_row_runs,
          above_row_runs, cookie));

        /* may have changed index... */
        cur_map_index = ccstate->map_index;

      } else {

        /* update neighbor info if new neighbors */
        tmp_ccstate = above_row_runs[tmp_run_index].ccstate;
        if (tmp_ccstate)
        {
        tmp_map_index = tmp_ccstate->map_index;

        cccmap_updateNeighbor(cccmap, prow_data, tmp_map_index, cur_map_index, 
          tmp_ccstate, cur_run->ccstate,
          ((ccstate_isNormal(tmp_ccstate) && ccstate_isNormal(cur_run->ccstate)) ? 
            lrgb_colorDist(cur_run->cccrun.colorInfo.rgbColor, 
                    above_row_runs[tmp_run_index].cccrun.colorInfo.rgbColor) : CCCMAP_MAX_DISTANCE));

        }
      }
    }

    /* we have a potential fill conflict when we have runs which are connected
       only by their corners (we're using 8-connectedness).  this can mess up
       the current fill algorithm because one ccstate which seems to be "inside"
       another ccstate might not be because it seeps outside of the outer ccstate
       by being connected to outside runs through one corner.  the potential_fill_conflict
       flag is set whenever the next run is corner aligned with the run above it.
       if it is, we check if the runs on the opposite sides of the corner could
       possibly be the same ccstate.  if so, set their potential fill conflict
       flags.  example (abcd are all pixels on a two by two grid):

          a b
          c d

       if a and d are in the same ccstate (or could be at some later point), and c and b
       are in the same ccstate (or could be...) then we have a potential fill problem. */
    if(potential_fill_conflict) {
      cccMapRowData *prow_data2;
      Dgp_Int32 tmp_map_index2, tmp_map_index3;

      /* this should be true in order for us to get here */
      DGP_ASSERT((above_row_runs[cur_start_above_run].xpos + 
                  above_row_runs[cur_start_above_run].cccrun.length) == cur_run->xpos);

      tmp_map_index = above_row_runs[cur_start_above_run].ccstate->map_index;
      tmp_map_index2 = above_row_runs[cur_start_above_run + 1].ccstate->map_index;
      tmp_map_index3 = cur_row_runs[cur_run_index - 1].ccstate->map_index;
      prow_data = cccmap_getRowData(cccmap, tmp_map_index, cur_map_index);
      prow_data2 = cccmap_getRowData(cccmap, tmp_map_index2, tmp_map_index3);

    }

    /* set potential fill conflict flag for next run */
    potential_fill_conflict = (step_back == 2);

    /* update run pointers */
    cur_start_above_run = cur_end_above_run - step_back;
  
    cur_run_index++;
    cookie->is_firstLine = DGP_FALSE;
    
  }

  /* the last run is invalid */
  rim_deleteLastRun(cur_row);

#if DGP_DBUG
  DGP_ASSERT(cccmap_validate(cookie) == cSuccessful);
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  /* no cleanup necessary */
  ERR_RETURN;
}
