/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, February 1, 2000
**
**  Filename: colorCCutil_t.h
**
**  Description of what's implemented in this file:
**    utility structure defines.
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCutil_t.h,v 1.15 2001/06/21 21:14:05 mkrocken Exp $
 */


#ifndef __COLORCCUTIL_T_H
#define __COLORCCUTIL_T_H

#include <util.h>
#include "colorCC.h"
#include "dgp_state.h"
#include "dgp_utils.h"

#ifdef __cplusplus
extern "C"
{
#endif

/********
 * semi-compact runs
 ********/

#define SC_RUN_NUM_RUNS  4

/* semi-compact run */
typedef struct {
  Dgp_UInt16 xpos, length;
} sc_run;

/* semi-compact eXtended run 
   (for washes and sampled images) */
typedef struct {
  Dgp_UInt16 xpos, length;
} sc_xrun;

/* semi-compact run segment
   (a "group" of sc_runs) */
typedef struct _sc_run_seg {
  void *next;
  Dgp_Int32 num_runs;
  sc_run runs[SC_RUN_NUM_RUNS];
} sc_run_seg;

/* semi-compact eXtended run segment
   (a "group" of sc_runs, with a 
    userdata field for each run) */
typedef struct _sc_xrun_seg {
  void *next;
  Dgp_Int32 num_runs;
  sc_run runs[SC_RUN_NUM_RUNS];  

  /* this is deliberately stuck as a separate field on the end
     of the structure, so that this structure can be treated
     like an sc_run_seg in the algorithms */
  Dgp_UInt32 userdata[SC_RUN_NUM_RUNS];
} sc_xrun_seg;

#define SCRS_POOL_ID  DGP_SMALLPOOL_MIN24
#define SCXRS_POOL_ID DGP_SMALLPOOL_MIN44


/********
 * CCstate
 ********/
typedef struct _CCstate {
  /* neighbor map info for this ccstate */
  Dgp_Int32 map_index;
  Dgp_UInt32 num_neighbors;
  Dgp_UInt16 num_potential_washes;

  /* the connected component info for this ccstate */
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  DgpAdt_List sc_lines;
  DgpAdt_List prev_run_line;
  DgpAdt_List cur_run_line;

  /* color, for solid color ccstates, otherwise 0 */
  Dgp_UInt32 color;

  Dgp_UInt32 area;

  /* various state flags for this ccstate */
  struct {
    Dgp_UInt8 sampled : 1; /* is this ccstate sampled */
    Dgp_UInt8 wash    : 1; /* is this ccstate a wash */
    Dgp_UInt8 background : 1; /* is this ccstate considered "background" */
    Dgp_UInt8 active  : 2; /* active state, one of: inactive, active, or "new inactive" */
  } state;

} CCstate;

#define DGPCC_CCSTATE_SAMPLED_COLOR 0

#define CCSTATE_ACTIVE         ((Dgp_UInt8)0x00)
#define CCSTATE_NEW_INACTIVE   ((Dgp_UInt8)0x01)
#define CCSTATE_INACTIVE       ((Dgp_UInt8)0x02)

/* some macros to test the states */
#define ccstate_isNormal(ccstate)  ((!(ccstate)->state.sampled) && (!(ccstate)->state.wash) && (!(ccstate)->state.background))
#define ccstate_isSampled(ccstate) ((ccstate)->state.sampled)
#define ccstate_isWash(ccstate) ((ccstate)->state.wash)
#define ccstate_isBackground(ccstate) ((ccstate)->state.background)

#define ccstate_isInactive(ccstate)    ((ccstate)->state.active == CCSTATE_INACTIVE)
#define ccstate_isActive(ccstate)      ((ccstate)->state.active == CCSTATE_ACTIVE)
#define ccstate_isNewInactive(ccstate) ((ccstate)->state.active == CCSTATE_NEW_INACTIVE)

/* convenience */
#define ccstate_isWashOrSampled(ccstate) (ccstate_isSampled(ccstate) || ccstate_isWash(ccstate))


/********
 * Run
 ********/
typedef struct _Run {
  DgpCC_CccRun cccrun;
  CCstate *ccstate;
  Dgp_UInt16 xpos;
} Run;

#define run_isSampled(run) ((run)->cccrun.sampled)


/********
 * RunImage
 ********/
typedef struct {
  Run *runs;
  Dgp_Int32 num_runs;
  Dgp_Int32 max_runs;
} RunRow;

typedef struct {
  RunRow *prev_row;
  RunRow *cur_row;

  /* you don't really use these directly,
     you access them through the cur_row and
     next_row ptrs above (doing this trick
     means you don't need to allocate each
     RunRow individually) */
  RunRow rows[2];
} RunImage;

#define RIM_DEF_SIZE   32
#define RIM_INCREMENT  16


/********
 * cccMap
 ********/

typedef struct {
  /* state of the relationship between these two elements */
  /* (this looks more complicated than it is...) */
  union {

    /* individual bit fields */
    struct {
      Dgp_UInt8 neighbor     : 1;  /* are these two neighbors */
      Dgp_UInt8 potential_wash : 1; /* there is a potential wash relationship here */
    } flags;

    /* all of the fields (for easy combining) */
    Dgp_UInt8 all;

  } state;

  /* color distance between these two elements */
  Dgp_UInt16 distance;
  
} cccMapRowData;

#define cccmaprd_isNeighbor(prow_data) ((prow_data)->state.flags.neighbor)
#define cccmaprd_isPotentialWash(prow_data) ((prow_data)->state.flags.potential_wash)


#define CCCMAP_MAX_DISTANCE 0xFFFF

#define CCCMAP_INVALID_INDEX    -1

typedef struct {
  cccMapRowData *row;
  CCstate *ccstate;
} cccMapColData;

typedef struct {
  cccMapColData *col;
  Dgp_Int32 num_els;
  Dgp_Int32 max_num_els;
  Dgp_Int32 cached_index;
} cccMap;

#define CCCMAP_DEF_SIZE    64
#define CCCMAP_INCREMENT   32


/********
 * colorCC_state
 ********/
typedef struct {
  Dgp_State *dgp_state;

  Dgp_Int32 width, height;
  DgpCC_PFNReadRun readRun;
  DgpCC_RunCookie runCookie;

  RunImage *rim;
  cccMap *cccmap;

  Dgp_UInt32 num_new_inactive;

  Dgp_Int32 curLine;
  Dgp_Int32 minSampSize;
  Dgp_UInt16 wThresh;

  DgpAdt_List new_ccs;
  DgpAdt_List new_sampled_ccs;

  Dgp_Bool is_firstLine;

} colorCC_state;


#ifdef __cplusplus
}
#endif

#endif /* __COLORCCUTIL_T_H */
