/********************************************** 
 * compressUtil.c
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/compressUtil.c,v 1.21 2003/02/10 22:53:39 mkrocken Exp $
 ***********************************************/
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>

#include "dgp_state.h"
#include "mmcompress.h"
#include "list.h"
#ifndef _WIN32
#include "endianness.h"
#endif
#include "g4.h"
#include "compress.h"
#include "compressUtil.h"

#include "deskew.h"
#include "libpixrw.pub"
#include "libutilw.pub"
#include "libipw.pub"
#include "imagerep.h"
#include "io.pub"
#include "dgp_allocators.h"
#include "PsgFix.h"

/* For lossless createTokensNew()... */
#include "dict-util.h"

#define	BITS_UCHAR	8

/*
 * Encode it in big-endian (network) order
 * No matter what system the program is running on.
 */
#ifdef _WIN32
#define u_Dgp_Int16 Dgp_UInt16
#define u_Dgp_Int32 Dgp_UInt32
#endif

/*
 * Change the coordinates of all the boxes to
 * be the lower left corner rather than the upper left corner. Also add in
 * the offsets. Also, assign each member cc's toknum field to toknum
 *
 * The member list here is assumed to have offsets (per TOK_OPT_RETURNOFFSET).
 * The userdata fields should point at the original cc structure.
 */
static void
adjustULtoLL(tokClass *class)
{
  DGP_VOID_PROLOGUE("adjustULtoLL()", NULL);
  DgpAdt_ListNode member;
  tokOffset *offset;
  dgplssy_cc *membercc;
  tokBitmap *exemplar;

  exemplar = &class->exemplar;

  /* Fix up all the coords. */
  DGPADT_FOREACH (member, class->members) {
    offset = (tokOffset *)DGPADT_LIGET(member);
    DGP_VOID_ASSERT(offset != (tokOffset *)NULL);

    membercc = (dgplssy_cc *)offset->userdata;
    DGP_VOID_ASSERT(membercc != (dgplssy_cc *)NULL);

    /*
     * membercc was originally found with its UL corner at
     * (membercc->x, membercc->y), and so its LL corner at
     * (x, y + height - 1).
     * It is being replaced by an instance of the exemplar. The best
     * match of the original bitmap to the exemplar occurred when the
     * bitmap's UL corner was shifted by (offset->xOffset, offset->yOffset)
     * wrt the exemplar's UL corner. The best match therefore occurred when
     * the bitmap's LL corner was positioned at
     * (xOffset, yOffset + height - exemplar->height) wrt the exemplar's
     * LL corner. Thus, the exemplar's LL corner should be positioned at
     * (x - xOffset, y + height - 1 - yOffset - height + exemplar->height)
     * or (x - xOffset, y + exemplar->height - 1 - yOffset) - note that
     * the original cc's height doesn't need to be known.
     */
    membercc->x -= offset->xOffset;
    membercc->y += class->exemplar.height - 1 - offset->yOffset;

    /* Set the token. */
    membercc->tok = (token *)class->userdata;

    /*
     * Patch the bitmap pointer - either they'd set TOK_OPT_COPYBITMAPS,
     * in which case they'd better have freed it already, or they hadn't
     * set it, in which case its value is garbage anyway.
     *
     * In fact, we should copy over the entire bitmap structure.
     */
    membercc->bitmap = class->exemplar;
  }

}


/*
 * Scan the list.  Ignore singletons.
 * For each real class (not a singleton), call adjustULtoLL on it.  This
 * not only adjusts the origin of all the boxes belonging to that class, but
 * also tags them with the class' token number (assigned by makeTokens, and
 * stored in the userdata field).
 *
 * Singletons also get adjustULtoLL called on them - the coordinate
 * systems still need to be corrected.  They get assigned the magic token
 * NULL - this means that you can tell by looking at a box whether it 
 * belongs to a singleton or not. Kindof anyway. I mean it could still
 * be an outlier. Anyway...
 */
void 
adjustCoords(DgpAdt_List tokOutput)
{
  DGP_VOID_PROLOGUE("adjustCoords()", NULL);
  tokClass *c;
  DgpAdt_ListNode classi;
    
  /* Fix up the borders, sizes, offsets, token etc. */
  DGPADT_FOREACH (classi, tokOutput) {
    c = (tokClass *)DGPADT_LIGET(classi);
    DGP_VOID_ASSERT(c != (tokClass *)NULL);

    adjustULtoLL( c); 
  }
}



/* Some bit fiddling macros */
/* Set the b rightmost (LS) bits of a Dgp_Int8. b can be 0..8 */
#define	rightbits(b)	(((b) < DGP_BITS_UINT32) ? ((1ul << (b)) - 1) : (~0ul))

/*
 * Take the bits from the runs and drop them into the bitmap. Put the
 * UL corner at (x, y).
 */
void
fillruns(Dgp_UInt32 *bits, Dgp_Int32 row_width,
	 Dgp_Int16 *runs, Dgp_Int32 nruns, Dgp_Int32 x, Dgp_Int32 y)
{
  DGP_VOID_PROLOGUE("fillruns()", NULL);
  Dgp_Int16 left, right, lastright;
  Dgp_UInt32 *row;
  Dgp_Int32 rownum = 0;
  Dgp_Int32 l, r;
  Dgp_Int32 lb, rb;

  left = runs[0];
  right = runs[1];
  runs += 2;
  while (nruns) {
    row = bits + row_width * (rownum + y);
    while (nruns) {
      /* Fill in the range (left..right, rownum) */

      /* What words get touched? */
      l = (left + x) / DGP_BITS_UINT32;
      lb = (left + x) % DGP_BITS_UINT32;
      r = (right + x) / DGP_BITS_UINT32;
      rb = (right + x) % DGP_BITS_UINT32;
      if (l == r) {
		  DGP_VOID_ASSERT(lb <= rb);
		  /* Set bits lb to rb inclusive in row[l] */
		  row[l] |= rightbits(DGP_BITS_UINT32 - lb) - rightbits(DGP_BITS_UINT32 - rb - 1);
      }
      else {
		  row[l] |= rightbits(DGP_BITS_UINT32 - lb);
		  for (l++; l < r; l++) {
			  row[l] = ~0ul;
		  }
		  row[r] |= ~rightbits(DGP_BITS_UINT32 - rb - 1);
      }

      nruns--;
      if (nruns) {
		  lastright = right;
		  left = runs[0];
		  right = runs[1];
		  runs += 2;

		  if (lastright + 1 >= left) {
			  break;
		  }
      }
    }
    rownum++;
  }
}

/*
 * Make up an array of positions.  Return the count in *numPos, and the
 * array in *_pos.  Use the numBoxes ccs in the list starting at firstbox.
 *
 * The backend gets unhappy if we feed it any box whose left side is off
 * either side of the page, or whose top or bottom lie outside the page.
 * (Note that even though you pass in the bottom left, which may lie on the
 * page, you need to be concerned about the top left as well, as the backend
 * may have to encode that if it decides that the page is transposed).
 *
 * 18.4.2001 / dhtesq.
 * Modified to include positions output by the lossless tokenizer as well.
 * I can't find any position-adjusting code like this in the lossless libraries,
 * so I'll assume for now this problem applies to the lossy tokenizer only
 * (and possibly for deskewing).
 */
ErrCode
makePosArray(	DgpAdt_ListNode firstbox, 
				DgpTdc_PageInfo *llPage,
				tokpos **_pos, 
				Dgp_Int32 numBoxes,
				Dgp_Int32 *numPos, 
				Dgp_Int32 pgWidth, 
				Dgp_Int32 pgHeight, 
				Dgp_State *dgp_state)
{
  DGP_PROLOGUE("makePosArray()", DGP_PERRSTATE);
  tokpos *pos = (tokpos *)NULL;
  Dgp_Int32 i;
  DgpAdt_ListNode box;
  Dgp_Int32 boxnum;
  dgplssy_cc *boxcc;
 
  DGP_STOP_WHINE;
  /* We've aleady locked down the memory for contiguous position blocks for lossless and lossy cc's --  */
  /* just point to it.  Lower part is occupied. */
  /*DGP_CHECK_ALLOC_CLEANUP(pos = (tokpos *)(DGP_MALLOC)((numBoxes + llPage->numBoxes) * sizeof(tokpos))); */
  pos = &llPage->boxes[llPage->numBoxes];

  /* Shift the boxes around as required. */
  for (i = 0, box = firstbox, boxnum = 0; i < numBoxes; i++, box = DGPADT_LINEXT(box)) {
    DGP_ASSERT(box != DGP_NULLLISTNODE);
    boxcc = (dgplssy_cc *)DGPADT_LIGET(box);
    DGP_ASSERT(boxcc != (dgplssy_cc *)NULL);
	
    /* Everything should have a token by now. */
    DGP_ASSERT( boxcc->tok != (token *)NULL);
    
    pos[boxnum].tok = boxcc->tok;
    pos[boxnum].extdata = (void*) &boxcc->color;

    /* Hack: just shift problem boxes */
    /* changed by jahlborn to check if box extends off right or top of page */
    if (boxcc->x < 0) {
      pos[boxnum].xpos = 0;
    }
    else if ((boxcc->x + (boxcc->bitmap.width)) > pgWidth) {
      pos[boxnum].xpos = (pgWidth - 1) - (boxcc->bitmap.width - 1);
    }
    else {
      pos[boxnum].xpos = (Dgp_Int16)boxcc->x;
    }
	    
    if (boxcc->y - (boxcc->bitmap.height - 1) < 0) {
      pos[boxnum].ypos = boxcc->bitmap.height - 1;
    }
    else if (boxcc->y >= pgHeight) {
      pos[boxnum].ypos = pgHeight - 1;
    }
    else {
      pos[boxnum].ypos = (Dgp_Int16)boxcc->y;
    }

    boxnum++;

    /*
     * Should shift boxes that extend off right/top? FIX?
     *   FIXED by jahlborn (hopefully)
     */
  }

  boxnum += llPage->numBoxes;

  *numPos = boxnum;
  *_pos = llPage->boxes;
  pos = NULL;

  NORMAL_RETURN;
  
}
  
/*
 * Scan the list.
 * For each real class (not a singleton), assign it a token.
 *
 * It's a little more complicated than that: some of the classes might
 * already have had a token assigned, and in that case we don't
 * want to assign them a new one (and don't want to put them in the
 * array of tokens).  Fortunately, we can tell which ones these
 * are by their userdata fields - they'll be non-NULL, and pointing at (guess
 * what) a pointer to their respective token struct.
 *
 * 18.4.2001 / dhtesq.
 * Extended to make lossless tokens too.  On return, **_toks will
 * refer to all tokens, not just lossy, with the lossless tokens at the 
 * lower array indices.
 */
ErrCode
makeTokens(	TCLYSectionBlock	*sb,
			DgpTdc_SectionBlock	*llSb,
			DeskewStruct		*deskewCookie,
			token				**_toks, 
			Dgp_Int32			*num_toks, 
			Dgp_Int32			*_llNonSingletonCount,
			Dgp_Bool			final,
			Dgp_State			*dgp_state)
{
  DGP_PROLOGUE( "makeTokens()", DGP_PERRSTATE);
  DgpAdt_List tokOutput = sb->classesFound;
  token *toks = (token *)NULL;
  Dgp_Int32 i;
  tokClass *c;
  DgpAdt_ListNode classi, member;
  Dgp_Int32 ntok;
  token *llToks = NULL;
  Dgp_Int32 llNonSingletonCount = 0;
  Dgp_Int32 *nElemPerPagePadding = NULL;
  tokOffset *offset;
  dgplssy_cc *membercc;
    
  /* Go to a predictable state */
  *_toks = (token *)NULL;

  /* How many new, non-singleton, (lossy) classes are there? */
  ntok = 0;
  DGPADT_FOREACH (classi, tokOutput) {
    c = (tokClass *)DGPADT_LIGET(classi);
    DGP_ASSERT(c != (tokClass *)NULL);

    if ((DGPADT_LILEN(c->members) > 1) && (c->userdata == (void *)NULL)) {
		ntok++;
    } else if((DGPADT_LILEN(c->members) == 1) && (c->userdata == NULL)){
		/* We can effectively get the sizes of all the residual dictionaries here. */
		member = DGPADT_LIFIRST(c->members);
		DGP_ASSERT(member);
		offset = (tokOffset *)DGPADT_LIGET(member);
		DGP_ASSERT(offset);
		membercc = (dgplssy_cc *)offset->userdata;

		sb->pages[membercc->pgnum - sb->sectionStartedAtThisPage].residualTokenCount++;
	}
  }
  /* Count up deskew tokens... */
  if(deskewCookie){
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
	  DGP_ERR_PASS_CLEANUP(countTokensToBeDeskewed(deskewCookie, sb));
#endif
  }

  /* Add on per-page deskew count... */
  for(i = 0; i < sb->numPages; i++){
	  sb->pages[i].residualTokenCount += sb->pages[i].skewedTokenPerPage;
  }

  /* Now:  to save effort, createTokensNew() has to process (lossless) global and residual tokens at the same time. */
  /* Later, I will have to feed a contiguous array of residual tokens to libjbig2, including both the */
  /* lossy and the lossless.  Since this array is allocated in createTokensNew(), I have to report the */
  /* size of the lossy residual dictionaries *now*, even though this isn't an intuitive place for it. */
  /* Sorry... / dhtesq. */
  DGP_CHECK_ALLOC_CLEANUP(nElemPerPagePadding = DGP_MALLOC(sb->numPages * sizeof(*nElemPerPagePadding)));
  for(i = 0; i < sb->numPages; i++){
	  nElemPerPagePadding[i] = sb->pages[i].residualTokenCount;
  }

  /* This will retrieve the *lossless* tokens, and allocate enough space for lossy ones. */
  DGP_ERR_PASS_CLEANUP(createTokensNew(dgp_state, llSb->classesFound, llSb, &llToks, &llNonSingletonCount, final, ntok, nElemPerPagePadding));
  toks = &llToks[llNonSingletonCount];

  /* If there are no tokens, there is nothing to do. Make sure to clean up memory. */
  if ((llNonSingletonCount + ntok) == 0) {
    *num_toks = 0;
	*_llNonSingletonCount = 0;
    goto ERR_CLEANUP;
  }    

  /* Clear up toks, for predictable error recovery */
  for (i = 0; i < ntok; i++) {
    toks[i].bits = (Dgp_UInt32 *)NULL;
  }

  /* Fix up the borders, sizes, offsets etc. Read into *toks. */
  i = 0;
  DGPADT_FOREACH (classi, tokOutput) {
    c = (tokClass *)DGPADT_LIGET(classi);
    DGP_ASSERT(c != (tokClass *)NULL);

    if( c->userdata != (void *)NULL) {
      /* It's got a token already */
    }
    else {
      if (DGPADT_LILEN(c->members) > 1) {
		  /* Give it a new token */
		  c->userdata = (void *)&toks[i];
		      
		  /* and add it to the token array */
		  toks[i].width = c->exemplar.width;
		  toks[i].height = c->exemplar.height;
		  toks[i].uses = DGPADT_LILEN(c->members);
		  
		  toks[i].committed = DGP_FALSE;
		  toks[i].extdata = (void *)NULL;
		  
		  DGP_ASSERT(c->exemplar.bits != (Dgp_UInt32 *)NULL);
		  toks[i].bits = c->exemplar.bits;
		  toks[i].longsperline = c->exemplar.longsperline;
		  
		  i++;
      }
      else {
		  /* It's a singleton. Leave its userdata set to NULL. */
      }
    }
  }
		
  DGP_ASSERT(i == ntok);
  
  DGP_FREE(nElemPerPagePadding);

  *num_toks = ntok;
  *_llNonSingletonCount = llNonSingletonCount;
  *_toks = llToks;
  toks = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:
  if (toks != (token *)NULL) {
    DGP_FREE((void *)toks);
  }
  if(nElemPerPagePadding) DGP_FREE(nElemPerPagePadding);        
  ERR_RETURN;
}


/**********************************************************************
 * When a residual is not written, outliers and singletons get put into a 
 * per-page dictionary.
 * Any outliers must have their runs converted to bitmaps first, then both
 * outliers and singletons are put into a dictionary and written to the file.
 *
 * 18.4.2001 / dhtesq.
 * Modified to include singletons (outliers not necessary) left over
 * from lossless tokenization.  These are present in the lower part of the
 * passed toks array.
 *
 ***********************************************************************/
ErrCode
writeResDict(	TCLYOutCookie		toc, 
				TCLYPageInfo		cpage,
				DgpTdc_PageInfo		llPage,
				DgpAdt_List			skewedTokens, 
				Dgp_Int32			ntok,
				Dgp_Int32			nLlTok,
				token				*toks, 
				Dgp_Bool			g4_toks)
{
  Dgp_State *dgp_state = toc->dgp_state;
  DGP_PROLOGUE( "writeResDict()", DGP_PERRSTATE);
  DgpAdt_ListNode firstbox = DGPADT_LIFIRST(cpage.boxes); 
  Dgp_Int32 numBoxes = cpage.numBoxes; 
  DgpAdt_ListNode box;
  Dgp_Int32 i, j;
  
  tokBitmap *bitmap;
  dgplssy_cc *boxcc;
  token *tok;
  extdataFork *fork;
  void *newDictstate;
  Dgp_UInt32 singleton_uses = ((g4_toks) ? 0 : 1);
  OutputCallbacks * ocb = toc->callbacks;
      
  (void)llPage;
  /* Clear up toks, for predictable error recovery */
  /* shrink wrap the jpeg data */
  for (i = nLlTok; i < ntok + nLlTok; i++) {
    toks[i].bits = (Dgp_UInt32 *)NULL;
  }
           
  /* add each one to the dictionary */
  j = nLlTok;
  for (i = 0, box = firstbox; i < numBoxes; i++, box = DGPADT_LINEXT(box)) {
    DGP_ASSERT(box != DGP_NULLLISTNODE);
    boxcc = (dgplssy_cc *)DGPADT_LIGET(box);
    DGP_ASSERT(boxcc != (dgplssy_cc *)NULL);
    bitmap = &(boxcc->bitmap);

    /* Some outliers haven't had runs converted to bitmap yet... */
    if (boxcc->runs != (Dgp_Int16 *)NULL) {

      /* If it has runs, then it better not have a token! */      
      DGP_ASSERT( boxcc->tok == (token *)NULL);
	    	    
      /* Make bitmap from runs. */
      DGP_CHECK_ALLOC_CLEANUP(bitmap->bits = (Dgp_UInt32 *)DGP_CALLOC((size_t)(bitmap->height), 
		  (size_t)(bitmap->longsperline * sizeof(*(bitmap->bits)))));
	    
      fillruns(bitmap->bits, bitmap->longsperline, boxcc->runs, boxcc->nruns, 0, 0);
	    
      /* Free the runs- we don't want to have both runs and a 
       * bitmap hanging around. */
      DGP_FREE( (void *)boxcc->runs);
      boxcc->runs = (Dgp_Int16 *)NULL;
	    
      /* convert orgin from UL to LL */
      boxcc->y += bitmap->height - 1;
    }
    
    if (boxcc->tok == (token *)NULL) { /* singleton or outlier */	
	  
      toks[j].width = bitmap->width;
      toks[j].height = bitmap->height;
      toks[j].longsperline = bitmap->longsperline;
	    
      /* Set uses to 0 if you want the backend to produce G4 tokens. */
      toks[j].uses = singleton_uses;
      toks[j].committed = DGP_FALSE;
      toks[j].extdata = (void *)NULL;

      DGP_ASSERT(bitmap->bits != (Dgp_UInt32 *)NULL);
      toks[j].bits = bitmap->bits;
	   
      /* Note that we _don't_ add this bitmap to the garbage list.
       * If it's an outlier then its bitmap will be freed by
       * pageCleanup. If it's a singleton, then its bitmap is 
       * owned by the tokenizer. Finally if it's a deskewed
       * singleton, it's freed by pageCleanup. */
	    
      /* Now set the boxcc token to point to the real token, so that 
       * when positions are written out, the boxcc will have the correct
       * pointer. */
      boxcc->tok = &toks[j];    
      j++;
    } else if( boxcc->tok->committed == DGP_FALSE){ /* deskewed class */
        /* This cc's original token was deskewed for this page. 
		 * The deskewer saved some inportant data in the extension. */
        fork = (extdataFork *)boxcc->tok->extdata;
		if( fork->newToken == (token *)NULL) {
			DgpAdt_ListNode retnode;
			tok = boxcc->tok;

			toks[j].width = tok->width;
			toks[j].height = tok->height;
			toks[j].longsperline = tok->longsperline;
			
			toks[j].uses = tok->uses;
			toks[j].committed = DGP_FALSE;
			
			/* The original extdata is also in the fork. */
			toks[j].extdata = fork->oldExtdata;
			
			DGP_ASSERT(tok->bits != (Dgp_UInt32 *)NULL);
			toks[j].bits = tok->bits;
			
			/* Save a pointer to the new token for other boxcc's. */
			/* This is the only place that fork->newToken is diddled. / dhtesq. */
			fork->newToken = &toks[j];
			
			/* Make sure the boxcc points to the _correct_ token. */
			boxcc->tok = &toks[j];
			
			/* Hold on to the old token so that it can be freed later. */
			DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(skewedTokens, tok, &retnode));
			
			j++;
			
		} else {
		/* This token has already been copied but we need to
	     * update the boxcc pointer, luckily we saved that pointer
	     * in the defunct token's extdata. */
			boxcc->tok = (token *)fork->newToken;
		}
	}
    /* else... it has a token and the token is committed, so no need to
     * write it to this dictionary- it's already been written. */
  } /* end for */

  /* Sanity... */
  DGP_ASSERT(j == ntok + nLlTok);
  
  /* Write the dictionary to file. */
  DGP_ERR_PASS_CLEANUPF0(ocb->writeOutputPageDictionary(toks, 
							ntok + nLlTok, 
							cpage.width,
							cpage.height,
							cpage.xRes,
							cpage.yRes,
							DGP_FALSE,
							toc->dictState,
							&newDictstate, 
							(DictExtension *)NULL, 
							0, 
							toc->outputState,
							ocb->userData),
			 "Unable to write dictionary.");
  /* Update the dictstate so that the position block can be correctly
   * written out. */
  toc->dictState = newDictstate;

  /* Note that we don't throw out the toks array. We need to hold on 
   * to it until the position block for this page is written out. */
  
  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}


