/******************************************
 * deskew.c
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/deskew.c,v 1.15 2003/02/10 22:53:40 mkrocken Exp $
 ******************************************/
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include "dgp_state.h"
#include "deskew.h"
#include "deskew_bmp.h"
#include "skew.pub"
#include "stdio.h"
#include "compress.h"
#include "math.h"
#include "compressUtil.h"
#include <limits.h>
#include "rotatop.pub"
#include "PsgFix.h"

#include "image.h"

#define MIN_CONFIDENCE 5
#define PI_6 0.523598776
#define DIVBY_2 >> 1
#define PIXEL_DIFF 2


/********************************************************************* 
 * create the deskew structure, giving it the number of pages in the file 
 ***********************************************************************/
ErrCode
deskewInit(	DeskewStruct **deskew_cookie, 
		Dgp_Int32 skewAccuracy, 
		Dgp_Int32 baselineOrient, 
           	Dgp_State *dgp_state)
{
  DGP_PROLOGUE("deskewInit()", DGP_PERRSTATE);
  DeskewStruct *dsq = NULL;


  DGP_CHECK_NULL(deskew_cookie);

  /* create the deskew structure */
  DGP_CHECK_ALLOC_CLEANUP(dsq = (DeskewStruct *)(DGP_MALLOC)(sizeof(*dsq)));
  dsq->dgp_state = dgp_state;
  dsq->baselineOrient = baselineOrient;

  /* skew accuracy should be between 0 and 5 */
  if(skewAccuracy > 5) {
    skewAccuracy = 5;
  } else if(skewAccuracy < 0) {
    skewAccuracy = 0;
  }
  dsq->skewAccuracy = ((Dgp_Int32)1) << (5 - skewAccuracy);

  /* create buffer to store skew angles */
  *deskew_cookie = dsq;
  dsq = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(dsq != NULL) (DGP_FREE)(dsq);

  ERR_RETURN;
}

/**********************************************************
 *  clean up any memory used in the deskew structure 
 ***********************************************************/
void
deskewCleanup(DeskewStruct *deskew_cookie)
{
  Dgp_Free tmp_free;
  DeskewStruct *dsq = deskew_cookie;

  /* if the structure is NULL (meaning we didn't deskew) just return */
  if(dsq == NULL) return;
  tmp_free = dsq->DGP_FREE;
  tmp_free(dsq);
}


/********************************************************
 * find and collect the skew angles for each page of 
 * the document 
 **********************************************************/
ErrCode
getSkewAngle(	DeskewStruct *deskew_cookie, 
		PIXR *pixr, 
		Float32 *angle)
{
  DeskewStruct *dsq = (DeskewStruct *)deskew_cookie;
  DGP_PROLOGUE("getSkewAngle()", dsq->DGP_PERRSTATE);
  Float32 skew_angle;
  Int32 confidence;
  PIXR *tmp_pixr = (PIXR *)NULL;
  IPWState *ipwstate = dsq->DGP_PIPWSTATE;

  /* rotate pixr because w_findSkewSample() wants a page with baseline down */
  switch(dsq->baselineOrient) {
  case BO_UP:
    DGP_ERR_PASS_CLEANUP(w_rotate180(ipwstate, pixr, &tmp_pixr));
    pixr = tmp_pixr;
    break;
  case BO_RIGHT:
    DGP_ERR_PASS_CLEANUP(w_rotate90(ipwstate, pixr, &tmp_pixr));
    pixr = tmp_pixr;
    break;
  case BO_LEFT:
    DGP_ERR_PASS_CLEANUP(w_rotate270(ipwstate, pixr, &tmp_pixr));
    pixr = tmp_pixr;
    break;
  }

  /* find the skew angle */
  DGP_ERR_PASS_CLEANUP(w_findSkewSample(ipwstate, 
					pixr, 
					dsq->skewAccuracy, 
					&confidence, 
					&skew_angle));

  if(tmp_pixr) w_destroyPixr(ipwstate, &tmp_pixr);

  /* if it's legitimate, save it, otherwise, discard it */
  if((confidence > MIN_CONFIDENCE) && (skew_angle <= PI_6) &&
     (skew_angle >= -(PI_6))) {
      *angle = skew_angle;
  } else {
      *angle = 0.0;
  }

  NORMAL_RETURN;

ERR_CLEANUP:
    
  if(tmp_pixr) w_destroyPixr(ipwstate, &tmp_pixr);  

  ERR_RETURN;
}

/*
 * 25.4.2001 / dhtesq.
 * Since we're going to be putting lossy, lossless, and deskewed tokens into one contiguous array
 * for libjbig2, we need to know the size of the token array beforehand.  This function counts up
 * the number of nonsingleton deskewed tokens in a section we'll have to make space for, but doesn't do any
 * deskewing.  Call this after completing a section, before writing the global dictionary.
 *
 * With some extra ambition, this function could prepare a list of tokens to be deskewed, saving some
 * work in deskewTokens().
 */
ErrCode
countTokensToBeDeskewed(DeskewStruct		*deskewCookie,
						TCLYSectionBlock	*sb)
{
  DeskewStruct *dsq = deskewCookie;
  Dgp_State *dgp_state = dsq->dgp_state;
  DGP_STATE_PROLOGUE("countTokensToBeDeskewed()", dgp_state);
  Dgp_Int32 min_rotate_dim, i, *rotate_dims = (Dgp_Int32 *)NULL;
  Dgp_Int32 rel_page, page, uses_this_page;
  Dgp_Float32 skew_angle;
  DgpAdt_List newMembers = DGP_NULLLIST;
  DgpAdt_ListNode ln, member, theClass, next, first, retnode;
  tokClass *c;
  DgpAdt_List tokOutput = sb->classesFound;
  Dgp_Int32 numPages = sb->numPages;

  min_rotate_dim = INT_MAX;
  DGP_CHECK_ALLOC_CLEANUP(rotate_dims = (Dgp_Int32 *)DGP_MALLOC(numPages * sizeof(*rotate_dims)));

  for(i = 0; i < numPages; i++) {
	skew_angle = sb->pages[i].skewAngle;
    if(skew_angle != 0) {
      rotate_dims[i] = (Dgp_Int32)fabs(PIXEL_DIFF / sin(skew_angle));
      if(rotate_dims[i] < min_rotate_dim) min_rotate_dim = rotate_dims[i];
    } else {
      rotate_dims[i] = 0;
    }
  }

  if ((tokOutput != DGP_NULLLIST) && (DGPADT_LILEN(tokOutput) > 0)) {
    DGPADT_FOREACH(theClass, tokOutput) {
      c = (tokClass *)DGPADT_LIGET(theClass);
      /* check if a token class's dimensions is bigger than the minimum *
       * deskew dimension of all the deskew	 dimensions */
	  
	  /* We don't want to count singletons here... */
      if ((DGPADT_LILEN(c->members) > 1) &&
		  ((c->exemplar.width > min_rotate_dim) || (c->exemplar.height > min_rotate_dim))) {
        DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&newMembers, dgp_state));
		next = DGPADT_LIFIRST(c->members);
		while (next != DGP_NULLLISTNODE) {
			first = member = next;
			next = DGP_NULLLISTNODE;
			uses_this_page = 0;
			page = ((dgplssy_cc *)(((tokOffset *) DGPADT_LIGET(member))->userdata))->pgnum;

			/* find number of uses of this token on this page */
			while(member != DGP_NULLLISTNODE) {
				if (((dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(member))->userdata))->pgnum == page) {
					uses_this_page++;
				} else if(next == DGP_NULLLISTNODE) {
					next = member;
				}
				member = DGPADT_LINEXT(member);
			}

			rel_page = page - sb->pages[0].pgNum;

			/* check if we'd REALLY want to deskew these tokens */
			if ((rotate_dims[rel_page]) &&
				((c->exemplar.width > rotate_dims[rel_page]) ||
				(c->exemplar.height > rotate_dims[rel_page]))) {

				/*************************************/
				sb->pages[rel_page].skewedTokenPerPage++;
				/*************************************/

			} /* end if */

			ln = first;

			for(i = 0; i < uses_this_page; i++) {
				/* Remove this offset from the old member list. 
				* This is important so that the offset won't be 
				* considered again. Add it to the new member list 
				* so that it will still be part of the class. */
				DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newMembers, DGPADT_LIGET(ln), &retnode));
				DgpAdt_liRemAndNext(ln, &ln);		  
				while((ln != DGP_NULLLISTNODE) && 
					(((dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(ln))->userdata))->pgnum != page)) {
					ln = DGPADT_LINEXT(ln);
				}
			} /* end for */

		} /* end while */
	   /* Now that we're finished with this class, move all the members back
		* to where they should be. */
		DGP_ASSERT( DGPADT_LILEN( c->members) == 0);
		DgpAdt_liFree( c->members);
		c->members = newMembers;
		newMembers = DGP_NULLLIST;	
      } /* end if big enough */
    } /* end DGPADT_FOREACH class */
  } /* end if there is tokenizer output */

  dsq->rotateDims = rotate_dims;
  rotate_dims = NULL;
  dsq->minRotateDim = min_rotate_dim;

  NORMAL_RETURN;
ERR_CLEANUP:
  if(rotate_dims) DGP_FREE(rotate_dims);
  ERR_RETURN;
}



/*************************************************************
 * Deskew tokens of a section block
 *************************************************************/
ErrCode 
deskewTokens(DeskewStruct		*deskew_cookie, 
			 DgpAdt_List		newBitmaps, 
			 TCLYSectionBlock	*sb)
{
  DeskewStruct *dsq = deskew_cookie;
  DGP_PROLOGUE("deskewTokens()", dsq->DGP_PERRSTATE);
  Dgp_State *dgp_state = dsq->dgp_state;
  Dgp_Allocators *a = &(dsq->DGP_ALLOCATORS);
  Dgp_Int32 min_rotate_dim, i, *rotate_dims = (Dgp_Int32 *)NULL;
  Dgp_Int32 new_w, new_h, new_lpl, rel_page, page, off_x, off_y, uses_this_page;
  dgplssy_cc *boxcc = (dgplssy_cc *)NULL;
  DgpAdt_List newMembers = DGP_NULLLIST;
  DgpAdt_ListNode ln, member, theClass, next, first, retnode;
  Dgp_UInt32 *bits = (Dgp_UInt32 *)NULL;
  tokClass *c;
  tokBitmap *bitmap;
  token *tmptok = NULL, *tok;
  extdataFork *fork = NULL;
  DgpAdt_List tokOutput = sb->classesFound;

  /*
  Allocation & setting of rotate dims etc. moved up to countTokensToBeDeskewed()... / dhtesq. 
  */
  rotate_dims = dsq->rotateDims;
  min_rotate_dim = dsq->minRotateDim;

  if ((tokOutput != DGP_NULLLIST) && (DGPADT_LILEN(tokOutput) > 0)) {
    DGPADT_FOREACH(theClass, tokOutput) {
      c = (tokClass *)DGPADT_LIGET(theClass);
      /* check if a token class's dimensions is bigger than the minimum *
       * deskew dimension of all the deskew	 dimensions */
      if ((c->exemplar.width > min_rotate_dim) || (c->exemplar.height > min_rotate_dim)) {
	   /* Create a new list to hold the members of the class as we check
		* them for deskewing. Once a member has been deskewed (or it has
		* been determined that it will not be deskewed) it is removed from
		* the original member list and added to this one. Later on, once
		* everything has been considered and moved to the new list, we
		* simply swap the lists. */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&newMembers, dgp_state));
		next = DGPADT_LIFIRST(c->members);
		while (next != DGP_NULLLISTNODE) {
			first = member = next;
			next = DGP_NULLLISTNODE;
			uses_this_page = 0;
			page = ((dgplssy_cc *)(((tokOffset *) DGPADT_LIGET(member))->userdata))->pgnum;

			/* find number of uses of this token on this page */
			while(member != DGP_NULLLISTNODE) {
				if (((dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(member))->userdata))->pgnum == page) {
					uses_this_page++;
				} else if(next == DGP_NULLLISTNODE) {
					next = member;
				}
				member = DGPADT_LINEXT(member);
			}

			rel_page = page - sb->pages[0].pgNum;

			/* check if we REALLY want to deskew these tokens */
			if ((rotate_dims[rel_page]) &&
				((c->exemplar.width > rotate_dims[rel_page]) ||
				(c->exemplar.height > rotate_dims[rel_page]))) {
				bitmap = &(c->exemplar);
				/* deskew bitmap */
				DGP_ERR_PASS_CLEANUPF0(DgpDeskew_bmp(	bitmap->bits, 
														bitmap->width, 
														bitmap->height, 
														bitmap->longsperline, 
														&bits, 
														&new_w, 
														&new_h, 
														&new_lpl,
														sb->pages[rel_page].skewAngle, 
														sb->pages[rel_page].width, 
														sb->pages[rel_page].height,
														dgp_state),
					"deskew_bmp failed on token.");
				if (new_lpl != ((new_w + 31) / 32)) {
					DGP_ASSERT(new_lpl == ((new_w + 31) / 32));
				}

				/* keep center of token in same place */
				off_x = (new_w - bitmap->width) DIVBY_2;
				off_y = (new_h - bitmap->height) DIVBY_2;
				boxcc = (dgplssy_cc *)(((tokOffset *) DGPADT_LIGET(first))->userdata);
				DGP_ASSERT(boxcc != (dgplssy_cc *)NULL);
				if(boxcc->tok == (token *)NULL) {
				   /* It's a singleton. So we don't want to make 
					* a new token.  Instead, just update the 
					* bitmap and related info. */
					DGP_ASSERT( DGPADT_LILEN(c->members) < 2);
				   /* Here we _don't_ free the original bits, 
					* since they belong to the tokenizer. 
					* Just set the cc's bitmap to the deskewed 
					* version. (The old bitmap is also the 
					* exemplar bitmap.) */
					boxcc->bitmap.bits = bits;
					/* And then hold on to a pointer to the new bitmap. */
					DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newBitmaps, boxcc, &retnode));
					/* assign new cc info */
					boxcc->x -= off_x;
					boxcc->y += off_y;
					boxcc->bitmap.width = (Dgp_UInt16)new_w;
					boxcc->bitmap.height = (Dgp_UInt16)new_h;
					boxcc->bitmap.longsperline = (Dgp_UInt16)new_lpl;
		    
				   /* There is only one member of this class, 
					* but just to be consistent, move this 
					* member to the new list. */
					DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newMembers, DGPADT_LIGET(first), &retnode));
					DgpAdt_liRem(first);
				} else {
					/* Create a new token to hold the new info. */
					DGP_CHECK_ALLOC_CLEANUP(tok = (a->malloc)( sizeof( *tok)));
					tmptok = tok;
					tok->extdata = NULL;
					tok->bits = bits;

					tok->width = (Dgp_UInt16)new_w;
					tok->height = (Dgp_UInt16)new_h;
					tok->longsperline = (Dgp_UInt16)new_lpl;

					tok->uses = (Dgp_UInt16)uses_this_page;

				   /* Save the userdata and make room for the forwarding
					* pointer. */
					DGP_CHECK_ALLOC_CLEANUP(fork = (extdataFork *)(a->malloc)( sizeof( extdataFork)));
					fork->oldExtdata = boxcc->tok->extdata;
					fork->newToken = (void *)NULL;
					tok->extdata = (void *)fork;
					fork = (extdataFork *)NULL;

				   /* Mark the token as NOT committed so that it will
					* be added to the per page residual dictionary. */
					tok->committed = DGP_FALSE;
					/* This increment moved up to countTokensToBeDeskewed(). / dhtesq. */
					/*sb->pages[rel_page].skewedTokenPerPage++; */
		        
					ln = first;
					tmptok = NULL;
					for(i = 0; i < uses_this_page; i++) {
						boxcc = (dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(ln))->userdata);
						bitmap = &(boxcc->bitmap);
						/* assign new cc info */
						boxcc->x -= off_x;
						boxcc->y += off_y;
						/* give it a new token */
						boxcc->tok = tok;
						/* assign new bitmap info */
						if( bitmap->bits != c->exemplar.bits) {
							(a->free)(bitmap->bits);
						}
						bitmap->bits = bits;
						bitmap->width = (Dgp_UInt16)new_w;
						bitmap->height = (Dgp_UInt16)new_h;
						bitmap->longsperline = (Dgp_UInt16)new_lpl;
	        
					   /* Remove this offset from the old member list. 
						* This is important so that the offset won't be 
						* considered again. Add it to the new member list 
						* so that it will still be part of the class. */
						DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newMembers, DGPADT_LIGET(ln), &retnode));
						DgpAdt_liRemAndNext(ln, &ln);		  
						while((ln != DGP_NULLLISTNODE) && 
							(((dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(ln))->userdata))->pgnum != page)) {
							ln = DGPADT_LINEXT(ln);
						}
					} /* end for */
				} /* end if */
      
				bits = (Dgp_UInt32 *)NULL;
			} /* end if */
			else {
			   /* This token was too small to be deskewed on this page.
				* we move it to the new list, so that it won't be counted
				* again later. Also move another other instances of this
				* token on this particular page. */
				ln = first;
				for(i = 0; i < uses_this_page; i++) {
					DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newMembers, DGPADT_LIGET(ln), &retnode));
					DgpAdt_liRemAndNext(ln, &ln);
					while((ln != DGP_NULLLISTNODE) &&
						(((dgplssy_cc *)(((tokOffset *)DGPADT_LIGET(ln))->userdata))->pgnum != page)) {
						ln = DGPADT_LINEXT(ln);
					}
				}
			}
		} /* end while */
	   /* Now that we're finished with this class, move all the members back
		* to where they should be. */
		DGP_ASSERT( DGPADT_LILEN( c->members) == 0);
		DgpAdt_liFree( c->members);
		c->members = newMembers;
		newMembers = DGP_NULLLIST;	
      } /* end if big enough */
    } /* end DGPADT_FOREACH class */
  } /* end if there is tokenizer output */

  if(rotate_dims) (a->free)(rotate_dims);
  dsq->rotateDims = NULL;
  dsq->minRotateDim = INT_MAX;
  NORMAL_RETURN;

ERR_CLEANUP:
  if(rotate_dims) (a->free)(rotate_dims); 
  if(tmptok != NULL) {
    if(tmptok->extdata != NULL) (a->free)(tmptok->extdata);
    (a->free)(tmptok);
  }
  if(newMembers != NULL) (a->free)(newMembers);
  if(fork != NULL) (a->free)(fork);

  ERR_RETURN;
}


/*************************************************************
 * Deskew outliers for a page
 *************************************************************/
ErrCode
deskewOutliers(	DeskewStruct *deskew_cookie, 
		TCLYPageInfo cPage)
{
  DeskewStruct *dsq = deskew_cookie;
  DGP_PROLOGUE("deskewOutliers()", dsq->DGP_PERRSTATE);
  Dgp_Float64 sin_angle, cos_angle;
  Dgp_Int32 rotate_dim;
  Dgp_Int32 new_w, new_h, new_lpl;

  Float32 skew_angle;
  dgplssy_cc *boxcc = (dgplssy_cc *)NULL;
  tokBitmap* bitmap;
  Dgp_Int32 pgWidth  = cPage.width;
  Dgp_Int32 pgHeight = cPage.height;

  DgpAdt_ListNode ln;
  Dgp_UInt32 *bits = (Dgp_UInt32 *)NULL;

  Dgp_State *dgp_state = dsq->dgp_state;
  Dgp_Allocators *a = &(dsq->DGP_ALLOCATORS);
    
  skew_angle = cPage.skewAngle;
      
  if( skew_angle != 0) {
    sin_angle = sin( skew_angle);
    cos_angle = cos( skew_angle);
    rotate_dim = (Dgp_Int32)fabs(PIXEL_DIFF / sin_angle);
  } else {
    /* No skewing for this page. */
    NORMAL_RETURN;
  }
  
  /* Check all the outliers on the current page. */
  DGPADT_FOREACH( ln, cPage.largeCCList ) {
    boxcc = (dgplssy_cc *)DGPADT_LIGET(ln);
    if((rotate_dim) && ((boxcc->bitmap.width > rotate_dim) ||
			(boxcc->bitmap.height > rotate_dim))) {
      bitmap = &(boxcc->bitmap);
		
      /*make bitmap from runs*/
      DGP_CHECK_ALLOC_CLEANUP(	bitmap->bits = (Dgp_UInt32 *)
				(a->calloc)((size_t)(bitmap->height), 
				(size_t)(bitmap->longsperline * 
						sizeof(*(bitmap->bits)))));
      fillruns(	bitmap->bits, 
		bitmap->longsperline, 
		boxcc->runs, 
	       	boxcc->nruns, 
		0, 
		0);
      /* free runs */
      free(boxcc->runs);
      boxcc->runs = (Dgp_Int16 *)NULL;
        
      /* convert orgin from UL to LL */
      boxcc->y += bitmap->height - 1;

      /* deskew bitmap */
      if (bitmap->height > 1 && bitmap->width > 1)
      DGP_ERR_PASS_CLEANUPF0(DgpDeskew_bmp(bitmap->bits, 
					bitmap->width, 
					bitmap->height,
					bitmap->longsperline, 
					&bits, 
					&new_w, 
					&new_h, 
					&new_lpl, 
					skew_angle, 
					pgWidth, 
					bitmap->height, 
					dgp_state),
			     		"deskew_bmp failed on outlier.");
      else
      {
          bits = bitmap->bits;
          new_w = bitmap->width;
          new_h = bitmap->height;
          new_lpl = bitmap->longsperline;
      }
      if(new_lpl != ((new_w + 31) / 32)) {
	/* printf("page = %d\n", boxcc->pgnum);
	 * printf("new_lpl = %d, lpl = %d\n", new_lpl,((new_w + 31) / 32)); 
	*/
	DGP_ASSERT(new_lpl == ((new_w + 31) / 32));
      }

      /* keep center of token in same place */
      boxcc->x -= (new_w - bitmap->width) DIVBY_2;
      while (boxcc->x + new_w > pgWidth) boxcc->x--;
      boxcc->y += (new_h - bitmap->height) DIVBY_2;
      while (boxcc->y + new_h > pgHeight) boxcc->y--;

      /* free old bitmap and assign new info to outlier */
      if (bits != bitmap->bits)
      {
          (a->free)(bitmap->bits);
          bitmap->bits = bits;
      }
      bits = (Dgp_UInt32 *)NULL;

      bitmap->width = (Dgp_UInt16)new_w;
      bitmap->height = (Dgp_UInt16)new_h;
      bitmap->longsperline = (Dgp_UInt16)new_lpl;

      /* This cc is still an outlier and so has no token. */
      boxcc->tok = (token *)NULL;
    }
  }


  NORMAL_RETURN;

ERR_CLEANUP:
  /* WRITEME */

  ERR_RETURN;
}


/*************************************************************
 * Deskew position blocks for a page
 *************************************************************/
ErrCode
deskewPositions(DeskewStruct *deskew_cookie, 
		TCLYPageInfo cpage)
{    
  DeskewStruct *dsq = deskew_cookie;
  DGP_PROLOGUE("deskewPositions()", dsq->DGP_PERRSTATE);
  Float32 skew_angle;
  Dgp_Float64 sin_angle, cos_angle;
  DgpAdt_ListNode firstBox = DGPADT_LIFIRST(cpage.boxes); 
  Dgp_Int32 numBoxes = cpage.numBoxes;
  Dgp_Int32 pgWidth = cpage.width; 
  Dgp_Int32 pgHeight = cpage.height;
    
  Dgp_Int32 i, rotate_dim;
  Dgp_Int32 orig_x, orig_y, nu_cntr_x, nu_cntr_y; 
    
  DgpAdt_ListNode ln;
  dgplssy_cc *boxcc;
      
  DGP_CHECK_NULL(deskew_cookie);
    
  skew_angle = cpage.skewAngle;
    
  if( skew_angle != 0) {
    sin_angle = sin( skew_angle);
    cos_angle = cos( skew_angle);
    rotate_dim = (Dgp_Int32)fabs(PIXEL_DIFF / sin_angle);
  } else {
    /* No skewing for this page. */
    NORMAL_RETURN;
  }

  orig_x = pgWidth DIVBY_2;
  orig_y = pgHeight DIVBY_2;

  ln = firstBox;
  for( i = 0; i < numBoxes; i++, ln = DGPADT_LINEXT( ln)) {

    /* Rotate each position around the center of the page. */
    boxcc = (dgplssy_cc *)DGPADT_LIGET(ln);
	
    nu_cntr_x = boxcc->x + ((boxcc->bitmap.width) DIVBY_2) - orig_x;
    nu_cntr_y = boxcc->y - ((boxcc->bitmap.height) DIVBY_2) - orig_y;

    /* We don't need to worry about bounds checking because it's done
     * in makePosArray(). */
    boxcc->x += (Dgp_Int32)((nu_cntr_x * cos_angle) - (nu_cntr_y * sin_angle) 
		      + 0.5) - nu_cntr_x;
    boxcc->y += (Dgp_Int32)((nu_cntr_y * cos_angle) + (nu_cntr_x * sin_angle)
		      + 0.5) - nu_cntr_y;
  }

  NORMAL_RETURN;
}



