/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/class.c,v 1.14 2003/01/27 21:48:41 awilliam Exp $
 *
 */

/*
 * This holds a bunch of routines that deal with tokeniser classes.
 *
 * A Dgp_Int16 discourse on classes and their states:
 *
 * Each class can be in one of three primary states:
 * - normal
 * - committed
 * - immutable
 * Every class also has (in concept) a histogram (tokCPixelSum) associated with
 * it.  In some cases, this histogram can be missing.
 *
 * A regular class contains bitmaps for all its members.  It may, or may not,
 * have a tokCPixelSum allocated for it; only singleton classes are allowed to
 * not have a tokCPixelSum allocated.  Its exemplar may be the first member's
 * bitmap, or might be a synthetic bitmap.  It is possible to add members to
 * a regular class, and to delete members from the class.
 *
 * A committed class does not contain bitmaps for any of its members.  It
 * is possible to add members to a committed class, but not to delete members.
 * A committed class must have a tokCPixelSum allocated, and must have a synthetic
 * exemplar.  Its exemplar may change as the histogram changes.  Committed
 * classes are marked with TOKCLASS_COMMITTED.
 *
 * An immutable class does not contain bitmaps for any of its members.  It
 * is possible to add members to an immutable class, but not to delete members.
 * An immutable class must have a tokCPixelSum allocated, and must have a synthetic
 * exemplar.  Its exemplar will never change, no matter what.  Immutable
 * classes are marked with TOKCLASS_IMMUTABLE.  Every immutable
 * class is also committed (marked with TOKCLASS_COMMITTED).
 *
 * Other status bits:
 * TOKCLASS_FROZENHIST: Some classes have "frozen" histograms: the values in
 * 		     the histogram don't change with member addition.
 *		     Only committed and immutable classes can have frozen
 *		     histograms.
 * TOKCLASS_EXSYNTHETIC: This marks synthetic exemplars; committed and immutable
 *		      classes must have this bit set.
 * TOKCLASS_EXCLEAN: This is only applicable to classes with synthetic
 *		  exemplars.  It means that the class's histogram has
 *		  not changed since the exemplar was built, and so the
 *		  does not need to be rebuilt.
 * TOKCLASS_CHECKCLEAN: This is only applicable to regular classes.  It means
 *		     that all of the class members are known to match the
 *		     class exemplar.
 * TOKCLASS_MERGECLEAN: This class has been checked against all other
 *		     TOKCLASS_MERGECLEAN classes.  This means that in a subsequent
 *		     merge pass, two TOKCLASS_MERGECLEAN classes need not be
 *		     compared.
 * TOKCLASS_OKSOFAR: This class' exemplar has not changed since the start of
 *		  the current merge phase.  After a merge phase finishes,
 *		  all TOKCLASS_OKSOFAR classes are converted into
 *		  TOKCLASS_MERGECLEAN.
 */

#include "util.h"
#include "tok-lossy.h"
#include "tokprivate.h"
#include "miscalloc.h"
#include "template.h"
#include "hauschex.h"
#include "class.h"
#include "classtable.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/class.c,v 1.14 2003/01/27 21:48:41 awilliam Exp $");

static ErrCode mergeBox(tokCClass c, tokCBoxType *box, Dgp_Int32 tx, Dgp_Int32 ty,
			tokOptions *opt);
static ErrCode commitClass(TokState state, tokClassTable classes, tokCClass c,
			Dgp_Bool forceCommit);
static ErrCode mergeClasses(TokState state, tokClassTable classes, tokCClass a,
                            tokCClass b, Dgp_Int32 tx, Dgp_Int32 ty, Dgp_Bool *merged);

/*
 * Take a box, and make it into its own class.
 */
ErrCode
tokCBoxToClass(tokCBoxType *box, tokCClass *result, tokOptions *opt)
{
  Dgp_State *dgp_state = opt->dgp_state;  
  tokCClass newClass = (tokCClass)NULL;
	DgpAdt_ListNode retnode;

  DGP_PROLOGUE("tokCBoxToClass()", DGP_PERRSTATE);

  TOKCLASS_SEC_MALLOC(&newClass);

  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&newClass->members, dgp_state));

    newClass->exemplar = box;
    newClass->modelsum = (tokCPixelSum *)NULL;
    newClass->status = TOKCLASS_CHECKCLEAN;
    newClass->userdata = (void *)NULL;

    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(newClass->members, (void *)box, &retnode));

    *result = newClass;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (newClass != (tokCClass)NULL) {
	if (newClass->members != DGP_NULLLIST) {
	    DgpAdt_liFree(newClass->members);
	}
	TOKCLASS_FREE(newClass);
    }

    ERR_RETURN;
}

/*
 * Merge the given box into the given class.
 *
 * This handles accumulating the histogram. It also frees the bitmap
 * of the box if the class has been committed. It does not commit classes
 * itself.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. On failure, the box has not
 * been merged into the class.
 */
static ErrCode
mergeBox(tokCClass c, tokCBoxType *box, Dgp_Int32 tx, Dgp_Int32 ty, tokOptions *opt)
{
    tokCBoxType *member1;
	DgpAdt_ListNode retnode;

DGP_PROLOGUE("mergeBox()", opt->DGP_PERRSTATE);

    if (c->modelsum == (tokCPixelSum *)NULL) {
	/*
	 * This must be the first merge - create the histogram
	 */
	DGP_ASSERT(DGPADT_LILEN(c->members) == 1);
	member1 = tokCMemberBox(DGPADT_LIFIRST(c->members));

DGP_ERR_PASS_CLEANUP(
        ModelToSum(member1->model, opt, &c->modelsum));
    }

    DGP_ASSERT(c->modelsum != (tokCPixelSum *)NULL);

DGP_ERR_PASS_CLEANUP(
    InsertModel(c, box->model, tx, ty, opt));

    /* Add it to the list of members */
	DgpAdt_liAdd(c->members, (void *)box, &retnode);
    if (retnode == DGP_NULLLISTNODE) {
	/* We need to make the class consistent again */
	if (c->modelsum != (tokCPixelSum *)NULL) {
	    DeleteModel(c, box->model, tx, ty, opt);
	}
	DGP_ERR_CLEANUP(cErrNomem);
    }

    /* Set the box's offset */
    box->xOffset = (Dgp_Int16)tx;
    box->yOffset = (Dgp_Int16)ty;

    if (c->status & TOKCLASS_COMMITTED) {
	/* Delete the bitmap */
	FreeModel(box->model, opt);
	FreeImage(box->image, opt);
	box->model = (tokCModelInfo *)NULL;
	box->image = (tokCImageInfo *)NULL;
    }

    /*
     * We *don't* clear TOKCLASS_CHECKCLEAN here. Why not? Well, if it was
     * clean before (everything known to match the exemplar), and we've
     * just added something that matches the exemplar, then it's still
     * clean.
     */

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * This is like checkBoxToBin, except it checks against a single class.
 *
 * It has the same caveats about failure.
 */
ErrCode
tokCMaybeMergeBox(TokState state, tokClassTable classes, tokCClass c, 
                  tokCBoxType *box, Dgp_Bool *matched)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    tokCBoxType *exemplar;
    Dgp_Int32 tx, ty;

DGP_PROLOGUE("tokCMaybeMergeBox()", opt->DGP_PERRSTATE);

    exemplar = c->exemplar;
    DGP_ASSERT(exemplar != (tokCBoxType *)NULL);

    tx = ty = 0;
    if (EdgeHausdorffMatch(exemplar->model, exemplar->image,
			   box->model, box->image, &tx, &ty, opt)) {
	/* They do match - try for the best shift */
	ShiftEdgeHausdorffMatch(exemplar->model, exemplar->image,
				box->model, box->image,
				&tx, &ty, opt);
DGP_ERR_PASS_CLEANUP(
        mergeBox(c, box, tx, ty, opt));

    }
    else {
	*matched = DGP_FALSE;
	NORMAL_RETURN;
    }

    *matched = DGP_TRUE;

    /* We've merged a box into the class. We might want to commit the class. */
    if ((!(c->status & TOKCLASS_COMMITTED)) && (opt->commitThresh > 0) &&
	    (DGPADT_LILEN(c->members) >= (Dgp_UInt32)opt->commitThresh)) {
	/* It's a candidate */
DGP_ERR_PASS_CLEANUP(
	commitClass(state, classes, c, DGP_FALSE));
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Commit the given class.
 *
 * This involves creating a synthetic exemplar, checking to see that every
 * member of the class still matches the exemplar, and throwing out those that
 * don't. The ones that got thrown out just get made into their own classes;
 * they don't get merged into anything else (yet... that happens in a later
 * merge phase). If the class still has enough members, it gets marked
 * as "committed" and all its members' bitmaps get freed. The histogram
 * may also get freed in some cases (based on opt->flags).
 *
 * If forceCommit is DGP_TRUE, then the committing happens even when the class
 * no Dgp_Int32er has enough members after stuff getting thrown out.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. Failure results in consistent
 * data structures, but possibly inconsistent classifications:
 * suppose that rejecting a member fails, due to being unable to create
 * a new class for the rejected member. In this case, the routine recovers
 * by not rejecting the member, and returning an error status (and keeping
 * the class marked as uncommitted). This may result in a class where some
 * members have been rejected, but not others; another possibility is that
 * the exemplar will not represent the class properly.
 */
static ErrCode
commitClass(TokState state, tokClassTable classes, tokCClass c, 
            Dgp_Bool forceCommit)
{
    _tokState *intstate = (_tokState *)state;
    tokOptions *opt = &(intstate->options);
    DgpAdt_ListNode ln;
    tokCBoxType *member;

DGP_PROLOGUE("commitClass()", opt->DGP_PERRSTATE);

    if (c->status & TOKCLASS_COMMITTED) {
	/* What's to do? */
	NORMAL_RETURN;
    }

    /*
     * A strange state that classes can sometimes get into: they're
     * marked as immutable, but not marked as committed.  How can this
     * happen?  Well, if you merge an immutable class into an uncommitted
     * class, the uncommitted class gets marked as immutable and then passed
     * into commitClass.  The expectation is that commitClass will
     * just go through and throw out all the member bitmaps, but not mess
     * around with anything else.  Actually, it's safe to check all the
     * members against the exemplar: the only members with bitmaps used to
     * beDgp_Int32 to the uncommitted class, and so they can safely change their
     * class allegiance.
     */
    if (c->status & TOKCLASS_IMMUTABLE) {
	/* It NEEDS to end up committed, and they should have known that */
	DGP_ASSERT(forceCommit);
    }

    /*
     * Make sure it has an up-to-date exemplar.  This is safe even when the
     * class is immutable: tokCGenerateExemplar will recognise that.
     */
DGP_ERR_PASS_CLEANUP(
    tokCGenerateExemplar(state, c));

    /* and check all the members against the new exemplar */
DGP_ERR_PASS_CLEANUP(
    tokCCheckMembers(classes, c, opt));

    /* All checked. Still worth committing? */
    if ((forceCommit) || ((opt->commitThresh > 0) &&
			  (DGPADT_LILEN(c->members) >= (Dgp_UInt32)opt->commitThresh))) {
	/* Yep - do it. */
	c->status |= TOKCLASS_COMMITTED;

	/* Toast the bitmaps. */
	DGPADT_FOREACH (ln, c->members) {
	    member = tokCMemberBox(ln);

	    /*
	     * It's possible some of the bitmaps are already gone - this
	     * could be the commitClass called after merging a committed
	     * class into an uncommitted one.
	     */
	    if (member->model != (tokCModelInfo *)NULL) {
		FreeModel(member->model, opt);
		member->model = (tokCModelInfo *)NULL;
		}

	    if (member->image != (tokCImageInfo *)NULL) {
		FreeImage(member->image, opt);
		member->image = (tokCImageInfo *)NULL;
		}
	    }

	/*
	 * Make very sure that there's a synthetic exemplar.  The exemplar
	 * might have been some member's bitmap, which no Dgp_Int32er exists.
	 */
	if (!(c->status & TOKCLASS_EXSYNTHETIC)) {
	    /*
	     * Call tokCGenerateExemplar again - since c is now
	     * marked committed, this is a forced generate.
	     */
DGP_ERR_PASS_CLEANUP(
	    tokCGenerateExemplar(state, c));
	}

    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Create a new class for the rejected member. If it's the only member of the
 * class, change the class instead, and return DGP_FALSE in *rejected.
 *
 * Beware that the class could be marked immutable, but not committed yet -
 * this is valid in some annoying merge cases.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. In case of failure, nothing
 * should have been changed.
 */
ErrCode
rejectMember(tokClassTable classes, tokCClass c, tokCBoxType *member, 
	     Dgp_Bool *rejected, tokOptions *opt)
{
  Dgp_State *dgp_state = opt->dgp_state;
    tokCClass newClass = (tokCClass)NULL;

DGP_PROLOGUE("rejectMember()", DGP_PERRSTATE);

    DGP_ASSERT(!(c->status & TOKCLASS_COMMITTED));

    if ((DGPADT_LILEN(c->members) < 2) && (!(c->status & TOKCLASS_IMMUTABLE))) {
	/*
	 * We're rejecting the only member of the class. This means that
	 * not a single thing in the class matches the exemplar. While this is
	 * odd, it's possible. We don't actually throw the member out, but
	 * rather throw out the exemplar.
	 *
	 * Note that this code is absolutely not allowed to fail, since
	 * it's making unrecoverable changes.
	 */
	if (c->status & TOKCLASS_EXSYNTHETIC) {
	    FreeExemplar(c, opt);
	}
	c->exemplar = member;
	c->status &= ~(TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN |
		       TOKCLASS_MERGECLEAN | TOKCLASS_OKSOFAR);
	c->status |= TOKCLASS_CHECKCLEAN;

	/*
	 * If the offsets are non-zero, fry the histogram... it's easier
	 * than changing its offsets at this point. If the offsets are zero,
	 * then the histogram and the (single) member are properly aligned,
	 * so it can live.
	 */
	if ((member->xOffset != 0) || (member->yOffset != 0)) {
	    member->xOffset = member->yOffset = 0;
	    /* How could we get here without a histogram? */
	    DGP_ASSERT(c->modelsum != (tokCPixelSum *)NULL);
	    PixelSumFree(c->modelsum, opt);
	    c->modelsum = (tokCPixelSum *)NULL;
	}

	*rejected = DGP_FALSE;
	NORMAL_RETURN;
	}

DGP_ERR_PASS_CLEANUP(
    tokCBoxToClass(member, &newClass, opt));

DGP_ERR_PASS_CLEANUP(
    tokCTAddToTable(classes, newClass, opt));

    /* Adjust the histogram - only done after all ops that could fail */
    DGP_ASSERT(c->modelsum != (tokCPixelSum *)NULL);
    DeleteModel(c, member->model, member->xOffset, member->yOffset, opt);

    member->xOffset = member->yOffset = 0;

    *rejected = DGP_TRUE;
    NORMAL_RETURN;

ERR_CLEANUP:

    if (newClass != (tokCClass)NULL) {
	/*
	 * Use free_tokCClass, not ClassFree, because ClassFree gets rid of
	 * the stuff hung off the tokCClass, which would be bad because whoever
	 * called us expects the member box to be around after a failure.
	 */
	TOKCLASS_FREE(newClass);
    }

    ERR_RETURN;
}

/*
 * Generate a new exemplar for the given class. Check to see if the old one
 * is still valid.
 */
ErrCode
tokCGenerateExemplar(TokState state, tokCClass c)
{
    _tokState *intstate = (_tokState *)state;
    tokOptions *opt = &(intstate->options);
    tokCBoxType *oldExemplar = c->exemplar;
    Dgp_Int32 oldStatus = c->status;

DGP_PROLOGUE("tokCGenerateExemplar()", opt->DGP_PERRSTATE);

    if (((c->status & TOKCLASS_EXSYNTHETIC) && (c->status & TOKCLASS_EXCLEAN))
	    || (c->status & TOKCLASS_IMMUTABLE)) {
	/* It exists and is clean, or can't or shouldn't be built. Done. */
	NORMAL_RETURN;
    }

    /*
     * Make the exemplar. Force the make if the class is committed - there's
     * no first member bitmap to fall back on.
     */
DGP_ERR_PASS_CLEANUP(
    MakeExemplarFromPixelSum(state, c, (Dgp_Bool)((c->status & TOKCLASS_COMMITTED) ? DGP_TRUE : DGP_FALSE)));

    if (c->exemplar == (tokCBoxType *)NULL) {
	/* MakeExemplarFromPixelSum failed to make a good enough one. */
	DGP_ASSERT(!(c->status & TOKCLASS_COMMITTED));
	c->exemplar = tokCMemberBox(DGPADT_LIFIRST(c->members));
	c->status &= ~(TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN);

    }
    else {
	c->status |= TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN;
    }

    if (c->exemplar != oldExemplar) {
	/* The exemplar changed - the class is no Dgp_Int32er checked clean */
        c->status &= ~(TOKCLASS_CHECKCLEAN | TOKCLASS_MERGECLEAN | 
		       TOKCLASS_OKSOFAR);
    }

    /* Get rid of the old one. Do it after the last failure point. */
    if (oldStatus & TOKCLASS_EXSYNTHETIC) {
	FreeBox(oldExemplar, opt);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    c->exemplar = oldExemplar;
    c->status = oldStatus;

    ERR_RETURN;
}

/*
 * Try to merge b into a.
 * This isn't as simple as it looks: a and b might be committed (and thus be
 * missing all the bitmaps of their members), or even might be committed and
 * have had their histograms removed.
 * - If both a and b are uncommitted (have hists and member bitmaps):
 *	Merge the members of b into a, and possibly commit the resulting class.
 *	Use the exemplar from a (or the one generated by the commit).
 * - If a is uncommitted, but b is committed (doesn't have bitmaps; may not
 *   have hist):
 *	Merge the members of b into a, and merge the histograms if possible.
 *	Commit the resulting class (get rid of a's bitmaps). Use the exemplar
 *	from *b*. This is essentially merging a into b, but the tokCClass structure
 *	left at the end is a's.
 * - If a is committed, but b is uncommitted:
 *	Merge the members of b into a, merge the histograms if possible,
 *	get rid of b's bitmaps, and use the exemplar from a.
 * - If a and b are both committed:
 *	Merge the members and histograms. Use the exemplar from a.
 *
 * a or b might also have been declared immutable (which is kind of a
 * stronger form of committed, where the exemplar is not allowed to change).
 * In this case:
 * - if both a and b are immutable, die horribly - if they get merged,
 *   then some immutable class's element will be represented by a
 *   different exemplar than it had before, which is not allowed.  Actually,
 *   it's theoretically possible that a and b have the identical exemplars,
 *   but due to the way that immutable classes are built, this shouldn't
 *   happen either.
 * - if neither a nor b is immutable, then proceed as above.
 * - if only one of a or b is immutable, then merge them in the same
 *   way that was done when only one class was committed.
 *
 * Combining the committed logic with the immutable logic, and relying on
 * the fact that any immutable class is also committed, the logic goes like:
 * - if both are uncommitted, merge b into a and possibly commit
 * - if only one is committed, merge the other one into it (but always use
 *   a's class structure)
 * - if both are committed, merge them, and use the exemplar from a, unless
 *   b is immutable and a isn't, in which case use b's exemplar; die if both
 *   are immutable.
 *
 * b's tokCClass structure and associated stuff is freed.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. On failure, the two classes
 * may have already been merged - see *merged for whether or not this
 * happened. *merged will always be DGP_TRUE on a successful return.
 */
static ErrCode
mergeClasses(TokState state, tokClassTable classes, tokCClass a, tokCClass b, 
             Dgp_Int32 tx, Dgp_Int32 ty, Dgp_Bool *merged)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    DgpAdt_ListNode ln;
    tokCBoxType *exemplarA, *exemplarB;

DGP_PROLOGUE("mergeClasses()", opt->DGP_PERRSTATE);

    exemplarA = a->exemplar;
    exemplarB = b->exemplar;

    *merged = DGP_FALSE;

    /*
     * Create a's histogram if it's a singleton, and we're not just
     * going to toast it.
     */
    if (a->modelsum == (tokCPixelSum *)NULL) {
	DGP_ASSERT(DGPADT_LILEN(a->members) == 1);

DGP_ERR_PASS_CLEANUP(
	ModelToSum(exemplarA->model, opt, &a->modelsum));
    }

    if (!(a->status & TOKCLASS_COMMITTED)) {
	if (!(b->status & TOKCLASS_COMMITTED)) {
	    /* Oh boy - an easy one */

	    /* a must have a histogram by now */
	    DGP_ASSERT(a->modelsum != (tokCPixelSum *)NULL);

	    /* b might be a singleton */
	    if (b->modelsum == (tokCPixelSum *)NULL) {
		/* A Dgp_Int16-cut to merge in a singleton class */
		DGP_ASSERT(DGPADT_LILEN(b->members) == 1);
		DGP_ASSERT(tokCMemberBox(DGPADT_LIFIRST(b->members))->xOffset == 0);
		DGP_ASSERT(tokCMemberBox(DGPADT_LIFIRST(b->members))->yOffset == 0);
DGP_ERR_PASS_CLEANUP(
		mergeBox(a, tokCMemberBox(DGPADT_LIFIRST(b->members)),
			      tx, ty, opt));
	    }
	    else {
		/* Merge the histograms */
DGP_ERR_PASS_CLEANUP(
		MergePixelSumsWithOffset(a, b, tx, ty, opt));

		/* Change the offsets of b's members */
		DGPADT_FOREACH (ln, b->members) {
		    tokCMemberBox(ln)->xOffset += (Dgp_Int16)tx;
		    tokCMemberBox(ln)->yOffset += (Dgp_Int16)ty;
		}
		
		DgpAdt_liApp(a->members, b->members);
		b->members = DGP_NULLLIST;
	    }

	    ClassFree(b, opt);

	    /* They're merged. Try to commit the new class. */
	    *merged = DGP_TRUE;
	    if ((opt->commitThresh > 0) &&
    		    (DGPADT_LILEN(a->members) >= (Dgp_UInt32)opt->commitThresh)) {
		/* It's a candidate */
DGP_ERR_PASS_CLEANUP(		
	        commitClass(state, classes, a, DGP_TRUE));
	    }
	}
	else {
	    /* b is committed */
	    DGP_ASSERT(b->modelsum != (tokCPixelSum *)NULL);

	    /* Merge a's histogram into b's histogram */
DGP_ERR_PASS_CLEANUP(
	    MergePixelSumsWithOffset(b, a, -tx, -ty, opt));

	    PixelSumFree(a->modelsum, opt);
	    a->modelsum = b->modelsum;
	    b->modelsum = (tokCPixelSum *)NULL;

	    /*
	     * Change the offsets of a's members because we're going to use
	     * b's exemplar.
	     */
	    DGPADT_FOREACH (ln, a->members) {
		tokCMemberBox(ln)->xOffset -= (Dgp_Int16)tx;
		tokCMemberBox(ln)->yOffset -= (Dgp_Int16)ty;
	    }

	    DgpAdt_liApp(a->members, b->members);
	    b->members = DGP_NULLLIST;

	    FreeExemplar(a, opt);
	    a->exemplar = b->exemplar;
	    b->exemplar = (tokCBoxType *)NULL;

	    /* Copy the state of the exemplar */
	    DGP_ASSERT(b->status & TOKCLASS_EXSYNTHETIC);
	    a->status |= TOKCLASS_EXSYNTHETIC;

	    /*
	     * Copy b's histogram status, etc.
	     * The new "a" class is really just a renamed "b", so its
	     * merge-clean status should match b's.
	     */
	    a->status &= ~(TOKCLASS_FROZENHIST | TOKCLASS_MERGECLEAN |
			   TOKCLASS_OKSOFAR);
	    a->status |= b->status & (TOKCLASS_FROZENHIST | 
				      TOKCLASS_MERGECLEAN |
				      TOKCLASS_OKSOFAR);

	    /* and make sure a gets marked as immutable if b was */
	    a->status |= (b->status & TOKCLASS_IMMUTABLE);
	    DGP_ASSERT(a->userdata == (void *)NULL);
	    a->userdata = b->userdata;

	    /* Mark it as dirty *before* the class commit */
	    a->status &= ~(TOKCLASS_EXCLEAN);

	    ClassFree(b, opt);

	    *merged = DGP_TRUE;

	    /* Now commit the result */
DGP_ERR_PASS_CLEANUP(
	    commitClass(state, classes, a, DGP_TRUE));
	    /* Note that the commit may generate a new, clean, exemplar. */
	}
    }
    else {
	/* a is committed. */
	DGP_ASSERT(a->modelsum != (tokCPixelSum *)NULL);

	if (b->status & TOKCLASS_IMMUTABLE) {
	    /* We must merge a into b */
	    DGP_ASSERT(!(a->status & TOKCLASS_IMMUTABLE));

DGP_ERR_PASS_CLEANUP(
	    MergePixelSumsWithOffset(b, a, -tx, -ty, opt));

	    PixelSumFree(a->modelsum, opt);
	    a->modelsum = b->modelsum;
	    b->modelsum = (tokCPixelSum *)NULL;

	    /*
	     * Change the offsets of a's members because we're going to use
	     * b's exemplar.
	     */
	    DGPADT_FOREACH (ln, a->members) {
		tokCMemberBox(ln)->xOffset -= (Dgp_Int16)tx;
		tokCMemberBox(ln)->yOffset -= (Dgp_Int16)ty;
	    }

	    DgpAdt_liApp(a->members, b->members);
	    b->members = DGP_NULLLIST;

	    FreeExemplar(a, opt);
	    a->exemplar = b->exemplar;
	    b->exemplar = (tokCBoxType *)NULL;

	    a->status = b->status;
	    DGP_ASSERT(a->userdata == (void *)NULL);
	    a->userdata = b->userdata;

	    ClassFree(b, opt);

	    *merged = DGP_TRUE;
	}
	else {
	    if (b->modelsum == (tokCPixelSum *)NULL) {
		DGP_ASSERT(DGPADT_LILEN(b->members) == 1);
		DGP_ASSERT(tokCMemberBox(DGPADT_LIFIRST(b->members))->xOffset == 0);
		DGP_ASSERT(tokCMemberBox(DGPADT_LIFIRST(b->members))->yOffset == 0);
DGP_ERR_PASS_CLEANUP(
		mergeBox(a, tokCMemberBox(DGPADT_LIFIRST(b->members)),
			      tx, ty, opt));
	    }
	    else {
DGP_ERR_PASS_CLEANUP(
		MergePixelSumsWithOffset(a, b, tx, ty, opt));

		/* Change the offsets of b's members, and free their images */
		DGPADT_FOREACH (ln, b->members) {
		    tokCBoxType *member = tokCMemberBox(ln);

		    member->xOffset += (Dgp_Int16)tx;
		    member->yOffset += (Dgp_Int16)ty;

		    if (!(b->status & TOKCLASS_COMMITTED)) {
			DGP_ASSERT(member->model != (tokCModelInfo *)NULL);
			FreeModel(member->model, opt);
			member->model = (tokCModelInfo *)NULL;

			DGP_ASSERT(member->image != (tokCImageInfo *)NULL);
			FreeImage(member->image, opt);
			member->image = (tokCImageInfo *)NULL;
			}
		    else {
			DGP_ASSERT(member->model == (tokCModelInfo *)NULL);
			DGP_ASSERT(member->image == (tokCImageInfo *)NULL);
			}
		    }

		DgpAdt_liApp(a->members, b->members);
		b->members = DGP_NULLLIST;
		}

	    ClassFree(b, opt);

	    *merged = DGP_TRUE;
	}
    }
    
    /*
     * Even if b's exemplar matched a's exemplar, possibly not every member
     * of b matches a's exemplar.
     */
    a->status &= ~TOKCLASS_CHECKCLEAN;

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Find the best translation of every member's bitmap wrt the exemplar.
 *
 * Make sure to update the histogram, but don't rebuild the exemplar..
 */
ErrCode
tokCAlignClass(tokCClass c, tokOptions *opt)
{
    DgpAdt_ListNode ln;
    tokCBoxType *exemplar;
    tokCBoxType *member;
    Dgp_Int32 tx, ty;

    DGP_PROLOGUE("tokCAlignClass()", opt->DGP_PERRSTATE);
    DGP_STOP_WHINE;

    if (c->status & TOKCLASS_COMMITTED) {
	/* No bitmaps - no alignment. */
	NORMAL_RETURN;
    }

    exemplar = c->exemplar;
DGP_ASSERT(exemplar != (tokCBoxType *)NULL);

    DGPADT_FOREACH (ln, c->members) {
	member = tokCMemberBox(ln);
	DGP_ASSERT(member->model != (tokCModelInfo *)NULL);
	DGP_ASSERT(member->image != (tokCImageInfo *)NULL);

	tx = member->xOffset;
	ty = member->yOffset;
	ShiftEdgeHausdorffMatch(exemplar->model, exemplar->image,
				member->model, member->image,
				&tx, &ty, opt);
	/* Found a good/better offset */
	if ((tx != member->xOffset) || (ty != member->yOffset)) {
	    /* Insert before delete so the histogram is never empty */
	    InsertModel(c, member->model, tx, ty, opt);
	    DeleteModel(c, member->model,
			member->xOffset, member->yOffset, opt);

	    member->xOffset = (Dgp_Int16)tx;
	    member->yOffset = (Dgp_Int16)ty;
	}
    }

    NORMAL_RETURN;   
}

/*
 * This checks to see if class b can be merged into a. If so, it does it.
 *
 * It has the same caveats about failure as checkBoxToBin.
 */
ErrCode
tokCMaybeMergeClasses(TokState state, tokClassTable classes, tokCClass a, 
                      tokCClass b, Dgp_Bool *merged)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    tokCBoxType *exemplarA, *exemplarB;
    Dgp_Int32 tx, ty;

DGP_PROLOGUE("tokCMaybeMergeClasses()", opt->DGP_PERRSTATE);

    exemplarA = a->exemplar;
    exemplarB = b->exemplar;
    DGP_ASSERT(exemplarA != (tokCBoxType *)NULL);
    DGP_ASSERT(exemplarB != (tokCBoxType *)NULL);
    if ((a->status & TOKCLASS_IMMUTABLE) && (b->status & TOKCLASS_IMMUTABLE)) {
	*merged = DGP_FALSE;
	NORMAL_RETURN;
    }

/*    DGP_WARNF3("compare %08lx(%2d,%2d) ",
	  a, a->exemplar->width, a->exemplar->height);
	DGP_WARNF3("with %08lx(%2d,%2d)",
	  b, b->exemplar->width, b->exemplar->height); */
    tx = ty = 0;
    if (EdgeHausdorffMatch(exemplarA->model, exemplarA->image,
			   exemplarB->model, exemplarB->image,
			   &tx, &ty, opt)) {
	/* They match. Find the best shift. */
	ShiftEdgeHausdorffMatch(exemplarA->model, exemplarA->image,
				exemplarB->model, exemplarB->image,
				&tx, &ty, opt);
DGP_ERR_PASS_CLEANUP(
	mergeClasses(state, classes, a, b, tx, ty, merged));
    }
    else {
	*merged = DGP_FALSE;
    }

/*    if (*merged) {
	DGP_WARNF0(" merged\n");
	}
    else {
	DGP_WARNF0(" not merged\n");
	} */

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * Check that every member of the class matches the exemplar. If any
 * do not, throw them out into their own classes.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure; see the caveat in commitClass
 * about the consequences of failure.
 */
ErrCode
tokCCheckMembers(tokClassTable classes, tokCClass c, tokOptions *opt)
{
    DgpAdt_ListNode ln;
    tokCBoxType *member;
    tokCBoxType *exemplar;
    Dgp_Int32 tx, ty;
    Dgp_Bool rejected;

DGP_PROLOGUE("tokCCheckMembers()", opt->DGP_PERRSTATE);

    if (c->status & TOKCLASS_COMMITTED) {
	/* Can't reject anything from a committed class */
	NORMAL_RETURN;
    }

    if (c->status & TOKCLASS_CHECKCLEAN) {
	/* It was checked, and hasn't changed since */
	NORMAL_RETURN;
    }

    exemplar = c->exemplar;

    for (ln = DGPADT_LIFIRST(c->members); ln != DGP_NULLLISTNODE; ) {
	member = tokCMemberBox(ln);
	/*
	 * In some cases, we can get here and find that the member's model
	 * doesn't exist. This happens when a committed class gets merged
	 * into an uncommitted class, and the result gets committed.
	 */
	if (member->model != (tokCModelInfo *)NULL) {
	    tx = member->xOffset;
	    ty = member->yOffset;
	    if (EdgeHausdorffMatch(exemplar->model, exemplar->image,
				   member->model, member->image,
				   &tx, &ty, opt)) {
		/* It's fine. */
		ln = DGPADT_LINEXT(ln);
		/*
		 * We could check to see if tx and ty have changed from
		 * xoffset, yoffset...
		 */
		}
	    else {
		/* Time to get rid of it. */
DGP_ERR_PASS_CLEANUP(
		rejectMember(classes, c, member, &rejected, opt));

		if (rejected) {
		    DgpAdt_liRemAndNext(ln, &ln);
		}
		else {
		    ln = DGPADT_LINEXT(ln);
		    DGP_ASSERT(ln == DGP_NULLLISTNODE);
		}
	    }
	}
	else {
	    ln = DGPADT_LINEXT(ln);
	}
    }

    /* It's clean now. */
    c->status |= TOKCLASS_CHECKCLEAN;

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Publically visible functions from tokclass.h:
 */

ErrCode
tokCMergeClasses(TokState state, tokCClass a, tokCClass b, Dgp_Int32 tx, Dgp_Int32 ty, 
		 Dgp_Bool *merged) 
{
    _tokState *intstate = (_tokState *)state;

DGP_PROLOGUE("tokCMergeClasses()", intstate->options.dgp_state->err_state);

DGP_ERR_PASS_CLEANUP(
    mergeClasses( state, &intstate->classes, a, b, tx, ty, merged));

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


ErrCode
tokCFreeClass(TokState state, tokCClass c) 
{
    _tokState *intstate = (_tokState *)state;

    ClassFree(c, &(intstate->options));

    return cSuccessful;
}

ErrCode
tokCNewClass(TokState state, tokCBoxType *box, tokCClass *c) 
{
    _tokState *intstate = (_tokState *)state;
    tokCClass newClass = (tokCClass)NULL;

DGP_PROLOGUE("tokCNewClass()", intstate->options.dgp_state->err_state);

DGP_ERR_PASS_CLEANUP(
    tokCBoxToClass(box, &newClass, &(intstate->options)));

    *c = newClass;
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
tokCSetExemplar(TokState state, tokCClass c, tokCBoxType* box, Dgp_Int32 tx, Dgp_Int32 ty,
		Dgp_Bool tokOwner) 
{
    DgpAdt_ListNode ln;
    tokCBoxType *member;
    tokOptions *opt = &((_tokState *)state)->options;

    DGP_PROLOGUE("tokCSetExemplar()", opt->DGP_PERRSTATE);
    DGP_STOP_WHINE;

    /* Can't change an exemplar once a class is marked immutable */
    DGP_ASSERT(!(c->status & TOKCLASS_IMMUTABLE));

    /* Switch the exemplar and free the old one */
    if( (c->exemplar != (tokCBoxType*)NULL) &&
	    (c->status & TOKCLASS_EXSYNTHETIC)) {
	FreeExemplar( c, opt);
    }
    c->exemplar = box;

    if( (tx != 0) || (ty != 0))
    {
	/* Adjust offsets */
	DGPADT_FOREACH( ln, c->members) {
	    member = (tokCBoxType*)DGPADT_LIGET(ln);

	    InsertModel(c, member->model, tx, ty, opt);
	    DeleteModel(c, member->model, member->xOffset, 
			member->yOffset, opt);

	    member->xOffset = (Dgp_Int16)tx;
	    member->yOffset = (Dgp_Int16)ty;
	}
    }

    /* Update the status bits */
    c->status &= ~(TOKCLASS_CHECKCLEAN | TOKCLASS_MERGECLEAN | TOKCLASS_OKSOFAR
	    | TOKCLASS_EXCLEAN);

    /* If we are to assume ownership of the exemplar, then set 
     * the synthetic flag. */
    if( tokOwner) {
    	c->status |= TOKCLASS_EXSYNTHETIC;
    }

    NORMAL_RETURN;
}
