/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/pick.c,v 1.3 2002/10/11 17:55:40 jlin Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992, 1994 Xerox Corporation.  All Rights Reserved.
 *
 * Use, reproduction, preparation of derivative works, and distribution
 * of this software is permitted.  Any copy of this software or of any
 * derivative work must include both the above copyright notices of
 * Cornell University and Xerox Corporation and this paragraph.  Any
 * distribution of this software or derivative works must comply with all
 * applicable United States export control laws.  This software is made
 * available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
 * LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#include "dgp_err_control.h"
#include "util.h"
#include "pick.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/pick.c,v 1.3 2002/10/11 17:55:40 jlin Exp $");

/*
 * Pick the (whichval)'th smallest element from vals. If whichval = 0,
 * pick the smallest; if whichval = 1, pick the second smallest, etc.
 * If whichval < 0, return something small.
 */
Dgp_Int32
pickNth(Dgp_Int32 *vals, Dgp_Int32 nvals, Dgp_Int32 whichval){
    Dgp_Int32 low;
    Dgp_Int32 high;
    Dgp_Int32 pivot;
    Dgp_Int32 bottom;
    Dgp_Int32 top;
    Dgp_Int32 tmp;

    DGP_PROLOGUE ("pickNth", NULL);
    DGP_STOP_WHINE;

    /*
     * This is similar to a quicksort, but we zoom in on the
     * value requested, and do not do a complete sort.
     */
    low = 0;
    high = nvals - 1;
    if (low > high) {
	/* Nothing to do - return something arbitrary */
	return(0);
	}

    if (whichval > high) {
	whichval = high;
	}
    if (whichval < low) {
	return(0);
	}

    while (DGP_TRUE) {
	DGP_ASSERT(low <= high);

	if (whichval == low) {
	    /* We want the minimum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = MIN(pivot, vals[low]);
		}
	    return(pivot);
	    }
	else if (whichval == high) {
	    /* We want the maximum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = MAX(pivot, vals[low]);
		}
	    return(pivot);
	    }

	/* Pick a pivot */
	pivot = vals[low];
	/* Make sure that there is at least one thing smaller than pivot */
	for (bottom = low + 1; bottom <= high; bottom++) {
	    if (vals[bottom] > pivot) {
		pivot = vals[bottom];
		break;
		}
	    else if (vals[bottom] < pivot) {
		break;
		}
	    }
	if (bottom == high + 1) {
	    /* All the same - return the (single) value */
	    return(pivot);
	    }

	bottom = low;
	top = high;
	while (bottom < top) {
	    /*
	     * Skip over elements that are in the right place.
	     * We know that these loops will terminate safely since there
	     * is at least one value in the array smaller than pivot,
	     * and it is below top, so the first loop will terminate,
	     * and there is at least one value equal to pivot,
	     * and it is at or higher than bottom, so the second loop
	     * will terminate.
	     */
	    while (vals[top] >= pivot) {
		top--;
		}
	    while (vals[bottom] < pivot) {
		bottom++;
		}

	    if (top >= bottom) {
		/* Swap the values at top and bottom */
		tmp = vals[top];
		vals[top] = vals[bottom];
		vals[bottom] = tmp;
		}
	    }

	/* At this poDgp_Int32, top and bottom have just crossed */
	DGP_ASSERT(top == bottom - 1);

	/* vals[low..top] are < pivot; vals[bottom..high] are >= pivot */
	if (top >= whichval) {
	    high = top;
	    }
	else {
	    low = bottom;
	    }
	}
    }

/*
 * Exactly the same, but with floats. Sigh.
 */
Dgp_Float32
pickNthFloat(Dgp_Float32 *vals, Dgp_Int32 nvals, Dgp_Int32 whichval)
{
    Dgp_Int32 low;
    Dgp_Int32 high;
    Dgp_Float32 pivot;
    Dgp_Int32 bottom;
    Dgp_Int32 top;
    Dgp_Float32 tmp;

	DGP_VOID_PROLOGUE ("pickNthFloat", NULL);

    /*
     * This is similar to a quicksort, but we zoom in on the
     * value requested, and do not do a complete sort.
     */
    low = 0;
    high = nvals - 1;
    if (low > high) {
	/* Nothing to do - return something arbitrary */
	return(0);
	}

    if (whichval > high) {
	whichval = high;
	}
    if (whichval < low) {
	return(0);
	}

    while (DGP_TRUE) {
	DGP_VOID_ASSERT(low <= high);

	if (whichval == low) {
	    /* We want the minimum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = MIN(pivot, vals[low]);
		}
	    return(pivot);
	    }
	else if (whichval == high) {
	    /* We want the maximum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = MAX(pivot, vals[low]);
		}
	    return(pivot);
	    }

	/* Pick a pivot */
	pivot = vals[low];
	/* Make sure that there is at least one thing smaller than pivot */
	for (bottom = low + 1; bottom <= high; bottom++) {
	    if (vals[bottom] > pivot) {
		pivot = vals[bottom];
		break;
		}
	    else if (vals[bottom] < pivot) {
		break;
		}
	    }
	if (bottom == high + 1) {
	    /* All the same - return the (single) value */
	    return(pivot);
	    }

	bottom = low;
	top = high;
	while (bottom < top) {
	    /*
	     * Skip over elements that are in the right place.
	     * We know that these loops will terminate safely since there
	     * is at least one value in the array smaller than pivot,
	     * and it is below top, so the first loop will terminate,
	     * and there is at least one value equal to pivot,
	     * and it is at or higher than bottom, so the second loop
	     * will terminate.
	     */
	    while (vals[top] >= pivot) {
		top--;
		}
	    while (vals[bottom] < pivot) {
		bottom++;
		}

	    if (top >= bottom) {
		/* Swap the values at top and bottom */
		tmp = vals[top];
		vals[top] = vals[bottom];
		vals[bottom] = tmp;
		}
	    }

	/* At this poDgp_Int32, top and bottom have just crossed */
	DGP_VOID_ASSERT(top == bottom - 1);

	/* vals[low..top] are < pivot; vals[bottom..high] are >= pivot */
	if (top >= whichval) {
	    high = top;
	    }
	else {
	    low = bottom;
	    }
	}
    }


/*
 * Exactly the same, but with a general comparison function. The above aren't
 * coded in terms of this one because they're used in inner loops in some
 * important applications.
 *
 * Pass this an array of poDgp_Int32ers and let it find the whichval'th element
 * (numbered 0==smallest).
 *
 * The function "compare" should return <0, 0, or >0 depending on how its
 * first argument is lexicographically smaller than, equal to, or greater
 * than the second.
 *
 * The result is a poDgp_Int32er to the chosen element. The order of the elements
 * in the array is not preserved: this routine will shuffle things around.
 */
void *
pickNthGeneral(void **vals, Dgp_Int32 nvals, Dgp_Int32 whichval,
	       Dgp_Int32 (*compare)(void *p1, void *p2))
{
    Dgp_Int32 low;
    Dgp_Int32 high;
    void *pivot;
    Dgp_Int32 bottom;
    Dgp_Int32 top;
    void *tmp;

	DGP_VOID_PROLOGUE ("pickNthGeneral", NULL);

    /*
     * This is similar to a quicksort, but we zoom in on the
     * value requested, and do not do a complete sort.
     */
    low = 0;
    high = nvals - 1;
    if (low > high) {
	/* Nothing to do - return something arbitrary */
	return((void *)NULL);
	}

    if (whichval > high) {
	/*
	 * This is common enough that it shouldn't be an error:
	 * people pass in n elements and ask for number n, whereas the
	 * highest is really number n-1.
	 */
	whichval = high;
	}
    if (whichval < low) {
	return((void *)NULL);
	}

    while (DGP_TRUE) {
	DGP_VOID_ASSERT(low <= high);

	if (whichval == low) {
	    /* We want the minimum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = ((*compare)(pivot, vals[low]) < 0) ? pivot : vals[low];
		}
	    return(pivot);
	    }
	else if (whichval == high) {
	    /* We want the maximum of vals[low..high] */
	    pivot = vals[low];
	    for (low++; low <= high; low++) {
		pivot = ((*compare)(pivot, vals[low]) > 0) ? pivot : vals[low];
		}
	    return(pivot);
	    }

	/* Pick a pivot */
	pivot = vals[low];
	/* Make sure that there is at least one thing smaller than pivot */
	for (bottom = low + 1; bottom <= high; bottom++) {
	    if ((*compare)(vals[bottom], pivot) > 0) {
		pivot = vals[bottom];
		break;
		}
	    else if ((*compare)(vals[bottom], pivot) < 0) {
		break;
		}
	    }
	if (bottom == high + 1) {
	    /* All the same - return the (single) value */
	    return(pivot);
	    }

	bottom = low;
	top = high;
	while (bottom < top) {
	    /*
	     * Skip over elements that are in the right place.
	     * We know that these loops will terminate safely since there
	     * is at least one value in the array smaller than pivot,
	     * and it is below top, so the first loop will terminate,
	     * and there is at least one value equal to pivot,
	     * and it is at or higher than bottom, so the second loop
	     * will terminate.
	     */
	    while ((*compare)(vals[top], pivot) >= 0) {
		top--;
		}
	    while ((*compare)(vals[bottom], pivot) < 0) {
		bottom++;
		}

	    if (top >= bottom) {
		/* Swap the values at top and bottom */
		tmp = vals[top];
		vals[top] = vals[bottom];
		vals[bottom] = tmp;
		}
	    }

	/* At this poDgp_Int32, top and bottom have just crossed */
	DGP_VOID_ASSERT(top == bottom - 1);

	/* vals[low..top] are < pivot; vals[bottom..high] are >= pivot */
	if (top >= whichval) {
	    high = top;
	    }
	else {
	    low = bottom;
	    }
	}
    }

Dgp_Float64
frac_lower(Dgp_Int32 *vals, Dgp_Int32 nvals, Dgp_Int32 upper)
{
    Dgp_Int32 i;
    Dgp_Int32 lowercount;
    DGP_PROLOGUE ("frac_lower", NULL);
    DGP_STOP_WHINE;

    DGP_ASSERT(vals != (Dgp_Int32 *)NULL);

    if (nvals == 0) {
	return(1.);
	}

    lowercount = 0;
    for (i = 0; i < nvals; i++) {
	if (vals[i] <= upper) {
	    lowercount++;
	    }
	}

    return((Dgp_Float64)lowercount / (Dgp_Float64)nvals);
    }

