/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/template.c,v 1.48 2003/02/10 22:53:41 mkrocken Exp $
 *
 */

/*
 * Assorted authors: Mike Leventon, William Rucklidge
 */
#include "tokprivate.h"
#include "template.h"
#include "pick.h"
#include "miscalloc.h"
#include "image.h"
#include "hauschex.h"
#include "cross.h"
#include "image_defs.h"
#include "PsgFix.h"
#include <stdlib.h>

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/template.c,v 1.48 2003/02/10 22:53:41 mkrocken Exp $");

#ifdef	DBUG
#include "dbug.h"

Dgp_Int32 maxhist = 0;
Dgp_Int32 curhist = 0;

extern Dgp_Int32 maxbitmap;
extern Dgp_Int32 curbitmap;
#endif

/* How much slop to put around the edges of the histograms */
#define MAXXDIFF         2
#define MAXYDIFF         2

#define maskBit(o)     (0x80000000 >> (o))

/*
 * Some note on histogram states:
 * tokCClass histograms can be in one of a number of states.
 * 1.  No tokCPixelSum allocated, no histogram image allocated
 * 2.  tokCPixelSum allocated, no histogram image allocated, member
 *     bitmaps present.
 * 3.  tokCPixelSum allocated, no histogram image allocated, member
 *     bitmaps absent
 * 4.  tokCPixelSum allocated, histogram image allocated, member bitmaps
 *     present.
 * 5.  tokCPixelSum allocated, histogram image allocated, member bitmaps
 *     absent.
 *
 * State 1 is reserved for singleton, uncommitted, classes: they have one
 * member, and that member's bitmap is still around, so the tokCPixelSum and
 * histogram image can be allocated as needed.
 *
 * State 2 is the state that uncommitted classes are in when they have too
 * few members to create the histogram image (lazy image creation).
 *
 * State 3 is the state that immutable classes are in when tokResume is handed
 * a class that has no histogram attached to it: it has nothing to put in the
 * image, and there are (initially) no members to be used to create it.  This
 * can only happen if the user didn't specify TOK_OPT_RETURNHIST.  In this
 * case, they don't want a histogram back at the end, and there's no point
 * in our tracking the immutable class' histogram... we'll never generate
 * an exemplar using it.  If the user wanted us to track immutable classes'
 * histograms, they should have specified TOK_OPT_RETURNHIST, and therefore
 * passed classes with histograms back into tokResume.
 *
 * State 4 is the state that uncommitted classes are in when they have
 * enough members to create the histogram image.
 *
 * State 5 is the state that committed, mutable, classes are normally in.
 *
 */

/*
 * Take the given model and create a tokCPixelSum containing it, of some
 * appropriate size. The histogram's size is based on the width and
 * height of the model *without* the dilation padding on the left, but
 * with its origin corresponding to the top left corner of the model's
 * bitmap.
 */
ErrCode
ModelToSum(tokCModelInfo *model, tokOptions *opt, tokCPixelSum **result)
{
    Dgp_State *dgp_state = opt->dgp_state;
    tokCPixelSum *newSum = (tokCPixelSum *)NULL;
    Dgp_Int32 width = model->xsize - opt->dilationRadius;
    Dgp_Int32 height = model->ysize;
    DgpImage_Gray newbyte = (DgpImage_Gray)NULL;
    Dgp_Int32 x, y;

  DGP_PROLOGUE("ModelToSum()", DGP_PERRSTATE);

  TOKPXL_SEC_MALLOC(&newSum);

    /* Do we create the histogram now, or defer it? */
    if (opt->makeHistThresh <= 1) {
	/*
	 * Create the image. Put some border around it.
	 */
  DGP_ERR_PASS_CLEANUP(
	DgpImage_newOffset(opt->dgp_state, DGPIMAGE_GRAY,
					 (unsigned)width + 2 * MAXXDIFF,
					 (unsigned)height + 2 * MAXYDIFF,
					 opt->dilationRadius - MAXXDIFF,
					 -MAXYDIFF, (void **)&newbyte));
#ifdef	DBUG
	curhist += ((width + 2 * MAXXDIFF) + sizeof(Dgp_Int8 *)) * (height + 2 * MAXYDIFF) * sizeof(Dgp_Int8) + sizeof(*newbyte);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	newSum->sum.bytesum = newbyte;
    }
    else {
	newSum->sum.bytesum = (DgpImage_Gray)NULL;
    }

    newSum->sumisbyte = DGP_TRUE;

    newSum->minx = (Dgp_Int16)opt->dilationRadius;
    newSum->miny = 0;
    newSum->maxx = model->xsize - 1;
    newSum->maxy = model->ysize - 1;

    newSum->nummodels = 1;

    if (newSum->nummodels < opt->makeHistThresh) {
	/* We don't create anything for it */
	DGP_ASSERT(newbyte == (DgpImage_Gray)NULL);
	*result = newSum;
        NORMAL_RETURN;
    }

    /* We created one - stick in the bits. */
    DGP_ASSERT(newbyte != (DgpImage_Gray)NULL);
    for (y = 0; y < model->ysize; y++) {
	for (x = 0; x < model->xsize; x++) {
	    if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
		(1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
		DGP_ASSERT(x >= opt->dilationRadius);
		DGPIMG_REF(newbyte, x, y)++;
            }
        }
    }

    *result = newSum;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (newSum != (tokCPixelSum *)NULL) {
	TOKPXL_FREE(newSum);
    }

    if (newbyte != (DgpImage_Gray)NULL) {
	DgpImage_free(newbyte);
    }

    ERR_RETURN;
}

/*
 * Create the histogram, for real; accumulate the info from the class into it.
 */
ErrCode
makeRealHist(tokCClass c, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    DgpImage_Short newShort;
    DgpImage_Gray newbyte;
    tokCPixelSum *sum = c->modelsum;
    DgpAdt_ListNode ln;
    tokCBoxType *member;
    tokCModelInfo *model;
    Dgp_Int32 tx, ty;

DGP_PROLOGUE("makeRealHist()", DGP_PERRSTATE);

    DGP_ASSERT(sum != (tokCPixelSum *)NULL);
    DGP_ASSERT(sum->sumisbyte ?
	   (sum->sum.bytesum == (DgpImage_Gray)NULL) :
	   (sum->sum.Shortsum == (DgpImage_Short)NULL));

    if (c->status & TOKCLASS_IMMUTABLE) {
	/* You don't want one - really */
	NORMAL_RETURN;
    }

    DGP_ASSERT(!(c->status & TOKCLASS_COMMITTED));

    if (sum->nummodels < DGPIMG_COLRNG) {
	sum->sumisbyte = DGP_TRUE;
    }
    else {
	sum->sumisbyte = DGP_FALSE;
    }

    /* Create the histogram */
    if (sum->sumisbyte) {
DGP_ERR_PASS_CLEANUP(
	DgpImage_newOffset(dgp_state, DGPIMAGE_GRAY,
					 ((unsigned)(sum->maxx - sum->minx) +
					  1 + 2 * MAXXDIFF),
					 ((unsigned)(sum->maxy - sum->miny) +
					  1 + 2 * MAXYDIFF),
					 sum->minx - MAXXDIFF,
					 sum->miny - MAXYDIFF, (void**) &newbyte));
#ifdef	DBUG
	curhist += ((sum->maxx - sum->minx + 1 + 2 * MAXXDIFF) + sizeof(Dgp_Int8 *)) * (sum->maxy - sum->miny + 1 + 2 * MAXYDIFF) * sizeof(Dgp_Int8) + sizeof(*newbyte);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	/* Accumulate into it */
	DGPADT_FOREACH (ln, c->members) {
	    member = (tokCBoxType *)DGPADT_LIGET(ln);
	    model = member->model;
	    tx = member->xOffset;
	    ty = member->yOffset;
	    for (y = 0; y < model->ysize; y++) {
		for (x = 0; x < model->xsize; x++) {
		    if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
			(1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
			DGP_ASSERT(x >= opt->dilationRadius);
			DGPIMG_REF(newbyte, x + tx, y + ty)++;
                    }
                }
            }
        }

	sum->sum.bytesum = newbyte;
    }
    else {
DGP_ERR_PASS_CLEANUP(
	DgpImage_newOffset(dgp_state, DGPIMAGE_SHORT,
					   ((unsigned)(sum->maxx - sum->minx) +
					    1 + 2 * MAXXDIFF),
					   ((unsigned)(sum->maxy - sum->miny) +
					    1 + 2 * MAXYDIFF),
					   sum->minx - MAXXDIFF,
					   sum->miny - MAXYDIFF, (void **)&newShort));
#ifdef	DBUG
	curhist += ((sum->maxx - sum->minx + 1 + 2 * MAXXDIFF) + sizeof(Dgp_Int16 *)) * (sum->maxy - sum->miny + 1 + 2 * MAXYDIFF) * sizeof(Dgp_Int16) + sizeof(*newShort);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	/* Accumulate into it */
	DGPADT_FOREACH (ln, c->members) {
	    member = (tokCBoxType *)DGPADT_LIGET(ln);
	    model = member->model;
	    tx = member->xOffset;
	    ty = member->yOffset;
	    for (y = 0; y < model->ysize; y++) {
		for (x = 0; x < model->xsize; x++) {
		    if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
			(1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
			DGP_ASSERT(x >= opt->dilationRadius);
			DGPIMG_REF(newShort, x + tx, y + ty)++;
                    }
                }
            }
        }

	sum->sum.Shortsum = newShort;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    
    ERR_RETURN;
}

/*
 * Get rid of the thing's histogram - it's under the needs-histogram threshold.
 *
 * You can only do this when the member bitmaps are still around; otherwise,
 * you're destroying information.
 */
static void
destroyRealHist(tokCPixelSum *sum, tokOptions *opt)
{
DGP_VOID_PROLOGUE( "destroyRealHist()", opt->DGP_PERRSTATE);

    DGP_VOID_ASSERT(sum->nummodels < opt->makeHistThresh);

    if (sum->sumisbyte) {
#ifdef	DBUG
	curhist -= (DGPIMG_WIDTH(sum->sum.bytesum) + sizeof(Dgp_Int8 *)) * DGPIMG_HEIGHT(sum->sum.bytesum) * sizeof(Dgp_Int8) + sizeof(*sum->sum.bytesum);
#endif
	DgpImage_free(sum->sum.bytesum);
	sum->sum.bytesum = (DgpImage_Gray)NULL;
    }
    else {
#ifdef	DBUG
	curhist -= (DGPIMG_WIDTH(sum->sum.Shortsum) + sizeof(Dgp_Int16 *)) * DGPIMG_HEIGHT(sum->sum.Shortsum) * sizeof(Dgp_Int16) + sizeof(*sum->sum.Shortsum);
#endif
	DgpImage_free(sum->sum.Shortsum);
	sum->sum.Shortsum = (DgpImage_Short)NULL;
    }
}

/*
 * Enlarge the size of the histogram.
 * Take the new bounds from sum->(min|max)[xy].
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure; on failure, nothing has changed.
 */
static ErrCode
PixelRealloc(tokCPixelSum *sum, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    DgpImage_Short oldDgp_Int16, newShort;
    DgpImage_Gray oldbyte, newbyte;

DGP_PROLOGUE("PixelRealloc()", DGP_PERRSTATE);

    DGP_ASSERT(sum != (tokCPixelSum *)NULL);

    if (sum->sumisbyte) {
	oldbyte = sum->sum.bytesum;
	DGP_ASSERT(oldbyte != (DgpImage_Gray)NULL);

DGP_ERR_PASS_CLEANUP(
	DgpImage_newOffset(dgp_state, DGPIMAGE_GRAY,
					 ((unsigned)(sum->maxx - sum->minx) +
					  1 + 2 * MAXXDIFF),
					 ((unsigned)(sum->maxy - sum->miny) +
					  1 + 2 * MAXYDIFF),
					 sum->minx - MAXXDIFF,
					 sum->miny - MAXYDIFF, (void**) &newbyte));
#ifdef	DBUG
	curhist += ((sum->maxx - sum->minx + 1 + 2 * MAXXDIFF) + sizeof(Dgp_Int8 *)) * (sum->maxy - sum->miny + 1 + 2 * MAXYDIFF) * sizeof(Dgp_Int8) + sizeof(*newbyte);
	curhist -= (DGPIMG_WIDTH(oldbyte) + sizeof(Dgp_Int8 *)) * DGPIMG_HEIGHT(oldbyte) * sizeof(Dgp_Int8) + sizeof(*oldbyte);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	for (y = DGPIMG_YBASE(oldbyte); y <= DGPIMG_YMAX(oldbyte); y++) {
	    for (x = DGPIMG_XBASE(oldbyte); x <= DGPIMG_XMAX(oldbyte); x++) {
		DGPIMG_REF(newbyte, x, y) = DGPIMG_REF(oldbyte, x, y);
            }
        }

	sum->sum.bytesum = newbyte;
	DgpImage_free(oldbyte);
    }
    else {
	oldDgp_Int16 = sum->sum.Shortsum;
	DGP_ASSERT(oldDgp_Int16 != (DgpImage_Short)NULL);

DGP_ERR_PASS_CLEANUP(
	DgpImage_newOffset(dgp_state, DGPIMAGE_SHORT,
					   ((unsigned)(sum->maxx - sum->minx) +
					    1 + 2 * MAXXDIFF),
					   ((unsigned)(sum->maxy - sum->miny) +
					    1 + 2 * MAXYDIFF),
					   sum->minx - MAXXDIFF,
					   sum->miny - MAXYDIFF, (void**) &newShort));
#ifdef	DBUG
	curhist += ((sum->maxx - sum->minx + 1 + 2 * MAXXDIFF) + sizeof(Dgp_Int16 *)) * (sum->maxy - sum->miny + 1 + 2 * MAXYDIFF) * sizeof(Dgp_Int16) + sizeof(*newShort);
	curhist -= (DGPIMG_WIDTH(oldDgp_Int16) + sizeof(Dgp_Int16 *)) * DGPIMG_HEIGHT(oldDgp_Int16) * sizeof(Dgp_Int16) + sizeof(*oldDgp_Int16);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	for (y = DGPIMG_YBASE(oldDgp_Int16); y <= DGPIMG_YMAX(oldDgp_Int16); y++) {
	    for (x = DGPIMG_XBASE(oldDgp_Int16); x <= DGPIMG_XMAX(oldDgp_Int16); x++) {
		DGPIMG_REF(newShort, x, y) = DGPIMG_REF(oldDgp_Int16, x, y);
            }
        }
	sum->sum.Shortsum = newShort;
	DgpImage_free(oldDgp_Int16);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    
    ERR_RETURN;
}

/*
 * Put the given model into the tokCClass by incrementing all the places
 * where it's turned on. tx and ty give an offset for the model in
 * the tokCClass's tokCPixelSum.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure.
 */
ErrCode
InsertModel(tokCClass c, tokCModelInfo *model, Dgp_Int32 tx, Dgp_Int32 ty, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    DgpImage_Short sumShort;
    DgpImage_Gray sumbyte;
    tokCPixelSum *modelsum = c->modelsum;

DGP_PROLOGUE("InsertModel()", DGP_PERRSTATE);

    if (c->status & TOKCLASS_FROZENHIST) {
	/* Nothing to be done. */
	NORMAL_RETURN;
    }

    DGP_ASSERT(modelsum != (tokCPixelSum *)NULL);
    DGP_ASSERT(model != (tokCModelInfo *)NULL);

    modelsum->maxx = (Dgp_Int16)MAX(modelsum->maxx, model->xsize - 1 + tx);
    modelsum->maxy = (Dgp_Int16)MAX(modelsum->maxy, model->ysize - 1 + ty);
    modelsum->minx = (Dgp_Int16)MIN(modelsum->minx, tx + opt->dilationRadius);
    modelsum->miny = (Dgp_Int16)MIN(modelsum->miny, ty);
    modelsum->nummodels++;
    DGP_ASSERT(modelsum->nummodels <= 32767);	/* Too many models kills it */

    /* The exemplar no Dgp_Int32er represents the histogram */
    if (c->status & TOKCLASS_EXSYNTHETIC) {
        c->status &= ~TOKCLASS_EXCLEAN;
    }

    if (modelsum->nummodels == opt->makeHistThresh) {
	/* We just hit the threshold where a real histogram gets made */
	DGP_ASSERT(modelsum->sumisbyte ?
	       (modelsum->sum.bytesum == (DgpImage_Gray)NULL) :
	       (modelsum->sum.Shortsum == (DgpImage_Short)NULL));
DGP_ERR_PASS_CLEANUP( 
        makeRealHist(c, opt));
    }
    else if (modelsum->nummodels < opt->makeHistThresh) {
	/* There shouldn't be a histogram now. Our job is done. */
	DGP_ASSERT(modelsum->sumisbyte ?
	       (modelsum->sum.bytesum == (DgpImage_Gray)NULL) :
	       (modelsum->sum.Shortsum == (DgpImage_Short)NULL));
	NORMAL_RETURN;
    }

    if ((modelsum->nummodels == DGPIMG_COLRNG) && (modelsum->sumisbyte)) {
	/* We just went over the limit. */
	sumbyte = modelsum->sum.bytesum;
DGP_ERR_PASS_CLEANUP(
	 DgpImage_convert(sumbyte, DGPIMAGE_SHORT, (void**) &sumShort));

	DgpImage_free(sumbyte);
	modelsum->sum.Shortsum = sumShort;
	modelsum->sumisbyte = DGP_FALSE;
    }

    if (modelsum->sumisbyte) {
	sumbyte = modelsum->sum.bytesum;
	DGP_ASSERT(sumbyte != (DgpImage_Gray)NULL);

	/*
	 * Enlarge the area if we need to.
	 */
	if ((model->xsize + tx > DGPIMG_XMAX(sumbyte) + 1) ||
	    (model->ysize + ty > DGPIMG_YMAX(sumbyte) + 1) ||
	    (tx + opt->dilationRadius < DGPIMG_XBASE(sumbyte)) ||
	    (ty < DGPIMG_YBASE(sumbyte))) {
DGP_ERR_PASS_CLEANUP(     
           PixelRealloc(modelsum, opt));

	    /* Reload - it changed */
	    sumbyte = modelsum->sum.bytesum;
        }

	for (y = 0; y < model->ysize; y++) {
	    for (x = 0; x < model->xsize; x++) {
		if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
		    (1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
		    DGP_ASSERT(x >= opt->dilationRadius);
		    DGPIMG_REF(sumbyte, x + tx, y + ty)++;
                }
            }
        }
    }
    else {
	sumShort = modelsum->sum.Shortsum;
	DGP_ASSERT(sumShort != (DgpImage_Short)NULL);

	if ((model->xsize + tx > DGPIMG_XMAX(sumShort) + 1) ||
	    (model->ysize + ty > DGPIMG_YMAX(sumShort) + 1) ||
	    (tx + opt->dilationRadius < DGPIMG_XBASE(sumShort)) ||
	    (ty < DGPIMG_YBASE(sumShort))) {
DGP_ERR_PASS_CLEANUP(
	    PixelRealloc(modelsum, opt));

	    sumShort = modelsum->sum.Shortsum;
        }

	for (y = 0; y < model->ysize; y++) {
	    for (x = 0; x < model->xsize; x++) {
		if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
		    (1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
		    DGP_ASSERT(x >= opt->dilationRadius);
		    DGPIMG_REF(sumShort, x + tx, y + ty)++;
                }
            }
        }
    }
    
    /* Is it at the histogram ceiling? */
    if (modelsum->nummodels >= opt->freezeHistThresh) {
	/* It's at or above it - must have been below it before */
	DGP_ASSERT(modelsum->nummodels == opt->freezeHistThresh);
	c->status |= TOKCLASS_FROZENHIST;
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * Remove the given model from the tokCPixelSum by decrementing all the places
 * that it contributed to. tx and ty give an offset for the model in
 * the tokCPixelSum.
 */
ErrCode
DeleteModel(tokCClass c, tokCModelInfo *model, Dgp_Int32 tx, Dgp_Int32 ty, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    tokCPixelSum *modelsum;
    DgpImage_Short sumShort;
    DgpImage_Gray sumbyte;

    DGP_PROLOGUE("DeleteModel()", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_ASSERT(c != (tokCClass)NULL);
    modelsum = c->modelsum;
    DGP_ASSERT(modelsum != (tokCPixelSum *)NULL);
    DGP_ASSERT(model != (tokCModelInfo *)NULL);

    /*
     * In some situations, when you merge a committed class into
     * an uncommitted class, you can end up with a frozen histogram.  The
     * uncommitted class then gets committed, and all the previous
     * uncommitted members (still having bitmaps) get checked against the
     * exemplar, and some may get rejected; in this case, DeleteModel
     * gets called on a frozen histogram.
     */
    if (c->status & TOKCLASS_FROZENHIST) {
	NORMAL_RETURN;
    }

    DGP_ASSERT(modelsum->nummodels < opt->freezeHistThresh);

    modelsum->nummodels--;

    /* Don't let them delete the only thing in a histogram */
    DGP_ASSERT(modelsum->nummodels > 0);

    /* Mark the exemplar dirty */
    if (c->status & TOKCLASS_EXSYNTHETIC) {
	c->status &= ~TOKCLASS_EXCLEAN;
    }

    if (modelsum->nummodels == opt->makeHistThresh - 1) {
	/* It just fell below the threshold for a histogram. */
	DGP_ASSERT(!(c->status & TOKCLASS_COMMITTED));
	destroyRealHist(modelsum, opt);
	/* That's all - no values to update */
	NORMAL_RETURN;
    }
    else if (modelsum->nummodels < opt->makeHistThresh - 1) {
	/* There shouldn't be a sum there at all */
	DGP_ASSERT(modelsum->sumisbyte ?
	       (modelsum->sum.bytesum == (DgpImage_Gray)NULL) :
	       (modelsum->sum.Shortsum == (DgpImage_Short)NULL));
	/* and there's nothing to do */
	NORMAL_RETURN;
    }

    if (modelsum->sumisbyte) {
	sumbyte = modelsum->sum.bytesum;
	DGP_ASSERT(sumbyte != (DgpImage_Gray)NULL);

	DGP_ASSERT(model->xsize + tx <= DGPIMG_XMAX(sumbyte) + 1);
	DGP_ASSERT(model->ysize + ty <= DGPIMG_YMAX(sumbyte) + 1);
	DGP_ASSERT(tx + opt->dilationRadius >= DGPIMG_XBASE(sumbyte));
	DGP_ASSERT(ty >= DGPIMG_YBASE(sumbyte));

	for (y = 0; y < model->ysize; y++) {
	    for (x = 0; x < model->xsize; x++) {
		if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
		    (1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
		    DGP_ASSERT(x >= opt->dilationRadius);
		    DGP_ASSERT(DGPIMG_REF(sumbyte, x + tx, y + ty) > 0);
		    DGPIMG_REF(sumbyte, x + tx, y + ty)--;
                }
            }
        }
    }
    else {
	sumShort = modelsum->sum.Shortsum;
	DGP_ASSERT(sumShort != (DgpImage_Short)NULL);

	DGP_ASSERT(model->xsize + tx <= DGPIMG_XMAX(sumShort) + 1);
	DGP_ASSERT(model->ysize + ty <= DGPIMG_YMAX(sumShort) + 1);
	DGP_ASSERT(tx + opt->dilationRadius >= DGPIMG_XBASE(sumShort));
	DGP_ASSERT(ty >= DGPIMG_YBASE(sumShort));

	for (y = 0; y < model->ysize; y++) {
	    for (x = 0; x < model->xsize; x++) {
		if (model->bits[y * model->row_width + (x / DGP_BITS_UINT32)] &
		    (1ul << ((DGP_BITS_UINT32 - 1) - (x & (DGP_BITS_UINT32 - 1))))) {
		    DGP_ASSERT(x >= opt->dilationRadius);
		    DGP_ASSERT(DGPIMG_REF(sumShort, x + tx, y + ty) > 0);
		    DGPIMG_REF(sumShort, x + tx, y + ty)--;
                }
            }
        }
    }

    /* The histogram has changed. */
    if (c->status & TOKCLASS_EXSYNTHETIC) {
	c->status &= ~TOKCLASS_EXCLEAN;
    }

    NORMAL_RETURN;
}

/*
 * Merge the two PixelSums from the two given classes. Merge "old" into "new"
 * (i.e., add all the pixel values from "old" into "new").
 * Offset "old" by the given translation.
 * "old" is not freed.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure; on failure, nothing has changed.
 * (Well, not quite - a byte-wide histogram might have been replaced by
 * a Dgp_Int16-wide histogram containing the same values).
 */
ErrCode
MergePixelSumsWithOffset(tokCClass newclass, tokCClass oldclass, Dgp_Int32 tx, 
                         Dgp_Int32 ty, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    DgpImage_Gray newsumbyte = (DgpImage_Gray)NULL, oldsumbyte;
    DgpImage_Short newsumShort = (DgpImage_Short)NULL, oldsumShort;
    tokCPixelSum *new = newclass->modelsum;
    tokCPixelSum *old = oldclass->modelsum;

DGP_PROLOGUE("MergePixelSumsWithOffset()", DGP_PERRSTATE);

    if (newclass->status & TOKCLASS_FROZENHIST) {
	/* Nothing to be done */
	NORMAL_RETURN;
    }

    new->maxx = (Dgp_Int16)MAX(new->maxx, old->maxx + tx);
    new->maxy = (Dgp_Int16)MAX(new->maxy, old->maxy + ty);
    new->minx = (Dgp_Int16)MIN(new->minx, old->minx + tx);
    new->miny = (Dgp_Int16)MIN(new->miny, old->miny + ty);
    new->nummodels += old->nummodels;
    DGP_ASSERT(new->nummodels <= 32767);    /* Too many models kills it */

    /* The exemplar no Dgp_Int32er represents the histogram */
    if (newclass->status & TOKCLASS_EXSYNTHETIC) {
	newclass->status &= ~TOKCLASS_EXCLEAN;
    }

    /*
     * Here's a problem: what if the old histogram was frozen?  Seems
     * harmless - this one will get pushed over the threshold and
     * frozen itself.  But what if it then gets used as the old histogram
     * in a future merge?  The number of models could expand without
     * bound.
     *
     * What we do in this case is quite simple: we throw away newclass'
     * current histogram and duplicate oldclass' histogram, applying
     * the offset as required.
     */
    if (oldclass->status & TOKCLASS_FROZENHIST) {
	/* Create the histogram copy */
	if (old->sumisbyte) {
DGP_ERR_PASS_CLEANUP(
	    DgpImage_dup(old->sum.bytesum, DGP_FALSE, (void**) &newsumbyte));

	    /* Adjust the origin */
	    DgpImage_setOffset(newsumbyte, DGPIMG_XBASE(old->sum.bytesum) + tx,
			DGPIMG_YBASE(old->sum.bytesum) + ty);
        }
	else {
DGP_ERR_PASS_CLEANUP( 
	    DgpImage_dup(old->sum.Shortsum, DGP_FALSE, (void**) &newsumShort));

	    DgpImage_setOffset(newsumShort, DGPIMG_XBASE(old->sum.Shortsum) + tx,
			DGPIMG_YBASE(old->sum.Shortsum) + ty);
        }

	/* nummodels must reflect histogram counts */
	new->nummodels = old->nummodels;
	new->maxx = (Dgp_Int16)(old->maxx + tx);
	new->maxy = (Dgp_Int16)(old->maxy + ty);
	new->minx = (Dgp_Int16)(old->minx + tx);
	new->miny = (Dgp_Int16)(old->miny + ty);

	if (new->sumisbyte ? (new->sum.bytesum != (DgpImage_Gray)NULL) :
	    (new->sum.Shortsum != (DgpImage_Short)NULL)) {
	    /* Free the old histogram */
	    if (new->sumisbyte) {
		DgpImage_free(new->sum.bytesum);
            }
	    else {
		DgpImage_free(new->sum.Shortsum);
            }
        }
        
	/* Put the copy in place */
	new->sumisbyte = old->sumisbyte;
	if (new->sumisbyte) {
	    new->sum.bytesum = newsumbyte;
        }
	else {
	    new->sum.Shortsum = newsumShort;
        }
        
	/* The copied histogram is frozen */
	newclass->status |= TOKCLASS_FROZENHIST;

	/* All done for this case */
	NORMAL_RETURN;
    }

    if (new->nummodels < opt->makeHistThresh) {
	/* Nothing to do. */
	NORMAL_RETURN;
    }

    /* If either is missing a histogram, create it */
    if (new->sumisbyte ? (new->sum.bytesum == (DgpImage_Gray)NULL) :
	(new->sum.Shortsum == (DgpImage_Short)NULL)) {
	DGP_ASSERT(new->nummodels - old->nummodels < opt->makeHistThresh);
DGP_ERR_PASS_CLEANUP(
	makeRealHist(newclass, opt));
    }

    if (old->sumisbyte ? (old->sum.bytesum == (DgpImage_Gray)NULL) :
	(old->sum.Shortsum == (DgpImage_Short)NULL)) {
	DGP_ASSERT(old->nummodels < opt->makeHistThresh);
DGP_ERR_PASS_CLEANUP(
	makeRealHist(oldclass, opt));
    }

    if ((new->nummodels > DGPIMG_COLRNG - 1) || (!new->sumisbyte)) {
	/* It needs to have a DgpImage_Short, or it's one already. */

	if (new->sumisbyte) {
	    /* It's not one now - make one */
	    newsumbyte = new->sum.bytesum;

	    /* This does the conversion. */
DGP_ERR_PASS_CLEANUP(
	    DgpImage_convert(newsumbyte, DGPIMAGE_SHORT, (void**) &newsumShort));
#ifdef	DBUG
	curhist -= (DGPIMG_WIDTH(newsumbyte) + sizeof(Dgp_Int8 *)) * DGPIMG_HEIGHT(newsumbyte) * sizeof(Dgp_Int8) + sizeof(*newsumbyte);
	curhist += (DGPIMG_WIDTH(newsumShort) + sizeof(Dgp_Int16 *)) * DGPIMG_HEIGHT(newsumShort) * sizeof(Dgp_Int16) + sizeof(*newsumShort);
	if (curhist > maxhist) {
	    maxhist = curhist;
	    DEBUG("histogram memory peak %d bytes, brk %d\n", maxhist, sbrk(0));
        }
#endif

	    /* Toast the old one */
	    DgpImage_free(newsumbyte);
	    new->sum.Shortsum = newsumShort;
	    new->sumisbyte = DGP_FALSE;
        }

	newsumShort = new->sum.Shortsum;

	if ((DGPIMG_XMAX(newsumShort) < new->maxx) ||
	    (DGPIMG_YMAX(newsumShort) < new->maxy) ||
	    (DGPIMG_XBASE(newsumShort) > new->minx) ||
	    (DGPIMG_YBASE(newsumShort) > new->miny)) {
DGP_ERR_PASS_CLEANUP(
            PixelRealloc(new, opt));

	    newsumShort = new->sum.Shortsum;
        }

	/* Two cases: old is DgpImage_Short, and old is DgpImage_Gray */
	if (old->sumisbyte) {
	    oldsumbyte = old->sum.bytesum;
	    for (y = old->miny; y <= old->maxy; y++) {
		for (x = old->minx; x <= old->maxx; x++) {
		    DGPIMG_REF(newsumShort, x + tx, y + ty) +=
			DGPIMG_REF(oldsumbyte, x, y);
                }
            }
        }
	else {
	    oldsumShort = old->sum.Shortsum;
	    for (y = old->miny; y <= old->maxy; y++) {
		for (x = old->minx; x <= old->maxx; x++) {
		    DGPIMG_REF(newsumShort, x + tx, y + ty) +=
			DGPIMG_REF(oldsumShort, x, y);
                }
            }
        }
    }
    else {
	/* It's a DgpImage_Gray, and that's fine. */

	newsumbyte = new->sum.bytesum;
	if ((DGPIMG_XMAX(newsumbyte) < new->maxx) ||
	    (DGPIMG_YMAX(newsumbyte) < new->maxy) ||
	    (DGPIMG_XBASE(newsumbyte) > new->minx) ||
	    (DGPIMG_YBASE(newsumbyte) > new->miny)) {
DGP_ERR_PASS_CLEANUP(
	    PixelRealloc(new, opt));

	    newsumbyte = new->sum.bytesum;
        }

	/* Two cases: old is DgpImage_Short, and old is DgpImage_Gray */
	if (old->sumisbyte) {
	    oldsumbyte = old->sum.bytesum;
	    for (y = old->miny; y <= old->maxy; y++) {
		for (x = old->minx; x <= old->maxx; x++) {
		    DGPIMG_REF(newsumbyte, x + tx, y + ty) +=
			DGPIMG_REF(oldsumbyte, x, y);
                }
            }
        }
	else {
	    oldsumShort = old->sum.Shortsum;
	    for (y = old->miny; y <= old->maxy; y++) {
		for (x = old->minx; x <= old->maxx; x++) {
		    DGPIMG_REF(newsumbyte, x + tx, y + ty) +=
			DGPIMG_REF(oldsumShort, x, y);
                }
            }
        }
    }

    /* Now get rid of the histograms we didn't really need */
    DGP_ASSERT(new->nummodels >= opt->makeHistThresh);	/* We need the new */
    if (old->nummodels < opt->makeHistThresh) {
	DGP_ASSERT(!(oldclass->status & TOKCLASS_COMMITTED));
	destroyRealHist(old, opt);
    }

    /* Is it at the histogram ceiling? */
    if (new->nummodels >= opt->freezeHistThresh) {
	/* It's at or above it */
	newclass->status |= TOKCLASS_FROZENHIST;
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    /* This is a bad bailout: it doesn't clean up the state properly. FIX. */
    ERR_RETURN;
}

#define PackedSet(ptr, x)                                                     \
  ((ptr)[(x) / DGP_BITS_UINT32] |= maskBit((x) % DGP_BITS_UINT32))

/*
 * Figure out how many points the thresholded histogram should have.
 * Do this by finding the median number of points in the class's members.
 *
 * Return the median number on success, -1 on failure.
 */
static ErrCode
CalcGoal(tokCClass class, Dgp_Int32 *ngoal_vals, Dgp_Int32 **goal_vals, 
         tokOptions *opt, Dgp_Int32 *goal)
{
    Dgp_State *dgp_state = opt->dgp_state;
    tokCPixelSum *modelsum; /* = class->modelsum; */
    Dgp_Int32 loMedian, hiMedian;
    Dgp_Int32 i;
    DgpAdt_ListNode member;
    Dgp_Int32 *lvals;

DGP_PROLOGUE("CalcGoal()", DGP_PERRSTATE);

    modelsum = class->modelsum;

    /* Keep around a playground to avoid constant malloc/free */
    if ((*ngoal_vals < (Dgp_Int32)DGPADT_LILEN(class->members)) || 
            (*goal_vals == (void *)NULL)) {
	if (*goal_vals != (void *)NULL) {
	    DGP_FREE((void *)*goal_vals);
        }

DGP_CHECK_ALLOC_CLEANUP(
	*goal_vals = DGP_MALLOC(DGPADT_LILEN(class->members) * sizeof(Dgp_Int32)));

	*ngoal_vals = DGPADT_LILEN(class->members);
    }

    lvals = (Dgp_Int32 *)*goal_vals;

    if (!(class->status & TOKCLASS_FROZENHIST)) {
	DGP_ASSERT(modelsum->nummodels == (Dgp_Int32) DGPADT_LILEN(class->members));
    }

    for (member = DGPADT_LIFIRST(class->members), i = 0;
	 member != DGP_NULLLISTNODE;
	 member = DGPADT_LINEXT(member), i++) {
	lvals[i] = tokCMemberBox(member)->npts;
    }
    loMedian = pickNth(lvals, i, i / 2);

    if (i & 1) {
        *goal = loMedian;
    }
    else {
	hiMedian = pickNth(lvals, i, i / 2 + 1);
	*goal = ((loMedian + hiMedian) / 2);
    }

    NORMAL_RETURN;

ERR_CLEANUP:

   ERR_RETURN;
}


/*
 * Figure out where to threshold the histogram array in order to get
 * "goal" pixels turned on in the result.
 *
 * Return the threshold on success, -1 on failure.
 */
static ErrCode
CalcMinSum(tokCPixelSum *modelsum, Dgp_Int32 goal, Dgp_Int32 *nPtsAway, 
           Dgp_Int32 *nsum_vals, Dgp_Int32 **sum_vals, tokOptions *opt, Dgp_Int32 *sum)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y;
    Dgp_Int32 minsum;
    Dgp_Int32 *ivals;
    DgpImage_Gray sumbyte;
    DgpImage_Short sumShort;
    Dgp_Int32 orig_goal = goal;

DGP_PROLOGUE("CalcMinSum()", DGP_PERRSTATE);

    /* Keep around a playground to avoid constant malloc/free */
    if ((*nsum_vals < modelsum->nummodels + 1)
            || (*sum_vals == (Dgp_Int32 *)NULL)) {
	if (*sum_vals != (void *)NULL) {
	    DGP_FREE((void *)*sum_vals);
        }
DGP_CHECK_ALLOC_CLEANUP(
	*sum_vals = DGP_MALLOC((modelsum->nummodels + 1) * sizeof(Dgp_Int32)));
	
	*nsum_vals = modelsum->nummodels + 1;
    }

    ivals = (Dgp_Int32 *)*sum_vals;

    /* Clear the histogram */
    memset(ivals, 0, (modelsum->nummodels + 1) * sizeof(Dgp_Int32));

    if (modelsum->sumisbyte) {
	sumbyte = modelsum->sum.bytesum;
	DGP_ASSERT(sumbyte != (DgpImage_Gray)NULL);
	for (y = modelsum->miny; y <= modelsum->maxy; y++) {
	    for (x = modelsum->minx; x <= modelsum->maxx; x++) {
		/*
		 * Put this assert here, because one wrong change when adding,
		 * deleting, or merging models in to the histogram,
		 * and KAPPLOOIE!
		 */
		DGP_ASSERT(DGPIMG_REF(sumbyte, x, y) <= modelsum->nummodels);
		ivals[DGPIMG_REF(sumbyte, x, y)]++;
		}
	    }
	}
    else {
	sumShort = modelsum->sum.Shortsum;
	DGP_ASSERT(sumShort != (DgpImage_Short)NULL);
	for (y = modelsum->miny; y <= modelsum->maxy; y++) {
	    for (x = modelsum->minx; x <= modelsum->maxx; x++) {
		/*
		 * Put this assert here, because one wrong change when adding,
		 * deleting, or merging models in to the histogram,
		 * and KAPPLOOIE!
		 */
		DGP_ASSERT(DGPIMG_REF(sumShort, x, y) <= modelsum->nummodels);
		DGP_ASSERT(DGPIMG_REF(sumShort, x, y) >= 0);
		ivals[DGPIMG_REF(sumShort, x, y)]++;
		}
	    }
	}

    for (minsum = modelsum->nummodels + 1;
	 (minsum > 0) && (goal >= 0);
	 goal -= ivals[--minsum]) {
	}

    /* Is over-thresholding or under-thresholding better? */
    if ((minsum > 0) && (-goal < goal + ivals[minsum])) {
	/* We're better off with our current value. */
	DGP_ASSERT(goal != orig_goal);/* We turned at least one pixel on */
	goal = -goal;
	}
    else {
	/* Back up one level */
	goal += ivals[minsum++];

	if (goal == orig_goal) {
	    /*
	     * This happens sometimes: we were told to threshold something to
	     * produce an exemplar with 1 pixel set, but our only choices are
	     * 0 or 2 pixels set (ivals[minsum - 1] == 2 right now).  In this
	     * case, we're going to under-threshold.
	     */
	    goal -= ivals[--minsum];
	    goal = -goal;
	    }
	}

    *nPtsAway = goal;
    *sum = minsum;
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * Generate a binary template from the histogram. Stick the result in
 * class->exemplar.
 *
 * This seems simple but can get hairy: the minimum (x,y) val in the exemplar
 * may not be at (dilationRadius,0) (remember the implicit offset of
 * dilationRadius at the left edge). This means that we need to adjust the
 * offsets of all the members of the class: those offsets are expressed wrt
 * the exemplar bitmap origin. Since the exemplar bitmap is a simple thing,
 * it *must* be built with its minimum x pixel at x==dilationRadius and
 * minimum y pixel at y==0. This means that if these aren't (dilationRadius,0)
 * in histogram coordinates (currently all the members'
 * offsets are expressed wrt the histogram), then we need to adjust the
 * offsets because they need to match the bitmap coordinate system. We also
 * need to adjust the histogram coordinate system so it matches the
 * bitmap coordinate system. Fortunately, this isn't too hard (the joys
 * of the image ADT).
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure; on failure, nothing has really
 * changed. Note that a NULL value in class->exemplar does not reflect failure:
 * it means that this routine decided that it couldn't make a good enough
 * exemplar, so it punted. If forceMake is DGP_TRUE, it isn't allowed to do this -
 * it must generate an exemplar (or die trying).
 */
ErrCode
MakeExemplarFromPixelSum(TokState state, tokCClass class, Dgp_Bool forceMake)
{
    _tokState *intstate = (_tokState *)state;
    tokOptions *opt = &(intstate->options);
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 x, y, goal;
    tokCBoxType *box = (tokCBoxType *)NULL;
    tokCModelInfo *model = (tokCModelInfo *)NULL;
    DgpAdt_ListNode member;
    Dgp_UInt32 *mptr;
    Dgp_Int32 nPtsAway;
    Dgp_Int32 minsum;
    tokCPixelSum *modelsum = class->modelsum;
    Dgp_Int32 npts;
    DgpImage_Gray sumbyte;
    DgpImage_Short sumShort;
    Dgp_Int32 miny = 0, maxy = 0, minx = 0, maxx = 0;

DGP_PROLOGUE("MakeExemplarFromPixelSum()", DGP_PERRSTATE);

    if (((modelsum == (tokCPixelSum *)NULL) || (modelsum->nummodels < 4)) &&
	(!forceMake)) {
	/* No histogram or too few things - punt. */
	class->exemplar = (tokCBoxType *)NULL;
	NORMAL_RETURN;
    }
    DGP_ASSERT(modelsum != (tokCPixelSum *)NULL);

    DGP_ASSERT(!(class->status & TOKCLASS_IMMUTABLE));

    /*
     * Note that modelsum->nummodels might not match DGPADT_LILEN(class->members):
     * frozen histograms won't match, nor will histograms beDgp_Int32ing to
     * immutable classes (some members might have been removed by a previous
     * tokPartial).  However, in all other cases, they should match.
     */
    if (!(class->status & TOKCLASS_FROZENHIST)) {
	DGP_ASSERT(modelsum->nummodels == (Dgp_Int32) DGPADT_LILEN(class->members));
    }

    /* Create the histogram for real */
    if (modelsum->nummodels < opt->makeHistThresh) {
	/* It shouldn't have one */
	DGP_ASSERT(modelsum->sumisbyte ?
	       (modelsum->sum.bytesum == (DgpImage_Gray)NULL) :
	       (modelsum->sum.Shortsum == (DgpImage_Short)NULL));
	/* so we make one */
DGP_ERR_PASS_CLEANUP(
	makeRealHist(class, opt));
    }

    /* How to threshold the histogram? */
DGP_ERR_PASS_CLEANUP(
    CalcGoal(class, &(intstate->ngoal_vals), &(intstate->goal_vals), 
             opt, &goal));
DGP_ERR_PASS_CLEANUP(
    CalcMinSum(modelsum, goal, &nPtsAway, &(intstate->nsum_vals), 
               &(intstate->sum_vals), opt, &minsum));

    if ((((Dgp_Float32)nPtsAway / goal) >= 0.10) && (!forceMake)) {
	/* Too far off - punt. */
	class->exemplar = (tokCBoxType *)NULL;
	if (modelsum->nummodels < opt->makeHistThresh) {
	    /* Destroy the temporary histogram. */
	    DGP_ASSERT(!(class->status & TOKCLASS_COMMITTED));
	    destroyRealHist(modelsum, opt);
        }
	NORMAL_RETURN;
    }

    TOKBOX_SEC_MALLOC(&box);

    box->model = (tokCModelInfo *)NULL;
    box->image = (tokCImageInfo *)NULL;

    /*
     * OK - we have the bounds of the non-zero parts of the histogram.
     * Scan the histogram to find the bounds of the region where it
     * exceeds the threshold.
     */

    if (modelsum->sumisbyte) {
	sumbyte = modelsum->sum.bytesum;

	/* Minimum y mark - scan rows down from the top */
	for (y = modelsum->miny; y <= modelsum->maxy; y++) {
	    for (x = modelsum->minx; x <= modelsum->maxx; x++) {
		if (DGPIMG_REF(sumbyte, x, y) >= minsum) {
		    miny = y;
		    minx = x;
		    break;
                }
            }
	    if (x != modelsum->maxx + 1) {
		break;
            }
        }
	DGP_ASSERT(y <= modelsum->maxy);

	/*
	 * Minimum x mark - scan columns from the left.
	 * Can start scanning at row miny without danger of missing anything.
	 * The current value of minx also bounds how far right we may have
	 * to scan.
	 */
	for (x = modelsum->minx; x <= minx; x++) {
	    for (y = miny; y <= modelsum->maxy; y++) {
		if (DGPIMG_REF(sumbyte, x, y) >= minsum) {
		    minx = x;
		    break;
                }
            }
	    if (y != modelsum->maxy + 1) {
		break;
            }
        }
	DGP_ASSERT(x <= minx);

	/* Maximum y mark - scan rows RtoL from the bottom. */
	for (y = modelsum->maxy; y >= miny; y--) {
	    for (x = modelsum->maxx; x >= minx; x--) {
		if (DGPIMG_REF(sumbyte, x, y) >= minsum) {
		    maxy = y;
		    maxx = x;
		    break;
                }
            }
	    if (x != minx - 1) {
		break;
            }
        }
	DGP_ASSERT(y >= miny);

	/* and maximum x mark - scan columns from the right */
	for (x = modelsum->maxx; x >= maxx; x--) {
	    for (y = miny; y <= maxy; y++) {
		if (DGPIMG_REF(sumbyte, x, y) >= minsum) {
		    maxx = x;
		    break;
                }
            }
	    if (y != maxy + 1) {
		break;
            }
        }
	DGP_ASSERT(x >= maxx);
	}
    else {
	sumShort = modelsum->sum.Shortsum;

	/* Minimum y mark - scan rows down from the top */
	for (y = modelsum->miny; y <= modelsum->maxy; y++) {
	    for (x = modelsum->minx; x <= modelsum->maxx; x++) {
		if (DGPIMG_REF(sumShort, x, y) >= minsum) {
		    miny = y;
		    minx = x;
		    break;
                }
            }
	    if (x != modelsum->maxx + 1) {
		break;
            }
        }
	DGP_ASSERT(y < modelsum->maxy + 1);

	/*
	 * Minimum x mark - scan columns from the left.
	 * Can start scanning at row miny without danger of missing anything.
	 * The current value of minx also bounds how far right we may have
	 * to scan.
	 */
	for (x = modelsum->minx; x <= minx; x++) {
	    for (y = miny; y <= modelsum->maxy; y++) {
		if (DGPIMG_REF(sumShort, x, y) >= minsum) {
		    minx = x;
		    break;
                }
            }
	    if (y != modelsum->maxy + 1) {
		break;
            }
        }

	/* Maximum y mark - scan rows RtoL from the bottom. */
	for (y = modelsum->maxy; y >= miny; y--) {
	    for (x = modelsum->maxx; x >= minx; x--) {
		if (DGPIMG_REF(sumShort, x, y) >= minsum) {
		    maxy = y;
		    maxx = x;
		    break;
                }
            }
	    if (x != minx - 1) {
		break;
            }
        }
	DGP_ASSERT(y > miny - 1);

	/* and maximum x mark - scan columns from the right */
	for (x = modelsum->maxx; x >= maxx; x--) {
	    for (y = miny; y <= maxy; y++) {
		if (DGPIMG_REF(sumShort, x, y) >= minsum) {
		    maxx = x;
		    break;
                }
            }
	    if (y != maxy + 1) {
		break;
            }
        }
    }

    /* OK - got the bounds. */

  TOKMDL_SEC_MALLOC(&model);

    model->bits = (Dgp_UInt32 *)NULL;

    model->xsize = maxx - minx + 1 + opt->dilationRadius;
    model->ysize = maxy - miny + 1;
    model->row_width = (model->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    npts = box->npts = model->npts = goal - nPtsAway;

DGP_CHECK_ALLOC_CLEANUP(
    model->bits =
	(Dgp_UInt32 *)(DGP_CALLOC)((unsigned)model->row_width * model->ysize,
				     sizeof(Dgp_UInt32)));

#ifdef	DBUG
    curbitmap += model->row_width * model->ysize * sizeof(Dgp_UInt32);
    if (curbitmap > maxbitmap) {
	maxbitmap = curbitmap;
	DEBUG("model bitmap memory peak %d bytes, brk %d\n", maxbitmap, sbrk(0));
    }
#endif

    if (modelsum->sumisbyte) {
	sumbyte = modelsum->sum.bytesum;

	for (y = miny; y <= maxy; y++) {
	    mptr  = model->bits + (y - miny) * model->row_width;
	    for (x = minx; x <= maxx; x++) {
		if (DGPIMG_REF(sumbyte, x, y) >= minsum) {
		    PackedSet(mptr, x - minx + opt->dilationRadius);
                }
            }
        }
    }
    else {
	sumShort = modelsum->sum.Shortsum;

	for (y = miny; y <= maxy; y++) {
	    mptr  = model->bits + (y - miny) * model->row_width;
	    for (x = minx; x <= maxx; x++) {
		if (DGPIMG_REF(sumShort, x, y) >= minsum) {
		    PackedSet(mptr, x - minx + opt->dilationRadius);
                }
            }
        }
    }

    box->width = model->xsize - opt->dilationRadius;
    box->height = model->ysize;

DGP_ERR_PASS_CLEANUP(
    MakeImCrossFromMod(model, opt, &(box->image)));
    box->xOffset = box->yOffset = 0;

    /*
     * Set this later than the last failure point so we don't try to free it
     * twice.
     */
    box->model = model;

    /*
     * Start patching coordinates. minx should have been dilationRadius
     * (all the things that went into the histogram had a dilationRadius
     * offset on the left column, so if things were simple that left column
     * would be empty). If it wasn't then we need to adjust the coordinate
     * system of the histogram.
     *
     * Want the pixel that used to be at (minx, miny) to be at
     * (dilationRadius,0) instead.
     */
    if ((minx != opt->dilationRadius) || (miny != 0)) {
	/* Gotta do the offsetting */
	if (modelsum->sumisbyte) {
	    DgpImage_setOffset(modelsum->sum.bytesum,
			DGPIMG_XBASE(modelsum->sum.bytesum) - minx +
			opt->dilationRadius,
			DGPIMG_YBASE(modelsum->sum.bytesum) - miny);
        }
	else {
	    DgpImage_setOffset(modelsum->sum.Shortsum,
			DGPIMG_XBASE(modelsum->sum.Shortsum) - minx +
			opt->dilationRadius,
			DGPIMG_YBASE(modelsum->sum.Shortsum) - miny);
        }
        
	modelsum->minx -= (Dgp_Int16)(minx - opt->dilationRadius);
	modelsum->maxx -= (Dgp_Int16)(minx - opt->dilationRadius);
	modelsum->miny -= (Dgp_Int16)miny;
	modelsum->maxy -= (Dgp_Int16)miny;

	/*
	 * and offset all the members too.
	 */
	DGPADT_FOREACH (member, class->members) {
	    tokCMemberBox(member)->xOffset -= (Dgp_Int16)(minx - opt->dilationRadius);
	    tokCMemberBox(member)->yOffset -= (Dgp_Int16)miny;
        }

    }

    if (modelsum->nummodels < opt->makeHistThresh) {
	/* Destroy the histogram we created above. */
	DGP_ASSERT(!(class->status & TOKCLASS_COMMITTED));
	destroyRealHist(modelsum, opt);
    }

    class->exemplar = box;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (box != (tokCBoxType *)NULL) {
	FreeBox(box, opt);
    }

    if (model != (tokCModelInfo *)NULL) {
	FreeModel(model, opt);
    }
    
    ERR_RETURN;
}
