/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/tok.c,v 1.40 2003/02/10 22:53:42 mkrocken Exp $
 *
 */

/*
 * This file contains top-level routines to interface to the tokeniser.
 *
 * Exports:
 *
 *	type TokState
 *	type tokOptions
 *	type tokBitmap
 *	type tokClass
 *	type tokClassTable
 *
 *	const NULLTOKSTATE
 *
 *	ErrCode tokInit(tokOptions *options, TokState *state) - initialise
 *		the tokeniser, and get back a state cookie.
 *
 *	void tokSetDefaults(tokOptions *options) - fill in all the options
 *		with their default values.
 *
 *	ErrCode tokAddBitmap(TokState state, tokBitmap *bitmap) - throw a
 *		bitmap at the tokeniser.
 *
 *	ErrCode tokMerge(TokState state) - tell the tokeniser to make a
 *		compaction pass.
 *
 *	ErrCode tokFinal(TokState state, DgpAdt_List *output) - complete the process
 *		and get the results back.
 *
 *	ErrCode tokPartial(TokState state, DgpAdt_List *output) - get a partial set
 *		of results.
 *
 *	ErrCode tokResume(TokState state, DgpAdt_List output) - resume after
 *		tokPartial, with a modified set of classes.
 *
 *	ErrCode tokFreeOutput(tokState state, DgpAdt_List output, Dgp_Bool hasOffsets)
 *      - free the result of tokFinal.  hasOffsets should be set
 *		according to the TOK_OPT_RETURNOFFSET flag in the initial
 *		tokOptions.
 *
 *	ErrCode tokFreeClass(tokState state, tokClass *class) - free one
 *		class from a list returned by tokPartial.
 *
 *	void tokAbort(TokState) - abort the process and get rid of the state.
 *
 *	void tokFreeState(TokState) - get rid of the state.
 *
 * The tokeniser takes a bunch of bitmaps and groups them into classes.
 * All the bitmaps within a class look like each other, and are represented
 * by an exemplar. It also produces a histogram formed by summing up all
 * the bitmaps in the class; this is a very good representation of the
 * overall shape represented by the class, and can be interpolated to
 * get a higher-representation of that shape.
 *
 * The tokeniser maintains a bunch of internal state; this is represented by
 * a "TokState" cookie. It is generally possible to have several different
 * tokeniser states floating around without them interfering with each
 * other (but see "Caveats" below).
 *
 * The way you use these routines:
 * 1. Create a tokOptions structure, fill it in, and call tokInit with it.
 *    This gives you back a TokState cookie, or an error if it can't
 *    create one.  tokSetDefaults will fill in the tokOptions structure
 *    with the default values (described below).  The only field in
 *    the structure that doesn't have a default value is dgp_state - this
 *    needs to be supplied before calling tokInit.
 * 2. Call tokAddBitmap as many times as you like with assorted bitmaps.
 *    Give each call the TokState value you got from tokInit.  If it fails,
 *    the TokState will be in some stable state. The bitmap may or may
 *    not have been added... but there's no way to tell which has happened.
 * 3. Call tokMerge whenever you feel like it. This is a time-consuming
 *    operation, but reduces the memory requirements.
 * 4. When you've run out of bitmaps, call tokFinal with the TokState.
 *    This will produce a DgpAdt_List of classes. Each entry on the list contains
 *    a (tokClass *) value. On failure, tokFinal returns an error, and has
 *    not deallocated any of the internal state (though the state may have
 *    changed in other ways, it's still valid, but don't try to do any
 *    tokAddBitmaps with it - just tokAbort or tokFinal again).
 *    tokFinal includes a tokMerge.
 * 5. When you're done with this stuff, call tokFreeOutput on the DgpAdt_List.
 *    You must free it with tokFreeOutput; freeing it yourself is bad.
 * 6. Finally, call tokFreeState to get rid of the state.
 * 7. If at any point you need to abort, call tokAbort on the TokState.
 *    This frees all the memory hung off the state.
 *
 * That's the simple way to do it... it can get more complicated.
 * The more complicated approaches involve using tokPartial, and maybe
 * tokResume, in order to get partial results and continue with the
 * tokenisation.
 *
 * As above, you create a TokState, and call tokAddBitmap a bunch of times.
 * You can call tokMerge too if you like.  When you're ready for some
 * results, but you want to continue adding bitmaps at a later time, you
 * call tokPartial.  This creates an output list, just like tokFinal.
 * You can process the list, deleting
 * some elements, and (slightly) modifying others, then pass the resulting
 * list to tokResume.  Now you can keep on calling tokAddBitmap; eventually,
 * you call tokFinal (or tokPartial again, and so on).
 *
 * What tokPartial does is
 * - it does a tokMerge, just like tokFinal.
 * - it commits every class, making it immutable; the class exemplars
 *   will never change, and that class will never be merged with another
 *   immutable class; thus, the caller can depend on every class in the
 *   output being around forever.
 * - it returns the same information that tokFinal would have returned.
 *
 * tokResume takes the output list that tokPartial gave you, and the state,
 * and continues on from there, using the classes that are present on the
 * list, but with each class being empty (having no members).  You are
 * required to call tokResume once after each tokPartial, to
 * give ownership of the list (and all it contains) back to the tokeniser.
 * Once you have called tokResume, that list becomes invalid.
 * You are allowed to make *some* changes to the list during the time
 * that you own it; these will affect how the tokeniser proceeds.  The
 * things you are allowed to do are:
 * - delete elements of the list.  To do this, call tokFreeClass on the
 *   tokClass you want to get rid of, then unhook the DgpAdt_ListNode that used to
 *   contain it from the list with liRem (or DgpAdt_liRemAndNext).  This is how
 *   you can get rid of (say) singletons that you don't want the tokeniser
 *   to retain for future use.
 * - alter userdata fields.  The userdata field in each tokClass is initially
 *   NULL.  You can change it to some other value if you want.  When you
 *   call tokResume, all the classes remaining on the list are immutable.
 *   Thus, any future call to tokPartial or tokFinal will (guaranteed) contain
 *   copies of these classes.  These copies will have the same exemplar as they
 *   had before, but possibly different histograms.  They will also have the
 *   same userdata value.  The actual addresses of the tokClass structures
 *   will NOT be the same from one call to another.  However, you can
 *   store any identifying pointer you like in the userdata field, and be
 *   sure you will get it back.  The tokeniser will not do anything with this
 *   pointer (other than copy it); it will not dereference it.  You are
 *   responsible for whatever memory the userdata field points at.
 * - You are not allowed to modify the members lists on any of the classes;
 *   after tokResume the classes are going to be empty, and tokResume will
 *   take care of cleaning these lists up.  For classes you remove from the
 *   list you passed back to tokResume, tokFreeClass will take care of the
 *   members list.
 * - If, after doing a tokPartial, you decide you're finished after all,
 *   call tokFreeOutput on the output list AND call tokFreeState on the
 *   TokState.  This converts the tokPartial into an (effective)
 *   tokFinal.
 *
 * Details of the structures and what they mean:
 *
 * tokOptions:
 *	This structure controls a number of parameters of the
 *	tokeniser. Some of these control space/time tradeoffs, some of them
 *	control accuracy/time or accuracy/space tradeoffs; others are simply
 *	tuning parameters. For each parameter, there is a standard value
 *	that gives good accuracy, possibly at the cost of time or memory.
 *
 *	fracHausd:	This is a value between 0 and 1. For two bitmaps to
 *			match, the partial Hausdorff distance between them,
 *			using this fraction, must be no more than 1. Values
 *			closer to 1 are more stringent (fewer things match).
 *
 *			The default value of fracHausd (the value set with
 *			tokSetDefaults) is 0.97.
 *
 *	maxDense:	If two shapes match according to fracHausd, then
 *			the tokeniser checks one shape against a dilated
 *			(by one pixel) version of the other. It then scans
 *			the parts of the first shape lying outside the dilated
 *			second shape. In any given 3x3 pixel area, there must
 *			be no more than maxDense pixels turned on. Thus,
 *			the first shape cannot contain a large contiguous
 *			block of "on" pixels not adjacent to any "on" pixel
 *			of the second shape.
 *
 *			The default value of maxDense is 3.
 *
 *	widthDiff:	Only shapes whose widths and heights are similar
 *	heightDiff:	are compared. These two parameters say how close
 *			the widths and heights must be for them to be compared.
 *
 *			The default values of widthDiff and heightDiff are
 *			1 and 1.
 *
 *	commitThresh:	Once a class has a certain number of members, it is
 *			considered "committed" and all its current members are
 *			fixed in that class. Before a class is committed,
 *			it is possible for some of its members to be rejected
 *			at some stage in the processing. The value of
 *			commitThresh is the number of members at which a class
 *			will be committed. If commitThresh is zero, then
 *			classes will never be committed (until tokFinal is
 *			called). Tuning commitThresh is an accuracy/memory
 *			tradeoff: a committed class uses less memory than
 *			an uncommitted class, but it might contain some
 *			incorrectly-classified members.
 *
 *			The default value of commitThresh is 10000.
 *
 *	freezeHistThresh:
 *			Above a certain number of members, a committed class
 *			will have its histogram frozen, so no further updates
 *			to the histogram will happen.  The idea is that this
 *			number specifies a threshold above which there should
 *			be no appreciable change to the class's histogram.
 *			Zero means "never" - no histogram will ever
 *			be frozen (this should be used with caution, as any
 *			class containing over 32K-1 members could have
 *			problems). This value should either
 *			be zero or at least commitThresh.  Note that this
 *			makes the TOK_OPT_COMMITKEEPHIST flag obsolete.
 *
 *			The default value of freezeHistThresh is 10000.
 *
 *	makeHistThresh:	There are a lot of small classes (with only a few
 *			members), and so it might make sense to create
 *			histograms only for larger classes; the smaller
 *			classes can have histograms created then deleted on
 *			the fly.  This sets the threshold: classes below
 *			this value don't have histograms created.  This is
 *			purely a memory/speed tradeoff and doesn't affect
 *			the output at all (i.e., if you set
 *			TOK_OPT_RETURNHIST, you get back histograms for all
 *			the classes regardless of the value of makeHistThresh).
 *
 *			The default value of makeHistThresh is 10.
 *
 *	returnHistThresh:	When you set TOK_OPT_RETURNHIST, don't
 *			return any histograms for classes with fewer than
 *			this many members.  Even when this is 1 or 0,
 *			singletons may or may not have histograms.
 *
 *			The default value of returnHistThresh is zero.
 *
 *	nneigh:		When the tokeniser decides to place a bitmap in a
 *			certain class, it tries a number of small translations
 *			of the bitmap relative to the class's exemplar.
 *			nneigh determines how many translations are considered.
 *			Valid values are 0, 4, 8, 24.
 *			Setting nneigh to smaller values decreases the time
 *			required. It degrades the positioning of the bitmaps;
 *			for applications that care about this, this can be
 *			significant. Even for applications that do not care
 *			about the absolute positioning of each bitmap,
 *			setting nneigh to a value greater than zero can be of
 *			use, as this improves the quality of the final exemplar
 *			and histogram of the class: the members of the class
 *			are better registered before they are histogrammed.
 *
 *			The default value of nneigh is 8.
 *
 *	xshift:		Small bitmaps can cause problems. The Hausdorff
 *	xslopechange:	fraction actually used is therefore not just
 *	newslope:	based on fracHausd. fracHausd determines the number
 *			of "mismatches" (pixels of one shape not lying under
 *			or adjacent to the other shape) allowed for two shapes
 *			to match. If this value is below xshift, it is
 *			set to zero. If it is between xshift and
 *			xslopechange, it is set to (value - xshift).
 *			If it is above xslopechange, it is set to
 *			(xslopechange - xshift) +
 *				(value - xslopechange) * newslope
 *			(but cannot go higher than the original number of
 *			allowable mismatches).
 *			The effect of this is to replace the linear
 *			mismatch function with a non-linear one:
 *
 *				|            .
 *				|           .
 *				|          .
 *				|         .+
 *				|        .
 *				|       . +
 *				|      .
 *				|     .  +
 *				|    .
 *				|   .   +
 *				|  .   +
 *				| .   +
 *				|.   +
 *				----*---*---------------
 *
 *			The '.' curve is the original mismatch-allowed
 *			function (a straight line, slope 1). The '+' curve
 *			is the modified function: zero for a while, up to
 *			xshift (the first '*'), then slope 1 for a while,
 *			up to xslopechange (the second '*'), then slope
 *			newslope from there until it catches up with the
 *			original curve.
 *
 *			The default value of xshift is 3.
 *			The default value of xslopechange is 6.
 *			The default value of newslope is 2.0.
 *
 *	dilationRadius:	This controls the dilation radius used for the
 *			shape comparison.
 *
 *			The default value of dilationRadius is 1.
 *
 *	mockRadius:	Ignore this field - it's only used internally.
 *
 *	flags:		Various options to control details of the tokeniser.
 *			The values that can be ORed into this are:
 *	    TOK_OPT_COPYBITMAPS:
 *				If this bit is set, then the
 *				tokeniser does not claim ownership
 *				of the bitmap memory passed in to
 *				tokAddBitmap(); instead, it makes copies
 *				of those bitmaps.
 *	    TOK_OPT_RETURNHIST:	If this bit is set, then the tokeniser
 *				returns histograms in the output. If it
 *				is clear, then it frees the histograms
 *				when doing a tokFinal(), and returns NULL
 *				in those fields. Note that if the class
 *				has only a single member, then there may not
 *				be a histogram generated for that class (but
 *				some singleton classes might have histograms).
 *				Any class with >1 member will always have
 *				a histogram; the value of makeHistThresh
 *				does not affect which classes have histograms
 *				returned.
 *	    TOK_OPT_RETURNOFFSET:
 *				If this bit is set, then the structure
 *				returned by tokFinal() is somewhat different:
 *				instead of each member of a class being
 *				represented by the (void *) value passed in
 *				with its original bitmap, each member is
 *				represented by a tokOffset structure,
 *				containing the offsets of the UL corner of
 *				that member from the UL corner of
 *				the exemplar, and the (void *) value.
 *	    TOK_OPT_ALIGN:	Just before it commits a class, the tokeniser
 *				can find the best positioning of each member
 *				of the class with respect to the class's
 *				exemplar. This is a moderately expensive
 *				operation, but gives significantly better
 *				positioning.
 *	    TOK_OPT_NONTPDIL	Normally, the tokeniser uses a topology-
 *				preserving dilation. Setting this flag
 *				makes it use a regular dilation.
 *				This also makes the dilationRadius
 *				value have special meanings:
 *				0-9 mean the same as usual, but
 *				10 means dilate by radius 1 (same as 1)
 *				14 means dilate by radius sqrt(2)
 *				20 means dilate by radius 2 (same as 2)
 *
 *			The default value of flags is zero.
 *
 *	dgp_state:	You need to supply a dgp_state for this code
 *			to work; it's used for the usual error handling
 *			and memory allocation routines.
 *
 *	mergeCallback:	If you want to, you can supply your own merge
 *			routine, which will be called by tokMerge (and
 *			tokFinal, tokPartial).  This is seriously
 *			advanced stuff; you're advised to avoid it unless
 *			you really know what you're doing.  Default value
 *			for this field is NULL, meaning that the
 *			normal merge routine is used.
 *
 *	callbackCookie:	If you supply a custom mergeCallback, this is a
 *			cookie for it.
 *
 * tokBitmap:
 *	This structure holds the bitmaps that the tokeniser works on.
 *
 *	width:		These fields give the size of the bitmap.
 *	height:		They should completely, and exactly, contain
 *			all the "set" (1) bits in the bitmap itself.
 *			In other words, the y-coordinate of the topmost
 *			set bit should be 0, the bottommost set bit
 *			height-1; the x-coordinate of the leftmost set
 *			bit should be 0, the x-coordinate of the
 *			rightmost set bit should be width-1.
 *
 *	bits:		This points to the bitmap itself. Each row of
 *			the bitmap should consist of an integral number
 *			of 32-bit words. The words should be in
 *			standard reading order: top to bottom, left to right.
 *			Within each word, the most significant bit is the
 *			leftmost (in the bitmap image).
 *			Note that on little-endian machines, this is *NOT*
 *			the same as filling each byte MSB-to-LSB. On big-endian
 *			machines, it is.
 *
 *	longsperline:	This says how many words there are in each row of the
 *			bitmap. This can be larger than the minimum number
 *			required according to "width".
 *
 *	userdata:	The tokeniser does not use the value of this field at
 *			all. It keeps track of it, and eventually hands it back
 *			as the representation of this bitmap in the final
 *			result.
 *
 * tokClass:
 *	This represents one of the final equivalence classes produced by
 *	the tokeniser.
 *
 *	exemplar:	This is the exemplar of the class. It is a tokBitmap
 *			like any other. The value of its userdata field is
 *			undefined.
 *
 *	histIsByte:	This specifies the size of the elements of the
 *			class's histogram: 8 or 16 bits per element.
 *
 *	histcount:	This specifies how many bitmaps are represented in
 *			the histogram.  This will normally be the number
 *			of members, unless this class hit the freezeHistThresh
 *			limit.
 *
 *	histogram:	This is the histogram of all the bitmaps that are
 *			members of this class. It is either a DgpImage_Gray
 *			(pixels are Dgp_UInt8) or a DgpImage_Short (pixels
 *			are Dgp_Int16). The origin of the histogram corresponds
 *			to the top left corner of the exemplar bitmap. Note
 *			that the histogram can contain non-zero counts
 *			outside the bounds of the exemplar bitmap. To
 *			make sure no non-zero entries are missed, the caller
 *			must scan every pixel of the histogram; note that some
 *			can have coordinates that are smaller than zero. See
 *			the image ADT documentation.
 *
 *	members:	This is a list of the members of the class. Each
 *			node of the list contains a "userdata" value from
 *			one of the bitmaps that were placed in this class,
 *			or a tokOffset * value if TOK_OPT_RETURNOFFSET was set.
 *			The number of elements in the class (and thus the
 *			maximum histogram count) is DGPADT_LILEN(members).
 *
 *	userdata:	This is another field that the tokeniser does not
 *			use, but keeps track of and eventually hands back.
 *			If the userdata field in a TokClass on an output
 *			list created by tokPartial is modified before being
 *			passed back to tokResume, then the tokeniser will
 *			place that same userdata value in the corresponding
 *			tokClass in any subsequent output list.
 *
 * Memory ownership:
 *	- tokInit does not own the options structure passed to it. The caller
 *	  is responsible for that memory. tokInit allocates the memory
 *	  used by the TokState information (using the allocator in
 *	  options->dgpState, of course).
 *
 *	- tokAddBitmap does not own the tokBitmap structure passed to it.
 *	  However, it claims ownership of the actual bitmap memory (the
 *	  "bits" pointer) in the tokBitmap, unless the TOK_OPT_COPYBITMAPS
 *	  flag was set in the initial call to tokInit(). If this bit was not
 *	  set, then the bitmap memory should not be freed by
 *	  the caller. Since the tokeniser will eventually be freeing it,
 *	  it must have been allocated using the same allocator that the
 *	  tokeniser is using.
 *
 *	- tokFreeState and tokAbort get rid of the memory used by the
 *	  TokState. That TokState becomes invalid and should not be used
 *	  further.
 *
 *	- tokFinal transfers ownership of the output list it creates to its
 *	  caller; the caller is responsible for freeing all the tokClass
 *	  structures and the associated bitmaps, DgpAdt_Lists and Images.
 *	  There is a convenience routine provided to do this
 *	  for you (tokFreeOutput()).  In fact, due to some peculiarities, this
 *	  is the only supported way of freeing such an output list.
 *
 *      - tokPartial also creates an output list; tokResume destroys it.
 *	  The output list created by tokPartial may be passed to tokFreeOutput,
 *	  but only if the TokState will not be used further, and is correctly
 *	  disposed of (by passing it to tokAbort or tokFreeState).
 *	  Note that in normal usage the tokeniser continues to own the
 *	  bitmaps on the output list, and will consume them in tokResume
 *	  even if TOK_OPT_COPYBITMAPS is specified.
 *
 *	- tokFreeClass can be used to free individual classes from an output
 *	  list; those classes should be unhooked from the list by the caller
 *	  before the list is passed back to tokResume.
 *
 * Caveats:
 *	- There are some option choices that should be made once only: once
 *	  you provide the tokeniser with a certain option value, you should
 *	  never provide it with a different value for the same option, either
 *	  in concurrent or subsequent runs. The options that have this
 *	  property are: widthDiff, heightDiff.
 *
 */

#include "tok-lossy.h"
#include "tokprivate.h"
#include "miscalloc.h"
#include "hauschex.h"
#include <math.h>
#include "cross.h"
#include "template.h"
#include "class.h"
#include "classtable.h"
#include "image_defs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/tok.c,v 1.40 2003/02/10 22:53:42 mkrocken Exp $");

#ifdef	DBUG
#include "dbug.h"
Dgp_Int32 curbitmap = 0;
Dgp_Int32 maxbitmap = 0;
Dgp_Int32 curimbitmap = 0;
Dgp_Int32 maximbitmap = 0;
#endif

static ErrCode tokBitmapToBox(tokBitmap *bitmap, tokOptions *opt, tokCBoxType **result);
static ErrCode addpad(tokCBoxType *box, tokBitmap *bitmap, tokOptions *opt);
static ErrCode unpad(tokCBoxType *box, tokOptions *opt);
static ErrCode checkBoxToBin(TokState state, tokClassTable classes, 
                             tokCBoxType *box, Dgp_Int32 bin_w, Dgp_Int32 bin_h, 
                             Dgp_Bool *matched);
static ErrCode
checkBoxToBinVerify(TokState      state, 
              tokClassTable classes, 
              tokCBoxType   *box, 
              Dgp_Int32     bin_w, 
              Dgp_Int32     bin_h, 
              Dgp_Bool      *matched, 
              DgpAdt_List   *classMatched);
static ErrCode generateOutput(TokState state, DgpAdt_List *output, Dgp_Bool merge);
static void freeState(_tokState *state);

ErrCode
tokInit(tokOptions *options, TokState *state)
{
    _tokState *newTokState = (_tokState *)NULL;

  DGP_PROLOGUE("tokInit()", options->DGP_PERRSTATE);

    /* validate small pool usage */
  DGP_SMALLPOOL_VALIDATE(tokCModelInfo, TOKMDL_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(tokCImageInfo, TOKIMG_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(tokCPixelSum, TOKPXL_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(struct _tokCClass, TOKCLASS_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(tokCBoxType, TOKBOX_POOL_ID);

    /* Now we can start allocating memory. */
  DGP_CHECK_ALLOC_CLEANUP(
    newTokState = (options->DGP_MALLOC)(sizeof(_tokState)));

    /* Copy over the options */
    newTokState->options.fracHausd = options->fracHausd;
    newTokState->options.maxDense = options->maxDense;
    newTokState->options.widthDiff = options->widthDiff;
    newTokState->options.heightDiff = options->heightDiff;
    newTokState->options.commitThresh = options->commitThresh;
    newTokState->options.freezeHistThresh = options->freezeHistThresh;
    newTokState->options.makeHistThresh = options->makeHistThresh;
    newTokState->options.returnHistThresh = options->returnHistThresh;
    newTokState->options.nneigh = options->nneigh;
    newTokState->options.xshift = options->xshift;
    newTokState->options.xslopechange = options->xslopechange;
	newTokState->options.lowerDilationLimit = options->lowerDilationLimit;
    newTokState->options.newslope = options->newslope;
    newTokState->options.dilationRadius = options->dilationRadius;
    newTokState->options.mockRadius = options->dilationRadius;
    newTokState->options.flags = options->flags;

    /* And the state. */
    newTokState->options.dgp_state = options->dgp_state;

    /* And the callback. */
    newTokState->options.mergeCallback = options->mergeCallback;
    newTokState->options.callbackCookie = options->callbackCookie;

    /* Initialise the class table to empty */
    newTokState->classes.elts = (DgpImage_Ptr) NULL;
    newTokState->classes.count = 0;

    /* Handle the special hack cases */
    if (options->flags & TOK_OPT_NONTPDIL) {

    	/* What is the real dilation radius (how much padding do we need?) */
    	if ((options->dilationRadius == 10) ||
        	    (options->dilationRadius == 14)) {
    	    newTokState->options.dilationRadius = 1;
        }
    	else if (options->dilationRadius == 20) {
    	    newTokState->options.dilationRadius = 2;
    	}
    }

    /* Clear other tokenizer state. */
    newTokState->search_order = (Dgp_Int32 *)NULL;
    newTokState->search_count = 0;
    newTokState->ngoal_vals = 0;
    newTokState->goal_vals = (Dgp_Int32 *)NULL;
    newTokState->nsum_vals = 0;
    newTokState->sum_vals = (Dgp_Int32 *)NULL;

    *state = (TokState)newTokState;

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
    }

/* 3.10.2000 / dhtesq. */
/* Although this function is still called, the defaults specified for the command-line machinery */
/* now take precedence. */
void 
tokSetDefaults(tokOptions *options)
{
    options->fracHausd = (Dgp_Float32)0.97;
    options->maxDense = 3;
    options->widthDiff = options->heightDiff = 1;
    options->commitThresh = 10000;
    options->freezeHistThresh = 10000;
    options->makeHistThresh = 10;
    options->returnHistThresh = 0;
    options->nneigh = 8;
    options->xshift = 3;
    options->xslopechange = 6;
	options->lowerDilationLimit = 6;
    options->newslope = 2.;
    options->dilationRadius = 1;
    options->mockRadius = 1;
    options->flags = 0;

    options->mergeCallback = NULL;
    options->callbackCookie = NULL;
}

ErrCode
tokGetCount(TokState state, Dgp_Int32 *count){
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    DGP_PROLOGUE( "tokGetCount()", DGP_PERRSTATE);
    DGP_STOP_WHINE;
    *count = intstate->classes.count;
    NORMAL_RETURN;
}

ErrCode
tokAddBitmap(TokState state, tokBitmap *bitmap)
{
    _tokState *intstate = (_tokState *)state;
    tokClassTable classes;
    Dgp_Int32 *order;
    Dgp_Int32 ordernum;
    Dgp_Bool matched;
    Dgp_Int32 i;
    tokCClass newClass = (tokCClass )NULL;
    tokCBoxType *newBox = (tokCBoxType *)NULL;
    DgpImage_Ptr newTable = (DgpImage_Ptr)NULL;
    DgpAdt_List newBin = DGP_NULLLIST;
    Dgp_State *dgp_state = intstate->options.dgp_state;

DGP_PROLOGUE( "tokAddBitmap()", DGP_PERRSTATE);

    /*
     * Convert it to our internal representation. This also builds
    } else {

     * the TP-dilated (+inverted) "image". It either copies it or
     * takes ownership, according to options->flags.
     */
DGP_ERR_PASS_CLEANUP(    
    tokBitmapToBox(bitmap, &(intstate->options), &newBox));

    classes = &(intstate->classes);

    /* Get the scan order */
    if (intstate->search_order == (Dgp_Int32 *)NULL) {
DGP_ERR_PASS_CLEANUP(
        tokCTCalcOrder(intstate->options.widthDiff,
		       intstate->options.heightDiff, &(intstate->search_count),
		       &(intstate->search_order), &(intstate->options)));
    }
    order = intstate->search_order;
    ordernum = intstate->search_count;

    /* Compare it against all the bins in order */
    for (i = 0; i < ordernum; i++)
    {
DGP_ERR_PASS_CLEANUP(
	    checkBoxToBin(state, classes, newBox, newBox->width + order[2 * i],
			   newBox->height + order[2 * i + 1],
			   &matched));

        if (matched) {
	        /* It matched something in that bin - we're done. */
	        NORMAL_RETURN;
	    }
	}

    /* It matched nothing - make a new class. */
DGP_ERR_PASS_CLEANUP(
    tokCBoxToClass(newBox, &newClass, &(intstate->options)));

    /* and put it in the table */
DGP_ERR_PASS_CLEANUP(
    tokCTAddToTable(classes, newClass, &(intstate->options)));

    /* All done - it's added. */
    NORMAL_RETURN;

ERR_CLEANUP:

    /* Free the stuff we've allocated so far */

    if (newBox != (tokCBoxType *)NULL) {
    	FreeBox(newBox, &(intstate->options));
	}

    if (newTable != (DgpImage_Ptr)NULL) {
    	DgpImage_free(newTable);
	}

    if (newClass != (tokCClass)NULL) {
    	ClassFree(newClass, &(intstate->options));
	}

    if (newBin != (DgpAdt_List)NULL) {
    	DgpAdt_liFree(newBin);
	}

    ERR_RETURN;
}

/* This is a cheesy version of the above function that notifies the caller */
/* if a new class is found;  used by verifier. */
/* It also returns a list of member tokCBoxType{}s for the matched class (or null if none). */
ErrCode
tokAddBitmapVerify(TokState state, tokBitmap *bitmap, Dgp_Bool *newClassFound, DgpAdt_List *classMatched)
{
    _tokState *intstate = (_tokState *)state;
    tokClassTable classes;
    Dgp_Int32 *order;
    Dgp_Int32 ordernum;
    Dgp_Bool matched;
    Dgp_Int32 i;
    tokCBoxType *newBox = (tokCBoxType *)NULL;
    Dgp_State *dgp_state = intstate->options.dgp_state;

    DGP_PROLOGUE( "tokAddBitmapVerify()", DGP_PERRSTATE);

    *newClassFound = DGP_FALSE;
    *classMatched = DGP_NULLLIST;
    /*
     * Convert it to our internal representation. This also builds
    } else {

     * the TP-dilated (+inverted) "image". It either copies it or
     * takes ownership, according to options->flags.
     */
DGP_ERR_PASS_CLEANUP(    
    tokBitmapToBox(bitmap, &(intstate->options), &newBox));

    classes = &(intstate->classes);

    /* Get the scan order */
    if (intstate->search_order == (Dgp_Int32 *)NULL) {
DGP_ERR_PASS_CLEANUP(
        tokCTCalcOrder(intstate->options.widthDiff,
		       intstate->options.heightDiff, &(intstate->search_count),
		       &(intstate->search_order), &(intstate->options)));
    }
    order = intstate->search_order;
    ordernum = intstate->search_count;

    /* Compare it against all the bins in order. */
    for (i = 0; i < ordernum; i++)
    {
DGP_ERR_PASS_CLEANUP(
	    checkBoxToBinVerify(state, classes, newBox, newBox->width + order[2 * i],
			   newBox->height + order[2 * i + 1],
			   &matched, classMatched));

        if(matched){
	        /* It matched something in that bin - we're done. */
            NORMAL_RETURN;
        }
	}
    *newClassFound = DGP_TRUE;
    /* Purification... since tokenizer made bitmap copies, free the image made */
    /* for the comparison. */
    /* If it matched, it'll be hung off the class table and standard tokenizer clean-up */
    /* will do it in. */
    DGP_FREE(newBox->image->bits);
    DGP_FREE(newBox->model->bits);
    NORMAL_RETURN;

ERR_CLEANUP:

    /* Free the stuff we've allocated so far. */

    if (newBox != (tokCBoxType *)NULL) {
    	FreeBox(newBox, &(intstate->options));
	}

    ERR_RETURN;
}

ErrCode
tokMerge(TokState state)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    tokClassTable classes;

DGP_PROLOGUE( "tokMerge()", DGP_PERRSTATE);

    classes = &(intstate->classes);

    /* Do the final merging */
DGP_ERR_PASS_CLEANUP(
    tokCTFinalMerge(state, classes));

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
tokFinalNoMerge(TokState state, DgpAdt_List *output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;

    DGP_PROLOGUE("tokFinal()", DGP_PERRSTATE);

    DGP_ERR_PASS(generateOutput(state, output, DGP_FALSE));

    NORMAL_RETURN;
}

ErrCode
tokFinal(TokState state, DgpAdt_List *output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;

DGP_PROLOGUE("tokFinal()", DGP_PERRSTATE);

DGP_ERR_PASS(
    generateOutput(state, output, DGP_TRUE));

    NORMAL_RETURN;
}

ErrCode
tokPartialNoMerge(TokState state, DgpAdt_List *output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;

    DGP_PROLOGUE("tokPartial()", DGP_PERRSTATE);

    DGP_ERR_PASS(
        generateOutput(state, output,DGP_FALSE));

    NORMAL_RETURN;

}

ErrCode
tokPartial(TokState state, DgpAdt_List *output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;

    DGP_PROLOGUE("tokPartial()", DGP_PERRSTATE);

    DGP_ERR_PASS(generateOutput(state, output, DGP_TRUE));

    NORMAL_RETURN;
}

ErrCode
tokResumeKeepMember(TokState state, DgpAdt_List output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    tokClass *oldclass;
    tokCClass newclass;
    tokOffset *offset;
    Dgp_UInt32 oldflags;
    DgpAdt_ListNode ln,ln2;

DGP_PROLOGUE( "tokResume()", DGP_PERRSTATE);

    /*
     * Run through the list they gave us, creating new classes.
     * We have to
     * - create a tokCClass for each one
     * - copy the bitmap and histogram, adding appropriate padding to each
     * - destroy the members list of each tokCClass
     * - destroy the list
     */
    /* We want to make sure that the bitmaps *aren't* copied */
    oldflags = intstate->options.flags;
    intstate->options.flags &= ~TOK_OPT_COPYBITMAPS;

    /* Deconstruct and convert the list */
    for (ln = DGPADT_LIFIRST(output); ln != DGP_NULLLISTNODE; DgpAdt_liRemAndNext(ln, &ln)) {
        oldclass = (tokClass *)DGPADT_LIGET(ln);

		TOKCLASS_SEC_MALLOC(&newclass);
        DGP_ERR_PASS_CLEANUP( DgpAdt_liNew(&newclass->members, dgp_state));
  		DGP_ERR_PASS_CLEANUP( tokBitmapToBox(&oldclass->exemplar,
                							&(intstate->options), 
											&newclass->exemplar));

        /*
         * Create a tokCPixelSum for this new class.  You need to create one
         * even if there's no histogram:  only singleton, uncommitted,
         * classes are allowed to be without one, and this doesn't count.
         */
  		TOKPXL_SEC_MALLOC(&newclass->modelsum);

        newclass->modelsum->sumisbyte = DGP_TRUE;
        newclass->modelsum->sum.bytesum = (DgpImage_Gray)NULL;
        newclass->modelsum->minx = newclass->modelsum->maxx = (Dgp_Int16)intstate->options.dilationRadius;
        newclass->modelsum->miny = newclass->modelsum->maxy = 0;
        newclass->modelsum->nummodels = 0;

        /* Copy the histogram, if any */
        if (oldclass->histIsByte ?
                (oldclass->histogram.byteHist != (DgpImage_Gray)NULL) :
                (oldclass->histogram.shortHist != (DgpImage_Short)NULL)) {
            if (oldclass->histIsByte) {
                newclass->modelsum->sum.bytesum =
                    oldclass->histogram.byteHist;
                /* Add the padding back into the histogram */
                DgpImage_setOffset(newclass->modelsum->sum.bytesum,
                    DGPIMG_XBASE(newclass->modelsum->sum.bytesum) +
                    intstate->options.dilationRadius,
                    DGPIMG_YBASE(newclass->modelsum->sum.bytesum));

                /* Put in conservative bounds on where the hist is non-zero */
                newclass->modelsum->minx =
                    (Dgp_Int16)DGPIMG_XBASE(newclass->modelsum->sum.bytesum);
                newclass->modelsum->miny =
                    (Dgp_Int16)DGPIMG_YBASE(newclass->modelsum->sum.bytesum);
                newclass->modelsum->maxx =
                    (Dgp_Int16)DGPIMG_XMAX(newclass->modelsum->sum.bytesum);
                newclass->modelsum->maxy =
                    (Dgp_Int16)DGPIMG_YMAX(newclass->modelsum->sum.bytesum);
		    }
            else
            {
                newclass->modelsum->sum.Shortsum = oldclass->histogram.shortHist;
                DgpImage_setOffset(newclass->modelsum->sum.Shortsum,
                    DGPIMG_XBASE(newclass->modelsum->sum.Shortsum) +
                    intstate->options.dilationRadius,
                    DGPIMG_YBASE(newclass->modelsum->sum.Shortsum));

                newclass->modelsum->minx =
                    (Dgp_Int16)DGPIMG_XBASE(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->miny =
                    (Dgp_Int16)DGPIMG_YBASE(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->maxx =
                    (Dgp_Int16)DGPIMG_XMAX(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->maxy =
                    (Dgp_Int16)DGPIMG_YMAX(newclass->modelsum->sum.Shortsum);
    		}

    	    newclass->modelsum->sumisbyte = oldclass->histIsByte;
    	    newclass->modelsum->nummodels = oldclass->histcount;
	    }

    	newclass->status = TOKCLASS_COMMITTED | TOKCLASS_IMMUTABLE |
    	    TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN | TOKCLASS_CHECKCLEAN;
        /*
         * If the histogram was previously frozen, or if there was no
         * histogram passed back (recognisable by nummodels being zero),
         * then the histogram is frozen.
         */
        if ((newclass->modelsum->nummodels >=
                intstate->options.freezeHistThresh) ||
                (newclass->modelsum->nummodels == 0)) {
            newclass->status |= TOKCLASS_FROZENHIST;
        }

    	newclass->userdata = oldclass->userdata;
		    /* Now start converting the members */
		    DGPADT_FOREACH (ln2, oldclass->members) {
                DgpAdt_ListNode retnode = NULL;
            	if (intstate->options.flags & TOK_OPT_RETURNOFFSET) {
                	tokCBoxType *newBox=NULL;
                	offset = (tokOffset *)DGPADT_LIGET(ln2);
                    TOKBOX_SEC_MALLOC (&newBox);
					newBox->xOffset = offset->xOffset;
					newBox->yOffset = offset->yOffset;
					newBox->userdata = offset->userdata;
                	newBox->width = 0;
                	newBox->height = 0;
                	newBox->image = NULL;
                	newBox->model = NULL;
			    	DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(newclass->members, (void*) newBox, &retnode));
			    	newBox = (tokCBoxType *)NULL;
				}else{
                	offset = (tokOffset *)DGPADT_LIGET(ln2);
			    	DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(newclass->members, (void*) offset, &retnode));
					offset = NULL;
				}
				retnode = NULL;
			}

            /* We've got the class - add it to the table */
            DGP_ERR_PASS_CLEANUP(
                tokCTAddToTable(&intstate->classes, newclass, &(intstate->options)));
            if (intstate->options.flags & TOK_OPT_RETURNOFFSET) {
                DGPADT_FOREACH (ln2, oldclass->members) {
    	            offset = (tokOffset *)DGPADT_LIGET(ln2);
		            DGP_ASSERT(offset != (tokOffset *)NULL);
    	            DGP_FREE((void *)offset);
                }
        }
    	DgpAdt_liFree(oldclass->members);
       	(DGP_FREE)((void *)oldclass);
    }
    
    intstate->options.flags = oldflags;
    
    DgpAdt_liFree(output);
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;	/*WRITEME*/
}



ErrCode
tokResume(TokState state, DgpAdt_List output)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    tokClass *oldclass;
    tokCClass newclass;
    tokOffset *offset;
    Dgp_UInt32 oldflags;
    DgpAdt_ListNode ln, ln2;

    DGP_PROLOGUE( "tokResume()", DGP_PERRSTATE);

    /*
     * Run through the list they gave us, creating new classes.
     * We have to
     * - create a tokCClass for each one
     * - copy the bitmap and histogram, adding appropriate padding to each
     * - destroy the members list of each tokCClass
     * - destroy the list
     */
    /* We want to make sure that the bitmaps *aren't* copied */
    oldflags = intstate->options.flags;
    intstate->options.flags &= ~TOK_OPT_COPYBITMAPS;

    /* Deconstruct and convert the list */
    for (ln = DGPADT_LIFIRST(output); ln != DGP_NULLLISTNODE; DgpAdt_liRemAndNext(ln, &ln)) {
        oldclass = (tokClass *)DGPADT_LIGET(ln);

          TOKCLASS_SEC_MALLOC(&newclass);

          DGP_ERR_PASS_CLEANUP(
                DgpAdt_liNew(&newclass->members, dgp_state));

          DGP_ERR_PASS_CLEANUP(
                tokBitmapToBox(&oldclass->exemplar,
                        &(intstate->options), &newclass->exemplar));

        /*
         * Create a tokCPixelSum for this new class.  You need to create one
         * even if there's no histogram:  only singleton, uncommitted,
         * classes are allowed to be without one, and this doesn't count.
         */
        TOKPXL_SEC_MALLOC(&newclass->modelsum);

        newclass->modelsum->sumisbyte = DGP_TRUE;
        newclass->modelsum->sum.bytesum = (DgpImage_Gray)NULL;
        newclass->modelsum->minx = newclass->modelsum->maxx = (Dgp_Int16)intstate->options.dilationRadius;
        newclass->modelsum->miny = newclass->modelsum->maxy = 0;
        newclass->modelsum->nummodels = 0;

        /* Copy the histogram, if any */
        if (oldclass->histIsByte ?
                (oldclass->histogram.byteHist != (DgpImage_Gray)NULL) :
                (oldclass->histogram.shortHist != (DgpImage_Short)NULL)) {
            if (oldclass->histIsByte) {
                newclass->modelsum->sum.bytesum =
                    oldclass->histogram.byteHist;
                /* Add the padding back into the histogram */
                DgpImage_setOffset(newclass->modelsum->sum.bytesum,
                    DGPIMG_XBASE(newclass->modelsum->sum.bytesum) +
                    intstate->options.dilationRadius,
                    DGPIMG_YBASE(newclass->modelsum->sum.bytesum));

                /* Put in conservative bounds on where the hist is non-zero */
                newclass->modelsum->minx =
                    (Dgp_Int16)DGPIMG_XBASE(newclass->modelsum->sum.bytesum);
                newclass->modelsum->miny =
                    (Dgp_Int16)DGPIMG_YBASE(newclass->modelsum->sum.bytesum);
                newclass->modelsum->maxx =
                    (Dgp_Int16)DGPIMG_XMAX(newclass->modelsum->sum.bytesum);
                newclass->modelsum->maxy =
                    (Dgp_Int16)DGPIMG_YMAX(newclass->modelsum->sum.bytesum);
		    }
            else
            {
                newclass->modelsum->sum.Shortsum = oldclass->histogram.shortHist;
                DgpImage_setOffset(newclass->modelsum->sum.Shortsum,
                    DGPIMG_XBASE(newclass->modelsum->sum.Shortsum) +
                    intstate->options.dilationRadius,
                    DGPIMG_YBASE(newclass->modelsum->sum.Shortsum));

                newclass->modelsum->minx =
                    (Dgp_Int16)DGPIMG_XBASE(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->miny =
                    (Dgp_Int16)DGPIMG_YBASE(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->maxx =
                    (Dgp_Int16)DGPIMG_XMAX(newclass->modelsum->sum.Shortsum);
                newclass->modelsum->maxy =
                    (Dgp_Int16)DGPIMG_YMAX(newclass->modelsum->sum.Shortsum);
    		}

    	    newclass->modelsum->sumisbyte = oldclass->histIsByte;
    	    newclass->modelsum->nummodels = oldclass->histcount;
	    }

    	newclass->status = TOKCLASS_COMMITTED | TOKCLASS_IMMUTABLE |
    	    TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN | TOKCLASS_CHECKCLEAN;
        /*
         * If the histogram was previously frozen, or if there was no
         * histogram passed back (recognisable by nummodels being zero),
         * then the histogram is frozen.
         */
        if ((newclass->modelsum->nummodels >=
                intstate->options.freezeHistThresh) ||
                (newclass->modelsum->nummodels == 0)) {
            newclass->status |= TOKCLASS_FROZENHIST;
        }

    	newclass->userdata = oldclass->userdata;

    	/* We've got the class - add it to the table */
        DGP_ERR_PASS_CLEANUP(
    	tokCTAddToTable(&intstate->classes, newclass, &(intstate->options)));

    	/* We've converted everything - delete the old one */
        DGP_ASSERT(oldclass->members != DGP_NULLLIST);

    	if (intstate->options.flags & TOK_OPT_RETURNOFFSET) {
    	    DGPADT_FOREACH (ln2, oldclass->members) {
    	    	offset = (tokOffset *)DGPADT_LIGET(ln2);
		        DGP_ASSERT(offset != (tokOffset *)NULL);
    	    	DGP_FREE((void *)offset);
	    }
        }
    	DgpAdt_liFree(oldclass->members);
       	(DGP_FREE)((void *)oldclass);
    }

    intstate->options.flags = oldflags;

    DgpAdt_liFree(output);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;	/*WRITEME*/
}

ErrCode
tokFreeOutput(TokState state, DgpAdt_List output, Dgp_Bool hasOffsets)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    DgpAdt_ListNode ln, ln2;
    tokClass *class;
    tokOffset *offset;

    DGP_PROLOGUE( "tokFreeOutput()", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGPADT_FOREACH (ln, output) {
        class = (tokClass *)DGPADT_LIGET(ln);
            DGP_ASSERT(class != (tokClass *)NULL);
            DGP_ASSERT(class->exemplar.bits != (Dgp_UInt32 *)NULL);

            DGP_FREE(class->exemplar.bits);

        if (class->histIsByte) {
            if (class->histogram.byteHist != (DgpImage_Gray)NULL) {
                DgpImage_free(class->histogram.byteHist);
            }
        }
        else {
            if (class->histogram.shortHist != (DgpImage_Short)NULL) {
                DgpImage_free(class->histogram.shortHist);
            }
        }

    	DGP_ASSERT(class->members != DGP_NULLLIST);

        if (hasOffsets) {
            DGPADT_FOREACH (ln2, class->members) {
                offset = (tokOffset *)DGPADT_LIGET(ln2);
                DGP_ASSERT(offset != (tokOffset *)NULL);
                DGP_FREE((void *)offset);
            }
        }
    	DgpAdt_liFree(class->members);
    	DGP_FREE((void *)class);
    }
    DgpAdt_liFree(output);

    NORMAL_RETURN;
}

ErrCode
tokFreeClass(TokState state, tokClass *class)
{
    DgpAdt_ListNode ln;
    tokOffset *offset;
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;

    DGP_PROLOGUE( "tokFreeClass()", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_ASSERT(class != (tokClass *)NULL);
    DGP_ASSERT(class->exemplar.bits != (Dgp_UInt32 *)NULL);

    (DGP_FREE)(class->exemplar.bits);


    if (class->histIsByte) {
	if (class->histogram.byteHist != (DgpImage_Gray)NULL) {
	    DgpImage_free(class->histogram.byteHist);
	}
    }
    else {
	if (class->histogram.shortHist != (DgpImage_Short)NULL) {
	    DgpImage_free(class->histogram.shortHist);
	}
    }

    DGP_ASSERT(class->members != DGP_NULLLIST);

    if (intstate->options.flags & TOK_OPT_RETURNOFFSET) {
	DGPADT_FOREACH (ln, class->members) {
	    offset = (tokOffset *)DGPADT_LIGET(ln);
	    DGP_ASSERT(offset != (tokOffset *)NULL);
	    (DGP_FREE)((void *)offset);
	}
    }
    DgpAdt_liFree(class->members);
    (DGP_FREE)((void *)class);

    NORMAL_RETURN;
}

void 
tokAbort(TokState state)
{
    _tokState *intstate = (_tokState *)state;

    /* Free up all the stuff hung off the state */
    freeState(intstate);
}

void
tokFreeState(TokState state)
{
  _tokState *intstate = (_tokState *)state;

  freeState(intstate);
}

/*
 * Walk through the table of classes, and generate an output list from them.
 * Unlink them all from the table, but don't destroy the table itself.
 * This routine is used by tokFinal and tokPartial to generate their output;
 * tokFinal then goes on and does some cleanup.
 */
static ErrCode
generateOutput(TokState state, DgpAdt_List *output, Dgp_Bool merge)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    tokClassTable classes;
    DgpAdt_List classlist = DGP_NULLLIST;
    tokClass *newClass = (tokClass *)NULL;
    tokOffset *newOffset = (tokOffset *)NULL;
    tokCClass oldClass;
    DgpAdt_ListNode classln, memberln, newclassln;
    DgpImage_Ptr elts;
    DgpAdt_List bin;
    DgpAdt_ListNode retnode;
    tokOffset *offset;
    tokCBoxType *member;
    Dgp_Int32 w, h;

DGP_PROLOGUE("generateOutput()", DGP_PERRSTATE);

    classes = &(intstate->classes);

    if (classes->elts == (DgpImage_Ptr)NULL) {
	/* They never gave us any bitmaps at all... */
	DGP_ASSERT(classes->count == 0);
	DGP_ERR_PASS (DgpAdt_liNew(&classlist, dgp_state));
	*output = classlist;
	NORMAL_RETURN;
    }

    /* Do the final merging if asked */
    if(merge)     
        DGP_ERR_PASS_CLEANUP(tokCTFinalMerge(state, classes));

    /* Do the final alignment */
    if (intstate->options.flags & TOK_OPT_ALIGN) {
DGP_ERR_PASS_CLEANUP(
      tokCTDoAlign(classes, &(intstate->options)));
    }

    /* Generate the final output */
	DGP_ERR_PASS (DgpAdt_liNew(&classlist, dgp_state));

    elts = intstate->classes.elts;

    DGP_ASSERT(DGPIMG_XBASE(elts) == 0);
    DGP_ASSERT(DGPIMG_YBASE(elts) == 0);

    /*
     * We're going to do this in two passes:
     * 1. Create the output list, filling in all the non-pointer fields.
     * 2. Fill in all the pointer fields.
     *
     * The reason for doing it this way is that if something fails when
     * we're creating the output list, we want to be able to free the
     * partial list without freeing any of the state that's still held in the
     * tokeniser state.
     */
    for (h = 0; h < DGPIMG_HEIGHT(elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(elts); w++) {
	    bin = (DgpAdt_List)DGPIMG_REF(elts, w, h);
	    if (bin != DGP_NULLLIST) {
		/* Convert each class in the bin */
		DGPADT_FOREACH (classln, bin) {
		    oldClass = tokCClassData(classln);
		    DGP_ASSERT(oldClass != (tokCClass)NULL);
		    DGP_ASSERT(oldClass->exemplar != (tokCBoxType *)NULL);
		    DGP_ASSERT(oldClass->members != DGP_NULLLIST);

            DGP_CHECK_ALLOC_CLEANUP(
		        newClass = (tokClass *)(DGP_MALLOC)(sizeof(*newClass)));

		    /*
		     * Create its histogram if it had been deferred.
		     * This should be done *before* unpad, since that would
		     * change the coordinate system.
		     */
		    if ((intstate->options.flags & TOK_OPT_RETURNHIST) && 
                ((Dgp_Int32)DGPADT_LILEN(oldClass->members) >= intstate->options.returnHistThresh)) {
			if (oldClass->modelsum != (tokCPixelSum *)NULL) {
			    /*
			     * It thought about having a histogram, and we
			     * need one.
			     */
			    if (oldClass->modelsum->sumisbyte ?
				(oldClass->modelsum->sum.bytesum ==  (DgpImage_Gray)NULL) :
				(oldClass->modelsum->sum.Shortsum ==  (DgpImage_Short)NULL)) {
				    /* but it didn't actually get one */
                    DGP_ERR_PASS_CLEANUP(makeRealHist(oldClass, &(intstate->options)));
			    }
			    newClass->histcount =
				oldClass->modelsum->nummodels;
			}
			else {
			    /*
			     * Only singleton classes shouldn't have even
			     * thought about building histograms.
			     */
			    DGP_ASSERT(DGPADT_LILEN(oldClass->members) == 1);
			    newClass->histcount = 1;
		        }
		    }

		    /* Get rid of the padding on this */
            DGP_ERR_PASS_CLEANUP(  unpad(oldClass->exemplar, &(intstate->options)));

		    newClass->exemplar.width = (Dgp_UInt16)oldClass->exemplar->width;
		    newClass->exemplar.height = (Dgp_UInt16)oldClass->exemplar->height;
		    newClass->exemplar.bits = (Dgp_UInt32 *)NULL;
		    newClass->exemplar.longsperline = (Dgp_UInt16)oldClass->exemplar->model->row_width;
		    newClass->exemplar.userdata = (void *)NULL;
		    newClass->histIsByte = DGP_TRUE;
		    newClass->histogram.byteHist = (DgpImage_Gray)NULL;
		    newClass->userdata = oldClass->userdata;

		    /* Now start converting the members */
            if(oldClass->members != DGP_NULLLIST){
			DGP_ERR_PASS (DgpAdt_liNew(&newClass->members, dgp_state));
		    DGPADT_FOREACH (memberln, oldClass->members) {
			member = tokCMemberBox(memberln);
			if (!(intstate->options.flags & TOK_OPT_RETURNOFFSET))                         {
			    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(newClass->members, member->userdata, &retnode));
			}
			else {
                DGP_CHECK_ALLOC_CLEANUP(
			    newOffset = (tokOffset *)
				(DGP_MALLOC)(sizeof *newOffset));

			    newOffset->xOffset = member->xOffset;
			    newOffset->yOffset = member->yOffset;
			    newOffset->userdata = member->userdata;
			    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(newClass->members, (void*) newOffset, &retnode));
			    newOffset = (tokOffset *)NULL;
			}
		    }
            }
		    /* Hook the new class onto the list we're building */
			DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(classlist, (void*) newClass, &retnode));
		    newClass = (tokClass *)NULL;
		}
	    }
	}
    }

    /* Now pass 2 - nothing in here can fail. */
    newclassln = DGPADT_LIFIRST(classlist);
    for (h = 0; h < DGPIMG_HEIGHT(elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(elts); w++) {
	    bin = (DgpAdt_List)DGPIMG_REF(elts, w, h);
	    if (bin != DGP_NULLLIST) {
		/* Convert each class in the bin */
		for (classln = DGPADT_LIFIRST(bin);
		     classln != DGP_NULLLISTNODE;
		     newclassln = DGPADT_LINEXT(newclassln),
		     classln = DGPADT_LINEXT(classln)) {

		    oldClass = tokCClassData(classln);

		    DGP_ASSERT(newclassln != DGP_NULLLISTNODE);
		    newClass = (tokClass *)DGPADT_LIGET(newclassln);

		    /*
		     * 2 bits of info to copy: the bits and the histogram.
		     * Set the oldClass versions to NULL once they've been
		     * copied.
		     */
		    newClass->exemplar.bits = oldClass->exemplar->model->bits;
		    oldClass->exemplar->model->bits = (Dgp_UInt32 *)NULL;

		    if (oldClass->modelsum != (tokCPixelSum *)NULL) {
			if ((intstate->options.flags & TOK_OPT_RETURNHIST) && ((Dgp_Int32)DGPADT_LILEN(oldClass->members) >=
			     intstate->options.returnHistThresh)) {
			    /*
			     * They want the histograms. Hook them into
			     * newClass (and unhook from oldClass).
			     *
			     * Also, the histograms were created in a
			     * coordinate frame that included the padding on
			     * the left hand side of the exemplar, which has
			     * since been stripped off by unpad. They
			     * should be adjusted accordingly.
			     */
			    newClass->histIsByte =
				oldClass->modelsum->sumisbyte;

			    if (newClass->histIsByte) {
				newClass->histogram.byteHist =
				    oldClass->modelsum->sum.bytesum;
				oldClass->modelsum->sum.bytesum =
				    (DgpImage_Gray)NULL;
				DgpImage_setOffset(newClass->histogram.byteHist,
				    DGPIMG_XBASE(newClass->histogram.byteHist) -
				    intstate->options.dilationRadius,
				    DGPIMG_YBASE(newClass->histogram.byteHist));
			    }
			    else {
				newClass->histogram.shortHist =
				    oldClass->modelsum->sum.Shortsum;
				oldClass->modelsum->sum.Shortsum =
				    (DgpImage_Short)NULL;
				DgpImage_setOffset(newClass->histogram.shortHist,
				    DGPIMG_XBASE(newClass->histogram.shortHist) -
				    intstate->options.dilationRadius,
				    DGPIMG_YBASE(newClass->histogram.shortHist));
			    }
			}
			else {
			    newClass->histIsByte = DGP_TRUE;
			    newClass->histogram.byteHist = (DgpImage_Gray)NULL;
			    /*
			     * Just leave the old histogram alone - it will
			     * get freed aDgp_Int32 with the rest of the state.
			     */
			}
		    }
		    else {
			newClass->histIsByte = DGP_TRUE;
			newClass->histogram.byteHist = (DgpImage_Gray)NULL;
		    }

		    /*
		     * We also free all the stuff hung off the table, but
		     * not the table itself, once everything's been copied
		     * off.  This means that tokPartial knows that it's
		     * returned all the info and doesn't have anything lying
		     * around internally that might confuse a future
		     * tokResume.
		     */
		    DGPADT_FOREACH (memberln, oldClass->members) {
				member = tokCMemberBox(memberln);
				FreeBox(member, &(intstate->options));
		    }
		    ClassFree(oldClass, &(intstate->options));
		}
		DgpAdt_liFree(bin);
		DGPIMG_REF(elts, w, h) = DGP_NULLLIST;
	    }
    }
	}
    /* Check the lists stayed in sync */
    DGP_ASSERT(newclassln == DGP_NULLLISTNODE);
    DGP_ASSERT(DGPADT_LILEN(classlist) == (unsigned)intstate->classes.count);
    intstate->classes.count = 0;

    /* All done - successfully converted. */
    *output = classlist;
    NORMAL_RETURN;

ERR_CLEANUP:

    /* Free any partially constructed things */

    if (classlist != DGP_NULLLIST) {
	    /* Free the partial results */
	    tokFreeOutput(state, 
                      classlist, 
                      (Dgp_Bool)((intstate->options.flags & TOK_OPT_RETURNOFFSET) ? DGP_TRUE : DGP_FALSE));
    }

    /*
     * Don't free any of the stuff hung off newClass - it's just
     * copied from the tokState so far. The exception is the member list
     * and any tokOffsets hung off that.
     */
    if (newClass != (tokClass *)NULL) {
	if (newClass->members != DGP_NULLLIST) {
	    if (intstate->options.flags & TOK_OPT_RETURNOFFSET) {
		DGPADT_FOREACH (memberln, newClass->members) {
		    offset = (tokOffset *)DGPADT_LIGET(memberln);
		    DGP_ASSERT(offset != (tokOffset *)NULL);
		    (DGP_FREE)((void *)offset);
		}
	    }
	    DgpAdt_liFree(newClass->members);
	}

	(DGP_FREE)((void *)newClass);
    }

    if (newOffset == (tokOffset *)NULL) {
	(DGP_FREE)((void *)newOffset);
    }

    ERR_RETURN;
}

/*
 * Take the given bitmap and add all the structures around it to make it
 * into a tokCBoxType. Add the magic one-bit padding at the left, and compute
 * the dilated "image".
 *
 */
static ErrCode
tokBitmapToBox(tokBitmap *bitmap, tokOptions *opt, tokCBoxType **result)
{
    Dgp_State *dgp_state = opt->dgp_state;
    tokCBoxType *newBox = (tokCBoxType *)NULL;
    
    DGP_PROLOGUE("tokBitmapToBox()", DGP_PERRSTATE);
    
    TOKBOX_SEC_MALLOC(&newBox);
    
    newBox->model = (tokCModelInfo *)NULL;
    newBox->image = (tokCImageInfo *)NULL;
    
    newBox->width = bitmap->width;
    newBox->height = bitmap->height;
    newBox->xOffset = newBox->yOffset = 0;
    newBox->userdata = bitmap->userdata;
    
    TOKMDL_SEC_MALLOC(&newBox->model);
    
    /* Add the magic padding to the model, copy the bitmap if nec */
    DGP_ERR_PASS_CLEANUP(
        addpad(newBox, bitmap, opt));
    
    /* and create the image */
    DGP_ERR_PASS_CLEANUP(
        MakeImCrossFromMod(newBox->model, opt, &newBox->image));
    
    *result = newBox; 
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (newBox != (tokCBoxType *)NULL) {
	if (newBox->model != (tokCModelInfo *)NULL) {
	    FreeModel(newBox->model, opt);
	}
	if (newBox->image != (tokCImageInfo *)NULL) {
	    TOKIMG_FREE(newBox->image);
	}
	TOKBOX_FREE(newBox);
    }

    ERR_RETURN;
}

/*
 * Fill out the fields in box->model.
 * Add a opt->dilationRadius pixel pad to the left side of the model bitmap.
 * (Remember the special meanings of 10, 14, 20 non-tp).
 *
 * While you're doing it, compute the number of points and the centroid
 * of the model.
 *
 * Copy the bitmap if the options tell you to; otherwise, use the original
 * one.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. On failure, no memory
 * has been allocated or freed.
 */
static ErrCode
addpad(tokCBoxType *box, tokBitmap *bitmap, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_UInt32 *p, *q, *rowp, *rowq;
    Dgp_UInt32 *newbits;
    Dgp_UInt32 left, t;
    Dgp_Int32 x, y;
    Dgp_Int32 b;
    tokCModelInfo *model = box->model;
    Dgp_Int32 row_used;
    Dgp_Int32 shift_used;
    Dgp_Int32 npts;
    Dgp_Int32 radius = opt->dilationRadius;

DGP_PROLOGUE("addpad()", DGP_PERRSTATE);

    model->xsize = bitmap->width;
    model->ysize = bitmap->height;
    model->row_width = bitmap->longsperline;

    /* I haven't written this case yet */
    DGP_ASSERT(radius < DGP_BITS_UINT32);

    /* We're going to increase the bitmap size */
    model->xsize += radius;
    /* Is there enough room in the row to do the shift in place? */
    if ((model->row_width < (model->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32) ||
	(opt->flags & TOK_OPT_COPYBITMAPS)) {
	/* Not enough room, or we need to copy anyway... realloc */

	model->row_width = (model->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;

DGP_CHECK_ALLOC_CLEANUP(
	newbits =
	    (Dgp_UInt32 *)(DGP_MALLOC)(sizeof(Dgp_UInt32) *
						 model->row_width *
						 model->ysize));

#ifdef	DBUG
	curbitmap += model->row_width * model->ysize * sizeof(Dgp_UInt32);
	if (curbitmap > maxbitmap) {
	    maxbitmap = curbitmap;
	    DEBUG("model bitmap memory peak %d bytes, brk %d\n", maxbitmap, sbrk(0));
	    }
#endif

	row_used = MIN(model->row_width, bitmap->longsperline);

	/* Copy it across. Could do the shift here, I suppose... */
	for (y = 0, rowp = bitmap->bits, rowq = newbits;
	     y < model->ysize;
	     y++, rowp += bitmap->longsperline, rowq += model->row_width) {
	    for (x = 0, p = rowp, q = rowq; x < row_used; x++, p++, q++) {
		*q = *p;
		}
	    /* Clear the extra words, if any */
	    for (; x < model->row_width; x++, q++) {
		*q = 0;
		}
	    }

	if (!(opt->flags & TOK_OPT_COPYBITMAPS)) {
	    /* We own bitmap->bits - we have to free it */
	    (DGP_FREE)((void *)bitmap->bits);
	    }
	model->bits = newbits;
	}
    else {
	/* Just use the bitmap pointer they gave us */
	model->bits = bitmap->bits;
#ifdef	DBUG
	curbitmap += model->row_width * model->ysize * sizeof(Dgp_UInt32);
	if (curbitmap > maxbitmap) {
	    maxbitmap = curbitmap;
	    DEBUG("model bitmap memory peak %d bytes, brk %d\n", maxbitmap, sbrk(0));
	    }
#endif
	}

    /* OK - do the shift. */
    /* How many words used per row? */
    npts = 0;
    /* Originally this many words used */
    row_used = (model->xsize - radius + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    /* and now this many */
    shift_used = (model->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    for (y = 0, rowp = model->bits;
	 y < model->ysize;
	 y++, rowp += model->row_width) {
	/* Shift if required */
	if (radius) {
	    left = 0;
	    for (x = 0, p = rowp; x < row_used; x++, p++) {
		t = *p;
		*p = (t >> radius) | (left << (DGP_BITS_UINT32 - radius));
		left = t;
		}
	    if (shift_used > row_used) {
		*p = (left << (DGP_BITS_UINT32 - radius));
		}
	    }

	/* Count the points and compute the centroid for this row */
	for (x = 0, p = rowp; x < shift_used; x++, p++) {
	    t = *p;
	    for (b = 0; b < DGP_BITS_UINT32; b++) {
		if (t & (1ul << (DGP_BITS_UINT32 - 1 - b))) {
		    npts++;
		    }
		}
	    }
	}

    /* They should know better... */
    DGP_ASSERT(npts > 0);
    box->npts = model->npts = npts;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * This is the opposite of addpad - it takes the dilationRadius pixel
 * padding off.
 * It gets called on the exemplars at the end of the process.
 *
 * In some cases, there might be zero padding on the left, not dilationRadius
 * This can happen when tokFinal fails partway through, and is called again
 * later. Those exemplars that got their padding removed the first time
 * shouldn't have more removed.
 *
 * Make sure that the 0/dilationRadius padding on the left really is the
 * only padding the bitmap has.
 */
static ErrCode
unpad(tokCBoxType *box, tokOptions *opt)
{
    tokCModelInfo *model;
    Dgp_UInt32 tmp, tmp2;
    Dgp_UInt32 *p;
    Dgp_Int32 pad;
    Dgp_Int32 roff;
    Dgp_Int32 i;
    Dgp_Int32 radius;
    
    DGP_PROLOGUE("unpad()", opt->DGP_PERRSTATE);
    DGP_STOP_WHINE;
    radius = opt->dilationRadius;
    
    model = box->model;
    DGP_ASSERT(model != (tokCModelInfo *)NULL);
    
    DGP_ASSERT((box->width == model->xsize - radius) ||
        (box->width == model->xsize));
    DGP_ASSERT(box->height == model->ysize);
    DGP_ASSERT(radius < DGP_BITS_UINT32);
    
    /* Check the left edge */
    tmp = 0;
    for (i = 0, p = model->bits;
	 i < model->ysize;
	 i++, p += model->row_width) {
	tmp |= *p;
	}

     /* Count zero MSBs */
     DGP_ASSERT(tmp != 0);
     for (i = 0, tmp2 = tmp;
     (tmp2 & (1ul << (DGP_BITS_UINT32 - 1))) == 0;
     i++, tmp2 <<= 1) {
     }
     pad = i;
     
     DGP_ASSERT(((pad == 0) || (pad == radius)) &&
         (box->width == model->xsize - pad));
     
     /* Check the right */
     roff = (model->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
     DGP_ASSERT(roff <= model->row_width);
     if (roff > 1) {
         /* It's a different word from the left word - re-OR. */
         tmp = 0;
         for (i = 0, p = model->bits + roff - 1;
         i < model->ysize;
         i++, p += model->row_width) {
             tmp |= *p;
         }
         
     }

    /* Check from the LSBs up */
    if (tmp != 0) {
	for (i = 0; (tmp & 1) == 0; i++, tmp >>= 1) {
	    }
	}
    else {
	i = DGP_BITS_UINT32;
	}

    /*
     * There should have been this many zero columns, if the model
     * was full.
     */
      DGP_ASSERT(i == (((model->xsize % DGP_BITS_UINT32) == 0) ?
		 0 :
		 (DGP_BITS_UINT32 - (model->xsize % DGP_BITS_UINT32))));

    /* Now make sure the top and bottom rows aren't empty */
    for (i = 0, p = model->bits; i < model->row_width; i++, p++) {
	if (*p) {
	    break;
	    }
	}
    DGP_ASSERT(i < model->row_width);

    for (i = 0, p = model->bits + model->row_width * (model->ysize - 1);
	 i < model->row_width;
	 i++, p++) {
	if (*p) {
	    break;
	    }
	}
DGP_ASSERT(i < model->row_width);

    /*
     * OK - we've checked the borders. The value of pad is the number of
     * columns of padding on the right.
     */
    if (pad) {
	/* Shift out the padding in the left boundary */
	tmp2 = 0;
	for (i = model->row_width * model->ysize - 1, p = model->bits + i;
	     i >= 0; i--, p--) {
	    /*
	     * It's a hack, but we're going to ignore row boundaries.
	     * This is safe, as we'll only ever be shifting the padding
	     * (zero) across rows.
	     */
	    tmp = *p;
	    *p = (*p << pad) | tmp2;
	    /* Figure out the carry-over */
	    tmp2 = (tmp >> (DGP_BITS_UINT32 - pad));
	    }
	}

    /* It's narrower now. */
    model->xsize -= pad;

    /*
     * The member offsets *do not* need to be adjusted. Imagine that
     * we shifted all the member bitmaps aDgp_Int32 with the exemplar...
     * the offsets (which are relative) wouldn't change. And since we
     * added the padding to the member bitmaps when we got fed them,
     * we're kinda removing them now by unpadding the exemplar.
     */
    
    NORMAL_RETURN;
}


/*
 * Check the given box against the given bin in the class table. Set
 * *matched to DGP_TRUE if the box matches some class in the bin, DGP_FALSE if
 * it doesn't.
 *
 * If it does match, merge it into that class. That may cause the class
 * to be committed.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure.
 * There are some interesting failure cases: suppose that the box gets
 * successfully merged into a class, which then gets committed, and
 * some part of that process fails. In this case, we return DGP_FALSE, but
 * set *matched to DGP_TRUE, to indicate that the box is still hooked into
 * the class table. On the other hand, if merging the box into the class
 * failed, we return DGP_FALSE and set *matched to DGP_FALSE.
 */
static ErrCode
checkBoxToBin(TokState state, tokClassTable classes, tokCBoxType *box, 
              Dgp_Int32 bin_w, Dgp_Int32 bin_h, Dgp_Bool *matched)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    DgpAdt_List bin;
    DgpAdt_ListNode ln;
    tokCClass c;

DGP_PROLOGUE( "checkBoxToBin()", opt->DGP_PERRSTATE);

    *matched = DGP_FALSE;

    if ((classes->elts == (DgpImage_Ptr)NULL) ||
	        (bin_w <= 0) || (bin_w >= DGPIMG_WIDTH(classes->elts)) ||
	        (bin_h <= 0) || (bin_h >= DGPIMG_HEIGHT(classes->elts))) {
	    NORMAL_RETURN;
	}

    bin = (DgpAdt_List)tokCTRef(classes, bin_w, bin_h);
    if (bin == DGP_NULLLIST) {
	    NORMAL_RETURN;
	}

    /* Found the bin - start checking. */
    DGPADT_FOREACH (ln, bin) {
    	c = tokCClassData(ln);

DGP_ERR_PASS_CLEANUP(
        tokCMaybeMergeBox(state, classes, c, box, matched));

	    if (*matched) {
	        /* Done. */
	        break;
	    }
	}

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;

}


/* A cheesy version of checkBoxToBin() that returns the class too in the event of a match. */
static ErrCode
checkBoxToBinVerify(TokState      state, 
              tokClassTable classes, 
              tokCBoxType   *box, 
              Dgp_Int32     bin_w, 
              Dgp_Int32     bin_h, 
              Dgp_Bool      *matched, 
              DgpAdt_List   *classMatched)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    DgpAdt_List bin;
    DgpAdt_ListNode ln;
    tokCClass c;

DGP_PROLOGUE( "checkBoxToBinVerify()", opt->DGP_PERRSTATE);

    *matched = DGP_FALSE;
    *classMatched = DGP_NULLLIST;

    if ((classes->elts == (DgpImage_Ptr)NULL) ||
	        (bin_w <= 0) || (bin_w >= DGPIMG_WIDTH(classes->elts)) ||
	        (bin_h <= 0) || (bin_h >= DGPIMG_HEIGHT(classes->elts))) {
	    NORMAL_RETURN;
	}

    bin = (DgpAdt_List)tokCTRef(classes, bin_w, bin_h);
    if (bin == DGP_NULLLIST) {
	    NORMAL_RETURN;
	}

    /* Found the bin - start checking. */
    DGPADT_FOREACH (ln, bin) {
    	c = tokCClassData(ln);

DGP_ERR_PASS_CLEANUP(
        tokCMaybeMergeBox(state, classes, c, box, matched));

	    if (*matched) {
	        /* Done. */
            *classMatched = c->members;
	        break;
	    }
	}

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;

}

/*
 * Free up everything that's hooked into the state. Watch out for NULL
 * pointers - they're valid in most fields.
 * Also be careful about the histograms: if TOK_OPT_RETURNHIST is set in
 * the state, then they've all been copied.  However,
 * the thing that copied them NULLed out the old fields, so it's not our
 * problem.
 */
static void
freeState(_tokState *state)
{
    Dgp_State *dgp_state = state->options.dgp_state;
    Dgp_Int32 w, h;
    DgpImage_Ptr elts;
    DgpAdt_List bin;
    DgpAdt_ListNode classln, memberln;
    tokCClass c;
    tokCBoxType *box;

DGP_VOID_PROLOGUE( "freeState()", DGP_PERRSTATE);

    elts = state->classes.elts;
    if (elts != (DgpImage_Ptr)NULL) {
	for (h = DGPIMG_YBASE(elts); h <= DGPIMG_YMAX(elts); h++) {
	    for (w = DGPIMG_XBASE(elts); w <= DGPIMG_XMAX(elts); w++) {
		bin = (DgpAdt_List)DGPIMG_REF(elts, w, h);
		if (bin != DGP_NULLLIST) {
		    DGPADT_FOREACH (classln, bin) {
			c = (tokCClass)DGPADT_LIGET(classln);
			DGP_VOID_ASSERT(c != (tokCClass)NULL);
			DGP_VOID_ASSERT(c->members != DGP_NULLLIST);

			DGPADT_FOREACH (memberln, c->members) {
			    box = (tokCBoxType *)DGPADT_LIGET(memberln);
			    DGP_VOID_ASSERT(box != (tokCBoxType *)NULL);

			    FreeBox(box, &(state->options));
			}
			
			ClassFree(c, &(state->options));
		    }

		    DgpAdt_liFree(bin);
		}
	    }
        }
	
	DgpImage_free(elts);
    }

    if (state->search_order) (DGP_FREE)((void *)state->search_order);
    if (state->goal_vals) (DGP_FREE)((void *)state->goal_vals);
    if (state->sum_vals) (DGP_FREE)((void *)state->sum_vals);

    (DGP_FREE)((void *)state);
}

