/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/tokclass.h,v 1.7 2002/10/11 17:55:40 jlin Exp $
 *
 */

/*
 * This header file (along with tokclasstable.h) defines a set of routines
 * and datastructures that allow the final set of class merge passes to
 * be replaced by a non-Hausdorff type merges (or a different Hausdorff at
 * least).
 *
 * When a mergeCallback is specified in the tokOptions struct, bitmaps are
 * added to the equivalence classes as usual as well as the initial merge
 * pass. After that, however, instead of iterating over the Hausdorff merge
 * until the classtable converges, the callback is made with the classtable
 * the tokenizer state, and a user-defined cookie for the callback.
 *
 *  o The classtable (tokClassTable) is a hashtable over the width and
 *	height of each exemplar. (Think of it as a 2D array where each element
 *	is a DgpAdt_List.) Use tokCTRef to access the table. Each entry may be a DgpAdt_List
 *	or DGP_NULLLIST.
 *
 *  o Each element of the classtable is a DgpAdt_List of classes or DGP_NULLLIST. (See
 *	list.h) (Sometimes called a bin or cluster.) The DgpAdt_List may contain
 *	zero or more elements.
 *
 *  o Each element of such a DgpAdt_List is a tokCClass. Note that this is different
 *	than a tokClass (with one less 'C'). tokClass is the type of each
 *	class in the final output (as defined in tok-lossy.h), while
 *	tokCClass is the type of each class during processing.
 *
 * Notes on memory ownership. The classtable is at all times owned by the
 * tokenizer. Note that whenever a call is made back into the tokenizer,
 * or after the merge callback return, everything hanging off the classtable
 * is also owned by the tokenizer.
 *
 * The functions below can be used to create and delete classes, but a newly
 * created class is not owned by the tokenizer until it is attached to the
 * classtable. Also, a class should be removed from the classtable before
 * it is deleted. The callback is responsible for all list structure
 * maintainence while it is processing the classtable.
 *
 * Exemplars are an exception. Initially all the exemplars are owned by the
 * the tokenizer. Use the tokCSetExemplar function to replace one of the
 * exemplars. This frees the orignal and replaces it with the new examplar.
 * If you wish to maintain ownership of the new exemplar (that is, if you
 * don't want the tokenizer to free it) set the tokOwner parameter to DGP_FALSE.
 * Otherwise the tokenizer will destroy the new exemplar when it, in turn,
 * is replaced, or when the classtable is destroyed. NB: If the new exemplar
 * is _already_ otherwise owned by the tokenizer (e.g. if it is one of the
 * class members) then tokOwner should be DGP_FALSE. This ensures that the
 * tokenizer will not prematurely free the exemplar/member.
 *
 * Never set or free one of the initial exemplars yourself.
 *
 * Also note that the classtable is an _approximate_ hashtable. When the
 * callback is made the tokenizer guarentees that each class will be in the
 * proper bin (that is, the bin cooresponding to the width and height of the
 * exemplar), but the tokenizer does not require that this state is maintained.
 *
 * Two flags are defined below, for committed and immutable classes. See notes
 * in tok-lossy.h for info. The some status bits may be used by the merge
 * callback function. Those bits are the bits which are NOT set in
 * TOKCLASS_RESERVED_BITS (obviously). These bits are guarenteed to be zero
 * when the callback is made. Do not change any of the reserved bits, as this
 * will lead to unpredictable behavior on the part of the tokenizer.
 *
 */

#ifndef	TOKCLASS_H

#define	TOKCLASS_H

#include "list.h"
#include "image.h"
#include "tok-lossy.h"

#ifdef __cplusplus
extern "C"
{
#endif


/* Info about a bitmap in internal format ("model") */
typedef struct {
    Dgp_Int32 xsize;			/* Width of the model (undilated) box */
    Dgp_Int32 ysize;			/* Height of the model (undilated) box */
    Dgp_Int32 row_width;		/* Width of each row in Dgp_Int32s */
    Dgp_UInt32 *bits;	/* Bit representation of the model */
    Dgp_Int32 npts;			/* The number of total model points */
    Dgp_Int32 nedgepts;		/* The number of total model edge pixels */
} tokCModelInfo;

/* Info about a dilated inverted bitmap ("image") */
typedef struct {
    Dgp_Int32 xsize;			/* Width of the image (dilated) box */
    Dgp_Int32 ysize;			/* Height of the image (dilated) box */
    Dgp_Int32 row_width;		/* Width of each row in Dgp_Int32s */
    Dgp_UInt32 *bits;	/* Bit representation of the image */
    Dgp_Int32 npts;			/* The number of total image points */
} tokCImageInfo;

/* Info about a histogram */
typedef struct {
    union {
	DgpImage_Short Shortsum;
	DgpImage_Gray bytesum;
	} sum;
    Dgp_Int16 minx, miny;
    Dgp_Int16 maxx, maxy;
    Dgp_Int32 nummodels;
    Dgp_Bool sumisbyte;
} tokCPixelSum;


/* Info about a bitmap */
typedef struct {
    Dgp_Int32 width, height;		/* Width, height of box (= model dimens) */
    tokCModelInfo *model;	/* Model representations of the box */
    tokCImageInfo *image;	/* Image representation (dilated) of box */
    Dgp_Int16 xOffset;		/* xOffset from class exemplar */
    Dgp_Int16 yOffset;		/* xOffset from class exemplar */
    Dgp_Int32 npts;			/* number of model points */
    void *userdata;
} tokCBoxType;


/* Info about an equivalence class */
typedef struct _tokCClass {
    DgpAdt_List members;
    tokCBoxType *exemplar;
    tokCPixelSum *modelsum;
    Dgp_UInt32 status;
    void *userdata;
} *tokCClass;


#define	TOKCLASS_COMMITTED	(0x01)	/* It's been committed */
#define	TOKCLASS_IMMUTABLE	(0x20)	/* No changes allowed */
#define TOKCLASS_RESERVED_BITS	(0x27)  /* These bits are reserved for the tokenizer */

#define tokCClassData(classlistelt)    ((tokCClass)DGPADT_LIGET(classlistelt))
#define tokCMemberBox(memberlistelt)   ((tokCBoxType*)DGPADT_LIGET(memberlistelt))


/*
 * Translates b by (tx,ty) and merges the two classes together. Sets
 * merged to DGP_TRUE if the two classes where successfully merged, DGP_FALSE
 * otherwise. Return DGP_FALSE on error.
 *
 * b will always be invalid if the merge was successful (i.e. merged == DGP_TRUE).
 * a will always point to the merged class if the merge was successful.
 *
 * Also, b (and everything below it) will automatically be freed if
 * the merge was successful. NB: The caller is responsible for freeing the list
 * node that contained b. Also the caller is responsible for maintaining the
 * count of the number of classes in the classtable. That is, if the merge
 * was successful, then the caller should decrement the classtable count.
 */
extern ErrCode tokCMergeClasses(TokState state, tokCClass a, tokCClass b,
				Dgp_Int32 tx, Dgp_Int32 ty, Dgp_Bool *merged);

/*
 * Free a class and all the stuff hanging below it.
 */
extern ErrCode tokCFreeClass(TokState state, tokCClass a);

/*
 * Create a new class using the specified box as the first and
 * only member and as the exemplar. Return DGP_FALSE on failure.
 */
extern ErrCode tokCNewClass(TokState state, tokCBoxType *box, tokCClass *a);

/*
 * Set the exemplar of a class. Translate all members of the the class by (tx,ty).
 * That is, (tx,ty) is the translation of the new exemplar wrt the old exemplar.
 * The tokOwner parameter allows you to specify whether or not the tokenizer should
 * assume ownership of the exemplar. Passing DGP_TRUE means that the tokenizer will
 * later free the exemplar. Initially the tokenizer owns all synthetic exemplars,
 * by definition, and all nonsynthetic ones (since they are attached to the table.)
 *
 * NOTE: Set tokOwner to DGP_FALSE if the new exemplar is already member of the class.
 * (That way the tokenizer won't prematurely free it.)
 *
 * If the old exemplar was owned by the tokenizer, it will be freed. After the
 * call, (since the new exemplar is attached to the tokClassTable structure and)
 * if the tokOwner flag was set to true, the exemplar is owned by the tokenizer and
 * should not be freed.
 */
extern ErrCode tokCSetExemplar(TokState state, tokCClass a, tokCBoxType* box,
				Dgp_Int32 tx, Dgp_Int32 ty, Dgp_Bool tokOwner);

#ifdef __cplusplus
}
#endif

#endif /* CLASS_H */
