Dictionary encoding routines:

Here's the expected flow of control:

You, the compressor (or the compressor's controller) have accumulated some
tokens and the positions of these tokens on the page. You want to commit
these to some stable storage. There are two parts to this: encoding the
tokens, and encoding the positions wrt that encoding of the tokens. These
two are somewhat independent.

Let's deal with the dictionary block only first. You have a bunch of
tokens, stored in an array in memory. You first pass this in to
tcmp_estimateTokDictStateSize which figures out how big a block of internal
state will be needed; you then allocate it, and pass it in to
tcmp_initializeTokDictState. This initialises it. You then pass this
initialised state block and the block of tokens into
tcmp_estimateEncodedTokDictSize. This estimates out how large a block is
needed to store the encoded tokens; you then allocate such a block and pass
it, the tokens, and the state block into tcmp_encodeTokDict. This can do
two things:

- It can return a code saying "oops - I need more memory than I thought"
  and an estimate of how much more memory; you enlarge the encoded token
  block by this amount and try again.
- It can succeed; in this case, it tells you how much of the encoded token
  block it actually used. It has modified the state block and parts of the
  token block.

Once you have a dictionary state block, you can keep feeding new token
blocks in: the procedure is the same as above, except that you pass the
dictionary state block in to tcmp_estimateTokDictStateSize and
tcmp_initializeTokDictState. You can therefore encode multiple token
blocks, with all the state required being accumulated in the same
dictionary state block (suitably extended etc by
tcmp_initializeTokDictState).

Given a bunch of token positions, and a dictionary state block (which must
encode all the token numbers occurring in the position block), you call
tcmp_estimateEncodedTokPosSize which estimates how much space is required
to encode those positions wrt that state block; you allocate this much
memory and pass it, plus the positions and dictionary state, to
tcmp_encodeTokPositions. As with tcmp_encodeTokDict, this can either say
"oops - more memory please" or "done - used this much".

Types:

struct {
    unsigned long *bits;
    void *extdata;
    unsigned int uses;
    unsigned short width, height;
    unsigned short longsperline;
    boolean committed;
    } token;

What this means:
This is a token.

It is "width" pixels wide by "height" pixels high.

The bitmap containing it is at location "bits", and contains "bytesperline"
bytes per line. There are no other frames: the top-left bit should be found
in the long at *bits (there may be horizontal frames; this is why
"longsperline" is specified rather than being implicit in "width"; vertical
frames can always be adjusted around).
The longs are filled MSB-to-LSB: the leftmost bit in each long (the one
appearing closest to the left of the page) is the one stored in the most
significant bit position.

NOTE: If width is non-zero, but longsperline is zero, this is a special
indication that the memory pointed at by bits is *NOT* a bitmap, but
instead is a G4-compressed representation of the token's bitmap.  It is
*NOT* legal to pass a token represented this way to tcmp_encodeTokDict, but
tcmp_decodeTokDict may return such tokens.  It will only return tokens
represented as G4-compressed bitmaps for single-use tokens; in these cases,
the "uses" field is hijacked to hold the length of the G4-compressed data -
see below.

This token is used "uses" times. This is not an absolute number of uses: it
may actually be used more or fewer times than this, but the relative "uses"
counts of the various tokens will be used to improve the coding. There is
an exception to this: the value 0 is interpreted, not as "this token is not
used at all", but rather as "this token is used once, and only ever once"
(contrast this with the value 1, meaning "this token is used once, or
something like that"). In the "uses==0" case, the thing generating the
token is guaranteeing that there is only one use of this token, ever; this
allows the decompressor to discard large known-unique tokens after they
have been used, saving memory.

(This paragraph no longer applies but is left for historical value. There
are no more toknums! The comment "NA:" will mark such paragraphs below.
    --spoons)  
(NA: This token's number is "toknum"; this number is a unique identifier for
this token. Within a token block, the numbers should be assigned
as a contiguous range of integers, not overlapping that of any other block.
They don't have to be in order, as long as there are no gaps.)

The flag "committed" should be initialised to FALSE when the token is
created. The routine tcmp_encodeTokDict will set it to TRUE when the token
has written to disk.

Any extension data associated with this position can be found in extdata.
This replaces the token_extension struct found below.

A note about token blocks: a number of the routines below take a dictionary
state and a block of tokens. The block of tokens can either include the
tokens which have contributed to the state, or not; the routines work
either way (this is the reason for the "committed" flag: so that they know
which tokens are new). (NA: The token numbers should never repeat in any case
(the same number should never be assigned to more than one token).)

typedef struct {
    token *tok;
    void *extdata;
    unsigned short xpos, ypos;
    } tokpos;

This encodes the position of a token on a page. The position of the token's
lower left corner is (xpos,ypos); the token that should be placed there is
pointed to by "tok". The token is not allowed to go outside the boundaries of
the page containing it. Any extension data associated with this position can
be found in extdata.

(NA: struct {
    token *tok;
    void *extdata;
    } token_extension;)

(NA: struct {
    tokpos *position;
    void *extdata;
    } tokpos_extension;)

(NA: These two structures are passed to a extension-building callbacks; see
below.)

The extension fields have been collapsed into the token and tokpos struct
above. 

Decoding routines:

The flow of control here is a bit simpler...

There are two things you can do:
- get a dictionary block, decode it, and add it to the current list of
  tokens
- get a position block and decode it wrt the current list of tokens

Dictionary block stuff:
- Read a dictionary block from the file.
- Feed it into tcmp_getTokenCount.
- If the result is zero, then something magic is happening - this case is
  currently undefined (it's a dictionary clearing code, and I haven't
  figured out what the control flow should be).
- Allocate a block of "token" of that size.
- Feed that block, plus the encoded dictionary block, into tcmp_decodeTokDict.
- This gives you a block of "token", including bitmaps. You can keep going,
  but it's not much use to you (you can't use it to decode position
  blocks). So you feed it into tcmp_estimateTokDictStateSize (along with your
  current token state block), allocate that memory, initialize it
  (tcmp_initializeTokDictState), and then make all the magic bits happy by
  feeding the "token" block and your initialised state block into
  tcmp_convertTokensToState, giving you a happy token state block.

Position block stuff is easy:
- Read a block from the file.
- Feed it into tcmp_getTokPosCount.
- Allocate a block of "tokpos" of that size.
- Feed that block, plus the encoded position block, plus the current token
  state block, into tcmp_decodeTokPositions.
- Done.

There's something else you might want to do: copy a dictionary state
block. This might be because you're reading a file and want to be able to
do random access, so you want to remember the dictionary state as of
various points in the file. You can do this by just duplicating the state
block (use memcpy or whatever), then feeding the old and new blocks into
tcmp_initializeTokDictState.

Routines:

size_t tcmp_estimateTokDictStateSize(void *oldstate, token *tokens, int ntok)

Given a dictionary state block ("oldstate"), which may be NULL, and a bunch
of tokens, this routine figures out how much dictionary state memory (in
bytes) is required to hold the new tokens *plus* the stuff in the old
state. This is what it returns.


void tcmp_initializeTokDictState(void *oldstate, void *newstatemem,
				 size_t memsize)

Given a dictionary state block ("oldstate"), some newly
allocated dictionary state memory, and the size of that memory block,
initialise the new state block, and copy over all the stuff from the old
state block. The old state block should then be discarded.

If oldstate==NULL, then it just initialises the new memory.

One way that this routine can be used: given an old state block, you call
tcmp_estimateTokDictStateSize as above, getting a new state block size. You
then call realloc on the old state block, extending it to the new size
(possibly via relocation). You then pass in the same pointer for "oldstate"
and "newstatemem"; the routine figures this out and knows what to do. In
cases where realloc didn't have to copy the memory (it was able to extend
it in place), this is more efficient.

You can also use this to clone a block: allocate the new block as the same
size as the old block, then pass in the old block as oldstate and the
newly-allocated memory as newstatemem.  Alternately, you can just copy the
old state block with a blind memory copy, then fix up the new clone by
passing it in as *both* oldstate and newstatemem.


size_t tcmp_estimateEncodedTokDictSize(void *dstate, token *tokens, int ntok)

This routine takes a dictionary state and a bunch of tokens and figures out
how much memory (in bytes) is required to encode those tokens. This is only
an estimate - see below.  It does not include the memory required to store
any dictionary extensions; the caller should estimate those requirements and
add that astimate to the value returned by this routine.


int tcmp_encodeTokDict(void *dstate, token *tokens, int ntok,
				 boolean (*extbuilder)(token **sorted_toks,
						       int ntok,
						       void *extcookie,
						       void **extension,
						       size_t *extlen),
				 void *extcookie,
				 void *outbuf, size_t outbuflen);


This routine takes a dictionary state, a bunch of tokens, and an output
buffer, and encodes the tokens into the buffer. It marks them all as
"committed", and also records various bits of information about them in the
dictionary state block. Any tokens already marked "committed" aren't
encoded.  If you aren't going to provide an extension, then just supply
NULL for extdata (of each token), extbuilder and extcookie.  If you are 
going to provide an extension, then read on...

The encoder can store extensions along with the dictionary.  It doesn't
specify what form the extensions take - that is up to the caller.  In many
cases, the extension data takes the form of some additional per-token
values: a second bitplane, an ASCII code, or whatever.  Since
tcmp_encodeTokDict sorts the tokens before writing them into the file, the
encoder must also sort the per-token data, so that they stay synchronised
and the decoder can re-associate them.

To generate an extension:
- convert each bit of per-token data to a void *
- set the extdata field of each token to point to this piece of data.
just like you'd expect!
(NA: - pass in an array of these void * values in extdata.  extdata[0] 
refers to the per-token extension data associated with the token in 
tokens[0] and so on.)
- pass in an extbuilder routine. The encoder sorts the tokens into an
_array_of_token_pointers. That is, in to "sorted_toks." This is the 
order that the tokens will eventually be written in (in the file) and
then later read out. "sorted_toks" will only contain tokens that have
NOT yet been committed. So you (as the extbuilder) don't have to worry
about which ones have been written and which ones haven't. See the note
about the cookie in the next paragraph.
(NA: - pass in an extbuilder routine.  The encoder sorts the tokens, and 
passes the extbuilder routine the sorted data: an array of token_extension
structures; each one contains a token * and a void * corresponding to the
per-token extension data associated with that token.  The order of the
tokens in sorted_extdata is the order that they will be stored in the file,
and the order in which tcmp_decodeTokDict will eventually read them out.)
- extbuilder is also passed the number of *new* tokens (the number of
tokens that were not marked committed initially), and a void * cookie,
which the caller passed to tcmp_encodeTokDict; this can provide extbuilder
with arbitrary additional context.
- extbuilder should allocate an extension block, and pass back its address
and length (in *extension and *extlen).
- extbuilder should return TRUE on success and FALSE on failure.
- you need to be careful if you're passing in tokens that have already been
encoded by a previous call to tcmp_encodeTokDict: all the tokens that were
marked "committed" are not represented in the "sorted_toks" array passed
to extbuilder.  The ntok value passed to extbuilder also does not represent
those tokens.
(NA: - If you pass in NULL for extdata, but non-NULL for extbuilder, then
extbuilder will be called, but the extdata fields in the token_extension
structures will all be NULL.)
- It's possible for tcmp_encodeTokDict to fail after the call to
extbuilder; extbuilder therefore shouldn't do anything that would make it
difficult to retry the encoding, rebuilding the same extension.
- tcmp_encodeTokDict does *not* free the extension block that extbuilder
returned to it.  If this needs to be freed, then the caller should do it;
the caller needs to find out where the block is, but extbuilder can pass
that back in *extcookie.
- if extbuilder decides that there is nothing to be put in the extension,
then it can pass back NULL and 0 for *extension and *extlen.
- N.B.: As described below, tcmp_encodeTokDict can fail due to lack of
buffer space, and then get called again with a larger buffer.  If this
happens, it will call extbuilder a second time on the second call;
extbuilder needs to be able to handle this.

tcmp_encodeTokDict returns:
- A non-negative number saying how many bytes of the output buffer were
actually used, or
- A negative number, saying how much *additional* memory it will need (i.e.
please extend the output buffer by -returnval bytes and try again).
- If extbuilder returns FALSE, then tcmp_encodeTokDict will return with
-1.  If you want to distinguish this from a "please retry" return, store
something in *extcookie.


size_t tcmp_estimateEncodedTokPosSize(void *dstate, tokpos *positions,
				      int npos)

This routine takes a dictionary state (which must have been properly filled
in by tcmp_encodeTokDict; one which has just been initialised by
tcmp_initializeTokDictState isn't enough), and a bunch of token
positions. It figures out an estimate of how much space is required to
encode those positions and returns that.


int tcmp_encodeTokPositions(void *dstate, tokpos *positions,
				      int npos,
				      int lowx, int lowy, int highx, int highy,
				      boolean (*extbuilder)(tokpos **sorted_tokpos,
							    int ntok,
							    void *extcookie,
							    void **extension,
							    size_t *extlen),
				      void *extcookie,
				      void *outbuf, size_t outbuflen);
				      
This routine takes a dictionary state, a bunch of token positions, and some
information about the page containing those tokens (the locations, in
pixels, of its corners), and encodes the positions into the given output
buffer. Its return values mean the same as those from tcmp_encodeTokDict.
lowx and lowy are inclusive; highx and highy are exclusive. To write out a
page that was WW by HH pixels, and whose top left corner was (0,0), you'd
say "tcmp_encodeTokPositions(..., 0, 0, WW, HH, ...)".

The extension mechanism works the same way as tcmp_encodeTokDict, but the
structures passed to extbuilder are tokpos, not token. Obviously...

Note: It is legal to pass in NULL for dstate, but only if npos==0.


int tcmp_getTokenCount(void *buffer, size_t buflen)

This routine takes an encoded dictionary block and returns the number of
tokens encoded in it. If the block contains a dictionary clearing code, it
returns zero. In this case, there is an (as-yet undefined) procedure to be
followed. It returns -1 in case of failure (this can be due to a malformed
buffer).

boolean tcmp_decodeTokDict(token *tokens, void *buffer, size_t buflen, 
				void *(*allocator)(size_t allocsize));
boolean tcmp_decodeTokDictMUsage( token *tokens, void *buffer, 
				size_t buflen, 
				void *(*allocator)(size_t allocsize), 
				long *tok_usage);		

This routine decodes the stuff in buffer, writing the results to *tokens
(which should have the correct number of entries). It allocates the memory
for the bitmaps using allocator; the only requirement it has of this
routine is that it give back memory that is at least 32-bit aligned. It
returns TRUE if it was successful, FALSE if it failed. 
(NA: The "toknum" fields are assigned values starting at firsttoknum.)

The MUsage version of the above function also accounts for the amount
of memory allocated in the token dictionary. tok_usage then counts the
number of bytes used:
 - in G4 encodings of single-use tokens (G4 Tokens)
 - in the bitmaps of other tokens
Note that this does not enclude the token structs themselves, nor any
extension data. You'll have to figure this out on your own.

As described above, for single-use tokens, this routine may (at its
discretion) return tokens whose bitmaps are G4-compressed.  These are
indicated by width>0 and longsperline==0.  In this case, the "bits" field
points at the start of the G4-compressed data.  The length of the data is
stored in the "uses" field.  This data has been allocated using allocator,
just like the regular bitmaps, and is the responsibility of the caller.
Note that the uses field's real value is *IMPLICITLY* zero - its actual
value is the data length.  This means that every token must be checked for
(width>0)&&(longsperline==0); those meeting this test have real "uses"
values of zero.  Don't just check "uses" to find single-use tokens - you
could miss some.

And this hack gets hairier: tcmp_convertTokensToState needs to know what
the real "uses" value is, in order to compute the Huffman codes.  If you
just pass in the unmodified token structure, tcmp_convertTokensToState will
figure out which tokens are G4-compressed, and get everything right.  If
you uncompress the token yourself, set longsperline to its real value AND
SET USES TO ZERO, tcmp_convertTokensToState will similarly get everything
right.  DONT'T change longsperline without changing uses also - this will
**seriously** confuse things.


boolean tcmp_convertTokensToState(token *tokens, int ntok, void *dstate)

This routine records various bits about the tokens in dstate and marks them
as "committed" (it ignores previously committed tokens). This is the same
procedure followed by tcmp_encodeTokDict, but without the actual encoding.
Also, tcmp_encodeTokDict re-orders the tokens to some canonical order prior
to writing them out; this routine does not do this (it assumes that the
tokens came out of an encoded dictionary block and so need no
re-ordering). As above, it returns TRUE on success, FALSE on failure.


size_t tcmp_estimateMergedTokDictStateSize(void **states, int nstates)

Given an array of void * pointers, each one referring to a dictionary state
block, this routine computes how much memory would be needed to hold their
union.

boolean tcmp_mergeTokDictStates(void *newstate, size_t newsize,
				void **states, int nstates)

Given an array of dictionary states, this computes a state that is their
union.  This is an *ordered* union: the tokens represented in states[0]
come before the tokens in states[1] and so on.  The old state blocks are
not modified.  The new state block need not be initialised in any way (you
can just call tcmp_estimateMergedTokDictStateSize then allocate a block
that big and hand it it tcmp_mergeTokDictStates).

In many cases, the position blocks in a file must be interpreted with
respect to the (ordered) union of a set of dictionary blocks.  It is
possible to read and merge those dictionaries one at a time, using
tcmp_decodeTokDict and tcmp_convertTokensToState for each block,
accumulating a single dictionary state.  Alternately, if you don't know up
front which subsets of the set of all dictionary state blocks are going to
be used to decode position blocks (you don't know ahead of time which
unions to compute), then you can just read in each dictionary separately,
and generate a dictionary state block for each; once you know which union
is needed, you can generate a merged dictionary state from the appropriate
state blocks.

All the toknum values in all the old states should be distinct; if two
tokens have the same token number, then it will be impossible to
distinguish between them.


int tcmp_getTokPosCount(void *buffer, int buflen)

This routine takes an encoded token position block and returns the number
of positions encoded in it. It returns -1 in case of failure (this can be
due to a malformed buffer).


boolean tcmp_decodeTokPositions(tokpos *positions,
			   int lowx, int lowy, int highx, int highy,
			   void *buffer, size_t buflen, void *dstate,

This routine decodes the stuff in buffer and writes the results to
*positions. It returns TRUE on success, FALSE on failure. The page
coordinates are as in tcmp_encodeTokPositions above; some higher layer of
the file needs to keep track of them. 

Note: it is legal to pass in NULL for dstate, but only if the number of
token positions stored in the buffer is zero (you can check this yourself
using tcmp_getTokPosCount).


void tcmp_setAllocator(void * (*newMalloc)(size_t), void (*newFree)(void *),
		       void * (*newCalloc)(size_t, size_t),
		       void * (*newRealloc)(void *, size_t))

This routine tells the token compressor/decompressor to use the given
routines instead of the standard versions. Note that tcmp_decodeTokDict
is also passed an allocator; that routine can be different from the
allocator given to this routine. This routine should only be called with
one set of values: if it is called more than once, the routines passed in
should be the same. If it is going to be called, it should be called before
the first call to any of the other tcmp_* functions.

Internally, the compressor/decompressor performs some allocations and
frees. It guarantees, however, that any memory allocated by any routine is
freed by the end of the routine, except for the bitmap memory allocated by
tcmp_decodeTokDict (that's what that routine has a separate allocator).
If you pass in NULL for calloc, it synthesises one.

Note for those programs using both these routines and the tokeniser
(libtok.a): the routines here *must* be the same ones as those passed into
tokInit.
