/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/zephdecode.h,v 1.8 2003/02/10 22:53:46 mkrocken Exp $ */

#ifndef ZEPHDECODE_H
#define ZEPHDECODE_H


#include "zephconst.h"
#include "resolution_t.h"
#include "orientation_t.h"

#include "uds.h"
#include "dgp_state.h"
#include "dgp_err_control.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * Maximum size of the decoder state. 
 */
#define ZEPH_MAX_DECODESTATE         8192


#define cErrMemBoundsTooLow          cDgpGen + 5


/* 
 * Prepare to decode an image. The decoder may allocate up to
 * ZEPH_MAX_DECODERSTATE bytes of storage for the "decodeState".
 * "decodeState" should reflect state for _this_image_only_.
 */
typedef ErrCode (*zephDecodeOpen)( Dgp_State *dgp_state,
                       zephImageInfo image, 
                       Dgp_Int32 srcW, Dgp_Int32 srcH,
                       DgpCoord_Resolution *srcRes, 
                       void **decodeState);

/*
 * Returns the amount of memory (in bytes) that is currently allocated by
 * the decoder for this image.
 */
typedef Dgp_UInt32 (*zephDecodeUsage)(void *decodeState);

/* 
 * Can the decoder decode a portion of the previously opened image with the
 * given type, orientation, and resolution into a buffer of the given size? 
 * If not, then the decoder should modify the appropriate arguments and set
 * the flags in "cantDo". Note that "type" and "orient" are bit-fields, so
 * the decoder should feel free to note all possible decoding options. There
 * are two other requirements of the decoder:
 *
 *   1) Simple Decoding. The decoder must be able to render the full image at 
 *      native image orientation and resolution in at least one output type.
 * 
 *   2) Convergence. Once the decoder has agreed to a particular output type
 *      (mono, RGB, etc.) for a particular image, all subsequent calls to
 *      the decoder for this image with this output type _must_not_ fail 
 *      because of that output type. Similarly, once the decoder has agreed
 *      to a render type and orientation, subsequent calls (with these two 
 *      values) must not fail because of these two parameters. The idea here
 *      is that the ordering defined by zephconst.h sets up some sort of
 *      converging behavior for this function so that once the decoder has
 *      agreed to the first 'n' parameters, it can't later fail because of
 *      one of these 'n' parameters. (Though it may certainly fail for only
 *      the type on the first call, only the orientation on the second, etc.)
 *      Also, the rectangle parameters must only be changed by the decoder to
 *      reflect an _increase_ in size. This condition is necessary to converge
 *      on a non-empty destination region.
 *
 * "maxMemRequired" does not include the size of the "decodeState" nor the 
 * output buffer. It should include the total size of all temporary buffers
 * allocated at any time during a call to "zephDecodeRead" with similar
 * parameters.
 * 
 * For an explanation of the coordinate variables and systems involved, see:
 * http:/*www3.cs.cornell.edu/Digipaper/coords.html */
 */
typedef ErrCode (*zephDecodeAbilities)( void *decodeState,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType *type,
                       DgpCoord_Orientation *orient,
                       DgpCoord_Resolution *dstRes,
                       Dgp_Int32 *dstX, Dgp_Int32 *dstY,
                       Dgp_Int32 *dstW, Dgp_Int32 *dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       Dgp_UInt8 *cantDo);

/*
 * Decode the image with parameters as previously agreed upon. If the image
 * cannot be decoded with memory constraints as delimited by "maxMemRequired"
 * then the decoder should free any temporary allocations (but not 
 * "decodeState"), set "maxMemRequired" to the new estimation of the total
 * required temporary buffer size, and return "cErrMemBoundsTooLow".
 *
 * After any call to "zephDecodeRead" all temporary buffers should be freed,
 * with the exception that any remaining buffers should be attached to 
 * "decodeState" and that the total size of "decodeState" (including these
 * buffers) should not exceed ZEPH_MAX_DECODESTATE.
 *
 * For packed monochrome reads, "dst" is guaranteed to fall on a word boundary,
 * and "dstBpl" will be a multiple of four.
 */
typedef ErrCode (*zephDecodeRead)( void *decodeState, 
                       void *dst, Dgp_Int32 dstBpl,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType type,
                       DgpCoord_Orientation orient,
                       DgpCoord_Resolution *dstRes,
                       Dgp_Int32 dstX, Dgp_Int32 dstY,
                       Dgp_Int32 dstW, Dgp_Int32 dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
					   void	*trans_src,
					   Dgp_Int32 trans_bpl,
                       Dgp_Int32 nominal_page_width, Dgp_Int32 nominal_page_height,
                       Dgp_Int32 nominal_strip_width, Dgp_Int32 nominal_strip_height);

/*
 * Free any memory associated with "decoderState".
 */
typedef ErrCode (*zephDecodeClose)( void *decodeState);
                       

#ifdef __cplusplus
}
#endif

#endif /* ZEPHDECODE_H */
