/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/zephimage.h,v 1.8 2003/02/10 22:53:46 mkrocken Exp $ */

#ifndef ZEPHIMAGE_H
#define ZEPHIMAGE_H

#include "zephpage_t.h"
#include "zephdictionary_t.h"
#include "zephimage_t.h"

#include "orientation_t.h"
#include "zephconst.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct _zephReadState * zephReadState;

extern ErrCode
zephOpenImage( zephPageInfo pageInfo, Dgp_UInt32 layerNum, 
               Dgp_UInt32 imageNum, zephImageInfo *pImageInfo);
extern ErrCode
zephCloseImage( zephImageInfo *pImageInfo);

extern Dgp_UInt32
zephGetImageMemoryUsage( zephImageInfo imageInfo);

/* 
 * (This contract is similar to the one between libzephyr and the image 
 * decoders below it. The major exception is for compressed reads, which are
 * documented below at ($).)
 *
 * Can the libzephyr decode a portion of the previously opened image with the
 * given type, orientation, and resolution into a buffer of the given size? 
 * If not, then the appropriate arguments will be modified and the flags in
 * "cantDo" will be set to indicate why not. Note that "type" and "orient" are
 * bit-fields, so libzephyr will note all possible decoding options. Callers
 * however, should set exactly one bit in each of these fields. There are
 * two other features that may be expected from this function:
 *
 *   1) Simple Decoding. libzephyr must be able to render the full image($) at
 *      native image orientation and resolution in at least one output type.
 * 
 *   2) Convergence. Once libzephyr has agreed to a particular output type
 *      (mono, RGB, etc.) for a particular image, all subsequent calls to
 *      libzephyr for this image with this output type _will_not_ fail 
 *      because of that output type. Similarly, once libzephyr has agreed
 *      to a render type and orientation, subsequent calls (with these two 
 *      values) will not fail because of these two parameters. The idea here
 *      is that the ordering defined by zephconst.h sets up some sort of
 *      converging behavior for this function so that once libzephyr has
 *      agreed to the first 'n' parameters, it can't later fail because of
 *      one of these 'n' parameters. (Though it may certainly fail for only
 *      the type on the first call, only the orientation on the second, etc.)
 *      Also, the rectangle parameters will only be changed by libzephyr to
 *      reflect an _increase_ in size (again ($) is the exception).
 *
 * ($) There is one exception here. Compressed reads (as specified by
 *     ZEPHRT_COMPRESSED) may not necessary succeed for the full image. This is
 *     the case only if the image is stripped. In this case, libzephyr
 *     will return parameters describing the region of the destination
 *     corresponding to the strip which contains the point (dstX, dstY). Note
 *     that this destination region may be smaller than the one originally
 *     specified by the caller, and so compressed reads of stripped images
 *     violates both conditions above.
 *
 * "maxMemRequired" does not include the size of the output buffer. It will 
 * include the total size of all temporary buffers allocated at any time during
 * and after a call to "zephReadImageState" and until "zephReadImageEnd" has
 * returned. Callers should specify a "reasonable" value here. Passing in zero
 * for this argument will almost certainly result in failure. Large numbers, 
 * however will give libzephyr (and the decoders below it) too much freedom.
 * Note that libzephyr is not required to make this value smaller if is it 
 * unnecessarily large; it will only complain if it is too small.
 *
 * In addition, (except where noted in ($) above), libzephyr abstracts the 
 * notion of an image strip so that callers to any functions in this module 
 * can ignore the "stripHeight" field if desired. This is not always a good 
 * idea, however. Though it may be convenient, this requires libzephyr to 
 * negotiate (with the image decoders below it) the intersection of each strip
 * with the desired region of the destination. In general, "stripHeight" 
 * should be used as a guide whenever possible.
 * 
 * For an  of the coordinate variables and systems involved, see:
 * http:/*www3.cs.cornell.edu/Digipaper/coords.html */
 */
extern ErrCode
zephReadImageAbilities( zephImageInfo imageInfo,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType *type,
                       DgpCoord_Orientation *orient,
                       DgpCoord_Resolution *dstRes,
                       Dgp_Int32 *dstX, Dgp_Int32 *dstY,
                       Dgp_Int32 *dstW, Dgp_Int32 *dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       Dgp_UInt8 *cantDo);

/* 
 * Start to read the given image with the specified parameters as agreed to
 * by the function above. This function returns asynchronously. Note that
 * the destination is in an undefined state until "zephReadImageEnd" is called.
 *
 * For packed monochrome reads, "dst" must fall on a word boundary, and 
 * "dstBpl" must be a multiple of four.
 */
extern ErrCode
zephReadImageStart( zephImageInfo imageInfo, 
                       void *dst, Dgp_Int32 dstBpl,
                       void *trans_src,
                       Dgp_Int32 trans_bpl,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType type,
                       DgpCoord_Orientation orient,
                       DgpCoord_Resolution *dstRes, 
                       Dgp_Int32 dstX, Dgp_Int32 dstY,
                       Dgp_Int32 dstW, Dgp_Int32 dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       zephReadState *pReadState);

/*
 * Finish reading an image. If the memory required to read the image exceeded
 * the amount estimated by "zephReadImageAbilities", then this function will
 * free all resources, return FIXME, and set "maxMemRequired" to the actual
 * amount required. The caller may then begin again by calling 
 * "zephReadImageStart" with similar parameters except for "maxMemRequired".

  Added nominal page width, height, so jpeg library could do coordinate rotations.
  21.6.2000 / dhtesq.
 */
extern ErrCode
zephReadImageEnd( zephImageInfo imageInfo, 
                       zephReadState *pReadState,
                       Dgp_UInt32 *maxMemRequired,
                       DgpCoord_Resolution *rres,
                       Dgp_Int32 nominal_page_width,
                       Dgp_Int32 nominal_page_height,
                       Dgp_Int32 nominal_strip_width,
                       Dgp_Int32 nominal_strip_height);

/*
 * 
 * Postconditions:
 *  - "pDictInfo" must be freed by calling zephCloseDictionary().
 */
extern ErrCode
zephGetDictionaryInfo( zephImageInfo imageInfo, 
                          zephDictionaryInfo *pDictInfo);

/*
 * Calculate the location and size of this image with respect to the page
 * coordinate system.
 */
extern ErrCode
zephCalcPagePosition( zephImageInfo imageInfo, Dgp_Int32 *pX, Dgp_Int32 *pY, 
                       Dgp_Int32 *pW, Dgp_Int32 *pH);

#ifdef __cplusplus
}
#endif


#endif /* ZEPHIMAGE_H */

