/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/zephtokenize.h,v 1.5 2000/12/29 18:23:29 mkrocken Exp $ */

#ifndef ZEPHTOKENIZE_H
#define ZEPHTOKENIZE_H

#include "zephdictionary_t.h"
#include "zephimage_t.h"

#include "token_t.h"
#include "symbol.h"
#include "extension_t.h"
#include "colortoken_t.h"
#include "color-extension.h"

#ifdef __cplusplus
extern "C"
{
#endif


/*
 * This functions reads more detailed information about the individual backend
 * symbol libraries  and their associated extensions. This function may be
 * considered "quick" in that it reads only the first few bytes of each symbol
 * library and doesn't pass through all the tokens of all of the libraries.
 *
 * Let N be the number dictionaries in "dictInfo".
 *
 * Preconditions:
 *  - "ntoks" points to an array of Dgp_Int32 of length N.
 *  - "backSizes" points to an array of Dgp_UInt32 of length N.
 *  - "nexts" points to an array of Dgp_Int32 of length N or is NULL.
 *  - "extSizes" points to an array of Dgp_Int32 of length N or is NULL.
 */
extern ErrCode
zephReadBinaryDictionaryDetails( zephDictionaryInfo dictInfo,
				Dgp_Int32 *ntoks, Dgp_UInt32 * backendSizes,
				Dgp_UInt32 * embDictSizes,
				Dgp_Int32 *nexts, Dgp_UInt32 * extSizes);

/*
 * This function reads in the entire backend dictionaries along with any 
 * extensions and asks the correct dictionary decoder to decode the tokens.
 * After this function is called all interal state associated with thte
 * backend dictionaries will be freed. All memory pointed to by the last
 * four arguments is owned by the caller. (This function mallocs memory.)
 *
 * Let N be the number dictionaries in "dictInfo".
 *
 * Preconditions:
 *  - "ntoks" points to an array of Dgp_Int32 of length N.
 *  - "toks" points to an array of (token *) of length N.
 *  - "nexts" points to an array of Dgp_Int32 of length N.
 *  - "exts" points to an array of (Extension *) of length N.
 */
extern ErrCode
zephReadBinaryDictionary ( zephImageInfo zephImage, Dgp_UInt32 dictId, 
			   jb2Segment **segment, Dgp_UInt32 *usage );

/* 
 * NOTE: Any of the four functions below will read in the compressed backend
 * dictionaries from the file and hold on to it until this dictionary is
 * closed. The size of these backend dictionaries is given by "backSizes" in 
 * zephReadDictionaryDetails above.
 */

/*
 * (See note above.)
 *
 * This function mallocs the necessary memory to hold the token structures
 * 
 */
extern ErrCode
zephReadBinaryDictionaryTokens( zephDictionaryInfo dictInfo,
                                Dgp_Int32 *ntoks, token **toks);

/*
 * (See note above.)
 *
 * After a call to "zephReadBinaryDictoinaryTokens", this function will
 * estimate the amount of memory required to hold the bitmaps for all
 * tokens in all libraries in this dictoinary.
 */
extern ErrCode
zephEstimateBinaryDictionaryBitmapSizes( zephDictionaryInfo dictInfo,
                                         Dgp_Int32 *ntoks, token **toks,
                                         Dgp_UInt32 *estimatedUsage);

/*
 * (See note above.)
 *
 * This function must only be called after "zephReadBinaryDictionaryTokens".
 * Also the arguments passed to this function must be identical to those 
 * passed to that function.
 */
extern ErrCode
zephReadBinaryDictionaryBitmaps( zephDictionaryInfo dictInfo,
                                 Dgp_Int32 *ntoks, token **toks);

/*
 * (See note above.)
 *
 */
extern ErrCode
zephReadBinaryDictionaryExtensions( zephDictionaryInfo dictInfo,
                                    Dgp_Int32 *nexts, Extension **exts);

/* 
 * Allocate a new "zephDictionaryState" structure and initialize it. 
 */
extern ErrCode
zephInitializeDictionaryState( zephDictionaryInfo dictInfo,
                          zephDictionaryState *pDictState,
                          Dgp_UInt32 *memUsage);
/* 
 * Note about "committed". 
 * "oldDictState" is optionally null.
 */
extern ErrCode
zephEstimateDecodedDictionaryStateSize( zephDictionaryState dictState,
                          zephDictionaryState oldDictState,
                          Dgp_UInt32* estimatedUsage);


/* Decode the state for the given set of tokens. 
 * If oldDictState is not NULL, then the new state will be the union of 
 * the old state and the state of the given set of tokens.
 * "memUsage" includes only memory allocated during this call. It does
 * NOT include memory allocated during "zephInitializeDictionaryState".
 * Note about "committed".
 */
extern ErrCode
zephDecodeBinaryDictionaryState( zephDictionaryState dictState, 
                          zephDictionaryState oldDictState,
                          Dgp_Int32 *ntoks, token **toks,
                          Dgp_UInt32 *memUsage);

extern void
zephFreeDictionaryState( zephDictionaryState *pDictState);

/*
 * Reads color dictionary info from the file.
 *
 * Preconditions:
 *  - "dictInfo" refers to one or more (N) color symbol libraries.
 *  - "type" indicates the desired output type, though this is
 *      not guaranteed, inspect the individual color token structs.
 *  - "ntoks" indicates the number of tokens in each of the N libraries.
 *  - "toks" points to a an array (of length N) of *color_token.
 */
extern ErrCode
zephReadColorDictionary( zephDictionaryInfo dictInfo,
                          zephReadType type, Dgp_Int32 *ntoks,
                          color_token **toks, Dgp_UInt32 *memUsage);


extern ErrCode
zephEstimatePositionBlockSize( zephImageInfo imageInfo,
                               Dgp_UInt32 *memUsage);

extern ErrCode
zephReadRawPositionBlock( zephImageInfo imageInfo, Dgp_UInt8 **data, Dgp_UInt32 *data_size ); 

extern ErrCode
zephReadPositionBlock( zephImageInfo imageInfo, DgpAdt_List seg_list,
                       jb2SymbolInstance **sym, Dgp_UInt32 *n_sym, 
                       Extension **exts, Dgp_UInt32 *memUsage);


extern ErrCode
zephReadBinaryAnnotations( zephImageInfo imageInfo, Dgp_UInt8** data, Dgp_UInt32 *size);


extern ErrCode
zephDecodeColorPositionExtension( zephImageInfo imageInfo,
                          Extension *ext, Dgp_Int32 *ntags, color_tag **tags, 
                          Dgp_UInt32 *memUsage);

#ifdef __cplusplus
}
#endif


#endif /* ZEPHTOKENIZE_H */
