#include "StdAfx.h"

#ifdef _WIN32
// Warning 4786 appears to be benign, and we get tons of them,
// so we disabled it.
#pragma warning(disable:4786)
#endif

#include "MrcFile.h"
#include "MrcIo.h"
#include "Jbig2Dict.h"

MrcFile::MrcFile ( const char* file )
: m_file( NULL ),
  m_pages(2),
  m_jbig2Dict(NULL),
  m_BGR(1)
{
    m_file = fopen( file, "rb" );

    if (m_file)
        m_pIO = new MrcIo ( m_file );
    else
        m_pIO = NULL;

    // set all pages to NULL
    for( std::vector<MrcPage*>::iterator page = m_pages.begin(); page != m_pages.end(); ++page )
    {
        *page = NULL;
    }
}

MrcFile::~MrcFile ( )
{
    purgeAllPages ();
    delete m_pIO;
	if(m_jbig2Dict) delete m_jbig2Dict;

    if (m_file) fclose( m_file );
}

TfxPage* MrcFile::getPage( Asd::UInt32 page )
{
    // check range
    if (page > numPages()) return NULL;

    // load if not already there
    if (m_pages.size() < page + 1 || m_pages[page] == NULL) loadPage (page);

    return m_pages[page];
}

void MrcFile::loadPage( Asd::UInt32 page )
{
    // check range
    if (page > numPages()) return;

    // check vector size
    if (m_pages.size () <= page + 1)
    {
        // keep the last non-NULL element
        TfxPage* last_page = *(m_pages.begin());

        for( std::vector<MrcPage*>::iterator page_it = m_pages.begin(); 
             page_it != m_pages.end(); ++page_it )
        {
            if (*page_it != NULL) last_page = *page_it;
        }

        // resize
        m_pages.resize (page + 1);

        // set all elements after the last non-NULL element to NULL
        for (std::vector<MrcPage*>::reverse_iterator page_rit = m_pages.rbegin(); 
             *page_rit != last_page; page_rit++)
        {
            *page_rit = NULL;
        }
    }

    if (m_pIO) loadData (page);

    return;
}

Asd::Point32 MrcFile::getPageSize (Asd::UInt32 page )
{
    Asd::Point32 errRet (-1, -1);

    // only attempt to load data if we have an IO object
    if (!m_pIO) return errRet;

    // how many layers are there
    Asd::UInt32 nLayers = m_pIO->pageNLayers( page );

    if (nLayers == 1) return m_pIO->layerSize (page, 1);
    if (nLayers >= 2) return m_pIO->layerSize (page, 2);

    return errRet;
}

void MrcFile::loadData( Asd::UInt32 page )
{

    // only attempt to load data if we have an IO object
    if (!m_pIO) return;

    Asd::UInt32 nLayers = m_pIO->pageNLayers( page );
    Asd::Point32 layerRes, relRes, pageRes = m_pIO->layerDPI ( page, 2 );

    // read all sel/fg layers
    std::vector<TfxSelLayer*> selLayers (nLayers / 2);
    for (Asd::UInt32 layer = 2; layer <= nLayers; layer+=2)
    {
        // find out properties of this selector layer
        CompressionMethod comp = m_pIO->layerCompression (page, layer);
        Asd::UInt32 nImages = m_pIO->layerNImages( page, layer );
        Asd::Point32 layerRes = m_pIO->layerDPI ( page, layer );

        // allocate a vector for the bands
        std::vector<TfxSelectorBand*> selBands (nImages);
        Asd::Int32 i;
		Asd::UInt8* data;
		Asd::UInt32 dataSize;
	    DictBufList::iterator where;
        for (Asd::UInt32 image = 0; image < nImages; image++)
        {
    		DictInfo dictInfo;
            Asd::Point32 origin = m_pIO->imageOrigin (page, layer, image);
            Asd::Point32 size = m_pIO->imageSize (page, layer, image);
         // create a selectorBand
            TfxSelectorBand* pBand = NULL;
            switch (comp)
            {
            case MMR:
				data = m_pIO->imageData (page, layer, image);
				dataSize = m_pIO->imageDataSize (page, layer, image);
                pBand = new TfxG4Band (origin, size, data, dataSize);
                break;
            case JBIG2:
				if(!m_jbig2Dict) m_jbig2Dict = new Jbig2Dict(this->numPages());
				//dictInfo = new DictInfo();
                m_pIO->dictInfo(page, layer, image, dictInfo);
                for(i = 0; i < dictInfo.getLibCount(); i++){
					Asd::UInt32* libIds = dictInfo.getLibIds();
                    where = (m_jbig2Dict->getDictBufList()).find(libIds[i]);
                    if( where == (m_jbig2Dict->getDictBufList()).end()){
                        (m_jbig2Dict->getDictBufList()).insert(DictBufList::value_type(libIds[i], m_pIO->dictData(libIds[i])));
                    }
                } 
	            data = m_pIO->jbig2Data (page, layer, image, dataSize);
				// jbig2Band communicate with MrcFile through jb2info and dictBufList; 
                pBand = new TfxJBig2Band (origin, size, data, dataSize, dictInfo);
                break;
            default:
                ;
            }
            // and add it to the bands vector
            selBands[image] = pBand;

        } /* loop over selector bands */
        
        // now read fg bands
        std::vector<TfxColorBand*> colBands = readColorBands (page, layer + 1);
        // copy the color bands into the selector band
        for (Asd::UInt32 nCol = 0; nCol < colBands.size(); nCol++)
        {
            TfxColorBand* col = colBands[nCol];
            Asd::UInt32 yPos = (col->getOrigin()).getY();
            // find the selector that lines up with this color image
            for (Asd::UInt32 nSel = 0; nSel < nImages; nSel++)
            {
                if ( ((selBands[nSel])->getOrigin()).getY() <= yPos &&
                     ((selBands[nSel])->getOrigin()).getY() + ((selBands[nSel])->getSize()).getY() > yPos)
                {
                    // add the color band here
                    (selBands[nSel])->addFgBand (col);
                    break;
                }
            }
        }

        // create a selector Layer from the selector bands
        layerRes = m_pIO->layerDPI ( page, layer + 1 );
        if (layerRes.getX() > 0)
            relRes = Asd::Point32(pageRes.getX() / layerRes.getX(), pageRes.getY() / layerRes.getY()); 
        else 
            relRes = layerRes;

        TfxSelLayer* pLayer = new TfxSelLayer (selBands, relRes);
		// set compression method
		pLayer->setCompressionMethod ( comp );
        
        // link the layer with  file
        pLayer->setFile(this);

		// set color spaces
		ColorSpace cSpace = m_pIO->layerColorSpace (page, layer + 1);
		pLayer->setFgColorSpace (cSpace);
        if (cSpace == ITULab)
        {
            // check for decode params
            pLayer->setDecParams (m_pIO->layerDecodeParam (page, layer + 1));
        }

		cSpace = m_pIO->layerColorSpace (page, layer);
		pLayer->setColorSpace (cSpace);
		
		// and add it to the layer vector
		selLayers[layer / 2 - 1] = pLayer;

    } /* loop over selector layers */

    // now read bg bands
    layerRes = m_pIO->layerDPI ( page, 1 );
    TfxBgLayer* bgLayer = NULL;

    // read layer if it exits
    if (layerRes.getX() > 0)
    { 
        // if there is no layer 2, get the resolution from this layer
        if (pageRes.getX() == -1)
        {
            pageRes = m_pIO->layerDPI ( page, 1 );
        }

        relRes = Asd::Point32(pageRes.getX() / layerRes.getX(), pageRes.getY() / layerRes.getY()); 
        // if there is no layer 2, then set relative resolution to 1
        if (pageRes.getX() == -1 || pageRes.getY() == -1) relRes = Asd::Point32 (1, 1);
        std::vector<TfxColorBand*> colBands = readColorBands ( page, 1 );
        if (colBands.size() > 0)
        {
            bgLayer = new TfxBgLayer (colBands, relRes);
    
            // link the layer with  file
            bgLayer->setFile(this);
            
            ColorSpace cSpace = m_pIO->layerColorSpace (page, 1);
            bgLayer->setColorSpace (cSpace);
            if (cSpace == ITULab)
            {
                // check for decode params
                bgLayer->setDecParams (m_pIO->layerDecodeParam (page, 1));
            }
        }
    }

    // create the page from bg and vector of selector Layers
    m_pages[page] = new MrcPage (this, bgLayer, selLayers, pageRes);

    return;
}

std::vector<TfxColorBand*> MrcFile::readColorBands ( Asd::UInt32 page, Asd::UInt32 layer )
{
    std::vector<TfxColorBand*> retVal(0);
    // we only deal with odd layers
    if (layer % 2 == 0) return retVal;

    // find out the number of images in this fg layer
    Asd::UInt32 nImages = m_pIO->layerNImages( page, layer );
    if (nImages == 0) return retVal;

    CompressionMethod comp = m_pIO->layerCompression (page, layer);
    // allocate a vector for the bands
    std::vector<TfxColorBand*> colBands (nImages);

    for (Asd::UInt32 image = 0; image < nImages; image++)
    {
        Asd::UInt32 dataSize;
        Asd::UInt8* data;
        switch(comp){
            case JBIG2:
                data = m_pIO->colorExtensionData(page, layer, image, dataSize);
                break;
            case JPEG:
                data = m_pIO->imageData (page, layer, image);
                dataSize = m_pIO->imageDataSize (page, layer, image);
            default:
                ;
        }
        Asd::Point32 origin = m_pIO->imageOrigin (page, layer, image);
        if (m_pIO->isStriped(page, layer))
        {
            // figure out the resolution relative to selector
            Asd::Point32 selRes, layerRes = m_pIO->layerDPI ( page, layer );
            if (layer == 1) 
            {
                selRes = m_pIO->layerDPI ( page, 2 );
            }
            else selRes = m_pIO->layerDPI ( page, layer - 1 );
            if (selRes.getX() > 0 && selRes.getY() > 0 &&
                layerRes.getX() > 0 && layerRes.getY() > 0)
            {
                origin.setX(origin.getX() * (selRes.getX() / layerRes.getX()));
                origin.setY(origin.getY() * (selRes.getY() / layerRes.getY()));
            }
        }
        Asd::Point32 size = m_pIO->imageSize (page, layer, image);
        
        // create a colorBand
        TfxColorBand* pBand = NULL;
        switch (comp)
        {
        case JPEG:
            pBand = new TfxJpegBand (origin, size, data, dataSize);
            break;
        case JBIG2:
            pBand = new TfxColorTokenBand (origin, size, data, dataSize);
            break;
        default:
            ;
        }
        // and add it to the bands vector
        colBands[image] = pBand;
    }

    return colBands;

}

void MrcFile::purgeAllPages ()
{
    // set all pages to NULL
    for( std::vector<MrcPage*>::iterator page = m_pages.begin(); page != m_pages.end(); ++page )
    {
        delete (*page);
        *page = NULL;
    }
}

Asd::UInt32 MrcFile::numPages( void ) 
{ 
    if (m_pIO) 
        return m_pIO->docNPages ();
    return 0; 
}
