#ifndef TfxBand_h_Included
#define TfxBand_h_Included

#include "AsdRect.h"
#include "AsdRgb.h"
#include "ImageAdapter.h"
#include <list>
#include "Jbig2Dict.h"
class TfxLayer;


/*
  TfxColorBand is used to represent the bands of color (odd) layers
  in a tiff-fx file.
*/

class TfxBand 
{
public:
	
    TfxBand (const Asd::Point32 & origin, const Asd::Point32 & size, 
        Asd::UInt8 * data, Asd::UInt32 dataSize) : 
	  m_data (data), m_dataSize (dataSize), m_origin (origin), m_size (size), m_layer(NULL), m_buffer(NULL) {};
	  virtual ~TfxBand () { delete [] m_data; };
	  
	  /* render into an existing image buffer */
	  virtual bool render( ImageAdapter * buf, const Asd::Rect & roi, const Asd::Point32 & scale ) = 0;
	  
	  /* origin of this band in PAGE coords */
	  const Asd::Point32 & getSize () const { return m_size; }
	  /* width and height of this band in LAYER coords */
	  const Asd::Point32 & getOrigin () const { return m_origin; }
	  
	  /* return the compressed data size */
	  Asd::UInt32 getMemoryUsage () const { return m_dataSize; }
	  
      TfxLayer* getLayer(){ return m_layer;}
      void setLayer(TfxLayer* layer){ m_layer = layer;}
	  
protected:
	
	/* compressed data pointer */
	Asd::UInt8 * m_data;
	/* compressed data size */
	Asd::UInt32 m_dataSize;
private:
	
	/* origin of this band in PAGE coords */
	Asd::Point32 m_origin;
	
	/* width and height of this band in LAYER coords */
	Asd::Point32 m_size;
    TfxLayer* m_layer;
    ImageAdapter* m_buffer;
};

class TfxColorBand : public TfxBand 
{
public:
	
	TfxColorBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : TfxBand (origin, size, data, dataSize) {}
	
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale ) = 0;
protected:
private:
};

class TfxEmptyBand : public TfxColorBand
{
public:
    
    TfxEmptyBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::RgbVal rgb) : 
	  TfxColorBand (origin, size, NULL, 0), m_rgb (rgb) {}
	  
	  /* Returns the color of this empty band. */
	  const Asd::RgbVal & getDefaultColor( void ) const { return m_rgb; };
	  
	  /* Well defined.  Just paints in default color. */
	  virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );
	  
protected:
private:
	Asd::RgbVal m_rgb;
};

class TfxUncompBand : public TfxColorBand
{
public:
	
	TfxUncompBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : TfxColorBand (origin, size, data, dataSize) {}
	
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );
	
protected:
private:
};

class TfxJpegBand : public TfxColorBand
{
public:
	
	TfxJpegBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : TfxColorBand (origin, size, data, dataSize) {}
	
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );
	
protected:
private:
};



/*

  TfxSelectorBand is used to represent the bands of selector
  (even) layers in a tiff-fx file.
  
*/
class TfxSelectorBand : public TfxBand 
{
public:
	
	TfxSelectorBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : TfxBand (origin, size, data, dataSize), m_fgBands () {}
    virtual ~TfxSelectorBand ();
	
    /* render the selector */
    virtual bool render ( ImageAdapter * buf, const Asd::Rect & roi, const Asd::Point32 & scale ) 
	{ return false;}
	
    void addFgBand (TfxColorBand* fg) { m_fgBands.push_back(fg); }
	
    std::list<TfxColorBand*>::iterator getFgBegin () { return m_fgBands.begin(); }
    std::list<TfxColorBand*>::iterator getFgEnd () { return m_fgBands.end(); }
	
    Asd::UInt32 nFgBands ( void ) { return m_fgBands.size(); }

protected:
    std::list<TfxColorBand*> m_fgBands;
private:
};

class TfxColorTokenBand : public TfxColorBand
{
public:
	
	TfxColorTokenBand (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : m_selBand(NULL), TfxColorBand (origin, size, data, dataSize) {}
	
    /* this object cannot render itself */
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );
    void setSelBand(TfxSelectorBand *band){ m_selBand = band;}
    TfxSelectorBand* getSelBand(){ return m_selBand;}

protected:
private:
    TfxSelectorBand* m_selBand;
};

class TfxG4Band : public TfxSelectorBand
{
public:
	
	TfxG4Band (const Asd::Point32 & origin, const Asd::Point32 & size, Asd::UInt8 * data, 
        Asd::UInt32 dataSize) : TfxSelectorBand (origin, size, data, dataSize) {};
	
    /* render the selector */
    virtual bool render ( ImageAdapter * buf, const Asd::Rect & roi, const Asd::Point32 & scale );
	
protected:
private:
};

class TfxJBig2Band : public TfxSelectorBand
{
public:
	TfxJBig2Band (  const Asd::Point32 & origin,
		                    const Asd::Point32 & size,
		                    Asd::UInt8 * data,
		                    Asd::UInt32 dataSize, 
		                    DictInfo &dictInfo) : m_dictInfo(dictInfo),m_jb2Info(NULL),
                            TfxSelectorBand (origin, size, data, dataSize) {};
    ~TfxJBig2Band();
	
    /* render the selector */
    virtual bool render ( ImageAdapter * buf, const Asd::Rect & roi, const Asd::Point32 & scale);
    bool decodeJBig2(bool keepAll, bool render, void** srcPixr);
    DictInfo & getDictInfo(){ return m_dictInfo;}	
    void* getJb2Info(){ return m_jb2Info;}

protected:
    bool decodeSymDict(Jbig2Dict* pDict, Asd::Int32 dictId, Asd::Int32 dictSize);

private:
    DictInfo m_dictInfo;
    void* m_jb2Info;
};


#endif
/*#ifndef TfxBand_h_Included */
