
#include "StdAfx.h"
#include "TfxLayer.h"
#include "DibImgBuf.h"
#include "AdapterFactory.h"

using namespace std;

Asd::Rect TfxLayer::getPageRoi(const Asd::Rect& roi, const Asd::Point32 scale){
    // bounding box of layer in scaled page res
    Asd::Rect scaledLayer = bBox (scale); 
    // intersection of bb with region
    Asd::Rect pageRoi = scaledLayer.intersection ( roi );
    return pageRoi;
}

Asd::Rect TfxLayer::getLayerRenderRoi(const Asd::Rect& roi, const Asd::Point32 pageScale){
    // Roi in unscaled layer coords
    Asd::Rect layerRoi = layerRegion(roi, pageScale);
    // get Layer scale in power of 2
    bool changed = false;
    Asd::Point32 layerScale =
        getLayerScalePower2(pageScale, changed);
    // convert the region to layer scale
    layerRoi.scaleDownMax (layerScale.getX(), layerScale.getY());
    return layerRoi;
}

Asd::Point32 TfxLayer::getLayerScalePower2(const Asd::Point32 scale, bool& scaleChanged){
    Asd::Rational <Asd::Int32> rscaleX(scale.getX(), m_resolution.getX());
    Asd::Rational <Asd::Int32> rscaleY(scale.getY(), m_resolution.getY());
    if( (Asd::Float32)rscaleX > 1 && (Asd::Float32)rscaleY > 1){
        Asd::Int32 xscale = rscaleX.toPower2();
        Asd::Int32 yscale = rscaleY.toPower2();
        if( (Asd::Float32)xscale == (Asd::Float32)rscaleX && 
            (Asd::Float32)yscale ==(Asd::Float32)rscaleY) {
            scaleChanged = false;
        }else scaleChanged = true;
        return   Asd::Point32(rscaleX.toPower2(), rscaleY.toPower2());
    }else{
        if((Asd::Float32)rscaleX == 1.0 && (Asd::Float32)rscaleY == 1.0)
            scaleChanged = false;
        else
            scaleChanged = true;
        return Asd::Point32(1,1);
    }
}

Asd::Rect TfxLayer::bBox (const Asd::Point32 pageScale)
{

    // get width and height in layer coords
    Asd::Rect boundingBox (m_size);

    boundingBox.scaleUp (m_resolution.getX(), m_resolution.getY());
    boundingBox.scaleDownMax (pageScale.getX(), pageScale.getY());

    // offset bounding box
    boundingBox.offsetXY (m_origin.getX() / pageScale.getX(), m_origin.getY() / pageScale.getY());

    return boundingBox;
}

Asd::Rect TfxLayer::layerRegion (const Asd::Rect pageRegion, const Asd::Point32 pageScale)
{
    Asd::Rect roi(pageRegion);
        
    // convert the region to layer scale
    roi.scaleUp(pageScale.getX(), pageScale.getY());
    roi.scaleDownMax (m_resolution.getX(), m_resolution.getY());

   // adjust the origin to layer coords
    Asd::Int32 offsetX = - m_origin.getX() / m_resolution.getX();
    Asd::Int32 offsetY = - m_origin.getY() / m_resolution.getY();
    roi.offsetXY (offsetX, offsetY);
    
    return roi;
}


TfxBgLayer::TfxBgLayer (const vector<TfxColorBand *>& bands, const Asd::Point32 & resolution) : 
TfxLayer (), 
m_bands(bands)
{
    // set the resolution of this layer (relative to PAGE resolution)
    setResolution (resolution);

    // find minX and maxX coordinate in this layer (PAGE coords)
    Asd::UInt32 minX = 0xFFFFFFFF, maxX = 0;
    vector<TfxColorBand*>::iterator band;
    for ( band = m_bands.begin();
         band != m_bands.end(); band++)
    {
        Asd::Point32 origin = (*band)->getOrigin();
        Asd::Point32 size = (*band)->getSize();

        // convert layer size to page coords
        size.setX (size.getX() * resolution.getX());
        size.setY (size.getY() * resolution.getY());

        if (origin.getX() < minX) minX = origin.getX();
        if (origin.getX() + size.getX() > maxX) 
            maxX = origin.getX() + size.getX();
    }

    // find minY coordinate in this layer ( PAGE coords )
    band = m_bands.begin();
    Asd::Point32 origin = (*band)->getOrigin();
    Asd::UInt32 minY = origin.getY();

    // find maxY coordinate in this layer ( PAGE coords )
    band = m_bands.end(); band--;
    origin = (*band)->getOrigin();
    Asd::Point32 size = (*band)->getSize();
    Asd::UInt32 maxY = origin.getY() + resolution.getY() * size.getY();

    // set Origin for this layer ( PAGE coords )
    setOrigin (Asd::Point32 (minX, minY));

    // set size for this layer ( LAYER coords )
    Asd::UInt32 width = (maxX - minX) / resolution.getX();
    Asd::UInt32 height = (maxY - minY) / resolution.getY();
    setSize (Asd::Point32 (width, height));

}


TfxBgLayer::~TfxBgLayer ()
{

    // delete all bands
    for (vector<TfxColorBand*>::iterator band = m_bands.begin(); 
         band != m_bands.end(); band++)
    {
        delete *band;
    }

}

bool TfxBgLayer::render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale )
{

    Asd::UInt32 scale_fac = scale.getX();

    AdapterFactory fac;

    bool status = true;
        
    // render all bands
    for (vector<TfxColorBand*>::iterator band = m_bands.begin(); 
         band != m_bands.end(); band++)
    {

        // get the Band coordinates in SCALED LAYER coords
        Asd::UInt32 bWidth = (*band)->getSize().getX();
        Asd::UInt32 bHeight = (*band)->getSize().getY();
        Asd::Rect rectBand (bWidth, bHeight);
        Asd::UInt32 bOriginY = ((*band)->getOrigin().getY() - getOrigin().getY()) / getResolution().getY();
        Asd::UInt32 bOriginX = ((*band)->getOrigin().getX() - getOrigin().getX()) / getResolution().getX();
        rectBand.offsetXY (bOriginX, bOriginY);
        rectBand.scaleDownMax (scale.getX(), scale.getY());

        // calculate the intersection of the current band with the ROI
        Asd::Rect roi = rectBand.intersection ( region );
        if (roi.isEmpty()) continue;

        // generate image buffer starting at roi.top
        Asd::Rect pageClipRoi (roi);
        pageClipRoi.offsetXY (-region.left(), -region.top());
        ImageAdapter *ibuf = fac.createSub (pageClipRoi, buf);

        // adjust the roi so that its top is aligned with the top of the band
        roi.offsetY (-rectBand.top());

        // link the band with the layer
        (*band)->setLayer(this);
        
        // render
        status = (*band)->render (ibuf, roi, scale);
        if (!status) return false;
    }

    return status;

}

TfxSelLayer::TfxSelLayer (vector<TfxSelectorBand *> & bands, const Asd::Point32 & fgResolution) : 
TfxLayer (), m_bands(bands)
{

    // set the resolution of this layer (relative to PAGE resolution)
    setFgResolution (fgResolution);

    // find minX and maxX coordinate in this layer (PAGE coords)
    Asd::UInt32 minX = 0xFFFFFFFF, maxX = 0;
    Asd::UInt32 fgMinX = 0xFFFFFFFF, fgMaxX = 0, fgMinY = 0xFFFFFFFF, fgMaxY;
    vector<TfxSelectorBand*>::iterator band = m_bands.begin(); 
    for (; band != m_bands.end(); band++)
    {
        Asd::Point32 origin = (*band)->getOrigin();
        Asd::Point32 size = (*band)->getSize();

        if (origin.getX() < minX) minX = origin.getX();
        if (origin.getX() + size.getX() > maxX) maxX = origin.getX() + size.getX();

        for (list<TfxColorBand*>::iterator fgBand = (*band)->getFgBegin(); 
             fgBand != (*band)->getFgEnd(); fgBand++)
        {
            Asd::Point32 fgOrigin = (*fgBand)->getOrigin();
            Asd::Point32 fgSize = (*fgBand)->getSize();

            // convert layer size to page coords
            fgSize.setX (fgSize.getX() * fgResolution.getX());
            fgSize.setY (fgSize.getY() * fgResolution.getY());

            if (fgOrigin.getX() < fgMinX) fgMinX = fgOrigin.getX();
            if (fgOrigin.getX() + fgSize.getX() > fgMaxX) fgMaxX = fgOrigin.getX() + fgSize.getX();
            fgMaxY = fgOrigin.getY() + fgSize.getY();
            if (fgOrigin.getY() < fgMinY) fgMinY = fgOrigin.getY();
        }
    }

    // find minY coordinate in this layer ( PAGE coords )
    band = m_bands.begin();
    Asd::Point32 origin = (*band)->getOrigin();
    Asd::UInt32 minY = origin.getY();

    // find maxY coordinate in this layer ( PAGE coords )
    band = m_bands.end(); band--;
    origin = (*band)->getOrigin();
    Asd::Point32 size = (*band)->getSize();
    Asd::UInt32 maxY = origin.getY() + size.getY();

    // set Origin for this layer ( PAGE coords )
    setOrigin (Asd::Point32 (minX, minY));

    // set size for this layer ( LAYER coords )
    Asd::UInt32 width = (maxX - minX);
    Asd::UInt32 height = (maxY - minY);
    setSize (Asd::Point32 (width, height));

    // set FG size and origin
    if (fgMinY == 0xFFFFFFFF)
    {
        // there was no fg band/layer
        setFgSize ( Asd::Point32 (0, 0) );
        setFgOrigin (Asd::Point32 (0, 0));
    }
    else
    {
        setFgOrigin (Asd::Point32 (fgMinX, fgMinY));
        width = (fgMaxX - fgMinX) / fgResolution.getX();
        height = (fgMaxY - fgMinY) / fgResolution.getY();
        setFgSize ( Asd::Point32 (width, height) );
    }

}

TfxSelLayer::~TfxSelLayer ()
{
    // delete all bands
    for (vector<TfxSelectorBand*>::iterator band = m_bands.begin(); 
         band != m_bands.end(); band++)
    {
        delete *band;
    }

}

bool TfxSelLayer::renderFg( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale )
{

    Asd::UInt32 scale_fac = scale.getX();

    AdapterFactory fac;

    bool status = true;
        
    // render all bands
    for (vector<TfxSelectorBand*>::iterator band = m_bands.begin(); 
         band != m_bands.end(); band++)
    {
        // link the band with the layer
        (*band)->setLayer(this);

        for (list<TfxColorBand*>::iterator fgBand = (*band)->getFgBegin(); 
             fgBand != (*band)->getFgEnd(); fgBand++)
        {

            // get the Band coordinates in SCALED LAYER coords
            Asd::UInt32 bWidth = (*fgBand)->getSize().getX();
            Asd::UInt32 bHeight = (*fgBand)->getSize().getY();
            Asd::Rect rectBand (bWidth, bHeight);
            Asd::UInt32 bOriginY = ((*fgBand)->getOrigin().getY() - getOrigin().getY()) / getFgResolution().getY();
            Asd::UInt32 bOriginX = ((*fgBand)->getOrigin().getX() - getOrigin().getX()) / getFgResolution().getX();
            rectBand.offsetXY (bOriginX, bOriginY);
            rectBand.scaleDownMax (scale.getX(), scale.getY());

            // calculate the intersection of the current band with the ROI
            Asd::Rect roi = rectBand.intersection ( region );
            if (roi.isEmpty()) continue;

            // generate image buffer starting at roi.top
            Asd::Rect pageClipRoi (roi);
            pageClipRoi.offsetXY (-region.left(), -region.top());
            ImageAdapter* ibuf = fac.createSub (pageClipRoi, buf);

            // adjust the roi so that its top is aligned with the top of the band
            roi.offsetY (-rectBand.top());

            // link the band with the layer
            (*fgBand)->setLayer(this);
        
            // to render color tokens we need to know about jbig2 band
            if(this->getCompressionMethod() == JBIG2){
                ((TfxColorTokenBand*)(*fgBand))->setSelBand(*band);
            }
            // render
            status = (*fgBand)->render (ibuf, roi, scale);
            if (!status) return false;
        }   
    }

    return status;

}


bool TfxSelLayer::renderSel( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale )
{

    AdapterFactory fac;

    Asd::UInt32 scale_fac = scale.getX();

    bool status = true;
        
    // render all bands
    for (vector<TfxSelectorBand*>::iterator selBand = m_bands.begin(); 
         selBand != m_bands.end(); selBand++)
    {
        // get the Band coordinates in SCALED PAGE coords
        Asd::UInt32 bWidth = (*selBand)->getSize().getX();
        Asd::UInt32 bHeight = (*selBand)->getSize().getY();
        Asd::Rect rectBand (bWidth, bHeight);
        Asd::UInt32 bOriginY = (*selBand)->getOrigin().getY();
        Asd::UInt32 bOriginX = (*selBand)->getOrigin().getX();
        rectBand.offsetXY (bOriginX, bOriginY);
        rectBand.scaleDownMax (scale.getX(), scale.getY());

        // calculate the intersection of the current band with the ROI
        Asd::Rect roi = rectBand.intersection ( region );
        if (roi.isEmpty()) continue;

        // generate image buffer starting at roi.top
        Asd::Rect pageClipRoi (roi);
        pageClipRoi.offsetXY (-region.left(), -region.top());
        ImageAdapter *ibuf = fac.createSub (pageClipRoi, buf);

        // adjust the roi so that its top is aligned with the top of the band
        roi.offsetY (-rectBand.top());

        // link the band with the layer
        (*selBand)->setLayer(this);

        // render
        status = (*selBand)->render (ibuf, roi, scale);

        if (!status) return false;
    }
    return status;
}


bool TfxSelLayer::render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale )
{

    AdapterFactory fac;

    // get the clip region for the page buffer
    Asd::Rect pageClipRoi;

    // two temporary image buffers for foreground and selector
    DibImgBuf dibFg, dibSel;
    bool status = true;
    bool fgUpscale = false;
        
    // figure out the size of the fg bands
    Asd::Point32 fgScale (1, 1);
    if (getFgResolution().getX() <= scale.getX())
    {
        fgScale.setX(scale.getX() / getFgResolution().getX());
        fgScale.setY(scale.getY() / getFgResolution().getY());
    }
    if (getFgResolution().getX() > scale.getX())
        fgUpscale = true;

    // render all bands
    for (vector<TfxSelectorBand*>::iterator band = m_bands.begin(); 
         band != m_bands.end(); band++)
    {
        // get the Band coordinates in SCALED PAGE coords
        Asd::UInt32 bWidth = (*band)->getSize().getX();
        Asd::UInt32 bHeight = (*band)->getSize().getY();
        Asd::Rect rectBand (bWidth, bHeight);
        Asd::UInt32 bOriginY = (*band)->getOrigin().getY();
        Asd::UInt32 bOriginX = (*band)->getOrigin().getX();
        rectBand.offsetXY (bOriginX, bOriginY);
        rectBand.scaleDownMax (scale.getX(), scale.getY());
        
        // calculate the intersection of the current band with the ROI
        Asd::Rect roi = rectBand.intersection ( region );
        if (roi.isEmpty()) continue;

        // fix page clip region
        pageClipRoi = roi;
        pageClipRoi.offsetXY (-region.left(), -region.top());

        // allocate a temporary image buffer and set it to black
        ImageAdapter *fgBuf = fac.createInstance (COLOR_24_BIT, roi, &dibFg);
        memset ((void*) dibFg.data(), 0x00, dibFg.dataSize());
        
        // link the band with the layer
        (*band)->setLayer(this);

        // render all foreground bands into a temporary buffer
        for (list<TfxColorBand*>::iterator fgBand = (*band)->getFgBegin(); 
             fgBand != (*band)->getFgEnd(); fgBand++)
        {
            // to render color tokens we need to know about jbig2 band
            if(this->getCompressionMethod() == JBIG2){
                ((TfxColorTokenBand*)(*fgBand))->setSelBand(*band);
            }
            // get the Band coordinates in unscaled page coords
            Asd::UInt32 bFgWidth = (*fgBand)->getSize().getX() * getFgResolution().getX();
            Asd::UInt32 bFgHeight = (*fgBand)->getSize().getY() * getFgResolution().getY();
            Asd::Rect rectFgBand (bFgWidth, bFgHeight);
            Asd::UInt32 bFgOriginY = (*fgBand)->getOrigin().getY();
            Asd::UInt32 bFgOriginX = (*fgBand)->getOrigin().getX();
            rectFgBand.offsetXY (bFgOriginX, bFgOriginY);

            // calculate the intersection of the current band with the ROI
            rectFgBand.scaleDownMax (scale.getX(), scale.getY());
            Asd::Rect fgRoi = rectFgBand.intersection ( roi );
            if (fgRoi.isEmpty()) continue;

            // adjust the roi so that its top is lining up with the selector band top
            fgRoi.offsetY (-rectBand.top());

            // link the band with the layer
            (*fgBand)->setLayer(this);
            // render
            if (fgUpscale)
            {
                // get a region for this layer
                Asd::Rect fgLayerRoi (fgRoi);
                fgLayerRoi.scaleDownMax (getFgResolution().getX() / scale.getX(),
                    getFgResolution().getY() / scale.getY());

                // need to decode into temporary buffer
                DibImgBuf tempDibBuf;
                ImageAdapter *tempFgBuf = fac.createInstance (COLOR_24_BIT, fgLayerRoi, &tempDibBuf);
                
                status = (*fgBand)->render(tempFgBuf, fgLayerRoi, Asd::Point32(1, 1));

                fgScale.setX(getFgResolution().getX() / scale.getX());
                fgScale.setY(getFgResolution().getY() / scale.getY());

               tempFgBuf->scaleUp (fgScale, fgBuf);

            }else
                status = (*fgBand)->render (fgBuf, fgRoi, fgScale);
            if (!status) return false;
        }

        // generate image buffer starting at roi.top
        ImageAdapter *ibuf = fac.createSub (pageClipRoi, buf);
 
        // we need to do an alpha blend between bg and fg
        if (scale.getX() > 1)
        {
            // need to down sample selector to gray
            ImageAdapter *graySel = fac.createInstance (COLOR_8_BIT, roi, &dibSel);

            Asd::Rect bandRoi = roi;
            // adjust the bandRoi so that its top is 0
            bandRoi.offsetY (-rectBand.top());
    
            // render and flip bytes to Little_Endian
            status = (*band)->render (graySel, bandRoi, scale);
            if (!status) return false;  

            ibuf->alphaBlend (graySel, fgBuf);

        }
        else // need to use the selector as a roi map on the existing buffer
        {
            // need to decode at full resolution
            ImageAdapter *monoSel = fac.createInstance (MONOCHROME, roi, &dibSel);

            Asd::Rect bandRoi = roi;
            // adjust the bandRoi so that its top is 0
            bandRoi.offsetY (-rectBand.top());
    
            // render and flip bytes to Little_Endian
            status = (*band)->render (monoSel, bandRoi, scale);
            if (!status) return false;  

            // do the masked copy
            ibuf->maskedCopy (monoSel, fgBuf);

        }
    } // loop over selector bands

    return status;

}

