#ifndef TfxLayer_h_Included
#define TfxLayer_h_Included

#include "TfxBand.h"

#include "MiscConst.h"
#include "AsdRect.h"
#include "AsdRational.h"
#include "ImageAdapter.h"

#include <vector>

class MrcFile;

class TfxLayer
{
public:

    /* generate an empty layer */
    TfxLayer () : m_origin (Asd::Point32()), m_size (Asd::Point32()), 
        m_comp(UnknownCompression), m_colorSpace(Default), m_resolution (1,1), m_file(NULL) {}
    virtual ~TfxLayer () {};

    /* public access methods */
    Asd::Point32 getSize( void ) const { return m_size; }
    Asd::Point32 getOrigin( void ) const { return m_origin; }
    void setCompressionMethod ( const CompressionMethod & comp) { m_comp = comp; return; }
    CompressionMethod getCompressionMethod( void ) const { return m_comp; }
    void setColorSpace ( const ColorSpace & color) { m_colorSpace = color; return; }
    ColorSpace getColorSpace ( void ) const { return m_colorSpace; }
    Asd::Point32 getResolution( void ) const { return m_resolution; }
    void setFile(MrcFile* file) { m_file = file;}
    MrcFile* getFile(){ return m_file;}
    /* rectangle of the outline of this layer in a page at a given resolution */
    Asd::Rect bBox (const Asd::Point32 pageScale);

    /* convert a page region into a region in this layer */
    Asd::Rect layerRegion (const Asd::Rect pageRegion, const Asd::Point32 pageScale);

    /* rendering interface */
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale ) = 0;

    /*get intersection of roi with layer region in scaled page res */
    Asd::Rect getPageRoi(const Asd::Rect& roi, const Asd::Point32 scale);
    /*get intersection of roi with layer region in actual scaled layer(power of 2) res */
    Asd::Rect getLayerRenderRoi(const Asd::Rect& roi, const Asd::Point32 scale);
    /*get actual layer scale(power of 2) */
    Asd::Point32 getLayerScalePower2(const Asd::Point32 scale, bool& scaleChanged);

protected:

    void setResolution( const Asd::Point32 & resolution ) { m_resolution = resolution; return; }
    void setSize( const Asd::Point32 & size ) { m_size = size; return; }
    void setOrigin( const Asd::Point32 & origin ) { m_origin = origin; return; }

private:

	/* origin of this band in PAGE coords */
	Asd::Point32 m_origin;

	/* width and height of this band in pixels */
	Asd::Point32 m_size;

    /* compression used in this layer */
    CompressionMethod m_comp;

    /* color space used in this layer */
    ColorSpace m_colorSpace;

    /* resolution of this layer relative to the page */
    Asd::Point32 m_resolution;
    
    /* A reference to TfxPage */
    MrcFile* m_file;
};


/*************************************************************************

    A TfxBgLayer is rather simple.  It is just a collection of color
    bands.

*************************************************************************/

class TfxBgLayer : public TfxLayer
{
public:

    TfxBgLayer () : TfxLayer (), m_bands(0) {}
    TfxBgLayer (const std::vector<TfxColorBand *> & bands, const Asd::Point32 & resolution);
    ~TfxBgLayer ();

    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );

    void setDecParams (const std::vector<Asd::Float32> & decParam) { m_decParam = decParam; }
    Asd::UInt32 numDecParams () { return m_decParam.size(); }
    Asd::Float32 getDecParam (Asd::UInt32 n) { if (m_decParam.size() > n) return m_decParam[n];
                                                return 0.; };

protected:

    /* return the number of bands */
    Asd::UInt32 getNBands () const { return m_bands.size(); }
    /* get a band by index */
    TfxColorBand* getBand (Asd::UInt32 index) { return m_bands[index]; }
    /* get a band by its y coordinate */
    TfxColorBand* getBand (Asd::Point32 origin);

private:

    /* All bands must be of the same type and cannot ever be TfxColorTokenBand. */
    std::vector<TfxColorBand *> m_bands;

    /* vector to hold decode parameters */
    std::vector<Asd::Float32> m_decParam;

};

/*************************************************************************

 A TfxSelLayer consists of a group of selector bands, which contain any 
 optional foreground bands

*************************************************************************/

class TfxSelLayer : public TfxLayer
{
public:

    TfxSelLayer () : TfxLayer(), m_bands(0) {}
    TfxSelLayer (std::vector<TfxSelectorBand *> & bands, const Asd::Point32 & fgResolution);
    ~TfxSelLayer ();
   
    /* render the foreground through the selector */
    virtual bool render( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );

    /* render the selector only */
    bool renderSel ( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );
    
    /* This method generally just calls render on the color band associated */
    /* with it.  However, if that color layer cannot be rendered because */
    /* it is JBIG2 color tokens then the selector and color tokens are used */
    /* to render only the tokens.  Pixels not touched by tokens are left */
    /* unchanged. */
    bool renderFg ( ImageAdapter * buf, const Asd::Rect & region, const Asd::Point32 & scale );

    /* get fg state variables */
    Asd::Point32 getFgSize( void ) const { return m_fgSize; }
    Asd::Point32 getFgOrigin( void ) const { return m_fgOrigin; }
    void setFgCompressionMethod ( const CompressionMethod & comp) { m_fgComp = comp; return; }
    CompressionMethod getFgCompressionMethod( void ) const { return m_fgComp; }
    void setFgColorSpace ( const ColorSpace & color) { m_fgColorSpace = color; return; }
    ColorSpace getFgColorSpace ( void ) const { return m_fgColorSpace; }
    Asd::Point32 getFgResolution ( void ) const { return m_fgResolution; }

    void setDecParams (const std::vector<Asd::Float32> & decParam) { m_decParam = decParam; }
    Asd::UInt32 numDecParams () { return m_decParam.size(); }
    Asd::Float32 getDecParam (Asd::UInt32 n) { if (m_decParam.size() > n) return m_decParam[n];
                                                return 0.; };

protected:

    /* get a band by index */
    TfxSelectorBand* getBand (Asd::UInt32 index) { return m_bands[index]; }
    /* get a band by its y coordinate */
    TfxSelectorBand* getBand (Asd::Point32 origin);

    /* set fg state variables */
    void setFgResolution( const Asd::Point32 & resolution ) { m_fgResolution = resolution; return; }
    void setFgSize( const Asd::Point32 & size ) { m_fgSize = size; return; }
    void setFgOrigin( const Asd::Point32 & origin ) { m_fgOrigin = origin; return; }
 
private:

    /* Note that we may need to use the TfxEmptyBand class here, although */
    /* NULL pointers may also work. */
    std::vector<TfxSelectorBand*> m_bands;

    /* vector to hold decode parameters */
    std::vector<Asd::Float32> m_decParam;

    /* fg state variables */
	Asd::Point32 m_fgOrigin;
	Asd::Point32 m_fgSize;
	Asd::Point32 m_fgResolution;
    CompressionMethod m_fgComp;
    ColorSpace m_fgColorSpace;

};

#endif
/*#ifndef TfxLayer_h_Included */
