#ifndef TfxPage_h_Included
#define TfxPage_h_Included

#include "TfxLayer.h"
#include "AdapterFactory.h"

#include "MiscConst.h"
#include "AsdRational.h"

#include <vector>


/*************************************************************************

    A TfxPage is typically created as a result of calling loadPage()
    on TfxFile.  It represents a page from TfxFile.  If the page
    is compressed, TfxPage will load the compressed page into memory.
    The page is only decompressed when render is called.  (Of course
    if the page was not compressed to begin with then the whole
    page will be loaded.)

*************************************************************************/

/************************************************************************

          C O O R D I N A T E S

  each object (Band, Layer, Page) has a render method.
  
  It is assumed that the imageBuffer into which the object is rendered has 
  its origin at the origin of the selected region.

  The region is in coordinates of the scaled object. This means that the object
  is first scaled by the amound specified in scale, and then the region of intrest
  is cut out. 

***************************************************************************/


class TfxFile;

class TfxPage
{
public:

    TfxPage( void ): m_fileOffset(NULL), m_file(NULL), m_res(0,0){}

    TfxPage( TfxFile * file, Asd::Int32 offset )
    : m_fileOffset(offset), m_file(file), m_res(0,0) {}

    /*/////////////////////////////////////////////////////////////////// */
    /* These attribute fetching methods refer to the recomposed page, */
    /* not necessarilly any of its components or bands. */

    /* "Resolution" is defined as the absolute number of pixels in each */
    /* direction.  For 8.5/11 at 600 dpi, resolution is 5100x6600. */
    virtual Asd::Point32 getSize( void ) const = 0;

    virtual ColorDepth getColorDepth( void ) { return COLOR_DEPTH_ERROR; }
   
    /* */
    /* */
    /*/////////////////////////////////////////////////////////////////// */

    /* the scale is negative for upsampling, e.g. scale.getX() == -4 means that  */
    /* the resulting image is 4x upsampled */
    /* renderBW outputs a halftoned BW image for printing */
    virtual bool render( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer = 0) = 0;
    virtual bool renderBW( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer = 0) = 0;

    virtual Asd::UInt32 getNumLayers( void ) const { return 0; }

    /* This method returns the amount of memory in bytes used by the */
    /* page.  Note that this is typically compressed, even for plain */
    /* tiff. */
    virtual Asd::UInt32 memoryUsage( void ) const { return 0;}

    virtual Asd::Int32 getXRes() { return m_res.getX();}
    virtual Asd::Int32 getYRes() { return m_res.getY();}
    virtual void setRes(Asd::Point32 &pageRes) { m_res = pageRes;}
protected:

    
    /* This is the pages location in the file.  If you are at the beginning */
    /* the file, you can fseek( stream, m_fileOffset, origin ) to get to  */
    /* the beginning of this page's IFD. */
    Asd::Int32 m_fileOffset;
    TfxFile * m_file;
    Asd::Point32 m_res;

private:

};


/*************************************************************************

 A typical page is actually a collection of MRC layers.  MRC page represents
 one such page.

 The implimentation of MrcPage is where most of the code that recombines
 layers for rendering is.

*************************************************************************/

class MrcPage : public TfxPage
{
public:

    MrcPage ( TfxFile * file, Asd::Int32 offset ) : TfxPage ( file, offset ), m_bg(NULL), m_sel(0) {} 
    MrcPage ( TfxFile * file, TfxBgLayer * bg, std::vector<TfxSelLayer *> sel, Asd::Point32 &pageRes) : 
    TfxPage ( file, -1 ), m_bg(bg), m_sel(sel){ setRes(pageRes); }
    virtual ~MrcPage ();

    virtual Asd::Point32 getSize( void ) const 
    { 
        if (m_sel.size()) return (m_sel[0])->getSize();
        if (m_bg) return m_bg->getSize();
        return Asd::Point32 (-1, -1);
    }
    virtual ColorDepth getColorDepth( void );

    virtual bool render( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer = 0);
    virtual bool renderBW( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer = 0);

    bool render( ImageAdapter * adpter, const Asd::Rect & region, const Asd::Point32 & scale );
	bool renderToFile(ImageBuf * buf, const Asd::Rect & region, Asd::Rational<Asd::Int32> zoom,
        const char * filename, const char * outtype, const Asd::Int32 & xRes, const Asd::Int32 & yRes);

	virtual Asd::UInt32 getNumLayers( void ) const ;

    Asd::UInt32 getNumFgs( void ) const { return m_sel.size(); }

    ColorSpace getColorSpace ( void ) ;

    /* May return NULL if there is only the default background. */
    TfxBgLayer * getBg( void ) { return m_bg; }
    
    /* fgIndex starts with 0.  Note that these "layers" are not the */
    /* "layers" from getNumLayers(), they are the layers from getNumFgs(). */
    /* A TfxFgLayer() is actually a selector-foreground pair. */
    TfxSelLayer * getSel ( Asd::UInt32 selIndex ) 
    { 
        if (selIndex < m_sel.size()) return m_sel[selIndex];
        return NULL;
    }


protected:
    bool renderInternal( ImageAdapter * outbuf, const Asd::Rect & region, 
                         Asd::Rational<Asd::Int32> zoom, Asd::Int32 layer);

private:

    /* This is not necessarilly a background.  If not, m_bg is NULL. */
    TfxBgLayer * m_bg;

    /* The first element (0) defines the image and is always present. */
    std::vector<TfxSelLayer *> m_sel;
};


#endif
/*#ifndef TfxPage_h_Included */
