#include "StdAfx.h"

// INTEL JPEG library
namespace intel
{
#include "ijl.h"
}

#include "TfxLayer.h"
#include "TfxBand.h"
#include "IntelImgBuf.h"

bool TfxJpegBand::render ( ImageAdapter * buf, const Asd::Rect & region, 
						  const Asd::Point32 & scale )
{

	// this implementation is for Intel Image Buffers only
	IntelImgBuf* ibuf = (IntelImgBuf*) buf;
	if (!ibuf) return false;

    ColorSpace colorSpace = getLayer()->getColorSpace();
    if (colorSpace == WhiteIsZero || colorSpace == BlackIsZero) 
    {
        TfxLayer *layer = getLayer();
        TfxSelLayer *sel = dynamic_cast<TfxSelLayer*> (layer);
        if (sel) colorSpace = sel->getFgColorSpace();
    }

    Asd::UInt32 width = ibuf->width();
    Asd::UInt32 height = ibuf->height();

    // check what we need to do
    bool needUpScale = false;
    bool needDownScale = false;
    bool needClip = false;
    if (scale.getX() < -1 ||  scale.getY() < -1) 
    {
        needUpScale = true;
    }
    if (scale.getX() > 1 ||  scale.getY() > 1) 
    {
        needDownScale = true;
    }
    if (region.width() < getSize().getX() / scale.getX() || 
        region.height() < getSize().getY() / scale.getY() )
    {
        needClip = true;
    }

	// Allocate the IJL JPEG_CORE_PROPERTIES structure.
	bool status = true;
	intel::IJLERR jerr;
	intel::JPEG_CORE_PROPERTIES jcprops;

	__try
	{
		// Initialize the Intel(R) JPEG Library.
		jerr = intel::ijlInit(&jcprops);
		if(intel::IJL_OK != jerr)
		{
			status = false;
			__leave;
		}
		// Get information on the JPEG image
		// (i.e., width, height, and channels).
		jcprops.JPGFile = NULL;
		jcprops.JPGBytes = m_data;
		jcprops.JPGSizeBytes = m_dataSize;
		jerr = intel::ijlRead(&jcprops, intel::IJL_JBUFF_READPARAMS);
		if(intel::IJL_OK != jerr)
		{
			status = false;
			__leave;
		}

		// Set up the info on the desired DIB properties.
		jcprops.DIBWidth = width;
		jcprops.DIBHeight = height;
		jcprops.DIBPadBytes = ibuf->bpl() - width*3;
        jcprops.DIBBytes = (Asd::UInt8 *) ibuf->data();
		jcprops.DIBChannels = 3;

		// Set the JPG color space ... this will always be
		// somewhat of an educated guess at best because JPEG
		// is "color blind" (i.e., nothing in the bit stream
		// tells you what color space the data was encoded from).
		// However, in this example we assume that we are
		// reading JFIF files which means that 3 channel images
		// are in the YCbCr color space and 1 channel images are
		// in the Y color space.

		switch(colorSpace)
		{
		case Gray:
			{
				jcprops.JPGColor = intel::IJL_G;
				jcprops.DIBColor = intel::IJL_BGR;
				break;
			}
		case YCbCr:
			{
				jcprops.JPGColor = intel::IJL_YCBCR;
				jcprops.DIBColor = intel::IJL_BGR;
				break;
			}
		default:
			{
				// This catches everything else, but no
				// color twist will be performed by the IJL.
				jcprops.JPGColor = intel::IJL_OTHER;
				jcprops.DIBColor = intel::IJL_OTHER;
				break;
			}
		}

        // was any region of intrest specified ?
        if (needClip)
        {

			// set the roi WATCH FOR OUR RECT PROPERTIES !!!!
			intel::IJL_RECT local_roi;
			local_roi.left = region.left();
			local_roi.top = region.top();
			local_roi.right = local_roi.left + region.width();
			local_roi.bottom = local_roi.top + region.height();
			jcprops.jprops.roi = local_roi;

        	if(region.width() > width || region.height() > height)
	       	{
		       	status = false;
		        __leave;
	        }
        }

        // was scaling requested ?
		if (needDownScale)
		{
			Asd::UInt32 scale_fac = scale.getX();
            if (scale.getY() > scale_fac) scale_fac = scale.getY();

			// Now actually get the JPEG image data into the pixel buffer.
			switch (scale_fac)
			{
			case 2:
				jerr = intel::ijlRead(&jcprops, intel::IJL_JBUFF_READONEHALF);
				break;
			case 4:
				jerr = intel::ijlRead(&jcprops, intel::IJL_JBUFF_READONEQUARTER);
				break;
			case 8:
				jerr = intel::ijlRead(&jcprops, intel::IJL_JBUFF_READONEEIGHTH);
				break;
			default:
				jerr = intel::IJL_UNSUPPORTED_SUBSAMPLING;
		    	status = false;
			    __leave;
			}
		}
		else
        {
            // decode the full buffer at nominal resolution
			jerr = intel::ijlRead(&jcprops, intel::IJL_JBUFF_READENTROPY);
		}
		
        if(intel::IJL_OK != jerr && intel::IJL_ROI_OK != jerr)
		{
			status = false;
			__leave;
		}
	} // __try
	__finally
	{
		// Clean up the Intel(R) JPEG Library.
		intel::ijlFree(&jcprops);
	} // __finally

	return status;
}

