#ifndef AsdRational_h_Included
#define AsdRational_h_Included

#include "AsdTypes.h"
#include "AsdMath.h"

#include <algorithm>

namespace Asd
{
    template <class I> class Rational;

/*************************************************************************

A Rational is a ratio of two objects that have integer semantics.  The 
I parameter to the this template is usually an atomic type, but it does
not have to be.  However, you will want to make sure that Euclid's
algorithm works on your objects.  Also note that I must be signed in the
sense that unary operator-() does something useful to it.

All basic arithmatic and copy operations are implemented.  Note that
reduce() is called immediatately upon construction.  This is because
we force two invariants:

  1.  There ratio is in lowest terms.
  2.  The sign of the numerator is the sign of the ration.

This means that we sometimes end up reducing fractions we do not need to
reduce.  However, the alternative is to write more complex and less
efficient code for other operations, such as operator==().  That is not
to say that we have really gone for absolute speed here.  If you are going
to be doing serious number crunching with lots of rationals, this may not
be the object for you.

Note that we have not written math operations for multiplying rationals
by integers, etc.  This is because there is no need.  You can always
turn an integer into a rational.

We do not check for overflow, but we do try to evaluate things in a manor
that is least likely to cause overflow.

*************************************************************************/

template <class I> class Rational
{
public:

    Rational( void ) : m_num( zero(m_num) ), m_den( AsdMath::ident(m_den) ) {}

    explicit Rational( const I & n )
    : m_num(n), m_den(AsdMath::ident(m_den)) {}

    Rational( const I & n, const I & d ) : m_num(n), m_den(d) { reduce(); }


    /* Default copy semantics. */
    /* Rational( const Rational<I> & rhs ); */
    /* Rational & operator=( const Rational<I> & rhs ); */

    Rational & operator=( const I & rhs )
    {
        return *this = Rational<I>(rhs);
    }

    ~Rational( void ) {} /* Not virtual */

    bool operator!=( const Rational<I> & rhs ) const {return !(rhs == *this);}
    bool operator==( const Rational<I> & rhs ) const
    {
        return (this->num() == rhs.num()) && (this->den() == rhs.den());
    }

    bool operator< ( const Rational<I> &     ) const;
    bool operator> ( const Rational<I> & rhs ) const {return rhs<*this;   }
    bool operator<=( const Rational<I> & rhs ) const {return !(*this>rhs);}
    bool operator>=( const Rational<I> & rhs ) const {return !(*this<rhs);}

    Rational<I> operator-( void ) const { return Rational<I>(-num(), den());}


    Rational<I> operator+( const Rational<I> &   ) const;
    Rational<I> operator-( const Rational<I> & r ) const {return *this + -r;}
    Rational<I> operator*( const Rational<I> &   ) const;
    Rational<I> operator/( const Rational<I> & r ) const
    {
        return *this * r.inverse();
    }

    Rational<I> & operator+=( const Rational<I> &   );
    Rational<I> & operator-=( const Rational<I> & r ) {return *this += -rhs;}
    Rational<I> & operator*=( const Rational<I> &   );
    Rational<I> & operator/=( const Rational<I> & r )
    {
        return *this *= r.invserse();
    }

    operator Float32( void ) { return Float32(m_num)/Float32(m_den); }
    operator Float64( void ) { return Float64(m_num)/Float64(m_den); }

    const I & num( void ) const { return m_num; }
    const I & den( void ) const { return m_den; }

    void invert( void ) { std::swap(m_num, m_den); fixSign(); }
    
    Rational<I> inverse( void ) const;
    
    Int32 toPower2();

private:

    /* The numerator and denominator of the rational number are always */
    /* kept in lowest terms with a positive denominator. */
    /* */
    I m_num; /* The numerator. */
    I m_den; /* The denominator. */

    void reduce( void );
    void fixSign( void );
};

template <class I> inline
Int32 Rational<I>::toPower2(){
    Int32 val=0;
    /* set the zoom factor for decoding */
    if(m_num > m_den)
    {
        val = (Asd::UInt32) floor(log (*this) / log (2.));
    }
    val = (Asd::UInt32) pow(2., double (val));
    return val;
}
template <class I> inline
void Rational<I>::reduce( void )
{
    fixSign();

    I factor = AsdMath::gcd( m_num, m_den );

    if( factor )
    {
        m_num /= factor;
        m_den /= factor;
    }
}

template <class I> inline
void Rational<I>::fixSign( void )
{
    /* Make sure denominator is non-negative. */
    if( m_den < 0 )
    {
        m_den *= -AsdMath::ident(m_den);
        m_num *= -AsdMath::ident(m_num);
    }
}

template <class I> inline
bool Rational<I>::operator<( const Rational<I> & rhs ) const
{
    I gcdNum = AsdMath::gcd( this->num(), rhs.num() );
    I gcdDen = AsdMath::gcd( this->den(), rhs.den() );

    I thisProd = this->num() / gcdNum;
    thisProd    *= rhs.  den() / gcdDen;

    I thatProd = rhs.  num() / gcdNum;
    thatProd    *= this->den() / gcdDen;

    return thisProd < thatProd;
}

template <class I> inline
Rational<I> & Rational<I>::
operator+=( const Rational<I> & rhs )
{
    I lcd = AsdMath::lcm( this->den(), rhs.den() );

    this->m_num = (lcd / this->den() ) * this->num();
    this->m_num +=  (lcd / rhs.den() ) * rhs.num();

    this->m_den = lcd;

    this->reduce();

    return *this;
}

template <class I> inline
Rational<I> Rational<I>::
operator+( const Rational<I> & rhs ) const
{
    Rational<I> sum( *this );
    return sum += rhs;
}

template <class I> inline
Rational<I> & Rational<I>::
operator*=( const Rational<I> & rhs )
{
    I gcd1 = AsdMath::gcd( this->num(), rhs.den() );
    I gcd2 = AsdMath::gcd( this->den(), rhs.num() );

    this->m_num /= gcd1;
    this->m_num *= (rhs.num() / gcd2);

    this->m_den /= gcd2;
    this->m_den *= (rhs.den() / gcd1);

    return *this;
}

template <class I> inline
Rational<I> Rational<I>::
operator*( const Rational<I> & rhs ) const
{
    Rational<I> prod( *this );
    return prod *= rhs;
}

template <class I> inline
Rational<I> Rational<I>::inverse( void ) const
{
    Rational<I> inv( *this );

    inv.invert();

    return inv;
}

template <class I> inline
std::ostream & operator<<( std::ostream & os, const Rational<I> & rhs )
{
    os << '(' << rhs.num() << '/' << rhs.den() << ')';

    return os;
}
};
#endif
/*AsdRational_h_Included */
