#ifndef __AsdRect_h_INCLUDED__
#define __AsdRect_h_INCLUDED__

#include <math.h>
#include <assert.h>
#include <algorithm>

#include "MiscConst.h"
#include "AsdTypes.h"
#include "AsdMath.h"

#ifndef _WIN32

typedef struct tagPOINT
{
    Asd::Int32 x;
    Asd::Int32 y;
}
POINT;

typedef struct tagRECT
{ 
    Asd::Int32 left; 
    Asd::Int32 top; 
    Asd::Int32 right; 
    Asd::Int32 bottom; 
}
RECT, *PRECT; 

#endif


/*************************************************************************
 
 Point32 is essentially just Point<Int32>  The only functionality
 that it adds are casting operators to and from POINT.
 The reason for this is to simplify interactions with Win32
 API calls and to simplify the Rect class that follows.

*************************************************************************/

class Asd::Point32 : public Asd::Point<Int32>
{
public:

    Point32( void ) {}

    Point32( Int32 x, Int32 y ) : Point<Int32>(x,y) {}

    explicit Point32( const POINT & p ) : Point<Int32>(p.x,p.y) {}

    /* Default copy semantics. */
    /* Point( const Point<T> & ); */
    /* Point<T> & operator=( const Point<T> & ); */

    ~Point32( void ) {} /* Not virtual. */

    POINT & asPOINT( void )
    {
        assert( sizeof(m_x) == sizeof(Int32) );
        assert( sizeof(Point32) == sizeof(POINT) );

        return *(reinterpret_cast<POINT*>(this));
    }

    const POINT & asPOINT( void ) const
    {
        assert( sizeof(m_x) == sizeof(Int32) );
        assert( sizeof(Point32) == sizeof(POINT) );

        return *(reinterpret_cast<const POINT*>(this));
    }
};

/*************************************************************************

  Rect represents a screen rectangle.  Coordinates are left-handed
  with x increasing to the right and y increasing downwards.  Thus, for
  a normal rectangle, (top < bottom) and (left < right).

  The "invariant" of (left < right) and (top < bottom) is not
  enforced automatically.  A public method fix() is exposed for any
  cases for which you have doubts.  However, you should note that
  isEmpty() returns true any time (left() > right()) or (top() > bottom()).

  Aside from being a rectangle, Rect also serves as a translator
  between ASD libraries and Windows.  In Windows, a RECT or CRect
  usually (always?) follows the convention that coordiates refer to the
  boundaries between pixels, not the pixels themselves.  Thus the 4x4 pixel
  screen region shown here 

      (0,0)   (4,0)
        |       |
        v       v
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |
        +-+-+-+-+<--(4,4)

 is represented in a RECT, CRect, etc. as:

 top=0, bottom=4, left=0, right=4.

 Heigth and width are defined by the formulae:

 height := bottom - top
 width  := right  - left

 (It is not defined what negetive height and width mean.)  It follows
 that heigth=4  and width=4 in our example.  An empty rectangle is
 represented with top==bottom and/or left==right.

 In many ASD libraries, including ipcore, we follow the convention that
 coordinates refer to pixels directly.  Thus, the same 4x4 screen region
 shown here

       (0,0) (3,0)
         |     |
         v     v
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |
        +-+-+-+-+
        | | | | |<--(3,3)
        +-+-+-+-+

 is represented as:

 top=0, bottom=3, left=0, right=3.

 Heigth and width are defined by the formulae:

 height := (bottom >= top )? (bottom - top  + 1) : 0
 width  := (right  >= left)? (right  - left + 1) : 0
 
 This, as you can well imagine, is the source of much strife!

 What to do?

 Rect does black magic that takes care of this for us.  The public
 interface of Rect supports the ASD convetion of pixel addressing.
 Privately, however, we store the data in a RECT using the Windows
 convention.  This allows us to easily (and efficiently) cast Rect
 objects to CRect or RECT objects for Windows calls while writing code
 that follows the non-windows convention.  You can also seemlessly
 construct a Rect out of CRect or RECT with the appropriate
 constructors.

 The implementation of private methods of Rect varies in convention.
 The reason is that some operations are simpler and/or more efficient
 in one representation than another.  Keep your eye's open if you 
 are looking at the implementation details.

 One final note.  Rect is not general part of the ASD math library the
 way point is.

**************************************************************************/

class Asd::Rect
{
public:

    /* */
    /* Constructors/Destructor */
    /* */

    Rect( void ) { setEmpty(); }

    Rect( Int32 t, Int32 l, Int32 b, Int32 r )
    {
        setTLBR( t, l, b, r );
    }

    Rect( const Asd::Point<Int32> & topLeft, const Asd::Point<Int32> & bottomRight );

    /* Construct a rectangle with upper left corner at origin of with the */
    /* desired height and width. */
    Rect( Int32 w, Int32 h ) { setTLBR( 0, 0, h-1, w-1 );}

    /* Similar to the two integer constructor above, but wh.getX() is the */
    /* width and wh.getY() is the height(). */
    explicit Rect( const Asd::Point<Int32> & wh )
    {
        setTLBR( 0, 0, wh.getY()-1, wh.getX()-1 );
    }

    /* Use default copy semantics. */
    /* Rect( const Rect & ); */
    /* Rect & operator=( const Rect & ); */

    /* */
    /* Conversions to and from RECT and CRect. */
    /* */

    explicit Rect( const RECT & r ) { *this = r; }

    Rect & operator=( const RECT & r ) { m_rect=r; return *this; }

    ~Rect( void ) {} /* Not virtual! */

    bool operator==( const Rect & rhs ) const;
    bool operator!=( const Rect & rhs ) const { return !(*this == rhs); }
    
    const RECT & asRECT ( void ) const { return m_rect; }   
          RECT & asRECT ( void )       { return m_rect; }   
       

    void fix( void );   /* forces top <= bottom and left <= right */

    /* */
    /* Methods for getting and setting the corners. */
    /* */
    /* Note that setRight() and setBottom() add 1 to the value passed in. */
    /* This converts to RECT format. */

    void setTop   ( Int32 t ) { m_rect.top    = t;   }
    void setLeft  ( Int32 l ) { m_rect.left   = l;   }
    void setBottom( Int32 b ) { m_rect.bottom = b+1; }
    void setRight ( Int32 r ) { m_rect.right  = r+1; }

    /* set "T"op "L"eft "B"ottom "R"ight. */
    void setTLBR( Int32 t, Int32 l, Int32 b, Int32 r );
    
    Int32 top   ( void ) const { return m_rect.top;        }
    Int32 left  ( void ) const { return m_rect.left;       }
    Int32 bottom( void ) const { return m_rect.bottom - 1; }
    Int32 right ( void ) const { return m_rect.right  - 1; }

    Point32 topLeft    (void) const { return Point32(left (),top   ()); }
    Point32 bottomRight(void) const { return Point32(right(),bottom()); }

    /* These will return 0 any time (top() > bottom()) or (left() > right()). */
    Int32 width ( void ) const;
    Int32 height( void ) const;
    Int32 area  ( void ) const { return height()*width(); }

    Float64 aspectRatio( void ) const
    {
        return height()? Float64(width())/Float64(height()) : Float64(0);
    }

    Point32 size( void ) const { return Point32( width(), height() ); }

    bool  isEmpty( void ) const { return !height() || !width(); }
    void setEmpty( void ) { setTLBR( 0, 0, -1, -1 ); }

    void inflateToHW( Int32 newHeight, Int32 newWidth );
    void inflateToHeight( Int32 newHeight );
    void inflateToWidth ( Int32 newWidth  );

    void offsetXY( Int32 xOff, Int32 yOff );
    void offsetX ( Int32 xOff );
    void offsetY ( Int32 yOff );
    void offset( const Asd::Point<Int32> & off ) { offsetXY(off.getX(), off.getY()); }

    void scaleUp( Int32 fac );
    void scaleDown( Int32 fac );
    void scaleUp( Int32 xfac, Int32 yfac );
    void scaleDown( Int32 xfac, Int32 yfac );
    void scaleDownMax( Int32 xfac, Int32 yfac );

    void scaleUp( Asd::Float64 );
    void scaleDown( Asd::Float64 );

    void rotate( Orientation, Int32 pageWidth, Int32 pageHeight );

    /* Two rectangles overlap if their intersection is not empty. */
    /* Note that an empty rectangle will not overlap a rectangle even */
    /* if it's corners are contained within the other rectangle. */
    bool overlaps( const Rect & r2 ) const;
    bool isSame  ( const Rect & r2 ) const;
    
    bool contains( Int32 x, Int32 y ) const;
    bool contains( const Asd::Point<Int32>  &   ) const;
    bool contains( const Rect       &   ) const;

    Rect intersection( const Rect & r2 ) const;

    /* We want a rectangle that meets the following three conditions: */
    /*    */
    /* 1.  Contains all of *this. */
    /* 2.  Contains no space that is not within either this or r1. */
    /* 3.  Contains as much of r1 as possible, given #1 and #2 above. */
    Rect getLargestUnionRect( const Rect & r1 ) const;

    
    /* We want a rectangle that meets the following three conditions: */
    /*    */
    /* 1.  Contains all of *this. */
    /* 2.  Contains all of r2. */
    /* 3.  Is as small as possible given #1 and #2. */
    Rect getSmallestUnionRect( const Rect & r1 ) const;


    static void makeBorders( const Rect & outer,
                             const Rect & inner,
                             Rect *    pTopBord,
                             Rect * pBottomBord,
                             Rect *   pLeftBord,
                             Rect *  pRightBord );
private:

    /* Note that asside from some of the obvious conveniences with using */
    /* the canned RECT struct here (instead of four variables) there is */
    /* also the advantage that there used to be (and may still be) places */
    /* in the code that assume that (RECT *) and (Rect*) can be used */
    /* interchangeably. */
    RECT m_rect;
};

/*//////////////////////////////////////////////////////////////////////// */
/*//////////////////////////////////////////////////////////////////////// */
/*//////////////////////////////////////////////////////////////////////// */

inline Asd::Rect::
Rect( const Asd::Point<Int32> & topLeft, const Asd::Point<Int32> & bottomRight )
{
    setTop   ( topLeft.    getY() );    
    setLeft  ( topLeft.    getX() );
    setBottom( bottomRight.getY() );    
    setRight ( bottomRight.getX() );
}

inline Asd::Int32 Asd::Rect::width( void ) const
{
    const Int32 width = (m_rect.right - m_rect.left);

    return (width > 0)?  width  :  0;
}

inline Asd::Int32 Asd::Rect::height( void ) const
{
    const Int32 height = (m_rect.bottom - m_rect.top);
    
    return (height > 0)?  height  :  0;
}

/*************************************************************************

  inflateToHeight() and inflateToWidth() each scale the size of the
  rectangle.  Instead of scaling by a factor, however, they instead
  scale to a specifc size.  inflateToHW() just calls inflateToHeight()
  and inflateToWidth() on its arguments.
  
  Note that special care is taken with command to allow inflation to sizes
  that differ from the current size by an odd number.

*************************************************************************/

inline void Asd::Rect::inflateToHeight( Int32 newHeight )
{
    const int extraHeight  = newHeight - height();
    const int yBaseInflate = int(floor(extraHeight/2));

    m_rect.top    -=               yBaseInflate;
    m_rect.bottom += extraHeight - yBaseInflate;
}

inline void Asd::Rect::inflateToWidth( Int32 newWidth )
{
    const int extraWidth   = newWidth - width();
    const int xBaseInflate = int(floor(extraWidth/2));

    m_rect.left  -=              xBaseInflate;
    m_rect.right += extraWidth - xBaseInflate;
}

inline void Asd::Rect::inflateToHW( Int32 newHeight, Int32 newWidth )
{
    inflateToHeight( newHeight );
    inflateToWidth ( newWidth );
}


inline void Asd::Rect::offsetX( Int32 xOff )
{
    m_rect.left  += xOff;
    m_rect.right += xOff;
}

inline void Asd::Rect::offsetY( Int32 yOff )
{
    m_rect.top    += yOff;
    m_rect.bottom += yOff;
}

inline void Asd::Rect::offsetXY( Int32 xOff, Int32 yOff )
{
    offsetX( xOff );
    offsetY( yOff );
}

inline void Asd::Rect::scaleUp( Int32 fac )
{
    m_rect.top    *= fac;
    m_rect.bottom *= fac;
    m_rect.left   *= fac;
    m_rect.right  *= fac;
}

inline void Asd::Rect::scaleDown( Int32 fac )
{
    if (fac == 0)
    {
        m_rect.top = m_rect.bottom = m_rect.left = m_rect.right = 0;
    }
    else
    {
        m_rect.top    /= fac;
        m_rect.bottom /= fac;
        m_rect.left   /= fac;
        m_rect.right  /= fac;
    }
}

inline void Asd::Rect::scaleUp( Int32 xfac, Int32 yfac )
{
    m_rect.top    *= yfac;
    m_rect.bottom *= yfac;
    m_rect.left   *= xfac;
    m_rect.right  *= xfac;
}

inline void Asd::Rect::scaleDown( Int32 xfac, Int32 yfac )
{
    if (yfac == 0)
    {
        m_rect.top = m_rect.bottom = 0;
    }
    else
    {
        m_rect.top    /= yfac;
        m_rect.bottom /= yfac;
    }

    if (xfac == 0) 
    {
        m_rect.left = m_rect.right = 0;
    }
    else
    {
        m_rect.left  /= xfac;
        m_rect.right /= xfac;
    }
}

inline void Asd::Rect::scaleDownMax( Int32 xfac, Int32 yfac )
{
    if (xfac == 0) 
    {
        m_rect.left = m_rect.right = 0;
    }
    else
    {
        m_rect.left /= xfac;

        m_rect.right += xfac - 1;
        m_rect.right /= xfac;
    }

    if (yfac == 0)
    {
        m_rect.top = m_rect.bottom = 0;
    }
    else
    {
        m_rect.top /= yfac;

        m_rect.bottom += yfac - 1;
        m_rect.bottom /= yfac;
    }
}

inline void Asd::Rect::scaleUp( Asd::Float64 fac )
{
    m_rect.top    = (Asd::Int32)(m_rect.top * fac);
    m_rect.bottom = (Asd::Int32)(m_rect.bottom * fac);
    m_rect.left   = (Asd::Int32)(m_rect.left * fac);
    m_rect.right  = (Asd::Int32)(m_rect.right * fac);
}

inline void Asd::Rect::scaleDown( Asd::Float64 fac )
{
    if( fac != 0.0 )
    {
        scaleUp( 1.0/fac );
    }
    else
    {
        m_rect.top    = 0;
        m_rect.bottom = 0;
        m_rect.left   = 0;
        m_rect.right  = 0;
    }
}

/*************************************************************************

  rotate rotates the rectangle to one of four relative orientations.
  The rotation is the transformation that would take place if the rectangle
  were a region of a page being rotated.  The page is assumed to have
  top-down coordinates.  (Top-down coordinates are left-handed.)  Rotation
  is in the clockwise direction.
  
  As an example:


      width 23
  +---------------------+
  |                     |
  |  (4,3)  (10,3)      |
  |   +-----+           |
  |   |     |           |
  |   |     |           |
  |   |     |           |  height 12
  |   |     |           |
  |   +-----+           |
  |  (4,8)   (10,8)     |
  |                     |
  +---------------------+ 
 
 rotated 90 degress becomes


    width 12
  +----------+
  |          |
  |          |
  |(3,4)(8,4)|
  |  +----+  |  
  |  |    |  |
  |  |    |  |
  |  |    |  |
  |  |    |  |
  |  |    |  |  height 23
  |  +----+  |  
  |(3,10)  (8,10)
  |          |
  |          |
  |          |
  |          |
  |          |
  |          |
  |          |
  |          |
  |          |
  |          |
  +----------+


 The orient argument is the new orientation relative to the current one.
 The pageWidth and pageHeight arguments are the height and widht of the
 page on which the rectangle resides before any rotations.

*************************************************************************/

inline void Asd::Rect::
rotate( Orientation orient, Int32 pageWidth, Int32 pageHeight )
{
    Rect temp( *this );

    switch( orient )
    {
        case Orient_0: break;

        case Orient_270:
        {
            this->setTop   ( pageWidth - temp.right() - 1 );
            this->setBottom( pageWidth - temp.left()  - 1 );
            this->setLeft  ( temp.top()    );
            this->setRight ( temp.bottom() );
        }
        break;

        case Orient_180:
        {
            this->setTop   ( pageHeight - temp.bottom() - 1 );
            this->setBottom( pageHeight - temp.top()    - 1 );
            this->setLeft  ( pageWidth  - temp.right()  - 1 );
            this->setRight ( pageWidth  - temp.left()   - 1 );
        }
        break;

        case Orient_90:
        {
            this->setTop   ( temp.left()   );
            this->setBottom( temp.right()  );
            this->setLeft  ( pageHeight - temp.bottom() - 1 );
            this->setRight ( pageHeight - temp.top() - 1 );
        }
        break;

        default: break;
    }
}

inline bool Asd::Rect::overlaps( const Rect & r2 ) const
{
    return !( this->intersection(r2).isEmpty() );
}

inline Asd::Rect Asd::Rect::intersection( const Rect & r2 ) const
{
    Rect ret;

    ret.m_rect.top    = AsdMath::MAX(this->m_rect.top   , r2.m_rect.top   );
    ret.m_rect.left   = AsdMath::MAX(this->m_rect.left  , r2.m_rect.left  );
    ret.m_rect.bottom = AsdMath::MIN(this->m_rect.bottom, r2.m_rect.bottom);
    ret.m_rect.right  = AsdMath::MIN(this->m_rect.right  ,r2.m_rect.right );

    return ret;
}

inline bool Asd::Rect::isSame( const Rect & r2 ) const
{
  return (this->m_rect.left   == r2.m_rect.left  ) && 
         (this->m_rect.bottom == r2.m_rect.bottom) &&
         (this->m_rect.top    == r2.m_rect.top   ) &&
         (this->m_rect.right  == r2.m_rect.right );
}

inline bool Asd::Rect::contains( const Asd::Point<Int32> & pt ) const
{
   return contains( pt.getX(), pt.getY() );
}

inline bool Asd::Rect::contains( Int32 x, Int32 y ) const
{
   return (x >= m_rect.left) && (x < m_rect.right ) &&
          (y >= m_rect.top ) && (y < m_rect.bottom);
}

inline bool Asd::Rect::contains( const Rect & r2 ) const
{
   return contains(r2.left (), r2.top   ())  &&
          contains(r2.right(), r2.bottom());
}

/*

  fix() "fixes" a rectangle by ensuring that top < bottom and
  left < right.  If these conditions are violated, values are
  swapped to "fix" the rectangle.


*/

inline void Asd::Rect::fix( void )
{
    if( m_rect.top  > m_rect.bottom )
    {
        std::swap( m_rect.top, m_rect.bottom );
    }

    if( m_rect.left > m_rect.right  )
    { 
        std::swap( m_rect.left, m_rect.right );
    }
}

inline bool Asd::Rect::operator==( const Rect & rhs ) const
{
    return (this->m_rect.top    == rhs.m_rect.top   ) &&
           (this->m_rect.bottom == rhs.m_rect.bottom) &&
           (this->m_rect.left   == rhs.m_rect.left  ) &&
           (this->m_rect.right  == rhs.m_rect.right );
}

inline void Asd::Rect::
setTLBR( Int32 t, Int32 l, Int32 b, Int32 r )
{
    setTop   ( t );
    setBottom( b );
    setLeft  ( l );
    setRight ( r );
}


/*========================================================================
   We want a rectangle that meets the following three conditions:
   
   1.  Contains all of r2.
   2.  Contains no space that is not within either r2 or r1.
   3.  Contains as much of r1 as possible, given #1 and #2 above.


   If *this is completely within r1 in a dimension, then we may be able
   to expand uRect in the other dimension.  For example, suppose you
   have:
       
               +-------+ <--r1
               |       |
           +---+----+  |
           |   |    |  |
  *this--> |   |    |  | 
           |   |    |  |
           +---+----+  |
               |       |
               +-------+

Then, with uRect starting out as *this, we can grow it....

               +-------+ <--r1
               |       |
           +---+-------+
           |   |       |
  uRect--> |   |       | 
           |   |       |
           +---+-------+
               |       |
               +-------+


   A similar set of drawings would work for the other dimension.  Also
   Note that we don't have to have separate cases for when both or
   neither dimension needs adjusting.  If, for example, *this is contained
   in r1, we will stretch on one axis and then the other.

=========================================================================*/

inline Asd::Rect Asd::Rect::getLargestUnionRect( const Rect & r1 ) const
{
    /* Since we need all of *this, start by guessing that *this is the */
    /* answer This is the case if r1 and *this do not overlap. */
    Rect uRect( *this );

    if( (this->m_rect.top    >= r1.m_rect.top   ) &&
        (this->m_rect.bottom <= r1.m_rect.bottom)    )
    {
        uRect.m_rect.left  = AsdMath::MIN(r1.m_rect.left , this->m_rect.left );
        uRect.m_rect.right = AsdMath::MAX(r1.m_rect.right, this->m_rect.right);
    }
        
    if( (this->m_rect.left  >= r1.m_rect.left ) && 
        (this->m_rect.right <= r1.m_rect.right)  )
    {
        uRect.m_rect.top    = AsdMath::MIN(r1.m_rect.top   , this->m_rect.top   );
        uRect.m_rect.bottom = AsdMath::MAX(r1.m_rect.bottom, this->m_rect.bottom);
    }

    return uRect;
}

inline Asd::Rect Asd::Rect::getSmallestUnionRect( const Rect & r1 ) const
{
    Rect uRect;

    uRect.m_rect.top    = AsdMath::MIN(this->m_rect.top   , r1.m_rect.top   );
    uRect.m_rect.left   = AsdMath::MIN(this->m_rect.left  , r1.m_rect.left  );
    uRect.m_rect.bottom = AsdMath::MAX(this->m_rect.bottom, r1.m_rect.bottom);
    uRect.m_rect.right  = AsdMath::MAX(this->m_rect.right , r1.m_rect.right );

    return uRect;
}

inline void Asd::Rect::
makeBorders( const Rect & outer,
             const Rect & inner,
             Rect *    pTopBord,
             Rect * pBottomBord,
             Rect *   pLeftBord,
             Rect *  pRightBord )
{
    assert(    pTopBord );
    assert( pBottomBord );
    assert(   pLeftBord );
    assert(  pRightBord );


    /* These four rectangles will represent the borders of the inner */
    /* area. */
    /* */
    /*  +---------------------------------+ */
    /*  |                                 | */
    /*  |           topBord               | */
    /*  |                                 | */
    /*  +------+-------------------+------+ */
    /*  |      |         ^         |      | */
    /*  | left-|         |         |right-| */
    /*  | Bord |<------inner------>| Bord | */
    /*  |      |         |         |      | */
    /*  |      |         v         |      | */
    /*  +------+-------------------+------+ */
    /*  |                                 | */
    /*  |           bottomBord            | */
    /*  |                                 |        */
    /*  +---------------------------------+<----- outer */
    /*                                             */
    /* */
    
    /* Three of the four sides of each of the top and bottom borders will be */
    /* from bounds, so just make copies of bounds to start. */
    *pTopBord    = outer;
    *pBottomBord = outer;

    /* Fix the last side of each.  If the image extends out of bounds, then */
    /* we want empty rectangles. */
       pTopBord->setBottom( AsdMath::MAX(inner.top   (), outer.top   ()) - 1 );
    pBottomBord->setTop   ( AsdMath::MIN(inner.bottom(), outer.bottom()) + 1 );

    /* leftBord and rightBord both have the same top and bottom. */

     pLeftBord->setTop   ( pTopBord->bottom() + 1 );
    pRightBord->setTop   ( pTopBord->bottom() + 1 );
     pLeftBord->setBottom( pBottomBord->top() - 1 );
    pRightBord->setBottom( pBottomBord->top() - 1 );

    
    /* Now set the remain portions of the borders.      */
     pLeftBord->setLeft ( outer.left () );
    pRightBord->setRight( outer.right() );
     pLeftBord->setRight( AsdMath::MAX(inner.left (), outer.left ()) - 1 );
    pRightBord->setLeft ( AsdMath::MIN(inner.right(), outer.right()) + 1 );

}

#endif
/*#ifndef __AsdRect_h_INCLUDED__ */
