#ifndef AsdUInt24_H_INCLUDED
#define AsdUInt24_H_INCLUDED

#include "AsdTypes.h"

namespace Asd
{
    class RgbPtr;
    class RgbVal;
};

class Asd::RgbVal
{
public:

    enum { R=0, G=1, B=2 };

    RgbVal( void ) { init(); setRGB(0,0,0); }

    RgbVal( const RgbPtr & );

    RgbVal( UInt8 r, UInt8 g, UInt8 b  ) { init(); setRGB(r,g,b); }

    /* Default copy semantics. */
    /* RgbVal( const RgbVal & ); */
    /* RgbVal & operator=( const RgbVal & ); */

    ~RgbVal( void ) {} /* Not virtual! */

    operator const UInt32 & ( void ) const
    {
        return *(reinterpret_cast<const UInt32 *>(m_data));
    };

    bool operator==( const RgbVal & rhs ) const
    {
        return UInt32(*this) == UInt32(rhs);
    }
    
    bool operator!=( const RgbVal & rhs ) const
    {
        return UInt32(*this) != UInt32(rhs);
    }

    /* init() must be called as the very first action of every constructor. */
    void init( void ) { m_data[3] = 0; }

    void setRGB( UInt8 r, UInt8 g, UInt8 b )
    {
        m_data[R] = r;
        m_data[G] = g;
        m_data[B] = b;
    }

    UInt8 r( void ) const { return m_data[R]; }
    UInt8 g( void ) const { return m_data[G]; }
    UInt8 b( void ) const { return m_data[B]; }

    UInt8 & r( void ) { return m_data[R]; }
    UInt8 & g( void ) { return m_data[G]; }
    UInt8 & b( void ) { return m_data[B]; }

private:

    /* The 4th slot is a place holder that is always 0.  It */
    /* allows simpler, more efficient code in some cases. */
    UInt8 m_data[4];
};

/*************************************************************************

 RgbPtr allows convenient access to 24-bit image data packed into
 an array of bytes.  operator++() moves the pointer 3 bytes, etc.

*************************************************************************/

class Asd::RgbPtr
{
public:
    
    explicit RgbPtr( UInt8 * p = NULL ) : m_ptr(p) {}

    /* Default copy semantics. */
    /* RgbPtr( const RgbPtr & ); */
    /* RgbPtr & operator=( const RgbPtr & ); */

    ~RgbPtr( void ) {}
    
    /* All of the following comparisons have the standard pointer */
    /* arithmatic meanings. */
    bool operator==(const RgbPtr &rhs) const {return m_ptr == rhs.m_ptr;}
    bool operator!=(const RgbPtr &rhs) const {return m_ptr != rhs.m_ptr;}
    bool operator< (const RgbPtr &rhs) const {return m_ptr <  rhs.m_ptr;}
    bool operator> (const RgbPtr &rhs) const {return m_ptr >  rhs.m_ptr;}
    bool operator<=(const RgbPtr &rhs) const {return m_ptr <= rhs.m_ptr;}
    bool operator>=(const RgbPtr &rhs) const {return m_ptr >= rhs.m_ptr;}
    
    /* Move our pointer 1 24-bit pixel. */
    RgbPtr & operator++( void ) { m_ptr += 3; return *this; }
    RgbPtr & operator--( void ) { m_ptr -= 3; return *this; }

    /* Move our pointer n 24-bit pixels. */
    RgbPtr & operator+=( size_t n ) { m_ptr += n*3; return *this; }
    RgbPtr & operator-=( size_t n ) { m_ptr -= n*3; return *this; }

    RgbPtr operator+( size_t n ) { return RgbPtr( m_ptr + (n*3) ); }
    RgbPtr operator-( size_t n ) { return RgbPtr( m_ptr - (n*3) ); }

    void reset( UInt8 * p ) { m_ptr = p; }
    
    /* Move our pointer n bytes. */
    RgbPtr & offset( size_t n ) { m_ptr += n; return *this; }

    UInt8 * bytePtr( void ) const { return m_ptr; }

    /* The rgb functions assume that r is the low-order byte and g */
    /* is the high-order byte. */

    UInt8 & r( void ) const { return m_ptr[0]; }
    UInt8 & g( void ) const { return m_ptr[1]; }
    UInt8 & b( void ) const { return m_ptr[2]; }

    void set( const RgbVal & v ) const { setRGB( v.r(), v.g(), v.g() ); }

    void setRGB( UInt8 r_, UInt8 g_, UInt8 b_ ) const
    {
        r() = r_;
        g() = g_;
        b() = b_;
    }

    /* Copy the data pointed to by another RgbPtr to our own data */
    /* area. */
    void copyDataFrom( const RgbPtr & src ) const
    {
        this->m_ptr[0] = src.m_ptr[0];
        this->m_ptr[1] = src.m_ptr[1];
        this->m_ptr[2] = src.m_ptr[2];
    }

    void swapData( RgbPtr & that ) const
    {
        RgbVal tmp( that );

        that.copyDataFrom( *this );

        this->set( tmp );
    }

    bool hasSameData( const RgbPtr & that ) const
    {
        return (this->r() == that.r()) &&
               (this->g() == that.g()) &&
               (this->b() == that.b());
    }

    bool isWhite( void ) const { return (r() & g() & b()) == 0xff; }
    bool isBlack( void ) const { return (r() | g() | b()) == 0x00; }

    bool isDontCare( void ) const { return !(r() | g() | (unsigned char)(~b())); }
    void markDontCare( void ) const { setRGB( 0, 0, 0xff ); }

    void forceCare( void ) const { if( isDontCare() ) { m_ptr[2] = 0xfe; } }

private:

    /* Pointer to the first byte of the current 24-bit pixel. */
    UInt8 * m_ptr;
};


inline Asd::RgbVal::RgbVal( const Asd::RgbPtr & ptr )
{
    init();
    setRGB( ptr.r(), ptr.g(), ptr.b() );
}

#endif
/*#ifndef AsdUInt24_H_INCLUDED */
