#ifndef __AsdTypes_h_INCLUDED__
#define __AsdTypes_h_INCLUDED__

#include <assert.h>
#include <iostream>

namespace Asd
{
    typedef signed char    Int8;
    typedef unsigned char UInt8;

    typedef short           Int16;
    typedef unsigned short UInt16;

    typedef long           Int32;
    typedef unsigned long UInt32;

    typedef float  Float32;
    typedef double Float64;

    /* zero<T> is a function that returns "0" for the type T.  The reason */
    /* we have this here is mostly to support template classes and functions */
    /* with mathematical semantics.  For built-in types, the default */
    /* constructor leaves the data undefined.  This is often inconveniant. */
    /* However, for built-in types we can just use plain old "0" and the */
    /* compiler will know what to do.  What if we have a class though? */
    /* We could argue that "foo()" (the no-arg constructor) was "0", but */
    /* then our template would not work properly for built-in types. */
    /* We could argue that "foo(0)" should be "0", but not all arithmatic */
    /* types should have 1-argument constructors.  Instead, we have a */
    /* convention that zero( object ) must return 0 for any object (that */
    /* we expect many of our templates to support.)  There are (will be */
    /* shortly) many similar concepts in the math files.  However, we do */
    /* not want to have to include math headers just to use zero(int). */
    /* The template below will expand properly for built-in types.  It */
    /* will also expand properly for classes that have 1-argumnent */
    /* constructors. */
    template <class T> T zero( const T & ) { return T(0); }

    template <class T> class Point;
    class Point32;
    class Rect;

    inline void checkTypes( void )
    {
        assert( sizeof(   Int8) == 1 );
        assert( sizeof ( UInt8) == 1 );
        assert( sizeof(  Int16) == 2 );
        assert( sizeof( UInt16) == 2 );
        assert( sizeof(  Int32) == 4 );
        assert( sizeof( UInt32) == 4 );
        assert( sizeof(Float32) == 4 );
        assert( sizeof(Float64) == 8 );
    }

/*************************************************************************
 
 The Point template class represents a simple 2-dimensional coordinate
 pair.  As is common, the coordinates are referred to as "x" and "y."
 
 In addition to bundling the two elements together, the Point class 
 exposes basic copy, comparison, and arithmatic semantics.  The only
 restrictions on T are that it must:

 -- Have copy semantics.
 -- Define operator += and -=
 -- Define unary operator-
 -- Define operator ==

 Since no multiplication or division occurs, Point can be used with any
 of the atomic types.

*************************************************************************/

template <class T> class Point
{
public:

    Point( void ) : m_x(zero(m_x)), m_y(zero(m_y)) {}

    Point( const T & x, const T & y ) : m_x(x), m_y(y) {}

    /* Default copy semantics. */
    /* Point( const Point<T> & ); */
    /* Point<T> & operator=( const Point<T> & ); */

    ~Point( void ) {} /* Not virtual. */

    bool operator!=( const Point<T> & rhs ) const { return !(*this == rhs); }
    bool operator==( const Point<T> & rhs ) const
    {
        return (this->m_x == rhs.m_x) && (this->m_y == rhs.m_y);
    }

    Point<T> & operator+=( const Point<T> & rhs );
    Point<T> & operator-=( const Point<T> & rhs );

    Point<T> operator+( const Point<T> & rhs ) const { return Point<T>(*this) += rhs; }
    Point<T> operator-( const Point<T> & rhs ) const { return Point<T>(*this) -= rhs; }

    Point<T> operator-( void ) const { return Point<T>(-m_x,-m_y); }

    const T & getX( void ) const { return m_x; }
    const T & getY( void ) const { return m_y; }

    void setX ( const T & x ) { m_x=x; }
    void setY ( const T & y ) { m_y=y; }

    void setXY( const T & x, const T & y ) { setX(x); setY(y); }

protected:

    T m_x, m_y;
};
  
template <class T> inline
std::ostream & operator<<( std::ostream & lhs, Point<T> rhs )
{
    return lhs << "(" << rhs.getX() << ", " << rhs.getY() << ")";
}

template <class T> inline
Point<T> & Point<T>::operator+=( const Point<T> & rhs )
{
    this->m_x += rhs.m_x;
    this->m_y += rhs.m_y;

    return *this;
}

template <class T> inline
Point<T> & Point<T>::operator-=( const Point<T> & rhs )
{
    this->m_x -= rhs.m_x;
    this->m_y -= rhs.m_y;

    return *this;
}

};
#endif
/*#ifndef __AsdTypes_h_INCLUDED__ */

