#ifndef OddPtr_H_INCLUDED
#define OddPtr_H_INCLUDED

namespace Asd
{
    class BitPtr;
    class NibblePtr;
};

/*************************************************************************

  BitPtr is a pointer-like object that allows you to conveniently access
  single bits in an array of bytes.  This is particularly useful for
  dealing with monochrome images.  Be warned however, it does not
  quite have true pointer semantics.

  For the most part, it behaves exactly as a pointer.  operator++() moves
  the pointer 1 bit.  operator+=() moves it an arbitrary number of bits.
  operator* is a const, non-reference operator, because there is no way
  to return a reference to a single bit.  (Actually, this is not quite
  true.  What I should say is that there is no way to do so efficiently.
  We could make a "Bit" object that contained a pointer and a mask that
  had bit semantics directly on the array, but that would mean extra
  steps in every bit manipulation.)
  
  Most bit manipultions wiil probably happen with set(), flip(), and
  clear().

  Because BitPtr was built for images, the bits go 0x80, 0x40 ... 0x02, 
  0x01.  That is to say, the high-order bit is the first one accessed
  in each byte.

*************************************************************************/

class Asd::BitPtr
{
public:

    explicit BitPtr( unsigned char * ptr = NULL ) { reset( ptr ); }

    /* Default copy semantics. */
    /* BitPtr( const BitPtr & rhs ); */
    /* BitPtr & operator=( const BitPtr & rhs ); */

    ~BitPtr( void ) {} /* Not virtual */

    /* Advance pointer 1 bit. */
    BitPtr & operator++( void )
    {       
        m_byte += (m_bit & 0x01);
        m_bit = ( (m_bit << 7) | (m_bit >> 1) );

        return *this;
    }

    /* Advance pointer 1 bit. */
    BitPtr & operator--( void )
    {       
        m_byte -= ((m_bit & 0x80) >> 7);
        m_bit = ( (m_bit >> 7) | (m_bit << 1) );

        return *this;
    }

    /* Advance pointer n bits. */
    BitPtr & operator +=( size_t n )
    {
        /* Move byte the number of whole bytes. */
        m_byte += (n/8);

        /* Find out the left-over number of bits. */
        n %= 8;

        if( n )
        {
            /* Put our bit in a larger integer. */
            unsigned int val = m_bit;

            /* Now copy the bit one byte over.  When we shift off remain */
            /* the bit that is in the low-order byte will be the one that */
            /* we want. */
            val |= (val << 8);

            /* Right shift the bit reamin spaces.  If this is enough to take */
            /* us to the next byte, the orginal bit is shifted off.  And the */
            /* one we just created ends up in the right place.  If it is not */
            /* enough, then the original bit ends up in the right position */
            /* and the new bit ends up staying in the higher byte. */
            val >>= n;

            /* Set bit to the new value. */
            m_bit = (val & 0xff);

            /* Increase byte if needed.  If both bits are still present, then */
            /* one of the bits will be in the 9th place or higher.  Both bits */
            /* still present indicates that we have not rolled over. */
            m_byte += (val <= 0x80);
        }
    
        return *this;
    }

    /* Read the bit currently pointed to. */
    bool operator*( void ) const { return !!((*m_byte) & m_bit); }
    
    /*  Each of set(), clear(), and flip() operate on the current bit. */
    void set( bool val ) { val? set() : clear(); }  
    void set  ( void ) { *m_byte |=  m_bit; }
    void clear( void ) { *m_byte &= ~m_bit; }
    void flip ( void ) { *m_byte ^=  m_bit; }

    /* Reassign this pointer to reference some other data. */
    void reset( unsigned char * p ) { m_byte = p; m_bit  = 0x80; }

    /* Move this pointer n bytes. */
    void offset( size_t n ) { m_byte += n; }

    /* Get the pointer to the current byte.  Beware that the current */
    /* bit may  be anywhere in this byte. */
    unsigned char * bytePtr( void ) const { return m_byte; }

    unsigned char bitMask( void ) const { return m_bit; }

    /* Bet the orderinal number of the current bit. */
    size_t bitNum( void ) const;

private:

    /* Points to the data we are currently accessing. */
    unsigned char * m_byte;
    
    /* A mask telling us which bit we are on.  m_bit should never be */
    /* anything other than a power of 2.  That is, exactly one bit */
    /* should always be set. */
    unsigned char m_bit;
};

inline size_t Asd::BitPtr::bitNum( void ) const
{
    switch( m_bit )
    {
        case 0x01 : return 7;
        case 0x02 : return 6;
        case 0x04 : return 5;
        case 0x08 : return 4;

        case 0x10 : return 3;
        case 0x20 : return 2;
        case 0x40 : return 1;

        case 0x80 : 
        default   : break;
    };

    return 0;
}


/*************************************************************************

 NibblePtr is a relative of BitPtr.  It's job is to simplify operations
 on groups of 4 bits.  Most of the comments about bit ptr apply here
 as well.
   
*************************************************************************/

class Asd::NibblePtr
{
public:

    explicit NibblePtr( unsigned char * ptr = NULL ) { reset( ptr ); }

    explicit NibblePtr( const BitPtr & bp )
    : m_byte( bp.bytePtr() ),
      m_nibble( (bp.bitMask() == 0x80)? 0xf0 : 0x0f )

    {
        /* This construction really only makes sense if the BitPtr is */
        /* lined up on a nibble. */
        assert( (bp.bitMask() == 0x08) ||
                (bp.bitMask() == 0x80)   );
    }

    /* Default copy semantics. */
    /* BitPtr( const BitPtr & rhs ); */
    /* BitPtr & operator=( const BitPtr & rhs ); */

    ~NibblePtr( void ) {} /* Not virtual */

    /* Advance our pointer 1 nibble. */
    NibblePtr & operator++( void )
    {       
        m_byte   += (m_nibble & 0x01);
        m_nibble ^= 0xff;

        return *this;
    }

    NibblePtr & operator +=( size_t n )
    {
        m_byte += (n/2);

        if( n % 2 ) { ++(*this); }
    
        return *this;
    }
    
    unsigned char operator*( void ) const
    {
        unsigned char val = (*m_byte) & m_nibble;

        val |= (val >> 4);

        return val & 0x0f;
    }
    
    /* All of these operate on all four bits together. */
    void setAll  ( void ) { *m_byte |=  m_nibble; }
    void clearAll( void ) { *m_byte &= ~m_nibble; }
    void flipAll ( void ) { *m_byte ^=  m_nibble; }

    /* set4() sets each of the bits in the nibble.  only the low-order */
    /* four bits of val are used regardless of whether we are on an even */
    /* or odd nibble.  It is illegal to have any of the high-order bits */
    /* set.  (Note the assertion.) */
    void set4( unsigned char val )
    {
        assert( !(val & 0xf0) );

        set4Fast( val | (val << 4) );
    }

    /* This is like set4(), but it assumes that you have put the data */
    /* into the correct nibble.  Because  it turns out that allowing */
    /* garbage to be in the other nibble is sometimes useful for this */
    /* optimization case, this method will safely ignore the nibble */
    /* that we are not interested in.  You can see how this is used */
    /* by set4() above. */
    void set4Fast( unsigned char val )
    {
        *m_byte &= ~m_nibble;
        *m_byte |= (val & m_nibble);
    }

    /* Make our pointer reference new data. */
    void reset( unsigned char * ptr )
    {
        m_byte   = ptr;
        m_nibble = 0xf0;
    }

    unsigned char * bytePtr( void ) const { return m_byte; }

private:

    /* Pointer to our data. */
    unsigned char * m_byte;
    
    /* A mask for the current nibble.  This must always be one of */
    /* 0xf0 or 0x0f. */
    unsigned char m_nibble;
};

#endif
/*#ifndef OddPtr_H_INCLUDED */
