/**********************************************************************
**  Copyright 1999-2000 Xerox Corporation.All rights reserved.
**
**  DigiPaper Codebase
**    
**  Ranjani Ramamurthy, May 2000
**
**  JPEG Compression Testing Code
**
***********************************************************************/
/*
 * cjpeg.c
 *
 * Copyright (C) 1991-1998, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains a command-line user interface for the JPEG compressor.
 * It should work on any system with Unix- or MS-DOS-style command lines.
 *
 * Two different command line styles are permitted, depending on the
 * compile-time switch TWO_FILE_COMMANDLINE:
 *	cjpeg [options]  inputfile outputfile
 *	cjpeg [options]  [inputfile]
 * In the second style, output is always to standard output, which you'd
 * normally redirect to a file or pipe to some other program.  Input is
 * either from a named file or from standard input (typically redirected).
 * The second style is convenient on Unix but is unhelpful on systems that
 * don't support pipes.  Also, you MUST use the first style if your system
 * doesn't do binary I/O to stdin/stdout.
 * To simplify script writing, the "-outfile" switch is provided.  The syntax
 *	cjpeg [options]  -outfile outputfile  inputfile
 * works regardless of which command line style is used.
 */

#ifdef _WIN32
#include "windows.h"
#endif

#ifndef _WIN32
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "cdjpeg.h"		/* Common decls for cjpeg/djpeg applications */
#include "jversion.h"		/* for version message */
#include "jpeg6blib.h"
#include "jerror.h"
#include "dgp_err_printers.h"
#include "dgp_err_control.h"
#include <memory.h>
#include <assert.h>

#ifdef USE_CCOMMAND		/* command-line reader for Macintosh */
#ifdef __MWERKS__
#include <SIOUX.h>              /* Metrowerks needs this */
#include <console.h>		/* ... and this */
#endif
#ifdef THINK_C
#include <console.h>		/* Think declares it here */
#endif
#endif


/* Create the add-on message string table. */

#define JMESSAGE(code,string)	string ,

static const char * const cdjpeg_message_table[] = {
#include "cderror.h"
  NULL
};


/*
 * This routine determines what format the input file is,
 * and selects the appropriate input-reading module.
 *
 * To determine which family of input formats the file belongs to,
 * we may look only at the first byte of the file, since C does not
 * guarantee that more than one character can be pushed back with ungetc.
 * Looking at additional bytes would require one of these approaches:
 *     1) assume we can fseek() the input file (fails for piped input);
 *     2) assume we can push back more than one character (works in
 *        some C implementations, but unportable);
 *     3) provide our own buffering (breaks input readers that want to use
 *        stdio directly, such as the RLE library);
 * or  4) don't put back the data, and modify the input_init methods to assume
 *        they start reading after the start of file (also breaks RLE library).
 * #1 is attractive for MS-DOS but is untenable on Unix.
 *
 * The most portable solution for file types that can't be identified by their
 * first byte is to make the user tell us what they are.  This is also the
 * only approach for "raw" file types that contain only arbitrary values.
 * We presently apply this method for Targa files.  Most of the time Targa
 * files start with 0x00, so we recognize that case.  Potentially, however,
 * a Targa file could start with any byte value (byte 0 is the length of the
 * seldom-used ID field), so we provide a switch to force Targa input mode.
 */

static boolean is_targa;	/* records user -targa switch */


/*
 * Argument-parsing code.
 * The switch parser is designed to be useful with DOS-style command line
 * syntax, ie, intermixed switches and file names, where only the switches
 * to the left of a given file name affect processing of that file.
 * The main program in this file doesn't actually use this capability...
 */


static const char * progname;	/* program name for error messages */
static char * outfilename;	/* for -outfile switch */
static char * defaultsfilename; /* for -defaults switch */
static char * inbufname;

LOCAL(void)
usage (void)
/* complain about bad command line */
{
  fprintf(stderr, "usage: %s [switches] ", progname);
#ifdef TWO_FILE_COMMANDLINE
  fprintf(stderr, "inputfile outputfile\n");
#else
  fprintf(stderr, "[inputfile]\n");
#endif

  fprintf(stderr, "Switches (names may be abbreviated):\n");
  fprintf(stderr, "  -quality N     Compression quality (0..100; 5-95 is useful range)\n");
  fprintf(stderr, "  -grayscale     Create monochrome JPEG file\n");
#ifdef ENTROPY_OPT_SUPPORTED
  fprintf(stderr, "  -optimize      Optimize Huffman table (smaller file, but slow compression)\n");
#endif
#ifdef C_PROGRESSIVE_SUPPORTED
  fprintf(stderr, "  -progressive   Create progressive JPEG file\n");
#endif
#ifdef TARGA_SUPPORTED
  fprintf(stderr, "  -targa         Input file is Targa format (usually not needed)\n");
#endif
  fprintf(stderr, "Switches for advanced users:\n");
#ifdef DCT_ISLOW_SUPPORTED
  fprintf(stderr, "  -dct int       Use integer DCT method%s\n",
	  (JDCT_DEFAULT == JDCT_ISLOW ? " (default)" : ""));
#endif
#ifdef DCT_IFAST_SUPPORTED
  fprintf(stderr, "  -dct fast      Use fast integer DCT (less accurate)%s\n",
	  (JDCT_DEFAULT == JDCT_IFAST ? " (default)" : ""));
#endif
#ifdef DCT_FLOAT_SUPPORTED
  fprintf(stderr, "  -dct float     Use floating-point DCT method%s\n",
	  (JDCT_DEFAULT == JDCT_FLOAT ? " (default)" : ""));
#endif
  fprintf(stderr, "  -restart N     Set restart interval in rows, or in blocks with B\n");
#ifdef INPUT_SMOOTHING_SUPPORTED
  fprintf(stderr, "  -smooth N      Smooth dithered input (N=1..100 is strength)\n");
#endif
  fprintf(stderr, "  -maxmemory N   Maximum memory to use (in kbytes)\n");
  fprintf(stderr, "  -outfile name  Specify name for output file\n");
  fprintf(stderr, "  -defaults name Specify name for defaults file\n");
  fprintf(stderr, "  -verbose  or  -debug   Emit debug output\n");
  fprintf(stderr, "Switches for wizards:\n");
#ifdef C_ARITH_CODING_SUPPORTED
  fprintf(stderr, "  -arithmetic    Use arithmetic coding\n");
#endif
  fprintf(stderr, "  -baseline      Force baseline quantization tables\n");
  fprintf(stderr, "  -qtables file  Use quantization tables given in file\n");
  fprintf(stderr, "  -qslots N[,...]    Set component quantization tables\n");
  fprintf(stderr, "  -sample HxV[,...]  Set component sampling factors\n");
#ifdef C_MULTISCAN_FILES_SUPPORTED
  fprintf(stderr, "  -scans file    Create multi-scan JPEG per script file\n");
#endif
  exit(EXIT_FAILURE);
}


ErrCode
parse_switches (j_compress_ptr cinfo, int argc, char **argv,
		int last_file_arg_seen, boolean for_real, int* ret_index,
        Dgp_State* dgp_state)
/* Parse optional switches.
 * Returns argv[] index of first file-name argument (== argc if none).
 * Any file names with indexes <= last_file_arg_seen are ignored;
 * they have presumably been processed in a previous iteration.
 * (Pass 0 for last_file_arg_seen on the first or only iteration.)
 * for_real is FALSE on the first (dummy) pass; we may skip any expensive
 * processing.
 */
{

  char * arg;
  int quality;			/* -quality parameter */
  int q_scale_factor;		/* scaling percentage for -qtables */
  boolean force_baseline;
  boolean simple_progressive;
  char * qtablefile = NULL;	/* saves -qtables filename if any */
  char * qslotsarg = NULL;	/* saves -qslots parm if any */
  char * samplearg = NULL;	/* saves -sample parm if any */
  char * scansarg = NULL;	/* saves -scans parm if any */
  int arg1n=0;

  DGP_PROLOGUE("parse_switches()",DGP_PERRSTATE);

  /* Set up default JPEG parameters. */
  /* Note that default -quality level need not, and does not,
   * match the default scaling for an explicit -qtables argument.
   */
  quality = 75;			/* default -quality value */
  q_scale_factor = 100;		/* default to no scaling for -qtables */
  force_baseline = FALSE;	/* by default, allow 16-bit quantizers */
  simple_progressive = FALSE;
  is_targa = FALSE;
  outfilename = NULL;
  cinfo->err->trace_level = 0;

  /* Scan command line options, adjust parameters */

  for (arg1n = 1; arg1n < argc; arg1n++) {
    arg = argv[arg1n];
    if (*arg != '-') { 
      /* Not a switch, must be a file name argument */
      if (arg1n <= last_file_arg_seen) {
	     outfilename = NULL;	/* -outfile applies to just one input file */
	     continue;		/* ignore this name if previously processed */
      }
      break;			/* else done parsing switches */
    }
    arg++;			/* advance past switch marker character */

    if (keymatch(arg, "arithmetic", 1)) {
      /* Use arithmetic coding. */
#ifdef C_ARITH_CODING_SUPPORTED
      cinfo->arith_code = TRUE;
#else
      fprintf(stderr, "%s: sorry, arithmetic coding not supported\n",
	      progname);
      ERR_RETURN;
#endif

    } else if (keymatch(arg, "baseline", 1)) {
      /* Force baseline-compatible output (8-bit quantizer values). */
      force_baseline = TRUE;

    } else if (keymatch(arg, "dct", 2)) {
      /* Select DCT algorithm. */
      if (++arg1n >= argc)	/* advance to next argument */
	     usage();
      if(keymatch(argv[arg1n], "int", 1)) {
	    cinfo->dct_method = JDCT_ISLOW;
      } else if (keymatch(argv[arg1n], "fast", 2)) {
	    cinfo->dct_method = JDCT_IFAST;
      } else if (keymatch(argv[arg1n], "float", 2)) {
	    cinfo->dct_method = JDCT_FLOAT;
      } else
	    usage();
    } else if (keymatch(arg, "debug", 1) || keymatch(arg, "verbose", 1)) {
      /* Enable debug printouts. */
      /* On first -d, print version identification */
      static boolean printed_version = FALSE;

      if (! printed_version) {
	    fprintf(stderr, "Independent JPEG Group's CJPEG, version %s\n%s\n",
		                 JVERSION, JCOPYRIGHT);
	    printed_version = TRUE;
      }
      cinfo->err->trace_level++;

    } else if (keymatch(arg, "grayscale", 2) || keymatch(arg, "greyscale",2)) {
      /* Force a monochrome JPEG file to be generated. */
      jpeg_set_colorspace(cinfo, JCS_GRAYSCALE);
    } else if (keymatch(arg, "maxmemory", 3)) {
      /* Maximum memory in Kb (or Mb with 'm'). */
      long lval;
      char ch = 'x';

      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      if (sscanf(argv[arg1n], "%ld%c", &lval, &ch) < 1)
	    usage();
      if (ch == 'm' || ch == 'M')
	    lval *= 1000L;
      cinfo->mem->max_memory_to_use = lval * 1000L;

    } else if (keymatch(arg, "optimize", 1) || keymatch(arg, "optimise", 1)) {
      /* Enable entropy parm optimization. */
#ifdef ENTROPY_OPT_SUPPORTED
      cinfo->optimize_coding = TRUE;
#else
      fprintf(stderr, "%s: sorry, entropy optimization was not compiled\n",
	      progname);
      ERR_RETURN;
#endif

    } else if (keymatch(arg, "outfile", 4)) {
      /* Set output file name. */
      if(++arg1n >= argc)	/* advance to next argument */
	     usage();
      outfilename = argv[arg1n];
      //strcpy(outfilename, argv[arg1n]);	/* save it away for later use */

    } else if (keymatch(arg,"defaults",5)) {
      /* Set the defaults file name */
      if(++ arg1n >= argc) /* advance to next argument */
        usage();
      defaultsfilename = argv[arg1n];

    } else if (keymatch(arg,"inbuf",5)){
      /* set the input buffer name */
      if(++ arg1n >= argc) /* advance to next argument */
        usage();
      inbufname  = argv[arg1n];
    } else if (keymatch(arg, "progressive", 1)) {
      /* Select simple progressive mode. */
#ifdef C_PROGRESSIVE_SUPPORTED
      simple_progressive = TRUE;
      /* We must postpone execution until num_components is known. */
#else
      fprintf(stderr, "%s: sorry, progressive output was not compiled\n",
	      progname);
      ERR_RETURN;
#endif

    } else if (keymatch(arg, "quality", 1)) {
      /* Quality factor (quantization table scaling factor). */
      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      if (sscanf(argv[arg1n], "%d", &quality) != 1)
	    usage();
      /* Change scale factor in case -qtables is present. */
      q_scale_factor = jpeg_quality_scaling(quality);

    } else if (keymatch(arg, "qslots", 2)) {
      /* Quantization table slot numbers. */
      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      qslotsarg = argv[arg1n];
      /* Must delay setting qslots until after we have processed any
       * colorspace-determining switches, since jpeg_set_colorspace sets
       * default quant table numbers.
       */

    } else if (keymatch(arg, "qtables", 2)) {
      /* Quantization tables fetched from file. */
      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      qtablefile = argv[arg1n];
      /* We postpone actually reading the file in case -quality comes later. */

    } else if (keymatch(arg, "restart", 1)) {
      /* Restart interval in MCU rows (or in MCUs with 'b'). */
      long lval;
      char ch = 'x';

      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      if (sscanf(argv[arg1n], "%ld%c", &lval, &ch) < 1)
	    usage();
      if (lval < 0 || lval > 65535L)
	    usage();
      if (ch == 'b' || ch == 'B') {
	    cinfo->restart_interval = (unsigned int) lval;
	    cinfo->restart_in_rows = 0; /* else prior '-restart n' overrides me */
      } else {
	    cinfo->restart_in_rows = (int) lval;
	    /* restart_interval will be computed during startup */
      }

    } else if (keymatch(arg, "sample", 2)) {
      /* Set sampling factors. */
      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      samplearg = argv[arg1n];
      /* Must delay setting sample factors until after we have processed any
       * colorspace-determining switches, since jpeg_set_colorspace sets
       * default sampling factors.
       */

    } else if (keymatch(arg, "scans", 2)) {
      /* Set scan script. */
#ifdef C_MULTISCAN_FILES_SUPPORTED
      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      scansarg = argv[arg1n];
      /* We must postpone reading the file in case -progressive appears. */
#else
      fprintf(stderr, "%s: sorry, multi-scan output was not compiled\n",
	      progname);
      ERR_RETURN;
#endif

    } else if (keymatch(arg, "smooth", 2)) {
      /* Set input smoothing factor. */
      int val;

      if (++arg1n >= argc)	/* advance to next argument */
	    usage();
      if (sscanf(argv[arg1n], "%d", &val) != 1)
	    usage();
      if (val < 0 || val > 100)
	    usage();
      cinfo->smoothing_factor = val;

    } else if (keymatch(arg, "targa", 1)) {
      /* Input file is Targa format. */
      is_targa = TRUE;
    } else {
      usage();			/* bogus switch */
    }
  }

  /* Post-switch-scanning cleanup */

  if (for_real) {

    /* Set quantization tables for selected quality. */
    /* Some or all may be overridden if -qtables is present. */
    jpeg_set_quality(cinfo, quality, force_baseline);

    if (qtablefile != NULL)	/* process -qtables if it was present */
      if (! read_quant_tables(cinfo, qtablefile,
			      q_scale_factor, force_baseline))
	     usage();

    if (qslotsarg != NULL)	/* process -qslots if it was present */
      if (! set_quant_slots(cinfo, qslotsarg))
	     usage();

    if (samplearg != NULL)	/* process -sample if it was present */
      if (! set_sample_factors(cinfo, samplearg))
	     usage();

#ifdef C_PROGRESSIVE_SUPPORTED
    if (simple_progressive)	/* process -progressive; -scans can override */
      jpeg_simple_progression(cinfo);
#endif

#ifdef C_MULTISCAN_FILES_SUPPORTED
    if (scansarg != NULL)	/* process -scans if it was present */
      if (! read_scan_script(cinfo, scansarg))
	     usage();
#endif
  }

  *(ret_index) = arg1n;
  NORMAL_RETURN;
}

#ifdef _WIN32
ErrCode
get_file_size(char* filename,int* size, Dgp_State* dgp_state)
{
  HANDLE fileHandle;
  DWORD  ret_val; BOOL close_ret;

  DGP_PROLOGUE("get_file_size()",DGP_PERRSTATE);

  fileHandle = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,
     NULL,OPEN_EXISTING,0,NULL);
  if(fileHandle == INVALID_HANDLE_VALUE){
    ERR_RETURN;
  }
  *(size) = (int) GetFileSize(fileHandle,NULL);
  ret_val = *(size);

  if(ret_val == 0xFFFFFFFF){
    ERR_RETURN;
  }
  
  close_ret = CloseHandle(fileHandle);
  if(close_ret == 0){
    ERR_RETURN;
  }

  NORMAL_RETURN;

}
#else
ErrCode
get_file_size(char* filename, int* size, Dgp_State* dgp_state)
{
 struct stat stbuf;

 DGP_PROLOGUE("get_file_size()",DGP_PERRSTATE);
 if(stat(filename,&stbuf) == -1){
   *(size) = -1;
   ERR_RETURN;
 }

 *(size) = stbuf.st_size;
 NORMAL_RETURN;
}
#endif

ErrCode
set_defaults_from_file(DGP_JPEGCompressDef* rgb_bufdef,
                       DGP_JPEGOutputDef* output_buf,
                       FILE* defaults_file, Dgp_State* dgp_state)
{
 char type_string[30];
 char string_holder[30]; int fseek_ret,fclose_ret;

  DGP_PROLOGUE("set_defaults_from_file()",DGP_PERRSTATE);

  fseek_ret = fseek(defaults_file,0L,SEEK_SET); //set it to the beginning
  if(fseek_ret != 0)
    ERR_RETURN;

  while(!feof(defaults_file)){
    fscanf(defaults_file,"%s",string_holder);
    if(keymatch(string_holder,"imgwidth",6)){
      fscanf(defaults_file,"%d\n",&rgb_bufdef->imgWidth);
    } else if(keymatch(string_holder,"imgheight",6)){
      fscanf(defaults_file,"%d\n",&rgb_bufdef->imgHeight);
    } else if(keymatch(string_holder,"imgtype",6)){
      fscanf(defaults_file,"%s\n",type_string);
      if(keymatch(type_string,"dgp_rgb",5)){
        rgb_bufdef->imgType = DGP_RGB;
      } else if(keymatch(type_string,"dgp_cmyk",6)){
        rgb_bufdef->imgType = DGP_CMYK;
      }
    } else if(keymatch(string_holder,"outbufsize",6)){
      // This again is for testing purposes only.
      fscanf(defaults_file,"%d\n",&output_buf->outLen);
      output_buf->buffer = calloc(1,output_buf->outLen);
      if(output_buf->buffer == NULL)
         ERR_RETURN;
    } /*else if(keymatch(string_holder,"outfile",7)){
      fscanf(defaults_file,"%s\n",outfilename);
    }*/
  } //end of the while loop


  if((fclose_ret = fclose(defaults_file)) != 0)
    ERR_RETURN;

  NORMAL_RETURN;
}

ErrCode
set_manual_defaults(DGP_JPEGCompressDef* rgb_bufdef,
                    DGP_JPEGOutputDef* output_buf,
                    Dgp_State* dgp_state)
{
    FILE *in_file; int file_size, bytes_read;

    DGP_PROLOGUE("set_manual_defaults()",DGP_PERRSTATE);

    if ((in_file = fopen("rgbbuf", READ_BINARY)) == NULL) {
      fprintf(stderr, "%s: can't open rgbbuf\n", progname);
      ERR_RETURN;
    }

    DGP_ERR_PASS(get_file_size("rgbbuf",&file_size,dgp_state));

    rgb_bufdef->imgHeight = 215;
    rgb_bufdef->imgType = DGP_RGB;
    rgb_bufdef->imgWidth = 327;

    rgb_bufdef->imgSrc.buffer = calloc(1,file_size);
    if(rgb_bufdef->imgSrc.buffer == NULL){
       ERR_RETURN;
    }

    // Now copy the contents of the file into the buffer
    bytes_read = fread(rgb_bufdef->imgSrc.buffer,1, file_size,in_file);

    if(bytes_read != file_size){
      ERR_RETURN;
    }
    rgb_bufdef->imgSrc.size = file_size;

    output_buf->buffer = calloc(1,16000);
    output_buf->outLen = 16000;
    NORMAL_RETURN;
}



/*
 * The main program.
 */

int
main (int argc, char **argv)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
#ifdef PROGRESS_REPORT
  struct cdjpeg_progress_mgr progress;
#endif
  int file_index, size_written, fclose_ret,file_size;
  int bytes_read;
  FILE *output_file, *fp;
  FILE * defaults_file;
  DGP_JPEGCompressDef rgb_bufdef;
  DGP_JPEGOutputDef output_buf;

  Dgp_State *dgp_state = NULL;

  DGP_MAIN_PROLOGUE("main()",stderr,Dgp_fileErrPrinter);


  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state,NULL, NULL,
                                stderr,Dgp_fileErrPrinter,
                                TRUE,FALSE));

  DGP_SET_ERR_STATE(DGP_PERRSTATE);

  /* On Mac, fetch a command line. */
#ifdef USE_CCOMMAND
  argc = ccommand(&argv);
#endif

  progname = argv[0]; defaultsfilename = NULL; inbufname = NULL;

  if (progname == NULL || progname[0] == 0)
    progname = "cjpeg";		/* in case C library doesn't provide it */

  /* Initialize the JPEG compression object with default error handling. */
  cinfo.err = jpeg_std_error(&jerr);

  DGP_ERR_PASS(DGP_JPEG_init_jcompress(&cinfo,dgp_state));

  /* Now safe to enable signal catcher. */
#ifdef NEED_SIGNAL_CATCHER
  enable_signal_catcher((j_common_ptr) &cinfo);
#endif


  /* Scan command line to find file names.
   * It is convenient to use just one switch-parsing routine, but the switch
   * values read here are ignored; we will rescan the switches after opening
   * the input file.
   */

  DGP_ERR_PASS(parse_switches(&cinfo, argc, argv, 0, FALSE, &file_index,
                              dgp_state));

  if(inbufname != NULL){
      if ((fp = fopen(inbufname, READ_BINARY)) == NULL) {
        fprintf(stderr, "Can't open file: %s\n",inbufname);
        goto ERR_CLEANUP;
      }
      DGP_ERR_PASS(get_file_size(inbufname,&file_size,dgp_state));
      // Now calloc a buffer and read the contents of the file
      // into the buffer.
      rgb_bufdef.imgSrc.buffer = calloc(1,file_size);
      if(rgb_bufdef.imgSrc.buffer == NULL)
         goto ERR_CLEANUP;
      // Now copy the contents of the file into the buffer
      bytes_read = fread(rgb_bufdef.imgSrc.buffer,1, file_size,fp);
      if(bytes_read != file_size)
        goto ERR_CLEANUP;
      rgb_bufdef.imgSrc.size = file_size;
      if((fclose_ret = fclose(fp)) != 0)
        goto ERR_CLEANUP;
  }

  if(defaultsfilename == NULL){
    DGP_ERR_PASS(set_manual_defaults(&rgb_bufdef,
                                     &output_buf,dgp_state));
  } else {
    if ((defaults_file = fopen(defaultsfilename, READ_BINARY)) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", progname, defaultsfilename);
      goto ERR_CLEANUP;
    }
    DGP_ERR_PASS(set_defaults_from_file(&rgb_bufdef,&output_buf,
                            defaults_file, dgp_state));
  }



#ifdef PROGRESS_REPORT
  start_progress_monitor((j_common_ptr) &cinfo, &progress);
#endif

  DGP_ERR_PASS_CLEANUP(Dgp_JPEG_create_jpeg(&rgb_bufdef,
               &output_buf,&cinfo,dgp_state));

  // Open the output file. 
  if (outfilename != NULL) {
    if ((output_file = fopen(outfilename, WRITE_BINARY)) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", progname, outfilename);
      goto ERR_CLEANUP;
    }
  } 

  size_written = 
     fwrite(output_buf.buffer,1,output_buf.outLen,output_file);

  if(size_written != output_buf.outLen){
    goto ERR_CLEANUP;
  } else {
    fflush(output_file);
    free(output_buf.buffer);
    free(rgb_bufdef.imgSrc.buffer);
  }

  DGP_ERR_PASS_CLEANUP(Dgp_JPEG_destroy_jcompress(&cinfo,dgp_state));

  /* Close files, if we opened them */
  if (output_file != stdout) {
    if((fclose_ret = fclose(output_file)) != 0){
      goto ERR_CLEANUP;
    }
  }

#ifdef PROGRESS_REPORT
  end_progress_monitor((j_common_ptr) &cinfo);
#endif

  Dgp_cleanupDgpState(&dgp_state);
  
  MAIN_NORMAL_RETURN;

  ERR_CLEANUP:
    if(output_buf.buffer != NULL){
      free(output_buf.buffer);
    }
    if(rgb_bufdef.imgSrc.buffer != NULL) {
      free(rgb_bufdef.imgSrc.buffer);
    }
    Dgp_JPEG_destroy_jcompress(&cinfo,dgp_state);
    Dgp_cleanupDgpState(&dgp_state);
    MAIN_ERR_RETURN;    
}
