/*
 *
 * $Header: /project/icons/cvsroot/digipaper/tools/libtok-lossy.help/makedil.c,v 1.2 2000/03/07 21:47:42 mkrocken Exp $
 *
 */

/*
 * This program generates a dilate lookup table for the topology-preserving
 * dilation ("cross dilation"). I stole it from Brian Freyburger, who in turn
 * got it out of code written by Mike Leventon.
 *
 * Mike's version generated the table at runtime; Brian made it generate
 * a table which would be included. I've hacked it to generate a .c file
 * rather than a .h file because I don't like including .h files that actually
 * define and initialise arrays.
 *
 * It actually generates two values: should this pixel be turned on in the
 * dilated image, and was this an edge pixel in the original image.
 *
 * The output values it produces are:
 * 0 - dilated output is 0. (i.e., turn this pixel off in
 *     the dilated image)
 * 1 - dilated output is 1. (turn this pixel on in the dilated image)
 * 3 - dilated output is 1. Furthermore, the original pixel
 *     was turned on, *and* at least one of its neighbours was off (it was
 *     an "edge pixel").
 *
 * It actually generates 3 tables:
 * - topology-preserving dilate by radius=1 (array name tpdilateLookup)
 * - regular dilate by radius=1 (array name rad1dilateLookup)
 * - regular dilate by radius=sqrt(2) (array name radroot2dilateLookup)
 */

/*
 * System include files
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

/*
 * Static function, type, and variable declarations.
 */

/*
 * This will be used in a table-lookup where the index will be based on the
 * current bit and 12 bits surrounding it. The bits are numbered
 *         0
 *     1   2   3
 * 4   5   6   7   8
 *     9  10  11
 *        12
 */
#define  BTN(i)    (((i) & Bit(2)) >> 2)
#define  BTS(i)    (((i) & Bit(10)) >> 10)
#define  BTW(i)    (((i) & Bit(5)) >> 5)
#define  BTE(i)    (((i) & Bit(7)) >> 7)
#define BTNW(i)    (((i) & Bit(1)) >> 1)
#define BTNE(i)    (((i) & Bit(3)) >> 3)
#define BTSW(i)    (((i) & Bit(9)) >> 9)
#define BTSE(i)    (((i) & Bit(11)) >> 11)

#define   T(i)    ((i) & Bit(6))
#define  TN(i)    ((i) & Bit(2))
#define  TS(i)    ((i) & Bit(10)) 
#define  TW(i)    ((i) & Bit(5))
#define  TE(i)    ((i) & Bit(7))
#define TNW(i)    ((i) & Bit(1))
#define TNE(i)    ((i) & Bit(3))
#define TSW(i)    ((i) & Bit(9))
#define TSE(i)    ((i) & Bit(11)) 
#define TNN(i)    ((i) & Bit(0))
#define TSS(i)    ((i) & Bit(12)) 
#define TWW(i)    ((i) & Bit(4))
#define TEE(i)    ((i) & Bit(8))

#define Bit(b)     (1 << (b))

/*
 * Public function definitions
 */

int
main(int argc, char **argv)
{
    FILE *fp;
    int i = 0;
    int dil;
  
    if (argc != 1) {
	fprintf(stderr, "Usage: %s > outfile\n", argv[0]);
	exit(1);
	}
  
    fp = stdout;

    fprintf(fp,"#include \"dgp_rcs.h\"\n");
    fprintf(fp,"#include \"dgp_types.h\"\n");
  
    fprintf(fp, "DGP_RCSID(\"@(#)cross dilation table created by $Header: /project/icons/cvsroot/digipaper/tools/libtok-lossy.help/makedil.c,v 1.2 2000/03/07 21:47:42 mkrocken Exp $\");\n");

    fprintf(fp, "const Dgp_UInt8 tpdilateLookup[1<<13] =\n");
    fprintf(fp, "{\n");
  
    for (i = 0; i < 1<<13; i++) {
	if (T(i)) {
	    /*
	     * This bit is on. However, we'll also want to look for
	     * edge pixels (on pixels with at least one off neighbour), so
	     * why not use this table for that too? Count the neighbours, and
	     * say "3" if edge, "1" if not edge. (In general, "1" or "3" will
	     * mean "turn this on when dilating"; "0" will mean "don't").
	     */
	    if (BTN(i) + BTW(i) + BTE(i) + BTS(i) +
		BTNE(i) + BTNW(i) + BTSE(i) + BTSW(i) < 8)  {
		dil = 3;
		}
	    else {
		dil = 1;
		}
	    }
	else {
	    if (		/*
				 * Dilate if:
				 *  X      X
				 * X?X    X?X
				 * OO      O
				 *         O
				 * (and their symmetries of course).
				 */
	      
	      (TN(i) && TW(i) && TE(i) && !TS(i) &&
	       (!TSE(i) || !TSW(i) || !TSS(i))) ||

	      (TS(i) && TW(i) && TE(i) && !TN(i) && 
	       (!TNE(i) || !TNW(i) || !TNN(i))) ||
	      
	      (TE(i) && TS(i) && TN(i) && !TW(i) && 
	       (!TNW(i) || !TSW(i) || !TWW(i))) ||
	      
	      (TW(i) && TS(i) && TN(i) && !TE(i) && 
	       (!TNE(i) || !TSE(i) || !TEE(i))) ||
	      
	      /*
	       * Dilate if:
	       *  X       XO         (The "0" in the S position is not required
	       * X?OO    X?O          because, if it's an X, we'll still dilate
	       *  0O      0O          by the above rule).
	       *  X       X
	       */
	      (TN(i) && TSS(i) &&
	       ((TW(i) && !TE(i) && !TSE(i) && 
		 (!TEE(i) || !TNE(i))) ||
		(TE(i) && !TW(i) && !TSW(i) &&
		 (!TWW(i) || !TNW(i))))) ||
	      
	      (TS(i) && TNN(i) &&
	       ((TW(i) && !TE(i) && !TNE(i) && 
		 (!TEE(i) || !TSE(i))) ||
		(TE(i) && !TW(i) && !TNW(i) &&
		 (!TWW(i) || !TSW(i))))) ||
	      
	      (TE(i) && TWW(i) &&
	       ((TS(i) && !TN(i) && !TNW(i) && 
		 (!TNN(i) || !TNE(i))) ||
		(TN(i) && !TS(i) && !TSW(i) &&
		 (!TSS(i) || !TSE(i))))) ||
	      
	      (TW(i) && TEE(i) &&
	       ((TS(i) && !TN(i) && !TNE(i) && 
		 (!TNN(i) || !TNW(i))) ||
		(TN(i) && !TS(i) && !TSE(i) &&
		 (!TSS(i) || !TSW(i)))))) {
	      dil = 1;
	      }
	  else {
	      /*
	       * Dilate When:
	       *	X?
	       *
	       * EXCEPT, Don't Dilate When:
	       *		 OX
	       * X?X    X?OX    X?
	       *
	       * EXCEPT, Do Dilate When the cases in the "if" clause above
	       * hold.
	       */
	      if (TN(i) || TS(i) || TW(i) || TE(i)) {
		  dil = 1;		/* By default, dilate */
	      
		  if ((TN(i) &&
		      (TS(i) || TSS(i) ||
		       (!TW(i) && TSW(i)) ||
		       (!TE(i) &&  TSE(i)))) ||
		      (TS(i) &&
		       (TNN(i) ||
			(TNW(i) && !TW(i)) ||
			(TNE(i) && !TE(i)))) ||
		      (TW(i) &&
		       (TE(i) || TEE(i) ||
			(TNE(i) && !TN(i)) ||
			(TSE(i) && !TS(i)))) ||
		      (TE(i) &&
		       (TWW(i) ||
			(TNW(i) && !TN(i)) ||
			(TSW(i) && !TS(i))))) {
		      dil = 0;
		      }
		  }
	      else {
		  dil = 0;
		  }
	      }
	  }

      if (i == 0) {
	  fprintf(fp, "  %d", dil);
	  }
      else if ((i % 16) == 0) {
	  fprintf(fp, ",\n  %d", dil);
	  }
      else {
	  fprintf(fp, ", %d", dil);
	  }
      }
  
    fprintf(fp, "\n};\n\n\n");
  
    fprintf(fp, "const Dgp_UInt8 rad1dilateLookup[1<<13] =\n");
    fprintf(fp, "{\n");
  
    for (i = 0; i < 1<<13; i++) {
	if (T(i)) {
	    if (BTN(i) + BTW(i) + BTE(i) + BTS(i) +
		BTNE(i) + BTNW(i) + BTSE(i) + BTSW(i) < 8)  {
		/* Edge pixel */
		dil = 3;
		}
	    else {
		dil = 1;
		}
	    }
	else {
	    if (		/*
				 * Dilate if:
				 * X?
				 * (and its symmetries of course).
				 */
	      
		TN(i) || TS(i) || TW(i) || TE(i)) {
		dil = 1;
		}
	    else {
		dil = 0;
		}
	    }

	if (i == 0) {
	    fprintf(fp, "  %d", dil);
	    }
	else if ((i % 16) == 0) {
	    fprintf(fp, ",\n  %d", dil);
	    }
	else {
	    fprintf(fp, ", %d", dil);
	    }
	}
  
    fprintf(fp, "\n};\n\n\n");
  
    fprintf(fp, "const Dgp_UInt8 radroot2dilateLookup[1<<13] =\n");
    fprintf(fp, "{\n");
  
    for (i = 0; i < 1<<13; i++) {
	if (T(i)) {
	    if (BTN(i) + BTW(i) + BTE(i) + BTS(i) +
		BTNE(i) + BTNW(i) + BTSE(i) + BTSW(i) < 8)  {
		/* Edge pixel */
		dil = 3;
		}
	    else {
		dil = 1;
		}
	    }
	else {
	    if (		/*
				 * Dilate if:
				 * X?		X
				 *		 ?
				 * (and their symmetries of course).
				 */
	      
		TN(i) || TS(i) || TW(i) || TE(i) ||
		TNE(i) || TNW(i) || TSW(i) || TSE(i)) {
		dil = 1;
		}
	    else {
		dil = 0;
		}
	    }

	if (i == 0) {
	    fprintf(fp, "  %d", dil);
	    }
	else if ((i % 16) == 0) {
	    fprintf(fp, ",\n  %d", dil);
	    }
	else {
	    fprintf(fp, ", %d", dil);
	    }
	}
  
    fprintf(fp, "\n};\n\n\n");
  
    exit(0);
    }
